//!
//! OUI
//! Built on: Wed Aug 24 2016 09:27:55
//! Library version: 3.8.7
//! Branch: master  |  Commit: 8bcae6d
//!
define('oui/utils/subscribable',['underscore'], function (_) {
    'use strict';

    var STATES = {
        play: 'play',
        pause: 'paused'
    };

    /**
     * @typedef {Object} Subscribable~EventOptions
     * @property {Boolean} [silent=false]
     * @property {String} [name]
     */

    /**
     * @method Subscribable~normalize
     *
     * @param {String} eventName
     *
     * @returns {Object}
     */
    function normalize(eventName) {
        var split = eventName.split('.');

        if (split.length === 1) {
            return { name: eventName };
        } else {
            return { name: split[1], namespace: split[0] };
        }
    }

    /**
     * @method Subscribable~argumentsSlice
     * @private
     *
     * @returns {Array} args
     */
    function argumentsSlice() {
        var args = new Array(arguments.length);

        for (var i = 0; i < args.length; ++i) {
            args[i] = arguments[i];
        }

        return args;
    }

    /**
     * The interface for all objects that can be subscribed to.
     *
     * @class Subscribable
     * @private
     */
    function Subscribable() {

        /**
         * The collection of events for the object.
         *
         * An event object has keys: name, namespace (optional), fn, context
         *
         * @memberof! Subscribable.prototype
         * @type {Array<Object>}
         * @default []
         * @private
         */
        this._events = {};
        this._pausedEvents = [];
        this._eventingState = STATES.play;
    }

    Subscribable.prototype = {
        constructor: Subscribable,
        /**
         * Calls each function that matches the argument.
         * Argument can either be namespace.event, or just event
         *
         * If argument is just event name than all functions will be called otherwise
         * only the namespaced event will be called.
         *
         * @example
         * <caption>Triggering a silent event with event args '1'</caption>
         *
         * this.trigger({ silent : true }, 'select', '1');
         *
         * @example
         * <caption>Triggering a default event with event args '1'</caption>
         *
         * this.trigger('select', '1');
         *
         * @example
         * <caption>Triggering a default namespaced event with event args '1'</caption>
         *
         * this.trigger('view.select', '1');
         *
         * @method Subscribable#trigger
         * @param {Subscribable~EventOptions} [options] The event options. Currently, only silent is an accepted property of the object
         * @param {String} [eventName] Either a string in the form of nameSpace.event or event
         * @param {...*} params The triggered events data
         */
        trigger: function () {
            var args = argumentsSlice.apply(null, arguments);
            var options = args[0];
            var specials = this._events['*'];
            var normalized;
            var eventName;
            var retVal;

            if (this._eventingState === STATES.pause) {
                this._pausedEvents.push(args);
                return;
            }

            //No options supplied or using combined options + eventName
            if (_.isString(options) || (_.isObject(options) && options.name)) {
                args.shift();

                eventName = options.name || options;

            //Options given to trigger
            } else {
                eventName = args.splice(0, 2)[1];
            }

            options = options || { silent: false };

            normalized = normalize(eventName);

            _.each(this._events[normalized.name], function (event) {
                if (!normalized.namespace || event.namespace === normalized.namespace) {
                    var intermReturn = event.fn.apply(event.context || options, args);
                    if (retVal == null) {
                        retVal = intermReturn;
                    }
                }
            });

            //There exist special listeners
            if (specials) {
                args = [eventName].concat(args);

                _.each(specials, function (event) {
                    var intermReturn = event.fn.apply(event.context || options, args);
                    if (retVal == null) {
                        retVal = intermReturn;
                    }
                });
            }

            return retVal;
        },
        /**
         * Sets the callback for the event name under the given namespace.
         *
         * '*' as the event name is a special property that will be
         * fired for any event called.
         *
         * @example
         * <caption>Listen to select event name and function</caption>
         *
         * this.on('select', function () {});
         *
         * @example
         * <caption>Listen to select event with specified context</caption>
         *
         * this.on('select', function () {}, that);
         *
         * @example
         * <caption>Listen to select through Object</caption>
         *
         * this.on({
         *    'select' : function () {}
         * })
         *
         * @example
         * <caption>Listen to select event through Object (variation)</caption>
         *
         * this.on({
         *     'select' : {
         *          fn : function () {},
         *          context : that
         *     }
         * })
         *
         * @method Subscribable#on
         * @chainable
         * @param {String|Object} eventName Either a string in the form of nameSpace.event or event
         * @param {Function} [callBack] The function to call when triggered.
         * @param {Object} [context] The context to use when calling the callBack.
         */
        on: function (eventName, callBack, context) {
            var that = this;

            /**
             * @method Subscribable#on~addEvent
             * @private
             *
             * @param {String} ev
             * @param {Function} cb
             * @param {Object} ctx
             */
            var addEvent = function (ev, cb, ctx) {
                var events = ev.split(' ');

                if (that.name && that.id) {
                    //If this is a widget then always use 'that' for the context
                    ctx = ctx || that;
                }

                _.each(events, function (event) {
                    that._events[normalize(event).name] = that._events[normalize(event).name] || [];
                    that._events[normalize(event).name].push(_.extend({ fn: cb, context: ctx }, normalize(event)));
                });
            };

            if (_.isObject(eventName)) {
                //Given a map
                //Ignore callback and context
                _.each(eventName, function (value, key) {
                    if (_.isFunction(value)) {
                        addEvent(key, value, undefined);
                    } else {
                        addEvent(key, value.fn, value.context);
                    }
                });
            } else if (_.isString(eventName)) {
                addEvent(eventName, callBack, context);
            }

            return this;
        },
        /**
         * Removes the function(s) for the given namespace. If no name space is given then all
         * functions are removed.
         *
         * @example
         * <caption>Remove all events</caption>
         *
         * this.off();
         *
         * @example
         * <caption>Remove all events of name 'select'</caption>
         *
         * this.off('select');
         *
         * @example
         * <caption>Remove only namespaced 'select' event</caption>
         *
         * this.off('view.select');
         *
         * @method Subscribable#off
         * @chainable
         * @param {String} [eventName] Either a string in the form of nameSpace.event or event
         */
        off: function (eventName) {
            if (eventName == null || eventName === '') {
                this._events = {};
            } else {
                var properties = normalize(eventName);
                this._events[properties.name] = _.reject(this._events[properties.name], function (event) {
                    return _.isEqual(properties, _.pick(event, _.keys(properties)));
                });
            }

            return this;
        },
        /**
         * @method Subscribable#pause
         */
        pause: function () {
            this._eventingState = STATES.pause;
        },
        /**
         * @method Subscribable#resume
         *
         * @returns {Object}
         */
        resume: function () {
            var clone = _.clone(this._pausedEvents);
            this._eventingState = STATES.play;

            this._pausedEvents = [];

            return _.groupBy(clone, function (eventArgs) {
                return eventArgs[0];
            });
        },
        /**
         * @method Subscribable#isPlaying
         *
         * @returns {Boolean}
         */
        isPlaying: function () {
            return this._eventingState === STATES.play;
        }
    };

    return Subscribable;

});

/**
 * @module uniqueId
 * @private
 */
define('oui/utils/uniqueId',[],function () {
    'use strict';

    var ids = {};

    return function (clz) {
        if (!ids[clz]) {
            ids[clz] = 0;
        }

        return ++ids[clz];
    };
});


define("oui/json/strings.json", function(){ return {
    "loading" : "Loading",
    "help" : "Help",
    "today": "Today",
    "ok": "OK",
    "cancel": "Cancel",
    "remove": "Remove",
    "close" : "Close",
    "expanded" : "Expanded",
    "collapsed": "Collapsed",
    "locked": "Locked",
    "fullScreen" : "Full Screen",
    "selectAll": "Select All",
    "newValue": "(New Value)",
    "add": "Add",
    "delete": "Delete",
    "calendar": {
        "nextYear": "Next Year",
        "prevYear": "Previous Year",
        "nextMonth": "Next Month",
        "prevMonth": "Previous Month",
        "selectDate": "Select Date",
        "standard" : "Standard",
        "unavailable": "Unavailable",
        "exception": "Exception",
        "previous" : "Previous",
        "next" : "Next",
        "noGroup" : "No Group"
    },
    "startDate" : "Start Date",
    "endDate" : "End Date",
    "event_create_title" : "Create Event",
    "event_update_title" : "Update Event",
    "views" : {
        "task-day" : "Day",
        "task-week" : "Week",
        "day" : "Day",
        "week" : "Week",
        "month" : "Month",
        "year" : "Year"
    },
    "more" : "More...",
    "spinner": {
        "increment": "Increment",
        "decrement": "Decrement"
    },
    "menu" : "Dropdown menu",
    "test": "Test",
    "clear": "Clear",
    "resize": "Resize",
    "collapse": {
        "up": "Collapse Up",
        "down": "Collapse Down",
        "left": "Collapse Left",
        "right": "Collapse Right"
    },
    "search" : "Search",
    "placeholder" : "Search",
    "notification" : "{0} selected. Click here to remove filter.",
    "no_results_found": "No Results Found",
    "empty_tree": "No Results",
    "picker" : {
        "select" : "Select"
    },
    "filter": {
        "applied": "Filter applied",
        "none": "No filter applied"
    },
    "browse" : "Browse",
    "editor" : {
        "title" : "Rich Text Editor",
        "fontColor" : "Font Color",
        "backgroundColor" : "Background Color",
        "bold" : "Bold",
        "italic" : "Italic",
        "underline" : "Underline",
        "strikeThrough" : "Strikethrough",
        "subscript" : "Subscript",
        "superscript" : "Superscript",
        "justifyLeft" : "Left Justify",
        "justifyCenter" : "Center",
        "justifyRight" : "Right Justify",
        "orderedList" : "Ordered List",
        "unorderedList" : "Unordered List",
        "increaseIndent" : "Increase Indent",
        "decreaseIndent" : "Decrease Indent",
        "insertLink" : "Insert Link",
        "removeLink" : "Remove Link",
        "fontSize" : "Font Size",
        "fontName" : "Font Name",
        "fontFormat" : "Font Format",
        "linkText" : "Link Text",
        "openIn" : "Open in",
        "openInCurrent" : "Current Tab",
        "openInNew" : "New Tab or Window",
        "url" : "URL"
    },
    "icon" : {
        "select" : "Select Icon"
    },
    "color" : {
        "select" : "Select Color",
        "empty" : "Empty",
        "000000" : "Black",
        "CCCCCC" : "Light Gray",
        "666666" : "Dark Gray",
        "C1CBD5" : "Light Blue Gray",
        "8B9AB2" : "Blue Gray",
        "3E577F" : "Dark Blue Gray",
        "0E2D5F" : "Darkest Blue Gray",
        "DBDCE0" : "Light Gray Blue",
        "B6BAC1" : "Gray Blue",
        "616977" : "Dark Gray Blue",
        "495363" : "Darkest Gray Blue",
        "7FCDEE" : "Light Blue",
        "66AAD7" : "Blue",
        "0072BC" : "Dark Blue",
        "0F527D" : "Darkest Blue",
        "99DDD8" : "Light Mint Green",
        "66C4BE" : "Mint Green",
        "009C92" : "Dark Mint Green",
        "00857C" : "Darkest Mint Green",
        "CBE5CB" : "Lightest Green",
        "AFDF8E" : "Light Green",
        "65B265" : "Green",
        "008000" : "Dark Green",
        "FFF59C" : "Light Yellow",
        "FFEF5E" : "Yellow",
        "FFDD00" : "Dark Yellow",
        "F2C019" : "Darkest Yellow",
        "FCD580" : "Light Orange",
        "FAAB01" : "Orange",
        "E3870A" : "Dark Orange",
        "BE5709" : "Brown",
        "F5C2C2" : "Light Pink",
        "EF9A9A" : "Pink",
        "D80202" : "Red",
        "A20202" : "Maroon",
        "EACEE7" : "Light Purple Pink",
        "D59ECF" : "Purple Pink",
        "AB3DA0" : "Dark Purple Pink",
        "710966" : "Darkest Purple Pink",
        "CBC0DB" : "Light Purple",
        "9881B7" : "Purple",
        "6B40AD" : "Dark Purple",
        "503082" : "Darkest Purple",
        "FFFFFF" : "White"
    },
    "uncategorizedLabel" : "Uncategorized",
    "errors": {
        "startsWith" : "{0} must start with either http:// or https://"
    },
    "instructions" : {
        "date" : "Date Field: Enter a date as text or press either F2 or Shift-F10 to open calendar popup.",
        "datetime" : "Date/Time Field: Enter a date with time as text or press either F2 or Shift-F10 to open the calendar popup.",
        "picker" : "Press F2 or Shift-F10 to open the dialog.",
        "typeahead" : "Enter a query to search.",
        "time" : "Time Field: Enter a time as text.",
        "number" : "Number Field: Enter a number as text.",
        "combo-input" : "Select Field: Press Down to show all options or type to find the value.",
        "combo-button" : "Select Field: Press Down to show all options or type a letter to filter values.",
        "spinner" : "Press up to increment the value and press down to decrement the value.",
        "calendar" : "Calendar: Use number keys (1 through 4) to switch between views, N for next date range, P for previous date range. When in Week or Day view C to create event, E to edit selected event, Up to move selected event earlier in the day, Down to move selected event later in the day, Left and Right to move between selected events, Delete to remove the selected event, Ctrl+Up and Ctrl+Down to modify the selected event's end date, Shift+Up and Shift+Down to modify the selected event's start date."
    },
    "logicalDate": {
        "customDay": "Custom date",
        "plusDay": "days",
        "plusWeek": "weeks",
        "plusMonth": "months",
        "plusQuarter": "quarters",
        "plusYear": "years"
    }
}
;});

define('oui/utils/language',[
    'underscore',
    'oui/json/strings.json'
], function (_, strings) {
    'use strict';

    /**
     * @module language
     * @private
     */

    /**
     * @typedef {Object} Locale
     * @global
     * @public
     * @desc These are the default strings provided by OUI. They are meant to be overriden within the product they are used
     * using the {@link module:oui.lang|lang} function of OUI. These same options can be used on a per-module basis to
     * customize various strings.
     *
     * @property {String} [loading="Loading"]
     * @property {String} [help="Help"]
     * @property {String} [today="Today"]
     * @property {String} [ok="OK"]
     * @property {String} [cancel="Cancel"]
     * @property {String} [remove="Remove"]
     * @property {String} [close="Close"]
     * @property {String} [expanded="Expanded"]
     * @property {String} [collapsed="Collapsed"]
     * @property {String} [locked="Locked"]
     * @property {String} [fullScreen="Full Screen"]
     * @property {String} [selectAll="Select All"]
     * @property {String} [newValue="(New Value)"]
     * @property {Object} [calendar]
     * @property {String} [calendar.nextYear="Next Year"]
     * @property {String} [calendar.prevYear="Previous Year"]
     * @property {String} [calendar.nextMonth="Next Month"]
     * @property {String} [calendar.prevMonth="Previous Month"]
     * @property {String} [calendar.selectDate="Select Date"]
     * @property {String} [calendar.standard="Standard"]
     * @property {String} [calendar.unavailable="Unavailable"]
     * @property {String} [calendar.exception="Exception"]
     * @property {String} [calendar.previous="Previous"]
     * @property {String} [calendar.next="Next"]
     * @property {String} [calendar.noGroup="No Group"]
     * @property {String} [startDate="Start Date"]
     * @property {String} [endDate="End Date"]
     * @property {String} [event_create_title="Create Event"]
     * @property {String} [event_update_title="Update Event"]
     * @property {Object} [views]
     * @property {String} [views.day="Day"]
     * @property {String} [views.week="Week"]
     * @property {String} [views.month="Month"]
     * @property {String} [views.year="Year"]
     * @property {String} [more="More..."]
     * @property {Object} [spinner]
     * @property {String} [spinner.increment="Increment"]
     * @property {String} [spinner.decrement="Decrement"]
     * @property {String} [menu="Dropdown menu"]
     * @property {String} [test="Test"]
     * @property {String} [clear="Clear"]
     * @property {String} [resize="Resize"]
     * @property {Object} [collapse]
     * @property {String} [collapse.up="Collapse Up"]
     * @property {String} [collapse.down="Collapse Down"]
     * @property {String} [collapse.left="Collapse Left"]
     * @property {String} [collapse.right="Collapse Right"]
     * @property {String} [search="Search"]
     * @property {String} [placeholder="Search"]
     * @property {String} [notification="{0} selected. Click here to remove filter."]
     * @property {String} [no_results_found="No Results Found"]
     * @property {String} [empty_tree="No Results"]
     * @property {Object} [picker]
     * @property {String} [picker.select="Select"]
     * @property {String} [browse="Browse"]
     * @property {Object} [editor]
     * @property {String} [editor.title="Rich Text Editor"]
     * @property {String} [editor.fontColor="Font Color"]
     * @property {String} [editor.backgroundColor="Background Color"]
     * @property {String} [editor.bold="Bold"]
     * @property {String} [editor.italic="Italic"]
     * @property {String} [editor.underline="Underline"]
     * @property {String} [editor.strikeThrough="Strikethrough"]
     * @property {String} [editor.subscript="Subscript"]
     * @property {String} [editor.superscript="Superscript"]
     * @property {String} [editor.justifyLeft="Left Justify"]
     * @property {String} [editor.justifyCenter="Center"]
     * @property {String} [editor.justifyRight="Right Justify"]
     * @property {String} [editor.orderedList="Ordered List"]
     * @property {String} [editor.unorderedList="Unordered List"]
     * @property {String} [editor.increaseIndent="Increase Indent"]
     * @property {String} [editor.decreaseIndent="Decrease Indent"]
     * @property {String} [editor.insertLink="Insert Link"]
     * @property {String} [editor.removeLink="Remove Link"]
     * @property {String} [editor.fontSize="Font Size"]
     * @property {String} [editor.fontName="Font Name"]
     * @property {String} [editor.fontFormat="Font Format"]
     * @property {String} [editor.linkText="Link Text"]
     * @property {String} [editor.openIn="Open in"]
     * @property {String} [editor.openInCurrent="Current Tab"]
     * @property {String} [editor.openInNew="New Tab or Window"]
     * @property {String} [editor.url="URL"]
     * @property {Object} [icon]
     * @property {String} [icon.select="Select Icon"]
     * @property {Object} [color]
     * @property {String} [color.select="Select Color"]
     * @property {String} [color.empty="Empty"]
     * @property {String} [color.000000="Black"]
     * @property {String} [color.CCCCCC="Light Gray"]
     * @property {String} [color.666666="Dark Gray"]
     * @property {String} [color.C1CBD5="Light Blue Gray"]
     * @property {String} [color.8B9AB2="Blue Gray"]
     * @property {String} [color.3E577F="Dark Blue Gray"]
     * @property {String} [color.0E2D5F="Darkest Blue Gray"]
     * @property {String} [color.DBDCE0="Light Gray Blue"]
     * @property {String} [color.B6BAC1="Gray Blue"]
     * @property {String} [color.616977="Dark Gray Blue"]
     * @property {String} [color.495363="Darkest Gray Blue"]
     * @property {String} [color.7FCDEE="Light Blue"]
     * @property {String} [color.66AAD7="Blue"]
     * @property {String} [color.0072BC="Dark Blue"]
     * @property {String} [color.0F527D="Darkest Blue"]
     * @property {String} [color.99DDD8="Light Mint Green"]
     * @property {String} [color.66C4BE="Mint Green"]
     * @property {String} [color.009C92="Dark Mint Green"]
     * @property {String} [color.00857C="Darkest Mint Green"]
     * @property {String} [color.CBE5CB="Lightest Green"]
     * @property {String} [color.AFDF8E="Light Green"]
     * @property {String} [color.65B265="Green"]
     * @property {String} [color.008000="Dark Green"]
     * @property {String} [color.FFF59C="Light Yellow"]
     * @property {String} [color.FFEF5E="Yellow"]
     * @property {String} [color.FFDD00="Dark Yellow"]
     * @property {String} [color.F2C019="Darkest Yellow"]
     * @property {String} [color.FCD580="Light Orange"]
     * @property {String} [color.FAAB01="Orange"]
     * @property {String} [color.E3870A="Dark Orange"]
     * @property {String} [color.BE5709="Brown"]
     * @property {String} [color.F5C2C2="Light Pink"]
     * @property {String} [color.EF9A9A="Pink"]
     * @property {String} [color.D80202="Red"]
     * @property {String} [color.A20202="Maroon"]
     * @property {String} [color.EACEE7="Light Purple Pink"]
     * @property {String} [color.D59ECF="Purple Pink"]
     * @property {String} [color.AB3DA0="Dark Purple Pink"]
     * @property {String} [color.710966="Darkest Purple Pink"]
     * @property {String} [color.CBC0DB="Light Purple"]
     * @property {String} [color.9881B7="Purple"]
     * @property {String} [color.6B40AD="Dark Purple"]
     * @property {String} [color.503082="Darkest Purple"]
     * @property {String} [color.FFFFFF="White"]
     * @property {Object} [errors]
     * @property {String} [errors.startsWith="{0} must start with either http:// or https://"]
     * @property {Object} [instructions]
     * @property {String} [instructions.date="Date Field: Enter a date as text or press either F2 or Shift-F10 to open calendar popup."]
     * @property {String} [instructions.datetime="Date/Time Field: Enter a date with time as text or press either F2 or Shift-F10 to open the calendar popup."]
     * @property {String} [instructions.picker="Press F2 or Shift-F10 to open the dialog."]
     * @property {String} [instructions.typeahead="Enter a query to search."]
     * @property {String} [instructions.time="Time Field: Enter a time as text."]
     * @property {String} [instructions.number="Number Field: Enter a number as text."]
     * @property {String} [instructions.combo-input="Select Field: Press Down to show all options or type to find the value."]
     * @property {String} [instructions.spinner="Press up to increment the value and press down to decrement the value."]
     * @property {String} [instructions.calendar="Calendar: Use number keys (1 through 4) to switch between views,
     *  N for next date range, P for previous date range. When in Week or Day view C to create event,
     *  E to edit selected event, Up to move selected event earlier in the day, Down to move selected event later in the day,
     *  Left and Right to move between selected events, Delete to remove the selected event,
     *  Ctrl+Up and Ctrl+Down to modify the selected event's end date, Shift+Up and Shift+Down to modify the selected event's start date."]
     */
    var _lang = strings;

    return {
        /**
         * Extends/returns the language used by OUI.
         *
         * If no arguments are given it will return the language object otherwise it will extend the defaults with
         * the provided object.
         *
         * @function lang
         * @param {Object} [language] key value pairs of localization key to localized string value
         * @returns {Object} The current language object used by OUI to construct templates
         */
        lang: function (language) {
            if (!arguments.length) {
                return _lang;
            }

            _lang = _.merge({}, strings, language);
        }
    };
});

/**
 * @module defaults
 * @private
 */
define('oui/utils/defaults',[
    'underscore'
], function (_) {
    'use strict';

    var _defaults = {};

    /**
     * @function defaults
     * @param {Object} defs The defaults to set for the library
     * @returns {Object} The defaults set after invocation
     */
    return function (defs) {
        if (!arguments.length) {
            return _defaults;
        }

        _defaults = _.merge(defs, _defaults);
    };
});

define('oui/utils/dom',[
    'jquery',
    'underscore',
    'logger'
], function ($, _, logger) {
    'use strict';

    /**
     * Gets the warning string for returning a widget with multiple values
     *
     * @method dom~getMultiValueWarning
     * @private
     *
     * @param {String} str
     *
     * @returns {String}
     */
    function getMultiValueWarning(str) {
        return [
            'Returning object with multiple values:',
            str + '.',
            'Be prepared to pick the widget out of the object or specify the name during the get call'
        ].join(' ');
    }

    /**
     * @module dom
     * @private
     */

    return {
        /**
         * Stores a widget with element to get them when the user has an element later.
         *
         * @method put
         * @param {jQuery} element The element to store the widget with
         * @param {Widget} widget The widget to store with the element.
         */
        put: function (element, widget) {
            var data = this.get(element, true) || {};
            data[widget.name()] = widget;

            if ($(element).hasClass('orcl-widget')) {
                $(element).data('oui-control', data);
            } else if ($(element).closest('.orcl-widget').length > 0) {
                $(element).closest('.orcl-widget').data('oui-control', data);
            }
        },
        /**
         * Gets the widget(s) associated with the element
         *
         * Will only turn the matching widget if name is specified otherwise returns an Object of name -> widget.
         *
         * **Note**: Boolean parameter for name is reserved for internal use only.
         *
         * @method get
         * @param {jQuery} element
         * @param {String|Boolean} [name] What specific widget to find.
         * @returns {?Object|Widget} Either a map of widgets keyed by name or the specific widget.
         */
        get: function (element, name) {
            var data;

            if ($(element).hasClass('orcl-widget')) {
                data = $(element).data('oui-control');
            } else if ($(element).closest('.orcl-widget').length > 0) {
                data = $(element).closest('.orcl-widget').data('oui-control');
            } else {
                return null;
            }

            if (name === true) {
                return data;
            }

            if (data && name) {
                return data[name];
            }

            //_.keys has protection for calling on non-object
            //Place _.isObject here as extra? protection
            if (_.isObject(data) && _.keys(data).length === 1) {
                return _.chain(data).values().flatten().value()[0];
            }

            if (data != null && _.keys(data).length > 1) {
                logger.warn(getMultiValueWarning(_.keys(data).join(' ')));
                if (data.picker) {
                    logger.info('Proxying functions ("data", "selected" and "value") for picker');
                    data.data = data.picker.data.bind(data.picker);
                    data.selected = data.picker.selected.bind(data.picker);
                    data.value = data.picker.value.bind(data.picker);
                }
            }
            return data;
        }
    };
});

define('oui/base/widget',['require','underscore','jquery','moment','oui/utils/subscribable','oui/utils/uniqueId','oui/utils/language','oui/utils/defaults','oui/utils/dom','logger'],function (require) { //jshint ignore:line
    'use strict';

    var _ = require('underscore');
    var $ = require('jquery');
    var moment = require('moment');
    var Subscribable = require('oui/utils/subscribable');
    var uniqueId = require('oui/utils/uniqueId');
    var language = require('oui/utils/language');
    var globalDefaults = require('oui/utils/defaults');
    var dom = require('oui/utils/dom');
    var logger = require('logger');

    var wp;

    /**
     * @method Widget~binder
     * @param {Object} ctx
     *
     * @returns {Function}
     */
    function binder(ctx) {
        return function (value) {
            if (_.isFunction(value)) {
                return _.bind(value, ctx);
            } else {
                return value;
            }
        };
    }

    /**
     * @event Widget#option-change
     *
     * @param {String} key The key of the option that was changed
     * @param {*} value The value the passed in option updated to.
     */

    /**
     * The jQuery object.
     *
     * @external "$"
     */

    //Monky-patching jQuery because they do not trigger an event
    //when an element is removed from the DOM.
    //DOMNodeRemoved is deprecated - and when people want to use $.detach it would
    //    assume destroyed
    //MutationObservers are the future - but like DOMNodeRemoved $.detach would not work.
    var _clean = $.cleanData;
    /**
     * @function external:"$".cleanData
     * @param {Array<HTMLElement>} elems The elements to clean
     * @returns {*} - The result of the extended function ($.cleanData)
     */
    $.cleanData = function (elems) {
        var events;
        var i;
        var elem;

        for (i = 0; (elem = elems[i]) != null; i++) {
            try {
                events = $._data(elem, 'events');
                if (events && events.destroy) {
                    $(elem).triggerHandler('destroy.oui');
                }
            } catch (e) {
            }
        }
        return _clean(elems);
    };

    /**
     * Determines whether the argument should be treated as initData.
     *
     * The default behavior is to ignore Objects (jquery, configs). The arguments would both be considered initData as well
     * as element or options which is why the callback exists. For widgets, like 'slider', the initData is an Object
     * ({ start: Number, end: Number}) and must be properly assigned to the initData variable.
     *
     * @method Widget~isInitData
     * @private
     *
     * @param {Function?} cb
     * @param {*} arg
     *
     * @returns {Boolean}
     */
    function isInitData(cb, arg) {
        return !_.isObject(arg) ||
                _.isArray(arg) ||
                moment.isMoment(arg) ||
                _.isFunction(arg) ||
                (cb && cb(arg));
    }

    /**
     * Customizes the merge operation.
     *
     * Throws away original arrays instead of unioning them together.
     *
     * See {@link https://github.com/lodash/lodash/blob/3.10.1/doc/README.md#_mergeobject-sources-customizer-thisarg|lodash}
     * for more information regarding the merge cusomizer.
     *
     * @method Widget~mergeCustomizer
     * @private
     *
     * @param {*} a
     * @param {*} b
     *
     * @returns {*}
     */
    function mergeCustomizer(a, b) {
        if (_.isArray(a) && _.isArray(b)) {
            return b;
        }
    }

    /**
     * Utility method to deal with optional arguments for the widget
     *
     * @method Widget~getArgs
     * @private
     *
     * @param {arguments} args
     * @param {Function} cb
     *
     * @returns {Object}
     */
    function getArgs(args, cb) {
        var length = args.length;
        var arg;
        var organizedArgs = {
            options: null,
            element: null,
            initData: null
        };

        for (var i = 0; i < length; ++i) {
            arg = args[i];

            if (arg == null) {
                arg = null;
            }

            if (organizedArgs.options === null && _.isObject(arg) && !arg.jquery && !moment.isMoment(arg)) {
                organizedArgs.options = arg;
            }

            if (organizedArgs.element === null && _.isObject(arg) && arg.jquery) {
                organizedArgs.element = arg;
            }

            if (organizedArgs.initData === null && isInitData(cb, arg)) {
                organizedArgs.initData = arg;
            }
        }

        if (organizedArgs.element != null && organizedArgs.element.jquery && organizedArgs.element.length === 0) {
            //Log about an invalid selector being passed in - setting element to null
            logger.warn('An invalid selector was supplied as the element - constructing new element for creation');
            organizedArgs.element = null;
        }

        return organizedArgs;
    }

    /**
     * Widget base class. Sets up the optional arguments (options, initData and element) and maintains a unique ID across all
     * widgets.
     *
     * Widgets trigger each event with Subscribable as well as DOM Events. For DOM events each event is prefixed with 'oui-'.
     *
     * @class Widget
     * @private
     * @extends Subscribable
     * @since 1.3.0
     * @param {Object} [options] The widget's options
     * @param {jQuery} [element] JQuery object that the UI will be built
     *   on top of.
     * @param {Number|Array|moment|String} [initData] The data to be initialized
     *  with.
     */
    function Widget() {
        var defaults = this.defaults();
        var options = null;
        var element = null;
        var initData = null;
        var val;

        //Super call
        Subscribable.call(this);
        var args = getArgs(arguments, this.isInitData.bind(this));
        options = args.options;
        element = args.element;
        initData = args.initData;

        if (!_.isNull(element) && _.isNull(initData) && element.is('input') && element.val() !== '') {
            val = element.val();
            if (_.has(options, 'formatter') && options.formatter.validate(val)) {
                initData = options.formatter.unformat(val);
            } else {
                initData = val;
            }
        }

        var defaultKey = this.name();

        if (defaults.type || (options && options.type)) {
            defaultKey += '.' + _.get(options, 'type', defaults.type);
        }

        /**
         * The options for the widget.
         *
         * @private
         * @type {Object}
         * @memberof! Widget.prototype
         */
        this._options = _.merge({}, defaults,
                                { locale: language.lang() },
                                _.get(globalDefaults(), defaultKey, {}),
                                options || {}, mergeCustomizer);
        /**
         * The element that was passed into when creating the widget.
         *
         * @private
         * @type {Object}
         * @memberof! Widget.prototype
         */
        this._element = element;
        /**
         * The data to use as the objects initial data.
         *
         * @private
         * @type {Object}
         * @memberof! Widget.prototype
         */
        this._initData = initData;
        /**
         * Each widget gets an id that will be unique across widgets on a page.
         *
         * @type {Number}
         * @private
         * @memberof! Widget.prototype
         */
        this._id = uniqueId('widget');

        Object.defineProperty(this, '_data', {
            enumerable: false,
            configurable: true,
            get: _.noop,
            set: _.noop
        });

        this._options.storeFns = _.mapValues(this._options.storeFns, binder(this));

        this._options.viewFns = _.mapValues(this._options.viewFns, binder(this));

        this._settings = {
            id: this._id,
            name: this.name()
        };
    }

    Widget.prototype = Object.create(Subscribable.prototype);
    wp = Widget.prototype;

    wp.constructor = Widget;

    /**
     * Abstract method for getting the widgets name.
     *
     * @method Widget#name
     * @returns {String} name The name of the widget (defaults to 'widget').
     */
    wp.name = function () {
        return 'widget';
    };

    /**
     * Abstract method for getting the widgets name.
     *
     * @deprecated
     * @see Widget#name
     * @method Widget#getName
     * @returns {String} name - The name of the widget.
     */
    wp.getName = function () {
        return this.name();
    };

    /**
     * Initializes the widget after it is on the DOM. This is necessary for all widgets to let them know that they can start
     * using computed CSS properties if needed. Not all widgets use CSS properties but dynamic positioning of additional DOM
     * structure will need to use some of these properties.
     *
     * Sets up the remove listener to clean itself up when it is removed from the DOM.
     *
     * **Note**: May not always return a reference to the widget. If init is supplied an object it will return the construction of the object instead.
     *
     * @method Widget#init
     * @chainable
     */
    wp.init = function () {
        var that = this;
        var store;

        this.element().on('destroy.oui', function () {
            that.destroy({ self: false });
        });

        if (!_.isUndefined(this._view) && _.isFunction(this._view.init)) {
            this._view.init();
        }

        dom.put(this.element(), this);

        store = this._init.apply(this, arguments);

        if (store != null) {
            //TODO: should we log here?
            return store;
        } else {
            return this;
        }
    };

    wp.isInitData = _.constant(false);

    /**
     * Empty function that each widget can override to provide
     * custom initialization logic.
     *
     * @method Widget#_init
     * @private
     */
    wp._init = _.noop;

    /**
     * Function to explicitly remove the remove event listener.
     *
     * @method Widget#removeOff
     * @private
     */
    wp.removeOff = function () {
        this.element().off('destroy.oui');
    };

    /**
     * Function to be called when the widget is either removed from the DOM
     * or is called directly.
     *
     * This method should not be overridden by subclasses.
     *
     * @method Widget#destroy
     * @param {Object} cleanUp Options for cleaning up widget.
     * @param {Boolean} [cleanUp.self=true] Whether to clean up the DOM element or false if jQuery is cleaning up the DOM
     * element.
     */
    wp.destroy = function (cleanUp) {
        if (_.isUndefined(cleanUp)) {
            cleanUp = { self: true };
        }
        this.removeOff();

        if (!_.isUndefined(this._view)) {
            this._view.destroy(cleanUp);
        }

        this._destroy(cleanUp);

        if (this._view) {
            this._view.trigger('destroy');
        }
    };

    /**
     * Empty function that each widget can implement to provide their default
     * destroy function.
     *
     * @method Widget#_destroy
     * @private
     */
    wp._destroy = _.noop;

    /**
     * Getter/setter for the current value of the widget.
     *
     * @method Widget#data
     * @param {*} [data] The new data/value to set on the widget
     * @returns {*} The current data/value of the widget.
     */
    wp.data = function (data) {
        if (!arguments.length) {
            if (!this.isValid() && this._controller && this._controller.error) {
                return this._controller.error().data;
            }
            return this._data;
        } else {
            this._data = data;
        }
    };

    /**
     * Returns the element that represents this widget.
     *
     * @method Widget#element
     * @returns {HTMLElement} The element that makes up the component for placement on the DOM
     */
    /**
     * Alias for element.
     *
     * @see Widget#element
     * @method Widget#getElement
     */
    wp.element = wp.getElement = function () {
        if (!_.isUndefined(this._view)) {
            return this._view.element();
        }
    };

    /**
     * Returns the default options for a widget.
     *
     * Used by widgets that contain other other widgets such as the
     * DatePicker.
     *
     * @method Widget#defaults
     */
    wp.defaults = function () {
        logger.error('"defaults" method must be implemented in subclass');
    };

    /**
     * Update a single option at the specific key.
     *
     * Currently, does not work with nested keys.
     *
     * @method Widget#updateOption
     * @param {String} key The attribute to update.
     * @param {String|Date|Number|null} value The value of the attribute.
     * @fires Widget#event:option-change
     */
    wp.updateOption = function (key, value) {
        if (_.isFunction(this[key])) {
            if (this[key]() !== value) {
                this[key](value);
                this.trigger('option-change', key, value);
            }
        } else if (_.has(this._options, key)) {
            if (_.get(this._options, key) !== value) {
                this._options = _.set(this._options, key, value);
                this.trigger('option-change', key, value);
            }
        } else {
            logger.warn('Attribute: ' + key + ' is not valid.');
        }
    };

    /**
     * Update multiple options at once.
     *
     * @method Widget#updateOptions
     * @param {Object} map The map of attributes to update.
     */
    wp.updateOptions = function (map) {
        var that = this;
        var convertMap = {};

        if (_.isString(map) || arguments.length === 2) {
            logger.warn('Please use updateOption when only setting a single option.');
            logger.info('Converting to object.');
            convertMap[map] = arguments[1];
            map = convertMap;
        }

        _.each(map, function (value, key) {
            that.updateOption(key, value);
        });
    };

    /**
     * Get/Set options within the widget.
     *
     * @method Widget#options
     * @param {Object} [map] The map of options to update
     * @returns {Object} - The current collection of options within the widget.
     */
    wp.options = function (map) {
        if (!arguments.length) {
            return _.cloneDeep(this._options);
        }

        this.updateOptions(map);
    };

    /**
     * Get specified option from current options within the widget
     *
     * @method Widget#getOption
     * @param {String} name The path to the option.
     * @returns {*|undefined} The found option or undefined if it was not found.
     */
    wp.getOption = function (name) {
        return _.get(this._options, name);
    };

    /**
     * Disables/Enables the widget
     *
     * @method Widget#disable
     * @param {Boolean} disabled Whether or not the widget should be disabled.
     */
    wp.disable = function (disabled) {
        if (!arguments.length) {
            if (this._view) {
                return this._view.isDisabled();
            }
            return null;
        }
        if (this._view && _.isBoolean(disabled)) {
            this._view.disable(disabled);
        }
    };

    /**
     * Locks or unlocks the widget.
     *
     * Used when the widget should be displayed but user does not have security rights to update the value.
     *
     * If the parameter is { locked.view : true } or locked is true then the widget will empty the data contained.
     * If the parameter is { locked.edit : true } then the widget applies the locked style (disables widget and displays
     * lock) without removing the data.
     *
     * @method Widget#lock
     * @param {Object|Boolean} locked Whether the widget is locked or not.
     * @param {Boolean} [locked.edit=false] Whether edit is allowed on the widget.
     * @param {Boolean} [locked.view=false] Whether the view is locked.
     */
    wp.lock = function (locked) {
        if (!arguments.length) {
            if (this._view) {
                return this._view.isLocked();
            }
            return null;
        }
        var emptyData = (locked === true) || locked.view || false;
        var lockView = emptyData || locked.edit || false;
        if (this._view) {

            this._view.lock(lockView);

            if (emptyData) {
                this.data(null);
            }
        }
    };

    /**
     * @method Widget#visible
     * @returns {Boolean}
     */
    /**
     * @method Widget#visible
     * @param {Boolean} visible
     */
    wp.visible = function (visible) {
        if (!arguments.length) {
            return !this.element().hasClass('hidden');
        }
        if (this._view && this._view.visible) {
            this._view.visible(visible);
        }
    };

    /**
     * @method Widget#readonly
     * @returns {Boolean?}
     */
    /**
     * @method Widget#readonly
     * @param {Boolean} readonly
     */
    wp.readonly = function (readonly) {
        if (!arguments.length) {
            if (this._view) {
                return this._view.isReadOnly();
            }

            return null;
        }
        if (this._view && this._view.readonly) {
            this._view.readonly(readonly);
        }
    };

    /**
     * Retrieves or sets the id attribute of the related DOM element that represents the widget.
     *
     * @method Widget#id
     * @since 2.7.3
     * @param {Object|String} [id] The id or settings to use when setting the attribute on the DOM element.
     * @returns {String} The id of the widget as used in the DOM
     */
    wp.id = function (id) {
        if (!arguments.length) {
            return this._view.id();
        }

        this._view.id(id);
    };

    /**
     * Used when triggering events and arguments need to be manipulated/translated.
     *
     * @method Widget#_translate
     * @since 2.7.4
     * @private
     * @returns {Array} arguments The translated arguments for the event being triggered.
     */
    wp._translate = function () {
        return arguments;
    };

    /**
     * Takes in the subscribable to subscribe to and sets up forwarding method.
     *
     * This function will not trigger any event if the event is described as 'silent'.
     *
     * Triggers DOM events with oui- prefixed to all events as well as triggering subscribable events.
     *
     * @method Widget#_setupEvents
     * @since 2.7.3
     * @protected
     * @param {Array<Subscribable>|Subscribable} [subscribable] The subscribables to add event listeners for.
     */
    wp._setupEvents = function (subscribable) {
        var viewEl = this._view.element();
        var that = this;

        /**
         * @method Widget#_setupEvents~addEvents
         * @private
         * @param {Subscribable} sub The subscribable to attach the special '*' listener on to for forwarding events.
         */
        function addEvents(sub) {
            sub.on('*', function (name) {
                if (!this.silent && !(that.name() === 'picker' && sub.name && sub.name() === 'tree-panel' && name === 'change')) {
                    var event;
                    var returnValue;
                    var args = _.chain(that._translate.apply(that, arguments))
                                .toArray()
                                .value();

                    //Reset name after _translate
                    name = args[0];
                    event = $.Event('oui-' + name);

                    if (that._element && that._element.length > 0) {
                        that._element.trigger(event, [that].concat(args));
                    } else {
                        viewEl.trigger(event, [that].concat(args));
                    }

                    if (that._options.events && that._options.events[name]) {
                        that._options.events[name].apply(that, _.rest(args, 1));
                    }

                    returnValue = that.trigger.apply(that, args);

                    if (returnValue == null) {
                        return event.result;
                    } else {
                        return returnValue;
                    }
                }
            });
        }

        if (_.isArray(subscribable)) {
            //Context will be undefined in addEvents after this call
            _.each(subscribable, addEvents);
        } else {
            addEvents(subscribable);
        }

    };

    /**
     * @method Widget#empty
     */
    wp.empty = function () {
        if (this.invalidate) {
            this.invalidate();
        }
        this.value(null);
    };

    /**
     * @method Widget#isValid
     * @returns {Boolean}
     */
    wp.isValid = function () {
        return true;
    };

    /**
     * @method Widget#focus
     *
     * @returns {null}
     */
    /**
     * @method Widget#focus
     * @param {Boolean} focused
     */
    wp.focus = function (focused) {
        if (!arguments.length) {
            return null;
        }
        if (this._view && this._view.focus) {
            this._view.focus(focused == null || focused === true);
        }
    };

    /**
     * This will trigger any event handlers on the error and change calls.
     *
     * @method Widget#validate
     */
    wp.validate = function () {
        if (this.isValid()) {
            this.data(this.data());
        } else {
            this.data(this.error().data);
        }
    };

    /**
     *
     * Add shortcuts to the widget.
     *
     * Shortcut keys should be separated by a hyphen. Modifiers can be 'CTRL', 'SHIFT', 'ALT', 'META'.
     *
     * @example
     *
     * widget.addShortcuts({
     *     'SHIFT-A': function (event) {
     *         //Do something on the calendar when Shift + A is down
     *     }
     * });
     *
     * @method Widget#addShortcuts
     *
     * @param {Object<String, Function>} shortcuts
     */
    wp.addShortcuts = function (shortcuts) {
        if (this._view && this._view.addShortcuts) {
            var nonConflicts = _.chain(shortcuts)
                .mapValues(this._view.bindShortcut.bind(this._view, this))
                .value();
            this._view.addShortcuts(nonConflicts);
        } else {
            logger.info('This widget does not support adding custom shortcuts');
        }
    };

    return Widget;

});

define('oui/base/input',[
    'underscore',
    'oui/base/widget'
], function (_, Widget) {
        'use strict';

        /**
         * @method Input~getRefList
         * @private
         *
         * @param {Array<HTMLElement>} arr
         * @returns {String}
         *
         */
        function getRefList(arr) {
            var map = _.map(arr, function (node) {
                if (_.isString(node)) {
                    return node;
                } else if (node.id) {
                    return node.id;
                }
            });

            return map.join(' ');
        }

        /**
         * @typedef Input~api
         *
         * @property {String} placeholder Getter/Setter for the placeholder property of an input
         * @property {String} labeledBy Getter/Setter for the aria-labelledby property of an input
         */

        /**
         * Used as a point of extension to specific input widgets.
         *
         * @class Input
         * @private
         * @extends Widget
         * @since 1.3.0
         */
        function Input(options, element, initData) {
            Widget.call(this, options, element, initData);
        }

        Input.prototype = Object.create(Widget.prototype);

        Input.prototype.constructor = Input;

        /**
         * Gets an API wrapper around the HTML input element within the control to update various properties.
         *
         * **Properties Supported**
         *
         *   * placeholder
         *   * labeledBy
         *
         * @example
         * var textField = new oui.TextField();
         * textField.getInput().placeholder = 'Search';
         *
         * @method Input#getInput
         * @returns {Input~api} A frozen API with a subset of properties the HTMLInputElement has
         */
        Input.prototype.getInput = function () {
            var input = this.element().find('input, textarea').first()[0];

            return Object.freeze({
                get placeholder() { //jscs:ignore jsDoc
                    return input.placeholder;
                },
                set placeholder(text) { //jscs:ignore jsDoc
                    input.placeholder = text;
                },
                get labeledBy() { //jscs:ignore jsDoc
                    return input.getAttribute('aria-labelledby');
                },
                set labeledBy(label) { //jscs:ignore jsDoc
                    if (_.isArray(label)) {
                        input.setAttribute('aria-labelledby', getRefList(label));
                    } else if (_.isString(label)) {
                        input.setAttribute('aria-labelledby', label);
                    } else if (label.id) {
                        input.setAttribute('aria-labelledby', label.id);
                    }
                }
            });
        };

        return Input;
    }
);


define('template-0', ['underscore'], function (_) { return function(data) {
var __t, __p = '';
__p += '<span class="spoken-text instructions" data-name="' +
((__t = ( data.name )) == null ? '' : __t) +
'">' +
((__t = ( data.instructions )) == null ? '' : __t) +
'</span>';
return __p
};});


define("oui/json/keyCode.json", function(){ return {
    "UP": 38,
    "DOWN": 40,
    "LEFT": 37,
    "RIGHT": 39,

    "SPACE": 32,
    "TAB": 9,
    "BACKSPACE": 8,
    "ENTER": 13,
    "ESC": 27,

    "SHIFT": 16,
    "CTRL": 17,
    "ALT": 18,

    "PAUSE": 19,
    "CAPS": 20,
    "PAGEUP": 33,
    "PAGEDOWN": 34,
    "END": 35,
    "HOME": 36,
    "INSERT": 45,
    "DELETE": 46,

    "F1": 112,
    "F2": 113,
    "F3": 114,
    "F4": 115,
    "F5": 116,
    "F6": 117,
    "F7": 118,
    "F8": 119,
    "F9": 120,
    "F10": 121,
    "F11": 122,
    "F12": 123,

    "NUMLOCK": 144,
    "SCROLLLOCK": 145,

    "LEFTSUPER": 91,
    "RIGHTSUPER": 92,
    "CONTEXTMENU": 93
}
;});

define('oui/utils/keyCode',['oui/json/keyCode.json'], function (keyCode) {
    'use strict';

    /**
     * @module keyCode
     * @private
     */

    /**
     * @exports keyCode
     * @type {Object}
     */
    return keyCode;
});

define('oui/utils/flexEqual',[
    'underscore'
], function (_) {
    'use strict';

    /**
     * @module flexEqual
     */

    /*
     *
     * @alias module:flexEqual
     * @method isEqual
     * @private
     *
     * @param {Object} node1
     * @param {Object} node2
     *
     * @returns {Boolean}
     */
    return function (node1, node2) {
        var keys1;
        var keys2;
        var comparisonKeys;

        //If they are the same reference return true
        if (node1 === node2) {
            return true;
        }

        keys1 = _.keys(node1);
        keys2 = _.keys(node2);
        comparisonKeys = _.intersection(keys1, keys2);

        //If there are no shared keys but at least one of the objects is not empty
        if (comparisonKeys.length === 0 && (keys1.length > 0 || keys2.length > 0)) {
            return false;
        }

        return _.isEqual(_.pick(node1, comparisonKeys), _.pick(node2, comparisonKeys));
    };
});

define('oui/mixins/shortcut',[
    'jquery',
    'underscore',
    'oui/utils/keyCode',
    'oui/utils/flexEqual'
], function ($, _, keyCode, flexEqual) {
    'use strict';

    /**
     * @method shortcut~getKey
     * @private
     * @param {Array<String>} keys
     *
     * @returns {Number}
     */
    function getKey(keys) {
        var shortcutKey = _.last(keys);
        return keyCode[shortcutKey] || shortcutKey.charCodeAt(0);
    }

    /**
     * @method shortcut~handler
     * @private
     *
     * @param {Array} allShortcuts
     * @param {jQuery.Event} event
     */
    var handler = function (allShortcuts, event) {
        var shortcuts;

        if (!event.ctrlKey && event.metaKey) {
            event.ctrlKey = true;
            event.metaKey = false;
        }

        shortcuts = _.find(allShortcuts, function (shortcuts, key) {
            var keys = key.toUpperCase().split('-');
            var triggerKeyCode = getKey(keys);
            var comparison = {
                ctrlKey: _.contains(keys, 'CTRL'),
                shiftKey: _.contains(keys, 'SHIFT') && (event.type === 'keydown' || (event.type === 'keyup' && triggerKeyCode !== keyCode.SHIFT)),
                altKey: _.contains(keys, 'ALT'),
                metaKey: _.contains(keys, 'META'),
                keyCode: triggerKeyCode
            };
            return flexEqual(comparison, event);
        });

        if (_.isArray(shortcuts)) {
            shortcuts = shortcuts.reverse();
            _.each(shortcuts, function (shortcut) {

                if (!shortcut.delegate || $(event.target).is(shortcut.delegate)) {
                    if (shortcut.preventDefault === true || shortcut.preventDefault.call(this)) {
                        event.preventDefault();
                    }
                    if (shortcut.stopPropagation === true || shortcut.stopPropagation.call(this)) {
                        event.stopPropagation();
                    }
                    shortcut.cb(event);

                    if (event.isImmediatePropagationStopped()) {
                        return false;
                    }
                }
            }.bind(this));
        }
    };

    /**
     * Mixin for views to quickly add shortcut handlers.
     *
     * @mixin shortcut
     * @private
     */
    return {
        /**
         * @method shortcut~_initShortcutProps
         * @private
         */
        _initShortcutProps: function () {
            this._keyDownShortcuts = {};
            this._keyUpShortcuts = {};

            this._$el
                .on('keydown', handler.bind(this, this._keyDownShortcuts))
                .on('keyup', handler.bind(this, this._keyUpShortcuts));
        },
        /**
         * Adds a shortcut using a parseable format to a single common handler.
         *
         * Splits the name on the '-' and grabs the modifier(s) to use for the character.
         *
         * **Limitation** - can only have one callback per shortcut key
         *
         * @function shortcut#addShortcut
         * @param {String} name The event to split and set up the event handler for.
         * @param {Function|Object} callback The function to call when the shortcut keys are held down. (keydown)
         * @param {Object} allShortcutSettings
         */
        addShortcut: function (name, callback, allShortcutSettings) {
            var up = name === 'ESC';
            var shortcut;
            var preventDefault;
            var delegate;
            var stopPropagation;

            if (_.isUndefined(this._shortcutInit)) {
                this._initShortcutProps();
                this._shortcutInit = true;
            }

            if (_.isObject(callback) && !_.isFunction(callback)) {
                preventDefault = callback.prevention;
                stopPropagation = callback.stopPropagation;
                delegate = callback.delegate;

                if (callback.up != null) {
                    up = callback.up;
                }
                callback = callback.fn;
            }

            if (delegate == null && allShortcutSettings && allShortcutSettings.delegate != null) {
                delegate = allShortcutSettings.delegate;
            }

            shortcut = {
                name: name,
                preventDefault: preventDefault || _.constant(true),
                stopPropagation: stopPropagation || _.constant(false),
                delegate: delegate,
                cb: _.bind(callback, this)
            };

            if (!up) {
                if (this._keyDownShortcuts[name] == null) {
                    this._keyDownShortcuts[name] = [];
                }
                this._keyDownShortcuts[name].push(shortcut);
            } else {
                if (this._keyUpShortcuts[name] == null) {
                    this._keyUpShortcuts[name] = [];
                }
                this._keyUpShortcuts[name].push(shortcut);
            }

            return this;
        },
        /**
         * Uses addShortcut to add a collection of shortcuts at one time.
         *
         * @function shortcut#addShortcuts
         * @param {Object} shortcuts An object of key value pairs where the key is the shortcut and the value
         *     is the callback
         * @param {Object} allShortcutSettings
         */
        addShortcuts: function (shortcuts, allShortcutSettings) {
            var addShortcutFn = _.bind(this.addShortcut, this);

            _.each(shortcuts, function (shortcut, name) {
                var multiple = name.split(' ');
                if (multiple.length > 1) {
                    _.each(multiple, function (singleName) {
                        addShortcutFn(singleName, shortcut, allShortcutSettings);
                    });
                } else {
                    addShortcutFn(name, shortcut, allShortcutSettings);
                }
            });
        },
        /**
         * Removes the shortcut from the collection of shortcuts.
         *
         * @function shortcut#removeShortcut
         * @param {String} name The name of the shortcut to remove.
         */
        removeShortcut: function (name) {
            delete this._keyDownShortcuts[name];
            delete this._keyUpShortcuts[name];
        },
        /**
         * @function shortcut#bindShortcut
         *
         * @param {Object} context
         * @param {Object|Function} shortcut
         *
         * @returns {Object|Function}
         */
        bindShortcut: function (context, shortcut) {
            if (_.isFunction(shortcut)) {
                shortcut = shortcut.bind(context);
            } else {
                shortcut.fn = shortcut.fn.bind(context);
            }

            return shortcut;
        }
    };
});

/**
 * @module inputTypes
 * @private
 * @type {Object}
 */
define('oui/utils/inputTypes',[],function () {
    'use strict';

    var available = {};
    var elem;

    [
        'date',
        'datetime',
        'color',
        'datetime-local',
        'email',
        'month',
        'number',
        'range',
        'search',
        'tel',
        'time',
        'url',
        'week',
        'file'
    ].forEach(function (key) {
        elem = document.createElement('input');
        elem.setAttribute('type', key);

        available[key] = elem.type === key;
    });

    return available;

});

/**
 * Utility module to determine if user is on a mobile browser.
 *
 * @module detectMobile
 * @private
 */
define('oui/utils/detectMobile',[],function () {
    'use strict';

    return {
        isMobile: (/Android|webOS|iPhone|iPad|iPod|BlackBerry|Mobile/i).test(navigator.userAgent)
    };
});

define('oui/base/view',['require','underscore','jquery','oui/utils/subscribable','template-0','oui/mixins/shortcut','oui/utils/inputTypes','oui/utils/detectMobile'],function (require) { //jshint ignore:line
    'use strict';

    var _ = require('underscore');
    var $ = require('jquery');
    var Subscribable = require('oui/utils/subscribable');
    var instructions = require('template-0');
    var shortcut = require('oui/mixins/shortcut');
    var supportedInputTypes = require('oui/utils/inputTypes');
    var detectMobile = require('oui/utils/detectMobile');

    var eventType = {
        focusout: 'blur',
        focusin: 'focus'
    };

    //Used to shorten the View.prototype call below.
    var vp;

    //TODO: look at these parameters. Reorganize to match all children constructors (model, controller, element, options)
    /**
     * Super class view that will subscribe to a subscribable object.
     * Not made to be instantiated by itself.
     *
     * Sets up the element to use as the widget. Might need to construct an entirely new DOM element or wrap an existing one.
     *
     * If the library is used on mobile devices it will use the native controls to allow the keyboards to change
     * appropriately.
     *
     * @class View
     * @extends Subscribable
     * @private
     * @mixes shortcut
     * @param {Object} options The widget's options. (They are not view specific)
     * @param {Object} [element] The element that will be used as the widget's UI.
     * @param {Object} [controller] The element the view will call into when the
     *   component is interacted with.
     */
    function View(options, element, controller, settings) {
        //Super call
        Subscribable.call(this);

        if (detectMobile.isMobile && ((this._supported && this._supported()) || supportedInputTypes[this.nativeType(options)])) {
            /**
             * The HTML element that gets created and returned to be appended to the page.
             *
             * It is a JQuery object.
             *
             * @type {Object}
             * @private
             * @memberof! View.prototype
             */
            this._$el = this.buildNativeElement(options, element, controller);

            this._$el.addClass('native');
        } else {
            this._$el = this.buildElement(options, element, controller, settings);
            this._$el.append(instructions({
                name: settings.name,
                instructions: options.locale.instructions[options.type || settings.name]
            }));
        }

        this._$el.toggleClass('mobile', detectMobile.isMobile);

        this.id(settings);

        this._$el.toggleClass('fluid', options.fluid);

        /**
         * The options for the widget.
         *
         * @private
         * @type {Object}
         * @memberof! View.prototype
         */
        this._options = options;
        /**
         * The controller to use for the view.
         *
         * @private
         * @type {Controller}
         * @memberof! View.prototype
         */
        this._controller = controller;
        /**
         * Settings built from the widget that are passed into the view.
         *
         * Contains the id and the name of the widget.
         *
         * @private
         * @type {Object}
         * @memberof! View.prototype
         */
        this._settings = settings;

        this._attrs = {
            locked: false,
            disabled: false,
            readonly: false
        };

        var that = this;

        this._visited = false;

        if (this._options.showErrors === 'immediate') {
            this._visited = true;
            this.showErrors();
        }

        this._$el.on('focusin focusout', function (event) {
            if (that._options.showErrors === eventType[event.type]) {
                that._visited = true;
                that.showErrors();
            }
        });

        return this;
    }

    View.prototype = Object.create(Subscribable.prototype);
    vp = View.prototype;

    _.extend(vp, shortcut);

    vp.constructor = View;

    /**
     * Returns the native type of the widget if it has one otherwise returns null.
     *
     * @method View#nativeType
     * @returns {?String} type The native type of the widget.
     */
    vp.nativeType = function () {
        return null;
    };

    /**
     * Initializes the view. Sets up any computed CSS properties (heights, widths) and
     * sets the initial state of the component based on the init data (error, valid...)
     *
     * @method View#init
     */
    vp.init = function () {};

    /**
     * Constructs the view's specific element that events are bound to.
     *
     * Overridden, in children classes when they need an element.
     *
     * @method View#buildElement
     * @returns {jQuery} The default element
     */
    vp.buildElement = function () {
        return $();
    };

    /**
     * Constructs the view's native element.
     *
     * @method View#buildNativeElement
     * @returns {jQuery} The native element of the widget.
     */
    vp.buildNativeElement = function () {
        return $();
    };

    /**
     * Alias to this._$el.find(selector)
     *
     * Utility to not issuing global DOM searches.
     *
     * @method View#$
     * @param {String} selector The selector to search for in the view's element.
     * @returns {jQuery} The jQuery object represented by the selector.
     */
    vp.$ = function (selector) {
        return this._$el.find(selector);
    };

    /**
     * Injects other elements into this views element.
     *
     * @method View#inject
     * @param {Object} $element The jQuery object to be injected.
     * @param {String} selector The place to inject the object.
     */
    vp.inject = function ($element, selector) {
        this._$el.find(selector).append($element);
    };

    /**
     * Returns the view's element.
     *
     * @method View#element
     * @returns {Object} The jQuery object used by the view.
     */
    vp.element = vp.getElement = function () {
        return this._$el;
    };

    /**
     * Cleans up the views created elements.
     *
     * @method View#destroy
     * @param {Object} clean
     * @param {Boolean} clean.self
     */
    vp.destroy = function (clean) {

        if (clean.self) {
            this.element().remove();
        }
    };

    /**
     * Returns the element that identifies the widget.
     *
     * @method View#getIdentifyingElement
     *
     * @returns {jQuery}
     */
    vp.getIdentifyingElement = function () {
        return this._$el.find('input, textarea').first();
    };

    /**
     * Default id function for every view that can be overwritten if a view is not made up of an input element.
     *
     * Gets/Sets the id of the DOM element.
     *
     * @method View#id
     * @param {String|Object} [settings] The id or settings to use to set the id attribute of the DOM element.
     * @returns {String} id The id of the DOM element.
     */
    vp.id = function (settings) {
        var input = this.getIdentifyingElement();

        if (!arguments.length) {
            return input.attr('id');
        }

        if (input.length > 0 && _.isUndefined(input.attr('id'))) {
            if (_.isObject(settings)) {
                input.attr('id', 'oui-' + settings.name + '-' + settings.id);
            } else if (_.isString(settings)) {
                input.attr('id', settings);
            }
        }
    };

    /**
     * Returns whether the widget is currently in a readonly state and additional popup elements should not be created.
     *
     * @method View#isReadOnly
     * @returns {Boolean} whether the element is in a readonly state.
     */
    vp.isReadOnly = function () {
        return this._attrs.readonly;
    };

    /**
     * @method View#isDisabled
     * @returns {Boolean}
     */
    vp.isDisabled = function () {
        return this._attrs.disabled;
    };

    /**
     * @method View#isLocked
     * @returns {Boolean}
     */
    vp.isLocked = function () {
        return this._attrs.locked;
    };

    /**
     * @method View#isEditable
     * @returns {Boolean}
     */
    vp.isEditable = function () {
        return !(this.isReadOnly() || this.isDisabled() || this.isLocked());
    };

    /**
     * @method View#visible
     * @param {Boolean} visible
     */
    vp.visible = function (visible) {
        this._$el.toggleClass('hidden', !visible);
    };

    /**
     * @method View#readonly
     * @param {Boolean} readonly
     */
    vp.readonly = function (readonly) {
        this._attrs.readonly = readonly;

        this._redrawAttrs();
    };

    /**
     * Renders the view as locked or unlocked.
     *
     * @method View#lock
     * @param {Boolean} locked Whether the view should render as locked or unlocked.
     */
    vp.lock = function (locked) {
        this._attrs.locked = locked;

        this._redrawAttrs();
    };

    /**
     * @method View#disable
     * @param {Boolean} disabled
     */
    vp.disable = function (disabled) {
        if (!arguments.length) {
            return this.isDisabled();
        }
        this._attrs.disabled = disabled;

        this._redrawAttrs();
    };

    /**
     * @method View#clearAttributes
     */
    vp.clearAttributes = function () {
        this._applyLocked(false);
        this._applyReadonly(false);
        this._applyDisabled(false);
    };

    /**
     * @method View#_applyLocked
     * @private
     *
     * @param {Boolean} locked
     */
    vp._applyLocked = function (locked) {
        this._$el.toggleClass('locked', locked);
        this._$el.find('input, textarea')
            .prop('disabled', locked)
            .toggleClass('disabled locked', locked);
        this._$el.find('.btn, .input')
            .toggleClass('disabled', locked)
            .prop('disabled', locked);
    };

    /**
     * @method View#_applyDisabled
     * @private
     *
     * @param {Boolean} disabled
     */
    vp._applyDisabled = function (disabled) {
        this._$el.find('input, textarea')
            .prop('disabled', disabled)
            .toggleClass('disabled', disabled);
        this._$el.find('.btn, .input')
            .toggleClass('disabled', disabled)
            .prop('disabled', disabled);
    };

    /**
     * @method View#_applyReadonly
     * @private
     *
     * @param {Boolean} readonly
     */
    vp._applyReadonly = function (readonly) {
        this._$el.find('input, textarea')
            .prop('readonly', readonly)
            .toggleClass('readonly-as-text readonly', readonly);
        this._$el.find('.btn, .input')
            .toggleClass('readonly disabled readonly-as-text', readonly)
            .prop('disabled', readonly);
    };

    /**
     * @method View#_redrawAttrs
     * @private
     *
     */
    vp._redrawAttrs = function () {
        this.clearAttributes();

        if (this._attrs.locked) {
            this._applyLocked(this._attrs.locked);
        } else if (this._attrs.readonly) {
            this._applyReadonly(this._attrs.readonly);
        } else if (this._attrs.disabled) {
            this._applyDisabled(this._attrs.disabled);
        }
    };

    /**
     * @method View#focus
     *
     * @param {Boolean} focused
     */
    vp.focus = function (focused) {
        if (focused) {
            this._$el.find('input, textarea').first().focus();
        }
    };

    /**
     * Returns the unescaped version of the string provided
     *
     * @method View#_getUnescapedValue
     *
     * @param {String} value
     * @returns {String}
     */
    vp._getUnescapedValue = function (value) {
        if (_.isObject(this._options.autoEscape)) {
            value = this._options.autoEscape.unescape(value);
        } else if (this._options.autoEscape) {
            value = _.unescape(value);
        }

        return value;
    };

    /**
     * @method View#showErrors
     */
    vp.showErrors = function () {
        if (this._controller && this._controller.isValid && !this._controller.isValid()) {
            this._$el.add(this._$el.closest('.control-group')).toggleClass('error', !this._controller.isValid() && this._visited);
        }
    };

    return View;

});

/**
 * @module positioning
 */
define('oui/utils/positioning',[
    'underscore',
    'jquery'
], function (_, $) {
    'use strict';

    /**
     * @typedef coordinates
     *
     * @property {Number} [top=0]
     * @property {Number} [left=0]
     */

    /**
     * @method getOffsetParentsUntil
     * @private
     * @param {JQuery} $relativeElement
     * @param {JQuery} $until
     *
     * @returns {Array<JQuery>} parents
     */
    function getOffsetParentsUntil($relativeElement, $until) {
        var parents = [];
        var offsetParent = $relativeElement.offsetParent();

        while (!offsetParent.is($until)) {
            parents.push(offsetParent);
            offsetParent = offsetParent.offsetParent();
        }

        return parents;
    }

    /**
     * @alias module:positioning
     * @method relativePosition
     * @param {JQuery} $relativeElement
     * @param {Boolean} parentIsBody
     * @param {JQuery} $popup
     *
     * @returns {module:positioning~coordinates} position
     */
    function relativePosition($relativeElement, parentIsBody, $popup) {
        var position;

        if (parentIsBody) {
            position = $relativeElement.offset();
        } else if ($relativeElement.parent().length !== 0) {

            position = $relativeElement.position();

            if (!$relativeElement.offsetParent().is($popup) &&
                !$relativeElement.offsetParent().is($popup.offsetParent()) &&
                $relativeElement.offsetParent().find($popup.offsetParent()).length === 0) {

                var parents = getOffsetParentsUntil($relativeElement, $popup.offsetParent());
                position = _.reduce(parents, function (memo, parent) {
                    var additionalPosition = $(parent).position();

                    position.top = position.top + additionalPosition.top;
                    position.left = position.left + additionalPosition.left;

                    return position;
                }, position);
            }

        } else {
            position = $relativeElement.position();
        }

        return position;
    }

    return relativePosition;
});

define('oui/mixins/popup',[
    'jquery',
    'underscore',
    'oui/utils/positioning',
    'logger'
], function ($, _, relativePosition, logger) {
    'use strict';

    /**
     * This is meant to be a mixin to provide common popup positioning.
     *
     * **NOT** meant to be constructed like the OUI objects.
     *
     * @mixin popup
     * @private
     */
    return {
        /**
         * @method popup#_positionPopup
         * @private
         *
         * @param {jQuery} $relativeElement
         * @param {Boolean} parentIsBody
         * @param {String} alignment
         */
        _positionPopup: function ($relativeElement, parentIsBody, alignment) {
            var offset = $relativeElement.offset();
            var popupHeight = this._$popup.outerHeight(true);
            var popupWidth = this._$popup.width();
            var top;
            var left;

            //Determine Base Relative Positioning
            var position = relativePosition($relativeElement, parentIsBody, this._$popup);

            alignment = alignment || 'left';

            //Remove this if it was once positioned above the element
            this._$popup.removeClass('up');

            //Set up base css top/left coordinates
            top = position.top + $relativeElement.outerHeight();
            left = position.left;

            if (alignment === 'right') {
                left = position.left + ($relativeElement.innerWidth() - popupWidth);
            }

            //Correct for window dimensions
            if (offset.top + popupHeight + $relativeElement.outerHeight() > window.pageYOffset + $(window).height()) {
                top -= (popupHeight + $relativeElement.outerHeight());
                this._$popup.addClass('up');
            }

            if (popupWidth > $(window).width()) {
                //Nothing we can do with either alignment if the popup is larger than the window
                this._$popup.width($(window).width() - offset.left - 2); //Keep it 2px smaller than the window to prevent body scrolling
                this._$popup.removeClass('grow-to-text');
                popupWidth = this._$popup.width();
            }

            if (alignment === 'left' && offset.left + popupWidth > window.pageXOffset + $(window).width()) {
                left -= (popupWidth - $relativeElement.innerWidth());
            }

            //Set the coordinates
            this._$popup.css({
                'top': top,
                'left': left
            });
        },
        /**
         * @method popup#_popupZindex
         * @private
         *
         * @param {jQuery} $relativeElement
         * @param {Boolean} parentIsBody
         */
        _popupZindex: function ($relativeElement, parentIsBody) {
            //Find the max z-index in the parent chain (only the parent chain) which
            //means this could conflict with a sibling element.
            if (parentIsBody) {
                var that = this;
                var currentZindex = parseInt(this._$popup.css('z-index'), 10);
                var offset = this._$popup.offset();
                var height = this._$popup.height();
                var width = this._$popup.width();
                var elements = $relativeElement.parents()
                    .add(document.elementFromPoint(offset.left + width, offset.top + height))
                    .add(document.elementFromPoint(offset.left, offset.top));

                var max = _.chain(elements)
                    .map(function (parent) {
                        if ($(parent).is(that._$popup)) {
                            return 0;
                        }
                        return parseInt($(parent).css('z-index'), 10) || 0;
                    })
                    .max()
                    .value();

                if (!_.isNaN(currentZindex) && currentZindex <= max) {
                    this._$popup.css('z-index', max + 1);
                }
            }
        },
        /**
         * Common showing logic for a popup. It takes care of positioning, z-indexing, aria attributes.
         *
         * Defaults to below the component and is positioned above or to the side if the window dimensions are too
         * small.
         *
         * @method popup#showPopup
         * @fires popup#showing
         * @fires popup#shown
         * @param {jQuery} $relativeElement The element that the popup is to be positioned around.
         * @param {String} alignment
         */
        showPopup: function ($relativeElement, alignment) {
            var that = this;

            /**
             * @method popup#showPopup~hideListener
             * @private
             *
             * @param {jQuery.Event} event
             */
            var hideListener = function (event) {
                if (!(event.target === that._$popup[0] || that._$popup[0].contains(event.target)) &&
                    (event.target.contains && (event.target.contains(that._$popup[0]) || event.target.contains($relativeElement[0])))) {
                    that.hidePopup();
                    document.removeEventListener('scroll', hideListener, true);
                }
            };

            if (!_.isUndefined(this._$popup)) {
                var parentIsBody;

                /**
                 * Fired before the popup is shown.
                 *
                 * @event popup#showing
                 */
                this._$popup.trigger('showing.popup');

                //If the popup is not appended - append it
                if (this._$popup.parent().length === 0 && this._options && this._options.appendTo) {
                    $(this._options.appendTo).append(this._$popup);
                }

                parentIsBody = this._$popup.parent().is('body');

                this._$popup.attr('aria-hidden', false);

                this._$popup.show();
                this._$el.addClass('open');
                this._positionPopup($relativeElement, parentIsBody, alignment);
                this._popupZindex($relativeElement, parentIsBody);

                document.addEventListener('scroll', hideListener, true);

                this._$popup.one('hidden.popup', function () {
                    document.removeEventListener('scroll', hideListener, true);
                });

                /**
                 * Fired after the popup is shown.
                 *
                 * @event popup#shown
                 */
                this._$popup.trigger('shown.popup');
            } else {
                logger.warn('_$popup is not defined.');
            }
        },
        /**
         * Hides the popup and fires events for the view to listen to.
         *
         * @method popup#hidePopup
         * @fires popup#hiding
         * @fires popup#hidden
         */
        hidePopup: function () {
            if (!_.isUndefined(this._$popup)) {
                /**
                 * Fired before the popup is hidden
                 *
                 * @event popup#hiding
                 */
                this._$popup.trigger('hiding.popup');
                this._$popup.hide();
                this._$el.removeClass('open');
                this._$popup.attr('aria-hidden', true);
                /**
                 * Fired after the popup is hidden.
                 *
                 * @event popup#hidden
                 */
                this._$popup.trigger('hidden.popup');
            }
        },
        /**
         * @method popup#isVisible
         */
        isVisible: function () {
            return this._$el.hasClass('open');
        }
    };
});

/**
 * @module stringUtils
 * @private
 */
define('oui/utils/stringUtils',[
    'underscore'
], function (_) {
    'use strict';

    /**
     * @typedef {Object} module:stringUtils#TokenTriggerOptions
     *
     * @property {String} trigger
     * @property {Object} formatted
     * @property {String} formatted.start
     * @property {String} formatted.end
     */

    /**
     * @method module:stringUtils~betweenFormattedStrings
     * @private
     *
     * @param {Object} queryOptions
     * @param {Object} selections
     *
     * @returns {Boolean}
     */
    function betweenFormattedStrings(queryOptions, selections) {
        var closestStart = queryOptions.closestStart;
        var closestEnd = queryOptions.closestEnd;

        var previousEnd = queryOptions.previousEnd;
        var nextStart = queryOptions.nextStart;

        return closestStart !== -1 && closestEnd !== -1 &&
            (closestStart > previousEnd) &&
                (closestEnd < nextStart || nextStart === -1) &&
                    selections.start > closestStart && selections.start < closestEnd;
    }

    /**
     * @method module:stringUtils~afterOpenFormattedStart
     * @private
     *
     * @param {Object} queryOptions
     * @param {Object} selections
     *
     * @returns {Boolean}
     */
    function afterOpenFormattedStart(queryOptions, selections) {
        var closestStart = queryOptions.closestStart;

        var previousEnd = queryOptions.previousEnd;
        var nextStart = queryOptions.nextStart;

        return closestStart !== -1 &&
            selections.start > closestStart &&
                nextStart === -1 &&
                    (previousEnd === -1 || previousEnd < closestStart);
    }

    /**
     * @method module:stringUtils~afterTrigger
     * @private
     *
     * @param {Object} queryOptions
     *
     * @returns {Boolean}
     */
    function afterTrigger(queryOptions) {
        var closestStart = queryOptions.startIndex;

        var previousEnd = queryOptions.previousEnd;

        return closestStart !== -1 &&
            (previousEnd === -1 || closestStart > previousEnd);
    }

    return {
        /**
         * @method module:stringUtils~getFullTokenQuery
         * @private
         *
         * @param {module:stringUtils#TokenTriggerOptions} options
         * @param {String} str
         * @param {Object} selections
         *
         * @returns {String}
         */
        getFullTokenQuery: function (options, str, selections) {
            var startIndex = str.lastIndexOf(options.trigger, selections.start);
            var closestStart = str.lastIndexOf(options.formatted.start, selections.start);
            var closestEnd = str.indexOf(options.formatted.end, selections.end);

            var previousEnd = str.lastIndexOf(options.formatted.end, selections.start);
            var nextStart = str.indexOf(options.formatted.start, selections.start);
            var queryOptions = {
                startIndex: startIndex,
                closestStart: closestStart,
                closestEnd: closestEnd,
                previousEnd: previousEnd,
                nextStart: nextStart
            };

            if (str.length === 0) {
                return '';
            }

            if (betweenFormattedStrings(queryOptions, selections)) {
                return str.substring(closestStart, closestEnd + 1);
            } else if (afterOpenFormattedStart(queryOptions, selections)) {
                return str.substring(closestStart);
            } else if (afterTrigger(queryOptions, selections)) {
                if (closestEnd === -1) {
                    return str.substring(startIndex, selections.end + 1);
                }
                return str.substring(startIndex, closestEnd + 1);
            }

            return '';
        },
        /**
         * @function module:stringUtils.getTokenTrigger
         *
         * @param {StringUtils#TokenTriggerOptions} options
         * @param {String} str
         * @param {Object} selections
         *
         * @returns {String}
         */
        getTokenTrigger: function (options, str, selections) {
            var fullQuery = this.getFullTokenQuery(options, str, selections);

            if (_.startsWith(fullQuery, options.trigger)) {
                fullQuery = fullQuery.substring(1);
            }

            if (_.startsWith(fullQuery, options.formatted.start)) {
                fullQuery = fullQuery.substring(2);
            }

            if (_.endsWith(fullQuery, options.formatted.end)) {
                fullQuery = fullQuery.substring(0, fullQuery.length - 1);
            }

            return fullQuery;
        },
        /**
         * @function module:stringUtils.getQuery
         *
         * @param {String} value
         * @param {Number} cursorLocation
         * @param {Boolean} multiSelect
         * @param {String} delimiter
         *
         * @returns {String}
         */
        getQuery: function (value, cursorLocation, multiSelect, delimiter) {
            if (!multiSelect || (multiSelect && value.indexOf(delimiter) === -1)) {
                return value;
            } else {
                var upperBound = value.indexOf(delimiter, cursorLocation);
                var lowerBound = value.lastIndexOf(delimiter, cursorLocation - 1) + 1;
                if (upperBound === -1) {
                    upperBound = value.length;
                }

                return value.substring(lowerBound, upperBound).trim();
            }
        },
        /**
         * @function module:stringUtils.trim
         *
         * @param {String} str
         * @returns {String}
         */
        trim: function (str) {
            return str.trim();
        },
        /**
         * @function module:stringUtils.union
         *
         * @param {String} str
         * @param {String} addedString
         *
         * @returns {String}
         */
        union: function (str, addedString) {
            str = str || '';
            var split = str.split(' ');
            split.push(addedString);
            return _.uniq(split).join(' ').trim();
        },
        /**
         * @function module:stringUtils.reject
         *
         * @param {String} str
         * @param {String} removedString
         *
         * @returns {String}
         */
        reject: function (str, removedString) {
            str = str || '';
            var split = str.split(' ');
            split.splice(split.indexOf(removedString), 1);
            return split.join(' ').trim();
        },
        /**
         * @function module:stringUtils.isEmpty
         *
         * @param {String} str
         * @returns {Boolean}
         */
        isEmpty: function (str) {
            return str === '\n' || this.trim(str) === '';
        },
        /**
         * @function module:stringUtils.pad
         * @param {String} str
         * @param {Number} length
         * @param {String} ch
         *
         * @returns {String}
         */
        pad: function (str, length, ch) {
            while (str.length < length) {
                str = ch + str;
            }

            return str;
        },
        /**
         * @function module:stringUtils.regexEscape
         *
         * @param {String} str
         *
         * @returns {String?}
         */
        regexEscape: function (str) {
            if (str != null) {
                return str.replace(/[\-\[\]\/\{\}\(\)\*\+\?\.\\\^\$\|]/g, '\\$&');
            } else {
                return null;
            }
        }
    };
});

/**
 * @module math
 * @private
 */
define('oui/utils/math',[
    'underscore'
], function (_) {
    'use strict';

    /**
     * @exports math
     */
    return {
        /**
         * @function step
         * @param {String} min The function to use as the rounding function.
         * @param {Number} scale The number to scale the value to.
         * @param {Number} num The number to apply the step function to.
         * @returns {Number} The rounded number
         */
        step: function (min, scale, num) {
            if (!_.isNumber(num)) {
                return NaN;
            }
            return Math[min](num / scale) * scale;
        },
        /**
         * @function valueFromPercent
         * @param {Number} percent The percent as a value over 100. For example, 50% should be passed in as .5
         * @param {Object} [options] The options to configure the scale (max and min)
         * @param {Number} [options.max=100] The max value of the scale
         * @param {Number} [options.min=0] The min value of the scale
         * @returns {Number} value - The converted value using the scale
         */
        valueFromPercent: function (percent, options) {
            var max = _.get(options, 'max', 100);
            var min = _.get(options, 'min', 0);

            return percent * (max - min) + min;
        },
        /**
         * @function percentFromValue
         * @param {Number} value The value to convert to a percent (as a mathematical value). For example 100% = 1.
         * @param {Object} [options] The options to configure the scale (max and min)
         * @param {Number} [options.max=100] The max value of the scale
         * @param {Number} [options.min=0] The min value of the scale
         * @returns {Number} value - The converted value using the scale
         */
        percentFromValue: function (value, options) {
            var max = _.get(options, 'max', 100);
            var min = _.get(options, 'min', 0);

            return (value - min) / (max - min);
        },
        /**
         * @function distance
         * @param {Object} p1 The first point to use for distance calculation
         * @param {Number} p1.x The x coordinate of the first point
         * @param {Number} p1.y The y coordinate of the first point
         * @param {Object} p2 The second point to use for distance calculation
         * @param {Number} p2.x The x coordinate of the second point
         * @param {Number} p2.y The y coordinate of the second point
         * @returns {Number} distance The distance between the two points.
         */
        distance: function (p1, p2) {
            var sqr1 = Math.pow(p2.x - p1.x, 2);
            var sqr2 = Math.pow(p2.y - p1.y, 2);
            return Math.sqrt(sqr1 + sqr2);
        },
        /**
         * @function mod
         *
         * @param {Number} number
         * @param {Number} modulo
         *
         * @returns {Number}
         */
        mod: function (number, modulo) {
            return (number % modulo + modulo) % modulo;
        },
        /**
         * @function getMantissa
         *
         * @param {Number} number
         *
         * @returns {Number}
         */
        getMantissa: function (number) {
            return Number(this.mod(Math.abs(number), 1).toFixed(14));
        },
        /**
         * Clamps a number between a lower and upper bound.
         *
         * @function clamp
         *
         * @param {Number} lowerBound
         * @param {Number} upperBound
         * @param {Number} number
         *
         * @returns {Number}
         */
        clamp: function (lowerBound, upperBound, number) {
            var clamped = number;
            if (lowerBound != null) {
                clamped = Math.max(lowerBound, number);
            }

            if (upperBound != null) {
                clamped = Math.min(clamped, upperBound);
            }

            return clamped;
        },
        /**
         * Returns the sign of a number, indicating whether the number is positive, negative or 0
         *
         * @function sign
         *
         * @param {Number|*} num The number to determine the sign of
         *
         * @returns {Number|NaN}
         */
        sign: function (num) {
            if (Math.sign) {
                return Math.sign(num);
            }

            num = +num;

            if (num === 0 || _.isNaN(num)) {
                return num;
            }

            if (num > 0) {
                return 1;
            } else {
                return -1;
            }
        }
    };
});

define('oui/utils/inputUtils',[],function () {
    'use strict';

    var api;

    /**
     * @method inputUtils~getInputSelection
     * @private
     * @param {HTMLInputElement} input
     * @returns {Object}
     */
    function getInputSelection(input) {
        if (input.type === 'textarea' || input.getAttribute('type') === 'text') {
            return {
                start: input.selectionStart,
                end: input.selectionEnd
            };
        }

        return {
            start: 0,
            end: 0
        };
    }

    /**
     * @method inputUtils~insertCharacter
     * @private
     * @param {HTMLInputElement} input
     * @param {Number} charCode
     * @param {Boolean} caps
     * @returns {String}
     */
    function insertCharacter(input, charCode, caps) {
        var str = input.value;
        var selections = api.getInputSelection(input);
        var character = String.fromCharCode(charCode);

        str = str.split('');

        if (!caps) {
            character = character.toLowerCase();
        }

        str.splice(selections.start, selections.end - selections.start, character);

        str = str.join('');

        return str;
    }

    /**
     * @method inputUtils~setInputSelection
     * @private
     * @param {HTMLInputElement} input
     * @param {Object} selections
     */
    function setInputSelection(input, selections) {
        if (document.activeElement !== input) {
            return;
        }

        if (selections.start == null) {
            selections.start = 0;
        }

        if (input.getAttribute('type') === 'text') {
            if (selections.end == null) {
                selections.end = selections.start;
            }

            if (document.body.contains(input)) {
                input.setSelectionRange(selections.start, selections.end);
            }
        }
    }

    api = {
        getInputSelection: getInputSelection,
        setInputSelection: setInputSelection,
        insertCharacter: insertCharacter
    };

    return api;
});


define('template-1', ['underscore'], function (_) { return function(data) {
var __t, __p = '', __j = Array.prototype.join;
function print() { __p += __j.call(arguments, '') }
__p += '<div class="orcl-widget orcl-text-field"> <output class="character-counter"></output>';
 if(data.multiline) { ;
__p += ' <textarea></textarea>';
 } else { ;
__p += ' <input type="text">';
 } ;
__p += '</div>';
return __p
};});


define('template-2',[],function () { return '<div class="orcl-widget orcl-text-field">\n</div>\n';});


define('template-3', ['underscore'], function (_) { return function(data) {
var __t, __p = '';
__p += '<ul role="listbox" aria-label="' +
((__t = ( data.locale.menu )) == null ? '' : __t) +
'" class="dropdown-menu combo-dropdown-menu"></ul>';
return __p
};});


define('template-4', ['underscore'], function (_) { return function(data) {
var __t, __p = '', __e = _.escape;
__p += '<li role="option" title="' +
__e( data.tooltip ) +
'" data-text="' +
__e( data.text ) +
'" id="' +
((__t = ( data.pageId )) == null ? '' : __t) +
'"    class="combo-item ' +
((__t = ( data.active ? 'active' : '' )) == null ? '' : __t) +
'"> <a href="#"> ' +
((__t = ( data.display )) == null ? '' : __t) +
' </a></li>';
return __p
};});

/**
 * @module escapingUtils
 * @private
 */
define('oui/utils/escapingUtils',[
    'underscore'
], function (_) {
    'use strict';

    return {
        /**
         * @method module:escapingUtils.escape
         *
         * @param {String} str
         * @param {Object} options
         *
         * @returns {String}
         */
        escape: function (str, options) {
            var value = str;

            if (_.isObject(options.autoEscape)) {
                value = options.autoEscape.escape(value);
            } else if (options.autoEscape) {
                value = _.escape(value);
            }

            return value;
        },
        /**
         * @method module:escapingUtils.unescape
         *
         * @param {String} str
         * @param {Object} options
         *
         * @returns {String}
         */
        unescape: function (str, options) {
            var value = str;

            if (_.isObject(options.autoEscape)) {
                value = options.autoEscape.unescape(value);
            } else if (options.autoEscape) {
                value = _.unescape(value);
            }

            return value;
        }
    };
});

define('oui/widgets/textField/textView',['require','jquery','underscore','oui/base/view','oui/utils/keyCode','oui/mixins/popup','oui/utils/stringUtils','oui/utils/math','oui/utils/inputUtils','template-1','template-2','template-3','template-4','oui/utils/escapingUtils'],function (require) { //jshint ignore:line
      'use strict';

      var $ = require('jquery');
      var _ = require('underscore');
      var View = require('oui/base/view');
      var keyCode = require('oui/utils/keyCode');
      var popupMixin = require('oui/mixins/popup');
      var stringUtils = require('oui/utils/stringUtils');
      var math = require('oui/utils/math');
      var inputUtils = require('oui/utils/inputUtils');
      var textFieldTemplate = require('template-1');
      var wrapTemplate = require('template-2');
      var menuTemplate = require('template-3');
      var itemTemplate = require('template-4');
      var escapingUtils = require('oui/utils/escapingUtils');
      var tvp;

      /**
       * Fired when the text has changed within the text field.
       *
       * @event TextView#text-changed
       * @param {String} text Changed text
       */
      /**
       * Fired when the attempted text fails a rule.
       *
       * @event TextView#error
       * @param {String} text Attempted text
       * @param {Object} error
       * @param {Boolean} error.valid
       * @param {String} error.type
       * @param {String} error.data
       */
      /**
       * Fired when the attempted text is null or empty string and the field is required.
       *
       * @event TextView#required-error
       * @param {String} text Attempted text
       */
      /**
       * Fired when the attempted text is out of the bounds of the min and max attributes.
       *
       * @event TextView#length-error
       * @param {String} text Attempted text
       */

      /**
       * @class TextView
       * @private
       * @mixes popup
       *
       * @param {TextModel} model
       * @param {TextController} controller
       * @param {jQuery} element
       * @param {TextField#Config} options
       * @param {Object} settings
       */
      function TextView(model, controller, element, options, settings) {
          var that = this;

          View.call(this, options, element, controller, settings);

          this._$input = this._$el.find('input, textarea').first();
          this._inputEl = this._$input[0];
          this._focused = false;
          this._$popup = $(menuTemplate({
              locale: this._options.locale
          }));

          model.on({
              'change': this.update.bind(this),
              'error': this.error.bind(this)
          });

          this._$el.on('change', 'input, textarea', function () {
              that._controller.text(escapingUtils.escape(this.value, that._options));
          });

          this._$el.on('focusin', this.focused.bind(this));
          this._$el.on('focusout', this.blurred.bind(this));

          this._$input.on('keypress', function (event) {
              var selections = inputUtils.getInputSelection(event.target);
              var str = inputUtils.insertCharacter(event.target, event.which, event.shiftKey);

              if (event.keyCode !== keyCode.ENTER &&
                  event.keyCode !== keyCode.TAB) {

                  if (_.isRegExp(that._options.restrictTo) && !that._options.restrictTo.test(str)) {
                      event.preventDefault();
                      that._options.restrictTo.lastIndex = 0;
                  } else if (_.isFunction(that._options.restrictTo) && !that._options.restrictTo(str)) {
                      event.preventDefault();
                  }

                  that.autoComplete(str, selections);
              }
          });

          this._$input.on('input', function (event, data) {
              var input = that._$input[0];
              var str = input.value;
              var selections = inputUtils.getInputSelection(input);

              if (!data || !data.keepLinked) {
                  that._$el.addClass('separated');
                  if (that._linked != null && !that._linked.element().hasClass('separated')) {
                      //Do not want to commit the value right away but wait until change has fired on the original element?
                      that._linked._view._inputEl.value = this.value;
                      that._linked._view._$input.trigger('input', { keepLinked: true });
                      that._linked.value(this.value);
                  }
              }

              that.updateCounter(this.value);
              that.autoComplete(str, selections);
          });

          this._$input.on('paste', function (event) {
              var attemptedString = event.originalEvent.clipboardData.getData('text/plain');
              if (attemptedString.length > that._options.maxLength) {
                  that._$el.one('animationend', function () {
                      that._$el.find('.character-counter').removeClass('shake');
                  });

                  that._$el.find('.character-counter').addClass('shake');
              }
          });

          this._$popup.on('mousedown', '.combo-item', function (event) {
              event.preventDefault();
          });

          this._$popup.on('click', '.combo-item', function (event) {
              event.preventDefault();

              that.autoCompleteConfirm($(this).attr('data-text'));
          });

          this._$input.on('keydown', function (event) {
              if (that.isVisible()) {
                  if (event.which === keyCode.ENTER) {
                      event.preventDefault();

                      that.autoCompleteConfirm(that._$popup.find('.active').attr('data-text'));
                  } else if (event.which === keyCode.DOWN) {
                      event.preventDefault();

                      that.moveActive(1);
                  } else if (event.which === keyCode.UP) {
                      event.preventDefault();

                      that.moveActive(-1);
                  } else if (event.which === keyCode.ESC) {
                      event.preventDefault();

                      that.hidePopup();
                  }
              }
          });
      }

      TextView.prototype = Object.create(View.prototype);
      tvp = TextView.prototype;

      tvp.constructor = TextView;

      _.extend(tvp, popupMixin);

      /**
       * @method TextView#buildElement
       *
       * @param {TextField#Config} opts
       * @param {jQuery} element
       *
       * @returns {jQuery}
       */
      tvp.buildElement = function (opts, element) {
          var $element;

          if (_.isNull(element)) {
              $element = $(textFieldTemplate({
                  maxLength: opts.maxLength,
                  multiline: opts.multiline
              }));
          } else if (element.jquery) {
              $element = element
                .wrap(wrapTemplate)
                .parent();

              $element.append('<output class="character-counter"></div>');
          }

          if (opts.maxLength) {
              $element.find('input, textarea').attr('maxlength', opts.maxLength);
              $element.find('.character-counter').text(opts.maxLength - $element.find('input, textarea').first().val().length);
          }

          $element.toggleClass('show-counter', opts.counter != null);

          return $element;
      };

      /**
       * @method TextView#focused
       */
      tvp.focused = function () {
          this._focused = true;

          if (!this.isVisible()) {
              this.autoComplete(this._$input.val(), inputUtils.getInputSelection(this._$input[0]));
          }
      };

      /**
       * @method TextView#updateCounter
       *
       * @param {String} value
       */
      tvp.updateCounter = function (value) {
          if (this._options.maxLength) {
              this._$el.find('.character-counter').text(this._options.maxLength - value.length);
          }
      };

      /**
       * @method TextView#blurred
       */
      tvp.blurred = function () {
          this.updateState();

          if (this.isVisible()) {
              this.hidePopup();
          }
      };

      /**
       * @method TextView#updateState
       */
      tvp.updateState = function () {
          if (this._focused) {
              this._$el.add(this._$el.closest('.control-group')).toggleClass('error', !this._controller.isValid() && this._visited);
          }
      };

      /**
       * @method TextView#link
       *
       * @param {TextField} textField
       */
      tvp.link = function (textField) {
          this._linked = textField;
      };

      /**
       * @method TextView#update
       *
       * @param {String} value
       */
      tvp.update = function (value) {
          value = escapingUtils.unescape(value, this._options);
          this._$el.add(this._$el.closest('.control-group')).toggleClass('error', !this._controller.isValid() && this._visited);

          this._inputEl.value = value;
          if (this._linked != null && !this._linked.element().hasClass('separated')) {
              this._linked.value(value);
          }

          this.updateCounter(value);
      };

      /**
       * @method TextView#error
       *
       * @param {String} attempted
       */
      tvp.error = function (attempted) {
          attempted = escapingUtils.unescape(attempted, this._options);
          this._$el.add(this._$el.closest('.control-group')).toggleClass('error', !this._controller.isValid() && this._visited);

          this._inputEl.value = attempted;
      };

      /**
       * @method TextView#populateMenu
       *
       * @param {Array} results
       */
      tvp.populateMenu = function (results) {
          if (results == null || results.length === 0) {
              this.hidePopup();
              this._$el.removeAttr('aria-activedescendant');
          } else {
              this._$popup.html(_.reduce(results, this.makeItem.bind(this), ''));
              this._$popup.width(this._$el.width());
              this.showPopup(this._$el);

              this._$el.attr('aria-activedescendant', this._$popup.find('.active').attr('id'));
          }
      };

      /**
       * @method TextView#makeItem
       *
       * @param {String} memo
       * @param {Object} item
       * @param {Number} idx
       *
       * @returns {String}
       */
      tvp.makeItem = function (memo, item, idx) {
          var tooltip;
          var text = this._options.autoComplete.text(item);

          if (this._options.autoComplete.tooltip) {
              tooltip = this._options.autoComplete.tooltip(item);
          } else {
              tooltip = text;
          }
          return memo += itemTemplate({
              display: this._options.autoComplete.display(item),
              pageId: this.id() + '-' + idx,
              text: text,
              tooltip: tooltip
          });
      };

      /**
       * @method TextView#autoComplete
       *
       * @param {String} str
       * @param {Object} selections
       */
      tvp.autoComplete = function (str, selections) {
          if (this._options.autoComplete.enabled) {
              var query = stringUtils.getTokenTrigger(this._options.autoComplete, str, selections).trim();
              if (query.length > 0) {
                  $.when(this._options.autoComplete.search(query))
                      .done(this.populateMenu.bind(this))
                      .fail(this.hidePopup.bind(this));
              } else {
                  this.hidePopup();
              }
          }
      };

      /**
       * @method TextView#autoCompleteConfirm
       *
       * @param {String} str
       */
      tvp.autoCompleteConfirm = function (str) {
          var selections = inputUtils.getInputSelection(this._$input[0]);
          var query = stringUtils.getFullTokenQuery(this._options.autoComplete, this._$input.val(), selections);

          var fullContent = this._$input.val();

          this._$input.val(fullContent.replace(new RegExp(_.escapeRegExp(query), 'g'), function (match, offset) {
              if (selections.start >= offset &&
                 selections.start <= offset + query.length) {
                  return this._options.autoComplete.formatted.start + str + this._options.autoComplete.formatted.end;
              } else {
                  return match;
              }
          }.bind(this)));

          this.hidePopup();

          this._$el.removeAttr('aria-activedescendant');
      };

      /**
       * @method TextView#moveActive
       *
       * @param {Number} number
       */
      tvp.moveActive = function (number) {
          var currentActive = this._$popup.find('.active');
          var allOptions = this._$popup.find('.combo-item');
          var indexOf = allOptions.index(currentActive);
          var $nextOption;

          currentActive.removeClass('active');

          if (indexOf === -1) {
              if (number === -1) {
                  $nextOption = allOptions.last();
              } else {
                  $nextOption = allOptions.first();
              }
          } else {
              $nextOption = $(allOptions.get(math.mod(indexOf + number, allOptions.length)));
          }

          if ($nextOption.length > 0) {
              this._$input.attr('aria-activedescendant', $nextOption.attr('id'));
              $nextOption.addClass('active');
          }
      };

      /**
       * @method TextView#insert
       *
       * @param {String} text
       */
      tvp.insert = function (text) {
          var selections = inputUtils.getInputSelection(this._$input[0]);

          var str = this._$input[0].value.split('');

          str.splice(selections.start,
                     selections.end - selections.start,
                     text);

          str = str.join('');

          this._controller.text(str);
      };

      return TextView;

  });

define('oui/widgets/textField/textController',[], function () {
    'use strict';

    /**
     * @class TextController
     * @private
     *
     * @param {TextModel} model
     * @param {TextField#Config} options
     */
    function TextController(model, options) {
        this._model = model;
        this._options = options;
    }

    TextController.prototype = {
        constructor: TextController,
        /**
         * @method TextController#text
         *
         * @returns {String}
         */
        /**
         * @method TextController#text
         *
         * @param {String} text
         * @param {Subscribable~EventOptions} eventParams
         */
        text: function (text, eventParams) {
            if (!arguments.length) {
                return this._model.text();
            }

            this._model.text(text, eventParams);
        },
        /**
         * @method TextController#isValid
         *
         * @returns {Boolean}
         */
        isValid: function () {
            return this._model.isValid();
        },
        /**
         * @method TextController#error
         *
         * @returns {Object}
         */
        error: function () {
            return this._model.error();
        }
    };

    return TextController;
});

define('oui/base/model',[
    'jquery',
    'underscore',
    'oui/utils/subscribable'
], function ($, _, Subscribable) {
    'use strict';

    /**
     * @typedef Model#EventParams
     *
     * @property {Boolean} [silent=false]
     */

    /**
     * @typedef Model#ValidationResult
     *
     * @property {*} data
     * @property {Boolean} valid
     * @property {String} type
     * @property {String} message
     */

    /**
     * Fired when the model is going to change has already passed validation.
     *
     * @event Model#pre-change
     *
     * @param {*} newData
     * @param {*} currentData
     */

    /**
     * Fired when the model has changed.
     *
     * @event Model#change
     *
     * @param {*} newData
     * @param {*} currentData
     */

    /**
     * Fired when the model is not able to update the data due to some validation
     * condition.
     *
     * Triggers with what is attempted and the validation result (includes data, type, validity, and an optional message).
     *
     * @event Model#error
     *
     * @param {*} attemptedData
     * @param {Model#ValidationResult} validation
     */

    /**
     * Fired when not updating but attempting to make a change (updating to the same value).
     *
     * @event Model#identical
     */

    /**
     * @class Model
     * @private
     * @extends Subscribable
     * @param {*} initData The initialization data used for the widget (dependent on subtype)
     * @param {*} options One of the various configuration options
     */
    function Model(initData, options) {
        Subscribable.call(this);

        this._options = options;

        this._data = null;
        this._validationResult = {
            valid: true
        };

        this.setData(this.init(initData, options));
    }

    var mp = Model.prototype = Object.create(Subscribable.prototype);

    /**
     * @method Model#name
     * @returns {String} 'default'
     */
    mp.name = _.constant('default');

    /**
     * @method Model#init
     * @param {*}
     * @returns {*}
     */
    mp.init = _.identity;

    /**
     * @method Model#setData
     * @param {*} newData
     * @param {Model#EventParams} eventParams
     */
    mp.setData = function (newData, eventParams) {
        var that = this;
        var validationResult = this.validate(newData);

        eventParams = eventParams || {};

        $.when(validationResult)
            .then(function (result) {

                if (result && result.valid != null) {
                    var dfd = $.Deferred();

                    if (result.valid) {
                        dfd.resolve();
                    } else {
                        dfd.reject(result);
                    }

                    return dfd.promise();
                }

                return arguments;
            })
            .then(function () {
                if (that._options.validators) {
                    //Convert array of functions to array of deferreds for jQuery execution
                    return $.when.apply($, _.invoke(that._options.validators, _.call, null, newData))
                        .then(function () {
                            var results = _(arguments).toArray().compact().value();

                            if (!_.some(results, _.property('then'))) {
                                var dfd = $.Deferred();

                                if (_.all(results, _.matchesProperty('valid', true))) {
                                    dfd.resolve();
                                } else {
                                    var error = _.find(results, _.matchesProperty('valid', false));

                                    if (!_.has(error, 'data')) {
                                        error.data = newData;
                                    }

                                    dfd.reject(error);
                                }

                                return dfd.promise();
                            }

                            return arguments;
                        }).promise();
                }

                return arguments;
            })
            .done(function () {
                $.when(this.presuccessHandler(eventParams, newData))
                    .done(this.successHandler.bind(this, eventParams))
                    .fail(this.rerender.bind(this));
            }.bind(this))
            .fail(this.failureHandler.bind(this, eventParams));
    };

    /**
     * @method Model#rerender
     */
    mp.rerender = function () {
        this.trigger('view.change', this._data);
    };

    /**
     * @method Model#getData
     * @returns {*}
     */
    mp.getData = function () {
        return this._data;
    };

    /**
     * @method Model#data
     * @returns {*} data
     */
    /**
     * @method Model#data
     * @param {*} newData
     * @param {Subscribable~EventOptions} eventParams
     */
    mp.data = function (newData, eventParams) {
        if (!arguments.length) {
            return this.getData();
        }

        this.setData(newData, eventParams);
    };

    /**
     * @method Model#isValid
     * @returns {Boolean} validity
     */
    mp.isValid = function () {
        return this._validationResult.valid;
    };

    /**
     * @method Model#validate
     * @param {*} data - The data to validate
     * @returns {Boolean} validity
     */
    mp.validate = function (data) {
        return $.Deferred().resolve(data).promise();
    };

    /**
     * @method Model#isSame
     * @param {*} newData - The data to check against the current data
     * @returns {Boolean} equality
     */
    mp.isSame = function (newData) {
        return this._data === newData;
    };

    /**
     * @method Model#isSet
     * @returns {Boolean} - whether or not the model is set
     */
    mp.isSet = function () {
        return this.isValid() && this._data != null;
    };

    /**
     * @method Model#presuccessHandler
     * @param {Model#EventParams} eventParams
     * @param {*} newData
     * @returns {Promise}
     */
    mp.presuccessHandler = function (eventParams, newData) {
        if (!this.isSame(newData)) {
            var value = this.trigger(eventParams, 'pre-change', newData, this._data);

            if (value && value.then) {
                return value;
            } else if (value === false) {
                return $.Deferred().reject(newData).promise();
            } else {
                return $.Deferred().resolve(newData).promise();
            }
        } else {
            return $.Deferred().resolve(newData).promise();
        }
    };

    /**
     * @method Model#successHandler
     * @param {Model#EventParams} eventParams
     * @param {*} newData
     *
     * @fires Model#identical
     * @fires Model#change
     */
    mp.successHandler = function (eventParams, newData) {
        var currentlyInvalid = !this.isValid();
        this._validationResult = {
            valid: true
        };

        if (!this.isSame(newData) || currentlyInvalid) {

            var prevData = this._data;
            this._data = newData;
            eventParams.name = this.name() + '-change';
            this.trigger(eventParams, this._data, prevData);
            eventParams.name = 'change';
            this.trigger(eventParams, this._data, prevData);
        } else {
            this.trigger('view.change', this._data);
            this.trigger('identical', this._data);
        }

    };

    /**
     * @method Model#failureHandler
     * @param {Model#EventParams} eventParams
     * @param {Model#ValidationResult} validationResult
     *
     * @fires Model#error
     */
    mp.failureHandler = function (eventParams, validationResult) {
        if (!_.isEqual(this._validationResult, validationResult)) {
            if (validationResult.valid == null) {
                this._validationResult = {
                    valid: false,
                    type: 'generic',
                    data: validationResult
                };
            } else {
                this._validationResult = validationResult;
            }

            this.trigger(_.extend({}, eventParams, { name: this._validationResult.type + '-error' }), this._validationResult.data, this._data);
            this.trigger(_.extend({}, eventParams, { name: 'error' }), this._validationResult.data, this._data, this._validationResult);
        }

    };

    /**
     * @method Model#error
     * @returns {Model#ValidationResult?} error
     */
    mp.error = function () {
        if (!this.isValid()) {
            return _.clone(this._validationResult);
        } else {
            return null;
        }
    };

    return Model;
});

/**
* @module typeValidator
*/
define('oui/utils/validators/typeValidator',[
    'underscore',
    'jquery',
    'moment'
], function (_, $, moment) {
    'use strict';

    /**
     * @typedef Type
     *
     * @property {String} Number
     * @property {String} moment
     * @property {String} String
     * @property {String} Null
     * @property {String} Undefined
     * @property {String} NaN
     * @property {String} Boolean
     * @property {String} Array
     * @property {String} Object
     */

    /**
     * @typedef typeValidatorOptions
     *
     * @property {Array<module:typeValidator~Type>} types
     */

    /**
     * @alias module:typeValidator
     * @method typeValidator
     * @param {module:typeValidator~typeValidatorOptions} options
     *
     * @returns {module:validator~ValidateFn}
     */
    return function (options) {

        return function (data) {
            var dfd = $.Deferred();
            var typeMatch = _.some(options.types, function (fn) {

                if (_['is' + fn]) {
                    return _['is' + fn](data);
                } else if (fn === 'moment') {
                    return moment.isMoment(data);
                }
            });

            if (typeMatch) {
                dfd.resolve(data);
            } else {
                dfd.reject({
                    type: 'type',
                    valid: false,
                    data: data
                });
            }

            return dfd.promise();
        };
    };
});

/**
* @module requiredValidator
*/
define('oui/utils/validators/requiredValidator',[
    'underscore',
    'jquery'
], function (_, $) {
    'use strict';

    var DEFAULTS = {
        required: true
    };

    /**
     * @typedef requiredValidatorOptions
     *
     * @property {Boolean} [required=true]
     */

    /**
     * @alias module:requiredValidator
     * @method requiredValidator
     * @param {module:requiredValidator~requiredValidatorOptions} options
     *
     * @returns {module:validator~ValidateFn}
     */
    return function (options) {
        options = _.defaults({}, options, DEFAULTS);

        return function (data) {
            var dfd = $.Deferred();

            if (options.required &&
                    (data == null || data === '' || (_.isArray(data) && data.length === 0))) {
                dfd.reject({
                    valid: false,
                    data: data,
                    type: 'required'
                });
            } else {
                dfd.resolve(data);
            }

            return dfd.promise();
        };
    };
});

/**
 * @module boundsValidator
 */
define('oui/utils/validators/boundsValidator',[
    'underscore',
    'jquery',
    'moment'
], function (_, $, moment) {
    'use strict';

    /**
     * @typedef boundsValidatorOptions
     *
     * @property {Number|moment|null} min
     * @property {Number|moment|null} max
     */

    /**
     * @alias module:boundsValidator
     * @method boundsValidator
     * @param {module:boundsValidator~boundsValidatorOptions} options
     *
     * @returns {module:validator~ValidateFn}
     */
    return function (options) {

        /**
         * @method module:boundsValidator~inBounds
         *
         * @param {Object} data
         * @returns {Boolean}
         */
        var inBounds = function (data) {
            //Infinite range
            if (data == null || (options.min == null && options.max == null)) {
                return true;
            }

            if (data.length != null) {
                data = data.length;
            }

            return (options.min == null || data >= options.min) &&
                (options.max == null || data <= options.max);
        };

        return function (data) {
            var dfd = $.Deferred();

            if (moment.isMoment(data)) {
                if (options.min != null) {
                    options.min = moment(options.min);
                }

                if (options.max != null) {
                    options.max = moment(options.max);
                }
            }

            if (inBounds(data)) {
                dfd.resolve(data);
            } else {
                dfd.reject({
                    valid: false,
                    data: data,
                    type: 'bounds'
                });
            }

            return dfd.promise();
        };
    };
});

/**
* @module patternValidator
*/
define('oui/utils/validators/patternValidator',[
    'underscore',
    'jquery'
], function (_, $) {
    'use strict';

    /**
     * @typedef patternValidatorOptions
     *
     * @property {RegExp} pattern
     */

    /**
     * @alias module:patternValidator
     * @method patternValidator
     * @param {module:patternValidator~patternValidatorOptions} options
     *
     * @returns {module:validator~ValidateFn}
     */
    return function (options) {

        /**
         * @method module:patternValidator~getPattern
         *
         * @param {module:patternValidator~patternValidatorOptions} options
         *
         * @returns {RegExp}
         */
        function getPattern(options) {
            if (options.pattern) {
                return options.pattern;
            }

            //Support constant regex?
        }

        return function (data) {
            var dfd = $.Deferred();

            var pattern = getPattern(options);

            if (pattern.test(data)) {
                dfd.resolve(data);
            } else {
                dfd.reject({
                    type: 'pattern',
                    data: data,
                    valid: false
                });
            }

            //Make sure we reset the regex in case the user used the global flag
            pattern.lastIndex = 0;

            return dfd.promise();
        };
    };
});

/**
 * @module ajaxValidator
 */
define('oui/utils/validators/ajaxValidator',[
    'underscore',
    'jquery'
], function (_, $) {
    'use strict';

    /**
     * @typedef ajaxValidatorOptions
     *
     * @property {Object} ajaxOptions
     * @property {String} url
     */

    /**
     * @alias module:ajaxValidator
     * @method ajaxValidator
     * @param {module:ajaxValidator~ajaxValidatorOptions} options
     *
     * @returns {module:validator~ValidateFn}
     */
    return function (options) {

        /**
         * @method module:ajaxValidator~getAjaxOptions
         * @private
         *
         * @param {module:ajaxValidator~ajaxValidatorOptions} options
         * @returns {Object}
         */
        var getAjaxOptions = function (options) {
            if (options.url) {
                return {
                    type: 'GET',
                    url: options.url
                };
            } else if (options.ajaxOptions) {
                return options.ajaxOptions;
            }
        };

        return function (data) {
            var dfd = $.Deferred();

            var ajaxOpts = getAjaxOptions(options);

            ajaxOpts.data = data;

            $.ajax(ajaxOpts)
                .done(function () {
                    dfd.resolve(data);
                })
                .fail(function () {
                    dfd.reject({
                        type: 'server',
                        data: data,
                        valid: false
                    });
                });

            return dfd.promise();
        };
    };
});

/**
* @module customValidator
*/
define('oui/utils/validators/customValidator',[
    'underscore',
    'jquery'
], function (_, $) {
    'use strict';

    /**
     * @typedef customValidatorOptions
     *
     * @property {Function} cb
     * @property {String} type The type of error to trigger when rejecting
     */

    /**
     * @alias module:customValidator
     * @method customValidator
     * @param {module:customValidator~customValidatorOptions} options
     *
     * @returns {module:validator~ValidateFn}
     */
    return function (options) {

        return function (data) {
            var dfd = $.Deferred();

            var validationResult = options.cb(data);

            if (validationResult === true || validationResult.valid) {
                dfd.resolve(data);
            } else {
                dfd.reject({
                    type: validationResult.type || options.type || 'generic',
                    data: data,
                    valid: false
                });
            }

            return dfd.promise();
        };
    };
});

define('oui/utils/validators/validator',['require','jquery','oui/utils/validators/typeValidator','oui/utils/validators/requiredValidator','oui/utils/validators/boundsValidator','oui/utils/validators/patternValidator','oui/utils/validators/ajaxValidator','oui/utils/validators/customValidator'],function (require) {
    'use strict';

    var $ = require('jquery');
    var typeValidator = require('oui/utils/validators/typeValidator');
    var requiredValidator = require('oui/utils/validators/requiredValidator');
    var boundsValidator = require('oui/utils/validators/boundsValidator');
    var patternValidator = require('oui/utils/validators/patternValidator');
    var ajaxValidator = require('oui/utils/validators/ajaxValidator');
    var customValidator = require('oui/utils/validators/customValidator');

    /**
     * @module validator
     */

    /**
     * @callback ValidateFn
     *
     * @param {*} data
     * @returns {Promise}
     */

    /**
     * @alias module:validator
     * @method validator
     * @param {*} data The data to validate
     *
     * @example
     *
     *  validator(newData)
     *      .then(validator.typeValidator({
     *          types: ['Null', 'Number', 'moment']
     *      }))
     *      .then(validator.requiredValidator({
     *         required: this._options.required
     *      }))
     *      .then(validator.boundsValidator({
     *          min: this._options.min,
     *          max: this._options.max
     *      })).promise();
     *
     * @returns {Promise}
     */
    function validator(data) {
        return $.when(data);
    }

    /**
     * @member {module:typeValidator} module:validator.typeValidator
     */
    validator.typeValidator = typeValidator;

    /**
     * @member {module:requiredValidator} module:validator.requiredValidator
     */
    validator.requiredValidator = requiredValidator;

    /**
     * @member {module:boundsValidator} module:validator.boundsValidator
     */
    validator.boundsValidator = boundsValidator;

    /**
     * @member {module:patternValidator} module:validator.patternValidator
     */
    validator.patternValidator = patternValidator;

    /**
     * @member {module:ajaxValidator} module:validator.ajaxValidator
     */
    validator.ajaxValidator = ajaxValidator;

    /**
     * @member {module:customValidator} module:validator.customValidator
     */
    validator.customValidator = customValidator;

    return validator;
});

define('oui/widgets/textField/textModel',[
    'underscore',
    'oui/base/model',
    'oui/utils/validators/validator'
], function (_, Model, validator) {
    'use strict';

    /**
     * @class TextModel
     * @extends Model
     * @private
     *
     * @param {String} initData
     * @param {TextField#Config} options
     */
    function TextModel(initData, options) {
        Model.call(this, initData, options);
    }

    var tmp = TextModel.prototype = Object.create(Model.prototype);

    /**
     * @method TextModel#name
     *
     * @returns {String}
     */
    tmp.name = function () {
        return 'text';
    };

    /**
     * @method TextModel#init
     *
     * @param {String} initData
     *
     * @returns {String}
     */
    tmp.init = function (initData) {
        return initData || '';
    };

    /**
     * @method TextModel#text
     *
     * @returns {String}
     */
    /**
     * @method TextModel#text
     *
     * @param {String} text
     * @param {Subscribable~EventOptions} eventParams
     */
    tmp.text = tmp.data;

    /**
     * @method TextModel#getData
     *
     * @returns {String}
     */
    tmp.getData = function () {
        return this._data || '';
    };

    /**
     * @method TextModel#validate
     *
     * @param {String} value
     *
     * @returns {Promise}
     */
    tmp.validate = function (value) {

        if (value !== _.unescape(value)) {
            value = _.unescape(value);
        }

        var dfd = validator(value)
            .then(validator.requiredValidator({
                required: this._options.required
            }))
            .then(validator.boundsValidator({
                min: this._options.minLength,
                max: this._options.maxLength
            }));

        if (this._options.validator) {
            dfd = dfd.then(validator.customValidator({
                cb: this._options.validator.validate
            }));
        }

        return dfd;
    };

    return TextModel;
});

define('oui/widgets/textField',[
    'underscore',
    'oui/base/input',
    'oui/widgets/textField/textView',
    'oui/widgets/textField/textController',
    'oui/widgets/textField/textModel'
], function (_, Input, View, Controller, Model) {
    'use strict';

    var __WIDGET_NAME__ = 'text-field';
    var tfp;

    /**
     * @typedef {Object} TextField#Config
     *
     * @property {Locale} [locale={}] The specific strings that are used by the widget and will use the defaults set from
     * the {@link module:oui.lang|lang}.
     * @property {Boolean|Object} [autoEscape=true] Whether the combobox should escape/unescape the values from the
     * input. When the property is an Object it should have two functions escape and unescape that take in the string and
     * provide the escaped/unescaped version of the string. When the property is true it will default to using _.escape
     * and _.unescape for the strings.
     * @property {Boolean} [fluid=false] Whether or not the text field should take up all available widget.
     * @property {Boolean} [fill=false] Whether or not the widget should take up all available height. **Note**: Only works
     * with text areas.
     * @property {Boolean} [multiline=false] Whether or not the text field supports editing text that is more than one line.
     * Uses a text area instead of an input.
     * @property {Boolean} [required=false] Whether or not the textfield is required.
     * @property {Number} [minLength=null] The minimum length the text is allowed to be
     * @property {Number} [maxLength=null] The maximum length the text is allowed to be
     * @property {Object} [validator] The validator to run after all basic validation has run. Must have `validate` function.
     * @property {RegExp|Function} [restrictTo] Places a restriction on the text field to only the valid values as described
     * by the RegExp or Function. Will not allow any additional characters to be typed as a result.
     * @property {Boolean} [count=false] Whether or not to show the remaining character count on the text field
     * @property {Object} [autoComplete]
     * @property {Boolean} [autoComplete.enabled=false] Whether or not the autocomplete functionality is enabled
     * @property {String} [autoComplete.trigger='@'] The trigger for autocomplete searching to be initialized.
     * @property {Object} [autoComplete.formatted]
     * @property {Object} [autoComplete.formatted.start='[~'] The start string to wrap the text in for display within the
     * text area.
     * @property {Object} [autoComplete.formatted.end=']'] The end string to wrap the text in for display within the text
     * area.
     * @property {Function} [autoComplete.search] Function to return results from a query.
     * @property {Function} [autoComplete.display] Function to transform the node into an HTML element.
     * @property {Function} [autoComplete.text] Function to transform the node into a string.
     * @property {Function} [autoComplete.tooltip] Function to use as the title attribute on the menu item (will fallback to
     * text function if no tooltip function is given).
     */

    var __DEFAULTS__ = {
        showErrors: 'blur',
        fluid: false,
        fill: false,

        //Only works with textareas
        multiline: false,
        required: false,
        minLength: null,
        maxLength: null,
        count: false,
        appendTo: 'body',
        autoEscape: true,
        autoComplete: {
            enabled: false,
            trigger: '@',
            formatted: {
                start: '[~',
                end: ']'
            },
            search: _.constant([]),
            display: function (node) { //jscs:ignore jsDoc
                return _.get(node, 'name', '');
            },
            text: function (node) { //jscs:ignore jsDoc
                return _.get(node, 'name');
            }
        }
    };

    /**
     * Basic text field widget to provide a familiar API to an
     * input field.
     *
     * ##### **Relevant Events**
     *
     * * {@link TextField#text-changed|text-changed}
     * * {@link TextField#error|error}
     * * {@link TextField#required-error|required-error}
     * * {@link TextField#length-error|length-error}
     *
     * @borrows TextView#event:text-changed as TextField#text-changed
     * @borrows TextView#event:error as TextField#error
     * @borrows TextView#event:required-error as TextField#required-error
     * @borrows TextView#event:length-error as TextField#length-error
     *
     * @class TextField
     * @extends Input
     * @param {TextField#Config} [options]
     * @param {jQuery} [element]
     * @param {String} [initData]
     */
    function TextField(options, element, initData) {
        Input.call(this, options, element, initData);

        var model = new Model(this._initData, this._options);

        this._controller = new Controller(model, this._options);
        this._view = new View(model, this._controller, this._element, this._options, this._settings);

        Object.defineProperty(this, '_data', {
            get: this.text,
            set: this.text
        });

        this._setupEvents([model, this._view]);
    }

    TextField.prototype = Object.create(Input.prototype);
    tfp = TextField.prototype;

    tfp.constructor = TextField;

    /**
     * Get the name of the widget.
     *
     * @method TextField#name
     * @returns {String} 'text-field'
     */
    tfp.name = function () {
        return __WIDGET_NAME__;
    };

    /**
     * Combined getter/setter for the text fields value.
     *
     * @method TextField#text
     * @param {String} [newText] The new value of the text field.
     * @returns {String} The current value of the text field.
     */
    tfp.text = function (newText) {
        if (!arguments.length) {
            return this._controller.text();
        }

        this._controller.text(newText);
    };

    /**
     * Combined getter/setter for the text fields value.
     *
     * Closely related to @{link TextField#text}
     *
     * **Note**: does not fire any public events.
     *
     * @method TextField#value
     * @param {String} [newText] The new value of the text field.
     * @returns {String} The current value of the text field.
     */
    tfp.value = function (newText) {
        if (!arguments.length) {
            return this._controller.text();
        }

        this._controller.text(newText, { silent: true });
    };

    /**
     * Get the defaults for the widget.
     *
     * @method TextField#defaults
     * @returns {Object}
     */
    tfp.defaults = function () {
        return __DEFAULTS__;
    };

    /**
     * Links to text fields together so they share the same content until separated.
     *
     * @method TextField#link
     * @param {TextField} textField The text field to link
     * @param {Object} modifiers
     */
    tfp.link = function (textField, modifiers) {
        this._view.link(textField);

        if (!modifiers || modifiers.bidirectional) {
            textField.link(this, { bidirectional: false });
        }
    };

    /**
     * Gets the current validity status of the control.
     *
     * @method TextField#isValid
     * @returns {Boolean} True if it is currently valid and false otherwise
     */
    tfp.isValid = function () {
        return this._controller.isValid();
    };

    /**
     * Inserts the given text at the currently selected cursor location.
     *
     * @method TextField#insert
     * @param {String} text The text to add to the text field
     */
    tfp.insert = function (text) {
        this._view.insert(text);
    };

    /**
     * @method TextField#error
     *
     * @returns {Object}
     */
    tfp.error = function () {
        return this._controller.error();
    };

    return TextField;
});


define('template-5',[],function () { return '<div class="orcl-widget orcl-date-picker">\n    <input type="date" spellcheck="false"/>\n</div>\n';});


define('template-6', ['underscore'], function (_) { return function(data) {
var __t, __p = '';
__p += '<button type="button" title="' +
((__t = ( data.locale.calendar.selectDate  )) == null ? '' : __t) +
'" class="btn btn-picker" tabindex="-1"> <i class="pgbu-icon-small pgbu-icon-calendar"></i></button>';
return __p
};});


define('template-7',[],function () { return '<div class="orcl-widget orcl-date-picker">\n</div>\n';});


define('template-8', ['underscore'], function (_) { return function(data) {
var __t, __p = '', __j = Array.prototype.join;
function print() { __p += __j.call(arguments, '') }
__p += '<div class="calendar dark-theme" aria-hidden="true"> <span class="button-controls"> <div class="prev-buttons"> <button type="button" data-field="year" data-value="prev" title="' +
((__t = (data.locale.calendar.prevYear)) == null ? '' : __t) +
'" tabindex="-1" class="btn btn-blue-light">&laquo;</button> <button type="button" data-field="month" data-value="prev" title="' +
((__t = (data.locale.calendar.prevMonth)) == null ? '' : __t) +
'" tabindex="-1" class="btn btn-blue-light">&lsaquo;</button> </div> <label id="' +
((__t = ( data.id  )) == null ? '' : __t) +
'" role="heading" aria-live="assertive" aria-atomic="true" class="title"></label> <div class="next-buttons"> <button type="button" data-field="month" data-value="next" title="' +
((__t = (data.locale.calendar.nextMonth)) == null ? '' : __t) +
'" tabindex="-1" class="btn btn-blue-light">&rsaquo;</button> <button type="button" data-field="year" data-value="next" title="' +
((__t = (data.locale.calendar.nextYear)) == null ? '' : __t) +
'" tabindex="-1" class="btn btn-blue-light">&raquo;</button> </div> </span> <div class="month"> <table aria-labelledby="' +
((__t = ( data.id  )) == null ? '' : __t) +
'"> <thead> <tr> <th scope="col"></th> <th scope="col"></th> <th scope="col"></th> <th scope="col"></th> <th scope="col"></th> <th scope="col"></th> <th scope="col"></th> </tr> </thead> ';
 for (var i = 0; i < 6; i++) { ;
__p += ' <tr> <td></td> <td></td> <td></td> <td></td> <td></td> <td></td> <td></td> </tr> ';
 } ;
__p += ' </table> </div> <span class="button-controls today-cancel"> ';
 if (data.type === 'datetime' ) { ;
__p += ' <input type="number"> ';
 } ;
__p += ' <div class="group"> <button type="button" tabindex="-1" title="' +
((__t = ( data.locale.today )) == null ? '' : __t) +
'" class="btn btn-small btn-blue-light today">' +
((__t = ( data.locale.today )) == null ? '' : __t) +
'</button> <div class="group pull-right"> <button type="button" tabindex="-1" title="' +
((__t = ( data.locale.cancel )) == null ? '' : __t) +
'" class="btn btn-small btn-blue-light cancel">' +
((__t = ( data.locale.cancel )) == null ? '' : __t) +
'</button> ';
 if (data.type === 'datetime' ) { ;
__p += ' <button type="button" tabindex="-1" title="' +
((__t = ( data.locale.ok )) == null ? '' : __t) +
'" class="btn btn-small btn-action ok">' +
((__t = ( data.locale.ok )) == null ? '' : __t) +
'</button> ';
 } ;
__p += ' </div> </div> </span></div>';
return __p
};});


define('template-9', ['underscore'], function (_) { return function(data) {
var __t, __p = '';
__p += '<input type="' +
((__t = ( data.type  )) == null ? '' : __t) +
'">';
return __p
};});


define('template-10', ['underscore'], function (_) { return function(data) {
var __t, __p = '';
__p += '<div class="oui-backdrop"></div>';
return __p
};});

define('oui/utils/dateHelper',[
    'underscore',
    'moment'
], function (_, moment) {
    'use strict';

    /**
     * Collection of functions to help with dealing with dates.
     *
     * @module dateHelper
     * @private
     */

    /**
     * @method numberComparator
     * @private
     * @param {Number} num1 The first number to compare
     * @param {Number} num2 The second number to compare
     */
    var numberComparator = function (num1, num2) {
        if (num1 > num2) {
            return 1;
        } else if (num1 < num2) {
            return -1;
        }

        return 0;
    };

    /**
     * @method flexComparator
     * @private
     * @param {Array<String>} fields
     * @param {Object} item1 The first item to compare
     * @param {Object} item2 The second item to compare
     */
    var flexComparator = function (fields, item1, item2) {
        var value;

        for (var i = 0; i < fields.length; i++) {
            value = numberComparator(item1[fields[i]](), item2[fields[i]]());

            if (value !== 0) {
                return value;
            }
        }

        return 0;
    };

    /**
     * @exports dateHelper
     */
    return {
        /**
         * Determine whether dates are before or after another date.
         *
         * Returns -1 when the first date is before the second
         * Returns 1 when the first date is after the second
         * Returns 0 when they are the same date.
         *
         * @method dateComparator
         * @param {moment} - The first moment to compare
         * @param {moment} - The second moment to compare
         * @returns num Whether the first date is an earlier date than the second.
         */
        dateComparator: _.partial(flexComparator, ['year', 'month', 'date']),
        /**
         * Determine whether dates are before or after another date.
         * Only looks at time fields (hours, minutes, seconds)
         *
         * Returns -1 when the first date is before the second
         * Returns 1 when the first date is after the second
         * Returns 0 when they are the same date.
         *
         * @method timeComparator
         * @param {moment} - The first moment to compare
         * @param {moment} - The second moment to compare
         * @returns num Whether the first date is an earlier date than the second.
         */
        timeComparator: _.partial(flexComparator, ['hours', 'minutes', 'seconds']),
        /**
         * Determines if a specific day of week or date is contained in an array.
         *
         * @method contains
         * @param {Array} dateArray The array of dates or integers (day of week) to check whether a date is contained in.
         * @param {moment} day The day to check
         * @returns {Boolean} contained Whether or not the day is contained in the array.
         */
        contains: function (dateArray, day) {
            //dataArray may contain days of week [0-6]
            var contained = _.contains(dateArray, day.day());

            //dateArray may contain specific days (such as holidays)
            var specificDay = _.find(dateArray, function (val) {
                return moment.isMoment(val) && val.isSame(day, 'day');
            });

            contained = contained || !_.isUndefined(specificDay);

            return contained;
        },

        /**
         * Clones the given object, using Moment's clone functionality for any moments.
         *
         * @method cloneWithDates
         * @param {Object} objectToClone The object to clone
         * @param {Boolean} deep A flag - if true, will do a cloneDeep, otherwise a clone
         * @returns {Object} The given object, cloned
         */
        cloneWithDates: function (objectToClone, deep) {
            return _[deep ? 'cloneDeep' : 'clone'](objectToClone, function (mom) {
                if (moment.isMoment(mom)) {
                    return mom.clone();
                }
            });
        }
    };

});

define('oui/widgets/spinner/numberModel',[
    'underscore',
    'oui/base/model',
    'moment',
    'oui/utils/dateHelper',
    'oui/utils/validators/validator'
], function (_, Model, moment, dateHelper, validator) {
        'use strict';

        /**
         * @method NumberModel~above
         * @private
         * @param {Number|moment|null} val1
         * @param {Number|moment|null} val2
         * @param {String} type
         *
         * @returns {Boolean}
         */
        function above(val1, val2, type) {
            if (val1 == null || val2 == null) {
                return true;
            }

            if (_.isNumber(val1) && _.isNumber(val2)) {
                return (val1 >= val2);
            } else if (type === 'time') {
                return dateHelper.timeComparator(val1, val2) >= 0;
            } else if (moment.isMoment(val1)) {
                return val1.isAfter(val2) || val1.isSame(val2);
            }
        }

        /**
         * @event NumberModel#non-number-error
         * @param {*} attempted The value that was attempted to be set on the number field.
         *  The type is anything that is not a number, null or moment object.
         * @param {moment|Number} currentValue The current value of the number field.
         */
        /**
         * @event NumberModel#number-change
         * @param {moment|Number|null} currentValue The current value of the number field.
         * @param {moment|Number|null} previousValue The value it changed from
         */
        /**
         * @event NumberModel#bounds-error
         * @param {moment|Number} newData The value that was attempted to be set on the number field.
         * @param {moment|Number} currentValue The current value of the number field.
         */
        /**
         * @event NumberModel#required-error
         * @param {moment|Number} newData The value that was attempted to be set on the number field.
         * @param {moment|Number} currentValue The current value of the number field.
         */
        /**
         * @event NumberModel#error
         * @param {*} attempted The value that was attempted to be set on the number field.
         *  The type is anything that is not a number, null or moment object.
         * @param {moment|Number} newData The value that was attempted to be set on the number field.
         * @param {Object} error
         * @param {String} error.type The type of error that has caused this event
         * @param {*} error.data The data that caused the event
         * @param {Boolean} error.valid The validity result (Always going to be false)
         */

        var nmp;

        /**
         * @class NumberModel
         * @private
         *
         * @param {null|Number|moment} initData The initial data on this number field.
         * @param {NumberField#Config} options The options for the number field.
         */
        function NumberModel(initData, options) {
            //Super call
            Model.call(this, initData, options);
        }

        NumberModel.prototype = Object.create(Model.prototype);
        nmp = NumberModel.prototype;

        nmp.constructor = NumberModel;

        /**
         * @method NumberModel#name
         * @returns {String}
         */
        nmp.name = function () {
            return 'number';
        };

        /**
         * @method NumberModel#init
         * @param {Number|moment|null} initData
         *
         * @returns {Number|moment|null}
         */
        nmp.init = function (initData) {
            if (initData === '') {
                initData = null;
            }

            if ((this._options.type === 'date' || this._options.type === 'month') && initData == null) {
                initData = moment();
            }

            return initData;
        };

        /**
         * @method NumberModel#inBounds
         * @param {Number|moment} data
         *
         * @returns {Boolean}
         */
        nmp.inBounds = function (data) {
            return above(data, this._options.min, this._options.type) &&
                above(this._options.max, data, this._options.type);
        };

        /**
         * @method NumberModel#validate
         * @param {Number|moment} newData
         *
         * @returns {Promise}
         */
        nmp.validate = function (newData) {
            //Null is a valid value of data which is the 'empty' value so only when there is no value
            if (_.isUndefined(newData)) {
                newData = this.data();
            }

            return validator(newData)
                .then(validator.typeValidator({
                    types: ['Null', 'Number', 'moment']
                }))
                .then(validator.requiredValidator({
                    required: this._options.required
                }))
                .then(validator.boundsValidator({
                    min: this._options.min,
                    max: this._options.max
                })).promise();

        };

        /**
         * @method NumberModel#getData
         * @returns {Number|moment|null}
         */
        nmp.getData = function () {
            //Clone the date before returning
            return moment.isMoment(this._data) ? moment(this._data) : this._data;
        };

        /**
         * @method NumberModel#isSame
         * @param {Number|moment|null} newData
         *
         * @returns {Boolean}
         */
        nmp.isSame = function (newData) {
            if (moment.isMoment(this._data) && moment.isMoment(newData)) {
                return this._data.isSame(newData);
            }

            return Model.prototype.isSame.call(this, newData);
        };

        /**
         * @method NumberModel#setData
         *
         * @param {Number|moment|null} newData
         * @param {Subscribable~EventOptions} eventParams
         */
        nmp.setData = function (newData, eventParams) {
            if (_.isString(newData) && this._options.formatter.validate(newData)) {
                newData = this._options.formatter.format(newData);
            }

            Model.prototype.setData.call(this, newData, eventParams);
        };

        /**
         * @method NumberModel#canUpdate
         * @param {Number|moment} value
         * @param {String} field
         * @param {Object} options
         *
         * @returns {Object}
         */
        nmp.canUpdate = function (value, field, options) {
            var incValue;
            var decValue;
            var strict = _.get(options, 'strict', true);
            var step = this._options.spinner.step;

            if (this._options.spinner.step[field]) {
                step = this._options.spinner.step[field];
            }

            //Make potentials and check those to the bounds
            if (moment.isMoment(value) || this._options.type === 'time' || this._options.type === 'date') {

                if (value == null) {
                    incValue = moment();
                    decValue = incValue.clone();
                } else {
                    incValue = value.clone();
                    decValue = value.clone();
                }

                if (field === 'half-day') {
                    field = 'hours';
                    step = 12;
                }

                incValue[field](incValue[field]() + step);
                decValue[field](decValue[field]() - step);

            } else {
                incValue = +(value + step).toFixed(12);
                decValue = +(value - step).toFixed(12);
            }

            if (strict) {
                return {
                    increment: this.inBounds(incValue),
                    decrement: this.inBounds(decValue)
                };
            } else {
                return {
                    increment: incValue <= this._options.max,
                    decrement: decValue >= this._options.min
                };
            }
        };

        return NumberModel;
    });

define('oui/widgets/spinner/numberController',[
    'underscore',
    'moment'
], function (_, moment) {
    'use strict';

    var ncp;

    var TIME_REGEX = /(\d+):(\d+)(?::(\d+))?(?: (\w{2}))?/;

    /**
     * @class NumberController
     * @private
     *
     * @param {NumberModel} model
     * @param {NumberField#Config} options
     */
    function NumberController(model, options) {

        this._model = model;
        this._options = options;
    }

    ncp = NumberController.prototype;

    ncp.constructor = NumberController;

    /**
     * Gets the next valid boundary value otherwise returns the given value.
     *
     * When below min value it will return the min boundary to snap to and
     * when above max value it will return the max boundary to snap to.
     *
     * @method NumberController#_getValidNumber
     * @private
     * @param {Number} newValue
     * @returns {Number}
     */
    ncp._getValidNumber = function (newValue) {
        var error = this._model.error();
        if (error.type === 'bounds') {
            if (error.data < this._options.min) {
                return this._options.min;
            } else if (error.data > this._options.max) {
                return this._options.max;
            }
        }

        return newValue;
    };

    /**
     * When the number field represents a number updates the underlying
     * number by the value provided for the field provided.
     *
     * @method NumberController#_updateNumber
     * @private
     * @param {String} field
     * @param {Number} value
     * @param {Number} data
     */
    ncp._updateNumber = function (field, value, data) {
        var newValue;

        if (data == null) {
            data = this._model.data();
        }
        if (this._options.spinner.step[field]) {
            value = this._options.spinner.step[field];
        }
        newValue = +(data + value).toFixed(12);
        if (!this._model.isValid()) {
            newValue = this._getValidNumber(newValue);
        }

        this._model.data(newValue);

    };

    /**
     * When the number field represents a moment (date/time), updates the underlying
     * number by the value provided for the field provided
     *
     * @method NumberController#_updateMoment
     * @private
     * @param {String} field
     * @param {moment} value
     * @param {moment} data
     */
    ncp._updateMoment = function (field, value, data) {
        var potentialTime = data || this._model.data() || moment().startOf('day');

        if (field === 'half-day') {
            potentialTime.add(12, 'hours');
        } else {
            potentialTime.add(value, field);
        }

        this._model.data(potentialTime);
    };

    /**
     * @method NumberController#update
     * @param {String} field
     * @param {Number|moment} value
     * @param {Number|moment} data
     */
    ncp.update = function (field, value, data) {
        if (this._options.type === 'number' || this._options.type === 'currency') {
            this._updateNumber(field, value, data);
        } else {
            this._updateMoment(field, value, data);
        }
    };

    /**
     * @method NumberController#canUpdate
     * @param {Number|moment} value
     * @param {String} field
     * @param {Object} options
     *
     * @returns {Boolean}
     */
    ncp.canUpdate = function (value, field, options) {
        return this._model.canUpdate(value, field, options);
    };

    /**
     * @method NumberController#increment
     * @param {String} field
     * @param {Number|moment} val
     */
    ncp.increment = function (field, val) {
        if (val != null && this._options.formatter.validate(val, { forgiving: true })) {
            val = this._options.formatter.unformat(val, { forgiving: true });
            if (this._model.canUpdate(val, field).increment) {
                this.update(field, this._options.spinner.step, val);
            } else if (val < this._options.min) {
                this.update(field, 0, this._options.min);
            } else if (val > this._options.max) {
                this.update(field, 0, this._options.max);
            }
        } else if (this._model.canUpdate(this._model.data(), field).increment) {
            this.update(field, this._options.spinner.step);
        }
    };

    /**
     * @method NumberController#decrement
     * @param {String} field
     * @param {Number|moment} val
     */
    ncp.decrement = function (field, val) {
        if (val != null && this._options.formatter.validate(val, { forgiving: true })) {
            val = this._options.formatter.unformat(val, { forgiving: true });
            if (this._model.canUpdate(val, field).decrement) {
                this.update(field, -this._options.spinner.step, val);
            } else if (val < this._options.min) {
                this.update(field, 0, this._options.min);
            } else if (val > this._options.max) {
                this.update(field, 0, this._options.max);
            }
        } else if (this._model.canUpdate(this._model.data(), field).decrement) {
            this.update(field, -this._options.spinner.step);
        }
    };

    /**
     * @method NumberController#data
     * @param {Number|moment} value
     * @param {Subscribable~EventOptions} eventParams
     */
    ncp.data = function (value, eventParams) {
        if (!arguments.length) {
            return this._model.data();
        }

        var newValue;
        if (value === '' && this._options.emptyValue != null) {
            newValue = this._options.emptyValue;
        } else if (_.isString(value) && this._options.formatter.validate(value, { forgiving: true })) {
            newValue = this._options.formatter.unformat(value, { forgiving: true });
        } else {
            newValue = value;
        }

        this._model.data(newValue, eventParams);
    };

    /**
     * @method NumberController#isValid
     * @returns {Boolean}
     */
    ncp.isValid = function () {
        return this._model.isValid();
    };

    /**
     * @method NumberController#error
     * @returns {Object}
     */
    ncp.error = function () {
        return this._model.error();
    };

    /**
     * @method NumberController#validate
     * @param {Subscribable~EventOptions} eventParams
     */
    ncp.validate = function (eventParams) {
        var data = this.data();

        if (!this.isValid()) {
            data = this.error().data;
        }

        this._model.data(data, eventParams || {});
    };

    /**
     * @method NumberController#getField
     *
     * @param {Number} loc
     * @param {String} value
     *
     * @returns {String}
     */
    ncp.getField = function (loc, value) {
        var matches = TIME_REGEX.exec(value) || [0, '', '', ''];
        var hours = 0;
        var minutes = 0;
        var seconds = 0;

        loc = loc || 0;

        if (_.isFunction(this._options.field)) {
            return this._options.field(loc, value);
        }

        hours = matches[1].length + 1;
        minutes = hours + matches[2].length + 1;
        if (matches[3]) {
            seconds = minutes + matches[3].length + 1;
        }

        if (loc < hours) {
            return 'hours';
        } else if (loc < minutes) {
            return 'minutes';
        } else if (loc < seconds) {
            return 'seconds';
        } else {
            return 'half-day';
        }
    };

    return NumberController;
});

define('oui/utils/repeater',['underscore'], function (_) {
    'use strict';

    /**
     * @module repeater
     * @private
     */

    /**
     * @typedef Object module:repeater~Repeater
     *
     * @property {Function} start
     * @property {Function} stop
     */

    /**
     * @alias module:repeater
     * @method
     *
     * @param {Function} fn
     * @param {Object} options
     *
     * @returns {module:repeater~Repeater}
     */
    return function (fn, options) {
        options = _.defaults(options || {}, { initial: 500, step: 50, min: 50 });
        var time = options.initial;
        var interval = null;
        var stopped = false;

        /**
         * @method module:repeater~recur
         * @private
         */
        function recur() {
            //TODO: In case of undefined
            fn();

            if (!stopped) {
                clearInterval(interval);
                interval = setInterval(recur, time);
            }

            if (time - options.step >= options.min) {
                time -= options.step;
            }
        }

        return {
            start: function () { //jscs:ignore jsDoc
                stopped = false;
                recur();
            },
            stop: function () { //jscs:ignore jsDoc
                clearInterval(interval);
                time = options.initial;
                stopped = true;
            }
        };
    };
});


define('template-11',[],function () { return '<div class="orcl-widget orcl-number-field">\n    <input type="text">\n</div>\n';});


define('template-12',[],function () { return '<div class="orcl-widget orcl-number-field">\n</div>\n';});


define('template-13', ['underscore'], function (_) { return function(data) {
var __t, __p = '';
__p += '<div class="spinner-controls"> <button type="button" data-action="increment" title="' +
((__t = ( data.locale.spinner.increment  )) == null ? '' : __t) +
'" class="btn btn-form btn-micro" tabindex="-1"> <i class="up"></i> </button> <button type="button" data-action="decrement" title="' +
((__t = ( data.locale.spinner.decrement  )) == null ? '' : __t) +
'" class="btn btn-form btn-micro" tabindex="-1"> <i class="down"></i> </button></div>';
return __p
};});

define('oui/widgets/spinner/numberView',['require','underscore','jquery','moment','oui/base/view','oui/utils/keyCode','oui/utils/repeater','oui/utils/inputUtils','template-9','template-11','template-12','template-13','template-0'],function (require) { //jshint ignore:line
      'use strict';

      var _ = require('underscore');
      var $ = require('jquery');
      var moment = require('moment');
      var View = require('oui/base/view');
      var keyCode = require('oui/utils/keyCode');
      var repeat = require('oui/utils/repeater');
      var inputUtils = require('oui/utils/inputUtils');
      var nativeElement = require('template-9');
      var spinnerTemplate = require('template-11');
      var wrapTemplate = require('template-12');
      var controlsTemplate = require('template-13');
      var instructions = require('template-0');

      var nvp;

      var nativeTimeFormat = 'HH:mm';

      /**
       * @method NumberView~getWheelDelta
       * @private
       *
       * @param {Event} e
       * @returns {Number}
       */
      function getWheelDelta(e) {
          if (!e.deltaY) {
              e = e.originalEvent;
          }

          return -e.deltaY;
      }

      /**
       * @class NumberView
       * @private
       *
       * @param {NumberModel} model
       * @param {NumberController} controller
       * @param {jQuery} element
       * @param {NumberField#Config} options
       * @param {Object} settings
       */
      function NumberView(model, controller, element, options, settings) {
          var that = this;

          /**
           * @method NumberView~fieldChangeHandler
           * @private
           */
          function fieldChangeHandler() {
              var update;
              var data = that._controller.data();
              that._cursorLoc = inputUtils.getInputSelection(that._$input[0]).start;

              if (!that._controller.isValid() && that._controller.error().type === 'bounds') {
                  data = that._controller.error().data;
                  update = that._controller.canUpdate(data, that._getField(that._cursorLoc, that._inputEl.value), { strict: false });
              } else {
                  update = that._controller.canUpdate(data, that._getField(that._cursorLoc, that._inputEl.value));
              }

              that._$incButton.prop('disabled', !update.increment);
              that._$decButton.prop('disabled', !update.decrement);
          }

          View.call(this, options, element, controller, settings);

          model.on({
              'view.change': this.update.bind(this),
              'view.error': this.error.bind(this),
              'view.bounds-error': this.boundsError.bind(this)
          });

          //Was calling find a lot so made it a variable
          this._$input = this._$el.find('input');
          this._inputEl = this._$input[0];
          this._$buttons = this._$el.find('button');
          this._$incButton = this._$buttons.filter('[data-action=increment]');
          this._$decButton = this._$buttons.filter('[data-action=decrement]');
          this.min(this._options.min);
          this.max(this._options.max);
          this._triggerChange = false;
          this._repeaters = {
              increment: repeat(_.bind(this.increment, this)),
              decrement: repeat(_.bind(this.decrement, this))
          };

          //Update the view if the model has some data
          if (!_.isNull(this._controller.data())) {
              this.update(this._controller.data());
          } else if (!this._controller.isValid()) {
              this.error(this._controller.error().data);
          }

          this._$input.on('change', function (event, data) {
              data = data || {
                  update: true
              };

              if (data == null || data.update) {
                  var value = this.value;

                  if (that._options.type === 'time' && that._$el.hasClass('native')) {
                      value = moment(this.value, nativeTimeFormat);
                  }
                  that._controller.data(value);
              }
          });

          this._$input.on('click touchend', function (event) {
              if (event.type === 'touchend' && $(document.activeElement).is(this)) {
                  event.preventDefault();
              }
              if (that._options.spinner.controls) {
                  fieldChangeHandler();
              }
          });

          this._$buttons.on('mousedown', function (event) {
              var action = $(this).attr('data-action');
              var repeater = that._repeaters[$(this).attr('data-action')];
              var namespace = ['oui', 'spinner', action].join('-');
              var events = 'mouseup.' + namespace + ' mouseleave.' + namespace;
              event.preventDefault();

              //Keep focus where it is
              that._$input.focus();

              if (!that.isReadOnly()) {
                  repeater.start();
                  $('body').one(events, function () {
                      repeater.stop();
                      $('body').off(events);
                      that.change();
                  });
              }
          })
          .on('click', function (event) {
              event.preventDefault();
          });

          if (this._options.spinner.scroll) {
              this._$input.on('wheel', function (event) {

                  //http://youmightnotneedjquery.com/#matches
                  if (!that.isReadOnly() &&
                    !that._$el.hasClass('native') && that._$input[0] === document.activeElement &&
                    !(event.shiftKey || event.ctrlKey || event.metaKey || event.altKey)) {

                      event.preventDefault();

                      if (getWheelDelta(event) < 0) {
                          that.decrement();
                      } else {
                          that.increment();
                      }

                      that.change();
                  }
              });
          }

          this._$input
          .on('focus', function () {
              var newValue;
              var selection = inputUtils.getInputSelection(that._inputEl);

              if (that._controller.isValid()) {
                  newValue = that._options.formatter.format(that._controller.data(), {
                      truncate: false
                  });

                  that._inputEl.value = newValue;
              } else if (that._controller.error().type === 'required') {
                  that.error(that._controller.error().data);
              }

              that._currentValue = that._inputEl.value;

              //Will wipe out the start if it was at something like 2
              if (selection.start !== selection.end && newValue != null) {
                  inputUtils.setInputSelection(that._inputEl, { start: 0, end: newValue.length });
              } else {
                  inputUtils.setInputSelection(that._inputEl, selection);
              }
          })
          .on('blur', function () {
              var newValue;

              if (that._controller.isValid()) {
                  newValue = that._options.formatter.format(that._controller.data(), {
                      truncate: true
                  });

                  that._inputEl.value = newValue;
              } else {
                  that.error(that._controller.error().data);
              }

          })
          .on('keydown', function (event, additionalData) {
              switch (event.keyCode) {
                  case keyCode.ENTER:

                      if (additionalData == null || additionalData.retrigger) {
                          //TODO: determine best way to handle primutils default-action="enter"
                          //when in keydown it needs this, when in keyup it will fire prematurely due to popups (combobox, date
                          //picker)
                          //https://jira.oraclecorp.com/jira/browse/PRIMCR-1547
                          if (that._inputEl.value !== that._currentValue) {
                              event.stopPropagation();
                              event.preventDefault();
                              that._$input.trigger('change');
                              var ev = $.Event('keydown', {
                                  keyCode: event.keyCode
                              });
                              that._$input.trigger(ev, { retrigger: false });
                          }
                      }
                      break;
                  case keyCode.UP:
                      if (!that._$el.hasClass('native') && that._options.spinner.controls) {
                          event.preventDefault();
                          that.increment();
                          that.change();
                      }
                      break;
                  case keyCode.DOWN:
                      if (!that._$el.hasClass('native') && that._options.spinner.controls) {
                          event.preventDefault();
                          that.decrement();
                          that.change();
                      }
                      break;
              }
          })
        .on('keypress', function (event) {
            var str  = inputUtils.insertCharacter(event.target, event.which, event.shiftKey);

            if (event.keyCode !== keyCode.ENTER &&
                event.keyCode !== keyCode.TAB) {

                if (_.isRegExp(that._options.restrictTo) && !that._options.restrictTo.test(str)) {
                    event.preventDefault();
                    that._options.restrictTo.lastIndex = 0;
                } else if (_.isFunction(that._options.restrictTo) && !that._options.restrictTo(str)) {
                    event.preventDefault();
                }
            }
        }).on('keyup', function (event) {
            if (that._options.spinner.controls) {
                switch (event.keyCode) {
                    case keyCode.END:
                        that._handleEnd(event);
                        break;
                    case keyCode.HOME:
                        that._handleHome(event);
                        break;
                    case keyCode.RIGHT:
                    case keyCode.LEFT:
                        fieldChangeHandler();
                        break;
                }
            }
        });
      }

      NumberView.prototype = Object.create(View.prototype);
      nvp = NumberView.prototype;

      nvp.constructor = NumberView;

      /**
       * @method NumberView#nativeType
       *
       * @param {NumberField#Config} options
       * @returns {String?}
       */
      nvp.nativeType = function (options) {
          if (options.type === 'number' || options.type === 'currency') {
              return 'number';
          } else if (options.type === 'time') {
              return options.type;
          }
          return null;
      };

      /**
       * @method NumberView#destroy
       * @param {Object} clean
       */
      nvp.destroy = function (clean) {
          //Remove all events on destroy
          //TODO: only remove oui namespaced events
          this._$el.add(this._$input).add(this._$buttons).off();

          if (clean.self) {
              this._$el.remove();
          } else {
              this._$el.find('.spoken-text').remove();
              this._$input.removeAttr('role aria-describedby aria-valuemin aria-valuemax aria-valuenow aria-valuetext');
              this._$buttons.parent().remove();
              this._$input.unwrap();
          }
      };

      /**
       * @method NumberView#_getField
       * @private
       *
       * @param {Number} loc
       * @param {Number|moment} value
       *
       * @returns {String}
       */
      nvp._getField = function (loc, value) {
          return this._controller.getField(loc, value);
      };

      /**
       * @method NumberView#buildElement
       * @param {NumberField#Config} opts
       * @param {jQuery} el
       *
       * @returns {jQuery}
       */
      nvp.buildElement = function (opts, el) {
          var $element;

          if (_.isNull(el)) {
              //Build element from scratch
              $element = $(spinnerTemplate);
          } else if (el.jquery) {
              //We were given a jquery element
              $element = el.attr('type', 'text')
                .wrap(wrapTemplate)
                .parent();
          }

          if (opts.spinner.controls) {
              $element.append($(controlsTemplate({
                  locale: opts.locale
              })));
              $element.addClass('spinner');
              $element.children('input').attr('role', 'spinbutton');
              $element.append(instructions({
                  instructions: opts.locale.instructions.spinner,
                  name: 'spinner'
              }));
          }

          $element.children('input').attr('required', opts.required);

          return $element;
      };

      /**
       * @method NumberView#buildNativeElement
       * @param {NumberField#Config} options
       * @param {jQuery} el
       *
       * @returns {jQuery}
       */
      nvp.buildNativeElement = function (options, el) {
          var $element;

          if (_.isNull(el)) {
              $element = $(nativeElement({
                  type: this.nativeType(options)
              }));
          } else if (el.jquery) {
              $element = el.attr('type', this.nativeType(options));
          }

          $element = $element.wrap(wrapTemplate)
            .parent();

          $element.find('input')
            .attr('max', options.max)
            .attr('min', options.min)
            .attr('step', options.spinner.step);

          return $element;
      };

      /**
       * @method NumberView#init
       */
      nvp.init = function () {
          var that = this;
          var input = this._$el.children('input');
          var instructions = this._$el.find('.instructions');

          var describedBy = _.reduce(instructions, function (memo, elem) {
              var id = that.id() + '-instructions-' + $(elem).attr('data-name');
              $(elem).attr('id', id);
              return memo + ' ' + id;
          }, '');
          input.attr('aria-describedby', ((input.attr('aria-describedby') || '') + ' ' + describedBy).trim());
      };

      /**
       * @method NumberView#error
       * @param {Number|moment} attempted
       */
      nvp.error = function (attempted) {
          var currentValue = this._inputEl.value;
          var newValue;
          this._$el.add(this._$el.closest('.control-group')).toggleClass('error', !this._controller.isValid() && this._visited);

          if (this._$el.hasClass('native') && moment.isMoment(attempted)) {
              //TODO: determine best way to determine native format for time field
              newValue = attempted.format('HH:mm');

            //Native format is required
          } else if (this._$el.hasClass('native') && _.isNumber(attempted)) {
              newValue = attempted.toString();
          } else if (_.isNumber(attempted) || moment.isMoment(attempted)) {
              newValue = this._options.formatter.format(attempted, {
                  truncate: document.activeElement !== this._inputEl
              });
          } else if (_.isString(attempted)) {
              newValue = attempted;
          } else if (attempted == null) {
              newValue = '';
          }

          if (_.isNaN(Number(newValue)) && this._inputEl.type === 'number') {
              this._$input.attr('type', 'text');
          }

          if (newValue != null && newValue !== currentValue) {
              this._inputEl.value = newValue;
              this._currentValue = this._inputEl.value;
          }
      };

      /**
       * @method NumberView#boundsError
       * @param {Number|moment} attempted
       */
      nvp.boundsError = function (attempted) {
          var update = this._controller.canUpdate(attempted, this._getField(this._cursorLoc, this._inputEl.value), { strict: false });
          this.updateButtons(update);
      };

      /**
       * @method NumberView#updateButtons
       * @param {Object} update
       */
      nvp.updateButtons = function (update) {

          if (!this.isDisabled()) {
              this._$incButton.prop('disabled', !update.increment);

              if (!update.increment) {
                  $('body').trigger('mouseup.oui-spinner-increment');
              }
              this._$decButton.prop('disabled', !update.decrement);
              if (!update.decrement) {
                  $('body').trigger('mouseup.oui-spinner-decrement');
              }
          }
      };

      /**
       * @method NumberView#min
       * @param {Number|moment} min
       */
      nvp.min = function (min) {
          if (!_.isNull(min)) {
              this._$input.attr('aria-valuemin', min);
          } else {
              this._$input.removeAttr('aria-valuemin');
          }
      };

      /**
       * @method NumberView#max
       * @param {Number|moment} max
       */
      nvp.max = function (max) {
          if (!_.isNull(max)) {
              this._$input.attr('aria-valuemax', max);
          } else {
              this._$input.removeAttr('aria-valuemax');
          }
      };

      /**
       * @method NumberView#increment
       */
      nvp.increment = function () {
          if (!this.isReadOnly() && !this.isDisabled()) {
              var input = this._$input[0];
              this._controller.increment(this._getField(this._cursorLoc, this._$input.val()), input.value);
              inputUtils.setInputSelection(input, {
                  start: this._cursorLoc
              });
          }
      };

      /**
       * @method NumberView#decrement
       */
      nvp.decrement = function () {
          if (!this.isReadOnly() && !this.isDisabled()) {
              var input = this._$input[0];
              this._controller.decrement(this._getField(this._cursorLoc, this._$input.val()), input.value);
              inputUtils.setInputSelection(input, {
                  start: this._cursorLoc
              });
          }
      };

      /**
       * @method NumberView#update
       * @param {Number|moment} value
       */
      nvp.update = function (value) {
          var update = this._controller.canUpdate(value, this._getField(this._cursorLoc, this._inputEl.value));
          var formattedValue;
          var currentValue = this._inputEl.value;

          this._$el.add(this._$el.closest('.control-group')).removeClass('error');

          if (_.isNull(value)) {
              this._inputEl.value = '';
          } else if (this._$el.hasClass('native') && moment.isMoment(value)) {
              //TODO: determine best way to determine native format for time field
              this._inputEl.value = value.format('HH:mm');

            //Native format is required
          } else if (this._$el.hasClass('native') && _.isNumber(value)) {
              formattedValue = this._options.formatter.format(value, {
                  truncate: document.activeElement !== this._inputEl
              });

              if (_.isNaN(Number(formattedValue)) && this._inputEl.type === 'number') {
                  this._$input.attr('type', 'text');
              }

              this._inputEl.value = formattedValue;
              this._$input.attr('aria-valuenow', value);
              this._$input.attr('aria-valuetext', formattedValue);
          } else {
              formattedValue = this._options.formatter.format(value, {
                  truncate: document.activeElement !== this._inputEl
              });
              this._inputEl.value = formattedValue;

              this._$input.attr('aria-valuenow', value);
              this._$input.attr('aria-valuetext', formattedValue);
          }

          inputUtils.setInputSelection(this._inputEl, {
              start: this._cursorLoc
          });

          //So the value was updated from what was currently in the input
          if (currentValue !== this._inputEl.value) {
              this._triggerChange = true;
          }

          this._currentValue = this._inputEl.value;

          this.updateButtons(update);
      };

      /**
       * @method NumberView#change
       */
      nvp.change = function () {
          if (this._triggerChange) {
              this._$input.trigger('change', {
                  update: false
              });
              this._triggerChange = false;
          }
      };

      /**
       * Updates the number field to the min value if it the input is editable and
       * the number field has a min value.
       *
       * @method NumberView#_handleEnd
       * @private
       *
       * @param {jQuery.Event} event
       */
      nvp._handleEnd = function (event) {
          event.preventDefault();
          if (!this.isReadOnly()) {
              if (!_.isNull(this._options.min)) {
                  this._controller.data(this._options.min);
              } else if (this._options.type === 'time') {
                  this._controller.data(this._controller.data().startOf('day'));
              }
          }
      };

      /**
       * Updates the number field to the max value if it the input is editable and
       * the number field has a max value.
       *
       * @method NumberView#_handleHome
       * @private
       *
       * @param {jQuery.Event} event
       */
      nvp._handleHome = function (event) {
          event.preventDefault();
          if (!this.isReadOnly()) {
              if (!_.isNull(this._options.max)) {
                  this._controller.data(this._options.max);
              } else if (this._options.type === 'time') {
                  this._controller.data(this._controller.data().endOf('day'));
              }
          }
      };

      return NumberView;
  }
);

define('oui/widgets/numberField',['require','oui/base/input','oui/widgets/spinner/numberModel','oui/widgets/spinner/numberController','oui/widgets/spinner/numberView'],function (require) {
      'use strict';

      var Input = require('oui/base/input');
      var NumberModel = require('oui/widgets/spinner/numberModel');
      var NumberController = require('oui/widgets/spinner/numberController');
      var NumberView = require('oui/widgets/spinner/numberView');

      var nfp;

      /**
       * @typedef {Object} NumberField#ValidationResult
       * @property {Boolean} valid
       * @property {String} type
       * @property {moment|Number} data
       */

      /**
       * @callback NumberField#format
       * @param {moment} moment - The moment object to format
       * @returns {String}
       */

      /**
       * @callback NumberField#unformat
       * @param {String} str - The string representation of a date
       * @returns {moment}
       */

      /**
       * @callback NumberField#formatValidate
       * @param {String} str - The string representation of a date
       * @returns {Boolean}
       */

      /**
       * @callback NumberField#validatorValidate
       * @param {moment|Number} newData - The new data attempted to be set on the number field
       * @returns {NumberField#ValidationResult} validationResult
       */

      /**
       * @typedef {Object} NumberField#Config
       *
       * @property {null|RegExp|Function} [restrictTo=null] Option to determine whether new character should be added to
       * the number field. Fired on keyDown and if it fails the test or the function returns false the event will be
       * prevented.
       * @property {String} [type='number'] The type of data that the number field contains and will represent. Valid
       * options are 'number', 'currency', 'time', 'date'.
       * @property {Boolean} [utc=false] If using a 'time' type number field, then this determines whether the time should
       * be parsed as UTC time or local time. Default is for local time. **Note**: This option is unused and reserved for
       * future use.
       * @property {Object} [spinner] Options for spinner functionality of the number field.
       * @property {Boolean} [spinner.controls=false] Whether or not the increment/decrement arrows should be drawn on the
       * component.
       * @property {Number} [spinner.step=1] The default step value when incrementing and decrementing a value.
       * @property {Boolean} [spinner.scroll=true] Whether or not the spinner should support scroll events on the control. These
       * only get fired when the component has focus.
       * @property {Array<NumberField#validatorValidate>} validators - The array of custom validators to run on the number
       * field.
       * @property {Object} [formatter={}] - The formatter the number field will use when attempting to convert data
       * types. Used when attempting to display the date/number in the input and when extracting the date/number from the input.
       * @property {NumberField#format} [formatter.format]
       * @property {NumberField#unformat} [formatter.unformat]
       * @property {NumberField#formatValidate} [formatter.validate]
       * @property {moment|Number|null} [max=null] The max value the number field can be or `null` for no max.
       * @property {moment|Number|null} [min=null] The min value the number field can be or `null` for no min.
       * @property {moment|Number|null} [emptyValue=null] The value the number field should become if it is not allowed to
       * be empty. If the number field is not allowed to be empty (no null value) then this option should be specified.
       * This can be used in conjunction with the {@link NumberField#Config.required} property.
       * @property {Boolean} [required=false] Whether the spinner is a required field or not.
       * @property {Boolean} [fallback=false] Whether the control should fallback to its native control. The control will
       * always fallback to the native control on a mobile device regardless of setting. **Note**: This property is
       * deprecated and should not be set.
       * @property {Boolean} [fluid=false] Whether the control should take up as much horizontal space as possible.
       * @property {Locale} [locale={}] The specific strings that are used by the widget and will use the defaults set from
       * the {@link module:oui.lang|lang}.
       */

      var __WIDGET_NAME__ = 'number';
      var __defaults__ = {
          showErrors: 'blur',
          restrictTo: null,
          type: 'number',
          utc: false,
          spinner: {
              controls: false,
              step: 1,
              scroll: true
          },
          formatter: {},
          max: null,
          min: null,
          emptyValue: null,
          required: false,
          fallback: false,
          fluid: false
      };

      /**
       * This class represents a number field which can handle both dates/times, and numbers.
       *
       *  ##### **Keyboard Shortcuts**
       *
       *  * <kbd>Up</kbd> - Increment the number field by step.
       *  * <kbd>Down</kbd> - Decrement the number field by step.
       *  * <kbd>Home</kbd> - Increments the number field to the maximum if provided.
       *  * <kbd>End</kbd> - Decrements the number field to the minimum if provided.
       *
       *
       *  ##### **Relevant Events**
       *
       *  * {@link NumberField#non-number-error|non-number-error}
       *  * {@link NumberField#number-change|number-change}
       *  * {@link NumberField#bounds-error|bounds-error}
       *  * {@link NumberField#required-error|required-error}
       *  * {@link NumberField#error|error}
       *
       *  @borrows NumberModel#event:non-number-error as NumberField#non-number-error
       *  @borrows NumberModel#event:number-change as NumberField#number-change
       *  @borrows NumberModel#event:bounds-error as NumberField#bounds-error
       *  @borrows NumberModel#event:required-error as NumberField#required-error
       *  @borrows NumberModel#event:error as NumberField#error
       *
       * @class NumberField
       * @extends Input
       * @since 1.2.0
       *
       * @param {NumberField#Config} [options] The widgets options.
       * @param {HTMLInputElement|jQuery} [element] The JQuery element to make a number field.
       * @param {moment|Number|null} [initData] The data to initialize the widget with.
       */
      function NumberField(options, element, initData) {
          var model;
          var view;
          var controller;

          //Super call
          Input.call(this, options, element, initData);

          model = new NumberModel(this._initData, this._options);
          controller = new NumberController(model, this._options);
          view = new NumberView(model, controller, this._element, this._options, this._settings);

          /**
           * The view that the field calls into to get the element.
           *
           * @private
           * @type {NumberView}
           * @memberof! NumberField.prototype
           */
          this._view = view;
          /**
           * The controller that the field calls into to change the value
           * of the element.
           *
           * @private
           * @type {NumberController}
           * @memberof! NumberField.prototype
           */
          this._controller = controller;

          this._setupEvents([model, view]);
      }

      NumberField.prototype = Object.create(Input.prototype);
      nfp = NumberField.prototype;

      nfp.constructor = NumberField;

      /**
       * Gets the widget name
       *
       * @method NumberField#name
       * @private
       * @returns {String} 'number' - widget name.
       */
      nfp.name = function () {
          return __WIDGET_NAME__;
      };

      /**
       * Getter/setter for the data of the number field.
       *
       * @method NumberField#data
       * @fires NumberModel#non-number-error
       * @fires NumberModel#number-change
       * @fires NumberModel#bounds-error
       * @fires NumberModel#required-error
       * @param {moment|Number|String|null} [newData] The new value of the number
       *   field.
       * @returns {moment|Number|String|null} The current value of the number field.
       */
      nfp.data = function (newData) {
          if (!arguments.length) {
              if (this.isValid()) {
                  return this._controller.data();
              } else {
                  return this._controller.error().data;
              }
          }

          this._controller.data(newData);
      };

      /**
       * Getter/setter for the data of the number field.
       *
       * Similar to {{#crossLink "NumberField/data"}}{{/crossLink}} except it does not
       * trigger an event after selection.
       *
       * @method NumberField#value
       * @param {moment|Number|String|null} [newData] The new value of the number
       *  field.
       * @returns {moment}
       */
      nfp.value = function (newData) {
          if (!arguments.length) {
              return this.data();
          }

          this._controller.data(newData, {
              silent: true
          });
      };

      /**
       * Getter/Setter for the formatter.
       *
       * Rerenders the view when the formatter is reset.
       *
       * @method NumberField#formatter
       * @param {Object} formatter A new formatter to use for the existing number spinner.
       * @returns {Object} - current formatter used by the number spinner.
       */
      nfp.formatter = function (formatter) {
          if (!arguments.length) {
              return this._options.formatter;
          }

          this._options.formatter = formatter;
          this.value(this.data());
          this._view.trigger('reformat');
      };

      /**
       * Getter/Setter for the required status of the component.
       *
       * Revalidates the component when the required attribute changes.
       *
       * @method NumberField#required
       * @param {Boolean} required Whether or not the component is to be marked as required.
       * @returns {Boolean} - The current status of the required option.
       */
      nfp.required = function (required) {
          if (!arguments.length) {
              return this._options.required;
          }

          if (this._options.required !== required) {
              this._options.required = required;

              this.validate();
          }
      };

      /**
       * Getter/Setter for the min value of the spinner.
       *
       * When using the setter it will rerender the view using the new min value.
       *
       * @method NumberField#min
       * @param {moment|Number} min The new min to set
       * @returns {moment|Number} - The current min
       */
      nfp.min = function (min) {
          if (!arguments.length) {
              return this._options.min;
          }

          this._options.min = min;
          this._view.min(min);
          this.validate();
      };

      /**
       * Getter/Setter for the max value of the spinner.
       *
       * When using the setter it will rerender the view using the new max value.
       *
       * @method NumberField#max
       * @param {moment|Number} max The new max to set
       * @returns {moment|Number} - The current max
       */
      nfp.max = function (max) {
          if (!arguments.length) {
              return this._options.max;
          }

          this._options.max = max;
          this._view.max(max);
          this.validate();
      };

      /**
       * Gets the last error of the number field.
       *
       * @method NumberField#error
       * @returns {OUIError|null} error The last error of the field.
       */
      nfp.error = function () {
          return this._controller.error();
      };

      /**
       * Revalidates a number field with the current data.
       *
       * @method NumberField#validate
       */
      nfp.validate = function () {
          this._controller.validate();
      };

      /**
       * Get the validity state of the number field.
       *
       * @method NumberField#isValid
       * @returns {Boolean} - Whether the field is currently in a valid state.
       */
      nfp.isValid = function () {
          return this._controller.isValid();
      };

      /**
       * Programmatically decrement the number field.
       *
       * Field is not required but is only used when the
       * number field is of type <code>'time'</code>.
       *
       * If no field is given it defaults to 'Hours'.
       *
       * Valid fields are:
       *
       *     'Hours',
       *     'Minutes',
       *     'Seconds',
       *     'half-day'
       *
       * @method NumberField#decrement
       * @param {String} [field] The field to increment.
       */
      nfp.decrement = function (field) {
          this._controller.decrement(field);
      };

      /**
       * Programmatically increment the number field.
       *
       * Field is not required but is only used when the
       * number field is of type <code>'time'</code>.
       *
       * If no field is given it will default to hours.
       *
       * Valid fields are:
       *
       *     'Hours',
       *     'Minutes',
       *     'Seconds',
       *     'half-day'
       *
       * @method NumberField#decrement
       * @param {String} [field] The field to increment.
       */
      nfp.increment = function (field) {
          this._controller.increment(field);
      };

      /**
       * @method NumberField#defaults
       * @returns {Object} The default options for the widget.
       */
      nfp.defaults = function () {
          return __defaults__;
      };

      return NumberField;

  }
);

define('oui/widgets/date/dateView',['require','underscore','jquery','oui/base/view','oui/utils/uniqueId','template-5','template-6','template-7','template-8','template-9','template-10','oui/widgets/numberField','oui/utils/keyCode','oui/utils/dateHelper','moment','oui/mixins/popup','oui/utils/math'],function (require) { //jshint ignore:line
    'use strict';

    var _ = require('underscore');
    var $ = require('jquery');
    var View = require('oui/base/view');
    var uniqueId = require('oui/utils/uniqueId');
    var datePickerTemplate = require('template-5');
    var buttonTemplate = require('template-6');
    var wrapTemplate = require('template-7');
    var calendarTemplate = require('template-8');
    var nativeElement = require('template-9');
    var backdrop = require('template-10');
    var NumberField = require('oui/widgets/numberField');
    var keyCode = require('oui/utils/keyCode');
    var dateHelper = require('oui/utils/dateHelper');
    var moment = require('moment');
    var popupMixin = require('oui/mixins/popup');
    var math = require('oui/utils/math');

    var dvp;

    /**
     * @method DateView~extractDate
     *
     * @param {HTMLElement} element
     *
     * @returns {moment}
     */
    function extractDate(element) {
        return moment(element.attr('data-date'), 'YYYY-MM-DD');
    }

    var nativeFormats = {
        month: 'YYYY-MM',
        datetime: 'YYYY-MM-DDTHH:mm:ssZ',
        date: 'YYYY-MM-DD'
    };

    /**
     * @class DateView
     * @private
     * @extends View
     * @mixes popup
     * @param {DateModel} model
     * @param {DateController} controller
     * @param {jQuery|null} element
     * @param {DateField#Config} options
     */
    function DateView(model, controller, element, options, settings) {
        var that = this;

        this._id = uniqueId('dateView');

        //Super call
        View.call(this, options, element, controller, settings);

        //TODO: This shouldn't be necessary
        /**
         * @private
         * @type {DateModel}
         * @memberof! DateView.prototype
         */
        this._model = model;

        //Bind events on the model
        model.on({
            'view.change': this.select.bind(this),
            'view.error': this.error.bind(this),
            'view.deselect': this.select.bind(this)
        });

        this.hideCal = true;
        this.calendar = null;
        this.spinner = null;
        this._calendarStartDate = null;
        this._triggerChange = false;

        if (this._$el.hasClass('mobile') && !this._$el.hasClass('native')) {
            this._$backdrop = $(backdrop());

            this._$backdrop.on('touchstart click', function (event) {
                event.preventDefault();
                that.hidePopup();
            });
            $(this._options.appendTo).append(this._$backdrop);

            if (this._options.type === 'datetime') {
                this._$el.find('input[type=text]').attr('readonly', true);
            }
        }

        //Useful instead of using find everywhere
        this._$button = this._$el.find('button');
        this._$input = this._$el.find('input');
        this._inputEl = this._$input[0];

        this.overButton = false;

        //Bind events to the elements
        this._$button
            .on('click', function (e) {
                e.preventDefault();
                that.toggleCalendar();
            }).hover(function () {
                that.hideCal = false;
                that.overButton = true;
            }, function () {
                that.hideCal = true;
                that.overButton = false;
            })
        .on('mousedown', function (event) {
            event.preventDefault();
            that._$input.focus();
        });

        this._$input
            .on('focus', function () {
                if (!that.isReadOnly() && !that.overButton &&
                    ((that._options.openOnFocus && !that.isVisible()) ||
                     (that._$el.hasClass('mobile') && !that._$el.hasClass('native')))) {
                    that.show(that._model.date());
                }
            })
            .on('input', function () {
                that.lastAction = 'input';
            })
            .on('blur', function () {
                if (that.hideCal && !_.isNull(that.calendar) && that.isVisible()) {
                    //TODO: try not to reference date popup
                    var date = that._$popup.find('.selected');

                    if (date.length === 0) {
                        date = that._$popup.find('.active');
                    }

                    if (that._controller.date() != null && that._controller.isValid() && that._options.type === 'datetime') {
                        that._controller.date(extractDate(date), that.spinner.value());
                    }
                    that.hidePopup();
                    that.updateCalendarAndInput(that._controller.date());
                } else {
                    _.defer(function () {
                        if (document.activeElement !== that._$input[0] && (that.calendar == null || !that.calendar[0].contains(document.activeElement))) {
                            that.hidePopup();
                        }
                    });
                }

                if (this.value === '' && that._options.required) {
                    that._controller.date(null);
                }
            })
            .hover(function () {
                that.hideCal = false;
            }, function () {
                that.hideCal = true;
            })
            .on('change', function (e, data) {
                data = data || {
                    update: true
                };
                that._triggerChange = true;

                if (data == null || data.update) {
                    var value = this.value;
                    if (that._$el.hasClass('native') && value !== '') {
                        value = moment(this.value, nativeFormats[that._options.type]);
                    }
                    that._controller.date(value);
                }
            });

        this.addShortcuts({
            //Opens the popup
            'F2 SHIFT-F10': function () { //jscs:ignore jsDoc
                if (that.isVisible()) {
                    that.cancel();
                } else if (!that.isDisabled() && !that.isReadOnly()) {
                    that.show(that._model.date());
                }

            },

            //Closes the popup
            'ESC': {
                fn: this.cancel,
                stopPropagation: this.isVisible
            },
            'ENTER': {
                fn: this.handleEnter,
                stopPropagation: this.isVisible,
                prevention: this.isVisible
            },
            'DELETE': {
                fn: this.handleDelete
            },
            'SPACE': {
                fn: this.handleSpace,
                prevention: _.constant(false)
            },
            'SHIFT-TAB TAB': {
                fn: this.handleTab.bind(this),
                prevention: _.constant(false)
            },

            //Move highlighted day
            'UP': {
                fn: _.partial(this.move, {
                    week: -1
                }),
                prevention: this.isVisible
            },
            'DOWN': {
                fn: _.partial(this.move, {
                    week: 1
                }),
                prevention: this.isVisible
            },
            'LEFT': {
                fn: _.partial(this.move, {
                    day: -1
                }),
                prevention: this.isVisible
            },
            'RIGHT': {
                fn: _.partial(this.move, {
                    day: 1
                }),
                prevention: this.isVisible
            },
            'PAGEUP': _.partial(this.move, {
                month: -1
            }),
            'PAGEDOWN': _.partial(this.move, {
                month: 1
            }),
            'HOME': {
                fn: this.jumpTo.bind(this, 'first-day'),
                prevention: this.isVisible
            },
            'END': {
                fn: this.jumpTo.bind(this, 'last-day'),
                prevention: this.isVisible
            },
            'alt-PAGEUP': _.partial(this.move, {
                year: -1
            }),
            'alt-PAGEDOWN': _.partial(this.move, {
                year: 1
            })
        });
    }

    DateView.prototype = Object.create(View.prototype);
    dvp = DateView.prototype;

    dvp.constructor = DateView;

    _.extend(dvp, popupMixin);

    /**
     * @method DateView#nativeType
     *
     * @param {DateField#Config} options
     *
     * @returns {String}
     */
    dvp.nativeType = function (options) {
        return options.type;
    };

    /**
     * Provides view specific destroy logic.
     *
     * Removes event listeners and elements.
     *
     * @method DateView#destroy
     *
     * @param {Object} cleanUp
     */
    dvp.destroy = function (cleanUp) {

        //Calendar should probably be appended to the DOM on create
        if (!_.isNull(this.calendar)) {
            this.calendar.find('button td').off();
            if (!_.isNull(this.spinner)) {
                this.spinner.destroy();
            }
            this.calendar.remove();
        }

        //Who needs to specify events when off and no parameters will remove everything
        //http://api.jquery.com/off
        this._$input.off();
        this._$button.off();
        this._$el.off();

        if (cleanUp.self) {
            this._$el.remove();
        }
    };

    /**
     * @method DateView#show
     *
     * @param {moment} date
     */
    dvp.show = function (date) {
        if (this.calendar == null && !this._$el.hasClass('native')) {
            this.calendar = this._$popup = this.buildCalendar();

            $(this._options.appendTo).append(this._$popup);
        }

        date = date || this._model.date();
        this.showDate(date);

        if (this._options.type === 'datetime') {
            this.spinner.value(this._model.date());
        }

        this.showPopup(this._$el);

        if (!this._$el.hasClass('native') && this._$el.hasClass('mobile')) {
            this._$backdrop.addClass('show');
        }
    };

    /**
     * @method DateView#handleEnter
     */
    dvp.handleEnter = function () {
        if (this.isVisible()) {
            this._triggerChange = true;

            var active = this._$popup.find('.active');
            var time = this.spinner != null ? this.spinner.data() : null;

            if (!active.hasClass('inactive') && this.lastAction === 'calendar') {
                this._controller.date(extractDate(active), time);
            } else if (this.lastAction === 'input') {
                var value = this._$input[0].value;
                if (this._$el.hasClass('native') && value !== '') {
                    value = moment(value, nativeFormats[this._options.type]);
                }
                this._controller.date(value);
            }
            this.hidePopup();
        }
    };

    /**
     * @method DateView#handleDelete
     */
    dvp.handleDelete = function () {
        if (this.isVisible()) {
            this._triggerChange = true;
            this._controller.date(null);

            this.hidePopup();
        }
    };

    //TODO: look into how similar this is to ENTER
    /**
     * @method DateView#handleSpace
     *
     * @param {Event} event
     */
    dvp.handleSpace = function (event) {
        if (this.isVisible() && this.lastAction === 'calendar') {
            var active = this._$popup.find('.active');
            var time = this.spinner != null ? this.spinner.data() : null;

            event.preventDefault();

            if (!active.hasClass('inactive')) {
                this._controller.date(extractDate(active), time);
            }
        }
    };

    /**
     * @method DateView#handleTab
     *
     * @param {jQuery.Event} event
     */
    dvp.handleTab = function (event) {
        if (this.isVisible() && this._options.type === 'datetime') {
            event.stopPropagation();
            event.preventDefault();
            this.hideCal = false;

            //Navgiate to lower views (time-spinner)
            this.calendar.find('input').focus();
        }
    };

    /**
     * @method DateView#move
     *
     * @param {Object<String, Number>} changes
     */
    dvp.move = function (changes) {
        if (!_.isNull(this._$popup) && this.isVisible()) {
            var active = this._$popup.find('.active');
            var date = extractDate(active);
            var newActive;

            newActive = active;

            _.each(changes, function (changeBy, field) {
                date.add(changeBy, field);
            });

            this.lastAction = 'calendar';

            newActive = this._$popup.find('[data-date=' + date.format('YYYY-MM-DD') + ']');

            if (newActive.hasClass('next-month') || newActive.hasClass('prev-month')) {
                date = extractDate(newActive);
                this.showDate(date, date);
            } else if (newActive.length === 0 && active.length === 1) {
                this.showDate(date, date);
            } else if (this._controller.isInBounds(extractDate(newActive))) {
                this.active(newActive);
            }
        }
    };

    /**
     * @method DateView#jumpTo
     *
     * @param {String} clz
     */
    dvp.jumpTo = function (clz) {
        if (!_.isNull(this._$popup) && this.isVisible()) {
            if (clz === 'first-day') {
                this.active(this._$popup.find('.current-month').first());
            } else if (clz === 'last-day') {
                this.active(this._$popup.find('.current-month').last());
            }
        }
    };

    /**
     * Constructs the view specific elements.
     *
     * @method DateView#buildElement
     * @param {Object} options
     * @param {Object} el
     */
    dvp.buildElement = function (options, el) {
        var $element;

        if (_.isNull(el)) {
            //Build element from scratch
            $element = $(datePickerTemplate)
                .append(buttonTemplate({
                    locale: options.locale
                }));
        } else if (el.jquery) {
            $element = el.attr('spellcheck', false)
                .wrap(wrapTemplate)
                .parent()
                .append(buttonTemplate({
                    locale: options.locale
                }));
        }

        $element.find('input')
            .attr('type', 'text')
            .attr('required', options.required)
            .attr('aria-required', options.required)
            .attr('aria-haspopup', true)
            .attr('aria-expanded', false)
            .attr('aria-activedescendant', $element.find('input').attr('id'));

        return $element;
    };

    /**
     * @method DateView#buildNativeElement
     *
     * @param {Object} options
     * @param {jQuery} el
     *
     * @returns {jQuery} $element
     */
    dvp.buildNativeElement = function (options, el) {
        var $element;

        if (_.isNull(el)) {
            $element = $(nativeElement({
                type: options.type
            }))
            .wrap(wrapTemplate)
                .parent();
        } else if (el.jquery) {
            $element = el
                .attr('type', options.type)
                .wrap(wrapTemplate)
                .parent();
        }

        if (!_.isNull(options.max)) {
            $element.find('input').attr('max', moment(options.max).format(nativeFormats[options.type]));
        }

        if (!_.isNull(options.min)) {
            $element.find('input').attr('min', moment(options.min).format(nativeFormats[options.type]));
        }

        return $element;
    };

    /**
     * @method DateView#markVisited
     */
    dvp.markVisited = function () {
        this._visited = true;
    };

    /**
     * @method DateView#error
     *
     * @param {moment|String|null} attempted
     */
    dvp.error = function (attempted) {
        var currentValue = this._inputEl.value;
        var newValue;
        this._$el.add(this._$el.closest('.control-group')).toggleClass('error', !this._controller.isValid() && this._visited);

        if (this._$el.hasClass('native') && moment.isMoment(attempted)) {
            this._inputEl.value = attempted.format(nativeFormats[this._options.type]);
        } else if (moment.isMoment(attempted)) {
            newValue = this._options.formatter.format(attempted);
        } else if (_.isString(attempted)) {
            newValue = attempted;
        } else if (attempted == null) {
            newValue = '';
        }

        if (newValue != null && newValue !== currentValue) {
            this._inputEl.value = newValue;
        }

    };

    /**
     * Toggles whether the calendar is visible.
     *
     * @method DateView#toggleCalendar
     */
    dvp.toggleCalendar = function () {
        if (this.isVisible()) {
            this.hidePopup();
        } else {
            this.show(this._model.date());
        }
    };

    //TODO: should these happen in the view or should the model/controller
    //fire some events saying that the buttons should be disabled.
    /**
     * Disables the next year and month as well as the previous year and month
     * when the date changes.
     *
     * @method DateView#toggleNextPrevButtons
     * @param {Date} date The date to determine whether the next/prev buttons should
     *    be disabled.
     */
    dvp.toggleNextPrevButtons = function (date) {
        var nextMonth = moment(date).add(1, 'month').startOf('month');
        var nextYear = moment(date).add(1, 'year').startOf('month');
        var prevMonth = moment(date).subtract(1, 'month').startOf('month');
        var prevYear = moment(date).subtract(1, 'year').startOf('month');
        var max;
        var min;

        if (!_.isNull(this._options.max)) {
            max = moment(this._options.max);

            this.calendar.find('.next-buttons button:eq(0)').prop('disabled', nextMonth.isAfter(max));
            this.calendar.find('.next-buttons button:eq(1)').prop('disabled', nextYear.isAfter(max));
        } else {
            this.calendar.find('.next-buttons button:eq(0)').prop('disabled', false);
            this.calendar.find('.next-buttons button:eq(1)').prop('disabled', false);
        }

        if (!_.isNull(this._options.min)) {
            min = moment(this._options.min);
            min.startOf('month');

            this.calendar.find('.prev-buttons button:eq(1)').prop('disabled', prevMonth.isBefore(min));
            this.calendar.find('.prev-buttons button:eq(0)').prop('disabled', prevYear.isBefore(min));
        } else {
            this.calendar.find('.prev-buttons button:eq(1)').prop('disabled', false);
            this.calendar.find('.prev-buttons button:eq(0)').prop('disabled', false);
        }

        this.calendar.find('.btn.today').prop('disabled', !this._controller.isInBounds(moment().startOf('day')));
    };

    /**
     * The handler for the select event that is fired by the model.
     *
     * @method DateView#select
     * @param {Date} date The date that is selected.
     */
    dvp.select = function (date) {
        this.changeValue(date);
    };

    /**
     * @method DateView#hidePopup
     * @override
     */
    dvp.hidePopup = function () {
        popupMixin.hidePopup.apply(this, arguments);

        if (this._$backdrop) {
            this._$backdrop.removeClass('show');
        }
        this.hideCal = true;
    };

    /**
     * Handles updating the input value and then triggering a change.
     *
     * @method DateView#changeValue
     * @param {Date} date The date to change the input to and fire a change event.
     */
    dvp.changeValue = function (date) {
        var currentValue = this._inputEl.value;

        this._$el.add(this._$el.closest('.control-group')).removeClass('error');

        this.updateCalendarAndInput(date);

        if (this._triggerChange && currentValue !== this._inputEl.value) {
            this._$input.trigger('change', {
                update: false
            });
            this._triggerChange = false;
        }
    };

    /**
     * @method DateView#updateCalendarAndInput
     *
     * @param {moment} date
     */
    dvp.updateCalendarAndInput = function (date) {

        if (_.isNull(date)) {
            this._inputEl.value = '';
        } else if (this._$el.hasClass('native')) {
            this._inputEl.value = date.format(nativeFormats[this._options.type]);
        } else {

            if (!_.isNull(this.calendar) && this.isVisible()) {
                this.showDate(date, date);
                this.calendar.find('.selected').removeClass('selected');
                this.calendar.find('[data-date=' + date.clone().startOf('day').format('YYYY-MM-DD') + ']').addClass('selected');
            }

            if (this.spinner != null && this._options.type === 'datetime') {
                this.spinner.value(date || this._model.date());
            }

            if (moment.isMoment(date)) {
                this._inputEl.value = this._options.formatter.format(date);
            } else if (_.isString(date)) {
                this._inputEl.value = date;
            }

            //Otherwise - it should never have gotten here
        }
    };

    /**
     * Initializes the widget after it is on the DOM since a widget can
     * be created before being on the DOM.
     *
     * Handles all fallback logic to the native HTML5 element.
     *
     * @method DateView#init
     */
    dvp.init = function () {
        //TODO: Move this to constructor - since the model is already set and created when passed into the view
        this.changeValue(this._model.date());

        if (!this._model.isValid()) {
            this.error(this._model.date(), this._model.error().data);
        }

        //TODO: find a way to set the ID before building the element?
        this._$el.find('.instructions').attr('id', this.id() + '-instructions');
        this._$input.attr('aria-describedby', ((this._$input.attr('aria-describedby') || '') + ' ' + this.id() + '-instructions').trim());
    };

    /**
     * Handles rendering each cell as well as the title.
     *
     * @method DateView#showDate
     * @param {moment} date The date to build the calendar for.
     * @param {moment} activeDate
     */
    dvp.showDate = function (date, activeDate) {
        var startMonth;
        var startDate;
        var today = moment().startOf('day');
        var selDate = this._model.date();
        var max = moment(this._options.max);
        var min = moment(this._options.min);
        var title;

        if (date == null || (!moment.isMoment(date) && !this._model.isValid())) {
            date = moment();
        }

        if (!this._model.selected()) {
            selDate = null;
        }

        //Only show valid dates - if the date given is out of bounds
        //then go to the bound it exceeds
        if (this._options.min != null && date.isBefore(min)) {
            activeDate = min;
            date = min;
        }

        if (this._options.max != null && date.isAfter(max)) {
            date = max;
            activeDate = max;
        }

        activeDate = activeDate || selDate || today;

        date = moment(date).startOf('month');
        startMonth = date.month();
        startDate = date.clone();

        if (_.isFunction(this._options.title)) {
            title = this._options.title(date);
        } else {
            title = startDate.format('MMMM YYYY');
        }
        this.calendar.find('.title').text(title).prop('title', title);

        //Might move to another month so cant be chained until after startMonth and year are set.
        date.startOf('week');

        this.calendar.find('tbody td').each(function (i, td) {
            var $el = $(td)
            .removeClass('selected active')
            .text(date.date())
            .attr({
                'data-date': date.format('YYYY-MM-DD'),
                'id': 'calendar-' + this._id + '-' + i,
                'aria-label': date.date(),
                'aria-selected': false
            });

            if (this._model.selected() && dateHelper.dateComparator(date, selDate) === 0) {
                $el.addClass('selected');
                $el.attr('aria-selected', true);
            }

            if (dateHelper.dateComparator(date, activeDate) === 0) {
                this.active($el);
            }

            $el.toggleClass('selected-month', this._model.selected() && date.month() === selDate.month() && date.year() === selDate.year());

            $el.toggleClass('today', this._options.type !== 'month' && dateHelper.dateComparator(date, today) === 0);

            $el.toggleClass('inactive',
                    ((!_.isNull(max) && date.isAfter(max)) ||
                     (!_.isNull(min) && date.isBefore(min)) ||
                     (!_.isUndefined(this._options.unselectable) && dateHelper.contains(this._options.unselectable, date))));

            $el.toggleClass('current-month', date.month() === startMonth);

            $el.toggleClass('prev-month', date.month() === math.mod(startMonth - 1, 12));

            $el.toggleClass('next-month', date.month() === math.mod(startMonth + 1, 12));

            date.add(1, 'day');
        }.bind(this));

        if (this._$popup.find('.active').length === 0) {
            //No active was added - move to day that was active of previous calendar
            this.active(this._$popup.find('[data-date=' + startDate.date(activeDate.date()).format('YYYY-MM-DD') + ']'));
        }

        this.toggleNextPrevButtons(startDate);
    };

    /**
     * @method DateView#active
     *
     * @param {jQuery} activeElement
     */
    dvp.active = function (activeElement) {
        this.calendar.find('.active').removeClass('active');

        this._$input.attr('aria-activedescendant', activeElement.attr('id'));
        activeElement.addClass('active');
    };

    /**
     * Adds the element to the DOM after it is built.
     *
     * @method DateView#buildCalendar
     */
    dvp.buildCalendar = function () {
        var that = this;
        var currentDate = this._model.date();
        var $calendar;

        /**
         * @method DateView#buildCalendar~addHover
         *
         * @param {String} clazz
         */
        function addHover(clazz) {
            $calendar.on('mouseenter', clazz, function () {
                $calendar.find(clazz).addClass('hover');
            }).on('mouseleave', clazz, function () {
                $calendar.find(clazz).removeClass('hover');
            });
        }

        if (_.isNull(currentDate) || (!moment.isMoment(currentDate) && !this._model.isValid())) {
            currentDate = moment();
        }

        $calendar = $(calendarTemplate({
            id: 'oui-calendar-label-' + this._id,
            locale: this._options.locale,
            type: this._options.type
        }));

        var $span = $('<span class="title" style="position:absolute; top:0; left:0;font-weight: bold; visibility: hidden;">');
        $('body').append($span);
        var width = _.chain(_.range(12))
            .map(function (month) {
                var title;
                var date = moment();
                date.month(month);

                if (_.isFunction(that._options.title)) {
                    title = that._options.title(date);
                } else {
                    title = date.format('MMMM YYYY');
                }

                $span.text(title);
                return $span.width();
            })
            .max()
            .value();
        $span.remove();

        $calendar.find('.title').width(width);

        var day = moment().startOf('week');
        _.each($calendar.find('th'), function (header) {
            $(header)
            .text(day.format('ddd'))
            .attr('aria-label', day.format('dddd'))
            .attr('title', day.format('dddd'));

            day.add(1, 'day');
        });

        //Position and show the calendar
        if (that._options.type === 'datetime') {
            //TODO: can this be moved out of here and into the DateField and inject it into the view
            //if it is a datetime date field.
            that.spinner = new NumberField(that._options.number, $calendar.find('input'), this._model.date())
                .init();

            $calendar.find('input')
                .on('keydown', function (event) {
                    if (event.which === keyCode.TAB) {
                        event.preventDefault();
                        event.stopPropagation();
                        that._$input.focus();
                        that.hidePopup();
                    } else if (event.which === keyCode.ENTER) {
                        event.preventDefault();
                        that._$input.focus();
                        that.handleEnter();
                    }
                }).on('oui-number-change', function (event, widget, eventType, time) {
                    that._triggerChange = true;
                    var date = extractDate($calendar.find('.selected'));

                    if ($calendar.find('.selected').length === 0) {
                        date = extractDate($calendar.find('.active'));
                    }

                    date.startOf('day');
                    date.add(time.hours(), 'hour');
                    date.add(time.minutes(), 'minute');
                    date.add(time.seconds(), 'seconds');
                    that.updateCalendarAndInput(date);
                });
        }

        $calendar.attr('id', 'oui-calendar-' + this._id);
        that._$input.attr('aria-owns', $calendar.attr('id'));

        if (that._options.type === 'month') {
            $calendar.addClass('month');
        }

        $calendar.hover(function () {
            that.hideCal = false;
        }, function () {
            that.hideCal = true;
        });

        addHover('.current-month');
        addHover('.prev-month');
        addHover('.next-month');

        $calendar.on('click', '.button-controls button', function () {
            var field = $(this).attr('data-field');
            var value = $(this).attr('data-value');

            var changes = {};
            changes[field] = value === 'next' ? 1 : -1;
            that.move(changes);
        }).on('click', '.today-cancel .today', function () {
            var date = moment();
            var time;

            if (that._options.type === 'datetime') {
                time = that.spinner.data() || moment();
            }

            that._triggerChange = true;
            that._controller.date(date, time || null);
            that.hidePopup();
        }).on('click', '.cancel', function () {
            that.cancel();
        }).on('click', '.ok', function () {
            if ($calendar.find('.selected').length > 0) {
                that._controller.date(extractDate($calendar.find('.selected')), that.spinner.value());
            }
            that.hidePopup();
        }).on('blur', 'input', function () {
            if (that.hideCal) {
                var date = $calendar.find('.selected');

                if (date.length === 0) {
                    date = $calendar.find('.active');
                }

                if (date.length !== 0) {
                    that._controller.date(extractDate(date), that.spinner.value());
                }
                that.hidePopup();
            }
        }).on('showing.popup', function () {
            if (that._model.isValid() && that._model.selected()) {
                currentDate = that._model.date();
                that._setCurrentDate(currentDate);
            } else {
                currentDate = moment().startOf('day');
                that._setCurrentDate(null);
            }
        }).on('shown.popup', function () {
            that._$input.attr('aria-expanded', true);
            that._$input.attr('aria-activedescendant', that._$popup.find('.active').attr('id'));
            that.lastAction = 'calendar';
        }).on('mousedown touchstart', function (event) {
            //Prevent default on any mousedown in the calendar except for inputs
            if ($(event.target).parents('.orcl-number-field').length === 0) {
                if (event.type !== 'touchstart') {
                    event.preventDefault();
                }

                //Keep focus on the date picker input if it isnt focused
                if (!$(document.activeElement).is(that._$input)) {
                    that._$input.focus();
                }
            }
        }).on('hidden.popup', function () {
            that._$input.attr('aria-activedescendant', that._$input.attr('id'));
            that._$input.attr('aria-expanded', false);
            that.lastAction = null;
        }).on('click touchend', 'td:not(.inactive)', function (ev) {
            var time = null;
            var date;
            ev.preventDefault();

            if (!(that._options.type === 'month' && $(this).hasClass('selected-month'))) {
                date = extractDate($(this));
                if (that._options.type === 'datetime') {
                    time = that.spinner.data();

                    if (moment.isMoment(time)) {
                        date.startOf('day');
                        date.add(time.hours(), 'hour');
                        date.add(time.minutes(), 'minute');
                        date.add(time.seconds(), 'seconds');
                    }
                }

                that._triggerChange = true;

                if (that._options.type !== 'datetime') {
                    that._controller.date(date, time || null);
                    that.hidePopup();
                } else {
                    that.updateCalendarAndInput(date);
                }
            }

        }).on('dblclick', 'td:not(.inactive)', function (ev) {
            ev.preventDefault();
            that._controller.date(extractDate($calendar.find('.selected')), that.spinner.value());
            that.hidePopup();
        });

        return $calendar;
    };

    /**
     * @method DateView#_setCurrentDate
     * @private
     *
     * @param {moment} date
     */
    dvp._setCurrentDate = function (date) {
        this._calendarStartDate = date;
    };

    /**
     * @method DateView#cancel
     */
    dvp.cancel = function () {
        this._controller.date(this._calendarStartDate, { silent: true });
        this.hidePopup();
    };

    return DateView;
});

define('oui/widgets/date/dateModel',[
    'underscore',
    'oui/base/model',
    'oui/utils/dateHelper',
    'moment'
], function (_, Model, dateHelper, moment) {
      'use strict';

      /**
       * Fired when the date field is not required, the current date is not null, and the new date is null.
       *
       * @event DateModel#deselect
       * @param {null} - The new date that the date field represents.
       */
      /**
       * Fired when the date field is required, the current date is not null, and the new date is null.
       *
       * @event DateModel#required-error
       * @deprecated
       * @param {moment|null} - The current date of the date field
       * @param {String} - The data that was attempted to be set on the date field.
       */
      /**
       * Fired when the new date is not a valid moment object.
       *
       * @event DateModel#invalid-error
       * @param {moment|Null} - The current date of the date field
       * @deprecated
       * @param {String} - The data that was attempted to be set on the date field.
       */
      /**
       * Fired when the new date exists in the unselectable dates array.
       *
       * @event DateModel#unselectable-error
       * @deprecated
       * @param {moment|Null} - The current date of the date field
       * @param {moment} - The date that resulted in the error
       */
      /**
       * Fired when the new date exists outside of the defined min and max bounds.
       *
       * @event DateModel#bounds-error
       * @deprecated
       * @param {moment|Null} - The current date of the date field
       * @param {moment} - The date that resulted in the error
       */
      /**
       * Fired when a valid date is supplied and it differs from the current date.
       *
       * @event DateModel#date-change
       * @deprecated
       * @param {moment|Null} - The current date of the date field
       */

      var dmp;

      /**
       * The class that holds the data used in the DateView.
       *
       * The date will never become invalid. It will always be the last valid value.
       *
       * @class DateModel
       * @extends Model
       * @private
       * @param {Date|null} initDate The date to start the control at.
       * @param {DateField#Config} options The widget options.
       */
      function DateModel(initDate, options) {
          //Super call
          Model.call(this, initDate, options);

      }

      DateModel.prototype = Object.create(Model.prototype);
      dmp = DateModel.prototype;

      dmp.constructor = DateModel;

      /**
       * @method DateModel#name
       *
       * @returns {String}
       */
      dmp.name = function () {
          return 'date';
      };

      /**
       * @method DateModel#isSame
       *
       * @param {moment} newDate
       *
       * @returns {Boolean}
       */
      dmp.isSame = function (newDate) {
          return Model.prototype.isSame.call(this, newDate) ||
              (moment.isMoment(this._data) && this._data.isSame(newDate)) || _.isEqual(this._data, newDate);
      };

      /**
       * @method DateModel#init
       *
       * @param {moment} initDate
       *
       * @returns {moment}
       */
      dmp.init = function (initDate) {
          //TODO: set up invalid data if given invalid data
          if (moment.isMoment(initDate) && !initDate.isValid()) {
              return initDate._i;
          }

          return initDate;
      };

      /**
       * Utility function to return the bounds validity of the supplied date.
       *
       * @method DateModel#isInBounds
       * @param {moment} date
       * @returns {Boolean} - Whether the moment is within the bounds of the date model
       */
      dmp.isInBounds = function (date) {
          var max = moment(this._options.max);
          var min = moment(this._options.min);

          return (this._options.max == null || !max.isValid() || (date.isSame(max) || date.isBefore(max))) &&
            (this._options.min == null || !min.isValid() || (date.isSame(min) || date.isAfter(min)));
      };

      /**
       * @method DateModel#validate
       *
       * @param {moment} date
       *
       * @returns {Promise}
       */
      dmp.validate = function (date) {

          if (date == null && !this._options.required) {
              return {
                  valid: true,
                  type: 'deselect',
                  data: null
              };
          } else if (date == null) {
              return {
                  valid: false,
                  type: 'required',
                  data: null
              };
          } else if (_.isString(date) || _.isNaN(date) || (moment.isMoment(date) && !date.isValid())) {
              return {
                  valid: false,
                  type: 'invalid',
                  data: _.has(date, '_i') ? date._i : date
              };
          } else if (!_.isUndefined(this._options.unselectable) &&
            dateHelper.contains(this._options.unselectable, date)) {
              return {
                  valid: false,
                  type: 'unselectable',
                  data: date
              };
          } else if (!this.isInBounds(date)) {
              return {
                  valid: false,
                  type: 'bounds',
                  data: date
              };
          }

          return {
              valid: true
          };
      };

      /**
       * @method DateModel#getData
       *
       * @returns {moment}
       */
      dmp.getData = function () {
          //Return a clone so caller can change without affecting model
          if (this.isValid()) {
              return (this._data == null || !this._data.isValid()) ? null : this._data.clone();
          } else {
              return this._validationResult.data;
          }
      };

      /**
       * @method DateModel#successHandler
       *
       * @param {Subscribable~EventOptions} eventParams
       * @param {moment} date
       */
      dmp.successHandler = function (eventParams, date) {
          if (date == null) {
              this._validationResult = {
                  valid: true
              };
              this._data = date;
              this.trigger(eventParams, 'deselect', null);
          } else {
              Model.prototype.successHandler.call(this, eventParams, date);
          }
      };

      /**
       * Getter/Setter for the date property.
       *
       * @method DateModel#date
       * @param {moment} date The new date moment.
       * @param {Subscribable~EventOptions} eventParams The parameters to pass to trigger. Can force an event to be silent.
       * @fires DateModel#deselect
       * @fires DateModel#required-error
       * @fires DateModel#invalid-error
       * @fires DateModel#unselectable-error
       * @fires DateModel#bounds-error
       * @fires DateModel#date-change
       * @returns {moment} The current date of the model.
       */
      dmp.date = dmp.data;

      /**
       * Getter for the current selected status of the model.
       *
       * Returns true if the model is valid and the current date is not null.
       *
       * @method DateModel#selected
       * @returns {Boolean} - Whether the model is currently selected.
       */
      dmp.selected = dmp.isSet;

      return DateModel;
  });

define('oui/widgets/date/dateController',[
    'underscore',
    'moment'
], function (_, moment) {
    'use strict';

    /**
     * @class DateController
     * @private
     *
     * @param {DateModel} model The model that the controller uses to set and get data.
     * @param {DateField#Config} options The widgets options object.
     */
    function DateController(model, options) {

        /**
         * The model for the controller.
         *
         * @private
         * @type {DateModel}
         * @memberof! DateController.prototype
         */
        this._model = model;
        /**
         * The widgets options.
         *
         * @type {Object}
         * @private
         * @memberof! DateController.prototype
         */
        this._options = options;
    }

    DateController.prototype = {
        constructor: DateController,
        /**
         * Getter/setter of the date in the model.
         *
         * @method DateController#date
         * @returns {Date} The model's date.
         * @param {Date|String|Number} dateMs The date to update to.
         * @param {Date|String|Number} [timeMs] The time object if the date field
         *   includes time.
         */
        date: function (dateMs, timeMs) {
            var date;
            var time;
            if (!arguments.length) {
                return this._model.date();
            }

            var eventParams;
            var newDate;

            if (timeMs != null && timeMs.silent) {
                eventParams = timeMs;
                timeMs = null;
            }

            if (moment.isMoment(dateMs)) {

                //Controller should handle date and time addition and not the model.
                date = moment(dateMs);

                if (date != null) {

                    //Since it is not possible to select the time from the picker - force it to the
                    //beginning of the date - regardless if it was given
                    if (timeMs != null && this._options.type === 'datetime') {
                        date.startOf('day');
                        time = moment(timeMs);
                        date.add(time.hours(), 'hour');
                        date.add(time.minutes(), 'minute');
                        date.add(time.seconds(), 'seconds');
                    } else if (this._options.type !== 'datetime') {
                        date.startOf('day');
                    }
                }

                newDate = date;
            } else if (_.isString(dateMs) && this._options.formatter.validate(dateMs, { forgiving: true })) {
                newDate = this._options.formatter.unformat(dateMs, { forgiving: true });
            } else {
                newDate = dateMs;
            }

            this._model.date(newDate, eventParams);
        },
        /**
         * @method DateController#isValid
         */
        isValid: function () {
            return this._model.isValid();
        },
        /**
         * @method DateController#error
         */
        error: function () {
            return this._model.error();
        },
        /**
         * @method DateController#isInBounds
         *
         * @param {moment} date
         *
         * @returns {Boolean}
         */
        isInBounds: function (date) {
            return this._model.isInBounds(date);
        }
    };

    return DateController;
});

define('oui/formatters/date',[
    'moment'
], function (moment) {
    'use strict';

    /**
     * @module formatters/date
     */

    return {
        /**
         * @method module:formatters/date.format
         *
         * @param {moment} date
         *
         * @returns {String}
         */
        format: function (date) {
            return date.format('YYYY-MM-DD');
        },
        /**
         * @method module:formatters/date.unformat
         *
         * @param {String} strDate
         *
         * @returns {moment}
         */
        unformat: function (strDate) {
            return moment(strDate, 'YYYY-MM-DD');
        },
        /**
         * @method module:formatters/date.validate
         *
         * @param {String} strDate
         *
         * @returns {Boolean}
         */
        validate: function (strDate) {
            return moment(strDate, 'YYYY-MM-DD', true).isValid();
        }
    };

});

define('oui/widgets/dateField',['require','underscore','oui/base/input','oui/widgets/date/dateView','oui/widgets/date/dateModel','oui/widgets/date/dateController','oui/widgets/numberField','oui/formatters/date'],function (require) {
      'use strict';

      var _ = require('underscore');
      var Input = require('oui/base/input');
      var DateView = require('oui/widgets/date/dateView');
      var DateModel = require('oui/widgets/date/dateModel');
      var DateController = require('oui/widgets/date/dateController');
      var NumberField = require('oui/widgets/numberField');
      var baseDateFormatter = require('oui/formatters/date');

      /**
       * Enum for valid type values in the DateField
       *
       * @typedef {Enum<String>} DateField#Types
       * @property {String} 'date'
       * @property {String} 'datetime'
       * @property {String} 'month''
       */

      /**
       * @callback DateField#formatFn
       * @param {moment} moment - The moment object to format
       * @returns {String}
       */

      /**
       * @callback DateField#unformatFn
       * @param {String} str - The string representation of a date
       * @returns {moment}
       */

      /**
       * @callback DateField#validateFn
       * @param {String} str - The string representation of a date
       * @returns {Boolean}
       */

      var __WIDGET_NAME__ = 'date';
      var dfp;

      /**
       * @typedef DateField#Config
       *
       * @property {Boolean} [openOnFocus=false] Whether or not the Calendar should open as soon as the input has received focus or not.
       * @property {Locale} [locale={}] The specific strings that are used by the widget and will use the defaults set from
       * the {@link module:oui.lang|lang}.
       * @property {moment|null} [min=null] The minimum value of the date picker
       * @property {moment|null} [max=null] The maximum value of the date picker.
       * @property {DateField#Types} [type='date'] The type of date picker to construct. Possible values are 'date', 'datetime', or 'month'.
       * @property {Boolean} [fallback=false] **Deprecated** Whether or not to force the fallback to native HTML5 element.
       * @property {Boolean} [required=false] Whether or not the date field is required.
       * @property {Boolean} [fluid=false] Whether or not the date field should take up the width of its parent container.
       * @property {String|jQuery} [appendTo='body'] The location to append the popup (calendar) to.
       * @property {NumberField#Config} [number] The options to pass to the number field. Used when the type of the field is 'datetime'.
       * @property {Boolean} [utc=false] **Unused** Whether or not the date field represents a date in UTC or local time.
       * @property {module:formatters} [formatter] The formatter to convert strings to moments and moments to strings.
       * @property {Array<Number|moment>} [unselectable=[]] The dates that are not selectable within the date field.
       */
      var __OVERRIDES__ = {
          number: {
              type: 'time',
              spinner: {
                  controls: true
              },
              fluid: true
          }
      };
      var __DEFAULTS__ = _.extend({
          showErrors: 'blur',
          openOnFocus: false,
          locale: {},
          min: null,
          max: null,
          type: 'date',
          fallback: false,
          required: false,
          fluid: false,
          appendTo: 'body',
          number: NumberField.prototype.defaults(),
          utc: false,
          formatter: baseDateFormatter,
          unselectable: []
      }, __OVERRIDES__);

      /**
       * The common date input element. Displays a UI element that can accept a
       * typed date as well as through a selection on the calendar popup.
       *
       * Each date field consists of an input and a calendar button wrapped by a div with class 'orcl-date-picker'. This
       * class is expected to be used as a selector and will be supported. Appending anything inside this div is not
       * supported. The styles only expect there to be an input and a button with hidden screen reader instructions. Any
       * additional labels should be placed outside the '.orcl-date-picker'.
       *
       * See {@link DateField#Config} for more information regarding options for the Date Picker.
       *
       * See {@link http://www.w3.org/TR/wai-aria-practices/#datepicker|ARIA Best Practices}
       * for information regarding keyboard shortcuts and general ARIA attributes and states.
       *
       *  ##### **Keyboard Shortcuts**
       *
       *  * <kbd>F2</kbd> - Opens the calendar
       *  * <kbd>Shift+F10</kbd> - Opens the calendar
       *  * <kbd>Esc</kbd> - Closes the calendar
       *  * <kbd>Enter</kbd> - When opened, selects the currently active date and closes the calendar.
       *  * <kbd>Space</kbd> - When opened, selects the currently active date.
       *  * <kbd>Tab</kbd> - When opened and type='datetime', shift between calendar and the spinner in the calendar.
       *  * <kbd>Up</kbd> - When opened, activates the date above the currently active date.
       *  * <kbd>Down</kbd> - When opened, activates the date below the currently active date.
       *  * <kbd>Left</kbd> - When opened, activates the date to the left of the currently active date.
       *  * <kbd>Right</kbd> - When opened, activate the date to the right of the currently active date.
       *  * <kbd>Page Up</kbd> - When opened, activate the date one month previous the currently active date.
       *  * <kbd>Page Down</kbd> - When opened, activate the date one month after the currently active date.
       *  * <kbd>Alt+Page Up</kbd> - When opened, activate the date one year previous to the currently active date.
       *  * <kbd>Alt+Page Down</kbd> - When opened, activate the date one year after the currently active date.
       *
       *
       *  ##### **Relevant Events**
       *
       *  * {@link DateField#deselect|deselect}
       *  * {@link DateField#required-error|required-error}
       *  * {@link DateField#invalid-error|invalid-error}
       *  * {@link DateField#unselectable-error|unselectable-error}
       *  * {@link DateField#bounds-error|bounds-error}
       *  * {@link DateField#date-change|date-change}
       *  * {@link DateField#change|change}
       *  * {@link DateField#error|error}
       *  * {@link DateField#pre-change|pre-change}
       *  * {@link DateField#identical|identical}
       *
       * @borrows DateModel#event:deselect as DateField#deselect
       * @borrows DateModel#event:required-error as DateField#required-error
       * @borrows DateModel#event:invalid-error as DateField#invalid-error
       * @borrows DateModel#event:unselectable-error as DateField#unselectable-error
       * @borrows DateModel#event:bounds-error as DateField#bounds-error
       * @borrows DateModel#event:date-change as DateField#date-change
       * @borrows Model#event:pre-change as DateField#pre-change
       * @borrows Model#event:change as DateField#change
       * @borrows Model#event:error as DateField#error
       * @borrows Model#event:identical as DateField#identical
       *
       * @example
       * <caption>Example construction (accepts all defaults - adds required options)</caption>
       *
       * var dateField = new oui.DateField({
       *    formatter : {
       *        format : function (moment) {
       *            return moment.format('LLL');
       *        },
       *        unformat : function (str) {
       *            return moment(str);
       *        },
       *        validate : function (str) {
       *            return moment(str).isValid();
       *        }
       *    }
       * });
       * $container.append(dateField.element());
       * dateField.init();
       *
       * @example
       * <caption>Example construction using existing date input</caption>
       *
       * var dateField = new oui.DateField({
       *    formatter : {
       *        format : function (moment) {
       *            return moment.format('LLL');
       *        },
       *        unformat : function (str) {
       *            return moment(str);
       *        },
       *        validate : function (str) {
       *            return moment(str).isValid();
       *        }
       *    }
       * }, $date);
       * dateField.init();
       *
       * @class DateField
       * @extends Input
       * @since 1.2.0
       *
       * @param {DateField#Config} [options] The options for the DateField. If not
       *  specified, all the defaults will be used.
       * @param {jQuery} [element] The JQuery element that this field should use.
       *  If not specified, DateField will construct its own element.
       * @param {moment|null} [initDate] The date to initialize the field with.
       */
      function DateField(options, element, initDate) {
          var model;

          //Super constructor
          Input.call(this, options, element, initDate);

          model = new DateModel(this._initData, this._options);
          /**
           * The controller object that handles the programmatic input from
           * the DateField.
           *
           * @memberof DateField
           * @name _controller
           * @instance
           * @private
           * @type {DateController}
           */
          this._controller = new DateController(model, this._options);
          /**
           * The view element.
           *
           * @memberof DateField
           * @instance
           * @name _view
           * @private
           * @type {DateView}
           */
          this._view = new DateView(model, this._controller, this._element || null, this._options, this._settings);

          Object.defineProperty(this, '_data', {
              get: this.date,
              set: this.date
          });

          this._setupEvents(model);
      }

      DateField.prototype = Object.create(Input.prototype);
      dfp = DateField.prototype;

      dfp.constructor = DateField;

      /**
       * Returns the name of the widget.
       *
       * @method DateField#name
       * @returns {String} - 'date'
       */
      dfp.name = function () {
          return __WIDGET_NAME__;
      };

      /**
       * Getter/Setter for the DateField's date. If no parameter is passed it
       * will be a getter otherwise it is a setter.
       *
       * @method DateField#date
       * @since 1.3.0
       * @param {moment|null} [date] The date to set for the DateField.
       * @fires DateModel#error
       * @fires DateModel#deselect
       * @fires DateModel#required-error
       * @fires DateModel#invalid-error
       * @fires DateModel#unselectable-error
       * @fires DateModel#bounds-error
       * @fires DateModel#date-change
       * @returns {moment|null} date The DateField's date when used as a getter.
       */
      dfp.date = function (date) {
          if (!arguments.length) {
              return this._controller.date();
          }
          this._view.markVisited();
          this._controller.date(date);
      };

      /**
       * Get the validity state of the date field.
       *
       * @method DateField#isValid
       * @returns {Boolean} - Whether the date field is currently in a valid state.
       */
      dfp.isValid = function () {
          return this._controller.isValid();
      };

      /**
       * Gets the last error of the date field.
       *
       * @method DateField#error
       * @returns {OUIError|null} The last error of the date field.
       */
      dfp.error = function () {
          return this._controller.error();
      };

      /**
       * Getter/Setter for the min property of the DateField.
       *
       * @method DateField#min
       * @since 1.3.0
       * @param {moment|null} [date] The date to set for the DateField's min.
       * @returns {Date} date The DateField's minimum date when used as a getter.
       */
      dfp.min = function (date) {
          if (!arguments.length) {
              return this._options.min;
          }
          this._options.min = date;
      };

      /**
       * Getter/Setter for the max property of the DateField.
       *
       * @method DateField#max
       * @since 1.3.0
       * @param {moment|null} [date] The date to set for the DateField's max.
       * @returns {moment} date The DateField's maximum date when used as a getter.
       */
      dfp.max = function (date) {
          if (!arguments.length) {
              return this._options.max;
          }
          this._options.max = date;
      };

      /**
       * Gets the default options for the date field.
       *
       * @method DateField#defaults
       * @private
       * @returns {Object} The default options for the widget
       */
      dfp.defaults = function () {
          return __DEFAULTS__;
      };

      /**
       * Getter/Setter for the value of the of the Date Field
       *
       * See {@link DateField#date} for information regarding the similar method.
       *
       * **Note**: does not trigger an event notifying of a change
       *
       * @method DateField#value
       * @param {moment|String|null} [val] The current value of the date field.
       * @returns {moment|String|null} - the current value of the date field.
       */
      dfp.value = function (val) {
          if (!arguments.length) {
              return this.data();
          }

          this._controller.date(val, {
              silent: true
          });
      };

      /**
       * Getter/Setter for the required option of the date field
       *
       * Rerenders the view when reset.
       *
       * @method DateField#required
       * @param {Boolean} [req] - Whether or not the date field is required.
       * @returns {Boolean} - The current value of the required option.
       */
      dfp.required = function (req) {
          if (!arguments.length) {
              return this._options.required;
          }

          this._options.required = req;
          this.value(this._controller.date());
      };

      /**
       * Getter/Setter for the formatter option of the date field
       *
       * Rerenders the view when set.
       *
       * @method DateField#formatter
       * @param {Object} [formatter] - The new formatter to use within the date field.
       * @returns {Object} - The formatter currently in use by the date field.
       */
      dfp.formatter = function (formatter) {
          if (!arguments.length) {
              return this._options.formatter;
          }

          this._options.formatter = formatter;
          this._view.changeValue(this._controller.date());
          this._view.trigger('reformat');
      };

      return DateField;
  }
);

define('oui/utils/prop',[
    'underscore'
], function (_) {
    'use strict';

    return function (property, elements, state) {
        _.each(elements, _.partial(_.set, _, property, state));
    };
});


define('template-14', ['underscore'], function (_) { return function(data) {
var __t, __p = '';
__p += '<div class="orcl-widget orcl-picker"> <div class="search ' +
((__t = ( data.options.search ? '' : 'hidden' )) == null ? '' : __t) +
'"> </div> <div class="tree-content"> </div></div>';
return __p
};});


define('template-15', ['underscore'], function (_) { return function(data) {
var __t, __p = '';
__p += '<button type="button" class="btn btn-picker" title="' +
((__t = ( data.locale.picker.select  )) == null ? '' : __t) +
'" aria-label="' +
((__t = ( data.locale.picker.select )) == null ? '' : __t) +
'" > <i class="pgbu-icon-small pgbu-icon-picker-dots"></i></button>';
return __p
};});


define('template-16', ['underscore'], function (_) { return function(data) {
var __t, __p = '';
__p += '<div class="orcl-widget orcl-picker-input"> <input type="text"> <button type="button" tabindex="-1" class="btn btn-picker" title="' +
((__t = ( data.locale.picker.select )) == null ? '' : __t) +
'"> <i class="pgbu-icon-small pgbu-icon-picker-dots"></i> </button></div>';
return __p
};});


define('template-17',[],function () { return '<div class="orcl-widget orcl-picker-input">\n</div>\n';});


define('template-18', ['underscore'], function (_) { return function(data) {
var __t, __p = '';
__p += '<label id="' +
((__t = ( data.id )) == null ? '' : __t) +
'" for="' +
((__t = ( data.treeId )) == null ? '' : __t) +
'" class="control-label">' +
((__t = ( data.locale.title  )) == null ? '' : __t) +
'</label>';
return __p
};});


define('template-19', ['underscore'], function (_) { return function(data) {
var __t, __p = '', __j = Array.prototype.join;
function print() { __p += __j.call(arguments, '') }
__p += '<div class="modal pgbu-modal hide fade orcl-picker-modal" data-backdrop="static" tabindex="-1" role="dialog" aria-hidden="true" aria-labelledby="' +
((__t = ( data.modalId )) == null ? '' : __t) +
'"> <div class="modal-header"> <a title="' +
((__t = ( data.locale.close )) == null ? '' : __t) +
'" class="modal-close" href="#" data-dismiss="modal"><i class="pgbu-icon pgbu-icon-dialog-close"></i></a> <h3 id="' +
((__t = ( data.modalId )) == null ? '' : __t) +
'" class="modal-title">' +
((__t = ( data.locale.title  )) == null ? '' : __t) +
'</h3> </div> <div class="modal-body"></div> <div class="modal-footer"> ';
 if(!data.required) { ;
__p += ' <button type="button" title="' +
((__t = ( data.locale.remove )) == null ? '' : __t) +
'" class="btn btn-blue-dark modal-clear pull-left" data-dismiss="modal" disabled="disabled">' +
((__t = ( data.locale.remove  )) == null ? '' : __t) +
'</button> ';
 } ;
__p += ' <button type="button" title="' +
((__t = ( data.locale.cancel )) == null ? '' : __t) +
'" class="btn btn-blue-dark modal-cancel" data-dismiss="modal">' +
((__t = ( data.locale.cancel  )) == null ? '' : __t) +
'</button> <button type="button" title="' +
((__t = ( data.locale.ok )) == null ? '' : __t) +
'" class="btn btn-action modal-accept disabled" data-dismiss="modal" disabled="disabled">' +
((__t = ( data.locale.ok  )) == null ? '' : __t) +
'</button> </div></div>';
return __p
};});

define('oui/utils/focus',[],function () {
    'use strict';

    /**
     * @module focus
     */
    return {
        /**
         * @method findFocusable
         *
         * @param {jQuery} $el
         *
         * @returns {jQuery}
         */
        findFocusable: function ($el) {
            return $el
                .find('a[href], button:not([disabled]), input:not([disabled]), textarea:not([disabled]), select:not([disabled]), [tabindex]')
                .filter(':visible:not([tabindex=-1]), input[type="file"]');
        }
    };
});

define('oui/widgets/picker/pickerView',['require','underscore','jquery','oui/base/view','oui/utils/prop','oui/utils/flexEqual','template-14','template-15','template-16','template-17','template-18','template-19','oui/utils/keyCode','oui/utils/focus','bootstrap'],function (require) { //jshint ignore:line
    'use strict';

    var _ = require('underscore');
    var $ = require('jquery');
    var View = require('oui/base/view');
    var prop = require('oui/utils/prop');
    var flexEqual = require('oui/utils/flexEqual');
    var pickerTemplate = require('template-14');
    var pickerButtonTemplate = require('template-15');
    var pickerInputTemplate = require('template-16');
    var pickerWrapInputTemplate = require('template-17');
    var titleTemplate = require('template-18');
    var pickerModalTemplate = require('template-19');
    var keyCode = require('oui/utils/keyCode');
    var focusUtils = require('oui/utils/focus');
    require('bootstrap');

    /**
     * @method PickerView~busy
     * @private
     *
     * @param {jQuery} $button
     */
    function busy($button) {
        $button.prop('disabled', true).addClass('disabled');

        var $icon = $button.children('i');
        if ($icon.length > 0 && $icon.hasClass('pgbu-icon-picker-dots')) {
            $icon.removeClass('pgbu-icon-picker-dots').addClass('pgbu-icon-loading');
        } else if ($icon.length === 0) {
            //Strictly a text button?
            $button.attr('data-orig-text', $button.text());
            $button.text('');
            $button.append('<div class="pgbu-icon throbber"><div class="dot"></div><div class="dot"></div><div class="dot"></div></div>');
        }
    }

    /**
     * @method PickerView~shake
     * @private
     *
     * @param {jQuery} $element
     */
    function shake($element) {
        $element.on('animationend', function () {
            $element.removeClass('shake');
        });

        $element.addClass('shake');
    }

    /**
     * @method PickerView~restore
     * @private
     *
     * @param {jQuery} $button
     */
    function restore($button) {
        $button.prop('disabled', false).removeClass('disabled');
    }

    /**
     * @method PickerView~available
     * @private
     *
     * @param {jQuery} $button
     */
    function available($button) {
        var $icon = $button.children('i');
        if ($icon.length > 0 && $icon.hasClass('pgbu-icon-loading')) {
            $icon.removeClass('pgbu-icon-loading').addClass('pgbu-icon-picker-dots');
        } else if ($icon.length === 0) {
            //Strictly a text button?
            $button.empty();
            $button.text($button.attr('data-orig-text'));
            $button.removeAttr('data-orig-text');
        }
    }

    /**
     * @method PickerView~handle
     * @private
     *
     * @param {Boolean} busy
     * @param {Boolean?} value
     *
     * @returns {Promise}
     */
    function handle(busy, value) {
        if (value == null || value) {
            return $.Deferred().resolve(value);
        } else {
            return $.Deferred().reject(value);
        }

    }

    /**
     * Fired when the picker is throwing away any changes made during the last time it was opened up.
     *
     * @event PickerView#cancel
     */
    /**
     * Fired when the picker is committing the currently selected nodes.
     *
     * @event PickerView#accept
     * @param {Array<Object>} nodes The accepted nodes of the picker.
     */
    /**
     * Fired when the picker modal is about to be shown.
     *
     * @event PickerView#pre-show
     * @param {Array<Object>} nodes The currently selected nodes on the picker.
     */
    /**
     * Fired when the picker is about to accept the currently selected nodes.
     *
     * @event PickerView#pre-accept
     * @param {Array<Object>} nodes The currently selected nodes on the picker.
     */

    var pvp;

    /**
     * @class PickerView
     * @private
     * @extends View
     * @since 1.3.0
     *
     */
    function PickerView(options, element, treePanel, typeAhead, settings) {
        var that = this;
        this._model = {
            data: [],
            init: false
        };

        this._typeAhead = typeAhead;
        this._treePanel = treePanel;

        View.call(this, options, element, null, settings);

        //When the typeahead issues the search event forwards that to the
        //treePanel.
        typeAhead.on('search invalidate clear', function (query, scope) {
            var scopeCriteria = scope;
            if (_.isUndefined(query) || _.isEmpty(query)) {
                query = null;
            }

            if (query == null && !that._options.filterOnScope) {
                scopeCriteria = null;
            }

            treePanel.filter(query, null, scopeCriteria);
            that._filter(query);
        });

        typeAhead.on('scope-change', function (query, scope) {
            if (that._options.filterOnScope) {
                if (_.isUndefined(query) || _.isEmpty(query)) {
                    query = null;
                }

                treePanel.filter(query, null, scope);
                that._filter(query);
            }
        });

        treePanel.on('filter', function (results) {
            if (_.isNull(results)) {
                typeAhead.invalidate();
                that._filter(null);
            }
        });

        typeAhead.element().find('input')
            .attr('aria-label', typeAhead.element().find('input').attr('placeholder'))
            .attr('aria-controls', treePanel.id());

        treePanel.on('invalidate', function () {
            that._model.data = [];
        });

        this.element().one('oui-search', function (event, widget) {
            var searchFn = widget.getOption('storeFns.search');
            if (searchFn) {
                widget.updateOption('storeFns.search', _.wrap(searchFn, function (fn) {
                    var unselectable = treePanel.unselectable();
                    var disabled = treePanel.disabled();
                    var recordKeyProp = treePanel.getOption('recordKeys.id');
                    var ids = _.union(_.pluck(unselectable, recordKeyProp),
                                      _.pluck(disabled, recordKeyProp));
                    var attrs = treePanel.getOption('storeFns.attrs');
                    return $.when(fn.apply(this, _.rest(arguments)))
                        .then(function (data) {
                            return _.reject(data, function (node) {
                                var nodeAttrs = [];
                                if (attrs) {
                                    nodeAttrs = attrs(node);
                                }
                                return nodeAttrs.indexOf('disabled') !== -1 ||
                                    nodeAttrs.indexOf('unselectable') !== -1 ||
                                        _.contains(ids, _.get(node, recordKeyProp));
                            });
                        });
                }));
            }
        });

        this.element().on('oui-select oui-required-error', function (event, widget, eventName, selected) {
            if (selected == null) {
                selected = [];
            }

            if (widget.name() === 'typeahead') {
                var nodesToAdd = _.filter(selected, function (node) {
                    return treePanel.node(node) == null;
                });

                if (widget.options().allowNewValues) {
                    treePanel.add(nodesToAdd);
                }
                treePanel.value(selected);
                that._model.data = selected;
                that.trigger('accept', selected);
            }
        });

        if (options.modal || _.isObject(options.modal)) {
            this._setupModal(that._model, settings);
        } else {
            this.inject(treePanel.element(), '.tree-content');
            this.inject(typeAhead.element(), '.search');
        }

        return this;
    }

    PickerView.prototype = Object.create(View.prototype);
    pvp = PickerView.prototype;

    pvp.constructor = PickerView;

    /**
     * Sets up all the modal listeners and events
     *
     * @method PickerView#_setupModal
     * @private
     *
     * @param {Object} model
     * @param {Object} settings
     */
    pvp._setupModal = function (model, settings) {
        var treePanel = this._treePanel;
        var typeAhead = this._typeAhead;
        var that = this;

        /**
         * @method PickerView~isDataEqual
         * @private
         *
         * @param {Array<Object>} newData
         *
         * @returns {Boolean}
         */
        function isDataEqual(newData) {
            var selected = newData || treePanel.selected();

            if (model.data.length !== selected.length) {
                return false;
            }

            return _.every(selected, function (node) {
                return _.find(model.data, flexEqual.bind(null, node));
            });
        }

        /**
         * @method PickerView~resetHandler
         * @private
         *
         * @param {Array<Object>} data
         */
        function resetHandler(data) {
            var disabled = isDataEqual(data);

            if (disabled && that._options.repeatable) {
                disabled = false;
            }

            that._$modal.find('.modal-accept')
            .prop('disabled', disabled)
            .toggleClass('disabled', disabled);

            that._$modal.find('.modal-clear')
                .prop('disabled', _.isEqual(model.data, []))
                .toggleClass('disabled', _.isEqual(model.data, []));
        }

        /**
         * @method PickerView~accept
         * @private
         */
        function accept() {
            treePanel.trigger('accept', treePanel.selected());
            treePanel.trigger('change', treePanel.selected());
            that._$el.trigger('change');
            resetHandler();

            that._$modal.modal('hide');
        }

        /**
         * @method PickerView~showModalHandler
         * @private
         *
         * @param {jQuery.Event} event
         */
        function showModalHandler(event) {
            event.preventDefault();

            if (!that._$modal.hasClass('in')) {
                $.when(busy(that._$el.find('.btn-picker')), that.trigger('pre-show', model.data || []))
                    .then(handle)
                    .done(function () {
                        that._$modal.modal('show');

                        that._$modal.one('shown', function () {
                            focusUtils.findFocusable(that._$modal.find('.modal-body'))[0].focus();
                        });
                    })
                    .fail(shake.bind(null, that._$modal))
                    .always(restore.bind(null, that._$el.find('.btn-picker')))
                    .always(available.bind(null, that._$el.find('.btn-picker')));
            }
        }

        /**
         * @method PickerView~cancelHandler
         * @private
         */
        function cancelHandler() {
            that.trigger('cancel');
            treePanel.selected(model.data);
            resetHandler();
        }

        if (_.isObject(this._options.modal)) {
            this._$modal = this._options.modal;
        } else {
            this._$modal = $(pickerModalTemplate({
                modalId: 'oui-' + settings.name + '-' + settings.id + '-modal-label',
                required: this._options.required,
                locale: this._options.locale
            }));
            this._$modal.find('.modal-body').append($(pickerTemplate({
                options: this._options
            })));
        }

        this._$modal.addClass(this._options.modalClasses);

        this._$modal.on('hidden', function () {
            that._$modal.detach();
        });

        this._$modal.on('show', function () {
            if (that._$modal.parent().length === 0) {
                $(that._options.appendTo).append(that._$modal);
            }
        });

        if (this._$el.is('button')) {
            this._$el.on('click', showModalHandler);
        } else {
            this._$el.find('button')
            .on('mousedown', function (event) {
                event.preventDefault();
            })
            .on('click', showModalHandler);

            this.addShortcuts({
                'F2 SHIFT-F10': showModalHandler
            });

            this._$el.on('click', 'input[readonly]', function (event) {
                if (this._options.openFromInput) {
                    showModalHandler(event);
                }
            }.bind(this));
        }

        this._$modal.find('.tree-content').append(this._treePanel.element());
        this._$modal.find('.search').append(this._typeAhead.element());

        this._treePanel.on('picker.select', function (data) {
            resetHandler(data);
        });

        this._treePanel.on('accept', function (data) {
            model.data = data;

            if (!that._$el.hasClass('orcl-type-ahead') && _.isFunction(that._options.viewFns.display)) {
                that._$el.find('input').val(that._options.viewFns.display(data));
            }
        });

        this._treePanel.on('picker.accept picker-accept', function (data) {
            model.data = data;
            resetHandler();
        });

        this._treePanel.on('details-loaded', function () {
            model.data = that._treePanel.selected();
            resetHandler();
        });

        this._$modal.on('click', '.modal-accept', function (event) {
            event.stopPropagation();
            event.stopImmediatePropagation();

            //TODO: look into accept - disabling button, available - enabling button, resetHandler - disabling again
            $.when(busy(that._$modal.find('.modal-accept')), that.trigger('pre-accept', treePanel.selected()))
                .then(handle)
                .done(accept)
                .fail(shake.bind(null, that._$modal))
                .always(available.bind(null, that._$modal.find('.modal-accept')))
                .always(resetHandler);

        }).on('click', '.modal-clear', function (e) {
            e.stopPropagation();
            e.stopImmediatePropagation();

            $.when(busy(that._$modal.find('.modal-clear')), that.trigger('pre-accept', []))
                .then(handle)
                .done(function () {
                    that._treePanel.selected([]);
                })
                .done(accept)
                .fail(shake.bind(null, that._$modal))
                .always(restore.bind(null, that._$modal.find('.modal-clear')))
                .always(available.bind(null, that._$modal.find('.modal-clear')))
                .always(resetHandler);

        }).on('click', '[data-dismiss=modal]', function (e) {
            e.stopPropagation();
            e.preventDefault();
            that._$modal.modal('hide');

            typeAhead.invalidate();
            that._filter(null);

            if (!$(this).hasClass('modal-accept')) {
                cancelHandler();
            }
        }).on('keydown', '.orcl-picker', function (e) {
            switch (e.keyCode) {
                case keyCode.ENTER:
                    if (!that._$modal.find('.modal-accept').hasClass('disabled')) {
                        e.stopPropagation();
                        e.stopImmediatePropagation();

                        $.when(busy(that._$modal.find('.modal-accept')), that.trigger('pre-accept', treePanel.selected()))
                            .then(handle)
                            .done(accept)
                            .fail(shake.bind(null, that._$modal))
                            .always(available.bind(null, that._$modal.find('.modal-accept')))
                            .always(resetHandler);
                    }
                    break;
            }
        }).on('keyup', function (e) {
            switch (e.keyCode) {
                case keyCode.ESC:
                    e.stopPropagation();
                    that._$modal.modal('hide');
                    cancelHandler();
                    break;
            }
        })
        .on('shown.treepanel-init', function () {
            treePanel.init();
            typeAhead.init();
            that._$modal.off('shown.treepanel-init');
        });

        this._treePanel.on('dblclick', function () {
            if (!that._$modal.find('.modal-accept').hasClass('disabled')) {
                $.when(busy(that._$modal.find('.modal-accept')), that.trigger('pre-accept', treePanel.selected()))
                    .then(handle)
                    .done(accept)
                    .fail(shake.bind(null, that._$modal))
                    .always(available.bind(null, that._$modal.find('.modal-accept')))
                    .always(resetHandler);
            }
        });
    };

    /**
     * @method PickerView#buildElement
     * @param {Picker#Config} options
     * @param {jQuery} el
     *
     * @returns {jQuery}
     */
    pvp.buildElement = function (options, el) {
        var $element;
        var isTypeAhead;

        if (options.modal) {
            if (_.isNull(el) && options.buttonTriggerOnly) {
                $element = $(pickerButtonTemplate({
                    locale: options.locale
                }));
            } else if (_.isNull(el)) {
                $element = $(pickerInputTemplate({
                    locale: options.locale
                }));
            } else if (el.jquery && el.is('input')) {
                //Input given
                if (el.parent().hasClass('orcl-widget')) {
                    el.parent().addClass('orcl-picker-input').addClass('orcl-picker');
                } else {
                    el.wrap(pickerWrapInputTemplate);
                }
                $element = el.parent();
                $element.append($(pickerButtonTemplate({
                    locale: options.locale
                })).attr('tabindex', '-1'));
            } else if (el.jquery && (el.is('button') || el.is('a'))) {
                //Button given
                $element = el;
            }

            isTypeAhead = !$element.hasClass('orcl-type-ahead');
            $element.find('input').attr('readonly', isTypeAhead).attr('aria-haspopup', isTypeAhead);
        } else if (!_.isNull(el)) {
            return el.addClass('orcl-picker');
        } else {
            $element = $(pickerTemplate({
                options: options
            })).addClass('orcl-picker');

            if (_.isString(options.locale.title)) {
                $element.find('.tree-content').prepend(titleTemplate({
                    id: this._treePanel.id() + '-label',
                    treeId: this._treePanel.id(),
                    locale: options.locale
                }));

                this._treePanel.element().attr('aria-labelledby', this._treePanel.id() + '-label');
            }
        }

        $element.addClass('orcl-picker');

        return $element;
    };

    /**
     * @method PickerView#init
     */
    pvp.init = function () {
        if (this._$el.hasClass('orcl-picker-input')) {
            var input = this._$el.children('input');
            this._$el.find('.instructions[data-name=' + this._settings.name + ']').attr('id', this.id() + '-instructions-' + this._settings.name);
            input.attr('aria-describedby', ((input.attr('aria-describedby') || '') + ' ' + this.id() + '-instructions-' + this._settings.name).trim());
        }
    };

    /**
     * @method PickerView#destroy
     *
     * @param {Object} clean
     */
    pvp.destroy = function (clean) {
        if (this._options.modal) {
            //We have to get rid of the modal
            this._$modal.one('hidden shown', function () {
                $(this).remove();
            });

            this._$el.find('button, .instructions').remove();
            this._$el.find('input').unwrap();
            this._$el.off();
            if (this._$modal.hasClass('in')) {
                this._$modal.modal('hide');
            } else {
                this._$modal.remove();
            }
        }

        if (clean.self) {
            this._$el.remove();
        }
    };

    /**
     * @method PickerView#_applyDisabled
     * @private
     *
     * @param {Boolean} disabled
     */
    pvp._applyDisabled = function (disabled) {
        var elems;

        if (this._options.buttonTriggerOnly) {
            elems = this._$el.toggleClass('disabled');

            prop('disabled', elems, disabled);
        } else if (this._options.modal) {
            elems = this._$el.find('input, button, .btn')
            .toggleClass('disabled', disabled);

            prop('disabled', elems, disabled);
        } else {
            this._typeAhead.disable(disabled);
            this._treePanel.disable(disabled);
        }
    };

    /**
     * @method PickerView#_filter
     * @private
     *
     * @param {String?} query
     */
    pvp._filter = function (query) {
        this._typeAhead.element().find('input').toggleClass('empty', _.isEmpty(query));
    };

    /**
     * Returns the element that identifies the widget.
     *
     * @method PickerView#getIdentifyingElement
     *
     * @returns {jQuery}
     */
    pvp.getIdentifyingElement = function () {
        var picker = this._$el;
        var pickerInput = this._$el.hasClass('orcl-picker-input');
        var element = picker;

        if (pickerInput) {
            element = picker.find('input');
        }

        return element;
    };

    /**
     * @method PickerView#show
     */
    pvp.show = function () {
        this._$modal.modal('show');
    };

    /**
     * @method PickerView#hide
     */
    pvp.hide = function () {
        this._$modal.modal('hide');
    };

    /**
     * @method PickerView#_applyDisabled
     * @private
     *
     * @param {Boolean} disabled
     */
    pvp._applyDisabled = function (disabled) {
        if (this._options.modal) {
            this._$el.find('.btn').prop('disabled', disabled);
            this._$el.find('input').prop('disabled', disabled);
            if (!this._$el.hasClass('orcl-type-ahead')) {
                this._$el.find('input').attr('readonly', true).attr('aria-haspopup', true);
            }
        } else {
            this._treePanel.disable(disabled);
        }
    };

    /**
     * @method PickerView#isValid
     *
     * @returns {Boolean}
     */
    pvp.isValid = function () {
        return !(this._options.required && this._model.data.length === 0);
    };

    return PickerView;
});

define('oui/widgets/tree/groupingUtils',[
    'underscore'
], function (_) {
    'use strict';

    /**
     * @module groupingUtils
     * @private
     */

    return {
        /**
         * @method groupingUtils#deflate
         * @static
         *
         * @param {Object} obj
         * @returns {String}
         */
        deflate: function (obj) {
            return obj.toString();
        },
        /**
         * @method groupingUtils#inflate
         * @static
         *
         * @param {String} id
         * @returns {Object}
         */
        inflate: function (id) {
            var self = _.last(id.split('@,@'));
            var parents = _.dropRight(id.split('@,@'), 1);

            var parentGroup;

            if (parents.length > 0) {
                parentGroup = this.inflate(parents.join('@,@'));
            }

            return this._makeGroup(self.split(':')[0], self.split(':')[1], parentGroup);

        },
        /**
         * @method groupingUtils#flatten
         * @static
         *
         * @param {Object} group
         * @returns {Array<Object>}
         */
        flatten: function (group) {

            if (group.parent) {
                return _.union([group], this.flatten(group.parent));
            } else {
                return [group];
            }
        },
        /**
         * @method groupingUtils#_makeGroup
         * @private
         * @static
         *
         * @param {String} key
         * @param {String} value
         * @param {Object} parentGroup
         *
         * @returns {Object}
         */
        _makeGroup: function (key, value, parentGroup) {
            return {
                group: true,
                parent: parentGroup,
                value: value,
                field: key,
                flatten: function () { //jscs:ignore jsDoc
                    if (this.parent) {
                        return _.union([this], this.parent.flatten());
                    } else {
                        return [this];
                    }
                },
                toString: function () { //jscs:ignore jsDoc
                    var str = '';

                    if (this.parent) {
                        str += this.parent.toString() + '@,@';
                    }

                    return str + this.field + ':' + this.value;
                }
            };
        }
    };
});

define('oui/utils/binarySearch',[
], function () {
    'use strict';

    /**
     * @module binarySearch
     */

    /**
     * @alias module:binarySearch
     *
     * @method search
     *
     * @param {Array} collection
     * @param {*} value
     * @param {Function} converter
     *
     * @returns {Number}
     */
    return function (collection, value, converter) {
        if (collection.length === 0) {
            return 0;
        }

        var min = 0;
        var max = collection.length - 1;

        while (min <= max) {
            var mid = Math.floor((max + min) / 2);
            var valueAtMid = converter(collection[mid]);
            if (value > valueAtMid) {
                min = mid + 1;
            } else if (value < valueAtMid) {
                max = mid - 1;
            } else {
                return mid + 1;
            }
        }

        if (max < 0) {
            return 0;
        } else {
            return min;
        }
    };

});

define('oui/widgets/tree/collection',[
    'underscore',
    'oui/base/model',
    'oui/widgets/tree/groupingUtils',
    'oui/utils/flexEqual',
    'oui/utils/math',
    'oui/utils/binarySearch',
    'logger'
], function (_, Model, groupingUtils, isEqual, math, search, logger) { //jshint ignore:line
    'use strict';

    /**
     * Returns the warning message for non-unique ids.
     *
     * @method Collection~getNodeWarning
     * @private
     *
     * @param {String} recordId
     *
     * @returns {String}
     */
    function getNodeWarning(recordId) {
        return [
            'OUI: Node not added because there was a node found with id: ' + recordId,
            'Please ensure that the recordKeys are mapped to a unique id'
        ].join('\n');
    }

    /**
     * @method Collection~groupEquality
     * @private
     *
     * @param {Object} group1
     * @param {Object} group2
     *
     * @returns {Boolean}
     */
    var groupEquality = function (group1, group2) {
        return group1.toString() === group2.toString();
    };

    var KEEP_CACHE_EVENTS = [
        'active',
        'pre-change',
        'default-change',
        'change'
    ];

    /**
     * @enum {String}
     * @memberof! Collection.prototype
     * @readonly
     */
    var STATES = {
        initialized: 'initialized', //Pre loaded
        normal: 'normal',
        filtered: 'filtered'
    };

    var ACTIVE_QUERY_KEY = '__OUI_ACTIVE_QUERY__';

    /**
     * @class Collection
     * @private
     * @extends Model
     *
     * @param {TreePanel#Config} options
     * @param {Array<Object>|Object<String,Array>|Function} initData
     */
    function Collection(options, initData) {
        Model.call(this, initData, options);

        this._nodes = {};
        this._filters = {};
        this._sorts = {
            'default': {
                ids: [],
                order: null
            }
        };
        this._attrs = {
            hidden: [],
            disabled: [],
            unselectable: [],
            expanded: [],
            selected: [],
            required: [],
            loaded: [],
            empty: [] //Specifically told there are no children
        };
        this._state = STATES.initialized;
        this._focused = null;
        this._partials = {};
        this._tempPartials = {};
        this._cachedNodes = null;

        Object.defineProperty(this, '_data', {
            get: function () { //jscs:ignore jsDoc
                return this._attrs.selected;
            },
            set: function (newSelected) { //jscs:ignore jsDoc
                this._attrs.selected = _.difference(newSelected, this.getAttr('unselectable'), this.getAttr('disabled'));
            }
        });

        this._initialize(initData);

        if (!this._options.lazyLoad) {
            this._state = STATES.normal;
        }

    }

    Collection.prototype = Object.create(Model.prototype);

    _.extend(Collection.prototype, {
        /**
         * @method Collection#trigger
         *
         * @param {String} eventName
         *
         * @returns {*}
         */
        trigger: function (eventName) {
            if (this._cachedNodes != null && KEEP_CACHE_EVENTS.indexOf(eventName) === -1) {
                this.clearVisibleNodes();
            }

            return Model.prototype.trigger.apply(this, arguments);
        },
        /**
         * @method Collection#init
         */
        init: function () {},
        /**
         * Populate the collection from the provided data.
         *
         * * If the data is an array just add all of those objects to the collection.
         * * If the data is an Object with the function getAll, treat that as a mediator and set it up. Adding all known
         * objects to the collection.
         * * If the data is a plain Object with available and selected properties add the available to the collection and set
         * the selected attributes on the provided data.
         *
         * @method Collection#_initialize
         * @private
         * @param {Array<Object>|Object|Function} initData
         */
        _initialize: function (initData) {
            if (_.isArray(initData)) {
                this.addAll(initData);
            } else if (_.isObject(initData) && initData.getAll) {
                this._mediator = initData;
                this.addAll(initData.getAll());
                this.data(initData.selected());
            } else if (_.isPlainObject(initData)) { //Support setting both the data set and the selected nodes
                this.addAll(initData.available || []);
                this.data(_.map(initData.selected || [], function (initNode) {
                    return this._getId(this.node(initNode));
                }.bind(this)));
            }
        },
        /**
         * Adds the field to the collection of fields to strip off since they are dynamically added
         *
         * @method Collection#addField
         * @param {Object} node The node to add the field to.
         * @param {String } field The field to add to the node.
         */
        addField: function (node, field) {
            if (node._ouiFields == null) {
                node._ouiFields = ['_ouiFields'];
            }

            node._ouiFields.push(field);
        },
        /**
         * Gets the id for the record.
         *
         * If the id is not found using the recordKeys option and the idPrefix option is supplied an ID will be generated and
         * put onto the record on the property supplied in recordKeys.
         *
         * @method Collection#_getId
         * @private
         * @param {Object} node The record to get the id for.
         * @returns {String} id The identifier for the record.
         */
        _getId: function (node) {
            var id = node[this._options.recordKeys.id];

            if (id == null && this._options.idPrefix) {
                logger.warn('Record ID not found on object!');
                this.addField(node, this._options.recordKeys.id);
                node[this._options.recordKeys.id] = _.uniqueId(this._options.idPrefix);
                id = node[this._options.recordKeys.id];
                this._tempPartials[id] = node;
            }

            return id;
        },
        /**
         * Gets the node's value at the field provided.
         *
         * @method Collection#_getNodeValue
         * @private
         * @param {Object} node The record to look through
         * @param {String} field The field to get. Can also be a path specifier for '_.get'
         * @param {*} defaultValue The default value to return if the value is not found.
         * @returns {*} The value of the node for the field requested.
         */
        _getNodeValue: function (node, field, defaultValue) {
            if (node[field]) {
                return node[field];
            }
            return _.get(node, field, defaultValue);
        },
        /**
         * Utility method for determining if the record is a group.
         *
         * @method Collection#isGroup
         * @param {String|Object} id The id to determine if it is a group
         * @returns {Boolean} Whether or not the record id maps to a group.
         */
        isGroup: function (id) {
            if (_.isObject(id)) {
                return true;
            }

            return false;
        },
        /**
         * Utility method to find all the parents for a given node.
         *
         * @method Collection#getParents
         * @private
         * @param {Object|String} id The id to look the parents up for.
         * @returns {Array} List of parent ids that are ancestors of the given id.
         */
        getParents: function (id) {
            var path = [];
            var treeConfig = _.findLast(this._options.groupBy, function (groupConfig) {
                return groupConfig.asTree;
            });
            var isGroup = this.isGroup(id);

            if (this._options.flat || (this.isFiltered() && !this._options.filterShowHierarchy)) {
                return [];
            }

            if (isGroup) {
                var groups = [];
                var currentGroup = id.parent;
                while (currentGroup != null) {
                    groups.push(currentGroup);
                    currentGroup = currentGroup.parent;
                }
                return groups;
            }

            if (treeConfig) {
                var reverseField = '_' + treeConfig.field;
                var parentId = _.get(this._filters, [reverseField, id]);

                while (parentId != null && parentId !== treeConfig.rootValue) {
                    path.unshift(parentId);
                    parentId = _.get(this._filters, [reverseField, parentId]);
                }
            }

            _.each(this._options.groupBy, function (groupConfig) {
                if (groupConfig.asTree == null || !groupConfig.asTree) {
                    var filter = this._filters[groupConfig.field];
                    var value = this._getGroupValue(this.node(id), groupConfig);
                    if (filter[value]) {
                        path = [groupingUtils._makeGroup(groupConfig.field, value, _.last(path))].concat(path);
                    }
                }
            }.bind(this));

            return _(path).compact().value();
        },
        /**
         * @method Collection#getParent
         *
         * @param {String} id
         * @returns {String}
         */
        getParent: function (id) {
            return _.last(this.getParents(id));
        },
        /**
         * Retrieves the siblings for the given id or empty array if none are found.
         *
         * @method Collection#_getSiblings
         * @private
         * @param {Object|String} id The id to look up siblings for (not including itself)
         * @returns {Array} The siblings if it has any or empty array
         */
        _getSiblings: function (id) {
            var treeConfig = _.findLast(this._options.groupBy, 'asTree');
            var siblings = [];

            if (this._options.flat || this._options.groupBy.length === 0) {
                return _.without(this._sorts['default'].ids, id);
            }

            if (treeConfig) {
                var reverseField = '_' + treeConfig.field;
                siblings = _.without(this._filters[treeConfig.field][this._filters[reverseField][id]], id);

            } else {
                var node = this.node(id);
                _.each(this._options.groupBy, function (group) {
                    var filterValue = this._filters[group.field][this._getGroupValue(node, group)];
                    if (siblings.length) {
                        siblings = _.intersection(siblings, filterValue);
                    } else {
                        siblings = _.without(filterValue, id);
                    }
                }.bind(this));
            }

            return _.intersection(this._sorts['default'].ids, siblings);
        },
        /**
         * Retrieves the level of the record based on grouping.
         *
         * This method is 1 based, therefore the initial level is 1.
         *
         * Mainly used as a method to apply aria-level property which is 1 based. See {@link
         * http://www.w3.org/TR/wai-aria/states_and_properties#aria-level} for more information.
         *
         * **Note:** if this changes the calculation for spacers on a node as well as the aria-level property should be
         * examined.
         *
         * @method Collection#_getLevel
         * @private
         * @param {Object|String} id The identifier to look up level
         * @returns {Number} The level of the record.
         */
        _getLevel: function (id) {
            return this.getParents(id).length + 1;
        },
        /**
         * Determines if a node matches the regular expression.
         *
         * The method uses the display property as a part of viewFns to determine matching.
         *
         * @method Collection#_matches
         * @private
         * @param {Object} node The record object to determine if it matches.
         * @param {RegExp} regex The regex to use for testing.
         */
        _matches: function (node, regex) {
            return regex.test(this._options.viewFns.display(node)) || _.some(this._options.defaultFilterFields, function (field) {
                return regex.test(_.get(node, field));
            }.bind(this));
        },
        /**
         * Utility function to look through the children (not currently configurable) property of the records to add them to
         * the collection. The collection maintains a flat list of records and builds up hierarchy dynamically.
         *
         * @method Collection#_walk
         * @private
         * @param {Object} node The node to walk down its descendants and add them to the collection.
         */
        _walk: function (node) {
            var treeConfig;
            var id;
            var children = _.get(node, this._options.recordKeys.children);

            if (children) {
                treeConfig = _.find(this._options.groupBy.reverse(), function (groupConfig) {
                    return groupConfig.asTree;
                });
                id = this._getId(node);

                if (treeConfig) {
                    _.each(children, function (child) {
                        child = _.clone(child);

                        if (child[treeConfig.field] == null) {
                            this.addField(child, treeConfig.field);
                            child[treeConfig.field] = id;
                        } else if (child[treeConfig.field] !== id) {
                            child[treeConfig.field] = id;
                        }

                        this.add(child);
                    }.bind(this));
                }
            }

        },
        /**
         * Gets the filter/group for the field if it exists.
         *
         * @method Collection#_getGroup
         * @private
         * @param {String} field The field to find a filter for.
         * @returns {Object<String, Array>} filter The filter if found or undefined if not found.
         */
        _getGroup: function (field) {
            if (this._filters[field]) {
                return this._filters[field];
            }
        },
        /**
         * Gets a representation of a group.
         *
         * Unique per parent.
         *
         * For example if there are two levels of groups (category, type) the second level will have a unique identifier
         * based on their parents.
         *
         * @method Collection#_makeGroup
         * @private
         * @param {String} field The field to make a group for.
         * @param {String} key The grouping value for the group.
         * @param {Object} [parentGroup] The parent group if one exists.
         * @returns {Object} The group for the field/value pair.
         */
        _makeGroup: function (field, key, parentGroup) {
            return groupingUtils._makeGroup(field, key, parentGroup);
        },

        /**
         * Utility function to determine whether the new data (selected attribute) is the same as the current value.
         *
         * @method Collection#isSame
         * @param {Array<Object>} newData - The new data to compare equality with existing data.
         * @returns {Boolean} Whether or not the existing data matches the new data
         */
        isSame: function (newData) {
            return _.isEqual(this._data, newData);
        },
        selected: Model.prototype.data,

        /**
         * @method Collection#_updatePartial
         * @private
         *
         * @param {Object} data
         * @param {String} recordId
         */
        _updatePartial: function (data, recordId) {
            if (_.keys(this._partials).length > 0) {
                var id = _.findKey(this._partials, function (partialData) {
                    return isEqual(_.omit(partialData, partialData._ouiFields), _.omit(data, partialData._ouiFields));
                }.bind(this));

                if (id != null && id !== recordId) {
                    _.each(this._attrs, function (ids) {
                        if (_.contains(ids, id)) {
                            var idx = _.indexOf(ids, id);
                            ids.splice(idx, 1);
                            if (!_.contains(ids, recordId)) {
                                ids.push(recordId);
                            }
                        }
                    }.bind(this));
                    delete this._nodes[id];
                    delete this._partials[id];

                    this.trigger('details-loaded');
                }
            }
        },

        //Methods relating to collections (CRUD - add/remove/update...)
        /**
         * Adds a node to the collection. If the node already exists based on id it will update the node instead with the new
         * data.
         *
         * Adds the record id to any of the filters, sorts, or attributes that are set up in the options.
         *
         * @method Collection#add
         * @param {Object} node The record to add to the collection.
         * @param {Number} [modifiers] The options for the function (currently, just supporting add index for placing a node
         * @param {Boolean} [skipPartials=false]
         * between nodes).
         *
         * @fires Collection#add
         *
         * @returns {String} The node id that was added to the collection.
         */
        add: function (node, modifiers, skipPartials) { //jshint ignore:line
            var recordId;

            if (this._options.cloneOnAdd) {
                node = _.clone(node);
            }
            recordId = this._getId(node);

            if (!this.node(recordId)) {
                this._nodes[recordId] = node;

                if (!this._options.flat) {
                    //Old attribute required to store the children on.
                    this._walk(node);
                }

                if (this._options.groupBy != null || this._options.indexes != null) {
                    this.addToGroups(node);
                }

                this.addToSorts(node, modifiers);

                this.addAttrs(node);

                this.trigger('add', recordId);
            } else if (this._options.extendExisting && isEqual(this._nodes[recordId], node)) {
                this.update(recordId, node);
            } else {
                logger.warn(getNodeWarning(recordId));
            }
            //Log when node was found and we are not extending?

            if (!skipPartials) {
                _.merge(this._partials, this._tempPartials);
                this._tempPartials = {};
            }

            this._updatePartial(node, recordId);

            return recordId;
        },
        /**
         * Adds the collection of nodes to the collection.
         *
         * Pauses any events that may be emitted as a result of this call (add) and fires its own addAll event once it
         * resumes.
         *
         * @method Collection#addAll
         * @param {Array<Object>} nodes The nodes to add to the collection.
         * @param {Number} [modifiers] The modifiers to use for each add operation.
         *
         * @fires Collection#will-change
         * @fires Collection#addAll
         */
        addAll: function (nodes, modifiers) {
            this.trigger('will-change');
            this.pause();

            var ids = _.map(nodes, function (node) {
                var id = this.add(node, modifiers, true);
                if (modifiers != null) {
                    modifiers++;
                }

                return id;
            }.bind(this));

            var attemptedEvents = this.resume();
            this.trigger('addAll', ids);

            if (!this._options.flat && attemptedEvents['group-added'] != null) {
                this.trigger('rerender');
            }

            if (attemptedEvents['details-loaded'] != null) {
                this.trigger('details-loaded');
            }

            //Once you have added items - it has move to the normal state
            if (this._state === STATES.initialized) {
                this._state = STATES.normal;
            }

            _.merge(this._partials, this._tempPartials);
            this._tempPartials = {};
        },
        /**
         * Gets the nodes within the collection as an array.
         *
         * @method Collection#nodes
         *
         * @param {Object} modifiers
         * @returns {Array} The array of nodes that currently exist within the collection.
         */
        nodes: function (modifiers) {
            var nodes = _.toArray(this._nodes);

            if (_.get(modifiers, 'sorted', false)) {
                nodes = _.sortBy(nodes, function (node) {
                    return this._sorts['default'].ids.indexOf(this._getId(node));
                }.bind(this));
            }

            if (modifiers && modifiers.strip) {
                return _.map(nodes, function (node) {
                    return _.omit(node, node._ouiFields);
                });
            }

            return nodes;
        },
        /**
         * Gets the node for a given id.
         *
         * @method Collection#node
         * @param {String} id The identifier to look for a node.
         * @param {Object} modifiers
         * @returns {Object} The found node or undefined if none exists for a given id.
         */
        node: function (id, modifiers) {
            var node = null;

            if (_.isString(id) || _.isNumber(id) || _.isBoolean(id)) {
                node = this._nodes[id];
            } else if (_.isObject(id)) {
                if (_.has(id, this._options.recordKeys.id)) {
                    node = this._nodes[_.get(id, this._options.recordKeys.id)];
                } else {
                    node =  _.find(this._nodes, function (node) {
                        return isEqual(id, node);
                    });
                }
            }

            if (node && modifiers && modifiers.strip) {
                return _.omit(node, node._ouiFields);
            }

            return node;
        },
        /**
         * Updates the node at a given id with the new data.
         *
         * @method Collection#update
         * @param {String} id The id of the record to update
         * @param {Object} data The new data to add/overwrite on the node.
         *
         * @fires Collection#update
         *
         * @returns {String} The id of the node updated.
         */
        update: function (id, data) {
            if (this.node(id)) {
                this._nodes[id] = _.extend({}, this.node(id), data);

                if (this._options.groupBy != null || this._options.indexes != null) {
                    this.addToGroups(this._nodes[id]);
                }

                _.each(this._attrs, function (list, key) {
                    if (key !== 'selected') {
                        this.removeFromAttr(key, id);
                    }
                }.bind(this));

                this.addAttrs(this._nodes[id]);

                this.trigger('update', id);

                return id;
            }
        },
        /**
         * Updates all the objects within the map.
         *
         * @method Collection#updateAll
         * @param {Object<String, Object>} map The collection of records to update. The key is the record id and the value is
         * the new data to be applied.
         *
         * @fires Collection#will-change
         * @fires Collection#update
         */
        updateAll: function (map) {
            this.trigger('will-change');

            this.pause();
            var ids = _.map(map, function (data, id) {
                return this.update(id, data);
            }, this);
            this.resume();

            this.trigger('update', ids);
        },
        /**
         * Removes the node with the given id.
         *
         * @method Collection#remove
         * @param {String} id The id of the node to remove
         *
         * @fires Collection#remove
         */
        remove: function (id) {
            if (this._nodes[id]) {
                _.each(_.union(this._options.groupBy, this._options.indexes), function (groupConfig) {
                    var groupValue = this._getNodeValue(this.node(id), groupConfig.field);
                    var parents = this.getParents(id);

                    this._filters[groupConfig.field][groupValue] = _.without(this._filters[groupConfig.field][groupValue], id);

                    if (this._filters[groupConfig.field][groupValue].length === 0) {
                        this.trigger('group-removed', _.findWhere(parents, { field: groupConfig.field, value: groupValue }));
                    }
                }.bind(this));

                _.each(this._attrs, function (value, attr) {
                    this.removeFromAttr(attr, id);
                }.bind(this));

                this._sorts['default'].ids.splice(_.indexOf(this._sorts['default'].ids, id), 1);

                delete this._nodes[id];

                this.trigger('remove', id);
            }

            return id;

        },

        /**
         * Removes the supplied nodes from the collection.
         *
         * @method Collection#removeAll
         * @param {Array<String>} ids The ids of the nodes to remove
         *
         * @fires Collection#removeAll
         */
        removeAll: function (ids) {
            this.trigger('will-change');
            this.pause();

            var removed = _.map(ids, this.remove.bind(this));

            this.resume();

            this.trigger('removeAll', removed);
        },

        //Methods relating to Trees...roots, levels, children...
        /**
         * Gets the first level of nodes from within the Collection.
         * Takes into account grouping, filtering, sorting configurations.
         *
         * @method Collection#getRoots
         * @returns {Array<String|Object>} The sorted array of objects to render in at the first level
         */
        getRoots: function () {
            var roots = [];

            if (this.isFiltered() && this._filters[ACTIVE_QUERY_KEY].length === 0) {
                //Nothing to draw
                return [];
            }

            if (!this._options.flat && this._options.groupBy) {
                var rootGroupBy = _.first(this._options.groupBy);

                if (rootGroupBy.asTree === true) {
                    var rootValue = rootGroupBy.rootValue || 'null';
                    roots = _.get(this._filters, [rootGroupBy.field, rootValue], []);
                } else {
                    return this._getGroupRoots(rootGroupBy);
                }
            } else {
                roots = _.pluck(this.nodes(), this._options.recordKeys.id);
            }

            if (this.isFiltered()) {
                if (this._options.filterShowHierarchy) {
                    roots = _.intersection(roots, this._filters[ACTIVE_QUERY_KEY]);
                } else {
                    roots = this._filters[ACTIVE_QUERY_KEY];
                }
            }

            return _.intersection(this._sorts['default'].ids, roots);
            /*
            //Must be all of them if there is no grouping
            return _.sortBy(roots, function (id) {
                return this._sorts['default'].ids.indexOf(id);
            }.bind(this));*/
        },
        /**
         * Gets the root elements if the collection is grouped.
         *
         * @method Collection#_getGroupRoots
         * @private
         * @param {Object} rootGroupBy The root group of the collection
         * @returns {Array<Object>} The root groups of the collection
         */
        _getGroupRoots: function (rootGroupBy) {
            var definesGroupValues = _.has(rootGroupBy, 'groupValues');
            var groupValues = _.indexBy(rootGroupBy.groupValues, 'id');
            var result;

            if (this._options.showEmptyGroups) {
                // This could be slow because it forces the whole evaluation of _.keys()
                result = _.chain(groupValues).pluck('id').union(_.keys(this._filters[rootGroupBy.field]));
            } else {
                result = _.chain(this._filters[rootGroupBy.field]).keys();
            }

            return result
                .sortBy(function (key) {
                    return (groupValues[key] && groupValues[key].order) || 0;
                })
                .map(function (key) {
                    var group = this._makeGroup(rootGroupBy.field, key);
                    if (_.has(groupValues, key) || (this._options.showEmptyGroups && !definesGroupValues) || this._getGroupChildren(group).length > 0) {
                        return group;
                    }
                }.bind(this))
                .compact()
                .value();
        },
        /**
         * Gets the child groups for a group.
         *
         * @method Collection#_getGroupChildGroups
         * @private
         * @param {Object} nextGroup The group config to get children for.
         * @param {Object} id The current group object
         * @returns {Array<Object>} The child groups below to requested group
         */
        _getGroupChildGroups: function (nextGroup, id) {
            var currentGroup = this._filters[id.field];
            var isFiltered = this.isFiltered();
            var definesGroupValues = _.has(nextGroup, 'groupValues');
            var allNextGroupKeys = _.chain(nextGroup.groupValues).filter(function (groupValue) {
                return groupValue[id.field] === id.value;
            }).indexBy('id').value();
            var matchingGroups = _.chain(this._filters[nextGroup.field])
                .reduce(function (memo, group, key) {
                    if (this._options.showEmptyGroups && !definesGroupValues) {
                        memo.push(key);
                    } else {
                        var intersection = _.intersection(group, currentGroup[id.value]);
                        if (isFiltered) {
                            intersection = _.intersection(intersection, this._filters[ACTIVE_QUERY_KEY]);
                        }
                        if (intersection.length > 0) {
                            memo.push(key);
                        }
                    }
                    return memo;
                }.bind(this), [])
                .value(); // Slows this down somewhat because we need to evaluate here (lodash cannot union two _.chain'd arrays)
            var groups;

            if (this._options.showEmptyGroups) {
                groups = _.chain(allNextGroupKeys).keys().union(matchingGroups);
            } else {
                groups = _.chain(matchingGroups);
            }

            return groups
                .sortBy(function (key) {
                    return (allNextGroupKeys[key] && allNextGroupKeys[key].order) || 0;
                })
                .map(function (key) {
                    return this._makeGroup(nextGroup.field, key, id);
                }.bind(this))
                .value();
        },
        /**
         * Gets the children for the group. Either the next level of group nodes or the children that match the current
         * group. Takes into account the parents to make sure the children match all filters.
         *
         * @method Collection#_getGroupChildren
         * @private
         * @param {Object} id The group config to get children for.
         * @param {{sorted: Boolean}} [modifiers] Modifiers for changing how the collection is returned
         * @returns {Array<Object|String>} The children below to requested group (either groups or ids)
         */
        _getGroupChildren: function (id, modifiers) {
            var nextGroup = this._options.groupBy[_.findIndex(this._options.groupBy, function (config) {
                return config.field === id.field;
            }) + 1];
            var isFiltered = this.isFiltered();

            if (nextGroup) { //Gets the next level of groups from groupByConfig
                return this._getGroupChildGroups(nextGroup, id);
            } else {
                var arrays = _.map(groupingUtils.flatten(id), function (group) {
                        return this._filters[group.field] ? this._filters[group.field][group.value] : [];
                    }.bind(this));

                if (isFiltered) {
                    arrays.push(this._filters[ACTIVE_QUERY_KEY]);
                }
                if (modifiers && modifiers.sorted) {
                    arrays.unshift(this._sorts['default'].ids);
                }

                return _.intersection.apply(_, arrays);
            }
        },
        /**
         * Gets the children for the requested group.
         *
         * Takes into account whether it is filtered, and grouped.
         *
         * @method Collection#getChildren
         * @param {String|Object} id Either a group or a tree to get the children for.
         * @param {{sorted: Boolean}} [modifiers] Modifiers for changing how the collection is returned
         * @returns {Array<String|Object>} The children for the group or record id.
         */
        getChildren: function (id, modifiers) {
            var children = [];

            if (this._options.groupBy) {
                var treeConfig = _.findLast(this._options.groupBy, function (groupConfig) {
                    return groupConfig.asTree;
                });

                if (this.isGroup(id)) {
                    return this._getGroupChildren(id, modifiers);
                } else if (treeConfig != null) {
                    children = _.get(this._filters, [treeConfig.field, id], []);
                    if (modifiers && modifiers.sorted) {
                        children = _.intersection(this._sorts['default'].ids, children);
                    }
                }

                if (this.isFiltered()) {

                    if (this._options.filterShowHierarchy) {
                        return _.intersection(children, this._filters[ACTIVE_QUERY_KEY]);
                    } else {
                        return [];
                    }
                }

            }

            return children;
        },
        /**
         * Gets a programmer-defined representation of the group suitable for displaying back to the user.
         *
         * @method Collection#getGroupDisplayValue
         *
         * @param {Object} group
         *
         * @returns {String}
         */
        getGroupDisplayValue: function (group) {
            var groupConfig = _.find(this._options.groupBy, function (groupConfig) {
                return groupConfig.field === group.field;
            });
            var displayFromNodeFn = groupConfig.displayFromNode;
            var displayFn = groupConfig.display;
            var groupValueItem = _.find(groupConfig.groupValues, 'id', group.value);
            var result;

            var withDisplay = function () {//jscs:ignore jsDoc
                return displayFn({
                    field: group.field,
                    key: group.value
                });
            };

            var withFromNode = function () {//jscs:ignore jsDoc
                var filtersGroup;
                if (_.has(this._filters, [group.field]) && _.has(this._filters[group.field], [group.value])) {
                    filtersGroup = this._filters[group.field][group.value];
                }
                if (filtersGroup && filtersGroup.length > 0) {
                    var node = this.node(_.first(filtersGroup));
                    if (node) {
                        return displayFromNodeFn(_.cloneDeep(node));
                    }
                }
            }.bind(this);

            if (groupValueItem != null) {
                result = groupValueItem.display;
            } else if (displayFromNodeFn) {
                result = withFromNode();
            } else if (displayFn) {
                result = withDisplay();
            }
            if (result != null) {
                return result;
            }
            return group.value;
        },
        /**
         * @method Collection#_getGroupValue
         * @private
         *
         * @param {Object} node
         * @param {Object} group
         *
         * @returns {String}
         */
        _getGroupValue: function (node, group) {
            var defaultGroupValue;

            if (this._options.lazyLoad) {
                defaultGroupValue = undefined;
            } else {
                defaultGroupValue = null;
            }

            if (group.getGroupValue) {
                return group.getGroupValue(node);
            } else {
                return this._getNodeValue(node, group.field, defaultGroupValue);
            }

        },
        /**
         * Adds the node to all of the groups specified in both groupBy and indexes.
         *
         * @method Collection#addToGroups
         * @param {Object} node The node to add to the groups.
         */
        addToGroups: function (node) {
            _.each(_.union(this._options.groupBy, this._options.indexes), function (groupConfig) {
                var currentGroup = _.get(this._filters, [groupConfig.field], {});
                var groupValue = this._getGroupValue(node, groupConfig);
                var recordId = this._getId(node);

                if (groupConfig.asTree) {
                    if (recordId === groupValue) {
                        logger.warn('OUI: Group value (' + groupConfig.field + ') and record id match meaning the node belongs in its own children group');
                        return;
                    }

                    _.set(this._filters, ['_' + groupConfig.field, recordId], groupValue);
                }

                if (_.isEmpty(currentGroup[groupValue])) {
                    currentGroup[groupValue] = [recordId];
                    this.trigger('group-added', currentGroup, groupValue);
                } else if (!_.contains(currentGroup[groupValue], recordId)) {
                    currentGroup[groupValue] = currentGroup[groupValue].concat(recordId);
                }
                this._filters[groupConfig.field] = currentGroup;

            }, this);
        },
        /**
         * Adds the node to the sorts specified in the sortBy option.
         *
         * @method Collection#addToSorts
         * @param {Object} node The node to add to its location in the sortBy config
         * @param {Object|Number} modifiers The options to change how sort behaves (ascending/descending, insert)
         */
        addToSorts: function (node, modifiers) {

            if (modifiers != null) {
                if (this._options.recordKeys.sequenceNumber != null) {
                    throw new Error('Cannot insert into sequenced collection');
                } else {
                    this._sorts['default'].ids.splice(modifiers, 0, this._getId(node));
                }
            } else if (this._options.recordKeys.sequenceNumber != null) {
                var value = this._getNodeValue(node, this._options.recordKeys.sequenceNumber);
                var idx = search(this._sorts['default'].ids, value, function (id) {
                    return this._getNodeValue(this.node(id), this._options.recordKeys.sequenceNumber);
                }.bind(this));
                this._sorts['default'].ids.splice(idx, 0, this._getId(node));
            } else {
                this._sorts['default'].ids.push(this._getId(node));
            }

            if (this._options.sortBy != null) {
                _.each(this._options.sortBy, function (sortConfig) {
                    var field = sortConfig.field;
                    var order = sortConfig.order;
                    var value = this._getNodeValue(node, field);

                    if (this._sorts[field] == null) {
                        this._sorts[field] = {
                            order: order,
                            ids: []
                        };
                    }

                    var idx = search(this._sorts[field].ids, value, function (id) {
                        return this._getNodeValue(this.node(id), field);
                    }.bind(this));

                    this._sorts[field].ids.splice(idx, 0, this._getId(node));
                }, this);
            }
        },
        /**
         * Adds the node to all the attribute caches.
         *
         * @method Collection#addAttrs
         * @param {Object} node The node to get the attributes for
         */
        addAttrs: function (node) {
            var attrs;

            if (this._options.storeFns && this._options.storeFns.attrs) {
                attrs = this._options.storeFns.attrs(node);

                _.each(attrs, function (attr) {
                    this.addToAttr(attr, this._getId(node));
                }.bind(this));

            }
        },
        /**
         * Finds the nodes that match the group by values.
         *
         * @method Collection#matches
         * @param {Array<Object>} values The group configs to match against.
         */
        matches: function (values) {
            var arrs = _.map(values, function (groupFilter) {
                return this._getGroup(groupFilter.field)[groupFilter.value];
            }, this);

            return _.intersection.apply(_, arrs);
        },
        /**
         * @method Collection#addNodeToGroup
         *
         * @param {Object} node
         * @param {String} field
         */
        addNodeToGroup: function (node, field) { //jshint ignore:line
            //TODO: implement
        },
        /**
         * Adds a filter (dynamic group) based on a string query to the collection.
         *
         * Updates the collections state to 'filtered'.
         *
         * @method Collection#addFilter
         * @param {String} query The filter value to search with.
         * @param {String} field The field to use for searching on.
         *
         * @fires Collection#filter
         */
        addFilter: function (query, field) {

            if (field != null && !this._filters[field]) {
                _.each(this._nodes, function (node) {
                    this.addNodeToGroup(node, field);
                });
            }

            if (query) {
                var matchingRegex = new RegExp(_.escapeRegExp(query), 'i');

                if (this._options.useGroupsToFilter) {
                    logger.log('TODO: implement use groups to filter');
                } else {
                    this.pause();
                    this._filters[ACTIVE_QUERY_KEY] = _(this._nodes).map(function (node) {
                        var parents;
                        if (this._matches(node, matchingRegex)) {
                            if (this._options.filterShowHierarchy) {
                                parents = this.getParents(this._getId(node));
                                this.addToAttr('expanded', parents);
                                return _.union([this._getId(node)], parents);
                            } else {
                                return this._getId(node);
                            }
                        }
                    }.bind(this)).compact().flatten().uniq().value();
                    this.resume();
                }

                this._state = STATES.filtered;
            }

            this.trigger('filter', this._filters[ACTIVE_QUERY_KEY]);
        },

        //TODO: temporary for filtering - adds it to nodes so it shows up when they remove the filter?
        /**
         * @method Collection#addAsFilter
         *
         * @param {Array<Object>} data
         */
        addAsFilter: function (data) {
            this._state = STATES.filtered;

            this._filters[ACTIVE_QUERY_KEY] = [];
            var parents = {};

            _.each(data, function (node) {
                var recordId;
                recordId = this._getId(node);

                if (!this.node(recordId)) {
                    this._nodes[recordId] = node;

                    //Old attribute required to store the children on.
                    if (!this._options.flat && _.get(node, this._options.recordKeys.children)) {
                        this._walk(node);
                    }

                    this.addAttrs(node);

                    this.addToSorts(node);

                    this.addToGroups(node);
                }

                if (this._options.filterShowHierarchy) {
                    _.each(this.getParents(recordId), function (parent) {
                        parents[parent] = parent;
                    });
                }

                this._filters[ACTIVE_QUERY_KEY].push(this._getId(node));
            }.bind(this));

            this._filters[ACTIVE_QUERY_KEY] = _.union(this._filters[ACTIVE_QUERY_KEY], _.values(parents));

            this.trigger('filter', this._filters[ACTIVE_QUERY_KEY]);
        },

        /**
         * Sets the state to 'normal' removing the filter.
         *
         * @method Collection#removeFilter
         *
         * @fires Collection#filter
         */
        removeFilter: function () {
            this._state = STATES.normal;

            this.trigger('filter', null);
        },

        //regroup and resort both do a full rerender -
        //wrap them in a deferred and then rerender?
        //how to handle both cases in one loop?
        /**
         * Removes all the filters as they exist from the collection and re-adds them to their respective groups.
         *
         * @method Collection#regroup
         *
         * @fires Collection#rerender
         */
        regroup: function () {
            this._filters = {};

            _.each(this._nodes, this.addToGroups.bind(this));

            this.trigger('rerender');
        },
        /**
         * Removes all sorts except the 'default' as added sort.
         *
         * @method Collection#resort
         *
         * @fires Collection#rerender
         */
        resort: function () {
            _.each(this._sorts, function (values, key) {
                if (key !== 'default') {
                    this._sorts[key] = {};
                }
            }, this);

            _.each(this._nodes, this.addToSorts.bind(this));

            this.trigger('rerender');

        },

        /**
         * @method Collection#_preCheck
         * @private
         *
         * @param {String} attr The attribute to test.
         * @param {String|Object} id The node id or group config to test if it matches.
         *
         * @returns {?Boolean} Whether or not the node matches the given attribute.
         */
        _preCheck: function (attr, id) {
            if (attr === 'loaded' && this.isFiltered() && this._options.lazyLoad) {
                return _.contains(this._filters[ACTIVE_QUERY_KEY], id);
            }

            if (attr === 'loaded' && this._options.flat && this._nodes[id] != null) {
                return true;
            }
        },

        //Methods relating to attributes
        /**
         * Tests whether a node matches an attribute specified.
         *
         * @method Collection#is
         * @param {String} attr The attribute to test.
         * @param {String|Object} id The node id or group config to test if it matches.
         * @param {Boolean} nested Whether the node is a hierarchical attribute
         *
         * @returns {Boolean} Whether or not the node matches the given attribute.
         */
        is: function (attr, id, nested) {
            nested = nested || false;

            var checked = this._preCheck(attr, id, nested);

            if (checked != null) {
                return checked;
            }

            if (this._attrs[attr] && this._attrs[attr].length > 0) {

                if (_.isObject(id)) {
                    return _.find(this._attrs[attr], groupEquality.bind(null, id)) != null;
                }

                if (!nested) {
                    return _.contains(this._attrs[attr], id);
                } else {
                    return _.contains(this._attrs[attr], id) || _.intersection(this._attrs[attr], this.getParents(id)).length > 0;
                }
            }

            return false;
        },
        /**
         * Gets all the attributes for the provided node.
         *
         * @method Collection#getAttributesForNode
         * @param {String|Object} id The node to get the attributes for.
         *
         * @returns {Object<String, Boolean>} Maps attribute name to its boolean value.
         */
        getAttributesForNode: function (id) {
            return _.mapValues(this._attrs, function (values, attr) {
                return this.is(attr, id, _.get(this._options, ['attrs', attr, 'nested'], false));
            }.bind(this));
        },
        /**
         * Gets the collection of ids within the attribute requested.
         *
         * @method Collection#getAttr
         * @param {String} attr The attribute to get.
         *
         * @returns {Array<String|Object>} The collection of ids within the attribute.
         */
        getAttr: function (attr) {
            return this._attrs[attr];
        },
        /**
         * Adds an id to the attribute within the collection.
         *
         * @method Collection#addToAttr
         * @param {String} attr The attribute to add to
         * @param {String|Object} id The identifier to add.
         *
         * @fires Collection#attribute-updated
         * @fires Collection#attribute-updated:disabled
         * @fires Collection#attribute-updated:unselectable
         * @fires Collection#attribute-updated:hidden
         * @fires Collection#attribute-updated:expanded
         * @fires Collection#attribute-updated:selected
         * @fires Collection#attribute-updated:required
         * @fires Collection#attribute-updated:empty
         */
        addToAttr: function (attr, id) {
            var newIds = [].concat(id);

            if (this._attrs[attr] == null) {
                this._attrs[attr] = [];
            }

            if (this._attrs[attr].indexOf(id) === -1) {
                this._attrs[attr] = this._attrs[attr].concat(newIds);

                this.trigger('attribute-updated', attr, this._attrs[attr], newIds);
                this.trigger('attribute-updated:' + attr, this._attrs[attr], newIds);
            }
        },
        /**
         * Removes an id to the attribute within the collection.
         *
         * @method Collection#removeFromAttr
         * @param {String} attr The attribute to remove from
         * @param {String|Object} id The identifier to remove.
         *
         * @fires Collection#attribute-updated
         * @fires Collection#attribute-updated:disabled
         * @fires Collection#attribute-updated:unselectable
         * @fires Collection#attribute-updated:hidden
         * @fires Collection#attribute-updated:expanded
         * @fires Collection#attribute-updated:selected
         * @fires Collection#attribute-updated:required
         * @fires Collection#attribute-updated:empty
         */
        removeFromAttr: function (attr, id) {
            var newIds = [].concat(id);
            var length = this._attrs[attr].length;

            if (_.isObject(id)) {
                this._attrs[attr] = _.reject(this._attrs[attr], groupEquality.bind(null, id));
            } else {
                this._attrs[attr] = _.difference(this._attrs[attr], newIds);
            }

            if (length !== this._attrs[attr].length) {
                this.trigger('attribute-updated', attr, this._attrs[attr], [], newIds);
                this.trigger('attribute-updated:' + attr, this._attrs[attr], [], newIds);
            }
        },
        /**
         * Forces the attribute to match the supplied identifiers
         *
         * @method Collection#attr
         * @param {String} attr The attribute to set.
         * @param {Array<String|Object>} ids The id to set to the attribute.
         *
         * @fires Collection#attribute-updated
         * @fires Collection#attribute-updated:disabled
         * @fires Collection#attribute-updated:unselectable
         * @fires Collection#attribute-updated:hidden
         * @fires Collection#attribute-updated:expanded
         * @fires Collection#attribute-updated:selected
         * @fires Collection#attribute-updated:required
         * @fires Collection#attribute-updated:empty
         *
         * @returns {Array<String|Object>} The attribute collection
         */
        attr: function (attr, ids) {
            var currentAttrs = this._attrs[attr];

            var removed = _.difference(currentAttrs, ids);
            var added = _.difference(ids, currentAttrs);

            this._attrs[attr] = ids;

            if (added.length > 0 || removed.length > 0) {
                this.trigger('attribute-updated', attr, this._attrs[attr], added, removed);
                this.trigger('attribute-updated:' + attr, this._attrs[attr], added, removed);
            }

            return this._attrs[attr];
        },

        //Methods relating to clearing datasets
        /**
         * Resets the collection. Clearing out all attributes.
         *
         * @method Collection#reset
         */
        reset: function () {
            this._attrs = _.mapValues(this._attrs, function () {
                return [];
            });
        },
        /**
         * Invalidates the entire collection. Resetting all attributes and all data (sorts, groups, filters, nodes).
         *
         * @method Collection#invalidate
         */
        invalidate: function () {
            this._nodes = {};
            this._filters = {};
            this._state = STATES.normal;
            this._sorts = {
                'default': {
                    ids: [],
                    order: null
                }
            };
            this._attrs = {
                hidden: [],
                disabled: [],
                unselectable: [],
                expanded: [],
                selected: [],
                required: []
            };

            this.trigger('invalidate');
            this.trigger('rerender');
        },

        /**
         * @method Collection#recache
         */
        recache: function () {
            var nodes = _.toArray(this._nodes);

            this.invalidate();
            this.addAll(nodes);
            this.trigger('rerender');
        },

        //Methods regarding state checks
        /**
         * Whether or not the collection is filtered.
         *
         * @method Collection#isFiltered
         * @returns {Boolean} whether it is currently filtered.
         */
        isFiltered: function () {
            return this._state === STATES.filtered;
        },

        /**
         * Whether the collection has a filter with data currently applied.
         *
         * @method Collection#hasFilterResults
         *
         * @returns {Boolean} whether the collection has a filter with data
         */
        hasFilterResults: function () {
            return _.get(this._filters, ACTIVE_QUERY_KEY, []).length > 0;
        },
        /**
         * Whether or not the collection matches the given state
         *
         * @method Collection#isState
         * @param {String} state The state to check.
         * @returns {Boolean} whether or not the collection matches that state.
         */
        isState: function (state) {
            return this._state === state;
        },

        //Methods manipulating active
        /**
         * Gets the currently marked active node
         *
         * @method Collection#getActive
         * @returns {String|Object} The current active node within the collection.
         */
        getActive: function () {
            return this._focused;
        },
        /**
         * Sets the current active node within the collection.
         *
         * @method Collection#setActive
         * @param {String|Object} id The id to mark as active.
         */
        setActive: function (id) {
            this._focused = id;
        },
        /**
         * Whether or not the node is the focused node.
         *
         * @method Collection#isActive
         * @param {String|Object} id The id to check if it is active
         */
        isActive: function (id) {
            return this._focused === id;
        },
        /**
         * Combined getter/setter for the active node within the collection.
         *
         * @method Collection#active
         * @param {Object|String} [id] The id to set if supplied.
         *
         * @fires Collection#active
         */
        active: function (id) {
            if (!arguments.length) {
                return this.getActive();
            }

            this.setActive(id);
            this.trigger('active', id);
        },
        /**
         * Moves the given ids by a given number.
         *
         * @method Collection#move
         * @param {Object|String|Array<Object|String>} ids The ids to move.
         * @param {Number} number The number to move the ids by.
         *
         * @fires Collection#node-moved
         */
        move: function (ids, number) {
            if (!_.isArray(ids)) {
                ids = [ids];
            }

            ids = _.sortBy(ids, function (id) {
                return this._sorts['default'].ids.indexOf(id);
            }.bind(this));

            if (number > 0) {
                ids.reverse();
            }

            _.each(ids, function (id) {
                var currentIndex = this._sorts['default'].ids.indexOf(this._getId(this.node(id)));
                var sortedIndex = -1;
                if ((currentIndex > 0 && number < 0) ||
                    (currentIndex < this._sorts['default'].ids.length && number > 0)) {
                    sortedIndex = math.mod(currentIndex + number, this._sorts['default'].ids.length);

                    this._sorts['default'].ids.splice(currentIndex, 1);
                    this._sorts['default'].ids.splice(sortedIndex, 0, id);

                    this.trigger('node-moved', _.last(this.getParents(id)) || null, id);
                }

            }.bind(this));
        },
        /**
         * Sets the given ids to a certain position within the collection.
         *
         * @method Collection#moveTo
         * @param {Array<String|Object>|String|Object} ids The ids to move
         * @param {Number} idx The index to add to.
         *
         * @fires Collection#node-moved
         */
        moveTo: function (ids, idx) {
            if (!_.isArray(ids)) {
                ids = [ids];
            }

            ids = _.sortBy(ids, function (id) {
                return this._sorts['default'].ids.indexOf(id);
            }.bind(this));

            //TODO: investigate when dragging with multiple levels of data
            _.each(ids, function (id, additive) {
                var currentIndex = this._sorts['default'].ids.indexOf(id);

                this._sorts['default'].ids.splice(currentIndex, 1);
                if (idx > currentIndex) {
                    this._sorts['default'].ids.splice(idx, 0, id);
                } else {
                    this._sorts['default'].ids.splice(idx + additive, 0, id);
                }

                this.trigger('node-moved', _.last(this.getParents(id)) || null, id);

            }.bind(this));
        },
        /**
         * @method Collection#getViewConfig
         *
         * @returns {Object}
         */
        getViewConfig: function () {
            return {
                attrs: _.cloneDeep(this._attrs),
                filters: _.cloneDeep(this._filters),
                sorts: _.cloneDeep(this._sorts),
                state: _.cloneDeep(this._state)
            };
        },
        /**
         * @method Collection#applyViewConfig
         *
         * @param {Object} viewConfig
         */
        applyViewConfig: function (viewConfig) {
            this._attrs = _.cloneDeep(viewConfig.attrs);
            this._sorts = _.cloneDeep(viewConfig.sorts);
            this._filters = _.cloneDeep(viewConfig.filters);

            this.trigger('rerender');
        },

        /**
         * Gets all the nodes that are reachable from the roots.
         *
         * @method Collection#getVisibleNodes
         *
         * @returns {Array<String>}
         */
        getVisibleNodes: function () {
            var stack;
            var visibleNodes = [];

            if (this._cachedNodes != null) {
                return this._cachedNodes;
            }

            stack = this.getRoots();

            if (this._options.flat) {
                var disabledHiddenNodes = [];
                if (this._options.disabledStyle === 'hidden') {
                    disabledHiddenNodes = this._attrs.disabled;
                }
                visibleNodes = _.difference(stack, disabledHiddenNodes, this._attrs.hidden);
            } else {
                while (stack.length > 0) {
                    var top = stack.shift();
                    var visible = this.isGroup(top) || (!this.is('hidden', top) && !(this._options.disabledStyle === 'hidden' && this.is('disabled', top)));

                    if (visible && this.is('expanded', top)) {
                        stack = this.getChildren(top).concat(stack);
                    }

                    if (visible) {
                        visibleNodes.push(top);
                    }
                }
            }

            this._cachedNodes = visibleNodes;
            return visibleNodes;
        },

        /**
         * @method Collection#clearVisibleNodes
         */
        clearVisibleNodes: function () {
            this._cachedNodes = null;
        }
    });

    return Collection;
});

define('oui/widgets/tree/treeController',[
    'underscore',
    'jquery'
], function (_, $) { //jshint ignore:line
    'use strict';

    /**
     * @event TreeModel#node-moved
     * @param {Object} parent
     * @param {Object} activeNode
     */
    /**
     * @event TreeModel#loading
     * @param {String} id The id of the node that is being loaded
     */
    /**
     * @event TreeModel#node-loaded
     * @param {Object} node The node that has finished loading
     */
    /**
     * @event TreeModel#node-loaded
     * @param {Object} node The node that has finished loading
     */
    /**
     * @event TreeModel#loading
     */

    var tcp;

    /**
     * The controller used for the tree panel.
     *
     * @class TreeController
     * @private
     * @since 1.2.0
     *
     * @param {TreeModel} model      The model that the controller will act on.
     * @param {Object}    options    The options for the tree. The controller only takes what it needs.
     */
    function TreeController(model, options) {

        /**
         * The model for the tree panel.
         *
         * @type {TreeModel}
         * @private
         * @memberof! TreeController.prototype
         */
        this._model = model;
        this._options = options;

        this._loadDeferreds = {};
        this._filteredDeferreds = {};
        this._linkedTrees = [];
        this._pivot = null;
        return this;
    }

    tcp = TreeController.prototype;

    tcp.constructor = TreeController;

    /**
     * Takes the id and expands it regardless of if it is already expanded.
     *
     * @method TreeController#expand
     * @param {Array} ids The id of the node to be expanded.
     */
    tcp.expanded = function (ids) {
        var that = this;
        if (!arguments.length) {
            return this._model.getAttr('expanded');
        }

        ids = _.filter(ids, this.hasChildren.bind(this));

        ids = this._model.attr('expanded', ids);

        _.each(ids, function (id) {
            if (!that._model.is('loaded', id)) {
                that.load(id);
            }
        });
    };

    /**
     * Takes an id and marks it as active.
     *
     * @method TreeController#active
     * @param {String} id The id of the node to mark as active.
     */
    tcp.active = function (id) {
        if (!arguments.length) {
            return this._model.getActive();
        }

        this._model.active(id);
    };

    /**
     * @method TreeController#moveUp
     */
    tcp.moveUp = function () {
        this._move(-1);
    };

    /**
     * @method TreeController#moveDown
     */
    tcp.moveDown = function () {
        this._move(1);
    };

    /**
     * @method TreeController#_move
     * @private
     *
     * @param {Number} number
     */
    tcp._move = function (number) {
        if (!this._model.isFiltered()) {
            this._model.move(this._model.selected(), number);
        }
    };

    /**
     * @method TreeController#_shift
     * @private
     *
     * @param {Number} number
     */
    tcp._shift = function (number) {
        if (!this._model.isFiltered()) {
            this._model.moveTo(this._model.selected(), number);
        }
    };

    /**
     * Loads a node if it is lazily loaded.
     *
     * @method TreeController#load
     * @param {String} id The id of the node to load it's children of.
     */
    tcp.load = function (id) {
        var that = this;
        var deferred;
        var loadFn;
        var node = this._model.node(id);

        if (this._options.lazyLoad && (this._model.isState('initialized') || !this._model.is('loaded', id))) {

            loadFn = this._options.storeFns.load;
            that._model.trigger('loading', id);
            deferred = loadFn.call(this, id, node);

            if (deferred && deferred.promise && deferred.abort) {
                this._loadDeferreds[id] = deferred;
            }

            $.when(deferred).done(function (data) {
                var groupBy = _.find(that._options.groupBy, function (group) {
                    return group.asTree;
                });
                data = _.map(data, function (node) {
                    node = _.clone(node);
                    if (groupBy && !_.has(node, groupBy.field)) {
                        this._model.addField(node, groupBy.field);
                        node[groupBy.field] = id;
                    }

                    return node;
                }.bind(that));
                that._model.addAll(data);
                that._model.addToAttr('loaded', id);
            }).fail(function (jqXHR, args) {
                if (!(args && args.filtered)) {
                    that._model.addToAttr('loaded', id);
                } else {
                    that._model.expanded(_.without(that._model.expanded(), id));
                }
            }).always(function () {
                delete that._loadDeferreds[id];
            });
        }
    };

    /**
     * @method TreeController#selected
     *
     * @returns {Array<String>}
     */
    /**
     * @method TreeController#selected
     *
     * @param {Array<String>} ids
     * @param {Subscribable~EventOptions} eventParams
     */
    tcp.selected = function (ids, eventParams) {
        if (!arguments.length) {
            var selected = this._model.selected();

            return _.chain(selected)
                .map(function (node) {
                    if (this.getNodeWithAttributes(node).attrs.group) {
                        return this._model.matches(node.flatten());
                    }
                    return node;
                }.bind(this))
                .flatten()
                .uniq()
                .value();
        }

        this._model.selected(ids, eventParams);

        var groups = _.filter(ids, _.isObject);
        _.each(groups, this.toggle.bind(this, 'expanded'));
    };

    /**
     * Takes a query and filters the data based on the fields given.
     * Currently only supports filtering already loaded data.
     *
     * @method TreeController#filter
     * @param {String} query The string to match against.
     * @param {Array} fields The fields to check.
     * @param {Object} scope
     */
    tcp.filter = function (query, fields, scope) {
        var that = this;

        if ((_.isNull(query) || query === '') && scope == null) {
            if (this._model.isFiltered()) {
                this._model.removeFilter();
            }

            _.each(this._filteredDeferreds, function (jqXHR) {
                jqXHR.abort({
                    filtered: false
                });
            });
        } else if (this._options.lazyLoad || !this._options.clientFilter) {
            var filterFn = this._options.storeFns.filter;
            this._model.trigger('loading', null);
            _.each(this._loadDeferreds, function (jqXHR) {
                jqXHR.abort({
                    filtered: true
                });
            });

            var deferred = filterFn.call(this, query, fields, scope);

            if (deferred && deferred.promise && deferred.abort) {
                this._filteredDeferreds[query] = deferred;
            }
            $.when(deferred).done(function (data) {
                that._model.addAsFilter(data);
            }).fail(function () {
                that._model.removeFilter();
            }).always(function () {
                delete that._filteredDeferreds[query];
            });
        } else {
            this._model.addFilter(query);
        }
    };

    /**
     * Takes data and adds it to the proper location in the tree.
     *
     * @method TreeController#add
     * @param {Array} data The nodes to be added. The nodes to be added should have the information regarding
     * their groups already within their data.
     * @param {Number} [elder] The sibling to add under.
     */
    tcp.add = function (data, elder) {
        if (_.isArray(data)) {
            this._model.addAll(data, elder);
        } else {
            this._model.add(data, elder);
        }
    };

    /**
     * Removes the subtree from the model.
     *
     * @method TreeController#remove
     * @param {String|Object} data The id of the subtree to remove.
     */
    tcp.remove = function (data) {
        if (_.isArray(data)) {
            _.each(data, this.remove.bind(this));
        } else if (_.isObject(data)) {
            var node = this.node(data);

            this._model.remove(_.get(node, this._options.recordKeys.id));
        } else {
            //This should be the id of the node to remove (client-side id)
            this._model.remove(data);
        }
    };

    /**
     * @method TreeController#tree
     *
     * @param {Object} modifiers
     *
     * @returns {Object}
     */
    tcp.tree = function (modifiers) {
        var roots = _.map(this._model.getRoots(), function (id) {
            var children = _.map(this._model.getChildren(id), function (id) {
                return this.node(id, modifiers);
            }.bind(this));
            if (children.length > 0) {
                return _.extend({}, this.node(id, modifiers), {
                    children: children
                });
            } else {
                return _.extend({}, this.node(id, modifiers));
            }
        }.bind(this));

        if (modifiers.showRoot) {
            return { children: roots };
        } else {
            return roots;
        }
    };

    /**
     * @method TreeController#nodes
     *
     * @param {Object} modifiers
     *
     * @returns {Array<Object>}
     */
    tcp.nodes = function (modifiers) {
        return this._model.nodes(modifiers);
    };

    /**
     * @method TreeController#node
     *
     * @param {String} id
     * @param {Object} modifiers
     *
     * @returns {Object}
     */
    tcp.node = function (id, modifiers) {
        return this._model.node(id, modifiers);
    };

    /**
     * @method TreeController#disabled
     *
     * @returns {Array<String>}
     */
    /**
     * @method TreeController#disabled
     *
     * @param {Array<String>} nodes
     */
    tcp.disabled = function (nodes) {
        if (!arguments.length) {
            return this._model.getAttr('disabled');
        }

        this._model.attr('disabled', _.pluck(nodes, this._options.recordKeys.id));
    };

    /**
     * @method TreeController#required
     *
     * @returns {Array<String>}
     */
    /**
     * @method TreeController#required
     *
     * @param {Array<String>} nodes
     */
    tcp.required = function (nodes) {
        if (!arguments.length) {
            return this._model.getAttr('required');
        }

        this._model.attr('required', nodes);
    };

    /**
     * @method TreeController#unselectable
     *
     * @returns {Array<String>}
     */
    /**
     * @method TreeController#unselectable
     *
     * @param {Array<String>} nodes
     */
    tcp.unselectable = function (nodes) {
        if (!arguments.length) {
            return this._model.getAttr('unselectable');
        }

        this.selected(_.difference(this.selected(), nodes));
        this._model.attr('unselectable', nodes);
    };

    /**
     * @method TreeController#invalidate
     */
    tcp.invalidate = function () {
        this._model.invalidate();

        this.load(null);
    };

    /**
     * @method TreeController#updateNodes
     *
     * @param {Array<Object>} nodes
     */
    tcp.updateNodes = function (nodes) {
        var that = this;
        var isArray = _.isArray(nodes);

        _.each(nodes, function (node, id) {
            var actualNode;

            //Sample Idea for a better way of getting the nodes
            if (!isArray) {
                actualNode = that.node(id);
            } else {
                actualNode = that.node(node);
            }
            that._model.update(_.get(actualNode, that._options.recordKeys.id), node);
        });
    };

    /**
     * @method TreeController#isFiltered
     *
     * @returns {Boolean}
     */
    tcp.isFiltered = function () {
        return this._model.isFiltered();
    };

    /**
     * @method TreeController#hasFilterResults
     *
     * @returns {Boolean}
     */
    tcp.hasFilterResults = function () {
        return this._model.hasFilterResults();
    };

    /**
     * @method TreeController#getVisibleNodes
     *
     * @returns {Array}
     */
    tcp.getVisibleNodes = function () {
        return this._model.getVisibleNodes();
    };

    /**
     * @method TreeController#filtered
     *
     * @returns {Array}
     */
    tcp.filtered = function () {
        return [];
    };

    /**
     * @method TreeController#linked
     *
     * @returns {Array<TreePanel>}
     */
    /**
     * @method TreeController#linked
     *
     * @param {Array<TreePanel>} linkTrees
     */
    tcp.linked = function (linkTrees) { //jshint ignore:line
        if (!arguments.length) {
            return this._linkedTrees;
        }

        this._linkedTrees = linkTrees;
    };

    /**
     * @method TreeController#reset
     *
     * @param {Object} attrs
     */
    tcp.reset = function (attrs) {
        this._model.reset(attrs);
    };

    /**
     * @method TreeController#roots
     *
     * @returns {Array<Object>}
     */
    tcp.roots = function () {
        return this._model.getRoots();
    };

    /**
     * @method TreeController#isExpanded
     *
     * @param {String} id
     *
     * @returns {Boolean}
     */
    tcp.isExpanded = function (id) {
        return this._model.is('expanded', id);
    };

    /**
     * @method TreeController#hasChildren
     *
     * @param {String} id
     *
     * @returns {Boolean}
     */
    tcp.hasChildren = function (id) {
        if (id == null) {
            return true;
        }

        if (this._model.isGroup(id)) {
            return true;
        }

        if (this._options.lazyLoad) {
            if (this._model.is('loaded', id)) {
                return this._model.getChildren(id).length > 0;
            }

            if (_.has(this.node(id), 'hasChildren')) {
                return this.node(id).hasChildren;
            }
            return true;
        } else {
            return this._model.getChildren(id).length > 0;
        }
    };

    /**
     * @method TreeController#isDisabled
     *
     * @param {String} id
     *
     * @returns {Boolean}
     */
    tcp.isDisabled = function (id) {
        return this._model.is('disabled', id, _.get(this._options, 'attrs.disabled.nested', true));
    };

    /**
     * @method TreeController#isRequired
     *
     * @param {String} id
     *
     * @returns {Boolean}
     */
    tcp.isRequired = function (id) {
        return this._model.is('required', id);
    };

    /**
     * @method TreeController#isSelectable
     *
     * @param {String} id
     *
     * @returns {Boolean}
     */
    tcp.isSelectable = function (id) {
        return !this._model.is('unselectable', id);
    };

    /**
     * @method TreeController#isSelected
     *
     * @param {String} id
     *
     * @returns {Boolean}
     */
    tcp.isSelected = function (id) {
        return this._model.is('selected', id);
    };

    /**
     * @method TreeController#children
     *
     * @param {String} id
     *
     * @returns {Array<Object>}
     */
    tcp.children = function (id) {
        return this._model.getChildren(id);
    };

    /**
     * @method TreeController#isLoaded
     *
     * @returns {Boolean}
     */
    tcp.isLoaded = function () {
        return !this._options.lazyLoad || (this._model.isState('normal') || this._model.isState('filtered'));
    };

    /**
     * @method TreeController#level
     *
     * @param {String} id
     * @returns {Number}
     */
    tcp.level = function (id) {
        return this._model._getLevel(id) - 1;
    };

    /**
     * @method TreeController#getNodeWithAttributes
     *
     * @param {String} id
     *
     * @returns {Object}
     */
    tcp.getNodeWithAttributes = function (id) {
        if (_.isObject(id) && id.group) {
            return {
                attrs: {
                    unselectable: false,
                    group: true,
                    hasChildren: true,
                    visible: _.some(this._model.matches([id]), function (children) {
                        return !this.isDisabled(children);
                    }.bind(this)),
                    expanded: this.isExpanded(id),
                    level: this._model._getLevel(id) - 1,
                    displayValue: this._model.getGroupDisplayValue(id)
                },
                data: {
                    field: id.field,
                    key: id.value
                }
            };
        }

        var data = this._model.node(id);
        var attrs = this._model.getAttributesForNode(id);

        attrs.active = this._model.isActive(id);
        attrs.level = this._model._getLevel(id) - 1;

        if (attrs.empty || this._options.flat) {
            attrs.hasChildren = false;
        } else {
            attrs.hasChildren = this.hasChildren(id);
        }

        return {
            data: data,
            attrs: attrs
        };
    };

    /**
     * @method TreeController#getParents
     *
     * @param {String} id
     * @returns {Array<String>}
     */
    tcp.getParents = function (id) {
        return this._model.getParents(id);
    };

    /**
     * @method TreeController#getParent
     *
     * @param {String} id
     *
     * @returns {String}
     */
    tcp.getParent = function (id) {
        return this._model.getParent(id);
    };

    /**
     * @method TreeController#recache
     */
    tcp.recache = function () {
        this._model.recache();
    };

    /**
     * @method TreeController#getViewConfig
     *
     * @returns {Object}
     */
    tcp.getViewConfig = function () {
        return this._model.getViewConfig();
    };

    /**
     * @method TreeController#applyViewConfig
     *
     * @param {Object} viewConfig
     */
    tcp.applyViewConfig = function (viewConfig) {
        this._model.applyViewConfig(viewConfig);
    };

    /**
     * @method TreeController#pivot
     *
     * @returns {String?}
     */
    /**
     * @method TreeController#pivot
     *
     * @param {String} pivot
     */
    tcp.pivot = function (pivot) {
        if (!arguments.length) {
            return this._pivot;
        }

        this._pivot = pivot;
    };

    /**
     * @method TreeController#toggle
     *
     * @param {String} attr
     * @param {String} id
     */
    tcp.toggle = function (attr, id) {
        switch (attr) {
            case 'selected':
                if (this._model.is(attr, id)) {
                    if (!this._options.required || this.selected().length > 1) {
                        this.selected(_.without(this.selected(), id));
                    }
                } else if (this._options.multiSelect) {
                    this.selected(_.union(this.selected(), [id]));
                } else {
                    this.selected([id]);
                }
                break;
            case 'expanded':
                if (this._model.is('expanded', id)) {
                    this._model.removeFromAttr('expanded', id);
                } else {
                    this._model.addToAttr('expanded', id);
                    if (!this._model.is('loaded', id)) {
                        this.load(id);
                    }
                }
                break;
            default:
                if (this._model.is(attr, id)) {
                    this._model.removeFromAttr(attr, id);
                } else {
                    this._model.addToAttr(attr, id);
                }
        }
    };

    /**
     * @method TreeController#clearVisibleNodes
     */
    tcp.clearVisibleNodes = function () {
        this._model.clearVisibleNodes();
    };

    return TreeController;

});


define('template-20', ['underscore'], function (_) { return function(data) {
var __t, __p = '', __j = Array.prototype.join;
function print() { __p += __j.call(arguments, '') }
__p += '<div role="tree" class="orcl-widget tree" tabindex="0"> ';
 if (data.columns.length > 0) { ;
__p += ' <div class="headers"> ';
 for (var i = 0, len = data.columns.length; i < len; i++) { ;
__p += ' <div class="col" title="' +
((__t = ( data.columns[i].display )) == null ? '' : __t) +
'" style="flex-basis: ' +
((__t = ( data.columns[i].width )) == null ? '' : __t) +
'"> ' +
((__t = ( data.columns[i].display )) == null ? '' : __t) +
' </div> ';
 } ;
__p += ' </div> ';
 } ;
__p += ' <div class="scroll-container"> <div class="scroller"></div> </div> <div class="pseudo-scroller vertical"></div> <div class="tree-items"> <div class="filter-notif" aria-live="polite" aria-atomic="true"></div> <div class="no-results" aria-live="polite" aria-atomic="true">' +
((__t = ( data.locale.no_results_found )) == null ? '' : __t) +
'</div> <i title="' +
((__t = ( data.locale.loading )) == null ? '' : __t) +
'" aria-live="polite" aria-atomic="true" class="loader pgbu-icon pgbu-icon-loading"></i> <ul class="root" aria-live="polite"> </ul> <div class="empty-tree">' +
((__t = ( data.locale.empty_tree )) == null ? '' : __t) +
'</div> </div> <div class="disable-overlay"></div></div>';
return __p
};});


define('template-21', ['underscore'], function (_) { return function(data) {
var __t, __p = '', __j = Array.prototype.join;
function print() { __p += __j.call(arguments, '') }

    var selected = data.attrs.selected ? 'selected' : null;    var disabled = data.attrs.disabled ? 'tree-disabled': null;    var unselectable = data.attrs.unselectable ? 'unselectable' : null;    var active = data.attrs.active ? 'active': null;    var expanded = data.attrs.expanded ? 'expanded' : 'collapsed';    var hidden = data.attrs.hidden ? 'invisible': null;    var titleString = data.collapsed ? 'collapsed' : 'expanded';    var withColumns = data.columns ? 'w-columns' : '';    var flat = data.flat ? 'flat' : null;    var empty = data.attrs.empty || !data.attrs.hasChildren ? 'empty' : null;    var inlineStyle = (data.flat == null || data.flat === false) ? 'style="padding-left: ' + (data.attrs.level * 15) + 'px;"' : '';    var ariaExpanded = '';    if (!data.flat && (!data.attrs.empty || !data.attrs.hasChildren)) {        if (data.attrs.expanded) {            ariaExpanded = 'aria-expanded="true"';        } else {            ariaExpanded = 'aria-expanded="false"';        }    };
__p += '<li ' +
((__t = ( !data.columns ? inlineStyle : '' )) == null ? '' : __t) +
' role="treeitem" aria-selected="' +
((__t = ( data.attrs.selected )) == null ? '' : __t) +
'" ' +
((__t = ( ariaExpanded )) == null ? '' : __t) +
' id="' +
((__t = ( data.pageId )) == null ? '' : __t) +
'" data-treeid="' +
((__t = ( data.id)) == null ? '' : __t) +
'"' +
((__t = ( data.group ? 'data-tree-group=true' : '' )) == null ? '' : __t) +
'class="' +
((__t = ( _.compact(['item', selected, disabled, unselectable, active, expanded, hidden, flat, empty, withColumns]).join(' ') )) == null ? '' : __t) +
'"> ' +
((__t = ( data.itemMarkup )) == null ? '' : __t) +
'</li>';
return __p
};});

/**
 * @module arrayUtils
 */
define('oui/utils/arrayUtils',[
], function () {
    'use strict';

    /**
     * @exports arrayUtils
     */
    return {
        /**
         * Rotates a method around the index given.
         *
         * If idx is `null`, `undefined` or a negative number the original array is returned.
         * Otherwise, the slice of the array before the index inclusive will be moved to the end.
         *
         * Returns a shallow copy of the array due to slice operations when the operation is performed.
         *
         * @function rotate
         *
         * @param {Array} arr The array to rotate
         * @param {Number?} idx The index to pivot the array on
         *
         * @returns {Array}
         *
         * @example
         *
         *     var a = arrayUtils.rotate([1, 2, 3, 4, 5, 6], 1); // returns [3, 4, 5, 6, 1, 2]
         */
        rotate: function (arr, idx) {
            if (idx == null || idx < 0) {
                return arr;
            }
            return arr.slice(idx + 1).concat(arr.slice(0, idx + 1));
        }
    };
});

/**
 * @module globalListeners
 * @private
 */
define('oui/utils/globalListeners',[
    'jquery',
    'underscore'
], function ($, _) {
    'use strict';

    var events = {};

    /**
     * @method addListener
     * @private
     *
     * @param {EventTarget} elem
     * @param {String} eventName
     * @param {Function} cb
     */
    function addListener(elem, eventName, cb) {
        if (events[eventName] == null) {
            events[eventName] = [].concat(cb);
            $(elem).on(eventName, function (event) {
                _.each(events[eventName], function (fn) {
                    fn(event);
                });
            });
        } else {
            events[eventName].push(cb);
        }
    }

    /**
     * @exports globalListeners
     */
    return {
        /**
         * @method addDocumentListener
         *
         * @param {String} eventName
         * @param {Function} cb
         */
        addDocumentListener: _.partial(addListener, document),
        /**
         * @method addWindowListener
         *
         * @param {String} eventName
         * @param {Function} cb
         */
        addWindowListener: _.partial(addListener, window),
        /**
         * @method clear
         */
        clear: function () {
            _.chain(events)
                .toArray()
                .flatten()
                .each(function (cb) {
                    if (cb.cancel) {
                        cb.cancel();
                    }
                })
                .run();
            events = {};
        }
    };
});

define('oui/utils/scrollbarWidth',[], function () {
    'use strict';

    /**
     * Determines the browser specific scrollbar width.
     *
     * @module scrollbarWidth
     * @private
     */

    /**
     * Determines what the width of the scrollbar would be if it is needed.
     *
     * @exports getScrollbarWidth
     * @private
     * @type Function
     *
     * @returns {Number}
     */
    return function () {
        var outer = document.createElement('div');
        outer.style.visibility = 'hidden';
        outer.style.position = 'absolute';
        outer.style.top = '-10000px';
        outer.style.left = '-10000px';
        outer.style.width = '100px';

        document.body.appendChild(outer);

        var widthNoScroll = outer.offsetWidth;
        // force scrollbars
        outer.style.overflow = 'scroll';

        // add innerdiv
        var inner = document.createElement('div');
        inner.style.width = '100%';
        outer.appendChild(inner);

        var widthWithScroll = inner.offsetWidth;

        // remove divs
        outer.parentNode.removeChild(outer);

        return widthNoScroll - widthWithScroll;
    };
});

define('oui/widgets/tree/treeView',['require','underscore','jquery','oui/base/view','template-20','template-21','oui/utils/keyCode','oui/utils/stringUtils','oui/widgets/tree/groupingUtils','oui/utils/math','oui/utils/arrayUtils','logger','oui/utils/globalListeners','oui/utils/scrollbarWidth'],function (require) { //jshint ignore:line
    'use strict';

    var _ =  require('underscore');
    var $ = require('jquery');
    var View = require('oui/base/view');
    var treeTemplate = require('template-20');
    var nodeTemplate = require('template-21');
    var keyCode = require('oui/utils/keyCode');
    var stringUtils = require('oui/utils/stringUtils');
    var groupingUtils = require('oui/widgets/tree/groupingUtils');
    var math = require('oui/utils/math');
    var arrayUtils = require('oui/utils/arrayUtils');
    var logger = require('logger');
    var globalListeners = require('oui/utils/globalListeners');
    var getScrollbarWidth = require('oui/utils/scrollbarWidth');

    var tvp;

    var DRAG_THRESHOLD_PX = 10;

    /**
     * Checks if the coordinates are within the boundaries of the element.
     *
     * @method TreeView~isContained
     * @private
     *
     * @param {jQuery.Event} event
     * @param {jQuery} $el
     *
     * @returns {Boolean}
     */
    function isContained(event, $el) {
        return !((event.pageX < $el.offset().left ||
                event.pageX > $el.offset().left + $el.width()) ||
                    (event.pageY < $el.offset().top ||
                     event.pageY > $el.offset().top + $el.height()));
    }

    /**
     * The view that constructs the element and binds all the events for the TreePanel.
     *
     * @class TreeView
     * @private
     * @extends View
     * @since 1.2.0
     *
     * @param {TreeModel} model The data for the tree panel.
     * @param {TreeController} controller The object to accept all UI calls.
     * @param {TreePanel#Config} options Object that contains all the functions that can be overrideable.
     * @param {Object} settings
     */
    function TreeView(model, controller, options, settings) {
        var that = this;

        View.call(this, options, null, controller, settings);

        //TODO: look into better way of doing this
        //need to wait until after buildElement and id are called
        this._$el.find('.no-results').attr('id', this.id() + '-no-results');

        this._cachedHeight = 0;

        this._rowHeightValidated = false;
        this._renderedNodeIds = [];
        this._renderRequest = null;

        this._addGlobalListeners();

        model.on({
            'view.load': this.render.bind(this),
            'view.remove': this.render.bind(this),
            'view.attribute-updated:expanded': this.render.bind(this),
            'view.attribute-updated:disabled': this.render.bind(this),
            'view.attribute-updated:unselectable': this.selectable.bind(this),
            'view.change': this.select.bind(this),
            'view.active': this.active.bind(this),
            'view.attribute-updated:loaded': this.render.bind(this),
            'view.loading': this.loading.bind(this),
            'view.addAll': this.render.bind(this),
            'view.add': this.render.bind(this),
            'view.filter': this.filter.bind(this),
            'view.invalidate': this.render.bind(this),
            'view.node-moved': this.render.bind(this),
            'view.reset': this.render.bind(this),
            'view.rerender': this.render.bind(this),
            'view.group-removed': this.groupRemoved.bind(this)
        });

        if (this._controller.isLoaded()) {
            this.render();
        }

        this._$el.on('click', '.unselectable', function (e) {
            e.preventDefault();
            e.stopPropagation();
            controller.toggle('expanded', that._getTreeId($(this)));
        });

        this._$el.on('click', '[data-attr][data-toggle=true]', function (event) {
            var attr = event.currentTarget.getAttribute('data-attr');
            event.preventDefault();
            event.stopPropagation();
            controller.toggle(attr, that._getTreeId($(event.target).closest('[data-treeid]')));
        });

        this._$scrollContainer = this._$el.find('.scroll-container');
        this._$el.on('wheel', function (event) {
            that._$scrollContainer.scrollTop(that._$scrollContainer.scrollTop() + (math.sign(event.originalEvent.deltaY) * 53));
        });

        var throttledRender = _.throttle(this.render.bind(this, { quick: true }), 0);
        this._$scrollContainer.on('scroll', throttledRender);

        this._$el.on('scroll wheel', function (event) {
            if (this._$scrollContainer[0].scrollHeight > this._$scrollContainer[0].clientHeight) {
                event.preventDefault();
                event.stopPropagation();
            }
        }.bind(this));

        this._$scrollContainer.css({
            width: getScrollbarWidth() + 1
        });

        if (window.MutationObserver) {
            var mutationObserver = new window.MutationObserver(function () {
                if (this._$el[0].parentNode != null) {
                    this.render();
                }
            }.bind(this));

            mutationObserver.observe(this._$el[0], {
                attributes: true,
                attributeOldValue: true,
                attributeFilter: ['style']
            });
        }

        this._$el.on('click', '[data-trigger]', function (e) {
            e.preventDefault();
            e.stopPropagation();
            var $element = $(this);
            var id = $element.parents('li').attr('data-treeid');
            var triggerName = $element.attr('data-trigger');
            that.trigger('data-trigger-' + triggerName, $(this), controller.node(id));
            that.trigger(triggerName, $(this), controller.node(id));
        });

        this._$el.on('click', '.filter-notif', function () {
            that._controller.filter(null);
        });

        this.addShortcuts({
            'PAGEUP': _.partial(this.page, -1),
            'PAGEDOWN': _.partial(this.page, 1),
            'UP': _.partial(this._moveActive, -1),
            'DOWN': _.partial(this._moveActive, 1),
            'LEFT': that._handleCollapse,
            'RIGHT': that._handleExpand,
            'HOME': that._top,
            'END': that._bottom,
            'ESC': {
                fn: function () { //jscs:ignore jsDoc
                    if (that._controller.isFiltered()) {
                        that._controller.filter(null);
                    }
                },
                stopPropagation: that._controller.isFiltered.bind(that._controller)
            },
            'SPACE': that._select,
            'shift-UP': _.partial(this._extendSelection, -1),
            'shift-DOWN': _.partial(this._extendSelection, 1),
            'ctrl-UP': _.partial(this._moveActive, -1),
            'ctrl-DOWN': _.partial(this._moveActive, 1),
            'ctrl-shift-UP': {
                fn: this.moveUp,
                stopPropagation: true
            },
            'ctrl-shift-DOWN': {
                fn: this.moveDown,
                stopPropagation: true
            },
            'ctrl-SPACE': function () { //jscs:ignore jsDoc
                that._controller.toggle('selected', that._controller.active());
            },
            'ctrl-F1': function (event) {
                event.preventDefault();
                if (this._options.viewFns.actions) {
                    this._findTree(this._controller.active()).find('[data-trigger]:eq(0)').click();
                }
            }.bind(this)
        });

        this._$el
            .on('keydown keyup', function (event) {
                if (that._options.multiSelect && event.keyCode === keyCode.SHIFT) {
                    //Put on disable select class
                    $('body').toggleClass('oui-disable-select', event.type === 'keydown');
                }
            }).on('blur', function () {
                $('body').removeClass('oui-disable-select');
            }).on('keypress', function (event) {
                var string = String.fromCharCode(event.keyCode);
                if (string.length > 0) {
                    that._find(string);
                }
            }).on('focus', '.tree-items', function () {
                that._$el.focus();
            }).on('click', '.item:not(".tree-disabled"):not(".unselectable")', function (event) {
            event.preventDefault();

            var id = that._getTreeId($(this));
            var $target = $(event.target);
            var prevSelected = that._controller.active();
            var nodes;

            var selectedIndex;
            var prevSelectedIndex;
            var selected;

            if (!$target.is('[data-attr]') && !$target.is('[data-trigger]')) {
                if (event.ctrlKey || that._$el.hasClass('mobile') || event.metaKey) { //Look at metaKey for MAC command click
                    that._controller.toggle('selected', id);
                } else if (event.shiftKey && that._options.multiSelect) {
                    nodes = that._controller.getVisibleNodes();
                    selectedIndex = nodes.indexOf(id);
                    prevSelectedIndex = nodes.indexOf(prevSelected);

                    if (selectedIndex < prevSelectedIndex) {
                        selected = nodes.slice(selectedIndex, prevSelectedIndex + 1);
                    } else {
                        selected = nodes.slice(prevSelectedIndex, selectedIndex + 1);
                    }

                    controller.selected(selected);
                } else {
                    controller.selected([id]);
                }
            }
        })
        .on('dblclick', '.item:not(".tree-disabled"):not(".unselectable")', function () {
            var id = that._getTreeId($(this));
            that.trigger('dblclick', $(this), controller.node(id));

            /**
             * @method TreeView~stopContinuing
             * @private
             */
            function stopContinuing() {
                that._$el.off('.continue');
            }

            var timeout = setTimeout(stopContinuing, 200);

            that._$el.on('click.continue', '.item:not(".tree-disabled"):not(".unselectable")', function () {
                var id = that._getTreeId($(this));
                if (id != null) {
                    that.trigger('dblclick', $(this), controller.node(id));
                }

                //Reset current timeout
                clearTimeout(timeout);
                timeout = setTimeout(stopContinuing, 200);
            });
        })
        .on('mousedown', '.item', function (event) {
            if (!(event.shiftKey || event.ctrlKey || event.metaKey)) {
                var id = that._getTreeId($(this));
                that._controller.active(id);
            }

            if (!($(event.target).attr('tabindex') != null ||
                    $(event.target).is('button') ||
                    $(event.target).is('a') ||
                    $(event.target).attr('draggable') != null)) {
                //Prevents scrolling on drag/drop when overflowing
                event.preventDefault();

                if (!that._$el.is(document.activeElement)) {
                    //See: https://connect.microsoft.com/IE/Feedback/Details/1028411 for the focus bug in IE
                    //Need to refocus tree without this IE will give focus to div
                    that._$el[0].focus();
                }

            }
        })
        .on('mousedown', '.item:not(".tree-disabled"):not(".unselectable")', function (event) {
            var dragging = false;
            var internal = true;
            var linkedTrees = that._controller.linked();
            var original = { x: event.clientX, y: event.clientY };
            var $target;
            var $dropTarget;

            var nodes;
            if (that._controller.selected().length === 0) {
                nodes = [that._controller.active()];
                $target = that._$el.find('.active');
            } else {
                nodes = that._controller.selected();
                $target = that._$el.find('.selected');
            }

            $dropTarget = $('<li class="drop-target" style="height: ' + ($target.height() * $target.length) + 'px;"></li>');

            if (that._options.drag.enabled && event.button === 0) {
                $('body').addClass('oui-disable-select');

                $(document).on('mousemove.dragging', function (event) {
                    if (math.distance({ x: event.clientX, y: event.clientY }, original) > DRAG_THRESHOLD_PX || dragging) {
                        dragging = true;

                        that._controller.selected(nodes);

                        if (!$(event.target).is($target)) {
                            that._controller.expanded(_.without(that._controller.expanded(), that._getTreeId($target)));

                            var offset = that._$el.offset().left;

                            var items = that._$el.find('.tree-items');
                            if (items[0].scrollHeight > items.height() &&
                               event.pageY < that._$el.offset().top + 15) {
                                items.scrollTop(items.scrollTop() - 15);
                            }

                            if (items[0].scrollHeight > items.height() &&
                               event.pageY > that._$el.offset().top + that._$el.height() - 15) {
                                items.scrollTop(items.scrollTop() + 15);
                            }

                            if (!isContained(event, that._$el)) {
                                offset = event.pageX - $target.width() / 2;
                                internal = false;
                            } else {
                                internal = true;
                            }

                            _.each($target, function (el, index) {
                                $(el).addClass('dragging')
                                .css({
                                    width: that._$el.width(),
                                    top: event.pageY - ($(el).height() / 2) + $(el).height() * index,
                                    left: offset
                                });

                            });

                            if (!internal) {
                                $dropTarget.remove();
                                _.each(linkedTrees, function (tree) {
                                    tree.sync(event);
                                });
                            }
                        }
                    }
                });

                $(document).on('mousemove.dragging', 'li:not(.drop-target)', function (event) {
                    var $over = $(event.target);

                    var append = Math.abs($over.offset().top - event.pageY) < ($over.height() / 2);
                    if (internal && that._options.drag.over === 'move' &&
                        (math.distance({ x: event.clientX, y: event.clientY }, original) > DRAG_THRESHOLD_PX || dragging)) {
                        if (append) {
                            $over.closest('li').after($dropTarget);
                        } else {
                            $over.closest('li').before($dropTarget);
                        }
                    }
                });

                $(document).one('mouseup.dragging mouseleave.dragging', function () {
                    $(document).off('.dragging');
                    $('body').removeClass('oui-disable-select');

                    if (!internal) {
                        _.each(linkedTrees, function (tree) {
                            tree.drop(_.map(that._controller.selected(), that._controller.node.bind(that._controller)));
                        });
                    }

                    if (dragging) {

                        $target.removeClass('dragging').css({
                            width: ''
                        });

                        if (internal && that._options.drag.over === 'move') {
                            var items = $target.siblings();
                            var moveNumber = items.index($dropTarget);

                            that._controller._shift(moveNumber);
                        } else if (that._options.drag.out === 'remove') {
                            that._controller.remove(that._controller.selected());
                        }
                    }

                    if ($dropTarget) {
                        $dropTarget.remove();
                    }
                });
            }
        });

        return this;
    }

    TreeView.prototype = Object.create(View.prototype);
    tvp = TreeView.prototype;

    tvp.constructor = TreeView;

    /**
     * @method TreeView#_addGlobalListeners
     */
    tvp._addGlobalListeners = function () {
        globalListeners.addDocumentListener('shown.virtualized-tree', function () {
            if (document.body.contains(this.element()[0])) {
                this.render();
            }
        }.bind(this));

        globalListeners.addWindowListener('resize.virtualized-tree', _.debounce(function () {
            if (this._$el[0].parentNode != null) {
                this.resize();
            }
        }.bind(this), 50, { leading: true }));
    };

    /**
     * @method TreeView#_findTree
     * @private
     *
     * @param {String|Array<String>} id
     *
     * @returns {jQuery}
     */
    tvp._findTree = function (id) {
        if (_.isArray(id)) {
            var stringIds = id.map(String); //Going to be comparing against the DOM which can only be strings
            var matched = _.filter(this._$el[0].getElementsByTagName('li'), function (node) {
                return stringIds.indexOf(node.getAttribute('data-treeid')) !== -1;
            });

            return $(matched);
        } else {
            return this._$el.find('[data-treeid="' + id + '"]');
        }
    };

    /**
     * @method TreeView#_getTreeId
     * @private
     *
     * @param {jQuery} elem
     *
     * @returns {String}
     */
    tvp._getTreeId = function (elem) {
        if (elem.attr('data-tree-group') != null) {
            return groupingUtils.inflate(elem.attr('data-treeid'));
        }

        var id = elem.closest('[data-treeid]').attr('data-treeid');

        return _.get(this._controller.node(id), this._options.recordKeys.id);
    };

    /**
     * @method TreeView#moveUp
     */
    tvp.moveUp = function () {
        if (this._options.move) {
            this._controller.moveUp();
        }
    };

    /**
     * @method TreeView#moveDown
     */
    tvp.moveDown = function () {
        if (this._options.move) {
            this._controller.moveDown();
        }
    };

    /**
     * @method TreeView#page
     *
     * @param {Number} delta
     */
    tvp.page = function (delta) {
        var renderedNodeIds = this.getRenderedNodeIds();

        this._$scrollContainer.scrollTop(this._$scrollContainer.scrollTop() + (delta * (renderedNodeIds.length * this._options.rowHeight)));
    };

    /**
     * @method TreeView#sync
     *
     * @param {jQuery.Event} event
     */
    tvp.sync = function (event) {
        this._$el.find('.drop-target').remove();

        var offset = this._$el.offset();
        var bottom = {};

        bottom.bottom = offset.top + this._$el.height();
        bottom.right = offset.left + this._$el.width();

        if ((event.clientX < bottom.right &&
                    event.clientX > offset.left) &&
                (event.clientY < bottom.bottom &&
                 event.clientY > offset.top)) {
            this._over = true;
            this._$el.find('.root').append($('<li class="drop-target" style="height: 22px;"></li>'));
        } else {
            this._over = false;
        }
    };

    /**
     * @method TreeView#drop
     * @param {Array<Object>} nodes
     */
    tvp.drop = function (nodes) {
        this._$el.find('.drop-target').remove();

        if (this._over) {
            //TODO: determine how to properly drop nodes
            this.trigger('nodes-dropped', nodes);
        }

        this._over = false;
    };

    /**
     * @method TreeView#_extendSelection
     * @private
     *
     * @param {Number} delta
     */
    tvp._extendSelection = function (delta) {
        var items = this._controller.getVisibleNodes();
        var pivot = this._controller.pivot();
        var selectionIndex = items.indexOf(pivot);
        var active = this._controller.active();
        var activeIndex = items.indexOf(active);
        var selected;

        if (this._options.multiSelect) {

            if (selectionIndex === -1) {
                selectionIndex = activeIndex;
            }

            if ((delta < 0 && selectionIndex + delta >= 0) ||
                (delta > 0 && selectionIndex + delta < items.length)) {
                selectionIndex += delta;
            }

            if (selectionIndex < activeIndex) {
                selected = items.slice(selectionIndex, activeIndex + 1);
            } else {
                selected = items.slice(activeIndex, selectionIndex + 1);
            }

            this._controller.selected(selected);

            this._controller.pivot(items[selectionIndex]);

            this.scrollIntoView(delta < 0, selectionIndex);
        }
    };

    /**
     * @method TreeView#_top
     * @private
     */
    tvp._top = function () {
        var item = _.first(this._controller.getVisibleNodes());
        if (item) {
            this._controller.active(item);
            this.scrollIntoView(true, 0);
        }
    };

    /**
     * @method TreeView#_bottom
     * @private
     */
    tvp._bottom = function () {
        var nodes = this._controller.getVisibleNodes();
        var item = _.last(nodes);
        if (item) {
            this._controller.active(item);
            this.scrollIntoView(false, nodes.length);
        }
    };

    /**
     * @method TreeView#_select
     * @private
     */
    tvp._select = function () {
        var activeId = this._controller.active();

        if (activeId != null && (this._controller.isSelectable(activeId) && !this._controller.isDisabled(activeId))) {
            this._controller.selected([activeId]);
        }
    };

    /**
     * @method TreeView#_find
     * @private
     *
     * @param {String} letter
     */
    tvp._find = function (letter) {
        var items = this._controller.getVisibleNodes();
        var regex = new RegExp('^' + stringUtils.regexEscape(letter), 'gi');
        var activeIndex = items.indexOf(this._controller.active());
        var safeEl = document.createElement('span');
        var nextActive;

        items = arrayUtils.rotate(items, activeIndex);
        nextActive = _.find(items, function (item) {
            safeEl.innerHTML = this._options.viewFns.display(this._controller.node(item));
            return regex.test(safeEl.textContent);
        }.bind(this));

        if (nextActive != null) {
            this._controller.active(nextActive);

            var index = items.indexOf(nextActive);
            this.scrollIntoView(activeIndex > index, index);
        }
    };

    /**
     * @method TreeView#scrollIntoView
     * @private
     *
     * @param {Number} top
     * @param {Number} currentIndex
     */
    tvp.scrollIntoView = function (top, currentIndex) {
        var numberOfNodes = this._$scrollContainer.height() / this._options.rowHeight;
        var totalVisible = Math.ceil(numberOfNodes);
        var fullyVisible = Math.floor(numberOfNodes);
        var topIndex = this._controller.getVisibleNodes().indexOf(this._getTreeId(this._$el.find('[data-treeid]:eq(0)')));
        if (currentIndex - topIndex < 0 || currentIndex - topIndex + 1 > fullyVisible) {
            if (top) {
                this._$scrollContainer.scrollTop(currentIndex * this._options.rowHeight);
            } else {

                if (numberOfNodes !== Math.round(numberOfNodes) || currentIndex - topIndex < fullyVisible) {
                    currentIndex++;
                }
                this._$scrollContainer.scrollTop((currentIndex - totalVisible + 1) * this._options.rowHeight);
            }
        }
    };

    /**
     * @method TreeView#_moveActive
     * @private
     *
     * @param {Number} delta
     */
    tvp._moveActive = function (delta) {
        var nodes = this._controller.getVisibleNodes();
        var activeElement = this._controller.active();
        var nextIndex = math.mod(nodes.indexOf(activeElement) + delta, nodes.length);
        var nextItem;
        nextItem = nodes[nextIndex];

        if (nextItem != null) {
            this._controller.active(nextItem);

            this.scrollIntoView(delta < 0, nextIndex);
        }

    };

    /**
     * @method TreeView#_handleCollapse
     * @private
     */
    tvp._handleCollapse = function () {
        var active = this._controller.active();
        var parent = this._controller.getParent(this._controller.active());

        if (active != null) {
            if (this._controller.isExpanded(active)) {
                this._controller.toggle('expanded', active);
            } else if (parent != null) {
                this._controller.active(parent);
            }
        }
    };

    /**
     * @method TreeView#_handleExpand
     * @private
     */
    tvp._handleExpand = function () {
        var active = this._controller.active();
        if (active != null) {

            if (this._controller.hasChildren(active) &&
                   this._controller.isExpanded(active)) {
                this._controller.active(this._controller.children(active)[0]);
            } else {
                this._controller.expanded(_.union(this._controller.expanded(), [active]));
            }
        }
    };

    /**
     * @method TreeView#buildElement
     *
     * @param {Tree#Config} options
     * @returns {jQuery}
     */
    tvp.buildElement = function (options) {
        var $el = $(treeTemplate({
            locale: options.locale,
            columns: _.map(options.columns.order, function (column) {
                var lookup = {};
                lookup.id = column;

                var columnDef = _.find(options.columns.definitions, lookup);

                return {
                    display: columnDef.header(),
                    width: columnDef.width
                };
            })
        }));

        $el.toggleClass('picker', options.picker);

        $el.toggleClass('fill', options.fill);

        $el.attr('aria-multiselectable', options.multiSelect);

        return $el;
    };

    /**
     * Returns the element that identifies the widget.
     *
     * @method TreeView#getIdentifyingElement
     *
     * @returns {jQuery}
     */
    tvp.getIdentifyingElement = function () {
        return this._$el;
    };

    /**
     * @method TreeView#groupRemoved
     *
     * @param {String} group
     */
    tvp.groupRemoved = function (group) {
        this._$el.find('[data-treeid="' + group + '"]').remove();
    };

    /**
     * @method TreeView#loading
     *
     * @param {String} id
     */
    tvp.loading = function (id) {
        if (_.isNull(id)) {
            this._$el.addClass('loading');
            this._$el.find('.root').empty();

            //Loading spinner should not be on the screen at the same time
        } else {
            this._findTree(id).addClass('loading');
        }

    };

    /**
     * Called after the selectable event fires.
     *
     * Marks the nodes passed in as unselectable.
     *
     * @method TreeView#selectable
     * @param {Array} ids The unselectable ids to be marked as unselectable.
     */
    tvp.selectable = function (ids) {
        var that = this;

        this._$el.find('.unselectable').removeClass('unselectable').attr('aria-selected', false);
        _.each(ids, function (id) {
            that._findTree(id).children('.item').addClass('unselectable').removeAttr('aria-selected');
        });
    };

    /**
     * @method TreeView#disable
     * @param {Boolean} disabled
     */
    tvp.disable = function (disabled) {

        this._$el.toggleClass('disabled', disabled);

        if (disabled) {
            this._$el.removeAttr('tabindex');
        } else {
            this._$el.attr('tabindex', 0);
        }
    };

    /**
     * Called after the select event fires.
     *
     * Marks each element as selected.
     *
     * @method TreeView#select
     * @param {Array} selectedIds The ids of the nodes that are selected.
     */
    tvp.select = function (selectedIds) {
        this._controller.pivot(null);
        this._$el.find('.selected').removeClass('selected').attr('aria-selected', false);

        this._$el.toggleClass('has-selection', selectedIds.length > 0);

        this._$el.find('.filter-notif').text(this._options.locale.notification.replace(/{\d+}/g, function () {
            return selectedIds.length;
        }));

        this._findTree(selectedIds).attr('aria-selected', true).addClass('selected');

    };

    /**
     * Called after the active event fires.
     *
     * Marks the node as active. (The node from where all keyboard interaction takes place)
     * There can only ever be on active element.
     *
     * @method TreeView#active
     * @param {String} id The id of the element to mark as active.
     */
    tvp.active = function (id) {

        this._$el.find('.active').removeClass('active');
        var activeElement = this._findTree(id)
            .addClass('active');

        if (activeElement.length !== 0) {
            this._$el.attr('aria-activedescendant', activeElement.attr('id'));
        } else {
            this._$el.removeAttr('aria-activedescendant');
        }

    };

    /**
     * Called after the filter event fires.
     *
     * Renders the results as a flat list.
     *
     * @method TreeView#filter
     * @param {Array} results The nodes that match the query.
     */
    tvp.filter = function (results) {
        var that = this;
        var newDescribedBy;

        this._$el.toggleClass('filtered', !_.isNull(results)).toggleClass('not-found', !_.isNull(results) && results.length === 0);

        if (!_.isNull(results) && results.length === 0) {
            newDescribedBy = stringUtils.union(this._$el.attr('aria-describedby'), this._$el.find('.no-results').attr('id'));
            this._$el.attr('aria-describedby', newDescribedBy);
        } else {
            newDescribedBy = stringUtils.reject(this._$el.attr('aria-describedby'), this._$el.find('.no-results').attr('id'));

            if (newDescribedBy === '') {
                this._$el.removeAttr('aria-describedby');
            } else {
                this._$el.attr('aria-describedby', newDescribedBy);
            }
        }

        if (this._controller.selected().length > 0) {
            this._$el.find('.filter-notif').text(this._options.locale.notification.replace(/{\d+}/g, function () {
                return that._controller.selected().length;
            }));
        }

        this.render(results);
    };

    /**
     * @method TreeView#getRenderedNodeIds
     *
     * @returns {Array}
     */
    tvp.getRenderedNodeIds = function () {
        if (this._$el[0].clientHeight === 0) {
            return [];
        }

        var rowHeight = this._options.rowHeight;
        var scroller = this._$scrollContainer.children('.scroller');
        var visibleNodes = this._controller.getVisibleNodes();
        var renderNodes = Math.ceil(this._$el.find('.tree-items')[0].clientHeight / rowHeight);
        window.requestAnimationFrame(function () {
            scroller.css({
                height: visibleNodes.length * rowHeight
            });
        });
        var startIndex = Math.ceil(this._$scrollContainer[0].scrollTop / (visibleNodes.length * rowHeight) * visibleNodes.length);
        this._renderedNodeIds = visibleNodes.slice(startIndex, startIndex + renderNodes);

        return this._renderedNodeIds;
    };

    /**
     * Creates the markup that represent the tree (current node) in the current
     * state (expanded/collapsed/selected/active).
     *
     * Recursive if the node is expanded - to render all of its children.
     *
     * @method TreeView#createTree
     * @param {Array} ids
     * @returns {String} markup The HTML markup representing the tree.
     */
    tvp.createTree = function (ids) {
        return _.reduce(ids, function (memo, id) {
            var node = this._controller.getNodeWithAttributes(id);
            var data = {
                id: id,
                attrs: node.attrs,
                flat: this._options.flat,
                locale: this._options.locale,
                pageId: this.id() + '-node-' + id,
                itemMarkup: this.makeNodeItem(node),
                group: node.attrs.group || false,
                columns: !_.isEmpty(_.get(this._options, 'columns.order'))
            };

            return memo + nodeTemplate(data);
        }.bind(this), '');
    };

    /**
     * @method TreeView#_generateGroupId
     * @private
     *
     * @param {Object} group
     */
    tvp._generateGroupId = function (group) {
        return group.toString();
    };

    /**
     *
     * @method TreeView#makeNodeItem
     * @param {Tree} node The node to make
     */
    tvp.makeNodeItem = function (node) {
        if (node.attrs.group) {
            return this._options.groupRenderer.render(node, this._options);
        } else {
            return this._options.renderer.render(node, this._options);
        }
    };

    /**
     * @method TreeView#resize
     */
    tvp.resize = function () {
        if (this._cachedHeight !== this._$el[0].clientHeight) {
            this.render();
        }
    };

    /**
     * Emptys the root node and renders the tree as in its current state.
     *
     * @method TreeView#render
     *
     * @param {Object} [modifiers]
     * @param {Boolean} [modifiers.quick]
     */
    tvp.render = function (modifiers) {
        var $node = this._$el.find('ul.root');
        var quick = (modifiers && modifiers.quick) || false;
        var currentRenderedNodes = this._renderedNodeIds;
        var removed;
        var added;
        var append;

        this._$el.toggleClass('loading', !this._controller.isLoaded());
        this._$el
            .toggleClass('filtered', this._controller.isFiltered())
            .toggleClass('not-found', this._controller.isFiltered() && !this._controller.hasFilterResults());

        this._cachedHeight = this._$el[0].clientHeight;

        if (this._renderRequest) {
            window.cancelAnimationFrame(this._renderRequest);
        }

        this._renderRequest = window.requestAnimationFrame(function () {
            var renderedNodes = this.getRenderedNodeIds();
            if (!quick) {
                $node[0].innerHTML = this.createTree(renderedNodes);
            } else {
                removed = _.difference(currentRenderedNodes, renderedNodes);
                added = _.difference(renderedNodes, currentRenderedNodes);
                append = renderedNodes[0] !== added[0];
                var trees = this._findTree(removed);
                var toAttach = this.createTree(added);

                trees.remove();
                if (append) {
                    $node.append(toAttach);
                } else {
                    $node.prepend(toAttach);
                }

            }

            $node.toggleClass('empty', renderedNodes.length === 0);
            var children = $node.children();

            this._validateRowHeight(children);
            this.trigger('dom-nodes-added', children);

            if (this._timeout) {
                clearTimeout(this._timeout);
            }

            this._timeout = setTimeout(function () {
                var idToElement = _.transform(renderedNodes, function (memo, key) {
                    memo[key] = {
                        $element: this._findTree(key),
                        node: this._controller.node(key)
                    };
                }.bind(this), {});

                this.trigger('dom-rendered', idToElement);
            }.bind(this), 100);

        }.bind(this));
    };

    /**
     * @method TreeView#_validateRowHeight
     * @private
     *
     * @param {Array<HTMLElement>} nodes
     */
    tvp._validateRowHeight = function (nodes) {
        if (!this._rowHeightValidated) {
            var rowHeight = nodes.first().outerHeight();
            if (rowHeight && rowHeight !== this._options.rowHeight) {
                logger.warn('Row height does not match the node\'s height - updating option');
                this._options.rowHeight = rowHeight;
                this.render();
            }

            this._rowHeightValidated = true;
        }
    };

    /**
     * @method TreeView#destroy
     *
     * @param {Object} clean
     */
    tvp.destroy = function (clean) {
        this._$el.off('click keydown');

        if (clean.self) {
            this._$el.remove();
        }
    };

    return TreeView;

});

define('oui/base/renderer',[
    'underscore',
], function (_) {
    'use strict';

    /**
     * @class Renderer
     */
    function Renderer() {}

    Renderer.prototype = _.extend(Renderer.prototype, {
        /**
         * @method Renderer#render
         *
         * @param {Array} nodes
         * @param {Object} options
         *
         * @returns {String}
         */
        render: _.noop,
        /**
         * @method Renderer#extend
         *
         * @param {Object} extension
         *
         * @returns {Object}
         */
        extend: function (extension) {
            _.each(extension, function (value, key) {
                if (key.charAt(0) !== '_' || this[key] == null) {
                    this[key] = value;
                }
            }.bind(this));
            return this;
        }
    });

    return Renderer;
});


define('template-22', ['underscore'], function (_) { return function(data) {
var __t, __p = '', __e = _.escape, __j = Array.prototype.join;
function print() { __p += __j.call(arguments, '') }

    var collapsed = !data.attrs.expanded ? 'collapsed' : 'expanded';    var hidden = !data.attrs.hasChildren ? 'tree-hidden': null;    var titleString = !data.attrs.expanded ? 'collapsed' : 'expanded';    var inlineStyle = (data.flat == null || data.flat === false) ? 'style="padding-left: ' + (data.attrs.level * 15) + 'px;"' : '';;

 if (data.columns != null ) { ;
__p += ' ';
 for(var i = 0, len = data.columns.length; i < len; i++) { ;
__p += ' ';
 if (i === 0) { ;
__p += ' <div class="col" title="' +
((__t = ( data.columns[i].title )) == null ? '' : __t) +
'" style="flex-basis: ' +
((__t = ( data.columns[i].width )) == null ? '' : __t) +
'; padding-left: ' +
((__t = ( data.attrs.level * 15 )) == null ? '' : __t) +
'px;"> <div class="spacer" style="width: ' +
((__t = ( data.attrs.level * 15 )) == null ? '' : __t) +
'px;"></div> <i title="' +
((__t = ( data.locale[titleString] )) == null ? '' : __t) +
'" data-attr="expanded" data-toggle="true" class="' +
((__t = ( ['tree-icon', collapsed, hidden].join(' ') )) == null ? '' : __t) +
'"></i> ' +
((__t = ( data.iconMarkup )) == null ? '' : __t) +
' ';
 } else { ;
__p += ' <div class="col" title="' +
((__t = ( data.columns[i].title )) == null ? '' : __t) +
'" style="flex-basis: ' +
((__t = ( data.columns[i].width )) == null ? '' : __t) +
';"> ';
 } ;
__p += ' ' +
((__t = ( data.columns[i].display )) == null ? '' : __t) +
' </div> ';
 } ;

 } else { ;
__p += '<div class="item-content ' +
((__t = ( data.attrs.required ? 'required' : '' )) == null ? '' : __t) +
'" title="' +
__e( data.title ) +
'"> <i title="' +
((__t = ( data.locale[titleString] )) == null ? '' : __t) +
'" data-attr="expanded" data-toggle="true" class="' +
((__t = ( ['tree-icon', collapsed, hidden].join(' ') )) == null ? '' : __t) +
'"></i> ' +
((__t = ( data.iconMarkup )) == null ? '' : __t) +
' ' +
((__t = ( data.valueMarkup )) == null ? '' : __t) +
'</div>';
 if (data.actions && data.actions.length > 0) { ;
__p += ' <div class="actions"> ';
 for (var i = 0, length = data.actions.length; i < length; i++ ) { ;
__p += ' ';
 if (_.isFunction(data.actions[i])) { ;
__p += ' ' +
((__t = ( data.actions[i]() )) == null ? '' : __t) +
' ';
 } else { ;
__p += ' ';
 var when = data.actions[i].when ? 'when-' + data.actions[i].when : ''; ;
__p += ' ';
 var classes = data.actions[i].classes ? data.actions[i].classes.join(' ') : ''; ;
__p += ' <span class="action ' +
((__t = ( when )) == null ? '' : __t) +
' ' +
((__t = ( classes )) == null ? '' : __t) +
'" title="' +
((__t = ( data.actions[i].title || '' )) == null ? '' : __t) +
'" data-trigger="' +
((__t = ( data.actions[i].action )) == null ? '' : __t) +
'"> <i class="pgbu-icon-small ' +
((__t = ( data.actions[i].icon )) == null ? '' : __t) +
'"></i> </span> ';
 } ;
__p += ' ';
 } ;
__p += ' </div>';
 } ;

 } ;

return __p
};});

define('oui/widgets/tree/treeNodeRenderer',[
    'underscore',
    'oui/base/renderer',
    'template-22'
], function (_, Renderer, nodeItemTemplate) {
    'use strict';

    /**
     * @typedef TreeNodeRenderer#RenderData
     *
     * @property {Object} attrs
     * @property {Boolean} attrs.unselectable
     * @property {Boolean} attrs.hidden
     * @property {Boolean} attrs.required
     * @property {Boolean} attrs.disabled
     * @property {Boolean} attrs.loaded
     * @property {Boolean} attrs.empty
     * @property {Boolean} attrs.group
     * @property {Boolean} attrs.hasChildren
     * @property {Boolean} attrs.visible
     * @property {Boolean} attrs.expanded
     * @property {Number} attrs.level
     * @property {String?} attrs.displayValue
     *
     * @property {Object} data
     */

    /**
     * @class TreeNodeRenderer
     * @extends Renderer
     */
    function TreeNodeRenderer() {
        Renderer.apply(this, arguments);
    }

    TreeNodeRenderer.prototype = Object.create(Renderer.prototype);

    TreeNodeRenderer.prototype = _.extend(TreeNodeRenderer.prototype, {
        /**
         * @method TreeNodeRenderer#getColumns
         *
         * @param {TreeNodeRenderer#RenderData} node
         * @param {TreePanel#Config} options
         *
         * @returns {Array}
         */
        getColumns: function (node, options) {
            var safeEl = document.createElement('span');

            if (options.columns && options.columns.order && options.columns.order.length > 0) {
                return _.map(options.columns.order, function (column) {
                    var lookup = {};
                    lookup.id = column;

                    var columnDef = _.find(options.columns.definitions, lookup);

                    safeEl.innerHTML = columnDef.display(node.data);
                    return {
                        title: safeEl.textContent.trim(),
                        display: safeEl.innerHTML,
                        width: columnDef.width
                    };
                });
            }

        },
        /**
         * @method TreeNodeRenderer#getMarkup
         *
         * @param {TreeNodeRenderer#RenderData} node
         * @param {TreePanel#Config} options
         *
         * @returns {String}
         */
        getMarkup: function (node, options) {
            var safeEl = document.createElement('span');

            if (!node.attrs.group) {
                safeEl.innerHTML = options.viewFns.display(node.data);
            } else {
                safeEl.innerHTML = node.attrs.displayValue;
            }

            safeEl.normalize();

            return safeEl.innerHTML;
        },
        /**
         * @method TreeNodeRenderer#getTitle
         *
         * @param {TreeNodeRenderer#RenderData} node
         * @param {TreePanel#Config} options
         *
         * @returns {String}
         */
        getTitle: function (node, options) {
            var safeEl = document.createElement('span');

            if (!node.attrs.group) {
                safeEl.innerHTML = options.viewFns.display(node.data);
            } else {
                safeEl.innerHTML = node.attrs.displayValue;
            }

            safeEl.normalize();

            return safeEl.textContent.trim();
        },
        /**
         * @method TreeNodeRenderer#getClasses
         *
         * @param {TreeNodeRenderer#RenderData} node
         * @param {TreePanel#Config} options
         *
         * @returns {Array<String>}
         */
        getClasses: function (node, options) {
            var classes = [];

            if (options.viewFns.classes) {
                classes = options.viewFns.classes(node.data);

                if (node.attrs.group) {
                    classes = _.without(classes, 'new-item');
                }
            }

            return classes;
        },
        /**
         * @method TreeNodeRenderer#getIcons
         *
         * @param {TreeNodeRenderer#RenderData} node
         * @param {TreePanel#Config} options
         *
         * @returns {Array<String>?}
         */
        getIcons: function (node, options) {
            if (options.viewFns.icons) {
                return options.viewFns.icons(node.data);
            }
        },
        /**
         * @method TreeNodeRenderer#getActions
         *
         * @param {TreeNodeRenderer#RenderData} node
         * @param {TreePanel#Config} options
         *
         * @returns {Array<Object>?}
         */
        getActions: function (node, options) {
            if (options.viewFns.actions) {
                return options.viewFns.actions(node.data);
            }
        },
        /**
         * @method TreeNodeRenderer#render
         *
         * @param {TreeNodeRenderer#RenderData} node
         * @param {TreePanel#Config} options
         *
         * @returns {String}
         */
        render: function (node, options) {
            var data = {
                title: this.getTitle(node, options),
                columns: this.getColumns(node, options),
                classes: this.getClasses(node, options),
                valueMarkup: this.getMarkup(node, options),
                attrs: node.attrs,
                locale: options.locale,
                flat: options.flat,
                actions: this.getActions(node, options),
                iconMarkup: this.getIcons(node, options)
            };

            return nodeItemTemplate(data);
        },
    });

    return TreeNodeRenderer;
});

define('oui/widgets/treePanel',['require','underscore','oui/base/widget','oui/widgets/tree/collection','oui/widgets/tree/treeController','oui/widgets/tree/treeView','oui/widgets/tree/treeNodeRenderer','logger'],function (require) { //jshint ignore:line
      'use strict';

      var _ = require('underscore');
      var Widget = require('oui/base/widget');
      var TreeModel = require('oui/widgets/tree/collection');
      var TreeController = require('oui/widgets/tree/treeController');
      var TreeView = require('oui/widgets/tree/treeView');
      var TreeNodeRenderer = require('oui/widgets/tree/treeNodeRenderer');
      var logger = require('logger');

      /**
       * @callback TreePanel#load
       * @param {Number|String} id The nodes client-side internally generated id.
       * @param {Object} data The data the node represents.
       */

      /**
       * @typedef TreePanel#Attributes
       * @desc Used in {@link TreePanel#Config|Config.attrs}
       *
       * @property {String} disabled
       */

      /**
       * @typedef TreePanel#actionDefinition
       *
       * @property {String} action The action to trigger as 'data-trigger-<X>'
       * @property {String} icon The icon to use for the action
       * @property {String} [when] When the action icon should show up. Possible values are 'selected', or 'active'. When not
       * given the icon will show up all the time
       * @property {String} [title] The title attribute to use for the icon
       * @property {Array<String>} [classes] The additional classes to apply to the action icon.
       */

      /**
       * @callback TreePanel#filter
       * @param {String} query The query to search the node's data for.
       * @param {Array<String>} fields The fields to look over in the node's data.
       */

      /**
       * @callback TreePanel#attrs
       * @param {Object} data The node's data.
       * @returns {Array<String>} The current nodes attributes. Valid strings are 'unselectable', 'disabled', 'expanded',
       * and 'selected'
       */

      /**
       * @callback TreePanel#actions
       * @param {Object} data The node's data.
       * @returns {Array<TreePanel#actionDefinition>} The action definitions.
       */

      /**
       * @callback TreePanel#display
       * @param {Object} data The node's data.
       * @returns {String} The node's string representation to use on the DOM.
       */

      /**
       * @typedef TreePanel#ColumnDefinition
       *
       * @property {String} width
       * @property {String} attr
       * @property {String} id
       * @property {Function} header
       * @property {Function} display
       */

      /**
       * @typedef TreePanel#GroupConfig
       *
       * @property {String} field The field to use as the group field
       * @property {Boolean} asTree Whether to the group is hierarchichal. Uses the property as a children based group
       * @property {Function} [getGroupValue] How to retrieve the group value (String) from the node.
       *
       * @property {Function} [display] How to display the value of the group for a node.
       */

      /**
       * @typedef TreePanel#AttrConfig
       * @desc Currently, **only** disabled is supported
       *
       * @property {Boolean} nested Whether or not the attr affects children.
       */

      /**
       *
       * @example
       *     //Assumed this property is included in the options
       *     viewFns.icons : function (data) {
       *         // Can return markup based on data of the current node requested as well
       *         // as multiple icons
       *         return '<i class="pgbu-icon pgbu-icon-ent-user"></i>';
       *     }
       *
       * @callback TreePanel#icons
       * @param {Object} data the node's data.
       * @returns {HTMLString} The markup to use for displaying the node's icon.
       */

      /**
       * @typedef {Object} TreePanel#Config
       *
       * @property {Locale} [locale={}] The specific strings that are used by the widget and will use the defaults set from
       * the {@link module:oui.lang|lang}.
       * @property {String} [disabledStyle='italic'] Style to use within the tree when a node is disabled. Valid options
       * are 'hidden' or 'italic'.
       * @property {Boolean} [extendExisting=true] Whether or not to add new data to the node when asked to add nodes to the
       * collection.
       * @property {Boolean} [cloneOnAdd=true] Clones the nodes when adding to prevent methods from adding properties to the
       * objects.
       * @property {Boolean} [hideEmpty=false] Whether to remove disabled nodes from the filtered list and also their
       * parents.
       * @property {Boolean} [move=false] Whether to allow move up/down in the tree. Necessary when allowing drag and drop
       * operations within the tree.
       *
       * @property {Boolean} [lazyLoad=false] Whether or not the tree is lazy loaded. Will call the load function when
       * initialized. When false the tree will not call load as it is assumed that the data is already loaded.
       *
       * @property {Boolean} [fluid=true] Whether or not the tree should take up all available width of the parent
       * container.
       *
       * @property {Boolean} [fill=true] Whether or not the tree should take up all available height of the parent
       * container.
       *
       * @property {Boolean} [clientFilter=false] Whether or not the tree will produce its own filtered list off the data.
       *
       * @property {Boolean} [multiSelect=false] Whether or not the tree supports multiselect operations.
       *
       * @property {Boolean} [displayRoot=false] Whether or not the tree should display the root node of the tree.
       * **Deprecated**: all nodes in the model are displayed.
       *
       * @property {Boolean} [flat=false] Whether or not the tree should render as a flat list.
       *
       * @property {Boolean} [picker=false] Whether or not the tree should apply picker styling. **Note**: This is an
       * internal option.
       *
       * @property {Array<String>} [defaultFilterFields=[]] The fields the tree should filter on in the data when
       * `clientFilter=true`.
       *
       * @property {Boolean} [filterShowHierarchy=false] Whether or not the tree should show the hierarchy when in a
       * filtered state.
       *
       * @property {Object} [storeFns={}] The functions to use as callbacks when loading, filtering, and retrieving current
       * data attributes.
       *
       * @property {TreePanel#load} [storeFns.load]
       *
       * @property {TreePanel#attrs} [storeFns.attrs]
       *
       * @property {TreePanel#filter} [storeFns.filter]
       *
       * @property {Object} [viewFns={}] The functions to use as callbacks when displaying the node.
       *
       * @property {TreePanel#display} [viewFns.display]
       *
       * @property {TreePanel#icons} [viewFns.icons]
       * @property {TreePanel#actions} [viewFns.actions]
       *
       * @property {Object} [columns] Collection of column definitions within the Tree.
       * @property {Array<String>} [columns.order] The order the columns should be displayed. Uses the 'id' property of the TreePanel#ColumnDefinition.
       * @property {Array<TreePanel#ColumnDefinition>} [columns.definitions]
       *
       * @property {Object} [recordKeys] The map of tree identifiers to data fields.
       *
       * @property {String} [recordKeys.id='id']
       * @property {String} [recordKeys.parentId='parentId']
       * @property {String} [recordKeys.children='children']
       *
       * @property {Object} [drag={}] The configuration for the drag operations within the tree
       *
       * @property {Boolean} [drag.enabled=false] Whether or not dragging is enabled. Dragging works on selected nodes
       * (blue border to the side)
       *
       * @property {String} [drag.out='ignore'] What behavior the tree should exhibit when the element is dropped
       * outside the tree. Valid options are 'ignore' and 'remove'.
       *
       * @property {String} [drag.over='ignore'] What behavior the tree should exhibit when the element is moved within the
       * tree itself. Valid options are 'ignore' and 'move'.
       *
       * @property {Array<TreePanel#GroupConfig>} groupBy How the tree should be grouped.
       *
       * @property {Object<TreePanel#Attributes, TreePanel#AttrConfig>} attrs How attrs should be configured.
       * @property {TreePanel#AttrConfig} attrs.disabled
       */

      var tpp;
      var __WIDGET_NAME__ = 'tree-panel';
      var __DEFAULTS__ = {
          disabledStyle: 'italic',
          hideEmpty: true,
          extendExisting: true,
          cloneOnAdd: true,

          //hidden, italic
          move: false,
          lazyLoad: false,
          fluid: true,
          fill: true,
          clientFilter: false,
          multiSelect: false,
          displayRoot: false,
          flat: false,
          picker: false,
          defaultFilterFields: [],
          filterShowHierarchy: false,
          storeFns: {},
          rowHeight: 22,
          groupRenderer: new TreeNodeRenderer(),
          renderer: new TreeNodeRenderer(),
          viewFns: {
              classes: function () { //jscs:ignore jsDoc
                  return [];
              }
          },
          columns: {},
          idPrefix: 'oui-',
          recordKeys: {
              id: 'id',
              parentId: 'parentId',
              children: 'children'
          },
          drag: {
              enabled: false,
              out: 'ignore',
              over: 'ignore'
          },
          groupBy: [
              {
                  field: 'parentId',
                  asTree: true
              }
          ],
          attrs: {
              disabled: {
                  nested: true
              }
          }
      };

      /**
      * Tree Panel constructs a typical tree element to be placed on the DOM.
      * It is a UI element that supports both tree-structured data as well as flat data.
      *
      * It supports typical expand/collapse, selected, keyboard navigation, disabled and unselectable nodes.
      *
      * See {@link http://www.w3.org/TR/wai-aria-practices/#TreeView|ARIA Best Practices}
      * for information regarding keyboard shortcuts and general ARIA attributes and states.
      *
      *  **Keyboard Shortcuts**
      *
      * * <kbd>Up</kbd> - Changes the focus to the node above
      * * <kbd>Down</kbd> - Changes the focus to the node below
      * * <kbd>Left</kbd> - Collapse Group Node (will collapse if expanded, otherwise will change the focused node to the parent)
      * * <kbd>Right</kbd> - Expand Group Node (will expand if collapsed, otherwise will change the focused node to the first child)
      * * <kbd>Home</kbd> - Mark first node in Tree as focused
      * * <kbd>End</kbd> - Mark last visible node in Tree as focused
      * * <kbd>Escape</kbd> - Remove current filter from Tree
      * * <kbd>Space</kbd> - Select currently focused node
      * * <kbd>Shift + Up</kbd> - Extends the selection up by 1
      * * <kbd>Shift + Down</kbd> - Extends the selection down by 1
      * * <kbd>Ctrl + Shift + Up</kbd> - Move the currently selected nodes (or focused node if there are no selected nodes) up by 1
      * * <kbd>Ctrl + Shift + Down</kbd> - Move the currently selected nodes (or focused node if there are no selected nodes) down by 1
      * * <kbd>Ctrl + Space</kbd> - Toggle selection of currently focused node
      *
      *  ##### **Relevant Events**
      *
      *  * {@link TreePanel#node-added|node-added}
      *  * {@link TreePanel#node-loaded|node-loaded}
      *  * {@link TreePanel#filter|filter}
      *  * {@link TreePanel#invalidate|invalidate}
      *  * {@link TreePanel#reset|reset}
      *  * {@link TreePanel#node-updated|node-updated}
      *  * {@link TreePanel#active|active}
      *  * {@link TreePanel#disabled-node|disabled-node}
      *  * {@link TreePanel#expand|expand}
      *  * {@link TreePanel#collapse|collapse}
      *  * {@link TreePanel#node-removed|node-removed}
      *  * {@link TreePanel#select|select}
      *  * {@link TreePanel#selectable|selectable}
      *
      * @borrows TreeModel#event:node-added as TreePanel#node-added
      * @borrows TreeModel#event:node-loaded as TreePanel#node-loaded
      * @borrows TreeModel#event:filter as TreePanel#filter
      * @borrows TreeModel#event:invalidate as TreePanel#invalidate
      * @borrows TreeModel#event:reset as TreePanel#reset
      * @borrows TreeModel#event:node-updated as TreePanel#node-updated
      * @borrows TreeModel#event:active as TreePanel#active
      * @borrows TreeModel#event:disabled-node as TreePanel#disabled-node
      * @borrows TreeModel#event:expand as TreePanel#expand
      * @borrows TreeModel#event:collapse as TreePanel#collapse
      * @borrows TreeModel#event:node-removed as TreePanel#node-removed
      * @borrows TreeModel#event:select as TreePanel#select
      * @borrows TreeModel#event:selectable as TreePanel#selectable
      *
      * @example
      *
      *      var treePanel = new TreePanel({
      *          storeFns : {
      *              load : function(id) {
      *                  if(id === null) {
      *                      return $.ajax({
      *                          url: 'mocks/topLevel.json',
      *                          type: 'GET'
      *                      });
      *                  }else{
      *                      return $.ajax({
      *                          url: 'mocks/personel' + id + '.json',
      *                          type: 'GET'
      *                      });
      *                  }
      *              },
      *              filter : function(query) {
      *                  console.log(query);
      *              }
      *          },
      *          viewFns : {
      *              display : function(node) {
      *                  //To be filled in with custom logic
      *                  return node.data.code + ' - ' + node.data.name;
      *              },
      *              icons : function(node) {
      *                  return '<i class="pgbu-icon pgbu-icon-ent-user"></i>';
      *              }
      *          }
      *          displayRoot : false,
      *          lazyLoad : false,
      *          flat : false,
      *          multiSelect : true
      *      }, initData);
      *
      * @class TreePanel
      * @since 1.2.0
      * @extends Widget
      *
      * @param {TreePanel#Config} options The options to override the defaults.
      * @param {Array} [initData] The initial data if it has already been loaded somewhere.
      */
      function TreePanel(options, initData) {
          var model;

          Widget.call(this, options, initData);

          if (this._options.flat) {
              this._options.groupBy = _.reject(this._options.groupBy, function (group) {
                  return group.asTree === true;
              });
          }

          model = new TreeModel(this._options, this._initData);
          /**
           * The controller object that the panel keeps a reference to.
           *
           * @private
           * @type {TreeController}
           * @memberof! TreePanel.prototype
           */
          this._controller = new TreeController(model, this._options);
          /**
           * The view element that the panel keeps a reference to.
           *
           * Used for hide/show otherwise everything should go
           * through the controller.
           *
           * @private
           * @type {TreeView}
           * @memberof! TreePanel.prototype
           */
          this._view = new TreeView(model, this._controller, this._options, this._settings);

          Object.defineProperty(this, '_data', {
              get: this.selected,
              set: this.selected
          });

          this._setupEvents([model, this._view]);

          var that = this;
          model.on('change', function (ids) {
              if (!this.silent) {
                  that.trigger('select', _.map(ids, that.node.bind(that)));
              }
          });
      }

      TreePanel.prototype = Object.create(Widget.prototype);

      tpp = TreePanel.prototype;

      tpp.constructor = TreePanel;

      /**
       * Used to prefix options.
       *
       * Currently unused due to the change in options.
       *
       * Plan on this name being used for public event prefixes.
       *
       * @method TreePanel#name
       * @returns {String} 'tree-panel' The name used in the widget for events
       */
      tpp.name = function () {
          return __WIDGET_NAME__;
      };

      /**
       * Calls initial load on the controller.
       *
       * @method TreePanel#_init
       * @private
       */
      tpp._init = function () {
          //Used for first render after append
          if (this._options.lazyLoad) {
              this._controller.load(null);
          } else {
              this._view.render();
          }
      };

      /**
       * @method TreePanel#_translate
       * @private
       *
       * @param {String} name
       */
      tpp._translate = function (name) {
          var that = this;

          /**
           * @method TreePanel#_translate~getDataForIds
           * @private
           *
           * @param {Array<String>} arr
           * @returns {Array}
           */
          function getDataForIds(arr) {
              return _.map(arr || [], that.node.bind(that));
          }

          if (name === 'change') {
              return [name, getDataForIds(arguments[1]), getDataForIds(arguments[2])];
          } else if (name === 'filter') {

              if (arguments[1] == null) {
                  return arguments;
              }

              var filteredData = getDataForIds(arguments[1]);

              return _.union([name, filteredData], _.slice(arguments, 2));
          }

          return arguments;
      };

      /**
       * Getter/Setter for the selected nodes in the TreePanel. When a parameter
       * is passed in the method will behave like a setter. If no parameter is
       * passed in then it will return the selected nodes.
       *
       * @since 1.3.0
       *
       * @method TreePanel#selected
       * @param {Array} [nodes] The nodes to set as selected in the TreePanel.
       * @returns {Array} selected When used as a getter the data of the selected nodes.
       *
       * @example
       *     var treePanel;
       *
       *     //Initialization...
       *
       *     treePanel.selected(['1', '2', '3']); //Will use the setter
       *
       *     treePanel.selected(); //Returns ['1', '2', '3']
       *
       *     //If you want to add to the current selection
       *     treePanel.selected(_.union(treePanel.selected(), ['4', '5', '6']));
       */
      tpp.selected = function (nodes) {
          if (!arguments.length) {
              return _.map(this._controller.selected(), this.node.bind(this));
          }

          nodes = _.compact(nodes);

          nodes = _.map(nodes, function (node) {
              if (!this._controller.node(node)) {
                  this._controller.add(node);
              }

              return _.get(this._controller.node(node), this._options.recordKeys.id);
          }.bind(this));

          this._controller.selected(nodes);
          this._view.trigger('picker-accept', this.selected());
          this.trigger('picker.select', this.selected());
      };

      /**
       * Getter/Setter for the selected nodes in the TreePanel. When a parameter
       * is passed in the method will behave like a setter. If no parameter is
       * passed in then it will return the selected nodes.
       *
       * **Note**: This function will not fire events. For more information on the alternate function see {@link
       * TreePanel#selected}.
       *
       * @method TreePanel#value
       * @param {Array} [nodes] The nodes to set as selected in the TreePanel.
       * @returns {Array} selected When used as a getter the data of the selected nodes.
       *
       * @example
       *     var treePanel;
       *
       *     //Initialization...
       *
       *     treePanel.selected(['1', '2', '3']); //Will use the setter
       *
       *     treePanel.selected(); //Returns ['1', '2', '3']
       *
       *     //If you want to add to the current selection
       *     treePanel.selected(_.union(treePanel.selected(), ['4', '5', '6']));
       */
      tpp.value = function (nodes) {
          if (!arguments.length) {
              return this.selected();
          }

          nodes = _.compact(nodes);

          nodes = _.map(nodes, function (node) {
              if (!this._controller.node(node)) {
                  this._controller.add(node);
              }

              return _.get(this._controller.node(node), this._options.recordKeys.id);
          }.bind(this));

          this._controller.selected(nodes, {
              silent: true
          });
          this._view.trigger('picker-accept', this.selected());
          this.trigger('picker.select', this.selected());
      };

      /**
       * Used to transform a client-side id into the JSON representation
       * of the node.
       *
       * @method TreePanel#node
       * @param {String|Object} id The client-side id or an unique attribute or set of
       *  attributes of the data.
       * @returns {Object} node The JSON representation of the node. Returns null if a node was not
       *  found.
       */
      tpp.node = function (id) {
          var node = this._controller.node(id, { strip: true });

          if (node != null) {
              //Returns the node based on the string id the user has.
              return node;
          } else {
              return id;
          }
      };

      /**
       * Transforms nodes into their client-side ids.
       *
       * @method TreePanel#ids
       * @param {Array} nodes The nodes to get the client-side ids for.
       * @returns {Array<Number>} ids The client-side ids of the nodes.
       */
      tpp.ids = function (nodes) {
          return _.map(nodes, function (node) {
              return _.get(this.node(node), this._options.recordKeys.id);
          }.bind(this));
      };

      /**
       * Adds data to either the root node or a specified node if parentNode is given.
       *
       * @since 1.3.0
       *
       * @example
       *     var tp = new TreePanel(); //Assuming correct parameters
       *
       *     //... Initialization here
       *
       *     //Adding to the root node
       *     tp.add([
       *         { name : 'Example obj1'},
       *         { name : 'Example obj2'}
       *     ]);
       *
       *     //Adding to the previously added example obj1
       *     tp.add({name : 'Example obj1'}, [
       *         { name : 'Example of child of obj1'}
       *     ]);
       *
       * @method TreePanel#add
       * @param {Array} nodes Array of objects that represent the nodes in the tree.
       * @param {Number} [index] The index of the parent to add the new node at. Added at the end when not specified.
       */
      tpp.add = function (nodes, index) {
          this._controller.add(nodes, index);
      };

      /**
       * Removes elements from the tree and all their respective children.
       *
       * @method TreePanel#remove
       * @param {Array} nodes The array of nodes to be removed. Can be either client-side ids or
       *  Objects representing the data of the node to remove.
       */
      tpp.remove = function (nodes) {
          _.each(nodes, this._controller.remove.bind(this._controller));
      };

      /**
       * Returns the JSON representation of the tree.
       *
       * @example
       *      //Example return format.
       *      { //Root node
       *          children : [
       *              {
       *                  "name" : "Example name 1", //name came back on load
       *                  //Will have whatever attributes were returned
       *                  //from the load function
       *                  "children" : [] //Can have nested objects
       *              },
       *              {
       *                  "name" : "Example name 2",
       *                  "children" : []
       *              }
       *          ]
       *      }
       *
       * @method TreePanel#tree
       * @param {Object} modifiers
       *
       * @returns {Object} - The JSON representation of the tree.
       */
      tpp.tree = function (modifiers) {

          if (modifiers === true) {
              modifiers = {
                  showRoot: false
              };
          }

          return this._controller.tree(_.defaults(modifiers || {}, {
              showRoot: true,
              strip: true
          }));
      };

      /**
       * @method TreePanel#nodes
       *
       * @returns {Array}
       */
      /**
       * @method TreePanel#nodes
       *
       * @param {Array} nodes
       */
      tpp.nodes = function (nodes) {
          if (!arguments.length) {
              return this._controller.nodes({ strip: true, sorted: true });
          }

          return _.map(nodes, function (id) {
              return this._controller.node(id, { strip: true });
          }.bind(this));
      };

      /**
       * Gets the defaults for the tree panel.
       *
       * @method TreePanel#defaults
       * @returns {Object} The defaults for the tree panel.
       */
      tpp.defaults = function () {
          return __DEFAULTS__;
      };

      /**
       * Filters the treePanel given the query and the array of fields.
       *
       * @method TreePanel#filter
       * @param {String} query The string to match to the fields of the nodes
       * @param {Array} fields The array of strings that match the fields to match on.
       * @param {String} scope The scope the filter is to be run on
       */
      tpp.filter = function (query, fields, scope) {
          var filterFields = fields || this._options.defaultFilterFields;

          if (_.isUndefined(query)) {
              this._controller.filter(null, filterFields, scope);
          } else {
              this._controller.filter(query, filterFields, scope);
          }
      };

      /**
       *
       * @method TreePanel#scrollIntoView
       *
       * @param {Object} node
       * @param {Boolean} [atTop=true]
       */
      tpp.scrollIntoView = function (node, atTop) {
          var id = _.get(this._controller.node(node), this._options.recordKeys.id);
          var parents = this._controller.getParents(id);

          if (atTop == null) {
              atTop = true;
          }

          this._controller.expanded(_.union(this._controller.expanded(), parents));

          this._view.scrollIntoView(atTop, this._controller.getVisibleNodes().indexOf(id));
      };

      /**
       * Disables the nodes that are supplied or returns the currently disabled nodes
       * if no parameter is specified.
       *
       * Overwrites currently disabled nodes with new nodes.
       *
       * The disabled attribute affects the node supplied as well as all of its children.
       *
       * Useful when shifting a node to another tree and the children are not allowed to be selected
       * after the parent has already been selected.
       *
       * @example
       *     //Example comment
       *     var treePanel = new TreePanel(); //Assuming correct construction.
       *
       *     //...Assuming initialization and loading
       *
       *     treePanel.disabled([
       *         {
       *             "name" : "example loaded name"
       *         }
       *     ]); //Will disable the node that has that name.
       *
       *     treePanel.disabled([]); //Will remove all disabled nodes.
       *
       *     treePanel.disabled(); //Will use the getter and return array of nodes' data.
       *
       *     treePanel.disabled(_.union(treePanel.disabled(), [
       *         {
       *             "name" : "additional node to disable"
       *         }
       *     ])); //Adds to the currently disabled list.
       *
       * @since 1.3.0
       * @method TreePanel#disabled
       * @param {Array} [nodes] The nodes to mark as disabled.
       * @returns {Array} - The array of disabled nodes' data.
       */
      tpp.disabled = function (nodes) {
          if (!arguments.length) {
              return _.map(this._controller.disabled(), this.node.bind(this));
          }

          nodes = _.map(nodes, function (node) {
              var realNode = this._controller.node(node);
              if (realNode) {
                  return realNode;
              } else {
                  return node;
              }

          }.bind(this));

          this._controller.disabled(_.compact(nodes));
      };

      /**
       * @method TreePanel#required
       *
       * @returns {Array}
       */
      /**
       * @method TreePanel#required
       * @param {Array} nodes
       */
      tpp.required = function (nodes) {
          if (!arguments.length) {
              return _.map(this._controller.required(), this.node.bind(this));
          }

          nodes = _.map(nodes, function (node) {
              var realNode = this._controller.node(node);
              if (realNode) {
                  return realNode;
              } else {
                  return node;
              }

          }.bind(this));

          this._controller.required(_.compact(nodes));
      };

      /**
       * Marks the nodes that are supplied as unselectable or returns the currently disabled nodes
       * if no parameter is specified.
       *
       * Overwrites currently unselectable nodes with new nodes.
       *
       * Only works on supplied node(s). The attribute does **not** affect children in the way that the
       * disabled attribute does.
       *
       * @example
       *     //Example comment
       *     var treePanel = new TreePanel(); //Assuming correct construction.
       *
       *     //...Assuming initialization and loading
       *
       *     treePanel.unselectable([
       *         {
       *             "name" : "example loaded name"
       *         }
       *     ]); //Will mark the node that has that name as unselectable.
       *
       *     treePanel.unselectable([]); //Will remove all unselectable nodes.
       *
       *     treePanel.unselectable(); //Will use the getter and return array of nodes' data.
       *
       *     treePanel.unselectable(_.union(treePanel.unselectable(), [
       *         {
       *             "name" : "additional node to mark"
       *         }
       *     ])); //Adds to the currently unselectable list.
       *
       * @since 1.3.0
       * @method TreePanel#unselectable
       * @param {Array} [nodes] The nodes to mark as unselectable.
       * @returns {Array} - The array of unselectable nodes' data.
       */
      tpp.unselectable = function (nodes) {
          if (!arguments.length) {
              return _.map(this._controller.unselectable(), this.node.bind(this));
          }

          if (_.isFunction(nodes)) {
              this._controller.unselectable(_.chain(this._controller.nodes())
                    .filter(nodes)
                    .pluck(this._options.recordKeys.id)
                    .value());
          } else {
              this._controller.unselectable(_.map(nodes, function (node) {
                  return _.get(this.node(node), this._options.recordKeys.id);
              }.bind(this)));
          }
      };

      /**
       * Marks the nodes that are supplied as expanded or returns the currently disabled nodes
       * if no parameter is specified.
       *
       * Overwrites currently expanded nodes with new nodes.
       *
       * @example
       *     //Example comment
       *     var treePanel = new TreePanel(); //Assuming correct construction.
       *
       *     //...Assuming initialization and loading
       *
       *     treePanel.expanded([
       *         {
       *             "name" : "example loaded name"
       *         }
       *     ]); //Will mark the node that has that name as expanded.
       *
       *     treePanel.expanded([]); //Will remove all expanded nodes.
       *
       *     treePanel.expanded(); //Will use the getter and return array of nodes' data.
       *
       *     treePanel.expanded(_.union(treePanel.expanded(), [
       *         {
       *             "name" : "additional node to mark"
       *         }
       *     ])); //Adds to the currently expanded list.
       *
       * @since 1.3.0
       * @method TreePanel#expanded
       * @param {Array} [nodes] The nodes to mark as expanded.
       * @returns {Array} - The array of expanded nodes' data.
       */
      tpp.expanded = function (nodes) {
          if (!arguments.length) {
              return _.map(this._controller.expanded(), function (id) {
                  return this.node(id);
              }.bind(this));
          }

          this._controller.expanded(_.map(nodes, function (node) {
              return _.get(this.node(node), this._options.recordKeys.id);
          }.bind(this)));
      };

      /**
       * Moves the currently active node up in its parent's subtree.
       *
       * The node will not move to another tree.
       *
       * @method TreePanel#moveUp
       */
      tpp.moveUp = function () {
          this._controller.moveUp();
      };

      /**
       * Moves the currently active node down in its parent's subtree.
       *
       * The node will not move to another subtree/parent.
       *
       * @method TreePanel#moveDown
       */
      tpp.moveDown = function () {
          this._controller.moveDown();
      };

      /**
       * Remove all data from the model (expanded, selected, disabled,...) and reloads
       * from the root node.
       *
       * @method TreePanel#invalidate
       */
      tpp.invalidate = function () {
          this._controller.invalidate();
      };

      /**
       * Updates existing nodes with new data.
       *
       * @method TreePanel#updated
       * @param {Array} nodes The array of nodes to update
       */
      tpp.updated = function (nodes) {
          this._controller.updateNodes(nodes);
      };

      /**
       * @method TreePanel#filtered
       * @returns {Array} The filtered nodes
       */
      tpp.filtered = function () {
          return _.map(this._controller.filtered(), function (tree) {
              return tree.toJSON();
          });
      };

      /**
       * @method TreePanel#isFiltered
       * @returns {Boolean} Whether or not the tree panel is currently filtered
       */
      tpp.isFiltered = function () {
          return this._controller.isFiltered();
      };

      /**
       * Getter/Setter for the linked tree panels.
       *
       * @method TreePanel#link
       * @returns {Array<TreePanel>} The linked trees associated with this tree panel.
       *
       */
      /**
       * Getter/Setter for the linked tree panels.
       *
       * @method TreePanel#link
       * @param {Array<TreePanel>} links
       *
       */
      tpp.link = function (links) {
          if (!arguments.length) {
              return this._controller.linked();
          }

          this._controller.linked(links);
      };

      /**
       * Synchronizes mouseover event with the original tree.
       *
       * Used internally with drag operations.
       *
       * @method TreePanel#sync
       * @param {jQuery.Event} event The mousemove event associated with the source tree to sync coordinates on.
       */
      tpp.sync = function (event) {
          this._view.sync(event);
      };

      /**
       * Synchronizes with original tree on mouseup when the item has been dropped.
       *
       * @method TreePanel#drop
       * @param {Array} nodes The nodes that were dropped on the tree.
       */
      tpp.drop = function (nodes) {
          this._view.drop(nodes);
      };

      /**
       * Resets attributes and recalls attributes.
       *
       * If no attributes are given it will reset all attributes.
       *
       * Differs from invalidating because it keeps the same data.
       *
       * @method TreePanel#reset
       * @param {Array<String>} attrs The attributes to reset.
       */
      tpp.reset = function (attrs) {
          this._controller.reset(attrs);
      };

      /**
       * @method TreePanel#groupBy
       *
       * @param {Array<Object>} groups
       */
      tpp.groupBy = function (groups) {
          if (!arguments.length) {
              return this._options.groupBy;
          }

          this._options.groupBy = groups;

          this._options.flat = groups.length === 0;

          this._controller.clearVisibleNodes();

          this._view.render();
      };

      /**
       * @method TreePanel#recordKeys
       *
       * @param {Object} newKeys
       */
      tpp.recordKeys = function (newKeys) {
          logger.warn('Updating recordKeys is an expensive operation. Can this be done during initialization?');

          this._options.recordKeys = _.extend(this._options.recordKeys, newKeys);
          this._controller.recache();
      };

      /**
       * @method TreePanel#getViewConfig
       * @returns {Object}
       */
      tpp.getViewConfig = function () {
          return this._controller.getViewConfig();
      };

      /**
       * @method TreePanel#applyViewConfig
       *
       * @param {Object} viewConfig
       */
      tpp.applyViewConfig = function (viewConfig) {
          this._controller.applyViewConfig(viewConfig);
      };

      /**
       * @method TreePanel#active
       *
       * @returns {Object}
       */
      /**
       * @method TreePanel#active
       *
       * @param {String} id
       */
      tpp.active = function (id) {
          if (!arguments.length) {
              return this.node(this._controller.active());
          }

          this._controller.active(id);
      };

      /**
       * @method TreePanel#resize
       */
      tpp.resize = function () {
          this._view.resize();
      };

      return TreePanel;
  }
);

define('oui/widgets/typeAhead/typeAheadModel',['require','jquery','underscore','moment','oui/utils/subscribable','oui/utils/uniqueId','oui/utils/flexEqual'],function (require) { //jshint ignore:line
      'use strict';

      var $ = require('jquery');
      var _ = require('underscore');
      var moment = require('moment');
      var Subscribable = require('oui/utils/subscribable');
      var uniqueId = require('oui/utils/uniqueId');
      var flexEqual = require('oui/utils/flexEqual');

      /**
       * @method TypeAheadModel~renderedEquality
       * @private
       * @param {Function} display
       * @param {String} str
       * @param {Object} node
       *
       * @returns {Boolean}
       */
      function renderedEquality(display, str, node) {
          return display(node.data) === str;
      }

      /**
       * Fired after a search has occurred or results were manually added to the model.
       *
       * @event TypeAheadModel#result-added
       * @param {Array} results The results that match the query
       */
      /**
       * Fired when a query has finished.
       *
       * @event TypeAheadModel#loaded
       * @param {Array} nodes The nodes used internally in the typeahead model
       * @param {String} query The query for the loaded nodes
       */
      /**
       * Fired when more nodes than are allowed have been selected.
       *
       * @event TypeAheadModel#selection-error
       */
      /**
       * Fired when no nodes have been selected but the field is marked as required.
       *
       * @event TypeAheadModel#required-error
       */
      /**
       * Fired after nodes have been selected.
       *
       * @event TypeAheadModel#select
       * @param {Array} selectedNodes The currently selected nodes of the typeahead
       */
      /**
       * Fired when the data has been cleared.
       *
       * @event TypeAheadModel#invalidate
       */
      /**
       * Fired after the active query or node changes.
       *
       * @event TypeAheadModel#active
       * @param {Object} active The data that is now active
       * @param {String} active.query The currently active query
       * @param {Number} active.node The currently active node
       */

      var tmp;

      /**
       * The data associated with the with the typeahead component.
       *
       * @class TypeAheadModel
       * @private
       * @since 1.3.0
       * @extends Subscribable
       *
       * @param {TypeAhead#Config} options Options passed from the widget.
       * @param {Array<Object>} initData The originally selected nodes within the typeahead
       */
      function TypeAheadModel(options, initData) {
          Subscribable.call(this);

          /**
           * Options for typeahead
           *
           * @memberof! TypeAheadModel.prototype
           * @type {Object}
           * @private
           */
          this._options = options;

          /**
           * @memberof! TypeAheadModel.prototype
           *
           * @type {Array<Object>}
           * @default []
           * @private
           */
          this._nodes = [];
          /**
           * @memberof! TypeAheadModel.prototype
           * @type {Array<Object>}
           * @default []
           * @private
           */
          this._queries = [];

          /**
           * @memberof! TypeAheadModel.prototype
           * @default null
           * @type {String}
           * @private
           */
          this._activeQuery = null;
          /**
           * @memberof! TypeAheadModel.prototype
           * @default null
           * @type {Object}
           * @private
           */
          this._activeNode = null;

          /**
           * The selected nodes in the model.
           *
           * @memberof! TypeAheadModel.prototype
           * @type {Array}
           * @private
           * @default []
           */
          this._selectedNodes = [];
          this._validationResult = null;

          this.selected(initData || []);
      }

      TypeAheadModel.prototype = Object.create(Subscribable.prototype);
      tmp = TypeAheadModel.prototype;

      tmp.constructor = TypeAheadModel;

      /**
       * Adds the results and sets an expire time to keep the results current.
       *
       * @method TypeAheadModel#addResult
       * @param {String} query The key to add to the result map
       * @param {Array} results The values in the map.
       *
       * @fires TypeAheadModel#result-added
       * @fires TypeAheadModel#loaded
       */
      tmp.addResult = function (query, results) {
          var that = this;
          var exactRegex;
          var exactMatch;

          this._queries.push({
              value: query,
              expires: moment().add(5, 'm')
          });

          if (this._options.allowNewValues) {
              exactRegex = new RegExp('^' + _.escapeRegExp(query) + '$', 'gi');
              exactMatch = _.some(results, function (data) {
                  return exactRegex.test(that._options.viewFns.display(data));
              });

              if (!exactMatch) {
                  this.addTempNode(query);
              }

          }

          _.each(results, _.bind(this.addNode, this, query));

          this.trigger('result-added', results);
          this.trigger('loaded', this.getNodesByQuery(query), query);
      };

      /**
       * Adds a temporary node to the typeahead with display value set and the data equaling the query.
       *
       * @method TypeAheadModel#addTempNode
       * @param {String} query The query that will also be used as the display value of the node
       *
       */
      tmp.addTempNode = function (query) {
          this._tempNode = {
              id: 'temp-id',
              displayValue: query + ' ' + this._options.locale.newValue,
              _oui: true,
              data: query,
              queries: [query]
          };
      };

      /**
       * Moves the temporary node into the collection of nodes within the store.
       *
       * @method TypeAheadModel#commitTempNode
       * @returns {Object} node The node after it has been added to the collection of nodes
       */
      tmp.commitTempNode = function () {
          var dfd = $.Deferred();
          if (this._options.storeFns.add) {
              $.when(this._options.storeFns.add(this._tempNode.data)).done(function (data) {
                  var node = this.addNode(this._tempNode.data, data);
                  dfd.resolve(node);
              }.bind(this));
          } else {
              var node = this.addNode(this._tempNode.data, this._tempNode.data);
              node.displayValue = this._tempNode.data;
              dfd.resolve(node);
          }

          dfd.always(function () {
              this._tempNode = null;
          }.bind(this));

          return dfd.promise();
      };

      /**
       * Adds a single node to the model.
       *
       * @method TypeAheadModel#addNode
       * @param {String} query The query this is being added under.
       * @param {Object} result The result to add to the model.
       * @returns {Object} The node added to the model or the existing node that was updated
       */
      tmp.addNode = function (query, result) {
          var currentNode = this.getNodeByData(result);

          if (currentNode == null) {
              //Push node to nodes
              currentNode = {
                  id: uniqueId('typeaheadnode-').toString(),
                  _oui: true,
                  data: result
              };

              this._nodes.push(currentNode);
          }

          if (query != null) {
              //Add query to node
              currentNode.queries = _.uniq([query].concat(currentNode.queries || []));
          }

          return currentNode;
      };

      /**
       * Check to see if the current data is after the time stamp expired tag.
       *
       * @method TypeAheadModel#isExpired
       * @param {String} query The results to see if they have expired.
       * @returns {Boolean} - true if the current data is after the time stamp.
       */
      tmp.isExpired = function (query) {
          return moment().isAfter(_.find(this._queries, { value: query }).expires);
      };

      /**
       * Whether or not the query has been loaded.
       *
       * @method TypeAheadModel#isLoaded
       * @param {String} query The query to check to see if it is loaded.
       * @returns {Boolean} - true if the query is undefined in the result map or if they query is expired.
       */
      tmp.isLoaded = function (query) {
          return _.find(this._queries, { value: query }) != null && !this.isExpired(query);
      };

      /**
       * Return a single node based on search object (map of keys and values that lodash can compare)
       *
       * @method TypeAheadModel#getNodeByData
       * @param {Object} data The data to search for within the store
       * @returns {Object|undefined} The found node
       */
      tmp.getNodeByData = function (data) {
          var queryData = data;
          if (_.has(this._options, 'recordKeys.id')) {
              queryData = _.pick(queryData, this._options.recordKeys.id);
          }
          return _.find(this.getAllNodes(), function (node) {
              return flexEqual(node.data, queryData);
          });
      };

      /**
       * Gets a single node if found from the collection of nodes.
       *
       * @method TypeAheadModel#getNodeById
       * @param {String|Number} id The id to match
       * @returns {Object|undefined} The found node
       */
      tmp.getNodeById = function (id) {
          return _.find(this.getAllNodes(), 'id', id);
      };

      /**
       * Gets a single node based off the rendered value.
       *
       * @method TypeAheadModel#getNodeByValue
       * @param {String} str The string value to match
       * @returns {Object|undefined} The found node
       */
      tmp.getNodeByValue = function (str) {
          return _.find(this.getAllNodes(), _.partial(renderedEquality, this._options.viewFns.display, str));
      };

      /**
       * Gets all the nodes currently in the model.
       *
       * @method TypeAheadModel#getAllNodes
       * @returns {Array<Object>} All the nodes known to the typeahead
       */
      tmp.getAllNodes = function () {
          return this._nodes;
      };

      /**
       * Returns the results for a given query.
       *
       * @method TypeAheadModel#getResults
       * @param {String} query The query to get the results for.
       * @returns {Object} - The results.
       */
      tmp.getResults = tmp.getNodesByQuery = function (query) {
          var that = this;
          var nodesToFilter = this._nodes;

          //Add query to nodes dynamically if it isnt already present
          _.each(nodesToFilter, function (node) {
              if (that.matches(node, query)) {
                  node.queries = _.uniq([query].concat(node.queries || []));
              }
          });

          if (this._tempNode) {
              nodesToFilter = [this._tempNode].concat(this._nodes);
          }

          return _.filter(nodesToFilter, function (node) {
              return _.contains(node.queries, query);
          });
      };

      /**
       * Determine based on type of data to call the appropriate getNodeBy{X} function.
       *
       * @method TypeAheadModel#node
       * @param {Object|String|Number} value
       * @returns {Object|undefined} found node
       */
      tmp.node = function (value) {
          var existingNode;

          if (_.isString(value) || _.isNumber(value)) {
              //Assuming this is a client side id.
              return this.getNodeById(value);
          } else if (_.isObject(value) && value._oui != null) {
              return value;
          } else if (_.isObject(value)) {
              existingNode = this.getNodeByData(value);

              if (_.isUndefined(existingNode)) {
                  return this.addNode(null, value);
              } else {
                  return existingNode;
              }
          }
      };

      //TODO: does this work?
      /**
       * @method TypeAheadModel#remove
       * @param {Object} node
       */
      tmp.remove = function (node) {
          this._selectedNodes = _.without(this._selectedNodes, node);
      };

      /**
       * Getter/setter for the selected nodes of the model
       *
       * @method TypeAheadModel#selected
       * @param {Array|Object} [nodes]
       * @param {Subscribable~EventOptions} eventParams The parameters to pass to trigger
       *
       * @fires TypeAheadModel#selection-error
       * @fires TypeAheadModel#required-error
       * @fires TypeAheadModel#select
       */
      tmp.selected = function (nodes, eventParams) {
          var dfd = $.Deferred().resolve(nodes).promise();

          if (!arguments.length) {
              //When calling the model directly when we say we want selected nodes
              //we mean it.
              return this._selectedNodes;
          }

          eventParams = eventParams || {};

          if (!this._options.multiSelect && nodes && nodes.length > 1) {
              //Fire error
              this.trigger('selection-error');
              return;
          }

          function always(committedNode) { //jscs:ignore jsDoc
              var index = _.indexOf(nodes, 'temp-id');
              if (committedNode == null) {
                  nodes.splice(index, 1);
              } else {
                  nodes.splice(index, 1, committedNode);
              }
              return nodes;
          }

          if (_.contains(nodes, 'temp-id')) {
              dfd = this.commitTempNode().then(always, always.bind(null, null));
          }

          dfd.always(function (nodes) {
              var selectedNodes = _.chain(nodes)
                    .map(this.node.bind(this))
                    .compact()
                    .uniq()
                    .value();

              if (selectedNodes.length === 0 && this._options.required) {
                  this._selectedNodes = selectedNodes;
                  var trigger = !(this._validationResult && this._validationResult.type === 'required');
                  if (trigger) {
                      this._validationResult = {
                          valid: false,
                          data: [],
                          type: 'required'
                      };

                      this.trigger(eventParams, 'required-error');
                  }
              } else if (!_.isEqual(this._selectedNodes, selectedNodes)) {
                  this._selectedNodes = selectedNodes;
                  this._validationResult = null;
                  eventParams.name = 'select';
                  this.trigger(eventParams, this._selectedNodes);
                  eventParams.name = 'change';
                  this.trigger(eventParams, this._selectedNodes);
              } else {
                  this._validationResult = null;
                  //Rerender view when the selected nodes are the same
                  this.trigger('view.select', this._selectedNodes);
              }
          }.bind(this));
      };

      /**
       * Sets and resets the active query and node within the model.
       *
       * @method TypeAheadModel#active
       * @param {Object|String} obj The active node or query to set as active within the model
       *
       * @fires TypeAheadModel#active
       */
      tmp.active = function (obj) {
          if (!arguments.length) {
              return {
                  query: this._activeQuery,
                  node: this._activeNode
              };
          }

          var results;

          if (_.isObject(obj)) {
              //We have a node
              this._activeNode = obj;
          } else if (_.isString(obj)) {
              this._activeQuery = obj;
              results = this.getNodesByQuery(obj);
              if (results.length > 0) {
                  this._activeNode = _.first(results);
              } else {
                  this._activeNode = null;
              }
          } else {
              this._activeQuery = null;
              this._activeNode = null;
          }

          this.trigger('active', {
              query: this._activeQuery,
              node: this._activeNode
          });
      };

      /**
       * Method to test if the rendered value matches the query.
       *
       * @method TypeAheadModel#matches
       * @param {Object} node The node to test equality
       * @param {String} query The query to compare to.
       * @returns {Boolean} Whether the node matches the query
       */
      tmp.matches = function (node, query) {
          if (_.has(this._options, 'viewFns.display')) {
              return new RegExp(_.escapeRegExp(query), 'ig').test(this._options.viewFns.display(node.data));
          } else {
              return false;
          }
      };

      /**
       * Clears the current models cached data.
       *
       * @method TypeAheadModel#invalidate
       *
       * @fires TypeAheadModel#invalidate
       */
      tmp.invalidate = function () {
          this._nodes = [];
          this._queries = [];
          this._activeQuery = null;
          this._activeNode = null;

          this.trigger('invalidate');

      };

      /**
       * @method TypeAheadModel#isValid
       *
       * @returns {Boolean}
       */
      tmp.isValid = function () {
          return !(this._validationResult != null && !this._validationResult.valid);
      };

      /**
       * @method TypeAheadModel#error
       *
       * @returns {Object?}
       */
      tmp.error = function () {
          if (!this.isValid()) {
              return _.clone(this._validationResult);
          } else {
              return null;
          }
      };

      return TypeAheadModel;
  }

);

define('oui/widgets/typeAhead/typeAheadController',[
    'underscore',
    'jquery',
    'oui/utils/stringUtils',
    'oui/utils/escapingUtils',
    'oui/utils/math'
], function (_, $, stringUtils, escapingUtils, math) {
    'use strict';

    var tcp;

    /**
     * Fired when the query is updated.
     *
     * @event TypeAheadModel#search
     *
     * @param {String} query
     */

    /**
     * @class TypeAheadController
     * @private
     * @since 1.3.0
     *
     * @param {Object} model The typeahead model to manipulate the data
     * @param {Object} controllerOptions The specific options regarding the controller.
     */
    function TypeAheadController(model, controllerOptions) {
        /**
         * The model object that the controller can manipulate.
         *
         * @private
         * @type {TypeAheadModel}
         * @memberof! TypeAheadController
         */
        this._model = model;
        this._options = controllerOptions;
    }

    tcp = TypeAheadController.prototype;

    tcp.constructor = TypeAheadController;

    /**
     * Either searches using the provided function into the options or uses
     * the results from the past search to return.
     *
     * @method TypeAheadController#search
     * @param {String} value The value of the text input
     * @param {Number} cursorLocation The index of the cursor location
     * @param {Object} scope The scope of the search
     * @returns {Boolean|undefined}
     */
    tcp.search = function (value, cursorLocation, scope) {
        var that = this;
        var query = stringUtils.getQuery(value, cursorLocation, this._options.multiSelect, this._options.delimiter);
        var trueQuery = escapingUtils.unescape(query, this._options);

        if ((_.isEmpty(trueQuery.trim()) && !this._options.allowEmpty) || trueQuery.trim().length < this._options.minLength) {
            this._model.active(null);

            if (!(this._options.multiSelect && this._model.selected().length > 0) && scope == null) {
                this._model.trigger('clear');
            } else if (scope != null) {
                this._model.trigger('search', value, scope);
            }
            /**
             * Triggered when data is loaded for a search query
             *
             * @event TypeAheadModel#loaded
             * @param {Array} results an empty array
             * @param {String} query the query that triggered the results
             */
            this._model.trigger('loaded', [], '');

            if (this._options.storeFns && this._options.storeFns.clear) {
                this._options.storeFns.clear();
            }

            return false;
        }

        if (query !== this._model.active().query) {
            this._model.trigger('search', query, scope);
        }

        if (!this._options.search) {
            this._model.active(query);
            return false;
        }

        if (this._model.isLoaded(query)) {
            this._model.trigger('loaded', this._model.getResults(query), query);
            this._model.active(query);
        } else {
            var searchFn = this._options.storeFns.search;
            /**
             * Triggered to tell the view that an asynchronous
             * event (ajax) request is taking place and to show the user.
             *
             * @event TypeAheadModel#loading
             */
            this._model.trigger('loading');
            $.when(searchFn.call(this, query, scope)).done(function (data) {
                that._model.addResult(query, data);
                that._model.active(query);
            });
        }
    };

    /**
     * @method TypeAheadController#getNodeById
     *
     * @param {String} id
     * @returns {Object}
     */
    tcp.getNodeById = function (id) {
        return this._model.getNodeById(id);
    };

    /**
     * @method TypeAheadController#node
     *
     * @param {String} str
     * @returns {Object}
     */
    tcp.node = function (str) {
        return this._model.node(str);
    };

    /**
     * @method TypeAheadController#remove
     *
     * @param {Object} node
     */
    tcp.remove = function (node) {
        this._model.remove(node);
    };

    /**
     * @method TypeAheadController#active
     * @returns {String}
     */
    /**
     * @method TypeAheadController#active
     *
     * @param {String} node
     */
    tcp.active = function (node) {
        if (!arguments.length) {
            return this._model.active();
        }

        this._model.active(node);
    };

    /**
     * Moves the active/focused node by the number provided.
     *
     * Typically called with 1 or -1.
     *
     * @method TypeAheadController#moveActive
     *
     * @param {Number} number
     */
    tcp.moveActive = function (number) {
        var active = this._model.active();
        var results = this._model.getResults(active.query);
        var index = _.indexOf(results, active.node);

        if (results.length > 0) {
            this._model.active(results[math.mod(index + number, results.length)]);
        }
    };

    /**
     * Selects a given node based on id and when no id is given it just
     * selects the currently active node.
     *
     * @method TypeAheadController#select
     * @param {String} id The id of the node to select
     * @param {Subscribable~EventOptions} eventParams
     */
    tcp.selected = function (id, eventParams) {
        if (!arguments.length) {
            return this._model.selected();
        }

        this._model.selected(id, eventParams);
    };

    /**
     * @method TypeAheadController#invalidate
     */
    tcp.invalidate = function () {
        this._model.invalidate();
    };

    /**
     * @method TypeAheadController#getQueryBounds
     *
     * @param {Object} selection
     * @param {String} value
     *
     * @returns {Object}
     */
    tcp.getQueryBounds = function (selection, value) {
        var bounds = {
            start: 0,
            end: 0
        };
        var firstDelimiter = value.lastIndexOf(this._options.delimiter, selection.start - 1);
        var lastDelimiter = value.indexOf(this._options.delimiter, selection.end);

        if (value.length > 0) {

            bounds.end = value.length;

            if (firstDelimiter !== -1 && this._options.multiSelect) {
                bounds.start = firstDelimiter + this._options.delimiter.length;
            }

            if (lastDelimiter !== -1 && this._options.multiSelect) {
                bounds.end = lastDelimiter;
            }
        }

        return bounds;
    };

    /**
     * @method TypeAheadController#isValid
     *
     * @returns {Boolean}
     */
    tcp.isValid = function () {
        return this._model.isValid();
    };

    /**
     * @method TypeAheadController#error
     *
     * @returns {Object?}
     */
    tcp.error = function () {
        return this._model.error();
    };

    return TypeAheadController;
});

define('oui/utils/scrollingUtils',['oui/utils/positioning'], function (relativePosition) {
    'use strict';

    /**
     * @module scrollingUtils
     * @private
     */

    return {
        /**
         * Modifies the containers scrollTop to force the element to be in the view.
         *
         * @method module:scrollingUtils.positionScrollTop
         *
         * @param {jQuery} $elementToPosition
         * @param {jQuery} $container
         */
        positionScrollTop: function ($elementToPosition, $container) {
            if ($elementToPosition.length > 0) {
                var position = relativePosition($elementToPosition, false, $container);
                var containerHeight = $container.height();
                var elementHeight = $elementToPosition.height();
                var containerScrollTop = $container.scrollTop();

                if (position.top + elementHeight > containerHeight) {
                    if (elementHeight > containerHeight) {
                        // Don't overscroll to the bottom of a large element
                        $container.scrollTop(containerScrollTop + position.top);
                    } else {
                        $container.scrollTop(containerScrollTop + position.top + elementHeight - containerHeight);
                    }
                } else if (position.top < 0) {
                    $container.scrollTop(containerScrollTop + position.top);
                }
            }
        },

        /**
         * Gets the relative positioning between one element and its ancestor
         *
         * @method module:scrollingUtils.relativePosition
         * @param {jQuery} $elem
         * @param {jQuery} $container
         * @returns {{top:Number,left:Number}}
         */
        relativePosition: function ($elem, $container) {
            return relativePosition($elem, false, $container);
        }
    };
});


define('template-23', ['underscore'], function (_) { return function(data) {
var __t, __p = '';
__p += '<input type="text" placeholder="' +
((__t = ( data.locale.placeholder  )) == null ? '' : __t) +
'" />';
return __p
};});


define('template-24', ['underscore'], function (_) { return function(data) {
var __t, __p = '';
__p += '<input type="search" class="empty" placeholder="' +
((__t = ( data.locale.placeholder )) == null ? '' : __t) +
'" title="' +
((__t = ( data.locale.search  )) == null ? '' : __t) +
'"><button class="btn btn-picker" title="' +
((__t = ( data.locale.search )) == null ? '' : __t) +
'"> <span class="search-icon"><i class="pgbu-icon-small pgbu-icon-magnifying-glass"></i></span></button><a title="' +
((__t = ( data.locale.clear )) == null ? '' : __t) +
'" class="search-clear-icon" href="#"> <i class="pgbu-icon pgbu-icon-find-clear"></i></a>';
return __p
};});


define('template-25', ['underscore'], function (_) { return function(data) {
var __t, __p = '';
__p += '<ul role="menu" class="orcl-type-ahead-menu dropdown-menu ' +
((__t = ( data.scroll ? 'scroll': '' )) == null ? '' : __t) +
'" style="max-height: ' +
((__t = (data.height)) == null ? '' : __t) +
'"> ' +
((__t = ( data.nodes  )) == null ? '' : __t) +
'</ul>';
return __p
};});


define('template-26', ['underscore'], function (_) { return function(data) {
var __t, __p = '', __e = _.escape;
__p += '<li role="option" id="' +
((__t = ( data.pageId )) == null ? '' : __t) +
'" data-id="' +
((__t = ( data.id)) == null ? '' : __t) +
'" class="orcl-type-ahead-node typeahead-item ' +
((__t = (data.active)) == null ? '' : __t) +
'" title="' +
__e(data.title) +
'"><a>' +
((__t = (data.text )) == null ? '' : __t) +
'</a></li>';
return __p
};});


define('template-27', ['underscore'], function (_) { return function(data) {
var __t, __p = '';
__p += '<span class="token-container" data-id="' +
((__t = ( data.id)) == null ? '' : __t) +
'" title="' +
((__t = (data.value )) == null ? '' : __t) +
'"> <span class="token"> ' +
((__t = ( data.value  )) == null ? '' : __t) +
' </span> <a class="token-remove pgbu-icon-small pgbu-icon-delete" href="#"></a></span>';
return __p
};});


define('template-28',[],function () { return '<div class="token-list-container">\n</div>\n';});


define('oui/widgets/tree/spinner.template',[],function () { return '<i class="loading-spinner pgbu-icon pgbu-icon-loading"></i>\n';});

define('oui/widgets/typeAhead/typeAheadTokenView',['require','underscore','jquery','oui/base/view','oui/utils/stringUtils','oui/utils/keyCode','oui/mixins/popup','oui/utils/scrollingUtils','template-23','template-24','template-25','template-26','template-27','template-28','oui/widgets/tree/spinner.template'],function (require) { //jshint ignore:line
      'use strict';

      var _ = require('underscore');
      var $ = require('jquery');
      var View = require('oui/base/view');
      var stringUtils = require('oui/utils/stringUtils');
      var keyCode = require('oui/utils/keyCode');
      var popupMixin = require('oui/mixins/popup');
      var scrollingUtils = require('oui/utils/scrollingUtils');
      var typeAheadTemplate = require('template-23');
      var typeAheadSearchTemplate = require('template-24');
      var menuTemplate = require('template-25');
      var nodeTemplate = require('template-26');
      var tokenTemplate = require('template-27');
      var tokenContainerTemplate = require('template-28');
      var loadingTemplate = require('oui/widgets/tree/spinner.template');

      var tvp;

      var PREVENT_DEFAULT_KEYS = [
          keyCode.DOWN,
          keyCode.UP,
          keyCode.LEFT,
          keyCode.RIGHT,
          keyCode.ENTER,
          keyCode.TAB,
          keyCode.SHIFT,
          keyCode.F2
      ];

      var PREVENT_DEFAULT_KEYS_UP = [
          keyCode.TAB,
          keyCode.SHIFT,
          keyCode.F2
      ];

      /**
       * Makes sure the element given is viewable within the container.
       *
       * @method TypeAheadTokenView~scrollIntoViewIfNeeeded
       * @private
       * @param {HTMLElement} element The element to scroll
       * @param {Boolean} [centerIfNeeded=true] Whether or not to scroll the element into the center
       */
      function scrollIntoViewIfNeeded(element, centerIfNeeded) {
          centerIfNeeded = arguments.length === 1 ? true : !!centerIfNeeded;

          var parent = element.parentNode;
          var parentComputedStyle = window.getComputedStyle(parent, null);
          var parentBorderTopWidth = parseInt(parentComputedStyle.getPropertyValue('border-top-width'));
          var parentBorderLeftWidth = parseInt(parentComputedStyle.getPropertyValue('border-left-width'));
          var overTop = element.offsetTop < parent.scrollTop;
          var overBottom = (element.offsetTop + element.clientHeight - parentBorderTopWidth) > (parent.scrollTop + parent.clientHeight);
          var overLeft = element.offsetLeft < parent.scrollLeft;
          var overRight = (element.offsetLeft + element.clientWidth - parentBorderLeftWidth) > (parent.scrollLeft + parent.clientWidth);
          var alignWithTop = overTop && !overBottom;

          if ((overTop || overBottom) && centerIfNeeded) {
              parent.scrollTop = element.offsetTop - parent.clientHeight / 2 - parentBorderTopWidth + element.clientHeight / 2;
          }

          if ((overLeft || overRight) && centerIfNeeded) {
              parent.scrollLeft = element.offsetLeft - parent.clientWidth / 2 - parentBorderLeftWidth + element.clientWidth / 2;
          }

          if ((overTop || overBottom || overLeft || overRight) && !centerIfNeeded) {
              element.scrollIntoView(alignWithTop);
          }
      }

      /**
       * The type ahead view is responsible for displaying the suggestions when they loaded.
       *
       * @class TypeAheadTokenView
       * @private
       * @since 2.1.0
       * @extends View
       * @mixes popup
       *
       * @param {TypeAheadModel} model The typeahead model that is responsible for holding all the past
       *  search results
       * @param {TypeAheadController} controller The type ahead controller that will modify the model.
       * @param {Object} element jQuery element for the view
       * @param {Object} viewOpts The options for the view.
       */
      function TypeAheadView(model, controller, element, viewOpts, settings) {
          var that = this;

          View.call(this, viewOpts, element, controller, settings);

          /**
           * The menu object for this typeahead
           *
           * A JQuery object.
           *
           * @private
           * @type {Object}
           * @memberof! TypeAheadTokenView.prototype
           */
          this._$menu = null;
          this._$input = this._$el.find('input');
          this._$tokListContainer = this._$el.find('.token-list-container');
          this._$tokList = this._$el.find('.token-list');

          this._inputEl = this._$input[0];

          model.on({
              'view.loaded': this.loaded.bind(this),
              'view.active': this.active.bind(this),
              'view.loading': this.loading.bind(this),
              'view.select': this.select.bind(this),
              'view.selection-change': this.select.bind(this),
              'view.invalidate': this.clear.bind(this),
              'view.required-error': this.required.bind(this)
          });

          this.element().on('oui-accept oui-picker-accept', function (event, widget, name, selected) {
              if (widget.name() === 'picker' || widget.name() === 'double-picker') {
                  that._controller.selected(selected, {
                      silent: true
                  });
              }
          });

          var debounced = _.debounce(controller.search.bind(controller), this._options.delay);
          this._debouncedSearch = debounced;

          this._$el.on('click', '.token-remove', function (event) {
              event.preventDefault();
              event.stopPropagation();
              //Remove token
              that._controller.selected(_.without(that._controller.selected(),
                that._controller.getNodeById($(this).parent('.token-container').attr('data-id'))));
              that._$input.focus();
          });

          this._$el.on('click', '.token', function (event) {
              event.preventDefault();
              //Select Token
              var parent = $(this).parent('.token-container');

              if (parent.position().left < 0) {
                  that.focusToken(parent);
              } else if (parent.position().left + parent.outerWidth() > that._$tokListContainer.width()) {
                  that.focusToken(parent);
              } else {
                  that._$el.find('.token-container.selected').removeClass('selected');
                  parent.addClass('selected');
                  that._$input.addClass('no-select');
              }

              that._$input.focus();
          });

          this._$el.on('mousedown', function (event) {
              event.preventDefault();
              that._$input.focus();
              that.focusText();
          });

          this._$input
        .on('keyup', function (event) {
            if (_.contains(PREVENT_DEFAULT_KEYS, event.keyCode)) {
                return;
            }

            switch (event.keyCode) {
                case keyCode.BACKSPACE:
                    if (!_.isEmpty(that._inputEl.value) &&
                      that._$el.find('.token-container.selected').length === 0) {
                        debounced.call(controller, this.value, this.selectionStart);
                    } else if (_.isEmpty(that._inputEl.value)) {
                        debounced.cancel();
                        if (that._$menu) {
                            that._$menu.remove();
                        }
                    }

                    break;
                default:
                    debounced.call(controller, this.value, this.selectionStart);
            }
        })
        .on('blur', function (event) {
            event.preventDefault();
            if ((_.isUndefined(that._hideMenu) || that._hideMenu) && !_.isNull(that._$menu)) {
                that._$menu.remove();
                that._$el.find('.token-container.selected').removeClass('selected');
            }
        })
        .on('keydown', function (event) {
            var selectedToken = that._$el.find('.token-container.selected');
            if (_.contains(PREVENT_DEFAULT_KEYS_UP, event.keyCode)) {
                return;
            }

            if (that._options.search) {
                switch (event.keyCode) {
                    case keyCode.DOWN:
                        event.preventDefault();
                        controller.moveActive(1);
                        break;
                    case keyCode.UP:
                        event.preventDefault();
                        controller.moveActive(-1);
                        break;
                    case keyCode.ENTER:
                        that._handleEnter(event);
                        break;
                    case keyCode.LEFT:
                        that._handleLeft(event, selectedToken);
                        break;
                    case keyCode.RIGHT:
                        that._handleRight(event, selectedToken);
                        break;
                    case keyCode.BACKSPACE:
                        that._handleBackspace(event, selectedToken);
                        break;
                    default:
                        that.focusText();
                        debounced.call(controller, this.value, this.selectionStart);
                }
            }
        });
      }

      TypeAheadView.prototype = Object.create(View.prototype);
      tvp = TypeAheadView.prototype;

      tvp.constructor = TypeAheadView;

      _.extend(tvp, popupMixin);

      /**
       * Removes the event listeners bound to the element and removes the element.
       *
       * @method TypeAheadTokenView#destroy
       */
      tvp.destroy = function () {
          this._$input.off('keyup blur focus keydown');
          this._$el.remove();
      };

      /**
       * Makes sure there is enough space to type the next search query in the input.
       *
       * @method TypeAheadTokenView#focusText
       */
      tvp.focusText = function () {
          //Move the text
          var selectedToken = this._$el.find('.token-container.selected');

          selectedToken.removeClass('selected');
          if (this._$input.hasClass('no-select')) {
              this._$input.removeClass('no-select');

              //In order to get the caret to reappear
              this._$input.blur();
              this._$input.focus();
          }

          if (this._$el.find('.token-container').length > 0) {
              var lastToken = _.last(this._$el.find('.token-container'));
              var firstOffset = lastToken.offsetWidth + lastToken.offsetLeft;
              this._$el.find('.token-list-container').css({
                  overflow: 'hidden'
              });
              var secondOffset = lastToken.offsetWidth + lastToken.offsetLeft;
              var width = this._$el.find('.token-list-container').width();

              if (secondOffset > firstOffset) {
                  width -= (firstOffset + 7);
              } else {
                  width -= secondOffset;
              }

              if (width < 75) {
                  this._$el.find('input').width(this._$el.find('.token-list-container').width());
              } else {
                  this._$el.find('input').width(width);
              }

              this._$el.find('.token-list-container').css({
                  overflow: 'auto'
              });
          }

          scrollIntoViewIfNeeded(this._$input[0]);
      };

      /**
       * Makes sure the token is viewable (scrolled into view) and selected.
       *
       * @method TypeAheadTokenView#focusToken
       *
       * @param {jQuery} $token
       */
      tvp.focusToken = function ($token) {
          //knows that it must move over to the right
          this._$el.find('.token-container.selected').removeClass('selected');
          $token.addClass('selected');
          scrollIntoViewIfNeeded($token[0]);
          this._$input.addClass('no-select');

          if (!_.isNull(this._$menu)) {
              this._$menu.remove();
          }

      };

      /**
       * Builds the DOM required for the TypeAhead.
       *
       * @method TypeAheadTokenView#buildElement
       * @param {TypeAhead#Config} options The options used within the TypeAhead
       * @param {?jQuery} $el The element to build around/wrap to become a TypAhead
       * @returns {jQuery} el The element representing the Type Ahead
       */
      tvp.buildElement = function (options, $el) {

          if (_.isNull($el)) {
              switch (options.style) {
                  case 'search':

                      $el = $(typeAheadSearchTemplate({
                          locale: options.locale
                      }));
                      $el.find('.search-icon').on('click', function () {
                          $el.find('input').focus();
                      });
                      break;
                  default:
                      $el = $(typeAheadTemplate({
                          locale: options.locale
                      }));
              }

          } else if ($el.jquery && $el.is('input')) {
              $el[0].type = 'text';
              $el[0].placeholder = options.locale.placeholder || '';
          }

          $el.wrap(tokenContainerTemplate);
          $el = $el.parent();

          if ($el.parent().hasClass('orcl-picker') || $el.parent().hasClass('orcl-double-picker')) {
              $el.parent().addClass('orcl-type-ahead w-tokens');
          } else {
              $el.wrap('<div class="orcl-type-ahead orcl-widget w-tokens"></div>');
          }

          $el = $el.parent();

          $el.find('input').removeAttr('readonly');

          if (options.search) {
              $el.find('input')
                .attr('role', 'combobox')
                .attr('aria-autocomplete', 'list')
                .attr('aria-expanded', 'false');
          }

          return $el;
      };

      /**
       * Empties out the current text and replaces it with a new string built
       * by calling into the supplied display function
       *
       * @method TypeAheadTokenView#select
       */
      tvp.select = function () {
          this._$el.add(this._$el.closest('.control-group')).removeClass('error');
          this._selectToken();

          this._$input.trigger('change', {
              update: false
          });

          if (!_.isNull(this._$menu)) {
              this._$menu.remove();
          }
      };

      /**
       * @method TypeAheadTokenView#required
       */
      tvp.required = function () {

          this.select();
          this._$el.add(this._$el.closest('.control-group')).addClass('error');
      };

      /**
       * Builds up the token markup.
       *
       * @method TypeAheadTokenView#_selectToken
       * @private
       */
      tvp._selectToken = function () {
          var tokenMarkup;
          var displayFn = this._options.viewFns.display;
          var selected = this._controller.selected();

          //Text is becoming a token - remove the current string value from the input
          this._inputEl.value = '';

          tokenMarkup = _.reduce(selected, function (memo, node) {
              return memo += tokenTemplate({
                  value: displayFn(node.data),
                  id: node.id
              });
          }, '');

          this._$el.find('.token-container').remove();
          this._$el.find('.token-list-container').prepend(tokenMarkup);

          this.focusText();
      };

      /**
       * Builds up the string that represents the selected items.
       *
       * @method TypeAheadtokenView#_selectNontoken
       */
      tvp._selectNontoken = function () {
          var that = this;

          this._inputEl.value = _.reduce(this._controller.selected(), function (memo, node, index) {
              if (!_.isUndefined(node.data)) {
                  return memo + (index > 0 ? that._options.delimiter + ' ' : '') + that._options.viewFns.display(node.data);
              } else {
                  return memo + (index > 0 ? that._options.delimiter + ' ' : '') + that._options.viewFns.display(node);
              }
          }, '');
      };

      /**
       * Displays the spinner when a search is happening but has not completed
       * yet.
       *
       * @method TypeAheadTokenView#loading
       */
      tvp.loading = function () {
          if (!_.isNull(this._$menu)) {
              this._$menu.remove();
          }

          //It needs a height because when you call spin - it tries to position itself in the center
          //TODO: this should not be here
          this._$menu = this._$popup = $('<ul class="orcl-type-ahead-menu dropdown-menu"><li style="height: 20px;"></li></ul>');

          $(this._options.appendTo).append(this._$menu);

          //Since JQuery objects are not really HTML elements we need to grab the element.
          this._$menu.find('li').append($(loadingTemplate));

          this.showPopup(this._$el);
      };

      /**
       * Marks the node as active and moves the viewport accordingly.
       *
       * @method TypeAheadTokenView#active
       * @param {Object} activeData The node that is active.
       */
      tvp.active = function (activeData) {
          var active;
          if (!_.isNull(this._$menu)) {
              this._$menu.find('li').removeClass('active');

              if (!_.isNull(activeData.node)) {
                  active = this._$menu.find('[data-id=' + activeData.node.id + ']');
                  active.addClass('active');

                  if (active.length > 0) {
                      scrollIntoViewIfNeeded(active[0]);
                  }
              }
          }

      };

      /**
       * Builds the markup for the menu.
       *
       * @method TypeAheadTokenView#buildMenu
       * @param {Array} data The data for the menu
       * @param {String} query The current query
       */
      tvp.buildMenu = function (data, query) {
          var that = this;
          var markup = '';
          var regex = new RegExp('(' + stringUtils.regexEscape(query) + ')', 'gi');
          var text;
          var highlightText;

          /**
           * Builds a nodes DOM representation from the data.
           *
           * @method TypeAheadTokenView~buildNode
           * @private
           *
           * @param {Object} node The node to build.
           * @returns {String} - The DOM representation of the node.
           */
          function buildNode(node) {
              text = node.displayValue || that._options.viewFns.display(node.data);
              highlightText = text;

              if (that._options.highlight) {
                  highlightText = text.replace(regex, '<strong>$1</strong>');
              }
              return nodeTemplate({
                  title: that._options.viewFns.text ? that._options.viewFns.text(node.data) : text,
                  text: highlightText,
                  id: node.id,
                  active: that._controller.active().node === node ? 'active' : ''
              });
          }

          /**
           * Constructs the menu
           *
           * @method TypeAheadTokenView~buildMenu
           * @returns {String} - The string representation of the menu.
           */
          function buildMenu() {
              return menuTemplate({
                  nodes: markup,
                  scroll: that._options.maxSuggestions < data.length ? 'scroll' : '',

                  //TODO: find a way to get 26 out of here (It is the height of an <li> in our dropdown
                  height: data.length < that._options.maxSuggestions ? 'auto' : (that._options.maxSuggestions * 26) + 'px'
              });
          }

          if (data.length > 0 && this._options.scroll) {
              //If it does support scrolling put them all in the markup
              _.each(data, function (node) {
                  markup += buildNode(node);
              });

              markup = buildMenu();
          } else if (data.length > 0 && !this._options.scroll) {
              //If no scrolling then only put their max suggestions in the markup
              var node;
              for (var i = 0; i < this._options.maxSuggestions; i++) {
                  node = data[i];
                  markup += buildNode(node);
              }

              markup = menuTemplate({
                  nodes: markup
              });
          }

          return markup;
      };

      /**
       * The method to take the markup and construct a DOM element from it.
       * Also since this is a new element we need to add the click events to the
       * menu.
       *
       * @method TypeAheadTokenView#render
       * @param {String} markup The markup representing the menu.
       */
      tvp.render = function (markup) {
          //Find proper z-indexing, append to the location
          var that = this;
          var active;

          //Since it is debounced - need to make sure we are still active/focused
          //TODO: use a boolean instead of is?
          if (this._$input.is(document.activeElement) && markup !== '') {

              if (!_.isNull(this._$menu)) {
                  this._$menu.remove();
              }

              //TODO: fix this rendering - create the menu once and only switch the items?
              this._$menu = this._$popup = $(markup);

              //Once we have the markup we need to put on the events - each time
              this._$menu.on('click', 'li', function (event) {
                  event.preventDefault();
                  that._$input.focus();

                  if (!that._options.multiSelect) {
                      that._controller.selected([$(this).attr('data-id')]);
                  } else {
                      that._controller.selected(that._controller.selected().concat($(this).attr('data-id')));
                  }
              });

              this._$menu.hover(function () {
                  that._hideMenu = false;
              }, function () {
                  that._hideMenu = true;
              });

              //Reposition popup after adding the nodes to it (giving it proper height to calculate with)
              this.showPopup(this._$el);

              active = this._$menu.find('.active');

              //Mess with scroll-top if it is scrollable
              if (this._options.scroll && active.length > 0) {
                  scrollingUtils.positionScrollTop(active, this._$menu);
              }
          } else {
              this.hidePopup();
          }
      };

      /**
       * @method TypeAheadTokenView#loaded
       * @param {Array} data The data for the menu
       * @param {String} query The query used to get the data.
       */
      tvp.loaded = function (data, query) {
          //Build menu, position menu, show menu
          var markup = this.buildMenu(data, query);
          this.render(markup);
      };

      /**
       * Removes all tokens from the typeahead and triggers change.
       *
       * @method TypeAheadTokenView#clear
       */
      tvp.clear = function () {
          this._$el.find('.token-list-container .token-container').remove();
          this._$input.trigger('change');
      };

      /**
       * Silently sets the value of the query on the input.
       *
       * @method TypeAheadTokenView#value
       *
       * @param {String} newQuery
       */
      tvp.value = function (newQuery) {
          this._$input.val(newQuery);
      };

      /**
       * Event handler for the left arrow action.
       *
       * Moves the focus to the next token to its left. If it is at the first character it will move
       * from text to a token.
       *
       * @method TypeAheadTokenView#_handleLeft
       * @private
       *
       * @param {jQuery.Event} event
       * @param {jQuery} selectedToken
       */
      tvp._handleLeft = function (event, selectedToken) {
          if (selectedToken.length !== 0) {
              if (!selectedToken.is(':first-child')) {
                  this.focusToken(selectedToken.prev('.token-container'));
                  event.preventDefault();
              }
          } else if (this.selectionStart === 0 && this._controller.selected().length > 0) {
              this.focusToken(this._$el.find('.token-container:last-of-type'));
              event.preventDefault();
          }

      };

      /**
       * Event handler for the right arrow action.
       *
       * Moves the focus to the next token to its right. If it is at the last character it will move
       * from tokens to the text issuing a new search for the menu to popup.
       *
       * @method TypeAheadTokenView#_handleRight
       * @private
       *
       * @param {jQuery.Event} event
       * @param {jQuery} selectedToken
       */
      tvp._handleRight = function (event, selectedToken) {
          if (selectedToken.length !== 0 && selectedToken.is(':last-of-type')) {
              //We are moving from the last token to the text input
              this.focusText();
              this._debouncedSearch(this.value, this.selectionStart);
              event.preventDefault();
          } else if (selectedToken.length !== 0) {
              //We are moving to the next token to the right
              this.focusToken(selectedToken.next('.token-container'));
              event.preventDefault();
          }

      };

      /**
       * Event handler for the backspace action.
       *
       * If there is a selected token it will remove the selected token. Otherwise it will mark the last token
       * as selected for the next time to backspace.
       *
       * @method TypeAheadTokenView#_handleBackspace
       * @private
       *
       * @param {jQuery.Event} event
       * @param {jQuery} selectedToken
       */
      tvp._handleBackspace = function (event, selectedToken) {
          //If token is already selected - remove it and move to the next one
          if (selectedToken.length !== 0) {
              var id = selectedToken.attr('data-id');
              this._controller.selected(_.without(this._controller.selected(), this._controller.getNodeById(id)));
              event.preventDefault();
          } else if (this.selectionStart === 0) {
              this._$el.find('.token-container:last-of-type').addClass('selected');
              event.preventDefault();
          }

      };

      /**
       * Event handler for the enter action.
       *
       * Selects an active node from the menu only if it is visible and exists.
       *
       * @method TypeAheadTokenView#_handleEnter
       * @private
       *
       * @param {jQuery.Event} event
       */
      tvp._handleEnter = function (event) {
          if (!_.isNull(this._$menu) && this.isVisible() && this._$menu.find('.active').length !== 0) {
              event.preventDefault();
              this._controller.selected(this._controller.selected().concat(this._$menu.find('.active').attr('data-id')));
          }

      };

      return TypeAheadView;
  }

);

define('oui/widgets/comboBox/comboBoxModel',[
    'underscore',
    'oui/widgets/tree/collection',
    'oui/utils/binarySearch'
], function (_, Collection, search) {
    'use strict';

    /**
     * @event ComboBoxModel#values-change
     * @param {Array} values The new values the combobox is now using to build the menu.
     */
    /**
     * @event ComboBoxModel#active-change
     * @param {Number} id The new id of the node to be marked as active
     */
    /**
     * @event ComboBoxModel#selection-change
     * @param {Number} id The new id of the node that is selected
     */
    /**
     * @event ComboBoxModel#invalid-selection
     */
    /**
     * @event ComboBoxModel#required-error
     */
    /**
     * @event ComboBoxModel#valueSelected
     * @deprecated
     * @param {Number} id The new id of the node that is selected
     */

    var cmp;

    /**
     * Data associated with the ComboBox component.
     *
     * @class ComboBoxModel
     * @private
     * @since 1.3.0
     * @extends Subscribable
     *
     * @param {String|null} initData Initial data for the widget
     * @param {ComboBox#Config} options Initial options passed to widget
     */
    function ComboBoxModel(initData, options) {
        // Super call
        Collection.call(this, options, initData);

        if (!this._options.multiSelect && !options.allowEmpty) {
            var firstNode = _.find(this.getVisibleNodes(), function (id) {
                var children =  this.getChildren(id);
                return children == null || children.length === 0;
            }.bind(this));

            if (firstNode != null) {
                this.setData([firstNode]);
            }

        }
    }

    ComboBoxModel.prototype = Object.create(Collection.prototype);
    cmp = ComboBoxModel.prototype;

    cmp.constructor = ComboBoxModel;

    /**
     * @method ComboBoxModel#trigger
     *
     * @param {Object} ev
     */
    cmp.trigger = function (ev) {
        Collection.prototype.trigger.apply(this, arguments);

        if (ev.name === 'change') {
            this.trigger(_.extend({}, ev, { name: 'selection-change' }), this._data);
            this.trigger(_.extend({}, ev, { name: 'valueSelected' }), this._data);
        }

        if (ev.name === 'invalid-error' && !ev.silent) {
            this.trigger('invalid-selection');
        }
    };

    /**
     * @method ComboBoxModel#validate
     *
     * @param {Array} data
     *
     * @returns {Object}
     */
    cmp.validate = function (data) {
        if (data == null && this._options.required) {
            return {
                valid: false,
                type: 'required',
                data: null
            };
        } else if (data != null && data.length !== _.compact(data).length) {
            return {
                valid: false,
                type: 'invalid',
                data: data
            };
        }

        return {
            valid: true
        };
    };

    /**
     * @method ComboBoxModel#is
     *
     * @param {String} attr
     *
     * @returns {Boolean}
     */
    cmp.is = function (attr) {
        if (attr === 'expanded') {
            return true;
        }

        return Collection.prototype.is.apply(this, arguments);
    };

    /**
     * @method ComboBoxModel#addAll
     *
     * @param {Array<Object|String>} nodes
     * @param {Number} modifiers
     */
    cmp.addAll = function (nodes, modifiers) {
        Collection.prototype.addAll.call(this, _.compact(_.map(nodes, function (node) {
            if (_.isString(node)) {

                if (node === 'divider') {
                    return undefined;
                }

                return {
                    _special: true,
                    value: node
                };
            }

            return node;
        })), modifiers);
    };

    /**
     * @method ComboBoxModel#setData
     *
     * @param {Array} newData
     * @param {Model#EventParams} eventParams
     */
    cmp.setData = function (newData, eventParams) {
        if (newData === undefined || newData.length === 0) {

            var existingData = this.getData();
            if (!this._options.multiSelect && !this._options.allowEmpty && existingData != null && existingData.length > 0) {
                return;
            }

            newData = [];

        }

        if (_.isArray(newData) && _.contains(newData, 'all')) {
            newData = _.difference(this._sorts['default'].ids, this.getAttr('special'));
        }

        Collection.prototype.setData.call(this, newData, eventParams);
    };

    /**
     * @method ComboBoxModel#addAttrs
     *
     * @param {Object} node
     */
    cmp.addAttrs = function (node) {
        Collection.prototype.addAttrs.apply(this, arguments);

        if (node._special) {
            this.addToAttr('special', this._getId(node));
        }

        if (this.getChildren(this._getId(node)).length > 0) {
            this.addToAttr('unselectable', this._getId(node));
        }
    };

    /**
     * Gets or sets the accepted values for the widget.
     *
     * @method ComboBoxModel#values
     * @protected
     * @param {Array|String} [values] The new acceptable values for the widget
     * @returns {Array|undefined}
     */
    cmp.values = function (values) {
        if (!arguments.length) {
            return _.sortBy(this.nodes(), function (node) {
                return this._sorts['default'].ids.indexOf(this._getId(node));
            }.bind(this));
        }

        if (this._options.selectAll) {
            values = ['all'].concat(values);
        }

        this.invalidate();
        this.addAll(values);

        var firstNode = _.find(this.getVisibleNodes(), function (id) {
            var children =  this.getChildren(id);
            return children == null || children.length === 0;
        }.bind(this));

        if (firstNode) {
            this.setData([firstNode], { silent: true });
        }

        this.trigger('values-change', this.nodes());
    };

    /**
     * @method ComboBoxModel#addToSorts
     *
     * @param {Object} node
     */
    cmp.addToSorts = function (node) {
        var value;
        var idx;

        if (this._options.autoSort && this._options.storeFns && this._options.storeFns.sortValue) {
            value = this._options.storeFns.sortValue(node);
            idx = search(this._sorts['default'].ids, value, function (id) {
                return this._options.storeFns.sortValue(this.node(id));
            }.bind(this));
            this._sorts['default'].ids.splice(idx, 0, this._getId(node));
        } else {
            Collection.prototype.addToSorts.apply(this, arguments);
        }
    };

    return ComboBoxModel;
});

define('oui/widgets/comboBox/comboBoxController',[
    'jquery',
    'underscore',
    'oui/utils/stringUtils',
    'oui/widgets/tree/treeController',
    'oui/utils/math'
], function ($, _, stringUtils, BaseController, math) {
    'use strict';

    /**
     * @method ComboBoxController~matches
     * @private
     *
     * @param {Object} node
     * @param {RegExp} regex
     * @param {Object} options
     *
     * @returns {Boolean}
     */
    function matches(node, regex, options) {

        if (options.viewFns.text) {
            return regex.test(options.viewFns.text(node));
        }
        return regex.test(options.viewFns.display(node));
    }

    /**
     * @class ComboBoxController
     * @extends TreeController
     * @private
     * @since 1.3.0
     *
     * @param {ComboBoxModel} model The ComboBoxModel to manipulate the data.
     * @param {ComboBox#Config} options The specific options regarding the controller.
     */
    function ComboBoxController(model, options) {
        BaseController.call(this, model, options);
    }

    ComboBoxController.prototype = Object.create(BaseController.prototype);

    ComboBoxController.prototype = _.extend(ComboBoxController.prototype, {

        /**
         * Gets or sets the accepted values for the widget.
         *
         * @method ComboBoxController#values
         * @protected
         * @param {Array|String} [values] The new acceptable values for the widget
         * @returns {Array|undefined}
         */
        values: function (values) {
            if (!arguments.length) {
                return this._model.values();
            }
            this._model.values(values);
        },

        /**
         * @method ComboboxContorller#convert
         * @param {jQuery} $el
         * @returns {Array<String|Object>}
         */
        convert: function ($el) {
            var optGroups = $el.children('optgroup');
            var vals = [];

            function buildChildren(group) { //jscs:ignore jsDoc
                return _.map($(group).children('option'), function (child) {
                    if (child.value === 'all') {
                        return 'all';
                    }
                    if (child.value === 'divider') {
                        return 'divider';
                    }
                    return {
                        name: $(child).text(),
                        value: child.value
                    };
                });
            }

            if (optGroups.length > 0) {
                //We have headers
                vals = _.map(optGroups, function (group) {
                    return {
                        name: $(group).attr('label'),
                        children: buildChildren(group)
                    };
                });

                vals = _.union(vals, buildChildren($el));
            } else {
                vals = buildChildren($el);
            }

            return vals;
        },
        /**
         * @method ComboBoxController#isValid
         *
         * @returns {Boolean}
         */
        isValid: function () {
            return this._model.isValid();
        },

        /**
         * @method ComboBoxController#error
         *
         * @returns {Object}
         */
        error: function () {
            return this._model.error();
        },

        /**
         * @method ComboBoxController#filter
         *
         * @param {String} query
         */
        filter: function (query) {
            var regex;
            var vals;

            if (query == null) {
                this._model.removeFilter();
            } else {
                regex = new RegExp(stringUtils.regexEscape(query), 'i');
                vals = _.filter(this._model.nodes(), function (node) {
                    var childMatches = false;

                    if (this._options.filterShowHierarchy) {
                        var children = this._model.getChildren(_.get(node, this._options.recordKeys.id));

                        childMatches = _.some(children, function (child) {
                            return matches(this._model.node(child), regex, this._options);
                        }.bind(this));
                    }

                    return matches(node, regex, this._options) || childMatches;
                }.bind(this));

                this._model.addAsFilter(vals);
            }
        },

        /**
         * @method ComboBoxController#shiftActive
         *
         * @param {Number} delta
         */
        shiftActive: function (delta) {
            //move the selection down one - if nothing is selected then select top
            //but start at the selected item
            var active = this.active();
            var comboItems = _.difference(this.getVisibleNodes(), this.unselectable());
            var activeIndex = _.indexOf(comboItems, active);

            this.active(comboItems[math.mod(activeIndex + delta, comboItems.length)]);

        },

        /**
         * @method ComboBoxController#getNodeByDisplayValue
         *
         * @param {String} query
         *
         * @returns {Object?}
         */
        getNodeByDisplayValue: function (query) {
            return _.find(this.nodes(), function (node) {
                return this._options.itemRenderer.asText(this.getNodeWithAttributes(_.get(node, this._options.recordKeys.id))) === query;
            }.bind(this));
        }

    });

    return ComboBoxController;
});


define('template-29',[],function () { return '<div class="orcl-widget orcl-combo-box combo-box-list">\n    <button type="button" role="combobox" aria-expanded="false" class="btn btn-blue-light combo-button combo-value">\n        <span></span>\n    </button>\n</div>\n';});


define('template-30', ['underscore'], function (_) { return function(data) {
var __t, __p = '', __e = _.escape, __j = Array.prototype.join;
function print() { __p += __j.call(arguments, '') }

    var classes = ['combo-item'];    var role;    if (data.attrs.selected) {        classes.push('selected');    }    if (data.attrs.active) {        classes.push('active');    }    if (data.attrs.unselectable || data.attrs.hasChildren) {        role = 'group';        classes.push('unselectable');    } else {        role = 'option';    };
__p += '<li role="' +
((__t = ( role )) == null ? '' : __t) +
'" id="' +
__e( data.pageId ) +
'" class="' +
((__t = ( classes.join(' ') )) == null ? '' : __t) +
'" data-combo-id="' +
__e( data.id ) +
'"> ' +
((__t = ( data.itemMarkup )) == null ? '' : __t) +
'</li>';
return __p
};});


define('template-31', ['underscore'], function (_) { return function(data) {
var __t, __p = '', __j = Array.prototype.join;
function print() { __p += __j.call(arguments, '') }

    var classes = ['combo-dropdown-menu'].concat(data.classes);    if (data.multiSelect) {        classes = classes.concat(['multi-select', 'select-menu']);    };
__p += '<ul role="listbox" id="' +
((__t = ( data.id )) == null ? '' : __t) +
'" aria-multiselectable="' +
((__t = ( data.multiSelect )) == null ? '' : __t) +
'" aria-label="' +
((__t = ( data.locale.menu )) == null ? '' : __t) +
'"    class="' +
((__t = ( classes.join(' ') )) == null ? '' : __t) +
'"></ul>';
return __p
};});


define('template-32',[],function () { return '<div class="orcl-widget orcl-combo-box combo-box-list">\n</div>\n';});

/**
 * @module inlineStyle
 * @private
 */
define('oui/utils/inlineStyle',[
    'underscore'
], function (_) {
    'use strict';

    return {
        /**
         * @method inlineStyle~getPropertyValue
         * @param {jQuery} $el The element to grab the inline style property off of
         * @param {String} prop The property to look for in the attribute
         *
         * @returns {String|undefined} The value of the property in the style attribute.
         */
        getPropertyValue: function ($el, prop) {
            var styles = $el.attr('style');
            var value;
            if (styles) {
                styles.split(';').forEach(function (e) {
                    var style = e.split(':');
                    if (style[0].trim() === prop) {
                        value = style[1].trim();
                    }
                });
            }
            return value;
        },

        /**
         * Forces the removal of a CSS property that cannot be removed with `$(...).css(property, '');`
         *
         * This can happen when a property is applied by manipulating the `style` attribute or using the `!important`
         * modifier.
         *
         * @method inlineStyle~setCalcProperty
         * @private
         * @param {*|HTMLElement} element The element to operate on
         * @param {String} property The property to add a calc'd field for
         * @param {String} value The calculated size, generally in the form "XX% +/- YYpx"
         */
        setCalcProperty: function (element, property, value) {
            var str = 'calc(' + value + ')';

            _.each(element, function (el) {
                this.forceRemoveProperty(el, property);
                el.style[property] = str;
            }.bind(this));
        },
        /**
         * Returns the specified CSS property as a pixel value
         *
         * @method inlineStyle~intCssProperty
         * @private
         * @param {*|HTMLElement} elem The element to operate on
         * @param {String} property The CSS property to retrieve
         * @returns {Number} the pixel size of the CSS property
         */
        intCssProperty: function (elem, property) {
            if (elem[property] && (property === 'outerWidth' || property === 'outerHeight')) {
                return elem[property](true);
            }
            return parseInt(elem.css(property).slice(0, -2), 10);
        },
        /**
         * Forces the removal of a CSS property that cannot be removed with `$(...).css(property, '');`
         *
         * This can happen when a property is applied by manipulating the `style` attribute or using the `!important`
         * modifier.
         *
         * @method inlineStyle~forceRemoveProperty
         * @private
         * @param {*|HTMLElement} element The element to operate on
         * @param {String} property the CSS property to remove
         */
        forceRemoveProperty: function (element, property) {
            if (element.length) {
                element = _.first(element);
            }
            element.style.removeProperty(property);
        }
    };
});


define('template-33', ['underscore'], function (_) { return function(data) {
var __t, __p = '';
__p += '<optgroup id="' +
((__t = ( data.id )) == null ? '' : __t) +
'" label="' +
((__t = ( data.display )) == null ? '' : __t) +
'">' +
((__t = ( data.children )) == null ? '' : __t) +
'</optgroup>';
return __p
};});


define('template-34', ['underscore'], function (_) { return function(data) {
var __t, __p = '';
__p += '<option id="' +
((__t = ( data.pageId )) == null ? '' : __t) +
'" data-combo-id="' +
((__t = ( data.id )) == null ? '' : __t) +
'" value="' +
((__t = ( data.id )) == null ? '' : __t) +
'">' +
((__t = ( data.display )) == null ? '' : __t) +
'</option>';
return __p
};});


define('template-35', ['underscore'], function (_) { return function(data) {
var __t, __p = '';
__p += '<li class="divider"></li>';
return __p
};});


define('template-36', ['underscore'], function (_) { return function(data) {
var __t, __p = '';
__p += '<li role="option" title="' +
((__t = ( data.locale.selectAll )) == null ? '' : __t) +
'"    class="combo-item ' +
((__t = ( data.active ? 'active' : '' )) == null ? '' : __t) +
'"    data-combo-id="all"> <a href="#"> <i class="check"></i> ' +
((__t = ( data.locale.selectAll )) == null ? '' : __t) +
' </a></li>';
return __p
};});

define('oui/widgets/comboBox/comboBoxButtonView',['require','underscore','jquery','oui/base/view','oui/utils/stringUtils','template-29','template-30','template-31','template-32','oui/utils/inlineStyle','template-33','template-34','template-35','template-36','oui/mixins/popup','oui/utils/scrollingUtils','oui/utils/arrayUtils'],function (require) { //jshint ignore:line
    'use strict';

    var _ = require('underscore');
    var $ = require('jquery');
    var View = require('oui/base/view');
    var stringUtils = require('oui/utils/stringUtils');
    var comboBoxListTemplate = require('template-29');
    var nodeTemplate = require('template-30');
    var menuMarkup = require('template-31');
    var wrapTemplate = require('template-32');
    var inlineStyle = require('oui/utils/inlineStyle');
    var nativeGroup = require('template-33');
    var nativeOption = require('template-34');
    var dividerTemplate = require('template-35');
    var allTemplate = require('template-36');
    var popupMixin = require('oui/mixins/popup');
    var scrollingUtils = require('oui/utils/scrollingUtils');
    var arrayUtils = require('oui/utils/arrayUtils');

    var cvp;

    /**
     * This view displays the combobox as a dropdown without an input element, so no text can be typed
     *
     * @class ComboBoxButtonView
     * @private
     * @since 1.3.0
     * @extends View
     * @mixes popup
     *
     * @param {ComboBoxModel} model The ComboBoxModel that is responsible for this view.
     * @param {ComboBoxController} controller The ComboBoxController that will modify the model.
     * @param {jQuery|null} element The default HTML input element that users can interact with.
     * @param {ComboBox#Config} options View specific options.
     */
    function ComboBoxButtonView(model, controller, element, options, settings) {
        var that = this;

        // Super call
        View.call(this, options, element, controller, settings);

        this._$menu = this._$popup = $(menuMarkup({
            multiSelect: options.multiSelect,
            id: this.id() + '-menu',
            locale: options.locale,
            classes: options.menuClasses || []
        }));

        this._$menu.on('click', '[data-trigger]', function (event) {
            var customEvent = $.Event($(this).attr('data-trigger'));
            that._$el.trigger(customEvent);

            if (customEvent.isDefaultPrevented()) {
                event.preventDefault();
            }

            if (customEvent.isPropagationStopped()) {
                event.stopPropagation();
            }
        });

        this._$originalElement = this._$el.children('.combo-original-element');
        this._$dropDownButton = this._$el.children('.combo-button');
        this._$valueEl = this._$el.children('.combo-value');

        model.on({
            'view.change': this.updateValue.bind(this),
            'view.active': this.updateActive.bind(this),
            'view.invalid-error': this.invalid.bind(this),
            'view.required-error': this.invalid.bind(this),
            'view.addAll': this._createDropdown.bind(this)
        });

        this._setupCustom();
        this._initListeners();

        this._createDropdown();

        this.updateValue(this._controller.selected());
    }

    ComboBoxButtonView.prototype = Object.create(View.prototype);
    cvp = ComboBoxButtonView.prototype;

    cvp.constructor = ComboBoxButtonView;

    _.extend(cvp, popupMixin);

    /**
     * @method ComboboxButtonView#invalid
     *
     * @param {Array} attempted
     */
    cvp.invalid = function (attempted) {
        this._$el.add(this._$el.closest('.control-group')).toggleClass('error', !this._controller.isValid() && this._visited);

        if (_.isArray(attempted)) {
            this.updateValue(attempted);
        } else {
            this.updateValue([]);
        }
    };

    /**
     * @method ComboboxButtonView#_supported
     * @private
     * @returns {Boolean}
     */
    cvp._supported = function () {
        return true;
    };

    /**
     * @method ComboboxButtonView#_setupCustom
     * @private
     */
    cvp._setupCustom = function () {
        var originalId = this._$originalElement.attr('id');

        this._$dropDownButton.attr('title', this._options.titleText);

        // Swap IDs on the button so that corresponding labels still work
        this._$originalElement
        .attr('data-original-id', originalId)
        .removeAttr('id');
        this._$dropDownButton.attr('id', originalId);
    };

    /**
     * @method ComboboxButtonView#template
     *
     * @returns {Function}
     */
    cvp.template = function () {
        return comboBoxListTemplate;
    };

    /**
     * @method ComboboxButtonView#buildNativeElement
     * @param {Combobox#Config} opts
     * @param {jQuery} el
     *
     * @returns {jQuery}
     */
    cvp.buildNativeElement = function (opts, el) {
        var $element;

        // We were given a jquery element
        // Whether it's an <input> or a <select> we need to replace it with our own thing
        if (!_.isNull(el) && el.jquery) {
            el.wrap(wrapTemplate);

            el.addClass('combo-original-element');

            if (el.is('input')) {
                //Hide the input and make a select
                el.parent().append('<select>');
            }

            $element = el.parent();
        } else {
            el = $('<select>').wrap(wrapTemplate);

            el.addClass('combo-original-element');

            $element = el.parent();
        }

        $element.find('select')
        .addClass('combo-value')
        .attr('multiple', opts.multiSelect);

        return $element;
    };

    /**
     * @method ComboboxButtonView#buildElement
     * @param {Combobox#Config} opts
     * @param {jQuery} el
     *
     * @returns {jQuery}
     */
    cvp.buildElement = function (opts, el) {
        var $element = $(this.template());
        var $value = $element.children('.combo-value');

        // We were given a jquery element
        // Whether it's an <input> or a <select> we need to replace it with our own thing
        if (!_.isNull(el) && el.jquery) {
            el.after($element);

            var attrs = {};
            if (el.length > 0) {
                _.each(el[0].attributes, function (key) {
                    if (!_.contains(['id', 'style', 'type', 'name'], key.nodeName)) {
                        attrs[key.nodeName] = key.value;
                    }
                });
            }

            _.each(attrs, function (value, key) {
                if (key !== 'class') {
                    $value.attr(key, value);
                    el.removeAttr(key);
                }
                el.attr('data-original-' + key, value);
            });

            // Move the element into the wrapper and hide it
            el.prependTo($element).css('display', 'none').addClass('combo-original-element');

            // Set the new element to the same width as the old one if it was applied specifically
            if (inlineStyle.getPropertyValue(el, 'width')) {
                $element.css('width', el.css('width'));
            }

            if (opts.copyClasses && attrs['class']) {
                $element.attr('class', $element.attr('class') + ' ' + attrs['class']);
            }
        }

        $element.toggleClass('shrink-to-fit', opts.shrinkToFit);

        return $element;
    };

    /**
     * @method ComboBoxButtonView#init
     */
    cvp.init = function () {
        var $origElement = this._$el.find('.combo-original-element');

        if ($origElement.is('select') && this._controller.values().length === 0) {
            var value = $origElement[0].value;
            this._controller.values(this._controller.convert($origElement));

            var option = _.find($origElement.find('option'), function (option) {
                return option.value === value;
            });

            if (option != null) {
                if (value === 'all') {
                    this._controller.selected([value]);
                } else {
                    this._controller.selected([option.value]);
                }
            }
        }

        //TODO: find a way to set the ID before building the element?
        this._$el.find('.instructions').attr('id', this.id() + '-instructions');
        this._$valueEl.attr('aria-describedby', ((this._$valueEl.attr('aria-describedby') || '') + ' ' + this.id() + '-instructions').trim());
    };

    /**
     * @method ComboBoxButtonView#destroy
     * @param {Object} clean
     */
    cvp.destroy = function (clean) {
        var that = this;

        this._$menu.remove();

        var attrs = {};

        if (this._$originalElement[0] != null) {
            _.each(this._$originalElement[0].attributes, function (key) {
                if (key.nodeName.match('data-original-')) {
                    attrs[key.nodeName.replace('data-original-', '')] = key.value;
                }
            });

            _.each(attrs, function (value, key) {
                that._$originalElement.attr(key, value);
                that._$originalElement.removeAttr('data-original-' + key);
            });
        }

        if (clean.self) {
            this._$el.remove();
        } else {
            this._$dropDownButton.remove();
            this._$valueEl.remove();
            this._$el.find('.instructions').remove();
        }
    };

    /**
     * @method ComboBoxButtonView#updateValue
     *
     * @param {Array} selected
     */
    cvp.updateValue = function (selected) {
        this._$el.add(this._$el.closest('.control-group')).toggleClass('error', !this._controller.isValid() && this._visited);

        this.active(this._controller.active());

        if (!this._$el.hasClass('native')) {
            var renderedNodes = this._options.renderer.render(_.compact(_.map(selected, function (id) {
                var node = this._controller.getNodeWithAttributes(id);
                if (node.data != null) {
                    return this._controller.getNodeWithAttributes(id);
                }
            }.bind(this))), this._options);

            if (renderedNodes !== this._$valueEl[0].innerHTML) {
                this._$valueEl[0].innerHTML = renderedNodes;
            }
        }

        if (this._$el.find('select').length > 0) {
            this._$el.find('select')[0].selectedIndex = -1;
            this._$el.find('option').removeAttr('selected');
            _.each(selected, function (select) {
                this._$el.find('option[value="' + select + '"]').prop('selected', true);
            }.bind(this));
        }

        this._$menu.find('.checked, .selected').removeClass('checked selected').attr('aria-selected', false);
        _.each(selected, function (id) {
            this._findMenuItem(id)
            .addClass('checked selected')
            .attr('aria-selected', true);
        }.bind(this));

        //Used by subclass
        return this._$valueEl[0].textContent.trim();
    };

    /**
     * @method ComboboxButtonView#updateActive
     *
     * @param {String} active
     */
    cvp.updateActive = function (active) {
        if (this._options.multiSelect) {
            this.active(this._controller.active());
        } else {
            this.updateValue([active]);
        }
    };

    /**
     * @method ComboboxButtonView#active
     *
     * @param {String} id
     */
    cvp.active = function (id) {
        var $menuItem;
        this._$menu.find('.active').removeClass('active');
        $menuItem = this._findMenuItem(id).addClass('active');

        if (this.isVisible()) {
            this._$valueEl.attr('aria-activedescendant', $menuItem.attr('id'));
        }

        scrollingUtils.positionScrollTop($menuItem, this._$menu);
    };

    /**
     * The function that handles the keyup, keydown, mousedown, mousein, and mouseout events of the ComboBox
     *
     * @private
     * @method ComboBoxButtonView#_initListeners
     */
    cvp._initListeners = function () {
        var that = this;
        var cancelBlur;

        this._$dropDownButton.on('mousedown', function (event) {
            //don't lose focus from anything
            event.preventDefault();
            that._$valueEl.focus();

            if (that.isEditable() && !that.isVisible()) {
                that._expand();
            } else {
                that.hidePopup();
            }

        });

        this._$dropDownButton.on('click', function (event) {
            event.preventDefault();
        });

        this._$valueEl.on('blur', function (event) {
            event.preventDefault();
            if (!cancelBlur && that.isVisible()) {
                that.hidePopup();
                if (that._controller.isValid()) {
                    that.updateValue(that._controller.selected());
                }
            }
        }).on('focus', function () {
            if (that.isEditable() && that._options.openOnFocus && !that.isVisible()) {
                that._expand();
            }
        });

        this.addShortcuts({
            'ESC': {
                fn: function () { //jscs:ignore jsDoc
                    that.hidePopup();
                    that.updateValue(that._controller.selected());

                    if (!that._options.multiSelect && that._controller.selected().length > 0) {
                        that._controller.active(that._controller.selected()[0]);
                    }
                },
                stopPropagation: this.isVisible
            },
            'ENTER': {
                fn: function () { //jscs:ignore jsDoc
                    if (that.isEditable()) {
                        if (!that.isVisible()) {
                            that._expand();
                            that._$valueEl.focus();
                        } else {
                            that._handleEnter();
                        }
                    }
                },
                prevention: true,
                stopPropagation: this.isVisible
            },
            'SPACE': {
                fn: function () { //jscs:ignore jsDoc
                    if (that.isEditable()) {
                        if (!that._$el.hasClass('combo-box-input') && !that.isVisible()) {
                            that._expand();
                            that._$valueEl.focus();
                        } else if (that.isVisible() && that._options.multiSelect && !that._options.keyboardInput) {
                            that._controller.selected([that._controller.active()]);
                            that.hidePopup();
                        }
                    }
                },
                prevention: function () { //jscs:ignore jsDoc
                    return !that._$el.hasClass('combo-box-input') && !that.isVisible() ||
                        (that.isVisible() && that._options.multiSelect && !that._options.keyboardInput);
                }
            },
            'shift-UP': function () { //jscs:ignore jsDoc
                that._controller.shiftActive(-1);
                that._select(that._$menu.find('.active'), true, false);
            },
            'shift-DOWN': function () { //jscs:ignore jsDoc
                that._controller.shiftActive(1);
                that._select(that._$menu.find('.active'), true, false);
            },
            'SHIFT': function () { //jscs:ignore jsDoc
                var active = that._findMenuItem(that._controller.active());
                active.addClass('axis');
            },
            'ctrl-SPACE': function () { //jscs:ignore jsDoc
                if (that.isVisible() && that._options.multiSelect) {
                    that._toggleSelect(that._controller.active());
                }
            },
            'UP': {
                fn: function () { //jscs:ignore jsDoc
                    if (that.isEditable()) {
                        if (that.isVisible()) {
                            that._controller.shiftActive(-1);
                        } else {
                            that._expand();
                            that._$valueEl.focus();
                        }
                    }
                },
                prevention: true
            },
            'DOWN': {
                fn: function () { //jscs:ignore jsDoc
                    if (that.isEditable()) {
                        if (that.isVisible()) {
                            that._controller.shiftActive(1);
                        } else {
                            that._expand();
                            that._$valueEl.focus();
                        }
                    }
                },
                prevention: true
            },
            'alt-DOWN': function () { //jscs:ignore jsDoc
                if (that.isEditable()) {
                    that._expand();
                }
            },
            'alt-UP': function () { //jscs:ignore jsDoc
                that.hidePopup();
            }
        });

        //TODO: look into shortcut mixin if it can have an up/down function
        this.addShortcut('SHIFT', {
            up: true,
            fn: function () { //jscs:ignore jsDoc
                that._$menu.find('.axis').removeClass('axis');
            }
        });

        if (!this._options.keyboardInput) {
            //http://api.jquery.com/keypress/
            //jQuery normalizes the .which property for us
            that._$el.on('keypress', function (event) {
                var string = String.fromCharCode(event.which);
                if (that.isEditable() && string.length > 0) {
                    that._find(string);
                }
            });
        }

        that._$menu.on('click', '[data-attr][data-toggle=true]', function (event) {
            var attr = event.currentTarget.getAttribute('data-attr');
            event.preventDefault();
            event.stopPropagation();
            var id = this._getId($(event.currentTarget).closest('[data-combo-id]'));
            this._controller.toggle(attr, id);
        }.bind(this));

        //set click behavior for list items
        that._$menu.on('click', '[role=option]', function (event) {
            event.preventDefault();

            that._select($(this), event.shiftKey, event.ctrlKey || event.metaKey);
        });

        that._$menu.on('click', 'a[href]', function (event) {
            event.preventDefault();
        });

        that._$el.on('change', 'select', function () {
            var selected = _.chain($(this).find('option'))
            .map(function (option) {
                if (option.selected) {
                    return option.value;
                }
            })
            .compact()
            .value();
            that._controller.selected(selected);
        });

        that._$menu.on({
            'mouseover mouseenter': function () { //jscs:ignore jsDoc
                cancelBlur = true;
            },
            'mousedown': function (event) { //jscs:ignore jsDoc
                event.preventDefault();
                that._$valueEl.focus();
            },
            'mouseleave hidden.popup': function () { //jscs:ignore jsDoc
                cancelBlur = false;
            }
        });

        this._$menu.on('hidden.popup', function () {
            this._$valueEl.removeAttr('aria-activedescendant');
            this._$valueEl.attr('aria-expanded', false);
        }.bind(this));

        this._$menu.on('showing.popup', function () {
            this._$menu
                .css('min-width', this._$el.innerWidth())
                .removeClass('grow-to-text');

            if (this._options.menuSize === 'input-width') {
                this._$menu.width(this._$el.innerWidth());
            } else if (this._options.menuSize === 'grow-to-text') {
                this._$menu.width('');
                this._$menu.addClass('grow-to-text');
            } else {
                this._$menu.width(this._options.menuSize);
            }
        }.bind(this));

        this._$menu.on('shown.popup', function () {
            this._$valueEl.attr('aria-expanded', true);
            this._$valueEl.attr('aria-activedescendant', this._$menu.find('.active').attr('id'));
            this._$valueEl.attr('aria-owns', this._$menu.attr('id'));
        }.bind(this));

        //IE9, IE10, and IE11 but not Edge
        //Super janky but need to refocus the element
        //when the scrollbar is mousedown'd.
        $(window).on('mousedown', function (event) {
            if ($(event.target).is(that._$menu)) {
                event.preventDefault();
                setTimeout(function () {
                    that._$valueEl.focus();
                }, 0);
            }
        });
    };

    /**
     * @method ComboBoxButtonView#_toggleSelect
     * @private
     *
     * @param {String} id
     */
    cvp._toggleSelect = function (id) {

        if (!this._options.multiSelect) {
            this._controller.selected([id]);
        } else {
            var selected = this._controller.selected();

            if (_.contains(selected, id)) {
                this._controller.selected(_.without(selected, id));
            } else {
                this._controller.selected(_.union(selected, [id]));
            }
        }
    };

    /**
     * @method ComboBoxButtonView#_findMenuItem
     * @private
     *
     * @param {String} id
     */
    cvp._findMenuItem = function (id) {
        return this._$menu.find('[data-combo-id="' + id + '"]');
    };

    /**
     * @method ComboBoxButtonView#_select
     * @private
     *
     * @param {Object} item
     * @param {Boolean} shiftKey
     * @param {Boolean} ctrlKey
     */
    cvp._select = function (item, shiftKey, ctrlKey) {
        var id = this._getId(item);

        this._controller.active(id);

        if (this._options.multiSelect && (ctrlKey || shiftKey || this._options.action === 'checkbox')) {
            if (ctrlKey || (!shiftKey && this._options.action === 'checkbox')) {
                this._toggleSelect(id);
            } else if (shiftKey) {
                var comboItems = this._$menu.find('.combo-item');
                var selectedIndex = comboItems.index(item);
                var prevSelectedIndex = comboItems.index(this._$menu.find('.axis'));
                var selected;

                if (selectedIndex < prevSelectedIndex) {
                    selected = comboItems.slice(selectedIndex, prevSelectedIndex + 1);
                } else {
                    selected = comboItems.slice(prevSelectedIndex, selectedIndex + 1);
                }

                selected = _.map(selected, function (el) {
                    return this._getId($(el));
                }.bind(this));

                this._controller.selected(selected);
            }
        } else {
            this._controller.selected([id]);
            this.hidePopup();
        }
    };

    /**
     * @method ComboBoxButtonView#_find
     * @private
     *
     * @param {String} searchString
     */
    cvp._find = function (searchString) {
        var display = this._options.viewFns.display;
        var regex = new RegExp('^' + stringUtils.regexEscape(searchString), 'gi');
        var ids = this._controller.getVisibleNodes();
        var found;

        ids = arrayUtils.rotate(ids, _.indexOf(ids, this._controller.active()));

        found = _.find(ids, _.flowRight(regex.test.bind(regex), display, this._controller.node.bind(this._controller)));

        if (found != null) {
            if (this._options.multiSelect) {
                this._expand();
                this._controller.active(found);
            } else {
                this._controller.active(found);
                this._toggleSelect(found);
            }
        }
    };

    /**
     * The function that creates (but does not actually reveal) the contents of the ComboBox dropdown.
     *
     * @private
     * @method ComboBoxButtonView#_createDropdown
     */
    cvp._createDropdown = function () {
        var that = this;
        var regex;

        var values = this._controller.getVisibleNodes();

        var span = document.createElement('span');

        /**
         * @method ComboBoxButtonView~buildMarkup
         * @private
         * @param {Array} vals
         *
         * @returns {String}
         */
        function buildMarkup(vals) {
            return _.reduce(vals, function (memo, val) {
                if (val === 'divider') {
                    return memo += dividerTemplate();
                } else if (val === 'all') {
                    return memo += allTemplate({
                        locale: that._options.locale
                    });
                }

                var node = that._controller.getNodeWithAttributes(val);
                var data = {
                    id: val,
                    pageId: that.id() + '-' + val,
                    attrs: node.attrs,
                    filter: regex,
                    itemMarkup: that._options.itemRenderer.render(node, that._options)
                };

                memo += nodeTemplate(data);

                return memo;
            }, '');
        }

        /**
         * @method ComboBoxButtonView~buildNativeMarkup
         * @private
         * @param {Array} vals
         *
         * @returns {String}
         */
        function buildNativeMarkup(vals) {
            return _.reduce(vals, function (memo, val) {
                var display = that._options.viewFns.display(val);
                var text;
                span.innerHTML = display;
                text = span.textContent;

                if (val === 'divider') {
                    return memo;
                } else if (val === 'all') {
                    return memo += nativeOption({
                        id: 'all',
                        pageId: that.id() + '-' + val,
                        display: that._options.locale.selectAll
                    });
                }

                if (that._options.viewFns.text) {
                    text = that._options.viewFns.text(val);
                }

                if (val._children !== undefined) {
                    memo += nativeGroup({
                        id: val,
                        display: text,
                        children: buildNativeMarkup(val._children)
                    });
                } else {
                    memo += nativeOption({
                        id: val,
                        pageId: that.id() + '-' + val,
                        display: text
                    });
                }

                return memo;
            }, '');
        }

        if (!this._$el.hasClass('native')) {
            this._$menu[0].innerHTML = buildMarkup(values);
        } else if (values.length !== 0) {
            this._$el.find('select')[0].innerHTML = buildNativeMarkup(values);
        }
    };

    /**
     * Expands the ComboBox drop down
     *
     * @private
     * @method ComboBoxButtonView#_expand
     *
     * @param {Object} [modifiers]
     * @param {Boolean} [modifiers.skipActive=false]
     */
    cvp._expand = function (modifiers) {
        this.showPopup(this._$el);

        if (modifiers == null || !modifiers.skipActive) {
            if (this._controller.active() == null) {
                if (this._controller.selected().length === 0) {
                    this._controller.active(this._controller.getVisibleNodes()[0]);
                } else {
                    this._controller.active(_.first(this._controller.selected()));
                }
            } else if (!this._options.multiSelect && this._controller.selected().length > 0 &&
                       this._controller.active() !== _.first(this._controller.selected())) {
                this._controller.active(_.first(this._controller.selected()));
            } else {
                this.updateActive(this._controller.active());
            }
        }
    };

    /**
     * @method ComboBoxButtonView#_getId
     * @param {jQuery} $el
     * @returns {Number|String}
     */
    cvp._getId = function ($el) {
        return $el.attr('data-combo-id');
    };

    /**
     * @method ComboboxButtonView#_handleEnter
     * @private
     */
    cvp._handleEnter = function () {
        this._controller.selected([this._getId(this._$menu.find('.combo-item.active'))]);

        //If opened - close the menu
        if (this.isVisible()) {
            this.hidePopup();
        }
    };

    /**
     * Returns the element that identifies the widget.
     *
     * @method ComboboxButtonView#getIdentifyingElement
     *
     * @returns {jQuery}
     */
    cvp.getIdentifyingElement = function () {
        return this._$el.find('.combo-value');
    };

    /**
     * @method ComboboxButtonView#_applyDisabled
     * @private
     * @param {Boolean} disabled
     */
    cvp._applyDisabled = function (disabled) {
        View.prototype._applyDisabled.apply(this, arguments);

        //Also toggle disable on the native element within the combobox - mobile
        this._$el.find('select')
            .toggleClass('disabled', disabled)
            .prop('disabled', disabled);
    };

    /**
     * @method ComboboxButtonView#_applyReadOnly
     * @private
     * @param {Boolean} readonly
     */
    cvp._applyReadonly = function (readonly) {
        View.prototype._applyReadonly.apply(this, arguments);

        if (readonly) {
            this._$el.find('.combo-original-element').prop('readonly', false);
            this._$el.find('.combo-value')
                .prop('disabled', false)
                .attr('aria-readonly', readonly);
        } else {
            this._$el.find('.combo-value')
                .removeAttr('aria-readonly');
        }
    };

    return ComboBoxButtonView;
});


define('template-37',[],function () { return '<div class="orcl-widget orcl-combo-box combo-box-input">\n    <input role="combobox" aria-expanded="false" aria-autocomplete="list" class="combo-value" type="text" />\n    <button type="button" class="btn btn-picker combo-button" tabindex="-1"><span class="caret"></span></button>\n</div>\n';});

define('oui/widgets/comboBox/comboBoxInputView',['require','underscore','oui/widgets/comboBox/comboBoxButtonView','oui/utils/inputUtils','oui/utils/stringUtils','template-37','oui/utils/keyCode','oui/utils/escapingUtils'],function (require) {
    'use strict';

    var _ = require('underscore');
    var View = require('oui/widgets/comboBox/comboBoxButtonView');
    var inputUtils = require('oui/utils/inputUtils');
    var stringUtils = require('oui/utils/stringUtils');
    var comboBoxInputTemplate = require('template-37');
    var keyCode = require('oui/utils/keyCode');
    var escapingUtils = require('oui/utils/escapingUtils');

    var cvp;

    /**
     * This view displays the combobox as a dropdown without an input element, so no text can be typed
     *
     * @class ComboBoxInputView
     * @private
     * @since 1.3.0
     * @extends ComboBoxButtonView
     *
     * @param {ComboBoxModel} model The ComboBoxModel that is responsible for this view.
     * @param {ComboBoxController} controller The ComboBoxController that will modify the model.
     * @param {jQuery|null} element The default HTML input element that users can interact with.
     * @param {ComboBox#Config} options View specific options.
     */
    function ComboBoxInputView(model, controller, element, options, settings) {
        // Super call
        View.call(this, model, controller, element, options, settings);

        this._typing = false;

        model.on({
            'view.active-change': this.updateActive.bind(this),
            'view.change': this.clearFilter.bind(this),
            'view.filter': this.filtered.bind(this)
        });

        this._$el.on('focusin', function () {
            if (this.isEditable() && !this._options.multiSelect) {
                this._$valueEl.select();
            }
        }.bind(this));
    }

    ComboBoxInputView.prototype = Object.create(View.prototype);
    cvp = ComboBoxInputView.prototype;

    cvp.constructor = ComboBoxInputView;

    /**
     * @method ComboboxInputView#template
     *
     * @returns {Function}
     */
    cvp.template = function () {
        return comboBoxInputTemplate;
    };

    /**
     * @method ComboboxInputView#updateActive
     */
    cvp.updateActive = function () {
        this.active(this._controller.active());
    };

    /**
     * @method ComboboxInputView#_setupCustom
     * @private
     */
    cvp._setupCustom = function () {
        var originalId = this._$originalElement.attr('id');

        this._$dropDownButton.attr('title', this._options.titleText);

        // Swap IDs on the button so that corresponding labels still work
        this._$originalElement
            .attr('data-original-id', originalId)
            .removeAttr('id');
        this._$valueEl.attr('id', originalId);
    };

    /**
     * @method ComboboxInputView#invalid
     */
    cvp.invalid = function () {
        this._$el.add(this._$el.closest('.control-group')).addClass('error');
    };

    /**
     * @method ComboboxInputView#updateValue
     * @param {String} id
     */
    cvp.updateValue = function (id) {
        //Super call
        var value = View.prototype.updateValue.call(this, id);

        if (!this._$el.hasClass('native')) {
            this._$valueEl[0].value = escapingUtils.unescape(value, this._options);
        }

    };

    /**
     * @method ComboboxInputView#_initListeners
     * @private
     */
    cvp._initListeners = function () {
        var that = this;
        var cancelInput = false;

        //Super call
        View.prototype._initListeners.call(this);

        //Setup input only listeners
        this._$el.on('keyup', function (event) {
            switch (event.keyCode) {
                case keyCode.BACKSPACE:
                    var value = escapingUtils.escape(that._$valueEl[0].value, that._options);

                    that._autoComplete(stringUtils.getQuery(value, that._$valueEl[0].selectionStart, that._options.multiSelect, that._options.delimiter.trim()));
                    cancelInput = false;
            }
        })
        .on('keydown', function (event) {
            switch (event.keyCode) {
                case keyCode.BACKSPACE:
                    cancelInput = true;
            }
        })

        //Grid triggers input event from the column to trigger the auto complete after
        //the column editor was shown from a character key
        .on('keypress', function (event) {
            var value = inputUtils.insertCharacter(event.target, event.which, event.shiftKey);
            var character = String.fromCharCode(event.which);
            if (!event.shiftKey) {
                character = character.toLowerCase();
            }

            if (that.isEditable() && !that._$el.hasClass('native') && !cancelInput) {

                if (event.keyCode !== keyCode.ENTER &&
                    event.keyCode !== keyCode.TAB &&
                    !(event.which == null || event.which === 0)) {

                    value = escapingUtils.escape(value, that._options);

                    //TODO: look into dealing with multi-character delimiters
                    if (character !== that._options.delimiter.trim()) {
                        that._autoComplete(stringUtils.getQuery(value, inputUtils.getInputSelection(event.target).start,
                                                    that._options.multiSelect, that._options.delimiter.trim()));
                    }
                }
            }
            cancelInput = false;
        })
        .on('change', 'input', function () {
            if (!that._$el.hasClass('open')) {
                that._handleEnter();
            }
        });

    };

    /**
     * @method ComboboxInputView#_autoComplete
     * @private
     * @param {String} text
     */
    cvp._autoComplete = function (text) {
        this._typing = true;

        this._controller.filter(text.trim() || null);

        if (this._controller.getVisibleNodes().length > 0) {
            this._expand({ skipActive: true });
            this._controller.active(_.difference(this._controller.getVisibleNodes(), this._controller.unselectable())[0]);
        } else {
            this._collapse();
        }

    };

    /**
     * @method ComboboxInputView#_collapse
     * @private
     */
    cvp._collapse = function () {
        //Super cal
        View.prototype.hidePopup.call(this);

        this._controller.filter(null);

        var active = this._controller.active();

        //Restore active if there was one
        if (!_.isNull(active)) {
            this._findMenuItem(active).addClass('active');
        }
    };

    /**
     * @method ComboboxInputView#filtered
     */
    cvp.filtered = function () {
        this._createDropdown();
    };

    /**
     * @method ComboboxInputView#_handleEnter
     * @private
     */
    cvp._handleEnter = function () {
        this.clearFilter();
        if (this._$el.hasClass('open') && this._$menu.find('.active').length > 0) {
            if (!this._typing || !this._options.multiSelect) {
                //Something in the menu was selected - select that
                View.prototype._handleEnter.call(this);
            } else {
                this._controller.selected(_.union(this._controller.selected(), [this._getId(this._$menu.find('.combo-item.active'))]));
                this._collapse();
            }
        } else {
            return this._controller.selected([this._controller.getNodeByDisplayValue(escapingUtils.escape(this._$valueEl[0].value, this._options))]);
        }

        this._typing = false;

    };

    /**
     * @method ComboboxInputView#clearFilter
     */
    cvp.clearFilter = function () {
        this._controller.filter(null);
    };

    return ComboBoxInputView;
});

define('oui/widgets/comboBox/comboBoxView',[
    'oui/widgets/comboBox/comboBoxButtonView',
    'oui/widgets/comboBox/comboBoxInputView'
], function (ComboBoxButtonView, ComboBoxInputView) {
    'use strict';

    /*
     * Convenience function to automatically pick the proper view type based on the initial options
     */
    return function (model, controller, element, options, settings) {
        var CorrectConstructor = options.keyboardInput ? ComboBoxInputView : ComboBoxButtonView;
        return new CorrectConstructor(model, controller, element, options, settings);
    };
});


define('template-38', ['underscore'], function (_) { return function(data) {
var __t, __p = '', __e = _.escape, __j = Array.prototype.join;
function print() { __p += __j.call(arguments, '') }

    var classes = data.classes;    if (data.attrs.hasChildren) {        classes.push('combo-item-header');    };
__p += '<a href="#" class="' +
((__t = ( classes.join(' ') )) == null ? '' : __t) +
'" title="' +
__e( data.title ) +
'" ' +
((__t = ( data.speakAs ? 'aria-label="' + data.speakAs + '";' : '' )) == null ? '' : __t) +
'> ' +
((__t = ( data.valueMarkup )) == null ? '' : __t) +
'</a>';
return __p
};});

define('oui/widgets/comboBox/comboBoxRenderer',[
    'underscore',
    'oui/base/renderer',
    'template-38'
], function (_, Renderer, nodeItemTemplate) {
    'use strict';

    /**
     * @typedef ComboBoxRenderer#RenderData
     *
     * @property {Object} attrs
     * @property {Boolean} attrs.unselectable
     * @property {Boolean} attrs.hasChildren
     * @property {Number} attrs.level
     * @property {Boolean} attrs.active
     *
     * @property {Object} data
     */

    /**
     * @class ComboBoxRenderer
     * @extends Renderer
     */
    function ComboBoxRenderer() {
        Renderer.apply(this, arguments);
    }

    ComboBoxRenderer.prototype = Object.create(Renderer.prototype);

    ComboBoxRenderer.prototype = _.extend(ComboBoxRenderer.prototype, {
        constructor: ComboBoxRenderer,
        /**
         * @method ComboBoxRenderer#asText
         *
         * @param {ComboBoxRenderer#RenderData} node
         *
         * @returns {String}
         */
        asText: function (node) {
            return node.data.name || node.data.display;
        },
        /**
         * @method ComboBoxRenderer#getMarkup
         *
         * @param {ComboBoxRenderer#RenderData} node
         * @param {ComboBox#Config} options
         *
         * @returns {String}
         */
        getMarkup: function (node, options) {
            var safeEl = document.createElement('span');

            safeEl.innerHTML = options.viewFns.display(node.data);

            safeEl.normalize();

            return safeEl.innerHTML;
        },
        /**
         * @method ComboBoxRenderer#getTitle
         *
         * @param {ComboBoxRenderer#RenderData} node
         * @param {ComboBox#Config} options
         *
         * @returns {String}
         */
        getTitle: function (node, options) {
            var safeEl = document.createElement('span');

            if (options.viewFns.text) {
                return options.viewFns.text(node.data);
            }

            safeEl.innerHTML = options.viewFns.display(node.data);

            safeEl.normalize();

            return safeEl.textContent.trim();
        },
        /**
         * @method ComboBoxRenderer#getSpokenText
         *
         * @param {ComboBoxRenderer#RenderData} node
         * @param {ComboBox#Config} options
         *
         * @returns {String?}
         */
        getSpokenText: function (node, options) {

            if (options.viewFns.speak) {
                return options.viewFns.speak(node.data);
            }

            return null;
        },
        /**
         * @method ComboBoxRenderer#getClasses
         *
         * @param {ComboBoxRenderer#RenderData} node
         * @param {ComboBox#Config} options
         *
         * @returns {Array<String>}
         */
        getClasses: function (node, options) {
            var classes = [];

            if (options.viewFns.classes) {
                classes = options.viewFns.classes(node.data);
            }

            if (node.attrs.hasChildren) {
                classes.push('header');
            }

            return classes;
        },
        /**
         * @method ComboBoxRenderer#render
         *
         * @param {ComboBoxRenderer#RenderData} node
         * @param {ComboBox#Config} options
         *
         * @returns {String}
         */
        render: function (node, options) {
            var data = {
                title: this.getTitle(node, options),
                speakAs: this.getSpokenText(node, options),
                classes: this.getClasses(node, options),
                valueMarkup: this.getMarkup(node, options),
                attrs: node.attrs,
                locale: options.locale
            };

            return nodeItemTemplate(data);
        }
    });

    return ComboBoxRenderer;
});


define('template-39', ['underscore'], function (_) { return function(data) {
var __t, __p = '', __e = _.escape;
__p += '<span title="' +
__e( data.title ) +
'"> ' +
((__t = ( data.display )) == null ? '' : __t) +
'</span>';
return __p
};});

define('oui/widgets/comboBox/comboBoxSelectedRenderer',[
    'underscore',
    'oui/base/renderer',
    'template-39'
], function (_, Renderer, selectedTemplate) {
    'use strict';

    /**
     * @class ComboBoxSelectedRenderer
     * @extends Renderer
     */
    function ComboBoxSelectedRenderer() {
        Renderer.apply(this, arguments);
    }

    ComboBoxSelectedRenderer.prototype = Object.create(Renderer.prototype);

    ComboBoxSelectedRenderer.prototype = _.extend(ComboBoxSelectedRenderer.prototype, {
        /**
         * @method ComboBoxSelectedRenderer#getDisplay
         *
         * @param {Array<ComboBoxRenderer#RenderData>} nodes
         * @param {TreePanel#Config} options
         *
         * @returns {String}
         */
        getDisplay: function (nodes, options) {
            var span = document.createElement('span');

            if (options.viewFns.displayAll) {
                return options.viewFns.displayAll(_.map(nodes, function (node) {
                    return node.data;
                }));
            }

            return _.map(nodes, function (node) {
                var intermValue = options.viewFns.display(node.data);
                span.innerHTML = intermValue;

                if (options.viewFns.text && !options.forceDisplay) {
                    intermValue = options.viewFns.text(node.data);
                }

                return intermValue;
            }).join(options.delimiter);

        },
        /**
         * @method ComboBoxSelectedRenderer#getTitle
         *
         * @param {Array<ComboBoxRenderer#RenderData>} nodes
         * @param {TreePanel#Config} options
         *
         * @returns {String}
         */
        getTitle: function (nodes, options) {
            var span = document.createElement('span');

            return _.map(nodes, function (node) {
                var intermValue;

                if (options.viewFns.speak) {
                    intermValue = options.viewFns.speak(node.data);
                }

                if (_.isEmpty(intermValue) && options.viewFns.text) {
                    intermValue = options.viewFns.text(node.data);
                }

                if (_.isEmpty(intermValue)) {
                    intermValue = options.viewFns.display(node.data);
                    span.innerHTML = intermValue;
                    intermValue = span.textContent.trim();
                }
                return intermValue;
            }).join(options.delimiter);
        },
        /**
         * @method ComboBoxSelectedRenderer#render
         *
         * @param {Array<ComboBoxRenderer#RenderData>} nodes
         * @param {TreePanel#Config} options
         *
         * @returns {String}
         */
        render: function (nodes, options) {

            return selectedTemplate({
                title: this.getTitle(nodes, options),
                display: this.getDisplay(nodes, options)
            });
        }
    });

    return ComboBoxSelectedRenderer;
});

define('oui/widgets/comboBox',['require','underscore','oui/base/input','oui/widgets/comboBox/comboBoxModel','oui/widgets/comboBox/comboBoxController','oui/widgets/comboBox/comboBoxView','oui/widgets/comboBox/comboBoxRenderer','oui/widgets/comboBox/comboBoxSelectedRenderer','logger'],function (require) {
      'use strict';

      var _ = require('underscore');
      var Input = require('oui/base/input');
      var ComboBoxModel = require('oui/widgets/comboBox/comboBoxModel');
      var ComboBoxController = require('oui/widgets/comboBox/comboBoxController');
      var ComboBoxView = require('oui/widgets/comboBox/comboBoxView');
      var ComboBoxRenderer = require('oui/widgets/comboBox/comboBoxRenderer');
      var ComboBoxSelectedRenderer = require('oui/widgets/comboBox/comboBoxSelectedRenderer');
      var logger = require('logger');

      function getValue(data) { //jscs:ignore jsDoc
          return _.get(data, 'value') || _.get(data, 'id', '');
      }

      /**
       * @callback ComboBox#display
       * @param {Object} data The nodes data to be used to construct a string
       * @returns {String|HTML} The string representation of the node.
       */

      /**
       * @callback ComboBox#text
       * @param {Object} data The nodes data to be used to construct a string
       * @returns {String} The string representation of the node.
       */

      /**
       * @callback ComboBox#speak
       * @param {Object} data The nodes data to be used to construct a string
       * @returns {String} The string representation of the node to speak.
       */

      /**
       * @callback ComboBox#value
       * @param {Object} data The nodes data to be used to construct a string
       * @returns {String} The unique identifier to use for the node.
       */

      /**
       * @typedef {Object} ComboBox#Config
       *
       * @property {Locale} [locale={}] The specific strings that are used by the widget and will use the defaults set from
       * the {@link module:oui.lang|lang}.
       * @property {Boolean} [ignoreInvalid=true] Allow combobox to ignore selections that result in invalid entries
       * @property {Boolean} [required=false] Whether or not the combobox is required. Much like the native select element
       * an element with a value of '' will result in a selection that will trigger an error.
       * @property {Boolean|Object} [autoEscape=true] Whether the combobox should escape/unescape the values from the input
       * element. When the property is an Object it should have two functions escape and unescape that take in the string
       * and provide the escaped or unescaped version of the string. When the property is a Boolean it will just used
       * underscores escape/unescape function.
       * @property {Boolean} [openOnFocus=false] Whether or not the combobox should open the list of values when focused.
       * @property {String|jQuery} [appendTo='body'] The jQuery object or selector to append the dropdown menu to.
       * @property {String} [titleText=''] The hover text display when mousing over the combobox.
       * @property {Boolean} [keyboardInput=true] Whether this combobox will allow the user to enter values with the
       * keyboard instead of just picking from a list. Setting to `true` will allow the user to type/auto-complete their
       * entries; setting to `false` means the user will need to search through a list in order to find the proper value.
       * @property {Boolean} [allowCustomValues=false] Whether the combo box allows custom values. This property is only
       * relevant if the keyboardInput property is set to `true`. Setting to `true` will allow a user to provide their own
       * value to the dropdown instead of just selecting from the provided options. **Currently Unused** **Note**: When
       * `false`, if the user attempts to provide a custom value, the value will revert to the old value when the focus is
       * lost.
       * @property {Boolean} [fluid=false] Whether the combo box takes up all available space inside the parent element.
       * @property {Boolean} [allowEmpty=false] Whether or not the combo box can be constructed with an empty value. If
       * false, the combo box will use the first selectable value in the values as the currently selected value. If true,
       * the combo box will allow the combobox to start as empty but once the user makes a selection they may not be able
       * to return to an empty state as there is no option that represents the empty state by default.
       * @property {Boolean} [copyClasses=false] Whether the combobox should take the classes from the original element and
       * place them on the combobox (.orcl-combobox)
       * @property {String|Number} [menuSize='grow-to-text'] The width of the menu displayed by the combobox.
       * @property {Boolean} [forceDisplay=false] Whether or not to force display function as the rendered value in the
       * combobox. Precedence defaults to 'display', 'text', 'speak' (if given)
       * @property {Array<String>} [menuClasses=[]] The classes to apply to the menu for additional styling.
       * @property {Boolean} [autoSort=false] Whether or not to sort the values each time they are updated.
       * @property {Object} viewFns The functions used during rendering of the component.
       * @property {ComboBox#display} viewFns.display
       * @property {ComboBox#text} [viewFns.text]
       * @property {ComboBox#speak} [viewFns.speak]
       * @property {Object} storeFns The functions used during manipulation of data in the model.
       * @property {ComboBox#value} [storeFns.value]
       * @property {Function} [storeFns.sortValue] The value of the node to use for sorting purposes. Only used when the
       * {@link ComboBox#Config.autoSort} is true.
       * @property {Boolean} [multiSelect=false] Whether or not the combo box supports multi-selecting.
       * @property {String} [delimiter=', '] The delimiter to use for multi selection rendering behavior.
       */

      var __DEFAULTS__ = {
          showErrors: 'blur',
          ignoreInvalid: true,
          cloneOnAdd: true,
          filterShowHierarchy: true,
          required: false,
          autoEscape: true,
          openOnFocus: false,
          appendTo: 'body',
          titleText: '',
          keyboardInput: true,
          allowCustomValues: false,
          fluid: false,
          shrinkToFit: false,
          allowEmpty: false,
          copyClasses: false,
          menuSize: 'grow-to-text',
          forceDisplay: false,
          selectAll: false,
          action: 'default',
          menuClasses: [],
          autoSort: false,
          renderer: new ComboBoxSelectedRenderer(),
          itemRenderer: new ComboBoxRenderer(),
          idPrefix: 'oui-',
          groupBy: [
              {
                  field: 'parentId',
                  asTree: true
              }
          ],
          recordKeys: {
              id: 'value',
              children: 'children',
              parentId: 'parentId'
          },
          viewFns: {
              display: function (data) { //jscs:ignore jsDoc
                  return (_.get(data, 'name', _.get(data, 'display', ''))).toString();
              }
          },
          storeFns: {
              sortValue: getValue,
              value: getValue
          },
          multiSelect: false,
          delimiter: ', '
      };

      /**
       * The ComboBox widget allows the user to pick from a pre-determined list of options, and optionally provide
       * their own values.
       *
       * See {@link ComboBox#Config} for more information regarding options.
       *
       * See {@link http://www.w3.org/TR/wai-aria-practices/#combobox|ARIA Best Practices}
       * for information regarding keyboard shortcuts and general ARIA attributes and states.
       *
       *  **Keyboard Shortcuts**
       *
       *  * <kbd>Escape</kbd> - Collapses the menu and rejects any changes made while it was open
       *  * <kbd>Enter</kbd> - If it is currently collapsed, expands the menu otherwise selects the value and closes the menu
       *  * <kbd>Space</kbd> - If it is currently collapsed, expands the menu otherwise selects the value and closes the menu
       *  * <kbd>Shift+Up</kbd> - Extends selection on multiselect comboboxes up by 1 (uses the original item that shift was initiated on as the axis)
       *  * <kbd>Shift+Down</kbd> - Extends selection on multiselect comboboxes down by 1 (uses the original item that shift was initiated on as the axis)
       *  * <kbd>Control+Space</kbd> - Toggles selection of the active node if it is a multiSelect combobox
       *  * <kbd>Up</kbd> - If it is currently collapsed, expands te menu otherwise the node above the active node is marked as active
       *  * <kbd>Down</kbd> - If it is currently collapsed, expands the menu otherwise the node below the active node is marked as active
       *  * <kbd>Alt+Up</kbd> - Collapses the menu
       *  * <kbd>Alt+Down</kbd> - Opens the menu
       *  * (Alphanumeric) - Finds a menu item that begins with the letter pressed.
       *
       *
       *  ##### **Relevant Events**
       *
       *  * {@link ComboBox#values-change|values-change}
       *  * {@link ComboBox#active-change|active-change}
       *  * {@link ComboBox#selection-change|selection-change}
       *  * {@link ComboBox#invalid-selection|invalid-selection}
       *  * {@link ComboBox#required-error|required-error}
       *  * {@link ComboBox#valueSelected|valueSelected}
       *
       *  @borrows ComboBoxModel#event:values-change as ComboBox#values-change
       *  @borrows ComboBoxModel#event:active-change as ComboBox#active-change
       *  @borrows ComboBoxModel#event:selection-change as ComboBox#selection-change
       *  @borrows ComboBoxModel#event:invalid-selection as ComboBox#invalid-selection
       *  @borrows ComboBoxModel#event:required-error as ComboBox#required-error
       *  @borrows ComboBoxModel#event:valueSelected as ComboBox#valueSelected
       *
       * @example
       * <caption>Creating a Combo Box (accepts all defaults, with no values)</caption>
       *
       * var combo = new oui.ComboBox();
       * $container.append(combo.element());
       * combo.init();
       *
       * @example
       * <caption>Creating a Combo Box (from `select` element)</caption>
       *
       * var combo = new oui.ComboBox($select);
       * combo.init();
       *
       * @example
       * <caption>Creating a Combo Box (with values)</caption>
       *
       * var combo = new oui.ComboBox([
       *     { name : 'Subaru', value : 'subaru' },
       *     { name : 'Ford', value : 'ford' },
       *     { name : 'Toyota', value : 'toyota' },
       *     { name : 'Honda', value : 'honda' },
       *     { name : 'Jeep', value : 'jeep' },
       * ]);
       * $container.append(combo.element());
       * combo.init();
       *
       * @class ComboBox
       * @since 1.3.0
       * @extends Input
       *
       * @param {ComboBox#Config} [options] Configurable options for the ComboBox widget
       * @param {jQuery} [element] The element to wrap into a ComboBox widget can contain either a select element or
       *                  an input element..
       * @param {Array<Object>} [initData] The values to populate the combobox with. The objects by default should be
       *                  in the form of { name : '', value : '' }. Of course, the display function can be updated
       *                  if the data is not keyed with name and value.
       *
       * @tutorial using-combobox
       */
      function ComboBox(options, element, initData) {
          var model;

          Input.call(this, options, element, initData);

          if (!_.isUndefined(this._options.values)) {
              logger.warn('options.values is deprecated, please move it to the initData parameter and the selected item to a method call');
              delete this._options.values;
          }

          this._options.type = this._options.keyboardInput ? 'combo-input' : 'combo-button';

          model = new ComboBoxModel(this._initData, this._options);
          this._controller = new ComboBoxController(model, this._options);
          this._view = new ComboBoxView(model, this._controller, this._element, this._options, this._settings);

          Object.defineProperty(this, '_data', {
              get: this.selected,
              set: this.selected
          });

          this._setupEvents(model);
      }

      ComboBox.prototype = Object.create(Input.prototype);
      var cbp = ComboBox.prototype;

      cbp.constructor = ComboBox;

      /**
       * Gets the widget's name.
       *
       * @method ComboBox#name
       * @returns {String} 'combobox'
       */
      cbp.name = function () {
          return 'combobox';
      };

      /**
       * @method ComboBox#_translate
       * @private
       *
       * @param {String} name The event to translate
       * @param {...*} args The arguments to translate
       * @returns {arguments|Array} The translated data
       */
      cbp._translate = function (name) {
          var that = this;

          if (name === 'selection-change' ||
              name === 'change' ||
            name === 'valueSelected') {
              var selected = arguments[1];

              if (selected.length === 1) {
                  return [name, this._controller.node(selected[0])];
              } else {
                  return [name, _.map(selected, function (node) {
                      return that._controller.node(node);
                  })];
              }
          }
          return arguments;
      };

      /**
       * Gets or sets the currently selected value from the ComboBox.
       *
       * When multiSelect is true this method accepts an Array of Strings or Objects otherwise it only accepts a single
       * item as the new value.
       *
       * When passing in a String it should be its display value. When passing in an Object the data should be a unique
       * identifier of the node that is selected.
       *
       * (A synonym for data)
       *
       * For the silent version of this method see {@link ComboBox#value}
       *
       * @fires ComboBoxModel#invalid-selection
       * @fires ComboBoxModel#required-error
       * @fires ComboBoxModel#selection-change
       * @fires ComboBoxModel#valueSelected
       *
       * @method ComboBox#selected
       * @param {Object|String|Array} [newSelected] The new value(s) to set in the ComboBox. Array may only be used when
       * `multiSelect` is true.
       * @returns {Object|Array<Object>}
       */
      cbp.selected = function (newSelected) {
          if (!arguments.length) {
              var that = this;
              var selected = this._controller.selected();

              if (!this._options.multiSelect && selected.length === 1) {
                  return that._controller.node(selected[0], { strip: true });
              }

              var select = _.map(selected, function (node) {
                  return that._controller.node(node, { strip: true });
              });

              if (select.length > 0) {
                  return select;
              } else {
                  return null;
              }
          }

          if (!_.isArray(newSelected)) {
              newSelected = [newSelected];
          }

          newSelected = _.compact(newSelected);

          newSelected = _.map(newSelected, function (node) {
              return _.get(this._controller.node(node), this._options.recordKeys.id);
          }.bind(this));

          this._controller.selected(newSelected);
      };

      /**
       * Gets or sets the currently selected value from the ComboBox.
       *
       * See {@link ComboBox#selected} for more details about this method.
       *
       * **Differences** from {@link ComboBox#selected|selected}/{@link ComboBox#data|data}:
       *
       *  * does not trigger event ({@link ComboBox#selection-change|selection-change})
       *
       * @method ComboBox#value
       * @param {Object|String|Array} [newSelected] The new value to set in the ComboBox
       * @returns {Object|Array<Object>}
       */
      cbp.value = function (newSelected) {
          if (!arguments.length) {
              return this.selected();
          }

          if (!_.isArray(newSelected)) {
              newSelected = [newSelected];
          }

          newSelected = _.compact(newSelected);

          newSelected = _.map(newSelected, function (node) {
              return _.get(this._controller.node(node), this._options.recordKeys.id);
          }.bind(this));

          this._controller.selected(newSelected, {
              silent: true
          });
      };

      /**
       * Gets or sets the accepted values for the ComboBox.
       *
       * @fires ComboBoxModel#values-change
       *
       * @method ComboBox#values
       * @param {Array<Object>} [newValues] The new values list for the ComboBox
       * @returns {Array<Object>} The data used for the combobox nodes
       */
      cbp.values = function (newValues) {
          if (!arguments.length) {
              return _.map(this._controller.getVisibleNodes(), function (id) {
                  return this._controller.node(id, { strip: true });
              }.bind(this));
          }
          this._controller.values(newValues);
      };

      /**
       * @method ComboBox#defaults
       * @returns {Object}
       */
      cbp.defaults = function () {
          return __DEFAULTS__;
      };

      /**
       * @method ComboBox#isValid
       *
       * @returns {Boolean}
       */
      cbp.isValid = function () {
          return this._controller.isValid();
      };

      return ComboBox;

  }
);


define('template-40', ['underscore'], function (_) { return function(data) {
var __t, __p = '';
__p += '<li role="group" title="' +
((__t = ( data.title )) == null ? '' : __t) +
'" class="typeahead-item-header"> <a href="#">' +
((__t = ( data.text )) == null ? '' : __t) +
'</a> <ul class="children"> ' +
((__t = ( data.children )) == null ? '' : __t) +
' </ul></li>';
return __p
};});


define('template-41', ['underscore'], function (_) { return function(data) {
var __t, __p = '';
__p += '<button class="btn btn-picker" tabindex="-1" title="' +
((__t = ( data.locale.search )) == null ? '' : __t) +
'" aria-label="' +
((__t = ( data.locale.search )) == null ? '' : __t) +
'"> <i class="pgbu-icon-small pgbu-icon-magnifying-glass"></i></button><a title="' +
((__t = ( data.locale.clear )) == null ? '' : __t) +
'" class="search-clear-icon" href="#"> <i class="pgbu-icon pgbu-icon-delete"></i></a>';
return __p
};});


define('template-42', ['underscore'], function (_) { return function(data) {
var __t, __p = '';
__p += '<span id="' +
((__t = ( data.id )) == null ? '' : __t) +
'" class="visually-hidden filter-status" role="status" aria-live="polite" aria-relevant="all" aria-atomic="true"></span>';
return __p
};});

define('oui/widgets/typeAhead/typeAheadNonTokenView',['require','underscore','jquery','oui/base/view','oui/utils/stringUtils','oui/utils/keyCode','oui/mixins/popup','oui/widgets/comboBox','oui/utils/dom','oui/utils/inputUtils','oui/utils/scrollingUtils','oui/utils/escapingUtils','template-23','template-25','template-26','template-40','template-41','template-42','oui/widgets/tree/spinner.template'],function (require) { //jshint ignore:line
      'use strict';

      var _ = require('underscore');
      var $ = require('jquery');
      var View = require('oui/base/view');
      var stringUtils = require('oui/utils/stringUtils');
      var keyCode = require('oui/utils/keyCode');
      var popupMixin = require('oui/mixins/popup');
      var ComboBox = require('oui/widgets/comboBox');
      var dom = require('oui/utils/dom');
      var inputUtils = require('oui/utils/inputUtils');
      var scrollingUtils = require('oui/utils/scrollingUtils');
      var escapingUtils = require('oui/utils/escapingUtils');
      var typeAheadTemplate = require('template-23');
      var menuTemplate = require('template-25');
      var nodeTemplate = require('template-26');
      var groupTemplate = require('template-40');
      var searchIcons = require('template-41');
      var filterStatusTemplate = require('template-42');
      var loadingTemplate = require('oui/widgets/tree/spinner.template');
      var tvp;

      /**
       * Fired when the combobox representing the scope of the filter has changed.
       *
       * @event TypeAheadNonTokenView#scope-change
       *
       * @param {String} query
       * @param {Object} scope
       */

      var SEARCH_CLASSES = [
          'search',
          'search-collapsible',
          'scoped-search'
      ];

      var PREVENT_DEFAULT_KEYS = [
          keyCode.DOWN,
          keyCode.UP,
          keyCode.TAB,
          keyCode.SHIFT,
          keyCode.ENTER,
          keyCode.F2,
          keyCode.F10
      ];

      /**
       * @method TypeAheadNonTokenView~getId
       *
       * @param {HTMLElement|jQuery} menuItem
       *
       * @returns {String}
       */
      function getId(menuItem) {
          return $(menuItem).attr('data-id');
      }

      /**
       * The type ahead view is responsible for displaying the suggestions when they loaded.
       *
       * @class TypeAheadNonTokenView
       * @private
       * @since 2.1.0
       * @extends View
       * @mixes popup
       *
       * @param {TypeAheadModel} model The typeahead model that is responsible for holding all the past
       *  search results
       * @param {TypeAheadController} controller The type ahead controller that will modify the model.
       * @param {Object} element jQuery element for the view
       * @param {TypeAhead#Config} viewOpts The options for the view.
       */
      function TypeAheadView(model, controller, element, viewOpts, settings) {
          var that = this;

          this._scope = null;

          View.call(this, viewOpts, element, controller, settings);

          /**
           * The menu object for this typeahead
           *
           * A JQuery object.
           *
           * @private
           * @type {Object}
           * @memberof! TypeAheadNonTokenView.prototype
           */
          this._$menu = null;
          this._$input = this._$el.find('input');
          this._selectedMap = {};

          this._inputEl = this._$input[0];

          model.on({
              'view.loaded': this.loaded.bind(this),
              'view.active': this.active.bind(this),
              'view.loading': this.loading.bind(this),
              'view.select': this.select.bind(this),
              'view.selection-change': this.select.bind(this),
              'view.invalidate': this.clear.bind(this),
              'view.clear': this.clear.bind(this),
              'view.search': this.update.bind(this),
              'view.required-error': this.error.bind(this)
          });

          if (this._options.style === 'search') {
              this._$el.append(filterStatusTemplate({
                  id: this.id() + '-filter-status'
              }));

          }

          if (this._scope) {
              this._scope.on('selection-change', function (scope) {
                  this._$el.find('.search-clear-icon').css({
                      right: that._scope.element().width() + 28
                  });
                  this.trigger('scope-change', this._inputEl.value || '', scope);
              }.bind(this));
          }

          var debounced = _.debounce(controller.search, this._options.delay);

          this._$menu = this._$popup = $(menuTemplate({
              scroll: this._options.scroll,

              //TODO: find a way to get 26 out of here (It is the height of an <li> in our dropdown
              height: (that._options.maxSuggestions * 26) + 'px'
          }));
          this._$menu.toggleClass('navigation', this._options.style === 'navigation');

          //Once we have the markup we need to put on the events - each time
          this._$menu.on('click', 'li', function (event) {
              event.preventDefault();

              var id = getId($(this));

              if (id != null) {
                  that._$input.focus();
                  if (!that._options.multiSelect) {
                      that._controller.selected([id]);
                  } else {
                      that._controller.selected(that._controller.selected().concat(id));
                  }
              }
          }).on('mousedown', function (event) {
              event.preventDefault();
          }).on('shown.popup', function () {
              that._$input.attr('aria-expanded', true);
          }).on('hidden.popup', function () {
              that._$input.removeAttr('aria-activedescendant').attr('aria-expanded', false);
          });

          this._$menu.on('mousedown', function () {
              that.mousedown = true;
              that._$menu.one('mouseup', function () {
                  that.mousedown = false;
              });
          });

          this.element().on('oui-accept oui-picker-accept', function (event, widget, name, selected) {
              if (widget.name() === 'picker' || widget.name() === 'double-picker') {
                  that._controller.selected(selected, {
                      silent: true
                  });
              }
          });

          this._$el.on('mousedown', '.btn-picker', function (event) {
              event.preventDefault();
              that._$input.focus();
          });

          this._$el.find('.search-clear-icon').on('click', function (event) {
              event.preventDefault();
              that._inputEl.value = '';
              that._controller.search(that._inputEl.value, null, that.getScope());
              that._$input.focus();
          }).on('keydown', function (event) {
              if (event.keyCode === keyCode.ENTER) {
                  event.preventDefault();
                  that._inputEl.value = '';
                  that._controller.search(that._inputEl.value, null, that.getScope());
                  that._$input.focus();
              }
          });

          var oldValue = this._inputEl.value;
          this._$input
        .on('keyup', function (event) {
            if (_.contains(PREVENT_DEFAULT_KEYS, event.keyCode)) {
                return;
            }

            switch (event.keyCode) {
                case keyCode.ESC:
                    that.hidePopup();

                    if (that._options.escToClear && !that._$el.hasClass('empty')) {
                        event.stopPropagation();
                        that._$input.val('');
                        that._controller.search(that._$input.val(), null, that.getScope());
                        that._$input.focus();
                    }

                    break;
                case keyCode.BACKSPACE:
                /* falls through */
                default:
                    if (this.selectionStart === this.selectionEnd) {
                        debounced.call(controller, escapingUtils.escape(this.value, that._options), this.selectionStart, that.getScope());
                    }
            }
        })
        .on('blur', function (event) {
            event.preventDefault();
            if (!that.mousedown) {
                if (that.isVisible()) {
                    that.hidePopup();
                }
                if (that._options.style !== 'search') {
                    that.updateValue();
                }

                if (that._inputEl.value === '' && that._controller.selected().length === 0 && that._options.required) {
                    that._controller.selected([]);
                }

                that.change();
            } else {
                that._inputEl.focus();
                that.mousedown = false;
            }
        })
        .on('input', function (event) {
            if (oldValue !== this.value) {
                if (this.selectionStart === this.selectionEnd) {
                    debounced.call(controller, escapingUtils.escape(this.value, that._options), this.selectionStart, that.getScope());
                }

                that.changeHandler(event);
                oldValue = this.value;
            }
        })
        .on('change', this.changeHandler.bind(this))
        .on('change', function () {
            that._triggerChange = true;
            that._$el.toggleClass('empty', that._inputEl.value === '');
        })
        .on('keypress', function (event) {
            var character = String.fromCharCode(event.which);
            var value = inputUtils.insertCharacter(event.target, event.which, event.shiftKey);

            if (!event.shiftKey) {
                character = character.toLowerCase();
            }

            if (event.keyCode !== keyCode.ENTER &&
                event.keyCode !== keyCode.TAB &&
                event.keyCode !== keyCode.ESC &&
                !(event.which == null || event.which === 0)) {
                if (this.selectionStart !== this.selectionEnd) {
                    that.handleBackspace(true);
                }

                value = escapingUtils.escape(value, that._options);

                //TODO: look into dealing with multi-character delimiters
                if (character !== that._options.delimiter.trim()) {
                    var query = stringUtils.getQuery(value, this.selectionStart,
                      that._options.multiSelect, that._options.delimiter.trim());

                    var node = that._controller.node(query);

                    //User is updating the currently selected query node
                    //remove it to allow for extensions like George to George Michael
                    if (node != null) {
                        controller.remove(node);
                    }
                }

                debounced.call(controller, value, this.selectionStart, that.getScope());
            }
        })
        .on('keydown', function (event) {
            if (that._options.search) {
                switch (event.which) {
                    case keyCode.DOWN:
                        event.preventDefault();
                        that._controller.moveActive(1);
                        break;
                    case keyCode.UP:
                        event.preventDefault();
                        that._controller.moveActive(-1);
                        break;
                    case keyCode.ENTER:
                        that._triggerChange = true;
                        if (!_.isNull(that._$menu) && that.isVisible() && that._$menu.find('.active').length !== 0) {
                            event.stopPropagation();
                            event.preventDefault();
                            controller.selected(controller.selected().concat(that._$menu.find('.active').attr('data-id')));

                        } else if (!that.isVisible()) {
                            that.change();
                        }

                        break;
                    case keyCode.BACKSPACE:
                    case keyCode.DELETE:

                        if (that.handleBackspace()) {
                            event.preventDefault();
                        }

                        break;
                }
            }
        });

          if (that._options.collapsible) {
              var collapseTimer;
              that._$el.find('.btn-picker').attr('tabindex', '0');
              that._$el.attr('tabindex', '0').on('focusin', function (e) {
                  if (collapseTimer) {
                      window.clearTimeout(collapseTimer);
                      collapseTimer = null;
                  }
                  var $this = $(this).removeClass('collapsed');
                  var $target = $(e.target);
                  $this.find('.btn-picker').attr('tabindex', '-1');
                  if ($target.is(that._$el) || $target.hasClass('btn-picker')) {
                      $this.find('input').focus();
                  }
                  $this.removeAttr('tabindex');
              });
              that._$el.find('.search-query, .search-clear-icon, .orcl-combo-box button').on('blur', function () {
                  collapseTimer = window.setTimeout(function () {
                      collapseTimer = null;
                      if (!(document.activeElement === that._$el.find('.search-query')[0] ||
                            document.activeElement === that._$el.find('.search-clear-icon')[0] ||
                            document.activeElement === that._$el.find('.orcl-combo-box button')[0])) {
                          that._$el
                            .addClass('collapsed')
                            .attr('tabindex', '0')
                            .find('.btn-picker').attr('tabindex', '0');
                      }
                  }, 0);
              });
          }
      }

      TypeAheadView.prototype = Object.create(View.prototype);
      tvp = TypeAheadView.prototype;

      tvp.constructor = TypeAheadView;

      _.extend(tvp, popupMixin);

      /**
       * @method TypeAheadNonTokenView#destroy
       */
      tvp.destroy = function () {
          this._$input.off();
          this._$el.off();
          this._$el.remove();

          if (this._$menu != null) {
              this._$menu.remove();
          }
      };

      /**
       * @method TypeAheadNonTokenView#changeHandler
       *
       * @param {jQuery.Event} event
       */
      tvp.changeHandler = function (event) {
          if (document.activeElement === event.target) {
              this._$el.toggleClass('empty', this._inputEl.value === '');

              if (this._scope != null) {
                  this._$el.find('.search-clear-icon').css({
                      right: this._scope.element().width() + 28
                  });
              }
          }
      };

      /**
       * @method TypeAheadNonTokenView#getScope
       *
       * @returns {String}
       */
      tvp.getScope = function () {
          if (this._scope) {
              return this._scope.value();
          }
      };

      /**
       * @method TypeAheadNonTokenView#handleBackspace
       *
       * @param {Boolean} force
       *
       * @returns {Boolean}
       */
      tvp.handleBackspace = function (force) {
          var selection = {
              start: this._inputEl.selectionStart,
              end: this._inputEl.selectionEnd
          };
          var value = this._getUnescapedValue(this._inputEl.value);

          var queryBounds = this._controller.getQueryBounds(selection, value);

          var selected = this._controller.selected();
          if (selected.length > 0) {

              //Move the selection ends to these indices
              this._inputEl.setSelectionRange(queryBounds.start, queryBounds.end);

              if (force || _.isEqual(selection, queryBounds)) {
                  var display = value.substring(queryBounds.start, queryBounds.end);

                  var filtered = _.filter(selected, function (item) {
                      return (item.displayValue !== display && display !== this._options.viewFns.display(item.data)) ||
                          this._options.viewFns.matches(item.data, display);
                  }.bind(this));

                  this._controller.selected(filtered);
              }

              return true;
          }

          return false;
      };

      /**
       * @method TypeAheadNonTokenView#buildElement
       *
       * @param {TypeAhead#Config} options
       * @param {jQuery} $el
       *
       * @returns {jQuery}
       */
      tvp.buildElement = function (options, $el) {
          if (_.isNull($el)) {
              $el = $(typeAheadTemplate({
                  locale: options.locale
              }));
          } else if ($el.jquery && $el.is('input')) {
              $el[0].type = 'text';
              $el[0].placeholder = options.locale.placeholder || '';
          }

          if ($el.parent().hasClass('orcl-widget')) {
              $el.parent().addClass('orcl-type-ahead');
          } else {
              $el.wrap('<div class="orcl-type-ahead orcl-widget"></div>');
          }

          $el = $el.parent();

          $el.find('input')
            .attr('required', options.required)
            .removeAttr('readonly');

          if (_.contains(SEARCH_CLASSES, options.style)) {
              $el.addClass('search');
              $el.toggleClass('scoped-search', options.scopes != null && options.scopes.length > 0);
              $el.find('input').addClass('search-query');
              $el.toggleClass('collapsible collapsed', options.collapsible);
              $el.toggleClass('empty', $el.find('input')[0].value === '');

              $el.on('mousedown', '.search-icon', function (event) {
                  event.preventDefault();
                  $el.find('input').focus();
              });

              if (options.scopes && options.scopes.length > 0) {
                  this._scope = new ComboBox(_.defaults(options.combobox, { appendTo: options.appendTo }));
                  this._scope.values(options.scopes);
                  $el.append(this._scope.element());
                  this._scope.init();
              }
              $el.append(searchIcons({
                  locale: options.locale
              }));
          } else if (options.style === 'navigation') {
              $el.addClass('navigation');
              $('body').append($el);
              $(document).on('keydown.navigation', function (event) {
                    if ((event.ctrlKey || event.metaKey) && event.altKey && event.which === keyCode.SPACE) {
                        $el.addClass('shown');
                        $el.find('input').focus();
                    }
                });

              $el.on('focusout oui-select', function () {
                    $el.removeClass('shown');
                });

              $el.on('keydown', function (event) {
                    if (event.which === keyCode.ESC) {
                        $el.removeClass('shown');
                    }
                });
          }

          if (options.search) {
              $el.find('input')
                .attr('role', 'combobox')
                .attr('aria-autocomplete', 'list')
                .attr('aria-expanded', 'false');
          }

          return $el;
      };

      /**
       * @method TypeAheadNonTokenView#init
       */
      tvp.init = function () {
          var input = this._$el.children('input');
          var picker = dom.get(input, 'picker');
          this._$el.find('.instructions[data-name=' + this._settings.name + ']').attr('id', this.id() + '-instructions-' + this._settings.name);
          input.attr('aria-describedby', ((input.attr('aria-describedby') || '') + ' ' + this.id() + '-instructions-' + this._settings.name).trim());

          input.attr('aria-describedby', stringUtils.union(input.attr('aria-describedby'), this._$el.find('.filter-status').attr('id')));

          if (picker && picker.selected().length > 0 && this._controller.selected().length === 0) {
              this._controller.selected(picker.selected(), { silent: true });
          }

          if (this._scope) {
              this.trigger('scope-change', this._inputEl.value || '', this.getScope());
          }
      };

      /**
       * Empties out the current text and replaces it with a new string built
       * by calling into the supplied display function
       *
       * @method TypeAheadNonTokenView#select
       */
      tvp.select = function () {
          this.updateValue();

          this.hidePopup();
      };

      /**
       * @method TypeAheadNonTokenView#error
       */
      tvp.error = function () {
          this.updateValue();
          this._$el.add(this._$el.closest('.control-group')).addClass('error');
      };

      /**
       * @method TypeAheadNonTokenView#updateValue
       */
      tvp.updateValue = function () {
          var that = this;

          var value = _.map(this._controller.selected(), function (node) {
              var data = node.data || node;

              if (_.isString(data)) {
                  return data;
              }
              return node.displayValue || that._options.viewFns.display(data);
          }).join(this._options.delimiter);

          if ((!this._options.required && this._controller.selected().length === 0) || this._controller.selected().length > 0) {
              this._$el.add(this._$el.closest('.control-group')).removeClass('error');
          }

          value = escapingUtils.unescape(value, this._options);

          if (this._inputEl.value !== value) {
              this._inputEl.value = value;
          }

          if (this._inputEl === document.activeElement) {
              this._inputEl.scrollLeft = this._inputEl.scrollWidth;
          }
      };

      /**
       * Displays the spinner when a search is happening but has not completed
       * yet.
       *
       * @method TypeAheadNonTokenView#loading
       */
      tvp.loading = function () {
          if (this._$input.is(document.activeElement)) {
              //Since JQuery objects are not really HTML elements we need to grab the element.
              this._$menu.empty().append($(loadingTemplate));

              this.showPopup(this._$el);
          }
      };

      /**
       * Marks the node as active and moves the viewport accordingly.
       *
       * @method TypeAheadNonTokenView#active
       * @param {Object} activeData The node that is active.
       */
      tvp.active = function (activeData) {
          var active;
          this._$menu.find('.active').removeClass('active');
          this._$input.removeAttr('aria-activedescendant');

          if (!_.isNull(activeData.node)) {
              active = this._$menu.find('[data-id=' + activeData.node.id + ']');
              active.addClass('active');
              this._$input.attr('aria-activedescendant', active.attr('id'));

              scrollingUtils.positionScrollTop(active, this._$menu);
          }

      };

      /**
       * Builds the markup for the menu.
       *
       * @method TypeAheadNonTokenView#buildMenu
       * @param {Array} data The data for the menu
       * @param {String} query The current query
       */
      tvp.buildMenu = function (data, query) {
          var that = this;
          var regex = new RegExp('(' + stringUtils.regexEscape(query) + ')', 'gi');
          var text;
          var highlightText;

          /**
           * @method TypeAheadNonTokenView#buildMenu~buildNode
           * @private
           *
           * @param {Object} node
           *
           * @returns {String}
           */
          function buildNode(node) {
              text = node.displayValue || that._options.viewFns.display(node.data);
              that._selectedMap[text] = node;
              highlightText = text;

              if (that._options.highlight) {
                  highlightText = text.replace(regex, '<strong>$1</strong>');
              }
              return nodeTemplate({
                  title: that._options.viewFns.text ? that._options.viewFns.text(node.data) : text,
                  text: highlightText,
                  pageId: that.id() + '-' + node.id,
                  id: node.id,
                  active: that._controller.active().node === node ? 'active' : ''
              });
          }

          /**
           * @method TypeAheadNonTokenView#buildMenu~buildGroup
           * @private
           *
           * @param {Object} group
           * @param {String} key
           *
           * @returns {String}
           */
          function buildGroup(group, key) {
              var children = _.reduce(group, function (memo, child) {
                  return memo += buildNode(child);
              }, '');

              return groupTemplate({
                  title: key,
                  text: key,
                  children: children
              });
          }

          if (data.length > 0 && !this._options.scroll) {
              //If no scrolling then only put their max suggestions in the markup
              data = _.take(data, this._options.maxSuggestions);
          }

          if (this._options.viewFns.groupBy) {
              data = _.groupBy(data, function (node) {
                  return that._options.viewFns.groupBy(node.data);
              });

              return _.reduce(data, function (memo, node, key) {
                  return memo + buildGroup(node, key);
              }, '');
          }

          return _.reduce(data, function (memo, node) {
              return memo + buildNode(node);
          }, '');
      };

      /**
       * The method to take the markup and construct a DOM element from it.
       * Also since this is a new element we need to add the click events to the
       * menu.
       *
       * @method TypeAheadNonTokenView#render
       * @param {String} markup The markup representing the menu.
       */
      tvp.render = function (markup) {
          //Find proper z-indexing, append to the location
          //Since it is debounced - need to make sure we are still active/focused
          if (this._inputEl === document.activeElement && markup !== '') {
              this._$menu.empty().append(markup);

              //Reposition popup after adding the nodes to it (giving it proper height to calculate with)
              this.showPopup(this._$el);

              this._$menu
                .css('min-width', this._$el.innerWidth())
                .removeClass('grow-to-text');

              if (this._options.menuSize === 'input-width') {
                  this._$menu.width(this._$el.innerWidth());
              } else if (this._options.menuSize === 'grow-to-text') {
                  this._$menu.width('');
                  this._$menu.addClass('grow-to-text');
              } else {
                  this._$menu.width(this._options.menuSize);
              }

              //Mess with scroll-top if it is scrollable
              if (this._options.scroll) {
                  scrollingUtils.positionScrollTop(this._$menu.find('.active'), this._$menu);
              }

          } else {
              this.hidePopup();
          }
      };

      /**
       * @method TypeAheadNonTokenView#loaded
       * @param {Array} data The data for the menu
       * @param {String} query The query used to get the data.
       */
      tvp.loaded = function (data, query) {
          //Build menu, position menu, show menu
          this.render(this.buildMenu(data, query));
      };

      /**
       * @method TypeAheadNonTokenView#clear
       */
      tvp.clear = function () {
          this._$el.addClass('empty');
          this._inputEl.value = '';
          this._$el.find('.filter-status').text(this._options.locale.filter.none).attr('aria-hidden', false);
      };

      /**
       * @method TypeAheadNonTokenView#update
       */
      tvp.update = function () {
          this._$el.toggleClass('empty', this._inputEl.value === '');

          if (this._options.style === 'search') {
              this._$el.find('.filter-status').text(this._options.locale.filter.applied);
          }
      };

      /**
       * @method TypeAheadNonTokenView#change
       */
      tvp.change = function () {
          if (this._triggerChange) {
              this._$input.trigger('change', {
                  update: false
              });
              this._triggerChange = false;
          }
      };

      /**
       * @method TypeAheadNonTokenView#value
       *
       * @param {String} newQuery
       */
      tvp.value = function (newQuery) {
          this._$input.val(newQuery);
      };

      return TypeAheadView;
  }

);

define('oui/widgets/typeAhead',['require','underscore','oui/base/input','oui/widgets/typeAhead/typeAheadModel','oui/widgets/typeAhead/typeAheadController','oui/widgets/typeAhead/typeAheadTokenView','oui/widgets/typeAhead/typeAheadNonTokenView'],function (require) {
      'use strict';

      var _ = require('underscore');
      var Input = require('oui/base/input');
      var TypeAheadModel = require('oui/widgets/typeAhead/typeAheadModel');
      var TypeAheadController = require('oui/widgets/typeAhead/typeAheadController');
      var TypeAheadTokenView = require('oui/widgets/typeAhead/typeAheadTokenView');
      var TypeAheadNonTokenView = require('oui/widgets/typeAhead/typeAheadNonTokenView');

      /**
       * @typedef {Object} TypeAhead#Style
       * @readonly
       *
       * @property {String} default
       * @property {String} search
       * @property {String} scoped-search
       * @property {String} search-collapsible
       */

      /**
       * @typedef {Object} TypeAhead#Config
       *
       * @property {Locale} [locale={}] The specific strings that are used by the widget and will use the defaults set from
       * the {@link module:oui.lang|lang}.

       * @property {Boolean|Object} [autoEscape=true] Whether the combobox should escape/unescape the values from the
       * input. When the property is an Object it should have two functions escape and unescape that take in the string and
       * provide the escaped/unescaped version of the string. When the property is true it will default to using _.escape
       * and _.unescape for the strings.
       *
       * @property {Boolean} [scroll=true] Whether the menu should be scrollable.
       *
       * @property {Number} [maxSuggestions=10] How many suggestions should be placed into the suggestion menu.
       *
       * @property {Number} [delay=250] The number, in ms, to wait before the input is used to query for suggestions.
       *
       * @property {Boolean} [escToClear=false] Whether or not the typeAhead should clear the query on ESC.
       *
       * @property {String} [delimiter=', '] Character or string to use as the delimiter between querys/inputs.
       *
       * @property {String|jQuery} [appendTo='body'] The selector or jQuery object to append the menu to. Useful when the
       * default results in a z-index issue.
       *
       * @property {Boolean} [multiSelect=false] Whether or not the typeahead component can hold multiple selections.
       *
       * @property {Boolean} [highlight=true] Whether or not it will bold the matching portion of the query within the menu.
       *
       * @property {Boolean} [allowNewValues=false] Whether or not new values that are not returned from a search are added
       * to the typeahead dynamically.
       *
       * @property {Boolean} [tokenize=true] Whether to use tokens in the typeahead component. Tokenize is only read when
       * the component is in multiselect mode.
       *
       * @property {Boolean} [search=true] Whether the component should perform the search itself. If false, an event is
       * fired and another component, such as a tree, should use the query and perform its own search.
       *
       * @property {Boolean} [allowEmpty=false] Whether to search when the query is the empty string.
       *
       * @property {TypeAhead#Style} [style='default'] Whether to use a standard input or a 'search' style with rounded corners and
       * a magnifying glass. Valid values are 'default', 'search', and 'scoped-search', 'search-collapsible'.
       *
       * @property {Boolean} [fluid=false] Whether the input should take up all available space it can.
       *
       * @property {Number} [minLength=0] The minimum string length before a search is issued.
       * @property {Boolean} [required=false] Whether or not the typeahead needs a selection to be valid.
       * @property {String} [menuSize=input-width] The size of the menu for the typeahead. Valid options are 'input-width',
       * 'grow-to-text' or a string with 'px'.
       * @property {Object} [viewFns] The view functions for the typeahead
       * @property {Function} [viewFns.matches] Function to determine whether an item matches the currently selected token.
       * Used when removing nodes that might have custom displays.
       *
       * @property {Object} [combobox] When using a scoped typeahead the combobox with values that represent what the current
       * scope is. Useful for a 'type' filter.
       *
       * @property {Array<Object>} [scopes] The scopes to display within the combobox when the style is 'scoped-search'. This
       * property is only required when using 'scoped-search'.
       *
       * @property {Object} [recordKeys] The keys to use for various attributes on the record.
       * @property {String|Array<String>} [recordKeys.id] Will use this attribute as the filter for matching existing nodes
       * to new nodes. Default behavior is to match complete object structures.
       */

      var tap;

      var __WIDGET_NAME__ = 'typeahead';
      var __DEFAULTS__ = {
          showErrors: 'blur',
          autoEscape: true,
          scroll: true,
          maxSuggestions: 10,
          delay: 250, //ms
          escToClear: false,
          delimiter: ', ',
          appendTo: 'body',
          multiSelect: false,
          highlight: true,
          collapsible: false,

          allowNewValues: false,

          // We set this to true, since it will be most likely that users will want tokenized if they do multiselect
          // Having tokenize:true, does not affect multiSelect:false
          tokenize: true,
          search: true,
          allowEmpty: false,
          style: 'default',
          fluid: false,
          minLength: 0,
          required: false,
          menuSize: 'input-width',
          viewFns: {
              /**
               * @see TypeAhead#Config.viewFns.matches
               */
              matches: function () {
                  return false;
              }
          },
          combobox: {
              shrinkToFit: true,
              menuSize: 'grow-to-text',
              multiSelect: false,
              keyboardInput: false,
              options: []
          }
      };

      //TODO: Update the picker options docs when these change.
      /**
       * The typeahead (aka autocomplete) component that recommends matches based on the
       * input to the component.
       *
       * It is navigable with the arrow keys and items can be selected with the enter key.
       *
       * See {@link http://www.w3.org/TR/wai-aria-practices/#autocomplete|ARIA Best Practices}
       * for information regarding keyboard shortcuts and general ARIA attributes and states.
       *
       *  ##### **Keyboard Shortcuts**
       *
       *  * <kbd>Down</kbd> - Highlights menu item below the currently highlighted item.
       *  * <kbd>Up</kbd> - Highlights menu item above the currently highlighted item.
       *  * <kbd>Enter</kbd> - Selects the currently highlighted item.
       *  * <kbd>Escape</kbd> - Closes the dropdown and if the escToClear option is turned on clears the selection.
       *
       *
       *  ##### **Relevant Events**
       *
       *  * {@link TypeAhead#result-added|result-added}
       *  * {@link TypeAhead#loaded|loaded}
       *  * {@link TypeAhead#selection-error|selection-error}
       *  * {@link TypeAhead#required-error|required-error}
       *  * {@link TypeAhead#select|select}
       *  * {@link TypeAhead#invalidate|invalidate}
       *  * {@link TypeAhead#active|active}
       *  * {@link TypeAhead#search|search}
       *  * {@link TypeAhead#scope-change|scope-change}
       *
       * @borrows TypeAheadModel#event:result-added as TypeAhead#result-added
       * @borrows TypeAheadModel#event:loaded as TypeAhead#loaded
       * @borrows TypeAheadModel#event:selection-error as TypeAhead#selection-error
       * @borrows TypeAheadModel#event:required-error as TypeAhead#required-error
       * @borrows TypeAheadModel#event:select as TypeAhead#select
       * @borrows TypeAheadModel#event:invalidate as TypeAhead#invalidate
       * @borrows TypeAheadModel#event:active as TypeAhead#active
       * @borrows TypeAheadModel#event:search as TypeAhead#search
       * @borrows TypeAheadNonTokenView#event:scope-change as TypeAhead#scope-change
       *
       * @class TypeAhead
       * @since 1.3.0
       * @extends Input
       *
       * @param {TypeAhead#Config} [options] The options to override defaults
       * @param {jQuery} [element] The jQuery element to use for typeAhead
       * @param {Object} [initData] The initial data passed to the typeAhead
       */
      function TypeAhead(options, element, initData) {
          var model;

          //Super call
          Input.call(this, options, element, initData);

          if (this._options.style === 'search-collapsible') {
              this._options.style = 'search';
              this._options.collapsible = true;
          }

          if (this._options.style === 'scoped-search') {
              this._options.style = 'search';
          }

          model = new TypeAheadModel(this._options, this._initData);
          /**
           * The controller that the component keeps a reference to.
           *
           * @private
           * @type {TypeAheadController}
           * @memberof! TypeAhead.prototype
           */
          this._controller = new TypeAheadController(model, this._options);

          //Setup proper view for the type ahead. Is it tokens or plain text.
          /**
           * The view element that the component keeps a reference to.
           *
           * @private
           * @type {TypeAheadView}
           * @memberof! TypeAhead.prototype
           */
          if (this._options.tokenize && this._options.multiSelect) {
              this._view = new TypeAheadTokenView(model, this._controller, this._element, this._options, this._settings);
          } else {
              this._view = new TypeAheadNonTokenView(model, this._controller, this._element, this._options, this._settings);
          }

          Object.defineProperty(this, '_data', {
              get: this.selected,
              set: this.selected
          });

          this._setupEvents([model, this._view]);
      }

      TypeAhead.prototype = Object.create(Input.prototype);
      tap = TypeAhead.prototype;

      tap.constructor = TypeAhead;

      /**
       * Returns the name of the widget that is used in eventing and id generation
       *
       * @method TypeAhead#name
       * @returns {String} - the widgets name
       */
      tap.name = function () {
          return __WIDGET_NAME__;
      };

      /**
       * @method TypeAhead#_translate
       * @private
       *
       * @param {String} name
       *
       * @returns {Array}
       */
      tap._translate = function (name) {
          if (name === 'select') {
              var data = arguments[1];

              return [name, _.pluck(data, 'data')];
          }

          return arguments;
      };

      /**
       * Combined getter/setter for the typeahead component.
       *
       * @method TypeAhead#selected
       * @fires TypeAheadModel#selection-error
       * @fires TypeAheadModel#select
       * @param {Array} id The nodes to select in the typeahead component.
       * @returns {Array} The selected objects in the typeahead.
       */
      tap.selected = function (id) {
          if (!arguments.length) {
              //Public get should not return the whole node. Just the part they care about.
              return _.pluck(this._controller.selected(), 'data');
          }

          this._controller.selected(id);
      };

      /**
       * Combined getter/setter for the selected nodes of the typeahead component.
       *
       * **Note**: This function is silent. For more information regarding its counterpart see {@link TypeAhead#selected}
       *
       * @method TypeAhead#selected
       * @param {Array} selected The nodes to select in the typeahead component.
       * @returns {Array} The selected objects in the typeahead.
       */
      tap.value = function (selected) {
          if (!arguments.length) {
              return this.selected();
          }

          this._controller.selected(selected, {
              silent: true
          });
      };

      /**
       * Ability to retrieve the widgets default options for use in merging with passed in options as
       * well as with composed widgets such as the date and time picker.
       *
       * @method TypeAhead#defaults
       * @returns {TypeAhead#Config} - the widgets default options
       */
      tap.defaults = function () {
          return __DEFAULTS__;
      };

      /**
       * Empties the typeahead's model and empties the view.
       *
       * @method TypeAhead#invalidate
       */
      tap.invalidate = function () {
          this._controller.invalidate();
      };

      /**
       * Issues a search on a typeahead programmatically
       *
       * **Note:** currently only works with typeaheads that do not show menus (search option is false)
       *
       * @method TypeAhead#search
       * @param {String} [newQuery] The query to search with
       */
      tap.search = function (newQuery) {
          if (newQuery != null) {
              this._view.value(newQuery);
          }

          //Currently not supporting the search option right now - as the menu only shows up when the input has focus
          //TODO: do we want to auto focus the input or just show the menu regardless? should showing the menu be an option
          //passed to this function?
          if (!this._options.search) {
              this._controller.search(newQuery || this._controller.active().query, 0);
          }
      };

      /**
       * @method TypeAhead#isValid
       *
       * @returns {Boolean}
       */
      tap.isValid = function () {
          return this._controller.isValid();
      };

      /**
       * @method TypeAhead#error
       *
       * @returns {Object?}
       */
      tap.error = function () {
          return this._controller.error();
      };

      return TypeAhead;
  }

);

define('oui/widgets/picker',['require','underscore','oui/base/widget','oui/widgets/picker/pickerView','oui/widgets/treePanel','oui/widgets/typeAhead'],function (require) { //jshint ignore:line
      'use strict';

      var _ = require('underscore');
      var Widget = require('oui/base/widget');
      var PickerView = require('oui/widgets/picker/pickerView');
      var TreePanel = require('oui/widgets/treePanel');
      var TypeAhead = require('oui/widgets/typeAhead');

      /**
       * @callback Picker#display
       * @param {Array} nodes The nodes to generate the string representation for.
       * @returns {String} - The display value
       */

      /**
       * @typedef Picker#Config
       * @property {Boolean} [repeatable=false] Whether or not the picker should allow selecting the items in the picker that
       * may already be selected.
       * @property {Object} viewFns
       * @property {Picker#display} viewFns.display
       * @property {Boolean|jQuery|String} [modal=false] Whether or not the picker should be displayed in a modal popup or
       * not. When passed a jQuery Object or a string it is used as the modal for the picker. There are two requirements to
       * the DOM object when using a custom modal: 1. The element that is going to hold the tree panel has class
       * 'tree-content' 2. The element that is going to hold the type ahead has class 'search'. This property assumes the
       * condition is true when passed in a jQuery object or a string.
       * @property {String} [modalClasses] Custom classes to add to the modal for additional styling.
       * @property {Boolean} [createTypeAhead=false] Whether or not the component will also generate a typeahead using the
       * supplied functions.
       * @property {Locale} [locale={}] The specific strings that are used by the widget and will use the defaults set from
       * the {@link module:oui.lang|lang}.
       * @property {Boolean} [required=false] Whether or not the field is required. If the field is not required then a
       * clear/remove button will show up in the modal and allow a user to deselect all of their previous selection.
       * @property {Boolean} [buttonTriggerOnly=false] Whether or not the widget is only a button instead of an input
       * element with a button.
       * @property {Boolean} [openFromInput=false] Whether the modal should be triggered from the input when clicked.
       * @property {String|jQuery} [appendTo='body'] Where to append the modal.
       * @property {Boolean} [fluid=false] Whether the input element should take up the widget of its container.
       * @property {Boolean} [search=false] Whether or not to show the search in the picker overridden if the Tree configuration contains a filter function.
       * @property {TreePanel#Config} [tree] The tree options.
       * @property {TypeAhead#Config} [typeahead] The typeahead options.
       */

      var __WIDGET_NAME__ = 'picker';
      var __OVERRIDES__ = {
          tree: {
              picker: true
          },
          typeAhead: {
              fluid: true,
              escToClear: true,
              style: 'search',
              search: false
          }
      };
      var __DEFAULTS__ = {
          repeatable: false,
          viewFns: {
              /**
               * @see {Picker#Config.viewFns.display}
               */
              display: function (nodes) {// jscs:ignore jsDoc
                  return _.map(nodes, function (node) {
                      return node.name || node.value;
                  }).join(', ');
              }
          },
          modal: false,
          modalClasses: null,
          createTypeAhead: false,
          locale: {},
          required: false,
          buttonTriggerOnly: false,
          openFromInput: false,
          appendTo: 'body',
          fluid: false,
          search: false,
          filterOnScope: true,
          tree: _.extend({}, TreePanel.prototype.defaults(), __OVERRIDES__.tree),
          typeAhead: _.extend({}, TypeAhead.prototype.defaults(), __OVERRIDES__.typeAhead)
      };
      var pp;

      /**
       * The element that represents a Picker. It is a container around the tree and typeahead
       * widgets.
       *
       * For more information on used classes
       *
       * * {@link TreePanel#Config}
       * * {@link TypeAhead#Config}
       *
       * ##### **Relevant Events**
       *
       * * {@link Picker#pre-show|pre-show}
       * * {@link Picker#pre-accept|pre-accept}
       * * {@link Picker#accept|accept}
       * * {@link Picker#cancel|cancel}
       *
       * @borrows PickerView#event:pre-show as Picker#pre-show
       * @borrows PickerView#event:pre-accept as Picker#pre-accept
       * @borrows PickerView#event:accept as Picker#accept
       * @borrows PickerView#event:cancel as Picker#cancel
       *
       * @class Picker
       * @extends Widget
       * @since 1.3.0
       *
       * @param {Picker#Config} options The options for the picker.
       * @param {Object} [element] The element to use for the picker. It will create an element if none is
       *    given.
       * @param {Array} [initData] The data to use before any loading takes place.
       */
      function Picker(options, element, initData) {
          //Super call
          Widget.call(this, options, element, initData);

          //Setting required to whatever the pickers required option is
          this._options.tree.required = this._options.required;

          this._options.search = this._options.search ||
            (this._options.tree && this._options.tree.storeFns && _.isFunction(this._options.tree.storeFns.filter)) ||
            (this._options.tree && this._options.tree.lazyLoad === false && this._options.tree.clientFilter);

          /**
           * The tree panel component that makes up a picker.
           *
           * @type {TreePanel}
           * @private
           * @memberof! Picker.prototype
           */
          this._treePanel = new TreePanel(this._options.tree, this._initData);

          /**
           * The type ahead component that combined with a tree makes up the entire
           * picker.
           *
           * @type {TypeAhead}
           * @private
           * @memberof! Picker.prototype
           */
          this._typeAhead = new TypeAhead(this._options.typeAhead);

          /**
           * The view for the picker.
           *
           * @type {PickerView}
           * @private
           * @memberof! Picker.prototype
           */
          this._view = new PickerView(this._options, this._element, this._treePanel, this._typeAhead, this._settings);

          if (this._options.createTypeAhead && this._options.modal && !this._options.buttonTriggerOnly) {
              this._outerTypeAhead = new TypeAhead({
                  multiSelect: this._options.tree.multiSelect,
                  appendTo: this._options.createTypeAhead.appendTo || this._options.appendTo,
                  fluid: this._options.fluid,
                  tokenize: false,
                  required: this._options.required,

                  //For now
                  storeFns: {
                      search: this._options.tree.storeFns.filter
                  },
                  viewFns: {
                      display: _.get(this._options, 'createTypeAhead.viewFns.display', this._options.tree.viewFns.display)
                  }
              }, this._view.element().find('input'), initData);
          }

          Object.defineProperty(this, '_data', {
              get: this.selected,
              set: this.selected
          });

          this._setupEvents([this._treePanel, this._view]);
      }

      Picker.prototype = Object.create(Widget.prototype);
      pp = Picker.prototype;

      pp.constructor = Picker;

      /**
       * Gets the widget name. Used to prefix some events.
       *
       * @method Picker#name
       * @returns {String} 'picker'
       */
      pp.name = function () {
          return __WIDGET_NAME__;
      };

      /**
       * Initialize the picker by initializing its subcomponents.
       *
       * @method Picker#init
       * @private
       */
      pp._init = function () {

          if (!(this._options.modal || _.isObject(this._options.modal))) {
              this._treePanel.init();
              this._typeAhead.init();
          }

          if (this._outerTypeAhead) {
              this._outerTypeAhead.init();
          }

          this._treePanel.removeOff();
          this._typeAhead.removeOff();
      };

      /**
       * Getter/Setter for the selected nodes in the picker.
       *
       * If no parameter is given then the method behaves like a getter
       * otherwise it will overwrite the currently selected nodes with
       * the given nodes.
       *
       * @example
       *
       *     var picker = new oui.Picker(); //Assuming valid options
       *
       *     //Initialization
       *
       *     //Finds the first nodes that match the fields given.
       *     picker.selected([
       *         { name : "Test name" },
       *         { id : "Test id" }
       *     ]);
       *
       *     picker.selected(); //Returns [
       *         {
       *             name : "Test name"
       *             //Any additional data that is in the node
       *         },
       *         {
       *             id : "Test id"
       *             //Any additional data that is in the node
       *         }
       *     ]
       *
       *     picker.selected([]); //Removes all selected nodes
       *
       * @method Picker#selected
       * @fires TreePanel#picker-accept
       * @param {Array} data The data to search for to make selected
       *   in the tree component.
       * @returns {Array} The selected nodes of the picker.
       */
      pp.selected = function (data) {
          if (!arguments.length) {
              return this._treePanel.selected();
          }

          this._treePanel.selected(data);
      };

      /**
       * Getter/Setter for the selected nodes in the picker.
       *
       * Similar to {@link Picker#selected} except it does
       * not trigger an event after selection.
       *
       * @method Picker#value
       * @param {Array} data The selected data of the tree component.
       * @returns {Array} The selected nodes of the picker.
       */
      pp.value = function (data) {
          if (!arguments.length) {
              return this._treePanel.selected();
          }

          this._treePanel.value(data);
      };

      /**
       * Getter/Setter for the unselectable nodes in the picker.
       *
       * If no parameter is given then the method behaves like a getter
       * otherwise it will overwrite the currently unselectable nodes with
       * the given nodes.
       *
       * @example
       *
       *     var picker = new oui.Picker(); //Assuming valid options
       *
       *     //Initialization
       *
       *     //Finds the first nodes that match the fields given.
       *     picker.unselectable([
       *         { name : "Test name" },
       *         { id : "Test id" }
       *     ]);
       *
       *     picker.unselectable(); //Returns [
       *         {
       *             name : "Test name"
       *             //Any additional data that is in the node
       *         },
       *         {
       *             id : "Test id"
       *             //Any additional data that is in the node
       *         }
       *     ]
       *
       *     picker.unselectable([]); //Removes all unselecable nodes
       *
       * @method Picker#unselectable
       * @param {Array} data The data to search for to make unselectable
       *   in the tree component.
       * @returns {Array} The unselectable nodes of the picker.
       */
      pp.unselectable = function (data) {
          if (!arguments.length) {
              return this._treePanel.unselectable();
          }

          this._treePanel.unselectable(data);
      };

      /**
       * Getter/Setter for the expanded nodes in the picker.
       *
       * If no parameter is given then the method behaves like a getter
       * otherwise it will overwrite the currently expanded nodes with
       * the given nodes.
       *
       * @example
       *
       *     var picker = new oui.Picker(); //Assuming valid options
       *
       *     //Initialization
       *
       *     //Finds the first nodes that match the fields given.
       *     picker.expanded([
       *         { name : "Test name" },
       *         { id : "Test id" }
       *     ]);
       *
       *     picker.expanded(); //Returns [
       *         {
       *             name : "Test name"
       *             //Any additional data that is in the node
       *         },
       *         {
       *             id : "Test id"
       *             //Any additional data that is in the node
       *         }
       *     ]
       *
       *     picker.expanded([]); //Removes all expanded nodes from the
       *                          //list of expanded nodes
       *
       * @method Picker#expanded
       * @param {Array} data The data to search for to make expanded
       *   in the tree component.
       * @returns {Array} The expanded nodes of the picker.
       */
      pp.expanded = function (data) {
          if (!arguments.length) {
              return this._treePanel.expanded();
          }

          this._treePanel.expanded(data);
      };

      /**
       * Getter/Setter for the disabled nodes in the picker.
       *
       * If no parameter is given then the method behaves like a getter
       * otherwise it will overwrite the currently disabled nodes with
       * the given nodes.
       *
       * @example
       *
       *     var picker = new oui.Picker(); //Assuming valid options
       *
       *     //Initialization
       *
       *     //Finds the first nodes that match the fields given.
       *     picker.disabled([
       *         { name : "Test name" },
       *         { id : "Test id" }
       *     ]);
       *
       *     picker.disabled(); //Returns [
       *         {
       *             name : "Test name"
       *             //Any additional data that is in the node
       *         },
       *         {
       *             id : "Test id"
       *             //Any additional data that is in the node
       *         }
       *     ]
       *
       *     picker.disabled([]); //Removes all disabled nodes from the
       *                          //list of disabled nodes
       *
       * @method Picker#disabled
       * @param {Array} data The data to search for to make disabled
       *   in the tree component.
       * @returns {Array} The disabled nodes of the picker.
       */
      pp.disabled = function (data) {
          if (!arguments.length) {
              return this._treePanel.disabled();
          }

          this._treePanel.disabled(data);
      };

      /**
       * Provides custom logic for the picker widget when destroy is called.
       *
       * @method Picker#_destroy
       * @private
       */
      pp._destroy = function () {
          this._treePanel.destroy();
          this._typeAhead.destroy();
      };

      /**
       * @method Picker#defaults
       * @returns {Object} The defaults for the picker.
       */
      pp.defaults = function () {
          return __DEFAULTS__;
      };

      /**
       * Returns the JSON representation of the Tree component.
       *
       * @method Picker#tree
       * @returns {Object} - The JSON representation of the tree.
       */
      pp.tree = function () {
          return this._treePanel.tree();
      };

      /**
       * Locks and disables the view as well as remove any data from the model.
       *
       * @method Picker#lock
       * @param {Boolean} locked Whether to render the view as locked or not
       */
      pp.lock = function (locked) {
          if (this._options.modal && !this._options.buttonTriggerOnly) {
              this._view.lock(locked);

              //TODO: remove all data from tree and typeahead.
          }
      };

      /**
       * Gets the tree panel used by the picker.
       *
       * @method Picker#treePanel
       * @returns {TreePanel} treePanel The tree panel used by the picker.
       */
      pp.treePanel = function () {
          return this._treePanel;
      };

      /**
       * Get the type ahead used by the picker.
       *
       * @method Picker#typeAhead
       * @returns {TypeAhead} ta The typeAhead used by the picker.
       */
      pp.typeAhead = function () {
          return this._typeAhead;
      };

      /**
       * Remove all data from the model (expanded, selected, disabled,...) and reloads
       * from the root node.
       *
       * @method Picker#invalidate
       */
      pp.invalidate = function () {
          this._treePanel.invalidate();
          this._typeAhead.invalidate();
      };

      /**
       * If the picker has the modal option set it will display the modal.
       *
       * @method Picker#show
       */
      pp.show = function () {
          if (this._options.modal) {
              this._view.show();
          }
      };

      /**
       * If the picker has the modal option set it will hide the modal.
       *
       * @method Picker#hide
       */
      pp.hide = function () {
          if (this._options.modal) {
              this._view.hide();
          }
      };

      /**
       * Resets attributes and recalls attributes.
       *
       * If no attributes are given it will reset all attributes.
       *
       * Differs from invalidating because it keeps the same data.
       *
       * @method Picker#reset
       * @param {Array<String>} attrs The attributes to reset.
       */
      pp.reset = function (attrs) {
          this._typeAhead.invalidate();
          this._treePanel.reset(attrs);
      };

      /**
       * @method Picker#empty
       */
      pp.empty = function () {
          this.invalidate();
          this.value([]);
      };

      /**
       * @method Picker#updateOption
       *
       * @param {String} option
       * @param {String|Boolean|Object} value
       */
      pp.updateOption = function (option, value) {
          if (option === 'tree') {
              this.treePanel().updateOptions(value);
          } else {
              Widget.prototype.updateOption.call(this, option, value);
          }
      };

      /**
       * @method Picker#nodes
       *
       * @returns {Array}
       */
      /**
       * @method Picker#nodes
       *
       * @param {Array} data
       */
      pp.nodes = function (data) {
          if (!arguments.length) {
              return this.treePanel().nodes();
          }

          return this.treePanel().nodes(data);
      };

      /**
       * @method Picker#isValid
       *
       * @returns {Boolean}
       */
      pp.isValid = function () {
          //TODO: migrate this when the picker is refactored for a more MVC style approach
          return this._view.isValid();
      };

      return Picker;
  }
);

define('oui/widgets/url/urlModel',['oui/base/model'], function (Model) {
    'use strict';

    var ump;

    /**
     * Fired when the url has changed.
     *
     * @event UrlModel#url-change
     * @param {String} url The current url.
     */
    /**
     * @event UrlModel#state-change
     * @param {String} state The current state of the url field.
     */
    /**
     * @event UrlModel#validation-state-change
     * @param {String} state The current state of the url field.
     */

    /**
     * Data associated with the UrlField component.
     *
     * @class UrlModel
     * @private
     * @since 1.3.0
     * @extends Subscribable
     *
     * @param {String} initData Initial URL for the widget
     */
    function UrlModel(initData, options) {
        // Super call
        Model.call(this, initData, options);

        /**
         * The state of the URL validation
         *
         * @type {String}
         * @private
         * @memberof! UrlModel.prototype
         * @default 'none'
         */
        this._state = 'none';
    }

    UrlModel.prototype = Object.create(Model.prototype);
    ump = UrlModel.prototype;

    ump.constructor = UrlModel;

    /**
     * @method UrlModel#name
     * @returns {String} 'url'
     */
    ump.name = function () {
        return 'url';
    };

    /**
     * Gets or sets the URL stored in the widget.
     *
     * @method UrlModel#url
     * @protected
     * @param {String} [url] If provided, the new URL to store in the model
     * @returns {String|undefined}
     */
    ump.url = ump.data;

    /**
     * Gets or sets the current validation state of the widget.
     *
     * States are: 'none', 'valid', 'invalid', 'testing'
     *
     * @method UrlModel#state
     * @protected
     * @param {String} [state] If provided, sets the state to the specified value
     * @returns {String|undefined}
     */
    ump.state = function (state) {
        if (!arguments.length) {
            return this._state;
        }

        if (this._state !== state) {
            this._state = state;
            this.trigger('state-change', this._state);
            this.trigger('validation-state-changed', this._state);
        }
    };

    return UrlModel;
});

define('oui/widgets/url/urlController',[], function () {
    'use strict';

    var ucp;

    /**
     * @class UrlController
     * @private
     * @since 1.3.0
     *
     * @param {UrlModel} model The UrlModel to manipulate the data.
     * @param {Object} options The specific options regarding the controller.
     */
    function UrlController(model, options) {

        this._model = model;
        this._options = options;
    }

    ucp = UrlController.prototype;

    ucp.constructor = UrlController;

    /**
     * Gets or sets the URL to be stored in the widget.
     *
     * @method UrlController#url
     * @protected
     * @param {String} [url] The url to set the widget to
     * @param {Object} [eventParams]
     * @returns {String|undefined}
     */
    ucp.url = function (url, eventParams) {
        if (!arguments.length) {
            return this._model.url();
        }
        this._model.url(url, eventParams);
    };

    /**
     * Tests the URL stored in the widget for validity and availability.
     *
     * @method UrlController#testUrl
     * @protected
     */
    ucp.testUrl = function () {
        var that = this;
        this._model.state('testing');

        this._options.testFn(this._model.url(), function (isValid) {
            if (isValid) {
                that._model.state('valid');
            } else {
                that._model.state('invalid');
            }
        });
    };

    /**
     * @method UrlController#state
     *
     * @returns {String} state
     */
    /**
     * @method UrlController#state
     *
     * @param {String} state
     */
    ucp.state = function (state) {
        if (!arguments.length) {
            return this._model.state();
        }

        this._model.state(state);
    };

    return UrlController;
});


define('template-43',[],function () { return '<div class="orcl-widget orcl-url-field">\n    <div class="input-container">\n        <input type="url" />\n    </div>\n</div>\n';});


define('template-44',[],function () { return '<div class="orcl-widget orcl-url-field">\n    <div class="input-container">\n    </div>\n</div>\n';});


define('template-45', ['underscore'], function (_) { return function(data) {
var __t, __p = '';
__p += '<div class="url-controls"> <button type="button" aria-label="' +
((__t = ( data.test )) == null ? '' : __t) +
'" title="' +
((__t = ( data.test )) == null ? '' : __t) +
'" class="btn btn-blue-light url-test-button">' +
((__t = ( data.test )) == null ? '' : __t) +
'</button></div>';
return __p
};});


define('template-46',[],function () { return '<i class="pgbu-icon pgbu-icon-loading"></i>\n';});

define('oui/widgets/url/urlView',['require','underscore','jquery','oui/base/view','template-43','template-44','template-45','template-46','oui/utils/escapingUtils'],function (require) {
      'use strict';

      var _ = require('underscore');
      var $ = require('jquery');
      var View = require('oui/base/view');
      var urlTemplate = require('template-43');
      var wrapTemplate = require('template-44');
      var controlsTemplate = require('template-45');
      var loadingTemplate = require('template-46');
      var escapingUtils = require('oui/utils/escapingUtils');

      var uvp;

      /**
       * The type ahead view is responsible for displaying the suggestions when they loaded.
       *
       * @class UrlView
       * @private
       * @since 1.3.0
       * @extends View
       *
       * @param {UrlModel} model The UrlModel that is responsible for this view.
       * @param {UrlController} controller The UrlController that will modify the model.
       * @param {Object} element The default HTML input element that users can interact with.
       * @param {Object} options View specific options.
       */
      function UrlView(model, controller, element, options, settings) {
          var that = this;

          // Super call
          View.call(this, options, element, controller, settings);

          //TODO: this should not be necessary in the view - should use controller
          /**
           * The model for this view to use.
           *
           * @private
           * @type {UrlModel}
           * @memberof! UrlView.prototype
           */
          this._model = model;

          model.on({
              'view.url-change': this.updateValue.bind(this),
              'view.state-change': this.updateState.bind(this)
          });

          this._inputEl = this._$el.find('input')[0];

          this.updateValue(model.url());

          this._$el.find('input')
              .on('change', function () {
                  that._controller.url(escapingUtils.escape(this.value, that._options));
              })
              .on('input', function () {
                  that.toggleTest(this.value);
                  that._controller.state('none');
              });

          this._$el.find('.url-test-button').on('click', this._controller.testUrl.bind(this._controller));
      }

      UrlView.prototype = Object.create(View.prototype);
      uvp = UrlView.prototype;

      uvp.constructor = UrlView;

      /**
       * Destroys the url field and any created elements as well as events used within the field.
       *
       * @method UrlView#destroy
       * @param {Object} cleanUp Option for what to clean up
       * @param {Boolean} cleanUp.self Whether or not to clean up the initial element
       */
      uvp.destroy = function (cleanUp) {
          this._$el.find('input').off('change keydown');

          if (cleanUp.self) {
              this._$el.remove();
          }
      };

      /**
       * Builds the url field given either no element or an input element.
       *
       * @method UrlView#buildElement
       * @param {Object} opts The options supplied to the widget
       * @param {?jQuery} el The element supplied to the widget to wrap
       * @returns {jQuery} - The url field
       */
      uvp.buildElement = function (opts, el) {
          var $element;

          if (_.isNull(el)) {
              //Build element from scratch
              $element = $(urlTemplate);
          } else if (el.jquery) {
              //We were given a jquery element
              el.wrap(wrapTemplate);
              $element = el.closest('.orcl-url-field');
          }
          $element.find('.input-container').append(loadingTemplate);

          $element.append($(controlsTemplate(opts.locale)));

          return $element;
      };

      /**
       * Updates the URL on the user visible text box.
       *
       * @method UrlView#updateValue
       * @protected
       * @param {String} value The new URL
       */
      uvp.updateValue = function (value) {
          value = escapingUtils.unescape(value, this._options);

          this._inputEl.value = value;
          this.updateState('none');
          this.toggleTest(value);
      };

      /**
       * Enables and disables the test url button depending on the value of the url field.
       *
       * @method UrlView#toggleTest
       * @param {?String} value The value of the url field to determine whether to disable the test button.
       */
      uvp.toggleTest = function (value) {
          var disabled = _.isNull(value) || value === '';

          this._$el.find('.url-test-button')
            .prop('disabled', disabled)
            .toggleClass('disabled', disabled);
      };

      /**
       * Updates the state of the user visible information.
       *
       * @method UrlView#updateState
       * @protected
       * @param {String} state The new state
       */
      uvp.updateState = function (state) {
          var $controlGroup = this._$el.add(this._$el.closest('.control-group')).removeClass('error success loading');

          switch (state) {
              case 'valid':
                  $controlGroup.addClass('success');
                  break;
              case 'invalid':
                  $controlGroup.addClass('error');
                  break;
              case 'testing':
                  this._$el.addClass('loading');
                  break;
          }
      };

      /**
       * @method UrlView#_applyReadonly
       * @private
       * @param {Boolean} readonly
       */
      uvp._applyReadonly = function (readonly) {
          View.prototype._applyReadonly.call(this, readonly);

          this._$el.toggleClass('readonly', readonly);
      };

      return UrlView;
  }
);

define('oui/widgets/urlField',[
    'oui/base/input',
    'oui/widgets/url/urlModel',
    'oui/widgets/url/urlController',
    'oui/widgets/url/urlView'
], function (Input, UrlModel, UrlController, UrlView) {
      'use strict';

      /**
       *
       * @example
       *
       *     function (url, isValid) {
       *         $.get(url)
       *             .done(function () { isValid(true);  })
       *             .fail(function () { isValid(false); });
       *     }
       *
       * @callback UrlField#testFn
       * @param {String} url The url to validate.
       * @param {Function} isValid The function to use to validate the url. The callback function called with the result of
       * the validation.
       */

      /**
       * @typedef {Object} UrlField#Config
       *
       * @property {Boolean|Object} [autoEscape=true] Whether the combobox should escape/unescape the values from the
       * input. When the property is an Object it should have two functions escape and unescape that take in the string and
       * provide the escaped/unescaped version of the string. When the property is true it will default to using _.escape
       * and _.unescape for the strings.
       *
       * @property {Locale} [locale={}] The specific strings that are used by the widget and will use the defaults set from
       * the {@link module:oui.lang|lang}.
       *
       * @property {UrlField#testFn} testFn The function used to validate a URL.
       *
       * @property {Boolean} [browserValidation=false] Remove browser regular expression checking or URL syntax. **Note**:
       * This option is currently not supported.
       */

      //Used to hold the UrlField.prototype to allow
      //for easier additions of functions.
      var ufp;

      var __DEFAULTS__ = {
          autoEscape: true,
          locale: {},
          browserValidation: false
      };

      var __WIDGET_NAME__ = 'url';

      /**
       * The UrlField widget that allows the entry of URLs into a text box.
       * The user can then click a test button to make sure that the entered URL is valid and working.
       *
       * ##### **Relevant Events**
       *
       * * {@link UrlField#url-change|url-change}
       * * {@link UrlField#state-change|state-change}
       * * {@link UrlField#validation-state-change|validation-state-change}
       *
       * @borrows UrlModel#event:url-change as UrlField#url-change
       * @borrows UrlModel#event:state-change as UrlField#state-change
       * @borrows UrlModel#event:validation-state-change as UrlField#validation-state-change
       *
       * @param {UrlField#Config} options Configurable options for the UrlField widget
       * @param {HTMLInputElement|jQuery} element The default `<input type="text">` element
       *                 to wrap into a UrlField widget. If one is not provided, it will be created.
       * @param {String} initData The initial URL to display.
       *
       * @class UrlField
       * @since 1.3.0
       * @extends Input
       *
       */
      function UrlField(options, element, initData) {
          var model;

          //Super call
          Input.call(this, options, element, initData);

          model = new UrlModel(this._initData, this._options);
          this._controller = new UrlController(model, this._options);
          this._view = new UrlView(model, this._controller, this._element, this._options, this._settings);

          this._setupEvents(model);
      }

      //Start with the base Input as the prototype for the url field.
      UrlField.prototype = Object.create(Input.prototype);
      ufp = UrlField.prototype;

      ufp.constructor = UrlField;

      /**
       * Returns the name of the widget that is used in eventing and id generation
       *
       * @method UrlField#name
       * @returns {String} The widget name.
       */
      ufp.name = function () {
          return __WIDGET_NAME__;
      };

      /**
       * Gets or sets the currently entered URL from the text box.
       *
       * @method UrlField#url
       * @fires UrlModel#url-change
       * @param {String} [url] The new URL to set the widget to
       * @returns {String|undefined}
       */
      ufp.url = function (url) {
          if (!arguments.length) {
              return this._controller.url();
          }
          this._controller.url(url);
      };

      /**
       * Ability to retrieve the widgets default options for use in merging with passed in options as
       * well as with composed widgets such as the date and time picker.
       *
       * @method UrlField#defaults
       * @returns {UrlField#Config} - the widgets default options
       */
      ufp.defaults = function () {
          return __DEFAULTS__;
      };

      /**
       * @method UrlField#value
       *
       * @returns {String} url
       */
      /**
       * @method UrlField#value
       * @param {String} url
       */
      ufp.value = function (url) {
          if (!arguments.length) {
              return this._controller.url();
          }
          this._controller.url(url, { silent: true });
      };

      return UrlField;

  }
);

define('oui/widgets/iconPicker/iconModel',[
    'underscore',
    'oui/base/model'
], function (_, Model) {
    'use strict';

    /**
     * @event IconModel#icon-changed
     * @param {String} selectedIcon The newly selected color of the color picker.
     */
    /**
     * @event IconModel#color-changed
     * @param {String} selectedIcon The newly selected color of the color picker.
     */
    /**
     * @event IconModel#deselect
     */

    /**
     * This class represents the model for the icon picker. It stores the
     * available icons as well as the selected.
     *
     * @class IconModel
     * @private
     *
     * @param {IconPicker#Config} options The options for the icon picker.
     * @param {String} initData The icon to start the picker off at.
     */
    function IconModel(options, initData) {
        Model.call(this, initData, options);
    }

    IconModel.prototype = Object.create(Model.prototype);

    IconModel.prototype.constructor = IconModel;

    /**
     * @method IconModel#name
     *
     * @returns {String}
     */
    IconModel.prototype.name = function () {
        return 'icon';
    };

    /**
     * The getter/setter for the model.
     *
     * @method IconModel#icon
     * @param {String} [icon] The icon to set.
     * @returns {String} The selected icon.
     */
    IconModel.prototype.icon = IconModel.prototype.data;

    /**
     * @method IconModel#successHandler
     *
     * @param {Subscribable~EventOptions} eventParams
     * @param {String} icon
     */
    IconModel.prototype.successHandler = function (eventParams, icon) {
        Model.prototype.successHandler.call(this, eventParams, icon);

        if (icon == null) {
            this.trigger(_.extend(eventParams, { name: 'deselect' }));
        }
    };

    /**
     * Getter for the available icons in the model.
     *
     * @method IconModel#availableIcons
     * @returns {Array} The icons available to choose from.
     */
    IconModel.prototype.availableIcons = function () {
        if (!arguments.length) {
            return this._options.map;
        }
    };

    return IconModel;
});

define('oui/widgets/colorPicker/colorModel',['oui/widgets/iconPicker/iconModel'], function (Model) {
    'use strict';

    /**
     * @class ColorModel
     * @extends Model
     * @private
     *
     * @param {ColorPicker#Config} options
     * @param {String} initData
     */
    function ColorModel(options, initData) {
        Model.call(this, options, initData);
    }

    ColorModel.prototype = Object.create(Model.prototype);

    /**
     * @method ColorModel#name
     * @returns {String} 'color'
     */
    ColorModel.prototype.name = function () {
        return 'color';
    };

    return ColorModel;
});


define('template-47', ['underscore'], function (_) { return function(data) {
var __t, __p = '';
__p += '<button type="button" class="orcl-widget input orcl-icon-picker" tabindex="0" role="combobox" aria-haspopup="true" aria-expanded="false"> <div class="icon-box"> <i class="thumbnail pgbu-icon-small"> <span class="spoken-text"></span> </i> </div> <div class="btn btn-picker" title="' +
((__t = ( data.locale.icon.select )) == null ? '' : __t) +
'"> <i class="caret"></i> </div></button>';
return __p
};});


define('template-48', ['underscore'], function (_) { return function(data) {
var __t, __p = '';
__p += '<div role="listbox" id="' +
((__t = ( data.id )) == null ? '' : __t) +
'" class="palette"> ' +
((__t = ( data.markup  )) == null ? '' : __t) +
'</div>';
return __p
};});


define('template-49', ['underscore'], function (_) { return function(data) {
var __t, __p = '';
__p += '<div class="row ' +
((__t = ( data.cls ? data.cls : ''  )) == null ? '' : __t) +
'"> ' +
((__t = ( data.row  )) == null ? '' : __t) +
'</div>';
return __p
};});


define('template-50', ['underscore'], function (_) { return function(data) {
var __t, __p = '';
__p += '<button type="button" class="btn btn-blue-light btn-small clear">' +
((__t = ( data.locale.clear  )) == null ? '' : __t) +
'</button>';
return __p
};});


define('template-51', ['underscore'], function (_) { return function(data) {
var __t, __p = '';
__p += '<div role="option" id="' +
((__t = ( data.id )) == null ? '' : __t) +
'" class="tile" title="' +
((__t = ( data.locale[data.icon] )) == null ? '' : __t) +
'" data-icon="' +
((__t = ( data.icon  )) == null ? '' : __t) +
'"> <i class="icon-preview pgbu-icon-small ' +
((__t = ( data.cls  )) == null ? '' : __t) +
'"></i></div>';
return __p
};});

define('oui/widgets/iconPicker/iconView',['require','underscore','jquery','oui/base/view','template-47','template-48','template-49','template-50','template-51','oui/mixins/popup','oui/utils/math'],function (require) { //jshint maxstatements:false
    'use strict';

    var _ = require('underscore');
    var $ = require('jquery');
    var View = require('oui/base/view');
    var buttonTemplate = require('template-47');
    var paletteTemplate = require('template-48');
    var rowTemplate = require('template-49');
    var clearButton = require('template-50');
    var tileTemplate = require('template-51');
    var popupMixin = require('oui/mixins/popup');
    var math = require('oui/utils/math');

    var cvp;

    /**
     * @method IconView~iconToClass
     * @private
     *
     * @param {IconPicker#Config} options
     * @param {String} key
     *
     * @returns {String}
     */
    function iconToClass(options, key) {
        if (options.viewFns && options.viewFns.iconClass) {
            return options.viewFns.iconClass(key);
        } else {
            return key;
        }
    }

    /**
     * @method IconView~buildRow
     * @private
     *
     * @param {Function} template
     * @param {String} id
     * @param {IconPicker#Config} options
     * @param {Array<String>} colors
     * @param {Number} rowNum
     *
     * @returns {String}
     */
    function buildRow(template, id, options, colors, rowNum) {
        return rowTemplate({
            cls: '',
            row: _.reduce(colors, function (memo, color, index) {
                return memo + template({
                    id: id + '-' + rowNum + '-' + index,
                    icon: color,
                    cls: iconToClass(options, color),
                    locale: options.locale
                });
            }, '')
        });
    }

    /**
     * @method IconView~buildPalette
     *
     * @param {IconPicker#Config} options
     * @param {String} id
     * @param {Function} tileTemplate
     *
     * @returns {jQuery}
     */
    function buildPalette(options, id, tileTemplate) {
        var rows = _.groupBy(options.map, function (item, index) {
            return Math.floor((index / options.columns)) % options.rows;
        });
        var markup = _.reduce(rows, function (memo, row, index) {
            return memo + buildRow(tileTemplate, id, options, row, index);
        }, '');

        if (options.clearable) {
            markup += rowTemplate({
                cls: 'clear-row',
                row: clearButton({
                    locale: options.locale
                })
            });
        }

        return $(paletteTemplate({
            id: id + '-palette',
            markup: markup
        }));
    }

    /**
     * @class IconView
     * @private
     * @extends View
     * @mixes popup
     *
     * @param {IconModel} model
     * @param {IconController} controller
     * @param {jQuery|null} element
     * @param {IconPicker#Config} options
     * @param {Object} settings
     */
    function IconView(model, controller, element, options, settings) {
        var that = this;

        //Super call
        View.call(this, options, element, controller, settings);

        this._$el.find('.thumbnail').attr('id', this.id() + '-thumbnail');
        this._$el.attr('aria-describedby', this.id() + '-thumbnail');

        /**
         * The JQuery object that represents the pop-up palette
         *
         * @private
         * @type {Object}
         * @memberof! IconView.prototype
         */
        this._palette = this._$popup = buildPalette(this._options, this.id(), this.tileTemplate);

        this._colors = _.reduce(this._palette.find('[data-icon]'), function (memo, elem) {
            memo[elem.getAttribute('data-icon')] = $(elem);
            return memo;
        }, {});

        //Setting up the events that the view will listen to on the model
        model.on({
            'view.change': this.update.bind(this),
            'view.deselect': this.deselect.bind(this)
        });

        this.update(model.icon());

        //Binding events to the elements
        this._$el.on('mousedown touchstart', function (e) {
            if (e.type === 'touchstart') {
                e.preventDefault();
            }

            if (that.isEditable() && !$(this).hasClass('native')) {
                e.stopPropagation();

                // prevent it from hiding itself (see this._hide)
                if (that.isVisible()) {
                    that._hide();
                } else {
                    that._show();
                    e.preventDefault();
                }
            }
        })
      .on('blur', function () {
          that._hide();
      })
      .on('focus', function () {
          if (that._options.openOnFocus) {
              that._show();
          }
      });

        this.addShortcuts({
            'ESC': {
                fn: this._hide,
                stopPropagation: this.isVisible.bind(this)
            },
            'LEFT': this._switchActive.bind(this, -1, 0),
            'RIGHT': this._switchActive.bind(this, 1, 0),
            'UP': this._switchActive.bind(this, 0, -1),
            'DOWN': function () { //jscs:ignore jsDoc
                that._show();
                that._switchActive(0, 1);
            },
            'SPACE': this._show.bind(this),
            'ENTER': function () { //jscs:ignore jsDoc
                if (!this.isVisible()) {
                    that._show();
                } else {
                    var $activeTile = that._palette.find('.active');
                    if ($activeTile.length) {
                        that._controller.icon($activeTile.attr('data-icon'));
                    }
                    that._hide();
                }
            }
        });

        this._palette
            .on('mousedown', function (event) {
                event.preventDefault();
            })
            .on('click', '[data-icon]:not(.disabled), .clear', function () {
                that._controller.icon($(this).attr('data-icon') || null);
                that._hide();
            });
    }

    IconView.prototype = Object.create(View.prototype);
    cvp = IconView.prototype;

    cvp.constructor = IconView;

    _.extend(cvp, popupMixin);

    cvp.tileTemplate = tileTemplate;

    /**
     * @method IconView#destroy
     *
     * @param {Object} cleanUp
     */
    cvp.destroy = function (cleanUp) {
        this._palette.off();
        this._palette.remove();

        this._$el.off();
        if (cleanUp.self) {
            this._$el.remove();
        }
    };

    /**
     * @method IconView#update
     *
     * @param {String} color
     */
    cvp.update = function (color) {
        if (this._options.thumbnail) {
            this._$el.find('.thumbnail')
              .attr('class', 'thumbnail pgbu-icon-small ' + iconToClass(this._options, color))
              .attr('title', this._options.locale[color]);

            this._$el.find('.spoken-text:not(.instructions)').text(this._options.locale[color]);
        }
    };

    /**
     * @method IconView#buildElement
     *
     * @param {IconPicker#Config} options
     * @param {jQuery} $el
     *
     * @returns {jQuery}
     */
    cvp.buildElement = function (options, $el) {
        this._existingElement = false;

        // Whether we should toggle active on the element when clicked

        if (_.isNull($el)) {
            //create our own element
            $el = $(buttonTemplate({
                locale: options.locale
            }));
        } else if ($el.jquery) {
            //Got a jquery element
            if ($el.is('button')) {
                $el.addClass('orcl-icon-picker');
                this._existingElement = true;
            } else {

                var $element = $(buttonTemplate({
                    locale: options.locale
                })).closest('.orcl-icon-picker').insertAfter($el);

                $element.append($el);

                $el = $element;
            }
        }

        return $el;
    };

    /**
     * @method IconView#deselect
     */
    cvp.deselect = function () {
        if (this._options.thumbnail) {
            this._$el.find('.thumbnail').addClass('empty');
        }
    };

    /**
     * @method IconView#_show
     * @private
     */
    cvp._show = function () {
        if (!this.isVisible()) {

            this.showPopup(this._$el, 'right');
            this._$el.attr('aria-owns', this._palette.attr('id'));
            this._$el.attr('aria-expanded', true);
            this._$el.focus();

            this._$el.toggleClass('active', this._existingElement);
        }
    };

    //TODO: remove references to this function and instead use hidePopup
    /**
     * @method IconView#_hide
     * @private
     */
    cvp._hide = function () {
        this.hidePopup();
    };

    /**
     * @method IconView#hidePopup
     */
    cvp.hidePopup = function () {
        popupMixin.hidePopup.apply(this, arguments);
        this._$el.attr('aria-expanded', false);
        this._$el.removeClass('active');
        this._palette.find('.tile').removeClass('active');
        this._$el.removeAttr('aria-activedescendant');
    };

    /**
     * @method IconView#_switchActive
     * @private
     *
     * @param {Number} dx
     * @param {Number} dy
     */
    cvp._switchActive = function (dx, dy) {
        var $tiles = this._palette.find('.tile:not(.disabled)');
        var $current = $tiles.filter('.active');

        $current.removeClass('active');

        if ($current.length === 0 && (dy > 0 || dx > 0)) {
            $tiles.first().addClass('active');
        } else if ($current.length === 0 && (dy < 0 || dx < 0)) {
            $tiles.last().addClass('active');
        } else if ($current.length > 0) {
            var currentIndex = $current.index();
            var rows = this._palette.find('.row');
            var currentRowIndex = rows.index($current.closest('.row'));
            var tiles = $current.siblings('.tile').add($current);

            var nextRowIndex = math.mod(currentRowIndex + dy, rows.length);
            var nextColumnIndex = math.mod(currentIndex + dx, tiles.length);

            var tempNext = rows.eq(nextRowIndex).find('.tile').eq(nextColumnIndex);

            while (tempNext.hasClass('disabled')) {
                if (dx !== 0) {
                    nextColumnIndex = math.mod(nextColumnIndex + dx, tiles.length);
                }

                if (dy !== 0) {
                    nextRowIndex = math.mod(nextRowIndex + dy, rows.length);
                }

                tempNext = rows.eq(nextRowIndex).find('.tile').eq(nextColumnIndex);
            }

            rows.eq(nextRowIndex).find('.tile').eq(nextColumnIndex).addClass('active');
        }

        this._$el.attr('aria-activedescendant', this._palette.find('.active').attr('id'));

    };

    /**
     * Returns the element that identifies the widget.
     *
     * @method IconView#getIdentifyingElement
     *
     * @returns {jQuery}
     */
    cvp.getIdentifyingElement = function () {
        return this._$el;
    };

    /**
     * @method ColorView#disabled
     *
     * @param {Array<String>} colors
     */
    cvp.disabled = function (colors) {
        this._palette.find('.tile.disabled')
          .removeClass('disabled')
          .removeAttr('aria-disabled');

        _.each(colors, function (color) {
            if (this._colors[color]) {
                this._colors[color]
                  .addClass('disabled')
                  .attr('aria-disabled', true);
            }
        }.bind(this));
    };

    /**
     * @method ColorView#_applyLocked
     * @private
     *
     * @param {Boolean} locked
     */
    cvp._applyLocked = function (locked) {
        var toDisable = this._$el.add(this.$('input, button, .btn'));

        View.prototype._applyLocked.call(this, locked);

        toDisable.prop('disabled', locked)
            .toggleClass('disabled', locked);
    };

    /**
     * @method ColorView#_applyDisabled
     * @private
     *
     * @param {Boolean} disabled
     */
    cvp._applyDisabled = function (disabled) {
        var toDisable = this._$el.add(this.$('input, button, .btn'));

        toDisable.toggleClass('disabled', disabled);
        toDisable.prop('disabled', disabled);

        if (disabled) {
            this._$el.removeAttr('tabindex');
        } else {
            this._$el.attr('tabindex', 0);
        }

    };

    /**
     * @method ColorView#_applyReadonly
     * @private
     *
     * @param {Boolean} readonly
     */
    cvp._applyReadonly = function (readonly) {
        this._$el.toggleClass('readonly', readonly);
        this._$el.add(this.$('input, button, .btn')).prop('readonly', readonly)
            .toggleClass('readonly-as-text', readonly)
            .prop('disabled', readonly);
    };

    return IconView;
}
);


define('template-52', ['underscore'], function (_) { return function(data) {
var __t, __p = '';
__p += '<button type="button" class="orcl-widget input orcl-color-picker" tabindex="0" role="combobox" aria-haspopup="true" aria-expanded="false"> <div class="color-box"> <i class="thumbnail"> <span class="spoken-text"></span> </i> </div> <div class="btn btn-picker" title="' +
((__t = ( data.locale.color.select  )) == null ? '' : __t) +
'"> <i class="pgbu-icon-small pgbu-icon-paint-bucket"></i> </div></button>';
return __p
};});


define('template-53', ['underscore'], function (_) { return function(data) {
var __t, __p = '';
__p += '<div role="option" id="' +
((__t = ( data.id )) == null ? '' : __t) +
'" class="tile" title="' +
((__t = ( data.locale.color[data.cls.toUpperCase()] )) == null ? '' : __t) +
'" data-icon="#' +
((__t = ( data.cls  )) == null ? '' : __t) +
'" > <div class="color-preview" style="background-color:#' +
((__t = ( data.cls  )) == null ? '' : __t) +
'"></div></div>';
return __p
};});

define('oui/widgets/colorPicker/colorView',['require','underscore','jquery','oui/widgets/iconPicker/iconView','template-52','template-53'],function (require) {
    'use strict';

    var _ = require('underscore');
    var $ = require('jquery');
    var View = require('oui/widgets/iconPicker/iconView');
    var buttonTemplate = require('template-52');
    var tileTemplate = require('template-53');

    var cvp;

    /**
     * @method ColorView~move
     *
     * @param {String} attr
     * @param {jQuery} source
     * @param {jQuery} dest
     */
    function move(attr, source, dest) {
        var val = source.attr(attr);

        if (val) {
            source.removeAttr(attr);
            source.attr('data-orig-' + attr, val);
        }

        dest.attr(attr, val);
    }

    /**
     * @class ColorView
     * @private
     * @extends IconView
     *
     * Does not use the parameters. Forwards them with an apply call.
     *
     * @param {IconModel} [model]
     * @param {IconController} [controller]
     * @param {jQuery|null} [element]
     * @param {ColorPicker#Config} [options]
     * @param {Object} [settings]
     */
    function ColorView(model) {
        View.apply(this, arguments);

        model.on('view.change', function (color) {
            if (this._options.changeFontColor) {
                this._$el.css('color', color);
            }
        }.bind(this));

        this._$thumbnail = this._$el.find('.thumbnail');

        this.update(this._controller.icon());
    }

    ColorView.prototype = Object.create(View.prototype);
    cvp = ColorView.prototype;

    /**
     * lodash/underscore template representing a tile
     * @type {Function}
     * @member ColorView.prototype.tileTemplate
     */
    cvp.tileTemplate = tileTemplate;

    /**
     * @method ColorView#nativeType
     * @returns {?String} 'color' The native type to create
     */
    cvp.nativeType = function () {
        return null;
    };

    /**
     * @method ColorView#buildElement
     * @param {ColorPicker#Config} options
     * @param {jQuery} $el
     */
    cvp.buildElement = function (options, $el) {
        this._existingElement = false;

        // Whether we should toggle active on the element when clicked

        if (_.isNull($el)) {
            //create our own element
            $el = $(buttonTemplate({
                locale: options.locale
            }));
        } else {
            //Got a jquery element
            if ($el.is('button')) {
                $el.addClass('orcl-color-picker');
                this._existingElement = true;
            } else {

                var $element = $(buttonTemplate({
                    locale: options.locale
                })).closest('.orcl-color-picker').insertAfter($el);

                move('id', $el, $element);
                move('aria-label', $el, $element);

                $element.append($el);

                $el = $element;
            }
        }

        return $el;
    };

    /**
     * @method ColorView#buildNativeElement
     * @param {ColorPicker#Config} options
     * @param {jQuery} $el
     */
    cvp.buildNativeElement = _.constant(null);

    /**
     * @method ColorView#deselect
     */
    cvp.deselect = function () {
        View.prototype.deselect.call(this);

        this._$el.find('.thumbnail').attr('title', this._options.locale.color.empty);
    };

    /**
     * @method ColorView#update
     * @param {String} color
     */
    cvp.update = function (color) {
        if (this._$thumbnail && this._$thumbnail.length > 0 && this._options.thumbnail && color) {
            var title = this._options.locale.color[color.substring(1).toUpperCase()];

            this._$el.find('.thumbnail')
              .removeClass('empty')
              .attr('title', title)[0].style.setProperty('background-color', color, 'important');

            this._$el.find('.spoken-text:not(.instructions)').text(title);

        }

        var input = this.$('input');
        if (input.length > 0) {
            input[0].value = color;
        }
    };

    return ColorView;
});

define('oui/widgets/iconPicker/iconController',[],function () {
    'use strict';

    var ccp;

    /**
     * This class represents the controller for the icon picker.
     *
     * @class IconController
     * @private
     *
     * @param {IconModel} model The model that the controller will call into
     *    to get/set data.
     * @param {IconPicker#Config} options The options object used for the widget.
     */
    function IconController(model, options) {

        /**
         * The model used in the controller.
         *
         * @private
         * @type {IconModel}
         * @memberof! IconController.prototype
         */
        this._model = model;
        /**
         * The options for the widget.
         *
         * @property _options
         * @private
         * @memberof! IconController.prototype
         */
        this._options = options;
    }

    ccp = IconController.prototype;
    ccp.constructor = IconController;
    /**
     * Getter for the icon of the model.
     *
     * @method IconController#icon
     * @returns {String} The selected icon of the model.
     */
    /**
     * Setter for the icon of the model.
     *
     * @method IconController#icon
     * @param {String} icon The icon to set on the model.
     * @param {Subscribable~EventOptions} eventParams
     */
    ccp.icon = function (icon, eventParams) {
        if (!arguments.length) {
            return this._model.icon();
        }

        this._model.icon(icon, eventParams);
    };

    return IconController;
});

define('oui/widgets/colorPicker',['require','oui/base/widget','oui/widgets/colorPicker/colorModel','oui/widgets/colorPicker/colorView','oui/widgets/iconPicker/iconController'],function (require) {
      'use strict';

      var Widget = require('oui/base/widget');
      var Model = require('oui/widgets/colorPicker/colorModel');
      var View = require('oui/widgets/colorPicker/colorView');
      var Controller = require('oui/widgets/iconPicker/iconController');

      /**
       * @typedef {Object} ColorPicker#Config
       *
       * @property {Boolean} [openOnFocus=false] Whether or not the color picker should open the popup when focused.
       * @property {Array<String>} [map] Map of colors (hex values) to use for the palette.
       * @property {Boolean} [fluid=false] Whether or not the component will take all available horizontal space of the
       * parent.
       * @property {String|jQuery} [appendTo='body'] The selector to append the palette to.
       * @property {Boolean} [thumbnail=true] Whether or not to display a preview/thumbnail of the selected color.
       * @property {Number} [rows=4] Specify how many rows the color picker should have. The color picker will not generate
       * empty rows.
       * @property {Number} [columns=11] Specify how many columns the color picker should have. The color picker uses this
       * property as the primary objected to match.
       * @property {Locale} [locale={}] The specific strings that are used by the widget and will use the defaults set from
       * the {@link module:oui.lang|lang}.
       * @property {Boolean} [clearable=false] Whether the color picker can accept a null value. When true it renders the
       * palette with a clear button to allow a user to remove their selection.
       * @property {Boolean} [changeFontColor=false] Whether or not the color picker should change its parent's elements font
       * color to match the selected color
       */

      var __DEFAULTS__ = {
          openOnFocus: false,
          map: [
            'ffffff', 'dbdce0', 'c1cbd5', '7fcdee', '99ddd8', 'cbe5cb', 'fff59c', 'fcd580', 'f5c2c2', 'eacee7', 'cbc0db',
            'cccccc', 'b6bac1', '8b9ab2', '66aad7', '66c4be', 'afdf8e', 'ffef5e', 'faab01', 'ef9a9a', 'd59ecf', '9881b7',
            '666666', '616977', '3e577f', '0072bc', '009c92', '65b265', 'ffdd00', 'e3870a', 'd80202', 'ab3da0', '6b40ad',
            '000000', '495363', '0e2d5f', '0f527d', '00857c', '008000', 'f2c019', 'be5709', 'a20202', '710966', '503082'
          ],
          fluid: false,
          appendTo: 'body',
          thumbnail: true,
          rows: 4,
          columns: 11,
          locale: {},
          clearable: false,
          changeFontColor: false
      };

      /**
       * This class creates a customizeable swatch of colors.
       *
       * See {@link ColorPicker#Config} for details regarding the options for the Color Picker
       *
       *  ##### **Keyboard Shortcuts**
       *
       *  * <kbd>ESC</kbd> - Closes an open color picker
       *  * <kbd>Left</kbd> - Moves focus to the color tile to the left
       *  * <kbd>Right</kbd> - Moves focus to the color tile to the right
       *  * <kbd>Up</kbd> - Moves focus to the color tile to the up
       *  * <kbd>Down</kbd> - Moves focus to the color tile to the down
       *  * <kbd>Space</kbd> - Shows the color picker if it is currently hidden
       *  * <kbd>Enter</kbd> - Shows the color picker if it is currently hidden or selects the currently active color tile if it is visible.
       *
       *
       *  ##### **Relevant Events**
       *
       *  * {@link ColorPicker#color-changed|color-changed}
       *  * {@link ColorPicker#deselect|deselect}
       *
       *  @borrows IconModel#event:deselect as ColorPicker#deselect
       *  @borrows IconModel#event:color-changed as ColorPicker#color-changed
       *
       * @example
       * <caption>Creating a Color Picker (accepts all defaults)</caption>
       *
       * var color = new oui.ColorPicker();
       * //Assumes parent container element is called $container
       * $container.append(color.element());
       * color.init();
       *
       * @example
       * <caption>Creating a Color Picker with existing element</caption>
       *
       * //Assumes element being used is already a jQuery object called $colorEl
       * var color = new oui.ColorPicker($colorEl);
       * color.init();
       *
       * @class ColorPicker
       * @extends Widget
       * @param {ColorPicker#Config} [options] The options object to use throughout the widget.
       * @param {HTMLInputElement|jQuery} [element] JQuery element to use as color picker on the DOM.
       * @param {String} [initColor='#FFFFFF'] The color to initialize the color picker with.
       */
      function ColorPicker(options, element, initColor) {
          var model;

          //Super call
          Widget.call(this, options, element, initColor);

          //Set up default _initData for the ColorPicker
          /**
           * The initial data used by the widget
           *
           * @memberof! ColorPicker.prototype
           * @type {String}
           * @private
           */
          this._initData = this._initData || '#FFFFFF';

          model = new Model(this._options, this._initData);
          /**
           * The controller object for the widget.
           *
           * @memberof! ColorPicker.prototype
           * @private
           * @type {ColorController}
           */
          this._controller = new Controller(model, this._options);
          /**
           * The view for the widget.
           *
           * @memberof! ColorPicker.prototype
           * @private
           * @type {ColorView}
           */
          this._view = new View(model, this._controller, this._element, this._options, this._settings);

          Object.defineProperty(this, '_data', {
              get: this.color,
              set: this.color
          });

          this._setupEvents(model);
      }

      ColorPicker.prototype = Object.create(Widget.prototype);
      var cpp = ColorPicker.prototype;

      cpp.constructor = ColorPicker;

      /**
       * Returns the name of the widget that is used in eventing and id generation
       *
       * @method ColorPicker#name
       * @returns {String} 'color' - the widgets name
       */
      cpp.name = function () {
          return 'color';
      };

      /**
       * Getter and setter for the selected color of the picker.
       *
       * @example
       *     var icon;
       *
       *     //Initialized
       *
       *     //Setter
       *     icon.color('000000');
       *
       *     //Getter
       *     icon.color(); //Returns '000000'
       *
       * @method ColorPicker#color
       * @fires IconModel#color-changed
       * @fires IconModel#deselect
       * @param {String} [color] The hex value to set.
       * @returns {String} The selected color.
       */
      cpp.color = function (color) {
          if (!arguments.length) {
              return this._controller.icon();
          }

          this._controller.icon(color);
      };

      /**
       * Getter and setter for the selected color of the picker.
       *
       * See {@link ColorPicker#color} for information regarding the similar method.
       *
       * **Note**: does not trigger an event notifying of a change.
       *
       * @method ColorPicker#value
       * @param {String} [color] The hex value to set.
       * @returns {String} The selected color.
       */
      cpp.value = function (color) {
          if (!arguments.length) {
              return this._controller.icon();
          }

          this._controller.icon(color, {
              silent: true
          });
      };

      /**
       * Programmaticaly open the color picker select popup.
       *
       * @method ColorPicker#show
       */
      cpp.show = function () {
          this._view._show();
      };

      /**
       * Programmatically close the color picker select popup.
       *
       * @method ColorPicker#hide
       */
      cpp.hide = function () {
          this._view._hide();
      };

      /**
       * Ability to retrieve the widgets default options for use in merging with passed in options as
       * well as with composed widgets such as the date and time picker.
       *
       * @method ColorPicker#defaults
       * @returns {ColorPicker#Config} - the widgets default options
       */
      cpp.defaults = function () {
          return __DEFAULTS__;
      };

      /**
       * Getter or setter for the disable state of the color picker
       *
       * @method ColorPicker#disable
       * @param {Boolean} [disabled] Whether to disable the view or now. True to disable and false to enable
       * @returns {Boolean} - Whether the view is currently disabled
       */
      cpp.disable = function (disabled) {
          if (!arguments.length) {
              return this._view.isDisabled();
          }

          this._view.disable(disabled);
      };

      /**
       * Setter for the disabled colors of the picker.
       *
       * @method ColorPicker#disabled
       * @param {Array<String>} colors The colors to disable
       */
      cpp.disabled = function (colors) {
          this._view.disabled(colors);
      };

      return ColorPicker;
  }
);

define('oui/widgets/iconPicker',[
    'oui/base/widget',
    'oui/widgets/iconPicker/iconModel',
    'oui/widgets/iconPicker/iconView',
    'oui/widgets/iconPicker/iconController'
], function (Widget, Model, View, Controller) {
      'use strict';

      /**
       * Function to get the CSS class representation of the icon.
       *
       * @callback IconPicker#iconClass
       * @param {String} key The key to return the CSS class for
       * @returns {String} - The css to use for the icon.
       */

      /**
       * @typedef {Object} IconPicker#Config
       *
       * @property {Boolean} [openOnFocus=false] Whether or not the color picker should open the popup when focused.
       * @property {Array<String>} [map] Map of colors (hex values) to use for the palette.
       * @property {Boolean} [fluid=false] Whether or not the component will take all available horizontal space of the
       * parent.
       * @property {String|jQuery} [appendTo='body'] The selector to append the palette to.
       * @property {Boolean} [thumbnail=true] Whether or not to display a preview/thumbnail of the selected color.
       * @property {Number} [rows=5] Specify how many rows the color picker should have. The color picker will not generate
       * empty rows.
       * @property {Number} [columns=8] Specify how many columns the color picker should have. The color picker uses this
       * property as the primary objected to match.
       * @property {Locale} [locale={}] The specific strings that are used by the widget and will use the defaults set from
       * the {@link module:oui.lang|lang}.
       * @property {Boolean} [clearable=false] Whether the color picker can accept a null value. When true it renders the
       * palette with a clear button to allow a user to remove their selection.
       * @property {Object} viewFns
       * @property {IconPicker#iconClass} viewFns.iconClass
       */

      var __WIDGET_NAME__ = 'icon';
      var __DEFAULTS__ = {
          openOnFocus: false,
          map: [],
          fluid: false,
          appendTo: 'body',
          thumbnail: true,
          locale: {},
          clearable: false
      };
      var ipp;

      /**
       * This class creates a palette of icons to choose from.
       *
       * There is no default map option so all the icons must be given.
       *
       * See {@link IconPicker#Config} for more information about options allowed by Icon Picker.
       *
       *  ##### **Keyboard Shortcuts**
       *
       *  * <kbd>ESC</kbd> - Closes an open color picker
       *  * <kbd>Left</kbd> - Moves focus to the color tile to the left
       *  * <kbd>Right</kbd> - Moves focus to the color tile to the right
       *  * <kbd>Up</kbd> - Moves focus to the color tile to the up
       *  * <kbd>Down</kbd> - Moves focus to the color tile to the down
       *  * <kbd>Space</kbd> - Shows the color picker if it is currently hidden
       *  * <kbd>Enter</kbd> - Shows the color picker if it is currently hidden or selects the currently active color tile if it is visible.
       *
       *
       *  ##### **Relevant Events**
       *
       *  * {@link IconPicker#icon-changed|icon-changed}
       *  * {@link IconPicker#deselect|deselect}
       *
       *  @borrows IconModel#event:deselect as IconPicker#deselect
       *  @borrows IconModel#event:icon-changed as IconPicker#icon-changed
       *
       * @class IconPicker
       * @extends Widget
       *
       * @param {IconPicker#Config} [options] The options object to use throughout the widget.
       * @param {Object} [element] JQuery element to use as picker on the DOM.
       * @param {String} [initIcon] The icon to initialize the picker with.
       */
      function IconPicker(options, element, initIcon) {
          var model;

          //Super call
          Widget.call(this, options, element, initIcon);

          model = new Model(this._options, this._initData);
          /**
           * The controller object for the widget.
           *
           * @private
           * @type {IconController}
           * @memberof! IconPicker.prototype
           */
          this._controller = new Controller(model, this._options);
          /**
           * The view for the widget.
           *
           * @private
           * @type {IconView}
           * @memberof! IconPicker.prototype
           */
          this._view = new View(model, this._controller, this._element, this._options, this._settings);

          Object.defineProperty(this, '_data', {
              get: this.icon,
              set: this.icon
          });

          this._setupEvents(model);
      }

      IconPicker.prototype = Object.create(Widget.prototype);
      ipp = IconPicker.prototype;

      ipp.constructor = IconPicker;

      /**
       * Returns the name of the widget that is used in eventing and id generation
       *
       * @method IconPicker#name
       * @returns {String} - the widgets name
       */
      ipp.name = function () {
          return __WIDGET_NAME__;
      };

      /**
       * Ability to retrieve the widgets default options for use in merging with passed in options as
       * well as with composed widgets such as the date and time picker.
       *
       * @method IconPicker#defaults
       * @returns {IconPicker#Config} - the widgets default options
       */
      ipp.defaults = function () {
          return __DEFAULTS__;
      };

      /**
       * Combined getter/setter for the disabled state of the icon picker.
       *
       * If an argument is supplied it will attempt to set the disabled status of the component.
       * If there is no argument the current status will be returned.
       *
       * @method IconPicker#disable
       * @param {Boolean} [disabled] whether or not the component should be disabled
       * @returns {Array} - disabled state of the icon picker.
       */
      ipp.disable = function (disabled) {
          if (!arguments.length) {
              return this._view.disable();
          }

          this._view.disable(disabled);
      };

      /**
       * Programmaticaly open the color picker select popup.
       *
       * @method IconPicker#show
       */
      ipp.show = function () {
          this._view._show();
      };

      /**
       * Programmatically close the color picker select popup.
       *
       * @method IconPicker#hide
       */
      ipp.hide = function () {
          this._view._hide();
      };

      /**
       * Getter and setter for the selected icon of the picker.
       *
       * @example
       *     var icon;
       *
       *     //Initialized
       *
       *     //Setter
       *     icon.icon(key);
       *
       *     //Getter
       *     icon.icon(); //Returns key
       *
       * @method IconPicker#color
       * @fires IconModel#icon-changed
       * @fires IconModel#deselect
       * @param {String} [newIcon] The hex value to set.
       * @returns {String} The selected color.
       */
      ipp.icon = function (newIcon) {
          if (!arguments.length) {
              return this._controller.icon();
          }

          this._controller.icon(newIcon);
      };

      /**
       * Getter and setter for the selected icon of the picker.
       *
       * See {@link IconPicker#icon} for information regarding the similar method.
       *
       * **Note**: does not trigger an event notifying of a change.
       *
       * @method IconPicker#value
       * @param {String} [newIcon] The icon value to set.
       * @returns {String} The selected icon.
       */
      ipp.value = function (newIcon) {
          if (!arguments.length) {
              return this._controller.icon();
          }

          this._controller.icon(newIcon, {
              silent: true
          });
      };

      /**
       * Setter for the disabled icons of the picker.
       *
       * @method IconPicker#disabled
       * @param {Array<String>} icons The icons to disable
       */
      ipp.disabled = function (icons) {
          this._view.disabled(icons);
      };

      return IconPicker;
  }
);


define('template-54', ['underscore'], function (_) { return function(data) {
var __t, __p = '';
__p += '<td data-event-id="' +
((__t = ( data.id )) == null ? '' : __t) +
'" data-group-id="' +
((__t = ( data.groupId )) == null ? '' : __t) +
'" data-doy="' +
((__t = ( data.doy )) == null ? '' : __t) +
'" class="task oui-draggable ' +
((__t = ( data.classes.join(' ') )) == null ? '' : __t) +
'" colspan="' +
((__t = ( data.duration )) == null ? '' : __t) +
'"> <div class="resize-handle resize-handle-left" data-resize="startDate"></div> ' +
((__t = ( data.event )) == null ? '' : __t) +
' <div class="resize-handle resize-handle-right" data-resize="endDate"></div></td>';
return __p
};});


define('template-55', ['underscore'], function (_) { return function(data) {
var __t, __p = '';
__p += '<td data-group-id="' +
((__t = ( data.id )) == null ? '' : __t) +
'" data-doy="' +
((__t = ( data.doy )) == null ? '' : __t) +
'" class="group" colspan="' +
((__t = ( data.duration )) == null ? '' : __t) +
'"> ' +
((__t = ( data.event )) == null ? '' : __t) +
'</td>';
return __p
};});

define('oui/widgets/calendar/event',[
  'underscore',
  'moment',
  'oui/utils/uniqueId'
], function (_, moment, uniqueId) {
    'use strict';

    /**
     * @class Event
     * @private
     *
     * @param {Object} data
     * @param {Object} keys
     */
    function Event(data, keys) {
        this._id = _.get(data, 'id', uniqueId('event'));

        this._data = data;
        this._keys = _.extend({
            startDate: 'startDate',
            endDate: 'endDate'
        }, keys);

        //TODO: where do we get the following data from?
        //config callback, config recordKeys, config mapper function
        this._recurring = false;

        this._schedule = {};
        this._dates = [];
        this._minDate = null;
        this._maxDate = null;

        this._startDate = data[this._keys.startDate];
        this._endDate = data[this._keys.endDate];
    }

    /**
     * @method isEvent
     * @memberof! Event
     * @static
     *
     * @param {Event} evToCheck
     *
     * @returns {Boolean}
     */
    Event.isEvent = function (evToCheck) {
        return evToCheck._ouiEvent === true;
    };

    var ep = Event.prototype = {
        _ouiEvent: true,

        /**
         * @method Event#id
         *
         * @returns {String} id
         */
        get id() {
            return this._id;
        },

        /**
         * @method Event#data
         *
         * @param {Object} newData
         */
        set data(newData) {
            if (!_.isEqual(newData, this._data)) {
                this._data = newData;
            }
        },

        /**
         * @method Event#data
         *
         * @returns {Object} data
         */
        get data() {
            return this._data;
        },

        /**
         * @method Event#startDate
         *
         * @param {moment} newSd
         */
        set startDate(newSd) {
            this._startDate = newSd;
        },

        /**
         * @method Event#startDate
         *
         * @returns {moment} start date
         */
        get startDate() {
            return this._startDate;
        },

        /**
         * @method Event#endDate
         *
         * @returns {moment} end date
         */
        get endDate() {
            return this._endDate;
        },

        /**
         * @method Event#endDate
         *
         * @param {moment} newEd
         */
        set endDate(newEd) {
            this._endDate = newEd;
        }

        //TODO: fill in other keys?
    };

    /**
     * @method Event#next
     */
    ep.next = function () {};

    /**
     * @method Event#isOccurring
     *
     * @param {moment} startDate
     * @param {moment} endDate
     *
     * @returns {Boolean}
     */
    ep.isOccurring = function (startDate, endDate) {

        if (startDate != null && endDate != null) {
            return +this.startDate < +endDate &&
              +this.endDate > +startDate;
        }

        if (startDate != null) {
            return this._endDate.isAfter(startDate) &&
              this._startDate.isBefore(startDate);
        }

        if (endDate != null) {
            return this._startDate.isBefore(endDate) &&
              this._endDate.isAfter(endDate);
        }

        return true;
    };

    /**
     * @method Event#set
     *
     * @param {*} value
     * @param {String} key
     */
    ep.set = function (value, key) {
        this[key] = value;
    };

    return Event;
});


define('template-56', ['underscore'], function (_) { return function(data) {
var __t, __p = '', __j = Array.prototype.join;
function print() { __p += __j.call(arguments, '') }

 if (data.events.length > 0) { ;
__p += '<tr data-group-id="' +
((__t = (  data.groupId )) == null ? '' : __t) +
'" class="' +
((__t = ( data.classes && data.classes.join(' ') )) == null ? '' : __t) +
'">' +
((__t = ( data.events.join('') )) == null ? '' : __t) +
'</tr>';
 } else { ;
__p += '<tr></tr>';
 } ;

return __p
};});


define('template-57', ['underscore'], function (_) { return function(data) {
var __t, __p = '';
__p += '<td data-doy="' +
((__t = ( data.doy )) == null ? '' : __t) +
'"></td>';
return __p
};});

define('oui/utils/tasks/tasks',['require','underscore','template-54','template-55','oui/widgets/calendar/event','template-56','template-57'],function (require) {
    'use strict';

    var _ = require('underscore');
    var taskTemplate = require('template-54');
    var taskGroupTemplate = require('template-55');
    var Event = require('oui/widgets/calendar/event');
    var rowTemplate = require('template-56');
    var emptyCell = require('template-57');

    var DOY_FORMAT = 'YYYY-MM-DD';
    var THRESHOLD = 1; // 1 ms
    var TO_MS = 864e5; // 1 day in ms (1000 * 60 * 60 * 24)

    /**
     * @typedef {Object} tasks#options
     * @property {Function} getStartDate
     * @property {Function} getVisibleDays
     * @property {Function} getMomentAttribute
     */

    /**
     * @method getDuration
     * @param {moment} end
     * @param {moment} start
     * @returns {Number}
     */
    function getDuration(end, start) {
        var difference = end.diff(start, 'day', true);
        var roundedDifference = Math.round(difference);
        if (Math.abs(roundedDifference * TO_MS - difference * TO_MS) <= THRESHOLD) {
            return roundedDifference;
        }
        return difference;
    }

    /**
     * @method prepareEvents
     * @param {Array} events
     * @param {moment} start
     */
    function prepareEvents(events, start) {
        return _.chain(events)
            .sortBy(function (task) {
                return +task.modified || 0;
            })
            .sortBy(function (event) {
                return -getDuration(event.endDate, event.startDate);
            })
            .groupBy(function (event) {
                if (start && event.startDate.isBefore(start)) {
                    return start.day();
                }
                return event.startDate.day();
            })
            .value();
    }

    /**
     * @method tasks#groupByDay
     * @param {Array} events
     * @param {moment} startDate
     * @returns {Object}
     */
    function groupByDay(events, startDate) {
        return _.groupBy(events, function (event) {
            if (startDate && event.startDate.isBefore(startDate)) {
                return startDate.day();
            }
            return event.startDate.day();
        });
    }

    /**
     * @method findEventRange
     * @param {Array} events
     */
    function findEventRange(events) {
        var earliest = null;
        var latest = null;
        _.each(events, function (e) {
            if (earliest === null || e.startDate.isBefore(earliest)) {
                earliest = e.startDate.clone();
            }
            if (latest === null || e.endDate.isAfter(latest)) {
                latest = e.endDate.clone();
            }
        });
        return {
            start: earliest,
            end: latest
        };
    }

    /**
     * @method taskUtil
     * @param {tasks#options} options
     */
    return function (options) {
        /**
         * @method getViewRange
         */
        function getViewRange() {
            return {
                start: options.getStartDate().clone(),
                end: options.getStartDate().clone().endOf(options.getMomentAttribute())
            };
        }

        /**
         * @method isOccurring
         * @param {Calendar#Event} event
         */
        function isOccurring(event) {
            var viewRange = getViewRange();
            return event.isOccurring(viewRange.start, viewRange.end);
        }

        /**
         * @method createGroup
         * @param {String} groupId
         * @param {moment} start
         * @param {moment} end
         */
        function createGroup(groupId, start, end) {
            return new Event(_.chain({})
                .set(options.eventKeys.parent.id, groupId)
                .set(options.eventKeys.parent.startDate, start)
                .set(options.eventKeys.parent.endDate, end)
                .value(), options.eventKeys.parent);
        }

        /**
         * @method getRenderOptions
         * @param {Calendar#Event} item
         */
        function getRenderOptions(item) {
            return {
                outOfView: item.outOfView,
                viewRange: getViewRange(),
                locale: options.locale
            };
        }

        /**
         * @method getRenderItem
         * @param {Calendar#Event} item
         */
        function getRenderItem(item) {
            if (item.outOfView && item.originalGroup) {
                return item.originalGroup.data;
            } else {
                return item.data;
            }
        }

        /**
         * @method getVisibleDuration
         * @param {moment} end
         * @param {moment} start
         */
        function getVisibleDuration(end, start) {
            var viewRange = getViewRange();
            var visibleStart = start.clone();
            var visibleEnd = end.clone();
            if (start.isBefore(viewRange.start)) {
                visibleStart = viewRange.start;
            }
            if (end.isAfter(viewRange.end)) {
                visibleEnd = viewRange.end;
            }
            return getDuration(visibleEnd, visibleStart);
        }

        /**
         * @method sortByDay
         * @param {Array} events
         */
        function sortByDay(events) {
            return _.sortBy(events, function (event) {
                return event.startDate;
            });
        }

        /**
         * @method getTasksByGroup
         * @param {Array} events
         */
        function getTasksByGroup(events) {
            return _.groupBy(events, function (task) {
                return _.get(task.data, options.eventKeys.parentKey);
            });
        }

        /**
         * @method getTable
         * @returns {jQuery}
         */
        function getTable() {
            return options.getTable();
        }

        /**
         * @method removeAllRows
         */
        function removeAllRows() {
            /* istanbul ignore next: just testing jquery find and remove */
            getTable()[0].innerHTML = '';
        }

        /**
         * @method toggleGroup
         * @param {Number|String} groupId
         * @param {Boolean} toggle
         */
        function toggleGroup(groupId, toggle) {
            var $table = getTable();
            var $group = $table.find('td[data-group-id=' + groupId + ']');
            var $taskRows = $table.find('tr[data-group-id=' + groupId + ']:not(.task-group-row)');
            $group.find('.selectable-group').toggleClass('collapsed', toggle);
            $taskRows.toggleClass('hidden', toggle);
        }

        /**
         * @method buildRow
         * @param {Object} itemsByDay
         * @param {Function} itemTemplate
         * @param {Object} args
         */
        function buildRow(itemsByDay, itemTemplate, args) {
            var isEditable = options.mediator && _.isFunction(options.mediator.canEdit) ? options.mediator.canEdit.bind(options.mediator) : _.constant(true);
            var startDate = options.getStartDate();
            var daysRemaining = options.getVisibleDays();
            var curDay = startDate.day();
            var days = [];
            var doy;
            var item;
            var duration;

            while (daysRemaining > 0) {
                item = _.get(itemsByDay, curDay, []).shift();
                doy = startDate.clone().add(curDay, 'day').format(DOY_FORMAT);
                if (item == null) {
                    days.push(emptyCell({
                        doy: doy
                    }));
                    curDay++;
                    daysRemaining--;
                } else {
                    duration = Math.ceil(getVisibleDuration(item.endDate, item.startDate));
                    days.push(itemTemplate({
                        id: args.isGroupRow ? _.get(item.data, options.eventKeys.parent.id) : _.get(item.data, options.eventKeys.id),
                        event: args.renderer.render(getRenderItem(item), getRenderOptions(item)),
                        duration: duration,
                        doy: doy,
                        groupId: args.groupId,
                        classes: isEditable(_.get(item.data, options.eventKeys.id)) ? ['editable'] : []
                    }));
                    curDay += duration;
                    daysRemaining -= duration;
                }
            }

            return rowTemplate({
                events: days,
                groupId: args.groupId,
                classes: args.isGroupRow ? ['task-group-row'] : []
            });
        }

        /**
         * @method buildTaskRows
         * @param {Object} eventsByDay
         * @param {String} [groupId]
         * @returns {String}
         */
        function buildTaskRows(eventsByDay, groupId) {
            var table = '';
            while (_.some(_.values(eventsByDay), 'length')) {
                table += buildRow(eventsByDay, taskTemplate, {
                    renderer: options.renderer,
                    groupId: groupId,
                    isGroupRow: false
                });
            }
            return table;
        }

        /**
         * @method buildGroupRow
         * @param {Object} groupsByDay
         * @param {String} groupId
         */
        function buildGroupRow(groupsByDay, groupId) {
            return buildRow(groupsByDay, taskGroupTemplate, {
                renderer: options.groupRenderer,
                isGroupRow: true,
                groupId: groupId
            });
        }

        /**
         * @method buildGroupSegment
         * @param {Array} events
         * @param {Calendar#Event} group
         * @param {String} groupId
         */
        function buildGroupSegment(events, group, groupId) {
            var startDate = options.getStartDate();
            var segment = [];
            segment.push(buildGroupRow(groupByDay([group], options.getStartDate()), groupId));
            segment.push(buildTaskRows(prepareEvents(events, startDate), groupId));
            return segment.join('');
        }

        /**
         * @method buildOutOfRangeSegment
         * @param {Array} events
         * @param {Calendar#Event} group
         * @param {String} groupId
         */
        function buildOutOfRangeSegment(events, group, groupId) {
            var viewRange = getViewRange();
            var eventRange = findEventRange(events);
            var outOfRangeGroup;
            var groupStart;
            var groupEnd;
            if (group.startDate.isAfter(viewRange.end)) {
                groupStart = eventRange.start.clone();
                groupEnd = viewRange.end.clone();
            } else if (group.endDate.isBefore(viewRange.start)) {
                groupStart = viewRange.start.clone();
                groupEnd = eventRange.end.clone();
            }
            outOfRangeGroup = createGroup(groupId, groupStart, groupEnd);
            outOfRangeGroup.outOfView = true;
            outOfRangeGroup.originalGroup = group;
            return buildGroupSegment(events, outOfRangeGroup, groupId);
        }

        /**
         * @method buildUngroupedSegment
         * @param {Array} events
         */
        function buildUngroupedSegment(events) {
            var groupRange = getViewRange();
            return buildGroupSegment(events, createGroup('null-group-id', groupRange.start, groupRange.end), 'null-group-id');
        }

        /**
         * @method append
         * @param {String} rows
         */
        function append(rows) {
            var $table = getTable();
            var $appendTo = options.getAppendTo();
            $table.detach();
            $table.append(rows);
            $appendTo.append($table);
        }

        /**
         * @method setActive
         * @param {String} activeId
         */
        function setActive(activeId) {
            var $table = getTable();
            $table.find('.task[data-event-id=' + activeId + ']').addClass('active');
        }

        /**
         * @method addByGroups
         * @param {Array} events
         * @param {Array} groups
         */
        function addByGroups(events, groups) {
            var tasksByGroup = getTasksByGroup(events, groups);
            var table = '';
            var tasks;
            var groupId;
            if (_.get(tasksByGroup, null, []).length) {
                table += buildUngroupedSegment(tasksByGroup[null]);
            }
            _.each(sortByDay(groups), function (group) {
                groupId = _.get(group.data, options.eventKeys.parent.id);
                tasks = _.get(tasksByGroup, groupId, []);
                if (isOccurring(group)) {
                    table += buildGroupSegment(tasks, group, groupId);
                } else if (tasks.length) {
                    table += buildOutOfRangeSegment(tasks, group, groupId);
                }
            });
            append(table);
        }

        /**
         * @method addAllTasks
         * @param {Array} events
         * @param {Array} groups
         * @param {Object} args
         */
        function addAllTasks(events, groups, args) {
            if (groups && groups.length) {
                addByGroups(events, groups);
                _.each(args.groupArgs, function (args, groupId) {
                    if (args.collapsed) {
                        toggleGroup(groupId, args.collapsed);
                    }
                });
            } else {
                append(buildTaskRows(prepareEvents(events, options.getStartDate())));
            }

            if (_.has(args, 'activeId')) {
                setActive(args.activeId);
            }
        }

        /**
         * @method redrawGroup
         * @param {Array} events
         * @param {Array} groups
         * @param {String} groupId
         */
        function redrawGroup(events, groups, groupId) {
            if (groupId === null) {
                groupId = 'null-group-id';
            }

            var $table = getTable();
            var groupSegment = '';
            var $groupRows = $table.find('tr[data-group-id=' + groupId + ']');
            var temporaryRow = '<tr class="temporary-task-group-row"></tr>';
            var tasksByGroup = getTasksByGroup(events, groups);
            var taskGroup = _.findWhere(groups, _.set({}, options.eventKeys.parent.id, groupId));
            var tasks = _.get(tasksByGroup, groupId === 'null-group-id' ? null : groupId, []);

            $groupRows.wrapAll(temporaryRow);

            if (groupId === 'null-group-id') {
                groupSegment += buildUngroupedSegment(tasksByGroup[null]);
            } else if (isOccurring(taskGroup)) {
                groupSegment += buildGroupSegment(tasks, taskGroup, groupId);
            } else if (tasks.length) {
                groupSegment += buildOutOfRangeSegment(tasks, taskGroup, groupId);
            }

            getTable().find('.temporary-task-group-row').replaceWith(groupSegment);
        }

        return {
            buildTaskRows: buildTaskRows,
            getDuration: getDuration,
            getVisibleDuration: getVisibleDuration,
            groupByDay: groupByDay,
            removeAllRows: removeAllRows,
            addAllTasks: addAllTasks,
            redrawGroup: redrawGroup,
            toggleGroup: toggleGroup
        };
    };
});


define('template-58', ['underscore'], function (_) { return function(data) {
var __t, __p = '', __j = Array.prototype.join;
function print() { __p += __j.call(arguments, '') }
__p += '<div class="day-time-container"> <div class="day-time-header"> <table> <tr class="days"> <td class="dummy-cell"></td> ';
 for (var i = 0; i < data.visibleDays; i++) { ;
__p += ' <th class="dow"><span class="view-day"></span></th> ';
 } ;
__p += ' <td class="scroll-dummy"></td> </tr> <tr class="all-day-events ' +
((__t = ( !data.allDayVisible ? 'hidden' : '' )) == null ? '' : __t) +
'"> <td class="dummy-cell"></td> <td colspan="' +
((__t = ( data.visibleDays )) == null ? '' : __t) +
'"> <table class="all-day-events-table task-table"> <tbody> <tr class="empty-row"> ';
 for (var i = 0; i < data.visibleDays; i++) { ;
__p += ' <td></td> ';
 } ;
__p += ' </tr> </tbody> </table> </td> <td class="scroll-dummy"></td> </tr> </table> </div> <div class="day-time-grid"> <table class="event-times"> <tr class="markers" height="1"> <td class="dummy-cell"> </td> <td colspan="' +
((__t = ( data.visibleDays )) == null ? '' : __t) +
'"> <div class="rel-container"> <div class="marker-container ' +
((__t = ( data.granularity )) == null ? '' : __t) +
'"> <div class="marker"></div> <div class="marker"></div> <div class="marker"></div> <div class="marker"></div> <div class="marker"></div> <div class="marker"></div> <div class="marker"></div> <div class="marker"></div> <div class="marker"></div> <div class="marker"></div> <div class="marker"></div> <div class="marker"></div> <div class="marker"></div> <div class="marker"></div> <div class="marker"></div> <div class="marker"></div> <div class="marker"></div> <div class="marker"></div> <div class="marker"></div> <div class="marker"></div> <div class="marker"></div> <div class="marker"></div> <div class="marker"></div> <div class="marker"></div> </div> </div> </td> </tr> <tr class="events"> <td class="times"> <div class="hour"> </div> <div class="hour"> </div> <div class="hour"> </div> <div class="hour"> </div> <div class="hour"> </div> <div class="hour"> </div> <div class="hour"> </div> <div class="hour"> </div> <div class="hour"> </div> <div class="hour"> </div> <div class="hour"> </div> <div class="hour"> </div> <div class="hour"> </div> <div class="hour"> </div> <div class="hour"> </div> <div class="hour"> </div> <div class="hour"> </div> <div class="hour"> </div> <div class="hour"> </div> <div class="hour"> </div> <div class="hour"> </div> <div class="hour"> </div> <div class="hour"> </div> <div class="hour"> </div> </td> ';
 for (var i = 0; i < data.visibleDays; i++) { ;
__p += ' <td class="day"> <div class="day-wrapper"> <div class="gutter"></div> </div> </td> ';
 } ;
__p += ' </tr> </table> </div></div>';
return __p
};});


define('template-59', ['underscore'], function (_) { return function(data) {
var __t, __p = '';
__p += '<div data-overlay-id="' +
((__t = ( data.id )) == null ? '' : __t) +
'" class="' +
((__t = ( ['overlay'].concat(data.classes || []).join(' ') )) == null ? '' : __t) +
'"></div>';
return __p
};});


define('template-60', ['underscore'], function (_) { return function(data) {
var __t, __p = '';
__p += '<div id="' +
((__t = ( data.pageId )) == null ? '' : __t) +
'" data-event-id="' +
((__t = ( data.id )) == null ? '' : __t) +
'" class="' +
((__t = ( ['event'].concat(data.classes).join(' ') )) == null ? '' : __t) +
'"> <div class="delete"><i class="pgbu-icon-small pgbu-icon-delete"></i></div> <div class="duration">' +
((__t = ( data.duration )) == null ? '' : __t) +
'</div> <div class="event-title" title="' +
((__t = ( data.title )) == null ? '' : __t) +
'">' +
((__t = ( data.title )) == null ? '' : __t) +
'</div> <div class="event-description"></div> <div class="resizer"> <i class="pgbu-icon pgbu-icon-drag-drop-dots-horizontal"></i> </div></div>';
return __p
};});


define('template-61', ['underscore'], function (_) { return function(data) {
var __t, __p = '';
__p += '<ul class="dropdown-menu" role="menu" data-doy="' +
((__t = ( data.doy )) == null ? '' : __t) +
'" id="' +
((__t = ( _.uniqueId('header-menu') )) == null ? '' : __t) +
'"></ul>';
return __p
};});


define('template-62', ['underscore'], function (_) { return function(data) {
var __t, __p = '';
__p += '<li class="' +
((__t = ( [].concat([ data.disabled, data.hideDisabled ]).join(' ') )) == null ? '' : __t) +
'"> <a role="menuitem" id="' +
((__t = ( _.uniqueId('action-item') )) == null ? '' : __t) +
'" href="#" data-trigger="' +
((__t = ( data.trigger )) == null ? '' : __t) +
'">' +
((__t = ( data.label )) == null ? '' : __t) +
'</a></li>';
return __p
};});

define('oui/widgets/calendar/dayView',['require','underscore','jquery','moment','oui/utils/math','oui/utils/keyCode','oui/utils/tasks/tasks','template-58','template-59','template-60','template-57','template-61','template-62'],function (require) {
    'use strict';

    var _ = require('underscore');
    var $ = require('jquery');
    var moment = require('moment');
    var math = require('oui/utils/math');
    var keyCode = require('oui/utils/keyCode');
    var taskUtil = require('oui/utils/tasks/tasks');
    var dayTimeViewTemplate = require('template-58');
    var overlayTemplate = require('template-59');
    var eventTemplate = require('template-60');
    var emptyCell = require('template-57');
    var menuTemplate = require('template-61');
    var actionTemplate = require('template-62');

    var DAY_TIME_VIEW = {
        ALL_DAY: 'all-day',
        SINGLE_DAY: 'single-day'
    };

    //TODO: determine how to keep this defined for granularity reasons
    var startHourFn = _.partial(math.step, 'floor');
    var endHourFn = _.partial(math.step, 'round');

    /**
     * Returns the hour for the given coordinates, scrolling state and row height.
     *
     * @method DayTimeView~computeHour
     * @private
     *
     * @param {Object} options
     * @param {Number} options.y
     * @param {Number} options.scrollTop
     * @param {Number} options.top
     * @param {Number} options.rowHeight
     *
     * @returns {Number}
     */
    function computeHour(options) {
        return (options.y + options.scrollTop - options.top) / options.rowHeight;
    }

    /**
     * @method DayTimeView~eventIndexOf
     * @private
     *
     * @param {Array} arr
     * @param {Object} event
     */
    function eventIndexOf(arr, event) {
        for (var i = 0; i < arr.length; i++) {
            if (arr[i].id === event.id) {
                return i;
            }
        }

        return -1;
    }

    /**
     * @method DayTimeView~getWidth
     * @private
     *
     * @param {Number} index
     * @param {Number} length
     */
    function getWidth(index, length) {
        if (index === 0) {
            return 1 - 0.1;
        } else if (index === length - 1) {
            return 1 - (index / length);
        } else {
            return 1 - (index / length) - 0.1;
        }
    }

    /**
     * @class DayTimeView
     * @private
     *
     * @param {Calendar#Config} options
     * @param {Object} controller
     * @param {jQuery} $container
     * @param {Object} settings
     */
    function DayTimeView(options, controller, $container, settings) {
        this._controller = controller;
        this._$container = $container;

        this._$element = null;
        this._options = options;
        this._startDate = moment().startOf(this._getMomentAttribute());
        this._setup = false;
        this._settings = settings;
        this._taskUtil = taskUtil(this.getTaskUtilOptions(), this.getElement.bind(this));
    }

    DayTimeView.prototype = {
        /**
         * @method DayTimeView#_getOptions
         * @private
         *
         * @returns {Object} dayOptions
         */
        _getOptions: function () {
            return this._options.day;
        },

        /**
         * @method DayTimeView#getTaskUtilOptions
         */
        getTaskUtilOptions: function () {
            var options = this._getOptions();
            return {
                getElement: this.getElement.bind(this),
                getAppendTo: this.getAppendTo.bind(this),
                getTable: this.getTable.bind(this),
                getStartDate: this.getStartDate.bind(this),
                getVisibleDays: this._getVisibleDays.bind(this),
                getMomentAttribute: this._getMomentAttribute.bind(this),
                renderer: options.renderer,
                eventKeys: this._options.eventKeys,
                locale: this._options.locale
            };
        },

        /**
         * @method DayTimeView#getElement
         * @returns {jQuery}
         */
        getElement: function () {
            return this._$element;
        },

        /**
         * @method DayTimeView#getTable
         * @returns {jQuery}
         */
        getTable: function () {
            return this._$allDayTableBody;
        },

        /**
         * @method DayTimeView#getAppendTo
         * @returns {jQuery}
         */
        getAppendTo: function () {
            return this._$allDayTable;
        },

        /**
         * @method DayTimeView#getStartDate
         * @returns {moment}
         */
        getStartDate: function () {
            return this._startDate;
        },

        /**
         * @method DayTimeView#_getMomentAttribute
         * @private
         *
         * @returns {String} 'day'
         */
        _getMomentAttribute: function () {
            return 'day';
        },

        /**
         * @method DayTimeView#_getVisibleDays
         * @private
         *
         * @returns {Number} 1
         */
        _getVisibleDays: function () {
            return 1;
        },

        /**
         * @method DayTimeView#_getEventElem
         * @private
         *
         * @param {Object} event
         *
         * @returns {jQuery} element
         */
        _getEventElem: function (event) {
            var id = event;
            if (event.id) {
                id = event.id;
            }
            return this._$element.find('[data-event-id=' + id + ']');
        },

        /**
         * @method DayTimeView#_getEventId
         * @private
         *
         * @param {Object} event
         *
         * @returns {Number} event ID
         */
        _getEventId: function (event) {
            return event.attr('data-event-id');
        },

        /**
         * @method DayTimeView#_getEvent
         * @private
         *
         * @param {Object} event
         *
         * @returns {Object} event
         */
        _getEvent: function (event) {
            var id = event;
            if (event.id) {
                id = event.id;
            }

            return this._controller.getEvent(id);
        },

        /**
         * @method DayTimeView#_isEditable
         * @private
         *
         * @param {*} options
         *
         * @returns {Boolean} editable
         */
        _isEditable: function (options) {
            if (_.isBoolean(this._options.editable)) {
                return this._options.editable;
            } else if (_.isFunction(this._options.editable)) {
                return this._options.editable(options);
            }
        },

        /**
         * @method DayTimeView#positionEvent
         *
         * @param {jQuery} element
         * @param {Object} ev
         * @param {Object} modifiers
         */
        positionEvent: function (element, ev, modifiers) {
            var $ev = element || this._getEventElem(ev);
            var minutes = ev.endDate.diff(ev.startDate, 'minutes');
            var start = ev.startDate;

            if (modifiers && modifiers.start) {
                start = modifiers.start;
                minutes = ev.endDate.diff(start, 'minutes', true);
            } else if (ev.endDate.day() !== ev.startDate.day()) {
                minutes = ev.startDate.clone().endOf('day').diff(ev.startDate, 'minutes', true);
            }

            $ev.css({
                top: (start.hours() / 24) * 100 + (start.minutes() / 1440) * 100 + '%',
                height: (minutes / 1440) * 100 + '%'
            });

            //Make sure DOM stays in correct order for stacking
            var $gutter = $ev.closest('.gutter');
            var $events = $gutter.find('.event');
            $events.sort(function (ev1, ev2) {
                var top1 = $(ev1).position().top;
                var top2 = $(ev2).position().top;
                if (top1 > top2) {
                    return 1;
                }

                return -1;
            });

            $events.detach().appendTo($gutter);
        },

        /**
         * @method DayTimeView#overlapEvent
         *
         * @param {Object} ev
         * @param {Array} overlap
         */
        overlapEvent: function (ev, overlap) {
            var $ev = this._getEventElem(ev);

            if (overlap.length > 1) {
                //Actually have overlaps
                var index = eventIndexOf(overlap, ev);

                $ev.toggleClass('event-overlap', index > 0);
                $ev.css({
                    left: (index / overlap.length) * 100 + '%',
                    width: getWidth(index, overlap.length) * 100 + '%'
                });
            } else {
                $ev.css({
                    left: 0,
                    width: '100%'
                });
            }
        },

        /**
         * @method DayTimeView#getOverlappingEvents
         *
         * @param {Object} ev
         * @param {Array} events
         *
         * @returns {Array} overlappingEvents
         */
        getOverlappingEvents: function (ev, events) {
            return _.filter(events, function (event) {
                return event.isOccurring(ev.startDate, ev.endDate);
            });
        },

        /**
         * @method DayTimeView#clear
         */
        clear: function () {
            //TODO: clean up all created elements for this view
            if (this._$element != null) {
                this._$element.addClass('hidden');
            }
        },

        /**
         * @method DayTimeView#_cacheElements
         * @private
         */
        _cacheElements: function () {
            this._$allDayTable = this._$element.find('.all-day-events-table');
            this._$allDayTableBody = this._$allDayTable.find('tbody');
        },

        /**
         * @method DayTimeView#setHeaders
         *
         * @param {moment} start
         */
        setHeaders: function (start) {
            var date = start.clone();
            var dow = this._$element.find('.day-time-header .days th');
            var today = moment();
            var that = this;
            var options = this._getOptions();
            var emptyRow = this._$element.find('.all-day-events-table .empty-row td');

            _.each(dow, function (day) {
                if (that._getOptions().headerCell) {
                    $(day)
                      .empty()
                      .attr('data-doy', date.format('YYYY-MM-DD'))
                      .append(that._getOptions().headerCell(date.clone()))
                      .removeAttr('tabindex')
                      .toggleClass('with-menu', options.headerMenu && options.headerMenu.enabled);

                } else {
                    var format = that._getOptions().getHeader(date.clone());
                    $(day).find('.view-day')
                      .text(format)
                      .attr('title', format)
                      .attr('data-doy', date.format('YYYY-MM-DD'))
                      .removeAttr('tabindex')
                      .toggleClass('with-menu', options.headerMenu && options.headerMenu.enabled);

                    if (options.headerMenu && options.headerMenu.enabled) {
                        $(day).find('.view-day').attr('tabindex', 0);
                    }
                }

                $(day).toggleClass('today', that._getOptions().highlightToday && date.clone().startOf('day').isSame(today.startOf('day')));
                date.add(1, 'day');
            });

            date = start.clone();
            _.each(emptyRow, function (day) {
                $(day).attr('data-doy', date.format('YYYY-MM-DD'));
                date.add(1, 'day');
            });
        },

        /**
         * @method DayTimeView#setDays
         *
         * @param {moment} start
         */
        setDays: function (start) {
            var date = start.clone();
            var days = this._$element.find('.day').removeClass('disabled');
            var that = this;

            that._gutters = {};

            //TODO: do we want to use another div inside the day <td>?
            _.each(days, function (day) {
                $(day)
                  .attr('data-doy', date.format('YYYY-MM-DD'))
                  .toggleClass('disabled', that._controller.isDayDisabled(date));
                $(day).find('.day-wrapper').height($(day).height());
                that._gutters[date.format('YYYY-MM-DD')] = $(day);
                date.add(1, 'day');
            });
        },

        /**
         * @method DayTimeView#setTimes
         */
        setTimes: function () {
            var startHour = moment().startOf('day');
            var hours = this._$element.find('.times .hour');

            var format = this._getOptions().timeFormat;
            _.each(hours, function (hour) {
                $(hour)
                  .text(startHour.format(format))
                  .attr('data-hour', startHour.hour());
                startHour.add(1, 'h');
            });
        },

        /**
         * @method DayTimeView#addEvent
         *
         * @param {Object} event
         * @param {Array} overlaps
         */
        addEvent: function (event, overlaps) {
            var that = this;

            if (this._options.events.isAllDay(event.data)) {
                var dateRange = this.getDateRange();
                this._taskUtil.removeAllRows();
                this._taskUtil.addAllTasks(_.filter(this._controller.getCachedEvents(dateRange.start, dateRange.end), function (event) {
                    return that._options.events.isAllDay(event.data);
                }));
            } else {
                var $day;

                //TODO: allow for customimzed
                var $event = $(eventTemplate({
                    pageId: this._settings.name + '-' + this._settings.id + '-' + event.id,
                    id: event.id,
                    duration: this._options.events.getEventDuration(event.data),
                    classes: this._options.events.getClasses(event.data),
                    title: this._options.events.getEventName(event.data)
                }));

                //TODO: temporary
                overlaps = _.reject(overlaps, function (event) {
                    return that._options.events.isAllDay(event.data);
                });

                //Determine if its an event that spans multiple days
                //Include overrides to position and overlap to tweak event start time since the start time will be midnight
                var gutter = this._gutters[event.startDate.format('YYYY-MM-DD')];
                if (gutter) {
                    $day = gutter.find('.gutter');
                    $day.append($event);

                    this.positionEvent($event, event);
                    this.overlapEvent(event, overlaps);
                }

                if (event.startDate.day() !== event.endDate.day() &&
                  event.endDate.diff(event.endDate.clone().startOf('day'), 'minutes', true) > 0) {
                    var $cont = $event.clone().addClass('event-cont');
                    $day = this._gutters[event.endDate.format('YYYY-MM-DD')];
                    if ($day != null) {
                        $day = $day.find('.gutter');
                        $day.append($cont);
                    }

                    this.positionEvent($cont, event, {
                        start: event.endDate.clone().startOf('day')
                    });
                }
            }
        },

        /**
         * @method DayTimeView#removeEvent
         *
         * @param {Object} event
         * @param {Array} overlaps
         */
        removeEvent: function (event, overlaps) {
            var $event = this._getEventElem(event);

            if ($event.is('td')) {
                $event.replaceWith(emptyCell({
                    doy: event.startDate.format('YYYY-MM-DD')
                }));
            } else {
                $event.remove();
            }

            var that = this;
            overlaps = _.reject(overlaps, function (ev) {
                return ev.id === event.id;
            });
            _.each(overlaps, function (ev) {
                that.overlapEvent(ev, overlaps);
            });
        },

        /**
         * @method DayTimeView#updateEvent
         *
         * @param {Object} event
         * @param {Array} overlaps
         * @param {Array} oldOverlaps
         */
        updateEvent: function (event, overlaps, oldOverlaps) {
            var that = this;

            var grouped = _.groupBy(overlaps, function (ev) {
                if (that._options.events.isAllDay(ev.data)) {
                    return DAY_TIME_VIEW.ALL_DAY;
                } else {
                    return DAY_TIME_VIEW.SINGLE_DAY;
                }
            });

            var oldGrouped = _.groupBy(oldOverlaps, function (ev) {
                if (that._options.events.isAllDay(ev.data)) {
                    return DAY_TIME_VIEW.ALL_DAY;
                } else {
                    return DAY_TIME_VIEW.SINGLE_DAY;
                }
            });

            //TODO: determine whether we need different forms of update (duration update, startDate update, details)
            that.removeEvent(event, oldGrouped[DAY_TIME_VIEW.SINGLE_DAY] || []);

            that.addEvent(event, grouped[DAY_TIME_VIEW.SINGLE_DAY] || []);
        },

        /**
         * @method DayTimeView#updateEvents
         *
         * @param {Array} events
         * @param {Array} overlaps
         * @param {Array} oldOverlaps
         */
        updateEvents: function (events, overlaps, oldOverlaps) {
            var that = this;
            _.each(events, function (event) {
                that.updateEvent(event, overlaps, oldOverlaps);
            });
        },

        /**
         * @method DayTimeView#addOverlap
         *
         * @param {Object} event
         */
        addOverlay: function (event) {
            var $day;
            var $event = $(overlayTemplate({
                classes: this._options.overlays.getClasses(event),
                id: event.id
            }));

            //Determine if its an event that spans multiple days
            //Include overrides to position and overlap to tweak event start time since the start time will be midnight
            //TODO: need to append a continuation if the startDate and endDate are not on the same day
            $day = this._$element.find('.day[data-doy=' + event.startDate.format('YYYY-MM-DD') + '] .day-wrapper');
            $day.append($event);

            this.positionEvent($event, event);
        },

        /**
         * @method DayTimeView#setEvents
         *
         * @param {Array} events
         */
        setEvents: function (events) {
            var that = this;

            var grouped = _.groupBy(events, function (ev) {
                if (that._options.events.isAllDay(ev.data)) {
                    return DAY_TIME_VIEW.ALL_DAY;
                } else {
                    return DAY_TIME_VIEW.SINGLE_DAY;
                }
            });

            _.each(grouped[DAY_TIME_VIEW.SINGLE_DAY], function (event) {
                that.addEvent(event, that.getOverlappingEvents(event, grouped[DAY_TIME_VIEW.SINGLE_DAY]));
            });

            //Should be a stable sort
            //First sort by length and then by startDate to get all the longest, earliest days first
            var allday = _.chain(grouped[DAY_TIME_VIEW.ALL_DAY])
                .sortBy(function (event) {
                    return event.endDate.diff(event.startDate, 'minutes', true);
                })
                .sortBy(function (event) {
                    return +event.startDate.clone().startOf('day');
                })
                .value();

            if (allday.length > 0) {
                this._taskUtil.removeAllRows();
                this._taskUtil.addAllTasks(allday);
            }
        },

        /**
         * @method DayTimeView#setOverlays
         *
         * @param {Array} overlays
         */
        setOverlays: function (overlays) {
            var that = this;

            _.each(overlays, function (overlay) {
                that.addOverlay(overlay);
            });
        },

        /**
         * @method DayTimeView#addViewEvent
         *
         * @param {Object} data
         *
         * @returns {jQuery} event element
         */
        addViewEvent: function (data) {
            var $day;
            var $event = $(eventTemplate({
                pageId: this._settings.name + '-' + this._settings.id + '-temp',
                id: 'temp',
                duration: this._options.events.getEventDuration(data),
                classes: ['temp'].concat(this._options.events.getClasses(data)),
                title: this._options.events.getEventName(data)
            }));

            //Make sure there is only one temp view event added
            this._$element.find('[data-event-id=temp]').remove();

            //Determine if its an event that spans multiple days
            //Include overrides to position and overlap to tweak event start time since the start time will be midnight
            var gutter = this._gutters[data.startDate.format('YYYY-MM-DD')];
            if (gutter) {
                $day = gutter.find('.gutter');
                $day.append($event);

                this.positionEvent($event, data);
            }

            return $event;
        },

        /**
         * @method DayTimeView#updateViewEvent
         *
         * @param {jQuery} $ev
         * @param {Object} data
         *
         * @returns {jQuery} event element
         */
        updateViewEvent: function ($ev, data) {
            $ev.remove();
            return this.addViewEvent(data);
        },

        /**
         * @method DayTimeView#getScale
         *
         * @returns {Number} scale
         */
        getScale: function () {
            return this._getOptions().granularity === 'hour' ? 1 : 0.5;
        },

        /**
         * @method DayTimeView#highlightNext
         *
         * @returns {jQuery} active
         */
        highlightNext: function () {
            var $active = this._$element.find('.active:visible');
            var $events = this._$element.find('.event:visible');

            $events.removeClass('active');

            if ($active.length === 0) {
                this._$element.find('.event:visible:eq(0)').addClass('active');
            } else if ($events.index($active) < $events.length - 1) {
                $events.eq($events.index($active) + 1).addClass('active');
            } else {
                $active.addClass('active');
            }

            return $active;
        },

        /**
         * @method DayTimeView#highlightPrev
         *
         * @returns {jQuery} active
         */
        highlightPrev: function () {
            var $active = this._$element.find('.active:visible');
            var $events = this._$element.find('.event:visible');

            $events.removeClass('active');

            if ($active.length === 0) {
                this._$element.find('.event:visible').last().addClass('active');
            } else if ($events.index($active) > 0) {
                $events.eq($events.index($active) - 1).addClass('active');
            } else {
                $active.addClass('active');
            }

            return $active;
        },

        /**
         * @method DayTimeView#shiftEnd
         *
         * @param {Number} up
         */
        shiftEnd: function (up) {
            var id = this._getEventId(this._$element.find('.active.event'));
            var correction = 1;
            var event;
            var editable;

            if (id) {
                event = this._getEvent(id);
                editable = this._isEditable({
                    event: event
                });
            }

            //TODO: determine if using a kill-switch is useful here
            if (!editable || _.isNaN(id)) {
                return;
            }

            up = up ? 1 : -1;

            if (this._getOptions().granularity === 'half-hour') {
                correction = 0.5;
            }

            var newEnd = event.endDate.clone().add(up * correction, 'hours');
            var newDuration = newEnd.diff(event.startDate, 'hours', true);

            if (newEnd.day() === event.startDate.day() && newDuration >= 0.5 && newDuration <= 24) {
                this._controller.updateEvent(id, {
                    endDate: event.endDate.clone().add(up * correction, 'hours')
                });
            }

            this._getEventElem(id).addClass('active');
        },

        /**
         * @method DayTimeView#shiftStart
         *
         * @param {Number} up
         */
        shiftStart: function (up) {
            var id = this._getEventId(this._$element.find('.active.event'));
            var correction = 1;
            var event;
            var editable;

            if (id) {
                event = this._getEvent(id);
                editable = this._isEditable({
                    event: event
                });
            }

            //TODO: determine if using a kill-switch is useful here
            if (!editable || _.isNaN(id)) {
                return;
            }

            up = up ? 1 : -1;

            if (this._getOptions().granularity === 'half-hour') {
                correction = 0.5;
            }

            var newStart = event.startDate.clone().add(up * correction, 'hours');
            var newDuration = event.endDate.diff(newStart, 'hours', true);

            if (newStart.day() === event.endDate.day() && newDuration >= 0.5 && newDuration <= 24) {
                this._controller.updateEvent(id, {
                    startDate: newStart
                });
            }

            this._getEventElem(id).addClass('active');
        },

        /**
         * @method DayTimeView#move
         *
         * @param {Number} up
         */
        move: function (up) {
            var id = this._getEventId(this._$element.find('.active.event'));
            var correction = 1;
            var event;
            var editable;

            if (id) {
                event = this._getEvent(id);
                editable = this._isEditable({
                    event: event
                });
            }

            //TODO: determine if using a kill-switch is useful here
            if (!editable || _.isNaN(id)) {
                return;
            }

            up = up ? 1 : -1;

            if (this._getOptions().granularity === 'half-hour') {
                correction = 0.5;
            }

            var currentDuration = event.endDate.diff(event.startDate, 'hours', true);
            var newStart = event.startDate.clone().add(up * correction, 'hours');

            if (newStart.day() === event.endDate.day()) {
                this._controller.updateEvent(id, {
                    startDate: newStart,
                    endDate: newStart.clone().add(currentDuration, 'hours')
                });
            }

            this._getEventElem(id).addClass('active');
        },

        /**
         * @method DayTimeView#deleteEvent
         */
        deleteEvent: function () {
            var activeEvent = this._$element.find('.active.event');
            var ev = this._getEventId(activeEvent) || this._getEventId(activeEvent.parent());
            this._controller.removeEvent(ev);
        },

        /**
         * @method DayTimeView#_withinDayBounds
         * @private
         *
         * @param {moment} day
         * @param {moment} start
         * @param {moment} end
         *
         * @returns {Boolean}
         */
        _withinDayBounds: function (day, start, end) {
            return (end.day() === day.day() &&
              start.day() === day.day()) ||
              (!end.isAfter(end.clone().startOf('day')) &&
              !start.isBefore(day.clone().startOf('day')));
        },

        /**
         * @method DayTimeView#setupView
         */
        setupView: function () {
            var that = this;
            this._$element = $(dayTimeViewTemplate({
                visibleDays: this._getVisibleDays(),
                allDayVisible: this._getOptions().allDay,
                granularity: this._getOptions().granularity
            }));
            this._setup = true;

            this._cacheElements();

            this._$element.on('click', '.view-day', function () {
                that._controller.currentDate(moment($(this).closest('[data-doy]').attr('data-doy'), 'YYYY-MM-DD'));
                that._$element.trigger('view-change.calendar', 'day');
            });

            this._$element.on('click', '.with-menu', function () {
                var $target = $(this);
                var options = that._getOptions();
                var $menu = $(menuTemplate({
                    doy: $target.attr('data-doy')
                }));

                that._$element.find('.dropdown-menu').remove();

                $target.attr('aria-owns', $menu.attr('id')).attr('data-menu', '#' + $menu.attr('id'));

                _.each(options.headerMenu.actionOrder, function (actionKey) {
                    var actionProps = options.headerMenu.actions[actionKey];
                    var enabled = options.headerMenu.menuFns.isActionEnabled(actionKey, moment($target.attr('data-doy'), 'YYYY-MM-DD'));
                    $menu.append(actionTemplate({
                        label: actionProps.label,
                        hideDisabled: !enabled && options.headerMenu.hideDisabled ? 'hidden' : '',
                        disabled: !enabled ? 'disabled' : '',
                        trigger: actionProps.trigger
                    }));
                });

                that._$element.append($menu);

                //TODO: use a class for static css properties
                $menu.css({
                    position: 'fixed',
                    display: 'block',
                    top: $target.offset().top + $target.height(),
                    left: $target.offset().left
                });
            });

            this._$element.on('blur', '.with-menu', function () {
                that._$element.find('.dropdown-menu').remove();
                $(this).removeAttr('aria-owns aria-activedescendant data-menu');
            });

            this._$element.on('keydown', '.with-menu', function (event) {
                var $target = $(this);
                var menu = that._$element.find($(this).attr('data-menu'));
                var items = menu.find('li:not(.divider,.disabled):visible a');
                var active = menu.find('li.active');
                var index = items.index(active.children('a'));

                active.removeClass('active');
                $target.removeAttr('aria-activedescendant');

                function moveTo(index) {// jscs:ignore jsDoc
                    if (index > -1 && index < items.length) {
                        var newActive = items.eq(index);

                        $target.attr('aria-activedescendant', newActive.attr('id'));

                        newActive
                          .parent()
                          .addClass('active');
                    }
                }

                switch (event.keyCode) {
                    case keyCode.ESC:
                        that._$element.find('.dropdown-menu').remove();
                        $target.removeAttr('aria-owns aria-activedescendant data-menu');
                        break;
                    case keyCode.ENTER:
                    case keyCode.SPACE:
                        event.preventDefault();
                        if (active.length === 1) {
                            active.find('[data-trigger]').click();
                        } else if (menu.length === 0) {
                            $target.click();
                        }
                        break;
                    case keyCode.DOWN:
                        event.preventDefault();
                        moveTo(index + 1);
                        break;
                    case keyCode.UP:
                        event.preventDefault();
                        moveTo(index - 1);
                        break;
                }
            });

            this._$element.on('mousedown', '.dropdown-menu', function (event) {
                event.stopPropagation();
                event.preventDefault();
            });

            this._$element.on('click', 'li:not(.disabled) [data-trigger]', function (event) {
                event.stopPropagation();
                that._controller._model.trigger($(this).attr('data-trigger'), moment($(this).closest('[data-doy]').attr('data-doy'), 'YYYY-MM-DD'));
                that._$element.find('[aria-owns=' + that._$element.find('.dropdown-menu').attr('id') + ']').removeAttr('aria-owns aria-activedescendant data-menu');
                that._$element.find('.dropdown-menu').remove();
            });

            this._$element.on('click', '.delete', function (event) {
                event.stopPropagation();
                var id = that._getEventId($(this).closest('[data-event-id]'));
                var editable;

                if (id) {
                    event = that._getEvent(id);
                    editable = that._isEditable({
                        event: event
                    });
                }

                if (editable) {
                    that._controller.removeEvent(id);
                }
            });

            this._$element.on('click', '.all-day-events td:empty', function () {
                var date = moment($(this).attr('data-doy'), 'YYYY-MM-DD');
                that._controller.newEvent(date, date.clone().add(1, 'day').startOf('day'));
            });

            this._$element.on('mousedown', '.delete', function (event) {
                event.stopPropagation();
            });

            this._$element.on('mousedown', '.event', function () {
                if (document.activeElement !== that._$container[0]) {
                    //IE is the only one to allow focus to move to .events
                    //Refocus the container
                    that._$container.focus();
                }
                that._$element.find('.event.active').removeClass('active');

                $(this).addClass('active');
            });

            this._$element.on('focusin', '.event', function () {
                //IE is the only one to allow focus to move to .events
                //Refocus the container
                that._$container.focus();
            });

            this._$element.on('mousedown.create', '.day:not(.disabled)', function (event) {
                $('body').addClass('oui-disable-select');
                var date = moment($(this).attr('data-doy'), 'YYYY-MM-DD');
                var startHour = startHourFn(that.getScale(), computeHour({
                    y: event.clientY,
                    scrollTop: that._$element.find('.day-time-grid').scrollTop(),
                    top: that._$element.find('.day-time-grid').offset().top,
                    rowHeight: that._getOptions().rowHeight
                }));
                var startMoment = date.clone().startOf('day').minutes(startHour * 60);
                var endMoment = startMoment.clone().add(that.getScale(), 'hours');
                var addedEvent;
                var isEditable = that._isEditable({
                    date: date
                });

                //TODO: determine if using a kill-switch is useful here
                if (event.which === 3 || !isEditable) {
                    return;
                }

                if (that._options.events.allowOverlaps || that._controller.getCachedSingleDayEvents(startMoment, endMoment).length === 0) {
                    addedEvent = that.addViewEvent({
                        startDate: startMoment,
                        endDate: endMoment
                    });

                    that._$element.off('mousemove.create').on('mousemove.create', function (event) {
                        var endHour = endHourFn(that.getScale(), computeHour({
                            y: event.clientY,
                            scrollTop: that._$element.find('.day-time-grid').scrollTop(),
                            top: that._$element.find('.day-time-grid').offset().top,
                            rowHeight: that._getOptions().rowHeight
                        }));
                        var tempMoment = date.clone().startOf('day').minutes(endHour * 60);
                        var events = [];

                        if (!that._options.events.allowOverlaps) {

                            if (endHour < startHour) {
                                events = that._controller.getCachedSingleDayEvents(tempMoment, startMoment);
                            } else if (endHour === startHour) {
                                events = that._controller.getCachedSingleDayEvents(startMoment, startMoment.clone().add(that.getScale(), 'hours'));
                            } else {
                                events = that._controller.getCachedSingleDayEvents(startMoment, tempMoment);
                            }
                        }

                        if (events.length === 0) {
                            endMoment = tempMoment;
                        }

                        if (endHour < startHour) {
                            addedEvent = that.updateViewEvent(addedEvent, {
                                startDate: endMoment,
                                endDate: startMoment
                            });
                        } else if (endHour === startHour) {
                            addedEvent = that.updateViewEvent(addedEvent, {
                                startDate: startMoment,
                                endDate: startMoment.clone().add(that.getScale(), 'hours')
                            });
                        } else {
                            addedEvent = that.updateViewEvent(addedEvent, {
                                startDate: startMoment,
                                endDate: endMoment
                            });
                        }

                    });
                    that._$element.off('mouseup.create').one('mouseup.create', function () {
                        $('body').removeClass('oui-disable-select');
                        that._$element.off('mousemove.create');
                        addedEvent.remove();
                        var temp;

                        if (startMoment.isAfter(endMoment)) {
                            temp = startMoment.clone();
                            startMoment = endMoment;
                            endMoment = temp;
                        }

                        //Only add after they have mouseup'd
                        that._controller.newEvent(startMoment.startOf('minute'), endMoment.startOf('minute'));
                    });

                }

            });

            this._$element.on('mousedown.dragging', '.event', function (event) {
                event.stopPropagation();
                if (that._options.events.draggable) {
                    $('body').addClass('oui-disable-select');
                    var editing = true;
                    var curTarget = $(event.target).closest('.event');
                    var curEvent = that._getEvent(that._getEventId(curTarget));
                    var startMoment = curEvent.startDate;
                    var endMoment = curEvent.endDate;
                    var startMouseHour = startHourFn(that.getScale(), computeHour({
                        y: event.clientY,
                        scrollTop: that._$element.find('.day-time-grid').scrollTop(),
                        top: that._$element.find('.day-time-grid').offset().top,
                        rowHeight: that._getOptions().rowHeight
                    }));
                    var finalPosition = null;
                    var finalStart = null;
                    var duration = endMoment.diff(startMoment, 'hours', true);
                    var editable = that._isEditable({
                        event: curEvent,
                        day: startMoment
                    });

                    //TODO: determine if using a kill-switch is useful here
                    if (event.which === 3 || !editable) {
                        return;
                    }

                    that._$element.off('mousemove.dragging').on('mousemove.dragging', function (event) {
                        event.stopPropagation();
                        editing = false;
                        var events = [];
                        var $day = $(document.elementFromPoint(event.clientX, event.clientY)).closest('.day');
                        var dayMoment = moment($day.attr('data-doy'), 'YYYY-MM-DD');
                        var endMouseHour = startHourFn(that.getScale(), computeHour({
                            y: event.clientY,
                            scrollTop: that._$element.find('.day-time-grid').scrollTop(),
                            top: that._$element.find('.day-time-grid').offset().top,
                            rowHeight: that._getOptions().rowHeight
                        }));
                        var potentialStart;
                        var potentialEnd;
                        var newStartHour;
                        curTarget.addClass('dragging');

                        if (dayMoment.isValid()) {
                            newStartHour = startMoment.hour() + (endMouseHour - startMouseHour);
                            potentialStart = dayMoment.clone().hours(Math.floor(newStartHour)).minutes(math.getMantissa(newStartHour) * 60);
                            potentialEnd = potentialStart.clone().add(duration, 'hours');

                            if (!that._options.events.allowOverlaps) {
                                events = _.reject(that._controller.getCachedSingleDayEvents(potentialStart, potentialEnd), function (event) {
                                    return event.id === curEvent.id;
                                });
                            }

                            if (curTarget != null && events.length === 0 &&
                              (!that._options.events.singleDay || that._withinDayBounds(dayMoment, potentialStart, potentialEnd))) {

                                finalPosition = newStartHour;
                                finalStart = potentialStart;

                                curTarget.css({
                                    zIndex: 1000,
                                    width: curTarget.closest('.day').width(),
                                    left: 0,
                                    top: finalPosition / 24 * 100 + '%'
                                });

                                if (!$day.is(curTarget.closest('.day'))) {
                                    curTarget.appendTo($day.find('.gutter'));
                                }

                            }
                        }
                    });

                    that._$element.one('mouseup.dragging', function (event) {
                        event.stopPropagation();
                        $('body').removeClass('oui-disable-select');
                        $(document).off('mouseup.oui-dragging');

                        if (curTarget) {
                            curTarget.removeClass('dragging');
                        }

                        //TODO: handle if editing === true
                        if (!editing && curTarget && finalStart != null) {
                            that._controller.updateEvent(that._getEventId(curTarget), {
                                startDate: finalStart,
                                endDate: {
                                    day: finalStart.clone().startOf('day'),
                                    hour: Math.floor(finalPosition + duration),
                                    minutes: math.getMantissa(finalPosition + duration) * 60
                                }
                            });
                        }
                        curTarget = null;
                        that._$element.off('mousemove.dragging');
                    });

                    $(document).one('mouseup.oui-dragging', function () {
                        that._$element.off('mousemove.dragging mouseup.dragging');
                        that.updateEvent(that._getEvent(that._getEventId(curTarget)));
                    });
                }
            });

            this._$element.on('mousedown.resize', '.resizer', function (event) {
                event.stopPropagation();
                $('body').addClass('oui-disable-select');
                var editing = true;
                var date = moment($(this).closest('.day').attr('data-doy'), 'YYYY-MM-DD');
                var curTarget = $(event.target).closest('.event');
                var entireEvent = that._getEventElem({
                    id: that._getEventId(curTarget)
                });
                var startHour = startHourFn(that.getScale(), computeHour({
                    y: curTarget.offset().top,
                    scrollTop: that._$element.find('.day-time-grid').scrollTop(),
                    top: that._$element.find('.day-time-grid').offset().top,
                    rowHeight: that._getOptions().rowHeight
                }));
                var startMoment = date.clone().startOf('day').minutes(startHour * 60);
                var curId = that._getEventId(curTarget);
                var endMoment = that._controller.getEvent(curId).endDate;
                var editable = that._isEditable({
                    event: that._controller.getEvent(curId),
                    date: date
                });

                //TODO: determine if using a kill-switch is useful here
                if (event.which === 3 || !editable) {
                    return;
                }

                that._$element.off('mousemove.resize').on('mousemove.resize', function (event) {
                    event.stopPropagation();
                    editing = false;
                    var endMouseHour = endHourFn(that.getScale(), computeHour({
                        y: event.clientY,
                        scrollTop: that._$element.find('.day-time-grid').scrollTop(),
                        top: that._$element.find('.day-time-grid').offset().top,
                        rowHeight: that._getOptions().rowHeight
                    }));
                    var tempMoment = startMoment.clone().startOf('day').minutes(endMouseHour * 60);
                    entireEvent.addClass('dragging');
                    var events = [];

                    if (!that._options.events.allowOverlaps) {
                        events = _.reject(that._controller.getCachedSingleDayEvents(startMoment, tempMoment), function (event) {
                            return event.id === curId;
                        });
                    }

                    if (events.length === 0 && tempMoment.isAfter(startMoment)) {
                        endMoment = tempMoment;
                    }

                    curTarget = that.updateViewEvent(curTarget, {
                        startDate: startMoment,
                        endDate: endMoment
                    });
                });

                that._$element.one('mouseup.resize', function (event) {
                    event.stopPropagation();
                    $('body').removeClass('oui-disable-select');

                    if (curTarget) {
                        entireEvent.removeClass('dragging');
                    }

                    //TODO: handle if editing is true
                    if (!editing && curTarget) {
                        that._controller.updateEvent(curId, {
                            endDate: endMoment
                        });
                        curTarget.remove();
                    }
                    curTarget = null;
                    that._$element.off('mousemove.resize');
                });
            });

            this._$container.append(this._$element);
            this._setup = true;
        },

        /**
         * @method DayTimeView#forceRedraw
         */
        forceRedraw: function () {
            this._currentDate = null;
        },

        /**
         * @method DayTimeView#render
         *
         * @param {moment} currentDate
         */
        render: function (currentDate) {
            var that = this;
            var startDate;
            currentDate = currentDate || this._controller.getCurrentDate();
            startDate = currentDate.clone().startOf(this._getMomentAttribute());
            this._startDate = startDate;

            if (!this._setup) {
                this.setupView();
            } else {
                this._$element.removeClass('hidden');
            }

            this.clean();

            if (this._currentDate == null || !this._currentDate.isSame(currentDate)) {
                this._currentDate = currentDate.clone();
                this.setHeaders(startDate);
                this.setDays(startDate);
                this.setTimes();
            }

            $.when(this._controller.getOverlays(startDate, startDate.clone().endOf(this._getMomentAttribute())))
                .done(function (overlays) {
                    that.setOverlays(overlays);
                });

            $.when(this._controller.getEvents(startDate, startDate.clone().endOf(this._getMomentAttribute())))
                .done(function (events) {
                    that.setEvents(events);
                });
        },

        /**
         * @method DayTimeView#removeAllEvents
         */
        removeAllEvents: function () {
            this._$element.find('.event').remove();
        },

        /**
         * @method DayTimeView#removeAllOverlays
         */
        removeAllOverlays: function () {
            this._$element.find('.overlay').remove();
        },

        /**
         * @method DayTimeView#clean
         */
        clean: function () {
            this.removeAllEvents();
            this.removeAllOverlays();
        },

        /**
         * @method DayTimeView#next
         */
        next: function () {
            var currentDate = this._controller.getCurrentDate();

            this._controller.currentDate(currentDate.clone().add(1, this._getMomentAttribute()));
        },

        /**
         * @method DayTimeView#prev
         */
        prev: function () {
            var currentDate = this._controller.getCurrentDate();

            this._controller.currentDate(currentDate.clone().subtract(1, this._getMomentAttribute()));
        },

        /**
         * @method DayTimeView#getDateRange
         *
         * @returns {CalendarView#DateRange} dateRange
         */
        getDateRange: function () {
            return {
                start: this._startDate,
                end: this._startDate.clone().endOf(this._getMomentAttribute())
            };
        },

        /**
         * @method DayTimeView#resize
         */
        resize: function () {
            var toolbarHeight = 0;
            var legendHeight = 0;
            var $grid = this._$element.find('.day-time-grid');
            var allDayEventsHeight = 0;

            if (this._options.layout.showToolbar) {
                toolbarHeight = this._options.layout.toolbarHeight;
            }

            if (this._getOptions().legend) {
                legendHeight = this._options.layout.legendHeight;
            }

            if (this._getOptions().allDay) {
                allDayEventsHeight = this._$element.find('.all-day-events')[0].clientHeight;
            }

            var borders = $grid.outerHeight() - $grid.height();

            $grid.addClass('hidden');

            //TODO: should this be passed in during resize so the weekView does not have to reach outside itself?
            var calHeight = this._$container.closest('.orcl-calendar').height();

            $grid.css('height', calHeight - toolbarHeight - this._getOptions().headerHeight - borders - allDayEventsHeight - legendHeight);

            $grid.removeClass('hidden');

            _.each($('.day-wrapper'), function (dayWrapper) {
                var $wrapper = $(dayWrapper);
                $wrapper.height($wrapper.parent().height());
            });
        },

        /**
         * @method DayTimeView#redraw
         *
         * @param {moment} startDate
         * @param {moment} endDate
         */
        redraw: function (startDate, endDate) { // jshint ignore:line
            this.render(this._controller.getCurrentDate());

            this.resize();
        },

        /**
         * @method DayTimeView#getActiveEvent
         *
         * @returns {Object} event
         */
        getActiveEvent: function () {
            return this._getEvent(this._getEventId(this._$element.find('.event.active')));
        }
    };

    return DayTimeView;
});

define('oui/widgets/calendar/weekView',[
  'oui/widgets/calendar/dayView'
], function (DayView) {
    'use strict';

    /**
     * @class WeekView
     * @private
     * @extends {DayTimeView}
     */
    function WeekView() {
        DayView.apply(this, arguments);
    }

    var wvp = WeekView.prototype = Object.create(DayView.prototype);

    /**
     * @method WeekView#_getOptions
     * @private
     *
     * @returns {Object} week options
     */
    wvp._getOptions = function () {
        return this._options.week;
    };

    /**
     * @method WeekView#_getMomentAttribute
     * @private
     *
     * @returns {String} 'week'
     */
    wvp._getMomentAttribute = function () {
        return 'week';
    };

    /**
     * @method WeekView#_getVisibleDays
     * @private
     *
     * @returns {Number} 7
     */
    wvp._getVisibleDays = function () {
        return 7;
    };

    return WeekView;
});


define('template-63', ['underscore'], function (_) { return function(data) {
var __t, __p = '', __j = Array.prototype.join;
function print() { __p += __j.call(arguments, '') }
__p += '<div class="month-container"> <div class="month-header ' +
((__t = ( data.showWeekNumbers ? 'show-week' : '' )) == null ? '' : __t) +
'"> <table> <tr class="days"> <th class="week"></th> <th class="dow"></th> <th class="dow"></th> <th class="dow"></th> <th class="dow"></th> <th class="dow"></th> <th class="dow"></th> <th class="dow"></th> </tr> </table> </div> <div class="month-grid ' +
((__t = ( data.showWeekNumbers ? 'show-week' : '' )) == null ? '' : __t) +
' ' +
((__t = ( data.switchToDay ? 'day-switch' : '' )) == null ? '' : __t) +
'"> <table class="borders"> ';
 for (var i = 0; i < 6; i++) { ;
__p += ' <tr> <td class="week"></td> <td><div class="cell"></div></td> <td><div class="cell"></div></td> <td><div class="cell"></div></td> <td><div class="cell"></div></td> <td><div class="cell"></div></td> <td><div class="cell"></div></td> <td><div class="cell"></div></td> </tr> ';
 } ;
__p += ' </table> <div class="month-row"> <table class="week-events"> <tr> <td class="week"></td> <td class="day"> <span class="header"> <span class="number"></span> </span> </td> <td class="day"> <span class="header"> <span class="number"></span> </span> </td> <td class="day"> <span class="header"> <span class="number"></span> </span> </td> <td class="day"> <span class="header"> <span class="number"></span> </span> </td> <td class="day"> <span class="header"> <span class="number"></span> </span> </td> <td class="day"> <span class="header"> <span class="number"></span> </span> </td> <td class="day"> <span class="header"> <span class="number"></span> </span> </td> </tr> </table> </div> <div class="month-row"> <table class="week-events"> <tr> <td class="week"></td> <td class="day"> <span class="header"> <span class="number"></span> </span> </td> <td class="day"> <span class="header"> <span class="number"></span> </span> </td> <td class="day"> <span class="header"> <span class="number"></span> </span> </td> <td class="day"> <span class="header"> <span class="number"></span> </span> </td> <td class="day"> <span class="header"> <span class="number"></span> </span> </td> <td class="day"> <span class="header"> <span class="number"></span> </span> </td> <td class="day"> <span class="header"> <span class="number"></span> </span> </td> </tr> </table> </div> <div class="month-row"> <table class="week-events"> <tr> <td class="week"></td> <td class="day"> <span class="header"> <span class="number"></span> </span> </td> <td class="day"> <span class="header"> <span class="number"></span> </span> </td> <td class="day"> <span class="header"> <span class="number"></span> </span> </td> <td class="day"> <span class="header"> <span class="number"></span> </span> </td> <td class="day"> <span class="header"> <span class="number"></span> </span> </td> <td class="day"> <span class="header"> <span class="number"></span> </span> </td> <td class="day"> <span class="header"> <span class="number"></span> </span> </td> </tr> </table> </div> <div class="month-row"> <table class="week-events"> <tr> <td class="week"></td> <td class="day"> <span class="header"> <span class="number"></span> </span> </td> <td class="day"> <span class="header"> <span class="number"></span> </span> </td> <td class="day"> <span class="header"> <span class="number"></span> </span> </td> <td class="day"> <span class="header"> <span class="number"></span> </span> </td> <td class="day"> <span class="header"> <span class="number"></span> </span> </td> <td class="day"> <span class="header"> <span class="number"></span> </span> </td> <td class="day"> <span class="header"> <span class="number"></span> </span> </td> </tr> </table> </div> <div class="month-row"> <table class="week-events"> <tr> <td class="week"></td> <td class="day"> <span class="header"> <span class="number"></span> </span> </td> <td class="day"> <span class="header"> <span class="number"></span> </span> </td> <td class="day"> <span class="header"> <span class="number"></span> </span> </td> <td class="day"> <span class="header"> <span class="number"></span> </span> </td> <td class="day"> <span class="header"> <span class="number"></span> </span> </td> <td class="day"> <span class="header"> <span class="number"></span> </span> </td> <td class="day"> <span class="header"> <span class="number"></span> </span> </td> </tr> </table> </div> <div class="month-row"> <table class="week-events"> <tr> <td class="week"></td> <td class="day"> <span class="header"> <span class="number"></span> </span> </td> <td class="day"> <span class="header"> <span class="number"></span> </span> </td> <td class="day"> <span class="header"> <span class="number"></span> </span> </td> <td class="day"> <span class="header"> <span class="number"></span> </span> </td> <td class="day"> <span class="header"> <span class="number"></span> </span> </td> <td class="day"> <span class="header"> <span class="number"></span> </span> </td> <td class="day"> <span class="header"> <span class="number"></span> </span> </td> </tr> </table> </div> </div></div>';
return __p
};});


define('template-64', ['underscore'], function (_) { return function(data) {
var __t, __p = '';
__p += '<tr> <td class="week"></td></tr>';
return __p
};});


define('template-65', ['underscore'], function (_) { return function(data) {
var __t, __p = '';
__p += '<td data-event="' +
((__t = ( data.id )) == null ? '' : __t) +
'" colspan="' +
((__t = ( data.duration )) == null ? '' : __t) +
'"> <div data-event-id="' +
((__t = ( data.id )) == null ? '' : __t) +
'" title="' +
((__t = ( data.time )) == null ? '' : __t) +
' ' +
((__t = ( data.title )) == null ? '' : __t) +
'" class="' +
((__t = ( ['event'].concat(data.classes).concat(data.boundary ? 'boundary' : '').join(' ') )) == null ? '' : __t) +
'"> <span class="event-duration">' +
((__t = ( data.time )) == null ? '' : __t) +
'</span> <span class="event-title" title="' +
((__t = ( data.title )) == null ? '' : __t) +
'">' +
((__t = ( data.title )) == null ? '' : __t) +
'</span> </div></td>';
return __p
};});


define('template-66', ['underscore'], function (_) { return function(data) {
var __t, __p = '';
__p += '<td class="overflow" data-doy="' +
((__t = ( data.doy )) == null ? '' : __t) +
'"> <a href="#" title="' +
((__t = ( data.locale.more )) == null ? '' : __t) +
'" data-action="more">' +
((__t = ( data.locale.more )) == null ? '' : __t) +
'</a></td>';
return __p
};});


define('template-67', ['underscore'], function (_) { return function(data) {
var __t, __p = '';
__p += '<div class="' +
((__t = ( ['overlay'].concat(data.classes).join(' ') )) == null ? '' : __t) +
'"></div>';
return __p
};});


define('template-68', ['underscore'], function (_) { return function(data) {
var __t, __p = '', __j = Array.prototype.join;
function print() { __p += __j.call(arguments, '') }
__p += '<span class="indicators"> ';
 for (var i = 0; i < data.indicators.length; i++) { ;
__p += ' <i class="pgbu-icon-small ' +
((__t = ( data.indicators[i].classes.join(' ') )) == null ? '' : __t) +
'" title="' +
((__t = ( data.indicators[i].title )) == null ? '' : __t) +
'"></i> ';
 } ;
__p += '</span>';
return __p
};});


define('template-69', ['underscore'], function (_) { return function(data) {
var __t, __p = '', __j = Array.prototype.join;
function print() { __p += __j.call(arguments, '') }
__p += '<div class="overflow-popup"> <div class="overflow-day"> ' +
((__t = ( data.day )) == null ? '' : __t) +
' </div> <table class="events"> ';
 for (var i = 0, length = data.events.length; i < length; i++) { ;
__p += ' <tr> ' +
((__t = ( data.events[i] )) == null ? '' : __t) +
' </tr> ';
 } ;
__p += ' </table></div>';
return __p
};});

define('oui/widgets/calendar/monthView',['require','underscore','jquery','moment','oui/utils/keyCode','template-63','template-64','template-65','template-57','template-66','template-67','template-68','template-69','oui/utils/positioning'],function (require) {
    'use strict';

    var _ = require('underscore');
    var $ = require('jquery');
    var moment = require('moment');
    var keyCode = require('oui/utils/keyCode');
    var monthViewTemplate = require('template-63');
    var weekTemplate = require('template-64');
    var eventTemplate = require('template-65');
    var emptyCell = require('template-57');
    var overflowTemplate = require('template-66');
    var monthOverlayTemplate = require('template-67');
    var indicatorTemplate = require('template-68');
    var overflowPopup = require('template-69');
    var relativePosition = require('oui/utils/positioning');

    /**
     * @class MonthView
     * @private
     *
     * @param {Calendar#Config} options
     * @param {Object} controller
     * @param {jQuery} $container
     * @param {Object} settings
     */
    function MonthView(options, controller, $container, settings) {
        this._controller = controller;
        this._$container = $container;

        this._$element = null;
        this._options = options;
        this._startDate = moment().startOf('month');
        this._setup = false;
        this._settings = settings;
    }

    MonthView.prototype = {
        /**
         * @method MonthView#_positionPopup
         * @private
         *
         * @param {jQuery} $relativeElement
         * @param {Boolean} parentIsBody
         */
        _positionPopup: function ($relativeElement, parentIsBody) {
            var top;
            var left;
            var position;
            var offset = $relativeElement.offset();
            var popupHeight = this.$menu.outerHeight(true);
            var popupWidth = this.$menu.width();

            //Determine Base Relative Positioning
            position = relativePosition($relativeElement, parentIsBody, this.$menu);

            //Set up base css top/left coordinates
            top = position.top;
            left = position.left;

            //Correct for window dimensions
            if (offset.top + popupHeight + $relativeElement.outerHeight() > window.pageYOffset + $(window).height()) {
                top -= (popupHeight + $relativeElement.outerHeight());
            }

            if (offset.left + popupWidth > window.pageXOffset + $(window).width()) {
                left -= (popupWidth - $relativeElement.innerWidth());
            }

            //Set the coordinates
            this.$menu.css({
                'top': top,
                'left': left
            });
        },

        /**
         * @method MonthView#_getEvent
         * @private
         *
         * @param {Object} event
         *
         * @returns {jQuery} event element
         */
        _getEvent: function (event) {
            var id = event;
            if (event.id) {
                id = event.id;
            }
            return this._$element.find('[data-event-id=' + id + ']');
        },

        /**
         * @method MonthView#_getEventId
         * @private
         *
         * @param {jQuery} event
         *
         * @returns {Number} event ID
         */
        _getEventId: function (event) {
            return parseInt(event.attr('data-event-id'), 10);
        },

        /**
         * @method MonthView#clear
         */
        clear: function () {
            //TODO: clean up all created elements for this view
            if (this._$element != null) {
                this._$element.addClass('hidden');
            }
        },

        /**
         * @method MonthView#setHeaders
         *
         * @param {moment} start
         */
        setHeaders: function (start) {
            var date = start.clone();
            var dow = this._$element.find('.month-header .dow');
            var that = this;

            this._$element.find('.month-header .week').text(that._options.locale.week);

            _.each(dow, function (day) {
                if (that._options.month.headerCell) {
                    $(day)
                      .empty()
                      .append(that._options.month.headerCell(date.clone()));
                } else {
                    var format = that._options.month.getHeader(date.clone());
                    $(day)
                      .text(format)
                      .attr('title', format)
                      .attr('data-date', date.format('YYYY-MM-DD'));
                }

                date.add(1, 'day');
            });
        },

        /**
         * @method MonthView#setDays
         *
         * @param {moment} start
         */
        setDays: function (start) {
            var date = start.clone();
            var week = start.clone();
            var days = this._$element.find('.day');
            var that = this;
            var today = moment();
            var weeks = this._$element.find('.week-events .week');

            that._$element.find('.diff-month').removeClass('diff-month');
            that._$element.find('.borders td').removeClass().filter(':first-child').addClass('week');

            //TODO: do we want to use another div inside the day <td>?
            _.each(days, function (day) {
                var borderCell = that._$element.find('.borders tr').eq($(day).closest('.month-row').index() - 1).find('td').eq($(day).index());
                borderCell.attr('data-day', date.format('YYYY-MM-DD'));

                //TODO: move this to a constant format
                $(day).attr('data-day', date.format('YYYY-MM-DD'));
                if (date.date() === 1) {
                    $(day).find('.number').text(date.format('MMM D'));
                } else {
                    $(day).find('.number').text(date.format('D'));
                }

                if (this._options.month.indicators) {
                    var indicators = this._options.month.indicators(date);
                    $(day).find('.indicators').remove();
                    $(day).children('.header').prepend(indicatorTemplate({
                        indicators: indicators
                    }));
                }
                $(day).add(borderCell).toggleClass('today', that._options.month.highlightToday && date.clone().startOf('day').isSame(today.startOf('day')));
                if (date.month() !== that._controller.getCurrentDate().month()) {
                    borderCell.addClass('diff-month');
                }
                date.add(1, 'day');
            }.bind(this));

            _.each(weeks, function (wk) {
                $(wk).text(week.format('w'));
                week.add(1, 'week');
            });
        },

        /**
         * @method MonthView#addEvent
         */
        addEvent: function () {
            /*
            var $day;
            //TODO: allow for customimzed
            var $event = $(eventTemplate({
                pageId : this._settings.name + '-' + this._settings.id + '-' + event.id,
                id : event.id,
                duration : this._options.events.getEventDuration(event.data),
                classes : this._options.events.getClasses(event.data),
                title : this._options.events.getEventName(event.data)
            }));
            */
        },

        /**
         * @method MonthView#removeEvent
         *
         * @param {Object} event
         */
        removeEvent: function (event) {
            this._getEvent(event).remove();
        },

        /**
         * @method MonthView#updateEvent
         *
         * @param {Object} event
         */
        updateEvent: function (event) {
            var that = this;

            //TODO: determine whether we need different forms of update (duration update, startDate update, details)
            that.removeEvent(event);

            that.addEvent(event);
        },

        /**
         * @method MonthView#updateEvents
         *
         * @param {Array} events
         * @param {Array} overlaps
         * @param {Array} oldOverlaps
         */
        updateEvents: function (events, overlaps, oldOverlaps) {
            var that = this;
            _.each(events, function (event) {
                that.updateEvent(event, overlaps, oldOverlaps);
            });
        },

        /**
         * @method MonthView#setupEvents
         *
         * @param {Array} events
         */
        setEvents: function (events) {
            var that = this;
            var startDate = this._startDate.clone();
            var grouped;
            var endDate;

            if (startDate.month() !== this._controller.getCurrentDate().month()) {
                endDate = this._controller.getCurrentDate().clone().startOf('month').add(5, 'week').endOf('week');
            } else {
                endDate = startDate.clone().endOf('month');
            }

            function buildWeek(wkEvents, index, startDate, endDate, maxRows) {// jscs:ignore jsDoc
                var rows = {};
                var minRowsToDraw;
                var drawnRows;

                function getEvent(row, day) {// jscs:ignore jsDoc
                    if (rows[day]) {
                        var ev = _.find(rows[day], function (event) {
                            return event.startDate.day() >= day;
                        }) || rows[day][row];

                        if (ev) {
                            _.each(rows, function (days, index) {
                                rows[index] = _.reject(rows[index], function (event) {
                                    return ev.id === event.id;
                                });
                            });
                        }

                        return ev;
                    }
                }

                function buildRowsDay(start) {// jscs:ignore jsDoc
                    _.each(_.range(7), function (day) {
                        _.each(wkEvents, function (event) {
                            var occurs = event.isOccurring(start, start.clone().endOf('day'));
                            if (occurs) {
                                if (rows[day] == null) {
                                    rows[day] = [];
                                }
                                rows[day].push(event);
                            }
                        });
                        start.add(1, 'day');
                    });
                }

                buildRowsDay(startDate.clone());
                minRowsToDraw = _.max(rows, function (day) {
                    return day.length;
                }).length;

                drawnRows = Math.min(minRowsToDraw, maxRows);

                _.each(_.range(drawnRows), function (row) {
                    var $curRow = $(weekTemplate({
                        showWeekNumber: that._options.month.showWeekNumbers
                    }));
                    var curDay = 0;

                    while (curDay < 7) {
                        var event = getEvent(row, curDay);
                        if (event == null) {
                            $curRow.append(emptyCell({
                                doy: startDate.clone().add(curDay, 'day').format('YYYY-MM-DD')
                            }));
                            curDay++;
                        } else if (event != null &&
                          $curRow.find('[data-event=' + event.id + ']').length === 0) {
                            curDay += Math.ceil(event.endDate.diff(event.startDate, 'days', true)) || 1;
                            $curRow.append(eventTemplate({
                                id: event.id,
                                time: that._options.events.getEventDuration(event.data),
                                title: that._options.events.getEventName(event.data),
                                classes: that._options.events.getClasses(event.data),
                                boundary: event.endDate.week() !== event.startDate.week(),
                                duration: Math.ceil(event.endDate.diff(event.startDate, 'days', true))
                            }));
                        }

                    }

                    that._$element.find('.month-row:eq(' + index + ') .week-events tbody').append($curRow);
                });

                if (drawnRows !== minRowsToDraw) {
                    var $curRow = $(weekTemplate({
                        showWeekNumber: that._options.month.showWeekNumbers
                    }));
                    var curDay = 0;

                    while (curDay < 7) {
                        var event = getEvent(drawnRows + 1, curDay);
                        if (event == null) {
                            $curRow.append(emptyCell({
                                doy: startDate.clone().add(curDay, 'day').format('YYYY-MM-DD')
                            }));
                            curDay++;
                        } else if (event != null) {
                            $curRow.append(overflowTemplate({
                                doy: startDate.clone().add(curDay, 'day').format('YYYY-MM-DD'),
                                locale: that._options.locale,
                                howMany: minRowsToDraw - drawnRows
                            }));
                            curDay++;
                        }

                    }

                    that._$element.find('.month-row:eq(' + index + ') .week-events tbody').append($curRow);
                }
            }

            this._$element.find('.month-row tr:not(:first-child)').remove();

            grouped = {};
            var weeks = Math.round(endDate.diff(startDate, 'week', true));

            _.each(events, function (event) {
                var startIndex = event.startDate.week() - startDate.week();
                var endIndex = event.endDate.week() - startDate.week();

                if (startIndex < 0) {
                    startIndex += 52;
                }
                startIndex = startIndex % weeks;

                if (endIndex < 0) {
                    endIndex += 52;
                }
                endIndex = endIndex % weeks;

                grouped[startIndex] = grouped[startIndex] || [];
                grouped[startIndex].push(event);

                if (endIndex !== startIndex) {
                    grouped[endIndex] = grouped[endIndex] || [];
                    grouped[endIndex].push(event);
                }
            });

            var maximumRowsAvailable = Math.floor(this._$element.find('.month-row').height() / this._options.month.rowHeight - 2);

            if (maximumRowsAvailable > 0) {
                _.each(grouped, function (wkEvents, week) {
                    buildWeek(wkEvents, parseInt(week, 10),
                              startDate.clone().add(week, 'week').startOf('week'),
                              startDate.clone().add(week, 'week').endOf('week'),
                              maximumRowsAvailable);
                });
            }
        },

        /**
         * @method MonthView#setOverlays
         *
         * @param {Array} overlays
         * @param {moment} startDate
         */
        setOverlays: function (overlays, startDate) {
            var that = this;
            var date = startDate.clone();
            var cells = this._$element.find('.borders td:not(.week) .cell');

            this._$element.find('.overlay').remove();

            _.each(cells, function (cell) {
                _.each(overlays, function (overlay) {
                    if (overlay.isOccurring(date.clone().startOf('day'), date.clone().endOf('day'))) {
                        $(cell).append(monthOverlayTemplate({
                            classes: that._options.overlays.getClasses(date)
                        }));
                    }
                });
                date.add(1, 'day');
            });
        },

        /**
         * @method MonthView#_showMore
         * @private
         *
         * @param {moment} day
         */
        _showMore: function (day) {
            var relativeElement = this._$element.find('.borders [data-day=' + day.format('YYYY-MM-DD') + ']');
            var events = this._controller.getCachedEvents(day.clone().startOf('day'), day.clone().endOf('day'));
            var that = this;

            that._closePopup();

            var markup = _.map(events, function (event) {
                return eventTemplate({
                    id: event.id,
                    time: that._options.events.getEventStart(event.data),
                    title: that._options.events.getEventName(event.data),
                    classes: that._options.events.getClasses(event.data),
                    boundary: event.endDate.week() !== event.startDate.week(),
                    duration: 1
                });
            });

            if (this.$menu) {
                this.$menu.remove();
            }

            this.$menu = $(overflowPopup({
                day: day.format('LL'),
                events: markup
            })).on('mousedown', function (event) {
                event.preventDefault();
                event.stopPropagation();
            });

            $('body').append(this.$menu).one('mousedown.calendar-overflow', function () {
                that._closePopup();
            });

            $(window).one('scroll.calendar-overflow resize.calendar-overflow', function () {
                that._closePopup();
            });

            this._positionPopup(relativeElement, true);
        },

        /**
         * @method MonthView#_closePopup
         * @private
         */
        _closePopup: function () {
            $('body').off('.calendar-overflow');
            $(window).off('.calendar-overflow');

            if (this.$menu) {
                this.$menu.remove();
            }
        },

        /**
         * @method MonthView#setupView
         */
        setupView: function () {
            var that = this;

            this._$element = $(monthViewTemplate({
                showWeekNumbers: this._options.month.showWeekNumbers,
                switchToDay: this._options.month.numbersSwitchToDayView
            }));

            this._$element.on('click', '.day-switch .number', function (event) {
                event.stopPropagation();

                that._controller.currentDate(moment($(this).closest('[data-day]').attr('data-day'), 'YYYY-MM-DD'));
                that._$element.trigger('view-change.calendar', 'day');
            });

            this._$element.on('click', '[data-action=more]', function (event) {
                event.preventDefault();
                var doy = moment($(this).closest('[data-doy]').attr('data-doy'), 'YYYY-MM-DD');
                that._showMore(doy);
            }).on('blur', '[data-action=more]', function () {
                that._closePopup();
            }).on('keydown', '[data-action=more]', function (event) {
                switch (event.keyCode) {
                    case keyCode.ESC:
                        that._closePopup();
                        break;
                    case keyCode.SPACE:
                    case keyCode.ENTER:
                        event.preventDefault();
                        var doy = moment($(this).closest('[data-doy]').attr('data-doy'), 'YYYY-MM-DD');
                        that._showMore(doy);
                        break;
                }
            });

            this._$element.on('click', '[data-day]', function (event) {
                var $target = $(event.target);
                var offset = $target.offset();
                var day = $target.attr('data-day');
                var monthGrid = that._$element.find('.month-grid');
                var args = {};
                var coords = {};
                var bounds = {};

                coords.top = offset.top;
                coords.left = offset.left;
                coords.right = offset.left + $target.outerWidth();

                bounds.top = monthGrid.offset().top;
                bounds.bottom = monthGrid.offset().top + monthGrid.height();
                bounds.height = monthGrid.innerHeight();

                args.day = moment(day, 'YYYY-MM-DD');
                args.coordinates = coords;
                args.bounds = bounds;

                if (that._options.month.showDetails) {
                    that._options.month.showDetails(args);
                }
            });

            this._$container.append(this._$element);
            this._setup = true;
        },

        /**
         * @method MonthView#render
         *
         * @param {moment} currentDate
         */
        render: function (currentDate) {
            var that = this;
            var startDate;
            var endDate;
            var weeks;
            currentDate = currentDate || this._controller.getCurrentDate();
            startDate = currentDate.clone().startOf('month').startOf('week');

            //TODO: when using dynamic weeks may not need to add 5 each time
            endDate = currentDate.clone().startOf('month').add(5, 'week').endOf('week');

            this._startDate = startDate;
            if (!this._setup) {
                this.setupView();
            } else {
                this._$element.removeClass('hidden');
            }

            if (!this._startDate.isSame(currentDate)) {
                this.clean();
            }

            if (this._options.month.dynamicWeeks) {
                weeks = Math.ceil(currentDate.clone().endOf('month').endOf('week').diff(startDate.clone(), 'weeks', true));

                this._$element.find('.month-grid').removeClass('wk-6 wk-5 wk-4').addClass('wk-' + weeks);
            }

            this.setHeaders(startDate);
            this.setDays(startDate);

            $.when(this._controller.getEvents(startDate, endDate))
                .done(function (events) {
                    that.setEvents(events);
                });

            $.when(this._controller.getOverlays(startDate, endDate))
                .done(function (overlays) {
                    that.setOverlays(overlays, startDate);
                });
        },

        /**
         * @method MonthView#redraw
         */
        redraw: function () {
            this.render(this._controller.getCurrentDate());
        },

        /**
         * @method MonthView#removeAllEvents
         */
        removeAllEvents: function () {
            this._$element.find('.month-row tr:not(:first-child)').remove();
        },

        /**
         * @method MonthView#clean
         */
        clean: function () {
            this.removeAllEvents();
        },

        /**
         * @method MonthView#next
         */
        next: function () {
            var currentDate = this._controller.getCurrentDate();

            this._controller.currentDate(currentDate.clone().add(1, 'month'));
        },

        /**
         * @method MonthView#prev
         */
        prev: function () {
            var currentDate = this._controller.getCurrentDate();

            this._controller.currentDate(currentDate.clone().subtract(1, 'month'));
        },

        /**
         * @method MonthView#getDateRange
         *
         * @returns {CalendarView#DateRange} dateRange
         */
        getDateRange: function () {
            return {
                start: this._startDate,
                end: this._startDate.clone().add(5, 'week').endOf('week')
            };
        },
        /**
         * @method MonthView#resize
         */
        resize: function () {
            var toolbarHeight = 0;
            var legendHeight = 0;
            var $grid = this._$element.find('.month-grid');

            if (this._options.layout.showToolbar) {
                toolbarHeight = this._options.layout.toolbarHeight;
            }

            if (this._options.month.legend) {
                legendHeight = this._options.layout.legendHeight;
            }

            $grid.addClass('hidden');

            //TODO: should this be passed in during resize so the weekView does not have to reach outside itself?
            var calHeight = this._$container.closest('.orcl-calendar').height();

            $grid.css('height', calHeight - toolbarHeight - this._options.month.headerHeight - legendHeight);

            $grid.removeClass('hidden');

            this.redraw();
        }
    };

    return MonthView;
});


define('template-70', ['underscore'], function (_) { return function(data) {
var __t, __p = '';
__p += '<div class="year-container ' +
((__t = ( data.switchToDay ? 'day-switch' : '' )) == null ? '' : __t) +
' ' +
((__t = ( data.switchToMonth ? 'month-switch' : '' )) == null ? '' : __t) +
'"> <div class="row-fluid"> <div class="month span4"> <div class="month-header"> </div> <table class="days"> <thead class="days-of-week"> <tr> <th></th> <th></th> <th></th> <th></th> <th></th> <th></th> <th></th> </tr> </thead> <tbody> <tr class="week"> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> </tr> <tr class="week"> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> </tr> <tr class="week"> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> </tr> <tr class="week"> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> </tr> <tr class="week"> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> </tr> <tr class="week"> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> </tr> </tbody> </table> </div> <div class="month span4"> <div class="month-header"> </div> <table class="days"> <thead class="days-of-week"> <tr> <th></th> <th></th> <th></th> <th></th> <th></th> <th></th> <th></th> </tr> </thead> <tbody> <tr class="week"> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> </tr> <tr class="week"> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> </tr> <tr class="week"> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> </tr> <tr class="week"> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> </tr> <tr class="week"> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> </tr> <tr class="week"> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> </tr> </tbody> </table> </div> <div class="month span4"> <div class="month-header"> </div> <table class="days"> <thead class="days-of-week"> <tr> <th></th> <th></th> <th></th> <th></th> <th></th> <th></th> <th></th> </tr> </thead> <tbody> <tr class="week"> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> </tr> <tr class="week"> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> </tr> <tr class="week"> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> </tr> <tr class="week"> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> </tr> <tr class="week"> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> </tr> <tr class="week"> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> </tr> </tbody> </table> </div> </div> <div class="row-fluid"> <div class="month span4"> <div class="month-header"> </div> <table class="days"> <thead class="days-of-week"> <tr> <th></th> <th></th> <th></th> <th></th> <th></th> <th></th> <th></th> </tr> </thead> <tbody> <tr class="week"> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> </tr> <tr class="week"> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> </tr> <tr class="week"> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> </tr> <tr class="week"> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> </tr> <tr class="week"> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> </tr> <tr class="week"> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> </tr> </tbody> </table> </div> <div class="month span4"> <div class="month-header"> </div> <table class="days"> <thead class="days-of-week"> <tr> <th></th> <th></th> <th></th> <th></th> <th></th> <th></th> <th></th> </tr> </thead> <tbody> <tr class="week"> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> </tr> <tr class="week"> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> </tr> <tr class="week"> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> </tr> <tr class="week"> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> </tr> <tr class="week"> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> </tr> <tr class="week"> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> </tr> </tbody> </table> </div> <div class="month span4"> <div class="month-header"> </div> <table class="days"> <thead class="days-of-week"> <tr> <th></th> <th></th> <th></th> <th></th> <th></th> <th></th> <th></th> </tr> </thead> <tbody> <tr class="week"> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> </tr> <tr class="week"> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> </tr> <tr class="week"> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> </tr> <tr class="week"> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> </tr> <tr class="week"> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> </tr> <tr class="week"> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> </tr> </tbody> </table> </div> </div> <div class="row-fluid"> <div class="month span4"> <div class="month-header"> </div> <table class="days"> <thead class="days-of-week"> <tr> <th></th> <th></th> <th></th> <th></th> <th></th> <th></th> <th></th> </tr> </thead> <tbody> <tr class="week"> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> </tr> <tr class="week"> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> </tr> <tr class="week"> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> </tr> <tr class="week"> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> </tr> <tr class="week"> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> </tr> <tr class="week"> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> </tr> </tbody> </table> </div> <div class="month span4"> <div class="month-header"> </div> <table class="days"> <thead class="days-of-week"> <tr> <th></th> <th></th> <th></th> <th></th> <th></th> <th></th> <th></th> </tr> </thead> <tbody> <tr class="week"> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> </tr> <tr class="week"> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> </tr> <tr class="week"> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> </tr> <tr class="week"> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> </tr> <tr class="week"> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> </tr> <tr class="week"> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> </tr> </tbody> </table> </div> <div class="month span4"> <div class="month-header"> </div> <table class="days"> <thead class="days-of-week"> <tr> <th></th> <th></th> <th></th> <th></th> <th></th> <th></th> <th></th> </tr> </thead> <tbody> <tr class="week"> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> </tr> <tr class="week"> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> </tr> <tr class="week"> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> </tr> <tr class="week"> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> </tr> <tr class="week"> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> </tr> <tr class="week"> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> </tr> </tbody> </table> </div> </div> <div class="row-fluid"> <div class="month span4"> <div class="month-header"> </div> <table class="days"> <thead class="days-of-week"> <tr> <th></th> <th></th> <th></th> <th></th> <th></th> <th></th> <th></th> </tr> </thead> <tbody> <tr class="week"> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> </tr> <tr class="week"> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> </tr> <tr class="week"> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> </tr> <tr class="week"> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> </tr> <tr class="week"> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> </tr> <tr class="week"> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> </tr> </tbody> </table> </div> <div class="month span4"> <div class="month-header"> </div> <table class="days"> <thead class="days-of-week"> <tr> <th></th> <th></th> <th></th> <th></th> <th></th> <th></th> <th></th> </tr> </thead> <tbody> <tr class="week"> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> </tr> <tr class="week"> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> </tr> <tr class="week"> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> </tr> <tr class="week"> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> </tr> <tr class="week"> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> </tr> <tr class="week"> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> </tr> </tbody> </table> </div> <div class="month span4"> <div class="month-header"> </div> <table class="days"> <thead class="days-of-week"> <tr> <th></th> <th></th> <th></th> <th></th> <th></th> <th></th> <th></th> </tr> </thead> <tbody> <tr class="week"> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> </tr> <tr class="week"> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> </tr> <tr class="week"> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> </tr> <tr class="week"> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> </tr> <tr class="week"> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> </tr> <tr class="week"> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> <td><span class="day"></span></td> </tr> </tbody> </div> </div> </div></div>';
return __p
};});

define('oui/widgets/calendar/yearView',[
  'underscore',
  'jquery',
  'moment',
  'template-70'
], function (_, $, moment, yearViewTemplate) {
    'use strict';

    var VIEW_FORMATS = {
        day: 'YYYY-MM-DD',
        month: 'YYYY-MM'
    };

    var SWITCH_ATTR = {
        day: 'numbersSwitchToDayView',
        month: 'monthsSwitchToMonthView'
    };

    /**
     * @method YearView~groupByDay
     *
     * @param {Event} event
     *
     * @returns {String}
     */
    function groupByDay(event) {
        return event.startDate.format(VIEW_FORMATS.day);
    }

    /**
     * @method YearView~getClasses
     *
     * @param {Array} collection
     * @param {Function} fn
     *
     * @returns {Array}
     */
    function getClasses(collection, fn) {
        return _.chain(collection)
        .map(function (item) {
            return fn(item.data);
        })
        .flatten()
        .uniq()
        .value();
    }

    /**
     * @method YearView~getMoment
     * @private
     *
     * @param {jQuery} $elem
     * @param {String} view
     *
     * @returns {moment}
     */
    function getMoment($elem, view) {
        return moment($elem.closest('[data-' + view + ']').attr('data-' + view), VIEW_FORMATS[view]);
    }

    /**
     * @class YearView
     * @private
     *
     * @param {Object} options
     * @param {Object} controller
     * @param {jQuery} $container
     * @param {Object} settings
     */
    function YearView(options, controller, $container, settings) {
        this._controller = controller;
        this._$container = $container;

        this._$element = null;
        this._options = options;
        this._startDate = moment().startOf('year');
        this._setup = false;
        this._settings = settings;
    }

    YearView.prototype = {
        /**
         * @method YearView#setDays
         *
         * @param {moment} startDate
         */
        setDays: function (startDate) {
            var that = this;
            var year = this._controller.getCurrentDate().year();
            var currentDate = startDate.clone().year(year).startOf('year');

            _.each(this._$element.find('.month'), function (month) {

                $(month).find('.month-header')
                  .text(that._options.year.getMonthHeader(currentDate))
                  .attr('data-month', currentDate.format(VIEW_FORMATS.month));
                var curMonth = currentDate.month();

                currentDate.startOf('month').startOf('week');

                _.each($(month).find('.day'), function (day) {

                    $(day).parent()
                        .toggleClass('weekend', currentDate.day() === 0 || currentDate.day() === 6)
                        .toggleClass('diff-month', currentDate.month() !== curMonth);

                    $(day)
                      .attr('data-day', currentDate.format(VIEW_FORMATS.day))
                      .text(currentDate.format('D'));

                    currentDate.add(1, 'day');
                });
            });
        },

        /**
         * @method YearView#setHeaders
         *
         * @param {moment} startDate
         */
        setHeaders: function (startDate) {
            var that = this;
            startDate = startDate.clone();

            _.each(this._$element.find('.days-of-week'), function (header) {
                startDate.startOf('week');

                _.each($(header).find('th'), function (day) {
                    $(day).text(that._options.year.getDayHeader(startDate));
                    startDate.add(1, 'day');
                });
            });
        },

        /**
         * @method YearView#setEvents
         *
         * @param {Array} events
         */
        setEvents: function (events) {
            this._$element.find('.day').attr('class', 'day');

            _(events)
                .groupBy(groupByDay)
                .each(function (events, key) {
                    var classes = getClasses(events, this._options.events.getClasses).join(' ');

                    this._$element.find('[data-day=' + key + ']').addClass(classes);
                }.bind(this))
                .value();

        },

        /**
         * @method YearView#setOverlays
         *
         * @param {Array} overlays
         */
        setOverlays: function (overlays) {
            _(overlays)
                .groupBy(groupByDay)
                .each(function (overlays, key) {
                    var classes = getClasses(overlays, this._options.overlays.getClasses).join(' ');

                    this._$element.find('[data-day=' + key + ']').parent().addClass(classes);
                }.bind(this))
                .value();
        },

        /**
         * @method YearView#prev
         */
        prev: function () {
            var currentDate = this._controller.getCurrentDate();

            this._controller.currentDate(currentDate.clone().subtract(1, 'year'));
        },

        /**
         * @method YearView#next
         */
        next: function () {
            var currentDate = this._controller.getCurrentDate();

            this._controller.currentDate(currentDate.clone().add(1, 'year'));
        },

        /**
         * @method YearView#render
         *
         * @param {moment} currentDate
         */
        render: function (currentDate) {
            var startDate;
            var endDate;

            currentDate = currentDate || this._controller.getCurrentDate();

            startDate = currentDate.clone().startOf('year').startOf('week');
            endDate = currentDate.clone().endOf('year').endOf('week');

            this._startDate = startDate;

            if (!this._setup) {
                this.setupView();
            } else {
                this._$element.removeClass('hidden');
            }

            if (!this._startDate.isSame(currentDate)) {
                this.clean();
            }

            this.setDays(startDate);
            this.setHeaders(startDate);

            $.when(this._controller.getEvents(startDate, endDate))
                .done(this.setEvents.bind(this));

            $.when(this._controller.getOverlays(startDate, endDate))
                .done(this.setOverlays.bind(this));
        },

        /**
         * @method YearView#clean
         */
        clean: function () {
            this._$element.find('.day').removeClass().addClass('day');
        },

        /**
         * @method YearView#setupView
         */
        setupView: function () {
            this._$element = $(yearViewTemplate({
                switchToDay: this._options.year.numbersSwitchToDayView,
                switchToMonth: this._options.year.monthsSwitchToMonthView
            }));

            this._$element.on('click', '[data-day]', this._switchHandler.bind(this, 'day'));

            this._$element.on('click', '[data-month]', this._switchHandler.bind(this, 'month'));

            this._$container.append(this._$element);
            this._setup = true;
        },

        /**
         * @method YearView#_switchHandler
         * @private
         *
         * @param {String} view
         * @param {jQuery.Event} event
         */
        _switchHandler: function (view, event) {
            this._switch(view, getMoment($(event.target), view));
        },

        /**
         * @method YearView#_switch
         * @private
         *
         * @param {String} view
         * @param {moment} date
         */
        _switch: function (view, date) {
            if (this._canSwitch(view)) {
                this._controller.currentDate(date);
                this._$element.trigger('view-change.calendar', view);
            }
        },

        /**
         * @method YearView#_canSwitch
         * @private
         *
         * @param {String} view
         *
         * @returns {Boolean}
         */
        _canSwitch: function (view) {
            return this._options.year[SWITCH_ATTR[view]];
        },

        /**
         * @method YearView#getDateRange
         *
         * @returns {CalendarView#DateRange} dateRange
         */
        getDateRange: function () {
            return {
                start: this._startDate.clone().startOf('year').startOf('week'),
                end: this._startDate.clone().endOf('year').endOf('week')
            };
        },

        /**
         * @method YearView#resize
         */
        resize: function () {
            var toolbarHeight = 0;
            var legendHeight = 0;
            var $grid = this._$element;

            if (this._options.layout.showToolbar) {
                toolbarHeight = this._options.layout.toolbarHeight;
            }

            if (this._options.year.legend) {
                legendHeight = this._options.layout.legendHeight;
            }

            var borders = $grid.outerHeight() - $grid.height();

            $grid.addClass('hidden');

            //TODO: should this be passed in during resize so the weekView does not have to reach outside itself?
            var calHeight = this._$container.closest('.orcl-calendar').height();

            $grid.css('height', calHeight - toolbarHeight - borders - legendHeight);

            $grid.removeClass('hidden');
        },

        /**
         * @method YearView#clear
         */
        clear: function () {
            if (this._$element != null) {
                this._$element.addClass('hidden');
            }
        },

        /**
         * @method YearView#redraw
         */
        redraw: function () {
            this.render(this._controller.getCurrentDate());
        }
    };

    return YearView;
});

define('oui/utils/draggable',[
    'underscore',
    'jquery',
    'oui/utils/math',
    'oui/utils/keyCode'
], function (_, $, math, keyCode) {
    'use strict';

    /**
     * @module {Function} draggable
     * @private
     */

    /**
     * Event to trigger once the drag operation starts. Only triggers if it moves farther than the
     * distance as configurated by [triggerDistance]{@link module:draggable~Config}
     *
     * @event module:draggable#drag-start
     */
    /**
     * Event to trigger once the drag operation ends.
     *
     * @event module:draggable#drag-end
     */
    /**
     * Event to trigger while the drag operation is occurring.
     * @event module:draggable#dragging
     */
    /**
     * Event to trigger when the user cancels the drag operation.
     * @event module:draggable#drag-cancel
     */

    var $body = $('body');

    /**
     * @method module:draggable~matches
     * @private
     *
     * @param {HTMLElement} el
     * @param {String} selector
     *
     * @returns {Boolean}
     */
    var matches = function (el, selector) {
        var p = Element.prototype;
        var f = p.matches || p.webkitMatchesSelector || p.mozMatchesSelector || p.msMatchesSelector || function (s) {
            return [].indexOf.call(document.querySelectorAll(s), this) !== -1;
        };
        return f.call(el, selector);
    };

    /**
     * @typedef {Object} module:draggable~Point
     *
     * @property {Number} x
     * @property {Number} y
     */

    /**
     * @method module:draggable~getCoordinates
     * @private
     *
     * @param {jQuery.Event} event
     *
     * @returns {module:draggable~Point}
     */
    var getCoordinates = function (event) {
        if (_.has(event, 'clientX')) {
            return {
                x: event.clientX,
                y: event.clientY
            };
        } else {
            return {
                x: event.originalEvent.touches[0].clientX,
                y: event.originalEvent.touches[0].clientY
            };
        }
    };

    /**
     * @typedef module:draggable~Config
     *
     * @property {Number} [triggerDistance=0]
     * @property {Object} [throttle]
     * @property {Number} [throttle.move=15]
     * @property {String} [selector]
     * @property {?Object} [$element] The element to use as the handle
     * @property {String} [axis='both'] What axis the element is allowed to be dragged on
     * @property {Object} constrainTo The element to use for bounds calculations
     * @property {Boolean} [cancelOffDocument=false] Whether or not the drag operation should be canceled if they take their
     * mouse off the document.
     * @property {Boolean} [cancellable=false] Whether the drag operation can be cancelled with the ESC key.
     */
    var DEFAULTS = {
        triggerDistance: 0,
        axis: 'both',
        cancelOffDocument: false,
        throttle: {
            move: 15
        },
        cancellable: false
    };

    /**
     * Adds common listeners to the element supplied to support drag and drop operation and triggers events while the
     * operation changes state (start, dragging, end).
     *
     * @function module:draggable
     * @param {Object} $el
     * @param {module:draggable~Config} [options]
     */
    return function ($el, options) {
        var dragging = false;
        var dragStartState;

        options = _.defaults({}, options, DEFAULTS);

        /**
         * @method module:draggable~repositionClone
         * @private
         *
         * @param {*} initialData
         */
        function repositionClone(initialData) {
            initialData.$clone.css({
                'position': 'fixed',
                'margin-left': initialData.offset.left - initialData.position.left,
                'margin-top': initialData.offset.top - initialData.position.top,
                'z-index': 10000,
                'opacity': 0.5
            });
        }

        /**
         * @method module:draggable~getOverElement
         * @private
         *
         * @param {jQuery} $handle
         * @param {jQuery} $clone
         * @param {{x: Number, y: Number}} coordinates
         */
        function getOverElement($handle, $clone, coordinates) {
            var handlePEvents;
            var clonePEvents;
            var overElement;
            if (document.elementsFromPoint) {
                var overAll = document.elementsFromPoint(coordinates.x, coordinates.y);
                overElement = overAll.shift();
                while (overElement === $handle[0] || $handle[0].contains(overElement) || overElement === $clone[0] || $clone[0].contains(overElement)) {
                    overElement = overAll.shift();
                }
            } else {
                handlePEvents = $handle.css('pointer-events') || '';
                clonePEvents = $clone.css('pointer-events') || '';
                $handle.css('pointer-events', 'none');
                $clone.css('pointer-events', 'none');
                overElement = document.elementFromPoint(coordinates.x, coordinates.y);
                $handle.css('pointer-events', handlePEvents);
                $clone.css('pointer-events', clonePEvents);
            }
            return overElement;
        }

        /**
         * @method module:draggable~finish
         * @private
         *
         * @param {jQuery} $handle
         * @param {jQuery} $clone
         * @param {Object} initialData
         * @param {jQuery.Event} origEvent
         */
        function finish($handle, $clone, initialData, origEvent) {
            var coordinates = getCoordinates(origEvent);
            var dropTarget = getOverElement($handle, $clone, coordinates);
            var end = $.Event(['drag-end', _.get(options, 'namespace', '')].join('.'), {
                $handle: $handle,
                dragging: dragging,
                coordinates: coordinates,
                dropTarget: dropTarget,
                deltaX: coordinates.x - initialData.x,
                deltaY: coordinates.x - initialData.x
            });
            $(document).off('.dragging');
            $(window).off('.dragging');
            $body.removeClass('oui-disable-select');
            $el.removeClass('dragging');
            $handle.removeClass('dragging');
            $clone.remove();

            if (dragging) {
                $el.trigger(end, dragStartState);
            }
            dragging = false;
        }

        /**
         * @method module:draggable~cancel
         * @private
         *
         * @param {jQuery} $handle
         * @param {jQuery} $clone
         */
        function cancel($handle, $clone) {
            var end = $.Event(['drag-cancel', _.get(options, 'namespace', '')].join('.'), {
                $handle: $handle
            });
            $(document).off('.dragging');
            $body.removeClass('oui-disable-select');
            $el.removeClass('dragging');
            $handle.removeClass('dragging');
            $clone.remove();

            $el.trigger(end, dragStartState);
        }

        /**
         * @method module:draggable~constrainCoordinates
         * @private
         *
         * @param {module:draggable~Config} options
         * @param {Object} point
         * @param {Object} initialData
         * @param {module:draggable~Point} coordinates
         *
         * @returns {Object}
         */
        function constrainCoordinates(options, point, initialData, coordinates) {
            var top = point.top;
            var left = point.left;
            var $handle = initialData.$handle;

            if (options.constrainTo) {
                var constrainOffset = options.constrainTo.offset();
                var constrainHeight = options.constrainTo.height();
                var constrainWidth = options.constrainTo.width();
                var offsetParentIsConstrain = $handle[0].offsetParent === options.constrainTo[0];

                if (initialData.offset.left + (coordinates.x - initialData.coordinates.x) < constrainOffset.left) {
                    left = 0;
                }

                if (initialData.offset.left + (coordinates.x - initialData.coordinates.x) + $handle.width() > constrainOffset.left + constrainWidth) {
                    if (offsetParentIsConstrain) {
                        left = constrainWidth - $handle.width();
                    } else {
                        left = constrainOffset.left + constrainWidth - $handle.width();
                    }
                }

                if (initialData.offset.top + (coordinates.y - initialData.coordinates.y) < constrainOffset.top) {
                    top = 0;
                }

                if (initialData.offset.top + (coordinates.y - initialData.coordinates.y) + $handle.height() > constrainOffset.top + constrainHeight) {
                    if (offsetParentIsConstrain) {
                        top = constrainHeight - $handle.height();
                    } else {
                        top = constrainOffset.top + constrainHeight - $handle.height();
                    }
                }

            }

            return {
                top: top,
                left: left
            };
        }

        /**
         * @method module:draggable~triggerDraggingEvent
         * @private
         *
         * @param {*} opts
         */
        function triggerDraggingEvent(opts) {
            var draggingEvent = $.Event(['dragging', _.get(options, 'namespace', '')].join('.'), {
                dragData: opts.startEvent.dragData,
                start: opts.initialData.coordinates,
                current: opts.coordinates,
                $handle: opts.$handle,
                deltaX: opts.coordinates.x - opts.initialData.coordinates.x,
                deltaY: opts.coordinates.y - opts.initialData.coordinates.y,
                originalEvent: opts.event
            });
            $el.trigger(draggingEvent, dragStartState);
            return draggingEvent;
        }

        /**
         * @method module:draggable~triggerDraggingOverEvent
         * @private
         *
         * @param {*} opts
         */
        function triggerDraggingOverEvent(opts) {
            var overElement = getOverElement(opts.$handle, opts.$clone, opts.coordinates);
            var draggingOverEvent = $.Event(['dragging-over', _.get(options, 'namespace', '')].join('.'), {
                dropTarget: overElement,
                current: opts.coordinates,
                $handle: opts.$handle,
                originalEvent: opts.event
            });
            $el.trigger(draggingOverEvent, dragStartState);
            return draggingOverEvent;
        }

        /**
         * @method module:draggable~setupClone
         *
         * @param {Object} initialData
         */
        function setupClone(initialData) {
            var $clone = initialData.$clone;
            if (options.clone) {
                $clone.appendTo(options.appendCloneTo || options.$element);
                $clone.addClass('dragging');
            }
            if (options.positionFixedClone) {
                repositionClone(initialData);
            }
            if (options.modifyClone) {
                options.modifyClone($clone, _.pick(initialData, 'coordinates', 'offset', 'position'));
            }
        }

        /**
         * @method module:draggable~moveHandler
         *
         * @param {Object} initialData
         * @param {jQuery.Event} event
         */
        function moveHandler(initialData, event) {
            var coordinates = getCoordinates(event);
            var left = initialData.position.left + (coordinates.x - initialData.coordinates.x);
            var top = initialData.position.top + (coordinates.y - initialData.coordinates.y);
            var draggingEvent;
            var $handle = initialData.$handle;
            var $clone = initialData.$clone;

            if (dragging || math.distance(coordinates, initialData.coordinates) > options.triggerDistance) {
                var startEvent = initialData.start(initialData);

                if (startEvent.isDefaultPrevented()) {
                    return;
                }

                $handle.addClass('dragging');
                dragging = true;

                setupClone(initialData);

                var constrained = constrainCoordinates(options, {
                    left: left,
                    top: top
                }, initialData, coordinates);

                left = constrained.left;
                top = constrained.top;

                if (options.axis === 'y') {
                    left = initialData.position.left;
                } else if (options.axis === 'x') {
                    top = initialData.position.top;
                }

                var eventOpts = {
                    startEvent: startEvent,
                    initialData: initialData,
                    coordinates: coordinates,
                    $handle: $handle,
                    $clone: $clone,
                    event: event
                };
                draggingEvent = triggerDraggingEvent(eventOpts);
                triggerDraggingOverEvent(eventOpts);

                if (!draggingEvent.isDefaultPrevented()) {
                    $clone.css({
                        left: left,
                        top: top
                    });
                    if (!options.staticHandle) {
                        $handle.css({
                            left: left,
                            top: top
                        });
                    }
                }
            }
        }

        $el.on('mousedown.draggable touchstart.draggable', function (event) {
            var initialCoordinates = getCoordinates(event);
            var moving;
            var $handle = options.$element;
            var $clone;
            var matched = matches(event.target, options.selector);

            /**
             * @method endMove
             * @param {jQuery.Event} ev
             */
            function endMove(ev) {
                moving.cancel();
                finish($handle, $clone, initialCoordinates, ev);
            }

            if (!matched) {
                matched = $(event.target).closest(options.selector).length !== 0;
                $handle = $(event.target).closest(options.selector);
            }

            if (event.which === 1 && matched) {
                $body.addClass('oui-disable-select');
                $(document).off('.dragging'); //Make sure there are not any pre-existing dragging listeners
                if ($handle == null) {
                    $handle = $(event.target);
                }

                $clone = $handle.clone();
                $clone.css({
                    width: $handle.width(),
                    height: $handle.height()
                });

                var initialData = {
                    $handle: $handle,
                    $clone: $clone,
                    position: $handle.position(),
                    offset: (function () {
                        var offset = $handle.offset();
                        return {
                            top: offset.top - $(window).scrollTop(),
                            left: offset.left - $(window).scrollLeft()
                        };
                    }()),
                    coordinates: initialCoordinates,
                    start: _.once(function (startData) {
                        var startEvent = $.Event(['drag-start', _.get(options, 'namespace', '')].join('.'), _.extend({}, { originalEvent: event }, startData));
                        $el.trigger(startEvent);
                        dragStartState = startEvent.result;

                        return startEvent;
                    })
                };
                moving = _.throttle(moveHandler.bind(null, initialData), options.throttle.move);

                $(document)
                    .on('keydown.dragging', function (event) {
                        if (event.which === keyCode.ESC && options.cancellable) {
                            cancel($handle, $clone, event);
                        }
                    })
                    .on('mousemove.dragging touchmove.dragging', moving)
                    .on('mouseout.dragging touchout.dragging', function (event) {
                        var target = event.relatedTarget || event.toElement;
                        if (options.cancelOffDocument && (target == null || target.nodeName === 'HTML')) {
                            endMove(event);
                        }
                    })
                    .on('mouseup.dragging touchend.dragging', endMove);

                if (options.positionFixedClone) {
                    $(window).on('scroll.dragging', repositionClone.bind(null, initialData));
                }
            }
        });

    };
});


define('template-71', ['underscore'], function (_) { return function(data) {
var __t, __p = '', __j = Array.prototype.join;
function print() { __p += __j.call(arguments, '') }
__p += '<div class="task-container day-time-container"> <div class="task-header day-time-header"> <table> <tr class="days"> ';
 for (var i = 0; i < data.visibleDays; i++) { ;
__p += ' <th class="dow"><span class="view-day"></span></th> ';
 } ;
__p += ' <th class="scroll-fix"></th> </tr> </table> <div class="drag-area"> <table class="drag-area-table"><tbody></tbody></table> </div> </div> <div class="task-grid-overlays"> <div class="task-overlay gutter-overlay"> <table> <tr class="gutters"> ';
 for (var i = 0; i < data.visibleDays; i++) { ;
__p += ' <td class="gutter"></td> ';
 } ;
__p += ' </tr> </table> </div> <div class="task-overlay drag-overlay"> <span></span> </div> </div> <div class="task-grid"> <table class="task-table"><tbody></tbody></table> </div></div>';
return __p
};});

define('oui/widgets/calendar/taskDayView',['require','underscore','jquery','moment','oui/utils/draggable','oui/utils/tasks/tasks','template-71','oui/utils/math','oui/widgets/calendar/event','oui/utils/scrollbarWidth'],function (require) {
    'use strict';

    var _ = require('underscore');
    var $ = require('jquery');
    var moment = require('moment');
    var draggable = require('oui/utils/draggable');
    var taskUtil = require('oui/utils/tasks/tasks');
    var taskViewTemplate = require('template-71');
    var math = require('oui/utils/math');
    var Event = require('oui/widgets/calendar/event');
    var getScrollbarWidth = require('oui/utils/scrollbarWidth');

    var DOY_FORMAT = 'YYYY-MM-DD';

    /**
     * @class TaskDayView
     * @mixes <tasks>
     * @private
     *
     * @param {Calendar#Config} options
     * @param {Object} controller
     * @param {jQuery} $container
     * @param {Object} settings
     */
    function TaskDayView(options, controller, $container, settings) {
        this._controller = controller;
        this._$container = $container;
        this._options = options;
        this._settings = settings;
        this._startDate = moment().startOf(this._getMomentAttribute());
        this._$element = null;
        this._setup = false;
        this._taskUtil = taskUtil(this.getTaskUtilOptions(), this.getElement.bind(this));
    }

    TaskDayView.prototype = _.extend({}, {
        /**
         * @method TaskDayView#_setStore
         * @private
         */
        _setStore: function () {
            var options = this._getOptions();
            options.renderer._setStore(this.getMediator());
        },
        /**
         * @method TaskDayView#_initListeners
         * @private
         */
        _initListeners: function () {
            this._$element.on('click', '.selectable-group > button.collapse-icon', this.toggleGroup.bind(this));
            this._$element.on('mousedown', '.selectable-group > button.collapse-icon', function (ev) { ev.preventDefault(); });
            this._$element.on('mouseover', '.group', this.highlightGroup.bind(this, true));
            this._$element.on('mouseout', '.group', this.highlightGroup.bind(this, false));
            this._$element.on('mousedown', '.task', this.focusTask.bind(this));
            this._$element.on('dblclick', 'td:not(.task, .group)', this.clickHandler.bind(this));
            this._$element.on('dblclick', '.task-grid', this.clickHandler.bind(this));
            this._draggableListeners();
        },
        /**
         * @method TaskDayView#_draggableListeners
         * @private
         */
        _draggableListeners: function () {
            var $document = $(document.body);
            draggable(this._$element, {
                clone: true,
                appendCloneTo: $document,
                positionFixedClone: true,
                staticHandle: true,
                triggerDistance: 10,
                selector: '.task.active',
                cancellable: true,
                namespace: 'move'
            });
            draggable(this._$element, {
                constrainTo: this._$taskGrid,
                triggerDistance: 10,
                selector: '.task.active .resize-handle',
                cancellable: true,
                axis: 'x',
                namespace: 'resize'
            });
            this._$element.on('drag-start.move', this.beginMove.bind(this));
            this._$element.on('dragging.move', this.dragMove.bind(this));
            this._$element.on('drag-end.move', this.dragEnd.bind(this));
            this._$element.on('drag-start.resize', this.beginResize.bind(this));
            this._$element.on('dragging.resize', this.dragResize.bind(this));
            this._$element.on('drag-end.resize', this.endResize.bind(this));
            this._$element.on('drag-end.move drag-end.resize', this.clearDraggableStyle.bind(this));

        },
        /**
         * @method TaskDayView#setDragData
         */
        setDragData: function () {
            this.dragData = {
                rowWidth: this.getRowWidth()
            };
        },
        /**
         * @method TaskDayView#beginMove
         * @param {Object} draggableEvent
         */
        beginMove: function (draggableEvent) {
            if (!this._controller.canEditTask()) {
                draggableEvent.preventDefault();
            }
            var active = this.getActiveTask();
            var viewRange = this.getDateRange();
            var start = active.startDate <= viewRange.start ? viewRange.start.clone() : active.startDate.clone();
            var end = active.endDate >= viewRange.end ? viewRange.end.clone() : active.endDate.clone();
            this.setDragData();
            return {
                id: active.id,
                startDate: start,
                endDate: end,
                active: active,
                duration: this._taskUtil.getDuration(active.endDate, active.startDate)
            };
        },
        /**
         * @method TaskDayView#endMove
         * @param {Object} draggableEvent
         * @param {Object} startState
         */
        dragMove: function (draggableEvent, startState) {
            var range = this.getDateRange();
            var startDate = range.start.clone();
            var rowWidth = this.dragData.rowWidth;
            var deltaX = (Math.round(draggableEvent.deltaX / (rowWidth / this._getVisibleDays())));
            var deltaDays = this._taskUtil.getDuration(startState.startDate, startDate) + deltaX;
            var draggedStartDate = startState.active.startDate.clone().add(deltaX, 'day');
            var draggedEndDate = draggedStartDate.clone().add(startState.duration, 'day');
            if ((draggedStartDate >= range.start || draggedEndDate > range.start) &&
                (draggedStartDate < range.end || draggedEndDate <= range.end) &&
                this._currentDragDelta !== deltaDays) {
                this._temporaryRender(this.getActiveTask(), deltaDays, startState.duration);
                this._currentDragDelta = deltaDays;
            }
        },
        /**
         * @method TaskDayView#dragOut
         */
        dragOut: function () {
            var result = this._controller._model.trigger('task-remove', {
                task: this.getActiveTask().data
            });
            $.when(result).done(function () {
                this._controller.clean();
                this.redraw();
                this._controller._model.trigger('repopulate-hopper.view');
            }.bind(this));
        },
        /**
         * @method TaskDayView#dragEnd
         * @param {Object} dragEndEvent
         * @param {Object} startState
         */
        dragEnd: function (dragEndEvent, startState) {
            var item = this.getActiveTask();
            var delta = (Math.round(dragEndEvent.deltaX / (this.dragData.rowWidth / this._getVisibleDays())));
            var $dropTarget = $(dragEndEvent.dropTarget);

            // If we're not dropping back into the calendar
            if ($dropTarget.closest('.task-container').closest(this._$element).length === 0) {
                var event = $.Event('oui-calendar-drop', {
                    item: item.data
                });
                $dropTarget.trigger(event);
                if (!event.isDefaultPrevented()) {
                    this.dragOut();
                } else {
                    this._controller.clean();
                    this.redraw();
                }
            } else {
                if (delta !== 0) {
                    this.updateTask({
                        startDate: startState.active.startDate.clone().add(delta, 'day'),
                        endDate: startState.active.endDate.clone().add(delta, 'day')
                    });
                } else {
                    this.redraw();
                }
            }
        },
        /**
         * @method TaskDayView#beginResize
         * @param {jQuery.Event} draggableEvent
         */
        beginResize: function (draggableEvent) {
            if (!this._controller.canEditTask()) {
                draggableEvent.preventDefault();
            }
            var active = this.getActiveTask();
            var resizeKey = draggableEvent.$handle.attr('data-resize');
            var initialDate = active[resizeKey].clone();
            this.setDragData();
            return {
                resizeKey: resizeKey,
                initialDate: initialDate
            };
        },
        /**
         * @method TaskDayView#dragResize
         * @param {jQuery.Event} draggableEvent
         * @param {Object} startState
         */
        dragResize: function (draggableEvent, startState) {
            var delta = (Math.round(draggableEvent.deltaX / (this.dragData.rowWidth / this._getVisibleDays())));
            var activeTask = this.getActiveTask();
            var temporaryItem = {
                id: activeTask.id,
                startDate: activeTask.startDate.clone(),
                endDate: activeTask.endDate.clone(),
                data: activeTask.data
            };

            temporaryItem[startState.resizeKey] = startState.initialDate.clone().add(delta, 'day');

            if (temporaryItem.startDate.isBefore(temporaryItem.endDate)) {
                this._removeTemporary();
                this._addTemporary(temporaryItem);
            }
        },
        /**
         * @method TaskDayView#endResize
         * @param {jQuery.Event} draggableEvent
         * @param {Object} startState
         */
        endResize: function (draggableEvent, startState) {
            var rowWidth = this.dragData.rowWidth;
            var delta = (Math.round(draggableEvent.deltaX / (rowWidth / this._getVisibleDays())));
            var update = {};

            update[startState.resizeKey] = startState.initialDate.clone().add(delta, 'day');

            this._removeTemporary();
            this.updateTask(update);
            this.redraw();
        },

        /**
         * @method TaskDayView#clearDraggableStyle
         * @param {jQuery.Event} draggableEvent
         */
        clearDraggableStyle: function (draggableEvent) {
            draggableEvent.$handle.css({
                left: 0,
                top: 0
            });
        },

        /**
         * @method TaskDayView#toggleGroup
         * @param {jQuery.Event} ev
         */
        toggleGroup: function (ev) {
            var $group = $(ev.target).closest('.group');
            var groupId = $group.attr('data-group-id');
            var isCollapsed = this._controller.toggleGroup(groupId);
            this._taskUtil.toggleGroup(groupId, isCollapsed);
        },

        /**
         * @method TaskDayView#getActiveTask
         */
        getActiveTask: function () {
            return this._controller.getEvent(this._controller.active());
        },

        /**
         * @method TaskDayView#focusTask
         * @param {Object} ev
         */
        focusTask: function (ev) {
            var $task = $(ev.currentTarget);
            var taskStart = $task.position().top;
            var calStart = this._$taskGrid.scrollTop();

            this._controller.active($task.attr('data-event-id'), taskStart <= calStart);
        },

        /**
         * @method TaskDayView#getMediator
         * @returns {Object}
         */
        getMediator: function () {
            return this._controller.getMediator();
        },

        /**
         * @method TaskDayView#_getOptions
         * @private
         *
         * @returns {Object} dayOptions
         */
        _getOptions: function () {
            return this._options['task-day'];
        },

        /**
         * @method TaskDayView#getEventOptions
         * @param {HTMLElement} target
         */
        getEventOptions: function (target) {
            var mediator = this.getMediator();
            var task = null;
            if (_.isFunction(mediator.getTask)) {
                task = mediator.getTask($(target).closest('[data-event-id]').attr('data-event-id'));
            }
            return {
                task: task
            };
        },

        /**
         * @method TaskDayView#getTaskUtilOptions
         */
        getTaskUtilOptions: function () {
            var options = this._getOptions();
            return {
                getAppendTo: this.getAppendTo.bind(this),
                getTable: this.getTable.bind(this),
                getStartDate: this.getStartDate.bind(this),
                getVisibleDays: this._getVisibleDays.bind(this),
                getMomentAttribute: this._getMomentAttribute.bind(this),
                renderer: options.renderer,
                groupRenderer: options.groupRenderer,
                eventKeys: this._options.eventKeys,
                locale: this._options.locale,
                mediator: this.getMediator()
            };
        },

        /**
         * @method TaskDayView#getElement
         * @returns {jQuery}
         */
        getElement: function () {
            return this._$element;
        },

        /**
         * @method TaskDayView#getElement
         * @returns {jQuery}
         */
        getTable: function () {
            return this._$tableBody;
        },

        /**
         * @method TaskDayView#getElement
         * @returns {jQuery}
         */
        getAppendTo: function () {
            return this._$taskTable;
        },

        /**
         * @method TaskDayView#getRowWidth
         */
        getRowWidth: function () {
            return this._$element.find('tr').last()[0].getBoundingClientRect().width;
        },

        /**
         * @method TaskDayView#_getMomentAttribute
         * @private
         *
         * @returns {String} 'day'
         */
        _getMomentAttribute: function () {
            return 'day';
        },

        /**
         * @method TaskDayView#_getVisibleDays
         * @private
         *
         * @returns {Number} 1
         */
        _getVisibleDays: function () {
            return 1;
        },

        /**
         * @method TaskDayView#getStartDate
         * @returns {moment}
         */
        getStartDate: function () {
            return this._startDate;
        },

        /**
         * @method TaskDayView#getDateRange
         *
         * @returns {CalendarView#DateRange} dateRange
         */
        getDateRange: function () {
            return {
                start: this._startDate,
                end: this._startDate.clone().endOf(this._getMomentAttribute())
            };
        },

        /**
         * @method TaskDayView#taskIsOccurring
         * @param {Calendar#Event} task
         */
        taskIsOccurring: function (task) {
            var dateRange = this.getDateRange();
            return task.isOccurring(dateRange.start, dateRange.end);
        },

        /**
         * @method TaskDayView#setupView
         */
        setupView: function () {
            this._$element = $(taskViewTemplate({ visibleDays: this._getVisibleDays() }));
            this._$element.find('.scroll-fix').css({
                width: getScrollbarWidth()
            });
            this._$container.append(this._$element);
            this._cacheElements();
            this._setStore();
            this._initListeners();
            this._setup = true;
        },

        /**
         * @method TaskDayView#_cacheElements
         * @private
         */
        _cacheElements: function () {
            this._$calendar = this._$container.closest('.orcl-calendar');
            this._$dayOfWeekHeaders = this._$element.find('.task-header .days th');
            this._$gutters = this._$element.find('td.gutter');
            this._$dragArea = this._$element.find('.drag-area');
            this._$dragAreaTable = this._$dragArea.find('.drag-area-table tbody');
            this._$taskGrid = this._$element.find('.task-grid');
            this._$overlays = this._$element.find('.task-grid-overlays');
            this._$taskTable = this._$taskGrid.find('.task-table');
            this._$tableBody = this._$taskTable.find('tbody');
        },

        /**
         * @method TaskDayView#setHeaders
         *
         * @param {moment} start
         */
        setHeaders: function (start) {
            var today = moment().startOf('day');
            var date = start.clone();
            var options = this._getOptions();

            _.each(this._$dayOfWeekHeaders, function (day) {
                var format = options.getHeader(date.clone());
                $(day).find('.view-day')
                    .text(format)
                    .attr('title', format)
                    .attr('data-doy', date.format(DOY_FORMAT));
                $(day).toggleClass('today', options.highlightToday && date.clone().startOf('day').isSame(today));
                date.add(1, 'day');
            });
        },

        /**
         * @method TaskDayView#setGutters
         *
         * @param {moment} start
         */
        setGutters: function (start) {
            var date = start.clone();
            var disabledDays = this._controller.getDisabledDays(this.getDateRange());
            var $day;

            _.each(this._$gutters, function (day) {
                $day = $(day);
                $day.attr('data-doy', date.format(DOY_FORMAT));
                $day.toggleClass('disabled-day', _.contains(disabledDays, date.day()));
                date.add(1, 'day');
            });
        },

        /**
         * @method TaskDayView#setEvents
         *
         * @param {Array} events
         * @param {Array} groups
         */
        setEvents: function (events, groups) {
            var activeId = this._controller.active();
            var $active;
            this.clean();
            this._removeTemporary();
            this._taskUtil.addAllTasks(events, groups, {
                groupArgs: this._controller.getTaskGroups(),
                activeId: activeId
            });
            $active = this.findTask(activeId);
            if ($active.length) {
                this.scrollIntoView(activeId, true);
            }
        },

        /**
         * @method TaskDayView#redrawGroup
         *
         * @param {String} groupId
         * @param {Array} events
         * @param {Array} groups
         */
        redrawGroup: function (groupId, events, groups) {
            this._removeTemporary();
            this._taskUtil.redrawGroup(events, groups, groupId, this._controller.getTaskGroups());
            this.active(this._controller.active());
        },

        /**
         * @method TaskDayView#forceRedraw
         */
        forceRedraw: function () {
            this._currentDate = null;
        },

        /**
         * @method TaskDayView#clean
         */
        clean: function () {
            this._taskUtil.removeAllRows();
        },

        /**
         * @method TaskDayView#clean
         */
        clear: function () {
            if (this._$element != null) {
                this._$element.addClass('hidden');
            }
        },

        /**
         * @method TaskDayView#active
         * @param {Number|String} activeId
         * @param {Boolean} [top]
         */
        active: function (activeId, top) {
            this._$element.find('.task').removeClass('active');
            if (activeId) {
                this.findTask(activeId).addClass('active');
            }
            if (top != null) {
                this.scrollIntoView(activeId, top);
            }
        },

        /**
         * @method TaskDayView#findTask
         * @param {Number|String} id
         * @returns {jQuery}
         */
        findTask: function (id) {
            return this._$element.find('.task[data-event-id=' + id + ']');
        },

        /**
         * @method TaskDayView#scrollIntoView
         * @param {Number|String} id
         * @param {Boolean} top
         */
        scrollIntoView: function (id, top) {
            var $task = this.findTask(id);
            var taskStart;
            var viewStart;
            var taskEnd;
            var viewEnd;
            var scrollTo;
            var task = this.getActiveTask();

            if (task && !this.taskIsOccurring(task)) {
                return $.when(this._controller.currentDate(task.startDate)).done(this.active.bind(this, id, top));
            }
            if ($task.length) {
                taskStart = $task.position().top;
                viewStart = this._$taskGrid.scrollTop();
                taskEnd = taskStart + $task.height();
                viewEnd = viewStart + this._$taskGrid.height();
                if (taskStart <= viewStart || taskEnd >= viewEnd) {
                    if (top) {
                        scrollTo = taskStart;
                    } else {
                        scrollTo = viewStart + (taskEnd - viewEnd);
                    }
                    this._$taskGrid.scrollTop(scrollTo);
                }
            }
        },

        /**
         * @method TaskDayView#render
         *
         * @param {moment} currentDate
         */
        render: function (currentDate) {
            currentDate = currentDate || this._controller.getCurrentDate();
            var range = this._getStartAndEndDate(currentDate);

            if (!this._setup) {
                this.setupView();
            } else {
                this._$element.removeClass('hidden');
            }

            if (this._currentDate == null || !this._currentDate.isSame(currentDate)) {
                this._currentDate = currentDate.clone();
                this.setHeaders(range.startDate);
                this.setGutters(range.startDate);
            }

            this.getItems(range.startDate, range.endDate).done(this.setEvents.bind(this));
        },

        /**
         * @method TaskDayView#renderGroup
         * @param {String} groupId
         */
        renderGroup: function (groupId) {
            var currentDate = this._controller.getCurrentDate();
            var range = this._getStartAndEndDate(currentDate);

            if (this._currentDate == null || !this._currentDate.isSame(currentDate)) {
                this._currentDate = currentDate.clone();
                this.setHeaders(range.startDate);
                this.setGutters(range.startDate);
            }

            this.getItems(range.startDate, range.endDate).done(this.redrawGroup.bind(this, groupId));
        },

        /**
         * @method TaskDayView#getItems
         * @param {moment} startDate
         * @param {moment} endDate
         * @returns {$.Deferred}
         */
        getItems: function (startDate, endDate) {
            return $.when(this._controller.getEvents(startDate, endDate));
        },

        /**
         * @method TaskDayView#redraw
         * @param {moment} startDate
         */
        redraw: function (startDate) {
            var currentDate = this._controller.getCurrentDate();
            var range = this._getStartAndEndDate(currentDate);
            startDate = startDate || range.startDate;
            this.render(startDate);
        },

        /**
         * @method TaskDayView#_getStartAndEndDate
         * @param {moment} currentDate
         * @returns {{startDate: moment, endDate: moment}}
         * @private
         */
        _getStartAndEndDate: function (currentDate) {
            currentDate = currentDate || this._controller.getCurrentDate();
            var startDate = currentDate.clone().startOf(this._getMomentAttribute());
            var endDate = startDate.clone().endOf(this._getMomentAttribute());
            this._startDate = startDate;
            return {
                startDate: startDate,
                endDate: endDate
            };
        },

        /**
         * @method TaskDayView#resize
         */
        resize: function () {
            var toolbarHeight = 0;
            var calHeight = this._$calendar.height();
            var borders = this._$taskGrid.outerHeight() - this._$taskGrid.height();

            if (this._options.layout.showToolbar) {
                toolbarHeight = this._options.layout.toolbarHeight;
            }
            calHeight = calHeight - toolbarHeight - this._getOptions().headerHeight - borders;

            this._$taskGrid.addClass('hidden');
            this._$taskGrid.css('height', calHeight);
            this._$overlays.css('height', calHeight);
            this._$taskGrid.removeClass('hidden');
        },

        /**
         * @method TaskDayView#clickHandler
         * @param {jQuery} ev
         */
        clickHandler: function (ev) {
            var $currentTarget = $(ev.currentTarget);
            var position = $currentTarget.position();
            var group = null;
            var startDate;
            var $group;
            var $allGroups = this._$element.find('.group');

            if ($currentTarget.hasClass('task-grid') && !$(ev.target).is($currentTarget)) {
                return;
            }
            ev.stopPropagation();
            startDate = ev.currentTarget.getAttribute('data-doy');
            if ($allGroups.length && $currentTarget.hasClass('task-grid')) {
                group = $allGroups.last().attr('data-group-id');
            } else {
                _.each($allGroups, function (groupEl) {
                    $group = $(groupEl);
                    if (position.top >= $group.position().top) {
                        group = $group.attr('data-group-id');
                    }
                });
            }
            if (startDate) {
                startDate = moment(startDate);
            } else {
                startDate = this.findDay(this.findProgress(ev.offsetX));
            }
            this.addTask(startDate, group);
        },

        /**
         * @method TaskDayView#addTask
         * @param {moment} startDate
         * @param {Number|String} groupId
         */
        addTask: function (startDate, groupId) {
            this._controller.newTask(startDate, groupId);
        },

        /**
         * @method TaskDayView#updateTask
         * @param {Object<field, values>} fields
         */
        updateTask: function (fields) {
            var taskStart;
            this._controller.updateTask(fields);
            taskStart = this.getActiveTask().startDate;
            if (!this.taskIsOccurring(this.getActiveTask())) {
                this._controller.currentDate(taskStart);
            }
        },

        /**
         * @method TaskDayView#update
         * @param {Object} args
         */
        update: function (args) {
            var groupId = _.get(args.task.data, this._options.eventKeys.parentKey);
            this._controller.getGroups().done(function (groups) {
                if (!_.isUndefined(groupId) && groups.length) {
                    this.renderGroup(groupId);
                } else {
                    this.render();
                }
            }.bind(this));
        },

        /**
         * @method TaskDayView#findProgress
         * @param {Number} offsetPosition
         * @returns {Number}
         */
        findProgress: function (offsetPosition) {
            return offsetPosition / this._$element.width();
        },

        /**
         * @method TaskDayView#findDay
         * @param {Number} progress
         * @returns {moment}
         */
        findDay: function (progress) {
            var day = Math.floor(progress * this._getVisibleDays());
            return this.getStartDate().clone().add(day, 'day');
        },

        /**
         * @method TaskDayView#addEvent
         */
        addEvent: function () {
            this.render();
        },

        /**
         * @method TaskDayView#next
         */
        next: function () {
            var currentDate = this._controller.getCurrentDate();
            this._controller.currentDate(currentDate.clone().add(1, this._getMomentAttribute()));
        },

        /**
         * @method TaskDayView#prev
         */
        prev: function () {
            var currentDate = this._controller.getCurrentDate();

            this._controller.currentDate(currentDate.clone().subtract(1, this._getMomentAttribute()));
        },

        /**
         * @method TaskDayView#highlightGroup
         * @param {Boolean} toggle
         * @param {jQuery.Event} event
         */
        highlightGroup: function (toggle, event) {
            var groupId = event.currentTarget.getAttribute('data-group-id');
            var $tasksInGroup = this._$element.find('.task[data-group-id=' + groupId + ']');

            $tasksInGroup.toggleClass('hovering', toggle);
        },

        /**
         * @method TaskDayView#highlightNext
         */
        highlightNext: function () {
            this.highlight(1, false);
        },

        /**
         * @method TaskDayView#highlightPrev
         */
        highlightPrev: function () {
            this.highlight(-1, true);
        },

        /**
         * @method TaskDayView#highlight
         * @param {Number} delta
         * @param {Boolean} top
         */
        highlight: function (delta, top) {
            var $tasks = this._$element.find('.task:visible');
            var activeId = this._controller.active();
            var $activeTask = $tasks.filter('[data-event-id=' + activeId + ']');
            var activeIndex = $tasks.index($activeTask);

            activeIndex = this.findActiveIndex(activeIndex, delta, $tasks.length);
            activeId = $tasks[activeIndex].getAttribute('data-event-id');
            this._controller.active(activeId, top);
        },

        /**
         * @method TaskDayView#_highlightGutter
         * @private
         * @param {moment} start
         * @param {moment} end
         */
        _highlightGutter: function (start, end) {
            var $highlight = $();
            var curDay = start.clone();
            for (var i = 0; i < this._taskUtil.getDuration(end, start); i++) {
                $highlight = $highlight.add(this._$gutters.filter('[data-doy=' + curDay.format(DOY_FORMAT) + ']'));
                curDay.add(1, 'day');
            }
            this._$gutters.removeClass('highlighted-day');
            $highlight.addClass('highlighted-day');
        },

        /**
         * @method TaskDayView#calculateActiveIndex
         * @param {Number} activeIndex
         * @param {Number} delta
         * @param {Number} numTasks
         */
        findActiveIndex: function (activeIndex, delta, numTasks) {
            if (activeIndex === -1 && delta === -1) {
                return numTasks - 1;
            }
            return math.mod(activeIndex + delta, numTasks);
        },

        /**
         * Renders a temporary item on the view
         *
         * @method TaskDayView#_renderTemporaryItem
         * @param {Object} item
         * @param {{pageX,pageY}} position
         * @returns {{startDate: moment, endDate: moment}}
         * @private
         */
        _renderTemporaryItem: function (item, position) {
            if (this.dragData == null) {
                this.setDragData();
            }

            var rowWidth = this.dragData.rowWidth;
            var offset = this._$element.offset();
            var relativeX = position.pageX - offset.left;
            var visibleDays = this._getVisibleDays();
            var delta = (Math.floor(relativeX / (rowWidth / visibleDays)));
            delta = (delta < 0) ? 0 : (delta > (visibleDays - 1)) ? (visibleDays - 1) : delta;
            return this._temporaryRender({
                data: item,
                id: item.id
            }, delta, this._controller._getTaskDuration(item));
        },

        /**
         * Handles rendering when dragging events on the calendar or dragging in from the hopper
         *
         * @method TaskDayView#_temporaryRender
         * @param {{data: Object}} event
         * @param {Number} delta Offset from first visible day in view
         * @param {Number} duration Length in days of the event
         * @returns {{startDate: moment, endDate: moment}}
         * @private
         */
        _temporaryRender: function (event, delta, duration) {
            var range = this.getDateRange();
            var startDate;
            var endDate;
            var displayItem;

            //create temporary event range regardless of view
            if (moment.isMoment(event.startDate) && event.startDate.isBefore(range.start)) {
                startDate = event.startDate.clone().add(delta, 'day');
            } else {
                startDate = range.start.clone().add(delta, 'day');
            }
            endDate = startDate.clone().add(duration, 'day');

            //truncate start or end dates out of view
            if (startDate <= range.start) {
                startDate = range.start.clone();
            }
            if (endDate >= range.end) {
                endDate = range.end.clone();
            }

            displayItem = {
                id: event.id,
                startDate: startDate,
                endDate: endDate,
                data: event.data,
                modified: Number.MAX_VALUE
            };

            this._removeTemporary();
            this._addTemporary(displayItem);

            return _.pick(displayItem, 'startDate', 'endDate');
        },

        /**
         * @method TaskDayView#_addTemporary
         * @param {Object} displayItem
         * @private
         */
        _addTemporary: function (displayItem) {
            var startDate = displayItem.startDate;
            var eventsByDay = {};
            var $temporaryRow;

            if (this.taskIsOccurring(new Event(displayItem))) {
                this._highlightGutter(startDate, displayItem.endDate);
                eventsByDay[startDate.day()] = [displayItem];
                $temporaryRow = $(this._taskUtil.buildTaskRows(eventsByDay)).addClass('temporary-row');
                $temporaryRow.find('td').addClass('active');
                this._$taskGrid.addClass('drag-active');
                this._$dragAreaTable.append($temporaryRow);
                this._$dragArea.addClass('drag-active').height(this._$dragAreaTable.outerHeight());
            }
        },

        /**
         * @method TaskDayView#_removeTemporary
         * @private
         */
        _removeTemporary: function () {
            this._$dragAreaTable[0].innerHTML = '';
            this._$dragArea.removeClass('drag-active').height(0);
            this._$gutters.removeClass('highlighted-day');
            this._$taskGrid.removeClass('drag-active');
        },
        /**
         * @method TaskDayView#callOut
         */
        callOut: function () {
            var toCallout = this._$element.find('.task.active');

            this._$element.one('animationend', function () {
                toCallout.removeClass('oui-callout');
            });
            toCallout.addClass('oui-callout');
        }
    });

    return TaskDayView;
});

define('oui/widgets/calendar/taskWeekView',['require','jquery','underscore','oui/widgets/calendar/taskDayView'],function (require) {
    'use strict';

    var $ = require('jquery');
    var _ = require('underscore');
    var TaskDayView = require('oui/widgets/calendar/taskDayView');

    /**
     * @class TaskWeekView
     * @extends TaskDayView
     * @private
     */
    function TaskWeekView() {
        TaskDayView.apply(this, arguments);
    }

    TaskWeekView.prototype = _.extend(Object.create(TaskDayView.prototype), {
        /**
         * @method TaskWeekView#_getOptions
         * @private
         *
         * @returns {Object} dayOptions
         */
        _getOptions: function () {
            return this._options['task-week'];
        },
        /**
         * @method TaskWeekView#_getMomentAttribute
         * @private
         *
         * @returns {String} 'week'
         */
        _getMomentAttribute: function () {
            return 'week';
        },

        /**
         * @method TaskWeekView#_getVisibleDays
         * @private
         *
         * @returns {Number} 7
         */
        _getVisibleDays: function () {
            return 7;
        },

        /**
         * @method TaskWeekView#getItems
         * @param {moment} startDate
         * @param {moment} endDate
         * @returns {$.Deferred}
         */
        getItems: function (startDate, endDate) {
            return $.when(this._controller.getEvents(startDate, endDate))
                .then(function (events) {
                    var dfd = $.Deferred();
                    $.when(this._controller.getGroups(startDate, endDate, events))
                        .then(function (groups) {
                            dfd.resolve(events, groups);
                        });

                    return dfd.promise();
                }.bind(this));
        }
    });

    return TaskWeekView;
});

/**
 * @module fullscreen
 * @private
 */
define('oui/utils/fullscreen',[],function () {
    'use strict';

    var api = {};

    /**
     * Put the element into full screen mode.
     *
     * @method open
     * @param {HTMLElement} elem The element to make full screen
     */
    function open(elem) {
        if (elem.requestFullscreen) {
            elem.requestFullscreen();
        } else if (elem.msRequestFullscreen) {
            elem.msRequestFullscreen();
        } else if (elem.mozRequestFullScreen) {
            elem.mozRequestFullScreen();
        } else if (elem.webkitRequestFullscreen) {
            elem.webkitRequestFullscreen(Element.ALLOW_KEYBOARD_INPUT);
        }
    }

    /**
     * Close the document's full screen element.
     *
     * @method close
     * @param {HTMLDocument} [doc]
     */
    function close(doc) {
        doc = doc || document;
        if (doc.exitFullscreen) {
            doc.exitFullscreen();
        } else if (doc.msExitFullscreen) {
            doc.msExitFullscreen();
        } else if (doc.mozCancelFullScreen) {
            doc.mozCancelFullScreen();
        } else if (doc.webkitExitFullscreen) {
            doc.webkitExitFullscreen();
        }
    }

    /**
     * Does the document currently have a full screen element.
     *
     * @method isOpen
     * @param {document} [doc]
     * @returns {Boolean} The current state of the documents full screen element
     */
    function isOpen(doc) {
        doc = doc || document;
        return doc.fullscreenElement != null ||
            doc.mozFullScreenElement != null ||
            doc.webkitFullscreenElement != null ||
            doc.msFullscreenElement != null;
    }

    /**
     * @method toggle
     * @param {HTMLElement} elem
     * @param {document} [doc]
     */
    function toggle(elem, doc) {
        if (api.isOpen(doc)) {
            api.close();
        } else {
            api.open(elem);
        }
    }

    api = {
        EVENT_NAME: 'webkitfullscreenchange mozfullscreenchange MSFullscreenChange fullscreenchange',
        open: open,
        close: close,
        isOpen: isOpen,
        toggle: toggle
    };

    /**
     * @exports fullscreen
     */
    return api;
});


define('template-72', ['underscore'], function (_) { return function(data) {
var __t, __p = '', __j = Array.prototype.join;
function print() { __p += __j.call(arguments, '') }
__p += '<div class="orcl-widget orcl-calendar ' +
((__t = ( data.editable ? 'editable' : '' )) == null ? '' : __t) +
'"> <div class="calendar-toolbar ' +
((__t = ( !data.showToolbar ? 'hidden' : '' )) == null ? '' : __t) +
'" aria-hidden="' +
((__t = ( !data.showToolbar )) == null ? '' : __t) +
'"> <div class="toolbar-left"> <div class="btn-toolbar nav-toolbar"> <button class="btn btn-today btn-blue-light">' +
((__t = ( data.locale.today )) == null ? '' : __t) +
'</button> <div class="btn-group"> <button class="btn btn-prev btn-blue-light" title="' +
((__t = ( data.locale.calendar.previous )) == null ? '' : __t) +
'"> <i class="pgbu-icon-small pgbu-icon-menu-back"></i> </button> <button class="btn btn-next btn-blue-light" title="' +
((__t = ( data.locale.calendar.next )) == null ? '' : __t) +
'"> <i class="pgbu-icon-small pgbu-icon-menu-forward"></i> </button> </div> </div> <span class="calendar-title"></span> </div> <div class="toolbar-right"> <div class="btn-toolbar pull-right"> <div class="btn-group btn-toggle-control" data-toggle="buttons-radio"> ';
 data.availableViews.forEach(function (view) { ;
__p += ' <button class="btn btn-blue-light" data-view="' +
((__t = ( view )) == null ? '' : __t) +
'">' +
((__t = ( data.locale.views[view] )) == null ? '' : __t) +
'</button> ';
 }); ;
__p += ' </div> <button data-action="fullscreen" class="btn btn-blue-light btn-fullscreen" title="' +
((__t = ( data.locale.fullScreen )) == null ? '' : __t) +
'"> <i class="pgbu-icon pgbu-icon-fullscreen"></i> </button> </div> </div> </div> <div class="view" tabindex="0"> </div> <div class="calendar-footer"> <div class="calendar-legend"> <span class="legend-item legend-normal">' +
((__t = ( data.locale.calendar.standard )) == null ? '' : __t) +
'</span> <span class="legend-item legend-overlay">' +
((__t = ( data.locale.calendar.unavailable )) == null ? '' : __t) +
'</span> <span class="legend-item legend-exception">' +
((__t = ( data.locale.calendar.exception )) == null ? '' : __t) +
'</span> </div> </div></div>';
return __p
};});


define('template-73', ['underscore'], function (_) { return function(data) {
var __t, __p = '';
__p += '<span class="legend-item ' +
((__t = ( data.key )) == null ? '' : __t) +
'">' +
((__t = ( data.name )) == null ? '' : __t) +
'</span>';
return __p
};});

define('oui/widgets/calendar/calendarView',['require','jquery','underscore','moment','oui/base/view','oui/widgets/calendar/weekView','oui/widgets/calendar/monthView','oui/widgets/calendar/dayView','oui/widgets/calendar/yearView','oui/widgets/calendar/taskDayView','oui/widgets/calendar/taskWeekView','oui/utils/fullscreen','template-72','template-73'],function (require) { //jshint ignore:line
    'use strict';

    var $ = require('jquery');
    var _ = require('underscore');
    var moment = require('moment');
    var View = require('oui/base/view');
    var WeekView = require('oui/widgets/calendar/weekView');
    var MonthView = require('oui/widgets/calendar/monthView');
    var DayView = require('oui/widgets/calendar/dayView');
    var YearView = require('oui/widgets/calendar/yearView');
    var TaskDayView = require('oui/widgets/calendar/taskDayView');
    var TaskWeekView = require('oui/widgets/calendar/taskWeekView');
    var fullscreen = require('oui/utils/fullscreen');
    var calendarTemplate = require('template-72');
    var legendItem = require('template-73');

    /**
     * @typedef {Object} CalendarView#DateRange
     * @property {moment} start
     * @property {moment} end
     */

    var VIEW_TYPES = {
        'week': WeekView,
        'month': MonthView,
        'day': DayView,
        'year': YearView,
        'task-day': TaskDayView,
        'task-week': TaskWeekView
    };

    /**
     * @class CalendarView
     * @extends {View}
     * @private
     *
     * @param {Object} model
     * @param {Calendar#Controller} controller
     * @param {HTMLElement} element
     * @param {Object} options
     * @param {Object} settings
     */
    function CalendarView(model, controller, element, options, settings) {
        var that = this;

        View.call(this, options, element, controller, settings);

        model.on({
            'view.rerender': this.dateChanged.bind(this),
            'view.event-added': this.added.bind(this),
            'view.event-removed': this.removed.bind(this),
            'view.event-updated': this.updated.bind(this),
            'view.task-updated': this.taskUpdate.bind(this),
            'view.active': this.active.bind(this)
        });

        this._options = options;
        this._controller = controller;
        this._views = {};
        this._currentView = null;

        this.changeView(options.defaultView);
        this.applyLayout(options.layout);

        this.addShortcuts({
            '1': {
                fn: function () {// jscs:ignore jsDoc
                    that.changeView(that._options.availableViews[0]);
                },
                prevention: true
            },
            '2': {
                fn: function () {// jscs:ignore jsDoc
                    that.changeView(that._options.availableViews[1]);
                },
                prevention: true
            },
            '3': {
                fn: function () {// jscs:ignore jsDoc
                    that.changeView(that._options.availableViews[2]);
                },
                prevention: true
            },
            '4': {
                fn: function () {// jscs:ignore jsDoc
                    that.changeView(that._options.availableViews[3]);
                },
                prevention: true
            },
            'K P': {
                fn: this.prev.bind(this)
            },
            'J N': {
                fn: this.next.bind(this)
            },
            'LEFT': {
                fn: function () {// jscs:ignore jsDoc
                    if (that._currentView.highlightPrev) {
                        that._currentView.highlightPrev();
                    }
                }
            },
            'ALT-LEFT': function () {// jscs:ignore jsDoc
                if (that._currentView.updateTask) {
                    var subtractDay = { day: -1 };
                    that._currentView.updateTask({
                        startDate: subtractDay,
                        endDate: subtractDay
                    });
                }
            },
            'RIGHT': {
                fn: function () {// jscs:ignore jsDoc
                    if (that._currentView.highlightNext) {
                        that._currentView.highlightNext();
                    }
                }
            },
            'ALT-RIGHT': function () {// jscs:ignore jsDoc
                if (that._currentView.updateTask) {
                    var addDay = { day: 1 };
                    that._currentView.updateTask({
                        startDate: addDay,
                        endDate: addDay
                    });
                }
            },
            'DELETE': {
                fn: function () {// jscs:ignore jsDoc
                    if (that._currentView.deleteEvent && that._options.editable) {
                        that._currentView.deleteEvent();
                    }
                }
            },
            'CTRL-SHIFT-UP': {
                fn: function () {// jscs:ignore jsDoc
                    if (that._currentView.shiftEnd) {
                        that._currentView.shiftEnd(false);
                    }
                },
                stopPropagation: true
            },
            'CTRL-SHIFT-RIGHT': {
                fn: function () {// jscs:ignore jsDoc
                    if (that._currentView.updateTask) {
                        that._currentView.updateTask({
                            endDate: {
                                day: 1
                            }
                        });
                    }
                },
                stopPropagation: true
            },
            'ALT-SHIFT-RIGHT': {
                fn: function () {// jscs:ignore jsDoc
                    if (that._currentView.updateTask) {
                        that._currentView.updateTask({
                            startDate: {
                                day: 1
                            }
                        });
                    }
                },
                stopPropagation: true
            },
            'CTRL-SHIFT-DOWN': {
                fn: function () {// jscs:ignore jsDoc
                    if (that._currentView.shiftEnd) {
                        that._currentView.shiftEnd(true);
                    }
                },
                stopPropagation: true
            },
            'ALT-SHIFT-LEFT': {
                fn: function () {// jscs:ignore jsDoc
                    if (that._currentView.updateTask) {
                        that._currentView.updateTask({
                            startDate: {
                                day: -1
                            }
                        });
                    }
                },
                stopPropagation: true
            },
            'CTRL-SHIFT-LEFT': {
                fn: function () {// jscs:ignore jsDoc
                    if (that._currentView.updateTask) {
                        that._currentView.updateTask({
                            endDate: {
                                day: -1
                            }
                        });
                    }
                },
                stopPropagation: true
            },
            'SHIFT-UP': {
                fn: function () {// jscs:ignore jsDoc
                    if (that._currentView.shiftStart) {
                        that._currentView.shiftStart(false);
                    }
                }
            },
            'SHIFT-DOWN': {
                fn: function () {// jscs:ignore jsDoc
                    if (that._currentView.shiftStart) {
                        that._currentView.shiftStart(true);
                    }
                }
            },
            'UP': {
                fn: function () {// jscs:ignore jsDoc
                    if (that._currentView.move) {
                        that._currentView.move(false);
                    }
                }
            },
            'DOWN': {
                fn: function () {// jscs:ignore jsDoc
                    if (that._currentView.move) {
                        that._currentView.move(true);
                    }
                }
            },
            'C': {
                fn: function () {// jscs:ignore jsDoc
                    if (that._options.editable && that._options.events.creator) {
                        $.when(that._options.events.creator('create', _.defaults({
                            $view: this._$el
                        }, this._options)))
              .done(function (config) {

                  if (that._options.events.allowOverlaps || that._controller.getCachedEvents(config.startDate, config.endDate).length === 0) {
                      that._controller.newEvent(config.startDate, config.endDate);
                  }
              });
                    }
                }
            },
            'E': {
                fn: function () {// jscs:ignore jsDoc
                    if (that._options.editable && that._options.events.creator) {
                        if (that._currentView.getActiveEvent) {
                            var activeEvent = that._currentView.getActiveEvent();
                            $.when(that._options.events.creator('update', _.defaults({
                                    $view: this._$el
                                }, this._options), {
                                    startDate: activeEvent.startDate,
                                    endDate: activeEvent.endDate
                                }))
                            .done(function (config) {
                                that._controller.updateEvent(activeEvent.id, config);
                            });
                        }
                    }
                }
            }
        }, { delegate: '.view' });

        var hoverTimeout = null;
        this._$el.on('mouseenter', '.event, .task', function (event) {
            hoverTimeout = window.setTimeout(function () {
                that.trigger('hover', event, that._controller.getEvent($(event.currentTarget).attr('data-event-id')).data);
            }, 2000);
        });

        this._$el.on('mouseleave', '.event, .task', function () {
            window.clearTimeout(hoverTimeout);
        });
    }

    var cvp = CalendarView.prototype = Object.create(View.prototype);

    /**
     * @method CalendarView#buildElement
     *
     * @param {Calendar#Config} options
     * @param {jQuery} element
     */
    cvp.buildElement = function (options, element) {
        if (element == null) {
            return $(calendarTemplate({
                editable: options.editable,
                locale: options.locale,
                availableViews: options.availableViews,
                showToolbar: options.layout.showToolbar
            }));
        } else if (element.jquery) {
            //TODO: figure out what to do here
            return null;
        }
    };

    /**
     * @method CalendarView#applyLayout
     *
     * @param {Object} layout
     */
    cvp.applyLayout = function (layout) {
        this._$el.css({
            height: layout.height,
            width: layout.width
        });
    };

    /**
     * Returns the element that identifies the widget.
     *
     * @method IconView#getIdentifyingElement
     *
     * @returns {jQuery}
     */
    cvp.getIdentifyingElement = function () {
        return this._$el;
    };

    /**
     * @method CalendarView#init
     */
    cvp.init = function () {
        var that = this;

        this._$el.find('.instructions').attr('id', this.id() + '-instructions');
        this._$el.find('.view').attr('aria-describedby', ((this._$el.attr('aria-describedby') || '') + ' ' + this.id() + '-instructions').trim());

        this._$el.on('click.calendar', '.btn-next', this.next.bind(this));
        this._$el.on('click.calendar', '.btn-prev', this.prev.bind(this));
        this._$el.on('click.calendar', '.btn-today', function () {
            that._controller.currentDate(moment());
        });
        this._$el.on('click.calendar', '[data-action=fullscreen]', fullscreen.toggle.bind(null, this._$el[0], null));

        $(document).on(fullscreen.EVENT_NAME, function () {
            var open = fullscreen.isOpen();

            this._$el.toggleClass('oui-fullscreen', open);

            this._$el.find('[data-action=fullscreen] i')
                .toggleClass('pgbu-icon-fullscreen', !open)
                .toggleClass('pgbu-icon-restore', open);

        }.bind(this));

        this._$el.on('click.calendar', '.btn[data-view]', function () {
            that.changeView($(this).attr('data-view'));
        });

        this._$el.on('view-change.calendar', function (event, view) {
            that.changeView(view);
        });

        $(window).on('resize.calendar' + this._settings.id, _.throttle(function () {
            that.resize();
        }, 200));

        //TODO: think about additional event types
        this._$el.on('click', '[data-trigger]', function (event) {
            var type = event.currentTarget.getAttribute('data-trigger-type');
            if (type === 'click' || type == null) {
                event.stopPropagation();
                this.trigger(event.currentTarget.getAttribute('data-trigger'), that._currentView.getEventOptions(event.currentTarget));
            }
        }.bind(this));

        this._$el.on('dblclick', '[data-trigger-type="dblclick"][data-trigger]', function (event) {
            event.stopPropagation();
            this.trigger(event.currentTarget.getAttribute('data-trigger'), that._currentView.getEventOptions(event.currentTarget));
        }.bind(this));

        this.resize();
    };

    /**
     * @method CalendarView#_destroy
     * @private
     */
    cvp._destroy = function () {
        this._$el.off('.calendar');
        $(window).off('.calendar' + this._settings.id);
    };

    /**
     * @method CalendarView#_isInView
     * @private
     *
     * @returns {Boolean} true
     */
    cvp._isInView = function () {
        return true;
    };

    /**
     * @method CalendarView#_filterViewableEvents
     * @private
     *
     * @param {Array} events
     *
     * @returns {Array} list of viewable events
     */
    cvp._filterViewableEvents = function (events) {
        var that = this;

        return _.filter(events, function (event) {
            return that._isInView(event);
        });
    };

    /**
     * @method CalendarView#dateChanged
     *
     * @param {moment} newDate
     */
    cvp.dateChanged = function (newDate) {
        this._currentView.render(newDate);
        this.updateTitle(this._currentView.getDateRange());
    };

    /**
     * @method CalendarView#added
     *
     * @param {Object} event
     * @param {Array} overlaps
     */
    cvp.added = function (event, overlaps) {
        this._currentView.addEvent(event, overlaps);
    };

    /**
     * @method CalendarView#removed
     *
     * @param {String} id
     */
    cvp.removed = function (id) {
        this._currentView.removeEvent(id);
    };

    /**
     * @method CalendarView#updated
     *
     * @param {Array|Object} events
     * @param {Object} overlaps
     * @param {Object} oldOverlaps
     */
    cvp.updated = function (events, overlaps, oldOverlaps) {
        if (!_.isArray(events)) {
            events = [events];
        }

        this._currentView.updateEvents(events, overlaps, oldOverlaps);
    };

    /**
     * @method CalendarView#taskUpdate
     * @param {Object} args
     */
    cvp.taskUpdate = function (args) {
        this._currentView.update(args);
    };

    /**
     * @method CalendarView#getView
     *
     * @param {String} viewType
     *
     * @returns {DayTimeView|MonthView|WeekView|YearView} view
     */
    cvp.getView = function (viewType) {
        var view = this._views[viewType];

        if (viewType != null && view == null) {
            view = new VIEW_TYPES[viewType](this._options, this._controller, this._$el.find('.view'), this._settings);
            this._views[viewType] = view;
        }

        return view;
    };

    /**
     * @method CalendarView#updateTitle
     *
     * @param {CalendarView#DateRange} dateRange
     */
    cvp.updateTitle = function (dateRange) {
        var sameMonth = dateRange.start.month() === dateRange.end.month();
        var sameDay = dateRange.start.dayOfYear() === dateRange.end.dayOfYear();

        //TODO: use an option to format?
        if (sameDay) {
            this._$el.find('.calendar-title').text(dateRange.start.format('ddd, MMMM DD, YYYY'));
        } else if (this._viewType === 'month') {
            //TODO: determine better way to format a single month when both start and end are outside the bounds
            this._$el.find('.calendar-title').text(this._controller.getCurrentDate().format('MMMM YYYY'));
        } else if (this._viewType === 'year') {

            this._$el.find('.calendar-title').text(this._controller.getCurrentDate().format('YYYY'));
        } else if (sameMonth) {
            this._$el.find('.calendar-title').text(dateRange.start.format('MMM DD') + ' - ' + dateRange.end.format('DD, YYYY'));
        } else {
            this._$el.find('.calendar-title').text(dateRange.start.format('MMM DD') + ' - ' + dateRange.end.format('MMM DD, YYYY'));
        }
    };

    /**
     * @method CalendarView#changeView
     *
     * @param {String} viewType
     */
    cvp.changeView = function (viewType) {
        var newView = this.getView(viewType);
        if (newView != null && newView !== this._currentView) {
            if (this._currentView != null) {
                this._currentView.clear();
            }

            this._$el.find('[data-view]').removeClass('active').filter('[data-view=' + viewType + ']').addClass('active');
            this._$el.find('.calendar-legend').empty();

            var legendMarkup = _.reduce(this._options[viewType].legend, function (memo, item) {
                memo += legendItem(item);
                return memo;
            }, '');

            this._$el.find('.calendar-legend').append(legendMarkup);
            this._currentView = newView;
            this._viewType = viewType;
            this._currentView.render(this._controller.getCurrentDate());
            this.resize();
            this.updateTitle(this._currentView.getDateRange());
            this.trigger('view-change', this._viewType);
        }
    };

    /**
     * @method CalendarView#getDateRange
     *
     * @returns {CalendarView#DateRange}
     */
    cvp.getDateRange = function () {
        if (_.isFunction(this._currentView.getDateRange)) {
            return this._currentView.getDateRange();
        }
    };

    /**
     * @method CalendarView#prev
     */
    cvp.prev = function () {
        this._currentView.prev();
    };

    /**
     * @method CalendarView#prev
     */
    cvp.next = function () {
        this._currentView.next();
    };

    /**
     * @method CalendarView#resize
     */
    cvp.resize = function () {
        this._currentView.resize();
    };

    /**
     * @method CalendarView#redraw
     *
     * @param {moment} startDate
     * @param {moment} endDate
     */
    cvp.redraw = function (startDate, endDate) {
        this._currentView.redraw(startDate, endDate);
    };

    /**
     * @method CalendarView#disable
     *
     * @param {Boolean} disabled
     */
    cvp.disable = function (disabled) {
        this._$el.toggleClass('disabled', disabled);
    };

    /**
     * @method CalendarView#active
     * @param {Number} activeId
     * @param {Boolean} [top]
     */
    cvp.active = function (activeId, top) {
        if (this._currentView.active) {
            this._currentView.active(activeId, top);
        }
    };

    /**
     * @method CalendarView#forceRedraw
     */
    cvp.forceRedraw = function () {
        if (this._currentView.forceRedraw) {
            this._currentView.forceRedraw();
        }
    };

    /**
     * Associates a hopper with the calendar
     *
     * @method CalendarView#_associateHopper
     * @param {Hopper} hopper
     * @returns {Boolean} whether a hopper is already associated
     * @private
     */
    cvp._associateHopper = function (hopper) {
        if (this._associatedHopper != null && this._associatedHopper !== hopper) {
            throw new Error('Calendar can only be associated with one hopper');
        } else if (this._associatedHopper) {
            return true;
        }

        this._associatedHopper = hopper;

        this._controller._model.on('repopulate-hopper.view', function () {
            this._associatedHopper.repopulate();
        }.bind(this));
    };

    /**
     * Adds a "virtual" item to the calendar
     *
     * Used for drag support from the hopper
     *
     * @method CalendarView#_renderTemporaryItem
     * @param {Object} item
     * @param {{pageX,pageY}} position
     * @returns {{startDate,endDate}} Information about the start and end date where the item is being dragged
     * @private
     */
    cvp._renderTemporaryItem = function (item, position) {
        if (this._currentView._renderTemporaryItem) {
            return this._currentView._renderTemporaryItem(item, position);
        }
    };

    /**
     * Gets the start and finish date in the current view
     *
     * @method CalendarView#getViewRange
     * @returns {{startDate: moment, endDate: moment}}
     */
    cvp.getViewRange = function () {
        return this._currentView._getStartAndEndDate();
    };

    /**
     * @method CalendarView#focus
     */
    cvp.focus = function () {
        this._$el.find('.view').focus();
    };

    /**
     * @method CalendarView#callOut
     */
    cvp.callOut = function () {
        if (this._currentView.callOut) {
            this._currentView.callOut();
        }
    };

    return CalendarView;
});

define('oui/widgets/calendar/calendarModel',[
  'jquery',
  'underscore',
  'moment',
  'oui/utils/subscribable',
  'oui/widgets/calendar/event'
], function ($, _, moment, Subscribable, Event) { //jshint ignore:line
    'use strict';

    /**
     * @typedef {Object} CalendarModel#Overlay
     * @property {moment} start
     * @property {moment} end
     */
    /**
     * Fired when the current date changed within the Calendar.
     *
     * @event CalendarModel#date-change
     * @param {moment} curDate The current date expressed by the calendar
     */
    /**
     * Fired when the Calendar has issued a request with a deferred result.
     *
     * @event CalendarModel#busy
     */
    /**
     * Fired when processing of a deferred has finished (success or failure).
     *
     * @event CalendarModel#available
     */
    /**
     * Fired when the Calendar has an event added.
     *
     * @event CalendarModel#event-added
     * @param {Event} event The event that has been added to the Calendar
     * @param {Array<Event>} existingEvents The existing events within the calendar.
     *
     */
    /**
     * Fired when an event is removed from the calendar.
     *
     * @event CalendarModel#event-removed
     * @param {Event} event The event that was removed
     * @param {Object} data The events internal data
     */
    /**
     * Fired when the start and/or end time of the event has been updated.
     *
     * @event CalendarModel#event-updated
     * @param {Event} event The event that was updated
     * @param {Array<Event>} overlaps The events that currently overlap with the updated event
     * @param {Array<Event>} oldOverlaps The events that were overlapped but after updating no longer overlap
     */

    var MOMENT_FIELDS = [
      'startDate',
      'endDate'
    ];

    /**
     * @class CalendarModel
     * @private
     *
     * @param {Object} options
     * @param {Object|Array} initData
     */
    function CalendarModel(options, initData) {
        Subscribable.call(this);

        var that = this;

        this._options = options;
        this._calendarEvents = {};
        this._taskGroups = {};
        this._currentDate = moment();

        this._store = null;
        if (_.isArray(initData)) {
            _.each(initData, function (event) {
                that.addEvent(event);
            });
        } else if (_.isObject(initData)) {
            //jscs:disable
            this._store = new initData(); //jshint ignore:line
            //jscs:enable
        }
    }

    var cmp = CalendarModel.prototype = Object.create(Subscribable.prototype);

    /**
     * @method CalendarModel#getCurrentDate
     *
     * @returns {moment} current date
     */
    /**
     * @method CalendarModel#currentDate
     *
     * @param {Object} curDate
     */
    cmp.getCurrentDate = cmp.currentDate = function (curDate) {
        if (!arguments.length) {
            return this._currentDate;
        }

        if (!this._currentDate.isSame(curDate)) {
            this._currentDate = curDate;
            this.trigger('rerender', curDate);
        }
    };

    /**
     * @method CalendarModel#_getFunction
     * @private
     *
     * @param {String} key
     */
    cmp._getFunction = function (key) {
        if (this._store && this._store[key]) {
            return this._store[key].bind(this._store);
        } else if (this._options.storeFns && this._options.storeFns[key]) {
            return this._options.storeFns[key];
        }
    };

    /**
     * @method CalendarModel#_execute
     * @private
     *
     * @param {Function} fn
     */
    cmp._execute = function (fn) {
        var that = this;

        var timeout = setTimeout(function () {
            that.trigger('busy');
        });

        return $.when(fn())
      .done(function () {
          clearTimeout(timeout);

          that.trigger('available');
      });
    };

    /**
     * @method CalendarModel#newEvent
     * @param {Object} config The properties that will be used to construct the event.
     *
     * @returns {$.Deferred} The deferred representing the status of the new event creation.
     */
    cmp.newEvent = function (config) {
        var fn;
        var that = this;

        fn = this._getFunction('newEvent');

        if (fn) {
            return $.when(this._execute(fn.bind(this, config)))
        .done(function (event) {
            that.addEvent(event);
        });
        } else {
            return that.addEvent(config);
        }
    };

    /**
     * @method CalendarModel#newTask
     * @param {moment} startDate
     * @param {Number|String} groupId
     */
    cmp.newTask = function (startDate, groupId) {
        var fn;
        var that = this;

        fn = this._getFunction('newTask');

        if (fn) {
            return $.when(this._execute(fn.bind(this, startDate, groupId)))
                .done(function (event) {
                    that.addEvent(event);
                });
        } else {
            return that.addEvent(startDate);
        }
    };

    /**
     * @method CalendarModel#addEvent
     *
     * @param {Object} event
     * @param {Object} params
     *
     * @returns {Object} event
     */
    cmp.addEvent = function (event, params) {
        var calEvent;

        var eventId = this.getEventId(event);

        if (eventId) {
            calEvent = this.findEvent(eventId);
        }

        if (calEvent == null) {
            calEvent = new Event(event, this._options.eventKeys);

            this._calendarEvents[calEvent.id] = calEvent;

            if (this._store && this._store.addEvent && (!params || (params && !params.cache))) {
                this._store.addEvent(calEvent.data);
            }

            if (!params || (params && !params.silent)) {
                this.trigger('event-added', calEvent, this.getEvents(calEvent.startDate, calEvent.endDate, {
                    internal: true
                }));
            }
        }

        return calEvent;
    };

    /**
     * @method CalendarModel#addGroup
     * @param {Object} group
     */
    cmp.addGroup = function (group) {
        var groupEvent = new Event(group, this._options.eventKeys.parent);
        if (!this._taskGroups[groupEvent.id]) {
            this._taskGroups[groupEvent.id] = {
                collapsed: false
            };
        }
        return groupEvent;
    };

    /**
     * @method CalendarModel#findEvent
     *
     * @param {Object|String|Number} id
     *
     * @returns {Object} event
     */
    cmp.findEvent = function (id) {
        var that = this;
        if (_.has(id, that._options.eventKeys.id)) {
            return this._calendarEvents[id[that._options.eventKeys.id]];
        } else if (_.isString(id) || _.isNumber(id)) {
            return this._calendarEvents[id];
        }
        return _.find(this._calendarEvents, function (event) {
            if (_.isObject(id)) {
                if (_.has(id, that._options.eventKeys.id)) {
                    return event.data[that._options.eventKeys.id] === id[that._options.eventKeys.id];
                } else {
                    return _.isEqual(id, _.pick(event.data, _.keys(id)));
                }
            } else {
                return event.data[that._options.eventKeys.id] === id;
            }
        });
    };

    /**
     * Gets the properties of the event which are identifying properties.
     *
     * Traditionally, this only needs to be a single property like 'id' but it may
     * need to handle multiple types where the id is a composite object if [ 'type', 'id' ].
     *
     * The function defaults to calling the getId function from the store, failing that it will fallback to the
     * storeFns.getId if it exists and as a final attempt will use the eventKeys property.
     *
     * @method CalendarModel#getEventId
     * @param {Object} event The client provided data to retrieve the id properties of.
     *
     * @returns {Array} The keys of the event object that identify the event uniquely.
     */
    cmp.getEventId = function (event) {
        if (this._store && this._store.getId) {
            return this._store.getId(event);
        } else if (this._options.storeFns.getId) {
            return this._options.storeFns.getId(event);
        } else {
            var id = _.pick(event, this._options.eventKeys.id);

            if (_.keys(id).length !== 0) {
                return id;
            }
        }
    };

    /**
     * @method CalendarModel#removeEvent
     *
     * @param {String} id
     */
    cmp.removeEvent = function (id) {
        var ev;

        if (this._calendarEvents[id]) {
            ev = this._calendarEvents[id];

            if (this._store.removeEvent) {
                this._store.removeEvent(ev.data);
            }

            delete this._calendarEvents[id];
            this.trigger('event-removed', ev, ev.data);
        }
    };

    /**
     * @method CalendarModel#updateTask
     * @param {Object} fields
     */
    cmp.updateTask = function (fields) {
        if (this._active) {
            var task = this.getEvent(this._active);
            var newData = {
                startDate: task.startDate.clone(),
                endDate: task.endDate.clone()
            };
            var oldData = {
                startDate: task.startDate.clone(),
                endDate: task.endDate.clone()
            };
            _.each(fields, function (newValue, field) {
                if (moment.isMoment(newValue)) {
                    newData[field] = newValue;
                } else {
                    _.each(newValue, function (qty, unit) {
                        newData[field].add(qty, unit);
                    });
                }
            });
            if (Math.round(newData.endDate.diff(newData.startDate, 'day', true)) >= 1 &&
                !newData.startDate.isAfter(newData.endDate, 'day') &&
                !(newData.startDate.isSame(oldData.startDate, 'day') && newData.endDate.isSame(oldData.endDate, 'day'))) {
                task.startDate = newData.startDate;
                task.endDate = newData.endDate;
                task.modified = Date.now();
                if (this._store && this._store.update) {
                    this._store.update(task.data, task, oldData);
                }
                this.trigger('task-updated', {
                    task: task,
                    oldData: oldData
                });
            }
        }
    };

    /**
     * @method CalendarController#addEventToFirstVisibleDay
     * @param {Calendar#eventData} eventData
     * @param {{startDate: moment, endDate: moment}} fields
     */
    cmp.addEventToFirstVisibleDay = function (eventData, fields) {
        var event = new Event(eventData, this._options.eventKeys);
        event.startDate = fields.startDate;
        event.endDate = fields.endDate;
        if (this._store && this._store.update) {
            this._store.update(eventData, event, eventData);
        }
        this.addEvent(eventData);
    };

    /**
     * @method CalendarModel#canEdit
     */
    cmp.canEditTask = function () {
        var canEdit = true;
        if (this._store && _.isFunction(this._store.canEdit)) {
            canEdit = this._store.canEdit(this._active);
        }
        return canEdit;
    };

    /**
     * @method CalendarModel#updateEvent
     *
     * @param {String} id
     * @param {Object} newData
     */
    cmp.updateEvent = function (id, newData) {
        var that = this;

        var oldOverlaps = this.getEvents(that._calendarEvents[id].startDate, that._calendarEvents[id].endDate);
        var overlaps;

        _.each(newData, function (value, key) {
            if (_.contains(MOMENT_FIELDS, key)) {

                if (moment.isMoment(value)) {
                    that._calendarEvents[id][key] = value;
                } else {

                    //TODO: how do we determine base moment?
                    if (value.day) {
                        that._calendarEvents[id][key] = value.day.clone();
                    }

                    delete value.day;

                    _.each(value, function (val, unit) {
                        that._calendarEvents[id][key][unit](val);
                    });

                    newData[key] = that._calendarEvents[id][key];
                }
            }
        });

        if (this._store.updateEvent) {
            this._store.updateEvent(this._calendarEvents[id].data, newData);
        }

        overlaps = this.getEvents(that._calendarEvents[id].startDate, that._calendarEvents[id].endDate);

        $.when(overlaps, oldOverlaps)
      .done(function (overlapResult, oldOverlapResult) {
          that.trigger('event-updated', that._calendarEvents[id], overlapResult, oldOverlapResult);
      });
    };

    /**
     * @method CalendarModel#getEvents
     *
     * @param {moment} startDate
     * @param {moment} endDate
     * @param {Object} modifiers
     *
     * @returns {Array} events
     */
    cmp.getEvents = function (startDate, endDate, modifiers) {
        var that = this;
        var getFn;

        getFn = this._getFunction('getEvents');

        if (!(modifiers && modifiers.internal) && getFn != null) {
            return this._execute(getFn.bind(this._store, startDate, endDate))
                .then(function (data) {

                    return _.map(data, function (ev) {
                        that.mapEvent(ev);
                        return that.addEvent(ev, {
                            silent: true,
                            cache: true
                        });
                    });
                })
                .then(function (events) {
                    return _.chain(events)
                        .filter(that.eventOccurs.bind(that, startDate, endDate))
                        .sortBy(function (ev) {
                            return ev.startDate.diff(ev.endDate, 'ms', true);
                        })
                        .value();
                });
        } else {
            return _.chain(this._calendarEvents)
                .filter(that.eventOccurs.bind(that, startDate, endDate))
                .sortBy(function (event) {
                    return +event.startDate;
                })
                .value();
        }
    };

    /**
     * @method CalendarModel#getGroups
     *
     * @param {moment} startDate
     * @param {moment} endDate
     * @param {Array} events
     */
    cmp.getGroups = function (startDate, endDate, events) {
        var that = this;
        var getFn = this._getFunction('getGroups');

        if (getFn != null) {
            return $.when(this._execute(getFn.bind(this._store, startDate, endDate, events)))
                .then(function (groups) {
                    return _.map(groups, function (group) {
                        return that.addGroup(group);
                    });
                })
                .then(function (groups) {
                    return _.sortBy(groups, function (group) {
                        return group.startDate.valueOf();
                    });
                });
        }
        return $.Deferred().resolve();
    };

    /**
     * @method CalendarModel#toggleGroup
     * @param {Number|String} groupId
     */
    cmp.toggleGroup = function (groupId) {
        var group = this._taskGroups[groupId];
        if (group && group.collapsed != null) {
            group.collapsed = !group.collapsed;
            return group.collapsed;
        }
        this._taskGroups[groupId] = { collapsed: false };
        return this._taskGroups[groupId].collapsed;
    };

    /**
     * @method CalendarModel#getTaskGroups
     */
    cmp.getTaskGroups = function () {
        return this._taskGroups;
    };

    /**
     * @method CalendarModel#getDisabledDays
     *
     * @param {Calendar#DateRange} dateRange
     *
     * @returns {Array}
     */
    cmp.getDisabledDays = function (dateRange) {
        if (this._store && _.isFunction(this._store.getDisabledDays)) {
            return this._store.getDisabledDays(dateRange);
        }
        return [];
    };

    /**
     * @method CalendarModel#getOverlays
     *
     * @param {moment} startDate
     * @param {moment} endDate
     *
     * @returns {Array<CalendarModel#Overlay>}
     */
    cmp.getOverlays = function (startDate, endDate) {
        var that = this;
        var getFn;

        //$.when and $.then to do the mapping of external data structure to internal?
        if (this._store && this._store.getOverlays) {
            getFn = this._store.getOverlays;
        } else if (this._options.storeFns.getOverlays) {
            getFn = this._options.storeFns.getOverlays;
        }

        if (getFn != null) {
            return $.when(this._execute(getFn.bind(this._store, startDate, endDate)))
                .then(function (data) {
                    return _.map(data, function (ev) {
                        that.mapEvent(ev);
                        return new Event(ev, that._options.eventKeys);
                    });
                })
                .then(function (events) {
                    return _.chain(events)
                        .filter(that.eventOccurs.bind(that, startDate, endDate))
                        .sortBy(function (ev) {
                            return ev.startDate.valueOf();
                        })
                        .value();
                });
        }
    };

    /**
     * @method CalendarModel#setStore
     *
     * @param {Object} Store
     *
     * @returns {Object} store
     */
    cmp.setStore = function (Store) {
        if (Store != null) {
            this._store = new Store();
            return this._store;
        }
    };

    /**
     * @method CalendarModel#getMediator
     * @returns {Object}
     */
    cmp.getMediator = function () {
        return this._store;
    };

    /**
     * @method CalendarModel#clean
     */
    cmp.clean = function () {
        this._calendarEvents = {};
    };

    /**
     * @method CalendarModel#getEvent
     *
     * @param {String|Number} id
     *
     * @returns {Object} event
     */
    cmp.getEvent = function (id) {
        if (this._calendarEvents[id]) {
            return this._calendarEvents[id];
        }
    };

    /**
     * @method CalendarModel#disabled
     *
     * @returns {Array} days
     */
    /**
     * @method CalendarModel#disabled
     *
     * @param {Array} days
     */
    cmp.disabled = function (days) {
        if (!arguments.length) {
            return this._disabled;
        }

        this._disabled = days;
    };

    /**
     * @method CalendarModel#active
     * @param {Number} [id]
     * @param {Boolean} [top]
     */
    cmp.active = function (id, top) {
        if (!arguments.length) {
            return this._active;
        }
        if (this._active !== id) {
            this._active = id;
        }
        this.trigger('active', this._active, top);
    };

    /**
     * Gets the end date for a task that is not yet on the calendar
     *
     * @method CalendarModel#_getTaskDuration
     * @param {Object} task
     * @private
     */
    cmp._getTaskDuration = function (task) {
        if (this._store.getTaskDuration) {
            return this._store.getTaskDuration(task);
        }
        var event = new Event(task, this._options.eventKeys);
        if (event.startDate && event.endDate) {
            return event.endDate.diff(event.startDate, 'day');

        }
        throw new Error('Event missing start and end date');
    };

    /**
     * @method CalendarModel#eventOccurs
     * @param {moment} start
     * @param {moment} end
     * @param {Event} ev
     *
     * @returns {Boolean}
     */
    cmp.eventOccurs = function (start, end, ev) {
        return ev.isOccurring(start, end);
    };

    /**
     * @method CalendarModel#mapEvent
     * @param {Event} ev
     */
    cmp.mapEvent = function (ev) {
        if (this._options.mapper && this._options.mapper.to) {
            ev = this._options.mapper.to(ev);
        }
        return ev;
    };

    return CalendarModel;
});

define('oui/widgets/calendar/calendarController',[
  'underscore'
], function (_) {
    'use strict';

    /**
     * @class CalendarController
     * @private
     *
     * @param {Calendar#Config} options
     * @param {CalendarModel} model
     */
    function CalendarController(options, model) {

        this._options = options;
        this._model = model;
    }

    CalendarController.prototype = {
        /**
         * @method CalendarController#load
         */
        load: function () {},

        /**
         * @method CalendarController#getEvent
         *
         * @param {String} id
         *
         * @returns {Object} event
         */
        getEvent: function (id) {
            return this._model.getEvent(id);
        },

        /**
         * @method CalendarController#getEvents
         *
         * @param {String} startDate
         * @param {String} endDate
         *
         * @returns {Array} events
         */
        getEvents: function (startDate, endDate) {
            return this._model.getEvents(startDate, endDate);
        },

        /**
         * @method CalendarController#getGroups
         *
         * @param {String} startDate
         * @param {String} endDate
         * @param {Array} events
         *
         * @returns {Array} events
         */
        getGroups: function (startDate, endDate, events) {
            return this._model.getGroups(startDate, endDate, events);
        },

        /**
         * @method CalendarController#getCachedEvents
         *
         * @param {moment} startDate
         * @param {moment} endDate
         *
         * @returns {Array} events
         */
        getCachedEvents: function (startDate, endDate) {
            return this._model.getEvents(startDate, endDate, {
                internal: true
            });
        },

        /**
         * @method CalendarController#getCachedSingleDayEvents
         *
         * @param {moment} startDate
         * @param {moment} endDate
         *
         * @returns {Array} events
         */
        getCachedSingleDayEvents: function (startDate, endDate) {
            return _.filter(this.getCachedEvents(startDate, endDate), function (ev) {
                //TODO: determine best way to handle all-day events. attribute or time period
                return ev.endDate.diff(ev.startDate, 'minutes') < 1439;
            });
        },

        /**
         * @method CalendarController#getOverlays
         *
         * @param {moment} startDate
         * @param {moment} endDate
         *
         * @returns {Array} overlays
         */
        getOverlays: function (startDate, endDate) {
            return this._model.getOverlays(startDate, endDate);
        },

        /**
         * @method CalendarController#getCurrentDate
         *
         * @returns {moment} current date
         */
        getCurrentDate: function () {
            return this._model.getCurrentDate();
        },

        /**
         * @method CalendarController#currentDate
         *
         * @param {moment} date
         * @returns {moment} current date
         */
        currentDate: function (date) {
            return this._model.currentDate(date);
        },

        /**
         * @method CalendarController#newEvent
         *
         * @param {moment} startDate
         * @param {moment} endDate
         *
         * @returns {Object} event
         */
        newEvent: function (startDate, endDate) {
            var config = {};

            config[this._options.eventKeys.startDate] = startDate;
            config[this._options.eventKeys.endDate] = endDate;
            return this._model.newEvent(config);
        },

        /**
         * @method CalendarController#newTask
         * @param {moment} startDate
         * @param {Number|String} groupId
         */
        newTask: function (startDate, groupId) {
            this._model.newTask(startDate, groupId);
        },

        /**
         * @method CalendarController#addEvent
         *
         * @param {Object} eventData
         *
         * @returns {Object} event
         */
        addEvent: function (eventData) {
            return this._model.addEvent(eventData);
        },

        /**
         * @method CalendarController#addEventToFirstVisibleDay
         * @param {Calendar#eventData} eventData
         * @param {moment} startDate
         */
        addEventToFirstVisibleDay: function (eventData, startDate) {
            var duration = this._model._getTaskDuration(eventData);
            var endDate;
            if (this._options.events.endOfDay) {
                endDate = startDate.clone().add(duration - 1, 'day').endOf('day');
            } else {
                endDate = startDate.clone().add(duration, 'day');
            }
            this._model.addEventToFirstVisibleDay(eventData, {
                startDate: startDate,
                endDate: endDate
            });
        },

        /**
         * @method CalendarController#removeEvent
         *
         * @param {String} eventId
         */
        removeEvent: function (eventId) {
            this._model.removeEvent(eventId);
        },

        /**
         * @method CalendarController#updateTask
         * @param {Object} fields
         */
        updateTask: function (fields) {
            if (this.active() && this.canEditTask()) {
                this._model.updateTask(fields);
            }
        },

        /**
         * @method CalendarController#canEditTask
         */
        canEditTask: function () {
            return this._model.canEditTask();
        },

        /**
         * @method CalendarController#updateEvent
         *
         * @param {String} id
         * @param {Object} newData
         */
        updateEvent: function (id, newData) {
            this._model.updateEvent(id, newData);
        },

        /**
         * @method CalendarController#changeView
         */
        changeView: function () {},

        /**
         * @method CalendarController#setStore
         *
         * @param {Object} Store
         *
         * @returns {Object} store
         */
        setStore: function (Store) {
            return this._model.setStore(Store);
        },

        /**
         * @method CalendarController#getMediator
         *
         * @returns {Object}
         */
        getMediator: function () {
            return this._model.getMediator();
        },

        /**
         * @method CalendarController#getDisabledDays
         * @param {Calendar#DateRange} dateRange
         *
         * @returns {Object}
         */
        getDisabledDays: function (dateRange) {
            return this._model.getDisabledDays(dateRange);
        },

        /**
         * @method CalendarController#clean
         */
        clean: function () {
            this._model.clean();
        },

        /**
         * @method CalendarController#disabled
         *
         * @param {Array} days
         */
        disabled: function (days) {
            if (!arguments.length) {
                return this._model.disabled();
            }
            this._model.disabled(days);
        },

        /**
         * @method CalendarController#active
         * @param {Number} [id]
         * @param {Boolean} [top]
         */
        active: function (id, top) {
            if (!arguments.length) {
                return this._model.active();
            }

            if (_.isObject(id)) {
                this._model.active(this._model.findEvent(id).id, top);
            } else {
                this._model.active(id, top);
            }
        },

        /**
         * @method CalendarController#isDayDisabled
         *
         * @param {moment} day
         */
        isDayDisabled: function (day) {
            var dates = this.disabled();

            return _.find(dates, function (date) {
                return date.format('YYYY-MM-DD') === day.format('YYYY-MM-DD');
            }) != null;
        },

        /**
         * Gets the end date for a task that is not yet on the calendar
         *
         * @method CalendarController#_getTaskDuration
         * @param {Object} task
         * @private
         */
        _getTaskDuration: function (task) {
            return this._model._getTaskDuration(task);
        },

        /**
         * @method CalendarController#toggleGroup
         * @param {String|Number} groupId
         * @returns {Boolean}
         */
        toggleGroup: function (groupId) {
            return this._model.toggleGroup(groupId);
        },

        /**
         * @method CalendarController#getTaskGroups
         * @returns {Object}
         */
        getTaskGroups: function () {
            return this._model.getTaskGroups();
        }
    };

    return CalendarController;
});


define('template-74', ['underscore'], function (_) { return function(data) {
var __t, __p = '';
__p += '<div class="modal pgbu-modal hide fade orcl-modal" data-backdrop="static" tabindex="-1" role="dialog" aria-hidden="true" aria-labelledby="oui-calendar-event-creator"> <div class="modal-header"> <a title="' +
((__t = ( data.locale.close )) == null ? '' : __t) +
'" class="modal-close" href="#" data-dismiss="modal"><i class="pgbu-icon pgbu-icon-dialog-close"></i></a> <h3 id="oui-calendar-event-creator" class="modal-title">' +
((__t = ( data.locale.title  )) == null ? '' : __t) +
'</h3> </div> <div class="modal-body"> <div class="control-group"> <label class="control-label" for="oui-calendar-event-creator-start-date">' +
((__t = ( data.locale.startDate )) == null ? '' : __t) +
'</label> <div class="controls"> <input id="oui-calendar-event-creator-start-date" data-field="startDate" type="date"> </div> </div> <div class="control-group"> <label class="control-label" for="oui-calendar-event-creator-end-date">' +
((__t = ( data.locale.endDate )) == null ? '' : __t) +
'</label> <div class="controls"> <input id="oui-calendar-event-creator-end-date" data-field="endDate" type="date"> </div> </div> </div> <div class="modal-footer"> <button type="button" title="' +
((__t = ( data.locale.cancel )) == null ? '' : __t) +
'" class="btn btn-blue-dark modal-cancel" data-dismiss="modal">' +
((__t = ( data.locale.cancel  )) == null ? '' : __t) +
'</button> <button type="button" title="' +
((__t = ( data.locale.ok )) == null ? '' : __t) +
'" class="btn btn-action modal-accept disabled" data-dismiss="modal" disabled="disabled">' +
((__t = ( data.locale.ok  )) == null ? '' : __t) +
'</button> </div></div>';
return __p
};});


define('template-75', ['underscore'], function (_) { return function(data) {
var __t, __p = '';
__p += '<div class="' +
((__t = ( ['event'].concat(data.classes).join(' ') )) == null ? '' : __t) +
'"> ' +
((__t = ( data.name )) == null ? '' : __t) +
'</div>';
return __p
};});

define('oui/utils/tasks/BaseTaskRenderer',['require','underscore','oui/base/renderer','template-75'],function (require) {
    'use strict';

    var _ = require('underscore');
    var BaseRenderer = require('oui/base/renderer');
    var defaultTaskTemplate = require('template-75');

    /**
     * @class BaseTaskRenderer
     * @extends Renderer
     * @constructor
     */
    function BaseTaskRenderer() {
        BaseRenderer.apply(this, arguments);
        this._mediator = null;
    }

    BaseTaskRenderer.prototype = Object.create(BaseRenderer.prototype);

    _.extend(BaseTaskRenderer.prototype, {
        constructor: BaseTaskRenderer,
        /**
         * @method BaseTaskRenderer#render
         * @param {Object} task
         */
        render: function (task) {
            return defaultTaskTemplate({
                classes: this.getClasses(task),
                name: _.get(task, 'name', '')
            });
        },
        /**
         * @method BaseTaskRenderer#setStore
         * @private
         * @param {Object} mediator
         */
        _setStore: function (mediator) {
            this._mediator = mediator;
        },
        getClasses: _.noop,
    });

    return BaseTaskRenderer;
});


define('template-76', ['underscore'], function (_) { return function(data) {
var __t, __p = '';
__p += '<div class="' +
((__t = ( ['group-card selectable-group'].concat(data.classes).join(' ') )) == null ? '' : __t) +
'"> ' +
((__t = ( data.name )) == null ? '' : __t) +
' <button type="button" class="not-btn collapse-icon"></button></div>';
return __p
};});

define('oui/utils/tasks/BaseGroupRenderer',['require','underscore','oui/utils/tasks/BaseTaskRenderer','template-76'],function (require) {
    'use strict';

    var _ = require('underscore');
    var BaseTaskRenderer = require('oui/utils/tasks/BaseTaskRenderer');
    var defaultGroupTemplate = require('template-76');

    /**
     * @class BaseGroupRenderer
     * @extends BaseTaskRenderer
     * @constructor
     */
    function BaseGroupRenderer() {
        BaseTaskRenderer.apply(this, arguments);
    }

    BaseGroupRenderer.prototype = _.extend(Object.create(BaseTaskRenderer.prototype), {
        constructor: BaseGroupRenderer,
        /**
         * @method BaseGroupRenderer#render
         * @param {Calendar#Event} group
         * @param {Object} args
         */
        render: function (group, args) {
            return defaultGroupTemplate({
                classes: this.getClasses(args),
                name: _.get(group, 'name', args.locale.calendar.noGroup)
            });
        },

        /**
         * @method BaseGroupRenderer#getClasses
         * @param {Object} args
         */
        getClasses: function (args) {
            var classes = [];
            if (args.outOfView) {
                classes.push('out-of-view');
            }
            return classes;
        }
    });

    return BaseGroupRenderer;
});


define('template-77', ['underscore'], function (_) { return function(data) {
var __t, __p = '';
__p += '<div class="' +
((__t = ( ['event'].concat(data.classes).join(' ') )) == null ? '' : __t) +
'" data-event-id="' +
((__t = ( data.id )) == null ? '' : __t) +
'"> <div class="delete"><i class="pgbu-icon-small pgbu-icon-delete"></i></div> <div class="event-title" title="' +
((__t = ( data.title )) == null ? '' : __t) +
'">' +
((__t = ( data.title )) == null ? '' : __t) +
'</div></div>';
return __p
};});

define('oui/widgets/calendar/allDayEventRenderer',['require','underscore','oui/utils/tasks/BaseTaskRenderer','template-77'],function (require) {
    'use strict';

    var _ = require('underscore');
    var BaseRenderer = require('oui/utils/tasks/BaseTaskRenderer');
    var allDayEventTemplate = require('template-77');

    /**
     * @class AllDayEventRenderer
     * @extends BaseTaskRenderer
     * @constructor
     */
    function AllDayEventRenderer() {
        BaseRenderer.apply(this, arguments);
        this._mediator = null;
    }

    AllDayEventRenderer.prototype = Object.create(BaseRenderer.prototype);

    _.extend(AllDayEventRenderer.prototype, {
        constructor: AllDayEventRenderer,
        /**
         * @method BaseTaskRenderer#render
         * @param {Object} task
         */
        render: function (task) {
            return allDayEventTemplate({
                id: task.id,
                classes: this.getClasses(task),
                title: _.get(task, 'name', '')
            });
        }
    });

    return AllDayEventRenderer;
});

define('oui/widgets/calendar',['require','underscore','moment','jquery','logger','oui/base/widget','oui/widgets/calendar/calendarView','oui/widgets/calendar/calendarModel','oui/widgets/calendar/calendarController','oui/widgets/dateField','template-74','oui/utils/tasks/BaseTaskRenderer','oui/utils/tasks/BaseGroupRenderer','oui/widgets/calendar/allDayEventRenderer'],function (require) { //jshint ignore:line
    'use strict';

    var _ = require('underscore');
    var moment = require('moment');
    var $ = require('jquery');
    var logger = require('logger');
    var Widget = require('oui/base/widget');
    var View = require('oui/widgets/calendar/calendarView');
    var Model = require('oui/widgets/calendar/calendarModel');
    var Controller = require('oui/widgets/calendar/calendarController');
    var DateField = require('oui/widgets/dateField');
    var eventModal = require('template-74');
    var BaseTaskRenderer = require('oui/utils/tasks/BaseTaskRenderer');
    var BaseGroupRenderer = require('oui/utils/tasks/BaseGroupRenderer');
    var AllDayEventRenderer = require('oui/widgets/calendar/allDayEventRenderer');

    /**
     * @event Calendar#view-changed
     * @param {String} newView The newly applied view
     * @param {String} oldView The view that was overwritten
     */
    /**
     * @typedef {Object} Calendar#eventData
     * @property {String} id The unique identifier for an individual event
     * @property {String} [name] The name of the event
     * @property {String} [description] The description of the event
     * @property {moment} startDate The month, day, year and time the event will start
     * @property {moment} endDate The month, day, year and time the event will end
     * @property {*} [*] Additional properties may be attached to events
     */
    /**
     * @typedef {Object} Calendar#Config
     *
     * @property {String} [defaultView='week'] What the starting view of the calendar should use to render. Valid properties
     * are 'week', 'month', 'day'.
     * @property {Array<String>} [availableViews=['day','week','month','year']] The collection of available views for the calendar
     * to render.
     * @property {Boolean|Function} [editable=false] Whether the calendar is editable or used primarily as a readonly view.
     * @property {Boolean} [editor=true] Whether to use the Calendar's internal editor for event creation and updates.
     * @property {Object} [locale] The locale to use within the calendar. This property is completely optional due to the
     * customization of header, toolbar, and event rendering.
     * @property {Locale} [locale] The locale to use within the calendar. The specific strings that are used by the widget and will use the defaults set from
     * the {@link module:oui.lang|lang}.
     * @property {Object} [layout] The layout configuration options.
     * @property {Boolean} [layout.autoResize=true] Whether to automatically resize the view on window resize events.
     * @property {String|Number} [layout.height='100%'] The height to size the calendar when recalculating height.
     * @property {String|Number} [layout.width='100%'] The width to size the calendar when recalculating width.
     * @property {Number} [layout.headerHeight=25] The height of the header row (used for displaying the day/day of week)
     * @property {Boolean} [layout.showToolbar=true] Whether or not to show the toolbar element of the calendar. The toolbar
     * consists of next/prev/today buttons, change view buttons, as well as a global title.
     * @property {Number} [layout.toolbarHeight=48] The default height used for the toolbar when recalculating heights and
     * widths.
     * @property {Object} [eventKeys] The properties of the event that are needed by the calendar component but are provided
     * by the user.
     * @property {String|Array<String>} [eventKeys.id='id'] The identification of an event. Defaults to a single property but
     * may be a composite key using an array.
     * @property {String|Array<String>} [eventKeys.startDate='startDate'] The startDate attribute of the event.
     * @property {String|Array<String>} [eventKeys.endDate='endDate'] The end date attribute of the event.
     * @property {String|Array<String>} [eventKeys.duration='duration'] The duration attribute of the event.
     * @property {Object} storeFns The collection of functions used as a fallback if a store is not provided as the first
     * parameter to constructing the calendar
     * @property {Function} storeFns.getEvents
     * @property {Function} storeFns.getOverlays
     * @property {Function} storeFns.getExceptions
     * @property {Object} storeFns.mapper
     * @property {Object} month The collection of attributes and functions used by the month view.
     * @property {Boolean} [month.highlightToday=true] Whether or not to highlight today on the month view.
     * @property {Number} [month.headerHeight=25] The header height of the month view. Used during resize calculations.
     * @property {Number} [month.rowHeight=22] The row height of an event in the month view.
     * @property {Boolean} [month.numbersSwitchToDayView=false] Whether or not the numbers should switch views when clicked.
     * @property {Boolean} [month.showWeekNumbers=false] Whether or not to show the year's week number in an additional column on the left of the month
     * @property {Boolean} [month.dynamicWeeks=false] Whether or not the month view should hide full weeks that do not belong to the current month.
     *  Default is to show 6 weeks.
     * @property {Function} month.getHeader Function to customize the display of the header in the month view. Default is to print out the long form of the day.
     * @property {Function} [month.showDetails]
     * @property {Object} week The collection of attributes and function used by the week view.
     * @property {Boolean} [week.allDay=false] Whether to show the allDay bar for long events
     * @property {Number} [week.headerHeight=25] The header height of the week view. Used during resize calculations.
     * @property {Boolean} [week.highlightToday=true] Whether or not to highlight today on the week view.
     * @property {Number} [week.rowHeight=42] The row height of the time periods within the week view.
     * @property {String} [week.granularity='half-hour'] What the granularity is for snapping and displaying times. Valid values are 'half-hour' or 'hour'.
     * @property {Function} [week.getHeader] Called to get the format of the header for the view.
     * @property {Object} [week.headerMenu] The options for the header menu within the week view
     * @property {Boolean} [week.headerMenu.enabled=false] Whether or not the header menu is enabled for the week view.
     * @property {Boolean} [week.headerMenu.hideDisabled=false] Whether or not to hide the disabled actions within the header menu.
     * @property {Array} [week.headerMenu.actionOrder=[]] The order of actions of keys used within actions.
     * @property {Object} [week.headerMenu.actions={}] The actions for the header menu.
     * @property {Object} [week.headerMenu.menuFns] The functions used within the menu.
     * @property {Function} [week.headerMenu.menuFns.isActionEnabled] Function called during construction of the menu to determine if the action is enabled.
     * @property {Object} day The collection of attributes and function used by the day view.
     * @property {Boolean} [day.allDay=false] Whether to show the allDay bar for long events
     * @property {Number} [day.headerHeight=25] The header height of the day view. Used during resize calculations.
     * @property {Boolean} [day.highlightToday=true] Whether or not to highlight today on the day view.
     * @property {Number} [day.rowHeight=42] The row height of the time periods within the day view.
     * @property {String} [day.granularity='half-hour'] What the granularity is for snapping and displaying times. Valid values are 'half-hour' or 'hour'.
     * @property {Function} [day.getHeader\ Called to get the format of the header for the view.
     * @property {Object} [day.headerMenu] The options for the header menu within the day view
     * @property {Boolean} [day.headerMenu.enabled=false] Whether or not the header menu is enabled for the day view.
     * @property {Boolean} [day.headerMenu.hideDisabled=false] Whether or not to hide the disabled actions within the header menu.
     * @property {Array} [day,headerMenu.actionOrder=[]] The order of actions of keys used within actions.
     * @property {Object} [day.headerMenu.actions={}] The actions for the header menu.
     * @property {Object} [day.headerMenu.menuFns] The functions used within the menu.
     * @property {Function} [day.headerMenu.menuFns.isActionEnabled] Function called during construction of the menu to determine if the action is enabled.
     * @property {Object} year The collection of attributes and functions used by the year view.
     * @property {Function} [year.getDayHeader] The function to get the day format for the months' week header.
     * @property {Function} [year.getMonthHeader] The function called when rendering a month.
     * @property {Boolean} [year.numbersSwitchToDayView=true] Whether or not the numbers should switch views to the day view.
     * @property {Boolean} [year.monthsSiwtchToMonthView=false] Whether or not the month header should switch to the month view.
     * @property {Object} [dateFormatter] Formatter to use for dates when using the creator dialog.
     * @property {Object} [timeFormatter] Formatter to use for times when using the creator dialog.
     * @property {Object} events The collection of attributes and functions used by event rendering.
     * @property {Boolean} [events.draggable=false] Whether the events are draggable or not within the calendar.
     *  Used globally within the events object instead of within the view specific options to provide a consistent feel.
     * @property {Boolean} [events.allowOverlaps=true] Whether or not the calendar allows creating/updating events to conflict with each other.
     *  Used globally within the events object instead of within the view specific options because
     *  it does not make sense if one view allows conflicts but another does not.
     * @property {Function} [events.isAllDay] Callback to determine if an event belongs in the all day events.
     * @property {Function} [events.getEventName] Used within the rendering process for an event.
     * @property {Function} [events.getEventDuration] Used within the rendering process for an event.
     * @property {Function} [events.getEventStart] Used within the rendering process for an event.
     * @property {Function} [events.getEventDescription] Used within the rendering process for an event.
     * @property {Function} events.getClasses Used to customize the event's during the rendering process.
     * @property {Function} [events.creator] The creator used when a user issues a 'C' or 'E' key press within the calendar.
     * @property {Object} overlays The collection of attributes and functions used by overlay rendering.
     * @property {Function} overlays.getClasses
     * @property {Object} exceptions The collection of attributes and functions used by exception rendering.
     * @property {Function} exceptions.getClasses
     */

    var __DEFAULTS__ = {
        defaultView: 'week',
        availableViews: [
          'day',
          'week',
          'month',
          'year'
        ],
        editable: false,
        editor: true,
        layout: {
            autoResize: true,
            height: '100%',
            width: '100%',
            headerHeight: 25,
            showToolbar: true,
            toolbarHeight: 39,
            legendHeight: 28
        },
        eventKeys: {
            id: 'id',
            //Also allow [ 'id', 'another_key' ]
            startDate: 'startDate',
            endDate: 'endDate',
            duration: 'duration',
            parentKey: 'groupId',
            parent: {
                id: 'id',
                startDate: 'startDate',
                endDate: 'endDate'
            }
        },
        storeFns: {
            getEvents: _.noop,
            mapper: {
                to: _.noop,
                from: _.noop
            }
        },
        month: {
            highlightToday: true,
            headerHeight: 25,
            rowHeight: 22,
            numbersSwitchToDayView: false,
            showWeekNumbers: false,
            dynamicWeeks: false,
            getHeader: function (day) {// jscs:ignore jsDoc
                return day.format('dddd');
            },
            showDetails: _.noop
        },
        week: {
            allDay: false,
            timeFormat: 'ha',
            headerHeight: 25,
            header: null,
            highlightToday: true,
            rowHeight: 42,
            granularity: 'half-hour',

            //valid values are half-hour and hour
            getHeader: function (day) {// jscs:ignore jsDoc
                return day.format('ddd MM/DD');
            },
            headerMenu: {
                enabled: false,
                hideDisabled: false,
                actionOrder: [],
                actions: {},
                menuFns: {
                    isActionEnabled: _.constant(true)
                }
            },
            renderer: new AllDayEventRenderer()
        },
        'task-day': {
            headerHeight: 25,
            highlightToday: true,
            getHeader: function (day) { // jscs:ignore jsDoc
                return day.format('ddd DD');
            },
            renderer: new BaseTaskRenderer(),
            groupRenderer: new BaseGroupRenderer()
        },
        'task-week': {
            headerHeight: 25,
            highlightToday: true,
            getHeader: function (day) { // jscs:ignore jsDoc
                return day.format('ddd DD');
            },
            renderer: new BaseTaskRenderer(),
            groupRenderer: new BaseGroupRenderer()
        },
        day: {
            allDay: false,
            timeFormat: 'ha',
            headerHeight: 25,
            header: null,
            highlightToday: true,
            rowHeight: 42,
            granularity: 'half-hour',

            //valid values are half-hour and hour
            getHeader: function (day) { // jscs:ignore jsDoc
                return day.format('ddd MM/DD');
            },
            headerMenu: {
                enabled: false,
                hideDisabled: false,
                actionOrder: [],
                actions: {},
                menuFns: {
                    isActionEnabled: _.constant(true)
                }
            },
            renderer: new AllDayEventRenderer()
        },
        year: {
            getDayHeader: function (day) { // jscs:ignore jsDoc
                return day.format('ddd');
            },
            getMonthHeader: function (month) { // jscs:ignore jsDoc
                return month.format('MMMM');
            },
            numbersSwitchToDayView: true,
            monthsSwitchToMonthView: false
        },
        dateFormatter: {
            format: function (momentDate) { // jscs:ignore jsDoc
                return momentDate.format('MM/DD/YYYY hh:mm A');
            },
            unformat: function (str) { // jscs:ignore jsDoc
                return moment(str, 'MM/DD/YYYY hh:mm A');
            },
            validate: function (str) { // jscs:ignore jsDoc
                return moment(str, 'MM/DD/YYYY hh:mm A').isValid();
            }
        },
        timeFormatter: {
            format: function (momentDate) {// jscs:ignore jsDoc
                return momentDate.format('hh:mm A');
            },
            unformat: function (str) {// jscs:ignore jsDoc
                return moment(str, 'hh:mm A');
            },
            validate: function (str) {// jscs:ignore jsDoc
                return moment(str, 'hh:mm A').isValid();
            }
        },
        events: {
            draggable: false,
            allowOverlaps: true,
            endOfDay: true,
            singleDay: true,
            isAllDay: function (event) {// jscs:ignore jsDoc
                return event.endDate.diff(event.startDate, 'hours', true) >= 24;
            },
            getEventName: function (event) {// jscs:ignore jsDoc
                return event.name;
            },
            getEventDuration: function (event) {// jscs:ignore jsDoc
                return event.startDate.format('h:mm A') + ' - ' + event.endDate.format('h:mm A');
            },
            getEventStart: function (event) {// jscs:ignore jsDoc
                if (event.startDate.minutes() === 0) {
                    return event.startDate.format('hA');
                }

                return event.startDate.format('h:mmA');
            },
            getEventDescription: _.noop,
            getClasses: _.constant([]),
            // jscs:disable jsDoc
            creator: function (mode, options, initData) { //jshint ignore:line
                // jscs:enable jsDoc
                var deferred = $.Deferred();

                //jscs:disable
                options.locale.title = mode === 'create' ?
                  options.locale.event_create_title :
                  options.locale.event_update_title;

                //jscs:enable

                var $modal = $(eventModal({
                    locale: options.locale
                })).on('hidden', function () {
                    $modal.remove();
                }).appendTo(options.$view);

                var fields = initData || {};

                _.map($modal.find('[type="date"]'), function (elem) {
                    var data;

                    if (initData) {
                        data = initData[$(elem).attr('data-field')];
                    }

                    return new DateField({
                        appendTo: $modal,
                        type: 'datetime',
                        formatter: options.dateFormatter,
                        number: {
                            formatter: options.timeFormatter
                        }
                    }, $(elem), data).init();
                });

                $modal.on('oui-date-change', function (event, widget, name, moment) {
                    fields[$(event.target).attr('data-field')] = moment;

                    if (fields.startDate && fields.endDate) {
                        $modal.find('.modal-accept')
                          .removeAttr('disabled')
                          .removeClass('disabled');
                    } else {
                        $modal.find('.modal-accept')
                          .attr('disabled', 'disabled')
                          .addClass('disabled');
                    }
                });

                $modal.modal('show')
                    .on('click', '.modal-accept', function () {
                        deferred.resolve(fields);
                    });

                return deferred;
            }
        },
        overlays: {
            getClasses: _.constant([])
        },
        exceptions: {
            getClasses: _.constant([])
        }
    };

    /**
     *
     *  ##### **Keyboard Shortcuts**
     *
     *  * <kbd>1</kbd> - Switches to the first available view
     *  * <kbd>2</kbd> - Switches to the second available view
     *  * <kbd>3</kbd> - Switches to the third available view
     *  * <kbd>4</kbd> - Switches to the fourth available view
     *  * <kbd>K</kbd> - Previous unit of time based on view (example. Month moves back 1 month)
     *  * <kbd>P</kbd> - Previous unit of time based on view (example. Month moves back 1 month)
     *  * <kbd>J</kbd> - Next unit of time based on view (example. Month moves forward 1 month)
     *  * <kbd>N</kbd> - Next unit of time based on view (example. Month moves forward 1 month)
     *  * <kbd>Left</kbd> - Highlight previous event
     *  * <kbd>Right</kbd> - Highlight next event
     *  * <kbd>Delete</kbd> - Remove focused event if the calendar is editable
     *  * <kbd>Ctrl+Shift+Up</kbd> - Shift the end of the event up by unit of time (example. 30 minutes)
     *  * <kbd>Ctrl+Shift+Down</kbd> - Shift the end of the event down by a unit of time (example. 30 minutes)
     *  * <kbd>Shift+Up</kbd> - Shift the start of the event up by a unit of time (example. 30 minutes)
     *  * <kbd>Shift+Down</kbd> - Shift the start of the event down by a unit of time (example. 30 minutes)
     *  * <kbd>Up</kbd> - Move the entire event (start and end) up by a unit of time (example. 30 minutes)
     *  * <kbd>Down</kbd> - Move the entire event (start and end) down by a unit of time (example. 30 minutes)
     *  * <kbd>C</kbd> - Creates an event if the calendar is editable and is using the built in creator
     *  * <kbd>E</kbd> - Edits the currently focused event if the calendar is editable and is using the built in creator.
     *  * <kbd>Ctrl+Shift+Right</kbd> - Shift the end date of a task forward by 1 day
     *  * <kbd>Ctrl+Shift+Left</kbd> - Shift the end date of a task backward by 1 day
     *  * <kbd>Alt-Shift-Left</kbd> - Shift the start date of a task backward by 1 day
     *  * <kbd>Alt-Shift-Right</kbd> - Shift the start date of a task forward by 1 day
     *
     *
     *  ##### ** Relevant Events**
     *
     *  * {@link Calendar#date-change|date-change}
     *  * {@link Calendar#busy|busy}
     *  * {@link Calendar#available|available}
     *  * {@link Calendar#event-added|event-added}
     *  * {@link Calendar#event-removed|event-removed}
     *  * {@link Calendar#event-updated|event-updated}
     *
     *  @borrows CalendarModel#event:date-change as Calendar#date-change
     *  @borrows CalendarModel#event:busy as Calendar#busy
     *  @borrows CalendarModel#event:available as Calendar#available
     *  @borrows CalendarModel#event:event-added as Calendar#event-added
     *  @borrows CalendarModel#event:event-removed as Calendar#event-removed
     *  @borrows CalendarModel#event:event-updated as Calendar#event-updated
     *
     * @class Calendar
     * @extends Widget
     *
     * @param {Calendar#Config} [options] The calendar configuration
     * @param {HTMLElement} [element] The element to use for the calendar
     * @param {Function|Array} [initData] Either an array of events to use within the calendar or a constructable Store
     *
     * @tutorial UsingCalendarStore
     */
    function Calendar(options, element, initData) {
        Widget.call(this, options, element, initData);

        var model = new Model(this._options, this._initData);
        this._controller = new Controller(this._options, model);
        this._view = new View(model, this._controller, this._element, this._options, this._settings);

        this._setupEvents([model, this._view]);
    }

    Calendar.TaskRenderer = BaseTaskRenderer;
    Calendar.GroupRenderer = BaseTaskRenderer;
    Calendar.AllDayEventRenderer = AllDayEventRenderer;

    var cp = Calendar.prototype = Object.create(Widget.prototype);

    /**
     * Getter for the calendar widget name
     *
     * @method Calendar#name
     *
     * @returns {String} 'calendar'
     */
    cp.name = function () {
        return 'calendar';
    };

    /**
     * Getter for the default options for the calendar widget
     *
     * @method Calendar#defaults
     *
     * @returns {Object}
     */
    cp.defaults = function () {
        return __DEFAULTS__;
    };

    /**
     * Initializes the calendar widget
     *
     * @method Calendar#_init
     * @private
     *
     * @param {CalendarStore} Store
     *
     * @returns {Object}
     */
    cp._init = function (Store) {
        var instance = this._controller.setStore(Store);
        this._view.forceRedraw();
        this._controller.load(null);

        return instance;
    };

    /**
     * Add an event to the calendar
     *
     * @method Calendar#addEvent
     * @param {Calendar#eventData} eventData The event object to add.
     */
    cp.addEvent = function (eventData) {
        this._controller.addEvent(eventData);
    };

    /**
     * Adds an event to the calendar, but first sets its start and end date properties based on the first visible date in the view
     *
     * @method Calendar#addEventToFirstVisibleDay
     * @param {Calendar#eventData} eventData
     */
    cp.addEventToFirstVisibleDay = function (eventData) {
        this._controller.addEventToFirstVisibleDay(eventData, this._view.getViewRange().startDate);
    };

    /**
     * Add a task to the calendar
     *
     * @method Calendar#addTask
     * @param {moment} startDate The start date of the task.
     * @param {String|Number} [groupId] The ID of the group the task will be in (if any)
     */
    cp.addTask = function (startDate, groupId) {
        this._controller.newTask(startDate, groupId);
    };

    /**
     * Removes an event from the calendar.
     *
     * @method Calendar#removeEvent
     * @param {Number} eventId The id of the event to remove
     */
    cp.removeEvent = function (eventId) {
        this._controller.removeEvent(eventId);
    };

    /**
     * Updates an event that already exists within the calendar
     *
     * @method Calendar#updateEvent
     * @param {Number} eventId The unique ID of the event to update
     * @param {Object} newData The data to merge onto the existing event's data
     */
    cp.updateEvent = function (eventId, newData) {
        this._controller.updateEvent(eventId, newData);
    };

    /**
     * Updates a task that already exists within the calendar
     *
     * @method Calendar#updateTask
     * @param {Number} eventId The unique ID of the event to update
     * @param {Object} newData The data to merge onto the existing task's data
     */
    cp.updateTask = function (eventId, newData) {
        this._controller.updateTask(eventId, newData);
    };

    /**
     * Change the view to the provided view as long as it is a part of the availableViews option.
     *
     * Will not change view to an unavailable view.
     *
     * @method Calendar#changeView
     * @param {String} view - The view to change to
     *
     * @fires Calendar#before-view-change
     * @fires Calendar#view-changed
     */
    cp.changeView = function (view) {
        this._view.changeView(view);
    };

    /**
     * Retrieves the calendar's current date
     * @method Calendar#currentDate
     * @returns {moment}
     */
    /**
     * Updates the calendar's date to the provided new date
     *
     * @method Calendar#currentDate
     * @param {moment} newDate
     */
    cp.currentDate = function (newDate) {
        if (!arguments.length) {
            return this._controller.getCurrentDate();
        } else if (moment.isMoment(newDate)) {
            this._controller.currentDate(newDate);
        } else {
            logger.warn('Calendar.currentDate must be provided a moment object');
        }
    };

    /**
     * Retrieves the calendar's visible start and end dates
     *
     * @method Calendar#getDateRange
     */
    cp.getDateRange = function () {
        return this._view.getDateRange();
    };

    /**
     * Advances the calendar to the next date range in the current view
     *
     * @method Calendar#next
     */
    cp.next = function () {
        this._view.next();
    };

    /**
     * Takes the calendar back to the previous date range in the current view
     *
     * @method Calendar#prev
     */
    cp.prev = function () {
        this._view.prev();
    };

    /**
     * Redraws the current view between the supplied start and end dates.
     *
     * If start and end date are not provided it will redraw the entire view.
     *
     * @method Calendar#redraw
     * @param {moment} [startDate] - The start date to contain the redraw to
     * @param {moment} [endDate] - The end point to contain the redraw to
     */
    cp.redraw = function (startDate, endDate) {
        this._view.redraw(startDate, endDate);
    };

    /**
     * Forces a resize when the window was not resized but the container the calendar belongs to resizes.
     *
     * @method Calendar#resize
     */
    cp.resize = function () {
        this._view.resize();
    };

    /**
     * Cleans cached events and redraws the current view.
     *
     * @method Calendar#repopulate
     */
    cp.repopulate = function () {
        this._controller.clean();
        this._view.redraw();
    };

    /**
     * Disables certain days within the calendar to prevent any manipulation.
     *
     * @method Calendar#disabled
     * @param {Array<moment>} days The days to disable within the view.
     */
    cp.disabled = function (days) {
        if (!arguments.length) {
            return this._controller.disabled();
        }
        this._controller.disabled(days);
    };

    /**
     * Whether or not the calendar is editable.
     *
     * @method Calendar#isEditable
     * @returns {Boolean} whether or not the calendar allows edits.
     */
    cp.isEditable = function () {
        return this._options.editable;
    };

    /**
     * Links the calendar to a hopper component for drag support
     *
     * @method Calendar#linkHopper
     * @param {Hopper} hopper
     */
    cp.linkHopper = function (hopper) {
        var alreadyLinked = this._view._associateHopper(hopper);
        if (!alreadyLinked) {
            hopper.linkCalendar(this);
        }
    };

    /**
     * Sets the task with a given id as the new active and ensures it is in view
     *
     * @method Calendar#focusTask
     * @param {String} id
     * @param {Boolean} [top]
     */
    cp.scrollIntoView = function (id, top) {
        if (top == null) {
            top = true;
        }
        this._controller.active(id, top);
        this._view.callOut();
    };

    /**
     * @method Calendar#active
     *
     * @returns {Object}
     */
    /**
     * @method Calendar#active
     *
     * @param {String|Number} id
     */
    cp.active = function (id) {
        if (!arguments.length) {
            return this._controller.getEvent(this._controller.active()).data;
        }

        this._controller.active(id);
    };

    return Calendar;
});


define('template-78', ['underscore'], function (_) { return function(data) {
var __t, __p = '';
__p += '<div class="modal pgbu-modal hide fade" data-backdrop="static" aria-role="dialog" aria-labelledby="linkDialogHeader"> <div class="modal-header"> <a class="modal-close" href="#" data-dismiss="modal" aria-hidden="true"><i class="pgbu-icon pgbu-icon-dialog-close"></i></a> <h3 id="linkDialogHeader">' +
((__t = ( data.locale.editor.insertLink )) == null ? '' : __t) +
'</h3> </div> <div class="modal-body"> <div class="control-group"> <label class="control-label" for="modal-text-input">' +
((__t = ( data.locale.editor.linkText )) == null ? '' : __t) +
'</label> <div class="controls"> <input class="link-text" type="text" id="modal-text-input"/> </div> </div> <div class="control-group"> <label class="control-label" for="modal-url-input">' +
((__t = ( data.locale.editor.url )) == null ? '' : __t) +
'</label> <div class="controls"> <input class="link-url" type="text" id="modal-url-input"/> <span class="help-block url-help"></span> </div> </div> <div class="control-group"> <label class="control-label" for="link-location"> ' +
((__t = ( data.locale.editor.openIn )) == null ? '' : __t) +
' </label> <div class="controls"> <select id="link-location"> <option value="_blank">' +
((__t = ( data.locale.editor.openInNew )) == null ? '' : __t) +
'</option> <option value="_self">' +
((__t = ( data.locale.editor.openInCurrent )) == null ? '' : __t) +
'</option> </select> </div> </div> </div> <div class="modal-footer"> <button type="button" class="btn btn-blue-dark remove-link pull-left">' +
((__t = ( data.locale.editor.removeLink )) == null ? '' : __t) +
'</button> <button type="button" class="btn btn-blue-dark cancel" data-dismiss="modal" aria-hidden="true">' +
((__t = ( data.locale.cancel )) == null ? '' : __t) +
'</button> <button type="button" class="btn btn-action accept">' +
((__t = ( data.locale.editor.insertLink )) == null ? '' : __t) +
'</button> </div></div>';
return __p
};});

define('oui/widgets/editor/linkDialog',[
    'jquery',
    'underscore',
    'oui/widgets/comboBox',
    'oui/widgets/textField',
    'template-78',
    'bootstrap'
], function ($, _, ComboBox, TextField, dialogTemplate) {
      'use strict';

      var __DEFAULTS__ = {
          'cancel': $.noop,
          'accept': $.noop,
          'remove': $.noop
      };

      var beginsWithRegex = /http(?:s)?:\/\/.*/;

      /**
       * @class LinkDialog
       * @private
       *
       * @param {Editor#Config} options
       */
      function LinkDialog(options) {
          var that = this;

          options = _.defaults(options, __DEFAULTS__);
          this._options = options;

          this._$el = $(dialogTemplate({
              locale: options.locale
          }));

          this._combo = new ComboBox({
              keyboardInput: false,
              appendTo: this._$el
          }, this._$el.find('#link-location')).init();

          var $url = this._$el.find('.link-url');
          var $urlHelp = this._$el.find('.url-help');

          this._url = new TextField({
              validators: [
                  function (text) {
                      if (text.length === 0 || beginsWithRegex.test(_.unescape(text))) {
                          return {
                              valid: true
                          };
                      }

                      return {
                          valid: false,
                          type: 'starts-with',
                          data: text
                      };
                  }
              ]
          }, $url).init();

          this._url.on('error', function () {

              var startsWithError = this._options.locale.errors.startsWith.replace(/{\d+}/g, function () {
                  return $url.closest('.control-group').children('label').text();
              });

              $urlHelp.text(startsWithError);
          }.bind(this));

          this._url.on('change', function () {
              $urlHelp.text('');
          }.bind(this));

          this._$el.on('click', '.accept', function () {
              options.accept({
                  text: that._$el.find('.link-text')[0].value,
                  link: that._$el.find('.link-url')[0].value,
                  location: that._combo.value().value
              });
              that._$el.modal('hide');
          }).on('click', '[data-dismiss=modal]', function (event) {
              event.stopPropagation();
              options.cancel();
          }).on('click', '.remove-link', function () {
              that._$el.modal('hide');
              options.remove();
          }).on('change input populated.oui', function () {
              var disabled =
                  that._$el.find('.link-text').val() === '' ||
                  (that._url.data() === '' || !that._url.isValid());
              that._$el.find('.accept')
                .toggleClass('disabled', disabled)
                .prop('disabled', disabled);
          });
      }

      LinkDialog.prototype = {
          /**
           * @method LinkDialog#show
           */
          show: function () {

              //If the popup is not appended - append it
              if (this._$el.parent().length === 0 && this._options && this._options.appendTo) {
                  $(this._options.appendTo).append(this._$el);
              }

              this._$el.modal('show');
          },
          /**
           * @method LinkDialog#populate
           *
           * @param {Object} linkData
           */
          populate: function (linkData) {
              this._$el.find('.link-text').val(linkData.text);
              this._url.data(linkData.link);
              if (linkData.location !== '') {
                  this._combo.value({
                      value: linkData.location
                  });
              }
              this._$el.trigger('populated.oui');
          },
          /**
           * @method LinkDialog#element
           *
           * @returns {jQuery}
           */
          element: function () {
              return this._$el;
          }
      };

      return LinkDialog;
  });


define('template-79', ['underscore'], function (_) { return function(data) {
var __t, __p = '';
__p += '<div class="orcl-widget orcl-editor"> <div class="toolbar-container"> </div> <div class="iframe-container"> <iframe src="about:blank" seamless="true" class="editor-frame" title="' +
((__t = ( data.locale.editor.title )) == null ? '' : __t) +
'"> </iframe> </div></div>';
return __p
};});


define('template-80', ['underscore'], function (_) { return function(data) {
var __t, __p = '', __j = Array.prototype.join;
function print() { __p += __j.call(arguments, '') }
__p += '<div class="editor-toolbar btn-toolbar"> ';
 if (data.hidden.formatting == null || data.hidden.formatting) { ;
__p += ' <div class="btn-group formatting"> <select class="format"> <option value="h1">Heading 1</option> <option value="h2">Heading 2</option> <option value="h3">Heading 3</option> <option value="h4">Heading 4</option> <option value="h5">Heading 5</option> <option value="h6">Heading 6</option> <option value="p">Paragraph</option> <option value="blockquote">Block Quote</option> <option value="div">Normal</option> </select> </div> ';
 } ;
__p += ' ';
 if (data.hidden.fontName == null || data.hidden.fontName) { ;
__p += ' <div class="btn-group font-name"> <select> <option value="Times New Roman">Times New Roman</option> <option value="Arial">Arial</option> <option value="Georgia">Georgia</option> <option value="Verdana">Verdana</option> <option value="Courier New">Courier New</option> </select> </div> ';
 } ;
__p += ' ';
 if (data.hidden.fontSize == null || data.hidden.fontSize) { ;
__p += ' <div class="btn-group font-size"> <select> <option value="1">8</option> <option value="2">10</option> <option value="3">12</option> <option value="4">14</option> <option value="5">18</option> <option value="6">24</option> <option value="7">36</option> </select> </div> ';
 } ;
__p += ' ';
 if (data.hidden.color == null || data.hidden.color) { ;
__p += ' <div class="btn-group color"> <button type="button" title="' +
((__t = ( data.locale.editor.fontColor )) == null ? '' : __t) +
'" aria-label="' +
((__t = ( data.locale.editor.fontColor )) == null ? '' : __t) +
'" class="btn btn-blue-light font-color"> <i class="pgbu-icon-small pgbu-icon-font-color"></i> </button> <button type="button" title="' +
((__t = ( data.locale.editor.backgroundColor )) == null ? '' : __t) +
'" aria-label="' +
((__t = ( data.locale.editor.backgroundColor )) == null ? '' : __t) +
'" class="btn btn-blue-light background-color"> <i class="pgbu-icon-small pgbu-icon-background-color"></i> </button> </div> ';
 } ;
__p += ' ';
 if (data.hidden.style == null || data.hidden.style) { ;
__p += ' <div class="btn-group style" data-toggle="buttons-checkbox"> <button type="button" title="' +
((__t = ( data.locale.editor.bold )) == null ? '' : __t) +
'" aria-label="' +
((__t = ( data.locale.editor.bold )) == null ? '' : __t) +
'" data-command="bold" class="btn btn-blue-light bold"> <i class="pgbu-icon-small pgbu-icon-bold"></i> </button> <button type="button" title="' +
((__t = ( data.locale.editor.italic )) == null ? '' : __t) +
'" aria-label="' +
((__t = ( data.locale.editor.italic )) == null ? '' : __t) +
'" data-command="italic" class="btn btn-blue-light italic"> <i class="pgbu-icon-small pgbu-icon-italic"></i> </button> <button type="button" title="' +
((__t = ( data.locale.editor.underline )) == null ? '' : __t) +
'" aria-label="' +
((__t = ( data.locale.editor.underline )) == null ? '' : __t) +
'" data-command="underline" class="btn btn-blue-light underline"> <i class="pgbu-icon-small pgbu-icon-underline"></i> </button> <button type="button" title="' +
((__t = ( data.locale.editor.strikeThrough )) == null ? '' : __t) +
'" aria-label="' +
((__t = ( data.locale.editor.strikeThrough )) == null ? '' : __t) +
'" data-command="strikeThrough" class="btn btn-blue-light strikethrough"> <i class="pgbu-icon-small pgbu-icon-strikeout"></i> </button> </div> ';
 } ;
__p += ' ';
 if (data.hidden.script == null || data.hidden.script) { ;
__p += ' <div class="btn-group script" data-toggle="buttons-checkbox"> <button type="button" title="' +
((__t = ( data.locale.editor.subscript )) == null ? '' : __t) +
'" aria-label="' +
((__t = ( data.locale.editor.subscript )) == null ? '' : __t) +
'" data-command="subscript" class="btn btn-blue-light subscript"> <i class="pgbu-icon-small pgbu-icon-subscript"></i> </button> <button type="button" title="' +
((__t = ( data.locale.editor.superscript )) == null ? '' : __t) +
'" aria-label="' +
((__t = ( data.locale.editor.superscript )) == null ? '' : __t) +
'" data-command="superscript" class="btn btn-blue-light superscript"> <i class="pgbu-icon-small pgbu-icon-superscript"></i> </button> </div> ';
 } ;
__p += ' ';
 if (data.hidden.alignment == null || data.hidden.alignment) { ;
__p += ' <div class="btn-group alignment" data-toggle="buttons-radio"> <button type="button" title="' +
((__t = ( data.locale.editor.justifyLeft )) == null ? '' : __t) +
'" aria-label="' +
((__t = ( data.locale.editor.justifyLeft )) == null ? '' : __t) +
'" data-command="justifyLeft" class="btn btn-blue-light left"> <i class="pgbu-icon-small pgbu-icon-align-left"></i> </button> <button type="button" title="' +
((__t = ( data.locale.editor.justifyCenter )) == null ? '' : __t) +
'" aria-label="' +
((__t = ( data.locale.editor.justifyCenter )) == null ? '' : __t) +
'" data-command="justifyCenter" class="btn btn-blue-light center"> <i class="pgbu-icon-small pgbu-icon-align-center"></i> </button> <button type="button" title="' +
((__t = ( data.locale.editor.justifyRight )) == null ? '' : __t) +
'" aria-label="' +
((__t = ( data.locale.editor.justifyRight )) == null ? '' : __t) +
'" data-command="justifyRight" class="btn btn-blue-light right"> <i class="pgbu-icon-small pgbu-icon-align-right"></i> </button> </div> ';
 } ;
__p += ' ';
 if (data.hidden.paragraph == null || data.hidden.paragraph) { ;
__p += ' <div class="btn-group list-indent"> <button type="button" title="' +
((__t = ( data.locale.editor.orderedList )) == null ? '' : __t) +
'" aria-label="' +
((__t = ( data.locale.editor.orderedList )) == null ? '' : __t) +
'" data-command="insertOrderedList" class="btn btn-blue-light ordered-list"> <i class="pgbu-icon-small pgbu-icon-list-ordered"></i> </button> <button type="button" title="' +
((__t = ( data.locale.editor.unorderedList )) == null ? '' : __t) +
'" aria-label="' +
((__t = ( data.locale.editor.unorderedList )) == null ? '' : __t) +
'" data-command="insertUnorderedList" class="btn btn-blue-light unordered-list"> <i class="pgbu-icon-small pgbu-icon-list-unordered"></i> </button> <button type="button" title="' +
((__t = ( data.locale.editor.increaseIndent )) == null ? '' : __t) +
'" aria-label="' +
((__t = ( data.locale.editor.increaseIndent )) == null ? '' : __t) +
'" data-command="indent" class="btn btn-blue-light indent"> <i class="pgbu-icon-small pgbu-icon-indent"></i> </button> <button type="button" title="' +
((__t = ( data.locale.editor.decreaseIndent )) == null ? '' : __t) +
'" aria-label="' +
((__t = ( data.locale.editor.decreaseIndent )) == null ? '' : __t) +
'" data-command="outdent" class="btn btn-blue-light outdent"> <i class="pgbu-icon-small pgbu-icon-outdent"></i> </button> </div> ';
 } ;
__p += ' ';
 if (data.hidden.link == null || data.hidden.link) { ;
__p += ' <div class="btn-group link"> <button type="button" title="' +
((__t = ( data.locale.editor.insertLink )) == null ? '' : __t) +
'" aria-label="' +
((__t = ( data.locale.editor.insertLink )) == null ? '' : __t) +
'" class="btn btn-blue-light link"> <i class="pgbu-icon-small pgbu-icon-link"></i> </button> </div> ';
 } ;
__p += '</div>';
return __p
};});


define('template-81', ['underscore'], function (_) { return function(data) {
var __t, __p = '';
__p += '<a href="' +
((__t = ( data.link )) == null ? '' : __t) +
'" target="' +
((__t = ( data.location )) == null ? '' : __t) +
'">' +
((__t = ( data.text  )) == null ? '' : __t) +
'</a>';
return __p
};});

define('oui/widgets/editor/editorView',['require','underscore','jquery','oui/base/view','oui/widgets/comboBox','oui/widgets/colorPicker','oui/widgets/editor/linkDialog','template-79','template-80','template-81'],function (require) { //jshint maxstatements:false
      'use strict';

      var _ = require('underscore');
      var $ = require('jquery');
      var View = require('oui/base/view');
      var ComboBox = require('oui/widgets/comboBox');
      var ColorPicker = require('oui/widgets/colorPicker');
      var LinkDialog = require('oui/widgets/editor/linkDialog');
      var editorMarkup = require('template-79');
      var toolbarTemplate = require('template-80');
      var linkTemplate = require('template-81');

      var HEADING_TO_VALUE = {
          'Heading 1': 'h1',
          'Heading 2': 'h2',
          'Heading 3': 'h3',
          'Heading 4': 'h4',
          'Heading 5': 'h5',
          'Heading 6': 'h6',
          'Bulleted List': 'div',
          'Numbered List': 'div'
      };

      /**
       * @method EditorView#_setupToolbar~replaceSelection
       * @private
       * @param {document} document
       * @param {Array} currentRanges
       */
      function replaceSelection(document, currentRanges) {
          var selection = document.getSelection();
          if (selection.isCollapsed) {
              selection.removeAllRanges();
              _.each(currentRanges, selection.addRange.bind(selection));

              if (currentRanges.length === 0) {
                  var range = document.createRange();
                  range.setStart(document.body, document.body.childNodes.length);
                  range.setEnd(document.body, document.body.childNodes.length);
                  selection.addRange(range);
              }
          }

          document.body.focus();
      }

      /**
       * @method EditorView~replaceSelectedText
       * @private
       * @param {document} document
       * @param {Array} ranges
       * @param {String} replacementText
       *
       * @returns {String}
       */
      function replaceSelectedText(document, ranges, replacementText) {
          var sel = document.getSelection();
          var range;

          if (sel.rangeCount) {
              range = sel.getRangeAt(0);
          } else if (ranges.length > 0) {
              range = ranges[0];
          }
          range.deleteContents();
          var textNode = document.createTextNode(replacementText);
          range.insertNode(textNode);

          //Set cursor to end of new text node
          range.setStartBefore(textNode);
          range.setEndAfter(textNode);
          sel.removeAllRanges();
          sel.addRange(range);

          return textNode;
      }

      /**
       * @method EditorView~getTrueFontName
       * @private
       *
       * @param {String} fontName
       *
       * @returns {String}
       */
      function getTrueFontName(fontName) {
          if (fontName == null || fontName === '' || fontName === 'Times' || fontName === 'serif') {
              fontName = 'Times New Roman';
          }
          return fontName.replace(/("|')/g, '');
      }

      /**
       * @method EditorView~sameSelection
       * @private
       * @param {document} document
       * @param {Array} currentRanges
       * @returns {Boolean}
       */
      function sameSelection(document, currentRanges) {

          var ranges = document.getSelection().rangeCount;

          if (ranges !== currentRanges.length) {
              return false;
          }

          for (var i = 0; i < ranges; i++) {
              for (var j = 0; j < currentRanges.length; j++) {
                  if (!(currentRanges[j].compareBoundaryPoints(window.Range.END_TO_END, document.getSelection().getRangeAt(i)) === 0 &&
                      currentRanges[j].compareBoundaryPoints(window.Range.START_TO_START, document.getSelection().getRangeAt(i)) === 0)) {

                      return false;
                  }
              }

          }

          return true;
      }

      /**
       * @method EditorView~getTrueFontSize
       * @private
       *
       * @param {String|Number} fontSize
       *
       * @returns {String}
       */
      function getTrueFontSize(fontSize) {
          if (fontSize == null || fontSize === '' || fontSize === 0) {
              fontSize = '3';
          }
          return fontSize.toString();
      }

      /**
       * @method EditorView~getTrueFormat
       * @private
       *
       * @param {String} formatBlock
       *
       * @returns {String}
       */
      function getTrueFormat(formatBlock) {
          if (formatBlock === '' || formatBlock === 'Normal' || formatBlock === 'x') {
              formatBlock = 'div';
          }

          return HEADING_TO_VALUE[formatBlock] || formatBlock;
      }

      var evp;

      /**
       * @class EditorView
       * @extends View
       * @private
       *
       * @param {EditorModel} model
       * @param {EditorController} controller
       * @param {jQuery} element
       * @param {Editor#Config} options
       * @param {Object} settings
       */
      function EditorView(model, controller, element, options, settings) {
          View.call(this, options, element, controller, settings);

          this._$document = null;
          this._$editorBody = null;
          this._$toolbar = this._$el.find('.toolbar-container');
          this._ranges = [];

          this._linkDialog = new LinkDialog({
              locale: options.locale,
              appendTo: options.appendAll,
              accept: this.insertLink.bind(this),
              remove: this.removeLink.bind(this)
          });

          //Set up events from buttons
          this._setupToolbar();

          model.on({
              'change': this.updateDocument.bind(this),
              'character-length-error': this.error.bind(this)
          });

          var iframe = this._$el.find('iframe')[0];

          //Set up listeners for the buttons that are active (queryState)
          //If iframe is completely loaded just call the functions
          //otherwise wait until it has loaded
          if (!_.isNull(iframe.contentWindow) && iframe.contentWindow.document.readyState === 'complete') {
              this.loaded();
          } else {
              this._$el.find('.editor-frame').one('load', this.loaded.bind(this));
          }
      }

      EditorView.prototype = Object.create(View.prototype);
      evp = EditorView.prototype;

      evp.constructor = EditorView;

      /**
       * @method EditorView#loaded
       */
      evp.loaded = function () {
          this.setupIframe(this._$el.find('iframe')[0].contentWindow);
          this._updateToolbar();
          this.updateDocument();
      };

      /**
       * @method EditorView#insertLink
       *
       * @param {Object} linkData
       */
      evp.insertLink = function (linkData) {

          if ((this._ranges.length > 0 && !this._ranges[0].collapsed) ||
              !this._$document[0].queryCommandSupported('insertHTML')) {

              replaceSelection(this._$document[0], this._ranges);

              //Need to unlink the current link if it exists otherwise browsers
              //will keep the old link <a> in the document with no text
              this._$document[0].execCommand('unlink', false);
              var text = replaceSelectedText(this._$document[0], this._ranges, linkData.text);
              this._$document[0].execCommand('createLink', false, linkData.link);
              var link = $(text).closest('a')[0];
              link.target = linkData.location;
              this._controller.document(this._$editorBody.html());
          } else if (this._$document[0].queryCommandSupported('insertHTML')) {
              this._$document[0].execCommand('insertHTML', false, linkTemplate(linkData).trim());
              this._controller.document(this._$editorBody.html());
          }

          this._$editorBody[0].normalize();
          this._$editorBody.focus();
      };

      /**
       * @method EditorView#removeLink
       */
      evp.removeLink = function () {
          this._$document[0].execCommand('unlink');
          this._$editorBody.focus();
          this._controller.document(this._$editorBody.html());
      };

      /**
       * @method EditorView#setupIframe
       *
       * @param {Window} iframeWindow
       */
      evp.setupIframe = function (iframeWindow) {
          var that = this;
          var iframeDoc = iframeWindow.document;

          this._$document = $(iframeDoc);

          this._$editorBody = this._$document.find('body');
          this._$editorBody[0].contentEditable = true;
          this._$editorBody.html(this._controller.document());

          this._$editorBody.on('blur', function () {
              that._$el.trigger('focusout');
              that._$el.removeClass('focused');
          }).on('focus', function () {
              that._$el.trigger('focusin');
              that._$el.addClass('focused');
              var sel = iframeDoc.getSelection();
              if (sel.focusNode == null) {
                  var range = iframeDoc.createRange();
                  range.setStart(iframeDoc.body, iframeDoc.body.childNodes.length);
                  range.setEnd(iframeDoc.body, iframeDoc.body.childNodes.length);
                  sel.addRange(range);
              }
          });

          //Since this is an iframe with no external access to stylesheets
          //just use css attributes inline (shouldn't need anymore than these)
          this._$editorBody.css({
              'margin': 0,
              'padding': '8px',
              'font-family': that._options.defaultFont != null ? that._options.defaultFont : '',
              'box-sizing': 'border-box'
          });

          var selectionchangeSupported = this._$document[0].onselectionchange !== undefined;

          //Setup Editor Selection event handlers
          this._$document.on('beforedeactivate', function () {
              that._updateRanges();
          }).on('selectionchange toolbar-update.oui', function () {
              //Update the toolbar
              that._updateRanges();
              that._updateToolbar();
          }).on('blur change input keyup keydown', function () {
              //IE10-11 does not support input on contenteditable divs
              //http://connect.microsoft.com/IE/feedback/details/794285/ie10-11-input-event-does-not-fire-on-div-with-contenteditable-set
              that._controller.document($(this).find('body').html());
          });

          if (!selectionchangeSupported) {
              this._$document.on('mousedown', function () {
                  //Firefox does not support the select event...so fake it.
                  that._$document.on('mousemove.editor mouseup.editor mouseleave.editor', function () {
                      if (!sameSelection(that._$document[0], that._ranges)) {
                          that._$document.trigger('toolbar-update.oui');
                      }
                  }).on('mouseup.editor mouseleave.editor', function () {
                      that._$document.off('.editor');
                  });
              }).on('keydown keyup', function () {
                  that._$document.trigger('toolbar-update.oui');
              });
          }

      };

      /**
       * @method EditorView#_updateRanges
       * @private
       */
      evp._updateRanges = function () {
          var selection = this._$document[0].getSelection();
          this._ranges = [];
          for (var i = 0; i < selection.rangeCount; i++) {
              this._ranges[i] = selection.getRangeAt(i).cloneRange();
          }

      };

      /**
       * @method EditorView#error
       * @param {Object} error
       */
      evp.error = function (error) {
          if (error == null) {
              error = true;
          }

          this._$el.closest('.control-group')
            .add(this._$el)
            .toggleClass('error', error);
      };

      /**
       * @method EditorView#resize
       */
      evp.resize = function () {
          var iframe = this._$el.find('.editor-frame');
          var parent = iframe.parent();
          var parentOffset = parent[0].offsetHeight;

          if (this._lastSet !== parentOffset && parentOffset !== (iframe[0].offsetHeight + 10)) {
              this._lastSet = parentOffset;
              iframe.addClass('hidden');
              iframe.height(parentOffset - 10);
              iframe.removeClass('hidden');
          }
      };

      /**
       * @method EditorView#_updateToolbar
       * @private
       */
      evp._updateToolbar = function () {
          var doc = this._$document[0];

          if (this._$editorBody[0].contentEditable === 'false' || document.activeElement !== this._$el.find('iframe')[0] || !this._$toolbar.is(':visible')) {
              return;
          }

          if (this._ranges.length === 0 && doc.getSelection() != null && doc.getSelection().rangeCount === 0 && doc.getSelection().isCollapsed) {
              doc.getSelection().removeAllRanges();
              var range = doc.createRange();
              range.setStart(doc.body, 0);
              range.setEnd(doc.body, 0);
              doc.getSelection().addRange(range);
          }

          _.each(this._$toolbar.find('.btn[data-command]'), function (element) {
              $(element).toggleClass('active', doc.queryCommandState($(element).attr('data-command')));
          });

          this._fontName.value({
              value: getTrueFontName(doc.queryCommandValue('fontName'))
          });

          this._fontSize.value({
              value: getTrueFontSize(doc.queryCommandValue('fontSize'))
          });

          this._format.value({
              value: getTrueFormat(doc.queryCommandValue('formatBlock'))
          });
      };

      /**
       * @method EditorView#_getLinkData
       * @private
       *
       * @returns {Object}
       */
      evp._getLinkData = function () {
          var linkData = {
              text: '',
              location: '',
              link: ''
          };

          //If the range is collapsed start with an empty link dialog
          if (this._ranges.length > 0 && !this._ranges[0].collapsed) {
              //We have something selected!
              var anchor = $(this._ranges[0].commonAncestorContainer).closest('a');

              //Needed for IE due to commonAncestorContainer not containing the <a>
              //but may cause issues with too much selected
              if (anchor.length === 0) {
                  anchor = $(this._ranges[0].commonAncestorContainer).children('a:eq(0)');
              }

              if (anchor.length > 0) {
                  //We have an anchor this the text is a part of!
                  linkData.link = anchor.attr('href');
                  linkData.location = anchor.attr('target');
              }

              linkData.text = this._ranges[0].toString();

              //IE does not properly return the string representation on toString
              if (linkData.text === '') {
                  var clone = this._ranges[0].cloneContents();
                  var div = document.createElement('div');
                  div.appendChild(clone);
                  linkData.text = div.textContent;
              }
          }

          return linkData;
      };

      /**
       * @method EditorView#_setupToolbar
       * @private
       */
      evp._setupToolbar = function () {
          var that = this;
          var options = this._options;

          this._$toolbar.on('mousedown', function (event) {
              event.preventDefault();
          });

          this._$toolbar.on('click', '.btn[data-command]', function () {
              that.updateAttribute($(this).attr('data-command'));
          })
        .on('click', '[data-trigger]', function (event) {
            that.trigger($(this).attr('data-trigger'), event);
        });

          this._fontName = new ComboBox({
              fluid: true,
              appendTo: options.appendAll,
              titleText: options.locale.editor.fontName
          }, this._$toolbar.find('.font-name select'))
            .init()
        .on('change', this.updateComboBox.bind(this, 'fontName'));

          this._fontSize = new ComboBox({
              fluid: true,
              appendTo: options.appendAll,
              titleText: options.locale.editor.fontSize
          }, this._$toolbar.find('.font-size select'))
            .init()
        .on('change', this.updateComboBox.bind(this, 'fontSize'));

          this._format = new ComboBox({
              fluid: true,
              appendTo: options.appendAll,
              titleText: options.locale.editor.fontFormat
          }, this._$toolbar.find('.formatting select'))
            .init()
            .on('change', function (value) {
                replaceSelection(that._$document[0], that._ranges);
                var trueValue = value.value;

                that._$document[0].execCommand('formatBlock', false, trueValue);

                if (that._$document[0].queryCommandValue('formatBlock') !== trueValue) {
                    trueValue = '<' + trueValue + '>';
                    that._$document[0].execCommand('formatBlock', false, trueValue);
                }
                that._controller.document(that._$editorBody.html());
            });

          this._background = new ColorPicker({
              appendTo: options.appendAll
          }, this._$toolbar.find('.background-color')).init()
          .on('change', this.updateAttribute.bind(this, 'backColor'));

          this._highlight = new ColorPicker({
              appendTo: options.appendAll
          }, this._$toolbar.find('.font-color')).init()
          .on('change', this.updateAttribute.bind(this, 'foreColor'));

          this._$el.find('.btn.link').on('click', function () {
              that._linkDialog.show();

              //Get name and link and send that to the dialog
              that._linkDialog.populate(that._getLinkData());
          });
      };

      /**
       * @method EditorView#updateAttribute
       *
       * @param {String} attribute
       * @param {String} value
       */
      evp.updateAttribute = function (attribute, value) {
          replaceSelection(this._$document[0], this._ranges);
          this._$document[0].execCommand(attribute, false, value);
          this._controller.document(this._$editorBody.html());
      };

      /**
       * @method EditorView#updateComboBox
       *
       * @param {String} attribute
       * @param {Object} value
       */
      evp.updateComboBox = function (attribute, value) {
          this.updateAttribute(attribute, value.value);
      };

      /**
       * @method EditorView#insert
       *
       * @param {String} text
       */
      evp.insert = function (text) {
          replaceSelection(this._$document[0], this._ranges);
          if (!this._$document[0].queryCommandSupported('insertHTML')) {
              replaceSelectedText(this._$document[0], this._ranges, text);
          } else {
              this._$document[0].execCommand('insertHTML', false, text);
          }

          this._controller.document(this._$editorBody.html());
      };

      /**
       * @method EditorView#buildElement
       *
       * @param {Editor#Config} options
       * @param {jQuery} el
       * @param {EditorController} controller
       *
       * @returns {jQuery}
       */
      evp.buildElement = function (options, el, controller) {
          var $element = $(editorMarkup({
              locale: options.locale
          }));
          var $toolbar;
          var existingMarkup;

          if (options.toolbar == null || options.toolbar.jquery) {
              $toolbar = options.toolbar;
          } else {
              $toolbar = $(toolbarTemplate({
                  hidden: options.toolbar,
                  locale: options.locale
              }));
          }

          if (!_.isNull(el) && el.jquery) {
              existingMarkup = el.html();
              if (existingMarkup.trim().length !== 0) {
                  controller.document(existingMarkup);
              }
              el.empty();
              $element = el.append($element.children());
              $element.addClass('orcl-editor');
          }

          $element.find('.toolbar-container').append($toolbar);

          $element.toggleClass('fill', options.fill);

          return $element;
      };

      /**
       * @method EditorView#_applyDisabled
       * @private
       *
       * @returns {Boolean}
       */
      /**
       * @method EditorView#_applyDisabled
       * @private
       *
       * @param {Boolean} disabled
       */
      evp._applyDisabled = function (disabled) {
          if (!arguments.length) {
              return this._$el.hasClass('disabled');
          }

          this._$el.toggleClass('disabled', disabled);
          this._fontSize.disable(disabled);
          this._fontName.disable(disabled);
          this._format.disable(disabled);
          this._$el.find('.editor-frame').toggleClass('disabled', disabled);
          this._$toolbar.find('.btn-group > .btn').toggleClass('disabled', disabled).prop('disabled', disabled);

          if (this._$editorBody) {
              this._$editorBody[0].contentEditable = !disabled;
          }
      };

      /**
       * @method EditorView#clearAttributes
       */
      evp.clearAttributes = function () {
          this._applyDisabled(false);
          this._applyReadonly(false);
      };

      /**
       * @method EditorView#_applyReadonly
       * @private
       *
       * @param {Boolean} readonly
       */
      evp._applyReadonly = function (readonly) {

          this._$el.toggleClass('readonly', readonly);
          this._fontSize.disable(readonly);
          this._fontName.disable(readonly);
          this._format.disable(readonly);
          this._$el.find('.editor-frame').toggleClass('readonly', readonly);
          this._$toolbar.find('.btn-group > .btn').toggleClass('disabled', readonly).prop('disabled', readonly);

          if (this._$editorBody) {
              this._$editorBody[0].contentEditable = !readonly;
          }
      };

      /**
       * @method EditorView#focus
       * @param {Boolean} focused
       */
      evp.focus = function (focused) {
          if (focused) {
              this._$editorBody.focus();
          }
      };

      /**
       * @method EditorView#updateDocument
       */
      evp.updateDocument = function () {
          var doc = this._controller.document();

          this._$el.closest('.control-group')
              .add(this._$el)
              .removeClass('error');

          //Update the toolbar each time we have a document
          if (doc && this._$document) {
              this._updateToolbar();
          }

          if (this._$editorBody != null && this._$editorBody.html() !== doc) {
              this._$editorBody.html(doc);

          }
      };

      return EditorView;
  }
);

define('oui/widgets/editor/editorController',[], function () {
    'use strict';

    var ecp;

    /**
     * @class EditorController
     * @private
     *
     * @param {EditorModel} model
     * @param {Editor#Config} options
     */
    function EditorController(model, options) {
        this._model = model;
        this._options = options;
    }

    ecp = EditorController.prototype;

    ecp.constructor = EditorController;

    /**
     * @method EditorController#document
     *
     * @returns {String}
     */
    /**
     * @method EditorController#document
     *
     * @param {String} doc
     */
    ecp.document = function (doc) {
        if (!arguments.length) {
            return this._model.document();
        }

        this._model.document(doc);
    };

    return EditorController;
});

define('oui/widgets/editor/editorModel',[
    'oui/base/model',
    'oui/utils/validators/validator'
], function (Model, validator) {
      'use strict';

      /**
       * Fired when the document is updated
       *
       * @event EditorModel#document-changed
       * @param {String} document The current document in the editor.
       */
      /**
       * Fired when the attempted document is longer than max characters
       * specified in the options.
       *
       * @event EditorModel#character-length-error
       * @param {String} doc The attempted document
       * @param {String} document The current document
       */

      var emp;

      /**
       * @class EditorModel
       * @extends Model
       * @private
       *
       *
       * @param {Editor#Config} options The options for the editor
       * @param {String} initData The initial document to set on the editor
       */
      function EditorModel(options, initData) {
          //Super call
          Model.call(this, initData, options);
      }

      EditorModel.prototype = Object.create(Model.prototype);
      emp = EditorModel.prototype;

      emp.constructor = EditorModel;

      /**
       * @method EditorModel#name
       *
       * @returns {String}
       */
      emp.name = function () {
          return 'document';
      };

      /**
       * @method EditorModel#init
       *
       * @param {String} initData
       *
       * @returns {String}
       */
      emp.init = function (initData) {
          return initData || '';
      };

      /**
       * @method EditorModel#getData
       *
       * @returns {String}
       */
      emp.getData = function () {
          return this.isValid() ? this._data : this._validationResult.data;
      };

      /**
       * @method EditorModel#document
       * @returns {String}
       */
      /**
       * @method EditorModel#document
       * @param {String} document
       */
      emp.document = emp.data;

      /**
       * @method EditorModel#validate
       *
       * @param {String} doc
       *
       * @returns {Promise}
       */
      emp.validate = function (doc) {
          var that = this;
          var docLength = this.characters(doc);

          return validator(doc)
              .then(validator.customValidator({
                  cb: function () { //jscs:ignore jsDoc
                      return that._options.maxCharacters == null || docLength <= that._options.maxCharacters;
                  },
                  type: 'character-length'
              }));
      };

      /**
       * @method EditorModel#characters
       *
       * @param {String} string
       *
       * @returns {Number}
       */
      emp.characters = function (string) {
          string = string || this._data;

          if (string == null) {
              return 0;
          }
          return string.replace(/<[^>]*>/g, '').replace(/&[^;]*;/g, 'a').length;
      };

      return EditorModel;
  }
);

define('oui/widgets/editor',[
    'oui/base/widget',
    'oui/widgets/editor/editorView',
    'oui/widgets/editor/editorController',
    'oui/widgets/editor/editorModel'
], function (Widget, View, Controller, Model) {
    'use strict';

    /**
     * @typedef {Object} Editor#ToolbarConfig
     * @property {String} formatting
     * @property {String} fontSize
     * @property {String} fontName
     * @property {String} color
     * @property {String} style
     * @property {String} script
     * @property {String} alignment
     * @property {String} paragraph
     * @property {String} link
     */

    /**
     * @typedef {Object} Editor#Config
     * @property {Locale} [locale={}] The specific strings that are used by the widget and will use the defaults set from
     * the {@link module:oui.lang|lang}.
     * @property {String|null} [defaultFont=null] Font to use in the editor iframe. Useful when there is no font selector in
     * the toolbar and the font should be constant within the editor. Should be a font-family value.
     * @property {Boolean} [showToolbar=true] Whether or not to show the toolbar for the component.
     * @property {Editor#ToolbarConfig|jQuery|null} [toolbar={}] Object that represents which components should be on the toolbar. If a
     * key is undefined it defaults to visible. If the property is a jQuery object it is appended in the toolbar's spot. This
     * allows for custom events.
     * @property {Boolean} [fluid=false] Whether or not the editor should take up as much horizontal space as it can.
     * @property {Boolean} [fill=false] Whether or not the editor should take up as much vertical space as it can.
     * @property {Number|null} [maxCharacters=null] The maximum number of characters allow in the editor.
     */

    var ep;
    var __DEFAULTS__ = {
        defaultFont: null,
        showToolbar: true,
        toolbar: {},
        fluid: false,
        fill: false,
        maxCharacters: null,
        appendAll: 'body'
    };
    var __NAME__ = 'editor';

    /**
     * The common rich text editor. Displays a UI element that can build a text document with
     * rich text features.
     *
     *  ##### **Keyboard Shortcuts**
     *
     *  * <kbd>Ctrl+B</kbd> - Bolds selected text
     *  * <kbd>Ctrl+I</kbd> - Italicizes selected text
     *  * <kbd>Ctrl+U</kbd> - Underlines selected text
     *  * <kbd>Ctrl+A</kbd> - Selects all text
     *
     *
     *  ##### **Relevant Events**
     *
     *  * {@link Editor#document-changed|document-changed}
     *  * {@link Editor#character-length-error|character-length-error}
     *
     *  @borrows EditorModel#event:document-changed as Editor#document-changed
     *  @borrows EditorModel#event:character-length-error as Editor#character-length-error
     *
     * @example
     * <caption>Constructing an Editor (accepts all defaults)</caption>
     *
     * var editor = new oui.Editor();
     * $container.append(editor.element());
     * editor.init();
     *
     * @example
     * <caption>Constructing an Editor with custom toolbar</caption>
     *
     * var editor = new oui.Editor({
     *     toolbar : $toolbar
     * });
     * $container.append(editor.element());
     * editor.init();
     *
     * @class Editor
     * @extends Widget
     * @since 1.3.0
     * @tutorial advanced-editor-usage
     *
     * @param {Editor#Config} [options] The options for the Editor. If not
     *  specified, all the defaults will be used.
     * @param {HTMLDivElement|jQuery} [element] The JQuery element that this field should use.
     *  If not specified, Editor will construct its own element.
     * @param {String} [initData] The document to use.
     */
    function Editor(options, element, initData) {
        var model;

        //Super call
        Widget.call(this, options, element, initData);

        model = new Model(this._options, this._initData);
        this._controller = new Controller(model, this._options);
        this._view = new View(model, this._controller, this._element, this._options, this._settings);

        this._setupEvents([model, this._view]);
    }

    Editor.prototype = Object.create(Widget.prototype);
    ep = Editor.prototype;

    ep.constructor = Editor;

    /**
     * Returns the widget name.
     *
     * @method Editor#name
     * @returns {String} 'editor'
     */
    ep.name = function () {
        return __NAME__;
    };

    /**
     * Gets/Sets the document in the editor.
     *
     * @method Editor#document
     * @fires EditorModel#document-changed
     * @fires EditorModel#character-length-error
     * @param {String} [doc] HTML document fragment to set on the editor.
     * @returns {String} doc The document that is rendered in the editor.
     */
    ep.document = function (doc) {
        if (!arguments.length) {
            return this._controller.document();
        }

        this._controller.document(doc);
    };

    /**
     * Inserts an HTML fragment into the editor document's current cursor location.
     *
     * @method Editor#insert
     * @fires EditorModel#document-changed
     * @fires EditorModel#character-length-error
     * @param {String} text The HTML fragment to set at the current editors cursor location.
     */
    ep.insert = function (text) {
        this._view.insert(text);
    };

    /**
     * Renders the view as an error state.
     *
     * @method Editor#error
     * @param {Boolean} error Whether or not the editor is in an error state.
     */
    ep.error = function (error) {
        this._view.error(error);
    };

    /**
     * @method Editor#defaults
     * @returns {Object} The defaults for the Editor.
     */
    ep.defaults = function () {
        return __DEFAULTS__;
    };

    /**
     * Programmtically resizes the editor when needed due to containers changing size.
     *
     * @method Editor#resize
     */
    ep.resize = function () {
        this._view.resize();
    };

    /**
     * @method Editor#value
     *
     * @returns {String}
     */
    /**
     * @method Editor#value
     *
     * @param {String} doc
     */
    ep.value = function (doc) {
        if (!arguments.length) {
            return this._controller.document();
        }

        this._controller.document(doc, { silent: true });
    };

    return Editor;

});

define('oui/widgets/slider/sliderModel',[
  'underscore',
  'oui/utils/subscribable',
  'oui/utils/math'
], function (_, Subscribable, math) {
    'use strict';

    /**
     * Fired when either of the properties has been updated.
     *
     * @event SliderModel#update
     * @param {Object} data The values of the slider
     * @param {Number} data.start The start value
     * @param {Number} data.end The end value of the slider.
     */
    /**
     * Debounced version of @{link SliderModel#update}
     * Listen to this event if the events are happening too frequently.
     *
     * @event SliderModel#debounced-update
     * @param {Object} data The values of the slider
     * @param {Number} data.start The start value
     * @param {Number} data.end The end value of the slider.
     */
    /**
     * Fired when the start property of the slider has been updated.
     *
     * @event SliderModel#update-start
     * @param {Object} data The values of the slider
     * @param {Number} data.start The start value
     * @param {Number} data.end The end value of the slider.
     */
    /**
     * Debounced version of @{link SliderModel#update-start}
     * Listen to this event if the events are happening too frequently.
     *
     * @event SliderModel#debounced-update-start
     * @param {Object} data The values of the slider
     * @param {Number} data.start The start value
     * @param {Number} data.end The end value of the slider.
     */
    /**
     * Fired when the end property of the slider has been updated.
     *
     * @event SliderModel#update-end
     * @param {Object} data The values of the slider
     * @param {Number} data.start The start value
     * @param {Number} data.end The end value of the slider.
     */
    /**
     * Debounced version of @{link SliderModel#update-end}
     * Listen to this event if the events are happening too frequently.
     *
     * @event SliderModel#debounced-update-end
     * @param {Object} data The values of the slider
     * @param {Number} data.start The start value
     * @param {Number} data.end The end value of the slider.
     */

    /**
     * @class SliderModel
     * @private
     * @param {Slider#Config} options The options for the slider
     * @param {Object} initData The data to initialize the slider with
     */
    function SliderModel(options, initData) {
        Subscribable.call(this);

        this.debouncedTrigger = _.debounce(this.trigger, options.debounce);

        initData = _.defaults({}, initData, {
            start: 0,
            end: 0
        });

        //Start them at default values
        this._startValue = 0;
        this._endValue = 0;

        this._options = options;

        this.data(initData);

        this.linked = [];
    }

    var smp = SliderModel.prototype = Object.create(Subscribable.prototype);

    smp.constructor = SliderModel;

    /**
     * @method SliderModel#_stepStart
     * @private
     *
     * @param {Number} start
     *
     * @returns {Number?}
     */
    smp._stepStart = function (start) {

        if (start == null) {
            return undefined;
        }

        if (this._options.snap != null) {
            var ret = math.step('round', this._options.snap, start);
            if (!_.isNaN(ret)) {
                return +ret.toFixed(12);
            } else {
                return undefined;
            }
        }

        return start;
    };

    /**
     * @method SliderModel#snap
     * @param {Object} data
     *
     * @returns {Object}
     */
    smp.snap = function (data) {
        return _.mapValues(data, this._stepStart.bind(this));
    };

    /**
     * @method SliderModel#link
     *
     * @param {Object} linkOpts
     */
    smp.link = function (linkOpts) {
        this.linked.push(linkOpts);
    };

    /**
     * @method SliderModel#data
     *
     * @returns {Object}
     */
    /**
     * @method SliderModel#data
     *
     * @param {Object} data
     * @param {Subscribable~EventOptions} eventParams
     */
    smp.data = function (data, eventParams) {
        if (!arguments.length) {
            return {
                start: this._startValue,
                end: this._endValue
            };
        }

        var orig = this.data();

        data = _.defaults(this.snap(data), this.data());

        //Clamp to bounds
        data.end = Math.max(Math.min(this._options.max, data.end), this._options.min);
        data.start = Math.min(Math.max(this._options.min, data.start), this._options.max);

        if (data.start != null && data.end < data.start) {
            data.end = this._endValue;
        }

        if (data.end != null && data.start > data.end) {
            data.start = this._startValue;
        }

        this._startValue = data.start;
        this._endValue = data.end;

        if (orig.start !== this._startValue) {
            this._notify('start', this._startValue, eventParams);
        }

        if (orig.end !== this._endValue) {
            this._notify('end', this._endValue, eventParams);
        }

        if (orig.start !== this._startValue || orig.end !== this._endValue) {
            this.trigger(eventParams, 'update', {
                start: this._startValue,
                end: this._endValue
            });
            this.debouncedTrigger(eventParams, 'debounced-update', {
                start: this._startValue,
                end: this._endValue
            });

        }
    };

    /**
     * @method SliderModel#_notify
     * @private
     *
     * @param {String} type
     * @param {Object} data
     * @param {Subscribable~EventOptions} eventParams
     */
    smp._notify = function (type, data, eventParams) {
        this.trigger(eventParams, 'update-' + type, {
            start: this._startValue,
            end: this._endValue
        });
        this.debouncedTrigger(eventParams, 'debounced-update-' + type, {
            start: this._startValue,
            end: this._endValue
        });

        _.chain(this.linked)
            .where({
                source: type
            })
            .each(function (linked) {
                var update = {};
                update[linked.dest] = data;
                linked.slider.value(update);
            })
            .value();
    };

    /**
     * @method SliderModel#max
     */
    /**
     * @method SliderModel#min
     */
    smp.min = smp.max = function () {
        this.data(this.data(), {
            silent: true
        });

        //Data may not change but still need to redraw
        this.trigger({
            silent: true
        }, 'view.update');
    };

    return SliderModel;
});

define('oui/widgets/slider/sliderController',[], function () {
    'use strict';

    /**
     * @class SliderController
     * @private
     *
     * @param {Slider#Config} options
     * @param {SliderModel} model
     */
    function SliderController(options, model) {
        this._options = options;
        this._model = model;
    }

    SliderController.prototype = {
        /**
         * @method SliderController#data
         *
         * @returns {Object}
         */
        /**
         * @method SliderController#data
         *
         * @param {Object} data
         * @param {Subscribable~EventOptions} eventParams
         */
        data: function (data, eventParams) {
            if (!arguments.length) {
                return this._model.data();
            }

            this._model.data(data, eventParams);
        },
        /**
         * @method SliderController#min
         *
         * @param {Number} min
         */
        min: function (min) {
            this._model.min(min);
        },
        /**
         * @method SliderController#max
         *
         * @param {Number} max
         */
        max: function (max) {
            this._model.max(max);
        },
        /**
         * @method SliderController#link
         *
         * @param {Object} linkOpts
         */
        link: function (linkOpts) {
            this._model.link(linkOpts);
        }
    };

    return SliderController;
});


define('template-82', ['underscore'], function (_) { return function(data) {
var __t, __p = '', __j = Array.prototype.join;
function print() { __p += __j.call(arguments, '') }
__p += '<div class="' +
((__t = ( [ 'orcl-widget', 'orcl-slider' ].concat( data.type, data.readonly, data.snap, data.showTicks, data.orientation, data.showBubble ).join(' ') )) == null ? '' : __t) +
'"> <div class="gutter"></div> <div class="scale"> ';
 if (data.showTicks) { ;
__p += ' ';
 for(var i = 0; i <= data.ticks - 1; i++) { ;
__p += ' <div class="slider-tick"></div> ';
 } ;
__p += ' ';
 } ;
__p += ' </div> <div class="handle handle-min" data-field="start" tabindex="0" role="slider"> <div class="value"> </div> </div> <div class="handle handle-max" data-field="end" tabindex="0" role="slider"> <div class="value"> </div> </div> <div class="range"> </div></div>';
return __p
};});


define('template-83', ['underscore'], function (_) { return function(data) {
var __t, __p = '';
__p += '<div class="' +
((__t = ( [ 'orcl-widget', 'orcl-slider' ].concat( data.type, data.readonly, data.snap, data.showTicks, data.orientation, data.showBubble ).join(' ') )) == null ? '' : __t) +
'"></div>';
return __p
};});


define('template-84', ['underscore'], function (_) { return function(data) {
var __t, __p = '', __j = Array.prototype.join;
function print() { __p += __j.call(arguments, '') }
__p += '<div class="gutter"></div><div class="scale"> ';
 if (data.showTicks) { ;
__p += ' ';
 for(var i = 0; i <= data.ticks - 1; i++) { ;
__p += ' <div class="slider-tick"></div> ';
 } ;
__p += ' ';
 } ;
__p += '</div><div class="handle handle-min" data-field="start" tabindex="0" role="slider"> <div class="value"> </div></div><div class="handle handle-max" data-field="end" tabindex="0" role="slider"> <div class="value"> </div></div><div class="range"></div>';
return __p
};});

define('oui/widgets/slider/sliderView',['require','jquery','underscore','oui/base/view','oui/utils/draggable','oui/utils/math','template-82','template-83','template-84','oui/utils/keyCode'],function (require) {
    'use strict';

    var $ = require('jquery');
    var _ = require('underscore');
    var View = require('oui/base/view');
    var draggable = require('oui/utils/draggable');
    var math = require('oui/utils/math');
    var sliderTemplate = require('template-82');
    var wrapTemplate = require('template-83');
    var controlsTemplate = require('template-84');
    var keyCode = require('oui/utils/keyCode');

    var HANDLE_CSS_PROPS = {
        horizontal: 'left',
        vertical: 'top'
    };

    var RANGE_CSS_PROPS = { //jshint ignore:line
        horizontal: {
            min: 'left',
            max: 'right'
        },
        vertical: {
            min: 'bottom',
            max: 'top'
        }
    };

    /**
     * Converts options to a class name used during construction of the slider
     * element
     *
     * @method SliderView~getTemplateOptions
     * @private
     * @param {Slider#Config} options
     * @returns {Object}
     */
    function getTemplateOptions(options) {
        return {
            snap: options.snap != null ? 'discrete' : 'continuous',
            ticks: options.snap != null ? (options.max - options.min) / options.snap : 0,
            showTicks: options.snap != null && options.ticks ? 'show-ticks' : '',
            readonly: !options.editable ? 'read-only' : '',
            orientation: options.orientation,
            showBubble: options.showTooltips && options.snap != null ? 'show-bubbles' : '',
            type: options.type
        };
    }

    /**
     * @class SliderView
     * @extends View
     * @private
     *
     * @param {SliderModel} model
     * @param {SliderController} controller
     * @param {jQuery} element
     * @param {Slider#Config} options
     * @param {Object} settings
     */
    function SliderView(model, controller, element, options, settings) {
        View.call(this, options, element, controller, settings);

        this._$max = this._$el.find('.handle-max');
        this._$min = this._$el.find('.handle-min');
        this._$range = this._$el.find('.range');

        draggable(this._$el, {
            selector: '.handle:not(.disabled), .range:not(.disabled)',
            constrainTo: this._$el
        });

        this._$el.on('drag-start', function (event) {
            if (!this._options.editable) {
                event.preventDefault();
            } else {

                if (_.get(event, 'originalEvent.type') === 'touchstart') {
                    event.originalEvent.preventDefault();
                    event.$handle.focus();
                }

                this._$el.addClass('dragging');
                return this._controller.data();
            }
        }.bind(this));

        this._$el.on('dragging', function (event, data) {
            event.preventDefault();
            var attr = event.$handle.attr('data-field');
            var update = {};
            var value = math.valueFromPercent(event.deltaX / this._$el.width(), {
                max: this._options.max,
                min: this._options.min
            });

            if (attr != null) {
                update[attr] = +(data[attr] + value).toFixed(12);
            } else {

                if (this._options.type !== 'min' && !this._$min.hasClass('disabled')) {
                    update.start = +(data.start + value).toFixed(12);
                }

                if (this._options.type !== 'max' && !this._$max.hasClass('disabled')) {
                    update.end = +(data.end + value).toFixed(12);
                }
            }
            this._controller.data(update);
        }.bind(this));

        this._$el.on('drag-end', function () {
            this._$el.removeClass('dragging');
        }.bind(this));

        this._$el.on('keydown', '.handle', function (event) {
            var attr = $(event.target).attr('data-field');
            var value = this._controller.data();

            var sign = 1;
            var snap = this._options.snap;

            if (snap == null) {
                snap = 1;
            }

            switch (event.keyCode) {
                case keyCode.LEFT:
                case keyCode.DOWN:
                    sign = -1;
                  /* falls through */
                case keyCode.RIGHT:
                case keyCode.UP:

                    this._$el.addClass('dragging');
                    value[attr] = +(value[attr] + (sign * snap)).toFixed(12);

                    this._controller.data(value);
                    break;
            }
        }.bind(this));

        this._$el.on('keyup', '.handle', function () {
            this._$el.removeClass('dragging');
        }.bind(this));

        model.on({
            'view.update': this.update.bind(this)
        });
    }

    var svp = SliderView.prototype = Object.create(View.prototype);

    svp.constructor = SliderView;

    /**
     * @method SliderView#buildElement
     *
     * @param {Slider#Config} options
     * @param {jQuery} element
     *
     * @returns {jQuery}
     */
    svp.buildElement = function (options, element) {
        var $el;

        var templateOptions = getTemplateOptions(options);
        if (element == null) {
            $el = $(sliderTemplate(templateOptions));
        } else {
            $el = element.wrap(wrapTemplate(templateOptions)).parent();
            $el.append(controlsTemplate(templateOptions));
        }

        return $el;
    };

    /**
     * @method SliderView#_getPercentage
     * @private
     *
     * @param {Number} value
     *
     * @returns {Number}
     */
    svp._getPercentage = function (value) {
        return math.percentFromValue(value, {
            max: this._options.max,
            min: this._options.min
        });
    };

    /**
     * @method SliderView#update
     *
     * @param {Object} newData
     */
    svp.update = function (newData) {
        newData = newData || this._controller.data();

        this._$min
          .css(HANDLE_CSS_PROPS[this._options.orientation],
            this._getPercentage(newData.start) * 100 + '%')
          .attr('aria-valuenow', newData.start)
          .attr('aria-valuemin', this._options.min)
          .find('.value')[0].textContent = this._options.formatter.format(newData.start);

        this._$max
          .attr('aria-valuemin', newData.start);

        this._$range.css({
            left: this._getPercentage(newData.start) * 100 + '%'
        });

        this._$max
          .css(HANDLE_CSS_PROPS[this._options.orientation],
            this._getPercentage(newData.end) * 100 + '%')
          .attr('aria-valuenow', newData.end)
          .attr('aria-valuemax', this._options.max)
          .find('.value')[0].textContent = this._options.formatter.format(newData.end);

        this._$min
          .attr('aria-valuemax', newData.end);

        this._$range.css({
            right: (1 - this._getPercentage(newData.end)) * 100 + '%'
        });
    };

    /**
     * @method SliderView#init
     */
    svp.init = function () {
        this.update(this._controller.data());
    };

    /**
     * @method SliderView#type
     *
     * @param {String} type
     */
    svp.type = function (type) {

        this._$el.removeClass(this._options.type);

        this._options.type = type;

        this._$el.addClass(type);
    };

    /**
     * @method SliderView#applyDisabled
     *
     * @param {Object} type
     */
    svp._applyDisabled = function (type) {
        this._$el.find('.disabled').removeClass('disabled');

        this._$min.toggleClass('disabled', type === 'all' || type === 'min' || type === 'handles');
        this._$max.toggleClass('disabled', type === 'all' || type === 'max' || type === 'handles');
        this._$range.toggleClass('disabled', type === 'all' || type === 'range');
    };

    return SliderView;
});

define('oui/formatters/number',[
    'underscore'
], function (_) {
    'use strict';

    /**
     * @module formatters/number
     */
    return {
        /**
         * @method module:formatters/number.format
         *
         * @param {Number} data
         *
         * @returns {String}
         */
        format: function (data) {
            return data.toString();
        },
        /**
         * @method module:formatters/number.unformat
         *
         * @param {String} str
         *
         * @returns {Number}
         */
        unformat: _.partialRight(parseInt, 10),
        /**
         * @method module:formatters/number.validate
         *
         * @returns {Boolean}
         */
        validate: _.constant(true)
    };
});

define('oui/widgets/slider',['require','oui/base/widget','oui/widgets/slider/sliderModel','oui/widgets/slider/sliderController','oui/widgets/slider/sliderView','oui/formatters/number'],function (require) {
    'use strict';

    var Widget = require('oui/base/widget');
    var Model = require('oui/widgets/slider/sliderModel');
    var Controller = require('oui/widgets/slider/sliderController');
    var View = require('oui/widgets/slider/sliderView');
    var baseNumberFormatter = require('oui/formatters/number');

    /**
     * @typedef {Object} Slider#Config
     *
     * @property {Boolean} [editable=true] Whether the slider is editable from the UI. This option does not prevent
     * programmatically updating the slider.
     * @property {String} [output=none] Whether the slider renders output values. Valid values are 'min', 'max', 'both' or
     * 'none'.
     * @property {String} [type='min'] The type of slider represented in the UI. Valid values are 'min', 'max', 'range' and
     * 'scale'.
     * @property {Number|moment} [min=0] The minimum value represented by the slider.
     * @property {Number|moment} [max=100] The maximum value represented by the slider.
     * @property {Boolean} [fluid=true] Whether the slider should take up all the available width of the parent.
     * @property {String} [orientation='horizontal'] Which direction should slider slide (left/right or up/down).
     * @property {Object} [formatter] The formatter used by the slider. Both for output values as well as tooltips.
     * @property {Number|moment} [snap=null] The increment to snap to. Used by keyboard operation to deal with incrementing
     * and decrementing.
     * @property {Boolean} [ticks=false] Whether or not to show snapping ticks within the slider.
     * @property {Boolean} [tooltips=false] Whether or not to display tooltips when sliding the handle.
     * @property {Number} [debounce=250] Speed at which the debounced events fire.
     * @property {Locale} [locale={}] The specific strings that are used by the widget and will use the defaults set from
     * the {@link module:oui.lang|lang}.
     */

    var _DEFAULTS_ = {
        editable: true,
        output: 'none',

        //min, max, both, none
        type: 'min',

        //min, max, range, scale
        min: 0,

        // number, date
        max: 100,
        fluid: true,
        orientation: 'horizontal',

        //vertical, horizontal
        formatter: baseNumberFormatter,
        snap: null,

        //what increment to snap to (if date ms to snap to)
        ticks: false,
        tooltips: false,
        debounce: 250
    };

    /**
     *
     * See {@link http://www.w3.org/TR/wai-aria-practices/#slider|Slider ARIA Best Practices} and
     * {@link http://www.w3.org/TR/wai-aria-practices/#slidertwothumb|Slider Mulit Thumb ARIA Best Practices}
     * for information regarding keyboard shortcuts and general ARIA attributes and states.
     *
     *  ##### **Keyboard Shortcuts**
     *
     *  * <kbd>Left</kbd> - Decrements the slider by step
     *  * <kbd>Down</kbd> - Decrements the slider by step
     *  * <kbd>Right</kbd> - Increments the slider by step
     *  * <kbd>Up</kbd> - Increments the slider by step
     *
     *
     *  ##### **Relevant Events**
     *
     *  * {@link Slider#update|update}
     *  * {@link Slider#debounced-update|debounced-update}
     *  * {@link Slider#update-start|update-start}
     *  * {@link Slider#debounced-update-start|debounced-update-start}
     *  * {@link Slider#update-end|update-end}
     *  * {@link Slider#debounced-update-end|debounced-update-end}
     *
     *  @borrows SliderModel#event:update as Slider#update
     *  @borrows SliderModel#event:debounced-update as Slider#debounced-update
     *  @borrows SliderModel#event:update-start as Slider#update-start
     *  @borrows SliderModel#event:debounced-update-start as Slider#debounced-update-start
     *  @borrows SliderModel#event:update-end as Slider#update-end
     *  @borrows SliderModel#event:debounced-update-end as Slider#debounced-update-end
     *
     * @class Slider
     * @extends Widget
     *
     * @param {Slider#Config} [options] The options used by the slider
     * @param {jQuery} [element] The element used as the basis for the slider
     * @param {Object} [initData] The initial data used by the slider
     * @param {Number} initData.start The start handle value for the slider
     * @param {Number} initData.end The end handle value for the slider
     *
     * @tutorial using-slider
     */
    function Slider(options, element, initData) {
        Widget.call(this, options, element, initData);

        var model = new Model(this._options, this._initData);
        this._controller = new Controller(this._options, model);

        this._view = new View(model, this._controller, this._element, this._options, this._settings);

        this._setupEvents(model);
    }

    var sp = Slider.prototype = Object.create(Widget.prototype);

    sp.constructor = Slider;

    /**
     * @method Slider#name
     *
     * @returns {String}
     */
    sp.name = function () {
        return 'slider';
    };

    /**
     * Gets the default options for the slider.
     *
     * @method Slider#defaults
     * @private
     * @returns {Object} The default options for the widget
     */
    sp.defaults = function () {
        return _DEFAULTS_;
    };

    /**
     * Updates the minimum value represented by the slider and rerenders the slider.
     *
     * @method Slider#min
     * @param {Number|moment} min The new min value represented by the slider.
     */
    sp.min = function (min) {
        this._options.min = min;
        this._controller.min(min);
    };

    /**
     * Updates the maximum value represented by the slider and rerenders the slider.
     *
     * @method Slider#max
     * @param {Number|moment} max The new max value represented by the slider.
     */
    sp.max = function (max) {
        this._options.max = max;
        this._controller.max(max);
    };

    /**
     * Silently updates the sliders data
     *
     * @method Slider#value
     * @param {Object} newData - The new data to set on the slider.
     * @param {Number|moment} newData.start The low value on the slider.
     * @param {Number|moment} newData.end The high value on the slider.
     */
    sp.value = function (newData) {
        if (!arguments.length) {
            return this._controller.data();
        }
        this._controller.data(newData, {
            silent: true
        });
    };

    /**
     * Updates the sliders data
     *
     * @method Slider#data
     * @param {Object} newData - The new data to set on the slider.
     * @param {Number|moment} newData.start The low value on the slider.
     * @param {Number|moment} newData.end The high value on the slider.
     *
     * @fires Slider#update
     */
    sp.data = function (newData) {
        if (!arguments.length) {
            return this._controller.data();
        }
        this._controller.data(newData);
    };

    /**
     * Link changes in slider to another slider.
     *
     * Link opts contains:
     *     dest, source, bidrectional
     *
     * @method Slider#link
     * @param {Object} linkOpts Destination, source and bidirectional attributes for the link.
     * @param {Boolean} [followBidirectional] Whether or not the link should exist on both sliders. Used internally to prevent
     * infinite linking between sliders.
     */
    sp.link = function (linkOpts, followBidirectional) {
        this._controller.link(linkOpts);

        if (linkOpts.bidirectional && (followBidirectional == null || followBidirectional === true)) {
            linkOpts.slider.link({
                slider: this,
                dest: linkOpts.source,
                source: linkOpts.dest,
                bidirectional: true
            }, false);
        }
    };

    /**
     * Changes the type of the slider between 'min', 'max', 'range', 'scale'.
     *
     * @method Slider#type
     * @param {String} type The new type of the slider.
     */
    sp.type = function (type) {
        if (type === 'min') {
            this._controller.data({
                start: this._options.min
            });
        } else if (type === 'max') {
            this._controller.data({
                end: this._options.max
            });
        }
        this._view.type(type);
    };

    /**
     * Disables portions of the slider.
     *
     * Valid types are 'all', 'min', 'max', 'range', 'handles'
     *
     * @method Slider#disable
     * @param {String} type Which portions of the slider are disabled
     */
    sp.disable = function (type) {
        //all, min, max, range, handles
        this._view.disable(type);
    };

    /**
     * Checks arguments for initData when mapping the optional arguments.
     *
     * @method Slider#isInitData
     * @private
     * @param {Object|*} args The argument to check.
     * @returns {Boolean} Whether or not the argument is the initialization data for the slider
     */
    sp.isInitData = function (args) {
        return !args.jquery && (args.start != null || args.end != null);
    };

    return Slider;
});


define('template-85', ['underscore'], function (_) { return function(data) {
var __t, __p = '';
__p += '<div class="orcl-widget orcl-double-picker ' +
((__t = ( data.sortable ? 'sortable' : 'no-sort' )) == null ? '' : __t) +
' ' +
((__t = ( data.vertical ? 'vertical' : 'horizontal' )) == null ? '' : __t) +
'"> <div class="source-picker"> <label id="' +
((__t = ( 'source-picker-label-' + data.id )) == null ? '' : __t) +
'" class="control-label header-label hidden"></label> </div> <div class="picker-controls picker-controls-center"> <button type="button" title="' +
((__t = ( data.locale.right )) == null ? '' : __t) +
'" aria-label="' +
((__t = ( data.locale.right )) == null ? '' : __t) +
'" class="not-btn btn-move-right" disabled> <i class="pgbu-icon pgbu-icon-menu-forward"></i> </button> <button type="button" title="' +
((__t = ( data.locale.left )) == null ? '' : __t) +
'" aria-label="' +
((__t = ( data.locale.left )) == null ? '' : __t) +
'" class="not-btn btn-move-left" disabled> <i class="pgbu-icon pgbu-icon-menu-back"></i> </button> </div> <div class="dest-picker"> <label id="' +
((__t = ( 'dest-picker-label-' + data.id )) == null ? '' : __t) +
'" class="control-label header-label hidden"></label> </div> <div class="picker-controls picker-controls-right"> <button type="button" title="' +
((__t = ( data.locale.up )) == null ? '' : __t) +
'" aria-label="' +
((__t = ( data.locale.up )) == null ? '' : __t) +
'" class="not-btn btn-move-up" disabled> <i class="pgbu-icon pgbu-icon-chevron-up"></i> </button> <button type="button" title="' +
((__t = ( data.locale.down )) == null ? '' : __t) +
'" aria-label="' +
((__t = ( data.locale.down )) == null ? '' : __t) +
'" class="not-btn btn-move-down" disabled> <i class="pgbu-icon pgbu-icon-chevron-down"></i> </button> </div></div>';
return __p
};});


define('template-86', ['underscore'], function (_) { return function(data) {
var __t, __p = '';
__p += '<button type="button" class="btn btn-picker" title="' +
((__t = ( data.locale.picker.select  )) == null ? '' : __t) +
'" aria-label="' +
((__t = ( data.locale.picker.select )) == null ? '' : __t) +
'" > <i class="pgbu-icon-small pgbu-icon-picker-dots"></i></button>';
return __p
};});


define('template-87', ['underscore'], function (_) { return function(data) {
var __t, __p = '';
__p += '<div class="orcl-widget orcl-double-picker orcl-picker-input"> <input type="text"> <button type="button" tabindex="-1" class="btn btn-picker" title="' +
((__t = ( data.locale.picker.select )) == null ? '' : __t) +
'"> <i class="pgbu-icon-small pgbu-icon-picker-dots"></i> </button></div>';
return __p
};});


define('template-88',[],function () { return '<div class="orcl-widget orcl-picker-input">\n</div>\n';});


define('template-89', ['underscore'], function (_) { return function(data) {
var __t, __p = '', __j = Array.prototype.join;
function print() { __p += __j.call(arguments, '') }
__p += '<div class="modal pgbu-modal hide fade orcl-picker-modal orcl-double-picker-modal" data-backdrop="static" tabindex="-1" role="dialog" aria-hidden="true" aria-labelledby="' +
((__t = ( data.modalId )) == null ? '' : __t) +
'"> <div class="modal-header"> <a title="' +
((__t = ( data.locale.close )) == null ? '' : __t) +
'" class="modal-close" href="#" data-dismiss="modal"><i class="pgbu-icon pgbu-icon-dialog-close"></i></a> <h3 id="' +
((__t = ( data.modalId )) == null ? '' : __t) +
'" class="modal-title">' +
((__t = ( data.locale.title  )) == null ? '' : __t) +
'</h3> </div> <div class="modal-body"></div> <div class="modal-footer"> ';
 if(!data.required) { ;
__p += ' <button type="button" title="' +
((__t = ( data.locale.remove )) == null ? '' : __t) +
'" class="btn btn-blue-dark modal-clear pull-left" data-dismiss="modal">' +
((__t = ( data.locale.remove  )) == null ? '' : __t) +
'</button> ';
 } ;
__p += ' <button type="button" title="' +
((__t = ( data.locale.cancel )) == null ? '' : __t) +
'" class="btn btn-blue-dark modal-cancel" data-dismiss="modal">' +
((__t = ( data.locale.cancel  )) == null ? '' : __t) +
'</button> <button type="button" title="' +
((__t = ( data.locale.ok )) == null ? '' : __t) +
'" class="btn btn-action modal-accept disabled" data-dismiss="modal" disabled="disabled">' +
((__t = ( data.locale.ok  )) == null ? '' : __t) +
'</button> </div></div>';
return __p
};});


define('template-90', ['underscore'], function (_) { return function(data) {
var __t, __p = '';
__p += '<div class="orcl-widget orcl-double-picker ' +
((__t = ( data.sortable ? 'sortable' : 'no-sort' )) == null ? '' : __t) +
'"></div>';
return __p
};});

define('oui/widgets/doublePicker/doublePickerView',['require','underscore','jquery','oui/base/view','oui/widgets/treePanel','oui/widgets/picker','oui/utils/flexEqual','template-85','template-86','template-87','template-88','template-89','template-90','oui/utils/keyCode','oui/mixins/shortcut','bootstrap'],function (require) { //jshint ignore:line
      'use strict';

      var _ = require('underscore');
      var $ = require('jquery');
      var View = require('oui/base/view');
      var TreePanel = require('oui/widgets/treePanel');
      var Picker = require('oui/widgets/picker');
      var flexEqual = require('oui/utils/flexEqual');
      var pickerTemplate = require('template-85');
      var pickerButtonTemplate = require('template-86');
      var pickerInputTemplate = require('template-87');
      var pickerWrapInputTemplate = require('template-88');
      var pickerModalTemplate = require('template-89');
      var pickerContainerTemplate = require('template-90');
      var keyCode = require('oui/utils/keyCode');
      var shortcutMixin = require('oui/mixins/shortcut');
      require('bootstrap'); //Unused as variable but required for bootstrap modal js

      /**
       * Fired when the selection of nodes has stopped and the results are committed.
       *
       * @event DoublePickerView#accept
       * @param {Array<Object>} data The accepted nodes from the Double Picker
       */
      /**
       * Fired when the picker actions taken while it has been opened are cancelled.
       *
       * @event DoublePickerView#cancel
       */
      /**
       * Fired when nodes move from the available side of the double picker to the selected side.
       *
       * @event DoublePickerView#nodes-selected
       * @param {Array<Object>} nodes The currently selected nodes of the Double Picker
       */
      /**
       * Fired when nodes move from the selected side of the double picker to the available side.
       *
       * @event DoublePickerView#nodes-deselected
       * @param {Array<Object>} nodes The deselected nodes
       */

      var dpvp;

      /**
       * @method DoublePickerView~isDataEqual
       * @private
       *
       * @param {Array} newData
       * @param {Array} currentData
       *
       * @returns {Boolean}
       */
      var isDataEqual = function (newData, currentData) {
          var selected = newData || this._dest.tree({ showRoot: false }) || [];

          if (currentData.length !== selected.length) {
              return false;
          }

          return _.every(selected, function (node) {
              return _.find(currentData, flexEqual.bind(null, node));
          });
      };

      /**
       * @method DoublePickerView~resetHandler
       * @private
       *
       * @param {Array} data
       */
      var resetHandler = function (data) {
          if (this._$modal) {
              var disabled = isDataEqual.call(this, data, this.model.data);

              if (disabled && this._options.repeatable) {
                  disabled = false;
              }

              if (!disabled && (this._options.required && (data == null || data.length === 0))) {
                  disabled = true;
              }

              this._$modal.find('.modal-accept')
                .prop('disabled', disabled)
                .toggleClass('disabled', disabled);

              this._$modal.find('.modal-clear')
                .prop('disabled', _.isEqual(this.model.data, []))
                .toggleClass('disabled', _.isEqual(this.model.data, []));
          }
      };

      //TODO: investigate this function
      /**
       * @method DoublePickerView~cancelHandler
       * @private
       */
      var cancelHandler = function () {
          var currentData = this.model.data;
          this.trigger('cancel');
          this._source.disabled(currentData);

          //Invalidating the picker currently wipes out the model data which
          //causes problems for add and updating value
          this._dest.invalidate();

          //Reset the model data to the current data
          this.model.data = currentData;
          this._dest.add(currentData);
          this._updateValue();
          resetHandler.call(this);
      };

      /**
       * @class DoublePickerView
       * @extends View
       * @private
       *
       * @param {DoublePicker#Config} options
       * @param {jQuery} $element
       * @param {Object} settings
       * @param {Array} initData
       */
      function DoublePickerView(options, $element, settings, initData) {
          var that = this;
          View.call(this, options, $element, null, settings);

          function isNew(node) {// jscs:ignore jsDoc
              var newNode = !_.find(that.model.data, function (modelNode) {
                  return _.isEqual(modelNode, _.pick(node, _.keys(modelNode)));
              });
              if (newNode) {
                  return ['new-item'];
              }
          }

          initData = _.defaults({}, initData, {
              selected: [],
              available: []
          });

          this.model = {
              _data: initData.selected,
              init: false
          };

          Object.defineProperty(this.model, 'data', {
              get: function () { //jscs:ignore jsDoc
                  return this._data;
              },
              set: function (newData) { //jscs:ignore jsDoc
                  this._data = newData;
                  resetHandler.call(that, newData);
              }
          });

          this._source = new Picker(options.source, initData.available);

          if (_.has(options, 'dest.viewFns.classes')) {
              options.dest.viewFns.classes = _.wrap(options.dest.viewFns.classes, function (fn, node) {
                  return _.union(fn(node), isNew(node));
              });
          } else {
              options.dest.viewFns.classes = function (node) {// jscs:ignore jsDoc
                  return isNew(node);
              };
          }

          this._dest = new TreePanel(options.dest, initData.selected);

          this._source.disabled(_.union(this._source.disabled(), initData.selected));

          this._$moveRight = this._$el.find('.btn-move-right');
          this._$moveLeft = this._$el.find('.btn-move-left');
          this._$moveUp = this._$el.find('.btn-move-up');
          this._$moveDown = this._$el.find('.btn-move-down');

          if (options.modal) {
              this._setupModal();
              this._updateValue();
          } else {
              this._appendElements();
              this._initListeners();
          }

          //TODO: handle bidirectional linking within the link call?
          this._source.treePanel().link([this._dest]);
          this._dest.link([this._source.treePanel()]);
      }

      dpvp = DoublePickerView.prototype = Object.create(View.prototype);

      _.extend(dpvp, shortcutMixin);

      /**
       * @method DoublePickerView#_setupModal
       * @private
       */
      dpvp._setupModal = function () {
          var that = this;
          var $doublePickerEl = $(pickerTemplate({
              id: this._settings.id,
              locale: this._options.locale,
              sortable: this._options.allowReordering,
              vertical: this._options.vertical
          }));

          if (_.isObject(this._options.modal)) {
              this._$modal = this._options.modal;
          } else {
              this._$modal = $(pickerModalTemplate({
                  modalId: 'oui-' + this._settings.name + '-' + this._settings.id + '-modal-label',
                  required: this._options.required,
                  locale: this._options.locale
              }));

          }

          this._$modal.addClass(this._options.modalClasses);

          resetHandler.call(this);
          this._$moveRight = $doublePickerEl.find('.btn-move-right');
          this._$moveLeft = $doublePickerEl.find('.btn-move-left');
          this._$moveUp = $doublePickerEl.find('.btn-move-up');
          this._$moveDown = $doublePickerEl.find('.btn-move-down');

          this._$modal.on('show', function () {
              if (that._$modal.parent().length === 0) {
                  $(that._options.appendTo).append(that._$modal);
              }
          });

          this._$modal.on('hidden', function () {
              that._$modal.detach();
          });

          this._$modal.on('click', '.modal-accept', function () {
              that.model.data = that._dest.tree({ showRoot: false }) || [];
              that.trigger('accept', that.model.data);
              that._$el.trigger('change');
              resetHandler.call(that);
              that._updateValue();

          }).on('click', '.modal-clear', function () {
              that.model.data = [];
              that._dest.invalidate();
              that.trigger('accept', []);
          }).on('click', '[data-dismiss=modal]', function (e) {
              e.stopPropagation();
              e.preventDefault();
              that.hide();

              if (!$(this).hasClass('modal-accept')) {
                  cancelHandler.call(that);
              }
          });

          this._$modal.on('keyup', function (e) {
              switch (e.keyCode) {
                  case keyCode.ESC:
                      e.stopPropagation();
                      that.hide();
                      cancelHandler.call(that);
                      break;
              }
          });

          //TODO: add listeners to enable/disable
          this._$modal.find('.modal-body').append($doublePickerEl);

          this._appendElements($doublePickerEl);
          this._initListeners($doublePickerEl);

          this._$el.find('.btn').addBack('.btn')
                .on('click', function () {
                    that.show();
                })
                .on('mousedown', function (event) {
                    event.preventDefault();
                });

          this._$el.on('click', 'input[readonly]', function () {
              if (that._options.openFromInput && !that._$modal.hasClass('in')) {
                  that.show();
              }
          });

          this.addShortcuts({
              'F2 SHIFT-F10': {
                  fn: function () {// jscs:ignore jsDoc
                      that.show();
                  },
                  prevention: true
              }
          });

          this.element().on('oui-select', function (event, widget, eventName, selected) {
              if (widget.name() === 'typeahead') {
                  var nodesToAdd = _.filter(selected, function (node) {
                      return that._source.treePanel().node(node) == null;
                  });
                  var nodesToRemove;

                  if (widget.options().allowNewValues) {
                      that._source.treePanel().add(nodesToAdd);
                  }

                  that._dest.add(_.cloneDeep(selected));

                  nodesToRemove = _.map(selected, function (node) {
                      return that._dest.node(node);
                  });

                  nodesToRemove = _.reject(that._dest.nodes(), function (node) {
                      return _.find(nodesToRemove, function (removeNode) {
                          return _.isEqual(node, removeNode);
                      });
                  });

                  that._dest.remove(_.cloneDeep(nodesToRemove));
                  that.model.data = that._dest.tree({ showRoot: false });
                  that._source.selected([]);
                  that._source.disabled(that._dest.nodes());
              }
          });

          this._$modal.one('shown.init', function () {
              that._source.init();
              that._dest.init();
          });
      };

      /**
       * @method DoublePickerView#_appendElements
       * @private
       *
       * @param {jQuery} $el
       */
      dpvp._appendElements = function ($el) {
          var element = $el || this._$el;

          if (this._options.labels.source != null && this._options.labels.source !== '') {
              element.find('.source-picker > .control-label')
                  .removeClass('hidden')
                  .text(this._options.labels.source)
                  .attr('title', this._options.labels.source)
                  .attr('for', this._source.treePanel().id());
          }

          if (this._options.labels.dest != null && this._options.labels.dest !== '') {
              element.find('.dest-picker > .control-label')
                  .removeClass('hidden')
                  .text(this._options.labels.dest)
                  .attr('title', this._options.labels.dest)
                  .attr('for', this._dest.id());
          }

          element.find('.source-picker .tree').attr('aria-labelledby', this._$el.find('.source-picker > .control-label').attr('id'));

          element.find('.dest-picker .tree').attr('aria-labelledby', this._$el.find('.dest-picker > .control-label').attr('id'));

          element.find('.source-picker').append(this._source.element());
          element.find('.dest-picker').append(this._dest.element());
      };

      /**
       * @method DoublePickerView#_initDOMListeners
       * @private
       *
       * @param {jQuery} $el
       */
      dpvp._initDOMListeners = function ($el) {
          var that = this;
          $el
            .on('click', '.btn-move-right', function () {
                that._moveRight();
                that._$moveDown.prop('disabled', !that._canMove().down);
            })
            .on('click', '.btn-move-left', function () {
                that._moveLeft();
            })
            .on('click', '.btn-move-up', function () {
                that._moveUp();
            })
            .on('click', '.btn-move-down', function () {
                that._moveDown();
            })
            .on('keydown', function (event) {
                if ((event.ctrlKey || event.metaKey) && event.which === keyCode.LEFT) {
                    that._moveLeft();
                }

                if ((event.ctrlKey || event.metaKey) && event.which === keyCode.RIGHT) {
                    that._moveRight();
                }
            });
      };

      /**
       * @method DoublePickerView#_initListeners
       * @private
       *
       * @param {jQuery} $el
       */
      dpvp._initListeners = function ($el) {
          var that = this;
          var element = $el || this._$el;

          this._initDOMListeners(element);
          this._source.on({
              'select invalidate': function () {// jscs:ignore jsDoc
                  that._$moveRight.prop('disabled', !that._canMove().right);
              },
              'dblclick': function () {// jscs:ignore jsDoc
                  that._moveRight();
              },
              'nodes-dropped': function () {// jscs:ignore jsDoc
                  that._moveLeft();
              },
              'node-loaded': function () {// jscs:ignore jsDoc
                  that._source.disabled(_.union(that._source.disabled(), that._dest.tree({ showRoot: false })));
              }
          });

          this._dest.on({
              'invalidate': function () {// jscs:ignore jsDoc
                  that.model.data = [];
                  resetHandler.call(that, that._dest.tree({ showRoot: false }));
                  that._updateValue();
              },
              'nodes-dropped': function () {// jscs:ignore jsDoc
                  that._moveRight();
              },
              'select': function () {// jscs:ignore jsDoc
                  that._$moveLeft.prop('disabled', !that._canMove().left);
                  that._$moveUp.prop('disabled', !that._canMove().up);
                  that._$moveDown.prop('disabled', !that._canMove().down);
              },
              'node-moved': function () {// jscs:ignore jsDoc
                  that._$moveUp.prop('disabled', !that._canMove().up);
                  that._$moveDown.prop('disabled', !that._canMove().down);
                  resetHandler.call(that, that._dest.tree({ showRoot: false }));
              },
              'dblclick': function () {// jscs:ignore jsDoc
                  that._moveLeft();
                  that._$moveLeft.prop('disabled', !that._canMove().left);
              }
          });
      };

      /**
       * @method DoublePickerView#_moveLeft
       * @private
       */
      dpvp._moveLeft = function () {
          if (this._canMove().left) {
              var selected = this._dest.selected();
              this._dest.selected([]);
              this._dest.remove(selected);

              //TODO: should I believe both trees recordKeys properties are the same?
              var sourceConverted = this._source.nodes(selected); //Convert the destination nodes to the source nodes
              var newDisabled = _.reject(this._source.disabled(), function (disabled) {
                  return _.find(sourceConverted, function (otherNode) {
                      return _.isEqual(otherNode, disabled);
                  });
              });
              this._source.disabled(newDisabled);
              this._$moveRight.prop('disabled', !this._canMove().right);
              this._$moveUp.prop('disabled', !this._canMove().up);
              this.trigger('nodes-deselected', selected);
              resetHandler.call(this, this._dest.tree({ showRoot: false }));

              this._$moveLeft.prop('disabled', !this._canMove().left);
              this._$moveUp.prop('disabled', !this._canMove().up);
              this._$moveDown.prop('disabled', !this._canMove().down);
          }
      };

      /**
       * @method DoublePickerView#_moveRight
       * @private
       */
      dpvp._moveRight = function () {
          if (this._canMove().right) {
              var selected = this._source.selected();
              var alreadySelected = this._dest.tree({ showRoot: false });
              var selection = this._dest.selected();
              var active = _.last(selection);
              var index = null;

              if (this._options.addAfterLastSelected) {
                  index = _.findIndex(alreadySelected || [], function (node) {
                      return _.isEqual(node, _.pick(active, _.keys(node)));
                  });

                  if (index !== -1) {
                      index++;
                  } else {

                      //Add at bottom if there is no index
                      index = alreadySelected.length;
                  }
              }

              this._dest.add(_.cloneDeep(selected), index);
              this._source.selected([]);
              this._source.disabled(_.union(this._source.disabled(), selected));
              this.trigger('nodes-selected', selected);
              resetHandler.call(this, this._dest.tree({ showRoot: false }));
          }
      };

      /**
       * @method DoublePickerView#_canMove
       * @private
       *
       * @returns {Object<{String}, {Boolean}>}
       */
      dpvp._canMove = function () {
          var nodes = this._dest.selected();
          var alreadySelected = this._dest.tree({ showRoot: false }) || [];
          var currentlyFocused = this._source.selected();

          return {
              right: this._source.selected().length > 0 && (this._options.max == null ||
                      (alreadySelected.length + currentlyFocused.length <= this._options.max)),
              left: nodes.length !== 0 && (!this._options.storeFns.canDeselect ||
                                           this._options.storeFns.canDeselect(nodes) &&
                                               _.intersection(nodes, this._dest.required()).length === 0),
              up: this._dest.options().groupBy.length === 0 &&
                  this._dest.selected().length > 0 &&
                      !(this._dest.tree({ showRoot: false }) && _.any(nodes, this._dest.tree({ showRoot: false })[0])),
              down: this._dest.options().groupBy.length === 0 && nodes.length > 0 && !_.any(nodes, _.last(this._dest.tree({ showRoot: false })))
          };
      };

      /**
       * @method DoublePickerView#_moveUp
       * @private
       */
      dpvp._moveUp = function () {
          this._dest.moveUp();
          resetHandler.call(this, this._dest.tree({ showRoot: false }));
      };

      /**
       * @method DoublePickerView#_moveDown
       * @private
       */
      dpvp._moveDown = function () {
          this._dest.moveDown();
          resetHandler.call(this, this._dest.tree({ showRoot: false }));
      };

      /**
       * @method DoublePickerView#_updateValue
       * @private
       */
      dpvp._updateValue = function () {
          var that = this;
          this._$el.find('input[readonly]').val(_.map(this.model.data, function (node) {
              return that._options.dest.viewFns.display(node);
          }).join(', '));
      };

      /**
       * @method DoublePickerView#buildElement
       *
       * @param {DoublePicker#Config} options
       * @param {jQuery} $el
       * @param {Object} controller
       * @param {Object} settings
       *
       * @returns {jQuery} element
       */
      dpvp.buildElement = function (options, $el, controller, settings) {
          var $element;

          if (options.modal && $el == null) {
              if (options.buttonTriggerOnly) {
                  $element = $(pickerButtonTemplate({
                      locale: options.locale
                  }));
              } else {
                  $element = $(pickerInputTemplate({
                      locale: options.locale
                  }));
              }

              if (!$element.hasClass('orcl-type-ahead')) {
                  $element.find('input').attr('readonly', true).attr('aria-haspopup', true);
              }

          } else if (options.modal) {
              if (!(options.buttonTriggerOnly || $el.is('button') || $el.is('a'))) {
                  $element = $el.wrap(pickerWrapInputTemplate).parent()
                    .addClass('orcl-double-picker')
                    .append(pickerButtonTemplate({
                        locale: options.locale
                    }));
              } else {
                  $element = $el.addClass('orcl-double-picker');
              }

              if (!$element.hasClass('orcl-type-ahead')) {
                  $element.find('input').attr('readonly', true).attr('aria-haspopup', true);
              }

          } else if ($el == null) {
              $element = $(pickerTemplate({
                  id: settings.id,
                  locale: options.locale,
                  sortable: options.allowReordering,
                  vertical: options.vertical
              }));
          }

          $element.toggleClass('fill', options.fill);

          return $element;
      };

      /**
       * @method DoublePickerView#init
       */
      dpvp.init = function () {
          if (this._$el.hasClass('orcl-picker-input')) {
              var input = this._$el.children('input');
              this._$el.find('.instructions[data-name=' + this._settings.name + ']').attr('id', this.id() + '-instructions-' + this._settings.name);
              input.attr('aria-describedby', ((input.attr('aria-describedby') || '') + ' ' + this.id() + '-instructions-' + this._settings.name).trim());
          }

          if (!this._$modal) {
              this._source.init();
              this._dest.init();

          }
      };

      /**
       * Returns the element that identifies the widget.
       *
       * @method DoublePickerView#getIdentifyingElement
       *
       * @returns {jQuery}
       */
      dpvp.getIdentifyingElement = function () {
          var picker = this._$el;
          var pickerInput = this._$el.hasClass('orcl-picker-input');
          var element = picker;

          if (pickerInput) {
              element = picker.find('input');
          }

          return element;
      };

      /**
       * @method DoublePickerView#show
       */
      dpvp.show = function () {
          if (this._$modal) {
              this._$modal.modal('show');
          }
      };

      /**
       * @method DoublePickerView#hide
       */
      dpvp.hide = function () {
          if (this._$modal) {
              this._$modal.modal('hide');
          }
      };

      /**
       * @method DoublePickerView#invalidate
       *
       * @param {Object} modifiers
       */
      dpvp.invalidate = function (modifiers) {
          if (modifiers == null) {
              modifiers = {
                  source: true,
                  dest: true
              };
          }

          if (modifiers.source) {
              this._source.invalidate();
          }

          if (modifiers.dest) {
              this._dest.invalidate();
              this._source.disabled([]);
          }
      };

      /**
       * @method DoublePickerView#available
       *
       * @returns {Object}
       */
      /**
       * @method DoublePickerView#available
       *
       * @param {Array} nodes
       */
      dpvp.available = function (nodes) {
          if (nodes == null) {
              return this._source.treePanel().tree({
                  showRoot: false
              });
          }

          this._source.invalidate();
          this._source.treePanel().add(nodes);
      };

      /**
       * @method DoublePickerView#appendRightTreeTo
       *
       * @param {jQuery} $target
       */
      dpvp.appendRightTreeTo = function ($target) {
          this._source.element().closest('.source-picker').addClass('separated');
          var $dest = this._dest.element().closest('.dest-picker').addClass('separated');
          var $up = this._$moveUp;

          var $separatedParent = $([]);
          if ($up.closest('.separated').length) {
              $separatedParent = $up.closest('.orcl-double-picker');
          }

          var $newTarget = $(pickerContainerTemplate({
              id: this._settings.id,
              locale: this._options.locale,
              sortable: this._options.allowReordering
          }));
          $newTarget.appendTo($target);

          $([]).add($dest).add($up.closest('.picker-controls-right'))
              .appendTo($newTarget)
              .addClass('separated');
          $separatedParent.remove();
          this._initDOMListeners($target);
      };

      /**
       * @method DoublePickerView#updateButtons
       */
      dpvp.updateButtons = function () {
          var canMove = this._canMove();

          this._$moveUp.prop('disabled', !canMove.up);
          this._$moveDown.prop('disabled', !canMove.down);
      };

      /**
       * @method DoublePickerView#_applyDisabled
       * @private
       *
       * @param {Boolean} disabled
       */
      dpvp._applyDisabled = function (disabled) {
          if (this._options.modal) {
              this._$el.find('.btn').prop('disabled', disabled);
              this._$el.find('input').prop('disabled', disabled);
              if (!this._$el.hasClass('orcl-type-ahead')) {
                  this._$el.find('input').attr('readonly', true).attr('aria-haspopup', true);
              }
          } else {
              this._dest.disable(disabled);
              this._source.disable(disabled);
          }
      };

      /**
       * @method DoublePickerView#isValid
       *
       * @returns {Boolean}
       */
      dpvp.isValid = function () {
          return !(this._options.required && this.model.data.length === 0);
      };

      return DoublePickerView;
  });

define('oui/widgets/doublePicker',['require','underscore','oui/base/widget','oui/widgets/picker','oui/widgets/treePanel','oui/widgets/doublePicker/doublePickerView','oui/widgets/typeAhead'],function (require) {
      'use strict';

      var _ = require('underscore');
      var Widget = require('oui/base/widget');
      var Picker = require('oui/widgets/picker');
      var TreePanel = require('oui/widgets/treePanel');
      var DoublePickerView = require('oui/widgets/doublePicker/doublePickerView');
      var TypeAhead = require('oui/widgets/typeAhead');

      /**
       * @typedef {Object} DoublePicker#Config
       *
       * @property {Locale} [locale={}] The specific strings that are used by the widget and will use the defaults set from
       * the {@link module:oui.lang|lang}.
       * @property {Object} shared The shared tree properties between the destination and source trees.
       * @property {Boolean} [modal=false] Whether or not the double picker should be displayed within a modal.
       * @property {String} [modalClasses] Custom classes to add to the modal for additional styling.
       * @property {Boolean} [vertical=false] Whether the picker should display in a vertical format.
       * @property {Picker#Config} [source] The picker options for the source picker.
       * @property {TreePanel#Config} [dest] The tree options for the destination.
       * @property {Number|null} [max=null] The maximum allowed selected records within the double picker.
       * @property {Boolean} [allowReordering=false] Whether the destination is allowed to be resorted using dragging or
       * up/down arrows.
       * @property {Boolean} [fluid=false] Whether the component should take up available width of the parent container.
       * @property {String|jQuery} [appendTo='body'] When using the double picker as a modal, the location to append the modal
       * to.
       * @property {Boolean} [required=false] Whether the double picker is required to have something selected.
       * @property {Boolean} [openFromInput=false] Whether the picker should be opened on a click from the input (when it is
       * readonly).
       * @property {Boolean} [addAfterLastSelected=true] Whether or not nodes moved right should be added after the last item
       * of the currently selected. If none are selected it adds to the end of the list.
       * @property {Object} [labels] The source and destination labels.
       * @property {String} [labels.source=''] The source label.
       * @property {String} [labels.dest=''] The destination label.
       */
      var __DEFAULTS__ = {
          shared: {},
          modal: false,
          modalClasses: null,
          vertical: false,
          repeatable: false,
          source: _.merge({}, Picker.prototype.defaults(), {
              tree: {
                  disabledStyle: 'hidden',
                  hideEmpty: true,
                  clientFilter: true,
                  filterShowHierarchy: true,
                  multiSelect: true,
                  drag: {
                      enabled: true
                  }
              }
          }),
          dest: _.merge({}, TreePanel.prototype.defaults(), {
              picker: true,
              multiSelect: true,
              drag: {
                  enabled: true
              }
          }),
          max: null,
          allowReordering: false,
          fluid: false,
          fill: false,
          appendTo: 'body',
          required: false,
          openFromInput: false,
          addAfterLastSelected: true,
          storeFns: {
              canDeselect: _.constant(true)
          },
          labels: {
              source: '',
              dest: ''
          }
      };

      /**
       * The Double Picker component. Displays a shuttle for the user to select items by moving them from the left side to the
       * right side.
       *
       * See {@link DoublePicker#Config} for more information regarding options for the Double Picker.
       *
       *
       *  ##### **Relevant Events**
       *
       *  * {@link DoublePicker#accept|accept}
       *  * {@link DoublePicker#cancel|cancel}
       *  * {@link DoublePicker#nodes-selected|nodes-selected}
       *  * {@link DoublePicker#nodes-deselected|nodes-deselected}
       *
       * @borrows DoublePickerView#event:accept as DoublePicker#accept
       * @borrows DoublePickerView#event:cancel as DoublePicker#cancel
       * @borrows DoublePickerView#event:nodes-selected as DoublePicker#nodes-selected
       * @borrows DoublePickerView#event:nodes-deselected as DoublePicker#nodes-deselected
       *
       * @class DoublePicker
       * @extends Widget
       * @since 2.28.0
       *
       * @param {DoublePicker#Config} [options] The options for the Double Picker. If not specified, all the defaults will be
       * used.
       * @param {jQuery} [element] The jQuery element that this field should use.
       * @param {Object} [initData] The initial data to use within the double picker.
       */
      function DoublePicker(options, element, initData) {
          Widget.call(this, options, element, initData);

          this._options.source.modal = false;

          this._options.dest.lazyLoad = false;

          if (this._options.shared) {
              _.merge(this._options.source.tree, this._options.shared);
              _.merge(this._options.dest, this._options.shared);
          }

          if (this._options.dest.groupBy != null &&
              this._options.dest.groupBy[0].asTree) {
              this._options.dest.groupBy = [];
          }

          if (this._options.dest.groupBy.length === 0) {
              this._options.dest.flat = true;
          }

          this._view = new DoublePickerView(this._options, this._element, this._settings, this._initData);

          this._setupEvents(this._view);

          var that = this;
          var methods = ['disabled', 'expanded', 'selected', 'tree', 'on', 'off', 'invalidate', 'add'];
          /**
           * A method that enables access to some methods on the source `TreePanel` managed by this double picker.
           * Unlike most other methods of the double picker, this method provides "raw" access to the tree, and
           * will not attempt to coordinate between the two trees, as such methods as {DoublePicker#selected} would.
           *
           * These methods are:
           *
           * * disabled
           * * expanded
           * * selected
           * * tree
           * * on
           * * off
           * * invalidate
           * * add
           *
           * And more methods may be added if the need for them is identified.
           *
           * @example
           *
           *     // On a tree with no selections:
           *     assert.deepEqual(doublePicker.available(), dp.source().tree(true));
           *
           * @method DoublePicker#source
           * @returns {Object} 'source' some functions of the source tree
           */
          var src = Object.freeze(_.reduce(methods, function (memo, method) {// jscs:ignore jsDoc
              memo[method] = that._view._source.treePanel()[method].bind(that._view._source.treePanel());
              return memo;
          }, {}));
          /**
           * A method that enables access to some methods on the destination `TreePanel` managed by this double picker.
           * Unlike most other methods of the double picker, this method provides "raw" access to the tree, and
           * will not attempt to coordinate between the two trees, as such methods as {DoublePicker#selected} would.
           *
           * These methods are:
           *
           * * disabled
           * * expanded
           * * selected
           * * tree
           * * on
           * * off
           * * invalidate
           * * add
           *
           * And more methods may be added if the need for them is identified.
           *
           * @example
           *
           *     doublePicker.dest().on('node-moved', function () {
           *         assert.deepEqual(doublePicker.dest().tree(true), doublePicker.selected());
           *     });
           *
           * @method DoublePicker#dest
           * @returns {Object} 'destination' some functions of the destination tree
           */
          var dest = Object.freeze(_.reduce(methods, function (memo, method) {// jscs:ignore jsDoc
              memo[method] = that._view._dest[method].bind(that._view._dest);
              return memo;
          }, {}));
          this.source = _.constant(src);
          this.dest = _.constant(dest);

          if (this._options.createTypeAhead && this._options.modal && !this._options.buttonTriggerOnly) {
              this._outerTypeAhead = new TypeAhead({
                  multiSelect: true,
                  appendTo: this._options.createTypeAhead.appendTo || this._options.appendTo,
                  fluid: this._options.fluid,
                  tokenize: false,
                  required: this._options.required,

                  //For now
                  storeFns: {
                      search: this._options.source.tree.storeFns.filter
                  },
                  viewFns: {
                      display: _.get(this._options, 'createTypeAhead.viewFns.display', this._options.source.tree.viewFns.display)
                  }
              }, this._view.element().find('input'), initData);
          }

      }

      var dpp = DoublePicker.prototype = Object.create(Widget.prototype);

      dpp.constructor = DoublePicker;

      /**
       * Method for returning the name of the widget.
       *
       * @method DoublePicker#name
       * @returns {String} 'double-picker'
       */
      dpp.name = function () {
          return 'double-picker';
      };

      /**
       * Initialize the picker by initializing its subcomponents.
       *
       * @method Picker#init
       * @private
       */
      dpp._init = function () {

          if (this._outerTypeAhead) {
              this._outerTypeAhead.init();

              if (this.selected().length > 0) {
                  this._outerTypeAhead.value(this.selected());
              }
          }
      };

      /**
       * Getter for the defaults of the component.
       *
       * @method DoublePicker#defaults
       * @returns {DoublePicker#Config} defaults
       */
      dpp.defaults = function () {
          return __DEFAULTS__;
      };

      /**
       * @method DoublePicker#updateOption
       *
       * @param {String} option
       * @param {*} value
       */
      dpp.updateOption = function (option, value) {
          Widget.prototype.updateOption.call(this, option, value);

          if (option === 'dest') {
              this._view._dest.updateOptions(value);
              this._view.updateButtons();
          }

          if (option === 'source') {
              this._view._source.updateOptions(value);
              this._view.updateButtons();
          }
      };

      /**
       * Getter for the selected nodes of the picker. Essentially, what is contained within the right side of the
       * component.
       *
       * @method DoublePicker#selected
       * @returns {Array<Object>} The currently selected nodes of the double picker.
       */
      /**
       * Setter for the selected nodes of the picker. Essentially, what is contained within the right side of the
       * component.
       *
       * @method DoublePicker#selected
       *
       * @param {Array} selected
       */
      dpp.selected = function (selected) {
          if (!arguments.length) {
              return this._view._dest.nodes();
          }
          this._view._dest.invalidate();
          this._view.model.data = selected;
          this._view._source.disabled(selected);
          this._view._dest.add(selected);
          this._view.model.data = selected;
      };

      /**
       * Getter/Setter for the value of the of the Double Picker
       *
       * See {@link DoublePicker#selected} for information regarding the similar method.
       *
       * **Note**: does not trigger an event notifying of a change
       *
       * @method DoublePicker#value
       * @param {Array<Object>} [selected] The selected values of the Double Picker.
       * @returns {Array<Object>} - the current value of the Double Picker.
       */
      dpp.value = function (selected) {
          if (!arguments.length) {
              return this.selected();
          }
          this._view._dest.invalidate();
          this._view._source.disabled(selected, {
              silent: true
          });
          this._view._dest.add(selected, {
              silent: true
          });
          this._view.model.data = selected;
      };

      /**
       * Checks the parameter for whether it is init data.
       *
       * @method DoublePicker#isInitData
       * @private
       * @param {Object|*} data The data to check existence of initialization data properties
       * @returns {Boolean} whether the data is initData
       */
      dpp.isInitData = function (data) {
          return data.available || data.selected;
      };

      /**
       * Shows the modal if the double picker is currently hidden and is a modal.
       *
       * @method DoublePicker#show
       */
      dpp.show = function () {
          this._view.show();
      };

      /**
       * Hides the modal if the double picker is currently showing and is a modal.
       *
       * @method DoublePicker#hide
       */
      dpp.hide = function () {
          this._view.hide();
      };

      /**
       * Clears the state of one or more of the pickers specified within the modifiers.
       *
       * By default it will clear both sides.
       *
       * @method DoublePicker#invalidate
       * @param {Object} modifiers Which pickers to invalidate
       */
      dpp.invalidate = function (modifiers) {
          this._view.invalidate(modifiers);
      };

      /**
       * Clears the source picker and adds the new nodes to the source picker.
       *
       * @method DoublePicker#available
       * @param {Array} nodes The new nodes that are available for selection.
       */
      dpp.available = function (nodes) {
          return this._view.available(nodes);
      };

      /**
       * Moves the double picker's right tree to another location.
       *
       * @method DoublePicker#appendRightTreeTo
       * @param {jQuery} $target The location to append the right tree to.
       */
      dpp.appendRightTreeTo = function ($target) {
          this._view.appendRightTreeTo($target);
      };

      /**
       * @method DoublePicker#isValid
       *
       * @returns {Boolean}
       */
      dpp.isValid = function () {
          return this._view.isValid();
      };

      return DoublePicker;
  });

define('oui/widgets/modal/modalView',['require','underscore','jquery','oui/base/view','template-15','template-16','template-17'],function (require) {
    'use strict';

    var _  = require('underscore');
    var $ = require('jquery');
    var View = require('oui/base/view');
    var pickerButtonTemplate = require('template-15');
    var pickerInputTemplate = require('template-16');
    var pickerWrapInputTemplate = require('template-17');

    var mvp;

    /**
     * @class ModalView
     * @extends View
     * @private
     *
     * @param {Modal#Config} options
     * @param {jQuery} element
     * @param {Object} settings
     */
    function ModalView(options, element, settings) {
        View.call(this, options, element, null, settings);

        this.modal(options.modal);

        this._$el.find('button')
            .on('mousedown', function (event) {
                event.preventDefault();
            })
            .on('click', this.showHandler.bind(this));

        this._$el.on('click', function (event) {
            if (this._$el.is('button') || this._$el.is('a')) {
                this.showHandler(event);
            }
        }.bind(this));

        this._$el.on('mousedown', 'button', function (event) {
            event.preventDefault();
        })
        .on('click', 'button', this.showHandler.bind(this));

        this.addShortcuts({
            'F2 SHIFT-F10': this.showHandler.bind(this)
        });

        this._$el.on('click', 'input[readonly]', function (event) {
            if (this._options.openFromInput && !this._$modal.hasClass('in')) {
                this.showHandler(event);
            }
        }.bind(this));
    }

    mvp = ModalView.prototype = Object.create(View.prototype);

    /**
     * @method ModalView#buildElement
     *
     * @param {Modal#Config} options
     * @param {jQuery} el
     *
     * @returns {jQuery}
     */
    mvp.buildElement = function (options, el) {
        var $element;

        if (_.isNull(el) && options.buttonTriggerOnly) {
            $element = $(pickerButtonTemplate({
                locale: options.locale
            }));
        } else if (_.isNull(el)) {
            $element = $(pickerInputTemplate({
                locale: options.locale
            })).addClass('orcl-modal-input');
        } else if (el.jquery && el.is('input')) {
            //Input given
            if (el.parent().hasClass('orcl-type-ahead')) {
                el.parent().addClass('orcl-modal-input').addClass('orcl-modal');
            } else {
                el.wrap(pickerWrapInputTemplate);
            }
            $element = el.parent();
            $element.append($(pickerButtonTemplate({
                locale: options.locale
            })).attr('tabindex', '-1'));
            $element.addClass('orcl-modal-input orcl-modal');
        } else if (el.jquery && (el.is('button') || el.is('a'))) {
            //Button given
            $element = el;
        }

        if (!$element.hasClass('orcl-type-ahead')) {
            $element.find('input').attr('readonly', true).attr('aria-haspopup', true);
        }

        $element.addClass('orcl-modal');

        return $element;
    };

    /**
     * @method ModalView#modal
     *
     * @returns {jQuery}
     */
    /**
     * @method ModalView#modal
     *
     * @param {jQuery} $modal
     */
    mvp.modal = function ($modal) {
        var that = this;
        if (!arguments.length) {
            return this._$modal;
        }

        this._$modal = $modal;

        this._$modal.on('hidden', function () {
            that._$modal.detach();
        });

        this._$modal.on('show', function () {
            if (that._$modal.parent().length === 0) {
                $(that._options.appendTo).append(that._$modal);
            }
        });

        this._$modal.on('click', '[data-dismiss]', function (event) {
            event.stopPropagation();
        });

    };

    /**
     * @method ModalView#showHandler
     *
     * @param {jQuery.Event} event
     */
    mvp.showHandler = function (event) {
        event.preventDefault();
        event.stopPropagation();

        this.show();
    };

    /**
     * @method ModalView#show
     */
    mvp.show = function () {
        this._$modal.modal('show');
    };

    /**
     * @method ModalView#hide
     */
    mvp.hide = function () {
        this._$modal.modal('hide');
    };

    return ModalView;
});

define('oui/widgets/modal/modal',[
    'underscore',
    'oui/base/widget',
    'oui/widgets/modal/modalView'
], function (_, Widget, ModalView) {
    'use strict';

    /**
     * @typedef Modal#Config
     *
     * @property {String|jQuery} [appendTo='body'] Where to append the modal if it is not already appended.
     * @property {Boolean} [buttonTriggerOnly=false] Whether or not the element is only supposed to be a button instead of an
     * input+button.
     * @property {Boolean} [openFromInput=false] Whether or not when the input is readonly if a click on the readonly input
     * should trigger the modal to show.
     * @property {Boolean} [fluid=false] Whether or not the input should take up the width of the parent.
     * @property {jQuery} modal=null The modal to use to launch from the button
     *
     */

    var mp;
    var __WIDGET_NAME__ = 'modal';
    var __DEFAULTS__ = {
        appendTo: 'body',
        buttonTriggerOnly: false,
        openFromInput: false,
        fluid: false,
        modal: null
    };

    /**
     * @class Modal
     * @extends Widget
     *
     * @param {Modal#Config} [options] The options for the Modal.
     * @param {jQuery} [element] The element to use as the base of the Modal.
     */
    function Modal(options, element) {
        Widget.call(this, options, element);

        this._view = new ModalView(this._options, this._element, this._settings);

        this._setupEvents([this._view]);
    }

    mp = Modal.prototype = Object.create(Widget.prototype);

    /**
     * Gets the widget name. Used to prefix some events.
     *
     * @method Modal#name
     * @returns {String} 'modal'
     */
    mp.name = function () {
        return __WIDGET_NAME__;
    };

    /**
     * Get the defaults for the widget options.
     *
     * @method Modal#defaults
     * @returns {Modal#Config}
     */
    mp.defaults = function () {
        return __DEFAULTS__;
    };

    /**
     * Shows the modal if it is hidden.
     *
     * @method Modal#show
     */
    mp.show = function () {
        this._view.show();
    };

    /**
     * Hides the modal if it is visible.
     *
     * @method Modal#hide
     */
    mp.hide = function () {
        this._view.hide();
    };

    /**
     * Getter/Setter for the modal within the widget.
     *
     * @method Modal#modal
     * @param {jQuery} [$modal] The modal to set
     * @returns {jQuery} - The current modal used by the widget
     */
    mp.modal = function ($modal) {
        if (!arguments.length) {
            return this._view.modal();
        }

        this._view.modal($modal);
    };

    return Modal;
});



define('template-91', ['underscore'], function (_) { return function(data) {
var __t, __p = '';
__p += '<div class="orcl-widget orcl-file-picker"> <input type="text" disabled aria-hidden="true"/> <label class="file-label"> <span class="btn btn-picker" aria-hidden="true" title="' +
((__t = ( data.locale.browse )) == null ? '' : __t) +
'"> <i class="pgbu-icon-small pgbu-icon-picker-dots"></i> </span> </label></div>';
return __p
};});


define('template-92',[],function () { return '<div class="orcl-widget orcl-file-picker">\n</div>\n';});

define('oui/widgets/file/fileView',['require','jquery','underscore','oui/base/view','template-91','template-9','template-92'],function (require) {
      'use strict';

      var $ = require('jquery');
      var _ = require('underscore');
      var View = require('oui/base/view');
      var pickerTemplate = require('template-91');
      var nativeTemplate = require('template-9');
      var wrapTemplate = require('template-92');

      var fvp;

      /**
       * @class FileView
       * @extends View
       * @private
       *
       * @param {FilePicker#Config} options
       * @param {jQuery} element
       * @param {Object} settings
       */
      function FileView(options, element, settings) {
          var that = this;
          View.call(this, options, element, null, settings);

          this._$el.find('label').attr('id', this.id() + '-label');

          this.addShortcuts({
              'ENTER': {
                  //TODO: look into refactoring shortcuts so the key goes to the callback
                  prevention: _.constant(false),
                  stopPropagation: _.constant(true),
                  fn: _.noop
              }
          });

          this._$el.find('[type=file]')
              .on('change', function (event) {
                  var fileList = that.getFileList(event.target);

                  if (that._$el.find('[type=text]').length > 0) {
                      that._$el.find('[type=text]')[0].value = _.map(fileList, _.property('name')).join(', ');
                  }

                  //Once the dialog is closed and the element changes make sure it retains focus
                  $(event.target).focus();
                  that.trigger('change', fileList);
              }).on('focus blur', function (event) {
                  that._$el.toggleClass('focus', event.type === 'focus');
              });
      }

      FileView.prototype = Object.create(View.prototype);
      fvp = FileView.prototype;

      /**
       * @method FileView#getFileList
       *
       * @param {HTMLInputElement} elem
       *
       * @returns {FileList}
       */
      fvp.getFileList = function (elem) {
          return elem.files;
      };

      /**
       * @method FileView#nativeType
       * @returns {String} 'file'
       */
      fvp.nativeType = function () {
          return 'file';
      };

      /**
       * @method FileView#buildNativeElement
       * @param {FilePicker#Config} options
       * @param {jQuery} el
       */
      fvp.buildNativeElement = function (options, el) {
          var $element;

          if (_.isNull(el)) {
              $element = $(nativeTemplate({
                  type: this.nativeType()
              })).wrap(wrapTemplate).parent();
          } else if (el.jquery) {
              $element = el.wrap(wrapTemplate).parent();
          }

          return $element;
      };

      /**
       * @method FileView#buildElement
       * @param {FilePicker#Config} options
       * @param {jQuery} el
       */
      fvp.buildElement = function (options, el) {
          var $element = $(pickerTemplate({
              locale: options.locale
          }));

          if (_.isNull(el)) {
              $element.find('label').prepend(nativeTemplate({
                  type: this.nativeType()
              }));
          } else if (el.jquery) {
              if (el.parent().length > 0) {
                  $element.insertAfter(el);
              }
              $element.find('label').prepend(el);
          }

          $element.find('[type=file]')
              .prop('multiple', options.multiple);

          return $element;
      };

      /**
       * Returns the element that identifies the widget.
       *
       * @method FileView#getIdentifyingElement
       *
       * @returns {jQuery}
       */
      fvp.getIdentifyingElement = function () {
          return this._$el.find('input[type=file]').first();
      };

      /**
       * @method FileView#value
       *
       * @returns {File}
       */
      /**
       * @method FileView#value
       * @param {File} file
       * @param {Subscribable~EventOptions} eventParams
       */
      fvp.value = function (file, eventParams) {
          if (!arguments.length) {
              return this._$el.find('[type=file]')[0].files;
          }

          if (file == null) {
              this._$el.find('[type=file]').val('');

              if (!eventParams || !eventParams.silent) {
                  this._$el.find('[type=file]').trigger('change');
              }
          }
      };

      /**
       * @method FileView#_applyReadonly
       * @private
       * @param {Boolean} readonly
       */
      fvp._applyReadonly = function (readonly) {
          View.prototype._applyReadonly.call(this, readonly);

          this._$el.find('[type=file]').prop('disabled', readonly);
      };

      return FileView;
  });

define('oui/widgets/file/filePicker',[
    'oui/base/input',
    'oui/widgets/file/fileView'
], function (Input, FileView) {
    'use strict';

    /**
     * @typedef {Object} FilePicker#Config
     * @property {Boolean} [fluid=false] Whether or not the file picker should take available widget.
     * @property {Boolean} [multiple=true] Whether or not the file picker should allow multiple files to be selected.
     * @property {Locale} [locale={}] The specific strings that are used by the widget and will use the defaults set from
     * the {@link module:oui.lang|lang}.
     */

    var fpp;
    var __WIDGET_NAME__ = 'file';

    var __DEFAULTS__ = {
        fluid: false,
        multiple: true
    };

    /**
     * The File Picker widget allows the user to select a file.
     *
     * Uses a native file input with additional styling over it to provide the look and feel.
     * The native file input must be used for security reasons. If at any point the browser believes the
     * user is being tricked into picking a file it will prevent that action from taking place.
     *
     * **Note**: File Picker does not allow setting of file programmatically due to security concerns. The browsers limit the
     * interaction of file inputs because a user may be tricked into selecting something faulty.
     *
     * @class FilePicker
     * @extends Input
     * @param {FilePicker#Config} [options]
     * @param {jQuery} [element]
     */
    function FilePicker(options, element) {
        var view;

        Input.call(this, options, element);

        view = new FileView(this._options, this._element, this._settings);

        this._view = view;

        this._setupEvents(view);
    }

    FilePicker.prototype = Object.create(Input.prototype);
    fpp = FilePicker.prototype;

    /**
     * Gets the name of the widget.
     *
     * @method FilePicker#name
     * @returns {String} 'file'
     */
    fpp.name = function () {
        return __WIDGET_NAME__;
    };

    /**
     * Gets the defaults for the widget.
     *
     * @method FilePicker#defaults
     * @returns {Object} The defaults for the widget.
     */
    fpp.defaults = function () {
        return __DEFAULTS__;
    };

    /**
     * @method FilePicker#value
     *
     * @returns {File}
     */
    /**
     * @method FilePicker#value
     * @param {File} file
     */
    fpp.value = function (file) {
        if (!arguments.length) {
            return this._view.value();
        }

        this._view.value(file, { silent: true });
    };

    return FilePicker;
});


define('template-93', ['underscore'], function (_) { return function(data) {
var __t, __p = '';
__p += '<div class="orcl-widget orcl-checkbox"></div>';
return __p
};});


define('template-94', ['underscore'], function (_) { return function(data) {
var __t, __p = '';
__p += '<div class="orcl-widget orcl-checkbox"> <input type="checkbox"> <i class="pgbu-icon-small pgbu-icon-selected"></i> <i class="pgbu-icon-small pgbu-icon-unselected"></i></div>';
return __p
};});


define('template-95', ['underscore'], function (_) { return function(data) {
var __t, __p = '';
__p += '<i class="pgbu-icon-small pgbu-icon-selected"></i><i class="pgbu-icon-small pgbu-icon-unselected"></i>';
return __p
};});


define('template-96', ['underscore'], function (_) { return function(data) {
var __t, __p = '';
__p += '<span class="only pull-right"> <a href="#" tabindex="-1" role="presentation" aria-hidden="true">Only</a></span>';
return __p
};});

define('oui/widgets/checkbox/checkboxView',['require','underscore','jquery','oui/base/view','template-93','template-94','template-95','template-96'],function (require) {
    'use strict';

    var _ = require('underscore');
    var $ = require('jquery');
    var View = require('oui/base/view');
    var checkboxWrapTemplate = require('template-93');
    var checkboxTemplate = require('template-94');
    var icons = require('template-95');
    var onlyLink = require('template-96');

    /**
     * @class CheckboxView
     * @extends View
     * @private
     *
     * @param {CheckboxModel} model
     * @param {CheckboxController} controller
     * @param {jQuery} element
     * @param {Checkbox#Config} options
     * @param {Object} settings
     */
    function CheckboxView(model, controller, element, options, settings) {
        View.call(this, options, element, controller, settings);

        this._controller = controller;
        this._$input = this._$el.find('input');
        this._value = this._$input.attr('value');

        this._$el.on('change', function () {
            var newValue = this._$input.prop('checked');
            this.value(this._controller.data());
            this._controller.data(newValue);
        }.bind(this));

        this._$el.on('click', '.only', function (event) {
            event.preventDefault();

            this.trigger('only');
            this._controller.data(true);
        }.bind(this));

        this._$el.on('mousedown', '.only', function (event) {
            event.preventDefault();
        });

        this._$el.on('mousedown', 'label', function (event) {
            event.preventDefault();
            this._$el.find('input').focus();
        }.bind(this));

        this.addShortcuts({
            'CTRL-SPACE': function (event) {
                event.preventDefault();
                this.trigger('only');
                this._controller.data(true);
            }.bind(this)
        });

        model.on({
            'view.change': this.value.bind(this)
        });

        this.value(this._controller.data());
    }

    var cvp = CheckboxView.prototype = Object.create(View.prototype);

    cvp.constructor = CheckboxView;

    /**
     * @method CheckboxView#buildElement
     *
     * @param {Checkbox#Config} options
     * @param {jQuery} $element
     */
    cvp.buildElement = function (options, $element) {

        if ($element != null) {
            $element = $element.wrap(checkboxWrapTemplate({})).parent();

            $element.find('label').append(icons()).append(onlyLink());
        } else {
            $element = $(checkboxTemplate({})).append(onlyLink());
        }

        if (options.useOnly) {
            $element.addClass('use-only');
        }

        return $element;
    };

    /**
     * @method CheckboxView#getFieldValue
     *
     * @returns {String}
     */
    cvp.getFieldValue = function () {
        return this._value;
    };

    /**
     * @method CheckboxView#value
     *
     * @returns {Boolean}
     */
    /**
     * @method CheckboxView#value
     *
     * @param {Boolean|Object} value
     */
    cvp.value = function (value) {
        if (!arguments.length) {
            return this._$input.prop('checked');
        } else {
            if (_.isBoolean(value)) {
                this._$input.prop('checked', value);
            } else if (_.isObject(value)) {
                this._$input.prop('checked', value[this._value]);
            } else {
                this._$input.prop('checked', value != null);
            }
        }
    };

    /**
     * @method CheckboxView#readonly
     *
     * @param {Boolean} readonly
     */
    cvp.readonly = function (readonly) {
        View.prototype.readonly.apply(this, arguments);
        this._$input.prop('disabled', readonly);
    };

    /**
     * @method CheckboxView#disable
     *
     * @param {Boolean} disabled
     */
    cvp.disable = function (disabled) {
        View.prototype.disable.apply(this, arguments);

        this._$el.toggleClass('disabled', disabled);
        this._$input.prop('disabled', disabled);
    };

    return CheckboxView;
});

define('oui/widgets/checkbox/checkboxController',[
], function () {
    'use strict';

    /**
     * @class CheckboxController
     * @private
     *
     * @param {CheckboxModel} model
     * @param {Checkbox#Config} options
     */
    function CheckboxController(model, options) {
        this._model = model;
        this._options = options;
    }

    CheckboxController.prototype = {
        /**
         * @method CheckboxController#data
         *
         * @returns {Boolean}
         */
        /**
         * @method CheckboxController#data
         *
         * @param {Boolean} value
         * @param {Subscribable~EventOptions} eventParams
         */
        data: function (value, eventParams) {
            if (!arguments.length) {
                return this._model.data();
            }
            this._model.data(value, eventParams);
        }
    };

    return CheckboxController;
});

define('oui/widgets/checkbox/checkboxModel',[
    'underscore',
    'oui/base/model'
], function (_, Model) {
    'use strict';

    /**
     * @class CheckboxModel
     * @private
     * @extends Model
     *
     * @param {Boolean} initData
     * @param {Checkbox#Config} options
     */
    function CheckboxModel(initData, options) {
        Model.call(this, initData === true, options);
    }

    var cmp = CheckboxModel.prototype = Object.create(Model.prototype);

    /**
     * @method CheckboxModel#name
     * @returns {String} 'checkbox'
     */
    cmp.name = _.constant('checkbox');

    return CheckboxModel;
});

define('oui/widgets/checkbox',[
    'underscore',
    'oui/base/widget',
    'oui/widgets/checkbox/checkboxView',
    'oui/widgets/checkbox/checkboxController',
    'oui/widgets/checkbox/checkboxModel'
], function (_, Widget, CheckboxView, CheckboxController, CheckboxModel) {
    'use strict';

    var cp;

    /**
     * @typedef Checkbox#Config
     *
     * @property {Locale} [locale={}] The specific strings that are used by the widget and will use the defaults set from
     * the {@link module:oui.lang|lang}.
     */
    var __DEFAULTS__ = {
        required: false,
        useOnly: false
    };

    /**
     * @class Checkbox
     * @extends Widget
     *
     * @param {Checkbox#Config} [options] The options for the checkbox
     * @param {jQuery} [element] The element to use as the checkbox
     * @param {Object} [data] The data to set on the checkbox. Keyed data based on the name attribute of the checkboxes and true/false.
     */
    function Checkbox(options, element, data) {
        var model;
        Widget.call(this, options, element, data);

        model = new CheckboxModel(this._initData, this._options);

        this._controller = new CheckboxController(model, this._options);

        this._view = new CheckboxView(model, this._controller, this._element, this._options, this._settings);

        this._relatedCheckboxes = [];
        this._setupEvents([model, this._view]);
    }

    cp = Checkbox.prototype = Object.create(Widget.prototype);

    cp.constructor = Checkbox;

    /**
     * Returns the name of the control
     *
     * @method Checkbox#name
     * @returns {String} 'checkbox' The name of the control
     */
    cp.name = function () {
        return 'checkbox';
    };

    /**
     * Returns the defaults of the checkbox controls
     *
     * @method Checkbox#defaults
     * @returns {Checkbox#Config} defaults The defaults for the checkbox
     */
    cp.defaults = function () {
        return __DEFAULTS__;
    };

    /**
     * @method Checkbox#getFieldValue
     *
     * @returns {String}
     */
    cp.getFieldValue = function () {
        return this._view.getFieldValue();
    };

    /**
     * Gets or sets the currently selected checkboxes in the set of checkboxes.
     *
     * @method Checkbox#value
     * @param {Object<String, Boolean>} [val] The values to set on the set of checkboxes
     * @returns {Object<String, Boolean>} values The current values set on the checkboxes
     */
    cp.value = function (val) {
        if (!arguments.length) {

            if (this._relatedCheckboxes.length > 0) {
                var initMemo = {};
                initMemo[this.getFieldValue()] = this._controller.data();
                return _.reduce(this._relatedCheckboxes, function (memo, checkbox) {
                    memo[checkbox.getFieldValue()] = checkbox._view.value();

                    return memo;
                }, initMemo);
            } else {
                return this._controller.data();
            }
        }

        if (_.isBoolean(val)) {
            this._controller.data(val, { silent: true });
        } else {
            this._controller.data(_.get(val, this._view.getFieldValue(), false), { silent: true });

            _.each(this._relatedCheckboxes, function (checkbox) {
                checkbox.value(_.get(val, checkbox.getFieldValue(), false));
            });
        }
    };

    /**
     * @method Checkbox#data
     *
     * @returns {Boolean|Object}
     */
    /**
     * @method Checkbox#data
     *
     * @param {Boolean|Object} val
     */
    cp.data = function (val) {
        if (!arguments.length) {
            return this.value();
        }

        if (_.isBoolean(val)) {
            this._controller.data(val);
        } else {
            this._controller.data(_.get(val, this._view.getFieldValue(), false));

            _.each(this._relatedCheckboxes, function (checkbox) {
                checkbox.data(_.get(val, checkbox.getFieldValue(), false));
            });
        }

    };

    /**
     * Adds a checkbox to the collection of linked checkboxes creating a checkbox set.
     *
     * @method Checkbox#add
     * @param {Checkbox} checkbox The checkbox to add to the collection of checkboxes
     * @param {Boolean} bidirectional Whether or not the current checkbox should be added to the passed in checkbox.
     */
    cp.add = function (checkbox, bidirectional) {
        if (bidirectional) {
            checkbox.add(this, false);

            _.each(this._relatedCheckboxes, function (related) {
                if (this._relatedCheckboxes.indexOf(checkbox) === -1) {
                    checkbox.add(related, true);

                }
            }.bind(this));

        }

        if (this._relatedCheckboxes.indexOf(checkbox) === -1) {
            this._relatedCheckboxes.push(checkbox);

            checkbox.on('only', function () {
                var value = this.value();
                _.each(this._relatedCheckboxes, function (related) {
                    if (checkbox !== related) {
                        related._controller.data(false, { silent: true });
                    }
                });
                this._controller.data(false, { silent: true });

                if (!_.isEqual(this.value(), value)) {
                    this.trigger('change', value);
                }
            }.bind(this));

            checkbox.on('change', function () {
                if (arguments.length !== 0) {
                    this.trigger('change');
                }
            }.bind(this));
        }
    };

    /**
     * @method Checkbox#disable
     *
     * @returns {Boolean}
     */
    /**
     * Setter for the disabled style on a checkbox.
     *
     * When passed a boolean it will disable only the current checkbox but if the parameter is `all` then all
     * checkboxes linked also get disabled. When passed `none` then all checkboxes are enabled.
     *
     * @method Checkbox#disable
     *
     * @param {Boolean|String} disabled
     */
    cp.disable = function (disabled) {
        if (!arguments.length) {
            return Widget.prototype.disable.call(this);
        }

        Widget.prototype.disable.call(this, disabled === true || disabled === 'all');

        if (disabled === 'all' || disabled === 'none') {
            _.invoke(this._relatedCheckboxes, 'disable', disabled === 'all');
        }
    };

    return Checkbox;
});


define('template-97', ['underscore'], function (_) { return function(data) {
var __t, __p = '';
__p += '<div class="orcl-widget orcl-radio"></div>';
return __p
};});


define('template-98', ['underscore'], function (_) { return function(data) {
var __t, __p = '';
__p += '<i class="pgbu-icon-small pgbu-icon-indicator-circle-partial"></i><i class="pgbu-icon-small pgbu-icon-indicator-circle-empty"></i>';
return __p
};});


define('template-99', ['underscore'], function (_) { return function(data) {
var __t, __p = '';
__p += '<div class="orcl-widget orcl-radio"> <input type="radio"> <i class="pgbu-icon-small pgbu-icon-indicator-circle-partial"></i> <i class="pgbu-icon-small pgbu-icon-indicator-circle-empty"></i></div>';
return __p
};});

define('oui/widgets/radio/radioView',['require','jquery','oui/widgets/checkbox/checkboxView','template-97','template-98','template-99'],function (require) {
    'use strict';

    var $ = require('jquery');
    var View = require('oui/widgets/checkbox/checkboxView');
    var radioWrapTemplate = require('template-97');
    var icons = require('template-98');
    var radioTemplate = require('template-99');

    /**
     * @class RadioView
     * @extends View
     * @private
     *
     * @param {null} model
     * @param {null} controller
     * @param {jQuery} $element
     * @param {Radio#Config} options
     * @param {Object} settings
     */
    function RadioView(model, controller, $element, options, settings) {
        View.call(this, model, controller, $element, options, settings);

    }

    var cvp = RadioView.prototype = Object.create(View.prototype);

    cvp.constructor = RadioView;

    /**
     * @method RadioView#buildElement
     *
     * @param {Radio#Config} options
     * @param {jQuery} $element
     */
    cvp.buildElement = function (options, $element) {

        if ($element != null) {
            $element = $element.wrap(radioWrapTemplate({})).parent();

            $element.find('label').append(icons());
        } else {
            $element = $(radioTemplate({}));
        }

        return $element;
    };

    return RadioView;
});

define('oui/widgets/radio',['require','underscore','oui/base/widget','oui/widgets/checkbox','oui/widgets/radio/radioView','oui/widgets/checkbox/checkboxController','oui/widgets/checkbox/checkboxModel'],function (require) {
    'use strict';

    var _ = require('underscore');
    var Widget = require('oui/base/widget');
    var Checkbox = require('oui/widgets/checkbox');
    var RadioView = require('oui/widgets/radio/radioView');
    var RadioController = require('oui/widgets/checkbox/checkboxController');
    var RadioModel = require('oui/widgets/checkbox/checkboxModel');

    var rp;

    /**
     * @typedef Radio#Config
     *
     * @property {Locale} [locale={}] The specific strings that are used by the widget and will use the defaults set from
     * the {@link module:oui.lang|lang}.
     */
    var __DEFAULTS__ = {

    };

    /**
     * @class Radio
     * @extends Widget
     *
     * @param {Radio#Config} [options] The options for the radio
     * @param {jQuery} [element] The element to use as the radio
     * @param {Object} [data] The data to set on the radio. Keyed data based on the name attribute of the radios and true/false.
     */
    function Radio(options, element, data) {
        var model;
        Widget.call(this, options, element, data);

        model = new RadioModel(this._initData, this._options);

        this._controller = new RadioController(model, this._options);

        this._view = new RadioView(model, this._controller, this._element, this._options, this._settings);

        this._relatedCheckboxes = [];
        this._setupEvents([model, this._view]);
    }

    rp = Radio.prototype = Object.create(Checkbox.prototype);

    rp.constructor = Radio;

    /**
     * Returns the name of the control
     *
     * @method Radio#name
     * @returns {String} 'radio' The name of the control
     */
    rp.name = function () {
        return 'radio';
    };

    /**
     * Returns the defaults of the radio controls
     *
     * @method Radio#defaults
     * @returns {Radio#Config} defaults The defaults for the radio
     */
    rp.defaults = function () {
        return __DEFAULTS__;
    };

    /**
     * Adds a radio to the collection of linked radios creating a radio set.
     *
     * @method Radio#add
     * @param {Radio} radio The radio to add to the collection of radios
     * @param {Boolean} bidirectional Whether or not the current radio should be added to the passed in radio.
     */
    rp.add = function (radio, bidirectional) {
        if (bidirectional) {
            radio.add(this, false);

            _.each(this._relatedCheckboxes, function (related) {
                if (this._relatedCheckboxes.indexOf(radio) === -1) {
                    radio.add(related, true);

                }
            }.bind(this));

        }

        if (this._relatedCheckboxes.indexOf(radio) === -1) {
            this._relatedCheckboxes.push(radio);

            this.on('change', function (data) {
                if (data === true) {
                    _.each(this._relatedCheckboxes, function (radio) {
                        radio.data(false);
                    });
                }
            }.bind(this));
        }
    };

    return Radio;
});

define('oui/widgets/form/fieldFactory',['require','underscore','oui/widgets/textField','oui/widgets/dateField','oui/widgets/numberField','oui/widgets/picker','oui/widgets/doublePicker','oui/widgets/typeAhead','oui/widgets/colorPicker','oui/widgets/comboBox','oui/widgets/file/filePicker','oui/widgets/editor','oui/widgets/urlField','oui/widgets/slider','oui/widgets/checkbox','oui/widgets/radio','oui/widgets/iconPicker'],function (require) {
    'use strict';

    var _ = require('underscore');
    var TextField = require('oui/widgets/textField');
    var DateField = require('oui/widgets/dateField');
    var NumberField = require('oui/widgets/numberField');
    var Picker = require('oui/widgets/picker');
    var DoublePicker = require('oui/widgets/doublePicker');
    var TypeAhead = require('oui/widgets/typeAhead');
    var ColorField = require('oui/widgets/colorPicker');
    var ComboBox = require('oui/widgets/comboBox');
    var FilePicker = require('oui/widgets/file/filePicker');
    var Editor = require('oui/widgets/editor');
    var UrlField = require('oui/widgets/urlField');
    var Slider = require('oui/widgets/slider');
    var Checkbox = require('oui/widgets/checkbox');
    var Radio = require('oui/widgets/radio');
    var IconPicker = require('oui/widgets/iconPicker');

    /**
     * @module fieldFactory
     */

    var FIELDS = {
        text: TextField,
        textarea: TextField,
        date: DateField,
        datetime: DateField,
        number: NumberField,
        picker: Picker,
        doublePicker: DoublePicker,
        typeahead: TypeAhead,
        color: ColorField,
        select: ComboBox,
        file: FilePicker,
        editor: Editor,
        url: UrlField,
        range: Slider,
        checkbox: Checkbox,
        radio: Radio,
        icon: IconPicker
    };

    return Object.freeze({
        /**
         * @alias module:fieldFactory
         * @method getTypeOf
         * @static
         *
         * @param {Widget} widget
         *
         * @returns {String} type of widget
         */
        getTypeOf: function (widget) {
            return _.findKey(FIELDS, function (fn) {
                return widget instanceof fn;
            });
        },

        /**
         * @alias module:fieldFactory
         * @method createField
         * @static
         *
         * @param {Object} definition
         * @param {String} type
         * @param {jQuery} $element
         *
         * @returns {Boolean|Widget} control
         */
        createField: function (definition, type, $element) {
            var supportedType = FIELDS[type] != null;
            var controlOptions;
            var control;

            if (!supportedType) {
                return false;
            }

            if (definition && definition.options) {
                controlOptions = definition.options;
            }

            controlOptions = _.defaults(controlOptions || {}, {
                fluid: $element.hasClass('input-block-level')
            });

            if (type === 'picker') {
                var pickerOptions = _.defaults(_.get(definition, 'options.picker', {}), {
                    fluid: $element.hasClass('input-block-level')
                });

                control = new FIELDS[type](pickerOptions, $element, definition.values || null);

                if (_.has(definition, 'options.typeAhead')) {
                    var typeAheadOptions = _.defaults(_.get(definition, 'options.typeAhead', {}), {
                        fluid: $element.hasClass('input-block-level')
                    });
                    new FIELDS.typeahead(typeAheadOptions, $element); // jshint ignore:line
                }
            } else {
                control = new FIELDS[type](controlOptions, $element, definition.values || null);
            }

            return control;
        }
    });
});

/* jshint maxdepth:2, maxcomplexity:6, maxparams:3, maxstatements:15 */
define('oui/widgets/form/formView',[
    'underscore',
    'jquery',
    'oui/widgets/form/fieldFactory',
    'oui/base/view'
], function (_, $, fieldFactory, View) { //jshint ignore:line
    'use strict';
    /**
     * @typedef {Object} FormView#ValidationResult
     * @property {String} message
     * @property {*} data
     * @property {Boolean} valid
     */

    var FIELD_ATTR = 'data-field';

    var DEFAULT_ATTRS = {
        required: false,
        visible: true,
        disable: false,
        readonly: false,
        lock: false,
        focus: false
    };

    /**
     * @method FormView~getFieldType
     * @private
     *
     * @param {jQuery} $field
     * @param {Object} definition
     *
     * @returns {String} field type
     */
    function getFieldType($field, definition) {
        if (definition && definition.type) {
            return definition.type;
        }

        if ($field.attr('data-oui-type') != null) {
            return $field.attr('data-oui-type');
        } else if ($field.is('input')) {
            return $field.attr('type');
        } else if ($field.is('select')) {
            return 'select';
        } else if ($field.is('textarea')) {
            return 'textarea';
        }
    }

    /**
     * @method FormView~mutationObserverSupported
     * @private
     *
     * @returns {Boolean}
     */
    function mutationObserverSupported() {
        return window.MutationObserver != null;
    }

    /**
     * @method FormView~addMutationObserver
     * @private
     *
     * @param {Form#Config} options
     * @param {HTMLElement} el
     * @param {Function} callback
     *
     * @returns {Object} mutation observer
     */
    function addMutationObserver(options, el, callback) {
        var mutationObserver = new window.MutationObserver(function (mutations) {
            if (options.autoCreateFields) {
                _.each(mutations, function (mutation) {
                    if (mutation.type === 'childList' && mutation.addedNodes.length > 0) {
                        var fields = _.filter(mutation.addedNodes, function (node) {
                            return (node.nodeName === 'INPUT' || node.nodeName === 'SELECT' ||
                                   node.nodeName === 'BUTTON') &&
                                node.getAttribute && node.getAttribute(FIELD_ATTR) !== '' &&
                                $(node).closest('.orcl-widget, .orcl-type-ahead').length === 0;
                        });

                        _.each(fields, callback);
                    }
                });
            }
        });

        mutationObserver.observe(el, {
            childList: true,
            subtree: true
        });

        return mutationObserver;
    }

    /**
     * @class FormView
     * @extends View
     * @private
     *
     * @param {Object} model
     * @param {Object} controller
     * @param {jQuery} $element
     * @param {Form#Config} options
     * @param {Object} settings
     */
    function FormView(model, controller, $element, options, settings) { //jshint ignore:line
        var that = this;
        View.call(this, options, $element, controller, settings);

        this._fields = {};
        this._customFields = {};
        this._help = {};

        this._$el.on('oui-change', this.changed.bind(this));
        this._$el.on('oui-error', this.error.bind(this));
        this._$el.on('oui-pre-change', this.prechanged.bind(this));

        this._$el.on('change', '[' + FIELD_ATTR + ']', function (event) {
            var fields = _.toArray(that._customFields);
            var contained = _.find(fields, function (elem) {
                return elem.is(event.target);
            });
            if (contained) {
                that.nativeChanged(event, event.target);
            }
        });

        if (mutationObserverSupported()) {
            addMutationObserver(options, this._$el[0], this.initializeField);
        }
    }

    var fvp = FormView.prototype = Object.create(View.prototype);

    fvp.constructor = FormView;

    /**
     * @method FormView#buildElement
     *
     * @param {Object} options
     * @param {jQuery} $element
     */
    fvp.buildElement = function (options, $element) {
        return $element;
    };

    /**
     * @method FormView#init
     */
    fvp.init = function () {
        _.each(this._$el.find('[' + FIELD_ATTR + ']'), this.initializeField.bind(this));
    };

    /**
     * @method FormView#_getFieldDefinition
     * @private
     *
     * @param {String} field
     *
     * @returns {Object} fieldDefinition
     */
    fvp._getFieldDefinition = function (field) {
        return _.get(this._options.fieldDefinitions, field, this._options.defaultFieldDefintiion);
    };

    /**
     * @method FormView#_addField
     * @private
     *
     * @param {String} name
     * @param {Widget} control
     */
    fvp._addField = function (name, control) {
        if (this._fields[name] != null && this._fields[name].add) {
            this._fields[name].add(control.init(), true);
        } else {
            this._fields[name] = control.init();
        }
    };

    /**
     * @method FormView#_addCustomField
     * @private
     *
     * @param {String} name
     * @param {jQuery} $field
     */
    fvp._addCustomField = function (name, $field) {
        if (this._customFields[name] != null) {
            this._customFields[name] = this._customFields[name].add($field);
        } else {
            this._customFields[name] = $field;
        }
    };

    /**
     * @method FormView#initializeField
     *
     * @param {HTMLElement} fieldEl
     */
    fvp.initializeField = function (fieldEl) {
        var $field = $(fieldEl);
        var name = $field.attr(FIELD_ATTR);
        var type = getFieldType($field, this._getFieldDefinition(name));
        var control;

        if (type === 'checkbox' || type === 'radio') {
            $field = $field.parent();
        }
        var definition = _.merge({}, this._options.typeDefinitions[type], this._getFieldDefinition(name));
        control = fieldFactory.createField(definition, type, $field);

        if (control) {
            this._addField(name, control);
        } else {
            this._addCustomField(name, $field);
        }

        if (definition.attributes) {
            this.setAttributes(name, _.defaults(definition.attributes, this._getDefaultAttributes()));
        }

        this.setFieldLabel($field, name);
        this.setInstructions(control, $field, name);
    };

    /**
     * @method FormView#setFieldLabel
     *
     * @param {jQuery} $field
     * @param {String} name
     */
    fvp.setFieldLabel = function ($field, name) {
        var label = $field.closest('.control-group').children('.control-label');
        var labelOption = _.get(this._options, ['fieldDefinitions', name, 'label'], '');

        if (label.length === 1 && (label.text() === '' || labelOption !== '')) {
            if (_.isFunction(labelOption)) {
                labelOption = labelOption();
            }
            label.text(labelOption);
        }
    };

    /**
     * @method FormView#setInstructions
     *
     * @param {Widget} control
     * @param {jQuery} $field
     * @param {String} name
     */
    fvp.setInstructions = function (control, $field, name) {
        var $instruction = control.element().siblings('.help-block');
        var labelOption = _.get(this._options, ['fieldDefinitions', name, 'help'], $instruction.text());

        if ($instruction.length === 0 && !_.isEmpty(labelOption)) {
            $instruction = $('<div class="help-block"></div>');

            $field.closest('.controls').append($instruction);
        }

        if ($instruction.text() !== labelOption) {

            if (_.isFunction(labelOption)) {
                labelOption = labelOption();
            }
            $instruction.text(labelOption)
                .attr('title', labelOption);
        }

        this._help[name] = labelOption;
    };

    /**
     * @method FormView#setStore
     *
     * @param {Function} Mediator
     *
     * @returns {Object|Undefined} mediator
     */
    fvp.setStore = function (Mediator) {
        if (Mediator != null) {
            this._mediator = new Mediator();

            var record = this._mediator.getRecord();
            if (record != null) {
                this.setRecord(record);
            }

            return this._mediator;
        }
    };

    /**
     * @method FormView#reset
     */
    fvp.reset = function () {
        var that = this;

        this.clearInvalidFields(true);
        _.invoke(this._fields, 'empty');

        _.each(this._customFields, function (value, key) {
            var fieldDefinition = this._getFieldDefinition(key);
            if (fieldDefinition && fieldDefinition.reset) {
                fieldDefinition.reset(value);
            }
        }.bind(this));

        if (this._mediator) {
            _.each(this._fields, function (value, field) {
                that.setFieldOptions(field, that._mediator.getOptions(field));
            });
        }
    };

    /**
     * @method FormView#setFieldOptions
     *
     * @param {String} field
     * @param {Object} options
     */
    fvp.setFieldOptions = function (field, options) {
        if (this._fields[field]) {
            this._fields[field].updateOptions(options);
        }
    };

    /**
     * @method FormView#getValues
     *
     * @returns {Object} record
     */
    fvp.getValues = function () {
        var that = this;
        var record = _.reduce(this._fields, function (memo, control, field) {
            var value = control.value();

            if (this._options.transformPickerData &&
                _.isArray(value) &&
                control.name() === 'picker' && control.options().tree.multiSelect === false) {
                value = _.first(value);
            }
            return _.set(memo, field, value);
        }.bind(this), {});

        record = _.reduce(this._customFields, function (memo, value, key) {
            return _.set(memo, key, that._getFieldDefinition(key).getValue(value));
        }, record);

        return record;
    };

    /**
     * @method FormView#getRecord
     *
     * @returns {Object} record
     */
    fvp.getRecord = function () {
        if (this._mediator) {
            return this._mediator.getRecord();
        }

        return this.getValues();
    };

    /**
     * @method FormView#_getDefaultAttributes
     * @private
     *
     * @returns {Object} default attributes
     */
    fvp._getDefaultAttributes = function () {
        return _.clone(DEFAULT_ATTRS);
    };

    /**
     * @method FormView#setRecord
     *
     * @param {Object} data
     */
    fvp.setRecord = function (data) {
        this.redraw(null, data);
    };

    /**
     * @method FormView#redraw
     *
     * @param {Object|Array<String>} fields
     * @param {Object} data
     */
    fvp.redraw = function (fields, data) {
        var that = this;

        if (fields == null) {
            fields = this._fields;
        }

        this.pause();

        _.each(fields, function (field) {
            var control;
            var value;

            if (_.isString(field)) {
                control = this._fields[field];
            } else {
                control = field;
                field = this._getField(field);
            }
            value = control.value();
            if (this._mediator) {
                value = this._mediator.getValue(field);
            } else if (_.has(data, field)) {
                value = _.get(data, field);
            }

            if (value == null) {
                value = null;
            }
            this.setValue(value, field);
        }.bind(this));

        if (this._mediator) {
            //Go over all known fields and get the attributes
            _.each(fields, function (value, field) {
                that.setAttributes(field, that._mediator.getAttributes(field, that._getDefaultAttributes(field)));
                that.setFieldOptions(field, that._mediator.getOptions(field));
            });
        }

        var invalidFields = this.getInvalidFields();

        var validFields = _.difference(_.keys(this._fields), _.keys(invalidFields));
        _.each(invalidFields, function (validationResult, key) {
            this.addError(this._fields[key], validationResult);
            this.trigger('error', key, validationResult);
        }.bind(this));

        this.resume();
        this.clearInvalidFields(validFields);
    };

    /**
     * @method FormView#redrawHelp
     */
    fvp.redrawHelp = function () {
        _.each(this._fields, function (field) {
            var control;
            if (_.isString(field)) {
                control = this._fields[field];
            } else {
                control = field;
                field = this._getField(field);
            }

            this.setInstructions(control, control.element(), field);
        }.bind(this));
    };

    /**
     * @method FormView#setValue
     *
     * @param {*} value
     * @param {String} field
     */
    fvp.setValue = function (value, field) {
        var control = this._fields[field];

        if (this._options.transformPickerData &&
            !_.isArray(value) &&
            control.name() === 'picker' && control.options().tree.multiSelect === false) {
            value = [value];
        }

        if (control && control.value) {
            control.value(value);
        } else if (this._customFields[field]) {
            this._getFieldDefinition(field).setValue(this._customFields[field], value);
        }
    };

    /**
     * @method FormView#getValue
     *
     * @param {String} field
     *
     * @returns {*} value
     */
    fvp.getValue = function (field) {
        if (this._mediator) {
            return this._mediator.getValue(field);
        } else {
            var control = this._fields[field];

            if (control && control.value) {
                return control.value();
            } else if (this._customFields[field]) {
                this._getFieldDefinition(field).getValue(this._customFields[field]);
            }

        }
    };

    /**
     * @method FormView#setAttributes
     *
     * @param {String} field
     * @param {Object} attributes
     */
    fvp.setAttributes = function (field, attributes) {

        if (attributes.visible == null) {
            attributes.visible = true;
        }

        this.updateAttributes(field, attributes);
        this.updateFieldAttributes(field, attributes);
    };

    /**
     * @method FormView#getDependencies
     *
     * @param {String} field
     * @param {Array} [canaries]
     */
    fvp.getDependencies = function (field, canaries) {
        var that = this;
        var deps = [];
        canaries = canaries || [];

        if (_.contains(canaries, field)) {
            return deps;
        }

        if (this._options.links && this._options.links[field]) {

            deps = this._options.links[field];

            //Visited
            canaries.push(field);

            _.each(this._options.links[field], function (linked) {
                deps = _.union(deps, that.getDependencies(linked, canaries));
            });
        }

        return deps;
    };

    /**
     * @method FormView#_getField
     *
     * @param {Widget} widget
     *
     * @returns {String} field
     */
    fvp._getField = function (widget) {
        var attr = widget.element().attr(FIELD_ATTR);

        if (attr != null) {
            return attr;
        }

        return widget.element().find('[' + FIELD_ATTR + ']').attr(FIELD_ATTR);
    };

    /**
     * @method FormView#prechanged
     *
     * @param {Object} event
     * @param {Widget} widget
     * @param {String} eventName
     * @param {*} newData
     * @param {*} oldData
     *
     * @returns {Boolean|Promise}
     */
    fvp.prechanged = function (event, widget, eventName, newData, oldData) { //jshint ignore:line
        if (widget.name() !== 'form') {
            return this.trigger('pre-change', this._getField(widget), newData, oldData);
        }
    };

    /**
     * @method FormView#changed
     *
     * @param {Object} event
     * @param {Widget} widget
     */
    fvp.changed = function (event, widget) {
        var field = this._getField(widget);
        var control = this._fields[field];
        var value = control.value();

        if (this.isPlaying() && widget.name() !== 'form') {

            if (this._options.transformPickerData &&
                _.isArray(value) &&
                control.name() === 'picker' && control.options().tree.multiSelect === false) {
                value = _.first(value);
            }

            this._clearErrorMessage(widget);

            if (this._mediator && field) {
                this._mediator.setValue(field, control.value());
                _.each(this.getDependencies(field), function (field) {
                    this.setValue(this._mediator.getValue(field), field);
                    this.setAttributes(field, this._mediator.getAttributes(field, this._getDefaultAttributes()));
                }.bind(this));
            }
            this.setInstructions(control, control.element(), field);

            if (this._options.redrawFormOnFieldUpdate) {
                this.redraw(_.omit(this._fields, field));
            }

            this.trigger('change', this._getField(widget));
        }
    };

    /**
     * @method FormView#nativeChanged
     *
     * @param {Object} event
     * @param {HTMLElement} target
     */
    fvp.nativeChanged = function (event, target) {
        var field = $(target).attr(FIELD_ATTR);
        if (this._mediator) {
            this._mediator.setValue(field, target.value);
        }
    };

    /**
     * @method FormView#error
     *
     * @param {Object} event
     * @param {Widget} widget
     * @param {String} name
     * @param {*} attempted
     * @param {*} current
     * @param {FormView#ValidationResult} validationResult
     */
    fvp.error = function (event, widget, name, attempted, current, validationResult) { //jshint ignore:line
        if (widget.name() !== 'form') {
            var field = this._getField(widget);
            if (this._mediator && field) {
                this._mediator.setValue(field, attempted);

            }

            if (this._options.redrawFormOnFieldUpdate) {
                this.redraw(_.omit(this._fields, field));
            }

            //TODO: look into if this error should accept a return value where the validation result gets a message?
            this.trigger('error', field, validationResult);
            this.addError(widget, validationResult);
        }
    };

    /**
     * @method FormView#_getErrorMessage
     * @private
     *
     * @param {String} field
     * @param {FormView#ValidationResult} result
     * @param {Widget} widget
     *
     * @returns {String|undefined} message
     */
    fvp._getErrorMessage = function (field, result, widget) {
        if (_.has(this._getFieldDefinition(field), 'validator.getMessage')) {
            return this._getFieldDefinition(field).validator.getMessage(field, result, widget.options());
        }
    };

    /**
     * @method FormView#_setErrorMessage
     * @private
     *
     * @param {Widget} widget
     * @param {FormView#ValidationResult} validationResult
     */
    fvp._setErrorMessage = function (widget, validationResult) {
        var $existingHelpBlock = widget.element().siblings('.help-block');
        var field = this._getField(widget);
        var message;
        if ($existingHelpBlock.length === 0) {

            $existingHelpBlock = $('<div class="help-block"></div>');
            widget.element().after($existingHelpBlock);
            this._help[field] = '';
        }

        if (this._help[field] == null) {
            this._help[field] = $existingHelpBlock.text();
        }

        message = validationResult.message || this._getErrorMessage(field, validationResult, widget);
        $existingHelpBlock
            .text(message)
            .attr('title', message);

        this.trigger('error-message-added', this._getField(widget));
    };

    /**
     * @method FormView#_clearErrorMessage
     * @private
     *
     * @param {Widget} widget
     */
    fvp._clearErrorMessage = function (widget) {
        var existingHelpBlock = widget.element().siblings('.help-block');
        var message;
        if (existingHelpBlock.length > 0) {
            message = this._help[this._getField(widget)] || '';
            existingHelpBlock.text(message)
                .attr('title', message);
        }

        this.trigger('error-message-cleared', this._getField(widget));
    };

    /**
     * @method FormView#isValid
     *
     * @returns {Boolean}
     */
    fvp.isValid = function () {
        return _.all(this._fields, function (field) {
            return field.isValid();
        });
    };

    /**
     * @method FormView#focusField
     *
     * @param {String} field
     */
    fvp.focusField = function (field) {
        var $field = this._$el.find('[' + FIELD_ATTR + '="' + field + '"]').first();

        $field.focus();

        if (document.activeElement !== $field[0]) {
            $field.one('transitionend', function () {
                $field.focus();
            }.bind(this));
        }
    };

    /**
     * @method FormView#updateForm
     *
     * @param {Object} attributes
     */
    fvp.updateForm = function (attributes) {
        _.each(this._fields, function (control) {
            if (control.updateOptions) {
                control.updateOptions(attributes);
            }
        });
    };

    /**
     * @method FormView#updateFieldAttributes
     *
     * @param {String} field
     * @param {Object} attributes
     */
    fvp.updateFieldAttributes = function (field, attributes) {
        var control = this._fields[field];

        if (control) {
            control.element().closest('.control-group')
                .toggleClass('required', control.name() !== 'checkbox' && control.name() !== 'radio' && _.get(attributes, 'required', false))
                .toggleClass('hidden', !attributes.visible);

            if (!attributes.visible) {
                this._clearErrorMessage(control);
            }

            if (attributes.focus) {
                this.focusField(field);
            }
        }
    };

    /**
     * @method FormView#updateAttributes
     *
     * @param {String} field
     * @param {Object} attributes
     */
    fvp.updateAttributes = function (field, attributes) {
        var control = this._fields[field];

        if (control && control.updateOptions) {
            control.updateOptions(attributes);
        }
    };

    /**
     * @method FormView#addError
     *
     * @param {Widget} widget
     * @param {FormView#ValidationResult} error
     */
    fvp.addError = function (widget, error) {
        if (_.isString(widget)) {
            widget = this._fields[widget];
        }
        this._setErrorMessage(widget, error);

        if (!error.valid && !widget.element().closest('.control-group').hasClass('error')) {
            widget.element().closest('.control-group').addBack().addClass('error');
        }
    };

    /**
     * @method FormView#setInvalidFields
     *
     * @param {Object<String, FormView#ValidationResult>} map
     */
    fvp.setInvalidFields = function (map) {
        var that = this;
        _.each(map, function (value, field) {
            that.addError(field, value);
        });
    };

    /**
     * @method FormView#clearInvalidFields
     *
     * @param {Boolean|Object} fields
     */
    fvp.clearInvalidFields = function (fields) {
        var that = this;

        if (_.isBoolean(fields) && fields) {
            _.each(this._fields, function (field) {
                that._clearErrorMessage(field);
            });
        } else {
            _.each(fields, function (field) {
                that._clearErrorMessage(that._fields[field]);
            });
        }
    };

    /**
     * @method FormView#getInvalidFields
     *
     * @returns {Object<String, FormView#ValidationResult>} invalid fields
     */
    fvp.getInvalidFields = function () {
        return _.transform(this._fields, function (memo, field, key) {
            if (!field.isValid()) {
                memo[key] = field.error();
            }
        }, {});
    };

    return FormView;
});

define('oui/widgets/form',[
    'underscore',
    'oui/base/widget',
    'oui/widgets/form/formView'
], function (_, Widget, FormView) {
    'use strict';

    var fp;

    /**
     * @typedef Form~Mediator#attributes
     *
     * @property {Boolean} visible
     * @property {Boolean} readonly
     * @property {Boolean} disable
     * @property {Boolean} lock
     * @property {Boolean} focus
     * @property {Boolean} required
     */

    /**
     * **Note**: no behavior is provided by OUI. Just a reference for documentation
     *
     * @class Form~Mediator
     * @abstract
     *
     */
    /**
     * Get the value for the field from the record in case any value is computed.
     *
     * **Note**: depending on the form this may be a {@link https://lodash.com/docs#get|lodash} path.
     *
     * @method Form~Mediator#getValue
     * @param {String} field The field to get the value for.
     * @returns {*} value The value of the field.
     */
    /**
     * Sets the value for the field. The values type is dependent on the field type. For example, date types will be moments,
     * numbers will be numbers, etc.
     *
     * **Note**: depending on the form the field may be of a {@link http://lodash.com/docs#set|lodash} path.
     *
     * @method Form~Mediator#setValue
     * @param {String} field The field to set the value for.
     * @param {*} value The value of the field as set by the control.
     */
    /**
     * Used when the underlying record changes and the options should be updated for controls. Mainly added to facilitate
     * fields that should have different loading functions.
     *
     * The control will only update the fields within the returning object. All others will be kept the same.
     *
     * @see Checkbox#Config
     * @see ColorPicker#Config
     * @see ComboBox#Config
     * @see DateField#Config
     * @see Editor#Config
     * @see FilePicker#Config
     * @see IconPicker#Config
     * @see NumberField#Config
     * @see Picker#Config
     * @see Radio#Config
     * @see Slider#Config
     * @see TextField#Config
     * @see TypeAhead#Config
     * @see UrlField#Config
     *
     * @method Form~Mediator#getOptions
     * @param {String} field The new options to get for the field.
     * @returns {Object} options required for the field.
     */
    /**
     * Returns the object containing all the data making up the form.
     *
     * @method Form~Mediator#getRecord
     * @returns {Object} record The object backing the form.
     */
    /**
     * Returns the attributes for the field. Used when redrawing the form, and on control updates.
     *
     * When redrawing a form, all fields attributes are requested in case any currently hidden fields should be updated
     * or vice versa as a result of the change..
     *
     * @method Form~Mediator#getAttributes
     * @param {String} field The field to get the attributes for.
     * @returns {Form~Mediator#attributes} The current attributes for the field.
     */

    /**
     * @typedef {enum} Form~FieldTypes
     *
     * @property {String} text
     * @property {String} textarea
     * @property {String} date
     * @property {String} datetime
     * @property {String} number
     * @property {String} picker
     * @property {String} typeahead
     * @property {String} color
     * @property {String} select
     * @property {String} file
     * @property {String} editor
     * @property {String} url
     * @property {String} range
     * @property {String} checkbox
     * @property {String} radio
     */

    /**
     * @typedef Form~FieldDefinition
     *
     * @property {String|Function} [label] The label text to use for the field.
     * @property {String|Function} [help] The help text to use for the field.
     * @property {Form~FieldTypes} [type] The type of field to create. If not provided it will be inferred from the given
     * element.
     * @property {Object} options The options that get forwarded to the OUI control.
     * @property {Function} getValue
     * @property {Function} reset
     * @property {Function} setValue
     * @property {Object} validator
     * @property {Function} validator.getMessage
     */

    /**
     * @typedef Form#Config
     *
     * @property {Boolean} [tranformPickerData=true] Wraps/Unwraps single object (not multiselect) pickers between an array
     * and an object.
     * @property {Boolean} [autoConfigureFields=false] Whether or not to read additional data properties from the field
     * during intialization.
     * @property {Boolean} [includeEmptyFields=false] Whether to include fields from the field definition that are not yet
     * included within the form.
     * @property {Boolean} [redrawFormOnFieldUpdate=true] Completely redraws the field when any field is updated in case
     * attributes or values on other fields are affected by the update.
     * @property {Boolean} [showErrors=true] Whether or not to show any errors within the form
     * @property {Object<String, Object>} [typeDefinitions={}] Map of types to a field definition. Works as defaults on a
     * type level.
     * @property {Object<String, Form~FieldDefinition>} [fieldDefinitions={}] Map of field name (data-field attribute) to configure the
     * options for the controls.
     * @property {Function} [getElement] Custom elements not directly supported by the library to be constructed by the
     * function and returned.
     * @property {Object<String, Array<String>>} [links={}] Map of field name to an array of field names that should be
     * updated whenever the field (key) is updated. This includes attributes, options, and values.
     * @property {Array<String>} [contextSensitive=[]] List of field names that are context sensitive. This means that when the supporting record
     * is changed the options should be re-requested as load paths may be changing.
     * @property {Function} [getOptions] Function to be called whenever dynamic options are requested.
     * @property {Object<String, String>} [recordKeys={label:'label'}] The keys to use within the record to find various
     * required data fields.
     * @property {Function} [getMessage] Function to get the hint/error text for the field whenever an update occurs (update
     * or error).
     */
    // jscs:disable jsDoc
    var __DEFAULTS__ = {
        transformPickerData: true,
        autoConfigureFields: false,
        includeEmptyFields: false,
        redrawFormOnFieldUpdate: true,
        showErrors: true,
        typeDefinitions: {},
        fieldDefinitions: {},
        getElement: _.constant(null),
        links: {},
        getOptions: _.constant(null),
        contextSensitive: [],
        recordKeys: {
            label: 'label'
        },
        getMessage: _.constant('')
    };
    // jscs:enable jsDoc

    /**
     * The form component handles linking of fields to their inputs. Currently, only inputs supported by OUI are supported
     * but future plans include custom components (maps, charts, grids...).
     *
     * Requirements for the form:
     *
     *  * All fields set to be tracked by the Form must have the attribute data-field on their input with the field they
     *  represent in the record.
     *  * Currently, must pass the element that represents the form into the component during construction.
     *
     * Limitations of the component:
     *
     *  * No custom elements currently supported
     *  * No layout is done from within this component. The element must be laid out from a template.
     *
     * @example <caption>Form with Mediator</caption>
     *
     *  function Mediator() {
     *      return {
     *          //Mediator API
     *      };
     *  }
     *
     *  var form = new Form({}, $form);
     *  form.init(Mediator);
     *
     * @class Form
     * @extends Widget
     * @see Form~Mediator
     *
     * @param {Form#Config} [options] The options for the form.
     * @param {jQuery} element The element to use for the forms construction. This is currently required during construction.
     * @param {Object} [data] The data/record to use for the values of the fields.
     */
    function Form(options, element, data) {
        Widget.call(this, options, element, data);

        //Nulls for 'model' and 'controller' to keep consistent with a base view call
        this._view = new FormView(null, null, this._element, this._options, this._settings);

        Object.defineProperty(this, '_data', {
            get: function () {// jscs:ignore jsDoc
                return this.getRecord();
            },
            set: function (record) {// jscs:ignore jsDoc
                this.setRecord(record);
            }
        });

        this._setupEvents([this._view]);
    }

    Form.prototype = Object.create(Widget.prototype);
    fp = Form.prototype;

    fp.constructor = Form;

    /**
     * Gets the name of the component.
     *
     * @method Form#name
     * @returns {String} 'form'
     */
    fp.name = function () {
        return 'form';
    };

    /**
     * Gets the defaults for the component.
     *
     * @method Form#defaults
     * @returns {Form#Config} defaults
     */
    fp.defaults = function () {
        return __DEFAULTS__;
    };

    /**
     * Returns the current validity of the form.
     *
     * @method Form#isValid
     * @returns {Boolean} - The validity of the entire form
     */
    fp.isValid = function () {
        return this._view.isValid();
    };

    //Named as init - to publicly define that the init function could accept a mediator
    /**
     * Initializes the form component when it has been added to the DOM.
     *
     * Can be set up to use a Mediator for more complicated forms with many rules, and linked fields.
     *
     * @method Form#init
     * @param {Form~Mediator|Object} [Mediator] The object to use to set/get values from the backing record. Useful in case there are any
     * rules that need to update multiple fields when another field is updated.
     *
     * @returns {Function} [instance] - Returns the constructed mediator instance.
     */
    fp._init = function (Mediator) {
        if (_.isFunction(Mediator)) {
            this.reset();
            var instance = this._view.setStore(Mediator);

            return instance;
        } else if (_.isObject(Mediator)) {
            this.setRecord(Mediator);
        }
    };

    /**
     * Forces a validation of the current state of the form.
     *
     * @method Form#validate
     */
    fp.validate = function () {
        this._view.validate();
    };

    /**
     * Empties out all controls on the form.
     *
     * @method Form#reset
     */
    fp.reset = function () {
        this._view.reset();
    };

    /**
     * Empties out all controls on the form.
     *
     * @method Form#invalidate
     */
    fp.invalidate = function () {
        this.reset();
    };

    /**
     * Gets the backing record of the form. If using a mediator, it will defer to that getRecord otherwise it
     * will build up an object with all the controls' fields as keys and values as values.
     *
     * @method Form#getRecord
     */
    fp.getRecord = function () {
        return this._view.getRecord();
    };

    /**
     * Resets the field and then sets the record on the form updating all controls in the process.
     *
     * @method Form#setRecord
     * @param {Object} record The data to set on the form
     */
    fp.setRecord = function (record) {
        this.reset();
        this._view.setRecord(record);
    };

    /**
     * Sets the field to the value provided.
     *
     * @method Form#setValue
     * @param {String} field The field to set
     * @param {*} value The value to set on that field. Type depends on the field being set.
     */
    fp.setValue = function (field, value) {
        //Internally used within the view as an iterator to _ hence the backwards params
        this._view.setValue(value, field);
    };

    /**
     * Gets the value for the field.
     *
     * @method Form#getValue
     * @param {String} field The field to get the value of
     * @returns {*} value The value of the field. Type depends on the field requested.
     */
    fp.getValue = function (field) {
        return this._view.getValue(field);
    };

    /**
     * Disables the entire form and all controls within the form.
     *
     * @method Form#disable
     * @param {Boolean} disabled Whether the form should be disabled as a whole.
     */
    fp.disable = function (disabled) {
        this._view.updateForm({ disable: disabled });
    };

    //Override base widget since this is not implemented on a 'form' component yet.
    /**
     * @method Form#lock
     * @todo Implement function for 'form' component if needed.
     */
    fp.lock = _.noop;

    /**
     * Disables or enables a field within the form.
     *
     * @method Form#disabled
     * @param {String} field The field to disable
     * @param {Boolean} disabled Whether or not to disable the field.
     */
    fp.disabled = function (field, disabled) {
        this._view.updateAttributes(field, { disable: disabled });
    };

    /**
     * Locks or unlocks a field within the form.
     *
     * @method Form#locked
     * @param {String} field The field to lock
     * @param {Boolean|Object} locked Whether or not to lock the field and how to lock the field.
     */
    fp.locked = function (field, locked) {
        this._view.updateAttributes(field, { lock: locked });
    };

    /**
     * Marks a field as readonly or editable within the form.
     *
     * @method Form#readonly
     * @param {String} field The field to mark readonly
     * @param {Boolean} readonly Whether or not the field is editable or not.
     */
    fp.readonly = function (field, readonly) {
        this._view.updateAttributes(field, { readonly: readonly });
    };

    /**
     * Changes the visibility attribute of the control within the form. The entire control group will be
     * marked as visible/invisible.
     *
     * @method Form#visible
     * @param {String} field The field to mark visible.
     * @param {Boolean} visible Whether or not to mark the field visible or hidden
     */
    fp.visible = function (field, visible) {
        this._view.updateAttributes(field, { visible: visible });
        this._view.updateFieldAttributes(field, { visible: visible });
    };

    /**
     * Redraws the form or the specific fields given.
     *
     * @method Form#redraw
     * @param {Array<String>} [fields] The fields to redraw or all of them if none are provided
     */
    fp.redraw = function (fields) {
        this._view.redraw(fields);
        this._view.redrawHelp();
    };

    /**
     * Focuses a specific field within the form.
     *
     * @method Form#focusField
     * @param {String} field The field to focus
     */
    fp.focusField = function (field) {
        this._view.focusField(field);
    };

    /**
     * Adds an error to the form for the field given.
     *
     * @method Form#addError
     * @param {String} field The field to add the error for.
     * @param {String} error The message to set for the field.
     */
    fp.addError = function (field, error) {
        this._view.addError(field, error);
    };

    /**
     * Sets multiple errors on the form. Uses the key as the field and the value as the message
     * to set as the error.
     *
     * @method Form#setInvalidFields
     * @param {Object<String, String>} errors The map of errors to messages to set
     */
    fp.setInvalidFields = function (errors) {
        this._view.setInvalidFields(errors);
    };

    /**
     * Clears the fields' errors or all the forms errors.
     *
     * @method Form#clearInvalidFields
     * @param {Array<String>|Boolean} fields The fields to clear or whether or not to clear all fields.
     */
    fp.clearInvalidFields = function (fields) {
        this._view.clearInvalidFields(fields);
    };

    /**
     * @method Form#getInvalidFields
     *
     * @returns {FormView#InvalidFields}
     */
    fp.getInvalidFields = function () {
        return this._view.getInvalidFields();
    };

    return Form;
});


define('template-100', ['underscore'], function (_) { return function(data) {
var __t, __p = '';
__p += '<div class="orcl-widget orcl-form-canvas form-canvas selected-border" tabindex="0"> <div class="grid-lines row-fluid"> <div class="span1"></div> <div class="span1"></div> <div class="span1"></div> <div class="span1"></div> <div class="span1"></div> <div class="span1"></div> <div class="span1"></div> <div class="span1"></div> <div class="span1"></div> <div class="span1"></div> <div class="span1"></div> <div class="span1"></div> </div> <div class="drag-overlay"> <span></span> </div> <div class="overlay"></div> <div class="form-editor"> </div></div>';
return __p
};});


define('template-101', ['underscore'], function (_) { return function(data) {
var __t, __p = '';
__p += '<div class="span' +
((__t = ( data.size )) == null ? '' : __t) +
'" data-width="' +
((__t = ( data.size )) == null ? '' : __t) +
'" data-column="' +
((__t = ( data.column )) == null ? '' : __t) +
'"></div>';
return __p
};});


define('template-102', ['underscore'], function (_) { return function(data) {
var __t, __p = '';
__p += '<div class="row-fluid" data-row=\'' +
((__t = ( data.id )) == null ? '' : __t) +
'\'></div>';
return __p
};});


define('template-103', ['underscore'], function (_) { return function(data) {
var __t, __p = '';
__p += '<div class="' +
((__t = ( ['orcl-draggable'].concat(data.selected ? 'selected' : '').concat(data.active ? 'active' :     '').concat(data.copied ? 'copied' : '').join(' ') )) == null ? '' : __t) +
'"     data-width="' +
((__t = ( data.width )) == null ? '' : __t) +
'" data-level="' +
((__t = ( data.level )) == null ? '' : __t) +
'" data-id="' +
((__t = ( data.id )) == null ? '' : __t) +
'"> <div class="resize-handle handle-vertical handle-left"></div> <div class="resize-handle handle-vertical handle-right"></div> <div class="drag-handle"></div></div>';
return __p
};});

define('oui/widgets/canvas/formEditorView',['require','underscore','jquery','oui/base/view','oui/utils/draggable','oui/utils/math','template-100','template-101','template-102','template-103'],function (require) { //jshint ignore:line
        'use strict';

        var _ = require('underscore');
        var $ = require('jquery');
        var View = require('oui/base/view');
        var draggable = require('oui/utils/draggable');
        var math = require('oui/utils/math');
        var canvasTemplate = require('template-100');
        var colTemplate = require('template-101');
        var rowTemplate = require('template-102');
        var handleTemplate = require('template-103');

        var modes = {
            SELECTION: 'selection',
            MOVE: 'move'
        };

        /**
         * @method FormEditorView~getId
         * @private
         *
         * @param {jQuery} $el
         *
         * @returns {Number}
         */
        function getId($el) {
            return $el.attr('data-id');
        }

        var clamp = math.clamp.bind(null, 1, 12);

        var fevp;

        /**
         * @class FormEditorView
         * @private
         *
         * @param {FormEditorModel} model
         * @param {FormEditorController} controller
         * @param {jQuery} element
         * @param {FormEditor~Config} options
         * @param {Object} settings
         */
        function FormEditorView(model, controller, element, options, settings) {
            var that = this;

            //Super call
            View.call(this, options, element, controller, settings);

            //Bind events on the model
            model.on({
                'view.add': this.add.bind(this),
                'view.update': this.update.bind(this),
                'view.remove view.removeAll': this.rerender.bind(this),
                'view.node-moved': this.move.bind(this),
                'view.node-collapsed': this.collapse.bind(this),
                'view.guides-changed': this.guides.bind(this),
                'view.attribute-updated:selected': this.select.bind(this),
                'view.active': this.active.bind(this),
                'view.rerender': this.rerender.bind(this),
                'view.attribute-updated:copy': this.copied.bind(this)
            });

            //During initialization - make sure to use whatever is the status
            //from the model
            this.guides(model.guides());
            this._mode = modes.SELECTION;

            this._editor = this._$el.find('.form-editor');

            this._$el.on('click', '[data-id]', function (event) {
                var id = getId($(this));
                event.preventDefault();
                //Need to stop propagation to prevent it from bubbling to the groupbox but still
                //need to trigger a click for everything above the formEditor that needs to listen to this
                //For example closing a dropdown menu in add
                event.stopPropagation();
                that._$el.parent().trigger('click');
                if (event.ctrlKey) {
                    var selected = that._controller.selected();

                    if (_.contains(selected, id)) {
                        that._controller.selected(_.without(selected, id));
                    } else {
                        that._controller.selected(_.union(selected, [id]));
                    }
                } else {
                    that._controller.selected([id]);
                }
            })
            .on('click', function () {
                if ($(this).is(':not([data-id])') && $(this).closest('[data-id]').length === 0) {
                    that._controller.selected([]);
                }
            })
            .on('mousedown', '[data-id]', function (event) {
                var id = getId($(event.target).closest('[data-id]'));
                that._controller.active(id);
            });

            draggable(this._$el, {
                constrainTo: this._$el.find('.form-editor'),
                clone: true,
                appendCloneTo: this._$el.find('.drag-overlay'),
                triggerDistance: 10,
                selector: '.orcl-draggable.active',
                cancellable: true,
                namespace: 'move'
            });

            draggable(this._$el, {
                constrainTo: this._$el.find('.form-editor'),
                triggerDistance: 10,
                selector: '.orcl-draggable .resize-handle.handle-vertical',
                cancellable: true,
                axis: 'x',
                namespace: 'resize'
            });

            this._$el.on('drag-start.move', function (event) {
                if (this.isEditable()) {
                    event.$handle.width(event.$handle.width());
                    event.$handle.height(event.$handle.height());
                } else {
                    event.preventDefault();
                }
            }.bind(this));

            this._$el.on('drag-start.move drag-start.resize', function (event) {
                if (this.isEditable()) {
                    var eventId = event.$handle.closest('[data-id]').attr('data-id');
                    return {
                        position: that._controller.position(that._controller.node(eventId)),
                        state: that._controller.nodes()
                    };
                } else {
                    event.preventDefault();
                }
            }.bind(this));

            this._$el.on('drag-cancel.move drag-cancel.resize', function (event, original) {
                event.$handle.css({
                    left: '',
                    top: '',
                    width: '',
                    height: ''
                });
                //Rerender the current model positions
                that._controller.restore(_.cloneDeep(original.state));
            });

            this._$el.on('dragging.move', function (event, original) {
                var eventId = event.$handle.closest('[data-id]').attr('data-id');
                var rowWidth = that.getDraggable(eventId).closest('[data-row]').width();
                var column = (Math.round((event.deltaX / (rowWidth / 12))));
                var row = parseInt($(document.elementFromPoint(event.current.x, event.current.y)).closest('[data-row]').attr('data-row'), 10);

                that._controller.restore(_.cloneDeep(original.state));
                that._controller.move(eventId, {
                    column: clamp(original.position.column + column),
                    row: row
                });
            });

            this._$el.on('dragging.resize', function (event, original) {
                var eventId = event.$handle.closest('[data-id]').attr('data-id');
                var rowWidth = that.getDraggable(eventId).closest('[data-row]').width();
                var column = (Math.round(event.deltaX / (rowWidth / 12)));

                that._controller.restore(_.cloneDeep(original.state));

                var newPos = {};
                if (event.$handle.hasClass('handle-left')) {
                    if (column < 0) {
                        newPos.column = clamp(original.position.column + column);
                        newPos.width = original.position.width - column;
                    } else {
                        newPos.width = original.position.width - column;
                        newPos.column = clamp(original.position.column + column);
                    }
                }

                if (event.$handle.hasClass('handle-right')) {
                    if (clamp(original.position.width + column) + original.position.column - 1 <= 12) {
                        newPos.width = clamp(original.position.width + column);
                    } else {
                        newPos.width = original.position.width;
                    }
                }

                that._controller.move(eventId, newPos);
            });

            this._$el.on('drag-end.move drag-end.resize', function (event, original) {
                if (original) {
                    that._controller.storeState(original.state);
                }
            });

            this.addShortcuts({
                'ESC': this.escapeHandler.bind(this),
                'DELETE': this.deleteHandler.bind(this),
                'DOWN': this.moveHandler.bind(this, { vertical: 1 }),
                'LEFT': this.moveHandler.bind(this, { horizontal: -1 }),
                'RIGHT': this.moveHandler.bind(this, { horizontal: 1 }),
                'UP': this.moveHandler.bind(this, { vertical: -1 }),
                'CTRL-Z': this._controller.undo.bind(this._controller),
                'CTRL-Y': this._controller.redo.bind(this._controller),
                'CTRL-C': function () {
                    if (this.isEditable()) {
                        this._controller.copy(this._controller.selected());
                    }
                }.bind(this),
                'CTRL-V': this._controller.paste.bind(this._controller, { row: 1 }),
                //'CTRL-SHIFT-V': this._controller.paste.bind(this._controller, { row: -1 }),
                'ALT-RIGHT': this.resizeHandler.bind(this, { left: 0, right: 1 }),
                'ALT-LEFT': this.resizeHandler.bind(this, { left: -1, right: 0 }),
                'ALT-SHIFT-RIGHT': this.resizeHandler.bind(this, { left: 1, right: 0 }),
                'ALT-SHIFT-LEFT': this.resizeHandler.bind(this, { left: 0, right: -1 }),
                'F2': this._toggleMode.bind(this)
            });

        }

        FormEditorView.prototype = Object.create(View.prototype);
        fevp = FormEditorView.prototype;

        fevp.constructor = FormEditorView;

        /**
         * @method FormEditorView#getDraggable
         *
         * @param {String} id
         *
         * @returns {jQuery}
         */
        fevp.getDraggable = function (id) {
            return this._editor.find('[data-id=' + id + ']');
        };

        /**
         * @method FormEditorView#_emptySelection
         * @private
         */
        fevp._emptySelection = function () {
            this._controller.selected([]);
        };

        /**
         * Provides view specific destroy logic.
         *
         * Removes event listeners and elements.
         *
         * @method FormEditorView#destroy
         * @param {Object} cleanUp
         */
        fevp.destroy = function (cleanUp) {
            if (cleanUp.self) {
                this._$el.remove();
            }
        };

        /**
         * @method FormEditorView#update
         * @param {String} id
         */
        fevp.update = function (id) {
            var node = this._controller.node(id);
            this.getDraggable(id).replaceWith(this.buildNode(node));

            this.select(this._controller.selected());
        };

        /**
         * @method FormEditorView#select
         * @param {Array<String>} ids
         */
        fevp.select = function (ids) {
            this._editor.find('.selected').removeClass('selected');
            this._$el.toggleClass('w-selection', ids.length > 0);

            _.each(ids, function (id) {
                this.getDraggable(id).addClass('selected');
            }.bind(this));
        };

        /**
         * @method FormEditorView#active
         *
         * @param {String} id
         */
        fevp.active = function (id) {
            this._editor.find('.active').removeClass('active');

            this.getDraggable(id).addClass('active');
        };

        /**
         * @method FormEditorView#buildRow
         *
         * @param {Array<Object>} rowNodes
         * @param {Number} rowNum
         *
         * @returns {jQuery}
         */
        fevp.buildRow = function (rowNodes, rowNum) {
            var columns = this.groupColumns(rowNodes);
            var $row = $(rowTemplate({
                id: rowNum
            }));

            _.each(columns, function (column) {
                $row.append(this.buildColumn(column));
            }.bind(this));

            return $row;
        };

        /**
         * @method FormEditorView#buildColumn
         *
         * @param {Object} column
         *
         * @returns {jQuery}
         */
        fevp.buildColumn = function (column) {
            var $column;

            $column = $(colTemplate({
                column: column.id,
                size: column.width
            }));

            $column.append(_.map(_.sortBy(column.nodes, 'level'), function (node) {
                return this.buildNode(node);
            }.bind(this)));

            return $column;
        };

        /**
         * @method FormEditorView#buildNode
         *
         * @param {Object} node
         *
         * @returns {jQuery}
         */
        fevp.buildNode = function (node) {
            var children = _.map(this._controller.getChildren(node.id), this._controller.node.bind(this._controller));
            var element = this.wrap(node);
            var rows = [];
            var fieldSetContent = element.find('.fieldset-content').first();
            var canHoldChildren = fieldSetContent.length > 0;
            var nodesByRow = _.groupBy(children, 'row');
            var maxRow = _.chain(children)
                .pluck('row')
                .max()
                .value();

            if (canHoldChildren && maxRow === 0) {
                maxRow = 1;
            }

            for (var i = 0; i < maxRow; i++) {
                rows[i] = this.buildRow(nodesByRow[i + 1] || [], i + 1);
            }
            fieldSetContent.append(rows);

            return element;
        };

        /**
         * @method FormEditorView#groupColumns
         *
         * @param {Array<Object>} rowNodes
         *
         * @returns {Object}
         */
        fevp.groupColumns = function (rowNodes) {
            var columnsMap = {};
            var column;

            /**
             * @method FormEditorView~getColumn
             * @private
             *
             * @param {Object} node
             *
             * @returns {?Object}
             */
            function getColumn(node) {
                return _.find(columnsMap, function (column) {
                    return node.column >= column.id && node.column <= column.id + column.width - 1;
                });
            }

            /**
             * @method FormEditorView~columnExists
             * @private
             *
             * @param {Number} col
             *
             * @returns {Boolean}
             */
            function columnExists(col) {
                return !_.isUndefined(_.find(columnsMap, function (column) {
                    return col >= column.id && col <= column.id + column.width - 1;
                }));
            }

            _.each(rowNodes, function (node) {
                column = getColumn(node);

                //Column does not exist
                if (_.isUndefined(column)) {
                    //Seen column for first time
                    columnsMap[node.column] = {
                        id: node.column,
                        width: node.width,
                        level: 1,
                        nodes: [node]
                    };
                //Column does exist
                } else {
                    //Add to columns nodes
                    column.nodes.push(node);

                    if (node.width > column.width) {
                        column.width = node.width;
                    }

                    if (node.level > column.level) {
                        column.level = node.level;
                    }
                }
            });

            var sum = _.reduce(columnsMap, function (memo, column) {
                return column.width + memo;
            }, 0);

            var min = 1;

            var curColumn = 0;
            var width;
            while (sum !== 12 && curColumn <= 12) {
                if (!columnExists(curColumn + 1)) {
                    width = min;
                    columnsMap[curColumn + 1] = {
                        id: curColumn + 1,
                        width: width,
                        level: 1,
                        nodes: []
                    };

                    sum += width;
                    curColumn += width;
                }else {
                    width = columnsMap[curColumn + 1].width;
                    curColumn += width;
                }
            }

            return columnsMap;
        };

        /**
         * @method FormEditorView#guides
         *
         * @param {Boolean} show
         */
        fevp.guides = function (show) {
            this._$el.find('.grid-lines').toggleClass('hidden', !show);
        };

        /**
         * @method FormEditorView#formNode
         */
        fevp.add = function () {
            this.rerender();
        };

        /**
         * @method FormEditorView#move
         * @param {Object} formNode
         */
        fevp.move = function (formNode) {
            //Find current node in DOM and remove it
            this.getDraggable(formNode.id).remove();

            this.add(formNode);
        };

        /**
         * @method FormEditorView#collapse
         *
         * @param {Object} formNode
         */
        fevp.collapse = function (formNode) {
            this.getDraggable(formNode.id).replaceWith(this.buildNode(formNode));
        };

        /**
         * @method FormEditorNode#wrap
         *
         * @param {Object} node
         *
         * @returns {jQuery}
         */
        fevp.wrap = function (node) {
            var attributes = this._controller.getAttributesForNode(node.id);
            var wrapTemplate = handleTemplate({
                selected: attributes.selected,
                copied: !!attributes.copy,
                active: this._controller.active() === node.id,
                id: node.id,
                width: node.width,
                level: node.level
            });

            return $(wrapTemplate).append(this._options.viewFns.element(node));
        };

        /**
         * Constructs the view specific elements.
         *
         * @method FormEditorView#buildElement
         * @param {Object} options
         * @param {Object} el
         *
         * @returns {jQuery}
         */
        fevp.buildElement = function (options, el) {
            var $element;

            if (_.isNull(el)) {
                //Build element from scratch
                $element = $(canvasTemplate());
            }else {
                $element = el;
            }

            return $element;
        };

        /**
         * Disables the input and button controls of the date view.
         *
         * @method FormEditorView#disable
         * @param {Boolean} disable Disable or enable the view.
         */
        fevp._applyDisabled = function (disable) { //jshint ignore:line
            this._$el.toggleClass('disabled', disable);
            this._$el.attr('tabindex', disable ? -1 : 0);
        };

        /**
         * Rerenders the current state of the form. Goes through row by row
         * to draw each row as it exists. Each row will render subrows as necessary.
         *
         * @method FormEditorView#rerender
         */
        fevp.rerender = function () {
            var nodesByRow = this._controller.getNodesByRow(null);
            var allNodes = this._controller.nodes({ parentId: null });
            var maxRow = _.chain(allNodes)
                .pluck('row')
                .max()
                .value();

            this._editor.empty();

            var rows = [];
            for (var i = 0; i < maxRow; i++) {
                rows[i] = this.buildRow(nodesByRow[i + 1] || [], i + 1);
            }

            this._editor.append(rows);
        };

        /**
         * @method FormEditorView#resizeHandler
         *
         * @param {Object} move
         */
        fevp.resizeHandler = function (move) {
            var position;

            var active = this._controller.active();

            if (this.isEditable() && this._mode === modes.MOVE && active != null) {
                position = this._controller.position(this._controller.node(active));

                if (move.left) {
                    position.column += move.left;
                    position.width += -move.left;
                }

                if (move.right) {
                    position.width += move.right;
                }
                this._controller.move(active, {
                    column: clamp(position.column),
                    width: clamp(position.width)
                });
            }
        };

        /**
         * @method FormEditorView#moveHandler
         *
         * @param {Object} move
         */
        fevp.moveHandler = function (move) {
            var position;

            if (this.isEditable() && this._mode === modes.MOVE) {
                position = this._controller.position(this._controller.node(this._controller.active()));

                if (move.horizontal) {
                    position.column += move.horizontal;
                }

                if (move.vertical) {
                    position.row += move.vertical;
                }

                this._controller.move(this._controller.active(), {
                    column: position.column,
                    row: position.row
                });
            }
        };

        /**
         * @method FormEditorView#_toggleMode
         * @private
         */
        fevp._toggleMode = function () {
            if (this.isEditable()) {
                this._$el.removeClass('mode-' + this._mode);
                if (this._mode === modes.SELECTION) {
                    this._mode = modes.MOVE;
                } else if (this._mode === modes.MOVE) {
                    this._mode = modes.SELECTION;
                }
                this._$el.addClass('mode-' + this._mode);
            }
        };

        /**
         * @method FormEditorView#copied
         *
         * @param {Array<String>} nodes
         */
        fevp.copied = function (nodes) {
            this._editor.find('.copied').removeClass('copied');

            _.each(nodes, function (id) {
                this.getDraggable(id).addClass('copied');
            }.bind(this));
        };

        /**
         * @method FormEditorView#escapeHandler
         */
        fevp.escapeHandler = function () {
            if (this._controller.copy().length !== 0) {
                this._controller.copy([]);
            } else {
                this._emptySelection();
            }
        };

        /**
         * @method FormEditorView#deleteHandler
         */
        fevp.deleteHandler = function () {
            if (this.isEditable()) {
                this._controller.removeAll(this._controller.selected());
            }
        };

        return FormEditorView;
    }
);

define('oui/widgets/canvas/formEditorModel',[
    'underscore',
    'oui/widgets/tree/collection',
    'oui/utils/flexEqual'
], function (_, Collection, isEqual) {
    'use strict';

    var femp;

    /**
     * @class FormEditorModel
     * @private
     *
     * @param {Array} initData
     * @param {FormEditor~Config} options
     */
    function FormEditorModel(initData, options) {
        //Super call
        Collection.call(this, options, initData);

        this._options = options;
        this._guides = false;

    }

    FormEditorModel.prototype = Object.create(Collection.prototype);
    femp = FormEditorModel.prototype;

    femp.constructor = FormEditorModel;

    /**
     * @method FormEditorModel#getNodesByData
     *
     * @param {Object} data
     *
     * @returns {Array<Object>}
     */
    femp.getNodesByData = function (data) {
        return _.filter(this._nodes, function (formNode) {
            return isEqual(formNode, data);
        });
    };

    /**
     * @method FormEditorModel#widths
     *
     * @param {Object} data
     *
     * @returns {Array}
     */
    femp.widths = function (data) {
        var node = this.node(data);
        var rowNodes = this.getNodesByRow(node.parentId)[node.row];
        var sum;
        var minWidth = 1;
        var columns = _.range(1, 13);

        sum = _.reduce(rowNodes, function (memo, node) {
            return memo + node.width;
        }, 0);

        var allChildrenFullWidth = node.children && _.every(node.children, _.matchesProperty('width', 12));

        if (node.children && node.children.length > 0 && !allChildrenFullWidth) {
            //This must be a group and might have a minimum width
            var max = _.chain(node.children)
                .sortBy('column')
                .max(function (child) {
                    return child.column + child.width;
                })
                .value();

            minWidth = Math.round((max.column + max.width) * (node.width / 12));
        }

        _.each(rowNodes, function (item) {
            columns = _.difference(columns, _.range(item.column, item.column + item.width));
        });

        var maxWidth = node.width;
        var startColumn = node.column - 1;
        while (_.contains(columns, startColumn)) {
            maxWidth++;
            startColumn--;
        }

        startColumn = node.column + node.width;
        while (_.contains(columns, startColumn)) {
            maxWidth++;
            startColumn++;
        }

        return _.union(_.range(minWidth, node.width + 1), _.range(node.width, maxWidth + 1));
    };

    /**
     * @method FormEditorModel#parents
     *
     * @param {Object} data
     *
     * @returns {Array}
     */
    femp.parents = function (data) {
        var node = this.node(data);
        var parent = this.node(node.parentId);

        if (parent == null) {
            return [];
        }
        return _.union([parent], this.parents(parent.id));
    };

    /**
     * @method FormEditorModel#getNodesByCoordinates
     *
     * @param {Object} position
     *
     * @returns {Array}
     */
    femp.getNodesByCoordinates = function (position) {
        var nodes = this.getNodesByRow(position.parentId)[position.row];
        var columns = _.range(position.column, position.column + position.width);

        return _.filter(nodes, function (node) {
            return _.intersection(columns, _.range(node.column, node.column + node.width)).length > 0;
        });

    };

    /**
     * @method FormEditorModel#getNodesByRow
     *
     * @param {Number} parentId
     *
     * @returns {Array}
     */
    femp.getNodesByRow = function (parentId) {
        parentId = parentId || null;

        var nodes = _.chain(this.getChildren(parentId))
                    .map(function (node) {
                        return this.node(node);
                    }.bind(this))
                    .groupBy('row')
                    .value();

        _.each(nodes, function (value, key) {
            nodes[key] = _.sortBy(value, 'col');
        });

        return nodes;
    };

    /**
     * @method FormEditorModel#position
     *
     * @param {Object} node
     * @param {Object} [position]
     *
     * @returns {Object}
     */
    femp.position = function (node, position) {
        if (arguments.length === 1) {
            //Get position
            return {
                column: node.column,
                row: node.row,
                width: node.width,
                height: node.height,
                level: node.level,
                parentId: node.parentId
            };
        }

        if (!_.isEqual(position, this.position(node))) {
            node = _.extend(node, position);

            this.trigger('node-moved', node, position);
        }
    };

    /**
     * @method FormEditorModel#_getAdjacentNodes
     *
     * @param {String} id
     *
     * @returns {Object}
     */
    femp._getAdjacentNodes = function (id) {
        var currentPosition = this.position(this.node(id));

        return {
            top: this.getNodesByCoordinates(_.extend({}, currentPosition, { row: currentPosition.row - 1 })),
            below: this.getNodesByCoordinates(_.extend({}, currentPosition, { row: currentPosition.row + 1 })),
            left: this.getNodesByCoordinates(_.extend({}, currentPosition, { column: currentPosition.column - 1, width: 1 })),
            right: this.getNodesByCoordinates(_.extend({}, currentPosition, { column: currentPosition.column + currentPosition.width, width: 1 }))
        };
    };

    /**
     * @method FormEditorModel#selected
     *
     * @returns {Array}
     */
    /**
     * @method FormEditorModel#selected
     *
     * @param {Array<String>} ids
     */
    femp.selected = function (ids) {
        if (!arguments.length) {
            return this._attrs.selected;
        }

        if (!_.isEqual(this._attrs.selected, ids)) {
            this.attr('selected', ids);
        }
    };

    /**
     * @method FormEditorModel#guides
     *
     * @returns {Boolean} show
     */
    /**
     * @method FormEditorModel#guides
     *
     * @param {Boolean} show
     */
    femp.guides = function (show) {
        if (!arguments.length) {
            return this._guides;
        }

        if (this._guides !== show) {
            this._guides = show;
            this.trigger('guides-changed', this._guides);
        }
    };

    /**
     * @method FormEditorModel#toJSON
     *
     */
    femp.toJSON = function () {
        /**
         * @method FormEditorModel~convertNode
         * @private
         *
         * @param {String} id
         *
         * @returns {Object}
         */
        var convertNode = function (id) {
            var node = this.node(id);
            var clone = _.cloneDeep(node);
            var children = this.getChildren(id);

            if (children.length > 0) {
                clone.children = _.map(children, convertNode.bind(this));
            }

            return clone;
        };

        return {
            children: _.map(this.getRoots(), convertNode.bind(this))
        };
    };

    /**
     * @method FormEditorModel#getAllNodes
     *
     * @returns {Array}
     */
    femp.getAllNodes = function () {
        return _.cloneDeep(_.toArray(this._nodes));
    };

    /**
     * @method FormEditorModel#nodes
     *
     * @param {Array} nodes
     */
    femp.nodes = function (nodes) {
        this._nodes = _.indexBy(nodes, 'id');

        this.trigger('rerender');
    };

    /**
     * @method FormEditorModel#copy
     *
     * @returns {Array<String>}
     */
    /**
     * @method FormEditorModel#copy
     *
     * @param {Array<String>} nodes
     */
    femp.copy = function (nodes) {
        if (!arguments.length) {
            return this.getAttr('copy') || [];
        }

        this.attr('copy', nodes);
    };

    femp._updatePartial = _.noop;

    return FormEditorModel;
});

define('oui/widgets/canvas/formEditorController',[
    'jquery',
    'underscore'
], function ($, _) { //jshint maxstatements:false
    'use strict';

    var MAX_WIDTH = 13;

    var fecp;

    /**
     * @method FormEditorController~setupNewPosition
     * @private
     *
     * @param {Object} node
     * @param {Object} position
     *
     * @returns {Object}
     */
    function setupNewPosition(node, position) {
        // Given a maximum number of columns, if column + width exceeds then go backward.
        var width  = position.width || node.width;
        var column = position.column || node.column;

        if (column + width > MAX_WIDTH) {
            column -= column + width - MAX_WIDTH;
        }

        return {
            height: position.height || node.height,
            width: width,
            level: position.level || node.level,
            parentId: _.get(position, 'parentId', node.parentId),
            row: position.row || node.row,
            column: column
        };
    }

    /**
     * @method FormEditorController~findMaxHeight
     *
     * @param {Array} nodes
     *
     * @returns {Number}
     */
    function findMaxHeight(nodes) {
        return _.max(nodes, function (node) {
            if (node.children && node.children.length > 0) {
                return node.height;
            }

            return -1;
        }).height || 1;
    }

    /**
     * @class FormEditorController
     * @private
     *
     * @param {FormEditorModel} model
     * @param {FormEditor~Config} options
     */
    function FormEditorController(model, options) {

        this._model = model;
        this._options = options;

        this._commands = [];
        this._redo = [];
        this._track = false;

        this._model.on('loaded', function () {
            this._track = true;
        }.bind(this));

        this._model.on('add', function (id) {
            //can move up
            this._collapse(id);

            //move other down
            var node = this.node(id);
            this._moveOverlaps(this.overlap(this._model.position(node), node));

            //move others up
        }.bind(this));

        this._model.on('node-moved', function (node, position) {

            var nodes = this._model.getNodesByRow(node.parentId);

            this._moveOverlaps(this.overlap(position, node));

            for (var i = 1; i < _.max(_.keys(nodes)) + 1; i++) {
                var row = nodes[i];
                _.each(row, function (node) {
                    this._collapse(node);
                }.bind(this));
            }
        }.bind(this));
    }

    fecp = FormEditorController.prototype;

    /**
     * @method FormEditorController#load
     */
    fecp.load = function () {
        var that = this;
        var loadFn = this._options.storeFns.load;
        var mapperToFn = _.get(this._options.storeFns, 'mapper.to', _.identity);

        $.when(loadFn.call(this))
            .then(mapperToFn)
            .done(function (data) {
                _.each(data, function (value) {
                    //get the element that this data represents
                    that.add(value);
                });

                that._model.trigger('loaded');
            });
    };

    /**
     * @method FormEditorController#_collapse
     * @private
     *
     * @param {String} id
     */
    fecp._collapse = function (id) {
        var adjacentNodes;
        var pos;

        if (this._options.collapse) {
            adjacentNodes = this._model._getAdjacentNodes(id);
            pos = this._model.position(this.node(id));

            if (adjacentNodes.top.length === 0 && pos.row > 1) {
                this._model.position(this.node(id), _.extend({}, pos, { row: pos.row - 1 }));
            }
        }

    };

    /**
     * @method FormEditorController#_moveOverlaps
     * @private
     *
     * @param {Array<Object>} overlaps
     */
    fecp._moveOverlaps = function (overlaps) {
        if (overlaps.length > 0) {
            _.each(overlaps, function (overlap) {
                var position = this._model.position(overlap);

                this._model.position(overlap, _.extend({}, position, { row: position.row + 1 }));
            }.bind(this));
        }
    };

    /**
     * @method FormEditorController#node
     *
     * @param {Object} element
     *
     * @returns {?Object}
     */
    fecp.node = function (element) {
        return this._model.node(element);
    };

    /**
     * @method FormEditorController#nodes
     *
     * @param {Object} data
     *
     * @returns {Array<Object>}
     */
    fecp.nodes = function (data) {
        if (data == null) {
            return this._model.getAllNodes();
        }
        return this._model.getNodesByData(data);
    };

    /**
     * @method FormEditorController#widths
     *
     * @param {Object} data
     *
     * @returns {Array<Number>}
     */
    fecp.widths = function (data) {
        return this._model.widths(data);
    };

    /**
     * @method FormEditorController#parents
     *
     * @param {Object} data
     *
     * @returns {Array}
     */
    fecp.parents = function (data) {
        return this._model.parents(data);
    };

    /**
     * @method FormEditorController#repositionNodes
     *
     * @param {Object} node
     */
    fecp.repositionNodes = function (node) {
        //If the max height of the row changed then continue.
        var maxHeight;
        var rowNodes = this._model.getNodesByRow(node.parentId)[node.row];
        var position;

        maxHeight = findMaxHeight(rowNodes);

        if (node.height >= maxHeight) {
            _.each(rowNodes, function (node) {
                if (node.level > maxHeight) {
                    position = this._model.position(node);
                    position.level = 1;
                    position.row += (node.level - maxHeight);
                    this.move(node, position);
                }
            }.bind(this));
        }
    };

    /**
     * @method FormEditorController#move
     *
     * @param {Object} element
     * @param {Object} newPosition
     */
    fecp.move = function (element, newPosition) {
        var parentNode;
        var prevParentNode;

        //TODO: this function has an issue when moving an element within the same parent
        //Where a duplication of the element occurs.

        element = this.node(element);
        newPosition = setupNewPosition(element, newPosition);

        parentNode = this.node(newPosition.parentId);
        prevParentNode = this.node(element.parentId);

        //accept new position
        //model triggers event
        this._model.position(element, newPosition);

        if (prevParentNode != null) {
            //Remove from previous parent
            prevParentNode.children = _.reject(prevParentNode.children, function (node) {
                return node.id === element.id;
            });

            this._model.trigger('node-collapsed', prevParentNode);

            this.repositionNodes(prevParentNode);
        }

        if (parentNode !== null) {
            if (_.find(parentNode.children, _.matchesProperty('id', element.id)) == null) {
                //Move to new parent
                parentNode.children = parentNode.children.concat(element);
            }

            this.repositionNodes(parentNode);
        }

    };

    /**
     * @method FormEditorController#selected
     *
     * @returns {Array}
     */
    /**
     * @method FormEditorController#selected
     *
     * @param {Array} ids
     */
    fecp.selected = function (ids) {
        var that = this;
        if (!arguments.length) {
            return this._model.selected();
        }

        this._model.selected(_.map(ids, function (id) {
            return _.get(that.node(id), 'id', null);
        }));
    };

    /**
     * @method FormEditorController#active
     *
     * @returns {Number}
     */
    /**
     * @method FormEditorController#active
     *
     * @param {Number} id
     */
    fecp.active = function (id) {
        if (!arguments.length) {
            return this._model.active();
        }

        this._model.active(id);
    };

    /**
     * @method FormEditorController#add
     *
     * @param {Object} data
     * @param {Object} parentNode
     */
    fecp.add = function (data, parentNode) {
        var node;

        this.storeState();
        parentNode = this.node(parentNode || null);

        if (data.row === -1) {
            data.row = _.size(this.getNodesByRow(_.get(parentNode, 'id', null))) + 1;
        }
        //New data takes location precedence
        node = data;

        if (!_.isNull(parentNode)) {
            node.parentId = parentNode.id;
        }

        this._model.add(node);
    };

    /**
     * Pushes either the supplied state onto the command stack or the current state
     * from the model if a state is not supplied.
     *
     * @method FormEditorController#storeState
     * @param {Object} [state]
     */
    fecp.storeState = function (state) {
        if (this._track) {
            state = state || _.cloneDeep(this._model.getAllNodes());
            this._commands.push(state);

            if (!_.isEqual(this._redo[0], state)) {
                this._redo = [];
            }
        }
    };

    /**
     * @method FormEditorController#undo
     */
    fecp.undo = function () {
        var state = this._commands.pop();

        if (state != null) {
            this._redo.push(_.cloneDeep(this._model.getAllNodes()));
            this._model.nodes(state);
        }
    };

    /**
     * Sets the state of the model using the redo collection.
     *
     * @method FormEditorController#redo
     */
    fecp.redo = function () {
        var state = this._redo.pop();

        if (state != null) {
            this._commands.push(_.cloneDeep(this._model.getAllNodes()));
            this._model.nodes(state);
        }
    };

    /**
     * Sets the state of the model to the given state.
     *
     * @method FormEditorController#restore
     *
     * @param {Object} modelState
     */
    fecp.restore = function (modelState) {
        this._model.nodes(modelState);
    };

    /**
     * @method FormEditorController#getNodesByRow
     *
     * @param {Number} parentId
     * @returns {Array}
     */
    fecp.getNodesByRow = function (parentId) {
        return this._model.getNodesByRow(parentId);
    };

    /**
     * @method FormEditorController#overlap
     *
     * @param {Object} position
     * @param {Object} curNode
     *
     * @returns {Array}
     */
    fecp.overlap = function (position, curNode) {
        var nodes = this.getNodesByRow(position.parentId)[position.row];
        var startColumn = position.column;
        var endColumn = startColumn + position.width;
        var startOrder = position.level;
        var endOrder = startOrder + position.height - 1;

        var posColRange = _.range(startColumn, endColumn);

        return _.filter(nodes, function (node) {
            var nodeColRange = _.range(node.column, node.column + node.width);
            return (node.id !== curNode.id) &&
                   (_.intersection(nodeColRange, posColRange).length > 0) &&
                   (node.level >= startOrder && node.level <= endOrder);
        });
    };

    /**
     * @method FormEditorController#removeAll
     *
     * @param {Array<String>} data
     */
    fecp.removeAll = function (data) {
        this._model.removeAll(_.map(data, function (node) {
            return this.node(node).id;
        }.bind(this)));
    };

    /**
     * @method FormEditorController#remove
     *
     * @param {Object} data
     */
    fecp.remove = function (data) {
        var node = this.node(data);

        this._model.remove(node.id);
    };

    /**
     * @method FormEditorController#update
     *
     * @param {Object} node
     * @param {Object} data
     */
    fecp.update = function (node, data) {
        this._model.update(node, data);
    };

    /**
     * @method FormEditorController#toJSON
     *
     * @returns {Object}
     */
    fecp.toJSON = function () {
        return this._model.toJSON();
    };

    /**
     * @method FormEditorController#guides
     *
     * @returns {Boolean}
     */
    /**
     * @method FormEditorController#guides
     *
     * @param {Boolean} show
     */
    fecp.guides = function (show) {
        if (!arguments.length) {
            return this._model.guides();
        }

        this._model.guides(show);
    };

    /**
     * @method FormEditorController#copy
     *
     * @returns {Array<String>}
     */
    /**
     * @method FormEditorController#copy
     *
     * @param {Array<String>} nodes The array of node ids to mark as copied.
     */
    fecp.copy = function (nodes) {
        if (!arguments.length) {
            return this._model.copy();
        }

        this._model.copy(nodes);
    };

    /**
     * @method FormEditorController#clone
     *
     * @param {Array<String>} ids
     * @param {Object} modifiers
     * @param {Array<String>} modifiers.strip
     *
     * @returns {Array<Object>}
     */
    fecp.clone = function (ids, modifiers) {
        return _.map(ids, function (id) {
            var node = this.node(id);
            var cloned = _.omit(_.cloneDeep(node), modifiers.strip);

            cloned.children = this.clone(this.getChildren(id), modifiers);

            return cloned;
        }.bind(this));
    };

    /**
     * @method FormEditorController#paste
     *
     * @param {Object} modifiers
     */
    fecp.paste = function (modifiers) {
        var copy = this.clone(this.copy(), { strip: [this._options.recordKeys.id, this._options.recordKeys.parentId, '_ouiFields'] });

        if (this.active() != null && copy.length > 0) {
            var activePosition = this._model.position(this.node(this.active()));
            _.each(_.sortBy(copy, 'row'), function (node, idx) {
                var newPos = _.extend(node, _.omit(activePosition, 'width', 'height'));
                newPos.row = newPos.row + modifiers.row + idx;
                this.add(newPos);
            }.bind(this));
        }
    };

    /**
     * @method FormEditorController#getAttributesForNode
     *
     * @param {String} id
     *
     * @returns {Object}
     */
    fecp.getAttributesForNode = function (id) {
        return this._model.getAttributesForNode(id);
    };

    /**
     * @method FormEditorController#position
     *
     * @param {Object} node
     *
     * @returns {Object}
     */
    fecp.position = function (node) {
        return this._model.position(node);
    };

    /**
     * @method FormEditorController#getChildren
     *
     * @param {String} id
     *
     * @returns {Array}
     */
    fecp.getChildren = function (id) {
        return this._model.getChildren(id);
    };

    /**
     * @method FormEditorController#data
     *
     * @param {Array<Object>} nodes
     */
    fecp.data = function (nodes) {

        this._model.invalidate();

        _.each(nodes, function (node) {
            this._model.add(node);
        }.bind(this));
    };

    return FormEditorController;

});

define('oui/widgets/canvas/formEditor',[
    'underscore',
    'oui/base/widget',
    'oui/widgets/canvas/formEditorView',
    'oui/widgets/canvas/formEditorModel',
    'oui/widgets/canvas/formEditorController'
], function (_, Widget, View, Model, Controller) {
        'use strict';

        var fep; //Used to shorten/store FormEditor.prototype
        var __DEFAULTS__ = {
            fields: {
                row: 'row',
                column: 'column',
                width: 'width',
                height: 'height',
                level: 'level',
                children: 'children'
            },
            lazyLoad: false,
            redrawFields: null,
            editable: true,
            idPrefix: 'oui-',
            recordKeys: {
                id: 'id',
                parentId: 'parentId',
                children: 'children'
            },
            groupBy: [
                {
                    field: 'parentId',
                    asTree: true
                }
            ]
        };

        /**
         * @typedef FormEditor~Config
         *
         * @property {Object} [fields] The fields to read positioning out of the data
         * @property {String} [fields.row=row] The row field attribute
         * @property {String} [fields.column=column] The column field attribute
         * @property {String} [fields.width=width] The width field attribute
         * @property {String} [fields.height=height] The height field attribute
         * @property {String} [fields.level=level] The level field attribute
         * @property {String} [fields.children=children] The children field attribute
         * @property {Boolean} [redrawFields] Whether or not to redraw the fields on update
         * @property {Boolean} [editable=true] Whether or not the form is editable
         * @property {Object} [recordKeys]
         * @property {String} [recordKeys.id='id']
         *
         * @property {Object} [storeFns]
         * @property {Object} [storeFns.mapper]
         * @property {Function} [storeFns.mapper.to]
         * @property {Function} [storeFns.mapper.from]
         * @property {Function} [storeFns.load]
         * @property {Function} [storeFns.element]
         */

        /**
         * The Form Editor object.
         *
         * Displays a canvas that can hold and manipulate various UI elements
         * (can be images or HTML elements).
         *
         * Example construction:
         *
         *     var formEditor = new FormEditor(\/*...*\/)
         *
         * Explanation of options:
         *
         * * **mapper** is a collection of functions that will convert a JSON structure to another JSON structure
         *
         *  * 'to' goes from the server to the client side structure.
         *  * 'from' goes the opposite direction.
         *
         * * **load** is a function that is called when the editor needs to grab the data from the server.
         *
         * * **element** is a 'factory' function that accepts data and returns the HTML element that represents that node.
         *
         * @class FormEditor
         * @extends Widget
         * @since 1.3.0
         *
         *
         * @param {FormEditor~Config} [options] The options for the Form Editor. If not
         *  specified, all the defaults will be used.
         * @param {Object} [element] The JQuery element that this field should use.
         *  If not specified, the Form Editor will construct its own element.
         * @param {Object} [initData] The data to initialize the editor with.
         */
        function FormEditor(options, element, initData) {
            var model;

            //Super constructor
            Widget.call(this, options, element, initData);

            model = new Model(initData, this._options);
            this._controller = new Controller(model, this._options);
            this._view = new View(model, this._controller, this._element || null, this._options, this._settings);

            this._setupEvents(model);
        }

        FormEditor.prototype = Object.create(Widget.prototype);
        fep = FormEditor.prototype;

        fep.constructor = FormEditor;

        /**
         * Returns the name of the widget.
         *
         * @method name
         * @private
         */
        fep.name = function () {
            return 'form-editor';
        };

        /**
         * Getter for the element that the DateField uses.
         *
         * This element is the div surrounding the input and button elements.
         *
         * @method element
         * @returns {Object} element The jQuery wrapped element.
         */
        fep.element = function () {
            return this._view.element();
        };

        /**
         * Initialization method used after the element is on the DOM.
         *
         * @method init
         */
        fep._init = function () {
            this._controller.load();
        };

        /**
         * Getter/Setter for the disabled attribute of the Form Editor.
         *
         * @method disable
         * @since 1.3.0
         * @returns {Boolean} disabled When used as a getter it represents if the
         *  Form Editor is disabled or not.
         * @param {Boolean} disabled Whether or not to disable the Form Editor.
         */
        fep.disable = function (disabled) {
            if (!arguments.length) {
                return this._view.disable();
            }
            this._view.disable(disabled);
        };

        /**
         * Gets the default options for the Form Editor.
         *
         * @method defaults
         * @private
         * @returns {Object} The default options for the widget
         */
        fep.defaults = function () {
            return __DEFAULTS__;
        };

        /**
         * Getter/Setter for the selected nodes of the Form Editor.
         *
         * @method selected
         * @returns {Array} The array of selected elements in the Form Editor
         * @param {Array} data The array of selected elements to set in the
         *  Form Editor
         */
        fep.selected = function (data) {
            if (!arguments.length) {
                return _.map(this._controller.selected(), function (id) {
                    return this._controller.node(id);
                }.bind(this));
            }
            this._controller.selected(data);
        };

        /**
         * Getter/Setter for the data of the Form Editor.
         *
         * When used as the setter it will wipe-out existing data and replace.
         *
         * @method data
         * @returns {Array} The array of nodes in the Form Editor.
         * @param {Array} data The array of nodes to set in the Form Editor.
         */
        fep.data = function (data) {
            if (!arguments.length) {

                if (this._options.storeFns.mapper && this._options.storeFns.mapper.from) {
                    return this._options.storeFns.mapper.from(this._controller.toJSON());
                } else {
                    return this._controller.toJSON();
                }
            }

            this._controller.data(data);
        };

        /**
         * Resizes a node and repositions nodes around it.
         * Elements move down if they are overlapped.
         *
         * Size is an object with width, height properties.
         *
         * @method resize
         * @param {Object|Number} element The element to resize.
         * @param {Object} size The new size of the element.
         */
        fep.resize = function (element, size) { //jshint ignore:line
            this._controller.move.apply(this._controller, arguments);
        };

        /**
         * Move takes in identifying data for a node and sets the
         * position of the node.
         *
         * Position is an object with row, column, width, height and level
         * properties.
         *
         * @method move
         * @param {Object|Number} element The element to move
         * @param {Object} pos The new position of the element
         */
        fep.move = function (element, pos) { //jshint ignore:line
            //TODO: should also be able to work on data (as single argument?)
            this._controller.move.apply(this._controller, arguments);
        };

        /**
         * Removes a node from the form editor.
         *
         * @method remove
         * @param {Object|Number} data The element to remove
         */
        fep.remove = function (data) {
            this._controller.remove(data);
        };

        /**
         * Add a node to the form editor.
         *
         * @method add
         * @param {Object} data The data for the new node
         * @param {Object} parentNode The parent node to add the element to.
         */
        fep.add = function (data, parentNode) {
            this._controller.add(data, parentNode);
        };

        /**
         * Updates information regarding a node in the model.
         *
         * @method update
         * @param {Object|Number} id The identifying information of a node.
         * @param {Object} data The new data to set on the node
         */
        fep.update = function (id, data) {
            this._controller.update(id, data);
        };

        /**
         * Get the JSON data of the form editor.
         *
         * @method toJSON
         * @returns {Object} The json representation of the data.
         */
        fep.toJSON = function () {
            return this._controller.toJSON();
        };

        /**
         * @method FormEditor#guides
         *
         * @returns {Boolean}
         */
        /**
         * @method FormEditor#guides
         *
         * @param {Boolean} show
         */
        fep.guides = function (show) {
            if (!arguments.length) {
                return this._controller.guides();
            }

            this._controller.guides(show);
        };

        /**
         * @method FormEditor#node
         *
         * @param {Object} data
         *
         * @returns {?Object}
         */
        fep.node = function (data) {
            return this._controller.node(data);
        };

        /**
         * @method FormEditor#nodes
         *
         * @param {Object} data
         *
         * @returns {Array}
         */
        fep.nodes = function (data) {
            return this._controller.nodes(data);
        };

        /**
         * @method FormEditor#widths
         *
         * @param {Object} data
         *
         * @returns {Array}
         */
        fep.widths = function (data) {
            return this._controller.widths(data);
        };

        /**
         * @method FormEditor#parents
         *
         * @param {Object} data
         *
         * @returns {Array}
         */
        fep.parents = function (data) {
            return this._controller.parents(data);
        };

        /**
         * @method FormEditor#active
         *
         * @returns {String}
         */
        /**
         * @method FormEditor#active
         *
         * @param {String} id
         */
        fep.active = function (id) {
            if (!arguments.length) {
                return this._controller.active();
            }

            this._controller.active(id);
        };

        return FormEditor;
    }
);

define('oui/widgets/toolbar/toolbarModel',[
    'underscore',
    'oui/base/model'
], function (_, Model) {
    'use strict';

    /**
     * @class ToolbarModel
     * @private
     *
     * @param {Array<Object>} initData
     * @param {Toolbar#Config} options
     */
    function ToolbarModel(initData, options) {
        Model.call(this, initData, options);

        this._available = initData;
        this._attrs = {
            hidden: [],
            disabled: []
        };
    }

    ToolbarModel.prototype = Object.create(Model.prototype);

    _.extend(ToolbarModel.prototype, {
        /**
         * @method ToolbarModel#init
         * @returns {null}
         */
        init: function () {
            return null;
        },
        /**
         * @method ToolbarModel#available
         * @returns {Array<Object>}
         */
        available: function () {
            return this._available;
        },
        /**
         * @method ToolbarModel#name
         *
         * @returns {String}
         */
        name: function () {
            return 'toolbar';
        },
        /**
         * @method ToolbarModel#hidden
         *
         * @returns {Array}
         */
        /**
         * @method ToolbarModel#hidden
         *
         * @param {Array<Object>} actions
         */
        hidden: function (actions) {
            if (!arguments.length) {
                return this._attrs.hidden;
            }

            if (!_.isEqual(actions, this._attrs.hidden)) {
                var nowShown = _.difference(this._attrs.hidden, actions);
                this._attrs.hidden = actions;
                this.trigger('attribute-updated:visibility', nowShown, this._attrs.hidden);
            }
        },
        /**
         * @method ToolbarModel#disabled
         *
         * @returns {Array}
         */
        /**
         * @method ToolbarModel#disabled
         *
         * @param {Array<Object>} actions
         */
        disabled: function (actions) {
            if (!arguments.length) {
                return this._attrs.disabled;
            }

            if (!_.isEqual(actions, this._attrs.disabled)) {
                var nowEnabled = _.difference(this._attrs.disabled, actions);
                this._attrs.disabled = actions;
                this.trigger('attribute-updated:disabled', nowEnabled, this._attrs.disabled);
            }
        }
    });

    return ToolbarModel;
});

define('oui/widgets/toolbar/toolbarController',[
    'underscore'
], function (_) {
    'use strict';

    /**
     * @class ToolbarController
     * @private
     *
     * @param {ToolbarModel} model
     * @param {Toolbar#Config} options
     */
    function ToolbarController(model, options) {
        this._model = model;
        this._options = options;
    }

    ToolbarController.prototype = {
        /**
         * @method ToolbarController#selected
         *
         * @returns {Object}
         */
        /**
         * @method ToolbarController#selected
         *
         * @param {Object} action
         * @param {Subscribable~EventOptions} eventParams
         */
        selected: function (action, eventParams) {
            if (!arguments.length) {
                return this._model.data();
            }

            this._model.data(action, eventParams);
        },
        /**
         * @method ToolbarController#getAction
         *
         * @param {String} id
         * @returns {Object}
         */
        getAction: function (id) {
            return _.find(this.actions(), _.matchesProperty('id', id)) || null;
        },
        /**
         * @method ToolbarController#actions
         *
         * @returns {Array<Object>}
         */
        actions: function () {
            return this._model.available();
        },
        /**
         * @method ToolbarController#hidden
         *
         * @returns {Object}
         */
        /**
         * @method ToolbarController#hidden
         *
         * @param {Array} actions
         */
        hidden: function (actions) {
            if (!arguments.length) {
                return this._model.hidden();
            }

            this._model.hidden(actions);
        },
        /**
         * @method ToolbarController#disabled
         *
         * @returns {Object}
         */
        /**
         * @method ToolbarController#disabled
         *
         * @param {Array} actions
         */
        disabled: function (actions) {
            if (!arguments.length) {
                return this._model.disabled();
            }

            this._model.disabled(actions);
        },
        /**
         * @method ToolbarController#isSelectable
         *
         * @param {String} actionId
         *
         * @returns {Boolean}
         */
        isSelectable: function (actionId) {
            var action = _.find(this.actions(), _.matchesProperty('id', actionId));

            if (action && action.type) {
                return action.type !== 'popup' && action.type !== 'action';
            }

            return true;
        }
    };

    return ToolbarController;
});


define('template-104', ['underscore'], function (_) { return function(data) {
var __t, __p = '', __j = Array.prototype.join;
function print() { __p += __j.call(arguments, '') }
__p += '<div class="orcl-widget orcl-toolbar ' +
((__t = ( data.floating ? 'floating' : '' )) == null ? '' : __t) +
' ' +
((__t = ( data.vertical ? 'vertical' : '' )) == null ? '' : __t) +
'"> <div class="drag-handle"></div> ';
 if (data.actions) { ;
__p += ' ';
 for (var i = 0; i < data.actions.length; i++) { ;
__p += ' <button type="button" class="not-btn toolbar-action pgbu-icon ' +
((__t = ( data.actions[i].icon )) == null ? '' : __t) +
'" title="' +
((__t = ( data.actions[i].name )) == null ? '' : __t) +
'"></button> ';
 } ;
__p += ' ';
 } ;
__p += '</div>';
return __p
};});


define('template-105', ['underscore'], function (_) { return function(data) {
var __t, __p = '', __e = _.escape;
__p += '<button type="button" class="not-btn toolbar-action pgbu-icon ' +
((__t = ( data.icon )) == null ? '' : __t) +
'" data-draggable="' +
((__t = ( data.draggable )) == null ? '' : __t) +
'" aria-label="' +
__e( data.name ) +
'" title="' +
((__t = ( data.name )) == null ? '' : __t) +
'" data-id="' +
((__t = ( data.id )) == null ? '' : __t) +
'"></button>';
return __p
};});


define('template-106', ['underscore'], function (_) { return function(data) {
var __t, __p = '';
__p += '<div class="toolbar-group"> ' +
((__t = ( data.children )) == null ? '' : __t) +
'</div>';
return __p
};});

define('oui/widgets/toolbar/toolbarView',['require','underscore','jquery','oui/base/view','oui/utils/draggable','oui/widgets/colorPicker','template-104','template-105','template-106'],function (require) {
    'use strict';

    var _ = require('underscore');
    var $ = require('jquery');
    var View = require('oui/base/view');
    var draggable = require('oui/utils/draggable');
    var ColorPicker = require('oui/widgets/colorPicker');
    var toolbarTemplate = require('template-104');
    var actionTemplate = require('template-105');
    var groupTemplate = require('template-106');

    /**
     * @class ToolbarView
     * @private
     *
     * @param {ToolbarModel} model
     * @param {ToolbarController} controller
     * @param {jQuery} element
     * @param {Toolbar#Config} options
     * @param {Object} settings
     */
    function ToolbarView(model, controller, element, options, settings) {
        View.call(this, options, element, controller, settings);

        model.on({
            change: this._changeHandler.bind(this),
            'attribute-updated:disabled': this._disabledHandler.bind(this),
            'attribute-updated:visibility': this._visibilityHandler.bind(this)
        });

        this._$dragHandle = this._$el.find('.drag-handle');
        this._addListeners();
        this._addActions(this._controller.actions());
    }

    ToolbarView.prototype = Object.create(View.prototype);

    _.extend(ToolbarView.prototype, {
        /**
         * @method ToolbarView#_$getAction
         * @private
         *
         * @param {String} id
         * @returns {jQuery}
         */
        _$getAction: function (id) {
            return this._$el.find('[data-id=' + id + ']');
        },
        /**
         * @method ToolbarView#_$getActions
         * @private
         *
         * @param {String} ids
         * @returns {jQuery}
         */
        _$getActions: function (ids) {
            var filter = _.map(ids, function (id) {
                return '[data-id=' + id + ']';
            });
            return this._$el.find(filter.join(', '));
        },
        /**
         * @method ToolbarView#_selectHandler
         * @private
         *
         * @param {jQuery.Event} event
         */
        _selectHandler: function (event) {
            var actionId = $(event.target).attr('data-id');

            if (actionId && this._controller.isSelectable(actionId)) {
                this._controller.selected($(event.target).attr('data-id'));
            }
        },
        /**
         * @method ToolbarView#_changeHandler
         * @private
         *
         * @param {String} action
         */
        _changeHandler: function (action) {
            this._$el.find('.toolbar-action.selected').removeClass('selected');

            this._$el.find('[data-id=' + action + ']').addClass('selected');
        },
        /**
         * @method ToolbarView#_hideActions
         * @private
         *
         * @param {Array<String>} ids
         */
        _hideActions: function (ids) {
            this._$getActions(ids).addClass('hidden');
        },
        /**
         * @method ToolbarView#_showActions
         * @private
         *
         * @param {Array<String>} ids
         */
        _showActions: function (ids) {
            this._$getActions(ids).removeClass('hidden');
        },
        /**
         * @method ToolbarView#_visibilityHandler
         * @private
         *
         * @param {Array<String>} visibleIds
         * @param {Array<String>} hiddenIds
         */
        _visibilityHandler: function (visibleIds, hiddenIds) {
            this._showActions(visibleIds);
            this._hideActions(hiddenIds);
        },
        /**
         * @method ToolbarView#_disableActions
         * @private
         *
         * @param {Array<String>} ids
         */
        _disableActions: function (ids) {
            this._$getActions(ids).prop('disabled', true).addClass('disabled');
        },
        /**
         * @method ToolbarView#_enableActions
         * @private
         *
         * @param {Array<String>} ids
         */
        _enableActions: function (ids) {
            this._$getActions(ids).prop('disabled', false).removeClass('disabled');
        },
        /**
         * @method ToolbarView#_disabledHandler
         * @private
         *
         * @param {Array<String>} enabledIds
         * @param {Array<String>} disabledIds
         */
        _disabledHandler: function (enabledIds, disabledIds) {
            this._enableActions(enabledIds);
            this._disableActions(disabledIds);
        },
        /**
         * @method ToolbarView#_addListeners
         * @private
         */
        _addListeners: function () {
            this._$el.on('click', '.toolbar-action:not([data-draggable=true])', this._selectHandler.bind(this));

            this._$el.on('mousedown', '[data-draggable=true]', function (event) {
                event.stopPropagation();
                event.stopImmediatePropagation();

                var action = this._controller.getAction($(event.target).attr('data-id'));

                if (action) {
                    action.dragging(event, this._options.constrainTo);
                }
            }.bind(this));

            draggable(this._$el, {
                selector: '.drag-handle',
                $element: this._$el,
                triggerDistance: 10,
                //jscs:disable
                snapTo: $('.snap-container'),
                //jscs:enable
                snapDistance: 10,
                axis: 'both',
                constrainTo: this._options.constrainTo,
                throttle: {
                    move: 10
                }
            });
        },
        /**
         * @method ToolbarView#_addActions
         * @private
         *
         * @param {Array<Object>} actions
         */
        _addActions: function (actions) {
            this._$el.find('.toolbar-action').remove();

            this._$el.append(_.reduce(actions, function (memo, action) {
                if (action.children) {
                    memo += groupTemplate({
                        id: action.id,
                        name: action.name,
                        children: _.reduce(action.children, function (memo, action) {
                            return memo += actionTemplate({
                                id: action.id,
                                draggable: _.get(action, 'draggable', false),
                                name: action.name,
                                icon: action.icon
                            });
                        }, '')
                    });
                } else {
                    memo += actionTemplate({
                        id: action.id,
                        draggable: _.get(action, 'draggable', false),
                        name: action.name,
                        icon: action.icon
                    });
                }

                return memo;
            }, ''));

            _.map(actions, function (action) {
                if (_.has(action, 'options.type') && action.options.type === 'color') {
                    return new ColorPicker({
                        map: action.options.colors,
                        changeFontColor: action.options.changeFontColor,
                        columns: actions.length % 3 + 1,
                        rows: 2,
                        appendTo: this._$el
                    }, this._$getAction(action.id)).init().on('change', function (color) {
                        this.trigger('action-option', {
                            color: color,
                            name: action.name
                        });
                    }.bind(this));
                }
            }.bind(this));
        },
        /**
         * @method ToolbarView#buildElement
         *
         * @param {Toolbar#Config} opts
         *
         * @returns {jQuery}
         */
        buildElement: function (opts) {
            return $(toolbarTemplate({
                locale: opts.locale,
                floating: opts.floating,
                vertical: opts.vertical
            }));
        }
    });

    return ToolbarView;
});

define('oui/widgets/toolbar/toolbar',[
    'underscore',
    'oui/base/widget',
    'oui/widgets/toolbar/toolbarModel',
    'oui/widgets/toolbar/toolbarController',
    'oui/widgets/toolbar/toolbarView'
], function (_, Widget, Model, Controller, View) {
    'use strict';

    /**
     * @typedef Toolbar#Action
     *
     * @property {String} id The identifier to use for the action.
     * @property {String} name The name of the action (used as the title attribute within the Toolbar).
     * @property {String} icon The icon class to display within the toolbar for the action.
     * @property {String} type The type of action. Types are 'popup', 'tool'.
     * @property {Boolean} [draggable] Whether or not the action can be dragged off the toolbar.
     * @property {Function} [dragging] Callback when the action is starting its drag operation.
     * @property {Object} [options] The options to supply the component that makes the popup. Currently, only 'color' is a
     * supported popup type.
     */

    /**
     * @typedef Toolbar#Config
     *
     * @property {Boolean} [floating=false] Whether or not the toolbar can be dragged around the screen.
     * @property {Boolean} [vertical=true] Whether or not the toolbar should lay itself out vertically.
     * @property {jQuery} [constrainTo] The element to use for the boundary during drag operations.
     * @property {Locale} [locale={}] The specific strings that are used by the widget and will use the defaults set from
     * the {@link module:oui.lang|lang}.
     */
    var __DEFAULTS__ = {
        floating: false,
        vertical: true
    };

    /**
     * @class Toolbar
     * @extends Widget
     * @since 3.1.0
     *
     * @param {Toolbar#Config} [options] The options for use within the Toolbar.
     * @param {jQuery} [element] The element to use as the toolbar.
     * @param {Array<Toolbar#Action>} [actions] The actions within the Toolbar.
     */
    function Toolbar(options, element, actions) {
        var model;

        Widget.call(this, options, element, actions);

        model = new Model(this._initData, this._options);

        this._controller = new Controller(model, this._options);
        this._view = new View(model, this._controller, this._element, this._options, this._settings);

        this._setupEvents([model, this._view]);

        Object.defineProperty(this, '_data', {
            get: function () { //jscs:ignore jsDoc
                return this.selected();
            },
            set: function (action) { //jscs:ignore jsDoc
                this.selected(action);
            }
        });
    }

    Toolbar.prototype = Object.create(Widget.prototype);

    _.extend(Toolbar.prototype, {
        name: _.constant('toolbar'),
        defaults: _.constant(__DEFAULTS__),
        /**
         * Hides an action in the toolbar.
         *
         * @method Toolbar#hideAction
         * @param {String} action The action to hide.
         */
        hideAction: function (action) {
            this._controller.hidden(_.union(this._controller.hidden(), [action]));
        },
        /**
         * Show an action in the toolbar.
         *
         * @method Toolbar#showAction
         * @param {String} action The action to show.
         */
        showAction: function (action) {
            this._controller.hidden(_.without(this._controller.hidden(), action));
        },
        /**
         * Disable an action in the toolbar.
         *
         * @method Toolbar#disableAction
         * @param {String} action The action to disable.
         */
        disableAction: function (action) {
            this._controller.disabled(_.union(this._controller.disabled(), [action]));
        },
        /**
         * Enable an action in the toolbar.
         *
         * @method Toolbar#enableAction
         * @param {String} action The action to enable.
         */
        enableAction: function (action) {
            this._controller.disabled(_.without(this._controller.disabled(), action));
        },
        /**
         * Get selected action.
         *
         * @method Toolbar#selected
         * @returns {Object} The action object selected.
         */
        /**
         * Select an action using the actions id.
         *
         * @method Toolbar#selected
         * @param {String} action The id of the action to select.
         */
        selected: function (action) {
            if (!arguments.length) {
                return this._controller.getAction(this._controller.selected());
            }

            this._controller.selected(action);
        },
        /**
         * Get selected action.
         *
         * @method Toolbar#value
         * @returns {Object} The action object selected.
         */
        /**
         * Select an action using the actions id.
         *
         * @method Toolbar#value
         * @param {String} action The id of the action to select.
         */
        value: function (action) {
            if (!arguments.length) {
                return this.selected();
            }

            this._controller.selected(action, { silent: true });
        }
    });

    return Toolbar;
});

define('oui/widgets/contextMenu/contextMenuModel',[
    'underscore',
    'oui/base/model'
], function (_, Model) {
    'use strict';

    /**
     * @class ContextMenuModel
     * @extends Model
     * @private
     *
     * @param {Array<ContextMenu#Action>?} initData
     * @param {ContextMenu#Config} options
     */
    function ContextMenuModel(initData, options) {
        Model.call(this, initData, options);

        this._available = [];

        this.addAll(initData);
        this._attrs = {
            hidden: [],
            disabled: [],
            active: null
        };
    }

    ContextMenuModel.prototype = Object.create(Model.prototype);

    _.extend(ContextMenuModel.prototype, {
        /**
         * @method ContextMenuModel#init
         * @returns {null}
         */
        init: function () {
            return null;
        },
        /**
         * @method ContextMenuModel#available
         * @returns {Array<ContextMenu#Actions>}
         */
        available: function () {
            return this._available;
        },
        /**
         * @method ContextMenuModel#addAll
         *
         * @param {Array<Object>} actions
         */
        addAll: function (actions) {
            _.each(actions, this.add.bind(this));
        },
        /**
         * @method ContextMenuModel#add
         *
         * @param {Object} action
         */
        add: function (action) {

            if (_.isObject(action) && action.id == null) {
                action.id = _.uniqueId();
            }
            this._available.push(action);
        },
        /**
         * @method ContextMenuModel#action
         *
         * @param {String} actionId
         *
         * @returns {Object}
         */
        action: function (actionId) {
            return _.find(this._available, _.matchesProperty('id', actionId));
        },
        /**
         * @method ContextMenuModel#name
         * @returns {String}
         */
        name: _.constant('context-menu'),
        /**
         * @method ContextMenuModel#active
         * @returns {Object}
         */
        /**
         * @method ContextMenuModel#active
         * @param {String} active
         */
        active: function (active) {
            if (!arguments.length) {
                return this._attrs.active;
            }

            if (active !== this._attrs.active) {
                this._attrs.active = active;
                this.trigger('attribute-updated:active', active);
            }
        },
        /**
         * @method ContextMenuModel#hidden
         *
         * @returns {Array<String>}
         */
        /**
         * @method ContextMenuModel#hidden
         *
         * @param {Array<String>} actions
         */
        hidden: function (actions) {
            if (!arguments.length) {
                return this._attrs.hidden;
            }

            if (!_.isEqual(actions, this._attrs.hidden)) {
                var nowShown = _.difference(this._attrs.hidden, actions);
                this._attrs.hidden = actions;
                this.trigger('attribute-updated:visibility', nowShown, this._attrs.hidden);
            }
        },
        /**
         * @method ContextMenuModel#disabled
         *
         * @returns {Array<String>}
         */
        /**
         * @method ContextMenuModel#disabled
         *
         * @param {Array<String>} actions
         */
        disabled: function (actions) {
            if (!arguments.length) {
                return this._attrs.disabled;
            }

            if (!_.isEqual(actions, this._attrs.disabled)) {
                var nowEnabled = _.difference(this._attrs.disabled, actions);
                this._attrs.disabled = actions;
                this.trigger('attribute-updated:disabled', nowEnabled, this._attrs.disabled);
            }
        }
    });

    return ContextMenuModel;
});

define('oui/widgets/contextMenu/contextMenuController',[
    'underscore',
    'oui/utils/math'
], function (_, math) {
    'use strict';

    /**
     * @class ContextMenuController
     * @private
     *
     * @param {ContextMenuModel} model
     * @param {ContextMenu#Config} options
     */
    function ContextMenuController(model, options) {
        this._model = model;
        this._options = options;
    }

    ContextMenuController.prototype = {
        /**
         * @method ContextMenuController#active
         * @returns {Object}
         */
        /**
         * @method ContextMenuController#active
         * @param {String} action
         * @param {Subscribable~EventOptions} eventParams
         */
        active: function (action, eventParams) {
            if (!arguments.length) {
                return this._model.active();
            }

            this._model.active(action, eventParams);
        },
        /**
         * @method ContextMenuView#_moveFocus
         * @private
         *
         * @param {Number} number The number to shift active by (negative for up, positive for down)
         */
        _moveFocus: function (number) {
            var items = _.chain(this.actions())
                .pluck('id')
                .compact()
                .difference(this.hidden(), this.disabled())
                .value();
            var idx = _.indexOf(items, this.active());

            if (idx >= 0 || number > 0) {
                idx += number;
            }

            this.active(items[math.mod(idx, items.length)]);
        },
        /**
         * @method ContextMenuView#_moveFocusUp
         * @private
         */
        moveFocusUp: function () {
            this._moveFocus(-1);
        },
        /**
         * @method ContextMenuView#_moveFocusDown
         * @private
         */
        moveFocusDown: function () {
            this._moveFocus(1);
        },
        /**
         * @method ContextMenuController#getAction
         * @param {String} id
         *
         * @returns {ContextMenu#Action}
         */
        getAction: function (id) {
            return this._model.action(id);
        },
        /**
         * @method ContextMenuController#actions
         *
         * @returns {Array<ContextMenu#Action>} actions
         */
        actions: function () {
            return this._model.available();
        },
        /**
         * @method ContextMenuController#hidden
         *
         * @returns {Array<String>}
         */
        /**
         * @method ContextMenuController#hidden
         *
         * @param {Array<String>} actions
         */
        hidden: function (actions) {
            if (!arguments.length) {
                return this._model.hidden();
            }

            this._model.hidden(actions);
        },
        /**
         * @method ContextMenuController#disabled
         *
         * @returns {Array<String>}
         */
        /**
         * @method ContextMenuController#disabled
         *
         * @param {Array<String>} actions
         */
        disabled: function (actions) {
            if (!arguments.length) {
                return this._model.disabled();
            }

            this._model.disabled(actions);
        },
        /**
         * @method ContextMenuController#isSelectable
         *
         * @param {String} actionId
         *
         * @returns {Boolean}
         */
        isSelectable: function (actionId) {
            var action = this.getAction(actionId);

            if (action && action.type) {
                return action.type !== 'popup';
            }

            return true;
        }
    };

    return ContextMenuController;
});


define('template-107', ['underscore'], function (_) { return function(data) {
var __t, __p = '';
__p += '<button aria-haspopup="true" class="orcl-widget orcl-context-menu btn not-btn"> <i class="pgbu-icon pgbu-icon-cog"></i></button>';
return __p
};});


define('template-108', ['underscore'], function (_) { return function(data) {
var __t, __p = '', __e = _.escape, __j = Array.prototype.join;
function print() { __p += __j.call(arguments, '') }
__p += '<ul role="menu" class="orcl-context-menu-menu dropdown-menu"> ';
 if (data.actions.length > 0) { ;
__p += ' ';
 for (var i = 0, length = data.actions.length; i < length; i++) { ;
__p += ' ';
 if (data.actions[i] === 'separator') { ;
__p += ' <li role="separator" class="menu-divider"></li> ';
 } else { ;
__p += ' <li role="menuitem" id="' +
((__t = ( data.actions[i].pageId )) == null ? '' : __t) +
'" data-id="' +
((__t = ( data.actions[i].id)) == null ? '' : __t) +
'"    class="item ' +
((__t = (data.actions[i].active)) == null ? '' : __t) +
'"    title="' +
__e(data.actions[i].title) +
'" data-trigger="' +
((__t = ( data.actions[i].trigger )) == null ? '' : __t) +
'" ><a>' +
((__t = (data.actions[i].text )) == null ? '' : __t) +
'</a></li> ';
 } ;
__p += ' ';
 } ;
__p += ' ';
 } ;
__p += '</ul>';
return __p
};});

define('oui/widgets/contextMenu/contextMenuView',['require','underscore','jquery','oui/base/view','oui/mixins/popup','template-107','template-108'],function (require) {
    'use strict';

    var _ = require('underscore');
    var $ = require('jquery');
    var View = require('oui/base/view');
    var popupMixin = require('oui/mixins/popup');
    var contextMenuTemplate = require('template-107');
    var menuTemplate = require('template-108');

    var dataAttribute = 'data-id';

    /**
     * @method ContextMenuView~selectorWrapper
     * @private
     *
     * @param {String} action
     *
     * @returns {String}
     */
    function selectorWrapper(action) {
        return '[' + dataAttribute + '="' + action + '"]';
    }

    /**
     * @method ContextMenuView~selectorMapper
     * @private
     *
     * @param {Array<String>} actions
     *
     * @returns {String}
     */
    function selectorMapper(actions) {
        return _.map(actions, selectorWrapper).join(', ');
    }

    /**
     * @class ContextMenuView
     * @extends View
     * @mixes popup
     * @private
     *
     * @param {ContextMenuModel} model
     * @param {ContextMenuController} controller
     * @param {jQuery} element
     * @param {ContextMenu#Config} options
     * @param {Object} settings
     */
    function ContextMenuView(model, controller, element, options, settings) {
        View.call(this, options, element, controller, settings);

        model.on({
            'attribute-updated:active': this.activeHandler.bind(this),
            'attribute-updated:disabled': this.disabledHandler.bind(this)
        });

        this._$popup = this.buildMenu(this._controller.actions());
        this._addListeners();
        this._addMenuListeners();
    }

    ContextMenuView.prototype = Object.create(View.prototype);

    _.extend(ContextMenuView.prototype, popupMixin, {
        /**
         * @method ContextMenuView#getIdentifyingElement
         *
         * @returns {jQuery}
         */
        getIdentifyingElement: function () {
            return this._$el;
        },
        /**
         * @method ContextMenu#destroy
         */
        destroy: function () {
            this._$popup.remove();
        },
        /**
         * @method ContextMenuView#_addListeners
         * @private
         */
        _addListeners: function () {
            this._$el.on('click', this.toggleMenu.bind(this));

            this._$el.on('blur', this.hidePopup.bind(this));

            this.addShortcuts({
                'ENTER': function () { //jscs:ignore jsDoc
                    if (this.isVisible()) {
                        this.triggerAction();
                    } else {
                        this.showPopup(this._$el);
                    }
                },
                'ESC': function () { //jscs:ignore jsDoc
                    if (this.isVisible()) {
                        this.hidePopup();
                    }
                },
                'UP': function () { //jscs:ignore jsDoc
                    if (this.isVisible()) {
                        this._controller.moveFocusUp();
                    }
                },
                'DOWN': function () { //jscs:ignore jsDoc
                    if (this.isVisible()) {
                        this._controller.moveFocusDown();
                    } else {
                        this.showPopup(this._$el);
                    }
                }
            });
        },
        /**
         * @method ContextMenuView#_addMenuListeners
         * @private
         */
        _addMenuListeners: function () {
            this._$popup.on('mousedown', '[data-id]', function (event) {
                event.preventDefault();
                this._controller.active($(event.currentTarget).attr(dataAttribute));
            }.bind(this));

            this._$popup.on('showing.popup', function () {
                var disabled = this._checkAction(this._options.isActionEnabled);
                var hidden = this._checkAction(this._options.isActionVisible);
                this._controller.disabled(disabled);
                this._controller.hidden(hidden);
            }.bind(this));

            this._$popup.on('shown.popup', function () {
                var activeId = this._controller.active();
                this._$el.attr('aria-owns', this._$popup.attr('id'));

                if (activeId != null) {
                    this._$el.attr('aria-activedescendant', this._$popup.find(selectorWrapper(activeId)).attr('id'));
                }
            }.bind(this));

            this._$popup.on('hidden.popup', function () {
                this._$el.removeAttr('aria-activedescendant');
            }.bind(this));

            this._$popup.on('click', '[data-trigger]', function (event) {
                this.triggerAction(event.currentTarget.getAttribute('data-trigger'));
            }.bind(this));
        },
        /**
         * Gets all actions and returns an array of ID's that fail the check function parameter
         *
         * @method ContextMenu#_checkAction
         * @private
         * @param {Function} checkFn
         * @returns {Array} an array of action ID's that failed checkFn
         */
        _checkAction: function (checkFn) {
            return _(this._controller.actions())
                .reject(checkFn)
                .compact()
                .pluck('id')
                .value();
        },
        /**
         * @method ContextMenuView#triggerAction
         * @param {String} trigger
         */
        triggerAction: function (trigger) {
            trigger = trigger || this._controller.getAction(this._controller.active()).trigger;

            this.trigger(trigger);
            this.hidePopup();
        },
        /**
         * @method ContextMenuView#buildMenu
         * @param {Array<ContextMenu#Action>} actions
         *
         * @returns {jQuery}
         */
        buildMenu: function (actions) {
            return $(menuTemplate({
                id: this.id() + '-menu',
                actions: _.map(actions, function (action) {
                    var id = action.id;
                    if (action === 'separator') {
                        return action;
                    }
                    return {
                        pageId: this.id() + '-' + id,
                        id: id,
                        active: '',
                        title: action.title,
                        text: action.title,
                        trigger: action.trigger
                    };
                }.bind(this))
            }));
        },
        /**
         * @method ContextMenuView#toggleMenu
         */
        toggleMenu: function () {
            if (this.isVisible()) {
                this.hidePopup();
            } else {
                this.showPopup(this._$el);
            }
        },
        /**
         * @method ContextMenuView#buildElement
         * @param {ContextMenu#Config} opts
         *
         * @returns {jQuery}
         */
        buildElement: function (opts) {
            return $(contextMenuTemplate({
                locale: opts.locale
            }));
        },
        /**
         * @method ContextMenuView#activeHandler
         * @param {String} id
         */
        activeHandler: function (id) {
            var newActive = this._$popup.find(selectorWrapper(id));
            this._$popup.find('.active').removeClass('active');

            newActive.addClass('active');
            this._$el.attr('aria-activedescendant', newActive.attr('id'));
        },
        /**
         * @method ContextMenuView#disabledHandler
         *
         * @param {Array<ContextMenu#Action>} enabledActions
         * @param {Array<ContextMenu#Action>} actions
         */
        disabledHandler: function (enabledActions, actions) {
            this._$popup.find('.disabled').removeClass('disabled hidden');

            this._$popup.find(selectorMapper(actions)).addClass('disabled hidden');
        }
    });

    return ContextMenuView;
});

define('oui/widgets/contextMenu/contextMenu',[
    'underscore',
    'oui/base/widget',
    'oui/widgets/contextMenu/contextMenuModel',
    'oui/widgets/contextMenu/contextMenuController',
    'oui/widgets/contextMenu/contextMenuView'
], function (_, Widget, Model, Controller, View) {
    'use strict';

    /**
     * @typedef {Object} ContextMenu#Action
     *
     * @property {String} id
     */

    /**
     * @typedef {Object} ContextMenu#Config
     *
     * @property {String|jQuery} [appendTo=body] Where to append the menu item when opened.
     * @property {Function} [isActionEnabled] Function callback to check if actions should be available
     * @property {Function} [isActionVisible] Function callback to check if actions should be visible
     */
    var __DEFAULTS__ = {
        appendTo: 'body'
    };

    /**
     * @class ContextMenu
     * @extends Widget
     *
     * @param {ContextMenu#Config} [options] The options for the context menu. If not specified, all the defaults will be
     * used.
     * @param {jQuery} [element] Currently, ignored and constructs its own element.
     * @param {Array<ContextMenu#Action>} [actions]
     */
    function ContextMenu(options, element, actions) {
        var model;

        Widget.call(this, options, element, actions);

        model = new Model(this._initData, this._options);

        this._controller = new Controller(model, this._options);
        this._view = new View(model, this._controller, this._element, this._options, this._settings);

        this._setupEvents([model, this._view]);

        //TODO: not implemented yet
        Object.defineProperty(this, '_data', {
            get: _.noop,
            set: _.noop
        });
    }

    ContextMenu.prototype = Object.create(Widget.prototype);

    _.extend(ContextMenu.prototype, {
        /**
         * Returns the name of the widget, 'context-menu'
         *
         * @method ContextMenu#name
         * @returns {String}
         */
        name: _.constant('context-menu'),
        /**
         * Returns the default configuration for the Context Menu
         *
         * @method ContextMenu#defaults
         * @returns {ContextMenu#Config}
         */
        defaults: _.constant(__DEFAULTS__),
        /**
         * Shows the pop-up menu from the Context Menu button
         *
         * @method ContextMenu#show
         */
        show: function () {
            this._view.showPopup(this._view.element());
        },
        /**
         * Hides the pop-up menu from the Context Menu button
         *
         * @method ContextMenu#hide
         */
        hide: function () {
            this._view.hidePopup();
        },
        /**
         * Gets an array of action ID's that are disabled
         *
         * @method ContextMenu#disabled
         *
         * @returns {Array}
         */
        /**
         * Takes an array of action ID's and disables them
         *
         * @method ContextMenu#disabled
         *
         * @param {Array} actions
         */
        disabled: function (actions) {
            if (!arguments.length) {
                return this._controller.disabled();
            }
            this._controller.disabled(actions);
        },
        /**
         * Gets an array of action ID's that are hidden
         *
         * @method ContextMenu#hidden
         *
         * @returns {Array}
         */
        /**
         * Takes an array of action ID's and hides them
         *
         * @method ContextMenu#hidden
         *
         * @param {Array} actions
         */
        hidden: function (actions) {
            if (!arguments.length) {
                return this._controller.hidden();
            }
            this._controller.hidden(actions);
        }
    });

    return ContextMenu;
});

define('oui/widgets/multiField/multiFieldModel',[
    'underscore',
    'oui/base/model'
], function (_, Model) {
    'use strict';

    /**
     * @class MultiFieldModel
     * @private
     *
     * @param {Array} initData
     * @param {MultiField#Config} options
     */
    function MultiFieldModel(initData, options) {
        Model.call(this, initData, options);

        this._rows = this._data;
        this._attrs = {
            active: null
        };

        Object.defineProperty(this, '_data', {
            set: function (newData) {// jscs:ignore jsDoc
                this._rows = newData;
            },
            get: function () {// jscs:ignore jsDoc
                return this._rows;
            }
        });
    }

    MultiFieldModel.prototype = _.extend(Object.create(Model.prototype), {
        /**
         * @method MultiFieldModel#init
         *
         * @param {Array} initData
         *
         * @returns {Array} initData
         */
        init: function (initData) {
            return initData || [];
        },
        /**
         * @method MultiFieldModel#_getId
         * @private
         *
         * @param {Object} rowConfig
         *
         * @returns {String} id
         */
        _getId: function (rowConfig) {
            var id = _.get(rowConfig, this._options.recordKeys.id);

            if (_.get(rowConfig, this._options.recordKeys.id) == null) {
                id = _.uniqueId('oui-');
                _.set(rowConfig, this._options.recordKeys.id, id);
            }

            return id;
        },
        /**
         * @method MultiFieldModel#findIndex
         *
         * @param {String} rowId
         *
         * @returns {Object} row
         */
        findIndex: function (rowId) {
            return _.findIndex(this._rows, function (row) {
                return row.id === rowId;
            });
        },
        /**
         * @method MultiFieldModel#addRow
         *
         * @param {Object} rowConfig
         * @param {Object} modifiers
         */
        addRow: function (rowConfig, modifiers) {
            if (this.canAdd()) {
                this._getId(rowConfig);
                var index = _.get(modifiers, 'at', this._data.length);

                if (modifiers && modifiers.at != null) {
                    this._data.splice(modifiers.at, 0, rowConfig);
                    this.setData(this._data);
                } else {
                    this.setData(this._data.concat(rowConfig));
                }

                this.trigger('add', rowConfig, index);
            }
        },
        /**
         * @method MultiFieldModel#updateRow
         *
         * @param {String} id
         * @param {Object} updatedConfig
         */
        updateRow: function (id, updatedConfig) {
            this.setData(_.map(this._data, function (row) {
                if (row.id === id) {
                    row = _.extend({}, row, updatedConfig);
                }

                return row;
            }));

            this.trigger('update', this.row(id), updatedConfig);
        },
        /**
         * @method MultiFieldModel#removeRow
         *
         * @param {String} id
         */
        removeRow: function (id) {
            if (this.canRemove()) {
                var index = this.findIndex(id);

                this.setData(_.reject(this._data, function (row) {
                    return id === row.id;
                }));

                this.trigger('remove', id, index);
            }
        },
        /**
         * @method MultiFieldModel#row
         *
         * @param {String} id
         *
         * @returns {Object} row
         */
        row: function (id) {
            return _.find(this._rows, function (row) {
                return row.id === id;
            });
        },
        /**
         * @method MultiFieldModel#active
         *
         * @returns {Array} active
         */
        /**
         * @method MultiFieldModel#active
         *
         * @param {String} id
         */
        active: function (id) {
            if (!arguments.length) {
                return this._attrs.active;
            }

            if (this._attrs.active !== id) {
                this._attrs.active = id;
                this.trigger('active', id);
            }
        },
        /**
         * @method MultiFieldModel#canAdd
         *
         * @returns {Boolean}
         */
        canAdd: function () {
            if (this._options.max != null) {
                return this._rows.length < this._options.max;
            }

            return true;
        },
        /**
         * @method MultiFieldModel#canRemove
         *
         * @returns {Boolean}
         */
        canRemove: function () {
            return this._rows.length > 1;
        }
    });

    return MultiFieldModel;
});

define('oui/widgets/multiField/multiFieldController',[
    'underscore'
], function (_) {
    'use strict';

    /**
     * @class MultiFieldController
     * @private
     *
     * @param {MultiField#Config} options
     * @param {MultiFieldModel} model
     */
    function MultiFieldController(options, model) {
        this._model = model;
        this._options = options;
    }

    MultiFieldController.prototype = {
        /**
         * @method MultiFieldController#isActionEnabled
         *
         * @param {String} action
         *
         * @returns {Boolean}
         */
        isActionEnabled: function (action) {
            switch (action) {
                case 'remove':
                    return this._model.canRemove();
                case 'add-below':
                    return this.canAdd();
                default:
                    return true;
            }
        },
        /**
         * @method MultiFieldController#canAdd
         *
         * @returns {Boolean}
         */
        canAdd: function () {
            return this._model.canAdd();
        },
        /**
         * @method MultiFieldController#_getDefaultRowConfig
         * @private
         *
         * @returns {Object} defaultRowConfig
         */
        _getDefaultRowConfig: function () {
            if (_.isFunction(this._options.defaultRow)) {
                return this._options.defaultRow();
            } else {
                return _.clone(this._options.defaultRow);
            }
        },
        /**
         * @method MultiFieldController#addRow
         *
         * @param {Object} rowConfig
         * @param {Object} modifiers
         */
        addRow: function (rowConfig, modifiers) {
            if (this.canAdd()) {
                this._model.addRow(rowConfig, modifiers);
            }
        },
        /**
         * @method MultiFieldController#newRow
         */
        newRow: function () {
            this.addRow(this._getDefaultRowConfig());
        },
        /**
         * @method MultiFieldController#duplicate
         *
         * @param {String} rowId
         */
        duplicate: function (rowId) {
            var index = this._model.findIndex(rowId);
            var data = _.clone(this._model.getRow(rowId));

            delete data.id;

            this.addRow(data, {
                at: index + 1
            });
        },
        /**
         * @method MultiFieldController#newRowAbove
         *
         * @param {String} rowId
         */
        newRowAbove: function (rowId) {
            var index = this._model.findIndex(rowId);

            this.addRow(this._getDefaultRowConfig(), {
                at: index
            });
        },
        /**
         * @method MultiFieldController#newRowBelow
         *
         * @param {String} rowId
         */
        newRowBelow: function (rowId) {
            var index = this._model.findIndex(rowId);
            this.addRow(this._getDefaultRowConfig(), {
                at: index + 1
            });
        },
        /**
         * @method MultiFieldController#updateRow
         *
         * @param {String} id
         * @param {Object} updateConfig
         */
        updateRow: function (id, updateConfig) {
            this._model.updateRow(id, updateConfig);
        },
        /**
         * @method MultiFieldController#removeRow
         *
         * @param {String} id
         */
        removeRow: function (id) {
            this._model.removeRow(id);
        },
        /**
         * @method MultiFieldController#data
         *
         * @returns {Array} data
         */
        /**
         * @method MultiFieldController#data
         *
         * @param {Array} value
         */
        data: function (value) {
            if (!arguments.length) {
                return this._model.data();
            }

            this._model.data(value);
        },
        /**
         * @method MultiFieldController#getValue
         *
         * @param {String} rowId
         * @param {String} fieldName
         *
         * @returns {*} value
         */
        getValue: function (rowId, fieldName) {
            return _.get(this._model.row(rowId), fieldName);
        },
        /**
         * @method MultiFieldController#row
         *
         * @param {String} rowId
         *
         * @returns {Object} row
         */
        row: function (rowId) {
            return this._model.row(rowId);
        },
        /**
         * @method MultiFieldController#active
         *
         * @returns {Array} active
         */
        /**
         * @method MultiFieldController#active
         *
         * @param {String} rowId
         */
        active: function (rowId) {
            if (!arguments.length) {
                return this._model.active();
            }

            this._model.active(rowId);
        }
    };

    return MultiFieldController;
});


define('template-109', ['underscore'], function (_) { return function(data) {
var __t, __p = '', __j = Array.prototype.join;
function print() { __p += __j.call(arguments, '') }
__p += '<div class="orcl-widget orcl-multi-field"> ';
 if (data.columns != null && data.columns.length > 0) { ;
__p += ' <div class="columns"> ';
 if (data.editable) { ;
__p += ' <div class="column action-column"></div> ';
 } ;
__p += ' ';
 for (var i=0; i < data.columns.length; i++) { ;
__p += ' <div class="column" id="' +
((__t = ( data.id + '-' + data.columns[i].id )) == null ? '' : __t) +
'"> <label class="control-label">' +
((__t = ( data.columns[i].display )) == null ? '' : __t) +
'</label> </div> ';
 } ;
__p += ' </div> ';
 } ;
__p += ' <div class="rows"></div></div>';
return __p
};});


define('template-110', ['underscore'], function (_) { return function(data) {
var __t, __p = '', __j = Array.prototype.join;
function print() { __p += __j.call(arguments, '') }
__p += '<div class="multi-field-row" data-id="' +
((__t = ( data.rowId )) == null ? '' : __t) +
'"> ';
 if (data.columns != null && data.columns.length > 0) { ;
__p += ' ';
 if (data.editable) { ;
__p += ' <div class="column action-column row-action" data-row="' +
((__t = ( data.rowId )) == null ? '' : __t) +
'"> </div> ';
 } ;
__p += ' ';
 for (var i=0; i < data.columns.length; i++) { ;
__p += ' <div class="column"> ';
 if (data.types[data.columns[i].id] === 'checkbox') { ;
__p += ' <label class="checkbox" data-field="' +
((__t = ( data.columns[i].id )) == null ? '' : __t) +
'"> <input type="checkbox" data-oui-type="' +
((__t = ( data.types[data.columns[i].id] )) == null ? '' : __t) +
'" data-id="' +
((__t = ( data.columns[i].id +                ',' + data.rowId )) == null ? '' : __t) +
'" data-row="' +
((__t = ( data.rowId )) == null ? '' : __t) +
'" aria-describedby="' +
((__t = ( data.id + '-' + data.columns[i].id )) == null ? '' : __t) +
'"> </label> ';
 } else { ;
__p += ' <input type="text" data-oui-type="' +
((__t = ( data.types[data.columns[i].id] )) == null ? '' : __t) +
'" data-id="' +
((__t = ( data.columns[i].id +                ',' + data.rowId )) == null ? '' : __t) +
'" data-row="' +
((__t = ( data.rowId )) == null ? '' : __t) +
'" data-field="' +
((__t = ( data.columns[i].id )) == null ? '' : __t) +
'" aria-describedby="' +
((__t = ( data.id + '-' + data.columns[i].id )) == null ? '' : __t) +
'"> ';
 } ;
__p += ' </div> ';
 } ;
__p += ' ';
 } ;
__p += '</div>';
return __p
};});

define('oui/widgets/multiField/multiFieldView',['require','jquery','underscore','oui/widgets/form/fieldFactory','oui/base/view','oui/widgets/contextMenu/contextMenu','oui/utils/dom','oui/utils/focus','template-109','template-110'],function (require) {
    'use strict';

    var $ = require('jquery');
    var _ = require('underscore');
    var fieldFactory = require('oui/widgets/form/fieldFactory');
    var View = require('oui/base/view');
    var ContextMenu = require('oui/widgets/contextMenu/contextMenu');
    var dom = require('oui/utils/dom');
    var focusUtils = require('oui/utils/focus');
    var multiFieldTemplate = require('template-109');
    var rowTemplate = require('template-110');

    /**
     * @class MultiFieldView
     * @private
     *
     * @param {MultiFieldModel} model
     * @param {MultiFieldController} controller
     * @param {jQuery} element
     * @param {MultiField#Config} options
     * @param {Object} settings
     */
    function MultiFieldView(model, controller, element, options, settings) {
        View.call(this, options, element, controller, settings);

        this._fields = [];
        this._focusField = null;

        this._$el.on('oui-change', this._changeHandler.bind(this));
        this._$el.on('oui-error', this._errorHandler.bind(this));

        this._$el.on('focusin', '.multi-field-row', function (event) {
            this._focusRow($(event.currentTarget).attr('data-id'));
            this._controller.active($(event.currentTarget).attr('data-id'));
        }.bind(this));

        this._$el.on('focusin', '[data-field]', function (event) {
            this._focusField = $(event.currentTarget).attr('data-field');
        }.bind(this));

        this._$el.on({
            'oui-add-below': function (event) {
                this._controller.newRowBelow($(event.target).closest('[data-row]').attr('data-row'));
            }.bind(this),
            'oui-add-above': function (event) {
                this._controller.newRowAbove($(event.target).closest('[data-row]').attr('data-row'));
            }.bind(this),
            'oui-remove-row': function (event) {
                this._controller.removeRow($(event.target).closest('[data-row]').attr('data-row'));
            }.bind(this)
        });

        this.addShortcuts({
            'alt-DOWN': function (event) {
                event.preventDefault();
                event.stopPropagation();
                this._controller.newRowBelow($(event.target).closest('[data-row]').attr('data-row'));
            }.bind(this),
            'alt-UP': function (event) {
                event.preventDefault();
                event.stopPropagation();
                this._controller.newRowAbove($(event.target).closest('[data-row]').attr('data-row'));
            }.bind(this),
            'ctrl-shift-D': function (event) {
                event.preventDefault();
                event.stopPropagation();
                this._controller.duplicate($(event.target).closest('[data-row]').attr('data-row'));
            }.bind(this),
            'ctrl-DELETE': function (event) {
                this._controller.removeRow($(event.target).closest('[data-row]').attr('data-row'));
            }.bind(this),
            'shift-F10 CONTEXTMENU': function (event) {
                var contextMenu = dom.get($(event.target).closest('.multi-field-row').find('.orcl-context-menu'));
                contextMenu.element().focus();
                contextMenu.show();
            }.bind(this),
            'TAB': {
                down: true,
                prevention: _.constant(false),
                fn: function (event) {
                    var focusableElements = focusUtils.findFocusable(this._$el);
                    if (this._options.newRowOnTab && $(event.target).is(_.last(focusableElements))) {
                        this._focusField = null;
                        this._controller.newRowBelow($(event.target).closest('[data-row]').attr('data-row'));
                        event.preventDefault();
                    }
                }.bind(this)
            }
        });

        model.on({
            'view.add': this._addHandler.bind(this),
            'view.update': this._updateHandler.bind(this),
            'view.remove': this._removeHandler.bind(this)
        });

        this._redrawHandler();

        this._$el.on('oui-destroy', function (event) {
            delete this._fields[$(event.target).attr('data-id')];
        }.bind(this));
    }

    MultiFieldView.prototype = _.extend(Object.create(View.prototype), {
        /**
         * @method MultiFieldView#buildElement
         *
         * @param {MultiField#Config} options
         * @param {jQuery} element
         *
         * @returns {jQuery} element
         */
        buildElement: function (options, element) {
            var $element;

            //TODO: handle given an element
            if (element == null) {
                $element = $(multiFieldTemplate({
                    locale: options.locale,
                    columns: options.columns,
                    editable: options.editable,
                    collapsible: options.collapsible
                }));
            }

            return $element;
        },
        /**
         * @method MultifieldView#getIdentifyingElement
         *
         * @returns {jQuery}
         */
        getIdentifyingElement: function () {
            return this._$el;
        },
        /**
         * @method MultiFieldModel#_addField
         * @private
         *
         * @param {Object} config
         */
        _addField: function (config) {
            var existingField = _.find(this._fields, function (fieldConfig) {
                return fieldConfig.field === config.field && fieldConfig.row === config.row;
            });

            if (!existingField) {
                this._fields.push(config);
                config.control.init();
            } else if (existingField.control.add) {
                config.control.add(config.control.init());
            }
        },
        /**
         * @method MultiFieldModel#initializeField
         *
         * @param {jQuery} fieldEl
         * @param {Object} modifiers
         */
        initializeField: function (fieldEl, modifiers) {
            var $field = $(fieldEl);
            var field = $(fieldEl).attr('data-field');
            var row = $(fieldEl).closest('.multi-field-row').attr('data-id');
            var control;
            var type;

            var column = _.find(this._options.columns, function (column) {
                return column.id === field;
            });

            type = this._getType(column, this._controller.row(row));

            var definition = _.merge({
                //TODO: take this out
                options: {
                    fluid: true
                }
            }, this._getOptions(field, this._controller.row(row)));
            control = fieldFactory.createField(definition, type, $field);

            control.element().find('[data-field]').attr('aria-labelledby', this.id() + '-' + column.id);

            if (_.get(modifiers, 'replace')) {
                this._replace({ row: row, field: field }, control);
                control.value(this._controller.row(row)[field]);
            } else {
                if (control) {
                    this._addField({
                        row: row,
                        field: field,
                        type: type,
                        control: control
                    });
                } else {
                    this._addCustomField({
                        row: row,
                        field: field,
                        type: type,
                        control: control
                    });
                }
            }

            return control;
        },
        /**
         * @method MultiFieldModel#init
         */
        init: function () {
            _.each(this._$el.find('.columns .column:not(.action-column)'), function (column, idx) {
                $(column).attr('id', this.id() + '-' + this._options.columns[idx].attr);
            }.bind(this));
        },
        /**
         * @method MultiFieldModel#_changeHandler
         * @private
         *
         * @param {Object} event
         * @param {Widget} widget
         */
        _changeHandler: function (event, widget) {
            if (widget.name() !== 'multi-field') {
                var field = widget.element().find('[data-field]').attr('data-field');
                var rowId = $(event.target).closest('.multi-field-row').attr('data-id');
                var updateObj = {};
                updateObj[field] = widget.value();
                this._controller.updateRow(rowId, updateObj);
            }
        },
        /**
         * @method MultiFieldModel#errorHandler
         * @private
         */
        _errorHandler: function () {

        },
        /**
         * @method MultiFieldModel#_getOptions
         * @private
         *
         * @param {String} field
         * @param {Object} row
         *
         * @returns {Object}
         */
        _getOptions: function (field, row) {
            var column = _.find(this._options.columns, 'id', field);

            if (_.isFunction(column.options)) {
                return column.options(row);
            } else {
                return column.options;
            }

        },
        /**
         * @method MultiFieldModel#_getType
         * @private
         *
         * @param {Object} column
         * @param {Object} row
         *
         * @returns {String} column type
         */
        _getType: function (column, row) {
            if (_.isFunction(column.type)) {
                return column.type(row);
            } else {
                return column.type;
            }
        },
        /**
         * @method MultiFieldModel#_findControl
         * @private
         *
         * @param {String} rowId
         * @param {String} fieldName
         *
         * @returns {Widget|undefined} control
         */
        _findControl: function (rowId, fieldName) {
            if (fieldName == null) {
                fieldName = _.first(this._options.columns).attr;
            }

            var found = _.find(this._fields, function (config) {
                return fieldName === config.field && rowId === config.row;
            });

            if (found) {
                return found.control;
            }
        },
        /**
         * @method MultFieldView#_getAttrs
         * @private
         *
         * @param {String} field
         * @param {Object} data
         *
         * @returns {Object?}
         */
        _getAttrs: function (field, data) {
            var attrsFn = _.get(this._options, 'storeFns.attrs');

            if (attrsFn) {
                return attrsFn(field, data);
            }
        },
        /**
         * @method MultiFieldModel#_addHandler
         * @private
         *
         * @param {Object} row
         * @param {Number} index
         */
        _addHandler: function (row, index) {
            var $row = $(rowTemplate({
                id: this.id(),
                rowId: row.id,
                columns: this._options.columns,
                editable: this._options.editable,
                types: _.transform(this._options.columns, function (memo, column) {
                    memo[column.id] = this._getType(column, row);
                    return memo;
                }.bind(this), {}),
                row: row
            }));

            if (index === 0) {
                this._$el.find('.rows').prepend($row);
            } else {
                this._$el.find('.rows .multi-field-row:eq(' + (index - 1) + ')').after($row);
            }

            var contextMenu = new ContextMenu({
                isActionEnabled: function (action) {
                    return this._controller.isActionEnabled(action.id);
                }.bind(this),
                appendTo: this._options.appendTo
            }, [
                {
                    id: 'add-below',
                    trigger: 'add-below',
                    title: this._options.locale.add
                },
                {
                    id: 'remove',
                    trigger: 'remove-row',
                    title: this._options.locale.delete
                }
            ]);
            $row.find('.row-action').append(contextMenu.element());
            contextMenu.init();

            _.each($row.find('[data-field]'), function (input) {
                var name = $(input).attr('data-field');
                var control = this.initializeField(input);
                control.value(this._controller.getValue(row.id, name));
                var attrs = this._getAttrs(name, row);

                control.disable(_.get(attrs, 'disable', false));
            }.bind(this));

            this._findControl(row.id, this._focusField).focus(true);
        },
        /**
         * @method MultiFieldModel#_updateHandler
         * @private
         *
         * @param {Object} row
         * @param {Object} updatedValues
         */
        _updateHandler: function (row, updatedValues) {
            _.each(this._getFieldsForRow(row.id), function (widget, key) {
                var column = _.find(this._options.columns, 'id', key);
                var newType;

                if (_.isFunction(column.type)) {
                    newType = column.type(row);
                } else {
                    newType = column.type;
                }

                if (!_.contains(_.keys(updatedValues), key) && fieldFactory.getTypeOf(widget) !== newType) {
                    var parent = widget.element().parent();
                    widget.destroy();
                    //TODO: replace these with templates
                    var $el = $('<input type="text" data-field="' + key + '">');

                    if (newType === 'checkbox' || newType === 'radio') {
                        $el = $('<label class="' + newType + '" data-field="' + key + '"><input type="' + newType + '" data-field="' + key + '"></label>');
                    }
                    parent.append($el);
                    widget = this.initializeField($el, { replace: true });
                } else {
                    widget.value(row[key]);
                }
                var attrs = this._getAttrs(key, row);

                widget.disable(_.get(attrs, 'disable', false));
            }.bind(this));
        },
        /**
         * @method MultiFieldModel#_removeHandler
         * @private
         *
         * @param {String} rowId
         * @param {String} index
         */
        _removeHandler: function (rowId, index) {
            //TODO: find field that is currently focused and focus same field in next row
            this._$el.find('.multi-field-row[data-id=' + rowId + ']').remove();

            var rowToFocus = this._$el.find('.multi-field-row:eq(' + index + ')');

            if (rowToFocus.length === 0) {
                rowToFocus = this._$el.find('.multi-field-row').last();
            }

            this._findControl(rowToFocus.attr('data-id'), this._focusField).focus(true);
        },
        /**
         * @method MultiFieldModel#_redrawHandler
         * @private
         */
        _redrawHandler: function () {
            var rows = _.reduce(this._controller.data(), function (memo, row) {
                return memo += rowTemplate({
                    id: this.id(),
                    rowId: row.id,
                    columns: this._options.columns,
                    editable: this._options.editable,
                    types: _.transform(this._options.columns, function (memo, column) {
                        memo[column.id] = this._getType(column, row);
                        return memo;
                    }.bind(this), {}),
                    row: row
                });
            }.bind(this), '');

            this._$el.find('.rows').empty().append(rows);

            _.each(this._$el.find('[data-field]'), function (input) {
                var control = this.initializeField(input);
                var rowId = $(input).closest('.multi-field-row').attr('data-id');
                var name = $(input).attr('data-field');
                control.value(this._controller.getValue(rowId, name));
                control.init();

                var attrs = this._getAttrs(name, this._controller.row(rowId));

                control.disable(_.get(attrs, 'disable', false));
            }.bind(this));

            _.each(this._$el.find('.multi-field-row'), function (row) {
                var $row = $(row);

                var contextMenu = new ContextMenu({
                    isActionEnabled: function (action) {
                        return this._controller.isActionEnabled(action.id);
                    }.bind(this),
                    appendTo: this._options.appendTo
                }, [
                    {
                        id: 'add-below',
                        trigger: 'add-below',
                        title: this._options.locale.add
                    },
                    {
                        id: 'remove',
                        trigger: 'remove-row',
                        title: this._options.locale.delete
                    }
                ]);
                $row.find('.row-action').append(contextMenu.element());
                contextMenu.init();
            }.bind(this));
        },
        /**
         * @method MultiFieldModel#_focusRow
         * @private
         *
         * @param {String} rowId
         */
        _focusRow: function (rowId) {
            this._$el.find('.multi-field-row.focused').removeClass('focused');

            this._$el.find('.multi-field-row[data-id=' + rowId + ']').addClass('focused');
        },
        /**
         * @method MultiFieldModel#_getFieldsForRow
         * @private
         *
         * @param {String} id
         *
         * @returns {Object} fields for row
         */
        _getFieldsForRow: function (id) {
            return _.transform(this._fields, function (memo, config) {
                if (config.row === id) {
                    memo[config.field] = config.control;
                }
                return memo;
            }, {});
        },
        /**
         * @method MultiFieldModel#replace
         * @private
         *
         * @param {Object} location
         * @param {Widget} widget
         */
        _replace: function (location, widget) {
            var found = _.find(this._fields, function (config) {
                return location.field === config.field && location.row === config.row;
            });

            found.control = widget;
        }
    });

    return MultiFieldView;
});

define('oui/widgets/multiField/multiField',[
    'underscore',
    'oui/base/widget',
    'oui/widgets/multiField/multiFieldModel',
    'oui/widgets/multiField/multiFieldController',
    'oui/widgets/multiField/multiFieldView'
], function (_, Widget, MultiFieldModel, MultiFieldController, MultiFieldView) {
    'use strict';

    /**
     * @typedef MultiField#Config
     *
     * @property {Boolean} [editable=false]
     * @property {Boolean} [collapse=false]
     * @property {Boolean} [allowEmpty=false]
     * @property {Array} [columns=[]]
     * @property {Object} [actions={}]
     * @property {Array} [actionOrder=[]]
     * @property {Number} [max=null]
     * @property {Object} [defaultRow={}]
     * @property {String} [factory='built-in']
     * @property {Object} [recordKeys]
     * @property {String} [recordKeys.id='id']
     */
    var __DEFAULTS__ = {
        editable: false,
        collapse: false,
        allowEmpty: false,
        columns: [],
        actions: {

        },
        actionOrder: [],
        max: null,
        defaultRow: {},
        factory: 'built-in',
        recordKeys: {
            id: 'id'
        }
    };

    /**
     * @class MultiField
     * @extends Widget
     *
     * @param {MultiField#Config} options
     * @param {jQuery} $element
     * @param {Array} initData
     */
    function MultiField(options, $element, initData) {
        var model;
        Widget.call(this, options, $element, initData);

        model = new MultiFieldModel(this._initData, this._options);
        this._controller = new MultiFieldController(this._options, model);
        this._view = new MultiFieldView(model, this._controller, this._element, this._options, this._settings);

        this._setupEvents([model, this._view]);
    }

    MultiField.prototype = _.extend(Object.create(Widget.prototype), {
        constructor: MultiField,
        name: _.constant('multi-field'),
        defaults: _.constant(__DEFAULTS__),
        /**
         * @method MultiField#addRow
         *
         * @param {Object} rowConfig
         */
        addRow: function (rowConfig) {
            this._controller.addRow(rowConfig);
        },
        /**
         * @method MultiField#newRow
         */
        newRow: function () {
            this._controller.newRow();
        },
        /**
         * @method MultiField#removeRow
         *
         * @param {String} id
         */
        removeRow: function (id) {
            this._controller.removeRow(id);
        },
        /**
         * @method MultiField#updateRow
         *
         * @param {String} id
         * @param {Object} updateConfig
         */
        updateRow: function (id, updateConfig) {
            this._controller.updateRow(id, updateConfig);
        },
        /**
         * @method MultiField#_init
         * @private
         *
         * @param {Function} Mediator
         *
         * @returns {Object|undefined} mediator
         */
        _init: function (Mediator) {
            if (!this._options.allowEmpty && (this._controller.data() == null || _.isEmpty(this._controller.data()))) {
                this._controller.newRow();
            }

            if (_.isFunction(Mediator)) {
                return this._controller.setStore(Mediator);
            }
        },
        /**
         * @method MultiField#data
         *
         * @param {Array} value
         */
        data: function (value) {
            if (!arguments.length) {
                return this._controller.data();
            }

            this._controller.data(value);
        }
    });

    return MultiField;
});

define('oui/widgets/hopper/hopperModel',[
    'jquery',
    'underscore',
    'oui/widgets/tree/collection'
], function ($, _, Collection) {
    'use strict';

    var defineNodes = function () { //jscs:ignore jsDoc
        Object.defineProperty(this, '_nodes', {
            get: function () { //jscs:ignore jsDoc
                return this.nodes();
            },
            set: _.noop // Don't want this to be overridden
        });
    };

    /**
     * @class HopperModel
     * @private
     * @extends Collection
     *
     * @param {TreePanel#Config} options
     * @param {Array<Object>|Object<String,Array>|Function} initData
     */
    var HopperModel = function (options, initData) {
        this._store = null;
        options.filterShowHierarchy = true; // Always needs to be true for the hopper
        Collection.call(this, options, initData);
    };

    HopperModel.prototype = _.create(Collection.prototype, {
        constructor: HopperModel,

        /**
         * Populate the collection from the provided data.
         *
         * * If the data is an array just add all of those objects to the collection.
         * * If the data is an Object with the function getAll, treat that as a mediator and set it up. Adding all known
         * objects to the collection.
         * * If the data is a plain Object with available and selected properties add the available to the collection and set
         * the selected attributes on the provided data.
         *
         * @method HopperModel#_initialize
         * @private
         * @param {Array<Object>|Object|Function} initData
         */
        _initialize: function (initData) {
            if (_.isFunction(initData)) {
                var Mediator = initData;
                this._store = new Mediator();
                _.each(this._options.storeFns, function (fn, name) {
                    this._store[name] = fn;
                }.bind(this));
                defineNodes.call(this);
                this._updateNodesData();
            } else {
                Collection.prototype._initialize.apply(this, arguments);
            }
        },

        /**
         * Adds the node to all the attribute caches.
         *
         * @method Collection#addAttrs
         * @param {Object} node The node to get the attributes for
         */
        addAttrs: function (node) {
            if (this._store) {
                if (this._store.attrs) {
                    var attrs = this._store.attrs(node);
                    var id = this._getId(node);

                    _.each(attrs, function (attr) {
                        this.addToAttr(attr, id);
                    }.bind(this));
                }
            } else {
                Collection.prototype.addAttrs.apply(this, arguments);
            }
        },

        /**
         * Repopulates the hopper
         *
         * @method _repopulate
         * @private
         */
        _repopulate: function () {
            this._updateNodesData();
            this.trigger('rerender');
        },

        /**
         * Gets the nodes within the collection as an array.
         *
         * @method Collection#nodes
         *
         * @param {Object} modifiers
         * @returns {Array} The array of nodes that currently exist within the collection.
         */
        nodes: function (modifiers) {
            if (this._store) {
                var nodes = this._store.getItems(modifiers);
                if (_.get(modifiers, 'sorted', false)) {
                    nodes = _.sortBy(nodes, function (node) {
                        return this._sorts['default'].ids.indexOf(this._getId(node));
                    }.bind(this));
                }
                return nodes;
            } else {
                return Collection.prototype.nodes.apply(this, arguments);
            }
        },

        /**
         * Gets the node for a given id.
         *
         * @method Collection#node
         * @param {String} id The identifier to look for a node.
         * @param {Object} modifiers
         * @returns {Object} The found node or undefined if none exists for a given id.
         */
        node: function (id, modifiers) {
            if (this._store) {
                return this._store.getItem(id, modifiers);
            } else {
                return Collection.prototype.node.apply(this, arguments);
            }
        },

        /**
         * Checks if an item can be added to the collection (in a specific group, if provided)
         *
         * @method canAddItem
         * @param {Object} [group]
         * @returns {Boolean}
         */
        canAddItem: function (group) {
            if (this._store && this._store.canAddItem) {
                return this._options.editable && this._store.canAddItem(group);
            }
            return this._options.editable;
        },

        /**
         * Checks if an item can be moved out of the hopper
         *
         * @method canMoveItem
         * @param {Object} item
         * @returns {Boolean}
         */
        canMoveItem: function (item) {
            if (this._store && this._store.canMoveItem) {
                return this._options.editable && this._store.canMoveItem(item);
            }
            return this._options.editable;
        },

        /**
         * Calls the store to add a new item
         *
         * @method newItem
         * @param {Object} [group] Specific group to add to
         */
        newItem: function (group) {
            var curGroup = group;
            var baseRecord = {};
            var value;

            while (curGroup != null) {
                // This is dangerous, but OUI doesn't yet provide a way to know if this is "null" or just `null`
                value = curGroup.value === 'null' ? null : curGroup.value;
                if (value !== undefined) {
                    _.set(baseRecord, curGroup.field, value);
                }
                curGroup = curGroup.parent;
            }

            if (this._options.recordKeys.sequenceNumber) {
                var lastId = _.last(this._sorts['default'].ids);
                var lastNode;
                if (lastId != null) {
                    lastNode = this.node(lastId);
                    _.set(baseRecord, this._options.recordKeys.sequenceNumber, this._getNodeValue(lastNode, this._options.recordKeys.sequenceNumber, -10) + 10);
                } else {
                    _.set(baseRecord, this._options.recordKeys.sequenceNumber, 10);
                }
            }

            if (this._store && this._store.newItem) {
                $.when(this._store.newItem(baseRecord)).done(this._repopulate.bind(this));
            } else if (!this._store) {
                this.add(baseRecord);
            }
        },

        /**
         * Updates the internal node data to rebuild grouping data for view
         *
         * @method _updateNodesData
         * @private
         */
        _updateNodesData: function () {
            var prevData = _.pick(this._attrs, 'expanded');
            this.pause();
            this.invalidate();
            this.resume();

            _.each(this.nodes(), function (node) {
                if (this._options.groupBy != null) {
                    this.addToGroups(node);
                }
                this.addToSorts(node);
                this.addAttrs(node);
            }.bind(this));

            _.extend(this._attrs, prevData);
        },

        /**
         * Updates the node at a given id with the new data.
         *
         * @method HopperModel#update
         * @param {String} id The id of the record to update
         * @param {Object} data The new data to add/overwrite on the node.
         *
         * @fires Collection#update (if not using a Mediator)
         *
         * @returns {String} The id of the node updated.
         */
        update: function (id, data) {
            if (this._store) {
                _.each(data, function (value, key) {
                    this._store.updateItem(this.node(id), key, value);
                }.bind(this));
                return id;
            } else {
                return Collection.prototype.update.apply(this, arguments);
            }
        }
    });

    return HopperModel;
});


define('template-111', ['underscore'], function (_) { return function(data) {
var __t, __p = '';
__p += '<div class="hopper oui-hopper orcl-widget orcl-hopper" tabindex="0"></div>';
return __p
};});


define('template-112', ['underscore'], function (_) { return function(data) {
var __t, __p = '';
__p += '<div class="hopper-group ' +
((__t = ( data.collapsed ? 'collapsed' : '' )) == null ? '' : __t) +
'" id="' +
((__t = ( data.id )) == null ? '' : __t) +
'" data-hopper-groupid="' +
((__t = ( data.groupId )) == null ? '' : __t) +
'" aria-labelledby="' +
((__t = ( data.id )) == null ? '' : __t) +
'-group-title"> <button type="button" class="not-btn collapse-icon" title="' +
((__t = ( data.locale.toggle )) == null ? '' : __t) +
'" tabindex="-1" aria-controls="' +
((__t = ( data.groupId )) == null ? '' : __t) +
'-group-items" aria-expanded="' +
((__t = ( data.collapsed ? 'false' : 'true' )) == null ? '' : __t) +
'"> <span class="spoken-text">' +
((__t = ( data.locale.toggle )) == null ? '' : __t) +
'</span> </button> <div class="group-title" id="' +
((__t = ( data.id )) == null ? '' : __t) +
'-group-title" aria-controls="' +
((__t = ( data.groupId )) == null ? '' : __t) +
'-group-items" role="tab" aria-expanded="' +
((__t = ( data.collapsed ? 'false' : 'true' )) == null ? '' : __t) +
'">' +
((__t = ( data.title )) == null ? '' : __t) +
'</div> <div class="group-items" role="tabpanel" id="' +
((__t = ( data.groupId )) == null ? '' : __t) +
'-group-items"><div role="tablist"> ' +
((__t = ( data.innerContent )) == null ? '' : __t) +
' </div></div></div>';
return __p
};});


define('template-113', ['underscore'], function (_) { return function(data) {
var __t, __p = '', __j = Array.prototype.join;
function print() { __p += __j.call(arguments, '') }
__p += '<div class="hopper-bucket ' +
((__t = ( data.hasTitle ? 'selectable-bucket' : '' )) == null ? '' : __t) +
' ' +
((__t = ( data.collapsed ? 'collapsed' : '' )) == null ? '' : __t) +
'" id="' +
((__t = ( data.id )) == null ? '' : __t) +
'" data-hopper-groupid="' +
((__t = ( data.groupId )) == null ? '' : __t) +
'"        ' +
((__t = ( data.hasTitle ? 'aria-labelledby="' + data.id + '-bucket-title"' : '')) == null ? '' : __t) +
'> ';
 if (data.hasTitle) { ;
__p += ' <button type="button" class="not-btn collapse-icon" title="' +
((__t = ( data.locale.toggle )) == null ? '' : __t) +
'" tabindex="-1" aria-controls="' +
((__t = ( data.groupId )) == null ? '' : __t) +
'-bucket-items" aria-expanded="' +
((__t = ( data.collapsed ? 'false' : 'true' )) == null ? '' : __t) +
'"> <span class="spoken-text">' +
((__t = ( data.locale.toggle )) == null ? '' : __t) +
'</span> </button> <div class="bucket-title" id="' +
((__t = ( data.id )) == null ? '' : __t) +
'-bucket-title" aria-controls="' +
((__t = ( data.groupId )) == null ? '' : __t) +
'-bucket-items" role="tab" aria-expanded="' +
((__t = ( data.collapsed ? 'false' : 'true' )) == null ? '' : __t) +
'">' +
((__t = ( data.title )) == null ? '' : __t) +
'</div> ';
 } ;
__p += ' <div class="bucket-items" role="tabpanel" id="' +
((__t = ( data.groupId )) == null ? '' : __t) +
'-bucket-items"> ';
 if (data.editable) { ;
__p += ' <div class="bucket-item add-bucket-item" id="' +
((__t = ( data.groupId )) == null ? '' : __t) +
'-add-item" data-hopper-groupaddid="' +
((__t = ( data.groupId )) == null ? '' : __t) +
'-add-item" data-hopper-group-parent-id="' +
((__t = ( data.groupId )) == null ? '' : __t) +
'"                aria-label="' +
((__t = ( data.locale.add )) == null ? '' : __t) +
'"> <div class="task-card new-task-card"></div> </div> ';
 } ;
__p += ' ' +
((__t = ( data.innerContent )) == null ? '' : __t) +
' </div></div>';
return __p
};});


define('template-114', ['underscore'], function (_) { return function(data) {
var __t, __p = '';
__p += '<div class="bucket-item ' +
((__t = ( data.isDraggable ? 'oui-draggable' : '' )) == null ? '' : __t) +
'" id="' +
((__t = ( data.id )) == null ? '' : __t) +
'" data-hopperid="' +
((__t = ( data.itemId )) == null ? '' : __t) +
'"> ' +
((__t = ( data.innerContent )) == null ? '' : __t) +
'</div>';
return __p
};});

define('oui/widgets/hopper/hopperView',['require','underscore','jquery','oui/base/view','template-111','template-112','template-113','template-114','oui/utils/scrollingUtils','oui/widgets/tree/groupingUtils','oui/utils/draggable','logger'],function (require) {//jshint ignore:line
    'use strict';

    var _ = require('underscore');
    var $ = require('jquery');
    var View = require('oui/base/view');
    var hopperTemplate = require('template-111');
    var groupTemplate = require('template-112');
    var bucketTemplate = require('template-113');
    var itemTemplate = require('template-114');
    var scrollingUtils = require('oui/utils/scrollingUtils');
    var groupingUtils = require('oui/widgets/tree/groupingUtils');
    var draggable = require('oui/utils/draggable');
    var logger = require('logger');

    var hvp;

    /**
     * @class HopperView
     * @private
     * @extends View
     *
     * Does not use the parameters. Forwards them with an apply call.
     *
     * @param {Hopper#Config} [options]
     * @param {jQuery|null} [element]
     * @param {HopperController} [controller]
     * @param {Object} [settings]
     * @param {Collection} [model]
     */
    function HopperView(options, element, controller, settings, model) {
        View.apply(this, arguments);

        if (this._controller.isLoaded()) {
            this.render();
        }

        model.on({
            'view.active': this.active.bind(this),
            'view.reset view.rerender view.invalidate view.change': this.render.bind(this),
            'view.filter': this.filter.bind(this),
            'view.attribute-updated:expanded': this._expandedChanged.bind(this),
            'view.call-out': this._callOut.bind(this)
        });

        this._initListeners();
    }

    HopperView.prototype = Object.create(View.prototype);
    hvp = HopperView.prototype;

    /**
     * Gets event arguments for click/dblclick data-triggers
     *
     * @method HopperView#_getEventArgs
     * @param {Object} event
     * @returns {{item: Object}}
     * @private
     */
    hvp._getEventArgs = function (event) {
        var id = this._getHopperId($(event.currentTarget).closest('.bucket-item'));
        return {
            item: id && this._controller.node(id)
        };
    };

    /**
     * Sets up listeners for the view
     *
     * @method HopperView#_initListeners
     * @private
     */
    hvp._initListeners = function () {
        this.addShortcuts({
            'UP LEFT': _.partial(this._moveActive, { previous: true }),
            'DOWN RIGHT': _.partial(this._moveActive, { next: true }),
            'ctrl-UP': _.partial(this._moveActive, { parent: true }),
            'ENTER': this._toggleCollapsedOrSelect
        });

        this._$el.on('click', '.collapse-icon, .group-title, .bucket-title', function (e) {
            this._triggerToggleCollapse($(e.currentTarget));
        }.bind(this));

        this._$el.on('mousedown', '.bucket-item', function (e) {
            this._controller.active(this._getHopperId($(e.currentTarget)));
        }.bind(this));

        this._$el.on('click', '.add-bucket-item', function (e) {
            var $this = $(e.currentTarget);
            this._triggerAdd(groupingUtils.inflate($this.attr('data-hopper-group-parent-id')));
        }.bind(this));

        //TODO: think about additional event types
        this._$el.on('click', '.bucket-item [data-trigger]', function (event) {
            var type = event.currentTarget.getAttribute('data-trigger-type');
            if (type === 'click' || type == null) {
                event.stopPropagation();
                this.trigger(event.currentTarget.getAttribute('data-trigger'), this._getEventArgs(event));
            }
        }.bind(this));

        this._$el.on('dblclick', '.bucket-item [data-trigger-type="dblclick"][data-trigger]', function (event) {
            event.stopPropagation();
            this.trigger(event.currentTarget.getAttribute('data-trigger'), this._getEventArgs(event));
        }.bind(this));

        this._$el.on('focusin', function (e) {
            if (!$(e.target).is(this._$el)) {
                this._$el.focus();
            }
        }.bind(this));

        var hoverTimeout = null;
        this._$el.on('mouseenter', '.bucket-item, .hopper-group > .group-title, .hopper-bucket > .bucket-title', function (event) {
            hoverTimeout = window.setTimeout(function () {
                var $node = $(event.currentTarget).closest('.bucket-item, .hopper-group, .hopper-bucket');
                var id = this._getHopperId($node);
                var triggerItem = id.group ? id : this._controller.node(id);
                this.trigger('hover', event, triggerItem);
            }.bind(this), 2000);
        }.bind(this));

        this._$el.on('mouseleave', '.bucket-item, .hopper-group > .group-title, .hopper-bucket > .bucket-title', function () {
            window.clearTimeout(hoverTimeout);
        });

        if (this._options.drag.enabled) {
            var $document = $(document.body);
            draggable(this._$el, {
                clone: true,
                appendCloneTo: $document,
                positionFixedClone: true,
                triggerDistance: 10,
                selector: '.bucket-item.active',
                cancellable: true,
                namespace: 'assign'
            });

            this._$el.on('drag-start.assign', this._dragStart.bind(this));

            if (this._options.drag.out) {
                this._$el.on('drag-end.assign', this._triggerDrop.bind(this));
            }

            if (this._options.drag.reorder) {
                this._$el.on('dragging-over.assign', this._dragReorder.bind(this));
                this._$el.on('drag-end.assign', this._dragReorderEnd.bind(this));
            }
        }
    };

    /**
     * @method HopperView#_dragStart
     * @param {Object} e
     * @returns {*}
     * @private
     */
    hvp._dragStart = function (e) {
        var dragState;

        if (e.$handle.hasClass('add-bucket-item')) {
            e.preventDefault();
        } else {
            dragState = {
                resetCalendar: this._associatedCalendar ? function () {
                    this._associatedCalendar._view._currentView.render();
                }.bind(this) : _.noop,
                inAssociatedCalendar: false,
                calendarRangeResults: null,
                dragItem: this._controller.node(this._getHopperId(e.$handle)),
                reordering: false
            };
            dragState.canDragOut = this._controller.canMoveItem(dragState.dragItem);

            // Immediately cancel drag if reordering is off and you cannot move the current item
            // We still check canMoveItem on item drop (see #_triggerDrop) in case we still currently support reorder
            if (!this._isDragReorderable() && !dragState.canDragOut) {
                e.preventDefault();
            }
        }

        return dragState;
    };

    /**
     * @method HopperView#_dragReorder
     * @param {Object} e
     * @param {Object} dragState
     * @private
     */
    hvp._dragReorder = function (e, dragState) {
        if (!this._isDragReorderable()) {
            return;
        }

        var $closestElement = $(e.dropTarget);
        var inHopper = $closestElement.closest(this._$el).length > 0;
        var $closestBucket = $closestElement.closest('.bucket-item').not('.add-bucket-item');
        var $closestBucketItems = $closestElement.closest('.bucket-items');
        if (inHopper) {
            // This assumes that you cannot reorder while grouped
            dragState.reordering = true;
            if (dragState.reorderOriginMarker == null) {
                dragState.reorderOriginMarker = $('<span style="display: none;" />').insertBefore(e.$handle);
            }
            if ($closestBucket.length) {
                var bucketPos = $closestBucket.offset();
                var closestCardHalfway = bucketPos.left + ($closestBucket.outerWidth(true) / 2);
                var pageX = $(window).scrollLeft() + e.current.x;

                if (pageX < closestCardHalfway) {
                    e.$handle.insertBefore($closestBucket);
                } else {
                    e.$handle.insertAfter($closestBucket);
                }
            } else if ($closestBucketItems.length === 0) {
                e.$handle.appendTo(e.$handle.closest('.bucket-items'));
            }
        } else if (!inHopper && dragState.reordering) {
            dragState.reordering = false;
            e.$handle.insertAfter(dragState.reorderOriginMarker);
        }
    };

    /**
     * @method HopperView#_dragReorderEnd
     * @param {Object} e
     * @param {Object} dragState
     * @private
     */
    hvp._dragReorderEnd = function (e, dragState) {
        if (dragState.reorderOriginMarker) {
            dragState.reorderOriginMarker.remove();
        }

        if (dragState.reordering) {
            var $bucketItems = e.$handle.closest('.bucket-items').children().not('.add-bucket-item');
            var orderedIds = _.map($bucketItems, function (bucketItem) {
                return this._getHopperId($(bucketItem));
            }.bind(this));
            this._controller._reorderNodes(orderedIds);
        }
    };

    /**
     * Associates a calendar to the hopper
     *
     * @method HopperView#_associateCalendar
     * @param {Calendar} calendar
     * @returns {Boolean} Whether a calendar is already associated
     * @private
     */
    hvp._associateCalendar = function (calendar) {
        if (this._associatedCalendar != null && this._associatedCalendar !== calendar) {
            throw new Error('Hopper can only be associated with one calendar');
        } else if (this._associatedCalendar) {
            return true;
        }

        this._associatedCalendar = calendar;

        this._$el.on('dragging-over.assign', this._dragOverCalendar.bind(this));
        this._$el.on('drag-end.assign', this._dragEndCalendar.bind(this));
    };

    /**
     * @method HopperView#_dragOverCalendar
     * @param {Object} e
     * @param {Object} dragState
     * @private
     */
    hvp._dragOverCalendar = function (e, dragState) {
        if ($(e.dropTarget).closest(this._associatedCalendar.element()).length) {
            dragState.inAssociatedCalendar = true;
            dragState.calendarRangeResults = this._associatedCalendar._view._renderTemporaryItem(dragState.dragItem, {
                pageX: e.originalEvent.pageX,
                pageY: e.originalEvent.pageY
            });
        } else if (dragState.inAssociatedCalendar) {
            dragState.inAssociatedCalendar = false;
            dragState.calendarRangeResults = null;
            dragState.resetCalendar();
        }
    };

    /**
     * @method HopperView#_dragEndCalendar
     * @param {Object} e
     * @param {Object} dragState
     * @private
     */
    hvp._dragEndCalendar = function (e, dragState) {
        if (dragState.inAssociatedCalendar) {
            if (dragState.dragItem && dragState.calendarRangeResults && dragState.canDragOut) {
                this._triggerAssign(dragState.dragItem, dragState.calendarRangeResults);
                this._controller.repopulate();
                if (this._associatedCalendar) {
                    this._associatedCalendar.active(_.get(dragState.dragItem, this._options.recordKeys.id));
                }
            }
            dragState.resetCalendar();
        }
    };

    /**
     * @method HopperView#buildElement
     * @param {Hopper#Config} options
     * @param {jQuery} $el
     */
    hvp.buildElement = function (options, $el) {
        if (_.isNull($el)) {
            //create our own element
            $el = $(hopperTemplate());
        }

        return $el;
    };

    /**
     * @method HopperView#filter
     * @param {Array} nodes
     */
    hvp.filter = function (nodes) {
        var that = this;
        this._$el.find('.search-result').removeClass('search-result')
            .filter('.bucket-item').replaceWith(function () {
                var id = that._getHopperId($(this));
                return that.renderNode(that._controller.node(id));
            });
        _.each(nodes, function (id) {
            var $searchResult = that._findItem(id).addClass('search-result');
            var $groupResult = $searchResult.filter('.hopper-group, .hopper-bucket');

            that._setCollapsed($groupResult, false);

            $searchResult.filter('.bucket-item').replaceWith(function () {
                var id = that._getHopperId($(this));
                return $(that.renderNode(that._controller.node(id), {
                    matchesSearch: true
                })).addClass('search-result');
            });
        });
    };

    /**
     * Renders the hopper as in its current state.
     *
     * @method HopperView#render
     */
    hvp.render = function () {
        var content;
        if (this._options.groupBy.length === 0) {
            this._$el.removeAttr('role');
            content = this.renderFlat(null, this._controller.nodes({
                unselectedOnly: true,
                sorted: true
            }));
        } else {
            this._$el.attr('role', 'tablist');
            content = this.renderTree(this._controller.roots());
        }

        this._$el.empty().append(content);
        this.active(this._controller.active());
    };

    /**
     * Renders a portion of the hopper.
     *
     * @method HopperView#renderTree
     * @param {Array<Object>} nodes
     * @protected
     */
    hvp.renderTree = function (nodes) {
        var nullNode;
        var state = {
            nullNode: null
        };
        var result = this._renderTree(nodes, state);
        if (state.nullNode != null) {
            nullNode = state.nullNode;
        } else if (this._options.showEmptyGroups) {
            nullNode = _.reduce(this._options.groupBy, function (memo, groupBy) {
                return this._controller._model._makeGroup(groupBy.field, null, memo);
            }.bind(this), null);
        }
        if (nullNode) {
            var nullChildren = this._controller.getChildren(nullNode, {
                unselectedOnly: true,
                sorted: true
            });
            var flatContent = _.reduce(nullChildren, function (memo, node) {
                return memo + this.renderNode(node);
            }.bind(this), '');
            result = this._createBucket(nullNode, this._options.locale.uncategorizedLabel, flatContent, {
                forceExpanded: true,
                hasTitle: true,

                // This programatically wrong but preserves the original behavior from when rendering the null node created
                // two nested buckets by calling renderFlat and sticking that in another bucket.
                // Consider changing post 16.2 release
                editable: this._controller.canAddItem(null)
            }) + result;
        }
        return result;
    };

    /**
     * Renders a single group
     *
     * @method HopperView#_createGroup
     * @param {Object} group
     * @param {String} groupName
     * @param {String} innerContent
     * @param {Object} [opts]
     * @returns {String}
     * @private
     */
    hvp._createGroup = function (group, groupName, innerContent, opts) {
        return groupTemplate({
            collapsed: this._controller._isCollapsed(group) && (!opts || !opts.forceExpanded),
            groupId: group.toString(),
            id: group.toString(),
            locale: this._options.locale,
            hasTitle: true,
            title: groupName,
            innerContent: innerContent
        });
    };

    /**
     * Renders a single bucket
     *
     * @method HopperView#_createBucket
     * @param {Object} bucket
     * @param {String} bucketName
     * @param {String} innerContent
     * @param {Object} opts
     * @returns {String}
     * @private
     */
    hvp._createBucket = function (bucket, bucketName, innerContent, opts) {
        return bucketTemplate({
            collapsed: this._controller._isCollapsed(bucket) && (!opts || !opts.forceExpanded),
            groupId: bucket.toString(),
            id: bucket.toString() + '-nodes',
            locale: this._options.locale,
            hasTitle: opts.hasTitle,
            title: bucketName,
            editable: opts.editable,
            innerContent: innerContent
        });
    };

    function groupChainNull(group) { //jscs:ignore jsDoc
        var g = group;
        while (g != null) {
            if (g.value !== 'null') {
                return false;
            }
            g = g.parent;
        }
        return true;
    }

    /**
     * @method HopperView#_renderCollapsedGroup
     * @param {Object} group
     * @param {Boolean} hasGroups
     * @param {Boolean} editable
     * @returns {String}
     * @private
     */
    hvp._renderCollapsedGroup = function (group, hasGroups, editable) {
        if (hasGroups) {
            return this._createGroup(group, this._controller._model.getGroupDisplayValue(group), '');
        } else {
            return this._createBucket(group, this._controller._model.getGroupDisplayValue(group), '', {
                hasTitle: !hasGroups,
                editable: editable
            });
        }
    };

    /**
     * @method HopperView#_maybeFindNullNOde
     * @param {Object} group
     * @param {Object} state
     * @private
     */
    hvp._maybeFindNullNode = function (group, state) {
        if (group.value === 'null') {
            var children = _.filter(this._controller.getChildren(group, {
                unselectedOnly: true
            }), function (child) {
                return !!child.group;
            });

            if (children.length) {
                _.each(children, function (child) {
                    this._maybeFindNullNode(child, state);
                }.bind(this));
            } else {
                state.nullNode = group;
            }
        }
    };

    /**
     * @method HopperView#_renderTreeWithGroups
     * @param {Object} group
     * @param {Array<Object>} subGroups
     * @param {Array<Object>} nodes
     * @param {Object} opts
     * @returns {String}
     * @private
     */
    hvp._renderTreeWithGroups = function (group, subGroups, nodes, opts) {
        subGroups = subGroups || [];
        nodes = nodes || [];

        if (opts.isCollapsed) {
            this._maybeFindNullNode(group, opts.state);
            return this._createGroup(group, this._controller._model.getGroupDisplayValue(group), '');
        }

        var subTree = this._renderTree(subGroups, opts.state, opts.level + 1);
        if (opts.editable || nodes.length) {
            subTree += this._createBucket(group, '',
                _.reduce(nodes, function (memo, node) {
                    return memo + this.renderNode(node);
                }.bind(this), ''), {
                    hasTitle: false,
                    editable: opts.editable
                });
        }
        if (opts.forceGroups || subTree.length) {
            return this._createGroup(group, this._controller._model.getGroupDisplayValue(group), subTree);
        }
        return '';
    };

    /**
     * Renders a portion of the hopper.
     *
     * @method HopperView#_renderTree
     * @param {Array<Object>} nodes
     * @param {Object} state
     * @param {Number} [level=0]
     * @private
     */
    hvp._renderTree = function (nodes, state, level) {
        level = level || 0;
        return _.reduce(nodes, function (memo, group) {
            var children = this._controller.getChildren(group, {
                unselectedOnly: true,
                sorted: true
            });
            var isCollapsed = this._controller._isCollapsed(group);
            var editable = this._controller.canAddItem(group);
            var categories = _.groupBy(children, function (child) {
                return child.group ? 'group' : 'node';
            }.bind(this));
            var forceGroups = level < this._options.groupBy.length - 1;
            var hasGroups = !!categories.group && categories.group.length > 0;
            var hasNodes = !!categories.node && categories.node.length > 0;
            var result = '';

            if (isCollapsed && (this._options.showEmptyGroups || children.length > 0) && !groupChainNull(group)) {
                return memo + this._renderCollapsedGroup(group, hasGroups || forceGroups, editable);
            }

            if (hasGroups || forceGroups) {
                result += this._renderTreeWithGroups(group, categories.group, categories.node, {
                    isCollapsed: isCollapsed,
                    editable: editable,
                    forceGroups: forceGroups,
                    state: state,
                    level: level
                });
            } else if (groupChainNull(group)) {
                logger.assert(state.nullNode == null, 'no null node has been set yet');
                state.nullNode = group;
                return memo;
            } else if (editable || hasNodes || this._options.showEmptyGroups) {
                result += this._createBucket(group, this._controller._model.getGroupDisplayValue(group),
                    _.reduce(categories.node, function (memo, node) {
                        return memo + this.renderNode(node);
                    }.bind(this), ''), {
                        hasTitle: true,
                        editable: editable
                    });
            }

            return memo + result;
        }.bind(this), '');
    };

    /**
     * Renders an ungrouped hopper.
     *
     * @method HopperView#renderFlat
     * @param {Object} parent
     * @param {Array<Object>} nodes
     * @protected
     */
    hvp.renderFlat = function (parent, nodes) {
        return bucketTemplate({
            locale: this._options.locale,
            hasTitle: false,
            groupId: parent ? parent.toString() : 'bucket-item-hopper-' + this._settings.id + '-non-group',
            editable: this._controller.canAddItem(null),
            innerContent: _.reduce(nodes, function (memo, node) {
                return memo + this.renderNode(node);
            }.bind(this), '')
        });
    };

    /**
     * Renders a single node.
     *
     * @method HopperView#renderNode
     * @param {Object} node
     * @param {Object} [opts]
     * @protected
     */
    hvp.renderNode = function (node, opts) {
        var id = this._controller._model._getId(node);
        return itemTemplate({
            itemId: id,
            id: 'bucket-item-hopper-' + this._settings.id + '-node-' + id,
            isDraggable: this._showAsDraggable(),
            innerContent: this._options.renderer.render({
                data: node,
                attrs: this._controller._model.getAttributesForNode(id),
                matchesSearch: !!opts && opts.matchesSearch
            }, this._options)
        });
    };

    /**
     * Whether drag to reorder has been activated and is applicable
     *
     * Cannot have any active groupBys
     *
     * @method HopperView#_isDragReorderable
     * @returns {Boolean}
     * @private
     */
    hvp._isDragReorderable = function () {
        return this._options.drag.enabled && this._options.drag.reorder && this._options.groupBy.length === 0;
    };

    /**
     * Whether to show a node as draggable
     *
     * @method HopperView#_showAsDraggable
     * @returns {Boolean}
     * @private
     */
    hvp._showAsDraggable = function () {
        var dragEnabled = this._options.drag.enabled;
        var dragReorder = this._isDragReorderable();
        var dragOut = this._options.drag.out;
        return dragEnabled && (dragReorder || dragOut);
    };

    /**
     * Gets the currently visible active element from the HTML
     *
     * @method HopperView#_getActiveElement
     * @returns {*}
     * @private
     */
    hvp._getActiveElement = function () {
        return this._$el.find('.bucket-item.active, .hopper-bucket.active, .hopper-group.active');
    };

    /**
     * Moves the active selection around the hopper
     *
     * @method HopperView#_moveActive
     * @param {{previous, next, parent}} options
     * @private
     */
    hvp._moveActive = function (options) {
        var selector = '.bucket-item, .hopper-bucket.selectable-bucket, .hopper-group';
        var $activeElement = this._getActiveElement();
        var $nextItem;
        var $child;

        function goVisible($originalElement, fnName) { //jscs:ignore jsDoc
            return $originalElement[fnName](selector).filter(':visible').first();
        }

        function search($originalElement, fnName) { //jscs:ignore jsDoc
            var $nextItem;
            do {
                $nextItem = goVisible($originalElement, fnName);
                if ($nextItem.length === 0) {
                    $originalElement = $originalElement.parent().closest(selector);
                    if (fnName === 'prevAll') {
                        $nextItem = $originalElement;
                    }
                }
            } while ($nextItem.length === 0 && $originalElement.length);
            return $nextItem;
        }

        if ($activeElement.length === 0) {
            $nextItem = goVisible(this._$el, 'find');
        } else {
            if (options.next) {
                $nextItem = goVisible($activeElement, 'find');
                if ($nextItem.length === 0) {
                    $nextItem = search($activeElement, 'nextAll');
                }
            } else if (options.previous) {
                $nextItem = search($activeElement, 'prevAll');
                if ($nextItem.is($activeElement.prev())) {
                    $child = $nextItem.find(selector).filter(':visible').last();
                }
                if ($child && $child.length) {
                    $nextItem = $child;
                }
            } else if (options.parent) {
                $nextItem = $activeElement.parent().closest(selector);
            }
        }

        if ($nextItem.length) {
            this._controller.active(this._getHopperId($nextItem));
        }
    };

    /**
     * Marks an item as active in the view
     *
     * @method HopperView#active
     * @param {String} id
     * @protected
     */
    hvp.active = function (id) {
        this._$el.find('.active').removeClass('active');
        var $activeElement = this._findItem(id)
            .addClass('active');

        //Only scroll on a tree with an active element
        //scrollIntoView scrolls tree and all scrolling container parents
        if ($activeElement.length !== 0) {
            scrollingUtils.positionScrollTop($activeElement, this._$el);
            this._$el.attr('aria-activedescendant', $activeElement.attr('id'));
        } else {
            this._$el.removeAttr('aria-activedescendant');
        }
    };

    /**
     * Calls out a particular element in the view
     *
     * @method HopperView#_callOut
     * @param {String|Object} id
     * @private
     */
    hvp._callOut = function (id) {
        var $item = this._findItem(id);
        if ($item.length > 0) {
            if (!$item.hasClass('bucket-item')) {
                this._controller._setCollapsed(this._getHopperId($item), false);
                $item = $item.children('.bucket-title, .group-title').addClass('oui-callout');
            }

            $item.addClass('oui-callout');
            window.setTimeout(function () {
                $item.removeClass('oui-callout');
            }, 450);
        }
    };

    /**
     * Find element on view
     *
     * @method HopperView#_findItem
     * @param {String} id
     * @returns {*}
     * @private
     */
    hvp._findItem = function (id) {
        if (_.isObject(id)) {
            return this._$el.find('[data-hopper-groupid="' + groupingUtils.deflate(id) + '"]');
        } else {
            return this._$el.find('[data-hopperid="' + id + '"], [data-hopper-groupaddid="' + id + '"]');
        }
    };

    /**
     * Get id from element on view
     *
     * @method HopperView#_getHopperId
     * @param {*|{attr, is}} $node
     * @returns {*}
     * @private
     */
    hvp._getHopperId = function ($node) {
        if ($node.is('[data-hopper-groupid]')) {
            return groupingUtils.inflate($node.attr('data-hopper-groupid'));
        } else if ($node.is('[data-hopper-groupaddid]')) {
            return $node.attr('data-hopper-groupaddid');
        }
        var id = $node.attr('data-hopperid');
        return this._controller._model._getId(this._controller.node(id));
    };

    /**
     * Handler for keyboard selecting an element on the view
     *
     * @method HopperView#_toggleCollapsedOrSelect
     * @private
     */
    hvp._toggleCollapsedOrSelect = function () {
        var $activeElement = this._getActiveElement();

        if ($activeElement.hasClass('bucket-item')) {
            if ($activeElement.hasClass('add-bucket-item')) {
                this._triggerAdd(groupingUtils.inflate($activeElement.attr('data-hopper-group-parent-id')));
            } else {
                this._triggerSelect(this._getHopperId($activeElement));
            }
        } else {
            this._triggerToggleCollapse($activeElement);
        }
    };

    /**
     * Mark a node as selected
     *
     * @method _triggerSelect
     * @param {String} id
     * @private
     */
    hvp._triggerSelect = function (id) {
        var item = this._controller.node(id);
        if (this._controller.canMoveItem(item)) {
            if (this._associatedCalendar) {
                this._moveActive({ previous: true });
                this._triggerAssign(item, {});
                this._controller.repopulate();
            } else {
                this._moveActive({ previous: true });
                this._controller.addToData(id);
            }
        }
    };

    /**
     * Toggle a node collapsed
     *
     * @method _triggerToggleCollapse
     * @param {jQuery} $activeElement
     * @private
     */
    hvp._triggerToggleCollapse = function ($activeElement) {
        var $elem = $activeElement.closest('.hopper-bucket, .hopper-group');
        var isCollapsed = $elem.hasClass('collapsed');
        this._controller._setCollapsed(this._getHopperId($elem), !isCollapsed);
    };

    /**
     * Trigger the add item sequence
     *
     * @method _triggerAdd
     * @param {Object} group
     * @private
     */
    hvp._triggerAdd = function (group) {
        this._controller.newItem(group);
    };

    /**
     * @method HopperView#_triggerAssign
     * @param {Object} item
     * @param {{startDate: moment, endDate: moment}|{}} rangeResults
     * @private
     */
    hvp._triggerAssign = function (item, rangeResults) {
        this._controller._triggerAssignItem(item, rangeResults);
    };

    /**
     * @method HopperView#_triggerDrop
     * @param {Object} dragEndEvent
     * @param {Object} dragState
     * @private
     */
    hvp._triggerDrop = function (dragEndEvent, dragState) {
        var item = dragState.dragItem;
        if (dragState.canDragOut) {
            var event = $.Event('oui-hopper-drop', {
                item: item
            });
            $(dragEndEvent.dropTarget).trigger(event);
        }
    };

    /**
     *
     * @method HopperView#_expandedChanged
     * @param {Array} allExpanded
     * @param {Array} newlyExpanded
     * @param {Array} [newlyCollapsed]
     * @private
     */
    hvp._expandedChanged = function (allExpanded, newlyExpanded, newlyCollapsed) {
        _.each(newlyExpanded, function (id) {
            this._setCollapsed(id, false);
        }.bind(this));
        _.each(newlyCollapsed, function (id) {
            this._setCollapsed(id, true);
        }.bind(this));
    };

    function getLevel(group) { //jscs:ignore jsDoc
        var result = -1;
        var g = group;
        while (g != null) {
            result++;
            g = g.parent;
        }
        return result;
    }

    /**
     *
     * @method HopperView#_setCollapsed
     * @param {Object} id
     * @param {Boolean} state
     * @private
     */
    hvp._setCollapsed = function (id, state) {
        var $elem = this._findItem(id);
        if (groupChainNull(id) && getLevel(id) === this._options.groupBy.length - 1) {
            $elem.toggleClass('collapsed', state);
            $elem.children('[aria-expanded]').attr('aria-expanded', !state);
        } else {
            var active = this._controller.active();
            var newContent = this._renderTree([id], {}, getLevel(id));
            $elem.replaceWith($(newContent));
            this._controller.active(active);
        }
    };

    return HopperView;
});

define('oui/widgets/hopper/hopperController',[
    'underscore'
], function (_) {
    'use strict';

    /**
     * @class HopperController
     * @private
     *
     * @param {Collection} model The model that the controller uses to set and get data.
     * @param {Hopper#Config} options The widgets options object.
     */
    function HopperController(model, options) {

        /**
         * The model for the controller.
         *
         * @private
         * @type {Collection}
         * @memberof! HopperController.prototype
         */
        this._model = model;
        /**
         * The widgets options.
         *
         * @type {Object}
         * @private
         * @memberof! HopperController.prototype
         */
        this._options = options;

        var idToNode = this.node.bind(this);
        model.on({
            'controller.change': function (newData, oldData) { //jscs:ignore jsDoc
                var nodesAdded = _.difference(newData, oldData);
                var nodesRemoved = _.difference(oldData, newData);
                model.trigger('change-delta', {
                    added: _.map(nodesAdded, idToNode),
                    removed: _.map(nodesRemoved, idToNode)
                });
            }
        });
    }

    HopperController.prototype = {
        constructor: HopperController,

        /**
         * Getter and setter for the data in the hopper.
         *
         * @method HopperController#data
         * @param {*} [data] The data to set.
         * @param {Subscribable~EventOptions} eventParams
         * @returns {*} The data.
         */
        data: function (data, eventParams) {
            if (!arguments.length) {
                return this._model.data();
            }

            this._model.data(data, eventParams);
        },

        /**
         * Adds an item to the selection, removing it from appearing in the hopper.
         *
         * @method HopperController#addToData
         * @param {*} id The data to add.
         */
        addToData: function (id) {
            this.data(_.union(this.data(), [id]));
        },

        /**
         * @method HopperController#nodes
         *
         * @param {Object} modifiers
         *
         * @returns {Array<Object>}
         */
        nodes: function (modifiers) {
            var results = this._model.nodes(modifiers);
            if (modifiers && modifiers.unselectedOnly) {
                results = _.reject(results, function (node) {
                    return this._model.is('selected', this._model._getId(node), false);
                }.bind(this));
            }
            return results;
        },

        /**
         * @method HopperController#node
         *
         * @param {String} id
         * @param {Object} modifiers
         *
         * @returns {Object}
         */
        node: function (id, modifiers) {
            return this._model.node(id, modifiers);
        },

        /**
         * Takes an id and marks it as active.
         *
         * @method HopperController#active
         * @param {String} id The id of the node to mark as active.
         */
        active: function (id) {
            if (!arguments.length) {
                return this._model.getActive();
            }

            this._model.active(id);
        },

        /**
         * @method HopperController#isLoaded
         *
         * @returns {Boolean}
         */
        isLoaded: function () {
            return this._model.isState('normal');
        },

        /**
         * @method HopperController#roots
         *
         * @returns {Array<Object>}
         */
        roots: function () {
            return this._model.getRoots();
        },

        /**
         * @method HopperController#getChildren
         *
         * @param {Object} node
         * @param {Object} modifiers
         *
         * @returns {Array<Object>}
         */
        getChildren: function (node, modifiers) {
            var ids;
            if (node.group) {
                ids = this._model.getChildren(node, modifiers);
            } else {
                ids = this._model.getChildren(this._model._getId(node), modifiers);
            }
            return _.reduce(ids, function (memo, id) {
                if (id.group) {
                    memo.push(id);
                } else if (modifiers && modifiers.unselectedOnly && !this._model.is('selected', id, false)) {
                    memo.push(this._model.node(id));
                }
                return memo;
            }.bind(this), []);
        },

        /**
         * Repopulates the hopper
         *
         * @method repopulate
         */
        repopulate: function () {
            this._model._repopulate();
        },

        /**
         * Is a node collapsed
         *
         * @method HopperController#_isCollapsed
         * @private
         *
         * @param {String} id
         *
         * @returns {Boolean}
         */
        _isCollapsed: function (id) {
            return !this._model.is('expanded', id);
        },

        /**
         * Set node collapsed state
         *
         * @method HopperController#_setCollapsed
         * @private
         *
         * @param {String} id
         * @param {Boolean} state
         *
         */
        _setCollapsed: function (id, state) {
            this._model[state ? 'removeFromAttr' : 'addToAttr']('expanded', id);
        },

        /**
         * Checks if a task can be added to a group
         *
         * @method canAddItem
         * @param {Object} group
         * @returns {Boolean}
         */
        canAddItem: function (group) {
            return this._model.canAddItem(group);
        },

        /**
         * Starts the new item adding sequence on the hopper
         *
         * @method newItem
         * @param {Object} group
         */
        newItem: function (group) {
            this._model.newItem(group);
        },

        /**
         * Filters the hopper
         *
         * @method filter
         * @param {String} query
         */
        filter: function (query) {
            if (query == null) {
                this._model.removeFilter();
            } else {
                this._model.addFilter(query);
            }
        },

        /**
         * Scrolls to a particular element
         *
         * @method scrollTo
         * @param {String|Object} id
         */
        scrollTo: function (id) {
            _.each(this._model.getParents(id), function (parent) {
                this._setCollapsed(parent, false);
            }.bind(this));
            this._model.active(id);
            this._model.trigger('view.call-out', id);
        },

        /**
         * Checks if an item can be moved out of the hopper
         *
         * @method canMoveItem
         * @param {Object} item
         * @returns {Boolean}
         */
        canMoveItem: function (item) {
            return this._model.canMoveItem(item);
        },

        /**
         * Triggers an event letting the user mark the item as added to a calendar
         *
         * This assumes that canMoveItem has already been checked.
         *
         * @method HopperController#_triggerAssignItem
         * @param {Object} item
         * @param {{startDate: moment, endDate: moment}|{}} rangeResults
         * @private
         */
        _triggerAssignItem: function (item, rangeResults) {
            this._model.trigger('assign-item', _.extend({
                item: item
            }, rangeResults));
        },

        /**
         * Reorders the nodes in the hopper
         *
         * Only calls updates on numbers that changed
         *
         * Only works if the hopper is ungrouped
         *
         * @method HopperController#_reorderNodes
         * @param {Array<String|Number>} newOrder
         * @private
         */
        _reorderNodes: function (newOrder) {
            // This assumes that you cannot reorder while grouped

            var originalSequenceNumbers = _.map(this.nodes({
                unselectedOnly: true,
                sorted: true
            }), function (node) {
                return this._model._getNodeValue(node, this._options.recordKeys.sequenceNumber, null);
            }.bind(this));

            _.chain(newOrder)
                .zip(originalSequenceNumbers)
                .each(function (zip) {
                    var node = this.node(zip[0]);
                    var originalSequence = this._model._getNodeValue(node, this._options.recordKeys.sequenceNumber, null);
                    if (originalSequence !== zip[1]) {
                        var newData = {};
                        newData[this._options.recordKeys.sequenceNumber] = zip[1];
                        this._model.update(zip[0], newData);
                    }
                }.bind(this))
                .run();

            this.repopulate();
        }
    };

    return HopperController;
});

define('oui/widgets/hopper/hopperNodeRenderer',['underscore'], function (_) {
    'use strict';

    /**
     * @typedef HopperNodeRenderer#RenderData
     *
     * @property {Object} attrs
     * @property {Boolean} attrs.visible
     * @property {Boolean} attrs.expanded
     *
     * @property {Object} data
     *
     * @property {Boolean} matchesSearch
     */

    /**
     * @class HopperNodeRenderer
     */
    function HopperNodeRenderer() {}

    HopperNodeRenderer.prototype = {
        constructor: HopperNodeRenderer,

        /**
         * @method HopperNodeRenderer#render
         *
         * @param {HopperNodeRenderer#RenderData} node
         * @param {Hopper#Config} options
         *
         * @returns {String}
         */
        render: function (node, options) {
            return options.viewFns.display(node.data);
        },

        /**
         * @method HopperNodeRenderer#extend
         *
         * @param {Object} extension
         *
         * @returns {Object}
         */
        extend: function (extension) {
            _.each(extension, function (value, key) {
                if (key.charAt(0) !== '_' || this[key] == null) {
                    this[key] = value;
                }
            }.bind(this));
            return this;
        }
    };

    /**
     * @method HopperNodeRenderer~extend
     *
     * @param {Object} extension
     *
     * @returns {Object}
     */
    HopperNodeRenderer.extend = function (extension) {
        var inst = new HopperNodeRenderer();
        return inst.extend(extension);
    };

    return HopperNodeRenderer;
});

define('oui/widgets/hopper/hopper',['require','underscore','oui/base/widget','oui/widgets/hopper/hopperModel','oui/widgets/hopper/hopperView','oui/widgets/hopper/hopperController','oui/widgets/hopper/hopperNodeRenderer'],function (require) {
    'use strict';

    var _ = require('underscore');
    var Widget = require('oui/base/widget');
    var HopperModel = require('oui/widgets/hopper/hopperModel');
    var View = require('oui/widgets/hopper/hopperView');
    var Controller = require('oui/widgets/hopper/hopperController');
    var Renderer = require('oui/widgets/hopper/hopperNodeRenderer');

    /**
     * @typedef {Object} Hopper#Config
     *
     * @property {Boolean} [sample=false] Testing
     */

    var __DEFAULTS__ = {
        locale: {},
        editable: false,
        recordKeys: {
            id: 'id',
            parentId: 'parentId',
            children: 'children'
        },
        drag: {
            enabled: false,
            out: true,
            reorder: true
        },
        groupBy: [],
        renderer: new Renderer(),
        viewFns: {
            display: function (node) {//jscs:ignore jsDoc
                return '<div>' + node + '</div>';
            }
        }
    };

    /**
     * This class creates a collection of elements which can be grouped and manager.
     *
     * See {@link Hopper#Config} for details regarding the options for the Hopper
     *
     *  ##### **Keyboard Shortcuts**
     *
     *
     *
     *  ##### **Relevant Events**
     *
     *
     *
     * @example
     * <caption>Creating a Hopper</caption>
     *
     * var hopper = new oui.Hopper();
     * hopper.init();
     *
     * @class Hopper
     * @extends Widget
     * @param {Hopper#Config} [options] The options object to use throughout the widget.
     * @param {HTMLInputElement|jQuery} [element] JQuery element to use as the hopper on the DOM.
     * @param {*} [initData] The data to initialize the hopper with.
     */
    function Hopper(options, element, initData) {
        var model;

        //Super call
        Widget.call(this, options, element, initData);

        //Set up default _initData for the Hopper
        /**
         * The initial data used by the widget
         *
         * @memberof! Hopper.prototype
         * @type {*}
         * @private
         */
        this._initData = this._initData || {
                available: [],
                selected: []
            };

        model = new HopperModel(this._options, this._initData);
        /**
         * The controller object for the widget.
         *
         * @memberof! Hopper.prototype
         * @private
         * @type {HopperController}
         */
        this._controller = new Controller(model, this._options);
        /**
         * The view for the widget.
         *
         * @memberof! Hopper.prototype
         * @private
         * @type {HopperView}
         */
        this._view = new View(this._options, this._element, this._controller, this._settings, model);

        Object.defineProperty(this, '_data', {
            get: this.data,
            set: this.data
        });

        this._setupEvents([model, this._view]);
    }

    Hopper.prototype = Object.create(Widget.prototype);
    var hp = Hopper.prototype;

    hp.constructor = Hopper;

    /**
     * Returns the name of the widget that is used in eventing and id generation
     *
     * @method Hopper#name
     * @returns {String} 'hopper' - the widgets name
     */
    hp.name = function () {
        return 'hopper';
    };

    /**
     * Getter and setter for the data in the hopper.
     *
     * @method Hopper#data
     * @param {*} [data] The data to set.
     * @returns {*} The data.
     */
    hp.data = function (data) {
        if (!arguments.length) {
            return _.map(this._controller.data(), this._controller.node.bind(this._controller));
        }

        data = _.compact(data);
        data = _.reduce(data, function (memo, node) {
            var fullNode = this._controller.node(node);
            if (fullNode) {
                memo.push(this._controller._model._getNodeValue(fullNode, this._options.recordKeys.id));
            }
            return memo;
        }.bind(this), []);

        this._controller.data(data);
    };

    /**
     * Adds an item to the selection, removing it from appearing in the hopper.
     *
     * @method Hopper#addToData
     * @param {*} id The data to add.
     */
    hp.addToData = function (id) {
        if (!_.isArray(id)) {
            id = [id];
        }
        id = _.compact(id);
        id = _.reduce(id, function (memo, node) {
            var fullNode = this._controller.node(node);
            if (fullNode) {
                memo.push(this._controller._model._getNodeValue(fullNode, this._options.recordKeys.id));
            }
            return memo;
        }.bind(this), []);
        this._controller.data(_.union(this._controller.data(), id));
    };

    /**
     * Removes an item from the selection, making it appear in the hopper.
     *
     * @method Hopper#removeFromData
     * @param {*} id The data to remove.
     */
    hp.removeFromData = function (id) {
        if (!_.isArray(id)) {
            id = [id];
        }
        id = _.compact(id);
        id = _.reduce(id, function (memo, node) {
            var fullNode = this._controller.node(node);
            if (fullNode) {
                memo.push(this._controller._model._getNodeValue(fullNode, this._options.recordKeys.id));
            }
            return memo;
        }.bind(this), []);
        this._controller.data(_.difference(this._controller.data(), id));
    };

    /**
     * @method Hopper#active
     *
     * @returns {Object}
     */
    /**
     * @method Hopper#active
     *
     * @param {String} id
     */
    hp.active = function (id) {
        if (!arguments.length) {
            return this._controller.node(this._controller.active());
        }

        this._controller.active(id);
    };

    /**
     * Getter and setter for the data in the hopper.
     *
     * See {@link Hopper#data} for information regarding the similar method.
     *
     * **Note**: does not trigger an event notifying of a change.
     *
     * @method Hopper#value
     * @param {*} [data] The data to set.
     * @returns {*} The selected data.
     */
    hp.value = function (data) {
        if (!arguments.length) {
            return _.map(this._controller.data(), this._controller.node.bind(this._controller));
        }

        data = _.compact(data);
        data = _.reduce(data, function (memo, node) {
            var fullNode = this._controller.node(node);
            if (fullNode) {
                memo.push(this._controller._model._getNodeValue(fullNode, this._options.recordKeys.id));
            }
            return memo;
        }.bind(this), []);

        this._controller.data(data, {
            silent: true
        });
    };

    /**
     * Ability to retrieve the widgets default options for use in merging with passed in options as
     * well as with composed widgets such as the date and time picker.
     *
     * @method Hopper#defaults
     * @returns {Hopper#Config} - the widgets default options
     */
    hp.defaults = function () {
        return __DEFAULTS__;
    };

    /**
     * Getter or setter for the disable state of the hopper
     *
     * @method Hopper#disable
     * @param {Boolean} [disabled] Whether to disable the view or now. True to disable and false to enable
     * @returns {Boolean} - Whether the view is currently disabled
     */
    hp.disable = function (disabled) {
        if (!arguments.length) {
            return this._view.isDisabled();
        }

        this._view.disable(disabled);
    };

    /**
     * Performs a search on the data in the hopper.
     *
     * By default only the class `search-result` is applied to the found item and not CSS styles that.
     * You can style that on a per-page basis or override the renderer customize the bucket item as the renderer
     * provides feedback on whether the current item matches the search results.
     *
     *     render: function (args, options) {
     *         if (args.matchesSearch) {
     *             // Do something different
     *         }
     *     }
     *
     * @method Hopper#filter
     * @param {String} query
     */
    hp.filter = function (query) {
        this._controller.filter(query);
    };

    /**
     * Checks the parameter for whether it is init data.
     *
     * @method Hopper#isInitData
     * @private
     * @param {Object|*} data The data to check existence of initialization data properties
     * @returns {Boolean} whether the data is initData
     */
    hp.isInitData = function (data) {
        return data.available || data.selected;
    };

    /**
     * Repopulates the hopper
     *
     * @method repopulate
     */
    hp.repopulate = function () {
        this._controller.repopulate();
    };

    /**
     * Links the hopper to a calendar component for drag support
     *
     * @method Hopper#linkCalendar
     * @param {Calendar} calendar
     */
    hp.linkCalendar = function (calendar) {
        var alreadyLinked = this._view._associateCalendar(calendar);
        if (!alreadyLinked) {
            calendar.linkHopper(this);
        }
    };

    /**
     * Scrolls to a particular item in the hopper
     *
     * @method Hopper#scrollIntoView
     * @param {String|Object} id
     * @param {Boolean} atTop
     */
    hp.scrollIntoView = function (id, atTop) {
        if (_.isObject(id)) {
            id = this._controller._model._getId(id);
        }
        this._controller.scrollTo(id, atTop);
    };

    /**
     * Scrolls to a particular group
     *
     * @method Hopper#scrollToGroup
     * @param {{field:String,id:String}|Array<{field:String,id:String}>} groupIdentifier
     * @param {Boolean} atTop
     */
    hp.scrollGroupIntoView = function (groupIdentifier, atTop) {
        var group;
        if (_.isArray(groupIdentifier)) {
            group = _.reduce(groupIdentifier, function (memo, groupId) {
                return this._controller._model._makeGroup(groupId.field, groupId.id, memo);
            }.bind(this), null);
        } else {
            group = this._controller._model._makeGroup(groupIdentifier.field, groupIdentifier.id);
        }
        this._controller.scrollTo(group, atTop);
    };

    /**
     * A reference to the default renderer for the hopper bucket items.
     *
     * @property Hopper~Renderer
     */
    Hopper.Renderer = Renderer;

    return Hopper;
});

define('oui/utils/searcher',['require','underscore','oui/utils/stringUtils'],function (require) {
    'use strict';

    var _ = require('underscore');
    var stringUtils = require('oui/utils/stringUtils');

    var __DEFAULTS__ = {
        smartSearch: true,
        text: _.identity
    };

    // jscs:disable maximumLineLength

    //At least a capital letter without matches
    var hasCapital = /(?:[\0-KM-km-oq-z\|~-\uD7FF\uE000-\uFFFF]|[\uD800-\uDBFF][\uDC00-\uDFFF]|[\uD800-\uDBFF](?![\uDC00-\uDFFF])|(?:[^\uD800-\uDBFF]|^)[\uDC00-\uDFFF])/;

    //All capitals with numbers and groupings
    var splitOn = /((?:[\0-/:-KM-km-oq-z\|~-\uD7FF\uE000-\uFFFF]|[\uD800-\uDBFF][\uDC00-\uDFFF]|[\uD800-\uDBFF](?![\uDC00-\uDFFF])|(?:[^\uD800-\uDBFF]|^)[\uDC00-\uDFFF]))/;

    //Only capital letters without grouping matches
    var onlyCapitals = /^(?:[\0-/:-KM-km-oq-z\|~-\uD7FF\uE000-\uFFFF]|[\uD800-\uDBFF][\uDC00-\uDFFF]|[\uD800-\uDBFF](?![\uDC00-\uDFFF])|(?:[^\uD800-\uDBFF]|^)[\uDC00-\uDFFF])+$/;

    // jscs:enable maximumLineLength

    return function (options) {

        options = _.merge({}, __DEFAULTS__, options);

        return function (query, collection) {
            var regexes = [];
            var strippedQuery = stringUtils.regexEscape(query);

            if (hasCapital.test(query) && options.smartSearch) {
                //Case sensitive
                regexes.push(new RegExp(strippedQuery, 'g'));
            } else {
                regexes.push(new RegExp(strippedQuery, 'gi'));
            }

            if (onlyCapitals.test(query) && options.smartSearch) {
                regexes.push(new RegExp(strippedQuery.split(splitOn).join('.*'), 'g'));
            }

            return _.filter(collection, function (item) {
                return _.any(regexes, function (regex) {
                    return options.text(item).match(regex);
                });
            });
        };
    };
});

define('oui/widgets/quickNav',[
    'underscore',
    'jquery',
    'oui/widgets/typeAhead',
    'oui/utils/stringUtils',
    'oui/utils/searcher'
], function (_, $, TypeAhead, stringUtils, searcher) {
    'use strict';

    /**
     * @module quickNav
     */

    /**
     * @typedef module:quickNav#Config
     *
     * @property {String} [style='navigation']
     * @property {Boolean} [tokenize=false]
     * @property {Boolean} [highlight=false]
     * @property {Object} storeFns
     * @property {Function} storeFns.search
     * @property {Object} viewFns
     * @property {Function} viewFns.display
     * @property {Function} viewFns.text
     * @property {Object} events
     * @property {Function} events.select
     */
    var __DEFAULTS__ = {
        style: 'navigation',
        tokenize: false,
        highlight: false,
        storeFns: {
            search: function (query) { //jscs:ignore jsDoc
                return searcher({
                    text: function (node) { //jscs:ignore jsDoc
                        return node.textContent;
                    }
                })($('a[href], button:visible'), query);
            }
        },
        viewFns: {
            display: function (node) { //jscs:ignore jsDoc
                return $(node).text();
            },
            text: function (node) { //jscs:ignore jsDoc
                return $(node).text();
            }
        },
        events: {
            select: function (nodes) { //jscs:ignore jsDoc
                if (nodes.length === 1) {
                    _.first(nodes).click();
                }
            }
        }
    };

    /**
     * @alias module:quickNav
     * @method quickNav
     * @private
     * @param {module:quickNav#Config} options
     */
    function quickNav(options) {
        options = _.merge({}, __DEFAULTS__, options, { tokenize: false, style: 'navigation' });

        var typeAhead = new TypeAhead(options).init();
        typeAhead.on('select', function () {
            this.value([]);
        });
    }

    return quickNav;
});

define('oui/utils/is',[
    'jquery',
    'underscore'
], function ($, _) {
    'use strict';

    /**
     * Module to determine whether an element is an oui control.
     *
     * @module is
     * @private
     */

    var OUI_CLASSES = [
        'orcl-date-picker',
        'orcl-picker',
        'orcl-picker-input',
        'orcl-typeahead',
        'tree',
        'orcl-number-field',
        'orcl-editor',
        'orcl-file-picker',
        'orcl-combo-box',
        'combo-box-list',
        'orcl-color-picker',
        'orcl-icon-picker',
        'orcl-url-field',
        'orcl-context-menu',
        'orcl-calendar',
        'orcl-toolbar',
        'orcl-slider',
        'orcl-form-canvas',
        'orcl-hopper'
    ];

    /**
     * Determine whether an element is an oui control
     *
     * @exports is
     * @type Function
     * @param {jQuery|HTMLElement} $elem The elem to check against for an existence of OUI class.
     * @returns {Boolean} - Whether the element was an OUI control.
     */
    return function ($elem) {
        if (!$elem.jquery) {
            $elem = $($elem);
        }

        if ($elem.length > 1) {
            //TODO: WARN there are multiple elements requested?
            return false;
        } else {
            var hasClass = _.bind($elem.hasClass, $elem);
            var parentHasClass = _.bind($elem.hasClass, $elem.parent());
            return _.any(OUI_CLASSES, hasClass) ||
                _.any(OUI_CLASSES, parentHasClass);
        }
    };
});


define('template-115', ['underscore'], function (_) { return function(data) {
var __t, __p = '';
__p += '<div class="row-fluid"> <div class="span6 field-path"></div> <div class="span3 field-arg"></div> <div class="span3 field-function"></div></div>';
return __p
};});

define('oui/widgets/logicalDate/logicalDateConstants',{
    TYPES: {
        PATH: 'PATH',
        USER_VALUE: 'USER_VALUE',
        COMPUTED_DATE: 'COMPUTED_DATE'
    },
});

define('oui/widgets/logicalDate/logicalDateModel',['require','underscore','moment','oui/base/model','./logicalDateConstants'],function (require) {
    'use strict';

    var _ = require('underscore');
    var moment = require('moment');
    var BaseModel = require('oui/base/model');
    var TYPES = require('./logicalDateConstants').TYPES;

    /**
     * Checks rules for numeric values associated with the number spinner of this input.
     *
     * @private
     * @param {Number} value The value of the spinner
     */
    function numericRules(value) {
        if (value == null) {
            return {
                valid: false,
                type: 'null',
                data: null
            };
        }
        if (!_.isNumber(value)) {
            return {
                valid: false,
                type: 'not-a-number',
                data: value
            };
        }
        if (Math.floor(value) !== value) {
            return {
                valid: false,
                type: 'fractional',
                data: value
            };
        }
    }

    /**
     * Coerces the given data into something valid for this input, if it wasn't already.
     *
     * @private
     * @param {Object} data Data given by user
     * @returns {Object} data Valid data
     */
    function defaultData(data) {
        if (!data || !TYPES[data.type]) {
            return {
                choice: TYPES.USER_VALUE,
                type: TYPES.USER_VALUE,
                data: moment().startOf('day')
            };
        }
        return data;
    }

    /**
     * Model class for the logical date widget.
     *
     * @private
     * @extends Model
     */
    function LogicalDateModel(initData, options) {
        BaseModel.call(this, defaultData(initData), options);
    }

    LogicalDateModel.prototype = _.create(BaseModel.prototype, {

        validate: function (data) {// jscs:ignore jsDoc
            if (!data || !TYPES[data.type]) {
                return {
                    valid: false,
                    type: 'illegal-type',
                    data: data
                };
            }

            if (data.type === TYPES.USER_VALUE) {
                if (data.data == null) {
                    return {
                        valid: false,
                        type: 'null',
                        data: null
                    };
                }

                if (!moment.isMoment(data.data) || !data.data.isValid()) {
                    return {
                        valid: false,
                        type: 'not-a-date',
                        data: data
                    };
                }
            } else {
                var counterValue = data.data;
                var error = numericRules(counterValue);
                if (error) {
                    error.data = data;
                    return error;
                }
            }
        }

    });

    LogicalDateModel.TYPES = TYPES;
    LogicalDateModel.numericRules = numericRules;

    return LogicalDateModel;
});

define('oui/widgets/logicalDate/logicalDateView',['require','underscore','jquery','moment','oui/base/view','template-115','oui/widgets/comboBox','oui/widgets/dateField','oui/widgets/numberField','oui/utils/dateHelper','./logicalDateConstants','./logicalDateModel'],function (require) {
    'use strict';

    var _ = require('underscore');
    var $ = require('jquery');
    var moment = require('moment');
    var BaseView = require('oui/base/view');
    var widgetTemplate = require('template-115');
    var ComboBox = require('oui/widgets/comboBox');
    var DateField = require('oui/widgets/dateField');
    var NumberField = require('oui/widgets/numberField');
    var dateHelper = require('oui/utils/dateHelper');

    var TYPES = require('./logicalDateConstants').TYPES;
    var numberValidator = require('./logicalDateModel').numericRules;

    /**
     * Returns an array of valid options for +- a given field path / logical date.
     *
     * @private
     * @param {LogicalDate#Config} options The widget options
     */
    function getDateArithmeticOptions(options) {
        var strings = _.get(options, 'locale.logicalDate');
        return [{
            type: 'DAYS',
            label: strings.plusDay
        }, {
            type: 'WEEKS',
            label: strings.plusWeek
        }, {
            type: 'MONTHS',
            label: strings.plusMonth
        }, {
            type: 'QUARTERS',
            label: strings.plusQuarter
        }, {
            type: 'YEARS',
            label: strings.plusYear
        }];
    }

    /**
     * View class for the logical date widget.
     *
     * @private
     * @extends View
     */
    function LogicalDateView(model, controller, element, options, settings) {
        BaseView.call(this, options, element, controller, settings);

        model.on({
            change: function () {
                this._renderArgFunction();
            }.bind(this),
            error: function () {// jscs:ignore jsDoc
                this._renderArgFunction();
            }.bind(this)
        });

        this._$path = this._$el.find('.field-path');
        this._$arg = this._$el.find('.field-arg');
        this._$function = this._$el.find('.field-function');
    }

    LogicalDateView.prototype = _.create(BaseView.prototype, {

        /**
         * @method LogicalDateView#buildElement
         *
         * @returns {jQuery}
         */
        buildElement: function () {
            return $(widgetTemplate());
        },

        /**
         * @method LogicalDateView#init
         */
        init: function () {
            this._initPath();
            this._renderArgFunction();
        },

        /**
         * @method LogicalDateView#destroy
         */
        destroy: function () {
            this._clear();
            if (this._pathCombo) {
                this._pathCombo.destroy();
                this._pathCombo = null;
            }
            this._$el.remove(); // TODO replace w/ call to view
        },

        /**
         * Updates data & triggers change for path changes.
         *
         * @private
         * @param {Object} comboChoice given by the path combo event
         */
        _onPathChanged: function (comboChoice) {
            this._controller.data(dateHelper.cloneWithDates(comboChoice));
        },

        /**
         * Updates data & triggers change for custom date changes.
         *
         * @private
         */
        _onCustomPickerChanged: function () {
            var newData = _.defaults({
                data: this._customPicker.data(),
            }, this._controller.data());
            this._controller.data(newData);
        },

        /**
         * Updates data & triggers change for custom numeric changes.
         *
         * @private
         */
        _onNumberArgChanged: function () {
            var newData = _.defaults({
                data: this._numberArg.data(),
            }, this._controller.data());
            this._controller.data(newData);
        },

        /**
         * Updates data & triggers change for date function type changes.
         *
         * @private
         */
        _onFunctionChanged: function () {
            var newData = _.defaults({
                dataFunction: this._functionCombo.data().type,
            }, this._controller.data());
            this._controller.data(newData);
        },

        /**
         * Typically called during render - destorys internal subinputs controlled by this input.
         *
         * @private
         */
        _clear: function () {
            if (this._customPicker) {
                this._customPicker.destroy(true);
                this._customPicker = null;
            }
            if (this._numberArg) {
                this._numberArg.destroy(true);
                this._numberArg = null;
            }
            if (this._functionCombo) {
                this._functionCombo.destroy(true);
                this._functionCombo = null;
            }

            this._$arg.empty();
            this._$function.empty();
        },

        /**
         * If the current path selection is not USER_VALUE, initializes the spinner/combo for this input.
         *
         * @private
         */
        _renderLogicalDateType: function () {
            this._clear();
            var data = this._controller.data();
            this._numberArg = new NumberField({
                spinner: {
                    controls: true
                },
                fluid: true,
                emptyValue: 0,
                validators: [numberValidator],
                formatter: this._options.formatter
            });
            data.data = data.data || 0;
            this._numberArg.data(data.data);
            this._$arg.append(this._numberArg.element());
            this._numberArg.on('change', this._onNumberArgChanged.bind(this));
            this._numberArg.on('error', this._onNumberArgChanged.bind(this));
            this._numberArg.init();
            var dateOptions = getDateArithmeticOptions(this._options);
            this._functionCombo = new ComboBox({
                keyboardInput: false,
                allowCustomValues: false,
                fluid: true,
                allowEmpty: false,
                viewFns: {
                    display: _.property('label')
                },
                recordKeys: {
                    id: 'type'
                },
                storeFns: {
                    value: _.property('type')
                }
            }, dateOptions);
            data.dataFunction = data.dataFunction || dateOptions[0].type;
            this._functionCombo.data({
                type: data.dataFunction
            });
            this._$function.append(this._functionCombo.element());
            this._functionCombo.on('selection-change', this._onFunctionChanged.bind(this));
            this._functionCombo.init();
        },

        /**
         * Initializes the argument input(s) associated with this input.
         *
         * @private
         */
        _renderArgFunction: function () {
            var data = this._controller.data();
            var newType = data.type;
            var newChoice = data.choice;
            if (newType !== this._currentArgType || newChoice !== this._currentArgChoice) {
                this._currentArgType = newType;
                this._currentArgChoice = newChoice;

                var isCustomDate = newType === TYPES.USER_VALUE;
                this._$function.toggleClass('hidden', isCustomDate).toggle(!isCustomDate);

                if (isCustomDate) {
                    this._$arg.removeClass('span3').addClass('span5');
                    this._renderCustomPicker();
                } else {
                    this._$arg.removeClass('span5').addClass('span3');
                    this._renderLogicalDateType();
                }
            }
        },

        /**
         * If the current path selection is USER_VALUE, initializes the custom date picker.
         *
         * @private
         */
        _renderCustomPicker: function () {
            this._clear();
            var data = this._controller.data();
            this._customPicker = new DateField({
                openOnFocus: true,
                type: 'date',
                fluid: true
            });
            this._customPicker.data(data.data);
            this._customPicker.on('change', this._onCustomPickerChanged.bind(this));
            this._customPicker.on('error', this._onCustomPickerChanged.bind(this));
            this._$arg.append(this._customPicker.element());
            this._customPicker.init();
        },

        /**
         * Initializes the path elements associated with this input.
         *
         * @private
         */
        _initPath: function () {
            var customDateChoice = {
                choice: TYPES.USER_VALUE,
                label: _.get(this._options, 'locale.logicalDate.customDay'),
                type: TYPES.USER_VALUE,
                data: moment().startOf('day')
            };
            var data = this._controller.data();
            if (data.type === TYPES.USER_VALUE) {
                customDateChoice.data = data.data && data.data.clone();
            }

            var pathChoices = [customDateChoice];
            if (this._options.choices) {
                pathChoices = pathChoices.concat(this._options.choices);
            }

            this._pathChoices = pathChoices;
            this._pathCombo = new ComboBox({
                keyboardInput: false,
                allowCustomValues: false,
                fluid: true,
                allowEmpty: false,
                viewFns: {
                    display: _.property('label')
                },
                recordKeys: {
                    id: 'choice'
                },
                storeFns: {
                    value: _.property('choice')
                }
            }, pathChoices);
            this._pathCombo.selected({
                choice: data.choice
            });
            this._pathCombo.on('selection-change', this._onPathChanged.bind(this));
            this._$path.append(this._pathCombo.element());
            this._pathCombo.init();
        },

    });

    return LogicalDateView;
});

define('oui/widgets/logicalDate/logicalDateController',[], function () {
    'use strict';

    /**
     * @class LogicalDateController
     * @private
     *
     * @param {LogicalDateModel} model
     * @param {LogicalDate#Config} options
     *
     */
    function LogicalDateController(model, options) {
        this._model = model;
        this._options = options;
    }

    LogicalDateController.prototype = {

        /**
         * @method LogicalDateController#isValid
         *
         * @returns {Boolean}
         */
        isValid: function () {
            return this._model.isValid();
        },

        /**
         * @method LogicalDateController#error
         *
         * @returns {Object}
         */
        error: function () {
            return this._model.error();
        },

        /**
         * @method LogicalDateController#data
         *
         * @returns {Object}
         */
        /**
         * @method LogicalDateController#data
         *
         * @param {Object} data
         * @param {Subscribable~EventOptions} [eventParams]
         */
        data: function (data, eventParams) {
            if (!arguments.length) {
                return this._model.data();
            }

            this._model.data(data, eventParams);
        }
    };

    return LogicalDateController;
});

define('oui/widgets/logicalDate/logicalDate',['require','underscore','oui/base/widget','logger','./logicalDateView','./logicalDateController','./logicalDateModel'],function (require) {
    'use strict';

    var _ = require('underscore');
    var Widget = require('oui/base/widget');
    var logger = require('logger');
    var LogicalDateView = require('./logicalDateView');
    var LogicalDateController = require('./logicalDateController');

    var Model = require('./logicalDateModel');
    var TYPES = Model.TYPES;

    /**
     * Options for the logical date widget.
     *
     * @typedef {Object} LogicalDate#Config
     *
     * @property {Locale} [locale={}] The specific strings that are used by the widget and will use the defaults set from
     * the {@link module:oui.lang|lang}.
     */
    var __userDefaults = {};

    /**
     * This class represents a logical date widget that can handle rolling date filters.
     *
     * Logical date widgets are used in filter components, and look like "'Custom Date' <date picker>" or "'Current Day' '-5' <day/week/etc combobox>".
     *
     * @class LogicalDate
     * @extends Widget
     *
     * @param {LogicalDate#Config} [options] The widget options
     * @param {HTMLInputElement|jQuery} [parentElem] An element to append this widget to
     * @param {Object} [data] Data to initialize with
     */
    function LogicalDateWidget(options, parentElem, data) {
        Widget.call(this, options, data);

        var model = new Model(data, this._options);
        Object.defineProperty(this, '_data', {
            get: model.data.bind(model),
            set: model.data.bind(model)
        });

        this._controller = new LogicalDateController(model, this._options);
        this._view = new LogicalDateView(model, this._controller, this._element, this._options, this._settings);

        this._setupEvents([model]);
    }

    LogicalDateWidget.prototype = _.create(Widget.prototype, {
        // docs inherited
        defaults: function () { // jscs:ignore jsDoc
            return _.defaults({}, __userDefaults);
        },

        // TODO replace with this._view functionality, inherit from superclass
        /**
         * Disabling the logical date widget is not supported.  Calls to this method will be ignored.
         *
         * @method LogicalDate#disable
         */
        disable: function () {
            logger.warn('Disabling logicalDate widgets is not supported - skipping call');
        },

        // docs inherited
        isValid: function () { // jscs:ignore jsDoc
            return this._controller.isValid();
        },

        /**
         * Gets the error of this field, if currently in an invalid state.
         *
         * @method LogicalDate#error
         * @returns {Object} error The error of the field.
         */
        error: function () {
            if (!this.isValid()) {
                return this._controller.error();
            }
        },
    });

    LogicalDateWidget.TYPES = TYPES;

    /**
     * Updates the default options used for future LogicalDate inputs.
     *
     * @static
     * @param {Object} newDefaults
     */
    LogicalDateWidget.defaults = function (newDefaults) {
        _.assign(__userDefaults, newDefaults);
    };

    return LogicalDateWidget;
});

/**
 * @namespace orcl.picker
 */
define('oui/widgets/picker/pickerWidget',[
    'jquery',
    'oui/widgets/picker',
    'jqueryui'
], function ($, Picker) {
      'use strict';

      $.widget('orcl.picker', {

          /**
           * @method orcl.picker#_create
           * @private
           */
          _create: function () {
              var that = this;

              this.widgetEventPrefix = this.widgetEventPrefix + '-';

              this._picker = new Picker(this.options, this.element);
              this._picker.init();

              this._picker.on('select', function (node) {
                  that._trigger('select', null, node);
              });

              this._picker.on('accept', function (data) {
                  //TODO: look into why the jquery event
                  //with multiselect unwraps the array to make a new
                  //array which then loses all but the first element
                  that._trigger('accept', null, [data]);
              });
          },

          /**
           * @method orcl.picker#selected
           * @param {Array} data The array of data to be selected in the picker.
           * @returns {Array} - The array of selected nodes in the picker.
           */
          selected: function (data) {
              if (!arguments.length) {
                  return this._picker.selected();
              }

              this._picker.selected(data);
          },

          /**
           * @method orcl.picker#unselectable
           *
           * @returns {Array}
           */
          /**
           * @method orcl.picker#unselectable
           *
           * @param {Array} data
           */
          unselectable: function (data) {
              if (!arguments.length) {
                  return this._picker.unselectable();
              }

              this._picker.unselectable(data);
          },

          /**
           * @method orcl.picker#expanded
           *
           * @returns {Array}
           */
          /**
           * @method orcl.picker#expanded
           *
           * @param {Array} data
           */
          expanded: function (data) {
              if (!arguments.length) {
                  return this._picker.expanded();
              }

              this._picker.expanded(data);
          },

          /**
           * @method orcl.picker#disable
           *
           * @param {Boolean} disabled
           */
          disable: function (disabled) {
              this._picker.disable(disabled);
          },

          /**
           * @method orcl.picker#disabled
           *
           * @returns {Array}
           */
          /**
           * @method orcl.picker#disabled
           *
           * @param {Array} data
           */
          disabled: function (data) {
              if (!arguments.length) {
                  return this._picker.disabled();
              }

              this._picker.disabled(data);
          },

          /**
           * @method orcl.picker#_destroy
           * @private
           */
          _destroy: function () {
              this._picker.destroy({
                  self: false
              });
              delete this._picker;
          }
      });

  }
);


define('template-116', ['underscore'], function (_) { return function(data) {
var __t, __p = '';
__p += '<div class="splitter" tabindex="0" role="separator"> <a href="#" class="splitter-inner collapse-handle"></a></div>';
return __p
};});

define('oui/widgets/splitPane/splitPaneWidget',['require','jquery','underscore','oui/utils/inlineStyle','template-116','oui/utils/language','oui/utils/keyCode','logger','jqueryui'],function (require) {
      'use strict';

      var $ = require('jquery');
      var _ = require('underscore');
      var inlineStyle = require('oui/utils/inlineStyle');
      var splitterTemplate = require('template-116');
      var language = require('oui/utils/language');
      var keyCode = require('oui/utils/keyCode');
      var logger = require('logger');
      require('jqueryui');

      /**
       * A basic split pane widget that allows dynamic resizing of two HTML elements in a container
       *
       * @namespace orcl.splitPane
       */

      /**
       * Removes a CSS property from an object and stores it to be retrieved later.
       *
       * @method orcl.splitPane~saveOriginalCSSValue
       * @private
       * @param {*|HTMLElement} element The element to operate on
       * @param {String} property The property to store
       */
      function saveOriginalCSSValue(element, property) {
          element.data('sp-original-' + property, inlineStyle.getPropertyValue(element, property));
      }

      /**
       * Saves the flex values internally to be restored when no longer collapsed
       *
       * @method orcl.splitPane~saveFlexValues
       * @private
       * @param {jQuery} element
       */
      function saveFlexValues(element) {
          _.each(['flex-basis', 'flex-shrink', 'flex-grow'], function (attr) {

              element.data('sp-original-' + attr, element.css(attr));
          });
      }

      /**
       * Restore the flex values onto fixed split panes when restored. The values may not exist and will
       * be set to their defaults.
       *
       * @method orcl.splitPane~restoreFlexValues
       * @private
       * @param {jQuery} element
       */
      function restoreFlexValues(element) {
          _.each(['flex-basis', 'flex-shrink', 'flex-grow'], function (attr) {
              element.css(attr, element.data('sp-original-' + attr) || '');
              $.removeData(element, 'sp-original-' + attr);
          });
      }

      /**
       * Restores a CSS property for an object that had its property stored earlier.
       *
       * @method orcl.splitPane~restoreOriginalCSSValue
       * @private
       * @param {*|HTMLElement} element The element to operate on
       * @param {String} property The property to retrieve
       */
      function restoreOriginalCSSValue(element, property) {
          element.css(property, element.data('sp-original-' + property) || '');
          $.removeData(element, 'sp-original-' + property);
      }

      /**
       * Checks to see if a CSS property is stored.
       *
       * @method orcl.splitPane~hasOriginalCSSValue
       * @private
       * @param {*|HTMLElement} element The element to operate on
       * @param {String} property The property to store
       */
      function hasOriginalCSSValue(element, property) {
          return !_.isUndefined(element.data('sp-original-' + property));
      }

      $.widget('orcl.splitPane', {

          /**
           * @typedef {Object} orcl.splitPane~Config
           *
           * @property {String} [orientation='horizontal'] The direction of the split pane. "horizontal" split panes can be
           * resied up or down; "vertical" split panes can be resized left or right.
           *
           * @property {Boolean} [collapseFirst=false] Determines whether the left or top child of the split pane can be
           * collapsed.
           *
           * @property {Boolean} [collapseSecond=false] Determines whether the right or bottom child of the split pane can
           * be collapsed.
           *
           * @property {Number} [firstSplitSize=0.5] The split ratio of the split pane, represented as the relative size of
           * the first child pane to the size of the container. For example, a 50:50 split would have a `firstSplitSize` of
           * 0.5. A 1:2 split would have a `firstSplitSize` of 0.33333333333.
           *
           * @property {Number} [throttleSpeed=500] The rate at which the secondary resize event, 'split-resize-throttled'
           * is fired, in milliseconds.
           *
           * @property {Locale} [locale={}] The specific strings that are used by the widget and will use the defaults set from
           * the {@link module:oui.lang|lang}.
           *
           * @property {Boolean} [enabled=true] Whether or not the splitter and it's components are enabled
           * @property {String} [collapseDirection] Can be left blank, 'right' or 'top'
           */

          /**
           * @memberof! orcl.splitPane
           * @property {orcl.splitPane~Config} options
           */
          options: {
              orientation: 'horizontal',
              collapseFirst: false,
              collapseSecond: false,
              firstSplitSize: 0.5,
              throttleSpeed: 500,
              locale: language.lang(),
              enabled: true
          },

          /**
           * Makes sure that a new split pane only contains two children.
           *
           * @method orcl.splitPane#_validateElement
           * @private
           */
          _validateElement: function () {
              if (this.element.children().length !== 2) {
                  var message = 'A split pane container must only contain two child elements';
                  logger.fatal(message);
                  throw new Error(message);
              }
          },

          /**
           * Returns the CSS property that matches the direction the split pane resizes.
           *
           *     horizontal => height
           *     vertical   => width
           *
           * @method orcl.splitPane#_getSplitSizeProperty
           * @returns {String}
           * @private
           */
          _getSplitSizeProperty: function () {
              return this.options.orientation === 'horizontal' ? 'height' : 'width';
          },

          /**
           * @method orcl.splitPane#_getSplitterSizeProperty
           * @private
           * @returns {String}
           */
          _getSplitterSizeProperty: function () {
              return this.options.orientation === 'horizontal' ? 'outerHeight' : 'outerWidth';
          },

          /**
           * Returns the CSS property that is the opposite direction that the split pane resizes.
           *
           *     horizontal => width
           *     vertical   => height
           *
           * @method orcl.splitPane#_getOppositeSizeProperty
           * @returns {String}
           * @private
           */
          _getOppositeSizeProperty: function () {
              return this.options.orientation === 'horizontal' ? 'width' : 'height';
          },

          /**
           * Creates the splitter `<div>`
           *
           * @method orcl.splitPane#_createSplitter
           * @returns {*|HTMLElement}
           * @private
           */
          _createSplitter: function () {
              return $(splitterTemplate(this.options.locale));
          },

          /**
           * @method orcl.splitPane#_removeMouseHandlers
           * @private
           */
          _removeMouseHandlers: function () {
              $(document).off(this.mouseNamespace);
          },

          /**
           * Sets up the event listeners for the splitter resizing
           *
           * @method orcl.splitPane#_setupHandlers
           * @private
           */
          _setupHandlers: function () {
              var that = this;
              var lastX;
              var lastY;

              this.$splitter
              .on('mousedown', function (event) {
                  event.preventDefault();
                  event.stopPropagation();
                  var $target = $(event.target);
                  var prevent = function (event) { //jscs:ignore jsDoc
                      event.preventDefault();
                      event.stopPropagation();
                  };

                  if (that.options.enabled && !that.options.preventDrag) {
                      $('body').addClass('oui-disable-select');
                      lastX = event.pageX;
                      lastY = event.pageY;

                      /**
                       *
                       * @method orcl.splitPane#_setupHandlers~moveHandler
                       * @private
                       *
                       * @param {jQuery.Event} event
                       */
                      var moveHandler = function (event) {
                          if (that.options.orientation === 'horizontal') {
                              that._mouseSetSplit(event.pageY - lastY);
                          } else if (that.options.orientation === 'vertical') {
                              that._mouseSetSplit(event.pageX - lastX);
                          }

                          if (!that._splitterEdge) {
                              lastX = event.pageX;
                              lastY = event.pageY;
                          }
                      };

                      $target.off(that.mouseNamespace);

                      $(document)
                          .on('mousemove' + that.mouseNamespace, moveHandler)
                          .on('mousemove' + that.mouseNamespace, _.once(function () {
                              $target.one('click' + that.mouseNamespace, prevent);
                              $target.on('mouseleave' + that.mouseNamespace, function () {
                                  $target.off('click' + that.mouseNamespace);
                              });
                              $target.on('mouseenter' + that.mouseNamespace, function () {
                                  $target.one('click' + that.mouseNamespace, prevent);
                              });

                          }))
                          .on('click' + that.mouseNamespace, '.collapse-handle', function (event) {
                              event.preventDefault();
                          })
                          .on('mouseleave' + that.mouseNamespace, function () {
                              $('body').removeClass('oui-disable-select');
                              that._removeMouseHandlers();
                          })
                          .on('mouseup' + that.mouseNamespace, function (event) {
                              event.preventDefault();
                              event.stopPropagation();
                              $('body').removeClass('oui-disable-select');
                              that._removeMouseHandlers();
                              $target.off('.dragging');
                          })
                          .on('click' + that.mouseNamespace, function (event) {
                              event.preventDefault();
                          });
                  }

              });

              this.$splitter.on('click', '.collapse-handle', function (event) {
                  event.preventDefault();
                  var collapsed = that.$splitter.hasClass('collapsed');

                  if (collapsed) {
                      that.restoreSplits();
                  } else {

                      if (that.options.collapseDirection == null) {
                          if (that.options.orientation === 'horizontal') {
                              that.collapseSecond();
                          } else {
                              that.collapseFirst();
                          }
                      } else if (that.options.collapseDirection === 'top') {
                          that.collapseFirst();
                      } else if (that.options.collapseDirection === 'right') {
                          that.collapseSecond();
                      }
                  }
              });

              this.$splitter.on('keydown', function (event) {
                  switch (event.keyCode) {
                      case keyCode.UP:
                      case keyCode.LEFT:
                          that.setSplit(that.options.firstSplitSize - 0.01);
                          break;
                      case keyCode.DOWN:
                      case keyCode.RIGHT:
                          that.setSplit(that.options.firstSplitSize + 0.01);
                          break;
                  }
              });

          },

          /**
           * Sets the proper classes on the collapse handles on the splitter bar
           *
           * @method orcl.splitPane#_setupCollapseHandles
           * @private
           */
          _setupCollapseHandles: function () {
              if (!this.options.collapseSecond &&
                  !this.options.collapseFirst) {
                  this.$splitter.find('.collapse-handle').hide();
              } else if (this.options.collapseFirst) {
                  this.$splitter.find('.collapse-handle').addClass('collapse-handle-up');
              } else if (this.options.collapseSecond) {
                  this.$splitter.find('.collapse-handle').addClass('collapse-handle-down');
              }

              if (this.options.collapseDirection === 'right') {
                  this.$splitter.find('.collapse-handle').addClass('direction-right');
              }

              if (this.options.collapseDirection === 'top') {
                  this.$splitter.find('.collapse-handle').addClass('direction-top');
              }
          },

          /**
           * Updates the DOM for the splitter
           *
           * @method orcl.splitPane#_setupSplit
           * @private
           */
          _setupSplit: function () {
              var splitSizeProp = this._getSplitSizeProperty();

              this.$splitContainer.append(this.$firstChild.addClass('split-child'));
              this.$splitContainer.append(this.$secondChild.addClass('split-child'));
              this.$splitter.insertAfter(this.$firstChild);

              saveOriginalCSSValue(this.$firstChild, splitSizeProp);
              saveOriginalCSSValue(this.$secondChild, splitSizeProp);
              saveOriginalCSSValue(this.$firstChild, 'max-' + splitSizeProp);
              saveOriginalCSSValue(this.$secondChild, 'max-' + splitSizeProp);

              inlineStyle.forceRemoveProperty(this.$firstChild, 'max-' + splitSizeProp);
              inlineStyle.forceRemoveProperty(this.$secondChild, 'max-' + splitSizeProp);

              this._setupCollapseHandles();

              this._setupHandlers();

              this.setSplit();

          },

          /**
           * A common location to update the call to the event 'split-resize-throttled' so that both the constructor
           * and subsequest 'option' calls can modify it.
           *
           * @method orcl.splitPane#_updateSplitResizeThrottle
           * @private
           */
          _updateSplitResizeThrottle: function () {
              var that = this;
              this._splitResizeThrottle = _.throttle(function () {
                  /**
                   * A rate-limited splitter resize event. Its frequency is determined by the `throttleSpeed`
                   * property
                   *
                   * @event orcl.splitPane#split-resize-throttled
                   */
                  that._trigger('split-resize-throttled');
              }, this.options.throttleSpeed);
          },

          /**
           * Overridden jQuery UI constructor function
           *
           * @method orcl.splitPane#_create
           * @override
           * @private
           */
          _create: function () {

              this.eventNamespace = this.eventNamespace + '-';

              this._uniqueId = _.uniqueId('orcl-sp-');
              this.mouseNamespace = '.sp-drag-handler' + this._uniqueId;

              this._validateElement();

              this._updateSplitResizeThrottle();

              this.$firstChild = this.element.children(':eq(0)');
              this.$secondChild = this.element.children(':eq(1)');

              this.$splitContainer = $('<div class="split-container"></div>').addClass(this.options.orientation).appendTo(this.element);
              this.$splitter = this._createSplitter().appendTo(this.$splitContainer);

              this._setupSplit();
              this.setSplit();

              // Second call to reflow after scrollbars go away!

              this.disable(!this.options.enabled);
          },

          /**
           * Internal convenience method to allow the splitter to change orientations
           *
           * @method orcl.splitPane#_reorient
           * @private
           */
          _reorient: function () {
              var oppositeSizeProperty = this._getOppositeSizeProperty();
              this.$splitContainer.removeClass('vertical').removeClass('horizontal').addClass(this.options.orientation);
              inlineStyle.forceRemoveProperty(this.$firstChild, oppositeSizeProperty);
              inlineStyle.forceRemoveProperty(this.$secondChild, oppositeSizeProperty);
              inlineStyle.forceRemoveProperty(this.$splitter, this.options.orientation === 'vertical' ? 'top' : 'left');
          },

          /**
           * Overridden jQuery UI _setOption method.
           *
           * @method orcl.splitPane#_setOption
           * @private
           * @param {String} key Option key
           * @param {*} value Option value
           */
          _setOption: function (key, value) {
              if (this._superApply) {
                  this._superApply(arguments);
              } else {
                  this.options[key] = value;
              }

              switch (key) {
                  case 'firstSplitSize':
                      this.setSplit();
                      break;
                  case 'orientation':
                      this._reorient();
                      this.setSplit();
                      break;
                  case 'collapseFirst':
                      this._setupCollapseHandles();
                      break;
                  case 'collapseSecond':
                      this._setupCollapseHandles();
                      break;
                  case 'throttleSpeed':
                      this._updateSplitResizeThrottle();
                      break;
              }
          },

          /**
           * Internal method called to resize splitter by a relative amount
           *
           * @method orcl.splitPane#_mouseSetSplit
           * @private
           * @param {Number} delta Amount to resize the splitter by, in **pixels**.
           */
          _mouseSetSplit: function (delta) {
              var available = inlineStyle.intCssProperty(this.$splitContainer, this._getSplitSizeProperty());
              this.setSplit(this.options.firstSplitSize + (delta / available));
          },

          /**
           * Internal method to revert the removed min-values when the splitter is collapsed
           *
           * @method orcl.splitPane#_restoreMinValues
           * @private
           * @param {Boolean} collapse Is the splitter being collapsed
           * @param {String} splitSizeProperty The property to restore values
           */
          _restoreMinValues: function (collapse, splitSizeProperty) {
              if (hasOriginalCSSValue(this.$firstChild, 'min-' + splitSizeProperty)) {
                  restoreOriginalCSSValue(this.$firstChild, 'min-' + splitSizeProperty);
              }

              if (hasOriginalCSSValue(this.$secondChild, 'min-' + splitSizeProperty)) {
                  restoreOriginalCSSValue(this.$secondChild, 'min-' + splitSizeProperty);
              }

              if (collapse) {
                  saveOriginalCSSValue(this.$firstChild, 'min-' + splitSizeProperty);
                  saveOriginalCSSValue(this.$secondChild, 'min-' + splitSizeProperty);
                  saveFlexValues(this.$firstChild);
                  saveFlexValues(this.$secondChild);
                  this.$firstChild.css('min-' + splitSizeProperty, '');
                  this.$secondChild.css('min-' + splitSizeProperty, '');
                  this.$firstChild.css('flex', '');
                  this.$secondChild.css('flex', '');
              } else {
                  restoreFlexValues(this.$firstChild);
                  restoreFlexValues(this.$secondChild);
              }
          },

          /**
           * Toggles enablement of the splitter.
           *
           * @method orcl.splitPane#disable
           * @param {Boolean} disable Whether or not to disable the splitter.
           */
          disable: function (disable) {
              this._setOption('enabled', !disable);
              if (disable) {
                  this.$splitter.addClass('disabled');
              } else {
                  this.$splitter.removeClass('disabled');
              }
          },

          /**
           * Sets the splitter division based on a pixel value.
           *
           * @method orcl.splitPane#setSplitPixels
           * @param {Number} newSplitSize The size (in **pixels**) of the first splitter child.
           * @param {Boolean} [firstSplit=true] Which splitter to size - defaults to the first splitter
           */
          setSplitPixels: function (newSplitSize, firstSplit) {
              var available = inlineStyle.intCssProperty(this.$splitContainer, this._getSplitSizeProperty());

              if (firstSplit == null || firstSplit === true) {
                  this.setSplit(newSplitSize / available);
              } else {
                  this.setSplit(1 - (newSplitSize / available));
              }
          },

          /**
           * Prevent the split pane from being draggable and then set the width of one of the containers
           * to be a constant pixel based width indepdendent of browser window sizing.
           *
           * @method orcl.splitPane#fixed
           * @param {Boolean} state Whether or not the splitPane is in a 'fixed' state.
           * @param {Object} sizes The sizes of the containers to set
           * @param {Number} sizes.first The size of the first child in pixels.
           * @param {Number} sizes.second The size of the second child in pixels.
           */
          fixed: function (state, sizes) {
              this.disable(state);

              this.$splitContainer.toggleClass('fixed', state);

              if (sizes && sizes.first) {
                  this.setSplitPixels(sizes.first);
                  this.$firstChild.css({
                      'flex-basis': sizes.first,
                      'flex-shrink': 0,
                      'flex-grow': 0
                  });
              } else {
                  this.$firstChild.css({
                      'flex': ''
                  });
              }

              if (sizes && sizes.second) {
                  this.setSplitPixels(sizes.second, false);
                  this.$secondChild.css({
                      'flex-basis': sizes.second,
                      'flex-shrink': 0,
                      'flex-grow': 0
                  });
              } else {
                  this.$secondChild.css({
                      'flex': ''
                  });

              }
          },

          /**
           * Sets the splitter division based on a percentage of the container.
           *
           * @method orcl.splitPane#setSplit
           * @param {Number} newSplitSize The size (as a **percentage**) of the split container
           * @param {Boolean} collapse (Internal) Used to signal that this is a collapse operation
           */
          setSplit: function (newSplitSize, collapse) {
              var splitSizeProperty = this._getSplitSizeProperty();
              var available = inlineStyle.intCssProperty(this.$splitContainer, splitSizeProperty);
              var splitterSize = inlineStyle.intCssProperty(this.$splitter, this._getSplitterSizeProperty());
              var splitterSizeHalf = splitterSize / 2;
              var absMin = splitterSizeHalf / available;
              var firstMin = collapse ?
                absMin :
                (Math.max(inlineStyle.intCssProperty(this.$firstChild, 'min-' + splitSizeProperty) || 0, absMin) + splitterSizeHalf) / available;
              var secondMin = collapse ?
                absMin :
                (Math.max(inlineStyle.intCssProperty(this.$secondChild, 'min-' + splitSizeProperty) || 0, absMin) + splitterSizeHalf) / available;
              var firstPercent;
              var secondPercent;

              if (!_.isUndefined(newSplitSize)) { // Need to handle setting to 0.0
                  this._oldSplitterSize = this.options.firstSplitSize;
                  this.options.firstSplitSize = newSplitSize;
              }

              this._restoreMinValues(collapse, splitSizeProperty);

              this._splitterEdge = false;

              if (this.options.firstSplitSize < firstMin) {
                  this.options.firstSplitSize = firstMin;
                  this._splitterEdge = true;
              } else if (this.options.firstSplitSize > (1 - secondMin)) {
                  this.options.firstSplitSize = (1 - secondMin);
                  this._splitterEdge = true;
              }

              firstPercent = this.options.firstSplitSize * 100;
              secondPercent = 100 - firstPercent;

              var start = inlineStyle.intCssProperty(this.$firstChild, splitSizeProperty);
              inlineStyle.setCalcProperty(this.$firstChild, splitSizeProperty, firstPercent + '% - ' + splitterSizeHalf + 'px');
              inlineStyle.setCalcProperty(this.$secondChild, splitSizeProperty, secondPercent + '% - ' + splitterSizeHalf + 'px');
              inlineStyle.setCalcProperty(this.$splitter, this.options.orientation === 'vertical' ? 'left' : 'top', firstPercent + '% - ' + splitterSizeHalf + 'px');
              var finish = inlineStyle.intCssProperty(this.$firstChild, splitSizeProperty);

              //Do not trigger event if it is the same as when we started
              if (start !== finish) {
                  /**
                   * Triggered when the splitter is resized.
                   *
                   * @event orcl.splitPane#split-resize
                   */
                  this._trigger('split-resize');
                  this._splitResizeThrottle();
              }
          },

          /**
           * @method orcl.splitPane#restoreSplits
           */
          restoreSplits: function () {
              this.setSplit(this._oldSplitterSize, false);
              this.$splitter.removeClass('collapsed');
              this.options.preventDrag = false;
          },

          /**
           * Collapses the first child to 0 height.
           * If pixelSize is specified, this becomes a synonym for `setSplitPixels`.
           *
           * @method orcl.splitPane#collapseFirst
           * @param {Number} [pixelSize] The size in pixels of the first split child.
           */
          collapseFirst: function (pixelSize) {
              this.$splitter.addClass('collapsed');
              this.options.preventDrag = true;
              if (!_.isUndefined(pixelSize)) {
                  this.setSplitPixels(pixelSize);
              } else if (this.options.collapseFirst) {
                  this.setSplit(0.0, true);

                  /**
                   * Triggered when the splitter is collapsed in either direction.
                   *
                   * @event orcl.splitPane#collapsed
                   * @param {Object} collapsing Has one property set depending on which child is collapsing. For example,
                   *                            if the first child is collapsed, the value will be `{ first : true }`
                   */
                  this._trigger('collapsed', {
                      first: true
                  });

                  /**
                   * Triggered when the first child in the splitter is collapsed.
                   *
                   * @event orcl.splitPane#collapsed-first
                   */
                  this._trigger('collapsed-first');
              }
          },

          /**
           * Collapses the second child to 0 height.
           * If pixelSize is specified, this becomes a synonym for `setSplitPixels` with a calculation based off of the
           * second child's new height.
           *
           * @method orcl.splitPane#collapseSecond
           * @param {Number} [pixelSize] The size in pixels of the second split child.
           */
          collapseSecond: function (pixelSize) {
              this.$splitter.addClass('collapsed');
              this.options.preventDrag = true;
              if (!_.isUndefined(pixelSize)) {
                  var available = inlineStyle.intCssProperty(this.$splitContainer, this._getSplitSizeProperty());
                  this.setSplitPixels(available - pixelSize);
              } else if (this.options.collapseSecond) {
                  this.setSplit(1.0, true);
                  this._trigger('collapsed', {
                      second: true
                  });

                  /**
                   * Triggered when the second child in the splitter is collapsed.
                   *
                   * @event orcl.splitPane#collapsed-second
                   */
                  this._trigger('collapsed-second');
              }
          },

          /**
           * Removes the splitter widget from it's container.
           *
           * @method orcl.splitPane#_destroy
           * @private
           */
          _destroy: function () {
              var splitSizeProperty = this._getSplitSizeProperty();

              this.element.append(this.$firstChild.removeClass('split-child'));
              this.element.append(this.$secondChild.removeClass('split-child'));

              restoreOriginalCSSValue(this.$firstChild, splitSizeProperty);
              restoreOriginalCSSValue(this.$secondChild, splitSizeProperty);
              restoreOriginalCSSValue(this.$firstChild, 'max-' + splitSizeProperty);
              restoreOriginalCSSValue(this.$secondChild, 'max-' + splitSizeProperty);

              $(document).off('.sp-drag-handler-' + this._uniqueId);

              this.$splitContainer.remove();

          }

      });

  });

/**
 * A widget that extends the JQuery-UI default widget and provides
 * a date picker ui element using an input html element.
 *
 * Example creation with default options:
 *
 *     $('selector').datePicker();
 *
 * @namespace orcl.datePicker
 *
 * @tutorial constructing-widget
 */
define('oui/widgets/date/dateWidget',[
    'jquery',
    'oui/widgets/dateField',
    'logger',
    'jqueryui'
], function ($, DateField, logger) {
      'use strict';

      $.widget('orcl.datePicker', {

          /**
           * @method orcl.datePicker#_create
           * @private
           */
          _create: function () {
              var that = this;

              logger.debug('"calling datePicker _create"');
              this.widgetEventPrefix = this.widgetEventPrefix + '-';
              logger.info('datePicker\'s event prefix: ' + this.widgetEventPrefix);

              this._dateField = new DateField(this.options, this.element, this.element[0].value);
              this._dateField.init();

              this._dateField.on('change', function (date) {
                  that._trigger('select', null, {
                      date: date
                  });
              });
          },

          /**
           * @method orcl.datePicker#disable
           *
           * @returns {Boolean}
           */
          /**
           * @method orcl.datePicker#disable
           * @param {Boolean} disabled
           */
          disable: function (disabled) {
              if (!arguments.length) {
                  return this._dateField.disable();
              }
              this._dateField.disable(disabled);
          },

          /**
           * @method orcl.datePicker#setDate
           *
           * @param {moment} date
           */
          setDate: function (date) {
              this._dateField.date(date);
          },

          /**
           * @method orcl.datePicker#getDate
           *
           * @returns {moment}
           */
          getDate: function () {
              return this._dateField.date();
          },

          /**
           * @method orcl.datePicker#date
           *
           * @returns {moment}
           */
          /**
           * @method orcl.datePicker#date
           * @param {moment} date
           */
          date: function (date) {
              if (!arguments.length) {
                  return this._dateField.date();
              }
              this._dateField.date(date);
          },

          /**
           * @method orcl.datePicker#value
           *
           * @returns {moment}
           */
          /**
           * @method orcl.datePicker#value
           * @param {moment} date
           */
          value: function (date) {
              if (!arguments.length) {
                  return this._dateField.value();
              }
              this._dateField.value(date);
          },

          /**
           * @method orcl.datePicker#setMinDate
           *
           * @param {moment} date
           */
          setMinDate: function (date) {
              this._dateField.min(date);
          },

          /**
           * @method orcl.datePicker#min
           *
           * @returns {moment}
           */
          /**
           * @method orcl.datePicker#min
           *
           * @param {moment} date
           */
          min: function (date) {
              if (!arguments.length) {
                  return this._dateField.min();
              }
              this._dateField.min(date);
          },

          /**
           * @method orcl.datePicker#setMaxDate
           *
           * @param {moment} date
           */
          setMaxDate: function (date) {
              this._dateField.max(date);
          },

          /**
           * @method orcl.datePicker#max
           *
           * @returns {moment}
           */
          /**
           * @method orcl.datePicker#max
           *
           * @param {moment} date
           */
          max: function (date) {
              if (!arguments.length) {
                  return this._dateField.max();
              }
              this._dateField.max(date);
          },

          /**
           * This method is called when new options are supplied. An example of this can be:
           *
           *      $('.some_class').datePicker({min : newMin});
           *
           * @private
           * @method orcl.datePicker#_setOption
           * @param {String} key The option key to update
           * @param {String} value The value the new key should hold
           */
          _setOption: function (key, value) {
              this._dateField.updateOption(key, value);
          },

          /**
           * Widget specific clean up and removal of element
           *
           * @private
           * @method orcl.datePicker# _destroy
           */
          _destroy: function () {
              this._dateField.destroy({
                  self: false
              });
          }

      });

  });

/**
 * A widget that extends the JQueryUI default widget and provides
 * a spinner ui element using an input html element.
 *
 * Example creation with default options:
 *
 *     $('selector').number();
 * @namespace orcl.number
 */
define('oui/widgets/spinner/spinnerWidget',[
    'jquery',
    'oui/widgets/numberField',
    'oui/formatters/number',
    'jqueryui'
], function ($, NumberField, numberFormatter) {
      'use strict';

      $.widget('orcl.number', {

          options: {
              type: 'number',

              //Valid options are number, currency, time
              max: null,
              min: null,
              fallback: false,
              fluid: false,
              spinner: {
                  controls: false,
                  step: 1,
                  scroll: true
              },
              formatter: numberFormatter
          },

          /**
           * @method orcl.number#_create
           * @private
           */
          _create: function () {
              var that = this;
              var value = this.element[0].value;
              this.widgetEventPrefix = this.widgetEventPrefix + '-';

              if (value === '') {
                  value = null;
              } else if (this.options.formatter && this.options.formatter.validate(value)) {
                  value = this.options.formatter.unformat(value);
              }
              this._number = new NumberField(this.options, this.element, value);
              this._number.init();

              this._number.on('change', function (val) {
                  that._trigger('change', null, {
                      value: val
                  });
              });
          },

          /**
           * @method orcl.number#setData
           * @param {Number|moment} data
           */
          setData: function (data) {
              this._number.data(data);
          },

          /**
           * @method orcl.number#getData
           */
          getData: function () {
              return this._number.data();
          },

          /**
           * @method orcl.number#data
           * @returns {Number|moment}
           */
          /**
           * @method orcl.number#data
           * @param {Number|moment} data
           */
          data: function (data) {
              if (!arguments.length) {
                  return this._number.data();
              }
              this._number.data(data);
          },

          /**
           * @method orcl.number#value
           * @returns {Number|moment}
           */
          /**
           * @method orcl.number#value
           * @param {Number|moment} data
           */
          value: function (data) {
              if (!arguments.length) {
                  return this._number.value();
              }
              this._number.value(data);
          },

          /**
           * @method orcl.number#increment
           * @param {String} field
           */
          increment: function (field) {
              this._number.increment(field);
          },

          /**
           * @method orcl.number#decrement
           * @param {String} field
           */
          decrement: function (field) {
              this._number.decrement(field);
          },

          /**
           * @method orcl.number#_setOption
           * @private
           * @param {String} key
           * @param {*} value
           */
          _setOption: function (key, value) {
              this._number.updateOption(key, value);
          },

          /**
           * @method orcl.number#disable
           * @param {Boolean} disabled
           */
          disable: function (disabled) {
              if (!arguments.length) {
                  return this._number.disable();
              }
              this._number.disable(disabled);
          },

          /**
           * @method orcl.number#_destroy
           * @private
           */
          _destroy: function () {
              // Destroys the underlying OUI number field
              this._number.destroy({
                  self: false
              });
          }

      });
  }
);

/**
* A widget that extends the JQuery-UI default widget and provides
* a Url testing field ui element using an input html element.
*
* Example creation with default options: (Note: `testFn` is required!)
*
*     $('selector').url({testFn: function (url, isValid) { ... }});
*
* @namespace orcl.url
*/
define('oui/widgets/url/urlWidget',[
    'jquery',
    'oui/widgets/urlField',
    'logger',
    'jqueryui'
], function ($, UrlField, logger) {
      'use strict';

      $.widget('orcl.url', {

          /**
           * @property {UrlField#Config} options
           * @memberof orcl.url
           */
          options: UrlField.prototype.defaults(),

          /**
           * Constructor for the url widget.
           *
           * @function orcl.url#_create
           * @private
           */
          _create: function () {
              var value = this.element[0].value;

              this.widgetEventPrefix = this.widgetEventPrefix + '-';

              if (value === '') {
                  value = null;
              }
              this._url = new UrlField(this.options, this.element, value);
              this._url.init();
          },

          /**
           * Getter/Setter for the url field.
           *
           * @function orcl.url#url
           * @param {String} url The url to set on the url field.
           * @returns {String} - The current url of the url field.
           */
          url: function (url) {
              if (!arguments.length) {
                  return this._url._controller.url();
              }
              this._url._controller.url(url);
          },

          /**
           * Disables/Enables the url field.
           *
           * @function orcl.url#disable
           * @param {Boolean} disabled Whether or not the url field is disable.
           */
          disable: function (disabled) {
              this._url._view.disable(disabled);
          },

          /**
           * Set/Reset options within the url field.
           *
           * @function orcl.url#_setOption
           * @private
           */
          _setOption: function () {
              //TODO: implement options function
              logger.warn('url._setOption: You cannot currently change any options in this widget.');
          },

          /**
           * Destroys the url field.
           *
           * @function orcl.url#_destroy
           * @private
           */
          _destroy: function () {
              this._url.destroy({
                  self: false
              });
          }

      });
  }
);


define('template-117', ['underscore'], function (_) { return function(data) {
var __t, __p = '';
__p += '<a href="#" title="' +
((__t = ( data.locale.close )) == null ? '' : __t) +
'" class="orcl-popover-close modal-close"> <i class="pgbu-icon pgbu-icon-dialog-close"></i></a>';
return __p
};});

define('oui/widgets/popover/popoverWidget',['require','jquery','underscore','oui/utils/keyCode','oui/utils/language','oui/utils/math','oui/utils/positioning','template-117','bootstrap'],function (require) {
    'use strict';

    var $ = require('jquery');
    var _ = require('underscore');
    var keyCode = require('oui/utils/keyCode');
    var lang = require('oui/utils/language');
    var math = require('oui/utils/math');
    var relativePosition = require('oui/utils/positioning');
    var closeButtonTemplate = require('template-117');
    require('bootstrap');

    var popoverClass = '.orcl-popover';
    var $arrow = $('<div class="arrow" />');

    /**
     * @class Popover
     * @private
     *
     * @param {HTMLElement} el
     */
    var Popover = function (el) {
        var $el = $(el);

        $el.find('.orcl-popover-outer').on('click.orcl-popover', this.toggle);
        $('html').on('click.orcl-popover', function () {
            $el.removeClass('open');
        });
    };

    /**
     * @method orcl.popover~createPopoverComponents
     * @private
     * @param {jQuery} $this
     */
    function createPopoverComponents($this) {
        if ($this.find('.arrow').length === 0) {
            $this.find('.orcl-popover-outer').append($arrow.clone());
        }

        if ($this.find('.orcl-popover-close').length === 0) {
            $this.find('.orcl-popover-inner').append(closeButtonTemplate({
                locale: lang.lang()
            }));
        }
    }

    /**
     * @method orcl.popover~closeAllPopovers
     * @private
     *
     * @param {?jQuery} $current
     */
    function closeAllPopovers($current) {
        $(popoverClass).add($current).each(function () {
            $(this).removeClass('open');
            $(this).find('.orcl-popover-content').attr('aria-hidden', true);
        });
    }

    /**
     * @method orcl.popover~determineOrientation
     * @private
     * @param {jQuery} $this
     */
    function determineOrientation($this) {
        var $arrow = $this.find('.arrow');
        var $inner = $this.find('.orcl-popover-inner');
        var rightEdge = $this.offset().left + $this.width() + $arrow.innerWidth() + $inner.innerWidth();
        var pageWidth = $(window).width();

        return (rightEdge < pageWidth) ? 'right' : 'left';
    }

    /**
     * @method orcl.popover~positionPopover
     * @private
     * @param {jQuery} $this
     */
    function positionPopover($this) {
        var $outer = $this.find('.orcl-popover-outer');
        var $inner = $this.find('.orcl-popover-inner');

        if ($this.hasClass('temporary-position')) {
            $this.removeClass('left right temporary-position');

            // We only attempt to auto position left or right
        }

        if (!($this.hasClass('left') || $this.hasClass('right') || $this.hasClass('top') || $this.hasClass('bottom'))) {
            $this.addClass(determineOrientation($this));
            $this.addClass('temporary-position');
        }

        var position = relativePosition($this, false, $inner);

        position.top = position.top - $(window).scrollTop();
        position.left = position.left - $(window).scrollLeft();

        //Right - default
        $inner.css('left', position.left + ($outer.outerWidth()));
        $inner.css('top', position.top - ($inner.outerHeight(true) / 2) + ($this.outerHeight() / 2));

        if ($this.hasClass('right align-top')) {
            $inner.css('top', position.top - 7);
        }

        if ($this.hasClass('left')) {
            $inner.css('left', position.left - $inner.outerWidth(true));
        }

        if ($this.hasClass('top')) {
            $inner.css('left', position.left - ($inner.outerWidth(true) / 2) + ($this.outerWidth() / 2));
            $inner.css('top', position.top - ($inner.outerHeight(true)));
        }

        if ($this.hasClass('bottom')) {
            $inner.css('left', position.left - ($inner.outerWidth(true) / 2) + ($this.outerWidth() / 2));
            $inner.css('top', position.top + $outer.height());
        }
    }

    Popover.prototype = {
        constructor: Popover,
        /**
         * @method Popover#focus
         * @param {Boolean} forward
         */
        focus: function (forward) {
            var focusable = $(this).find('a, button, input, [tabindex]:not([tabindex=-1])')
              .filter(':visible');
            var length = focusable.length;

            if ($.contains(this, document.activeElement)) {
                //Go on to the next one
                var index = focusable.index(document.activeElement);
                if (forward) {
                    index = math.mod(index + 1, length);
                } else {
                    index = math.mod(index - 1, length);
                }

                focusable.get(index).focus();
            } else if (forward && focusable.length > 0) {
                focusable.get(0).focus();
            } else if (!forward && focusable.length > 0) {
                focusable.get(focusable.length - 1).focus();
            }
        },
        /**
         * @method Popover#close
         */
        close: function () {
            var $this = $(this);

            if ($this.is('.disabled, :disabled')) {
                return;
            }

            closeAllPopovers($(this));

            $this.removeClass('open');

            $this.find('.orcl-popover-content').attr('aria-hidden', true);

            $this.focus();

            return false;
        },
        /**
         * @method Popover#toggle
         */
        toggle: function () {
            var $this = $(this);
            var isActive;

            if ($this.is('.disabled, :disabled')) {
                return;
            }

            isActive = $this.hasClass('open');

            closeAllPopovers($(this));

            if (!isActive) {
                createPopoverComponents($this);
                if (!$this.hasClass('no-position')) {
                    positionPopover($this);
                }
                $this.toggleClass('open');
            }

            var id = $this.find('.header').attr('id') || _.uniqueId('popover');
            $this.find('.header').attr('id', id);
            var contentIds = _.map($this.find('.orcl-popover-content').children('*:not(.header)'), function (elem) {
                var curId = $(elem).attr('id') || _.uniqueId('popover-content-');
                if (!$(elem).attr('id')) {
                    $(elem).attr('id', curId);
                }
                return curId;
            });
            $this.find('.orcl-popover-content')
                .attr('tabindex', 0)
                .attr('role', 'dialog')
                .attr('aria-labelledby', id);

            if (contentIds.length > 0) {
                $this.find('.orcl-popover-content')
                    .attr('aria-describedby', contentIds.join(' '));
            }

            if ($this.hasClass('open')) {
                $this.find('.orcl-popover-content').attr('aria-hidden', false).focus();
            } else {
                $this.focus();
            }

            return false;
        }
    };

    /**
     * @method $.fn.orclPopover
     *
     * @param {Object} option
     *
     * @returns {jQuery}
     */
    $.fn.orclPopover = function (option) {
        return this.each(function () {
            var $this = $(this);
            var data = $this.data('orclPopover');
            $(this).attr('tabindex', 0);
            if (!data) {
                $this.data('orclPopover', (data = new Popover(this)));
            }
            if (typeof option === 'string') {
                data[option].call($this);
            }
        });
    };

    $.fn.orclPopover.Constructor = Popover;

    $(document)
    .on('click.orclpopover', '.orcl-popover-close, .orcl-popover-close > .pgbu-icon', function (event) {
        event.preventDefault();
        event.stopPropagation();
        closeAllPopovers();
        $(this).closest(popoverClass).focus();
    })
    .on('keydown.orclpopover', '.orcl-popover-close', function (event) {
        switch (event.keyCode) {
            case keyCode.ENTER:
                event.preventDefault();
                event.stopPropagation();
                closeAllPopovers();
                $(this).closest(popoverClass).focus();
                break;
        }
    })
    .on('click.orcl-popover touchstart.orcl-popover', '.orcl-popover-inner *:not(.orcl-popover-close)', function (e) {
        e.stopPropagation();
    })
    .on('focusin', popoverClass, function (event) {
        var $popover = $(event.currentTarget);
        $popover.attr('aria-role', 'button');
        $popover.find('.orcl-popover-content').attr('aria-hidden', !$popover.hasClass('open'));
    })
    .on('click', closeAllPopovers)
    .on('keyup', popoverClass, function (event) {
        switch (event.keyCode) {
            case keyCode.ESC:
                if ($(this).hasClass('open')) {
                    event.stopPropagation();
                    Popover.prototype.close.call(this);
                }
                break;
        }
    })
    .on('keydown', popoverClass, function (event) {
        switch (event.keyCode) {
            case keyCode.TAB:
                if ($(this).hasClass('open')) {
                    Popover.prototype.focus.call(this, !event.shiftKey);
                    event.preventDefault();
                    event.stopPropagation();
                }
                break;
            case keyCode.ENTER:
            case keyCode.SPACE:
                event.preventDefault();
                Popover.prototype.toggle.call(this);
                break;
        }
    })
    .on('click.orcl-popover touchstart.orcl-popover', popoverClass, function (event) {
        event.preventDefault();
        event.stopPropagation();
        Popover.prototype.toggle.call(this);
    });

    $(window).on('resize scroll', _.debounce(closeAllPopovers, 250, {
        leading: true
    }));

    return Popover;
});

/**
* Custom styled file input.
*
* Hides the built-in file input but allows screen readers to read it as a
* file input.
*
* @namespace orcl.filePicker
*
*/
define('oui/widgets/file/fileWidget',[
    'jquery',
    'oui/widgets/file/filePicker',
    'jqueryui'
], function ($, FilePicker) {
      'use strict';

      $.widget('orcl.filePicker', {
          options: {
              /**
               * @property fluid
               * @default false
               * @type {Boolean}
               */
              fluid: false
          },
          /**
           * @method orcl.filePicker#_create
           * @private
           */
          _create: function () {
              this._filePicker = new FilePicker(this.options, this.element).init();
          },
          /**
           * @method orcl.filePicker#_destroy
           * @private
           */
          _destroy: function () {
              this._filePicker.destroy({
                  self: false
              });
          }
      });
  });

/**
* A widget that extends the JQuery-UI default widget and provides
* a ComboBox ui element using an input or select html element.
*
* Example creation with default options
*
*     $('selector').comboBox();
*
* @namespace orcl.comboBox
*/
define('oui/widgets/comboBox/comboBoxWidget',[
    'jquery',
    'oui/widgets/comboBox',
    'logger',
    'jqueryui'
], function ($, ComboBox, logger) {
      'use strict';

      $.widget('orcl.comboBox', {

          options: ComboBox.prototype.defaults(),

          /**
           * @method orcl.comboBox#_create
           * @private
           */
          _create: function () {
              var that = this;
              var value = this.options.defaultValue || this.element[0].value;

              // Don't want to pass this into the actual ComboBox widget
              delete this.options.defaultValue;

              this.widgetEventPrefix = this.widgetEventPrefix + '-';

              if (this.element.is('input') || this.element.is('select')) {
                  this._comboBox = new ComboBox(this.options, this.element);
                  this._comboBox.init();
                  this._comboBox.value({
                      'value': value
                  });
              } else {
                  var message = 'comboBox._create: Needs to be called on either an <input> or a <select>';
                  logger.fatal(message);
                  throw new Error(message);
              }

              // Attach widget to new root ComboBox element
              this._comboBox.element().data('orclComboBox', this);

              // Pass on the valueSelected event again!
              this._comboBox.on('valueSelected', function (selected) {
                  that._trigger('valueSelected', null, selected);
              });
          },

          /**
           * @method orcl.comboBox#data
           *
           * @returns {Array|String}
           */
          /**
           * @method orcl.comboBox#data
           *
           * @param {Array|String} newValue
           */
          data: function (newValue) {
              if (!arguments.length) {
                  return this.selected();
              }
              this.selected(newValue);
          },

          /**
           * @method orcl.comboBox#selected
           *
           * @returns {Array|String}
           */
          /**
           * @method orcl.comboBox#selected
           *
           * @param {Array|String} newValue
           */
          selected: function (newValue) {
              if (!arguments.length) {
                  return this._comboBox.selected();
              }
              this._comboBox.selected(newValue);
          },

          /**
           * @method orcl.comboBox#values
           *
           * @returns {Array}
           */
          /**
           * @method orcl.comboBox#values
           *
           * @param {Array} newValues
           */
          values: function (newValues) {
              if (!arguments.length) {
                  return this._comboBox.values();
              }
              this._comboBox.values(newValues);
          },

          /**
           * @method orcl.comboBox#value
           *
           * @returns {Array|String}
           */
          /**
           * @method orcl.comboBox#value
           *
           * @param {Array|String} newValue
           */
          value: function (newValue) {
              if (!arguments.length) {
                  return this._comboBox.value();
              }
              this._comboBox.value(newValue);
          },

          /**
           * @method orcl.comboBox#disabled
           *
           * @param {Boolean} disabled
           */
          disable: function (disabled) {
              this._comboBox.disable(disabled);
          },

          /**
           * @method orcl.comboBox#_setOption
           * @private
           *
           * @param {String} key
           * @param {*} value
           */
          _setOption: function (key, value) {
              switch (key) {
                  case 'values':
                      this._comboBox.values(value);
                      break;
                  default:
                      logger.warn('comboBox._setOption: You cannot currently change any options in this widget.');
              }
          },

          /**
           * @method orcl.comboBox#_destroy
           * @private
           */
          _destroy: function () {
              this._comboBox.destroy({
                  self: false
              });
          }

      });
  }
);

/**
 * A widget that extends the JQuery-UI default widget and provides
 * an editor ui element using an input html element.
 *
 * Example creation with default options:
 *
 *     $('selector').editor();
 *
 * @namespace orcl.editor
 */
define('oui/widgets/editor/editorWidget',[
    'jquery',
    'oui/widgets/editor',
    'jqueryui'
], function ($, Editor) {
      'use strict';

      $.widget('orcl.editor', {

          /**
           * @method orcl.editor#_create
           * @private
           */
          _create: function () {
              this.widgetEventPrefix = this.widgetEventPrefix + '-';

              this._editor = new Editor(this.options, this.element, this.element[0].value)
                .init();
          },

          /**
           * @method orcl.editor#disable
           *
           * @returns {Boolean}
           */
          /**
           * @method orcl.editor#disable
           *
           * @param {Boolean} disabled
           */
          disable: function (disabled) {
              if (!arguments.length) {
                  return this._editor.disable();
              }
              this._editor.disable(disabled);
          },

          /**
           * @method orcl.editor#text
           *
           * @returns {Boolean}
           */
          /**
           * @method orcl.editor#text
           *
           * @param {Boolean} text
           */
          text: function (text) {
              if (!arguments.length) {
                  return this._editor.document();
              }

              this._editor.document(text);
          },

          /**
           * This method is called when new options are supplied. An example of this can be:
           *
           *      $('.some_class').datePicker({min : newMin});
           *
           * @private
           * @method orcl.editor#_setOption
           * @param {String} key The option key to update
           * @param {String} value The value the new key should hold
           */
          _setOption: function (key, value) {
              this._editor.option(key, value);
          },

          /**
           * Widget specific clean up and removal of element
           *
           * @private
           * @method orcl.editor#_destroy
           */
          _destroy: function () {
              this._editor.destroy({
                  self: false
              });
          }

      });

  });

define('oui/utils/valHooks',[
    'jquery',
    'oui/utils/defaults',
    'oui/utils/is',
    'oui/utils/dom'
], function ($, defaults, is, dom) {
    'use strict';

    var oldTextHooks = $.valHooks.text || { get: function () {}, set: function () {} }; //jscs:ignore jsDoc

    /**
     * The jQuery val hooks namespace
     * Only hook onto input[type=text] for their value calls.
     * @external "$.valHooks.text"
     */

    $.valHooks.text = {
        /**
         * @function external:"$.valHooks.text".get
         * @param {HTMLInputElement} elem The element to get its current value.
         * @returns {Boolean|null} Whether or not the hook handled the call or not.
         */
        get: function (elem) {
            if (defaults().hooks && is(elem)) {
                return dom.get(elem).data();
            }

            return oldTextHooks.get.call(this, elem);
        },
        /**
         * @function external:"$.valHooks.text".set
         * @param {HTMLInputElement} elem The element to set its current value for.
         * @param {String} val The string to set the elem's value to.'
         * @returns {Boolean|null} Whether or not the hook was handled by the call or not.
         */
        set: function (elem, val) {
            //TODO: this picker condition is a bit janky
            //can be removed when the picker controls the value of the input
            if (defaults().hooks &&
               is(elem) && !($(elem).parent().hasClass('orcl-picker') && !$(elem).parent().hasClass('orcl-typeahead'))) {
                dom.get(elem).data(val);
                return true;
            }

            return oldTextHooks.set.call(this, elem, val);
        }
    };
});

define('oui/utils/propHooks',[
    'jquery',
    'oui/utils/defaults',
    'oui/utils/is',
    'oui/utils/dom'
], function ($, defaults, is, dom) {
    'use strict';

    var oldDisableHooks = $.propHooks.disabled || { get: function () { return null; }, set: function () { return undefined;} }; //jscs:ignore jsDoc

    /**
     * The jQuery property hooks namespace
     * @external "$.propHooks.disabled"
     */

    $.propHooks.disabled = {
        /**
         * @function external:"$.propHooks.disabled".get
         * @param {HTMLInputElement} elem The element to get its current disabled status.
         * @returns {Boolean|null} Whether or not the hook handled the call or not.
         */
        get: function (elem) {
            var widget;
            if (defaults().hooks && is(elem)) {
                widget = dom.get(elem);
                return widget.disable();
            }

            return oldDisableHooks.get.call(this, elem);
        },
        /**
         * @function external:"$.propHooks.disabled".set
         * @param {HTMLInputElement} elem The element to get its current disabled status.
         * @param {Boolean} val The new disabled state of the input element.
         * @returns {Boolean|null} Whether or not the hook handled the call or not.
         */
        set: function (elem, val) {
            var widget;
            if (defaults().hooks && is(elem)) {
                widget = dom.get(elem);
                widget.disable(val);
                return true;
            }

            return oldDisableHooks.set.call(this, elem, val);
        }
    };
});

define('oui', ['require','oui/widgets/textField','oui/widgets/dateField','oui/widgets/picker','oui/widgets/numberField','oui/widgets/urlField','oui/widgets/colorPicker','oui/widgets/iconPicker','oui/widgets/comboBox','oui/widgets/treePanel','oui/widgets/typeAhead','oui/widgets/calendar','oui/widgets/editor','oui/widgets/slider','oui/widgets/doublePicker','oui/widgets/modal/modal','oui/widgets/form','oui/widgets/checkbox','oui/widgets/canvas/formEditor','oui/widgets/radio','oui/widgets/toolbar/toolbar','oui/widgets/contextMenu/contextMenu','oui/widgets/multiField/multiField','oui/widgets/file/filePicker','oui/widgets/hopper/hopper','oui/widgets/quickNav','oui/utils/language','oui/utils/dom','oui/utils/defaults','oui/utils/is','oui/widgets/logicalDate/logicalDate','oui/widgets/picker/pickerWidget','oui/widgets/splitPane/splitPaneWidget','oui/widgets/date/dateWidget','oui/widgets/spinner/spinnerWidget','oui/widgets/url/urlWidget','oui/widgets/popover/popoverWidget','oui/widgets/file/fileWidget','oui/widgets/comboBox/comboBoxWidget','oui/widgets/editor/editorWidget','oui/utils/valHooks','oui/utils/propHooks'],function (require) { //jshint ignore:line
        'use strict';

        /**
         * Provides a single point of entry, 'oui', to require the inputs library.
         *
         * Returns an object with some useful fields:
         *
         *     define(['oui'], function (oui) {
         *         console.log(oui.version); // Would print out oui version, e.g. '1.2' or '1.3'.
         *     });
         *
         * @module oui
         **/

        var TextField = require('oui/widgets/textField');
        var DateField = require('oui/widgets/dateField');
        var Picker = require('oui/widgets/picker');
        var NumberField = require('oui/widgets/numberField');
        var UrlField = require('oui/widgets/urlField');
        var ColorPicker = require('oui/widgets/colorPicker');
        var IconPicker = require('oui/widgets/iconPicker');
        var ComboBox = require('oui/widgets/comboBox');
        var TreePanel = require('oui/widgets/treePanel');
        var TypeAhead = require('oui/widgets/typeAhead');
        var Calendar = require('oui/widgets/calendar');
        var Editor = require('oui/widgets/editor');
        var Slider = require('oui/widgets/slider');
        var DoublePicker = require('oui/widgets/doublePicker');
        var Modal = require('oui/widgets/modal/modal');
        var Form = require('oui/widgets/form');
        var Checkbox = require('oui/widgets/checkbox');
        var FormEditor = require('oui/widgets/canvas/formEditor');
        var Radio = require('oui/widgets/radio');
        var Toolbar = require('oui/widgets/toolbar/toolbar');
        var ContextMenu = require('oui/widgets/contextMenu/contextMenu');
        var MultiField = require('oui/widgets/multiField/multiField');
        var FilePicker = require('oui/widgets/file/filePicker');
        var Hopper = require('oui/widgets/hopper/hopper');
        var quickNav = require('oui/widgets/quickNav');
        var language = require('oui/utils/language');
        var dom = require('oui/utils/dom');
        var defaults = require('oui/utils/defaults');
        var is = require('oui/utils/is');
        var logicalDate = require('oui/widgets/logicalDate/logicalDate');
        require('oui/widgets/picker/pickerWidget');
        require('oui/widgets/splitPane/splitPaneWidget');
        require('oui/widgets/date/dateWidget');
        require('oui/widgets/spinner/spinnerWidget');
        require('oui/widgets/url/urlWidget');
        require('oui/widgets/popover/popoverWidget');
        require('oui/widgets/file/fileWidget');
        require('oui/widgets/comboBox/comboBoxWidget');
        require('oui/widgets/editor/editorWidget');
        require('oui/utils/valHooks');
        require('oui/utils/propHooks');

        /**
         * @exports oui
         */
        return {
            /**
             * The exported reference to the TextField constructor.
             *
             * See {@link TextField} for more information.
             *
             * @member TextField
             * @type {TextField}
             */
            TextField: TextField,
            /**
             * The exported reference to the DateField constructor.
             *
             * See {@link DateField} for more information.
             *
             * @member DateField
             * @type {DateField}
             */
            DateField: DateField,
            /**
             * The exported reference to the IconPicker constructor.
             *
             * See {@link IconPicker} for more information.
             *
             * @member IconPicker
             * @type {IconPicker}
             */
            IconPicker: IconPicker,
            /**
             * The exported reference to the Picker constructor.
             *
             * See {@link Picker} for more information.
             *
             * @member Picker
             * @type {Picker}
             */
            Picker: Picker,
            /**
             * The exported reference to the Double Picker constructor.
             *
             * See {@link DoublePicker} for more information.
             *
             * @member DoublePicker
             * @type {DoublePicker}
             */
            DoublePicker: DoublePicker,
            /**
             * The exported reference to the NumberField constructor.
             *
             * See {@link NumberField} for more information.
             *
             * @member NumberField
             * @type {NumberField}
             */
            NumberField: NumberField,
            /**
             * The exported reference to the UrlField constructor.
             *
             * See {@link UrlField} for more information.
             *
             * @member UrlField
             * @type {UrlField}
             */
            UrlField: UrlField,
            /**
             * The exported reference to the ColorPicker constructor.
             *
             * See {@link ColorPicker} for more information.
             *
             * @member ColorPicker
             * @type {ColorPicker}
             */
            ColorPicker: ColorPicker,
            /**
             * The exported reference to the ComboBox constructor.
             *
             * See {@link ComboBox} for more information.
             *
             * @member ComboBox
             * @type {ComboBox}
             */
            ComboBox: ComboBox,
            /**
             * The exported reference to the TreePanel constructor.
             *
             * See {@link TreePanel} for more information.
             *
             * @member TreePanel
             * @type {TreePanel}
             */
            TreePanel: TreePanel,
            /**
             * The exported reference to the TypeAhead constructor.
             *
             * See {@link TypeAhead} for more information.
             *
             * @member TypeAhead
             * @type {TypeAhead}
             */
            TypeAhead: TypeAhead,
            /**
             * The exported reference to the Editor constructor.
             *
             * See {@link Editor} for more information.
             *
             * @member Editor
             * @type {Editor}
             */
            Editor: Editor,
            /**
             * The exported reference to the Slider constructor.
             *
             * See {@link Slider} for more information.
             *
             * @member Slider
             * @type {Slider}
             */
            Slider: Slider,

            /**
             * The exported reference to the Calendar constructor.
             *
             * See {@link Calendar} for more information.
             *
             * @member Calendar
             * @type {Calendar}
             */
            Calendar: Calendar,
            /**
             * The exported reference to the Modal constructor.
             *
             * See {@link Modal} for more information.
             *
             * @member Modal
             * @type {Modal}
             */
            Modal: Modal,
            /**
             * The exported reference to the Form constructor.
             *
             * See {@link Form} for more information.
             *
             * @member Form
             * @type {Form}
             */
            Form: Form,
            /**
             * The exported reference to the Checkbox constructor.
             *
             * See {@link Checkbox} for more information.
             *
             * @member Checkbox
             * @type {Checkbox}
             */
            Checkbox: Checkbox,
            /**
             * The exported reference to the Radio constructor.
             *
             * See {@link Radio} for more information.
             *
             * @member Radio
             * @type {Radio}
             */
            Radio: Radio,
            /**
             * The exported reference to the FilePicker constructor.
             *
             * See {@link FilePicker} for more information.
             *
             * @member FilePicker
             * @type {FilePicker}
             */
            FilePicker: FilePicker,
            /**
             * The exported reference to the Toolbar constructor.
             *
             * See {@link Toolbar} for more information.
             *
             * @member Toolbar
             * @type {Toolbar}
             */
            Toolbar: Toolbar,
            MultiField: MultiField,
            LogicalDate: logicalDate,
            /**
             * The exported reference to the Context Menu constructor.
             *
             * See {@link ContextMenu}
             *
             * @member ContextMenu
             * @type {ContextMenu}
             */
            ContextMenu: ContextMenu,
            /**
             * The exported reference to the Hopper constructor.
             *
             * See {@link Hopper}
             *
             * @member Hopper
             * @type {Hopper}
             */
            Hopper: Hopper,
            quickNav: quickNav,
            FormEditor: FormEditor,
            /**
             * A function to set the default language for use within OUI and all widget locale options.
             *
             * @method lang
             * @static
             */
            lang: language.lang,
            /**
             * A single-use function to set the module-wide default options to use for widget constructions.
             *
             * Useful for default formatters or locale objects.
             *
             * @method defaults
             * @static
             */
            defaults: defaults,
            /**
             *
             * Function to get the OUI widget associated with the element.
             *
             * @method get
             * @static
             *
             * @param {HTMLElement} element
             * @returns Widget
             */
            get: dom.get,
            /**
             * Function to determine whether an element is an OUI controlled element.
             *
             * This function is to be used by [get]{@link module:oui.get} after it is determined to be an OUI element.
             *
             * @example
             *
             *  var value;
             *  var $elem = $('input:eq(0)');
             *  if(oui.isOUI($elem)) {
             *      value = oui.get($elem).value();
             *  }
             *
             * @method isOUI
             * @static
             * @param {HTMLElement} element
             * @returns {Boolean}
             */
            isOUI: is,
            /**
             * The version of OUI.
             *
             * @member version
             * @type {String}
             */
            version: '3.8.7'
        };
    }
);



// End OUI
