//!
//! Built on: Mon Dec 19 2016 11:36:51
//! Library version: 5.3.0
//! Branch: master  |  Commit: 24b72af9
//!
/**
 * A collection of static functions around dealing with Strings.
 *
 * For string methods that are uniquely impacted by our XSS escaping protections, see [`utils.escaping`](utils.escaping.html)
 *
 * @class utils.strings
 * @module utils
 */

define('primutils/utils/strings',['lodash'], function (_) {
    'use strict';

    var that = {
        /**
         * Simply and efficiently tests whether or not the given value is a string.
         *
         * @method isString
         * @param val
         * @static
         */
        isString : function (str) {
            /* eslint-disable lodash/prefer-lodash-typecheck */
            // compare lodash, which also invokes toString to test against "[object String]"
            return typeof str === 'string';
            /* eslint-enable lodash/prefer-lodash-typecheck */
        },

        /**
         * Simple templating function, mimicking very basic usage of _.template.
         *  Will ONLY do replacements on '<%=VAR_NAME%>'-style strings.  As examples,
         *  see the following:
         *
         *     // invalid!
         *     strings.template('<% if (cond) { return "hello"; } %>')(true); // no executable code allowed
         *     // invalid!
         *     strings.template('<%= NAME %>'); // spaces not allowed
         *     // valid:
         *     strings.template('<%=NAME%>');
         *
         * The original intention was for use in Prime, where [CSP](https://developer.mozilla.org/en-US/docs/Web/Security/CSP) protections are in place.
         *
         * @param {String} text the template, similar to first option of _.template
         * @param {Object} data optional, if specified, template will invoke the template
         *   and do replacement based on the arguments provided; if not, template will
         *   return a function that can be called with an Object to do the same thing.
         *
         * @method template
         * @static
         */
        template : function template(text, data) {
            function handleReplace(replaceData) {
                var response = text;
                _.forEach(replaceData, function (replaceVal, replaceName) {
                    var reReplace = new RegExp('<%=' + replaceName + '%>', 'g');
                    response = response.replace(reReplace, function () {
                        return replaceVal;
                    });
                });
                return response;
            }

            if (data) {
                return handleReplace(data);
            }

            return handleReplace;
        },

        /**
         * Formats a string with placeholders.
         *
         * A placeholder takes the form of `{n}` where `n` is a decimal number starting with 0.
         *
         *     strings.formatString("Testing {1} or {0} stuff", "fun", "exciting") === "Testing exciting or fun stuff";
         *
         * If there are more placeholders than passed arguments, the remaining placeholders will be replaced with the empty string.
         *
         * @method formatString
         * @param node {String} The format string
         * @param ... {String} Each value to replace in the format string, starting with index 0
         * @returns {String} The string with the format values replaced
         */
        formatString : function (node) {
            var params = _.rest(arguments);

            return node.replace(/\{([0-9]+)\}/g, function (match, g1) {
                var value = params[parseInt(g1, 10)];
                return (value != null ? value : '');
            });
        }
    };

    return that;
});

define('primutils/i18n/Localizer',[
    'lodash',
    'logger',
    'primutils/utils/strings'
], function (_, logger, strings) {
    'use strict';

    var __noSuchKey__ = 'KEY NOT FOUND: ',
        __pathDelimiter__ = '.';

    function findKey(stringMap, path) {
        if (stringMap == null || path == null) {
            return null;
        }

        // get last part of path
        var depthIndex = path.lastIndexOf(__pathDelimiter__);
        if (depthIndex === -1) {
            // return the value
            return stringMap[path];
        }
        // move down a level in the object
        var firstLevelIndex = path.indexOf(__pathDelimiter__);
        var depth = path.slice(0, firstLevelIndex);
        return findKey(stringMap[depth], path.slice(firstLevelIndex + 1));
    }

    /**
     * Allows for retrieving localized strings from a given string map of any depth.
     *
     * Also contains helper methods for formatting strings.
     *
     * -----
     *
     * Notice for developers in UP: the `'localizer'` module in UP uses this internally, but does not expose the `'setStringMap'` method.
     * <a href="i18n.emptyStrings.html">emptyStrings</a> talks about expectations for application-level developers in providing a set of strings for primutils.
     *
     * @class i18n.Localizer
     * @module i18n
     * @param [stringMap] {Object} A map of localization "keys" to their localized representation
     * @constructor
     */
    function Localizer(stringMap) {
        logger.assert(this && this instanceof Localizer, 'Localizer must be instantiated with new');
        this.setStringMap(stringMap);
        this.setStringMap = this.setStringMap.bind(this);
        this.getString = this.getString.bind(this);
        this.getFormattedString = this.getFormattedString.bind(this);
        this.formatString = this.formatString.bind(this);
    }

    Localizer.formatString = strings.formatString;

    Localizer.prototype = {
        constructor : Localizer,

        /**
         * Changes which string map is used to find localized values
         *
         * @method setStringMap
         * @param stringMap {Object} A map of localization "keys" to their localized representation
         */
        setStringMap : function (stringMap) {
            this.stringMap = stringMap;
        },

        /**
         * Retrieves a localized string from the map.
         *
         * The key is a "."-delimited string. Each part of the path represents on level of objects in the string map.
         * The map can be of relatively unlimited depth.
         *
         * For example, the key "foo.bar" would return "Hello, World!" in the following map
         *
         *     {
         *         "foo" : {
         *             "bar" : "Hello, World!"
         *         }
         *     }
         *
         * @method getString
         * @param key {String} The string to find
         * @returns {String} The localized value
         */
        getString : function (key) {
            var node = findKey(this.stringMap, key);

            if (!node) {
                logger.warn('Key: \'' + key + '\' not found.');
                return __noSuchKey__ + key;
            }

            return node;
        },

        /**
         * Reports whether or not the localizer has a string for the given key; reports `true` if so and `false` otherwise.
         *
         * (Is not impacted by any call to `prefixFallbacks`)
         *
         * @method hasString
         * @param key {String} the key to check
         * @returns true if the key is present, false otherwise
         */
        hasString : function (key) {
            return !!findKey(this.stringMap, key);
        },

        /**
         * Alias for <a href="utils.strings.html#method_formatString">strings.formatString</a>.
         *
         * @method formatString
         * @param node {String} The format string
         * @param ... {String} Each value to replace in the format string, starting with index 0
         * @returns {String} The string with the format values replaced
         */
        formatString : strings.formatString,

        /**
         * Formats the string given by the key with placeholders.
         *
         * See <a href="utils.strings.html#method_formatString">strings.formatString</a> for more on placeholders.
         *
         * @method getFormattedString
         * @param key {String} The string to find
         * @param ... {String} Each value to replace in the format string, starting with index 0
         * @returns {String} The string with the format values replaced
         */
        getFormattedString : function (key/*, arg1, arg2, ...*/) {
            var node = this.getString(key);

            if (node.indexOf(__noSuchKey__) === 0) {
                return node;
            }

            var params = _.toArray(arguments);
            params[0] = node;

            return this.formatString.apply(null, params);
        },

        /**
         * A helper method to support prefix fallback rules.
         *
         * Example: in Prime 15.1, a large number of strings were moved from the `message.` prefix to `application_error.`.
         * This method, then, allows basically the following to support that change, and hoping to avoid key not found errors from being found by your users:
         *
         *     myLz.prefixFallbacks({
         *         'message': 'application_error'
         *     })
         *
         * @method prefixFallbacks
         * @protected
         * @param rules {Object} the fallback rules; each key is the prefix that should fallback to the value
         */
        prefixFallbacks : function (rules) {
            var origGetString = Object.getPrototypeOf(this).getString;
            this.getString = function getString(key) {
                var str = origGetString.call(this, key);
                if (str.indexOf(__noSuchKey__) === 0) {
                    var splitStr = (key || '').split('.');
                    var prefix = splitStr[0];
                    var fallbackPrefix = rules[prefix];
                    if (fallbackPrefix) {
                        var newKey = fallbackPrefix + '.' + _.rest(splitStr).join('.');
                        var newStr = origGetString.call(this, newKey);
                        if (newStr.indexOf(__noSuchKey__) === -1) {
                            return newStr;
                        }
                    }
                }
                return str;
            };
        }
    };

    return Localizer;
});

/**
 * Utilities for dealing with regular expressions safely.
 *
 * @class primutils.utils.regex
 */
define('primutils/utils/regex',['lodash'], function (_) {
    'use strict';

    var LITERALS = ['$', '^', '*', '?', '{', '}', '[', '\\]', '|', '(', ')', '\\', '.', '+'];
    var JOINED = '[' + LITERALS.join('') + ']';
    var that = {
        /**
         * Returns a string with escaped regular expression control characters.
         *
         * Example:
         *
         *     regex.escapeLiterals('abc') // => 'abc'
         *     regex.escapeLiterals('a[b') // => 'a\\[b'
         *
         * @method escapeLiterals
         * @static
         */
        escapeLiterals : function escapeLiterals(str) {
            if (str == null || str === '') {
                return str;
            }
            var LITERALS_RE = new RegExp(JOINED, 'g');
            return str.replace(LITERALS_RE, '\\$&');
        }
    };

    return that;
});



define('primutils/formatters/helpers/scientific',[], function () {
    'use strict';

    var that = {
        repeatZeros : function repeatZeros(count) {
            return new Array(count + 1).join('0');
        },

        isScientific : function isScientific(num) {
            return String(num).indexOf('e') !== -1;
        },

        string : function string(num) {
            if (that.isScientific(num)) {
                return that.expandScientific(num);
            }
            return String(num);
        },

        toFixed : function toFixed(num, fixedVal) {
            if (that.isScientific(num)) {
                return that.expandScientific(num) + '.' + that.repeatZeros(fixedVal);
            }
            return num.toFixed(fixedVal);
        },

        countDecimalsInNumber : function countDecimalsInNumber(num) {
            if (that.isScientific(num)) {
                return 0;
            }
            return (String(num).split('.')[1] || '').length;
        },

        wholeDigits : function wholeDigits(val) {
            if (that.isScientific(val)) {
                return that.expandScientific(val);
            }
            return parseInt(val, 10);
        },

        expandScientific : function expandScientific(num) {
            var e = parseInt(num.toString().split('+')[1], 10);
            if (e > 20) {
                e -= 20;
                num /= Math.pow(10, e);
                return num + that.repeatZeros(e);
            } else {
                throw new Error('number was not scientific notation');
            }
        }
    };
    return that;
});



define('primutils/formatters/helpers/unformatIgnore',[
    'lodash'
], function (_) {
    'use strict';

    return function (options, input) {
        var ignored = options.unformatIgnore;
        if (_.isArray(ignored)) {
            _.forEach(ignored, function (ignoreRule) {
                input = input.replace(ignoreRule, '');
            });
        } else if (_.isRegExp(ignored) || _.isString(ignored)) {
            input = input.replace(ignored, '');
        }
        return input;
    };
});


define('primutils/formatters/helpers/roundValue',['primutils/utils/regex'], function (regex) {
    'use strict';

    return function roundValue(options, curValue, nextDigit, wholeDigits, decimalDigits, reDecimalOverride) {
        // If we end up rounding curValue, these two will be the replacement strings
        var newWholeDigits = false;
        var newDecimalDigits = false;

        if (nextDigit) {
            // Displaying decimals, rounding them up (and possibly rolling up that round to the
            //  whole values)
            if (decimalDigits && parseInt(nextDigit, 10) >= 5) {
                var newDecimals = parseInt(decimalDigits, 10) + 1;

                // e.g. if we changed from "999" to "1000", that means that the non-decimal
                //  portion of the Number should be incremented
                if (newDecimals.toString().length > decimalDigits.length) {
                    newWholeDigits = String(parseInt(wholeDigits, 10) + 1);
                    newDecimalDigits = decimalDigits.replace(/./g, '0');
                }
                // e.g. if we changed from "005" to "006", we don't need to increment the
                //  non-decimal portion, but have to preserve those two zeroes in the front
                else {
                    var newDigits = String(parseInt(decimalDigits, 10) + 1);
                    while (newDigits.length < decimalDigits.length) {
                        newDigits = '0' + newDigits;
                    }
                    newDecimalDigits = newDigits;
                }
            }
            // Not displaying decimals, but the decimal portion of the value causes us to
            //  round up anyway
            else if (parseInt(nextDigit, 10) >= 5) {
                newDecimalDigits = decimalDigits.replace(/./g, '0');
                newWholeDigits = String(parseInt(wholeDigits, 10) + 1);
            }
        }

        if (newWholeDigits) {
            curValue = curValue.replace(/\d+/, newWholeDigits);
        }
        if (newDecimalDigits) {
            var decimalSeparator = regex.escapeLiterals(reDecimalOverride || options.decimalSeparator);
            var reDecimal = new RegExp('(' + decimalSeparator + ')(\\d+)');
            curValue = curValue.replace(reDecimal, '$1' + newDecimalDigits);
        }

        return curValue;
    };
});


define('primutils/formatters/helpers/numbers',[
    'lodash',
    'logger',
    'primutils/formatters/helpers/unformatIgnore',
    'primutils/formatters/helpers/roundValue'
], function (_, logger, unformatIgnore, roundValue) {
    'use strict';

    var regexThousandsStyles = {
        by_threes : 3,
        by_twos : 2,
        by_fours : 4
    };
    var validRegexThousandsStyles = _.keys(regexThousandsStyles);

    var groupThousandsByRegex = function (options, formattedValue, reDecimalOverride, decimalDisplayOverride, countOverride) {
        var neg = '';
        var split = formattedValue.split(reDecimalOverride || options.decimalSeparator);
        var portionBeforeDecimal = split[0];
        if (portionBeforeDecimal[0] === '-') {
            neg = '-';
            portionBeforeDecimal = portionBeforeDecimal.substr(1);
        }

        var count = countOverride || regexThousandsStyles[options.thousandsGroupingStyle];
        var grouped = '';
        for (var i = portionBeforeDecimal.length - 1, loop = 0; i >= 0; i--) {
            if (loop && loop % count === 0) {
                grouped = options.thousandsSeparator + grouped;
            }
            grouped = portionBeforeDecimal[i] + grouped;
            loop++;
        }
        portionBeforeDecimal = grouped;

        var portionAfterDecimal = split[1];

        formattedValue = portionBeforeDecimal;
        if (portionAfterDecimal) {
            var separator = decimalDisplayOverride || options.decimalSeparator;
            formattedValue += separator + portionAfterDecimal;
        }

        return neg + formattedValue;
    };

    var groupThousandsByTwosThenThree = function (options, formattedValue, reDecimalOverride, decimalDisplayOverride) {
        var neg = '';
        var split = formattedValue.split(reDecimalOverride || options.decimalSeparator);
        var portionBeforeDecimal = split[0];
        if (portionBeforeDecimal[0] === '-') {
            neg = '-';
            portionBeforeDecimal = portionBeforeDecimal.substr(1);
        }

        var beforeDecimalLength = portionBeforeDecimal.length;
        var portionAfterDecimal = split[1];

        if (beforeDecimalLength <= 3) {
            formattedValue = portionBeforeDecimal;
        } else {
            var splitPoint = beforeDecimalLength - 3;
            var byTwosPart = portionBeforeDecimal.substr(0, splitPoint);
            var byThreesPart = portionBeforeDecimal.substr(splitPoint, beforeDecimalLength);

            formattedValue = groupThousandsByRegex(options, byTwosPart, reDecimalOverride, decimalDisplayOverride, 2);
            formattedValue += options.thousandsSeparator + byThreesPart;
        }

        if (portionAfterDecimal) {
            var separator = decimalDisplayOverride || options.decimalSeparator;
            formattedValue += separator + portionAfterDecimal;
        }

        return neg + formattedValue;
    };

    var makeZeros = function makeZeros(count) {
        if (count === false) {
            return '';
        }

        var result = '';
        for (var i = 0; i < count; i++) {
            result += '0';
        }
        return result;
    };

    var that = {
        // way more than we can reasonably handle, but want to reject outright oddities like PPB-3753
        maxInputLength : 200,
        reDecimalsInSymbolFormat : /\$0+/,
        reDecimalsInPlainFormat : /\.0+/,
        reIsSymbolFormat : /0+\$0+/,
        countDecimalsInFormat : function countDecimalsInFormat(format) {
            var countInFormat = that.reDecimalsInPlainFormat.exec(format);
            if (that.shouldReplaceDecimalWithSymbol(format)) {
                countInFormat = that.reDecimalsInSymbolFormat.exec(format);
            }

            if (countInFormat) {
                return countInFormat[0].length - 1;
            }

            return 0;
        },
        shouldReplaceDecimalWithSymbol : function shouldReplaceDecimalWithSymbol(format) {
            return that.reIsSymbolFormat.test(format);
        },
        unformatIgnore : unformatIgnore,
        roundValue : roundValue,
        groupThousands : function (options, formattedValue, reDecimalOverride, decimalDisplayOverride) {
            var groupingStyle = options.thousandsGroupingStyle;
            if (groupingStyle === 'none') {
                return formattedValue;
            } else if (_.includes(validRegexThousandsStyles, groupingStyle)) {
                return groupThousandsByRegex(options, formattedValue, reDecimalOverride, decimalDisplayOverride);
            } else if (groupingStyle === 'by_twos_then_three') {
                return groupThousandsByTwosThenThree(options, formattedValue, reDecimalOverride, decimalDisplayOverride);
            } else {
                throw new Error('unrecognized grouping strategy: ' + groupingStyle);
            }
        },
        useDecimalsInFormat : function (baseFormat, numDecimals, includeSymbolIfFormat, showDecimalsInPlainFormat) {
            var currencyZeros = makeZeros(numDecimals);
            var decimalZeros = currencyZeros;
            if (decimalZeros.length > 0) {
                decimalZeros = '.' + decimalZeros;
            }

            if (baseFormat.indexOf('0$0') === -1) {
                if (showDecimalsInPlainFormat) {
                    return baseFormat.replace('0,0', '0,0' + decimalZeros);
                }
                return baseFormat;
            } else if (includeSymbolIfFormat) {
                return baseFormat.replace('0,0$0', '0,0$' + currencyZeros);
            } else {
                return baseFormat.replace('0,0$0', '0,0');
            }
        },
        getTrueScale : function getTrueScale(options, unit) {
            if (unit.internalType && options.hoursPer && options.hoursPer[unit.internalType] != null) {
                return options.hoursPer[unit.internalType];
            }
            if (unit.internalType && options.scales && options.scales[unit.internalType] != null) {
                return options.scales[unit.internalType];
            }
            return unit.scale;
        },
        checkRatioOpts : function (options, showIn, type) {
                        logger.assert(options[type], 'unit type `' + type + '` did not exist in options');
            var units = options[type];
            logger.assert(units[showIn], 'showIn `' + showIn + '` did not exist in `' + type + '` unit type options');
                    }
    };
    return that;
});

define('primutils/formatters/baseFormatter',[
    'lodash'
], function (_) {
    'use strict';

    function maybeWrap(fn, wrapperName, opt) {
        return opt[wrapperName] ? _.wrap(fn, opt[wrapperName]) : fn;
    }

    return function (defaultOptions, serializeFn, unserializeFn, formatFn,
                     unformatFn, validateFn, modOptions) {
        modOptions = modOptions || _.noop;
        function doModifyOptions(opts) {
            modOptions(opts || {});
        }

        var formatterFactory = function formatterFactory(newOptions) {
            var formatter = {};

            var options = _.defaults({}, newOptions || {}, defaultOptions);
            doModifyOptions(options);

            formatter.withOptions = function (newOptions) {
                var newFormatterOptions = _.defaults({}, newOptions || {}, options);
                return formatterFactory(newFormatterOptions);
            };

            formatter.getOptions = function () {
                return _.cloneDeep(options);
            };

            var wrappedSerialize = maybeWrap(serializeFn, 'serializeWrapper', options);
            formatter.serialize = function (val) {
                return wrappedSerialize(options, val);
            };

            var wrappedUnserialize = maybeWrap(unserializeFn, 'unserializeWrapper', options);
            var wrappedFormat = maybeWrap(formatFn, 'formatWrapper', options);
            formatter.unserialize = function (valStr) {
                return wrappedUnserialize(options, valStr);
            };

            formatter.unserializeAndFormat = function (valStr, formatOptions) {
                return wrappedFormat(options, wrappedUnserialize(options, valStr), formatOptions);
            };

            formatter.format = function (val, formatOptions) {
                return wrappedFormat(options, val, formatOptions);
            };

            var wrappedUnformat = maybeWrap(unformatFn, 'unformatWrapper', options);
            formatter.unformat = function (valStr, unformatOptions) {
                return wrappedUnformat(options, valStr, unformatOptions);
            };

            var wrappedValidate = maybeWrap(validateFn, 'validateWrapper', options);
            formatter.validate = function (valStr) {
                return wrappedValidate(options, valStr);
            };

            return formatter;
        };

        formatterFactory.withOptions = function (newOptions) {
            return formatterFactory(newOptions);
        };

        formatterFactory.setDefaultOptions = function setDefaultOptions(newOptions) {
            defaultOptions = _.defaults({}, newOptions || {}, defaultOptions);
        };

        formatterFactory.serialize = function (val, newOptions) {
            var options = _.defaults({}, newOptions || {}, defaultOptions);
            doModifyOptions(options);
            var wrappedFn = maybeWrap(serializeFn, 'serializeWrapper', options);
            return wrappedFn(options, val);
        };

        formatterFactory.unserialize = function (valStr, newOptions) {
            var options = _.defaults({}, newOptions || {}, defaultOptions);
            doModifyOptions(options);
            var wrappedFn = maybeWrap(unserializeFn, 'unserializeWrapper', options);
            return wrappedFn(options, valStr);
        };

        formatterFactory.format = function (val, newOptions, formatOptions) {
            var options = _.defaults({}, newOptions || {}, defaultOptions);
            doModifyOptions(options);
            var wrappedFn = maybeWrap(formatFn, 'formatWrapper', options);
            return wrappedFn(options, val, formatOptions);
        };

        formatterFactory.unformat = function (valStr, newOptions, unformatOptions) {
            var options = _.defaults({}, newOptions || {}, defaultOptions);
            doModifyOptions(options);
            var wrappedFn = maybeWrap(unformatFn, 'unformatWrapper', options);
            return wrappedFn(options, valStr, unformatOptions);
        };

        formatterFactory.validate = function (valStr, newOptions) {
            var options = _.defaults({}, newOptions || {}, defaultOptions);
            doModifyOptions(options);
            var wrappedFn = maybeWrap(validateFn, 'validateWrapper', options);
            return wrappedFn(options, valStr);
        };

        formatterFactory.unserializeAndFormat = function (valString, newOptions) {
            return formatterFactory.format(formatterFactory.unserialize(valString, newOptions), newOptions);
        };

        return formatterFactory;
    };
});

/* eslint-disable no-useless-escape */
/**
 * All formatters abide by the same core interface and theory; see
 *  [Formatter Theory](formatters.FormatterTheory.html) for more information.
 *
 * The options available for number formatters are documented in the Properties tab, below.
 *
 * The number formatter uses a custom format flag specification, best illuminated by examples:
 *
 *     define(['primutils'], function(primutils) {
 *         var numberFormatter = primutils.numberFormatter;
 *
 *         formatter = numberFormatter({
 *             format : '0' //This says to print out a number
 *                          //with no decimal separator, currency symbol,
 *                          //thousands separator and no digits following the decimal
 *         });
 *
 *         formatter.format(0); //Returns '0'
 *         formatter.format(100); //Returns '100'
 *         formatter.format(100.00); //Also returns '100'
 *
 *         //Currency
 *         currencyFormatter = numberFormatter({
 *             format : '$0,0.00',
 *             symbol : '$'
 *         });
 *
 *         currencyFormatter.format(0); //Returns '$0.00'
 *         currencyFormatter.format(10000.01); //Returns '$10,000.01'
 *
 *         //Append suffix to number for presenting hours and days
 *         naiveDurationFormatter = numberFormatter({
 *             format : '0h'
 *         });
 *
 *         naiveDurationFormatter.format(0); //Returns '0h'
 *         naiveDurationFormatter.format(7); //Returns '7h'
 *     });
 *
 * @uses formatters.FormatterTheory
 * @uses formatters.numericOptions
 * @class formatters.numberFormatter
 * @module formatters
 */

define('primutils/formatters/numberFormatter',[
    'lodash',
    'logger',
    'primutils/utils/regex',
    'primutils/formatters/helpers/scientific',
    'primutils/formatters/helpers/numbers',
    'primutils/formatters/baseFormatter'
], function (_, logger, regex, scientific, numbers, baseFormatter) {
    'use strict';

    var defaults = {
        // documented in base numericOptions
        workableDecimalPlaces : 5,

        /**
         * Whether or not digits after the decimal separator are optional.  If true and using a
         *  format of '0.00', for example, 5 <=> '5' but 5.1 <=> '5.10'
         *
         * @default false
         * @property decimalsOptional {Boolean}
         * @optional
         */
        decimalsOptional : false,

        // documented in base numericOptions
        thousandsGroupingStyle : 'by_threes',

        // documented in base numericOptions
        decimalSeparator : '.',

        // documented in base numericOptions
        thousandsSeparator : ',',

        /**
         * The currency symbol, if any, to use.  Unused by default.
         *
         * @property symbol {String}
         * @optional
         */
        symbol : '',
        /**
         * If the symbol is present in the format, prints it out.
         *
         * @property printSymbolIfPresent
         * @type {Boolean}
         * @optional
         * @default true
         */
        printSymbolIfPresent : true,
        /**
         * A string to postfix any outputted number with; ignored when parsing user-entered data.
         *
         * Empty/unused by default.
         *
         * @property postfix
         * @optional
         */
        postfix : '',
        /**
         * A symbol to use to indicate negative numbers.  In the US, for example, this is a '-'
         *  sign, where negative one hundred gets written as: '-100'
         *
         * Only used with negative numbers, when no negativeFormat option has been specified.
         *
         * @property negativeSymbol {String}
         * @optional
         * @default '-'
         */
        negativeSymbol : '-',
        /**
         * The format flags used for displaying numbers as positive Strings to the user.
         *  See examples in the class description for more information.
         *
         * See also {{#crossLink "formatters.numberFormatter/negativeFormat:property"}}{{/crossLink}}
         *
         * Supported tokens:
         * * `$` - the currency symbol, also used for the percent symbol when `symbol` is `%`
         * * `,` - indicates that thousands grouping should be used
         * * `.` - indicates that decimal values should be printed out
         * * `<` - indicates where the denominator should go.  if `numeratorType` is specified and this isn't in the format (or `negativeFormat`) strings, this will be added to the end of the format automatically
         * * other chars - left in verbatim
         *
         * @property format
         * @type String
         * @optional
         * @default '0'
         */
        format : '0',
        /**
         * The format flags used for displaying negative numbers as Strings to the user.
         *
         * See also {{#crossLink "formatters.numberFormatter/format:property"}}{{/crossLink}}
         *
         * If not specified, will default to the 'format' option.
         *
         * When specified, an explicit '-' sign will not be added to the number.
         *
         * As examples.  If using a format of '0,0.00' and no negativeFormat,
         *  0.5 <=> '0.50' and -0.5 <=> '-0.50'.
         *  If using a format of '0,0.00' and a negativeFormat of '(0,0.00)',
         *  0.5 <=> '0.50' and -0.5 <=> '(0.50)'.
         *
         * @property negativeFormat
         * @type String
         * @optional
         **/
        negativeFormat : null,
        /**
         * A multipler to use when formatting/unformatting values.  If the multiplier is 100 and
         *  using a format of '0,0.0$' and a symbol of '%', for example, 0.5 <=> '50.0%' and
         *  1.0 <=> '100.0%'.
         *
         * @property formattedMultiplier
         * @type Number
         * @optional
         */
        formattedMultiplier : undefined,

        /**
         * A pattern (RegExp or String), or list of patterns, that will be ignored in any string
         *  the user types, for unformat() or validate() calls.
         *
         * @property unformatIgnore
         * @type String|RegExp|Array
         * @optional
         */
        //unformatIgnore = [],

        /**
         * What unit to show as the denominator, assuming `denominatorType` is not null.
         * As [ratioFormatter#numeratorShowIn](formatters.ratioFormatter.html#property_numeratorShowIn), this option must match one of the keys of this formatter's `unit` config.
         *
         * @property denominatorShowIn
         * @type String
         * @optional
         * @default 'h' (Hours)
         */
        denominatorShowIn : 'h',
        /**
         * As the [ratioFormatter `numeratorType`](formatters.ratioFormatter.html#property_numeratorType) option, but for the denominator.
         *
         * Unlike the ratioFormatter, a denominatorType of null indicates to the number formatter that the value is not a ratio at all - this is the default.
         *
         * @property denominatorType
         * @type String
         * @optional
         * @default null (indicating that there is no denominator)
         */
        denominatorType : null

        //serializeWrapper,
        //unserializeWrapper,
        //formatWrapper,
        //unformatWrapper,
        //validateWrapper
    };

    function getValue(options, opt) {
        if (_.isFunction(options[opt])) { return options[opt](); }
        return options[opt];
    }
    function isTrueOrUndefined(fmtOpts, opt) {
        return !fmtOpts || fmtOpts[opt] == null || fmtOpts[opt] === true;
    }

    var _maxPrecision_ = null;
    function digitFlag(value, decimal, printDecimals) {
        if (!decimal) {
            //First 0 - print all numbers before decimal
            if (value >= 0) {
                return scientific.wholeDigits(value);
            } else if (Math.abs(value) < 1) {
                return '0';
            } else {
                return scientific.wholeDigits(value).toString().substring(1);
            }
        } else if (decimal && printDecimals) {
            //These are the 0s after the decimal separator
            if (_maxPrecision_ && _maxPrecision_.length > 0) {
                var nextDigit = _maxPrecision_.slice(0, 1);
                _maxPrecision_ = _maxPrecision_.substring(1);
                return nextDigit;
            } else {
                return '0';
            }
        }

        return '';
    }

    function checkOpts(options) {
        var denominatorType = options.denominatorType;
        if (denominatorType != null) {
            if (!_.includes(options.format, '<')) {
                options.format += '<';
            }
            if (options.negativeFormat != null && !_.includes(options.negativeFormat, '<')) {
                options.negativeFormat += '<';
            }
        }
                if (options.denominatorType != null) {
            numbers.checkRatioOpts(options, options.denominatorShowIn, options.denominatorType);
        }
            }

    function serialize(options, val) {
        if (val == null) {
            return null;
        }

        if (!_.isNumber(val)) {
            logger.error('serialize: transfer form is a Number');
            return null;
        }

        if (_.isNaN(val)) {
            logger.error('numberFormatter should not be passed NaN');
            return null;
        }
        if (val === Number.POSITIVE_INFINITY || val === Number.NEGATIVE_INFINITY) {
            logger.error('numberFormatter should not be passed infinity');
            return null;
        }

        var modifier = Math.pow(10, options.workableDecimalPlaces);
        return Math.round(modifier * val) / modifier;
    }

    function unserialize(options, val) {
        if (val == null || val === '') {
            return null;
        }

        if (!_.isNumber(val)) {
            logger.error('unserialize: transfer form is a Number');
        }

        var modifier = Math.pow(10, options.workableDecimalPlaces);
        return Math.round(modifier * val) / modifier;
    }

    function groupThousands(options, formattedValue, shouldGroup, reDecimalOverride, decimalDisplayOverride) {
        if (shouldGroup) {
            return numbers.groupThousands(options, formattedValue, reDecimalOverride, decimalDisplayOverride);
        }

        return formattedValue;
    }

    function format(options, value, formatOptions) {
        if (options.decimalSeparator === options.thousandsSeparator) {
            throw new Error('decimalSeparator matches thousandsSeparator; values cannot be handled unambiguously');
        }

        if (value == null) {
            return '';
        }

        if (!_.isNumber(value)) {
            logger.fatal('format: workable form for numberFormatter is a Number');
            throw new Error('Must be a number');
        }

        if (_.isNaN(value)) {
            logger.error('numberFormatter should not be passed NaN');
            return '';
        }
        if (value === Number.POSITIVE_INFINITY || value === Number.NEGATIVE_INFINITY) {
            logger.error('numberFormatter should not be passed infinity');
            return '';
        }

        var usedFormat = getValue(options, 'format');
        var negFmt = getValue(options, 'negativeFormat');
        if (value < 0 && negFmt != null) {
            usedFormat = negFmt;
        }
        if (usedFormat == null) {
                        logger.error('No format given to formatter, returning empty string');
                        return '';
        }

        if (options.formattedMultiplier != null) {
            value = parseFloat((value * options.formattedMultiplier).toFixed(Math.max(options.workableDecimalPlaces, 12)));
        }
        var denominatorType = options.denominatorType;
        var denomMultiplier = null;
        if (denominatorType != null) {
            var units = options[denominatorType];
            denomMultiplier = numbers.getTrueScale(options, units[options.denominatorShowIn]);
        }
        if (denomMultiplier != null) {
            value = parseFloat((value / denomMultiplier).toFixed(Math.max(options.workableDecimalPlaces, 12)));
        }

        var isScientific = scientific.isScientific(value);
        if (isScientific && Math.abs(value) < 1) {
            // if it's this small, let's face it, not worth bothering about
            return format(options, 0, formatOptions);
        }

        var valueHasDecimals = Math.floor(value) !== value;
        var decimalsInFormat = numbers.countDecimalsInFormat(usedFormat);
        if (formatOptions && formatOptions.truncate === false) {
            decimalsInFormat = Math.max(decimalsInFormat, scientific.countDecimalsInNumber(value));
        }

        _maxPrecision_ = '';
        if (decimalsInFormat !== 0) {
            //This will handle rounding when you are given .99999999999999
            var valueFixed1 = scientific.toFixed(value, decimalsInFormat + 2);
            _maxPrecision_ = valueFixed1.split('.')[1];
            value = parseFloat(valueFixed1);
        } else if (valueHasDecimals) {
            var valueFixed2 = scientific.toFixed(value, 2);
            _maxPrecision_ = valueFixed2.split('.')[1];
            value = parseFloat(valueFixed2);
        }

        var printDecimals = !options.decimalsOptional || parseInt(_maxPrecision_, 10) > 0;
        var replaceDeciWithSymbol = numbers.shouldReplaceDecimalWithSymbol(usedFormat);
        var group = _.includes(usedFormat, ',');
        var symbol = getValue(options, 'symbol');
        usedFormat = usedFormat.replace(/[0,\\.]+/, '0'); // e.g. 0,0.000 -> 0
        if (replaceDeciWithSymbol) {
            usedFormat = usedFormat.replace(/\$0*/, ''); // e.g. altogether with above, 0$0 -> 0
        }

        var wholeDigits = '';
        var decimalDigits = '';
        var numericValuePart = '';
        numericValuePart += (wholeDigits = digitFlag(value, false, printDecimals));
        if (printDecimals && decimalsInFormat) {
            numericValuePart += options.decimalSeparator; // can be replace by symbol, below
        }
        for (var i = 0; i < decimalsInFormat; i++) {
            decimalDigits += digitFlag(value, true, printDecimals);
        }
        numericValuePart += decimalDigits;

        var roundingDigit = _maxPrecision_ && _maxPrecision_[0];
        numericValuePart = numbers.roundValue(options, numericValuePart, roundingDigit, wholeDigits, decimalDigits);
        numericValuePart = groupThousands(options, numericValuePart, group);

        if (replaceDeciWithSymbol) {
            numericValuePart = numericValuePart.replace(options.decimalSeparator, '$');
        }

        // Insert a negative symbol if we should have one, but didn't insert one already
        if (value < 0 && negFmt == null) {
            numericValuePart = numericValuePart.replace(/(\d)/, options.negativeSymbol + '$1');
        }


        var formattedValue = usedFormat
            .replace('0', numericValuePart)
            .replace(/\$/, function () {
                return options.printSymbolIfPresent && symbol || '';
            })
            .replace('<', function () {
                return '/' + options.denominatorShowIn;
            });
        if (options.postfix) {
            return formattedValue + options.postfix;
        }
        return formattedValue;
    }

    function replaceSymbolForRegex(options, format) {
        var symb = getValue(options, 'symbol');
        if (symb) {
            format = format.replace(symb, function () { return '$'; });
        }
        if (options.postfix) {
            format = format.replace(options.postfix, '');
        }
        return format;
    }

    function baseReplace(options, valStr) {
        var thousands = options.thousandsSeparator === '.' ? '\\.' : options.thousandsSeparator;
        var deci = options.decimalSeparator === '.' ? '\\.' : options.decimalSeparator;
        var replaced = valStr.replace(new RegExp(thousands, 'g'), '')
                              .replace(new RegExp(deci, 'g'), '.');
        return replaceSymbolForRegex(options, replaced);
    }

    function patternFromString(options, valStr) {
        var pattern = valStr.replace(/\d+/g, '0')
            .replace(options.thousandsSeparator, ',')
            .replace(options.decimalSeparator, '.')
            .replace(/(?:0,)+/, '0,');
        pattern = replaceSymbolForRegex(options, pattern);
        return pattern;
    }

    function reducedPattern(options, pattern) {
        if (getValue(options, 'negativeFormat') == null) {
            pattern = pattern.replace(/-/, '');
        }
        return pattern.replace(/,/g, '')
            .replace('$', '')
            .replace('.', '')
            .replace('+', '')
            .replace('<', '')
            .replace(/ /g, '') // TODO do we really want to ignore spaces?
            .replace(/\d+/g, '0');
    }

    function fuzzyMatch(options, patternToMatchAgainst, pattern) {
        if (patternToMatchAgainst == null) {
            return false;
        }
        return patternToMatchAgainst === pattern ||
            reducedPattern(options, patternToMatchAgainst) === reducedPattern(options, pattern);
    }

    function unformat(options, valStr, unformatOptions) {
        if (options.decimalSeparator === options.thousandsSeparator) {
            throw new Error('decimalSeparator matches thousandsSeparator; values cannot be handled unambiguously');
        }

        if (valStr == null || valStr === '') {
            return null;
        }

        valStr = _.unescape(valStr);
        valStr = numbers.unformatIgnore(options, valStr);

        if (valStr === '') {
            return null;
        }

        // TODO good idea, or match behavior with validate() function?
        var rawNumberString = baseReplace(options, valStr).replace(/[^\$\d\.\-\+]/g, '');
        if (rawNumberString === '') {
            return NaN;
        }

        var denominatorType = options.denominatorType;
        var denomMultiplier = null;
        if (denominatorType != null) {
            var units = options[denominatorType];
            denomMultiplier = numbers.getTrueScale(options, units[options.denominatorShowIn]);
            var unitCount = 0;
            _.forEach(units, function (unit, key) {
                if (_.includes(valStr, '/' + key)) {
                    unitCount++;
                    denomMultiplier = numbers.getTrueScale(options, units[key]);
                }
            });
            if (unitCount > 1) {
                return NaN;
            }
        }

        var negFmt = getValue(options, 'negativeFormat');
        var useNegative = negFmt != null && fuzzyMatch(options, negFmt, patternFromString(options, valStr));
        var usedFormat = useNegative ? negFmt : getValue(options, 'format');

        if (numbers.shouldReplaceDecimalWithSymbol(usedFormat)) {
            rawNumberString = rawNumberString.replace('$', '.');
        } else {
            rawNumberString = rawNumberString.replace('$', '');
        }

        var baseValue = parseFloat(rawNumberString);

        if (options.formattedMultiplier != null) {
            baseValue = baseValue / options.formattedMultiplier;
        }
        if (denomMultiplier != null) {
            baseValue = baseValue * denomMultiplier;
        }

        if (useNegative) {
            baseValue = -Math.abs(baseValue);
        }

        if (isTrueOrUndefined(unformatOptions, 'truncate')) {
            var modifier = Math.pow(10, options.workableDecimalPlaces);
            return Math.round(modifier * baseValue) / modifier;
        } else {
            return baseValue;
        }
    }

    function countMatches(str, toMatch) {
        var match = str.match(toMatch);
        return match ? match.length : 0;
    }

    function hasSurroundingNumbers(str, token) {
        if (!token) { return false; }

        var tester = new RegExp('\\d' + token + '\\d');
        return tester.test(str);
    }

    function validate(options, valueString) {
        if (options.decimalSeparator === options.thousandsSeparator) {
            return false;
        }

        if (valueString == null || valueString === '') {
            return true;
        }

        valueString = _.unescape(valueString);
        valueString = numbers.unformatIgnore(options, valueString);

        if (valueString === '') {
            return true;
        }
        if (valueString.length >= numbers.maxInputLength) {
            return false;
        }

        var symbol = getValue(options, 'symbol');
        var symbolEscaped = symbol && regex.escapeLiterals(symbol);
        var hasMultipleSymbols = symbol && countMatches(valueString, symbolEscaped) > 1;
        var checkDollarChar = !_.includes(symbol, '$');
        var hasDollarCharacter = countMatches(valueString, /\$/g) > 0;
        if (hasMultipleSymbols || (checkDollarChar && hasDollarCharacter)) {
            return false;
        }

        var postfixEscaped = options.postfix && regex.escapeLiterals(options.postfix);
        if (hasSurroundingNumbers(valueString, postfixEscaped)) {
            return false;
        }

        var replaced = baseReplace(options, valueString);
        var unlocalizedNumericString = replaced.replace(/[^\d\.\-\+]/g, '');

        // Only allow one decimal
        //  TODO: same for '-', '+' chars?  user could sneak them on at the end if they wanted
        if (countMatches(unlocalizedNumericString, /\./g) > 1) {
            return false;
        }

        // If there's some numeric value with random junk intermixed - like "4-0" - reject that
        if (countMatches(unlocalizedNumericString, /\d[^\d\.]+\d/) > 0) {
            return false;
        }

        var parsedVal = parseFloat(unlocalizedNumericString);
        if (!_.isNumber(parsedVal) || _.isNaN(parsedVal)) {
            return false;
        }

        var simplifiedValueString = valueString;
        var denominatorType = options.denominatorType;
        if (denominatorType != null) {
            var units = options[denominatorType];
            var unitCount = 0;
            var keyFound = null;
            _.forEach(units, function (unit, key) {
                if (_.includes(simplifiedValueString, '/' + key)) {
                    unitCount++;
                    keyFound = key;
                }
            });
            if (unitCount > 1) {
                return false;
            } else if (unitCount) {
                simplifiedValueString = simplifiedValueString.replace('/' + keyFound, '<');
            }
        }

        var negFmt = getValue(options, 'negativeFormat');
        var posFmt = getValue(options, 'format');
        var enteredPattern = patternFromString(options,
                baseReplace(options, simplifiedValueString));
        var checkNegativeFormat = negFmt != null;

        if (parsedVal > 0 && fuzzyMatch(options, '0', enteredPattern)) {
            return true;
        }
        if (parsedVal < 0 && fuzzyMatch(options, '-0', enteredPattern)) {
            return true;
        }

        var matchesNegativeFormat = checkNegativeFormat && fuzzyMatch(options, negFmt, enteredPattern);
        var matchesPositiveFormat = fuzzyMatch(options, posFmt, enteredPattern);
        if (matchesPositiveFormat || matchesNegativeFormat) {
            return true;
        }

        return false;
    }

    return baseFormatter(defaults, serialize, unserialize, format, unformat, validate, checkOpts);
});


define('primutils/utils/dates',[
    'lodash',
    'logger',
    'moment'
], function (_, logger, moment) {
    'use strict';

    var utcishAllowed = false;
    var patchMethods = [
        'isSame', 'isAfter', 'isBefore',
        'valueOf', 'toDate', 'toArray', 'toISOString', 'unix',
        'calendar', 'diff', 'from', 'fromNow',
        'add', 'subtract', 'endOf', 'startOf',
        'isDST', 'isDSTShifted'
    ];
    var originals = _.reduce(patchMethods, function (memo, fnName) {
        memo[fnName] = moment.fn[fnName];
        return memo;
    }, {});

    function utcishSame(m1, m2) {
        return m1.year() === m2.year() &&
            m1.month() === m2.month() &&
            m1.date() === m2.date() &&
            m1.hour() === m2.hour() &&
            m1.minute() === m2.minute() &&
            m1.seconds() === m2.seconds() &&
            m1.milliseconds() === m2.milliseconds();
    }

    function utcishAfter(m1, m2) {
        if (m1.year() !== m2.year()) {
            return m1.year() > m2.year();
        }
        if (m1.month() !== m2.month()) {
            return m1.month() > m2.month();
        }
        if (m1.date() !== m2.date()) {
            return m1.date() > m2.date();
        }
        if (m1.hour() !== m2.hour()) {
            return m1.hour() > m2.hour();
        }
        if (m1.minute() !== m2.minute()) {
            return m1.minute() > m2.minute();
        }
        if (m1.seconds() !== m2.seconds()) {
            return m1.seconds() > m2.seconds();
        }
        if (m1.milliseconds() !== m2.milliseconds()) {
            return m1.milliseconds() > m2.milliseconds();
        }
        return false;
    }

    function utcishBefore(m1, m2) {
        if (m1.year() !== m2.year()) {
            return m1.year() < m2.year();
        }
        if (m1.month() !== m2.month()) {
            return m1.month() < m2.month();
        }
        if (m1.date() !== m2.date()) {
            return m1.date() < m2.date();
        }
        if (m1.hour() !== m2.hour()) {
            return m1.hour() < m2.hour();
        }
        if (m1.minute() !== m2.minute()) {
            return m1.minute() < m2.minute();
        }
        if (m1.seconds() !== m2.seconds()) {
            return m1.seconds() < m2.seconds();
        }
        if (m1.milliseconds() !== m2.milliseconds()) {
            return m1.milliseconds() < m2.milliseconds();
        }
        return false;
    }

    function isSame(m1, m2) {
        // Only known usecase for UTC moments in Prime is the dateFormatter parser - so let's just compare the values
        if (utcishAllowed && (m1.isUTC() || m2.isUTC())) {
            return utcishSame(m1, m2);
        }
        return originals.isSame.call(m1, m2);
    }

    function isAfter(m1, m2) {
        // Only known usecase for UTC moments in Prime is the dateFormatter parser - so let's just compare the values
        if (utcishAllowed && (m1.isUTC() || m2.isUTC())) {
            return utcishAfter(m1, m2);
        }
        return originals.isAfter.call(m1, m2);
    }

    function isBefore(m1, m2) {
        // Only known usecase for UTC moments in Prime is the dateFormatter parser - so let's just compare the values
        if (utcishAllowed && (m1.isUTC() || m2.isUTC())) {
            return utcishBefore(m1, m2);
        }
        return originals.isBefore.call(m1, m2);
    }

    /**
     * Methods for dealing with dates in Prime.
     *
     * With dateFormatter's `fakeUTC` parse behavior, the only known usecase for UTC moment objects in Prime is for globalDateTime objects - in which case, the timezone offset doesn't matter, but individual fields (`year()`, `month()`, etc) are reported correctly.
     *
     * The methods in this class, therefore, mostly center around implementing isSame/Before/After checks that are 'UTC-safe', in that they check all the relevant fields, but not the milliseconds since epoch (which is, of course, affected by the timezone offset).
     *
     * @class utils.dates
     * @module protected
     * @protected
     */
    var that = {
        isSame : isSame,
        isAfter : isAfter,
        isBefore : isBefore,

        patchMomentWarnings : function () {
                        var oldToJSON = moment.fn.toJSON;
            moment.fn.toJSON = function () {
                logger.error('JSON.stringify(moment) will not result in correct serialization to the Prime rest project - more at https://confluence.oraclecorp.com/confluence/display/PRIMEENGINEERING/Using+Dates+in+Prime#UsingDatesinPrime-ExplicitSerialization/Unserialization');
                return oldToJSON.apply(this, arguments);
            };
                    },

        /**
         * Overrides the default Moment.js implementations of `isBefore`, `isSame`, `isAfter`, and `valueOf` to handle the UTC behaviors, as described in this class's description.
         *
         * Original Moment.js behaviors can be restored with the `unpatchMoment` function, also exposed on this class.
         * Unpatching these methods after dates have been obtained from the server and unserialized is...not advised.  Strongly.
         *
         * @method patchMoment
         * @protected
         * @static
         */
        patchMoment : function () {
            utcishAllowed = true;
            _.forEach(patchMethods, function (fnName) {
                moment.fn[fnName] = function () {
                    if (this.isUTC()) {
                        this.local(true);
                    }
                    return originals[fnName].apply(this, arguments);
                };
            });
        },
        unpatchMoment : function () {
            utcishAllowed = false;
            _.forEach(patchMethods, function (fnName) {
                moment.fn[fnName] = originals[fnName];
            });
        }
    };

    return that;
});


/**
 * All formatters abide by the same core interface and theory; see
 *  [Formatter Theory](formatters.FormatterTheory.html) for more information.
 *
 * The date formatter uses [MomentJS format flags](http://momentjs.com/docs/#/displaying/format/)
 *
 * Date formatters have three primary configuration options; format, unserializeFormat, and
 *  serializedFormat.  These options are documented in the Properties tab, below.
 *
 * *Note*: for developers in the PGBU, unserializeFormat and serializeFormat are not specified
 *  by most developers.  Instead, import the 'userFormatters' module, and if you need to override
 *  the display format option, create a new formatter using the 'withOptions' function.
 *  The documentation in the UP project has more information.
 *
 * @uses formatters.FormatterTheory
 * @class formatters.dateFormatter
 * @module formatters
 */
define('primutils/formatters/dateFormatter',['lodash',
    'moment',
    'logger',
    'primutils/utils/dates',
    'primutils/formatters/helpers/unformatIgnore',
    'primutils/formatters/baseFormatter'
], function (_, moment, logger, dateUtils, unformatIgnore, baseFormatter) {
    'use strict';

    var defaults = {
        /**
         * The format flags used for displaying dates as Strings to the user.  Please note that
         *  the date formatter uses
         *  [MomentJS format flags](http://momentjs.com/docs/#/displaying/format/).
         *
         * @property format
         * @type String
         * @optional
         * @default 'MM-DD-YYYY'
         */
        format : 'MM-DD-YYYY',
        /**
         * Alternate formats used when parsing values given by user when unformat is invoked with `{forgiving: true}` (as OUI will).
         *
         * @property alternateFormats
         * @type Array<String>
         * @optional
         * @default []
         */
        alternateFormats : [],

        // TODO bad format, probably, ever
        // Example: 2013-05-14T00:00:00.000-0400
        /**
         * The format used for unserialization of dates sent by the server.
         *
         * @property unserializeFormat
         * @type String
         * @optional
         * @default 'YYYY-MM-DD[T]HH:mm:ss.ZZ'
         */
        unserializeFormat : 'YYYY-MM-DD[T]HH:mm:ss.ZZ',

        /**
         * The format used for serialization of dates sent to the server.
         *
         * @property serializeFormat
         * @type String
         * @optional
         * @default 'YYYY-MM-DD[T]HH:mm:ss.ZZ'
         */
        serializeFormat : 'YYYY-MM-DD[T]HH:mm:ss.ZZ'

        /**
         * A pattern (RegExp or String), or list of patterns, that will be ignored in any string
         *  the user types, for unformat() or validate() calls.
         *
         * @property unformatIgnore
         * @type String|RegExp|Array
         * @optional
         */
        //unformatIgnore = [],

        //serializeWrapper,
        //unserializeWrapper,
        //formatWrapper,
        //unformatWrapper,
        //validateWrapper
    };

    var fakeUTCParsingEnabled = false;
    var cachedStrategies = {};
    var utcTzOffsetStr = moment.utc().format('Z');

    // Various strategies we can use to parse moments - some won't work in a given browser, hence we inherently have to them all until we find a good one on the first date we receive from the server
    var unserializeStrategies = {
        rawDate : function rawDateUnserialize(serverString) {
            return moment(new Date(serverString));
        },
        rawDatePlusTz : function rawDateModded(serverString) {
            var date = new Date(serverString);
            var dt = moment(date);
            var origOffset = -dt.utcOffset();
            dt.add(origOffset, 'minutes');

            var newOffset = -dt.utcOffset();
            if (newOffset !== origOffset) {
                dt.add(newOffset - origOffset, 'minutes');
            }

            return dt;
        },
        // The old, naive, but always correct behavior - used as a fallback and the tests for all other behaviors
        correct : function momentUnserialize(serverString, format) {
            return moment(serverString, format);
        },
        momentDirect : function momentDirect(serverString) {
            return moment(serverString);
        },
        rawDateWithZZReplaced : function rawDateWithZZReplaced(serverString) {
            return moment(new Date(serverString.replace(/\.([+-]\d\d)(\d\d)/, '$1:$2')));
        },
        // Chrome will parse '2015-02-21T12:14:35-0500', but not '2015-02-21T12:14:35.-0500'
        replacePeriodWithNothing : function replacePeriodWithNothing(serverString) {
            return moment(new Date(serverString.replace('.', '')));
        },
        fakeUTC : function fakeUTC(serverString) {
            return moment.utc(new Date(serverString + utcTzOffsetStr));
        },
        splitString : function manuallyToken(serverString) {
            // YYYY-MM-DDTHH:mm:ss
            var year = parseInt(serverString, 10);
            serverString = serverString.substr(5);
            var month = parseInt(serverString, 10);
            serverString = serverString.substr(3);
            var day = parseInt(serverString, 10);
            serverString = serverString.substr(3);
            var hour = parseInt(serverString, 10);
            serverString = serverString.substr(3);
            var minute = parseInt(serverString, 10);
            serverString = serverString.substr(3);
            var second = parseInt(serverString, 10);
            return moment(new Date(year, month - 1, day, hour, minute, second, 0));
        }
    };
    var testers = {
        fakeUTC : function (d1, d2, format) {
            if (!fakeUTCParsingEnabled || format.match(/\.Z/)) {
                return false;
            }
            d1.locale('en');
            d2.locale('en');
            return d1.isValid() && dateUtils.isSame(d1, d2);
        },
        splitString : function (d1, d2, format) {
            if (format !== 'YYYY-MM-DD[T]HH:mm:ss') {
                return false;
            }
            return testers.standard(d1, d2, format);
        },
        standard : function (d1, d2) {
            if (d1 == null) {
                return false;
            }
            d1.locale('en');
            d2.locale('en');
            return d1.isValid() && d1.isSame(d2) && d1.utcOffset() === d2.utcOffset();
        }
    };

    function testStrat(fn, serverString, format, stratName) {
        var stratDate = fn(serverString, format);
        var correctDate = unserializeStrategies.correct(serverString, format);
        // Many parsing strategies will be outright wrong, so we _must_ reject those strats
        var tester = testers[stratName] || testers.standard;
        if (tester(stratDate, correctDate, format)) {
            return true;
        }
        return false;
    }

    // Lets us try various unserialize strategies, such as deferring to native Date; gives us a chance to use the fastest strategy available
    function getUnserializeStrategy(serverString, format) {
        // Once a strategy is chosen, use that for the rest of the page lifetime
        if (cachedStrategies[format] !== undefined) {
            return cachedStrategies[format];
        }

        var strat = false;
        function applyStrat(stratName) {
            if (strat) { return; }

            if (testStrat(unserializeStrategies[stratName], serverString, format, stratName)) {
                strat = stratName;
            }
        }

        // Try each strategy in turn, in an loosely empirically driven attempt to prioritize parsing strategies I've seen that are faster
        applyStrat('rawDate');
        applyStrat('fakeUTC');
        applyStrat('splitString'); // surprisingly fast; actually wins on IE9 (limited testing), not that we need to support that browser.  /barely/ loses out to fakeUTC in FF/Chrome, so it gets #2
        applyStrat('replacePeriodWithNothing');
        applyStrat('rawDateWithZZReplaced');
        applyStrat('rawDatePlusTz');
        applyStrat('momentDirect');
        if (strat === false) {
            strat = 'correct';
        }

        cachedStrategies[format] = strat;
        return strat;
    }

    var _momentTester = moment();
    function checkFauxVariant(dt) {
        return dt.__pgbu_original_format_fn != null;
    }

    function checkMoment(dt) {
        if (dt != null && moment.isMoment(dt)) {
            return true;
        } else {
            logger.fatal('dateFormatter only works with Moment.js objects');
            throw new Error('dateFormatter only works with Moment.js objects');
        }
    }

    function serialize(options, dt) {
        if (dt == null) {
            return null;
        }

        checkMoment(dt);

        if (!dt.isValid()) {
            logger.error('dateFormatter should only be passed Moment objects that pass isValid()');
            return null;
        }

        // ...wut?  Allows dateFormatter to serialize properly with FAUX variant
        if (checkFauxVariant(dt)) {
            return dt.__pgbu_original_format_fn(options.serializeFormat);
        } else {
            return dt.format(options.serializeFormat);
        }
    }

    function unserialize(options, serializedDateString) {
        if (serializedDateString == null || serializedDateString === '') {
            return null;
        }

        var strat = 'correct';
        var unserFmt = options.unserializeFormat;

                strat = getUnserializeStrategy(serializedDateString, unserFmt);
        
        return unserializeStrategies[strat](serializedDateString, unserFmt);
    }

    function format(options, dt) {
        if (dt == null) {
            return '';
        }

        checkMoment(dt);

        if (!dt.isValid()) {
            logger.error('dateFormatter should only be passed Moment objects that pass isValid()');
            return '';
        }

        return dt.format(options.format);
    }

    var FAUX_PREFIX = 'X\'X_';
    var FAUX_SUFFIX = '_X\u00A5X\u59CB';
    var reFaux = new RegExp(FAUX_PREFIX + '|' + FAUX_SUFFIX, 'g');
    function correctFauxVariant(ds) {
        return ds.replace(reFaux, '');
    }

    function genForgivingZeroFormat(format) {
        /* eslint-disable no-useless-escape */
                if (format.match(/MMM|DDD|ddd|\[.*(MM|DD|dd|ww|HH|hh|mm|ss).*\]/)) {
            logger.warn('Format given for dates includes a pattern which may not work correctly with the "forgiveLeadingZeros" option.  File a request if fixing this is necessary and you\'re seeing issues.');
        }
                /* eslint-enable no-useless-escape */
        return format.replace(/MM(?!M)|DD(?!D)|dd(?!d)|ww|HH|hh|mm|ss/g, function (match) {
            return match[0];
        });
    }

    function unformat(options, dateString, formatOptions) {
        if (dateString == null || dateString === '') {
            return null;
        }

        dateString = unformatIgnore(options, dateString);
        if (checkFauxVariant(_momentTester)) {
            dateString = correctFauxVariant(dateString);
        }

        if (dateString === '') {
            return null;
        }

        var dt = moment(dateString, options.format, true);
        if (dt.isValid()) {
            return dt;
        }

        if (formatOptions && formatOptions.forgiving && options.alternateFormats.length) {
            var formats = options.alternateFormats;
            dt = moment(dateString, formats, true);
            if (dt && dt.isValid()) {
                return dt;
            }
        }

        if (options.forgiveLeadingZeros) {
            var format = genForgivingZeroFormat(options.format);
            dt = moment(dateString, format, true);
            if (dt.isValid()) {
                return dt;
            }
        }

        return false;
    }

    function validate(options, dateString, formatOptions) {
        if (dateString == null || dateString === '') {
            return true;
        }

        var unformatted = unformat(options, dateString, formatOptions);
        return unformatted == null || !!(unformatted && unformatted.isValid());
    }

    var fmt = baseFormatter(defaults, serialize, unserialize, format, unformat, validate);

    fmt.enableFakeUTCParsing = function () {
        fakeUTCParsingEnabled = true;
    };
    fmt.disableFakeUTCParsing = function () {
        fakeUTCParsingEnabled = false;
        cachedStrategies = {};
    };

    
    return fmt;
});


/* eslint-disable no-useless-escape */
define('primutils/formatters/helpers/duration',[
    'lodash',
    'logger',
    'primutils/formatters/helpers/numbers',
    'primutils/formatters/helpers/scientific'
], function (_, logger, numbers, scientific) {
    'use strict';

    var utils = {
        numbers : numbers,
        get reRatioChunk() {
            return /([+-]?[\d\. ]+)\s*(.*)\s*/g;
        },
        get reNextChunk() {
            return /([+-]?[\d\. ]+)\s*([^\d+-]*)\s*/g;
        },
        serialize : function serialize(options, val) {
            if (val == null) {
                return null;
            }

            if (!_.isNumber(val)) {
                logger.error('serialize: transfer form is a Number');
                return null;
            }

            if (_.isNaN(val)) {
                logger.error('formatters should not be passed NaN');
                return null;
            }
            if (val === Number.POSITIVE_INFINITY || val === Number.NEGATIVE_INFINITY) {
                logger.error('numberFormatter should not be passed infinity');
                return null;
            }

            return val;
        },
        unserialize : function unserialize(options, valStr) {
            if (valStr == null || valStr === '') {
                return null;
            }

            if (!_.isNumber(valStr)) {
                logger.error('unserialize: transfer form is a Number');
            }

            return valStr;
        },
        fixDecimal : function fixDecimal(val, places) {
            var multiplier = Math.pow(10, places);
            return scientific.toFixed(Math.round(multiplier * val) / multiplier, places);
        },
        stringy : function stringy(options, value) {
            var base = scientific.string(value);
            return numbers.groupThousands(options, base, '.');
        },
        clipResult : function clipResult(val, places) {
            var modifier = Math.pow(10, places);
            return Math.round(modifier * val) / modifier;
        },
        getTrueScale : numbers.getTrueScale,
        checkRatioOpts : numbers.checkRatioOpts,
        maxInputLength : numbers.maxInputLength
    };

    return utils;
});


/**
 * All formatters abide by the same core interface and theory; see
 *  [Formatter Theory](formatters.FormatterTheory.html) for more information.
 *
 * The options for duration formatters are documented in the Properties tab below.
 *
 * The duration formatter formats durations according to how the PGBU deals with durations.
 *  Examples:
 *
 *     define(['primutils'], function(primutils) {
 *         var durationFormatter = primutils.durationFormatter;
 *
 *         durationFormatter.format(5); //Returns '5h'
 *         durationFormatter.format(100); //Returns '100h'
 *
 *         durationFormatter.unformat('1y 1m 1w 1d 1h'); //Returns 2221
 *         durationFormatter.unformat('1.2y'); //Returns 2400
 *
 *         durationFormatter.validate('1h'); //Returns true
 *         durationFormatter.validate('pie'); //Returns false
 *     });
 *
 * @uses formatters.FormatterTheory
 * @uses formatters.numericOptions
 * @class formatters.durationFormatter
 * @module formatters
 */
define('primutils/formatters/durationFormatter',['lodash',
    'logger',
    'primutils/formatters/helpers/scientific',
    'primutils/formatters/helpers/duration',
    'primutils/formatters/helpers/unformatIgnore',
    'primutils/formatters/baseFormatter'
], function (_, logger, scientific, durationUtils, unformatIgnore, baseFormatter) {
    'use strict';

    var defaults = {
        // PGBU application settings
        /**
         * The units available to the duration formatter.
         *
         * Units are broken down by the strings the user configures in application settings (in Prime):
         * * n - Minutes
         * * h - Hours
         * * d - Days
         * * w - Weeks
         * * m - Months
         * * y - Years
         *
         * *Note* - in the above, 'h' is the `units` config only if the user has configured 'h' as the hourChar in Prime.
         * If they've configured, on the other hand, 'hh', then the `units` option would be keyed by 'hh' instead, and the `showIn` option would have to be 'hh' as well, not 'h'.
         *
         * Each unit defines its `scale` relative to hours and a `sub` which is the child unit
         *
         *     units : {
         *         h : {
         *             scale : 1, // 1 hour == 1 hour
         *             sub : 'n'  // Hours break down into minutes
         *         },
         *         d : {
         *             scale : 8, // 1 day == 8 hours (workday)
         *             sub : 'h'  // Days break down into minutes
         *         }
         *     }
         *
         * @property units
         * @type Object
         * @optional
         * @default 8 hours/day, 40 hours/week, 172 hours/month, 2000 hours/year
         */
        units : {
            n : { // minutes
                internalType : 'minutes',
                scale : 1 / 60,
                sub : null
            },
            h : { // hours
                internalType : 'hours',
                scale : 1,
                sub : 'n'
            },
            d : { // days
                internalType : 'days',
                scale : 8,
                sub : 'h'
            },
            w : { // weeks
                internalType : 'weeks',
                scale : 40,
                sub : 'd'
            },
            m : { // months
                internalType : 'months',
                scale : 172,
                sub : 'w'
            },
            y : { // years
                internalType : 'years',
                scale : 2000,
                sub : 'm'
            }
        },

        /**
         * Allows overriding the `units...scale` option without modifying other `units` options (i.e. the abbreviations and subs)
         *
         * For example, if you're using `userFormatters` in Prime (which uses formatterFactory in primutils), you might want to specialize the formatter like so:
         *
         *     var calendar = //...;
         *     var calendarDurFormatter = userFormatters.durationFormatter.withOptions({
         *         hoursPer: {
         *             hour: 1, // key roughly corresponds to `units['n'].internalType`
         *             day: calendar.hoursPerDay, // key roughly corresponds to `units['d'].internalType`
         *             week: calendar.hoursPerWeek, // key roughly corresponds to `units['w'].internalType`
         *             month: calendar.hoursPerMonth // key roughly corresponds to `units['m'].internalType`
         *
         *             // all of the above where, in `units[char]`, `char` are the abbreviations a user would use
         *         }
         *     });
         *
         * If the hoursPer option is present, the default scale provided by the `units` option may be used if `hoursPer` does not specify an override.
         * For example, if this duration formatter know of years, with a scale of 20, and the above withOptions call was used, without overriding the years hoursPer setting, 20 would still be used.
         * It is suggested that, instead of using this option directly, the static helper method `withSpecializedUnits` method is used, which will treat internalTypes/hoursPer mismatches as an error.
         *
         * Aliased as the `scales` option.
         *
         * *Note* to Prime developers: see the userFormatters documentation to see the internalTypes that must be used.
         *
         * @property hoursPer
         * @type Object
         * @optional
         * @default null
         */
        hoursPer : null,

        /**
         * @property flexibleScale
         * @type Boolean
         * @optional
         * @default false
         */
        flexibleScale : false,
        // PGBU user pref
        /**
         * Which unit is the base unit to display values in
         * This option must match one of the keys of this formatter's `unit` config option.
         *
         * @property showIn
         * @type String
         * @optional
         * @default 'h' (Hours)
         */
        showIn : 'h',              // any of the above units
                                  //  PGBU h/d/w/m/y only

        /**
         * Should sub units be shown
         *
         * @property showSubUnits
         * @type Boolean
         * @optional
         * @default false
         */
        showSubUnits : false,
        /**
         * Should units be shown as abbreviations
         *
         * @property showAbbr
         * @type Boolean
         * @optional
         * @default true
         */
        showAbbr : true,
        /**
         * The number of decimal places to show
         *
         * @property decimalsToShow
         * @type Number (0-2)
         * @optional
         * @default 1
         */
        decimalsToShow : 1,

        // documented in base numericOptions
        thousandsGroupingStyle : 'by_threes',

        // documented in base numericOptions
        thousandsSeparator : ',',

        // documented in base numericOptions
        decimalSeparator : '.',

        // documented in base numericOptions
        workableDecimalPlaces : 6,
        /**
         * Should all non-zero sub units be shown
         *
         * @property showAllNonzeroSubUnits
         * @type Boolean
         * @optional
         * @default false
         */
        showAllNonzeroSubUnits : false,

        /**
         * Whether or not to allow negative values to be entered by the user with this formatter.
         *
         * @property allowNegative
         * @type Boolean
         * @optional
         * @default false
         */
        allowNegative : false

        /**
         * A pattern (RegExp or String), or list of patterns, that will be ignored in any string
         *  the user types, for unformat() or validate() calls.
         *
         * @property unformatIgnore
         * @type String|RegExp|Array
         * @optional
         */
        //unformatIgnore = [],

        //serializeWrapper,
        //unserializeWrapper,
        //formatWrapper,
        //unformatWrapper,
        //validateWrapper
    };

    // TODO add validation options on create; showSubUnits: false + showAllNonzeroSubUnits: true
    //  makes no sense.  Probably other such things we could do as well, like check
    //  showIn/flexibleScale validity

    var stringy = durationUtils.stringy;
    var clipResult = durationUtils.clipResult;
    var getTrueScale = durationUtils.getTrueScale;
    var fixDecimal = durationUtils.fixDecimal;

    function format(options, val, formatOptions) {
        if (val == null) {
            return '';
        }

        if (!_.isNumber(val)) {
            logger.fatal('format: workableForm for durationFormatter is a Number');
            throw new Error('Must be a number');
        }

        if (_.isNaN(val)) {
            logger.error('numberFormatter should not be passed NaN');
            return '';
        }
        if (val === Number.POSITIVE_INFINITY || val === Number.NEGATIVE_INFINITY) {
            logger.error('numberFormatter should not be passed infinity');
            return '';
        }

        var showIn = options.showIn;
        if (options.flexibleScale) {
            showIn = null;
            var descendingUnitPairs = _.sortBy(_.pairs(options.units), function (unitSetting) {
                return -getTrueScale(options, unitSetting[1]);
            });
            var bestFit = _.find(descendingUnitPairs, function (setting) {
                return Math.floor(val / getTrueScale(options, setting[1])) > 0;
            });
            if (bestFit != null) {
                showIn = bestFit[0];
            } else {
                showIn = _.last(descendingUnitPairs)[0];
            }
        }

        var units = options.units;
        var isScientific = scientific.isScientific(val);
        if (isScientific && Math.abs(val) < 1) {
            // if it's this small, let's face it, not worth bothering about
            return format(options, 0, formatOptions);
        }

        var showInUnitOptions = units[showIn];
        var majorScale = getTrueScale(options, showInUnitOptions);
        var decimalsToShow = options.decimalsToShow;
        if (formatOptions && formatOptions.truncate === false) {
            decimalsToShow = Math.max(decimalsToShow, scientific.countDecimalsInNumber(val));
        }

        var scaledValue = val / majorScale;
        if (!options.showAbbr) {
            return stringy(options, fixDecimal(scaledValue, decimalsToShow));
        } else if (!options.showSubUnits) {
            return stringy(options, fixDecimal(scaledValue, decimalsToShow)) + showIn;
        } else if (options.showAllNonzeroSubUnits) {
            var portionVal = Math.floor(scaledValue);
            var portionUnit = showIn;
            var portionScale = majorScale;
            var nextPortionVal = val - (portionVal * portionScale);
            var ret = '';

            if (units[portionUnit].sub) {
                ret = stringy(options, portionVal) + portionUnit;
            } else {
                ret = stringy(options, fixDecimal(scaledValue, decimalsToShow)) + portionUnit;
            }
            while (units[portionUnit].sub && nextPortionVal > 0) {
                portionUnit = units[portionUnit].sub;

                var nextPortionOptions = units[portionUnit];
                portionScale = getTrueScale(options, nextPortionOptions);

                portionVal = Math.floor(nextPortionVal / portionScale);
                nextPortionVal -= (portionVal * portionScale);

                if (portionVal > 0) {
                    if (units[portionUnit].sub) {
                        ret += ' ' + stringy(options, portionVal) + portionUnit;
                    } else {
                        ret += ' ' + stringy(options, fixDecimal(portionVal, decimalsToShow)) + portionUnit;
                    }
                }
            }
            return ret;
        } else {
            var mathFn = (scaledValue >= 0) ? Math.floor : Math.ceil;
            var majorVal = mathFn(scaledValue);
            var result = stringy(options, mathFn(majorVal)) + showIn;

            if (!units[showIn].sub) {
                return stringy(options, fixDecimal(scaledValue, decimalsToShow)) + showIn;
            }

            if (majorVal !== scaledValue) {
                var subKey = units[showIn].sub;
                var minorScale = getTrueScale(options, units[subKey]);
                var subVal = Math.abs(((scaledValue - majorVal) * majorScale) / minorScale);
                result += ' ' + stringy(options, fixDecimal(subVal, decimalsToShow)) + subKey;
            }

            return result;
        }
    }

    function unformat(options, valStr) {
        if (valStr == null || valStr === '') {
            return null;
        }

        if (!_.isString(valStr)) {
            logger.fatal('unformat: typed form for all formatters is a String');
            throw new Error('Must be a string');
        }

        var thousands = options.thousandsSeparator === '.' ? '\\.' : options.thousandsSeparator;
        var deci = options.decimalSeparator === '.' ? '\\.' : options.decimalSeparator;
        valStr = _.unescape(valStr);

        if (thousands !== ' ') {
            valStr = valStr.trim()
                           .replace(new RegExp(thousands, 'g'), '')
                           .replace(new RegExp(deci, 'g'), '.');
        } else {
            valStr = valStr.trim()
                           .replace(new RegExp(deci, 'g'), '.');
        }

        valStr = unformatIgnore(options, valStr);

        if (valStr === '') {
            return null;
        }

        if (valStr.length >= durationUtils.numbers.maxInputLength) {
            return NaN;
        }

        var reNextChunk = durationUtils.reNextChunk;
        var foundChunk = NaN;
        var match = false;
        var baseValue = 0;
        var previousKeys = [];
        /* eslint-disable no-cond-assign */
        while ((match = reNextChunk.exec(valStr)) != null) {
            foundChunk = true;
            var partVal = Number(match[1]);
            if (thousands === ' ') {
                partVal = Number((match[1] || '').replace(' ', ''));
            }

            var partKey = (match[2] || '').trim();

            //use the default type if none is set
            if (partKey.length < 1) {
                partKey = options.showIn;
            }

            if (previousKeys.indexOf(partKey) !== -1) {
                return NaN;
            }
            previousKeys.push(partKey);

            var unitOptions = options.units[partKey];
            if (!_.isNaN(partVal) && unitOptions) {
                var scale = getTrueScale(options, unitOptions);
                baseValue += partVal * (scale);
            } else {
                return NaN;
            }
        }
        /* eslint-enable no-cond-assign */

        return foundChunk && clipResult(baseValue, options.workableDecimalPlaces);
    }

    function validate(options, valStr) {
        if (valStr == null || valStr === '') {
            return true;
        }

        if (!options.allowNegative && valStr.indexOf('-') !== -1) {
            return false;
        }

        if (valStr.length >= durationUtils.maxInputLength) {
            return false;
        }

        var wrappedUnformat = options.unformatWrapper ? _.wrap(unformat, options.unformatWrapper) : unformat;
        var unformatted = wrappedUnformat(options, valStr);

        if (unformatted == null) {
            return true;
        }
        if (_.isNaN(unformatted)) {
            return false;
        }

        return true;
    }

    function checkOpts(options) {
                var showIn = options.showIn;
        var units = options.units;
        var flexibleScale = options.flexibleScale;
        logger.assert(flexibleScale || units[showIn], 'No unit specified for the major scale ' + showIn + ', did you forget to specify the showIn or flexileScale option?');
        if (flexibleScale && !units[showIn]) {
            logger.warn('Using a flexibleScale, but the `showIn` option wasnt available in the given `units`.  Unformatting user-entered values without any units entered will not give any result');
        }
            }

    var formatter = baseFormatter(defaults, durationUtils.serialize, durationUtils.unserialize, format, unformat, validate, checkOpts);

    /**
     * A static helper method supporting the `hoursPer` option.
     *
     * In contrast to the option, this method is very strict - if the internalTypes of the formatter being specialized differ from the new hoursPer settings being provided, this method will throw an error instead of silently behaving incorrectly.
     *
     * @method withSpecializedUnits
     * @param formatter {DurationFormatter} an existing formatter to sub-specialize with `withOptions`
     * @param hoursPer {Object} the new `hoursPer` object to apply
     * @static
     */
    formatter.withSpecializedUnits = function (formatter, newUnits) {
        var oldOptions = formatter.getOptions();
                var oldInternalTypes = _.pluck(_.values(oldOptions.units), 'internalType');
        var newHoursPer = _.keys(newUnits);
        if (!_.isEqual(oldInternalTypes, newHoursPer)) {
            var err = 'Expected existing internalTypes to all have overrides specified by the given hoursPer overrides, but they differed.  ' + oldInternalTypes.toString() + newHoursPer.toString();
            logger.fatal(err);
            throw new Error(err);
        }
        
        return formatter.withOptions({
            hoursPer : newUnits
        });
    };

    return formatter;
});


define('primutils/validators/dateValidator',['lodash', 'moment'], function (_, moment) {
    'use strict';

    return function (opts) {
        var defaults = {
                minValue : null,
                maxValue : null
            }, options = _.defaults(opts || {}, defaults);

        if (options.message == null) {
            var message = 'Invalid date';
            if (options.maxValue != null && options.minValue != null) {
                message = 'LOCALIZE Date must be between ' + options.minValue + ' and ' + options.maxValue;
            } else if (options.maxValue != null) {
                message = 'LOCALIZE Date may be before ' + options.maxValue;
            } else if (options.minValue != null) {
                message = 'LOCALIZE Date must be after ' + options.minValue;
            }
            options.message = message;
        }

        return {
            options : options,
            validate : function (date) {
                if (date == null) {
                    return true;
                }
                if (moment.isMoment(date)) {
                    if (options.minValue != null && date.isBefore(options.minValue)) {
                        return false;
                    }
                    if (options.maxValue != null && date.isAfter(options.maxValue)) {
                        return false;
                    }
                    return true;
                } else {
                    return false;
                }
            },

            message : options.message
        };
    };
});

define('primutils/validators/numberValidator',['lodash'], function (_) {
    'use strict';


    return function (opts) {
        var defaults = {
                minValue : null,
                maxValue : null
            }, options = _.defaults(opts || {}, defaults);

        if (options.message == null) {
            var message = 'LOCALIZE Invalid number';
            if (options.maxValue != null && options.minValue != null) {
                message = 'LOCALIZE Number must be between ' + options.minValue + ' and ' + options.maxValue;
            } else if (options.maxValue != null) {
                message = 'LOCALIZE Number may be at most ' + options.maxValue;
            } else if (options.minValue != null) {
                message = 'LOCALIZE Number must be at least ' + options.minValue;
            }
            options.message = message;
        }

        return {
            options : options,
            validate : function (number) {
                if (_.isObject(number)) {
                    if (_.has(number, 'valueOf')) {
                        number = number.valueOf();
                    } else {
                        return true;
                    }
                }

                if (number == null) {
                    return true;
                }
                if (isNaN(number)) {
                    return false;
                }
                if (options.minValue != null && number < options.minValue) {
                    return false;
                }
                if (options.maxValue != null && number > options.maxValue) {
                    return false;
                }
                return true;
            },

            message : options.message
        };
    };
});

define('primutils/validators/stringValidator',['lodash'], function (_) {
    'use strict';


    return function (opts) {
        var defaults = {
                minValue : null,
                maxValue : null
            }, options = _.defaults(opts || {}, defaults);

        if (options.message == null) {
            var message = 'Invalid string';
            if (options.maxValue != null && options.minValue != null) {
                message = 'LOCALIZE String must be between ' + options.minValue + ' and ' + options.maxValue + ' characters.';
            } else if (options.maxValue != null) {
                message = 'LOCALIZE String may be no greater than ' + options.maxValue + ' characters.';
            } else if (options.minValue != null) {
                message = 'LOCALIZE String must be at least ' + options.minValue + ' characters.';
            }
            options.message = message;
        }

        return {
            options : options,
            validate : function (string) {
                if (string == null) {
                    return true;
                }
                var strLen = _.unescape(string).length;
                if (!_.isString(string)) {
                    return false;
                }
                if (options.minValue != null && strLen < options.minValue) {
                    return false;
                }
                if (options.maxValue != null && strLen > options.maxValue) {
                    return false;
                }
                return true;
            },

            message : options.message

        };
    };
});

define('primutils/validators/requiredValidator',['lodash'], function (_) {
    'use strict';

    return function (options) {
        var defaultOptions = {
            message : 'Field is required localizeme in options'
        };
        options = _.defaults(options || {}, defaultOptions);
        return {
            options : options,
            validate : function (value) {
                if (value == null) {
                    return false;
                }
                if (_.isString(value)) {
                    return value.trim().length > 0;
                } else if (_.isArray(value)) {
                    return value.length > 0;
                } else if (_.isNaN(value)) {
                    return false;
                }
                return true;
            },

            message : options.message
        };
    };
});

define('primutils/validators/uniquenessValidator',[
    'lodash',
    'primutils/utils/strings'
], function (_, strings) {
    'use strict';

    /**
     * @class validators.uniquenessValidator
     * @module validation
     */
    return function (opts) {
        var defaults = {
            caseSensitive : false,
            getValueFromRecord : function (record, key) {
                return record != null ? record[key] : null;
            },
            newValue : _.identity,

            // Overrides caseSensitive, getValueFromRecord, for fields by key
            keyOptions : {
            }
        };
        var globalOptions = _.defaults(opts || {}, defaults);

        return {
            /**
             * Check the array for records with duplicate values
             *
             * @method checkRecordUniqueness
             * @param {Object} newValue The new value to check for uniqueness
             * @param {Object} record The record to validate
             * @param {Object} key The field to load
             * @param {Array} arrayOfRecords The array of records to check
             * @return Array of records with the same non null value for a field
             */
            checkRecordUniqueness : function (newValue, record, key, arrayOfRecords) {
                var newValueComparable = newValue;

                var optNewValue = _.get(globalOptions.keyOptions, key + '.newValue', globalOptions.newValue);
                if (optNewValue) {
                    newValueComparable = optNewValue(newValue);
                }

                var optCaseSensitive = _.get(globalOptions.keyOptions, key + '.caseSensitive', globalOptions.caseSensitive);
                var optGetValue = _.get(globalOptions.keyOptions, key + '.getValueFromRecord', globalOptions.getValueFromRecord);

                if (!optCaseSensitive && strings.isString(newValueComparable)) {
                    newValueComparable = newValueComparable.toLocaleLowerCase();
                }

                var matchingRecords = [];
                for (var i = 0; i < arrayOfRecords.length; i++) {
                    var recordToCheck = arrayOfRecords[i];
                    if (recordToCheck === record) {
                        continue;
                    }

                    var valueToCheck = optGetValue(recordToCheck, key);
                    if (valueToCheck != null && valueToCheck !== '') {
                        if (!optCaseSensitive && strings.isString(valueToCheck)) {
                            valueToCheck = valueToCheck.toLocaleLowerCase();
                        }

                        if (valueToCheck === newValueComparable) {
                            matchingRecords.push(recordToCheck);
                        }
                    }
                }
                return matchingRecords;
            },

            checkStoreUniqueness : function (key, arrayOfRecords) {
                var valueToRecordMapping = {};
                var matchingRecords = [];
                var optCaseSensitive = _.get(globalOptions.keyOptions, key + '.caseSensitive', globalOptions.caseSensitive);
                var optGetValue = _.get(globalOptions.keyOptions, key + '.getValueFromRecord', globalOptions.getValueFromRecord);
                for (var i = 0, len = arrayOfRecords.length; i < len; ++i) {
                    var recordToCheck = arrayOfRecords[i];
                    var valueToCheck = optGetValue(recordToCheck, key);
                    if (valueToCheck != null && valueToCheck !== '') {
                        if (!optCaseSensitive && strings.isString(valueToCheck)) {
                            valueToCheck = valueToCheck.toLocaleLowerCase();
                        }

                        if (valueToRecordMapping[valueToCheck] == null) {
                            valueToRecordMapping[valueToCheck] = [];
                        }
                        valueToRecordMapping[valueToCheck].push(recordToCheck);
                    }
                }

                var matches = _.values(valueToRecordMapping);
                for (var j = 0, jlen = matches.length; j < jlen; ++j) {
                    if (matches[j].length > 1) {
                        matchingRecords.push(matches[j]);
                    }
                }

                return matchingRecords;
            }
        };
    };
});

define('primutils/validators/validatorFactory',[
    'lodash',
    'logger',
    'primutils/validators/dateValidator',
    'primutils/validators/numberValidator',
    'primutils/validators/stringValidator',
    'primutils/validators/requiredValidator',
    'primutils/validators/uniquenessValidator'
], function (_, logger, DateValidator, NumberValidator, StringValidator, RequiredValidator, UniquenessValidator) {
    'use strict';

            var __loggedTypeWarnings = {};
        
    var knownTypeMap = {
        string : {
            type : 'string',
            cons : StringValidator
        },
        number : {
            type : 'number',
            cons : NumberValidator
        },
        date : {
            cons : DateValidator
        },
        duration : {
            cons : NumberValidator
        },
        unit : {
            cons : NumberValidator
        }
    };
    knownTypeMap.text = knownTypeMap.string;
    knownTypeMap.multicurrency_cost = knownTypeMap.cost = knownTypeMap.long = knownTypeMap.double = knownTypeMap.int = knownTypeMap.integer = knownTypeMap.percent = knownTypeMap.number;
    knownTypeMap.hour = knownTypeMap.quantity = knownTypeMap.system_duration = knownTypeMap.duration;

        // We don't really know anything about them from a validation perspective...but don't want to uselessly error either
        // TODO should we include validations around these?
    var silentTypes = ['enum', 'key', 'boolean', 'timestamp6', 'start_date', 'finish_date', 'picker', 'localized_date', 'key', 'blob', 'year', 'url', 'join'];
    _.forEach(silentTypes, function (type) {
        knownTypeMap[type] = {};
    });

    var defaultOptions = {
            /**
             * The data type of the field being validated
             *
             * Supports: 'string', 'date', 'number'
             *
             * @property type {String}
             * @default 'string'
             */
        type : 'string',
            /**
             * The minimum value for the field being validated
             *
             * For strings, this is the minimum number of characters required
             *
             * @property minValue {*|Number}
             * @optional
             */
        minValue : null,
            /**
             * The maximum value for the field being validated
             *
             * For strings, this is the maximum number of characters allowed
             *
             * @property maxValue {*|Number}
             * @optional
             */
        maxValue : null,

            /**
             * Whether to perform a check that the field is not null
             *
             * @property required {Boolean}
             * @default false
             */
        required : false,
            /**
             * If false, the validator will perform the required validation.
             * If true, the developer will do manual validation, but the field will still show as required (asterisk and placeholder info in grid)
             *
             * @property requiredManualValidation {Boolean}
             * @default false
             */
        requiredManualValidation : false,
            /**
             * The message returned when the field is required but is null
             *
             * @property requiredMessage {String}
             */
        requiredMessage : 'options.requiredMessage Field is required',

            /**
             * Used to turn the value into an actual object.
             * If no formatter, the assumption is the value is already correct.
             *
             * The formatter should provide two functions:
             *   * validate: Takes a string to verify that it can be unformatted
             *   * unformat: Turn a string user entered value into a workable form
             *
             * @property formatter {Object}
             * @optional
             */
        formatter : null,
            /**
             * The message returned when the format is not valid
             *
             * @property formatMessage {String}
             */
        formatMessage : 'options.formatMessage Invalid Format',

            /**
             * A default validation message.
             * Messages that will be passed into the type validators, these should be localized.
             *
             * @property message {String}
             * @optional
             */
        message : null,

            /**
             * Whether to perform a check for uniqueness
             *
             * @property unique {Boolean}
             * @default false
             */
        unique : false,
            /**
             * If false, the validator will perform the unique validation.
             * If true, the developer will do manual validation, but the field will be marked as unique
             *
             * @property uniqueManualValidation {Boolean}
             * @default false
             */
        uniqueManualValidation : false,

            /**
             * Allows overriding caseSensitive, getValueFromRecord, newValue checks globally and per key.
             *
             * @property uniqueOptions
             * @default `{}`
             * @optional
             * @type {Object}
             */
        uniqueOptions : {},

            /**
             * Used for uniqueness tests.
             * record[dataIndex] is how it will pull the data off each record in the store
             *
             * @property dataIndex {String}
             * @optional
             */
        dataIndex : null,
            /**
             * Used for uniqueness.
             * This is if you have an underlying store you can give it to check; if not pass in getUniqueStore.
             *
             * @property uniqueStore {Object}
             * @optional
             */
        uniqueStore : null,
            /**
             * Used for uniqueness.
             * Callback to get the store for uniqueness checking.
             *
             * @property getUniqueStore {Function}
             * @optional
             */
        getUniqueStore : _.noop,
            /**
             * The message returned when the value is not unique
             *
             * @property uniqueMessage {String}
             */
        uniqueMessage : 'options.uniqueMessage Field value already exists. Enter a unique value.',
            /**
             * Callback to process unique response
             *
             * @propery processUniqueResponse {Function}
             * @optional
             */
        processUniqueResponse : _.noop,
            /**
             * Check entire store when checking uniqueness
             *
             * @property checkEntireStore {Boolean}
             * @default true
             */
        checkEntireStore : true
    };

        /**
         * This class generates a validator that will run through all possible permutations of standard required, formatting, type bounds checking and uniqueness.
         *
         * 1. Check required if option is set to true.
         * 2. Check the format is valid via formatter.validate(value)
         * 3. Check any bounds error based on minValue, and maxValues set
         * 4. Run uniqueness tests if unique is true.
         * 5. Call processUniqueResponse
         * 6. Return validationResponse with the errorType, message and valid
         *
         * If it fails at any step it will stop processing the data.
         *
         * Sample use:
         *
         *      var validator = validatorFactory({
         *          type : 'number',
         *          minValue : 100,
         *          message : 'Total Cost must be at least 100',
         *      });
         *      validator.validate(200); // would return {valid : true}
         *      validator.validate(50);  // would return {valid : false, errorType : 'type', message : 'Total Cost must be at least 100'}
         *
         * @class validators.validatorFactory
         * @module validation
         */
    var factory = function (options) {
                    logger.assert(_.isObject(options), 'validatorFactory | options passed in must be an object');
                    _.defaults(options, defaultOptions);
        var required, type, unique;
        if (options.required && !options.requiredManualValidation) {
            required = new RequiredValidator({ message : options.requiredMessage });
        }

        var typeKey = options.type.toLowerCase();
        var typeInfo = knownTypeMap[typeKey];

        if (typeInfo) {
            if (typeInfo.cons) {
                var Cons = typeInfo.cons,
                    typeValidatorOptions = (options.requiredManualValidation && Cons === StringValidator) ? _.defaults({ minValue : 0 }, options) : options;

                type = new Cons(typeValidatorOptions);

                if (typeInfo.type) {
                    options.type = typeInfo.type; // Reset to the base known type
                }
            }

                            else if (!__loggedTypeWarnings[typeKey]) {
                logger.debug('No known type validator for type : ' + options.type + '.  No type-specific validation rules will be run, though generic rules will still run.');
                __loggedTypeWarnings[typeKey] = true;
            }
                        }
                    else {
            logger.warn('Unrecognized type for validatorFactory: ' + options.type + '.  Known types are: ' + _.keys(knownTypeMap));
        }
            
        if (options.unique && !options.uniqueManualValidation) {
            unique = new UniquenessValidator(options.uniqueOptions);
        }

        return {
            options : options,
            validators : {
                type : type,
                unique : unique,
                required : required
            },
                /**
                 * (This is a function available on the returned validator)
                 *
                 * Validates the value
                 *
                 * Returns a validation result with the following format:
                 *
                 *     {
                 *         valid: {Boolean},
                 *         errorType: {String},
                 *         message: {String},
                 *         matchingRecords: {Array}
                 *     }
                 *
                 * @method validate
                 * @param rawValue The value to validate
                 * @param [record] The whole record
                 * @returns {Object}
                 */
            validate : function (rawValue, record) {
                var response = {
                    valid : true,
                    errorType : null,
                    message : null,
                    matchingRecords : null
                };
                var value = rawValue;

                if (required) {
                    if (!required.validate(value)) {
                        response.errorType = 'required';
                        response.message = required.message;
                        response.valid = false;
                        return response;
                    }
                }

                if (options.formatter != null) {
                    if (options.formatter.validate(value)) {
                        value = options.formatter.unformat(value);
                    } else {
                        response.errorType = 'format';
                        response.message = options.formatMessage;
                        response.valid = false;
                        return response;
                    }
                }

                if (type && !type.validate(value)) {
                    response.errorType = 'type';
                    response.message = type.message;
                    response.valid = false;
                }

                if (unique) {
                    var store = options.uniqueStore;
                    if (store == null) {
                        store = options.getUniqueStore();
                    }
                    if (store != null) {
                        var matchingRecords = [];
                        if (options.checkEntireStore) {
                            matchingRecords = unique.checkStoreUniqueness(options.dataIndex, store);
                        } else {
                            matchingRecords = unique.checkRecordUniqueness(value, record, options.dataIndex, store);
                        }
                        if (matchingRecords.length > 0) {
                            response.errorType = 'unique';
                            response.valid = false;
                            response.matchingRecords = matchingRecords;
                            response.message = options.uniqueMessage;
                        }

                        options.processUniqueResponse(matchingRecords);
                    }
                }

                return response;
            }
        };
    };

    return factory;
});


define('primutils/utils/defaultStrings',[], function () {
    'use strict';

    /**
     * An empty set of strings.
     * For the set of strings that primutils expects in production, see <a href="http://10.140.183.183:8080/job/primutils/javadoc/files/assets_primutils_utils_defaultStrings.js.html">the source of defaultStrings.js</a>, which is only used for primutils development.
     *
     * It is expected that someone, at an application level, calls something like the following:
     *
     *     require(['primutils'], function (primutils) {
     *         primutils.lang({
     *             // lots of localized strings for the current locale
     *         });
     *     });
     *
     * @class i18n.emptyStrings
     * @module i18n
     * @private
     */
    return {};
});

define('primutils/utils/language',['primutils/i18n/Localizer', 'primutils/utils/defaultStrings'], function (Localizer, defaultStrings) {
    'use strict';

    return new Localizer(defaultStrings);
});


/**
 * A shared map of key names to their keycode values.
 * See the <a href="http://10.140.183.183:8080/job/primutils/javadoc/files/assets_primutils_utils_keyCode.js.html">source</a> for available keycodes.
 *
 * @class utils.keyCode
 * @module utils
 */
define('primutils/utils/keyCode',{

    UP : 38,
    DOWN : 40,
    LEFT : 37,
    RIGHT : 39,

    SPACE : 32,
    TAB : 9,
    BACKSPACE : 8,
    ENTER : 13,
    ESC : 27,
    ESCAPE : 27,

    SHIFT : 16,
    CTRL : 17,
    ALT : 18,

    PAUSE : 19,
    CAPS : 20,
    PAGEUP : 33,
    PAGEDOWN : 34,
    END : 35,
    HOME : 36,
    INSERT : 45,
    DELETE : 46,

    F1 : 112,
    F2 : 113,
    F3 : 114,
    F4 : 115,
    F5 : 116,
    F6 : 117,
    F7 : 118,
    F8 : 119,
    F9 : 120,
    F10 : 121,
    F11 : 122,
    F12 : 123,

    NUMLOCK : 144,
    SCROLLLOCK : 145,

    LEFTSUPER : 91,
    RIGHTSUPER : 92,

    // All below since v2.13.1
    SELECT_KEY : 93,

    NUMROW_0 : 48,
    NUMROW_1 : 49,
    NUMROW_2 : 50,
    NUMROW_3 : 51,
    NUMROW_4 : 52,
    NUMROW_5 : 53,
    NUMROW_6 : 54,
    NUMROW_7 : 55,
    NUMROW_8 : 56,
    NUMROW_9 : 57,

    NUMPAD_0 : 96,
    NUMPAD_1 : 97,
    NUMPAD_2 : 98,
    NUMPAD_3 : 99,
    NUMPAD_4 : 100,
    NUMPAD_5 : 101,
    NUMPAD_6 : 102,
    NUMPAD_7 : 103,
    NUMPAD_8 : 104,
    NUMPAD_9 : 105,

    NUMPAD_DECIMAL : 110,
    NUMPAD_SUBTRACT : 109,
    NUMPAD_ADD : 107,
    NUMPAD_DIVIDE : 111,
    NUMPAD_MULTIPLY : 106,
    NUMPAD_ENTER : 108,

    DASH : 189,
    FF_DASH : 173,

    EQUALS : 187,
    FF_EQUALS : 61,

    PERIOD : 190,
    COMMA : 188,

    // all intended for use with `&& event.ctrlKey`
    LETTER_A : 65,
    LETTER_C : 67,
    LETTER_S : 83,
    LETTER_V : 86,
    LETTER_X : 88,
    LETTER_Y : 89,
    LETTER_Z : 90

});


define('primutils/a11y/collapsibleDropdown',['jquery', 'primutils/utils/keyCode'], function ($, keyCode) {
    'use strict';

    /**
     * What this will do is allow for left/right to open/close submenus and up/down will skip anchors in currently
     * hidden submenus.
     *
     * Warning: will only support one level of submenus
     * Warning: liberally uses "> a" selector to find direct focusable children
     *
     * In general, we try not to duplicate some of bootstrap's work here (such as up/down opening up the hidden
     * dropdown in the first place), and instead only replace behavior that could be problematic and stopPropagation().
     *
     * @class a11y.collapsibleDropdown
     * @module a11y
     * @param $menu {jQuery Object} The menu to wrap - e.g. `$('.dropdown-menu').parent()` - (though for performance and
     *   simplicity you should grab the parent directly as opposed to using `.parent()`)
     * @constructor
     */
    function CustomDropdown($menu) {
        this.$menu = $menu;
        this._start();
    }

    CustomDropdown.prototype = {
        constructor : CustomDropdown,

        _start : function () {
            this.$menu.on('keydown', this._handleMove.bind(this));
        },

        _move : function (moveDir, $node, ev) {
            //Bootstrap puts open on the parent of the menu
            if (this.$menu.parent().hasClass('open') && !$node.hasClass('dropdown-toggle') && $node.is('a, .dropdown-submenu-item')) {
                // .children("a, .dropdown-submenu-item") lets us be a bit more careful than bootstrap - don't want subitems of the dropdown-submenu that are
                //  invisible matching this
                var $nodes = this.$menu.find('li:visible').children('a, .dropdown-submenu-item');
                var index = $nodes.index($node);
                var newIndex = index + moveDir;
                if (newIndex >= 0 && newIndex < $nodes.length) {
                    ev.stopPropagation();
                    var $newNode = $nodes.eq(newIndex).focus();

                    // If up/down moved us out of the submenu, hide it
                    if ($node.parents('.dropdown-submenu').length &&
                        (!$newNode.parents('.dropdown-submenu').length || $newNode.parent().is('.dropdown-submenu'))) {
                        $node.parents('.dropdown-submenu').removeClass('open');
                    }
                }
            }
        },

        _maybeToggleSubmenu : function ($node, ev) {
            var $parent = $node.parent();
            if (!$parent.is('.dropdown-submenu') && $node.parents('.dropdown-submenu').length) {
                ev.stopPropagation();

                var $submenu = $node.parents('.dropdown-submenu');
                $submenu.removeClass('open');
                $submenu.children('a, .dropdown-submenu-item').first().focus();
            } else if ($parent.is('.dropdown-submenu')) {
                ev.stopPropagation();

                var isOpen = $parent.hasClass('open');
                if (!isOpen) {
                    $parent.addClass('open');
                    $parent.find('> .dropdown-menu > li').children('a, .dropdown-submenu-item').first().focus();
                } else {
                    $parent.removeClass('open');
                }
            }
        },

        _handleMove : function (ev) {
            var $node = $(ev.target);
            switch (ev.keyCode) {
            case keyCode.UP:
                ev.preventDefault();
                this._move(-1, $node, ev);
                break;
            case keyCode.DOWN:
                ev.preventDefault();
                this._move(1, $node, ev);
                break;
            case keyCode.LEFT:
            case keyCode.RIGHT:
                this._maybeToggleSubmenu($node, ev);
                break;
            default:
                return;
            }
        }
    };

    return function ($menu) {
        return new CustomDropdown($menu);
    };
});


define('primutils/a11y/accessibilityUtils',['jquery', 'lodash',
    'primutils/a11y/collapsibleDropdown',
    'primutils/utils/keyCode'
], function ($, _, collapsibleDropdown, keyCode) {
    'use strict';

    function getFocusableThings($collection, action) {
        return $collection[action]('a[href], button:not([disabled]), input:not([disabled]), textarea:not([disabled]), select:not([disabled]), [tabindex], iframe')
            .filter(':not([tabindex=-1]):visible') // catch visible tab-able things and OUI hidden inputs
            .filter(function () {
                return window.getComputedStyle(this).visibility !== 'hidden';
            });
    }

    /**
     * Defines a shared set of functionality around accessibility.
     *
     * Some functions are also available on jQuery. Generally those that take a jQuery object as their first
     * parameter are also bound to jQuery. You can access functions bound to jQuery with the prefix 'prim_'.
     *
     *     require(['jquery', 'primutils/a11y/accessibilityUtils'], function ($) {
     *         'use strict';
     *         $('li > a.someItem').prim_toggleListItemAnchor(true);
     *     });
     *
     * This is exactly equivalent to
     *
     *     require(['jquery', 'primutils/a11y/accessibilityUtils'], function ($, accessibilityUtils) {
     *         'use strict';
     *         var $item = $('li > a.someItem');
     *         accessibilityUtils.toggleListItemAnchor($item, true);
     *     });
     *
     * @class a11y.accessibilityUtils
     * @module a11y
     */

    var jqueryCompatibleUtils = {
        /**
         * Toggles the state of an `<a>` in a dropdown list item so that it gets the proper Bootstrap styling and
         * aria attributes.
         *
         * Warning: This does not prevent the user from triggering clicks on the element. This feature is pending.
         *
         * (Also available from jQuery)
         *
         * @method toggleListItemAnchor
         * @param $anchor {jQuery Object} The `<a>` to toggle
         * @param enable {Boolean} Whether the anchor should appear enabled
         */
        toggleListItemAnchor : function ($anchor, enable) {
            var $li = $anchor.parent('li');
            $li.toggleClass('disabled', !enable);
            $anchor.attr('aria-disabled', !enable).attr('tabindex', enable ? 0 : -1);
        },
        /**
         * Creates a collapsible dropdown from the given menu.
         * (See <a href="a11y.CustomDropdown.html">a11y.CustomDropdown</a> for more information.)
         *
         * (Also available from jQuery)
         *
         * @method collapsibleDropdown
         * @param $menu {jQuery Object} The menu to bind to
         */
        collapsibleDropdown : function ($menu) {
            return collapsibleDropdown($menu);
        },
        /**
         * Appends the given modal to the highest visible modal currently in the DOM, if any, to prevent `focus` competitions from causing an infinite loop and crashing the browser.
         *
         * Be aware that calling this will place your modal "inside" the DOM of the other modal.
         * If that parent modal is closed, your modal will be closed also.
         * This method should only be called if that is your expected behavior or you are an alert modal which locks out additional user activity until action is taken.
         *
         * @method modalAppend
         * @param $modal {jQuery Object} The modal to append
         * @param options {Object} if `options.allowClosingParent` is true, will allow 'hide' and 'hidden' events to propagate to the parent modal.
         *   By default, these two events WILL NOT propagate to the parent modal.
         */
        modalAppend : function ($modal, options) {
            options = options || {};

            // To avoid causing a infinite call stack by modals trying to compete for focus,
            //  we'll append to the topmost visible modal onscreen, if there currently is one.
            var appendPoint = document.body;
            var possibleModalAppend = $('.modal.in').last(); // purposely not doing .tray.in
            if (possibleModalAppend.length) {
                appendPoint = possibleModalAppend;

                if (!options.allowClosingParent) {
                    jqueryCompatibleUtils.stopModalHidePropagation($modal);
                }
            }
            $modal.appendTo(appendPoint);
        },
        /**
         * Prevents `'hide'` and `'hidden'` events from propagating from the given modal to any parent modals.
         *   Useful when using <a href="#method_modalAppend">modalAppend</a>, which will call this method by default.
         *
         * @method stopModalHidePropagation
         * @param $modal {jQuery Object} the modal to suppress events on
         */
        stopModalHidePropagation : function ($modal) {
            $modal.one('hide hidden', function (e) {
			alert("prim")
                e.stopPropagation();
            });
        },

        filterFocusable : function ($collection) {
            return getFocusableThings($collection, 'filter');
        },

        findFocusable : function ($parent) {
            return getFocusableThings($parent, 'find');
        }
    };

    function clickGlobalSave($content, ev) {
        var $save = $content.find('[accesskey=s]');
        if ($save.length && !$save.prop('disabled')) {
            $save.click();
            return true;
        }
    }

    var utils = {
        /**
         * Applies global keyboard shortcuts that apply across all Primavera apps, such as ctrl-s to save data (rather than page source/html).
         *
         * @method applyGlobalShortcuts
         * @param $content the main content div for finding e.g. save cancel bar and other divs
         * @static
         */
        applyGlobalShortcuts : function ($content) {
            var LETTER_S = keyCode.LETTER_S;
            $(document).on('keydown.global-shortcuts', function (ev) {
                if (ev.keyCode !== LETTER_S) {
                    return;
                }

                if ((ev.altKey && ev.shiftKey) || ev.metaKey || ev.ctrlKey) {
                    // 'Save as HTML' never useful, always preventDefault()
                    ev.preventDefault();

                    var $modal = $('.modal.in').last();
                    if ($modal.length) {
                        var $defaultAction = $modal.find('[data-default="enter"]');
                        if ($defaultAction.length && !$defaultAction.prop('disabled') && $defaultAction.is(':visible')) {
                            $defaultAction.click();
                        }
                    } else {
                        clickGlobalSave($content, ev);
                    }
                }
            });
        }
    };

    function makeJQueryFn(fn) {
        return function () {
            var $this = $(this);

            var args = _.toArray(arguments);
            args.unshift($this);

            var result = fn.apply(undefined, args);

            if (result === undefined) {
                return $this;
            } else {
                return result;
            }
        };
    }

    // Extend jQuery with jquery compatible fns
    $.fn.extend(_.reduce(jqueryCompatibleUtils, function (res, fn, name) {
        res['prim_' + name] = makeJQueryFn(fn);
        return res;
    }, {}));

    return _.assign({}, jqueryCompatibleUtils, utils);
});


define('primutils/templates/helpers/localizer',[
    'logger',
    'lodash',
    'primutils/i18n/Localizer'
], function (logger, _, Localizer) {
    'use strict';

    var localizer;

    /**
     * Gets a localized string from the registered localizer
     *
     * Available as: `\{{lz}}`.
     *
     * Localizer is a helper that does not have inner content.
     *
     * If only one argument is passed, it will be assumed that your string is not formatted.
     * If multiple arguments are passed, the will be inserted into the format placeholders.
     *
     * Example:
     *
     *     \{{lz 'global.hello_world'}}
     *     \{{lz 'global.hello_x' 'World'}}
     *
     * You can also use the localizer helper as a parameter to another helper.
     * Just use parentheses `()` instead of staches, for example:
     *
     *     \{{tab (lz 'global.my_tab_name') 'tab_id'}}
     *
     * To set the localizer used by this helper use the following method.
     * `myLocalizer` should be an instance of the `Localizer` class.
     *
     *     primutils.setLocalizerForLocalizerHelper(myLocalizer);
     *
     * See also: <a href="i18n.Localizer.html">i18n.Localizer</a>
     *
     * @class templates.helpers.lz
     * @module templates.helpers
     * @constructor
     * @param key {String} Localization key to read from the localizer
     * @param [...] {String} Values to insert into format placeholders
     * @param [options] {Object} Handlebars options
     * @returns {string} Localized string
     */
    var lz = function (key /*, options */) {
        logger.assert(!!localizer, 'Templates: Cannot use localizer helper until a localizer has been configured');
        logger.assert(arguments.length >= 1, 'Templates: Localizer helper requires a "key" argument');
        if (arguments.length <= 2) {
            return localizer.getString(key);
        } else {
            var localizerArgs = _.initial(_.toArray(arguments)); // Take everything but what would be the options arg
            return localizer.getFormattedString.apply(localizer, localizerArgs);
        }
    };

    /**
     * Sets the localizer used by the localizer helper
     * @method useLocalizer
     * @for templates.helpers.lz
     * @static
     * @param newLocalizer {Localizer} The localizer to use when retrieving keys
     */
    lz.useLocalizer = function (newLocalizer) {
        logger.assert(newLocalizer instanceof Localizer, 'Templates: Localizer must be an instance of the type Localizer');
        localizer = newLocalizer;
    };

    lz.handlebarsType = 'helper';
    lz.handlebarsName = 'lz';

    return lz;
});


define('primutils/templates/utils/utils',[
    'lodash',
    'jquery'
], function (_, $) {
    'use strict';

    var GLOBAL_ATTRIBUTES = ['accesskey', /*'class',*/ 'dir', 'dropzone', 'id', 'lang', 'spellcheck', 'style', 'tabindex', /*'title',*/ 'translate'],
        GLOBAL_BOOL_ATTRIBUTES = ['draggable', 'hidden'];

    /**
     * Provides utility functions that can be used inside helpers and partials
     *
     * @class templates.utils.utils
     * @module templates.utils
     */

    var utils = {
        /**
         * Filters the Handlebars hash to a list of attributes that can be applied to the generated element.
         * The user can specify additional attributes that are valid for their element in addition to the globally allowed attributes.
         *
         * Attributes are broken into two categories: string and boolean.
         * String attributes the pass the filter are written to the template exactly as passed.
         * Boolean attributes ignore the value and are always written as attribute="attribute" to ensure backwards compatibility.
         * This makes it easy for the template author to specify something like `{{myHelper required=true}}` and have it interpreted correctly.
         *
         * The default list of attributes is: accesskey, dir, draggable, dropzone, hidden, id, lang, spellcheck, style, tabindex, and translate.
         *
         * If you need to prevent users from specifying these attributes but still want to call `getElementAttributes`, delete them from the hash before passing it to this method.
         *
         * Some global attributes are missing from the default list, namely: class, contenteditable, and title.
         * Generally you do not want the user specifying these herself.
         * For 'class', consider using the <a href="#method_mergeClasses">mergeClasses</a> method passed to the <a href="templates.utils.renderer.html#method_renderElement">renderer.renderElement</a> method.
         * For the others, you can explicitly pass them in the first attributes parameter to the <a href="templates.utils.renderer.html#method_renderElement">renderer.renderElement</a> method or list them in the `passedAttrs` or `booleanPassedAttrs` parameters in this function.
         *
         * @method getElementAttributes
         * @param hash {Object} The Handlebars hash of additional properties passed to the helper in a template
         * @param [passedAttrs] {Array} List of attributes that should be kept when rendering an element
         * @param [booleanPassedAttrs] {Array} List of boolean attributes that should be kept when rendering an element
         * @returns {Object} A filtered key/value pair collection valid for converting into HTML attributes
         */
        getElementAttributes : function (hash, passedAttrs, booleanPassedAttrs) {
            return _.reduce(hash, function (memo, value, key) {
                var lowerKey = key.toLowerCase();
                if (_.includes(passedAttrs, lowerKey) || _.includes(GLOBAL_ATTRIBUTES, lowerKey)) {
                    memo[key] = value;
                } else if ((_.includes(booleanPassedAttrs, lowerKey) || _.includes(GLOBAL_BOOL_ATTRIBUTES, lowerKey)) && value !== false) {
                    // Do not match boolean attributes when they are explicitly set to false
                    memo[key] = key;
                } else if (key.match(/^data-/g)) {
                    memo[key] = value;
                } else if (key.match(/^aria-/g)) {
                    memo[key] = value;
                }
                return memo;
            }, {});
        },

        /**
         * Merges an array of class names with a string separated list of classes and converts it into a space separated list of classes.
         *
         * @method mergeClasses
         * @param defaults {Array} List of classes provided by the helper or partial author
         * @param [classes] {String} Space separated list of classes usually provided by the template author
         * @returns {string} Space separated list of classes
         */
        mergeClasses : function (defaults, classes) {
            return classes ? _.union(defaults, classes.split(' ')).join(' ') : defaults.join(' ');
        },

        /**
         * Converts a space separated list of classes into an array
         *
         * @method classList
         * @private
         * @param classes {String} Space separated list of classes
         * @returns {Array} Array of classes
         */
        classList : function (classes) {
            return (classes || '').split(' ');
        },

        /**
         * Checks to see if a particular class is provided in a space separated list of classes
         *
         * @method hasClass
         * @private
         * @param classes {String} Space separated list of classes
         * @param clz {String} Class name to find
         * @returns {Boolean} Whether the class exists inside the classes string
         */
        hasClass : function (classes, clz) {
            return _.includes(utils.classList(classes), clz);
        },

        /**
         * Gives the inner text of an element, trimmed, with newlines replaced as spaces.
         *
         * @method textifyElement
         * @private
         * @param text {String} element text to textify
         * @returns {String} more normalized string
         */
        textifyElement : function (elem) {
            return $('<div>' + elem + '</div>').text().trim().replace(/\s*\n\s*/g, ' ');
        }
    };

    return utils;
});

define('primutils/utils/escapingCommon',[
    'lodash'
], function (_) {
    'use strict';

    var entityMap = {
        unescape : {
            '&amp;' : '&',
            '&lt;' : '<',
            '&gt;' : '>',
            '&#x27;' : '\'',
            '&quot;' : '"',
            '&#x2F;' : '/'
        }
    };
    entityMap.escape = _.invert(entityMap.unescape);

    return {
        unescapeMap : entityMap.unescape,
        escapeMap : entityMap.escape,
        escapeDoubleQuoteInStr : function escapeDoubleQuoteInStr(str) {
            return str.replace(/"/g, function () {
                return entityMap.escape['"'];
            });
        }
    };
});


define('primutils/templates/utils/renderer',[
    'lodash',
    'primutils/utils/escapingCommon'
], function (_, escapingCommon) {
    'use strict';

    /**
     * Provides a set of methods for writing various HTML tags.
     *
     * Generally, the render methods provide two separate "attributes" parameters.
     * The first is used for HTML attributes that are provided by the helper.
     * The second is for attributes provided by the template author.
     * Both are always required, but you can pass `null` for either when a helper does not require them.
     *
     * There is currently no effort to de-dupe the attributes because there is no way to know whether the template author's or the helper author's attribute should be preferred.
     * If you need to merge the helper value with an author provided value, it is suggested that you do the merge, place it in the first attribute object, and delete the author's value from the hash.
     *
     * @class templates.utils.renderer
     * @module templates.utils
     */

    var renderer = {
        /**
         * Converts a key/value collection of attributes into a string of attributes to append to an HTML element
         *
         * @method renderAttributesString
         * @protected
         * @param attributes {Object} key/value collection of attributes
         * @returns {String} HTML attributes string
         */
        //TODO can we escape more than double quotes, i.e., utils.escaping#escapeForHTMLAttribute?  People might be depending on not-fully-escaped strings here...
        renderAttributesString : function (attributes) {
            return _.reduce(attributes, function (memo, value, key) {
                if (typeof value === 'string') {
                    return memo + ' ' + key + '="' + escapingCommon.escapeDoubleQuoteInStr(value.trim()) + '" ';
                } else if (value != null) {
                    return memo + ' ' + key + '="' + value + '" ';
                } else {
                    return memo;
                }
            }, '');
        },

        /**
         * Renders an HTML element with inner text.
         *
         * @method renderElementWithContent
         * @param tag {String} HTML element to render
         * @param attributes {Object|null} Key/value collection of attributes provided by the helper
         * @param dataAttributes {Object|null} Key/value collection of attributes provided by the template author
         * @param text {String} Text/HTML to place inside the element
         * @returns {string} The built HTML element
         */
        renderElementWithContent : function (tag, attributes, dataAttributes, text) {
            return '<' + tag + ' ' + renderer.renderAttributesString(attributes) + renderer.renderAttributesString(dataAttributes) + ' >' + text + '</' + tag + '>';
        },

        /**
         * Renders an HTML element as fully closed.
         *
         * This should be used for elements that do not contain inner text, but must have a proper closing tag.
         * Examples include empty `<div>`s (e.g. that you plan to fill later with jQuery) and icons `<i></i>`.
         *
         * @method renderElement
         * @param tag {String} HTML element to render
         * @param attributes {Object|null} Key/value collection of attributes provided by the helper
         * @param [dataAttributes] {Object|null} Key/value collection of attributes provided by the template author
         * @returns {string} The built HTML element
         */
        renderElement : function (tag, attributes, dataAttributes) {
            return renderer.renderElementWithContent(tag, attributes, dataAttributes, '');
        },

        /**
         * Renders an HTML element that is self closed.
         *
         * This should be used for elements that do not contain inner text and are always self closed.
         * Examples include image tags `<img />`.
         *
         * @method renderSelfClosingElement
         * @param tag {String} HTML element to render
         * @param attributes {Object|null} Key/value collection of attributes provided by the helper
         * @param dataAttributes {Object|null} Key/value collection of attributes provided by the template author
         * @returns {string} The built HTML element
         */
        renderSelfClosingElement : function (tag, attributes, dataAttributes) {
            return '<' + tag + ' ' + renderer.renderAttributesString(attributes) + renderer.renderAttributesString(dataAttributes) + ' />';
        }
    };

    return renderer;
});


define('primutils/templates/helpers/icon',[
    'logger',
    'lodash',
    'primutils/templates/utils/utils',
    'primutils/templates/utils/renderer'
], function (logger, _, utils, renderer) {
    'use strict';

    var baseIconName = 'pgbu-icon',
        baseIconPrefix = baseIconName + '-';

    try {
        /*global requirejs:true*/
        var requireConfig = requirejs.s.contexts._.config;
        baseIconName = requireConfig.markup.baseIconName || baseIconName;
        baseIconPrefix = requireConfig.markup.baseIconPrefix || baseIconName + '-';
    } catch (e) {
        // Do nothing
    }

    function spokenText(title) {
        return renderer.renderElementWithContent('span', {
            'class' : 'spoken-text'
        }, null, title);
    }

    var customIcons = {
        'throbber' : function (fixedIconClass, title, options, size) {
            var dot = renderer.renderElement('div', { 'class' : 'dot' }, {});
            return renderer.renderElementWithContent('i', {
                'class' : utils.mergeClasses([size, 'throbber'], options.hash.class),
                title : title
            }, utils.getElementAttributes(options.hash), dot + dot + dot + spokenText(title));
        }
    };

    function getFixedIconClass(iconClass) {
        return baseIconPrefix + iconClass;
    }

    /**
     * Creates an icon element.
     * See <a href="http://pgbulwf.us.oracle.com/style-guide/?section=web&topic=GraphicDesign&subtopic=Icons">Icons on the Blue Sky UX Guide</a>.
     *
     * Available as: `\{{icon}}`.
     *
     * Icon is a helper that does not have inner content.
     *
     * Example:
     *
     *     \{{icon 'email' 'Email' size="pgbu-icon-small"}}
     *
     * As shown, when specifying the icon type, you do not need to prepend 'pgbu-icon-'.
     * When specifying the icon size you must specify the full size class, e.g. "pgbu-icon-large".
     *
     * The base icon name can be controlled by specifying the following options in the require config:
     *
     *     markup : {
     *         baseIconName : 'pgbu-icon',
     *         baseIconPrefix : 'pgbu-icon-' // Optional, inferred from baseIconName if omitted
     *     }
     *
     * @class templates.helpers.icon
     * @module templates.helpers
     * @constructor
     * @param iconClass {String} The type of the icon
     * @param title {String} The title text of the icon as well as the text spoken by the screen reader
     * @param [options] {Object} Handlebars options
     * @returns {string} Icon element
     */
    var icon = function (iconClass, title, options) {
        logger.assert(arguments.length === 3, 'Templates: Link helper requires a "iconClass" and "title" arguments');

        var size = options.hash.size || baseIconName,
            fixedIconClass = getFixedIconClass(iconClass);

        if (_.has(customIcons, iconClass)) {
            return customIcons[iconClass].call(this, fixedIconClass, title, options, size);
        }

        if (options.data && options.data.inputRadioCheckboxContext && options.data.inputRadioCheckboxContext) {
            options.data.inputRadioCheckboxContext.iconSize = size;
        }

        return renderer.renderElementWithContent('i', {
            'class' : utils.mergeClasses([size, fixedIconClass], options.hash.class),
            title : title
        }, utils.getElementAttributes(options.hash), spokenText(title));
    };

    icon.handlebarsType = 'helper';
    icon.handlebarsName = 'icon';
    icon.getFixedIconClass = getFixedIconClass;

    return icon;
});


define('primutils/templates/helpers/button',[
    'logger',
    'handlebars',
    'lodash',
    'jquery',
    'primutils/templates/utils/utils',
    'primutils/templates/utils/renderer',
    'primutils/utils/language'
], function (logger, Handlebars, _, $, utils, renderer, lang) {
    'use strict';

    var BUTTON_TYPE_MAP = {
        'default' : 'blue-light',
        'primary' : 'action',
        'secondary' : 'blue-dark',
        'page' : 'primary',
        'grid' : 'blue-light',
        'toolbar' : 'blue-light',
        'form' : 'form',
        'not-btn' : 'not-btn'
    };

    var BUTTON_ATTRIBUTES = ['form', 'formaction', 'formenctype', 'formmethod', 'formtarget', 'name', /*'type',*/ 'value'];
    var BUTTON_BOOL_ATTRIBUTES = ['autofocus', 'disabled', 'formnovalidate'];

    function getType(type) {
        var newType = (BUTTON_TYPE_MAP[type] || type);
        return newType !== 'not-btn' ? 'btn-' + newType : newType;
    }

    /**
     * Allows the button to control the related dropdown menu.
     *
     * The button should either appear inside a <a href="templates.helpers.buttonGroup.html">\{{#buttonGroup}}</a> next to a <a href="templates.helpers.dropdownMenu.html">\{{#dropdownMenu}}</a> or
     * should specify a `data-target` attribute which is a jQuery selector to a dropdown menu.
     *
     * Example:
     *
     *     \{{#button 'toolbar' dropdownToggle=true data-target="#myDropdown"}}Menu\{{/button}}
     *
     *     \{{#buttonGroup}}
     *         \{{#button 'toolbar' dropdownToggle=true}}Menu\{{/button}}
     *         \{{#dropdownMenu}} ... \{{/dropdownMenu}}
     *     \{{/buttonGroup}}
     *
     * @property dropdownToggle
     * @for templates.helpers.button
     * @type {Boolean}
     * @default false
     */

    /**
     * Creates a button element.
     * See <a href="http://pgbulwf.us.oracle.com/style-guide/?section=web&topic=Buttons&subtopic=ButtonTypes">Buttons on the Blue Sky UX Guide</a>.
     *
     * Available as: `\{{#button}}\{{/button}}`.
     *
     * Button is a block helper and requires inner content.
     *
     * Example:
     *
     *     \{{#button 'primary' 'My Button Title'}}My Button\{{/button}}
     *
     * There are eight types of buttons: 'default', 'primary', 'secondary', 'page', 'grid', 'toolbar', 'form', and 'not-btn'.
     *
     * You can also use any of the attributes found on <a href="http://www.w3.org/wiki/HTML/Elements/button#HTML_Attributes">W3C Button Attributes</a>.
     *
     * See also: <a href="templates.helpers.buttonGroup.html">\{{#buttonGroup}}</a>, <a href="templates.helpers.toolbar.html">\{{#toolbar}}</a>
     *
     * @class templates.helpers.button
     * @module templates.helpers
     * @constructor
     * @param type {String} The type of the button
     * @param [title] {String} The title text of the button (Default: .text() of the button's inner text, stripped of newlines)
     * @param [options] {Object} Handlebars options
     * @returns {string} Button element
     */
    var button = function (type, title, options) {
        var dropdownClass, dataToggle, role;

        if (arguments.length === 2) {
            options = title;
            title = null;
        } else if (arguments.length === 1) {
            options = type;
            type = null;
            title = null;
        }

        logger.assert(!!options.fn, 'Templates: Button helper must be used as a block helper.');

        if (type != null && type !== '') { // Allow for empty types to just give 'btn'
            type = getType(type);
        }

        if (options.hash.dropdownToggle) {
            dropdownClass = 'dropdown-toggle';
            dataToggle = 'dropdown';
            role = 'button';
        }

        var frame = Handlebars.createFrame(options.data || {});
        var buttonContext = frame.buttonContext = {
            isDropdownToggle : options.hash.dropdownToggle
        };

        var text = options.fn(this, options).trim(),
            hasCaret = !!buttonContext.hasCaret;

        return renderer.renderElementWithContent('button', {
            type : options.hash.type || 'button',
            'class' : utils.mergeClasses([type !== 'not-btn' ? 'btn' : '', type, dropdownClass], options.hash.class),
            title : title != null ? title : utils.textifyElement(text),
            'data-toggle' : dataToggle,
            'aria-pressed' : utils.hasClass(options.hash.class, 'active') ? 'true' : undefined,
            'aria-expanded' : options.hash.dropdownToggle ? 'false' : undefined,
            'role' : role,
            'aria-haspopup' : role ? 'true' : undefined,
            'aria-owns' : role ? '{{menuId}}' : undefined,
            'aria-label' : (!options.hash['aria-label'] && hasCaret) ? lang.getString('templates.moreOptions') : undefined
        }, utils.getElementAttributes(options.hash, BUTTON_ATTRIBUTES, BUTTON_BOOL_ATTRIBUTES), text);
    };

    button.handlebarsType = 'helper';
    button.handlebarsName = 'button';
    button.getType = getType;

    return button;
});


define('primutils/templates/helpers/avatar',[
    'logger',
    'handlebars',
    'lodash',
    'jquery',
    'primutils/templates/utils/utils',
    'primutils/templates/utils/renderer',
    'primutils/templates/helpers/icon',
    'primutils/templates/helpers/button'
], function (logger, Handlebars, _, $, utils, renderer, icon, button) {
    'use strict';

    /**
     * Icon to use as the overlay on the avatar
     *
     * @property overlay
     * @for templates.helpers.avatar
     * @type {String}
     */

    /**
     * Text to describe the overlay
     *
     * @property [overlayLabel]
     * @for templates.helpers.avatar
     * @type {String}
     */

    /**
     * Additional classes to add to the overlay icon
     *
     * @property overlayClass
     * @for templates.helpers.avatar
     * @type {String}
     */

    /**
     * Icon to use as the badge on the avatar
     *
     * @property badge
     * @for templates.helpers.avatar
     * @type {String}
     */

    /**
     * Additional classes to add to the badge icon
     *
     * @property badgeClass
     * @for templates.helpers.avatar
     * @type {String}
     */

    /**
     * Create a button for the badge instead.
     *
     * So it can be clicked to toggle some value, for example a "favorite" state.
     *
     * @property badgeButton
     * @for templates.helpers.avatar
     * @type {Boolean}
     */

    /**
     * Creates an avatar element.
     *
     * Available as: `\{{avatar}}`.
     *
     * Avatar is a helper that does not have inner content.
     *
     * Example:
     *
     *     \{{avatar 'http://localhost/path/to/img'}}
     *
     * There are two additional attributes that can be specified to change the appearance of the avatar: overlay and badge.
     *
     * An overlay is an icon placed over the avatar to indicate the type of item being shown.
     * The attribute is the name of the icon to use as the overlay.
     * These are the same names that are accepted by the `\{{icon}}` helper.
     *
     *     \{{avatar 'path/to/img' overlay="ent-workspace"}}
     *
     * A badge is an icon placed at the corner of the avatar to indicate something about the item.
     * This is often whether or not it is a "favorite" item, but it can be used for other things.
     *
     *     \{{avatar 'path/to/img' badge="favorite"}}
     *
     * You can add classes to the badge or overlay by using the `badgeClass` or `overlayClass` respectively.
     * This can be used to toggle the state of the icon.
     * For examples, see the Blue Sky Style Guide page on icons.
     *
     *     \{{avatar 'path/to/img' badge="favorite" badgeClass="active"}}
     *
     * @class templates.helpers.avatar
     * @module templates.helpers
     * @constructor
     * @param imgUrl {String} The url to the avatar to display
     * @param [options] {Object} Handlebars options
     * @returns {string} Avatar (w/ optional overlay) element
     */
    var avatar = function (imgUrl, options) {
        var innerContent = '';

        logger.assert(arguments.length === 2, 'Templates: Avatar helper must have an imgUrl property. If the avatar is unknown at this time, use an empty string and set it later.');

        innerContent += renderer.renderSelfClosingElement('img', {
            'src' : imgUrl,
            alt : options.hash.alt
        });

        if (options.hash.overlay) {
            innerContent += icon.call(this, options.hash.overlay, options.hash.overlayLabel || '', {
                data : options.data,
                hash : {
                    size : 'overlay',
                    'class' : options.hash.overlayClass
                }
            });
        }

        if (options.hash.badge) {
            if (options.hash.badgeButton) {
                innerContent += button.call(this, 'not-btn', '', {
                    data : options.data,
                    hash : {
                        'class' : utils.mergeClasses(['avatar-badge', icon.getFixedIconClass(options.hash.badge)], options.hash.badgeClass)
                    },
                    fn : function () { return ''; }
                });
            } else {
                innerContent += renderer.renderElement('div', {
                    'class' : utils.mergeClasses(['avatar-badge', icon.getFixedIconClass(options.hash.badge)], options.hash.badgeClass)
                });
            }
        }

        return renderer.renderElementWithContent('div', {
            'class' : utils.mergeClasses(['avatar'], options.hash.class)
        }, utils.getElementAttributes(options.hash), innerContent);
    };

    avatar.handlebarsType = 'helper';
    avatar.handlebarsName = 'avatar';

    return avatar;
});


define('primutils/templates/helpers/link',[
    'logger',
    'primutils/templates/utils/utils',
    'primutils/templates/utils/renderer',
    'primutils/templates/helpers/button'
], function (logger, utils, renderer, button) {
    'use strict';

    var LINK_ATTRIBUTES = ['download', /*'href',*/ 'hreflang', 'media', 'name', 'rel', 'target', 'type'];

    /**
     * Allows the link to appear as a button.
     * You can specify any of the valid <a href="templates.helpers.button.html">button types</a>.
     *
     * @property asButton
     * @for templates.helpers.link
     * @type {String|null}
     * @default null
     */

    /**
     * Indicates that the link is to a file, rather than a simple navigation link.
     *
     * If this is set, will automatically use <a href="utils.fileSaver.html">fileSaver</a> to download the link, rather than navigation.
     *
     * @property data-fileDownload
     * @for templates.helpers.link
     * @type {Boolean}
     * @default false
     */

    /**
     * Creates a link element.
     *
     * Available as: `\{{#link}}\{{/link}}`.
     *
     * Link is a block helper and requires inner content.
     *
     * Example:
     *
     *     \{{#link 'http://www.google.com/' 'Google'}}Go To Google\{{/link}}
     *
     * You can also use any of the attributes found on <a href="http://www.w3.org/wiki/HTML/Elements/a#HTML_Attributes">W3C Anchor Attributes</a>.
     *
     * If only one arguments are passed, it is assumed to be href.
     * If you want to specify just the item's title, you should pass '#' as the first argument.
     *
     * See also: <a href="templates.helpers.button.html">\{{#button}}</a>
     *
     * @class templates.helpers.link
     * @module templates.helpers
     * @constructor
     * @param href {String} The end point of the anchor
     * @param [title] {String} The title text of the link
     * @param [options] {Object} Handlebars options
     * @returns {string} Link element
     */
    var link = function (href, title, options) {
        logger.assert(arguments.length >= 2, 'Templates: Link helper requires an "href" argument');

        if (arguments.length === 2) {
            options = title;
            title = null;
        }

        logger.assert(!!options.fn, 'Templates: Link helper must be used as a block helper.');

        var type;
        if (options.hash.asButton) {
            type = button.getType(options.hash.asButton);
        }

        var clz = options.hash.asButton ?
            utils.mergeClasses([type !== 'not-btn' ? 'btn' : '', type], options.hash.class) :
            options.hash.class;

        return renderer.renderElementWithContent('a', {
            href : href,
            title : title,
            class : clz,
            role : options.hash.asButton ? 'button' : options.hash._internalRole
        },
            utils.getElementAttributes(options.hash, LINK_ATTRIBUTES), options.fn(this, options));
    };

    link.handlebarsType = 'helper';
    link.handlebarsName = 'link';

    return link;
});


define('primutils/templates/helpers/accordion',[
    'logger',
    'handlebars',
    'lodash',
    'primutils/templates/utils/utils',
    'primutils/templates/utils/renderer',
    'primutils/templates/helpers/link'
], function (logger, Handlebars, _, utils, renderer, link) {
    'use strict';

    var TYPE_CLASS_MAP = {
        'twistie' : 'toggle-accordion',
        'grid' : 'grid-accordion',
        'folded' : 'folded-accordion'
    };

    var getAccordionHeaderHash = function (headerId, bodyId, parent, open, extraClasses) {
        return {
            'id' : headerId,
            'aria-expanded' : open ? 'true' : 'false',
            'aria-controls' : bodyId,
            'aria-selected' : open ? 'true' : 'false',
            'role' : 'tab',
            'class' : utils.mergeClasses(['accordion-toggle'], extraClasses || ''),
            'data-toggle' : 'collapse',
            'data-parent' : parent ? '#' + parent : undefined
        };
    };

    var accordionHeaderBase = function (content) {
        return renderer.renderElementWithContent('div', {
            'class' : 'accordion-heading'
        }, null, content);
    };

    var accordionHeader = function (name, headerId, bodyId, parent, open, data) {
        var accordionHeaderHash = getAccordionHeaderHash(headerId, bodyId, parent, open);
        return accordionHeaderBase(link.call(this, '#' + bodyId, null, {
            hash : accordionHeaderHash,
            fn : function () { return name; },
            data : data
        }));
    };

    var accordionBody = function (headerId, innerText, open, hash, foldAnimate, neverAriaHidden) {
        return renderer.renderElementWithContent('div', {
            'class' : 'accordion-body collapse' + (open ? ' in' : ' hidden') + (foldAnimate ? ' fold-animated' : '')
        }, utils.getElementAttributes(hash), renderer.renderElementWithContent('div', {
            'role' : 'tabpanel',
            'aria-labelledby' : headerId,
            'aria-hidden' : (open || neverAriaHidden) ? 'false' : 'true',
            'class' : utils.mergeClasses(['accordion-inner'], hash.class)
        }, utils.getElementAttributes(_.omit(hash, ['role', 'aria-labelledby', 'aria-hidden', 'id'])), innerText));
    };

    var userHeaderInner = function (accordionSectionContext, accordionContext, options) {
        var fixedHeaderHash = getAccordionHeaderHash(accordionSectionContext.headerId, accordionSectionContext.bodyId, accordionSectionContext.parent, accordionSectionContext.open, options.hash.class);
        fixedHeaderHash['data-target'] = '#' + accordionSectionContext.bodyId;
        if (accordionContext.type === 'grid' || accordionContext.type === 'folded') {
            fixedHeaderHash.tabindex = '0';
        }
        var fixedHeaderKeys = _.keys(fixedHeaderHash);

        return renderer.renderElementWithContent('div',
            fixedHeaderHash,
            utils.getElementAttributes(_.omit(options.hash, fixedHeaderKeys)),
            options.fn(this, options));
    };

    /**
     * Creates the header of an accordion section.
     * See <a href="http://pgbulwf.us.oracle.com/style-guide/?section=web&topic=Navigation&subtopic=ToggleAccordion">Toggle Accordion on the Blue Sky UX Guide</a>.
     *
     * Available as: `\{{#accordionHeader}}\{{/accordionheader}}`.
     *
     * Accordion section header is a block helper and requires inner content.
     *
     * Accordion section header can only be used inside of an <a href="templates.helpers.accordionSection.html">accordionSection</a>.
     *
     * @class templates.helpers.accordionHeader
     * @module templates.helpers
     * @protected
     * @constructor
     * @param [options] {Object} Handlebars options
     * @returns {string} Accordion section header element
     */
    var accordionUserHeader = function (options) {
        var accordionSectionContext = options.data && options.data.accordionSectionContext,
            accordionContext = options.data && options.data.accordionContext;
        logger.assert(accordionSectionContext, 'Templates: accordionHeader helper must be called from within a {{#accordionSection}}');
        logger.assert(!!options.fn, 'Templates: accordionHeader helper must be used as a block helper.');

        accordionSectionContext.headerContent = accordionHeaderBase(userHeaderInner.call(this, accordionSectionContext, accordionContext, options));
    };

    accordionUserHeader.handlebarsType = 'helper';
    accordionUserHeader.handlebarsName = 'accordionHeader';

    /**
     * Creates the body of an accordion section.
     * See <a href="http://pgbulwf.us.oracle.com/style-guide/?section=web&topic=Navigation&subtopic=ToggleAccordion">Toggle Accordion on the Blue Sky UX Guide</a>.
     *
     * Available as: `\{{#accordionBody}}\{{/accordionBody}}`.
     *
     * Note: To set the ID of this element, set the ID of the parent `\{{#accordionSection}}`.
     *
     * Accordion section body is a block helper and requires inner content.
     *
     * Accordion section body can only be used inside of an <a href="templates.helpers.accordionSection.html">accordionSection</a>.
     *
     * @class templates.helpers.accordionBody
     * @module templates.helpers
     * @protected
     * @constructor
     * @param [options] {Object} Handlebars options
     * @returns {string} Accordion section body element
     */
    var accordionUserBody = function (options) {
        var accordionSectionContext = options.data && options.data.accordionSectionContext;
        logger.assert(accordionSectionContext, 'Templates: accordionBody helper must be called from within a {{#accordionSection}}');
        logger.assert(!!options.fn, 'Templates: accordionBody helper must be used as a block helper.');

        if (_.has(options.hash, 'id')) {
            logger.error('Templates: accordionBody ID needs to be set on the accordionSection.');
        }

        options.hash.id = accordionSectionContext.bodyId;

        accordionSectionContext.bodyContent = accordionBody(
            accordionSectionContext.headerId,
            options.fn(this, options),
            accordionSectionContext.open,
            options.hash,
            false);
    };

    accordionUserBody.handlebarsType = 'helper';
    accordionUserBody.handlebarsName = 'accordionBody';

    /**
     * Opens this accordion section initially.
     * @property open
     * @for templates.helpers.accordionSection
     * @type {Boolean}
     * @default false
     */

    /**
     * Creates an accordion section.
     * See <a href="http://pgbulwf.us.oracle.com/style-guide/?section=web&topic=Navigation&subtopic=ToggleAccordion">Toggle Accordion on the Blue Sky UX Guide</a>.
     *
     * Available as: `\{{#accordionSection}}\{{/accordionSection}}`.
     *
     * Accordion section is a block helper and requires inner content.
     *
     * Accordion section can only be used inside of an <a href="templates.helpers.accordion.html">accordion</a>.
     *
     * @class templates.helpers.accordionSection
     * @module templates.helpers
     * @protected
     * @constructor
     * @param [name] {String} The name of the section
     * @param [options] {Object} Handlebars options
     * @returns {string} Accordion section element
     */
    var accordionSection = function (name, options) {
        if (arguments.length === 1) {
            options = name;
            name = '';
        }

        var accordionContext = options.data && options.data.accordionContext;
        logger.assert(accordionContext, 'Templates: Accordion section helper must be called from within a {{#accordion}}');
        logger.assert(!!options.fn, 'Templates: Accordion section helper must be used as a block helper.');

        options.hash.id = options.hash.id || _.uniqueId('pgbu-accordion-section-');

        var frame = Handlebars.createFrame(options.data || {}),
            sectionContext = frame.accordionSectionContext = {
                headerId : _.uniqueId('pgbu-accordion-header-'),
                bodyId : options.hash.id,
                parent : accordionContext.independent ? null : accordionContext.id,
                open : options.hash.open
            };

        var innerText = options.fn(this, { data : frame }),
            headerText = sectionContext.headerContent || accordionHeader.call(this, name, sectionContext.headerId, sectionContext.bodyId, sectionContext.parent, sectionContext.open, options.data),
            usesInnerHelpers = (sectionContext.headerContent || sectionContext.bodyContent),
            bodyText = usesInnerHelpers ? (sectionContext.bodyContent || '') : accordionBody(sectionContext.headerId, innerText, sectionContext.open, _.omit(options.hash, 'class'), false);

        return renderer.renderElementWithContent('div', {
            'class' : utils.mergeClasses(['accordion-group', (options.hash.open ? 'open' : '')], options.hash.class)
        }, usesInnerHelpers ? utils.getElementAttributes(_.pick(options.hash, function (value, key) {
            return key.indexOf('data-') === 0;
        })) : null, headerText + bodyText);
    };

    accordionSection.handlebarsType = 'helper';
    accordionSection.handlebarsName = 'accordionSection';

    /**
     * Makes the accordion "twistie" style instead of the Bootstrap default.
     * @property twistie
     * @for templates.helpers.accordion
     * @type {Boolean}
     * @default false
     */

    /**
     * Makes all of the accordion sections independent of one another.
     * Normally, only one section can be open at a time.
     * When true, this allows multiple sections to be open at once.
     * @property independent
     * @for templates.helpers.accordion
     * @type {Boolean}
     * @default false
     */

    /**
     * Creates an accordion component.
     * See <a href="http://pgbulwf.us.oracle.com/style-guide/?section=web&topic=Navigation&subtopic=ToggleAccordion">Toggle Accordion on the Blue Sky UX Guide</a>.
     *
     * Available as: `\{{#accordion}}\{{/accordion}}`.
     *
     * Accordion is a block helper and requires inner content.
     *
     * Accordion requires some sub-helpers:
     *  - <a href="templates.helpers.accordionSection.html">accordionSection</a>
     *
     * Example:
     *
     *     \{{#accordion twistie=true}}
     *         \{{#accordionSection "Section 1" open=true}}
     *             ... Section Content ...
     *         \{{/accordionSection}}
     *     \{{/accordion}}
     *
     * @class templates.helpers.accordion
     * @module templates.helpers
     * @constructor
     * @param [type] {Object} The type of accordion to create
     * @param [options] {Object} Handlebars options
     * @returns {string} Accordion element
     */
    var accordion = function (type, options) {
        if (arguments.length === 1) {
            options = type;
            type = null;
        }

        if (options.hash.twistie) {
            type = 'twistie';
            delete options.hash.twistie;
        }

        logger.assert(!!options.fn, 'Templates: Accordion helper must be used as a block helper.');

        options.hash.id = options.hash.id || _.uniqueId('pgbu-accordion-');

        var typeClass = _.has(TYPE_CLASS_MAP, type) ? TYPE_CLASS_MAP[type] : '',
            frame = Handlebars.createFrame(options.data || {}),
            accordionContext = frame.accordionContext = {
                id : options.hash.id,
                independent : options.hash.independent,
                type : type
            };

        var innerText = options.fn(this, { data : frame });

        return renderer.renderElementWithContent('div', {
            'class' : utils.mergeClasses(['accordion', typeClass], options.hash.class),
            'role' : 'tablist',
            'aria-multiselectable' : accordionContext.independent ? 'true' : 'false'
        }, utils.getElementAttributes(options.hash), innerText);
    };

    accordion.handlebarsType = 'helper';
    accordion.handlebarsName = 'accordion';

        if (require.nodeRequire && !_) {
        _ = require.nodeRequire('lodash');
    }
    
    var belowFoldBase = function (element, options) {
        var foldedAccordionSectionContext = options.data && options.data.accordionSectionContext,
            isFoldedSection = foldedAccordionSectionContext && foldedAccordionSectionContext.foldedSection;
        logger.assert(foldedAccordionSectionContext && isFoldedSection, 'Templates: Below fold helper must be called from within a {{#foldedAccordionSection}}');
        logger.assert(!!options.fn, 'Templates: Below fold helper must be used as a block helper.');

        return renderer.renderElementWithContent(element, {
            'class' : utils.mergeClasses(['folded-section'], options.hash.class)
        }, utils.getElementAttributes(options.hash), options.fn(this, options));
    };

    var belowFoldSpan = _.partial(belowFoldBase, 'span');
    belowFoldSpan.handlebarsType = 'helper';
    belowFoldSpan.handlebarsName = 'belowFoldSpan';

    var belowFoldDiv = _.partial(belowFoldBase, 'div');
    belowFoldDiv.handlebarsType = 'helper';
    belowFoldDiv.handlebarsName = 'belowFoldDiv';

    var foldedAccordionSectionToggle = function (options) {
        var foldedAccordionSectionContext = options.data && options.data.accordionSectionContext,
            isFoldedSection = foldedAccordionSectionContext && foldedAccordionSectionContext.foldedSection;
        logger.assert(foldedAccordionSectionContext && isFoldedSection, 'Templates: Accordion folded section toggle helper must be called from within a {{#foldedAccordionSection}}');
        logger.assert(!!options.fn, 'Templates: Accordion folded section toggle helper must be used as a block helper.');

        return userHeaderInner.call(this, foldedAccordionSectionContext, options.data.accordionContext, options);
    };

    foldedAccordionSectionToggle.handlebarsType = 'helper';
    foldedAccordionSectionToggle.handlebarsName = 'foldedAccordionSectionToggle';

    var foldedAccordionSection = function (options) {
        var accordionContext = options.data && options.data.accordionContext;
        logger.assert(accordionContext && accordionContext.type === 'folded', 'Templates: Accordion folded section helper must be called from within a {{#accordion \'folded\'}}');
        logger.assert(!!options.fn, 'Templates: Accordion folded section helper must be used as a block helper.');

        options.hash.id = options.hash.id || _.uniqueId('pgbu-accordion-section-');

        var frame = Handlebars.createFrame(options.data || {}),
            sectionContext = frame.accordionSectionContext = {
                headerId : _.uniqueId('pgbu-accordion-header-'),
                bodyId : options.hash.id,
                parent : accordionContext.independent ? null : accordionContext.id,
                open : options.hash.open,
                foldedSection : true
            };

        var innerText = options.fn(this, { data : frame }),
            foldAnimate = _.has(options.hash.foldAnimate) ? !!options.hash.foldAnimate : true,
            bodyText = accordionBody(sectionContext.headerId, innerText, sectionContext.open, _.omit(options.hash, 'class'), foldAnimate, true);

        return renderer.renderElementWithContent('div', {
            'class' : utils.mergeClasses(['accordion-group', (options.hash.open ? 'open' : '')], options.hash.class)
        }, null, bodyText);
    };

    foldedAccordionSection.handlebarsType = 'helper';
    foldedAccordionSection.handlebarsName = 'foldedAccordionSection';

    return {
        accordion : accordion,
        accordionSection : accordionSection,
        accordionHeader : accordionUserHeader,
        accordionBody : accordionUserBody,

        foldedAccordionSection : foldedAccordionSection,
        foldedAccordionSectionToggle : foldedAccordionSectionToggle,
        belowFoldSpan : belowFoldSpan,
        belowFoldDiv : belowFoldDiv
    };
});


define('primutils/templates/helpers/buttonGroup',[
    'logger',
    'primutils/templates/utils/utils',
    'primutils/templates/utils/renderer',
    'handlebars'
], function (logger, utils, renderer, Handlebars) {
    'use strict';

    /**
     * Creates a button group.
     * See <a href="http://pgbulwf.us.oracle.com/style-guide/?section=web&topic=Buttons&subtopic=ButtonTypes">Buttons on the Blue Sky UX Guide</a>.
     *
     * Available as: `\{{#buttonGroup}}\{{/buttonGroup}}`.
     *
     * Button group is a block helper and requires inner content, generally <a href="templates.helpers.button.html">\{{#button}}</a> and <a href="templates.helpers.dropdownMenu.html">\{{#dropdownMenu}}</a>.
     *
     * Using `data-` attributes, you can activate the <a href="http://getbootstrap.com/2.3.2/javascript.html#buttons">Bootstrap</a> checkbox or radio functionality.
     *
     * Example:
     *
     *     \{{#buttonGroup}}
     *         \{{#button 'toolbar'}}Edit\{{/button}}
     *     \{{/buttonGroup}}
     *
     * See also: <a href="templates.helpers.button.html">\{{#button}}</a>, <a href="templates.helpers.toolbar.html">\{{#toolbar}}</a>
     *
     * @class templates.helpers.buttonGroup
     * @module templates.helpers
     * @constructor
     * @param [options] {Object} Handlebars options
     * @returns {string} Button group element
     */
    var buttonGroup = function (options) {
        logger.assert(!!options.fn, 'Templates: Button group helper must be used as a block helper.');

        var frame = Handlebars.createFrame(options.data || {});
        var buttonGroupContext = frame.buttonGroupContext = {};

        var innerText = options.fn(this, { data : frame });

        if (buttonGroupContext.hasDropdownMenu) {
            innerText = innerText.replace(/\{\{menuId\}\}/g, buttonGroupContext.menuId);
        }

        return renderer.renderElementWithContent('div', {
            'class' : utils.mergeClasses(['btn-group'], options.hash.class)
        }, utils.getElementAttributes(options.hash), innerText);
    };

    buttonGroup.handlebarsType = 'helper';
    buttonGroup.handlebarsName = 'buttonGroup';

    return buttonGroup;
});


define('primutils/templates/helpers/card',[
    'logger',
    'lodash',
    'handlebars',
    'primutils/templates/utils/utils',
    'primutils/templates/utils/renderer',
    'primutils/templates/helpers/link',
    'primutils/templates/helpers/icon'
], function (logger, _, Handlebars, utils, renderer, link, icon) {
    'use strict';

    var cardHeader = function (header, iconClass, iconTitle) {
        logger.assert(iconClass == null || iconTitle != null, 'Templates: Card header must include iconTitle attribute when an icon is added');

        var iconMarkup = '';
        if (iconClass != null) {
            iconMarkup = renderer.renderElementWithContent('div', {
                'class' : 'pull-right'
            }, null, link.call(this, '#', iconTitle, {
                hash : {
                    'class' : 'flip-indicator'
                },
                fn : function (context) {
                    return icon.call(context, iconClass, iconTitle, {
                        hash : {}
                    });
                }
            }));
        }

        return renderer.renderElementWithContent('div', {
            'class' : 'card-header'
        }, null, header + iconMarkup);
    };

    var cardSideBase = function (side, header, iconClass, options) {
        if (arguments.length === 3) {
            options = iconClass;
            iconClass = null;
        } else if (arguments.length === 2) {
            options = header;
            header = null;
            iconClass = null;
        }

        var cardContext = options.data && options.data.cardContext;
        logger.assert(cardContext, 'Templates: Card side (' + side + ') helper must be called from within a {{#panelContainer}}');
        logger.assert(!!options.fn, 'Templates: Card side (' + side + ') helper must be used as a block helper.');

        cardContext.created[side] = true;

        var innerContent = renderer.renderElementWithContent('div', {
            'class' : 'card-content'
        }, null, options.fn(this, options));

        if (header != null) {
            innerContent = cardHeader.call(this, header, iconClass, options.hash.iconTitle) + innerContent;
        }

        return renderer.renderElementWithContent('div', {
            'class' : side
        }, null, innerContent);
    };

    if (require.nodeRequire && !_) {
        _ = require.nodeRequire('lodash');
    }

    var cardFront = _.partial(cardSideBase, 'front');
    cardFront.handlebarsType = 'helper';
    cardFront.handlebarsName = 'cardFront';

    var cardBack = _.partial(cardSideBase, 'back');
    cardBack.handlebarsType = 'helper';
    cardBack.handlebarsName = 'cardBack';

    var card = function (options) {
        logger.assert(!!options.fn, 'Templates: Card helper must be used as a block helper.');

        var frame = Handlebars.createFrame(options.data || {});
        var cardContext = frame.cardContext = {
            created : {
                front : false,
                back : false
            }
        };

        var innerContent = options.fn(this, { data : frame });

        if (!cardContext.created.front && !utils.hasClass(options.hash.class, 'card-back-static')) {
            logger.warn('Created card without front-side');
        }
        if (!cardContext.created.back && !utils.hasClass(options.hash.class, 'card-front-static')) {
            logger.warn('Created card without back-side');
        }

        return renderer.renderElementWithContent('div', {
            'class' : utils.mergeClasses(['card', options.hash.flipped ? 'flipped' : ''], options.hash.class)
        }, utils.getElementAttributes(options.hash), innerContent);
    };

    card.handlebarsType = 'helper';
    card.handlebarsName = 'card';

    return {
        'card' : card,
        'cardFront' : cardFront,
        'cardBack' : cardBack
    };
});


define('primutils/templates/helpers/caret',['primutils/templates/utils/renderer'], function (renderer) {
    'use strict';

    /**
     * Creates a down-facing caret.
     *
     * Generally used as the content of a dropdown toggle button.
     *
     * Available as: `\{{caret}}`.
     *
     * Caret is a helper that does not have inner content.
     *
     * Example:
     *
     *     \{{caret}}
     *
     * See also: <a href="templates.helpers.button.html">\{{#button}}</a>
     *
     * @class templates.helpers.caret
     * @module templates.helpers
     * @constructor
     * @param [options] {Object} Handlebars options
     * @returns {string} Caret element
     */
    var caret = function (options) {
        var buttonContext = options.data && options.data.buttonContext;
        if (buttonContext && buttonContext.isDropdownToggle) {
            buttonContext.hasCaret = true;
        }
        return renderer.renderElement('span', { 'class' : 'caret' });
    };

    caret.handlebarsType = 'helper';
    caret.handlebarsName = 'caret';

    return caret;
});


define('primutils/templates/helpers/controlGroup',[
    'logger',
    'handlebars',
    'lodash',
    'primutils/templates/utils/utils',
    'primutils/templates/utils/renderer'
], function (logger, Handlebars, _, utils, renderer) {
    'use strict';

    var controlLabel = function (options) {
        var controlGroupContext = options.data && options.data.controlGroupContext;
        logger.assert(controlGroupContext, 'Templates: controlLabel helper must be called from within a {{#controlGroup}}');

        var innerText = options.fn(this, options);

        controlGroupContext.labelContent = renderer.renderElementWithContent('label', {
            'class' : 'control-label',
            style : !controlGroupContext.useDefaultLabel ? 'overflow: inherit;' : undefined,
            for : controlGroupContext.inputs.length > 0 ? controlGroupContext.inputs[0] : controlGroupContext.for,
            id : controlGroupContext.labelId,
            title : controlGroupContext.label
        }, null, innerText);
    };

    controlLabel.handlebarsType = 'helper';
    controlLabel.handlebarsName = 'controlLabel';

    var controlContent = function (options) {
        var controlGroupContext = options.data && options.data.controlGroupContext;
        logger.assert(controlGroupContext, 'Templates: controlContent helper must be called from within a {{#controlGroup}}');

        var innerText = options.fn(this, options);

        controlGroupContext.controlContent = renderer.renderElementWithContent('div', {
            'class' : utils.mergeClasses(['controls'], controlGroupContext.readOnly ? 'controls-view-only' : '')
        }, null, innerText);
    };

    controlContent.handlebarsType = 'helper';
    controlContent.handlebarsName = 'controlContent';

    /**
     * Marks the control group as required.
     *
     * The property may be set automatically by contained `\{{input}}`s.
     * If the control group contains an `\{{input required=true}}`, the control group will also automatically be marked as required.
     * However, the reverse is not true.
     * If the control group is marked as required, contained `\{{input}}`s do not automatically become required as well.
     *
     * @property required
     * @for templates.helpers.controlGroup
     * @type {Boolean}
     * @default false
     */

    /**
     * Specifies that the contents on the control group is only plain text.
     * If you are only displaying a value and not creating an input, you should set this property to true.
     * This should **not** be set if the control group contains an input that has been set to readonly.
     *
     * See the difference between temporarily disabled inputs and "view-only fields" on the <a href="http://pgbulwf.us.oracle.com/style-guide/?section=web&topic=InputTypes&subtopic=ViewOnly">Blue Sky UX Guide</a>.
     *
     * @property readOnly
     * @for templates.helpers.controlGroup
     * @type {Boolean}
     * @default false
     */

    /**
     * Specifies that the contained inputs are a group of radio buttons that should be on one line instead of on top of each other.
     *
     * @property inlineRadio
     * @for templates.helpers.controlGroup
     * @type {Boolean}
     * @default false
     */

    /**
     * Specifies the HTML `name` attribute to be given to all of the contained inputs.
     * If any underlying inputs do not have their `name` attribute specified, they will get this one automatically.
     * Generally used when the contained inputs are all checkboxes or radio buttons.
     *
     * @property name
     * @for templates.helpers.controlGroup
     * @type {String|null}
     * @default null
     */

    /**
     * Specifies the ID attribute of the generated label.
     * Used in cases where the page author needs to access the `<label>` once its created.
     *
     * If not specified, it will receive an auto-incrementing ID so that `<input>`s can take advantage of it for accessibility purposes.
     *
     * @property labelId
     * @for templates.helpers.controlGroup
     * @type {String}
     * @default Auto-incrementing value
     */

    /**
     * Specifies the for attribute of the generated label.
     * Used in cases when the input is created after the `<label>` is created.
     *
     * If not specified, it will receive the ID of the first input in the control group so that `<input>`s can take advantage of it for accessibility purposes.
     *
     * @property for
     * @for templates.helpers.controlGroup
     * @type {String}
     * @default Auto-incrementing value
     */

    /**
     * Renders the control as inline with the label right next to the control.
     *
     * Useful for toolbars.
     *
     * @property inline
     * @for templates.helpers.controlGroup
     * @type {Boolean}
     * @default false
     */

    /**
     * Places icons in radio buttons and checkboxes on the left side of the input.
     *
     * @property leftIcons
     * @for templates.helpers.controlGroup
     * @type {Boolean}
     * @default false
     */

    /**
     * Creates a control group section.
     * See <a href="http://pgbulwf.us.oracle.com/style-guide/?section=web&topic=Forms&subtopic=LabelandFieldLayout">Forms on the Blue Sky UX Guide</a>.
     *
     * Available as: `\{{#controlGroup}}\{{/controlGroup}}`.
     *
     * Control group is a block helper and requires inner content.
     *
     * Example:
     *
     *     \{{#controlGroup 'My Input'}}
     *         \{{input 'text'}}
     *     \{{/controlGroup}}
     *
     * See also: <a href="templates.helpers.input.html">\\{{input}}</a>
     *
     * @class templates.helpers.controlGroup
     * @module templates.helpers
     * @constructor
     * @param label {String} The control group's label text
     * @param [options] {Object} Handlebars options
     * @returns {string} Control group element
     */
    var controlGroup = function (label, options) {
        logger.assert(arguments.length === 2, 'Templates: Control group helper requires a "label" argument.');
        logger.assert(!!options.fn, 'Templates: Control group helper must be used as a block helper.');

        options.hash.labelId = options.hash.labelId || _.uniqueId('pgbu-label-');

        var frame = Handlebars.createFrame(options.data || {});
        var controlGroupContext = frame.controlGroupContext = {
            labelId : options.hash.labelId,
            inputRequired : !!options.hash.required,
            inputs : [],
            name : options.hash.name,
            readOnly : !!options.hash.readOnly,
            for : options.hash.for || '{{forInputId}}',
            label : label,
            leftIcons : !!options.hash.leftIcons
        };

        var result = options.fn(this, { data : frame });

        if (!controlGroupContext.labelContent) {
            controlGroupContext.useDefaultLabel = true;
            controlLabel.call(this, {
                data : frame,
                fn : function () { return label; }
            });
        }

        if (!controlGroupContext.controlContent) {
            controlContent.call(this, {
                data : frame,
                fn : function () { return result; }
            });
        }

        controlGroupContext.labelContent = controlGroupContext.labelContent.replace('{{forInputId}}', controlGroupContext.inputs.length > 0 ? controlGroupContext.inputs[0] : '');

        return renderer.renderElementWithContent('div', {
            'class' : utils.mergeClasses(['control-group',
                controlGroupContext.inputRequired ? 'required' : '',
                options.hash.inlineRadio ? 'controls-yes-no' : '',
                options.hash.inline ? 'control-group-inline' : ''
            ], options.hash.class),
            'role' : controlGroupContext.isRadioGroup ? 'radiogroup' : undefined
        }, utils.getElementAttributes(options.hash), controlGroupContext.labelContent + controlGroupContext.controlContent);
    };

    controlGroup.handlebarsType = 'helper';
    controlGroup.handlebarsName = 'controlGroup';

    return {
        controlGroup : controlGroup,
        controlLabel : controlLabel,
        controlContent : controlContent
    };
});


define('primutils/templates/helpers/dock',[
    'logger',
    'lodash',
    'primutils/templates/utils/utils',
    'primutils/templates/utils/renderer',
    'primutils/templates/helpers/link',
    'primutils/templates/helpers/button',
    'primutils/templates/helpers/icon',
    'primutils/utils/language'
], function (logger, _, utils, renderer, link, button, icon, lang) {
    'use strict';

    var VALID_ORIENTATIONS = ['left', 'right', 'top', 'bottom'],
        SUPPORTS_ORIENTATION_SWITCHING = ['right', 'bottom'],
        ARIA_ORIENTATIONS = {
            left : 'vertical',
            right : 'vertical',
            top : 'horizontal',
            bottom : 'horizontal'
        };

    var ORIENTATION_LANG_STRINGS = {
        'vertical' : 'dock.dockRight',
        'horizontal' : 'dock.dockBottom'
    };

    var orientButton = function (orientation, active) {
        return button.call(this, '', lang.getString(ORIENTATION_LANG_STRINGS[orientation]), {
            hash : { 'class' : active ? 'active' : '', 'data-orientation' : orientation },
            fn : function () {
                return icon.call(this, 'details-' + orientation, lang.getString(ORIENTATION_LANG_STRINGS[orientation]), {
                    hash : { size : 'pgbu-icon-small' }
                });
            }
        });
    };

    var dockHeader = function (title, titleId, initialOrientation) {
        var titleBlock = renderer.renderElementWithContent('div', {
                'class' : 'dock-title',
                id : titleId
            }, null, title),
            dockControls = renderer.renderElementWithContent('div', {
                'class' : 'dock-controls',
                'data-toggle' : 'buttons-radio'
            }, null, orientButton.call(this, 'horizontal', initialOrientation === 'bottom') + orientButton.call(this, 'vertical', initialOrientation === 'right'));

        return renderer.renderElementWithContent('div', {
            'class' : 'dock-header'
        }, null, titleBlock + dockControls);
    };

    /**
     * Allows the user to change the orientation of the dock.
     *
     * This option only applies to docks that are docked right or bottom.
     * Docks that are docked left or top do not support reorientation.
     *
     * @property allowOrientation
     * @for templates.helpers.dock
     * @type {Boolean}
     * @default false
     */

    /**
     * Shows the dock title.
     *
     * @property showTitle
     * @for templates.helpers.dock
     * @type {Boolean}
     * @default false
     */

    /**
     * Prevents the user from collapsing the dock.
     *
     * @property noCollapse
     * @for templates.helpers.dock
     * @type {Boolean}
     * @default false
     */

    /**
     * Prevents the user from manually resizing the dock.
     *
     * @property noResize
     * @for templates.helpers.dock
     * @type {Boolean}
     * @default false
     */

    /**
     * Creates a dock component.
     *
     * Available as: `\{{#dock}}\{{/dock}}`.
     *
     * Dock is a block helper and requires inner content.
     *
     * In order for a dock to work properly, it must be placed inside of an element with the ".dock-parent" class.
     *
     * The most likely use case in Prime is for a [`\{{#panelContainer}}`](templates.helpers.panelContainer.html) to be placed inside the dock.
     *
     * Example:
     *
     *     <div class="dock-parent">
     *         ... content that should be tabbed through before the contents of the dock ...
     *
     *         \{{#dock 'My Dock' 'left'}}
     *             ... dock inner content ...
     *         \{{/dock}}
     *
     *         ... content that should be tabbed through after the contents of the dock ...
     *     </div>
     *
     * See Also: [\{{#panelContainer}}](templates.helpers.panelContainer.html)
     *
     * @class templates.helpers.dock
     * @module templates.helpers
     * @constructor
     * @param title {String} The title of the dock
     * @param [initialOrientation] {String} The initial location of the dock (Default: 'right')
     * @param [options] {Object} Handlebars options
     * @returns {string} Dock element
     */
    var dock = function (title, initialOrientation, options) {
        logger.assert(arguments.length >= 2, 'Templates: Dock helper requires a "title" argument');

        if (arguments.length === 2) {
            options = initialOrientation;
            initialOrientation = 'right';
        }

        logger.assert(!!options.fn, 'Templates: Dock helper must be used as a block helper.');
        logger.assert(_.includes(VALID_ORIENTATIONS, initialOrientation), 'Templates: Dock initialOrientation must be: left, right, top, or bottom');

        var collapsed = _.includes(utils.classList(options.hash.class), 'collapsed');

        if (!_.includes(SUPPORTS_ORIENTATION_SWITCHING, initialOrientation)) {
            options.hash.allowOrientation = false;
        }

        options.hash.id = options.hash.id || _.uniqueId('pgbu-dock-');
        var titleId = _.uniqueId('pgbu-dock-title-');

        var orientationClass = 'dock-' + initialOrientation,
            innerText = options.fn(this, options),
            resizeHandle = options.hash.noResize ? '' : renderer.renderElement('div', {
                'role' : 'separator',
                'aria-controls' : options.hash.id,
                'aria-orientation' : ARIA_ORIENTATIONS[initialOrientation],
                'aria-expanded' : !collapsed,
                tabindex : '0',
                'class' : 'resize-handle'
            }, null),
            collapseHandleString = collapsed ? lang.getString('dock.expand') : lang.getString('dock.collapse'),
            collapseHandle = options.hash.noCollapse ? '' : link.call(this, '#', collapseHandleString, {
                hash : { 'class' : 'collapse-handle', 'aria-labelledby' : titleId },
                fn : function () { return renderer.renderElementWithContent('span', { 'class' : 'spoken-text' }, null, collapseHandleString); }
            }),
            header = dockHeader.call(this, title, titleId, initialOrientation),
            body = renderer.renderElementWithContent('div', { 'class' : 'dock-body' }, null, innerText);

        return renderer.renderElementWithContent('div', {
            'class' : utils.mergeClasses(['dock', orientationClass,
                !options.hash.allowOrientation ? 'no-orientation' : '',
                !options.hash.showTitle ? 'no-title' : ''], options.hash.class),
            'aria-labelledby' : titleId
        }, utils.getElementAttributes(options.hash), resizeHandle + collapseHandle + header + body);
    };

    dock.handlebarsType = 'helper';
    dock.handlebarsName = 'dock';

    return dock;
});


define('primutils/templates/helpers/dropdown',[
    'logger',
    'handlebars',
    'lodash',
    'primutils/templates/utils/utils',
    'primutils/templates/utils/renderer',
    'primutils/templates/helpers/link'
], function (logger, Handlebars, _, utils, renderer, link) {
    'use strict';

    /**
     * Creates a divider for a dropdown menu.
     *
     * Available as: `\{{divider}}`.
     *
     * Divider is a helper that does not have inner content.
     *
     * Dividers should only be used inside of a <a href="templates.helpers.dropdownMenu.html">dropdownMenu</a>.
     *
     * @class templates.helpers.divider
     * @module templates.helpers
     * @protected
     * @constructor
     * @param [options] {Object} Handlebars options
     * @returns {string} Divider element
     */
    var divider = function (options) {
        return renderer.renderElement('li', {
            'class' : utils.mergeClasses(['divider'], options.hash.class),
            'role' : 'separator'
        }, utils.getElementAttributes(options.hash));
    };

    divider.handlebarsType = 'helper';
    divider.handlebarsName = 'divider';

    /**
     * Allows the developer to add classes to the generated `<li>` instead of the `<a>`.
     *
     * @property liClass
     * @for templates.helpers.dropdownItem
     * @type {String}
     * @default ''
     */

    /**
     * Displays the equivalent keyboard shortcut to perform the given menu action with the proper styling.
     *
     * NOTE: This does not wire up the action for you.
     * You can auto-wire actions using BPC's `data-action` attribute.
     *
     * @property shortcut
     * @for templates.helpers.dropdownItem
     * @type {String}
     * @default null
     */

    /**
     * Indicates that the dropdown item is to a file, rather than a simple navigation link.
     *
     * If this is set, will automatically use <a href="utils.fileSaver.html">fileSaver</a> to download the link, rather than navigation.
     *
     * @property data-fileDownload
     * @for templates.helpers.dropdownItem
     * @type {Boolean}
     * @default false
     */

    /**
     * Creates a menu item in a dropdown menu.
     *
     * Available as: `\{{#dropdownItem}}\{{/dropdownItem}}`.
     *
     * Dropdown item is a block helper and requires inner content.
     *
     * Dropdown items should only be used inside of a <a href="templates.helpers.dropdownMenu.html">dropdownMenu</a>.
     *
     * Additional attributes passed to the dropdownItem will be applied to the generated <a href="templates.helpers.link.html">Link `<a>`</a> and not the `<li>` with the exception of `disabled`.
     * If you pass `disabled=true`, the `<li>` will have the class 'disabled'.
     * This still does not prevent the menu item from being triggered.
     * This must still be controlled in code.
     * (There is an open issue for handling this automatically in primutils.)
     *
     * If only one arguments are passed, it is assumed to be href.
     * If you want to specify just the item's title, you should pass '#' as the first argument.
     *
     * @class templates.helpers.dropdownItem
     * @module templates.helpers
     * @protected
     * @constructor
     * @param [href] {String} The link where the menu will take the user (Default: '#' - expecting programatic changes)
     * @param [title] {String} The HTML title attribute
     * @param [options] {Object} Handlebars options
     * @returns {string} Dropdown item element
     */
    var dropdownItem = function (href, title, options) {
        if (arguments.length === 2) {
            options = title;
            title = null;
        }
        if (arguments.length === 1) {
            options = href;
            title = null;
            href = '#';
        }

        logger.assert(!!options.fn, 'Templates: Dropdown item helper must be used as a block helper.');

        var disabled = !!options.hash.disabled,
            shortcut = options.hash.shortcut,
            linkText;

        if (shortcut != null) {
            options.fn = _.wrap(options.fn, function (baseFn) {
                return '<span>' + baseFn.apply(this, _.rest(arguments)) + '</span><span class="shortcut">' + shortcut + '</span>';
            });
            options.hash.class = (options.hash.class || '') + ' with-shortcut';
        }
        linkText = link.call(this, href, title, options);

        return renderer.renderElementWithContent('li', {
            'class' : utils.mergeClasses([disabled ? 'disabled' : undefined], options.hash.liClass),
            'role' : 'menuitem'
        }, null, linkText);
    };

    dropdownItem.handlebarsType = 'helper';
    dropdownItem.handlebarsName = 'dropdownItem';

    /**
     * Causes the submenu to open to the left instead of to the right.
     *
     * @property openLeft
     * @for templates.helpers.dropdownSubmenu
     * @type {Boolean}
     * @default false
     */

    /**
     * Creates a submenu inside a dropdown menu.
     *
     * Available as: `\{{#dropdownSubmenu}}\{{/dropdownSubmenu}}`.
     *
     * Dropdown submenu item is a block helper and requires inner content.
     *
     * Dropdown submenus should only be used inside of a <a href="templates.helpers.dropdownMenu.html">dropdownMenu</a>.
     *
     * Additional attributes passed to the dropdownItem will be applied to the generated <a href="templates.helpers.link.html">Link `<a>`</a> and not the `<li>`.
     *
     * Dropdown submenus to not currently support linking to external pages.
     *
     * @class templates.helpers.dropdownSubmenu
     * @module templates.helpers
     * @protected
     * @constructor
     * @param name {String} The text that will appear on the submenu
     * @param [title] {String} The HTML title attribute
     * @param [options] {Object} Handlebars options
     * @returns {string} Dropdown submenu element
     */
    var dropdownSubmenu = function (name, title, options) {
        logger.assert(arguments.length >= 2, 'Templates: Dropdown submenu helper requires a "name" argument');

        if (arguments.length === 2) {
            options = title;
            title = null;
        }

        options.hash.id = options.hash.id || _.uniqueId('pgbu-submenu-item-');
        options.hash['aria-hasPopup'] = 'true';

        var submenuId = _.uniqueId('pgbu-dropdown-menu-');
        options.hash['aria-owns'] = submenuId;

        logger.assert(!!options.fn, 'Templates: Dropdown submenu helper must be used as a block helper.');

        var that = this,
            linkText = renderer.renderElementWithContent('div', {
                title : title,
                'class' : utils.mergeClasses(['dropdown-submenu-item'], options.hash.class),
                tabindex : 0,
                'data-toggle' : 'submenu'
            }, utils.getElementAttributes(options.hash), name),
            listText = dropdownMenu.call(this, {
                hash : {
                    'id' : submenuId,
                    'aria-labelledby' : options.hash.id
                },
                fn : function () {
                    return options.fn(that, options);
                },
                data : options.data
            }),
            liClasses = (options.hash.liClass || '') + (options.hash.openLeft ? ' pull-left' : '');

        return renderer.renderElementWithContent('li', {
            'class' : utils.mergeClasses(['dropdown-submenu'], liClasses),
            'role' : 'menuitem'
        }, null, linkText + listText);
    };

    dropdownSubmenu.handlebarsType = 'helper';
    dropdownSubmenu.handlebarsName = 'dropdownSubmenu';

    /**
     * Creates a dropdown menu element.
     * See <a href="http://pgbulwf.us.oracle.com/style-guide/?section=web&topic=Buttons&subtopic=ButtonTypes">Buttons on the Blue Sky UX Guide</a>.
     *
     * Available as: `\{{#dropdownMenu}}\{{/dropdownMenu}}`.
     *
     * Dropdown menu is a block helper and requires inner content.
     *
     * Dropdown can have several inner elements:
     *  - <a href="templates.helpers.dropdownItem.html">dropdownItem</a>
     *  - <a href="templates.helpers.dropdownSubmenu.html">dropdownSubmenu</a>
     *  - <a href="templates.helpers.divider.html">divider</a>
     *
     * Example:
     *
     *     \{{#dropdownMenu}}
     *         \{{#dropdownItem}}My Item\{{/dropdownItem}}
     *         \{{divider}}
     *         \{{#dropdownSubmenu "Submenu"}}
     *             \{{#dropdownItem}}My Subitem\{{/dropdownItem}}
     *         \{{/dropdownSubmenu}}
     *     \{{/dropdownMenu}}
     *
     * See also: <a href="templates.helpers.buttonGroup.html">\{{#buttonGroup}}</a>,
     *           <a href="templates.helpers.button.html">\{{#button}}</a>
     *
     * @class templates.helpers.dropdownMenu
     * @module templates.helpers
     * @constructor
     * @param [options] {Object} Handlebars options
     * @returns {string} Dropdown menu element
     */
    var dropdownMenu = function (options) {
        logger.assert(!!options.fn, 'Templates: Dropdown menu helper must be used as a block helper.');

        options.hash.id = options.hash.id || _.uniqueId('pgbu-dropdown-menu-');

        var buttonGroupContext = options.data && options.data.buttonGroupContext,
            tabContext = options.data && options.data.tabContext,
            frame = Handlebars.createFrame(options.data || {});

        if (buttonGroupContext) {
            buttonGroupContext.hasDropdownMenu = true;
            buttonGroupContext.menuId = options.hash.id;

            // Need to mask the menu Ids of submenus from the root menu item
            frame.buttonGroupContext = {};
        }

        if (tabContext) {
            tabContext.menuId = options.hash.id;
        }

        var innerText = options.fn(this, { data : frame });

        return renderer.renderElementWithContent('ul', {
            'class' : utils.mergeClasses(['dropdown-menu', tabContext ? 'needs-overflow' : ''], options.hash.class),
            role : 'menu',
            'aria-hidden' : 'true'
        }, utils.getElementAttributes(options.hash), innerText);
    };

    dropdownMenu.handlebarsType = 'helper';
    dropdownMenu.handlebarsName = 'dropdownMenu';

    return {
        dropdownMenu : dropdownMenu,
        dropdownSubmenu : dropdownSubmenu,
        dropdownItem : dropdownItem,
        divider : divider
    };
});


define('primutils/templates/helpers/emptyState',[
    'logger',
    'lodash',
    'primutils/templates/utils/utils',
    'primutils/templates/utils/renderer',
    'primutils/templates/helpers/icon',
    'primutils/templates/helpers/button'
], function (logger, _, utils, renderer, icon, button) {
    'use strict';

    var STATE_ICONS = {
        'error' : 'status-error',
        'warning' : 'status-warning',
        'success' : 'status-success',
        'info' : 'status-info',
        'restricted' : 'status-restricted',
        'in-progress' : 'throbber',
        'in-progress-static' : 'throbber'
    };

    var makeIcon = function (type) {
        var hash = {};
        if (type === 'in-progress-static') {
            hash.class = 'static';
        }
        return icon.call({}, STATE_ICONS[type], '', {
            hash : hash
        });
    };

    /**
     * Allow the empty state to be used statically (instead of relatively positioned) in places like modals.
     *
     * @property static
     * @for templates.helpers.emptyState
     * @type {Boolean}
     * @default false
     */

    /**
     * Sets the text that appears on the button at the bottom of the empty state.
     *
     * If this property is omitted then no button is generated.
     *
     * @property btnText
     * @for templates.helpers.emptyState
     * @type {String}
     * @optional
     * @default null
     */

    /**
     * Add additional classes to the generated button.
     *
     * This property is ignored if the `btnText` property is missing.
     *
     * @property btnClass
     * @for templates.helpers.emptyState
     * @type {String}
     * @optional
     * @default null
     */

    /**
     * The ID to give the empty state button.
     *
     * This property is ignored if the `btnText` property is missing.
     *
     * @property btnId
     * @for templates.helpers.emptyState
     * @type {String}
     * @optional
     * @default null
     */

    /**
     * Creates an empty static notification.
     *
     * Available as: `\{{#emptyState}}\{{/emptyState}}`.
     *
     * Empty state is a block helper and requires inner content.
     *
     * In order for an empty state to work properly, it must be placed inside of an element with the ".empty-state-parent" class or declare `static=true`.
     *
     * Example:
     *
     *     <div class="empty-state-parent">
     *         ... other content ...
     *
     *         \{{#emptyState 'info' 'Information' btnText="Call to Action..."}}
     *             ... Empty State Message ...
     *         \{{/emptyState}}
     *     </div>
     *
     *     \{{#emptyState 'in-progress' 'In progress' static=true}}
     *         ... Empty State Message ...
     *     \{{/emptyState}}
     *
     * @class templates.helpers.emptyState
     * @module templates.helpers
     * @constructor
     * @param type {String} The type of the empty state (info, warning, error, success, in-progress, restricted)
     * @param header {String} The header shown at the top of the empty state
     * @param [options] {Object} Handlebars options
     * @returns {string} Empty state element
     */
    var emptyState = function (type, header, options) {
        logger.assert(arguments.length === 3, 'Templates: Empty state helper requires a "type" and "header" argument');

        logger.assert(!!options.fn, 'Templates: Empty state helper must be used as a block helper.');
        logger.assert(_.includes(_.keys(STATE_ICONS), type), 'Templates: Empty state type must be one of: ' + _.keys(STATE_ICONS));

        var isStatic = (utils.hasClass(options.hash.class, 'empty-state-static')) || (!!options.hash.static);

        var headerElement = renderer.renderElementWithContent('div', {
            'class' : 'empty-state-header'
        }, null, header);

        var iconElement = makeIcon(type);

        // Now that we built the icon, use the default name
        if (type === 'in-progress-static') {
            type = 'in-progress';
        }

        var messageElement = renderer.renderElementWithContent('div', {
            'class' : 'empty-state-message'
        }, null, options.fn(this, options));

        var buttonElement = '';
        if (options.hash.btnText) {
            buttonElement = button.call({}, 'callout', '', {
                hash : {
                    'class' : utils.mergeClasses(['btn-huge'], options.hash.btnClass),
                    'id' : options.hash.btnId
                },
                fn : function () {
                    return options.hash.btnText;
                },
                data : options.data
            });
        }

        return renderer.renderElementWithContent('div', {
            'class' : utils.mergeClasses(['empty-state', type, (isStatic ? 'empty-state-static' : '')], options.hash.class),
            // Live region forces this to be read when this is added to the page --
            //   does not affect locked detail panel empty state because it started on the page invisible
            'aria-live' : 'polite',
            'aria-relevant' : 'text',
            'role' : 'alert',
            'aria-atomic' : 'true'
        }, utils.getElementAttributes(options.hash), iconElement + headerElement + messageElement + buttonElement);
    };

    emptyState.handlebarsType = 'helper';
    emptyState.handlebarsName = 'emptyState';

    return {
        'emptyState' : emptyState
    };
});


define('primutils/templates/helpers/fieldset',[
    'logger',
    'lodash',
    'handlebars',
    'primutils/templates/utils/utils',
    'primutils/templates/utils/renderer'
], function (logger, _, Handlebars, utils, renderer) {
    'use strict';

    var FIELDSET_ATTRIBUTES = ['form', 'name'],
        FIELDSET_BOOL_ATTRIBUTES = ['disabled'],
        BODY_CLASSES = 'fieldset-content fit-parent';

    /**
     * Creates the legend for a fieldset element.
     *
     * Available as: `\{{#fieldsetLegend}}\{{/fieldsetLegend}}`.
     *
     * Fieldset Legend is a block helper and requires inner content.
     *
     * The fieldsetLegend helper is generally optional.
     * It is only required when you want to create a custom legend for your fieldset that has more than just text.
     *
     * When using this helper, you must also use the `\{{#fieldsetBody}}` helper to identify the inner content of the fieldset.
     *
     * Example:
     *
     *     \{{#fieldset 'My Form Section'}}
     *         \{{#fieldsetLegend}} ... custom legend ... \{{/fieldsetLegend}}
     *         \{{#fieldsetBody}} ... my body ... \{{/fieldsetBody}}
     *     \{{/fieldset}}
     *
     * @class templates.helpers.fieldsetLegend
     * @module templates.helpers
     * @constructor
     * @param [options] {Object} Handlebars options
     * @returns {string} Fieldset legend element
     */
    var fieldsetLegend = function (options) {
        var fieldsetContext = options.data && options.data.fieldsetContext;
        logger.assert(fieldsetContext, 'Templates: fieldsetLegend helper must be called from within a {{#fieldset}}');
        logger.assert(!!options.fn, 'Templates: fieldsetLegend helper must be used as a block helper.');

        options.hash.id = options.hash.id || fieldsetContext.legendId;

        fieldsetContext.legendContent = renderer.renderElementWithContent('legend', {
            'id' : options.hash.id,
            'class' : options.hash.class,
            style : 'overflow: inherit;' // For dropdowns, etc.
        }, null, options.fn(this, options));
    };

    fieldsetLegend.handlebarsType = 'helper';
    fieldsetLegend.handlebarsName = 'fieldsetLegend';

    /**
     * Creates the body of a fieldset element.
     *
     * Available as: `\{{#fieldsetBody}}\{{/fieldsetBody}}`.
     *
     * Fieldset Body is a block helper and requires inner content.
     *
     * The fieldsetBody helper is generally optional.
     * It is only required when using the `\{{#fieldsetLegend}}` to customize the legend content.
     * It can also be used if you wish to add custom classes to the inner legend div.
     *
     * Example:
     *
     *     \{{#fieldset 'My Form Section'}}
     *         \{{#fieldsetLegend}} ... custom legend ... \{{/fieldsetLegend}}
     *         \{{#fieldsetBody}} ... my body ... \{{/fieldsetBody}}
     *     \{{/fieldset}}
     *
     * @class templates.helpers.fieldsetBody
     * @module templates.helpers
     * @constructor
     * @param [options] {Object} Handlebars options
     * @returns {string} Fieldset body element
     */
    var fieldsetBody = function (options) {
        var fieldsetContext = options.data && options.data.fieldsetContext;
        logger.assert(fieldsetContext, 'Templates: fieldsetBody helper must be called from within a {{#fieldset}}');
        logger.assert(!!options.fn, 'Templates: fieldsetBody helper must be used as a block helper.');

        fieldsetContext.bodyContent = renderer.renderElementWithContent('div', {
            'class' : BODY_CLASSES + ' ' + (options.hash.class || '')
        }, null, options.fn(this, options));
    };

    fieldsetBody.handlebarsType = 'helper';
    fieldsetBody.handlebarsName = 'fieldsetBody';

    /**
     * Specifies the message shown to the user directing them that all the inputs in the section are required.
     * This property is **not** controlled by any contained <a href="templates.helpers.controlGroup.html">\{{#controlGroup}}s</a> or <a href="templates.helpers.input.html">\{{input}}s</a>.
     *
     * @property requiredMessage
     * @for templates.helpers.fieldset
     * @type {String|null}
     * @default null
     */

    /**
     * Creates a fieldset element.
     * See <a href="http://pgbulwf.us.oracle.com/style-guide/?section=web&topic=Forms">Forms on the Blue Sky UX Guide</a>.
     *
     * Available as: `\{{#fieldset}}\{{/fieldset}}`.
     *
     * Fieldset is a block helper and requires inner content.
     *
     * Example:
     *
     *     \{{#fieldset 'My Form Section'}}
     *         ... Form Content ...
     *     \{{/fieldset}}
     *
     * You can also use any of the attributes found on <a href="http://www.w3.org/wiki/HTML/Elements/fieldset#HTML_Attributes">W3C Fieldset Attributes</a>.
     *
     * @class templates.helpers.fieldset
     * @module templates.helpers
     * @constructor
     * @param [name] {String} The label shown at the top of the fieldset
     * @param [options] {Object} Handlebars options
     * @returns {string} Fieldset element
     */
    var fieldset = function (name, options) {
        if (arguments.length === 1) {
            options = name;
            name = null;
        }

        logger.assert(!!options.fn, 'Templates: Fieldset helper must be used as a block helper.');

        options.hash.id = options.hash.id || _.uniqueId('pgbu-fieldset-');

        var classList = utils.classList(options.hash.class);

        var frame = Handlebars.createFrame(options.data || {});
        var fieldsetContext = frame.fieldsetContext = {
            legendId : _.uniqueId('pgbu-fieldset-legend-')
        };

        var innerText = options.fn(this, { data : frame });

        var legend = fieldsetContext.legendContent || renderer.renderElementWithContent('legend', {
            'id' : fieldsetContext.legendId,
            'class' : utils.hasClass(options.hash.class, 'small') ? 'small' : undefined,
            'title' : name
        }, null, name || '');

        var inner = (fieldsetContext.legendContent || fieldsetContext.bodyContent) ? (fieldsetContext.bodyContent || '') : renderer.renderElementWithContent('div', {
            'class' : BODY_CLASSES
        }, null, innerText);

        if (options.hash.requiredMessage) {
            classList.push('required');
            legend += renderer.renderElementWithContent('span', {
                'class' : 'required-note'
            }, null, options.hash.requiredMessage);
        }

        return renderer.renderElementWithContent('fieldset', {
            'class' : utils.mergeClasses(['bordered'], _.without(classList, 'small').join(' '))
        }, utils.getElementAttributes(options.hash, FIELDSET_ATTRIBUTES, FIELDSET_BOOL_ATTRIBUTES), legend + inner);
    };

    fieldset.handlebarsType = 'helper';
    fieldset.handlebarsName = 'fieldset';

    return {
        fieldset : fieldset,
        fieldsetLegend : fieldsetLegend,
        fieldsetBody : fieldsetBody
    };
});


define('primutils/templates/helpers/input',[
    'logger',
    'handlebars',
    'lodash',
    'primutils/templates/utils/utils',
    'primutils/templates/utils/renderer'
], function (logger, Handlebars, _, utils, renderer) {
    'use strict';

    //Note: 'title' is added here because it's ignored from the global list
    //      Since I don't want to do anything fancy with it now, we'll list it here

    var INPUT_ATTRIBUTES = ['accept', 'alt', 'autocomplete', 'formaction', 'formenctype', 'formmethod', 'formtarget',
            'height', 'list', 'max', 'maxlength', 'min', 'name', 'pattern', 'placeholder', 'size', 'src', 'step', 'title', 'width'],
        INPUT_BOOL_ATTRIBUTES = ['checked', 'disabled', 'formnovalidate', 'multiple', 'readonly', 'required'];

    /**
     * Sets the ID of the internal `<label>` generated for the checkboxes and radio buttons.
     * This option can only be used when the type of input is 'checkbox' or 'radio' or with the \{{checkbox}} or \{{radio}} helper.
     *
     * When set, an additional `<span>` is generated with the ID you provided.
     * This is generally the approach taken when changing the label text programatically.
     *
     *     \{{checkbox "Default Label" id="my-checkbox" labelId="my-checkbox-label}}
     *
     *     <label class="checkbox" id="pgbu-label-X" for="my-checkbox">
     *         <input type="checkbox" id="my-checkbox" aria-labelledby="pgbu-label-X" />
     *         <span id="my-checkbox-label">Default Label</span>
     *     </label>
     *
     * If you need to actually programatically adjust the actual `<label>`, you do a select on the parent of the generated input, e.g.
     *
     *     $('#my-checkbox').parent()
     *
     * @property labelId
     * @for templates.helpers.input
     * @type {String|null}
     * @default null
     */

    /**
     * Places a small description below the input.
     * See <a href="http://pgbulwf.us.oracle.com/style-guide/?section=web&topic=Forms&subtopic=Hints">Hint text on the Blue Sky UX Guide</a>.
     *
     * @property hint
     * @for templates.helpers.input
     * @type {String|null}
     * @default null
     */

    /**
     * Specifies the ID attribute of the generated hint text.
     * Used in cases where the page author needs to access the hint text once its created.
     *
     * If not specified, it will receive an auto-incrementing ID so that `<input>`s can take advantage of it for accessibility purposes.
     *
     * @property hintId
     * @for templates.helpers.input
     * @type {String}
     * @default Auto-incrementing value
     */

    /**
     * Hides the hint text by default using the Bootstrap class `hidden`.
     * Useful for error messages.
     *
     * Change the visibility by using:
     *
     *     $hint.toggleClass('hidden', !shouldShow);
     *
     * See also <a href="#property_hintNotDisplayed">hintNotDisplayed</a>.
     * It is *not* recommended to use both hintHidden and hintNotDisplayed.
     *
     * @property hintHidden
     * @for templates.helpers.input
     * @type {Boolean}
     * @default false
     */

    /**
     * Hides the hint text by default using `style="display:none;"`.
     * Useful for error messages.
     *
     * Change the visibility by using:
     *
     *     $hint.toggle(shouldShow);
     *
     * See also <a href="#property_hintNotDisplayed">hintNotDisplayed</a>.
     * It is *not* recommended to use both hintHidden and hintNotDisplayed.
     *
     * @property hintNotDisplayed
     * @for templates.helpers.input
     * @type {Boolean}
     * @default false
     */

    /**
     * Places the icon inside a radio button or checkbox input to the left of the input instead of the right.
     *
     * This only applies to checkbox and radio button inputs. This overrides the value from the control group if one was passed in.
     *
     * @property leftIcon
     * @for templates.helpers.input
     * @type {Boolean}
     * @default false
     */

    /**
     * Adds a class to the checkbox or radio button's surrounding label.
     *
     * This only applies to checkbox or radio button helpers.
     *
     * @property labelClass
     * @for templates.helpers.input
     * @type {String|null}
     * @default null
     */

    /**
     * Creates an input element.
     * See <a href="http://pgbulwf.us.oracle.com/style-guide/?section=web&topic=InputTypes">Inputs on the Blue Sky UX Guide</a>.
     * **Note:** This helper does not automatically create <a href="http://10.140.183.183:8080/view/Inputs%20(OUI)/job/oui/javadoc/">OUI elements</a>.
     * You will still need to do that through javascript.
     *
     * Available as: `\{{input}}`.
     *
     * Input can be used as either a regular helper or a block helper.
     *
     * Examples:
     *
     *     \{{input 'text' 'Default Value'}}
     *
     *     \{{input 'text' value="Default Value"}}
     *
     *     \{{#input 'text'}}Default Value\{{/input}}
     *
     * You can use any of the <a href="http://www.w3.org/wiki/HTML/Elements/input">HTML 5 input types</a>.
     * You can also specify `'textarea'` to create a text area.
     *
     * You can also use any of the attributes found on <a href="http://www.w3.org/wiki/HTML/Elements/input#HTML_Attributes">W3C Input Attributes</a>.
     * The helper is not smart enough to determine which attributes are valid for which types.
     * Use other attributes at your own peril.
     *
     * Checkboxes and radio buttons, in addition to specifying them with the type parameter, can be created with:
     *
     *     \{{checkbox 'Checkbox Label' 'checkbox_value'}}
     *     \{{checkbox 'Checkbox Label'}}checkbox_value\{{/checkbox}}
     *     \{{radio 'Radio Label'}}
     *
     * If created using the `\{{input 'checkbox'}}` syntax, you must specify the input's value explicitly
     *
     *     \{{#input 'radio' 'radio_value'}}Radio Label\{{/input}}
     *
     * **Important**:
     * If you attempt to hide a checkbox or a radio button using either the `hidden="hidden"` attribute or with
     * `class="hidden"`, neither or these will be applied to the `input`. Instead, a `class="hidden"` will be applied
     * to the surrounding label. This means that when you later want to unhide the checkbox or radio button, you
     * must remove the `hidden` class from that label instead.
     *
     * Generally, this can be accomplished with:
     *
     *     $('#myRadioButton').closest('.radio').toggleClass('hidden');
     *     $('#myCheckbox').closest('.checkbox').toggleClass('hidden');
     *
     * Textareas can also be created using:
     *
     *     \{{#textarea}}Value\{{/textarea}}
     *
     * See also: <a href="templates.helpers.controlGroup.html">\{{#controlGroup}}</a>
     *
     * @class templates.helpers.input
     * @module templates.helpers
     * @constructor
     * @param type {String} The type of the input
     * @param [value] {String} The default value of the input (Default: empty)
     * @param [options] {Object} Handlebars options
     * @returns {string} Button element
     */
    var input = function (type, value, options) {
        logger.assert(arguments.length >= 2, 'Templates: Input helper requires a "type" argument');

        if (arguments.length === 2) {
            options = value;
            value = null;
        }

        if (_.includes(['radio', 'checkbox'], type) && !options.hash._receivedLabel) {
            var frame = Handlebars.createFrame(options.data || {}),
                opts = Handlebars.Utils.extend({}, options, { data : frame });
            frame.inputRadioCheckboxContext = {};
            return checkboxRadioBase.call(this, type, options.fn(this, opts), value, opts);
        }

        options.hash.id = options.hash.id || _.uniqueId('pgbu-input-');

        // Prefer inner text over specified value
        value = (options.fn && (!options.hash._receivedLabel || value == null)) ? options.fn(this, options) : value;

        // See controlGroup.js helper for details on the controlGroupContext
        var controlGroupContext = options.data && options.data.controlGroupContext;
        if (controlGroupContext) {
            options.hash['aria-labelledby'] = controlGroupContext.labelId;
            options.hash.name = options.hash.name || controlGroupContext.name;
            if (options.hash.id) {
                controlGroupContext.inputs.push(options.hash.id);
            }
            if (options.hash.required) {
                controlGroupContext.inputRequired = controlGroupContext.inputRequired || !!options.hash.required;
            }
            if (type === 'radio') {
                controlGroupContext.isRadioGroup = true;
            }
        }

        var fieldsetContext = options.data && options.data.fieldsetContext;
        if (fieldsetContext) {
            options.hash['aria-labelledby'] = fieldsetContext.legendId + ' ' + (options.hash['aria-labelledby'] || '');
        }

        var input,
            hint = '',
            hintId = options.hash.hintId || _.uniqueId('pgbu-hint-text-');

        if (options.hash.hint != null) {
            hint = renderer.renderElementWithContent('span', {
                'class' : 'help-block' + (options.hash.hintHidden ? ' hidden' : ''),
                'style' : options.hash.hintNotDisplayed ? 'display: none;' : undefined,
                id : hintId
            }, null, options.hash.hint);
        }

        if (type !== 'textarea') {
            input = renderer.renderSelfClosingElement('input', {
                type : type,
                value : value,
                'class' : utils.mergeClasses([], options.hash.class), // Unnecessary initially, but if we ever have default classes it's ready
                'aria-required' : (!!options.hash.required).toString(),
                'aria-describedby' : (options.hash.hint != null) ? hintId : undefined,
                'role' : (type === 'radio') ? 'radio' : undefined
            }, utils.getElementAttributes(options.hash, INPUT_ATTRIBUTES, INPUT_BOOL_ATTRIBUTES));
        } else {
            input = renderer.renderElementWithContent('textarea', {
                'class' : utils.mergeClasses([], options.hash.class), // Unnecessary initially, but if we ever have default classes it's ready
                'aria-required' : (!!options.hash.required).toString(),
                'aria-describedby' : (options.hash.hint != null) ? hintId : undefined
            }, utils.getElementAttributes(options.hash, _.union(INPUT_ATTRIBUTES, ['rows', 'cols']), INPUT_BOOL_ATTRIBUTES), value ? value : '');
        }

        return input + hint;
    };

    input.handlebarsType = 'helper';
    input.handlebarsName = 'input';

    var checkboxRadioBase = function (type, label, value, options) {
        logger.assert(arguments.length >= 3, 'Templates: Checkbox/Radio helper requires a "label" argument');

        if (arguments.length === 3) {
            options = value;
            value = null;
        }

        var hasLabelId = !!options.hash.labelId;
        options.hash.outerLabelId = _.uniqueId('pgbu-label-');
        options.hash.labelId = options.hash.labelId || _.uniqueId('pgbu-label-');

        if (hasLabelId) {
            label = renderer.renderElementWithContent('span', {
                id : options.hash.labelId
            }, null, label);
        }

        var controlGroupContext = options.data && options.data.controlGroupContext;
        var frame = Handlebars.createFrame(options.data || {});
        frame.inputRadioCheckboxContext = frame.inputRadioCheckboxContext || {};
        frame.controlGroupContext = { inputs : [],
            labelId : (controlGroupContext ? controlGroupContext.labelId + ' ' : '') + options.hash.outerLabelId,
            inputRequired : false,
            name : controlGroupContext && controlGroupContext.name,
            isCheckboxRadio : true
        };

        options.hash = _.omit(options.hash, 'hint');
        options.hash._receivedLabel = true;

        var inputText = input.call(this, type, value, Handlebars.Utils.extend({}, options, { data : frame })),
            inputIsHidden = utils.hasClass(options.hash.class, 'hidden') || !!options.hash.hidden,
            leftIcons = (!!options.hash.leftIcon) || (controlGroupContext && controlGroupContext.leftIcons),
            leftIconsClass = (frame.inputRadioCheckboxContext.iconSize === 'pgbu-icon-large' ? 'left-icon-large' : 'left-icon');

        // Remove the hidden-y things
        options.hash.class = utils.mergeClasses([_.without(utils.classList(options.hash.class), 'hidden')]);
        options.hash.hidden = false;

        var result = renderer.renderElementWithContent('label', {
            'class' : type + (options.hash.disabled ? ' disabled ' : '') + (inputIsHidden ? ' hidden ' : '') + (leftIcons ? ' ' + leftIconsClass + ' ' : '') + ' ' + (options.hash.labelClass || ''),
            id : options.hash.outerLabelId,
            for : options.hash.id
        }, null, inputText + label);

        if (controlGroupContext) {
            controlGroupContext.isRadioGroup = frame.controlGroupContext.isRadioGroup;
        }

        return result;
    };

    if (require.nodeRequire && !_) {
        _ = require.nodeRequire('lodash');
    }

    var checkbox = _.partial(checkboxRadioBase, 'checkbox');

    checkbox.handlebarsType = 'helper';
    checkbox.handlebarsName = 'checkbox';

    var radio = _.partial(checkboxRadioBase, 'radio');

    radio.handlebarsType = 'helper';
    radio.handlebarsName = 'radio';

    var textarea = _.partial(input, 'textarea');

    textarea.handlebarsType = 'helper';
    textarea.handlebarsName = 'textarea';

    return {
        input : input,
        checkbox : checkbox,
        radio : radio,
        textarea : textarea
    };
});


define('primutils/templates/helpers/join',['lodash'], function (_) {
    'use strict';

    /**
     * Concatenates variables and string expressions together when needed in hashes.
     *
     * Example:
     *
     *     \{{#button 'toolbar' class=(_ myVariableClass " myOtherClass")}}Button\{{/button}}
     *
     * No spaces are inserted into the concatenation.
     * If you need to separate strings, insert an empty string between the two variables
     *
     *     \{{#button 'toolbar' class=(_ myVariableClass " " myOtherVariable)}}Button\{{/button}}
     *
     * @class templates.helpers._
     * @module templates.helpers
     * @constructor
     * @param ... {Array} Strings to join
     * @param [options] {Object} Handlebars options
     * @returns {string} Joined string
     */
    var join = function () {
        return _.initial(_.toArray(arguments)).join('');
    };

    join.handlebarsType = 'helper';
    join.handlebarsName = '_';

    return join;
});

define('primutils/templates/helpers/progress',['primutils/templates/utils/renderer', 'primutils/templates/utils/utils'], function (renderer, utils) {
    'use strict';

    /**
     * Creates a progress bar element.
     * See <a href="https://confluence.oraclecorp.com/confluence/display/PrimaveraUX/Progress+Bar">Progress Bars on the Blue Sky UX Guide</a>.
     *
     * Available as: `\{{progress}}`.
     *
     * Progress is a helper that does not have inner content.
     *
     * This does **not** create a `<progress>` element.
     * See the Style Guide.
     *
     * Example:
     *
     *     \{{progress initialPercent=0.4}}
     *     \{{progress indeterminate=true}}
     *
     * You can either specify an `initialPercent` or `indeterminate` to control the progress bar.
     * If you omit both, the progress bar will be determinate and the percent will be 0%.
     *
     * @class templates.helpers.progress
     * @module templates.helpers
     * @constructor
     * @param [options] {Object} Handlebars options
     * @returns {string} Progress element
     */
    var progress = function (options) {
        return renderer.renderElementWithContent('div', {
            'class' : utils.mergeClasses(['progress', options.hash.indeterminate ? 'indeterminate' : ''], options.hash.class)
        }, utils.getElementAttributes(options.hash), renderer.renderElement('div', {
            'class' : 'bar',
            'style' : 'width: ' + ((options.hash.initialPercent || 0) * 100) + '%'
        }, null));
    };

    progress.handlebarsType = 'helper';
    progress.handlebarsName = 'progress';

    return progress;
});

define('primutils/templates/helpers/kpi',[
    'logger',
    'primutils/templates/helpers/card',
    'primutils/templates/helpers/icon',
    'primutils/templates/helpers/progress',
    'primutils/templates/utils/utils',
    'primutils/templates/utils/renderer'
], function (logger, card, icon, progress, utils, renderer) {
    'use strict';

    /**
     * Apply a class to the stat element.
     * Often used if the value is not known at template runtime.
     *
     * @property statClass
     * @for templates.helpers.kpi
     * @type {String|null}
     * @default null
     */

    /**
     * Apply a class to the label element.
     * Often used if the value is not known at template runtime.
     *
     * @property labelClass
     * @for templates.helpers.kpi
     * @type {String|null}
     * @default null
     */

    /**
     * Apply an ID to the stat element.
     * Often used if the value is not known at template runtime.
     *
     * @property statId
     * @for templates.helpers.kpi
     * @type {String|null}
     * @default null
     */

    /**
     * Apply an ID to the label element.
     * Often used if the value is not known at template runtime.
     *
     * @property labelId
     * @for templates.helpers.kpi
     * @type {String|null}
     * @default null
     */

    /**
     * Creates a KPI stat element.
     *
     * Available as: `\{{kpi}}`.
     *
     * Caret is a helper that does not have inner content.
     *
     * Example:
     *
     *     \{{kpi '76' 'Trombones' 'ent-project'}}
     *
     * @class templates.helpers.kpi
     * @module templates.helpers
     * @constructor
     * @param stat {String} The statistic to show
     * @param label {String} The label below the statistic
     * @param [iconClass] {String} The icon used in the KPI
     * @param [options] {Object} Handlebars options
     * @returns {string} KPI stat element
     */
    var kpi = function (stat, label, iconClass, options) {
        logger.assert(arguments.length >= 3, 'Templates: KPI helper requires stat and label arguments');

        if (arguments.length === 3) {
            options = iconClass;
            iconClass = null;
        }

        var iconElem = iconClass != null ? icon.call(this, iconClass, '', {
            hash : {
                size : 'kpi-icon'
            },
            data : {}
        }) : '';
        var statElem = renderer.renderElementWithContent('span', {
            'class' : utils.mergeClasses(['kpi-stat'], options.hash.statClass),
            id : options.hash.statId
        }, null, stat);
        var labelElem = renderer.renderElementWithContent('span', {
            'class' : utils.mergeClasses(['kpi-label'], options.hash.labelClass),
            id : options.hash.labelId
        }, null, label);

        return renderer.renderElementWithContent('div', {
            'class' : utils.mergeClasses(['kpi', options.hash.class])
        }, utils.getElementAttributes(options.hash), iconElem + statElem + labelElem);
    };

    kpi.handlebarsType = 'helper';
    kpi.handlebarsName = 'kpi';

    var kpiInProgress = function (formattedCompleted, formattedTotal, percentComplete, label, options) {
        var formattedTotalElem = renderer.renderElementWithContent('span', { class : 'kpi-total' }, null, formattedTotal);
        var topLabel = renderer.renderElementWithContent('div', {
            'class' : 'kpi-progress-info'
        }, null, renderer.renderElementWithContent('span', { class : 'kpi-completed' }, null, formattedCompleted) + ' / ' + formattedTotalElem);
        var progressBar = progress.call(this, {
            hash : { initialPercent : percentComplete }
        });
        var bottomLabel = renderer.renderElementWithContent('div', {
            'class' : 'kpi-label'
        }, null, renderer.renderElementWithContent('span', { class : 'kpi-name' }, null, label) + ' - ' + formattedTotalElem);

        return renderer.renderElementWithContent('div', {
            'class' : utils.mergeClasses(['kpi-in-progress', options.hash.class])
        }, utils.getElementAttributes(options.hash), topLabel + progressBar + bottomLabel);
    };

    kpiInProgress.handlebarsType = 'helper';
    kpiInProgress.handlebarsName = 'kpiInProgress';

    return {
        kpi : kpi,
        kpiInProgress : kpiInProgress
    };
});


define('primutils/templates/helpers/modal',[
    'logger',
    'handlebars',
    'lodash',
    'primutils/templates/utils/utils',
    'primutils/templates/utils/renderer',
    'primutils/templates/helpers/link',
    'primutils/templates/helpers/icon',
    'primutils/templates/helpers/button',
    'primutils/utils/language'
], function (logger, Handlebars, _, utils, renderer, link, icon, button, lang) {
    'use strict';

        if (require.nodeRequire && !_) {
        _ = require.nodeRequire('lodash');
    }
    
    function factory(baseClass) {
        var modalCloseButton = function () {
            return link.call(this, '#', lang.getString('dialogs.close'), {
                hash : {
                    'class' : 'modal-close', // Should always be modal-close, not overridden by Blue Sky
                    'data-dismiss' : 'modal' // Should always be modal, from Bootstrap interactions
                },
                fn : function () {
                    return icon.call(this, 'dialog-close', lang.getString('dialogs.close'), { hash : {} });
                }
            });
        };

        /**
         * Hides the close button on the modal.
         *
         * Make sure that users have other ways of dismissing the modal: buttons in the footer or setting keyboard=true on the modal itself.
         *
         * @property hideClose
         * @for templates.helpers.modalControlBar
         * @type {Boolean}
         * @default false
         */

        /**
         * Sets the control bar in the modal header.
         *
         * Available as: `\{{#modalControlBar}}\{{/modalControlBar}}`.
         *
         * Modal control bar is a block helper.
         *
         * Modal header can only be used inside of an <a href="templates.helpers.modalHeader.html">modalHeader</a>.
         *
         * Unless you specify `hideClose` at either the \{{#modalHeader}} or \{{#modalControlBar}} level, the close button will be added for you automatically.
         *
         * @class templates.helpers.modalControlBar
         * @module templates.helpers
         * @protected
         * @constructor
         * @param [options] {Object} Handlebars options
         * @returns {string} Modal control bar element
         */
        var modalControlBar = function (options) {
            var modalContext = options.data && options.data.modalContext;
            logger.assert(modalContext, 'Templates: Control bar helper must be called from within a {{#' + baseClass + '}}');
            logger.assert(modalContext.headerContext, 'Templates: Control bar helper must be called from within a {{#' + baseClass + 'Header}}');
            logger.assert(!!options.fn, 'Templates: Control bar helper must be used as a block helper.');
            modalContext.headerContext.hasControlBar = true;
            var hideClose = modalContext.headerContext.hideClose = modalContext.headerContext.hideClose || !!options.hash.hideClose;

            var innerContent = options.fn(this, options);
            if (!hideClose) {
                innerContent += modalCloseButton.call(this);
            }

            return renderer.renderElementWithContent('div', {
                'class' : utils.mergeClasses(['control-bar'], options.hash.class)
            }, utils.getElementAttributes(options.hash), innerContent);
        };

        modalControlBar.handlebarsType = 'helper';
        modalControlBar.handlebarsName = baseClass + 'ControlBar';

        /**
         * Hides the close button on the modal.
         *
         * Make sure that users have other ways of dismissing the modal: buttons in the footer or setting keyboard=true on the modal itself.
         *
         * @property hideClose
         * @for templates.helpers.modalHeader
         * @type {Boolean}
         * @default false
         */

        /**
         * Hides the title on the modal.
         *
         * @property hideTitle
         * @for templates.helpers.modalHeader
         * @type {Boolean}
         * @default false
         */

        /**
         * Creates a modal header.
         *
         * Available as: `\{{#modalHeader}}\{{/modalHeader}}`.
         *
         * Modal header can be used as either a regular helper or a block helper.
         *
         * Modal header can only be used inside of an <a href="templates.helpers.modal.html">modal</a>.
         *
         * As a non-block helper, the modal header will insert the title of the modal unless `hideTitle` is set.
         *
         * As a block helper, any content passed inside the block will be rendered *after* the title.
         *
         * @class templates.helpers.modalHeader
         * @module templates.helpers
         * @protected
         * @constructor
         * @param [options] {Object} Handlebars options
         * @returns {string} Modal header element
         */
        var modalHeader = function (options) {
            var modalContext = options.data && options.data.modalContext;
            logger.assert(modalContext, 'Templates: Header must be called from within a {{#' + baseClass + '}}');
            modalContext.hasHeader = true;

            modalContext.headerContext = {
                hideClose : !!options.hash.hideClose,
                hideTitle : !!options.hash.hideTitle,
                hasControlBar : false
            };

            var innerContent = '',
                fnContent = '';

            if (!modalContext.headerContext.hideTitle) {
                innerContent += renderer.renderElementWithContent('h3', {
                    id : modalContext.labeledBy,
                    'class' : 'modal-title' // Should always be modal-title, not overridden by Blue Sky
                }, null, modalContext.title);
            }

            if (options.fn) {
                fnContent = options.fn(this, options);
            }

            if (!modalContext.headerContext.hasControlBar && !modalContext.headerContext.hideClose) {
                innerContent += modalCloseButton.call(this);
            }

            innerContent += fnContent;

            return renderer.renderElementWithContent('div', {
                'class' : utils.mergeClasses([baseClass + '-header'], options.hash.class)
            }, utils.getElementAttributes(options.hash), innerContent);
        };

        modalHeader.handlebarsType = 'helper';
        modalHeader.handlebarsName = baseClass + 'Header';

        /**
         * Creates a modal body.
         *
         * Available as: `\{{#modalBody}}\{{/modalBody}}`.
         *
         * Modal body is a block helper and requires inner content.
         *
         * Modal body can only be used inside of an <a href="templates.helpers.modal.html">modal</a>.
         *
         * @class templates.helpers.modalBody
         * @module templates.helpers
         * @protected
         * @constructor
         * @param [options] {Object} Handlebars options
         * @returns {string} Modal body element
         */
        var modalBody = function (options) {
            var modalContext = options.data && options.data.modalContext;
            logger.assert(modalContext, 'Templates: Header must be called from within a {{#' + baseClass + '}}');
            logger.assert(!!options.fn, 'Templates: Body helper must be used as a block helper.');

            modalContext.hasBody = true;

            return renderer.renderElementWithContent('div', {
                'class' : utils.mergeClasses([baseClass + '-body'], options.hash.class)
            }, utils.getElementAttributes(options.hash), options.fn(this, options));
        };

        modalBody.handlebarsType = 'helper';
        modalBody.handlebarsName = baseClass + 'Body';

        /**
         * Specifies whether the auto generated buttons should come before or after the inner block content.
         *
         * @property buttonsLast
         * @for templates.helpers.modalFooter
         * @type {Boolean}
         * @default false
         */

        /**
         * Creates a modal footer.
         *
         * Available as: `\{{#modalFooter}}\{{/modalFooter}}`.
         *
         * Modal footer can be used as either a regular helper or a block helper.
         *
         * Modal footer can only be used inside of an <a href="templates.helpers.modal.html">modal</a>.
         *
         * You can either specify your own content, placing it inside of the block helper, or
         * you can use the default buttons parameter to create dummy buttons in the footer that will close the modal.
         *
         * Valid button names: 'ok', 'cancel'
         *
         * @class templates.helpers.modalFooter
         * @module templates.helpers
         * @protected
         * @constructor
         * @param [defaultButtons] {String} A comma delimited string of buttons to create in the footer
         * @param [options] {Object} Handlebars options
         * @returns {string} Modal footer element
         */
        var modalFooter = function (defaultButtons, options) {
            var that = this,
                buttonText = '',
                innerText = '';

            if (arguments.length === 1) {
                options = defaultButtons;
                defaultButtons = '';
            }

            var modalContext = options.data && options.data.modalContext;
            logger.assert(modalContext, 'Templates: Footer must be called from within a {{#' + baseClass + '}}');
            modalContext.hasFooter = true;

            _.forEach(defaultButtons.split(','), function (buttonName) {
                buttonName = buttonName.trim();
                switch (buttonName) {
                case 'ok':
                    buttonText += button.call(that, 'primary', lang.getString('dialogs.ok'), {
                        hash : {
                            'data-dismiss' : 'modal' // Should always be modal, from Bootstrap interactions
                        },
                        fn : function () {
                            return lang.getString('dialogs.ok');
                        },
                        data : options.data
                    });
                    break;
                case 'cancel':
                    buttonText += button.call(that, 'secondary', lang.getString('dialogs.cancel'), {
                        hash : {
                            'data-dismiss' : 'modal' // Should always be modal, from Bootstrap interactions
                        },
                        fn : function () {
                            return lang.getString('dialogs.cancel');
                        },
                        data : options.data
                    });
                    break;
                }
            });

            if (options.fn) {
                if (options.hash.buttonsLast) {
                    innerText += options.fn(this, options) + buttonText;
                } else {
                    innerText += buttonText + options.fn(this, options);
                }
            } else {
                innerText = buttonText;
            }

            if (options.hash.omit === true) {
                return '';
            } else {
                return renderer.renderElementWithContent('div', {
                    'class' : utils.mergeClasses([baseClass + '-footer'], options.hash.class)
                }, utils.getElementAttributes(options.hash), innerText);
            }
        };

        modalFooter.handlebarsType = 'helper';
        modalFooter.handlebarsName = baseClass + 'Footer';

        /**
         * Let's assistive applications know that this dialog does not have any user editable content and is solely a displaying an alert.
         * Most of the shared dialogs (with the exception of ConfirmDialog) would be alert=true.
         *
         * @property alert
         * @for templates.helpers.modal
         * @type {Boolean}
         * @default false
         */

        /**
         * Controls how the backdrop behaves with the modal.
         * See <a href="http://getbootstrap.com/2.3.2/javascript.html#modals">Bootstrap modals</a>.
         *
         * Valid values: true, false, 'static'
         *
         * @property backdrop
         * @for templates.helpers.modal
         * @type {String|Boolean}
         * @default 'static'
         */

        /**
         * Controls whether escape can be used to close the modal
         * See <a href="http://getbootstrap.com/2.3.2/javascript.html#modals">Bootstrap modals</a>.
         *
         * Valid values: true, false
         *
         * @property keyboard
         * @for templates.helpers.modal
         * @type {Boolean}
         * @default true
         */

        /**
         * Creates a modal element.
         * See <a href="http://pgbulwf.us.oracle.com/style-guide/?section=web&topic=Dialogs">Dialogs on the Blue Sky UX Guide</a>.
         *
         * Available as: `\{{#modal}}\{{/modal}}`.
         *
         * Modal is a block helper and requires inner content.
         *
         * Modals should have several inner elements:
         *  - <a href="templates.helpers.modalHeader.html">modalHeader</a>
         *  - <a href="templates.helpers.modalBody.html">modalBody</a>
         *  - <a href="templates.helpers.modalFooter.html">modalFooter</a>
         *
         * Example:
         *
         *     \{{#modal 'My Sample Modal' alert=true}}
         *         \{{#modalHeader}}
         *             Additional Header Content
         *         \{{/modalHeader}}
         *         \{{#modalBody}}
         *             ... content ...
         *         \{{/modalBody}}
         *         \{{#modalFooter}}
         *             Footer content
         *         \{{/modalFooter}}
         *     \{{/modal}}
         *
         * If a modalHeader or modalFooter section is skipped inside the modal block, the default header and footer will be inserted automatically.
         * If no modalHeader, modalBody, or modalFooter section is found, the inner content of the modal will automatically be placed inside a modalBody section, and
         * the default header and footers appended automatically.
         * The default header includes a title and a close button.
         * The default footer includes an "Ok" button.
         *
         * @class templates.helpers.modal
         * @module templates.helpers
         * @constructor
         * @param title {String} The title of the modal
         * @param [options] {Object} Handlebars options
         * @returns {string} Modal element
         */
        var modal = function (title, options) {
            if (arguments.length === 1) {
                options = title;
                title = '';
            }

            logger.assert(!!options.fn, 'Templates: ' + baseClass + ' helper must be used as a block helper.');

            if (!options.hash.id) {
                options.hash.id = _.uniqueId('pgbu-' + baseClass + '-');
            }

            var frame = Handlebars.createFrame(options.data || {});
            var isModal = baseClass === 'modal';
            var modalContext = frame.modalContext = {
                id : options.hash.id,
                isModal : isModal,
                labeledBy : options.hash.id + '-label',
                type : options.hash.alert ? 'alertdialog' : 'dialog',
                title : title
            };

            var innerText = options.fn ? options.fn(this, { data : frame }) : '',
                bodyConverted = false;

            // If there's nothing from the context, assume that the whole thing is for the body
            if (!modalContext.hasHeader && !modalContext.hasFooter && !modalContext.hasBody) {
                bodyConverted = true;
                innerText = modalBody.call(this, {
                    hash : {},
                    fn : function () {
                        return innerText;
                    },
                    data : frame
                });
            }

            if (!modalContext.hasHeader) {
                logger.info('Templates: Template did not specify a ' + baseClass + '-header, adding default header');
                innerText = modalHeader.call(this, { hash : {}, data : frame }) + innerText;
            }

            if (!modalContext.hasFooter) {
                logger.info('Templates: Template did not specify a ' + baseClass + '-footer, adding default footer');
                innerText = innerText + modalFooter.call(this, bodyConverted ? 'ok' : '', { hash : {}, data : frame });
            }

            return renderer.renderElementWithContent('div', {
                'class' : utils.mergeClasses([baseClass, 'pgbu-' + baseClass, 'hide', 'fade'], options.hash.class),
                tabindex : '-1',
                role : isModal ? modalContext.type : undefined,
                'aria-hidden' : 'true',
                'aria-labelledby' : modalContext.labeledBy,
                'data-backdrop' : _.has(options.hash, 'backdrop') ? options.hash.backdrop.toString() : (isModal ? 'static' : undefined),
                'data-keyboard' : _.has(options.hash, 'keyboard') ? options.hash.keyboard.toString() : (isModal ? 'true' : undefined)
            }, utils.getElementAttributes(options.hash), innerText);
        };

        modal.handlebarsType = 'helper';
        modal.handlebarsName = baseClass;

        var result = {};
        result[baseClass + 'Header'] = modalHeader;
        result[baseClass + 'ControlBar'] = modalControlBar;
        result[baseClass + 'Body'] = modalBody;
        result[baseClass + 'Footer'] = modalFooter;
        result[baseClass] = modal;

        return result;
    }

    return _.assign({}, factory('modal'), factory('tray'));
});


define('primutils/templates/helpers/panel',[
    'logger',
    'lodash',
    'handlebars',
    'primutils/templates/utils/utils',
    'primutils/templates/utils/renderer',
    'primutils/templates/helpers/link',
    'primutils/templates/helpers/button',
    'primutils/templates/helpers/icon',
    'primutils/templates/helpers/caret',
    'primutils/templates/helpers/buttonGroup',
    'primutils/templates/helpers/dropdown',
    'primutils/utils/language'
], function (logger, _, Handlebars, utils, renderer, link, button, icon, caret, buttonGroup, dropdown, lang) {
    'use strict';

    var PANEL_TYPE_MAP = {
            'detail' : 'nav nav-tabs detail-panel-tabs',
            'page' : 'nav nav-tabs with-notch',
            'content' : 'nav nav-tabs content-tabs',
            'scenario' : 'nav nav-tabs scenario-tabs',
            'wizard' : 'steps'
        },
        PANEL_TYPE_NO_NAV_MAP = {
            'detail' : 'detail-panel-tabs',
            'page' : 'with-notch',
            'content' : 'content-tabs',
            'scenario' : 'scenario-tabs',
            'wizard' : 'steps'
        },
        PANEL_PANELS_MAP = {
            'detail' : 'detail-panel',
            'page' : 'page-panel',
            'content' : 'content-panel',
            'scenario' : 'scenario-panel',
            'wizard' : 'wizard-panel'
        };

    /**
     * Creates a panel.
     *
     * Available as: `\{{#panel}}\{{/panel}}`.
     *
     * Panel is a block helper and requires inner content.
     *
     * Panel can only be used inside of an <a href="templates.helpers.panels.html">panels</a>.
     *
     * If the ID parameter is omitted the panel element will not function with the <a href="http://getbootstrap.com/2.3.2/javascript.html#tabs">Bootstrap tab widget</a>.
     *
     * @class templates.helpers.panel
     * @module templates.helpers
     * @protected
     * @constructor
     * @param [id] {String} The HTML id of the panel
     * @param [options] {Object} Handlebars options
     * @returns {string} The panel element
     */
    var panel = function (id, options) {
        var active = '';

        if (arguments.length === 1) {
            options = id;
            id = null;
            logger.warn('Panel was not given an ID. It may never be visible.');
        }

        var panelContext = options.data && options.data.panelContext;
        logger.assert(panelContext, 'Templates: Tab helper must be called from within a {{#panels}}');
        logger.assert(!!options.fn, 'Templates: Panel helper must be used as a block helper.');

        if (id) {
            panelContext.panelsCreated.push(id);

            if (panelContext.activeTab === id) {
                active = 'active';
            }
        }

        if (!panelContext.noSizePanels) {
            if (!options.hash.style) {
                options.hash.style = 'height: 100%;';
            } else {
                options.hash.style = options.hash.style + '; height: 100%;';
            }
        }

        return renderer.renderElementWithContent('div', {
            'id' : id,
            'class' : utils.mergeClasses(['tab-pane', active], options.hash.class),
            'role' : 'tabpanel'
        }, utils.getElementAttributes(options.hash), options.fn(this, options));
    };

    panel.handlebarsType = 'helper';
    panel.handlebarsName = 'panel';

    /**
     * Creates a collection of panels.
     *
     * Available as: `\{{#panels}}\{{/panels}}`.
     *
     * Panels is a block helper and requires inner content.
     *
     * Panels can only be used inside of an <a href="templates.helpers.panelContainer.html">panelContainer</a>.
     *
     * Panels should contain several <a href="templates.helpers.panel.html">panel</a> elements.
     *
     * **Note:** The panels section should be created *after* the tabList.
     *
     * @class templates.helpers.panels
     * @module templates.helpers
     * @protected
     * @constructor
     * @param [options] {Object} Handlebars options
     * @returns {string} Collection of panels element
     */
    var panels = function (options) {
        var panelContext = options.data && options.data.panelContext;
        logger.assert(panelContext, 'Templates: Tab helper must be called from within a {{#panelContainer}}');
        logger.assert(!!options.fn, 'Templates: Panels helper must be used as a block helper.');

        return renderer.renderElementWithContent('div', {
            'class' : utils.mergeClasses(['tab-content', PANEL_PANELS_MAP[panelContext.type]], options.hash.class)
        }, utils.getElementAttributes(options.hash), options.fn(this, options));
    };

    panels.handlebarsType = 'helper';
    panels.handlebarsName = 'panels';

    /**
     * Creates a menu button for a tab.
     *
     * Available as: `\{{tabMenuButton}}`.
     *
     * Tab menu button is a helper that does not have inner content.
     *
     * Tab menu button can only be used inside of an <a href="templates.helpers.tab.html">tab</a>.
     *
     * Menus are only supported in 'content' and 'scenario' tabs.
     *
     * @class templates.helpers.tabMenuButton
     * @module templates.helpers
     * @protected
     * @constructor
     * @param [options] {Object} Handlebars options
     * @returns {string} Tab menu button element
     */
    var tabMenuButton = function (options) {
        var panelContext = options.data && options.data.panelContext,
            tabContext = options.data && options.data.tabContext;
        logger.assert(panelContext, 'Templates: Tab Menu Button helper must be called from within a {{#tab}}');
        logger.assert(tabContext, 'Templates: Tab Menu Button helper must be called from within a {{#tab}}');
        logger.assert(_.includes(['content', 'scenario'], panelContext.type), 'Templates: Menus are only supported in content and scenario tabs');

        tabContext.hasMenu = true;

        var linkHash = _.assign({}, options.hash, {
            'data-toggle' : 'dropdown',
            'aria-haspopup' : 'true',
            'aria-owns' : '{{menuId}}'
        });

        return link.call(this, '#', null, {
            hash : linkHash,
            fn : function () { return caret.call(this, { data : options.data }); },
            data : options.data
        });
    };

    tabMenuButton.handlebarsType = 'helper';
    tabMenuButton.handlebarsName = 'tabMenuButton';

    /**
     * Creates a notification bubble for a tab.
     *
     * Available as: `\{{#notification}}\{{/notification}}`.
     *
     * Notification is a block helper and requires inner content.
     *
     * Notification should only be used inside of an <a href="templates.helpers.tab.html">tab</a>.
     *
     * Notifications are only supported in 'content' and 'scenario' tabs.
     *
     * @class templates.helpers.notification
     * @module templates.helpers
     * @protected
     * @constructor
     * @param [options] {Object} Handlebars options
     * @returns {string} Notification bubble element
     */
    var notification = function (options) {
        var panelContext = options.data && options.data.panelContext;
        logger.assert(panelContext && _.includes(['content', 'scenario'], panelContext.type), 'Templates: Notifications are only supported in content and scenario tabs');
        logger.assert(!!options.fn, 'Templates: Notification helper must be used as a block helper.');

        return renderer.renderElementWithContent('span', {
            'class' : utils.mergeClasses(['notifications'], options.hash.class)
        }, utils.getElementAttributes(options.hash), options.fn(this, options));
    };

    notification.handlebarsType = 'helper';
    notification.handlebarsName = 'notification';

    /**
     * When used as an option on the block helper, it places the contents of the block inside the `<a>` instead of the `<li>`.
     *
     * @property isAnchorContent
     * @for templates.helpers.tab
     * @type {Boolean}
     * @default false
     */

    /**
     * Allows you to specify a different href than tab target.
     *
     * (Advanced feature)
     *
     * @property href
     * @for templates.helpers.tab
     * @type {Boolean|null}
     * @default null
     */

    /**
     * Creates a tab.
     *
     * Available as: `\{{#tab}}\{{/tab}}`.
     *
     * Tab can be used as either a regular helper or a block helper.
     *
     * Tab can only be used inside of an <a href="templates.helpers.tabList.html">tabList</a> or <a href="templates.helpers.filterTabs.html">filterTabs</a>.
     *
     * You can specify `class="active"` to make sure that this tab and its corresponding panel are default visible when executing the template.
     *
     * Tabs can contain the following elements:
     *  - <a href="templates.helpers.notification.html">notification</a>
     *  - <a href="templates.helpers.tabMenuButton.html">tabMenuButton</a>
     *  - <a href="templates.helpers.dropdownMenu.html">dropdownMenu</a>
     *
     * All attributes will be applied to the underlying anchor (`<a>`) tag
     * with the exception of four clases: 'header', 'active', 'disabled', 'dropdown'.
     * If specified, these classes will appear on the list item (`<li>`).
     *
     * When used as a block helper, you can specify content that will appear after the tab link.
     * This is most often used for either notifications or dropdown menus.
     *
     * The `controls` parameter is optional only if you are using the `\{{#filterTabs}}` helper.
     *
     * @class templates.helpers.tab
     * @module templates.helpers
     * @protected
     * @constructor
     * @param name {String} The name of the tab
     * @param controls {String} The HTML id of the panel
     * @param [options] {Object} Handlebars options
     * @returns {string} Tab element
     */
    var tab = function (name, controls, options) {
        logger.assert(arguments.length >= 2, 'Templates: Tab name is required');

        if (arguments.length === 2) {
            options = controls;
            controls = null;
        }

        var panelContext = options.data && options.data.panelContext;
        logger.assert(panelContext, 'Templates: Tab helper must be called from within a {{#tabList}}');
        logger.assert(controls != null || panelContext.isFilter, 'Templates: Tab must specify its controlled panel via "controls" parameter');

        var classList = utils.classList(options.hash.class),
            isHeader = _.includes(classList, 'header'),
            isActive = false,
            isAnchorContent = options.hash.isAnchorContent,
            linkName, innerText;

        if (options.hash.href) {
            linkName = options.hash.href;
        } else if (controls && !panelContext.isFilter) {
            linkName = '#' + controls;
        } else {
            linkName = '#';
        }

        var linkHash = _.assign({}, _.omit(options.hash, 'href'), {
            '_internalRole' : 'tab',
            'aria-controls' : panelContext.isFilter ? panelContext.tabsControl : controls
        });
        linkHash.class = _.without(classList, 'header', 'active', 'disabled', 'required').join(' ');

        if (controls || panelContext.isFilter) {
            linkHash['data-toggle'] = 'tab';
            if (!panelContext.isFilter) {
                linkHash['data-target'] = '#' + controls;
            } else {
                linkHash['data-filter-type'] = controls;
            }

            panelContext.tabsCreated.push(controls);

            if (_.includes(classList, 'active')) {
                if (panelContext.activeTab) {
                    logger.error('More than one tab was selected as the active tab, current active: ' + panelContext.activeTab + ', new tab: ' + controls);
                    classList = _.without(classList, 'active');
                } else {
                    panelContext.activeTab = controls;
                    isActive = true;
                }
            }
        }

        linkHash['aria-selected'] = isActive;

        if (isHeader || panelContext.type === 'wizard') {
            innerText = name;
        } else {
            innerText = link.call(this, linkName, options.hash.title || name, {
                hash : linkHash,
                fn : function () {
                    return (isAnchorContent && options.fn) ? options.fn(this, options) : name;
                },
                data : options.data
            });
        }

        if (options.fn) {
            var frame = Handlebars.createFrame(options.data || {});
            var tabContext = frame.tabContext = {};

            if (!isAnchorContent) {
                innerText += options.fn(this, { data : frame });
            }

            if (tabContext.hasMenu) {
                classList.push('dropdown');
                innerText = innerText.replace(/\{\{menuId\}\}/g, tabContext.menuId);
            }
        }

        return renderer.renderElementWithContent('li', {
            'class' : _.intersection(classList, ['header', 'active', 'disabled', 'dropdown', 'required']).join(' '),
            'style' : options.hash.hidden ? 'display: none;' : undefined,
            'data-wizard-step' : panelContext.type === 'wizard' ? controls : undefined,
            'data-target' : isHeader ? linkHash['data-target'] : undefined
        }, null, innerText);
    };

    tab.handlebarsType = 'helper';
    tab.handlebarsName = 'tab';

    var scrollers = function (scrollClass) {
        var scroller = function (clz, title) {
            return button.call(this, 'not-btn', title, {
                hash : { 'class' : clz, tabindex : '-1' },
                fn : function () {
                    return renderer.renderElementWithContent('span', {
                        'class' : 'spoken-text'
                    }, null, title);
                },
                data : {}
            });
        };

        return renderer.renderElementWithContent('span', {
            'class' : 'scrollers'
        }, null, scroller(scrollClass, lang.getString(scrollClass === 'left' ? 'panel.scrollLeft' : 'panel.scrollRight')));
    };

    var collapseHandle = function () {
        var collapseStr = lang.getString('panel.collapse');
        return renderer.renderElementWithContent('span', {
            'class' : 'collapse-handle'
        }, null, link.call(this, '#', collapseStr, {
            hash : {},
            fn : function () {
                return renderer.renderElementWithContent('i', { 'class' : 'collapse-icon' }, null,
                    renderer.renderElementWithContent('span', { 'class' : 'spoken-text' }, null, collapseStr));
            },
            data : {}
        }));
    };

    var tabsMenu = function () {
        return buttonGroup.call(this, {
            hash : {
                'class' : 'tabs-menu'
            },
            fn : function (ctx, options) {
                var linkText = button.call(this, 'not-btn', lang.getString('panel.tabsMenu'), {
                    hash : { dropdownToggle : true },
                    fn : function (ctx, options) {
                        return icon.call(this, 'cog', lang.getString('panel.tabsMenu'), { hash : {}, data : options.data });
                    },
                    data : options.data
                });

                var dropdownMenu = dropdown.dropdownMenu.call(this, {
                    hash : { 'class' : 'needs-overflow disclosure pull-right' },
                    fn : _.constant('<div class="dropdown-menu nested-dropdown-menu"></div>'),
                    data : options.data
                });

                return linkText + dropdownMenu;
            },
            data : {}
        });
    };

    var tabListNoSizing = function (options) {
        var panelContext = options.data && options.data.panelContext,
            innerText = options.fn(this, options);

        return renderer.renderElementWithContent(panelContext.type === 'wizard' ? 'ol' : 'ul', {
            'class' : utils.mergeClasses([PANEL_TYPE_MAP[panelContext.type], (panelContext.numbered ? 'numbered' : '')], options.hash.class),
            'role' : 'tablist'
        }, utils.getElementAttributes(options.hash), innerText);
    };

    var tabListSizing = function (options) {
        var panelContext = options.data && options.data.panelContext,
            mainTabs = tabListNoSizing.call(this, options),
            collapseHandleText = '', beforeTabs, afterTabs;

        if (panelContext.type === 'detail' && (options.hash.hasCollapse || panelContext.hasCollapse)) {
            collapseHandleText = collapseHandle();
        }

        beforeTabs = renderer.renderElementWithContent('div', {
            'class' : 'before-tabs'
        }, null, scrollers('left'));

        afterTabs = renderer.renderElementWithContent('div', {
            'class' : 'after-tabs'
        }, null, scrollers('right') + tabsMenu() + collapseHandleText);

        return renderer.renderElementWithContent('div', {
            'class' : 'nav-tabs-scroll-container has-before-tabs has-after-tabs ' + PANEL_TYPE_NO_NAV_MAP[panelContext.type]
        }, null, beforeTabs + mainTabs + afterTabs);
    };

    /**
     * Creates a tab list.
     *
     * Available as: `\{{#tabList}}\{{/tabList}}`.
     *
     * Tab list is a block helper and requires inner content.
     *
     * Tabs should list several <a href="templates.helpers.tab.html">tabs</a>.
     *
     * Tab list can only be used inside of an <a href="templates.helpers.panelContainer.html">panelContainer</a>.
     *
     *
     * @class templates.helpers.tabList
     * @module templates.helpers
     * @protected
     * @constructor
     * @param [options] {Object} Handlebars options
     * @returns {string} Tab list element
     */
    var tabList = function (options) {
        var panelContext = options.data && options.data.panelContext;

        logger.assert(panelContext, 'Templates: Tab List helper must be called from within a {{#panelContainer}}');
        logger.assert(!!options.fn, 'Templates: Tab list helper must be used as a block helper.');

        if (options.hash.noSizing || _.includes(['wizard', 'page'], panelContext.type)) {
            return tabListNoSizing.call(this, options);
        } else {
            return tabListSizing.call(this, options);
        }
    };

    tabList.handlebarsType = 'helper';
    tabList.handlebarsName = 'tabList';

    /**
     * Prevents a detail panel from automatically creating a collapse button.
     *
     * Only available when type='detail'
     *
     * @property hasCollapse
     * @for templates.helpers.panelContainer
     * @type {Boolean}
     * @default false
     */

    /**
     * Skips creating the outer <div class="panel-container">.
     *
     * While you can set this to true, you are still required to use the \{{#panelContainer}} helper to set up the tabs and panels.
     * It creates a context that is shared between the tabs and panels.
     *
     * @property noWrap
     * @for templates.helpers.panelContainer
     * @type {Boolean}
     * @default false
     */

    /**
     * Do not automatically set panels to 100% of the height of the panels contains (&lt;div class="tab-container"&gt;).
     *
     * @property noSizePanels
     * @for templates.helpers.panelContainer
     * @type {Boolean}
     * @default false
     */

    /**
     * Creates a panel container element.
     * See <a href="http://pgbulwf.us.oracle.com/style-guide/?section=web&topic=Navigation">Panels on the Blue Sky UX Guide</a>.
     *
     * Available as: `\{{#panelContainer}}\{{/panelContainer}}`.
     *
     * Panel container is a block helper and requires inner content.
     *
     * Panels should have the following inner elements (in this order):
     *  - <a href="templates.helpers.tabList.html">tabList</a>
     *  - <a href="templates.helpers.panels.html">panels</a>
     *
     * Example:
     *
     *     \{{#panelContainer 'detail'}}
     *         \{{#tabList}}
     *             \{{tab 'Foo' 'foo' class="active"}}
     *             \{{tab 'Bar' 'bar'}}
     *         \{{/tabList}}
     *         \{{#panels}}
     *             \{{#panel 'foo'}}
     *                 ...
     *             \{{/panel}}
     *             \{{#panel 'bar'}}
     *                 ...
     *             \{{/panel}}
     *         \{{/panels}}
     *     \{{/panelContainer}}
     *
     * See Also: [\{{#dock}}](templates.helpers.dock.html)
     *
     * @class templates.helpers.panelContainer
     * @module templates.helpers
     * @constructor
     * @param type {String} The type of panel to create
     * @param [options] {Object} Handlebars options
     * @returns {string} Panel container element
     */
    var panelContainer = function (type, options) {
        logger.assert(arguments.length === 2, 'Templates: Panel helper requires type argument');
        logger.assert(_.has(PANEL_TYPE_MAP, type), 'Templates: Panel type must be one of: ' + JSON.stringify(_.keys(PANEL_TYPE_MAP)));
        logger.assert(!!options.fn, 'Templates: Panel container helper must be used as a block helper.');

        var frame = Handlebars.createFrame(options.data || {});
        var panelsContext = frame.panelContext = {
            type : type,
            hasCollapse : options.hash.hasCollapse,
            noSizePanels : options.hash.noSizePanels,
            numbered : options.hash.numbered,
            tabsCreated : [],
            activeTab : null,
            panelsCreated : []
        };

        var innerText = options.fn(this, { data : frame });

        var missingTabs = _.difference(panelsContext.panelsCreated, panelsContext.tabsCreated),
            missingPanels = _.difference(panelsContext.tabsCreated, panelsContext.panelsCreated);

        if (missingTabs.length > 0) {
            logger.warn('Panels created without corresponding tab: ' + missingTabs.join(', '));
        }
        if (missingPanels.length > 0) {
            logger.warn('Tabs created without corresponding panel: ' + missingPanels.join(', '));
        }

        if (options.hash.noWrap) {
            return innerText;
        } else {
            return renderer.renderElementWithContent('div', {
                'class' : utils.mergeClasses([type === 'wizard' ? 'wizard-context' : 'panel-container'], options.hash.class)
            }, utils.getElementAttributes(options.hash), innerText);
        }
    };

    panelContainer.handlebarsType = 'helper';
    panelContainer.handlebarsName = 'panelContainer';

    /**
     * Creates a tab list whose purpose is to filter or control an existing element.
     * Instead of switching between different `<div>`s, the generated tabs simply trigger `show` and `shown` events
     * and leave the implementation up to the developer.
     *
     * Available as: `\{{#filterTabs}}\{{/filterTabs}}`.
     *
     * Filter tabs should use the [`tab`](templates.helpers.tab.html) helper to create tabs inside.
     *
     * Example:
     *
     *     \{{#filterTabs 'controlled-element-id'}}
     *         \{{tab 'Open' 'open' class="active"}}
     *         \{{tab 'Closed' 'closed'}}
     *     \{{/filterTabs}}
     *
     * The second argument for the tab helper is optional.
     * If it is set, it will be added to the link as `data-filter-type`.
     *
     * @class templates.helpers.filterTabs
     * @module templates.helpers
     * @constructor
     * @param controls {String} The ID of the element being controlled by the filter tabs
     * @param [options] {Object} Handlebars options
     * @return {String} Filter tabs element
     */
    var filtersTabs = function (controls, options) {
        logger.assert(arguments.length === 2, 'Templates: Filter Tab helper requires a "controls" parameter');
        logger.assert(!options.data.panelContext, 'Templates: Filter Tab helper *cannot* be used in {{#panel}} context');
        logger.assert(!!options.fn, 'Templates: Filter Tab helper must be used as a block helper.');

        var frame = Handlebars.createFrame(options.data || {});
        frame.panelContext = {
            type : 'detail',
            isFilter : true,
            tabsControl : controls,
            hasCollapse : false,
            tabsCreated : [],
            activeTab : null
        };

        return tabListNoSizing.call(this, _.merge({ data : frame }, options));
    };

    filtersTabs.handlebarsType = 'helper';
    filtersTabs.handlebarsName = 'filterTabs';

    return {
        panelContainer : panelContainer,
        tabList : tabList,
        tab : tab,
        notification : notification,
        tabMenuButton : tabMenuButton,
        panels : panels,
        panel : panel,
        filterTabs : filtersTabs
    };
});

define('primutils/templates/helpers/placeholder',[
    'lodash',
    'logger',
    'primutils/templates/utils/renderer',
    'primutils/templates/utils/utils'
], function (_, logger, renderer, utils) {
    'use strict';

    var saveCurrentContext = function (key, options) {
        var data = _.omit(options.data || {}, '_parent', 'root');
        if (_.has(options.data.root, key)) {
            logger.warn('Templates: In {{saveCurrentContext}}, ' + key + ' property was already defined. Overwritting.');
        }
        options.data.root[key] = data;
    };

    saveCurrentContext.handlebarsType = 'helper';
    saveCurrentContext.handlebarsName = 'saveCurrentContext';

    var placeholder = function (key, options) {
        saveCurrentContext.call(this, key, options);
        return renderer.renderElement('div', {
            'class' : options.hash.class
        }, utils.getElementAttributes(options.hash));
    };

    placeholder.handlebarsType = 'helper';
    placeholder.handlebarsName = 'placeholder';

    return {
        saveCurrentContext : saveCurrentContext,
        placeholder : placeholder
    };
});


define('primutils/templates/helpers/search',[
    'logger',
    'lodash',
    'jquery',
    'primutils/utils/language',
    'primutils/templates/utils/utils',
    'primutils/templates/utils/renderer',
    'primutils/templates/helpers/link',
    'primutils/templates/helpers/icon',
    'primutils/templates/helpers/input'
], function (logger, _, $, lang, utils, renderer, link, icon, input) {
    'use strict';

    var searchBase = function (baseClassName, value, options) {
        if (arguments.length === 2) {
            options = value;
            value = null;
        }

        // Prefer value from block helper body if available
        value = options.fn ? options.fn(this, options) : value;

        var innerText = '',
            inputHash = _.omit(options.hash, 'class', 'fluid', 'style'),
            outerHash = _.pick(options.hash, 'class', 'style');

        inputHash.class = 'filterbox search-query';

        innerText += input.input.call(this, 'search', value, {
            hash : inputHash,
            data : options.data
        });

        innerText += renderer.renderElementWithContent('span', {
            'class' : 'search-icon'
        }, null, icon.call(this, 'magnifying-glass', lang.getString('search.search'), { hash : { 'size' : 'pgbu-icon-small' }, data : {} }));

        innerText += link.call(this, '#', lang.getString('search.clear'), {
            hash : { 'class' : 'search-clear-icon' },
            fn : function () {
                return icon.call(this, 'find-clear', lang.getString('search.clear'), { hash : {}, data : {} });
            },
            data : options.data
        });

        return renderer.renderElementWithContent('div', {
            'class' : utils.mergeClasses([baseClassName, options.hash.fluid ? baseClassName + '-fluid' : ''], options.hash.class)
        }, utils.getElementAttributes(outerHash), innerText);
    };

    /**
     * Makes the search box take up the full width of its container
     *
     * @property fluid
     * @for templates.helpers.search
     * @type {Boolean}
     * @default false
     */

    /**
     * Creates a search element.
     * See <a href="http://localhost/style-guide/?section=web&topic=Navigation&subtopic=LeftHandInPageNavigation">Search on the Blue Sky UX Guide</a>.
     *
     * Input can be used as either a regular helper or a block helper.
     *
     * Examples:
     *
     *     \{{search 'value'}}
     *
     *     \{{search}}Value\{{/value}}
     *
     * See also: <a href="templates.helpers.input.html">\{{#input}}</a>
     *
     * @class templates.helpers.search
     * @module templates.helpers
     * @constructor
     * @param [value] {String} The default value of the search
     * @param [options] {Object} Handlebars options
     * @returns {string} Search element
     */
    var search = _.partial(searchBase, 'filter-div');
    search.handlebarsType = 'helper';
    search.handlebarsName = 'search';

    var globalSearch = _.partial(searchBase, 'search-div');
    globalSearch.handlebarsType = 'helper';
    globalSearch.handlebarsName = 'globalSearch';

    return {
        search : search,
        globalSearch : globalSearch
    };
});


define('primutils/templates/helpers/slider',[
    'logger',
    'lodash',
    'primutils/templates/utils/utils',
    'primutils/templates/utils/renderer',
    'primutils/templates/helpers/button'
], function (logger, _, utils, renderer, button) {
    'use strict';

    /**
     * Creates a toggle switch element.
     * See <a href="http://pgbulwf.us.oracle.com/style-guide/?section=web&topic=Buttons&subtopic=ToggleControl">Slider on the Blue Sky UX Guide</a>.
     *
     * Available as: `\{{slider}}`.
     *
     * Slider is a helper that does not have inner content.
     *
     * Example:
     *
     *     \{{slider 'On' 'Off'}}
     *
     * See also: <a href="templates.helpers.toggle.html">\{{#toggle}}</a>
     *
     * @class templates.helpers.slider
     * @module templates.helpers
     * @constructor
     * @param onText {String} The screen reader spoken text when the button is on
     * @param offText {String} The screen reader spoken text when the button is off
     * @param [options] {Object} Handlebars options
     * @returns {string} Slider element
     */
    var slider = function (onText, offText, options) {
        logger.assert(arguments.length === 3, 'Templates: Slider helper requires on/off text arguments');

        return button.call(this, 'slider', '', {
            hash : _.assign(options.hash, { 'data-toggle' : 'button' }),
            fn : function () { return ''; },
            data : options.data
        });
    };

    slider.handlebarsType = 'helper';
    slider.handlebarsName = 'slider';

    return slider;
});


define('primutils/templates/helpers/spinner',[
    'logger',
    'lodash',
    'primutils/templates/utils/utils',
    'primutils/templates/utils/renderer'
], function (logger, _, utils, renderer) {
    'use strict';

    /**
     * Concatenates variables and string expressions together when needed in hashes.
     *
     * Example:
     *
     *     \{{spinner}}
     *
     * @class templates.helpers.spinner
     * @module templates.helpers
     * @constructor
     * @param [size] {String|Number} Size of the spinner in font-compatible units (px, em, etc.)
     * @param [options] {Object} Handlebars options
     * @returns {string} Joined string
     */
    var spinner = function (size, options) {
        if (arguments.length === 1) {
            options = size;
            size = '60px'; // default size
        }

        if (_.isNumber(size)) {
            size = size.toString() + 'px';
        }

        // If we are making this an icon, do not set the size
        if (options.hash.class && options.hash.class.indexOf('pgbu-icon-') > -1) {
            size = null;
        }

        var dots = '';
        _.times(3, function (n) {
            dots += renderer.renderElement('div', { 'class' : 'dot dot' + (n + 1) }, {});
        });
        return renderer.renderElementWithContent('div', {
            'class' : utils.mergeClasses(['pulse-spinner'], options.hash.class),
            'style' : size != null ? ((options.hash.style || '') + ';font-size: ' + size + ';') : undefined
        }, utils.getElementAttributes(options.hash), dots);
    };

    spinner.handlebarsType = 'helper';
    spinner.handlebarsName = 'spinner';

    return spinner;
});


define('primutils/templates/helpers/steps',[
    'logger',
    'primutils/templates/utils/utils',
    'primutils/templates/utils/renderer'
], function (logger, utils, renderer) {
    'use strict';

    /**
     * Shows the step numbers of each step.
     * @property numbered
     * @for templates.helpers.stepList
     * @type {Boolean}
     * @default false
     */

    /**
     * Creates a stepList element.
     * See <a href="http://pgbulwf.us.oracle.com/style-guide/?section=web&topic=Dialogs&subtopic=StepIndicator">Step Lists on the Blue Sky UX Guide</a>.
     *
     * Available as: `\{{#stepList}}\{{/stepList}}`.
     *
     * Step list is a block helper and requires inner content.
     * Only <a href="templates.helpers.stepItem.html">\{{#stepItem}}s</a> should be placed inside step lists.
     *
     * Example:
     *
     *     \{{#stepList numbered=true}}
     *         \{{#stepItem}}First Step\{{/stepItem}}
     *         \{{#stepItem}}Second Step\{{/stepItem}}
     *     \{{/stepList}}
     *
     * @class templates.helpers.stepList
     * @module templates.helpers
     * @constructor
     * @param [options] {Object} Handlebars options
     * @returns {string} Step list element
     */
    var stepList = function (options) {
        logger.assert(!!options.fn, 'Templates: Step list helper must be used as a block helper.');

        return renderer.renderElementWithContent('ol', {
            class : utils.mergeClasses(['steps', options.hash.numbered ? 'numbered' : ''], options.hash.class)
        },
            utils.getElementAttributes(options.hash), options.fn(this, options));
    };

    stepList.handlebarsType = 'helper';
    stepList.handlebarsName = 'stepList';

    /**
     * Creates a stepItem element.
     * See <a href="http://pgbulwf.us.oracle.com/style-guide/?section=web&topic=Dialogs&subtopic=StepIndicator">Step Lists on the Blue Sky UX Guide</a>.
     *
     * Available as: `\{{#stepItem}}\{{/stepItem}}`.
     *
     * Step item is a block helper and requires inner content.
     *
     * Step items should only be used inside of a <a href="templates.helpers.stepList.html">stepList</a>.
     *
     * @class templates.helpers.stepItem
     * @module templates.helpers
     * @protected
     * @constructor
     * @param [options] {Object} Handlebars options
     * @returns {string} Step item element
     */
    var stepItem = function (options) {
        logger.assert(!!options.fn, 'Templates: Step item helper must be used as a block helper.');

        return renderer.renderElementWithContent('li', null, utils.getElementAttributes(options.hash, ['class']), options.fn(this, options));
    };

    stepItem.handlebarsType = 'helper';
    stepItem.handlebarsName = 'stepItem';

    return {
        stepList : stepList,
        stepItem : stepItem
    };
});


define('primutils/markup/primutils/templates/helpers/termDef.hbs', ['handlebars','primutils.markup','primutils/templates/helpers/icon'], function (H, u) {
var imports = u.buildImported(Array.prototype.slice.call(arguments, 2), 'primutils/templates/helpers/termDef.hbs'),
    func = H.template({"1":function(depth0,helpers,partials,data) {
  var stack1, lambda=this.lambda, buffer = "            <div class=\"header\">";
  stack1 = lambda((depth0 != null ? depth0.header : depth0), depth0);
  if (stack1 != null) { buffer += stack1; }
  return buffer + "</div>\n";
},"compiler":[6,">= 2.0.0-beta.1"],"main":function(depth0,helpers,partials,data) {
  var stack1, lambda=this.lambda, buffer = "<div class=\"orcl-popover ";
  stack1 = lambda((depth0 != null ? depth0.side : depth0), depth0);
  if (stack1 != null) { buffer += stack1; }
  buffer += " ";
  stack1 = lambda((depth0 != null ? depth0.classes : depth0), depth0);
  if (stack1 != null) { buffer += stack1; }
  buffer += "\" ";
  stack1 = lambda((depth0 != null ? depth0.attributes : depth0), depth0);
  if (stack1 != null) { buffer += stack1; }
  buffer += " tabindex=\"0\">\n    <div class=\"orcl-popover-outer\">\n        ";
  stack1 = helpers.icon.call(depth0, (depth0 != null ? depth0.icon : depth0), (depth0 != null ? depth0.title : depth0), {"name":"icon","hash":{},"data":data});
  if (stack1 != null) { buffer += stack1; }
  buffer += "\n    </div>\n    <div class=\"orcl-popover-inner\">\n        <div class=\"orcl-popover-content\">\n";
  stack1 = helpers['if'].call(depth0, (depth0 != null ? depth0.hasHeader : depth0), {"name":"if","hash":{},"fn":this.program(1, data),"inverse":this.noop,"data":data});
  if (stack1 != null) { buffer += stack1; }
  buffer += "            <div>";
  stack1 = lambda((depth0 != null ? depth0.content : depth0), depth0);
  if (stack1 != null) { buffer += stack1; }
  return buffer + "</div>\n        </div>\n    </div>\n</div>";
},"useData":true});
return function (context, options) {
    options = u.extendOptions(options || {}, imports);
    return func(context, options).trim();
};
});


define('primutils/templates/helpers/termDef',['primutils/templates/utils/renderer', 'primutils/templates/utils/utils', 'primutils/markup/primutils/templates/helpers/termDef.hbs', 'logger'], function (renderer, utils, markup, logger) {
    'use strict';

    /**
     * Creates a terminology definition.
     *
     * Available as: `\{{#termDef}}\{{/termDef}}`.
     *
     * Terminology definition is a block helper and requires inner content.
     *
     * Example:
     *
     *     \{{#termDef 'Header' 'Unexpanded hover text'}}\{{/termDef}}
     *
     * See also: <a href="templates.helpers.button.html">\{{#button}}</a>
     *
     * @class templates.helpers.termDef
     * @module templates.helpers
     * @constructor
     * @param header {String} Header inside terminology definition (Optional, pass `''` (empty string) to omit)
     * @param title {String} Hover text for the terminology definition
     * @param [options] {Object} Handlebars options
     * @returns {string} Caret element
     */
    var termDef = function (header, title, options) {
        if (!require.defined('oui')) {
            logger.warn('Terminology definition created without OUI being loaded. It may not function correctly');
        }
        logger.assert(!!options.fn, 'Templates: termDef helper must be used as a block helper.');
        return markup({
            icon : options.hash.icon || 'popover-help',
            side : options.hash.side || '',
            title : title,
            hasHeader : header && header.length > 0,
            header : header,
            content : options.fn(this, options),
            classes : options.hash.class,
            attributes : renderer.renderAttributesString(utils.getElementAttributes(options.hash))
        });
    };

    termDef.handlebarsType = 'helper';
    termDef.handlebarsName = 'termDef';

    return termDef;
});


define('primutils/templates/helpers/ternary',[], function () {
    'use strict';

    /**
     * Simple ternary helper, accessible by either `?` or `ternary`.
     *
     * Does an inline `if` check and returns the second parameter if true or the third parameter if false.
     *
     * Due to limitations in Handlebars, the condition (first parameter) must be a boolean value.
     * You cannot run a condition check (like `foo === 3`) inside the helper.
     *
     * The purpose of this helper if to allow you to use the `\{{#for}}` helper inside of attributes.
     *
     * Example:
     *
     *     <div class={{? myBoolean "true-class" "false-class"}}></div>
     *     {{#button 'toolbar' class=(? isVisible "" "hidden")}}Conditionally-visible Button{{/button}
     *
     * @class templates.helpers.ternary
     * @module templates.helpers
     * @constructor
     * @param {Boolean} bool True/False value
     * @param {String} ifTrue Value to output if `bool` is true
     * @param {String} ifFalse Value to output if `bool` is false
     * @param [options] {Object} Handlebars options
     * @returns {string}
     */
    var ternary = function (bool, ifTrue, ifFalse /*, options */) {
        return bool ? ifTrue : ifFalse;
    };

    ternary.handlebarsType = 'helper';
    ternary.handlebarsName = '?';

    var ternaryAlt = function () {
        return ternary.apply(this, arguments);
    };

    ternaryAlt.handlebarsType = 'helper';
    ternaryAlt.handlebarsName = 'ternary';

    return {
        '?' : ternary,
        'ternary' : ternaryAlt
    };
});



define('primutils/templates/helpers/toggle',[
    'logger',
    'primutils/templates/utils/utils',
    'primutils/templates/utils/renderer'
], function (logger, utils, renderer) {
    'use strict';

    var TOGGLE_TYPE_MAP = {
        'page' : 'btn-toggle-control',
        'data' : 'btn-toggle-switch'
    };

    /**
     * Creates a toggle button group.
     * See <a href="http://pgbulwf.us.oracle.com/style-guide/?section=web&topic=Buttons&subtopic=ToggleControl">Toggles on the Blue Sky UX Guide</a>.
     *
     * Available as: `\{{#toggle}}\{{/toggle}}`.
     *
     * Toggle button group is a block helper and requires inner content.
     * Two button elements should be specified.
     * Their 'type' parameter should be omitted.
     * If you need a title on your buttons, you should pass '' (empty string) as the button type.
     *
     * Example:
     *
     *     \{{#toggle 'page'}}
     *         \{{#button}}EPS View\{{/button}}
     *         \{{#button}}Project View\{{/button}}
     *     \{{/toggle}}
     *
     * There are two types of toggles: 'page' and 'data'.
     *
     * See also: <a href="templates.helpers.buttonGroup.html">\{{#buttonGroup}}</a>, <a href="templates.helpers.slider.html">\{{slider}}</a>
     *
     * @class templates.helpers.toggle
     * @module templates.helpers
     * @constructor
     * @param type {String} The type of toggle to create
     * @param [options] {Object} Handlebars options
     * @returns {string} Toggle element
     */
    var toggle = function (type, options) {
        logger.assert(arguments.length === 2, 'Templates: Toggle helper requires a "type" argument');
        logger.assert(!!options.fn, 'Templates: Toggle helper must be used as a block helper.');

        return renderer.renderElementWithContent('div', {
            'class' : utils.mergeClasses(['btn-group', TOGGLE_TYPE_MAP[type]], options.hash.class),
            'data-toggle' : 'buttons-radio'
        }, utils.getElementAttributes(options.hash), options.fn(this, options));
    };

    toggle.handlebarsType = 'helper';
    toggle.handlebarsName = 'toggle';

    return toggle;
});


define('primutils/templates/helpers/toolbar',[
    'handlebars',
    'logger',
    'primutils/templates/utils/utils',
    'primutils/templates/utils/renderer'
], function (Handlebars, logger, utils, renderer) {
    'use strict';

    /**
     * Creates a button toolbar.
     * See <a href="http://pgbulwf.us.oracle.com/style-guide/?section=web&topic=Buttons&subtopic=GridActionButtons">Toolbars on the Blue Sky UX Guide</a>.
     *
     * Available as: `\{{#toolbar}}\{{/toolbar}}`.
     *
     * Toolbar is a block helper and requires inner content, generally <a href="templates.helpers.button.html">\{{#button}}s</a> and <a href="templates.helpers.buttonGroup.html">\{{#buttonGroup}}s</a>.
     *
     * Example:
     *
     *     \{{#toolbar}}
     *         \{{#button 'toolbar'}}Add\{{/button}}
     *         \{{#buttonGroup}}
     *             \{{#button 'toolbar'}}Edit\{{/button}}
     *             \{{#button 'toolbar'}}Super Edit\{{/button}}
     *         \{{/buttonGroup}}
     *     \{{/toolbar}}
     *
     * **Note:** Buttons inside of a toolbar do not automatically get the 'toolbar' button type.
     *
     * See also: <a href="templates.helpers.button.html">\{{#button}}</a>, <a href="templates.helpers.buttonGroup.html">\{{#buttonGroup}}</a>
     *
     * @class templates.helpers.toolbar
     * @module templates.helpers
     * @constructor
     * @param [options] {Object} Handlebars options
     * @returns {string} Toolbar element
     */
    var toolbar = function (options) {
        logger.assert(!!options.fn, 'Templates: Toolbar helper must be used as a block helper.');

        var frame = Handlebars.createFrame(options.data || {});
        frame.toolbarContext = {}; // Placeholder for inline inputs
        var innerText = options.fn(this, { data : frame });

        return renderer.renderElementWithContent('div', {
            'class' : utils.mergeClasses(['btn-toolbar'], options.hash.class),
            'role' : 'toolbar'
        }, utils.getElementAttributes(options.hash), innerText);
    };

    toolbar.handlebarsType = 'helper';
    toolbar.handlebarsName = 'toolbar';

    return toolbar;
});

define('primutils/templates/helpers/wizard',['lodash', 'primutils/templates/helpers/button'], function (_, button) {
    'use strict';

    var wizardBaseButton = function (type) {
        var args = _.rest(_.toArray(arguments)),
            options = _.last(args);

        options.hash['data-wizard'] = type;

        return button.apply(this, args);
    };

    if (require.nodeRequire && !_) {
        _ = require.nodeRequire('lodash');
    }

    /**
     * @class templates.helpers.wizardNextButton
     * @extends templates.helpers.button
     * @module templates.helpers
     * @constructor
     * @param type {String} The type of the button
     * @param [title] {String} The title text of the button (Default: .text() of the button's inner text)
     * @param [options] {Object} Handlebars options
     * @returns {string} Button element
     */
    var wizardNextButton = _.partial(wizardBaseButton, 'next');
    wizardNextButton.handlebarsType = 'helper';
    wizardNextButton.handlebarsName = 'wizardNextButton';

    /**
     * @class templates.helpers.wizardPrevButton
     * @extends templates.helpers.button
     * @module templates.helpers
     * @constructor
     * @param type {String} The type of the button
     * @param [title] {String} The title text of the button (Default: .text() of the button's inner text)
     * @param [options] {Object} Handlebars options
     * @returns {string} Button element
     */
    var wizardPrevButton = _.partial(wizardBaseButton, 'prev');
    wizardPrevButton.handlebarsType = 'helper';
    wizardPrevButton.handlebarsName = 'wizardPrevButton';

    /**
     * @class templates.helpers.wizardCancelButton
     * @extends templates.helpers.button
     * @module templates.helpers
     * @constructor
     * @param type {String} The type of the button
     * @param [title] {String} The title text of the button (Default: .text() of the button's inner text)
     * @param [options] {Object} Handlebars options
     * @returns {string} Button element
     */
    var wizardCancelButton = _.partial(wizardBaseButton, 'cancel');
    wizardCancelButton.handlebarsType = 'helper';
    wizardCancelButton.handlebarsName = 'wizardCancelButton';

    /**
     * @class templates.helpers.wizardSaveButton
     * @extends templates.helpers.button
     * @module templates.helpers
     * @constructor
     * @param type {String} The type of the button
     * @param [title] {String} The title text of the button (Default: .text() of the button's inner text)
     * @param [options] {Object} Handlebars options
     * @returns {string} Button element
     */
    var wizardSaveButton = _.partial(wizardBaseButton, 'save');
    wizardSaveButton.handlebarsType = 'helper';
    wizardSaveButton.handlebarsName = 'wizardSaveButton';

    /**
     * @class templates.helpers.wizardResetButton
     * @extends templates.helpers.button
     * @module templates.helpers
     * @constructor
     * @param type {String} The type of the button
     * @param [title] {String} The title text of the button (Default: .text() of the button's inner text)
     * @param [options] {Object} Handlebars options
     * @returns {string} Button element
     */
    var wizardResetButton = _.partial(wizardBaseButton, 'reset');
    wizardResetButton.handlebarsType = 'helper';
    wizardResetButton.handlebarsName = 'wizardResetButton';

    return {
        wizardNextButton : wizardNextButton,
        wizardPrevButton : wizardPrevButton,
        wizardCancelButton : wizardCancelButton,
        wizardSaveButton : wizardSaveButton,
        wizardResetButton : wizardResetButton
    };
});

define('primutils/templates/helpers/chartTooltip',[
    'lodash',
    'handlebars',
    'primutils/templates/utils/utils',
    'primutils/templates/utils/renderer'
], function (_, Handlebars, utils, renderer) {
    'use strict';

    var WHITE = '#FFFFFF';
    var BLUE_DARK_GREY = '#495363';
    var GRAY_DARKER = '#333333';
    var FONT_WEIGHT_BOLD = 'font-weight:bold;';
    var FONT_SIZE_11 = 'font-size:11px;';
    var FONT_SIZE_13 = 'font-size:13px;';
    var LABEL_STYLE = 'color:' + BLUE_DARK_GREY + ';' + FONT_SIZE_11 + FONT_WEIGHT_BOLD + 'margin-right:5px';
    var TITLE_STYLE = 'color:' + BLUE_DARK_GREY + ';' + FONT_SIZE_13;
    var TEXT_STYLE = 'color:' + GRAY_DARKER + ';' + FONT_SIZE_11;

    /**
     * Creates chartTooltip DOM, suitable for use with JET or OCharts.
     *
     * Example:
     *
     *     \{{#chartTooltip}}
     *         \{{#chartTooltipTitle}}{{this.title}}{{/chartTooltipTitle}}
     *         \{{#chartTooltipRow}}
     *             \{{#chartTooltipLabel}}Values{{/chartTooltipLabel}}
     *         \{{/chartTooltipRow}}
     *         \{{#chartTooltipRow}}
     *             \{{#chartTooltipLabel}}Horizon:{{/chartTooltipLabel}}
     *             \{{#chartTooltipText}}{{this.x}}{{/chartTooltipText}}
     *         \{{/chartTooltipRow}}
     *         \{{#chartTooltipRow}}
     *             \{{#chartTooltipLabel}}Value:{{/chartTooltipLabel}}
     *             \{{#chartTooltipText}}{{this.y}}{{/chartTooltipText}}
     *         \{{/chartTooltipRow}}
     *     \{{/chartTooltip}}
     *
     * @class templates.helpers.chartTooltip
     * @module templates.helpers
     * @constructor
     * @returns {String} chart tooltip as string
     */
    var chartTooltip = function (options) {
        var frame = Handlebars.createFrame(options.data || {});
        var innerText = options.fn ? options.fn(this, { data : frame }) : '';

        return renderer.renderElementWithContent('div', {
            class : 'chart-tooltip',
            style : 'padding:13px;background-color:' + WHITE
        }, utils.getElementAttributes(options.hash), innerText);
    };
    chartTooltip.handlebarsType = 'helper';
    chartTooltip.handlebarsName = 'chartTooltip';

    var chartTooltipTitle = function (options) {
        var innerText = options.fn(this, options);
        return renderer.renderElementWithContent('div', {
            class : 'chart-tooltip-title',
            style : TITLE_STYLE
        }, utils.getElementAttributes(options.hash), innerText);
    };
    chartTooltipTitle.handlebarsType = 'helper';
    chartTooltipTitle.handlebarsName = 'chartTooltipTitle';

    var chartTooltipRow = function (options) {
        var innerText = options.fn(this, options);
        return renderer.renderElementWithContent('div', {
            class : 'chart-tooltip-row'
        }, utils.getElementAttributes(options.hash), innerText);
    };
    chartTooltipRow.handlebarsType = 'helper';
    chartTooltipRow.handlebarsName = 'chartTooltipRow';

    var chartTooltipLabel = function (options) {
        var innerText = options.fn(this, options);
        return renderer.renderElementWithContent('span', {
            class : 'chart-tooltip-label',
            style : LABEL_STYLE
        }, utils.getElementAttributes(options.hash), innerText);
    };
    chartTooltipLabel.handlebarsType = 'helper';
    chartTooltipLabel.handlebarsName = 'chartTooltipLabel';

    var chartTooltipText = function (options) {
        var innerText = options.fn(this, options);
        return renderer.renderElementWithContent('span', {
            class : 'chart-tooltip-text',
            style : TEXT_STYLE
        }, utils.getElementAttributes(options.hash), innerText);
    };
    chartTooltipText.handlebarsType = 'helper';
    chartTooltipText.handlebarsName = 'chartTooltipText';

    return {
        chartTooltip : chartTooltip,
        chartTooltipTitle : chartTooltipTitle,
        chartTooltipRow : chartTooltipRow,
        chartTooltipText : chartTooltipText,
        chartTooltipLabel : chartTooltipLabel
    };
});

define('primutils/utils/validator',[
    'jquery',
    'lodash',
    'logger',
    'moment'
], function ($, _, logger, moment) {
    'use strict';

    var __emptyValidator__ = {
            name : '',
            required : [],
            banned : [],
            customValidations : []
        },
        __defaultOptions__ = {
            name : 'Unnamed Validator',
            ignoredSubkeys : [],
            validators : {}
        };

    function loopDetectFactory(options) {
        if (!('Map' in window)) {
            return function noSetAvailable() { return false; };
        }

        var valPrefixes = new Map();
        return function hasLoop(val, prefix) {
            if (valPrefixes.has(val)) {
                var existingPrefix = valPrefixes.get(val);
                var matchingPrefixPortion = _.dropRight(prefix.split('.'), 1).join('.');
                if (matchingPrefixPortion.indexOf(existingPrefix) === 0) {
                    logger.error(options.name + ' validator | detected a circular loop in options passed to validator; consider using the ignoredSubkeys or other options as appropriate to avoid circular references and infinite loops in non-modern browsers.');
                    return true;
                }
                // else...probably just one object/func referenced multiple times
                return false; // so let's effectively just leave the first one in and not set again
            }
            valPrefixes.set(val, prefix);
        };
    }

    function optionsKeys(options, ignoredSubkeys, prefix, loopDetector) {
        var result = [], ignoresSubs;

        prefix = prefix || '';
        loopDetector = loopDetector || loopDetectFactory(options);

        _.forEach(_.keys(options), function (key) {
            var obj = options[key];
            ignoresSubs = _.some(ignoredSubkeys, function (ignoreKey) {
                return ignoreKey === prefix + key;
            });

            result.push(prefix + key);
            if (moment.isMoment(obj)) {
                // Skip over date (Moment, really) subobjects
                return;
            }
            if (moment.isDuration(obj)) {
                // Durations, too
                return;
            }
            if (obj instanceof $) {
                return;
            }
            if (loopDetector(obj, prefix + key)) {
                return;
            }

            if (!ignoresSubs && _.isObject(obj) && !_.isArray(obj)) {
                result = _.union(
                    result,
                    optionsKeys(options[key], ignoredSubkeys, (prefix + key) + '.', loopDetector)
                );
            }
        });

        return result;
    }

    function deepValue(rootOptions, key) {
        var path = key.split('.'), pathLength = path.length,
            obj, i;
        obj = rootOptions;

        for (i = 0; i < pathLength && !_.isUndefined(obj) && !_.isNull(obj); i++) {
            obj = obj[path[i]];
        }

        return obj;
    }

    function fnToString(fn) {
        var str = fn.toString();

        if (str.split('\n').length > 1) {
            return str.split('\n')[0] + '...}';
        } else {
            return str;
        }
    }

    function valueString(obj) {
        if (_.isFunction(obj) && obj.toString().split('\n').length > 1) {
            return fnToString(obj);
        } else {
            return ('' + obj);
        }
    }

    function checkType(type, value) {
        if (_.isArray(type)) {
            return _.some(type, function (possibleType) {
                return checkType(possibleType, value);
            });
        } else if (_.isFunction(type)) {
            return type(value);
        } else if (type === 'StrictObject') {
            return _.isObject(value) && !_.isArray(value) && !_.isFunction(value);
        } else if (type === 'Jquery' || type === 'jQuery') {
            return (value instanceof $) && value.jquery;
        } else if (type === 'Date' || type === 'Moment') {
            return moment.isMoment(value);
        } else if (type === 'Duration') {
            return moment.isDuration(value);
        } else if (_.isString(type) && _.has(_, 'is' + type)) {
            return _['is' + type](value);
        } else {
            // Some options can be {Number | 'none'}, this way we can give a list of types like ['Number', 'none']
            return value === type;
        }
    }

    function printError(value, type, keyName) {
        var message = 'Illegal value for \'' + keyName + '\' option. The option\'s value was \'' +
            valueString(value) + '\', which is of type \'' + (typeof value) + '\' but should match one of the following rules/types: ';

        if (_.isArray(type)) {
            message += '[';
            _.forEach(type, function (t) {
                message += '\'' + t + '\'';
                if (_.last(type) !== t) {
                    message += ', ';
                }
            });
            message += ']';
        } else {
            message += '\'' + type + '\'';
        }

        return message;
    }

    /**
     * A helper for running validation on options provided to your component.
     *
     * For simple cases, consider using the static method <a href="#method_checkType">checkType</a>.
     *
     *     validator.checkType('a', ['b', 'c', 'Number'], 'Key name');
     *         // Would logger.warn a message, and return false,
     *         //  because the option given by the user did not match:
     *         //  the string 'b',the string 'c', or pass the _.isNumber function.
     *
     * For a simple for checking everything in a given object (the following example works as of primutils 2.7.3):
     *
     *     var myValidator = validator({
     *         validOptions: {
     *             optionA: ['none', 'Array'],
     *             optionB: ['jQuery']
     *         }
     *     });
     *     myValidator({optionC: 'value', optionB: [], optionA: 'none'});
     *         // Would logger.warn that 'optionC' is unrecognized
     *         // Would logger.warn that 'optionB' doesn't pass any of the validOptions rules
     *         // Would accept optionA without any complaints, as it matches one of the rules (an exact match for the string 'none')
     *
     * For an example of a more complex case:
     *
     *     var myValidator = validator({
     *          // included in error messages when unrecognized/banned/missing options are given by the user
     *         name: 'Your component name',
     *          // all the options that could be recognized by your component; warnings will be fired for any options given that aren't indicated here
     *         validOptions: {
     *              // your component has a option called 'minValue' that must pass the check _.isNumber
     *             minValue: 'Number',
     *              // your component has an option called 'stringValue' that must pass the check _.isString
     *             stringValue: 'String',
     *              // your component has an option called 'anything' that can be of any type
     *             anything: undefined,
     *              // your component has an option called 'someObject' that must be an Object
     *             someObject: 'Object',
     *              // your component has an option called 'someObject', with a subkey of 'subKey', which must be a MomentJS date object
     *             'someObject.subKey': 'Date',
     *              // your component has an option called 'positiveNumeric' that must pass the given function
     *             positiveNumeric: function (n) { return _.isNumber(n) && n > 0; }
     *         },
     *          // any subkeys of the 'anything' option will not be type checked for correctness
     *         ignoredSubkeys: ['anything'],
     *         validators: {
     *              // Rules specific to when you call `myValidator.validate('initialize', {...options to check...});`
     *             initialize: {
     *                  // your component requires that the provided options hash specify the 'minValue' option when you call `myValidator.validate('initialize', userOptions);`
     *                 required: ['minValue'],
     *                  // your component will not allow that the provided options hash specify the 'anything' option when you call `myValidator.validate('initialize', userOptions);`
     *                 banned: ['anything'],
     *             },
     *             configure
     *         }
     *     });
     *     myValidator.validate('initialize', userOptions); // Checks all provided userOptions keys against the spec given on creation
     *
     * @class utils.validator
     * @module utils
     * @param options {Object} see class description example
     * @constructor
     */
    function validatorFactory(options) {
        var _options = {},
            _hasValidated = false;

        _options = _.defaults({}, options || {}, _options, __defaultOptions__);

                validatorFactory.checkType(_options.name, ['String'], 'name');
        validatorFactory.checkType(_options.validators, [undefined, 'Object'], 'validators');
        validatorFactory.checkType(_options.validOptions, ['Object'], 'validOptions');
        validatorFactory.checkType(_options.ignoredSubkeys, [undefined, 'Array'], 'ignoredSubkeys');
        validatorFactory.checkType(_options.required, [undefined, 'Array'], 'required');
        validatorFactory.checkType(_options.banned, [undefined, 'Array'], 'banned');
        

        /**
         * Runs the validation rules provided at validator creation time against the provided options hash.
         * See the more complex example in the class description for more details.
         *
         * @method validate
         * @param [validatorName] {String} the validator required/banned ruleset to run
         * @param options {Object} the object to run validation rules against
         * @returns true if any test failed, false otherwise
         */
        function that(validatorName, userOptions) {
            if (arguments.length === 1) {
                userOptions = validatorName;
                validatorName = __emptyValidator__.name;
            }

            var clonedValidator = _.clone(__emptyValidator__),
                validator = _.has(_options.validators, validatorName) ?
                    _.defaults({}, _options.validators[validatorName], __emptyValidator__) :
                    clonedValidator,
                keys = optionsKeys(userOptions, _options.ignoredSubkeys),
                unrecognized = _.difference(keys, _.keys(_options.validOptions)),
                banned = _.intersection(keys, _.union(validator.banned, _options.banned)),
                missing = _.difference(_.union(validator.required, _options.required), keys),
                custom = _.union(validator.customValidations, _options.customValidations),
                message;

            if (validator === clonedValidator && arguments.length === 2) {
                logger.error('No validator for ' + validatorName + ' found');
            }

            _.forEach(unrecognized, function (key) {
                message = 'Unrecognized ' + validatorName + ' option for ' + _options.name + ': ' +
                    key + ', value: ' + valueString(deepValue(userOptions, key));
                logger.warn(message);
            });

            _.forEach(banned, function (key) {
                message = 'Illegal ' + validatorName + ' option for ' + _options.name + ': ' +
                    key + ', value: ' + valueString(deepValue(userOptions, key));
                logger.fatal(message);
                throw new Error(message);
            });

            _.forEach(missing, function (key) {
                message = 'Missing required ' + validatorName + ' option for ' + _options.name + ': ' +
                    key + ', value: ' + valueString(deepValue(userOptions, key));
                logger.fatal(message);
                throw new Error(message);
            });

            _.forEach(keys, function (key) {
                var value, type;
                if (_.has(_options.validOptions, key)) {
                    value = deepValue(userOptions, key);
                    type = _options.validOptions[key];

                    if (type === undefined || checkType(type, value)) {
                        return;
                    }

                    message = printError(deepValue(userOptions, key), type, key);
                    logger.warn(message);
                }
            });

            _.forEach(custom, function (v) {
                var result, level;
                if (_.isFunction(v)) {
                    result = {
                        options : userOptions,
                        level : false,
                        message : 'Custom validation step failed'
                    };

                    v(result);

                    if (result.level !== false) {
                        message = result.message;
                        level = logger[result.level] ? result.level : 'warn';
                        logger[level](result.message);
                        if (level === 'fatal') {
                            throw new Error(result.message);
                        }
                    }
                } else {
                    logger.warn('Validator was not a function');
                }
            });

            return !!message;
        }
        that.validate = that;

        // First call will use 'initialize' validator options, second will use 'configure'
        that.autoValidate = function (autoValOptions) {
            if (!_hasValidated) {
                _hasValidated = true;
                return that.validate('initialize', autoValOptions);
            } else {
                return that.validate('configure', autoValOptions);
            }
        };

        /**
         * Creates a new validator, based on the current validator, with the provided new options overriding and/or augmenting existing options.
         *
         * @Example
         *
         *     var myValidator = validator({
         *         name: 'myValidator',
         *         validOptions: { jq: 'jQuery' }
         *     });
         *     var myExtendedValidator = myValidator.withOptions({
         *         validOptions: { dat: 'Date' }
         *     });
         *
         *     // Here, `myExtendedValidator` would recognize both the fields `jq` and `dat` to be jQuery selected objects and Moment objects, respectively.
         *
         * @method withOptions
         * @param newOptions {Object} the new options
         * @returns a new validator with new options overriding old ones
         * @since 2.7.2
         */
        that.withOptions = function withOptions(newOptions) {
            return validatorFactory($.extend(true, {}, _options, newOptions));
        };

        /**
         * Gets the valid options recognized by this validator instance.
         *
         * This returns the map of recognized keys to their valid types, so if all you want are the keys, `_.keys(myValidator.getValidOptions())` should suffice.
         *
         * @method getValidOptions
         * @returns the map of recognized options to their acceptable keys
         * @since 2.7.2
         */
        that.getValidOptions = function getValidOptions() {
            return _.clone(_options.validOptions);
        };

        return that;
    }

    /**
     * See simple example in class description.
     *
     * Simple static method to check a single value against a list of types.
     * Will logger.warn if the test fails.
     *
     * @method checkType
     * @static
     * @returns true if the value passed the test, false otherwise
     */
    validatorFactory.checkType = function (value, type, keyName) {
        if (!checkType(type, value)) {
            keyName = keyName || '';
            logger.warn(printError(value, type, keyName));
            return false;
        }
        return true;
    };

    return validatorFactory;
});


define('primutils/markup/primutils/ui/changeBar/saveControls.hbs', ['handlebars','primutils.markup'], function (H, u) {
var imports = u.buildImported(Array.prototype.slice.call(arguments, 2), 'primutils/ui/changeBar/saveControls.hbs'),
    func = H.template({"compiler":[6,">= 2.0.0-beta.1"],"main":function(depth0,helpers,partials,data) {
  var stack1, lambda=this.lambda, buffer = "<div class=\"save-cancel-controls\" aria-hidden=\"true\" data-error-count=\"0\" style=\"display: none;\">\n    ";
  stack1 = lambda((depth0 != null ? depth0.content : depth0), depth0);
  if (stack1 != null) { buffer += stack1; }
  return buffer + "\n</div>";
},"useData":true});
return function (context, options) {
    options = u.extendOptions(options || {}, imports);
    return func(context, options).trim();
};
});


define('primutils/markup/primutils/ui/changeBar/errorsList.hbs', ['handlebars','primutils.markup','primutils/templates/helpers/button','primutils/templates/helpers/buttonGroup','primutils/templates/helpers/dropdown','primutils/templates/helpers/join'], function (H, u) {
var imports = u.buildImported(Array.prototype.slice.call(arguments, 2), 'primutils/ui/changeBar/errorsList.hbs'),
    func = H.template({"1":function(depth0,helpers,partials,data) {
  return "";
},"compiler":[6,">= 2.0.0-beta.1"],"main":function(depth0,helpers,partials,data) {
  var stack1, lambda=this.lambda, buffer = "<div class=\"errors-list\" id=\"change-bar-error-list-container-";
  stack1 = lambda((depth0 != null ? depth0.changeBarId : depth0), depth0);
  if (stack1 != null) { buffer += stack1; }
  buffer += "\">\n    ";
  stack1 = helpers.button.call(depth0, "error", {"name":"button","hash":{
    'class': ("error-button"),
    'accesskey': ((depth0 != null ? depth0.accesskey : depth0)),
    'aria-relevant': ("text"),
    'aria-live': ("polite"),
    'aria-owns': (helpers._.call(depth0, "change-bar-errors-list-", (depth0 != null ? depth0.changeBarId : depth0), {"name":"_","hash":{},"data":data})),
    'aria-haspopup': ("true"),
    'data-toggle': ("dropdown")
  },"fn":this.program(1, data),"inverse":this.noop,"data":data});
  if (stack1 != null) { buffer += stack1; }
  buffer += "\n    <div class=\"dropdown-menu pull-center\">\n        <ul id=\"change-bar-errors-list-";
  stack1 = lambda((depth0 != null ? depth0.changeBarId : depth0), depth0);
  if (stack1 != null) { buffer += stack1; }
  buffer += "\" data-target=\"#change-bar-error-list-container-";
  stack1 = lambda((depth0 != null ? depth0.changeBarId : depth0), depth0);
  if (stack1 != null) { buffer += stack1; }
  return buffer + "\" class=\"notification-list error-list\" role=\"menu\" aria-region=\"log\" aria-live=\"polite\" aria-relevant=\"additions text\"></ul>\n    </div>\n</div>";
},"useData":true});
return function (context, options) {
    options = u.extendOptions(options || {}, imports);
    return func(context, options).trim();
};
});

/* eslint-disable quote-props */
define('primutils/ui/changeBar/helpers',['lodash', 'logger', 'primutils/utils/language',
    'primutils/utils/validator',
    'primutils/templates/helpers/button',
    'primutils/templates/helpers/icon',
    'primutils/templates/utils/utils',
    'primutils/templates/utils/renderer',
    'primutils/markup/primutils/ui/changeBar/saveControls.hbs',
    'primutils/markup/primutils/ui/changeBar/errorsList.hbs'
], function (_, logger, lang, validator, buttonBase, icon, utils, renderer, hbrSaveControls, hbrErrorsList) {
    'use strict';

    var saveControls = function (options) {
        logger.assert(!!options.fn, 'change-bar-controls must be used as a block helper');
        return hbrSaveControls({
            content : options.fn.call(this, this, options)
        });
    };
    saveControls.handlebarsType = 'helper';
    saveControls.handlebarsName = 'change-bar-controls';

    var errorList = function (type, options) {
        if (arguments.length === 1) { // implying that option 1 is the options hash
            options = type;
            type = 'default';
        }
                else {
            logger.assert(arguments.length === 2, 'change-bar-errors only has one required parameter (the type, either \'default\' or \'simple\', passing in more is an error');
            validator.checkType(type, ['default', 'simple']);
        }
        
        return hbrErrorsList({
            accesskey : type === 'default' ? 'e' : _.get(options, 'hash.accesskey'),
            changeBarId : (this && this.changeBarId) || _.uniqueId('')
        });
    };
    errorList.handlebarsType = 'helper';
    errorList.handlebarsName = 'change-bar-errors';

    var getCancelText = function () { return lang.getString('changeBar.cancel'); };
    var cancelButton = function (type, title, options) {
        if (arguments.length === 1) {
            // only one argument, type IS options
            return buttonBase.call(this, 'secondary', {
                hash : _.defaults({
                    'disabled' : true,
                    accesskey : 'r',
                    'class' : 'cancel-button'
                }, type.hash),
                fn : _.isFunction(type.fn) ? type.fn : getCancelText,
                data : type.data
            });
        }
        if (arguments.length === 2) {
            options = title;
            title = null;
        }

        var mergedOptions = _.defaults({
            hash : _.defaults({
                'disabled' : true,
                'class' : utils.mergeClasses(['cancel-button'], options.hash.class)
            }, options.hash)
        }, options, {
            fn : getCancelText
        });
        return buttonBase.call(this, type, title, mergedOptions);
    };
    cancelButton.handlebarsType = 'helper';
    cancelButton.handlebarsName = 'change-bar-cancel';

    var getSaveText = function (innerSaveTextFn, context) {
        var saveText = innerSaveTextFn ? innerSaveTextFn(context, {}) : lang.getString('changeBar.save');

        var unsaved = renderer.renderElementWithContent('div', {
            'class' : 'unsaved save-button-state'
        }, {}, saveText);

        var saving = renderer.renderElementWithContent('div', {
            'class' : 'saving save-button-state hidden'
        }, {}, icon.call(this, 'throbber', '', { hash : {} }));

        var saved = renderer.renderElementWithContent('div', {
            'class' : 'saved save-button-state hidden'
        }, {}, icon.call(this, 'checkmark-checked', '', { hash : {} }));

        return function () {
            return unsaved + saving + saved;
        };
    };

    var saveButtonInner = function (options) {
        return getSaveText(options.fn, this)();
    };

    saveButtonInner.handlebarsType = 'helper';
    saveButtonInner.handlebarsName = 'change-bar-save-inner';

    var saveButton = function (type, title, options) {
        if (arguments.length === 1) {
            // only one argument, type IS options
            return buttonBase.call(this, 'primary', {
                hash : _.defaults({
                    'disabled' : true,
                    accesskey : 's',
                    'class' : 'save-button'
                }, type.hash),
                fn : getSaveText(type.fn, this),
                data : type.data
            });
        }
        if (arguments.length === 2) {
            options = title;
            title = null;
        }

        var mergedOptions = _.defaults({
            hash : _.defaults({
                'disabled' : true,
                'class' : utils.mergeClasses(['save-button'], options.hash.class)
            }, options.hash),
            fn : getSaveText(options.fn, this)
        }, options);
        return buttonBase.call(this, type, title, mergedOptions);
    };
    saveButton.handlebarsType = 'helper';
    saveButton.handlebarsName = 'change-bar-save';

    var statusText = function () {
        return renderer.renderElement('span', {
            'class' : 'status-message',
            'aria-live' : 'polite'
        });
    };

    statusText.handlebarsType = 'helper';
    statusText.handlebarsName = 'change-bar-status';

    return {
        'change-bar-controls' : saveControls,
        'change-bar-errors' : errorList,
        'change-bar-cancel' : cancelButton,
        'change-bar-save' : saveButton,
        'change-bar-status' : statusText,
        'change-bar-save-inner' : saveButtonInner
    };
});



define('primutils/templates/templateUtilities',[
    'primutils.markup',
    'lodash',
    'handlebars',
    'logger',
    'primutils/templates/helpers/localizer', // This is here because it is used by this module, it is listed twice on purpose

    // All exported helpers and partials must be listed here
    'primutils/templates/helpers/avatar',
    'primutils/templates/helpers/accordion',
    'primutils/templates/helpers/button',
    'primutils/templates/helpers/buttonGroup',
    'primutils/templates/helpers/card',
    'primutils/templates/helpers/caret',
    'primutils/templates/helpers/controlGroup',
    'primutils/templates/helpers/dock',
    'primutils/templates/helpers/dropdown',
    'primutils/templates/helpers/emptyState',
    'primutils/templates/helpers/fieldset',
    'primutils/templates/helpers/icon',
    'primutils/templates/helpers/input',
    'primutils/templates/helpers/join',
    'primutils/templates/helpers/kpi',
    'primutils/templates/helpers/link',
    'primutils/templates/helpers/localizer',
    'primutils/templates/helpers/modal',
    'primutils/templates/helpers/panel',
    'primutils/templates/helpers/placeholder',
    'primutils/templates/helpers/progress',
    'primutils/templates/helpers/search',
    'primutils/templates/helpers/slider',
    'primutils/templates/helpers/spinner',
    'primutils/templates/helpers/steps',
    'primutils/templates/helpers/termDef',
    'primutils/templates/helpers/ternary',
    'primutils/templates/helpers/toggle',
    'primutils/templates/helpers/toolbar',
    'primutils/templates/helpers/wizard',
    'primutils/templates/helpers/chartTooltip',
    'primutils/ui/changeBar/helpers'

], function (hbr, _, Handlebars, logger, localizerHelper) {
    'use strict';

    var importModules = _.drop(_.toArray(arguments), 5),
        registered = false;

    /**
     * Utility functions for exporting Handlebars templates from primutils
     *
     * @class templates.templateUtilities
     * @module templates
     */

    return {
        /**
         * By default primutils will not register its Handlebars helpers and partials globally.
         * Calling this function assigns all of the known primutils helpers and partials globally to Handlebars.
         *
         * This function can only be called once.
         * Additional calls to this function will not do anything.
         *
         * @method registerImports
         */
        registerImports : function () {
            if (registered) {
                return;
            }

            var imports = hbr.buildImported(importModules, 'primutils');

            _.forEach(imports.helpers, function (fn, name) {
                Handlebars.registerHelper(name, fn);
            });

            _.forEach(imports.partials, function (fn, name) {
                Handlebars.registerPartial(name, fn);
            });

            registered = true;
        },

        /**
         * Handlebars provides no built-in checking to make sure that globally registered helpers and partials do not conflict with previously registered global handlers.
         * This method can be called to throw a warning in the console when a duplicate handler is attempted to register.
         *
         * To completely prevent duplicate handlers from overwrite each other, pass `true`.
         * Otherwise the new handler will overwrite the old one.
         *
         * Note: This does not prevent individual templates from "shadowing" global templates by using the "deps" option.
         * See the <a href="#">markup plugin</a> for more information.
         *
         * @method protectHandlebarsImports
         * @param [noOverwrite] {Boolean} Prevents a second handler with the same name from being registered at all.
         */
        protectHandlebarsImports : function (noOverwrite) {
            Handlebars.registerHelper = _.wrap(Handlebars.registerHelper, function (fn, name) {
                var names = _.isObject(name) ? _.keys(name) : [name],
                    existingHelpers = _.intersection(names, _.keys(Handlebars.helpers)),
                    hasDupes = existingHelpers.length > 0;

                if (hasDupes) {
                    logger[noOverwrite ? 'error' : 'warn']('The following helpers are already registered globally' +
                        (noOverwrite ? ' and will not be overwritten' : '') +
                        ': ' + existingHelpers.join(', '));
                }

                if (!noOverwrite || !hasDupes) {
                    return fn.apply(this, _.rest(_.toArray(arguments)));
                } else {
                    throw new Error('Cannot override previously registered helpers: ' + existingHelpers.join(', '));
                }
            });

            Handlebars.registerPartial = _.wrap(Handlebars.registerPartial, function (fn, name) {
                var names = _.isObject(name) ? _.keys(name) : [name],
                    existingPartials = _.intersection(names, _.keys(Handlebars.partials)),
                    hasDupes = existingPartials.length > 0;

                if (hasDupes) {
                    logger[noOverwrite ? 'error' : 'warn']('The following partials are already registered globally' +
                        (noOverwrite ? ' and will not be overwritten' : '') +
                        ': ' + existingPartials.join(', '));
                }

                if (!noOverwrite || !hasDupes) {
                    return fn.apply(this, _.rest(_.toArray(arguments)));
                } else {
                    throw new Error('Cannot override previously registered helpers: ' + existingPartials.join(', '));
                }
            });
        },

        /**
         * Sets the localizer to use in the localizer helper.
         *
         * @method setLocalizerForLocalizerHelper
         * @param localizer {Localizer} The localizer to use in the localizer helper
         */
        setLocalizerForLocalizerHelper : function (localizer) {
            // Type checking is done inside the helper
            localizerHelper.useLocalizer(localizer);
        }
    };
});


/**
 * A simple eventing system which supports namespacing.
 *
 *     var s = new Subscribable();
 *     s.on('test', function () { console.log('test'); }
 *     s.trigger('test');
 *     s.off('test');
 *
 * Other classes can inherit from Subscribable like below. It will create variables on your object to track listeners and options.
 *
 *     function MySubscribable () {
 *         // ...
 *         Subscribable.call(this, options); // options can be null/undefined
 *         // ...
 *     }
 *
 *     MySubscribable.prototype = Object.create(Subscribable.prototype);
 *     MySubscribable.prototype.otherMethod = function () { ... };
 *
 * ----------
 *
 * **Warning**: in new versions of `Subscribable`, attaching listeners to the same namespace/event is NOT a replacement operation.
 * In other words, if you do `sub.on('ns.ev', ...).on('ns.ev', ...)`, both listeners will be triggered.
 * (This is to better match the behavior of jQuery events, and occurred during a rework of the Subscribable internals to achieve better performance.)
 *
 * @class utils.Subscribable
 * @module utils
 * @constructor
 */
define('primutils/utils/Subscribable',['lodash', 'primutils/utils/validator', 'logger'], function (_, validator, logger) {
    'use strict';

    var splitCache = {};
    function normalize(eventName) {
        if (splitCache[eventName] == null) {
            var split = eventName.split('.');
            if (split.length === 1) {
                splitCache[eventName] = { name : eventName };
            } else {
                splitCache[eventName] = { name : split[1], namespace : split[0] };
            }
        }
        return splitCache[eventName];
    }

    function addEvent(sub, ev, cb, ctx, opts) {
        var events = ev.split(' ');

        _.forEach(events, function (event) {
            var normalized = normalize(event);

                        if (sub._subscribableOptions.allowedEvents && !checkEventValidity.call(sub, normalized.namespace, normalized.name)) {
                return;
            }
            
            sub._listeners[normalized.name] = sub._listeners[normalized.name] || [];
            sub._listeners[normalized.name].push(_.assign({ fn : cb, context : ctx }, normalized, opts));
        });
    }

    /**
     * Without first elem - assumed to be event name, and read from arguments directly in caller
     */
    function argumentsSlice() {
        var len = arguments.length;
        var args = new Array(len - 1);

        for (var i = 1; i < len; ++i) {
            args[i - 1] = arguments[i];
        }

        return args;
    }

    var removeAllListeners = function () {
        var listenerMap = this._listeners;
        _.forEach(listenerMap, function (listener, key) {
            delete listenerMap[key];
        });
    };

    function Subscribable(options) {
        logger.assert(this && this instanceof Subscribable, 'Subscribable must be created with \'new\'');

        /**
         * The map of event names to their proper functions.
         *
         * @example
         *      var _subscribableListeners = [
         *          { name : 'select', fn : callback1 },
         *          { name : 'select', fn : callback2 },
         *          { name : 'select', namespace : 'view', fn : callBack },
         *          { name : 'select', namespace : 'customNameSpace', fn : callBack }
         *      ];
         *
         * @property _listeners
         * @type {Array}
         * @private
         * @default {}
         */
        Object.defineProperty(this, '_listeners', {
            value : {},
            writable : false,
            enumerable : false
        });

        options = options || {};
        validator.checkType(options, ['Object']);
        this._validator(options);
        Object.defineProperty(this, '_subscribableOptions', {
            value : _.assign({}, options),
            writable : false,
            enumerable : false
        });
    }

    var checkEventValidity = function (evNs, evName) {
        if (!_.includes(this._subscribableOptions.allowedEvents, evName)) {
            logger.error('The event spec: ' + evName + ' is not allowed for this Subscribable.  Valid events are: [' + this._subscribableOptions.allowedEvents.join(',') + ']');
            return false;
        }

        return true;
    };

    Subscribable.prototype = {
        constructor : Subscribable,

        /**
         * Calls each function that matches the argument.
         * Argument can either be namespace.event, or just event
         *
         * If argument is just event name than all functions will be called otherwise
         * only listeners in the namespace will be called.
         *
         * Chainable since 2.7.3.
         *
         * @method trigger
         * @param {String} eventName Either a string in the form of `namespace.event` or `event`
         * @chainable
         */
        trigger : function () {
            var args = argumentsSlice.apply(null, arguments);
            var argsLen = args.length;
            var eventArg = arguments[0];
            var normalized = normalize(eventArg);
            var normalizedName = normalized.name;
            var normalizedNs = normalized.namespace;

                        if (this._subscribableOptions.allowedEvents) {
                var validEvent = checkEventValidity.call(this, normalizedNs, normalizedName);
                if (!validEvent) {
                    return this;
                }
            }
            
            var listeners = this._listeners[normalizedName];
            if (listeners) {
                for (var i = 0, len = listeners.length; i < len; i++) {
                    var listener = listeners[i];
                    if (!normalizedNs || listener.namespace === normalizedNs) {
                        if (listener.once) {
                            listeners.splice(i, 1);
                            i--;
                            len--;
                        }

                        var ctx = listener.context;
                        if (argsLen === 0) { listener.fn.call(ctx); } else if (argsLen === 1) { listener.fn.call(ctx, args[0]); } else if (argsLen === 2) { listener.fn.call(ctx, args[0], args[1]); } else if (argsLen === 3) { listener.fn.call(ctx, args[0], args[1], args[2]); } else { listener.fn.apply(listener.context, args); }
                    }
                }
            }

            if (this._listeners['*'] != null) {
                args = [eventArg].concat(args);

                _.forEach(this._listeners['*'], function (listener) {
                    listener.fn.apply(listener.context, args);
                });
            }

            return this;
        },

        /**
         * Registers a listener for the event name possibly under the given namespace.
         * Unlike the basic [`on`](#method_on) method, the listener attached with this method is guaranteed to only fire once.
         *
         * If no name space if given then it goes under the array 'noNamespace'.
         *
         * @method one
         * @chainable
         * @param {String} eventName Either a string in the form of `namespace.event` or `event`
         * @param {Function} callBack The function to call when triggered.
         * @param {Object} [context] The context to use when calling the callBack.
         */
        one : function (eventName, callBack, context) {
            var that = this;

            if (_.isObject(eventName)) {
                //Given a map
                //Ignore callback and context
                _.forEach(eventName, function (value, key) {
                    if (_.isFunction(value)) {
                        addEvent(that, key, value, undefined, { once : true });
                    } else {
                        addEvent(that, key, value.fn, value.context, { once : true });
                    }
                });
            } else if (_.isString(eventName)) {
                addEvent(that, eventName, callBack, context, { once : true });
            }

            return this;
        },

        /**
         * Registers a listener for the event name possibly under the given namespace.
         *
         * If no name space if given then it goes under the array 'noNamespace'.
         *
         * '*' as the event name is a special property that will be
         * fired for any event called.
         *
         * @method on
         * @chainable
         * @param {String} eventName Either a string in the form of `namespace.event` or `event`
         * @param {Function} callBack The function to call when triggered.
         * @param {Object} [context] The context to use when calling the callBack.
         */
        on : function (eventName, callBack, context) {
            var that = this;

            if (_.isObject(eventName)) {
                //Given a map
                //Ignore callback and context
                _.forEach(eventName, function (value, key) {
                    if (_.isFunction(value)) {
                        addEvent(that, key, value, undefined);
                    } else {
                        addEvent(that, key, value.fn, value.context);
                    }
                });
            } else if (_.isString(eventName)) {
                addEvent(that, eventName, callBack, context);
            }

            return this;
        },

        /**
         * Removes the function(s) for the given `namespace.event`.
         *
         * If no namespace is given - that is, the passed eventName does not include a '.' - AND no function is provided - all listeners for the given eventname, under any namespace, are removed.
         *
         * (Since 2.7.3) If a second parameter, the callback function, is provided, AND the eventName does not include a namespace, then the given function will be removed from the list of listeners not gathered under specific namespaces.
         *
         * @method off
         * @chainable
         * @param {String} eventName Either a string in the form of `namespace.event` or `event`
         * @param {Function} [fn] a specific function to remove from the list of non-namespaced listeners
         */
        off : function (eventName, fn) {
            var that = this,
                events = eventName.split(' '),
                hasFnToRemove = arguments.length === 2;

            _.forEach(events, function (eventName) {
                var normalized = normalize(eventName),
                    eventKeys = _.keys(normalized);

                if (hasFnToRemove) {
                    normalized = _.clone(normalized);
                    normalized.fn = fn;
                    eventKeys.push('fn');
                    eventKeys.push('namespace');
                }

                                if (that._subscribableOptions.allowedEvents && !checkEventValidity.call(that, normalized.namespace, normalized.name)) {
                    return;
                }
                
                that._listeners[normalized.name] = _.reject(that._listeners[normalized.name], function (listener) {
                    return _.isEqual(normalized, _.pick(listener, eventKeys));
                });
            });

            return this;
        },

        /**
         * Removes all events listeners tracked by this Subscribable.
         *
         * @method clear
         * @chainable
         * @since 2.0.8
         */
        clear : function () {
            if (!arguments.length) {
                removeAllListeners.call(this);
            }

            return this;
        },

        /**
         * The validator for this Subscribable.
         *
         * @property _validator
         * @protected
         * @since 2.7.2
         */
        _validator : validator({
            name : 'Subscribable',
            validOptions : {
                /**
                 * An array of event names that are valid for `on`/`off`/`trigger` method calls.
                 * If this option is present, any attempts to call any methods on this Subscribable for events not listed in the array will trigger logger.error calls and be ignored.
                 *
                 * @property allowedEvents
                 * @public
                 * @type {Array|undefined}
                 * @default undefined
                 */
                allowedEvents : ['Array', undefined]
            }
        })
    };

    /**
     * Alias for `.on`.
     *
     * @method subscribe
     * @chainable
     * @since 2.7.3
     */
    Subscribable.prototype.subscribe = Subscribable.prototype.on;

    /**
     * Alias for `.off`.
     *
     * @method unsubscribe
     * @chainable
     * @since 2.7.3
     */
    Subscribable.prototype.unsubscribe = Subscribable.prototype.off;

    /**
     * Alias for `.trigger`.
     *
     * @method publish
     * @chainable
     * @since 2.7.3
     */
    Subscribable.prototype.publish = Subscribable.prototype.trigger;

    return Subscribable;
});

/**
 * @module utils
 */
define('primutils/utils/objectUtils',['logger', 'lodash', 'moment'], function (logger, _, moment) {
    'use strict';

    var momentProto = Object.getPrototypeOf(moment()),
        warnedMoment = false;

    function checkForAndCloneMoment(a) {
        if (moment.isMoment(a)) {
            if (momentProto.isPrototypeOf(a)) {
                return a.clone();
            } else {
                                if (!warnedMoment) {
                    logger.warn('Moments should be cloned with the moment .clone method or though "primutils/utils/objectUtils" utilities');
                    warnedMoment = true;
                }
                                return moment(a);
            }
        } else if (moment.isDuration(a)) {
            return moment.duration(a); // durations do not have a clone method
        }
        // DURATIONS CLONED WITHOUT THIS METHOD WILL BE WRONG
        // BUT DURATIONS CLONED WITH JQUERY BE DARNED....
    }

    function makeCloneCallback(callback) {
        return _.isFunction(callback) ? _.wrap(callback, function (origCallback, a) {
            var origResult = origCallback.call(this, a);
            if (!_.isUndefined(origResult)) {
                return origResult;
            }
            return checkForAndCloneMoment.call(this, a);
        }) : checkForAndCloneMoment;
    }

    function isMomentOrDuration(value) {
        return moment.isMoment(value) || moment.isDuration(value);
    }

    function checkForAndMergeMoment(ignored, a) {
        if (moment.isMoment(a) || moment.isDuration(a)) {
            return a; // Inform lodash to not recurse into moment objects
        }
    }

    function makeMergeCallback(callback) {
        return _.isFunction(callback) ? _.wrap(callback, function (origCallback, value, source) {
            var origResult = origCallback.call(this, value, source);
            if (!_.isUndefined(origResult)) {
                return origResult;
            }
            return checkForAndMergeMoment.call(this, value, source);
        }) : checkForAndMergeMoment;
    }

    function assertUniqueKeys(dest, sources) {
                var destKeys = _.keys(dest);
        _.forEach(sources, function (src) {
            var srcKeys = _.keys(src);
            logger.assert(_.intersection(destKeys, srcKeys).length === 0, 'dest and one or more src objects had duplicated keys');
        });
            }

    var production = true;
        production = false;
    
    /**
     * @class utils.objectUtils
     */
    var api = {
        /**
         * Creates a clone of value.
         * If `isDeep` is true nested objects will also be cloned, otherwise they will be assigned by reference.
         * If a callback is provided it will be executed to produce the cloned values.
         * If the callback returns undefined cloning will be handled by the method instead.
         *
         * The callback is bound to `thisArg` and invoked with one argument; `(value)`.
         *
         * This version is exactly the same as the version in Lo-Dash, except that it has built-in support for moment objects using moments native `clone` function.
         * The moment check occurs after the provided callback, so you can override the default moment comparison as desired.
         *
         * @method clone
         * @param {*} value The value to clone
         * @param {Boolean} [isDeep] Specify a deepClone
         * @param {Function} [callback] The function to customize cloning values
         * @param {*} [thisArg] The `this` binding of the `callback`
         * @return {*} the cloned value
         */
        clone : _.wrap(_.clone, function (origFn, value, isDeep, callback, thisArg) {
            // isDeep is optional
            if (typeof isDeep !== 'boolean' && isDeep != null) {
                thisArg = callback;
                callback = isDeep;
                isDeep = false;
            }

            callback = makeCloneCallback(callback);

            return origFn.call(this, value, isDeep, callback, thisArg);
        }),

        /**
         * Creates a deep clone of value.
         * If a callback is provided it will be executed to produce the cloned values.
         * If the callback returns undefined cloning will be handled by the method instead.
         *
         * The callback is bound to `thisArg` and invoked with one argument; `(value)`.
         *
         * This version is exactly the same as the version in Lo-Dash, except that it has built-in support for moment objects using moments native `clone` function.
         * The moment check occurs after the provided callback, so you can override the default moment comparison as desired.
         *
         * @method cloneDeep
         * @param {*} value The value to clone
         * @param {Function} [callback] The function to customize cloning values
         * @param {*} [thisArg] The `this` binding of the `callback`
         * @return {*} the deep cloned value
         */
        cloneDeep : _.wrap(_.cloneDeep, function (origFn, value, callback, thisArg) {
            return origFn.call(this, value, makeCloneCallback(callback), thisArg);
        }),

        /**
         * Assigns own enumerable properties of source object(s) to the destination object.
         * Subsequent sources will overwrite property assignments of previous sources.
         * If a callback is provided it will be executed to produce the assigned values.
         *
         * The callback is bound to `thisArg` and invoked with two arguments; `(objectValue, sourceValue)`.
         *
         * In production builds, the function is exactly the Lo-Dash extend method.
         * In debug builds, it will warn if the user attempts to clone a moment using this function.
         *
         * @method extend
         * @param {Object} object The destination object
         * @param {...Object} [source] The source objects
         * @param {Function} [callback] The function to customize cloning values
         * @param {*} [thisArg] The `this` binding of the `callback`
         * @return {*} the destination object
         */
        extend : production ? _.extend : _.wrap(_.extend, function (origFn) {
            var args = _.rest(arguments),
                sources = _.rest(args),
                anyMoment = _.some(sources, isMomentOrDuration);
            if (anyMoment) {
                logger.warn('Moments should be cloned with the moment .clone method or though "primutils/utils/objectUtils" utilities');
            }
            return origFn.apply(this, args);
        }),

        /**
         * As [`_.extend`](http://devdocs.io/lodash/index#assign), but first checks that all the keys of the destination object are not present in any of the source objects.
         *
         * **Note**: this does /not/ perform a deep check on any keys, and does /not/ check that the keys from all source objects are unique against each other.
         *
         * @method safeExtend
         * @return {*} the destination object
         */
        safeExtend : function safeExtend() {
                        var dest = arguments[0];
            var sources = _.rest(arguments);
            assertUniqueKeys(dest, sources);
                        return _.extend.apply(_, arguments);
        },

        /**
         * Assigns own enumerable properties of source object(s) to the destination object for all destination properties that resolve to undefined.
         * Once a property is set, additional defaults of the same property will be ignored.
         *
         * In production builds, the function is exactly the Lo-Dash extend method.
         * In debug builds, it will warn if the user attempts to clone a moment using this function.
         *
         * @method defaults
         * @param {Object} object The destination object
         * @param {...Object} [source] The source objects
         * @return {*} the destination object
         */
        defaults : production ? _.defaults : _.wrap(_.defaults, function (origFn) {
            var args = _.rest(arguments),
                sources = _.rest(args),
                anyMoment = _.some(sources, isMomentOrDuration);
            if (anyMoment) {
                logger.warn('Moments should be cloned with the moment .clone method or though "primutils/utils/objectUtils" utilities');
            }
            return origFn.apply(this, args);
        }),

        /**
         * Recursively merges own enumerable properties of the source object(s), that don't resolve to undefined into the destination object.
         * Subsequent sources will overwrite property assignments of previous sources for values that are "value type" (leaves of the object's tree).
         * If a callback is provided it will be executed to produce the merged values of the destination and source properties.
         * If the callback returns undefined merging will be handled by the method instead.
         *
         * The callback is bound to `thisArg` and invoked with two arguments; `(objectValue, sourceValue)`.
         *
         * @method extendDeep
         * @param {Object} object The destination object
         * @param {...Object} [source] The source objects
         * @param {Function} [callback] The function to customize cloning values
         * @param {*} [thisArg] The `this` binding of the `callback`
         * @return {*} the destination object
         */
        extendDeep : _.wrap(_.merge, function (origFn) {
            var args = _.rest(arguments);
            if (args.length > 3 && _.isFunction(args[args.length - 2])) {
                args[args.length - 2] = makeMergeCallback(args[args.length - 2]);
            } else if (args.length > 2 && _.isFunction(args[args.length - 1])) {
                args[args.length - 1] = makeMergeCallback(args[args.length - 1]);
            } else {
                args.push(makeMergeCallback());
            }

            return origFn.apply(this, args);
        }),

        cloneWithOriginals : (function () {
            var SKIPPED_MARKER = new (function Skipped() {})(); // jshint ignore:line
            var cloneWithOriginals = function (newObj, oldObj) {
                return _.reduce(newObj, function nestSkippedCheck(memo, value, key) {
                    if (value === SKIPPED_MARKER) {
                        memo.result[key] = memo.currentObj[key];
                    } else if (_.isArray(value)) {
                        memo.result[key] = _.reduce(value, nestSkippedCheck, {
                            result : new Array(value.length),
                            currentObj : memo.currentObj[key] || []
                        }).result;
                    } else if (_.isObject(value)) {
                        memo.result[key] = cloneWithOriginals(value, memo.currentObj[key] || {});
                    } else {
                        memo.result[key] = value;
                    }
                    return memo;
                }, {
                    result : {},
                    currentObj : oldObj
                }).result;
            };
            cloneWithOriginals.markSkipped = _.constant(SKIPPED_MARKER);
            return cloneWithOriginals;
        }()),

        /**
         * Overrides the Lo-Dash default implementation of the methods provided by this utility to use the Primutils versions.
         *
         * @method mixinToLodash
         * @protected
         * @static
         */
        mixinToLodash : function () {
            _.mixin({
                clone : api.clone,
                cloneDeep : api.cloneDeep,

                assign : api.extend,
                extend : api.extend,

                merge : api.extendDeep,
                extendDeep : api.extendDeep,

                defaults : api.defaults,
                defaultsDeep : api.defaultsDeep
            });
        }
    };

    /**
     * Assigns own enumerable properties of source object(s) to the destination object for all destination properties that resolve to undefined.
     * Once a property is set, additional defaults of the same property will be ignored for values that are "value type" (leaves of the object's tree).
     *
     * @method defaultsDeep
     * @param {Object} object The destination object
     * @param {...Object} [source] The source objects
     * @return {*} the destination object
     */
    api.defaultsDeep = _.partialRight(api.extendDeep, function deep() {
        var args = _.toArray(arguments);
        args.push(deep);
        return api.extendDeep.apply(api, args);
    });

    // Aliases
    /**
     * Alias of [`extend`](#method_extend)
     * @method assign
     */
    api.assign = api.extend;

    /**
     * Alias of [`extendDeep`](#method_extendDeep)
     * @method merge
     */
    api.merge = api.extendDeep;

    return api;
});


define('primutils/markup/primutils/ui/changeBar/modalBackdrop.hbs', ['handlebars','primutils.markup'], function (H, u) {
var imports = u.buildImported(Array.prototype.slice.call(arguments, 2), 'primutils/ui/changeBar/modalBackdrop.hbs'),
    func = H.template({"compiler":[6,">= 2.0.0-beta.1"],"main":function(depth0,helpers,partials,data) {
  return "<div class=\"modal-backdrop fade\"></div>";
  },"useData":true});
return function (context, options) {
    options = u.extendOptions(options || {}, imports);
    return func(context, options).trim();
};
});

define('primutils/a11y/messageAnnouncer',['jquery', 'primutils/utils/strings'], function ($, strings) {
    'use strict';

    var template = strings.template('<div style="position: fixed; top: -10000px; left: -10000px;" class="announce-alert" role="alert" aria-live="assertive" aria-relevant="all" aria-atomic="true"><%=message%></div>');

    return {
        announce : function (message) {
            var $announceDiv = $(template({ message : message })).appendTo(document.body);
            setTimeout(function () {
                $announceDiv.remove();
            }, 100);
        }
    };
});


define('primutils/utils/grid',['jquery', 'lodash'], function ($, _) {
    'use strict';

    var GRID_PSEUDO_SELECTOR = 'pgbu-grid';

    var gridLoadedDfd = $.Deferred(),
        gridLoadedPromise = gridLoadedDfd.promise();

    $(window).one('gridapi-loaded', function () {
        gridLoadedDfd.resolve();
    });

    var that = {
        get gridDataName() {
            return GRID_PSEUDO_SELECTOR;
        },
        _pseudoSelectorDefined : function (pseudo) {
            return _.has($.expr[':'], pseudo);
        },
        gridApiLoaded : function () {
            return that._pseudoSelectorDefined(GRID_PSEUDO_SELECTOR);
        },
        whenGridApiLoaded : function (callback) {
            return gridLoadedPromise.done(callback);
        },
        selectGrids : function () {
            if (!that.gridApiLoaded()) {
                return $([]);
            }
            return $(':' + GRID_PSEUDO_SELECTOR);
        },
        findGrids : function ($parent) {
            if (!that.gridApiLoaded()) {
                return $([]);
            }
            return $parent.find(':' + GRID_PSEUDO_SELECTOR);
        },
        filterGrids : function ($collection) {
            if (!that.gridApiLoaded()) {
                return $([]);
            }
            return $collection.filter(':' + GRID_PSEUDO_SELECTOR);
        }
    };
    return that;
});

define('primutils/utils/emulateTransitionEnd',['jquery', 'bootstrap'], function ($) { //Require that Bootstrap be loaded first for $.support.transition.end
    'use strict';

    // Based on: http://api.jquery.com/jQuery.cssHooks/
    function venderPrefix(prop, style) {
        var vendorProp,
            // Capitalize first character of the prop to test vendor prefix
            capProp = prop.charAt(0).toUpperCase() + prop.slice(1),
            prefixes = ['Moz', 'Webkit', 'O', 'ms'];

        if (prop in style) { // Browser supports standard CSS property name
            return prop;
        } else {
            // Otherwise test support for vendor-prefixed property names
            for (var i = 0; i < prefixes.length; i++) {
                vendorProp = prefixes[i] + capProp;
                if (vendorProp in style) {
                    return vendorProp;
                }
            }
        }
    }

    function parseToMS(value) {
        value = value.trim();
        if (value.slice(-2) === 'ms') {
            return parseFloat(value);
        } else if (value.slice(-1) === 's') {
            return 1000 * parseFloat(value);
        }
    }

    function transitionDurationMS($elem, defaultValue) {
        var venderTransitionDuration = venderPrefix('TransitionDuration', $elem.get(0).style),
            value = $elem.css(venderTransitionDuration);
        if (value.indexOf(',') > -1) {
            return defaultValue;
        } else {
            return parseToMS(value) || defaultValue;
        }
    }

    // Based on: http://blog.alexmaccaw.com/css-transitions
    $.fn.emulateTransitionEnd = function (guessDuration) {
        if ($.support.transition) {
            var $node = $(this), called = false;
            $node.one($.support.transition.end, function () { called = true; });
            // Add 25ms to guessed duration to give slow animations time to finish
            setTimeout(function () {
                if (!called) {
                    $node.trigger($.support.transition.end);
                }
            }, transitionDurationMS($node, guessDuration ? (guessDuration + 25) : 10000));
        }
    };
});

/*eslint no-unused-expressions:0, no-use-before-define:0, no-unused-vars:1*/
define('primutils/ui/changeBar/ChangeBarView',['jquery', 'lodash',
    'primutils/utils/language', 'logger',
    'primutils/markup/primutils/ui/changeBar/modalBackdrop.hbs',
    'primutils/a11y/messageAnnouncer',
    'primutils/utils/keyCode', 'primutils/utils/grid',
    'primutils/utils/emulateTransitionEnd'
], function ($, _, lang, logger, modalBackdrop, messageAnnouncer, keyCode, gridUtils) {
    'use strict';

    // why?  https://jira.oraclecorp.com/jira/browse/PRIMCR-404
    //  debouncing save after click aloows grid validation, etc, to take place
    //  before trying to process the save
    var SAVE_DEBOUNCE_TIME = 25,
        BACKDROP_FADE_TIME = 100, // See modals.less (.1s in Blue Sky 3)
        SAVED_STATE_VISIBLE_TIMEOUT = 2000;

    var SAVE_STATES = {
        UNSAVED : 'unsaved',
        SAVING : 'saving',
        SAVED : 'saved'
    };

    var getValue = function (opt) {
        return _.isFunction(opt) ? opt.apply(this, _.rest(_.toArray(arguments))) : opt;
    };

    var removeListeners = function () {
        this.$errorList.off('click', 'li').off('scroll', 'li').off('keydown', 'li');
        this.buttons.$saveButton.off('click');
        this.buttons.$cancelButton.off('click');
        this.buttons.$errorButton.off('click');
    };

    var isWithoutTopmostModal = function () {
        var $modal = $('.modal.in').last(); // purposely not doing .tray.in
        if (!$modal.length) {
            return false;
        }

        var $cb = this.$el;
        if ($cb.parents().filter($modal).length) {
            return false;
        }

        return true;
    };

    var initListeners = function () {
        var that = this,
            controller = this.controller,
            refocus = null,
            debouncedSave;

        this.$errorList.on('click', 'li', function (e) {
            e.preventDefault();
            var error = $(this).data('error');
            controller._selectError(error);
        }).on('scroll', function () {
            scrollErrorsList.call(that);
            if (refocus === 'first') {
                that.$errorList.find('li:first-child a').focus();
            } else if (refocus === 'last') {
                that.$errorList.find('li').last().find('a').focus();
            }
            refocus = null;
        }).on('keydown', 'li', function (e) {
            var $this = $(this),
                preventDefault = true;

            if (e.keyCode === keyCode.UP) {
                if ($this.is(':first-child')) {
                    refocus = 'first';
                    that.$errorList.scrollTop(that.$errorList.scrollTop() - that.virtualHeights.errorItem);
                } else {
                    $this.prev().children('a').focus();
                }
            } else if (e.keyCode === keyCode.DOWN) {
                if ($this.next().is(':last-child')) { // .virtual-scroller is last child
                    refocus = 'last';
                    that.$errorList.scrollTop(that.$errorList.scrollTop() + that.virtualHeights.errorItem);
                } else {
                    $this.next().children('a').focus();
                }
            } else if (e.keyCode === keyCode.PAGEUP) {
                refocus = 'first';
                that.$errorList.scrollTop(that.$errorList.scrollTop() - that.virtualHeights.errorList);
            } else if (e.keyCode === keyCode.PAGEDOWN) {
                refocus = 'last';
                that.$errorList.scrollTop(that.$errorList.scrollTop() + that.virtualHeights.errorList);
            } else if (e.keyCode === keyCode.HOME) {
                refocus = 'first';
                that.$errorList.scrollTop(0);
            } else if (e.keyCode === keyCode.END) {
                refocus = 'last';
                that.$errorList.scrollTop(that.$errorList.get(0).scrollHeight);
            } else if (e.keyCode === keyCode.TAB && !e.shiftKey && $this.next().is(':last-child') && !$this.is(':first-child')) { // .virtual-scroller is last child
                preventDefault = false;
                that.$errorList.closest('.open').removeClass('open');
            } else {
                return;
            }

            if (preventDefault) {
                e.preventDefault();
                e.stopPropagation();
            }
        });

        this.buttons.$errorButton.on('click', function (e) {
            if (!that._repopulatedOnce) {
                repopulateErrorsList.call(that);
            } else {
                setupVirtualization.call(that);
            }
            _.defer(function () {
                that.$errorList.scrollTop(that._resetScrollPosition ? 0 : that._lastErrorScrollPosition);
                that._resetScrollPosition = false;
            });
        });

        this.$el.on('click', '[data-linkdisabled="save"]', function (e) {
            if (controller._unhideAnyErrors()) {
                e.stopImmediatePropagation();
            }
        });

        function baseHandleSave() {
            if (!getSaveEnabled.call(that)) {
                return;
            }
            controller._save();
        }
        debouncedSave = _.debounce(baseHandleSave, SAVE_DEBOUNCE_TIME);
        this.buttons.$saveButton.on('click', function (e) {
            if (isWithoutTopmostModal.call(that)) {
                // PPB-7037 skip save if a modal is up
                return;
            }

            // over-opinionated?  seems reasonable but...
            if (!that.buttons.$saveButton.is(':focus')) {
                that.buttons.$saveButton.focus();
            }

            // want to do this before the debounce, so grid/page have a chance to work
            if (controller._options.closeGridEditorsOnSave && gridUtils.gridApiLoaded()) {
                var $grids = gridUtils.selectGrids();
                if ($grids && $grids.length) {
                    $grids.grid('closeEditor', true);
                }
            }

            debouncedSave();
        });

        this.buttons.$cancelButton.on('click', function (e) {
            if (isWithoutTopmostModal.call(that)) {
                // PRIMCR-1885 if cancel dialog is already up...
                return;
            }

            controller._cancel();
        });

        this.$el
            .on('mousedown', function () {
                that._suspendRedraws = true;
            })
            .on('mouseup', function () {
                that._suspendRedraws = false;
                if (that._redrawRequested) {
                    that._redrawRequested = false;
                    that.redraw();
                }
            });
    };

    var initElement = function () {
                logger.assert(this.controller._options.changeBarMarkup || this.controller._options.fromExisting, 'ChangeBar must be instantiated with either DOM provided or with the fromExisting option');
                var $el = false;
        if (this.controller._options.fromExisting) {
            var $fromExisting = $(getValue.call(this, this.controller._options.fromExisting));
            if ($fromExisting.is('.save-cancel-controls')) {
                $el = $fromExisting.filter('.save-cancel-controls');
                logger.assert($el.length === 1, 'when using ChangeBar fromExisting, change-bar-controls to attach to must be unambiguous, but more than one was found');
            } else if ($fromExisting.has('.save-cancel-controls')) {
                $el = $fromExisting.find('.save-cancel-controls');
                logger.assert($el.length === 1, 'when using ChangeBar fromExisting, change-bar-controls to attach to must be unambiguous, but more than one was found');
            } else {
                var message = 'The `fromExisting` option was specified, but the provided selector doesn\'t appear to contain a save cancel bar created with the change-bar-controls helper.';
                logger.fatal(message);
                throw new Error(message);
            }
        } else {
            $el = $(this.controller._options.changeBarMarkup);
            if (!this.controller._options.useAccessKeys) {
                $el.find('[accesskey]').removeAttr('accesskey');
            }
        }
        this.$el = $el;
        this.$el.toggleClass('fade', this.controller._options.fadeChangeBar);

        this._savingButtonClass = _.compact(_.without((this.$el.find('.save-button').attr('class') || '').split(' '), 'btn', 'save-button')).join(' ');

        this.$modalBackdrop = $(modalBackdrop());

        this.clearStatusMessage();
    };

    var setupVirtualization = function () {
        var $dropdownContainer, isOpen, $errorItem;

        if (!this.virtualHeights) {
            this.virtualHeights = {
                canVirtualize : this.$errorList.css('max-height') !== 'none',
                hasValidErrorItemHeight : false,
                errorList : parseInt(this.$errorList.css('max-height'), 10)
            };
            this.virtualHeights.errorItem = this.virtualHeights.errorList;
        }

        if (this.$el.css('display') !== 'none' && !this.virtualHeights.hasValidErrorItemHeight && this.virtualHeights.canVirtualize) {
            $dropdownContainer = this.buttons.$errorButton.is('[data-toggle=dropdown]') ? this.buttons.$errorButton.parent() : $();
            isOpen = $dropdownContainer.hasClass('open');
            $errorItem = makeErrorItem.call(this, { message : 'sizing' }).appendTo(this.$errorList);

            $dropdownContainer.addClass('open');

            this.virtualHeights.errorItem = parseInt($errorItem.outerHeight(true), 10);
            this.virtualHeights.hasValidErrorItemHeight = true;

            $errorItem.remove();
            $dropdownContainer.toggleClass('open', isOpen);
        }
    };

    var repopulateErrorsList = function () {
        var that = this,
            errors = this.controller.getVisibleErrors();

        this._repopulatedOnce = true;

        setupVirtualization.call(this);

        if (this.virtualHeights.canVirtualize) {
            that._resetScrollPosition = true;
            scrollErrorsList.call(that);
        } else {
            this.$errorList.empty();
            _.forEach(errors, function (error) {
                that.$errorList.append(makeErrorItem.call(that, error));
            });
        }
    };

    // Makes save and cancel buttons the same width so switching the save state doesn't affect the save button's size
    // Also do this to the cancel button to make the change bar appear "even"
    var fixSaveButtonWidth = function () {
        if (this._fixedSaveButtonWidth) { return; }

        var barVisible = this.isVisible();

        toggle.call(this, true, true);

        this.buttons.$saveButton.children('.save-button-state').removeClass('hidden');

        var $buttons = this.buttons.$saveButton.add(this.buttons.$cancelButton);

        var maxWidth = _.max(_.map($buttons, function (btn) {
            return $(btn).outerWidth(true);
        }));
        $buttons.css('width', maxWidth);

        updateSavingThreeState.call(this, true);
        toggle.call(this, barVisible, true);

        this._fixedSaveButtonWidth = true;
    };

    var scrollErrorsList = function () {
        if (!this.virtualHeights || !this.virtualHeights.canVirtualize) {
            return;
        }

        if (this.$errorList.find('.virtual-scroller').length === 0) {
            this.$errorList.append('<div class="virtual-scroller"></div>');
        }

        var scrollTop = this.$errorList.scrollTop(),
            errors = this.controller.getVisibleErrors(),
            maxVisibleErrors = Math.ceil(this.virtualHeights.errorList / this.virtualHeights.errorItem),
            countOffset = Math.max(Math.min(Math.floor(scrollTop / this.virtualHeights.errorItem), errors.length - maxVisibleErrors), 0),
            maxCount = Math.min(maxVisibleErrors, errors.length),
            $virtualScroller = this.$errorList.find('.virtual-scroller');

        $virtualScroller.css('height', ((errors.length - countOffset - maxVisibleErrors) * this.virtualHeights.errorItem) + 'px');
        this.$errorList.children(':not(.virtual-scroller)').remove();
        for (var i = countOffset; i <= (countOffset + maxCount - 1); i++) {
            if (errors[i] != null) {
                makeErrorItem.call(this, errors[i]).css('margin-top', i === countOffset ? scrollTop : 0).insertBefore($virtualScroller);
            }
        }

        this._lastErrorScrollPosition = scrollTop;
    };

    function ChangeBarView(controller) {
        logger.assert(this && this instanceof ChangeBarView, 'ChangeBarVew must be created with new');
        this.controller = controller;

        this._savingState = SAVE_STATES.UNSAVED;

        initElement.call(this);
        toggle.call(this, false);

        this.errorItemMarkup = this.controller._options.errorItemMarkup;

        this.buttons = {
            $saveButton : this.$el.find('.save-button'),
            $cancelButton : this.$el.find('.cancel-button'),
            $errorButton : this.$el.find('.error-button')
        };
        this.$el.find('[data-linkdisabled="save"]').prop('disabled', this.buttons.$saveButton.prop('disabled'));

        this.buttons.saveButtonStates = {
            unsaved : this.buttons.$saveButton.find('.unsaved'),
            saving : this.buttons.$saveButton.find('.saving'),
            saved : this.buttons.$saveButton.find('.saved')
        };

        this.$errorList = this.$el.find('.error-list');
        this.$errorListParent = this.$el.find('.errors-list');

        initListeners.call(this);
    }

    var centerChangeBar = function () {
        if (!this.controller._options.centerChangeBar) { return; }
        this.$el.css('margin-left', this.$el.outerWidth(false) / -2);
    };

    var temporarilySlideIn = function () {
        this.$el.addClass('notification');
        _.delay(function () {
            this.$el.removeClass('notification');
        }.bind(this), 200 + 50); // See core/theme/mixins/scaffold.less
    };

    var delayedCheckForErrorsInner = _.debounce(function () {
        if (!this._previousErrors && this.$el.hasClass('has-errors')) {
            temporarilySlideIn.call(this);
        }
        this._previousErrors = false;
    });
    var delayedCheckForErrors = function (previousErrors) {
        this._previousErrors = previousErrors || this._previousErrors || false;
        delayedCheckForErrorsInner.call(this);
    };

    var toggle = function (visible, immediate) {
        var that = this;

        function doToggle() {
            var isFade = that.$el.hasClass('fade');

            if (visible) {
                that.$el.css('display', '');
                /* jshint -W030 */ //Suppress the fact that we aren't assigning this to anything
                that.$el[0].offsetWidth; // force reflow
                /* jshint +W030 */
            } else if ($.support.transition && isFade && !immediate) {
                that.$el.one($.support.transition.end, function () {
                    that.$el.css('display', that.isVisible() ? '' : 'none');
                }).emulateTransitionEnd(200);
            }

            if (visible && !that.$el.hasClass('in')) {
                temporarilySlideIn.call(that);
            }

            that.$el
                .toggleClass('in', visible)
                .attr('aria-hidden', !visible)
                .data('visible', visible);

            if (!visible && (!$.support.transition || !isFade || immediate)) {
                that.$el.css('display', 'none');
            }
        }

        this._cancelHide();

        if (visible || immediate) {
            doToggle();
        } else {
            this._hideTimeout = setTimeout(doToggle, SAVED_STATE_VISIBLE_TIMEOUT);
        }
    };

    var makeErrorItem = function (error) {
        var $item = $(this.errorItemMarkup({ message : error.message }));
        $item.data('error', error);
        return $item;
    };

    var getSaveEnabled = function () {
        var barVisible = this.$el.data('visible'),
            validSavingState = this._savingState === SAVE_STATES.UNSAVED,
            saveEnabled = this.controller._checkAllowSave() && barVisible && validSavingState;

        return saveEnabled;
    };

    var updateSavingThreeState = function (force) {
        if (force || this._savingState !== this._lastSavingThreeState) {
            this._lastSavingThreeState = this._savingState;
            this.buttons.$saveButton
                .toggleClass('success', this._savingState === SAVE_STATES.SAVED)
                .toggleClass('btn-blue-dark', this._savingState === SAVE_STATES.SAVING)
                .toggleClass(this._savingButtonClass, this._savingState !== SAVE_STATES.SAVING);
            this.buttons.$saveButton.children('.save-button-state').addClass('hidden');
            this.buttons.saveButtonStates[this._savingState].removeClass('hidden');
            this.$el.toggleClass(_.values(SAVE_STATES).join(' '), false).toggleClass(this._savingState, true);
        }
    };

    var toggleModal = function (visible, immediate) {
        var that = this;

        // Make is a boolean
        immediate = !!immediate;

        if (visible) {
            this.$modalBackdrop.appendTo(document.body);
            /*jshint -W030*/ //Suppress the fact that we aren't assigning this to anything
            this.$modalBackdrop[0].offsetWidth; // force reflow
            /*jshint +W030*/
            that.$modalBackdrop.addClass('in');
        } else if (this.$modalBackdrop.parents('html').length > 0) {
            if ($.support.transition && !immediate) {
                this.$modalBackdrop.one($.support.transition.end, function () {
                    that.$modalBackdrop.remove();
                }).removeClass('in').emulateTransitionEnd(BACKDROP_FADE_TIME);
            } else {
                this.$modalBackdrop.removeClass('in').remove();
            }
        }
    };

    ChangeBarView.prototype = {
        constructor : ChangeBarView,

        show : function () {
            logger.assert(this._savingState === SAVE_STATES.UNSAVED, 'ChangeBar.show should only be called when change bar is unsaved');

            this._suspendButtonState = false;
            if (!this.isVisible()) {
                toggle.call(this, true);
                toggleModal.call(this, false);
            } else {
                this._cancelHide();
            }
            this.updateButtonStates();
        },

        _cancelHide : function () {
            if (this._hideTimeout) {
                clearTimeout(this._hideTimeout);
                this._hideTimeout = null;
            }
        },

        updateButtonStates : function () {
            if (this._suspendButtonState || this._suspendRedraws) { return; }

            var errors = this.controller.getVisibleErrors(),
                barVisible = this.$el.data('visible'),
                saveEnabled = getSaveEnabled.call(this),
                errorEnabled = (errors.length !== 0) && this._savingState === SAVE_STATES.UNSAVED,
                errorVisible = (errors.length !== 0),
                cancelEnabled = barVisible && this._savingState === SAVE_STATES.UNSAVED;

            this.$el.find('[data-linkdisabled="save"]')
                .add(this.buttons.$saveButton)
                .prop('disabled', !saveEnabled);
            this.buttons.$cancelButton.prop('disabled', !cancelEnabled);
            this.buttons.$errorButton.prop('disabled', !errorEnabled).attr('aria-hidden', !errorEnabled);
            // Swap the button group visibility instead of the button itself otherwise the margin is still visible
            this.$errorListParent.css('display', errorVisible ? '' : 'none');
            this.$el.toggleClass('has-errors', errorVisible);

            updateSavingThreeState.call(this);
            centerChangeBar.call(this);
        },

        hide : function () {
            if (this.isVisible()) {
                logger.assert(this._savingState !== SAVE_STATES.SAVING, 'ChangeBar.hide should not be called while change bar is saving');

                this._savingState = SAVE_STATES.UNSAVED;
                toggle.call(this, false, true);
                toggleModal.call(this, false, true);
                this.updateButtonStates();
            }
        },

        saving : function () {
            logger.assert(this._savingState === SAVE_STATES.UNSAVED && this.isVisible(), 'ChangeBar.save should only be called when change bar is unsaved and visible');

            this._savingState = SAVE_STATES.SAVING;
            this.announceEvent('dialogs.savingInProgress');
            toggleModal.call(this, true);
            this.updateButtonStates();
        },

        saveSucceeded : function () {
            logger.assert(this._savingState === SAVE_STATES.SAVING, 'ChangeBar.saveSucceeded should only be called when change bar is saving');

            this._savingState = SAVE_STATES.SAVED;
            this.announceEvent('dialogs.savingCompleted');

            toggle.call(this, false, false);
            toggleModal.call(this, false, false);
            this.updateButtonStates();

            this._suspendButtonState = true;

            // Switch back to unsaved state after running all of the redraw
            this._savingState = SAVE_STATES.UNSAVED;
        },

        saveFailed : function () {
            logger.assert(this._savingState === SAVE_STATES.SAVING, 'ChangeBar.saveFailed should only be called when change bar is saving');

            this._savingState = SAVE_STATES.UNSAVED;
            toggleModal.call(this, false, false);
            this.updateButtonStates();
        },

        getSavingState : function () {
            return this._savingState;
        },

        redraw : function () {
            if (this._suspendRedraws) {
                this._redrawRequested = true;
                return;
            }

            var previousHasErrors = this.$el.hasClass('has-errors');
            var errors = this.controller.getVisibleErrors();

            if (this.$el.closest(document.documentElement).length > 0) {
                repopulateErrorsList.call(this);
                fixSaveButtonWidth.call(this);
            }

            var $modal = this.$el.closest('.modal');
            if ($modal.length > 0 && !this._modalShown) {
                $modal.one('shown', function (e) {
                    this._modalShown = this._modalShown || $(e.target).is($modal);
                    this._fixedSaveButtonWidth = false;
                    fixSaveButtonWidth.call(this);
                }.bind(this));
            }

            this.buttons.$errorButton
                .text(lang.formatString(this.controller._options.locale.errors, errors.length))
                .attr('data-error-count', errors.length);
            this.$el.attr('data-error-count', errors.length);

            this.updateButtonStates();

            delayedCheckForErrors.call(this, previousHasErrors);
        },

        isVisible : function () {
            return !!this.$el.data('visible');
        },

        removeListeners : function () {
            removeListeners.call(this);
        },

        setStatusMessage : function (message) {
            var $status = this.$el.find('.status-message');
            if (message !== $status.text()) {
                $status.text(message);
            }
            centerChangeBar.call(this);
        },

        clearStatusMessage : function () {
            this.setStatusMessage(this.controller._options.locale.defaultStatusMessage);
        },

        announceEvent : function (event) {
            messageAnnouncer.announce(lang.getString(event));
        }
    };

    return ChangeBarView;
});


define('primutils/markup/primutils/ui/changeBar/body.hbs', ['handlebars','primutils.markup','primutils/ui/changeBar/helpers'], function (H, u) {
var imports = u.buildImported(Array.prototype.slice.call(arguments, 2), 'primutils/ui/changeBar/body.hbs'),
    func = H.template({"1":function(depth0,helpers,partials,data) {
  var stack1, buffer = "    ";
  stack1 = helpers['change-bar-status'].call(depth0, {"name":"change-bar-status","hash":{},"data":data});
  if (stack1 != null) { buffer += stack1; }
  buffer += "\n    ";
  stack1 = helpers['change-bar-errors'].call(depth0, {"name":"change-bar-errors","hash":{},"data":data});
  if (stack1 != null) { buffer += stack1; }
  buffer += "\n    ";
  stack1 = helpers['change-bar-cancel'].call(depth0, {"name":"change-bar-cancel","hash":{},"data":data});
  if (stack1 != null) { buffer += stack1; }
  buffer += "\n    ";
  stack1 = helpers['change-bar-save'].call(depth0, {"name":"change-bar-save","hash":{},"data":data});
  if (stack1 != null) { buffer += stack1; }
  return buffer + "\n";
},"compiler":[6,">= 2.0.0-beta.1"],"main":function(depth0,helpers,partials,data) {
  var stack1, buffer = "";
  stack1 = helpers['change-bar-controls'].call(depth0, {"name":"change-bar-controls","hash":{},"fn":this.program(1, data),"inverse":this.noop,"data":data});
  if (stack1 != null) { buffer += stack1; }
  return buffer;
},"useData":true});
return function (context, options) {
    options = u.extendOptions(options || {}, imports);
    return func(context, options).trim();
};
});


define('primutils/markup/primutils/ui/changeBar/errorItem.hbs', ['handlebars','primutils.markup','primutils/templates/helpers/icon','primutils/templates/helpers/dropdown'], function (H, u) {
var imports = u.buildImported(Array.prototype.slice.call(arguments, 2), 'primutils/ui/changeBar/errorItem.hbs'),
    func = H.template({"1":function(depth0,helpers,partials,data) {
  var stack1, lambda=this.lambda, buffer = "<span class=\"message\" title=\"";
  stack1 = lambda((depth0 != null ? depth0.message : depth0), depth0);
  if (stack1 != null) { buffer += stack1; }
  buffer += "\">";
  stack1 = lambda((depth0 != null ? depth0.message : depth0), depth0);
  if (stack1 != null) { buffer += stack1; }
  return buffer + "</span>";
},"compiler":[6,">= 2.0.0-beta.1"],"main":function(depth0,helpers,partials,data) {
  var stack1, buffer = "";
  stack1 = helpers.dropdownItem.call(depth0, {"name":"dropdownItem","hash":{},"fn":this.program(1, data),"inverse":this.noop,"data":data});
  if (stack1 != null) { buffer += stack1; }
  return buffer;
},"useData":true});
return function (context, options) {
    options = u.extendOptions(options || {}, imports);
    return func(context, options).trim();
};
});

/**
 * A standardized, flexible ChangeBar, suitable for use anywhere save and cancel are used, optimized for use at the page level in Prime.
 *
 * ChangeBar attempts to namespace functions into logical groupings.
 * For example, the `errors` functions (`errors.all`, `errors.removeMatching`, etc) all deal with errors.
 * The `events` functions deal with events, while `visibility` deal with visibility.
 *
 * There are three main topics covered by this class description:
 *
 * * [Creating a ChangeBar](#creation)
 * * [Dealing with errors](#errors)
 * * [Customizing the ChangeBar DOM](#custom_dom)
 *
 * <a id="creation"></a>
 * #### Creating a ChangeBar
 *
 * ChangeBar, by default, does not attach to the DOM.
 * This makes it very flexible, and indeed, can mean that more than one ChangeBar can exist peacefully on a page.
 *
 * We have three options for attaching to the DOM.
 * The first, setting the `attachTo` option and calling `attach`, can be used like the following:
 *
 *     define(['primutils/ui/ChangeBar'], function (ChangeBar) {
 *         // ....
 *         var myChangeBar = new ChangeBar({
 *             attachTo: '.content .banner' // this is the default
 *         });
 *         myChangeBar.attach();
 *     });
 *
 * Or, making that shorter with chainable functions, assuming we like all the default options:
 *
 *     define(['primutils/ui/ChangeBar'], function (ChangeBar) {
 *         var myChangeBar = new ChangeBar().attach();
 *     });
 *
 * Alternatively, since ChangeBar by default creates and maintains it's own element, it can be attached more manually:
 *
 *     define(['primutils/ui/ChangeBar'], function (ChangeBar) {
 *         var myChangeBar = new ChangeBar();
 *         var $element = $(...);
 *         $element.append(myChangeBar.$element);
 *     });
 *
 * Finally, if we have already created some ChangeBar on some existing element [as described below](#custom_dom), we could avoid calling `attach()` or `$element` entirely:
 *
 *     define(['primutils/ui/ChangeBar'], function (ChangeBar) {
 *         var myChangeBar = new ChangeBar({
 *             // since version 2.0.6
 *             fromExisting: $('.my-changebar-parent'),
 *         });
 *     });
 *
 * In this final form, we create our ChangeBar by using the `fromExisting` option and passing either an element which `.is` a `{{#change-bar-controls}}` or can `.find` a `{{#change-bar-controls}}`.
 *
 * <a id="errors"></a>
 * #### Dealing with errors
 *
 * ChangeBar attempts to deal with errors in a very non-opinionated, flexible manner.
 * For every error, the only required field is `message` - any other fields to the object can be added according to your needs.
 * (Though note how ChangeBar uses <a href="#hidden_errors">hidden errors</a>).
 * Internally, all functions use <a href="http://devdocs.io/underscore/index#where">_.where</a> when trying to match particular errors.
 *
 * So, how do we use this?
 * Let's say that we have multiple detail panels and navigation tabs, and errors can be associated with each.
 * We might have error objects like the following:
 *
 *     myChangeBar.errors.add([
 *         {detailPane: 'pane-events', navTab: 'project-1', message: '...'},
 *         {detailPane: 'pane-discussions', navTab: 'project-1', message: '...'},
 *         {detailPane: 'pane-discussions', navTab: 'project-2', message: '...'},
 *         {detailPane: 'pane-events', navTab: 'project-3', message: '...'}
 *     ]);
 *
 * We could clear out all errors associated with project-1 with the following:
 *
 *     myChangeBar.errors.removeMatching({navTab: 'project-1'});
 *
 * We could find all the errors associated with project-1 by the following:
 *
 *     myChangeBar.errors.findMatching({navTab: 'project-1'});
 *
 * In addition, `errors.all` returns all errors, `errors.visible` returns all visible errors, and `errors.removeAll` can be useful for dealing with the `cancel` event.
 *
 * <a id="selecting_errors"></a>
 * ##### Selecting Errors (AKA focusing errors)
 *
 * When the user clicks on an error in the list of errors, pages should bring browser focus to the error in question.
 * The <a data-anchor="#event_select-error" data-tabid="#events" href="#event_select-error">select-error</a> event exists to support this functionality.
 *
 * <a id="hidden_errors"></a>
 * ##### Hidden Errors
 *
 * If errors include the property `{hidden: true}`, those errors are considered hidden.
 * Hidden errors do not increment the ChangeBar's error count and are not visible in the dropdown when the user clicks to see the errors.
 * However, if there are no visible errors, but there are hidden errors, and the user clicks save, the hidden errors will become visible (that is, ChangeBar will loop through the list of errors, update the hidden errors to have `{hidden: false}`, and repopulate the error button/list).
 *
 * <a id="custom_dom"></a>
 * #### Customizing the ChangeBar DOM (since 2.0.6)
 *
 * ChangeBar uses several helpers in order to support generating custom DOM.
 * The default is a Handlebars template with the following content:
 *
 *     \{{#change-bar-controls}}
 *         \{{change-bar-errors}}
 *         \{{change-bar-cancel}}
 *         \{{change-bar-save}}
 *     \{{/change-bar-controls}}
 *
 * The above is a typical page-level ChangeBar, with access keys and the normal text for all the three buttons.
 * If you were using ChangeBar in a modal, something like the following would be more appropriate:
 *
 *     \{{#change-bar-controls}}
 *         \{{change-bar-errors "simple"}}
 *         \{{change-bar-cancel "secondary"}}
 *         \{{change-bar-save "primary"}}
 *     \{{/change-bar-controls}}
 *
 * Since no accesskeys would be created, but the controls are otherwise suitable for ChangeBar to make use of.
 *
 * If custom text for the save/cancel buttons is necessary, something like the following would work:
 *
 *     \{{#change-bar-controls}}
 *         \{{change-bar-errors "simple"}}
 *         \{{#change-bar-cancel "secondary"}}My cancel text\{{/change-bar-cancel}}
 *         \{{#change-bar-save "primary"}}My save text\{{/change-bar-save}}
 *     \{{/change-bar-controls}}
 *
 * When generating custom DOM as any of the above examples, see the `fromExisting` and `changeBarMarkup` options on how to use them.
 * Basically, `changeBarMarkup` allows you to specify the entire string that ChangeBar will wrap (and subsequent `.attach()` calls would append to DOM).
 * `fromExisting` allows you to add the DOM whenever works best for you, and ChangeBar would then grab that DOM on creation.
 *
 * <a id="helper_details"></a>
 * ##### Some helper details
 *
 * * `change-bar-controls` requires block content
 * * `change-bar-errors:` only one parameter.  "default", assumed if no parameters provided, gives the error button an accesskey.  No accesskey is provided if "simple" is passed.
 * * `change-bar-cancel` if no parameters are provided, suitable for the page level in Prime (i.e. has an accesskey).
 *     If parameters are provided, the `button` helper is used, provided with default text/attributes as required for this ChangeBar, but otherwise options are passed right along to the `button` helper.
 * * `change-bar-save` if no parameters are provided, suitable for the page level in Prime (i.e. has an accesskey).
 *     If parameters are provided, the `button` helper is used, provided with default text/attributes as required for this ChangeBar, but otherwise options are passed right along to the `button` helper.
 *
 * In addition to all these helpers, any elements - under the `change-bar-controls` element, that match the selector `[data-linkdisabled="save"]` - will have their disabled property kept in sync with the save button of this changebar.
 *
 * @class ui.ChangeBar
 * @module ui
 * @param [options] {Object} a map of options; see 'Properties' tab below for available options
 * @param [actions] {Object} handlers for events. See the events tab for <a data-anchor="#event_save" data-tabid="#events" href="#event_save">save</a>, <a data-anchor="#event_cancel" data-tabid="#events" href="#event_cancel">cancel</a>, <a data-anchor="#event_select-error" data-tabid="#events" href="#event_select-error">selectError</a>, <a data-anchor="#event_pre-destroy" data-tabid="#events" href="#event_pre-destroy">preDestroy</a>, and <a data-anchor="#event_unhide-errors" data-tabid="#events" href="#event_unhide-errors">unhideErrors</a>.
 *     Passing any of these values as part of this actions hash is equivalent to attaching listeners to ChangeBar events later, e.g. `myChangeBar.on('select-error', actions.selectError)`.
 * @constructor
 */

define('primutils/ui/changeBar/ChangeBar',['jquery', 'lodash', 'primutils/utils/Subscribable', 'logger',
    'primutils/utils/objectUtils',
    'primutils/utils/language',
    'primutils/utils/validator',
    'primutils/ui/changeBar/ChangeBarView',
    'primutils/markup/primutils/ui/changeBar/body.hbs',
    'primutils/markup/primutils/ui/changeBar/errorItem.hbs'
], function ($, _, Subscribable, logger, objectUtils, lang, validator, ChangeBarView, changeBarTemplate, errorItemTemplate) {
    'use strict';

    var ALLOWED_EVENTS = ['save', 'cancel', 'select-error', 'unhide-errors', 'pre-destroy'];

    function failOnceDestroyed() {
        throw new Error('ChangeBar has been destroyed, calling methods is no longer legal');
    }

    function isJquerySelector($node) {
        return !!$node.jquery;
    }

    var __validator__ = validator({
        name : 'ChangeBar',
        validOptions : {
            id : 'String',
            /**
             * Specifies the element to attach to, in the event of using `myChangeBar.attach()`.
             * Can be a String, in which case changeBar will select for that selector using jQuery to attach to, a jQuery selector already, or a Function that should return either of the same two types.
             *
             * @property attachTo
             * @type {String|$jquerySelector|Function}
             */
            attachTo : ['String', 'Function', isJquerySelector],
            /**
             * Will be used to sort any errors before rendering them to the DOM.
             * Will be called every time errors are repopulated; ChangeBar does not attempt to detect that the errors have not been updated in order to avoid doing work.
             * The function should function as per the native <a href="http://devdocs.io/javascript/global_objects/array/sort">Array.sort</a>.
             *
             * @property sortErrors
             * @type {undefined|Function}
             */
            sortErrors : [undefined, 'Function'],
            /**
             * Whether or not save should be allowed.
             * Regardless of the result of this option, save will not be allowed if the ChangeBar is not visible.
             * An example choice for this function might be something like:
             *
             *     allowSave: function (cb) {
             *         return cb.util.defaultFns.allowSave() || myPage.allowSaveWithErrors();
             *     }
             *
             * See also the method `util.defaultFns.allowSave`
             *
             * @property allowSave
             * @type {undefined|Function}
             * @since 2.0.3
             * @default a function that returns true if there are no visible errors in the changebar and false if there are visible errors
             */
            allowSave : [undefined, 'Function'],
            /**
             * Called every time errors are populated in this changebar, to determine whether or not the changebar should be shown.
             * If this function returns true, `changebar.visibility.show()` will be called, otherwise it will not.
             *
             * An example choice for this function might be something like:
             *
             *     checkAutoShow: function (cb) {
             *         return myPage.isDirty();
             *     }
             *
             * See also the method `util.defaultFns.checkAutoShow`
             *
             * @property checkAutoShow
             * @type {undefined|Function}
             * @default a function that returns true if there are errors in the changebar, false otherwise
             * @since 2.0.3
             */
            checkAutoShow : [undefined, 'Function'],
            /**
             * Used to configure the markup created by this ChangeBar.
             * Not used if the option `fromExisting` is specified.
             * See the class description for more details on DOM customization.
             *
             * @property changeBarMarkup
             * @type {undefined|String}
             * @default {String} the default DOM suitable for the page level in Prime
             */
            changeBarMarkup : [undefined, 'String'],
            /**
             * Used to attach this ChangeBar to existing DOM.
             * If this is used, overrides the `changeBarMarkup` option.
             * Must be attached to existing DOM that uses the DOM as created by the ChangeBar helpers `change-bar-controls`, `change-bar-errors`, `change-bar-cancel`, and `change-bar-save`.
             * See the class description for more details on DOM customization.
             *
             * @since 2.0.6
             * @property fromExisting
             * @type {String|$jquerySelector|Function}
             * @default undefined
             */
            fromExisting : ['String', 'Function', isJquerySelector],
            /**
             * Whether the change bar should be centered over its DOM anchor point.
             *
             * This option is `true` by default because most change bars go in a page banner where Blue Sky dictates that change bars should be centered.
             * If you are using a change bar for a different reason, you will probably want to set this option to `false`.
             *
             * @since 2.7.11
             * @property centerChangeBar
             * @type {Boolean}
             * @default true
             */
            centerChangeBar : 'Boolean',
            /**
             * Whether the change bar should be faded in and out.
             *
             * This option is `true` by default because most change bars go in a page banner where Blue Sky dictates that change bars should be centered.
             * If you are using a change bar for a different reason, you will probably want to set this option to `false`.
             *
             * @since 4.4.0
             * @property fadeChangeBar
             * @type {Boolean}
             * @default true
             */
            fadeChangeBar : 'Boolean',
            /**
             * Whether or not to automatically close and commit any editors on grids on the page.
             * This typically only matters if the save button accesskey is used, and it's recommended to leave this as true.
             *
             * @since 2.6.10
             * @property closeGridEditorsOnSave
             * @type {Boolean}
             * @default true
             */
            closeGridEditorsOnSave : 'Boolean',
            /**
             * Localization values
             *
             * @property locale
             * @type {Object}
             */
            locale : 'Object',
            /**
             * Whether to add browser access keys to the buttons in the changebar
             *
             * @property useAccessKeys
             * @type {Boolean}
             */
            useAccessKeys : 'Boolean',

            // used by MultiViewChangeBar
            canSaveView : undefined
        },
        validators : {
            initialize : {}
        },
        ignoredSubkeys : ['attachTo', 'fromExisting', 'locale']
    });

    var __actionValidator = validator({
        name : 'ChangeBar Actions',
        validOptions : {
            save : 'Function',
            saveView : 'Function',
            saveViewAs : 'Function',
            cancel : 'Function',
            selectError : 'Function',
            unhideErrors : 'Function',
            resetView : 'Function'
        },
        validators : {
            initialize : {}
        }
    });

    var repopulateErrors = function () {
        if (_.isFunction(this._options.sortErrors)) {
            this._errors.sort(this._options.sortErrors);
        }

        this._view.redraw();
        if (this._options.checkAutoShow(this)) {
            this.show();
        }
    };

    var addErrorWithoutRepopulate = function (error) {
        validator.checkType(error.hidden, [undefined, 'Boolean']);
        if (!validator.checkType(error.message, 'String', 'changeBarError.message')) {
            error.message = lang.getString('changeBar.error');
        }

        this._errors.push(error);
    };

    var getValue = function (opt) {
        return _.isFunction(opt) ? opt.apply(this, _.rest(_.toArray(arguments))) : opt;
    };

    /**
     * The default function behavior for the `checkAutoShow` option.
     * Returns true if there are errors managed by this changebar, false otherwise.
     *
     * @method util.defaultFns.checkAutoShowss
     * @since 2.0.3
     */
    var checkAutoShow = function (cb) {
        return cb.errors.all().length > 0;
    };

    /**
     * The default function behavior for the `allowSave` option.
     * Returns true if there are no visible errors managed by this changebar, false otherwise.
     *
     * @method util.defaultFns.allowSave
     * @since 2.0.3
     */
    var allowSave = function (cb) {
        return cb.errors.visible().length === 0;
    };

    var globalDefaults = {};

    function getDefaults() {
        var id = _.uniqueId('change-bar-');
        return {
            id : id,
            attachTo : '.change-bar-control-area, .content .banner',
            centerChangeBar : true, // Gonna default this to true because most changebars are in the banner
            fadeChangeBar : true, // Gonna default this to true because most changebars are in the banner
            fromExisting : undefined,
            changeBarMarkup : changeBarTemplate({
                changeBarId : id
            }),
            errorItemMarkup : errorItemTemplate,
            locale : {
                errors : lang.getString('changeBar.errorsWithCounter'),
                defaultStatusMessage : lang.getString('changeBar.defaultStatusMessage')
            },
            allowSave : allowSave,
            checkAutoShow : checkAutoShow,
            closeGridEditorsOnSave : true,
            useAccessKeys : true
        };
    }

    function ChangeBar(options, actions) {
        logger.assert(this && this instanceof ChangeBar, 'ChangeBar must be created with new');
        Subscribable.call(this, {
            allowedEvents : ALLOWED_EVENTS
        });

        options = options || {};
        validator.checkType(options, ['Object']);
        __validator__.validate('initialize', options);
        Object.defineProperty(this, '_options', {
            value : objectUtils.extendDeep({}, getDefaults(), globalDefaults, options, function (dest, src) {
                if (src instanceof $) {
                    return src;
                }
            }),
            writable : false,
            enumerable : false
        });

        actions = actions || {};
        validator.checkType(actions, ['Object']);
        __actionValidator.validate('initialize', actions);
        if (_.isFunction(actions.save)) {
            this.on('save', actions.save);
        }
        if (_.isFunction(actions.cancel)) {
            this.on('cancel', actions.cancel);
        }
        if (_.isFunction(actions.selectError)) {
            this.on('select-error', actions.selectError);
        }
        if (_.isFunction(actions.unhideErrors)) {
            this.on('unhide-errors', actions.unhideErrors);
        }
        if (_.isFunction(actions.preDestroy)) {
            this.on('pre-destroy', actions.preDestroy);
        }

        Object.defineProperty(this, '_view', {
            value : new ChangeBarView(this),
            writable : false,
            enumerable : false
        });

        this._errors = [];

        this.visibility = this.vis = {};
        this.visibility.toggle = this.toggleChanged.bind(this);
        this.visibility.show = this.show.bind(this);
        this.visibility.hide = this.hide.bind(this);
        this.visibility.saving = this.saving.bind(this);
        this.visibility.saveSucceeded = this.saveSucceeded.bind(this);
        this.visibility.saveFailed = this.saveFailed.bind(this);
        this.visibility.up = this.isVisible.bind(this);
        this.visibility.setStatusMessage = this.setStatusMessage.bind(this);
        this.visibility.clearStatusMessage = this.clearStatusMessage.bind(this);

        this.errors = {};
        this.errors.all = this.getErrors.bind(this);
        this.errors.visible = this.getVisibleErrors.bind(this);
        this.errors.add = this.addError.bind(this);
        this.errors.findMatching = this.findMatchingErrors.bind(this);
        this.errors.removeMatching = this.removeMatchingErrors.bind(this);
        this.errors.removeAll = this.removeAllErrors.bind(this);

        this.util = {};
        this.util.attach = this.attach.bind(this);
        this.util.detach = this.detach.bind(this);
        this.util.destroy = this.destroy.bind(this);
        this.util.destroyWithoutDOM = this.destroyWithoutDOM.bind(this);
        this.util.defaultFns = {};
        this.util.defaultFns.checkAutoShow = checkAutoShow.bind(this);
        this.util.defaultFns.allowSave = allowSave.bind(this);
        this.util.updateButtons = this.updateButtons.bind(this);
        this.util.getSavingState = this.getSavingState.bind(this);

        this.events = {};
        /**
         * Attaches event handlers for this change bar.
         * For details on event syntax, see <a href="utils.Subscribable.html">Subscribable</a>.
         *
         * @method events.on
         */
        this.events.on = this.on.bind(this);
        /**
         * Detaches event handlers for this change bar.
         * For details on event syntax, see <a href="utils.Subscribable.html">Subscribable</a>.
         *
         * @method events.off
         */
        this.events.off = this.off.bind(this);
        this.events.trigger = this.trigger.bind(this);
    }

    var cbp = ChangeBar.prototype = Object.create(Subscribable.prototype);
    cbp.constructor = ChangeBar;

    Object.defineProperty(cbp, '$element', {
        get : function () {
            return this._view.$el;
        }
    });

    // maybe consider documenting this as `this.util.attach`.  Keep backwards compatability either way
    /**
     * Simple chainable utility for attaching this ChangeBar to the page.
     * Respects the option `attachTo` for what element to append to.
     *
     * @method attach
     * @chainable
     */
    cbp.attach = function () {
        logger.info(this._options.id + ' is attaching to: ' + this._options.attachTo);
        $(getValue.call(this, this._options.attachTo)).append(this.$element);
        repopulateErrors.call(this);
        return this;
    };

    /**
     * Simple chainable utility for detaching this ChangeBar from the DOM.
     * Note: uses $.fn.detach internally, so does not disconnect event listeners, etc.
     * For that purpose, see <a href="#method_destroy">destroy</a>.
     *
     * @method detach
     * @chainable
     */
    cbp.detach = function () {
        logger.info(this._options.id + ' is detaching from DOM');
        this.$element.detach();
        return this;
    };

    /**
     * Simple chainable utility for destroying this ChangeBar.
     * Note: uses $.fn.destroy internally, so does disconnect event listeners, etc.
     * If this is not desired, consider <a href="#method_detach">detach</a>.
     *
     * @method destroy
     */
    cbp.destroy = function () {
        logger.info(this._options.id + ' is getting destroyed');
        /**
         * Triggered immediately before this ChangeBar is destroyed.
         *
         * @event pre-destroy
         * @since 2.0.8
         * @param cb {ChangeBar} the ChangeBar that's getting destroyed (this)
         */
        this.trigger('pre-destroy', this);
        this.detach();
        this.$element.remove();

        this.events = failOnceDestroyed;
        this.on = failOnceDestroyed;
        Subscribable.prototype.clear.call(this);
        this.off = failOnceDestroyed;
        this.trigger = failOnceDestroyed;
    };

    cbp.destroyWithoutDOM = function () {
        logger.info(this._options.id + ' is getting destroyed (but leaving DOM in place)');
        this.trigger('pre-destroy', this);
        this._errors = [];
        repopulateErrors.call(this);
        this._checkAllowSave = function () {
            return false;
        };
        this.hide();
        this._view.removeListeners();
        this._view.$el = $([]);

        this.events = failOnceDestroyed;
        this.on = failOnceDestroyed;
        Subscribable.prototype.clear.call(this);
        this.off = failOnceDestroyed;
        this.trigger = failOnceDestroyed;
    };

    /**
     * Returns whether or not this ChangeBar is visible.
     * Please note: if the ChangeBar is `detach`'d from the DOM, it will still return true, while this would be false after `destroy`.
     *
     * @method visibility.up
     * @returns {Boolean} true if this ChangeBar is visible, false otherwise
     * @since 2.0.2
     */
    cbp.isVisible = function () {
        return !!this._view.isVisible();
    };

    cbp.setStatusMessage = function (message) {
        this._view.setStatusMessage(message);
    };

    cbp.clearStatusMessage = function () {
        this._view.clearStatusMessage();
    };

    /**
     * Toggles the visibility of the changeBar.
     * If any errors are present, the changeBar will not hide.
     *
     * @method visibility.toggle
     * @chainable
     */
    cbp.toggleChanged = function (changed) {
        if (changed) {
            this.show();
        } else {
            this.hide();
        }
        return this;
    };

    /**
     * Update the state of the buttons (save, cancel, errors) managed by this changebar.
     *
     * @method util.updateButtons
     * @chainable
     * @since 2.3.9
     */
    cbp.updateButtons = function () {
        this._view.updateButtonStates();
        return this;
    };

    /**
     * Shows the changeBar.
     *
     * @method visibility.show
     * @chainable
     */
    cbp.show = function () {
        this._view.show();
        return this;
    };

    /**
     * Sets the change bar status to saving.
     *
     * *This method is NOT chainable.*
     *
     * @method visibility.saving
     * @return {{succeeded:Function,failed:Function}}
     */
    cbp.saving = function () {
        this._view.saving();

        return {
            succeeded : this.saveSucceeded.bind(this),
            failed : this.saveFailed.bind(this)
        };
    };

    /**
     * Sets the change bar status to saved and hides the change bar after flashing the success button.
     *
     * @method visibility.saveSucceeded
     * @chainable
     */
    cbp.saveSucceeded = function () {
        this._view.saveSucceeded();
        return this;
    };

    /**
     * Sets the change bar status to unsaved and keeps the change bar visible.
     *
     * @method visibility.saveFailed
     * @chainable
     */
    cbp.saveFailed = function () {
        this._view.saveFailed();
        return this;
    };

    cbp.getSavingState = function () {
        return this._view.getSavingState();
    };

    /**
     * Hides the changeBar.
     *
     * @method visibility.hide
     * @chainable
     */
    cbp.hide = function (force) {
        if (this._errors.length === 0 || force === true) {
            this._view.hide();
        }
        return this;
    };

    /**
     * Gets all errors, including invisible ones.
     *
     * @method errors.all
     * @returns {Array} all errors, hidden or not, possibly empty
     */
    cbp.getErrors = function () {
        return this._errors;
    };

    /**
     * Gets all visible errors.
     *
     * @method errors.visible
     * @returns {Array} all visible errors, possibly empty
     */
    cbp.getVisibleErrors = function () {
        return _.difference(this._errors, this.findMatchingErrors({ hidden : true }));
    };

    /**
     * Registers an error, or array of errors, with this changeBar.
     * When registering errors, the `message` property is required.
     * If `hidden` is set to true, the error is considered an invisible one and will only show up if the user attempts to save before dealing with all invisible errors.
     *
     * Note: ChangeBar does *not* attempt to deduplicate errors.
     * If you add the same error object twice, that's your prerogative.
     * Consider `errors.removeMatching` for deduplicating errors, and TODO Brian Sacksteder's blog post for dealing with grid errors.
     *
     * @method errors.add
     * @chainable
     * @param errorItem {Array|Object} the error Object, or Array of error Objects, to be added to this changeBar
     */
    cbp.addError = function (errorItem) {
                logger.assert(errorItem && (_.isObject(errorItem) || _.isArray(errorItem)), 'changeBar.errors.add only takes objects or arrays');
        
        if (_.isArray(errorItem)) {
            this._addBulkErrors(errorItem);
        } else {
            addErrorWithoutRepopulate.call(this, errorItem);
            repopulateErrors.call(this);
        }

        return this;
    };

    cbp._addBulkErrors = function (errors) {
        if (!errors.length) {
            return this;
        }

        _.forEach(errors, function (err) {
                        logger.assert(err && _.isObject(err), 'When adding multiple errors, the array must be an array of objects');
                        addErrorWithoutRepopulate.call(this, err);
        }.bind(this));

        repopulateErrors.call(this);

        return this;
    };

    /**
     * Returns all errors matching the given example.
     * As <a href="http://devdocs.io/underscore/index#where">_.where</a> used on the list of errors.
     *
     * @method errors.findMatching
     * @param example {Object} an Object to be used as a filter for the errors with this change bar.
     *      For example, if the object `{controller:'foo', tab:'bar'}` is used as the example, any error with both of those properties present and matching will be included in the return of this function.
     * @returns {Array} all matching errors
     */
    cbp.findMatchingErrors = function (example) {
        return _.where(this._errors, example);
    };

    /**
     * Removes all errors matching the given example.
     * Internally uses <a href="http://devdocs.io/underscore/index#where">_.where</a> to find the matching errors to be removed.
     *
     * @method errors.removeMatching
     * @param example {Object} an Object to be used as a filter for the errors to be removed.
     *      For example, if the object `{controller:'foo', tab:'bar'}` is used as the example, any error with both of those properties present and matching will be removed.
     * @chainable
     */
    cbp.removeMatchingErrors = function (example) {
        var matchingErrors = this.findMatchingErrors(example);
        if (matchingErrors.length > 0) {
            this._errors = _.difference(this._errors, matchingErrors);
            repopulateErrors.call(this);
        }
        return this;
    };

    /**
     * Removes all the errors associated with this change bar.
     *
     * @method errors.removeAll
     * @chainable
     */
    cbp.removeAllErrors = function () {
        if (this._errors.length > 0) {
            this._errors.splice(0, this._errors.length);
            repopulateErrors.call(this);
        }
        return this;
    };

    /**
     * Triggered when an error is selected (aka clicked) by the user.
     *
     * @event select-error
     * @param {Object} error the error that was selected (clicked) by the user.
     */
    cbp._selectError = function (error) {
        this.trigger('select-error', error);
        return this;
    };

    /**
     * Triggered when the user clicks save.
     * Will only trigger from ChangeBar itself if no errors are present.
     *
     * @event save
     */
    cbp._save = function () {
        if (this.getVisibleErrors().length) {
            // not sure how this would happen, but doesn't hurt to be paranoid
            $.noop();
        } else if (this.getErrors().length) {
            this._unhideAnyErrors();
        } else {
            this.trigger('save');
        }
        return this;
    };

    cbp._unhideAnyErrors = function () {
        if (this.getErrors().length) {
            var unhiddenErrors = [];
            this._errors = _.map(this._errors, function (err) {
                if (err.hidden === true) {
                    unhiddenErrors.push(err);
                    err.hidden = false;
                }
                return err;
            });
            repopulateErrors.call(this);
            /**
             * Triggered when the user clicks save, when no visible errors were present but hidden errors were, and those errors have been made visible.
             *
             * @event unhide-errors
             * @param {Array} errors the errors that were now made visible
             */
            this.trigger('unhide-errors', unhiddenErrors);
            return unhiddenErrors.length > 0;
        }
        return false;
    };

    /**
     * Triggered when the user clicks cancel.
     *
     * @event cancel
     */
    cbp._cancel = function () {
        this.trigger('cancel');
        return this;
    };

    // This function is exposed on ChangeBar only for ChangeBarView to use it: buyer beware, do not use this private API if you're looking at this comment
    cbp._checkAllowSave = function () {
        return this._options.allowSave.call(undefined, this);
    };

    /**
     * Applies a level of defaults to all ChangeBars constructed after this is called.
     *
     * @method defaults
     * @static
     * @param newDefaults {Object}
     */
    ChangeBar.defaults = function (newDefaults) {
        _.assign(globalDefaults, newDefaults);
    };

    return ChangeBar;
});


define('primutils/ui/ChangeBar',['primutils/ui/changeBar/ChangeBar'], function (ChangeBar) {
    'use strict';

    return ChangeBar;
});



define('primutils/markup/primutils/ui/dialogs/alert.hbs', ['handlebars','primutils.markup','primutils/templates/helpers/modal','primutils/templates/helpers/button','primutils/templates/helpers/icon','primutils/templates/helpers/join','primutils/templates/helpers/ternary'], function (H, u) {
var imports = u.buildImported(Array.prototype.slice.call(arguments, 2), 'primutils/ui/dialogs/alert.hbs'),
    func = H.template({"1":function(depth0,helpers,partials,data) {
  var stack1, buffer = "    ";
  stack1 = helpers.modalHeader.call(depth0, {"name":"modalHeader","hash":{},"data":data});
  if (stack1 != null) { buffer += stack1; }
  buffer += "\n";
  stack1 = helpers.modalBody.call(depth0, {"name":"modalBody","hash":{},"fn":this.program(2, data),"inverse":this.noop,"data":data});
  if (stack1 != null) { buffer += stack1; }
  stack1 = helpers.modalFooter.call(depth0, {"name":"modalFooter","hash":{},"fn":this.program(11, data),"inverse":this.noop,"data":data});
  if (stack1 != null) { buffer += stack1; }
  return buffer;
},"2":function(depth0,helpers,partials,data) {
  var stack1, lambda=this.lambda, buffer = "";
  stack1 = helpers['if'].call(depth0, (depth0 != null ? depth0.hasIconType : depth0), {"name":"if","hash":{},"fn":this.program(3, data),"inverse":this.noop,"data":data});
  if (stack1 != null) { buffer += stack1; }
  stack1 = helpers['if'].call(depth0, (depth0 != null ? depth0.hasIconString : depth0), {"name":"if","hash":{},"fn":this.program(5, data),"inverse":this.noop,"data":data});
  if (stack1 != null) { buffer += stack1; }
  stack1 = helpers['if'].call(depth0, (depth0 != null ? depth0.uiContext : depth0), {"name":"if","hash":{},"fn":this.program(7, data),"inverse":this.noop,"data":data});
  if (stack1 != null) { buffer += stack1; }
  buffer += "        <p>";
  stack1 = lambda((depth0 != null ? depth0.message : depth0), depth0);
  if (stack1 != null) { buffer += stack1; }
  buffer += "</p>\n";
  stack1 = helpers['if'].call(depth0, (depth0 != null ? depth0.hasAdditionalDetails : depth0), {"name":"if","hash":{},"fn":this.program(9, data),"inverse":this.noop,"data":data});
  if (stack1 != null) { buffer += stack1; }
  return buffer;
},"3":function(depth0,helpers,partials,data) {
  var stack1, buffer = "        <span class=\"modal-icon\">";
  stack1 = helpers.icon.call(depth0, (depth0 != null ? depth0.iconType : depth0), "", {"name":"icon","hash":{},"data":data});
  if (stack1 != null) { buffer += stack1; }
  return buffer + "</span>\n";
},"5":function(depth0,helpers,partials,data) {
  var stack1, lambda=this.lambda, buffer = "        <span class=\"modal-icon\">";
  stack1 = lambda((depth0 != null ? depth0.iconString : depth0), depth0);
  if (stack1 != null) { buffer += stack1; }
  return buffer + "</span>\n";
},"7":function(depth0,helpers,partials,data) {
  var stack1, lambda=this.lambda, buffer = "        <h4 class=\"ui-context\">";
  stack1 = lambda((depth0 != null ? depth0.uiContext : depth0), depth0);
  if (stack1 != null) { buffer += stack1; }
  return buffer + "</h4>\n";
},"9":function(depth0,helpers,partials,data) {
  var stack1, lambda=this.lambda, buffer = "        <div class=\"additional-details\" style=\"display: block;\">\n            <a href=\"#\" class=\"additional-details-label\">";
  stack1 = lambda((depth0 != null ? depth0.additionalDetailsLabel : depth0), depth0);
  if (stack1 != null) { buffer += stack1; }
  buffer += "</a>\n            <div class=\"additional-details-body\">";
  stack1 = lambda((depth0 != null ? depth0.additionalDetails : depth0), depth0);
  if (stack1 != null) { buffer += stack1; }
  return buffer + "</div>\n        </div>\n";
},"11":function(depth0,helpers,partials,data) {
  var stack1, buffer = "        ";
  stack1 = helpers.button.call(depth0, "primary", {"name":"button","hash":{
    'data-dismiss': ("modal"),
    'data-default': ("enter"),
    'class': ("proceedModal")
  },"fn":this.program(12, data),"inverse":this.noop,"data":data});
  if (stack1 != null) { buffer += stack1; }
  return buffer + "\n";
},"12":function(depth0,helpers,partials,data) {
  var stack1, lambda=this.lambda;
  stack1 = lambda((depth0 != null ? depth0.ok : depth0), depth0);
  if (stack1 != null) { return stack1; }
  else { return ''; }
  },"compiler":[6,">= 2.0.0-beta.1"],"main":function(depth0,helpers,partials,data) {
  var stack1, buffer = "";
  stack1 = helpers.modal.call(depth0, (depth0 != null ? depth0.title : depth0), {"name":"modal","hash":{
    'class': (helpers._.call(depth0, helpers['?'].call(depth0, (depth0 != null ? depth0.hasIcon : depth0), "pgbu-modal-with-icon ", "", {"name":"?","hash":{},"data":data}), "message-modal", {"name":"_","hash":{},"data":data})),
    'id': ("alert-modal")
  },"fn":this.program(1, data),"inverse":this.noop,"data":data});
  if (stack1 != null) { buffer += stack1; }
  return buffer;
},"useData":true});
return function (context, options) {
    options = u.extendOptions(options || {}, imports);
    return func(context, options).trim();
};
});


define('primutils/markup/primutils/ui/dialogs/doNotShow.hbs', ['handlebars','primutils.markup'], function (H, u) {
var imports = u.buildImported(Array.prototype.slice.call(arguments, 2), 'primutils/ui/dialogs/doNotShow.hbs'),
    func = H.template({"compiler":[6,">= 2.0.0-beta.1"],"main":function(depth0,helpers,partials,data) {
  var stack1, lambda=this.lambda, buffer = "<div class=\"pull-left\">\n    <div class=\"controls\">\n        <label class=\"checkbox suppress-modal-checkbox\" for=\"doNotShowAgain\">\n            <input type=\"checkbox\" id=\"doNotShowAgain\">\n            ";
  stack1 = lambda((depth0 != null ? depth0.doNotShowAgain : depth0), depth0);
  if (stack1 != null) { buffer += stack1; }
  return buffer + "\n        </label>\n    </div>\n</div>";
},"useData":true});
return function (context, options) {
    options = u.extendOptions(options || {}, imports);
    return func(context, options).trim();
};
});

define('primutils/ui/dialogs/doNotShowHelper',[], function () {
    'use strict';

    var DO_NOT_SHOW_KEY = 'ORA_CPP_DIALOG_DO_NOT_SHOW_DATA';

    function getCurrentDoNotShowData() {
        var currentDoNotShow = {};
        if (window.localStorage && window.localStorage[DO_NOT_SHOW_KEY]) {
            currentDoNotShow = JSON.parse(window.localStorage[DO_NOT_SHOW_KEY]);
        }
        return currentDoNotShow;
    }

    return {
        toggleDoNotShow : function (key, state) {
            var currentDoNotShow = getCurrentDoNotShowData();
            currentDoNotShow[key] = state;
            if (window.localStorage) {
                window.localStorage[DO_NOT_SHOW_KEY] = JSON.stringify(currentDoNotShow);
            }
        },

        isDoNotShow : function (key) {
            return !!getCurrentDoNotShowData()[key];
        }
    };
});

define('primutils/ui/dialogs/dialogTypes',['primutils/utils/language'], function (lang) {
    'use strict';

    /**
     * @class ui.dialogs.dialogType
     * @module ui.dialogs
     * @type {{NONE: string, CUSTOM: string, ERROR: string, WARNING: string, RESTRICTED: string, INFORMATION: string, CONFIRM: string, SUCCESS: string}}
     */
    var types = {
        /**
         * @property NONE
         * @static
         */
        NONE : 'none',
        /**
         * @property CUSTOM
         * @private
         * @deprecated
         * @static
         */
        CUSTOM : 'custom',
        /**
         * @property ERROR
         * @static
         */
        ERROR : 'error',
        /**
         * @property WARNING
         * @static
         */
        WARNING : 'warning',
        /**
         * @property RESTRICTED
         * @static
         */
        RESTRICTED : 'restricted',
        /**c
         * @property INFORMATION
         * @static
         */
        INFORMATION : 'information',
        /**
         * @property CONFIRM
         * @static
         */
        CONFIRM : 'confirm',
        /**
         * @property SUCCESS
         * @static
         */
        SUCCESS : 'success'
    };

    var icons = {};
    icons[types.NONE] = '';
    icons[types.ERROR] = 'status-error';
    icons[types.WARNING] = 'status-warning';
    icons[types.RESTRICTED] = 'status-restricted';
    icons[types.INFORMATION] = 'status-info';
    icons[types.CONFIRM] = 'status-warning';
    icons[types.SUCCESS] = 'status-success';

    var defaultTitles = {},
        defaultTitleProperties = {};
    defaultTitleProperties[types.NONE] = { get : function () { return lang.getString('dialogTitles.alert'); } };
    defaultTitleProperties[types.ERROR] = { get : function () { return lang.getString('dialogTitles.error'); } };
    defaultTitleProperties[types.WARNING] = { get : function () { return lang.getString('dialogTitles.warning'); } };
    defaultTitleProperties[types.RESTRICTED] = { get : function () { return lang.getString('dialogTitles.restricted'); } };
    defaultTitleProperties[types.INFORMATION] = { get : function () { return lang.getString('dialogTitles.information'); } };
    defaultTitleProperties[types.CONFIRM] = { get : function () { return lang.getString('dialogTitles.confirm'); } };
    defaultTitleProperties[types.SUCCESS] = { get : function () { return lang.getString('dialogTitles.success'); } };
    Object.defineProperties(defaultTitles, defaultTitleProperties);

    Object.freeze(types);
    Object.freeze(icons);
    Object.freeze(defaultTitles);

    return {
        types : types,
        icons : icons,
        titles : defaultTitles
    };
});

/*jshint -W030*/
/**
 * Displays a message to the user that they must acknowledge by pressing "ok".
 * Additional information can be provided which is hidden by default.
 *
 * By specifying the `doNotShowAgain` option with a key specific to your modal, the end-user will be able to indicate that they do not want to see this modal again (in which case it will be auto-confirmed).
 *
 * To add a handler for when the user clicks "ok" or closes the modal, see the `ok` method or `okCallback` property.
 *
 * @class ui.dialogs.AlertDialog
 * @module ui.dialogs
 * @param options {Object} See properties tab
 * @constructor
 */
define('primutils/ui/dialogs/AlertDialog',[
    'jquery',
    'lodash',
    'primutils/utils/keyCode',
    'primutils/markup/primutils/ui/dialogs/alert.hbs',
    'primutils/markup/primutils/ui/dialogs/doNotShow.hbs',
    'primutils/utils/language',
    'primutils/ui/dialogs/doNotShowHelper',
    'primutils/ui/dialogs/dialogTypes',
    'primutils/a11y/accessibilityUtils',
    'logger'
], function ($, _, keyCode, alertTemplate, doNotShowTemplate, lang, doNotShowHelper, dialogTypes, accessibilityUtils, logger) {
    'use strict';

    var DEFAULT_OPTIONS = {
        /**
         * A key stored in local storage which can prevent a particular dialog from displaying again.
         * If the user checks "Do not show again", the <a href="#property_okCallback">okCallback</a> will automatically
         * be called on subsequent calls to dialogs using the same key.
         *
         * Using `false` means the dialog will always be shown.
         *
         * @property doNotShowAgain {String|Boolean}
         * @optional
         * @default false
         */
        doNotShowAgain : false,
        /**
         * Whether to automatically call `show` when the spinner is created
         *
         * @property autoDisplay {Boolean}
         * @optional
         * @default true
         */
        autoDisplay : true,
        /**
         * A method which will be called when the user clicks the Ok button. This method will *not* be called if the
         * user closes the modal through the Close (X) button in the corner or with the escape button.
         *
         * The callback takes no arguments.
         *
         * @property okCallback {Function}
         * @optional
         */
        //okCallback,
        /**
         * Information about what the UI was doing when the dialog needed to be shown.
         *
         * This is a brief sentence fragment that describes what the user was attempting to do when the dialog was shown.
         * This differs from the `message` which shows the cause and potential resolution steps to the users problem.
         *
         * @property uiContext {String}
         * @optional
         */
        //uiContext,
        /**
         * The type of the dialog.
         *
         * The type of the dialog controls the default title and icon, but both are overridable.
         *
         * Available types are exposed by the static `types` property.
         *
         * @property type
         * @type {ui.dialogs.dialogType}
         * @default NONE
         */
        type : dialogTypes.types.NONE
    };

    function getDefaultStrings() {
        return {
            /**
             * The title of the dialog.
             * @property title {String}
             * @optional
             * @default Depends on the dialog type, see `type` property
             */
            //title,
            /**
             * The message to display inside the dialog.
             *
             * This should include the cause of the error and potential resolution steps.
             * This differs from the `uiContext` which is a brief description of what the user was doing when the dialog was shown.
             *
             * @property message {String}
             * @optional
             * @default ""
             */
            message : '',
            /**
             * A string representing an icon to display. You must specify the full markup.
             *
             * This property is only used if you set the `type` to CUSTOM.
             *
             * @property icon {String}
             * @deprecated Use a provided icon by changing the `type` property.
             * @optional
             * @default Depends on the dialog type, see `type` property
             */
            icon : '<i class="pgbu-icon pgbu-icon-warning"></i>',
            /**
             * The text to display on the default Ok button
             * @property ok {String}
             * @optional
             * @default "Ok" - localized
             */
            ok : lang.getString('dialogs.ok'),
            /**
             * The text to display as the label for more details
             * @property additionalDetailsLabel {String}
             * @optional
             * @default "Details" - localized
             */
            additionalDetailsLabel : lang.getString('dialogs.additionalDetailsLabel')
            /**
             * The text to display in the additional details section
             * @property additionalDetails {String}
             * @optional
             */
            //additionalDetails
        };
    }

    function bindOptionsCallbacks(options, dfd) {
        if (_.isFunction(options.okCallback)) {
            dfd.done(options.okCallback);
        }
    }

    function AlertDialog(options) {
        var that = this,
            dfd = $.Deferred();

        options = _.defaults({}, options, DEFAULT_OPTIONS, getDefaultStrings(), {
            title : dialogTypes.titles[options.type || DEFAULT_OPTIONS.type]
        });

        if (options.iconClass) {
            logger.warn('The option "iconClass" is deprecated, use "icon" instead.');
            options.icon = '<i class="' + options.iconClass + '"></i>';
        }

        if (!_.isString(options.doNotShowAgain)) {
            options.doNotShowAgain = false;
        }

        /**
         * Adds a method which will be called when the user clicks the Ok button. This method will *not* be called if
         * the user closes the modal through the Close (X) button in the corner or with the escape button.
         *
         * The callback takes no arguments.
         *
         * @method ok
         * @param fn {Function} The callback function
         * @chainable
         * @returns {AlertDialog} The alert dialog instance
         */
        this.ok = dfd.promise().done;

        if (options.doNotShowAgain !== false && doNotShowHelper.isDoNotShow(options.doNotShowAgain)) {
            dfd.resolve();
            bindOptionsCallbacks(options, dfd);
            return; // Don't bother constructing the rest of the dialog if we're not showing it
        }

        this.$modal = $(alertTemplate({
            title : options.title,

            uiContext : options.uiContext,
            message : options.message,

            ok : options.ok,
            close : lang.getString('dialogs.close'),

            hasIcon : options.type !== dialogTypes.types.NONE,
            hasIconType : options.type !== dialogTypes.types.NONE,
            hasIconString : options.type === dialogTypes.types.CUSTOM,
            iconType : dialogTypes.icons[options.type],
            iconString : options.icon,

            hasAdditionalDetails : !!options.additionalDetails,
            additionalDetailsLabel : options.additionalDetailsLabel,
            additionalDetails : options.additionalDetails
        }));

        if (options.doNotShowAgain !== false) {
            this.$modal.find('.modal-footer').prepend(doNotShowTemplate({
                doNotShowAgain : lang.getString('dialogs.dontShowAgain')
            }));
        }

        this.$modal.find('.additional-details-label').on('click', function (event) {
            event.preventDefault();
            this.$modal.find('.additional-details').toggleClass('open');
        }.bind(this));

        this.$modal.find('[data-dismiss=modal]').on('click', function (event) {
            // Do not toggle the do not show
            dfd.resolve();
            event.preventDefault();
            event.stopPropagation();
        });

        this.$modal.one('keyup', function (ev) {
            if (ev.keyCode === keyCode.ESC) {
                dfd.resolve();
            }
        });

        this.$modal.find('.proceedModal').on('click', function (event) {
            if (options.doNotShowAgain !== false) {
                doNotShowHelper.toggleDoNotShow(options.doNotShowAgain, that.$modal.find('#doNotShowAgain').prop('checked'));
            }

            dfd.resolve();

            event.stopPropagation();
        });

        this.$modal.on('shown', function () {
            $('.proceedModal').focus();
        });

        this.$modal.one('hidden', function () {
            that.$modal.remove();
        });

        dfd.always(function () {
            that.$modal.modal('hide');
        });

        bindOptionsCallbacks(options, dfd);

        accessibilityUtils.modalAppend(this.$modal);

        this.$modal.modal({
            show : options.autoDisplay && (dfd.state() === 'pending'),
            backdrop : 'static'
        });
    }

    AlertDialog.prototype = {
        constructor : AlertDialog,

        /**
         * Shows the modal
         * @method show
         */
        show : function () {
            if (this.$modal) {
                this.$modal.modal('show');
            }
        }
    };

    /**
     * The types available to this dialog.
     *
     * See also the `type` property and [dialogTypes](ui.dialogs.dialogType.html).
     *
     * @property types
     * @static
     */
    Object.defineProperty(AlertDialog, 'types', {
        get : function () {
            return dialogTypes.types;
        }
    });

    return AlertDialog;
});

/**
 * @module ui.dialogs
 */
/**
 * This module provides a simple way to handle server-side errors of the Prime format with an alert dialog.
 *
 * Quick example of the most common usage:
 *
 *     define(['jquery', 'primutils/ui/dialogs/AJAXErrorHandler', function ($, AJAXErrorHandler) {
 *         $.ajax({
 *             // ...
 *             error: AJAXErrorHandler.callbackForTaskKey('message.my_task_specific_error_message_key'),
 *             // ...
 *         });
 *     });
 *
 * The key provided here is known as the UI Error Context.
 * It serves as a reminder of the context of what the user was attempting to do.
 *  'Unable to create new tab', 'Unable to save data', and so on.
 *  This is displayed alongside whatever message we get from the server (defaulting to unknown err msg).
 *  Generally, the message from the server provides the cause and/or action information that
 *  will help the user to know what to do now that the error has occurred.
 *
 * For a reference on writing error messages, please refer to the writing guide:
 *  http://pgbulwf.us.oracle.com/style-guide/assets/writing-guide/
 *
 * Full constructor example:
 *
 *     define(['jquery', 'localizer', 'primutils/ui/dialogs/AJAXErrorHandler',
 *             function ($, localizer, AJAXErrorHandler) {
 *
 *         var myErrorHandler = new AJAXErrorHandler({
 *             taskMessage: localizer.getString('message.my_task_specific_error_message_key')
 *         });
 *         $.ajax({
 *             // ...
 *             error: myErrorHandler.errorCallback
 *             // ...
 *         });
 *     });
 *
 * The options that can be passed into the full constructor are documented below, in the
 *  Properties tab.
 *
 * Both versions work just fine with promises, too.  As examples:
 *
 *     define(['jquery', 'primutils/ui/dialogs/AJAXErrorHandler'], ($, AJAXErrorHandler) {
 *         $.ajax(...)
 *             .fail(AJAXErrorHandler.callbackForTaskKey(...));
 *
 *         // or maybe
 *         $.when($.ajax(...), $.ajax(...))
 *             .fail(AJAXErrorHandler.callbackForTaskKey(...));
 *     });
 *
 * *Deprecation notice*: the option appendTo has been removed and AJAXErrorHandler will warn
 *  if it is present.
 *  Instead of relying on this option, AJAXErrorHandler will look for any visible modals, and append to them
 *  instead of document.body, if necessary, automatically.
 *
 * @class ui.dialogs.AJAXErrorHandler
 */
define('primutils/ui/dialogs/AJAXErrorHandler',[
    'jquery',
    'lodash',
    'primutils/utils/language',
    'primutils/i18n/Localizer',
    'logger',
    'primutils/ui/dialogs/AlertDialog'
], function ($, _, lang, Localizer, logger, AlertDialog) {
    'use strict';

    var localizer;
    var DEFAULT_OPTIONS = {
        /**
         * This option will be used as the UI Error Context.
         * The UI Error Context is shown "above the fold", to provide context as a reminder of what the user was attempting to do when the error occurred.
         *
         * *Note*: there should be no error code for UI Error Contexts.
         * See also: [Deprecation Notice on Confluence](https://confluence.oraclecorp.com/confluence/pages/viewpage.action?pageId=33854310)
         *
         * @property taskMessage
         * @type {String}
         */
        taskMessage : null,
        /**
         * This option can be used for two purposes.  First, it can be used to conveniently
         *  execute conditional logic for errors while taking advantage of AJAXErrorHandler's
         *  error object normalization work.  Second, it can be used to suppress the modal that
         *  would otherwise pop up.
         *
         * A value of false indicates that no error processing is needed and AJAXErrorHandler's
         *  default logic will be used.
         *
         * As an example, the following example would log an error and then prevent the error
         *  modal from appearing in some circumstances:
         *
         *     $.ajax({
         *         // ...
         *         error: AJAXErrorHandler.callbackForOptions({
         *             taskMessage: '...', // Fill me in with a localized message!
         *             processError: function (error, callbacks) {
         *                 if (error.someKeySentByServer === true) {
         *                     logger.error('Suppressing modal for some reason');
         *                     callbacks.suppressModal();
         *                 }
         *             }
         *         }),
         *         // ...
         *     });
         *
         * *Warning*: this option is not intended as a hook for you to try and make sense out of server errors, and to present something better.
         * If you want a better error message, that work has to be done server side.
         *
         * @property processError
         * @default false
         * @type {false|Function}
         */
        processError : false,
        /**
         * This option can be used to ignore errors if the AJAX request was aborted.
         *  (i.e. if the second argument to ajax's error function was 'abort').
         *  If this option is true (the default), such errors will be ignored and no modal
         *  will pop up.
         *
         * @property ignoreAborts
         * @default true
         * @type {true|false}
         */
        ignoreAborts : true,
        /**
         * Allows configuring the keys that will be present on error objects sent by server code.
         *
         * This **should not** be configured per error handler instance - only at the application level, through [`defaults`](#method_defaults).
         *
         * Recognized keys, with their defaults, as used in Prime:
         *
         *     recordKeys: {
         *         isLocalized: 'localizedException',
         *         isIgnored: 'isIgnored',
         *         errorCode: 'errorCode',
         *         message: 'message',
         *         exceptionType: 'exceptionType',
         *         wrappedExceptions: 'wrappedExceptions',
         *     }
         *
         * @property recordKeys
         * @type {Object}
         */
        recordKeys : {
            isLocalized : 'localizedException',
            isIgnored : 'isIgnored',
            errorCode : 'errorCode',
            message : 'message',
            exceptionType : 'exceptionType',
            wrappedExceptions : 'wrappedExceptions'
        }
    };
    var userDefaults = {};

    function AJAXErrorHandler(options) {
        this._options = _.defaults({}, options, this._options, userDefaults, DEFAULT_OPTIONS);
        this._keys = _.defaults({}, this._options.recordKeys, DEFAULT_OPTIONS.recordKeys);

        this.errorCallback = this.errorCallback.bind(this);
        this.errorCallbackObject = this.errorCallbackObject.bind(this);
    }

    var ehp = AJAXErrorHandler.prototype;

    /**
     * An errorCallback that can be handed directly to jQuery's AJAX error argument.
     *  This function will use the jqXHR object to parse the response from the server,
     *  and will use that message "below" the "show additional details" toggle.
     *
     * Example usage can be seen in the 'full constructor' example in the overall class
     *  description.
     *
     * @method errorCallback
     * @param {jqXHR}
     */
    ehp.errorCallback = function (jqXHR, textStatus, errorThrown) {
        var error = jqXHR;
        if (jqXHR && jqXHR.responseText && jqXHR.status) {
            this._handleAJAXError(jqXHR, textStatus, errorThrown);
        } else if (error instanceof Error) {
            this._handleJSError(error);
        } else if (_.isObject(error)) {
            this._handleObjectError(error);
        } else {
            logger.warn('Unrecognized error type for AJAXErrorHandler, going to popup unexpected error message');
            this._handleObjectError();
        }
    };

    /**
     * Deprecated.  Just use `errorCallback`.
     *
     * @method errorCallbackObject
     * @param {Object} obj corresponding to server-side LocalizedErrorRO
     * @deprecated
     */
    ehp.errorCallbackObject = function (errorObject) {
        this._handleObjectError(errorObject);
    };

    // When an $.ajax request fails honestly
    ehp._handleAJAXError = function (jqXHR, textStatus /*errorThrown*/) {
        var errorObject;

        try {
            errorObject = JSON.parse(jqXHR.responseText);
        } catch (err) {
            // Do nothing
        }

        errorObject = this._normalizeError(errorObject);

        // TODO https://jira.oraclecorp.com/jira/browse/PRIM-2365
        //  When we do the above, remove this code, since AJAXErrorHandler would then
        //  itself be responsible for wiping out page content
        if (errorObject[this._keys.isLocalized] && errorObject[this._keys.isIgnored]) {
            return;
        }

        if (this._options.ignoreAborts === true && textStatus === 'abort') {
            return;
        }

        this._handleError(errorObject);
    };

    // Really only applies to old file upload dialogs?, but a good fallback anyway maybe
    ehp._handleObjectError = function (errorObject) {
        errorObject = this._normalizeError(errorObject);
        this._handleError(errorObject);
    };

    // e.g. when used with 'primutils/utils/promisifyError', takes a JS error, throws up the 'unexpected error' message, and propagates the error to window for easier debugging
    ehp._handleJSError = function (err) {
        this._handleError(this._normalizeError());
        // if we bring in tracekit:
        //tracekit.report(err); // tracekit.report throws the error, so make sure to do it last
        throw err;
    };

    ehp._normalizeError = function (errorObject) {
        if (!_.isObject(errorObject)) {
            errorObject = {};
        }
        if (errorObject[this._keys.isLocalized] !== true || !errorObject[this._keys.message]) {
            if (lang.hasString('dialogs.unknownErrorCode')) {
                errorObject[this._keys.errorCode] = lang.getString('dialogs.unknownErrorCode');
            }
            errorObject[this._keys.message] = lang.getString('dialogs.unknownError');
            errorObject[this._keys.exceptionType] = 'ERROR';
        }
        return errorObject;
    };

    ehp._handleError = function (errorObject) {
        if (!_.isFunction(this._options.processError)) {
            this._prepareDialog(errorObject);
            return;
        }

        var suppressModal = false;
        var callbacks = {
            suppressModal : function () {
                suppressModal = true;
            }
        };
        this._options.processError.call(undefined, errorObject, callbacks);
        if (!suppressModal) {
            this._prepareDialog(errorObject);
            return;
        }
    };

    ehp._prepareDialog = function (errorObject) {
        var dialogOptions = {};
        var errorMessage = errorObject[this._keys.message];
        var postfix = '';

        if (this._options.replaceErrorCodeWithLink) {
            var errorCode = errorObject[this._keys.errorCode];
            errorMessage = this._options.errorCodeMessageStrip(errorMessage);
            postfix = Localizer.formatString(this._options.errorCodeLinkTemplate, errorCode);
        }
        if (errorObject[this._keys.wrappedExceptions] && errorObject[this._keys.wrappedExceptions].length) {
            errorMessage = this._nestMessages(errorObject);
        }

        if (errorObject.exceptionCause) {
            logger.error('Error info from failed request:');
            logger.error(errorObject.exceptionCause);
        }

        dialogOptions = {
            hideInitial : true,
            type : AlertDialog.types[errorObject[this._keys.exceptionType] || 'ERROR'],
            uiContext : this._options.taskMessage,
            message : errorMessage + postfix,
            okCallback : this._options.okCallback || $.noop
        };
        this._dialog = new AlertDialog(dialogOptions);

        if (this._options.appendTo) {
            logger.warn('The appendTo option of AJAXErrorHandler is deprecated; it will automatically detect any modals present on screen and append to them instead of body if necessary');
        }

        this._dialog.show();
    };

    // Nesting strategy; say we have a BulkError for excel imports.
    // Say row 2 has 2 errors, and row 3 has 4 errors; the end result is that we want the bulk error 'header' and then 6 errors in the dialog.
    // So, in effect, we're flattening out the bulk hierarchy here, on the client.
    ehp._nestMessages = function (errorObject, nowRecursive) {
        var errorMessage = '';
        if (!nowRecursive || !errorObject[this._keys.wrappedExceptions]) {
            errorMessage = '<p>' + errorObject[this._keys.message] + '</p>';
        }
        if (errorObject[this._keys.wrappedExceptions]) {
            errorObject[this._keys.wrappedExceptions].forEach(function (wrappedException) {
                errorMessage += '<p>' + this._nestMessages(this._normalizeError(wrappedException), true) + '</p>';
            }, this);
        }
        return errorMessage;
    };

    /**
     * A static helper method for constructing an error callback with all the given options.
     *  As ```new AJAXErrorHandler(options).errorCallback```.
     *
     * @method callbackForOptions
     * @static
     * @param {Object} options the options for constructing the error handler
     * @returns {Function} errorFn an error function that can be passed directly to jQuery's ajax
     *  error option
     */
    AJAXErrorHandler.callbackForOptions = function (options) {
        return new AJAXErrorHandler(options).errorCallback;
    };

    /**
     * A static helper method for getting an error callback with the given message.
     *  As AJAXErrorHandler.callbackForTaskKey, but assumes that the string given to it
     *  is already localized.
     *
     * @method callbackForMessage
     * @static
     * @param {String} message the UI Error Context message
     * @returns {Function} errorFn an error function that can be passed directly to jQuery's AJAX
     *  error option
     */
    AJAXErrorHandler.callbackForMessage = function (message) {
        if (arguments[1]) {
            logger.warn('The appendTo option of AJAXErrorHandler is deprecated; it will automatically detect any modals present on screen and append to them instead of body if necessary');
        }
        return AJAXErrorHandler.callbackForOptions({
            taskMessage : message
        });
    };

    /**
     * A static helper method for getting an error callback with the given message string key.
     *  See overall class description for an example.
     *
     * @method callbackForTaskKey
     * @static
     * @param {String} taskKey a key, to be passed through the localizer, representing the UI Error Context message
     * @returns {Function} errorFn an error function that can be passed directly to jQuery's AJAX
     *  error option
     */
    AJAXErrorHandler.callbackForTaskKey = function (taskKey) {
        if (arguments[1]) {
            logger.warn('The appendTo option of AJAXErrorHandler is deprecated; it will automatically detect any modals present on screen and append to them instead of body if necessary');
        }
        return AJAXErrorHandler.callbackForMessage(localizer.getString(taskKey));
    };

    AJAXErrorHandler.useLocalizer = function (newLocalizer) {
        logger.assert(newLocalizer instanceof Localizer, 'Templates: Localizer must be an instance of the type Localizer');
        localizer = newLocalizer;
    };

    /**
     * A static method that allows AJAXErrorHandler default options to be overriden at the application level.
     *
     * These defaults take higher precendence than the default hardcoded default options, but lower than any other options that may be imposed by a error handler subclass.
     *
     * @method defaults
     * @param defaults {Object} the defaults to be used
     * @static
     */
    AJAXErrorHandler.defaults = function (defaults) {
        _.assign(userDefaults, defaults);
    };

    AJAXErrorHandler.clearDefaults = function () {
        userDefaults = {};
    };

    Object.defineProperty(AJAXErrorHandler, 'types', {
        get : function () {
            return AlertDialog.types;
        }
    });

    return AJAXErrorHandler;
});



define('primutils/formatters/wrappers',[], function () {
    'use strict';

    return {
        timezoneSensitiveDateTimeUnserializeValidation : function (origFn, options, str) {
            if (str == null || str === '') {
                return null;
            } else if (typeof str !== (typeof '')) {
                throw new Error('All dates must be serialized as Strings');
            } else if (str && str.indexOf('.') === -1) {
                throw new Error('TimezoneSensitive dates must include timezone marker');
            } else {
                var result = origFn(options, str);
                if (!result || !result.isValid()) {
                    throw new Error('Invalid date string - \'' + str + '\'');
                }
                return result;
            }
        },

        globalDateUnserializeValidation : function (origFn, options, str) {
            if (str == null || str === '') {
                return null;
            } else if (typeof str !== (typeof '')) {
                throw new Error('All dates must be serialized as Strings');
            } else if (str && str.indexOf('.') !== -1) {
                throw new Error('Global dates must not include timezone marker');
            } else {
                var result = origFn(options, str);
                if (!result || !result.isValid()) {
                    throw new Error('Invalid date string - \'' + str + '\'');
                }
                return result;
            }
        },

        globalDateTimeUnserializeValidation : function (origFn, options, str) {
            if (str == null || str === '') {
                return null;
            } else if (typeof str !== (typeof '')) {
                throw new Error('All dates must be serialized as Strings');
            } else if (str && str.indexOf('.') !== -1) {
                throw new Error('Global dates must not include timezone marker');
            } else {
                var result = origFn(options, str);
                if (!result || !result.isValid()) {
                    throw new Error('Invalid date string - \'' + str + '\'');
                }
                return result;
            }
        },

        integerValidation : function (origFn, options, str) {
            var origResult = origFn(options, str);
            var decimalSeparator = options.decimalSeparator === '\\.' ? '.' : options.decimalSeparator;
            if (origResult === true && str.indexOf(decimalSeparator) !== -1) {
                return false;
            }
            return origResult;
        }
    };
});



define('primutils/formatters/formats',{
    timezoneSensitiveDateTimeSerialization : 'YYYY-MM-DD[T]HH:mm:ss.ZZ',
    timezoneSensitiveDateTimeUnserialization : 'YYYY-MM-DD[T]HH:mm:ss.ZZ',

    globalDateSerialization : 'YYYY-MM-DD[T00:00:00]',
    globalDateUnserialization : 'YYYY-MM-DD[T00:00:00]',
    globalDateTimeSerialization : 'YYYY-MM-DD[T]HH:mm:ss',
    globalDateTimeUnserialization : 'YYYY-MM-DD[T]HH:mm:ss'
});


/**
 * @module utils
 */
define('primutils/utils/isEqual',[
    'logger',
    'lodash',
    'moment',
    'primutils/utils/dates'
], function (logger, _, moment, dateUtils) {
    'use strict';

    var momentProto = Object.getPrototypeOf(moment()),
        durationProto = Object.getPrototypeOf(moment.duration()),
        warnedMoment = false,
        warnedDuration = false;

    function compareMoments(a, b) {
        return dateUtils.isSame(a, b);
    }

    function comparePseudoMoments(a, b) {
                if (!warnedMoment) {
            logger.warn('Moments should be cloned with the moment .clone method or though "primutils/utils/objectUtils" utilities');
            warnedMoment = true;
        }
                return compareMoments(moment(a), moment(b));
    }

    function compareMomentAndOther(value, other) {
        if (moment.isMoment(other)) {
            if (momentProto.isPrototypeOf(value) && momentProto.isPrototypeOf(other)) {
                return compareMoments(value, other);
            } else {
                return comparePseudoMoments(value, other);
            }
        }
        return false;
    }

    function compareDurations(a, b) {
        return (a.valueOf() === b.valueOf());
    }

    function comparePseudoDurations(a, b) {
                if (!warnedDuration) {
            logger.warn('Durations should be cloned with the moment.duration(val) or though "primutils/utils/objectUtils" utilities');
            warnedDuration = true;
        }
                return compareDurations(moment.duration(a), moment.duration(b));
    }

    function compareDurationAndOther(dur, other) {
        if (moment.isDuration(other)) {
            if (durationProto.isPrototypeOf(dur) && durationProto.isPrototypeOf(other)) {
                return compareDurations(dur, other);
            } else {
                return comparePseudoDurations(dur, other);
            }
        }
        // ALWAYS DEFER TO MOMENT FOR CHECKING, EVEN THOUGH IT RELIES ON A PROTOTYPE CHECK
        // DURATIONS CLONED WITH JQUERY BE DARNED...

        return false;
    }

    function checkForAndCompareMoments(a, b) {
        if (moment.isMoment(a)) {
            return compareMomentAndOther(a, b);
        } else if (moment.isMoment(b)) {
            return compareMomentAndOther(b, a);
        } else if (moment.isDuration(a)) {
            return compareDurationAndOther(a, b);
        } else if (moment.isDuration(b)) {
            return compareDurationAndOther(b, a);
        }
    }

    /**
     * @class utils.isEqual
     * @module protected
     * @protected
     */

    /**
     * Performs a deep comparison between two values to determine if they are equivalent to each other.
     * If a callback is provided it will be executed to compare values.
     * If the callback returns undefined comparisons will be handled by the method instead.
     *
     * The callback is bound to `thisArg` and invoked with two arguments; `(a, b)`.
     *
     * This version is exactly the same as the version in Lo-Dash, except that it has built-in support for moment objects using moments native `isSame` function.
     * The moment check occurs after the provided callback, so you can override the default moment comparison as desired.
     *
     * @method isEqual
     * @param {*} a The first value to compare
     * @param {*} b The second value to compare
     * @param {Function} [callback] The function to customize comparing values
     * @param {*} [thisArg] The `this` binding of the `callback`
     * @return {Boolean} whether the two objects are equivalent
     */
    var isEqual = _.wrap(_.isEqual, function (origFn, a, b, callback, thisArg) {
        callback = _.isFunction(callback) ? _.wrap(callback, function (origCallback, a, b) {
            var origResult = origCallback.call(this, a, b);
            if (!_.isUndefined(origResult)) {
                return origResult;
            }
            return checkForAndCompareMoments.call(this, a, b);
        }) : checkForAndCompareMoments;

        return origFn.call(this, a, b, callback, thisArg);
    });

    /**
     * Overrides the Lo-Dash default implementation of isEqual to use the Primutils version.
     *
     * @method mixinToLodash
     * @protected
     * @static
     */
    isEqual.mixinToLodash = function () {
        _.mixin({
            isEqual : isEqual
        });
    };

    return isEqual;
});

/**
 * Module containing support functions, and browser/library patches, for XSS protections.
 *
 * XSS protections are based on recommendations from [OWASP](https://www.owasp.org/index.php/XSS_%28Cross_Site_Scripting%29_Prevention_Cheat_Sheet#Output_Encoding_Rules_Summary).
 *
 * All methods documented here are static functions.
 * For example, `unescapeSingleChar` can be used as:
 *
 *     define(['primutils/urls/escaping'], function (escapes) {
 *         escapes.unescapeSingleChar('&amp;'); // would return '&' if patching has been enabled
 *     });
 *
 * -----
 *
 * **Warning** for Prime developers: matters of browser/library patching are already handled for you by our global bootstrapping infrastructure.
 * You will not need to call the `patch*` methods.
 *
 * @class utils.escaping
 * @module utils
 */
define('primutils/utils/escaping',[
    'jquery',
    'lodash',
    'primutils/utils/escapingCommon'
], function ($, _, common) {
    'use strict';

    // The only 'state' of this module - primutils internally will avoid some unescape/escape calls (e.g. in [`urls.template`](utils.urls#method_template)) if this is false
    var escapingEnabled = false;

    // List of HTML entities for escaping.
    var entityMap = {
        unescape : common.unescapeMap,
        escape : common.escapeMap
    };

    function maybeEscape(str) {
        if (escapingEnabled) {
            return _.escape(str);
        }
        return str;
    }
    function maybeUnescape(str) {
        if (escapingEnabled) {
            return _.unescape(str);
        }
        return str;
    }

    var origEscape = _.escape;
    var origUnescape = _.unescape;
    function unpatchLodashEscaping() {
        _.escape = origEscape;
        _.unescape = origUnescape;
    }
    function patchLodashEscaping() {
        // internal escape(), unescape, adapted from Underscore.js
        //  but updated with the actual entity encodes that the PGBU is using


        // Regexes containing the keys and values listed immediately above.
        var entityRegexes = {
            escape : new RegExp('[' + _.keys(entityMap.escape).join('') + ']', 'g'),
            unescape : new RegExp('(' + _.keys(entityMap.unescape).join('|') + ')', 'g')
        };

        // Functions for escaping and unescaping strings to/from HTML interpolation.
        _.forEach(['escape', 'unescape'], function (method) {
            _[method] = function (string) {
                if (string == null) {
                    return '';
                }
                return ('' + string).replace(entityRegexes[method], function (match) {
                    return entityMap[method][match];
                });
            };
        });
    }

    var origValHooks = _.cloneDeep($.valHooks);
    var origText = $.fn.text;
    function patchJQueryMethods() {
        $.extend($.valHooks, {
            search : {
                get : function (elem) {
                    return _.escape(elem.value);
                },
                set : function (elem, val) {
                    if (typeof val === 'string') {
                        elem.value = _.unescape(val);
                        return true;
                    }
                }
            },
            text : {
                get : function (elem) {
                    return _.escape(elem.value);
                },
                set : function (elem, val) {
                    if (typeof val === 'string') {
                        elem.value = _.unescape(val);
                        return true;
                    }
                }
            },
            textarea : {
                get : function (elem) {
                    return _.escape(elem.value);
                },
                set : function (elem, val) {
                    if (typeof val === 'string') {
                        elem.value = _.unescape(val);
                        return true;
                    }
                }
            }
        });

        var $textFn = $.fn.text;
        var $attrFn = $.fn.attr;
        $.fn.extend({
            text : function (value) {
                if (arguments.length === 1 && _.isString(value)) {
                    return $textFn.call(this, _.unescape(value));
                } else if (arguments.length === 1) {
                    return $textFn.call(this, value);
                } else {
                    return _.escape($textFn.call(this));
                }
            },
            attr : function (attr, value) {
                if (arguments.length === 2 && _.isString(value) && attr === 'title') {
                    return $attrFn.call(this, attr, _.unescape(value));
                } else if (arguments.length === 1 && attr === 'title') {
                    return _.escape($attrFn.call(this, attr));
                }
                return $attrFn.apply(this, arguments);
            }
        });
    }
    function unpatchJQueryMethods() {
        $.extend($.valHooks, origValHooks);
        $.fn.text = origText;
    }

    var nativeEncodeURI = window.encodeURI;
    var nativeEncodeURIComponent = window.encodeURIComponent;
    var nativeDecodeURI = window.decodeURI;
    var nativeDecodeURIComponent = window.decodeURIComponent;
    function patchGlobalMethods() {
        window.encodeURI = function unescapingEncodeURI(uri) {
            return nativeEncodeURI.call(window, maybeUnescape(uri));
        };
        window.encodeURIComponent = function unescapingEncodeURIComponent(uri) {
            return nativeEncodeURIComponent.call(window, maybeUnescape(uri));
        };
        window.decodeURI = function unescapingDecodeURI(uri) {
            return nativeDecodeURI.call(window, maybeUnescape(uri));
        };
        window.decodeURIComponent = function unescapingDecodeURIComponent(uri) {
            return nativeDecodeURIComponent.call(window, maybeUnescape(uri));
        };
    }
    function unpatchGlobalMethods() {
        window.encodeURI = nativeEncodeURI;
        window.encodeURIComponent = nativeEncodeURIComponent;
        window.decodeURI = nativeDecodeURI;
        window.decodeURIComponent = nativeDecodeURIComponent;
    }

    var that = {
        /**
         * The star function of this module - patches Lo-Dash, jQuery, and native browser methods to support our XSS protections.
         *
         * Make sure this is called before anything in your page happens - see [How Prime Works](https://confluence.oraclecorp.com/confluence/display/PRIMEENGINEERING/Global+Prepares) for more information.
         *
         * For informational purposes, this patches:
         * * `_.escape` and `_.unescape` to match the characters Prime encodes
         * * `$input.val()` returns encoded strings (for search, text, and textarea elements)
         * * `$input.val('abc&amp;');` will display as 'abc&' (for search, text, and textarea elements)
         * * `$elem.text()` returns escaped strings
         * * `$elem.text('abc&amp;')` will display as 'abc&'
         * * `$elem.attr('title')` returns escaped strings
         * * `$elem.attr('title', 'abc&amp;')` sets a tooltip title of 'abc&' on the element
         * * `window.encodeURI` et all unescape before the standard url encoding
         *
         * @method patchAll
         * @static
         */
        patchAll : function patchAll() {
            if (escapingEnabled) {
                return;
            }
            escapingEnabled = true;
            patchLodashEscaping();
            patchJQueryMethods();
            patchGlobalMethods();
        },

        unpatchAll : function unpatchAll() {
            if (!escapingEnabled) {
                return;
            }
            escapingEnabled = false;
            unpatchLodashEscaping();
            unpatchJQueryMethods();
            unpatchGlobalMethods();
        },

        /**
         * Unescapes a single character.  Useful, for example, in making sure that the date
         *  separator is an actual '/' and not the entity-encoded equivalent.
         *
         * @method unescapeSingleChar
         * @static
         * @param {String} character a single entity-encoded character
         */
        unescapeSingleChar : function (ch) {
            if (escapingEnabled && entityMap.unescape[ch] != null) {
                return entityMap.unescape[ch];
            }
            return ch;
        },

        /**
         * Returns the given string with all non-alphanumberic characters
         *  substituted with encoded equivalents.
         *
         * This corresponds to the second row in OWASP's recommendations located here:
         *  https://www.owasp.org/index.php/XSS_(Cross_Site_Scripting)_Prevention_Cheat_Sheet#Output_Encoding_Rules_Summary
         *
         * *Example*
         *
         * The primary usage of this function comes with our use of templates.
         *  For an example:
         *
         *     // myTemplateUsingCode.js:
         *     myTemplate({
         *         someValue: urls.escapeForHTMLAttribute(someValue)
         *     });
         *
         * @param {String} str
         * @return the string, with all non-alphanumeric characters encoded
         *
         * @method escapeForHTMLAttribute
         * @static
         */
        escapeForHTMLAttribute : function escapeForHTMLAttribute(str) {
            var unencoded = _.unescape(str);
            return unencoded.replace(/[^a-zA-Z0-9]/g, function (char) {
                return '&#x' + char.charCodeAt(0).toString(16) + ';';
            });
        },

        /**
         * Sets the page title to the given string, unescaping if necessary.
         *
         * @method setPageTitle
         * @static
         * @param {String} title the title for the page
         */
        setPageTitle : function (title) {
            window.document.title = maybeUnescape(title);
        },

        /**
         * Like the native [substr](http://devdocs.io/javascript/global_objects/string/substr), but takes caution to not get tripped up by the special characters that Prime protects per OWASP guidelines.
         *
         * See https://www.owasp.org/index.php/XSS_(Cross_Site_Scripting)_Prevention_Cheat_Sheet#Output_Encoding_Rules_Summary for more info.
         *
         * @method substr
         * @static
         * @param str {String} the string to substr
         * @param min {Number} the start point at which to begin extracting characters
         * @param length {Number} the number of characters to extract
         */
        substr : function substr(str, min, max) {
            return maybeEscape(maybeUnescape(str).substr(min, max));
        },

        /**
         * Like the native [slice](http://devdocs.io/javascript/global_objects/string/slice), but takes caution to not get tripped up by the special characters that Prime protects per OWASP guidelines.
         *
         * See https://www.owasp.org/index.php/XSS_(Cross_Site_Scripting)_Prevention_Cheat_Sheet#Output_Encoding_Rules_Summary for more info.
         *
         * @method slice
         * @static
         * @param str {String} the string to slice
         * @param min {Number} the start point at which to begin extracting characters
         * @param length {Number} the number of characters to extract
         */
        slice : function slice(str, min, max) {
            return maybeEscape(maybeUnescape(str).slice(min, max));
        },

        /**
         * Like the native [length](http://devdocs.io/javascript/global_objects/string/length), but takes caution to not get tripped up by escaped special characters.
         *
         * See https://www.owasp.org/index.php/XSS_(Cross_Site_Scripting)_Prevention_Cheat_Sheet#Output_Encoding_Rules_Summary for more info.
         *
         * @method length
         * @static
         * @param str {String} the string to get the length of
         * @return len {Number} the number of unescaped characters in the string
         */
        length : function length(str) {
            return maybeUnescape(str).length;
        }
    };

    return that;
});

define('primutils/ui/widgets/dock',[
    'jquery',
    'lodash',
    'primutils/utils/keyCode',
    'primutils/utils/language'
], function ($, _, keyCode, lang) {
    'use strict';

    var ORIENTATIONS = {
        horizontal : 'dock-bottom',
        vertical : 'dock-right'
    };

    function getOrientation($dock) {
        if ($dock.hasClass('dock-left') || $dock.hasClass('dock-right')) {
            return 'vertical';
        } else if ($dock.hasClass('dock-top') || $dock.hasClass('dock-bottom')) {
            return 'horizontal';
        }
    }

    // TODO Implement remainder of ARIA splitter recommendations
    // End - max size
    // Home - min size
    // Enter - Restore previous position

    function Dock(el) {
        var that = this;
        this.$el = $(el);
        this.throttledResize = _.throttle(function () {
            that.$el.trigger('dock-resized');
        }, 250, { trailing : true });
    }

    Dock.prototype = {
        constructor : Dock,

        toggle : function (state) {
            state = (state == null) ? this.$el.hasClass('collapsed') : state;
            return this[state ? 'expand' : 'collapse']();
        },

        expand : function () {
            if (!this.$el.hasClass('collapsed')) {
                return;
            }

            var e = $.Event('dock-expanding');
            this.$el.trigger(e);

            if (e.isDefaultPrevented()) {
                return;
            }

            this.$el.removeClass('collapsed')
                .find('> .collapse-handle')
                .attr('title', lang.getString('dock.collapse'))
                .find('> .spoken-text')
                .text(lang.getString('dock.collapse'));
            this.$el.find('.resize-handle').attr('aria-expanded', true);
            this.$el.trigger($.Event('dock-expanded'));
        },

        collapse : function () {
            if (this.$el.hasClass('collapsed')) {
                return;
            }

            var e = $.Event('dock-collapsing');
            this.$el.trigger(e);

            if (e.isDefaultPrevented()) {
                return;
            }

            this.$el.addClass('collapsed')
                .find('> .collapse-handle')
                .attr('title', lang.getString('dock.expand'))
                .find('> .spoken-text')
                .text(lang.getString('dock.expand'));
            this.$el.find('.resize-handle').attr('aria-expanded', false);
            this.$el.trigger($.Event('dock-collapsed'));
        },

        orient : function (orientation) {
            if ((orientation === 'horizontal' && this.$el.hasClass(ORIENTATIONS.horizontal)) || (orientation === 'vertical' && this.$el.hasClass(ORIENTATIONS.vertical))) {
                return;
            }

            var e = $.Event('dock-before-orientation');
            this.$el.trigger(e, { orientation : orientation });

            if (e.isDefaultPrevented()) {
                return;
            }

            if (orientation === 'vertical') {
                this._storedHeight = this.$el.css('height');
            } else if (orientation === 'horizontal') {
                this._storedWidth = this.$el.css('width');
            }

            this.$el.css({
                height : (orientation === 'vertical') ? '' : this._storedHeight,
                width : (orientation === 'horizontal') ? '' : this._storedWidth
            });

            this.$el.removeClass('dock-left dock-right dock-top dock-bottom').addClass(ORIENTATIONS[orientation])
                .find('.resize-handle').attr('aria-orientation', orientation);
            this.$el.find('.dock-controls button').removeClass('active').filter('[data-orientation="' + orientation + '"]').addClass('active');

            this.$el.trigger($.Event('dock-orientation-changed'), { orientation : orientation });
        }
    };

    $.fn.extend({
        prim_dock : function (option) {
            var applyArgs = Array.prototype.slice.call(arguments, 1);
            return this.each(function () {
                var $this = $(this),
                    data = $this.data('prim_dock');

                if (!data) {
                    data = new Dock(this);
                    $this.data('prim_dock', data);
                }

                if (typeof option === 'string') {
                    data[option].apply(data, applyArgs);
                }
            });
        }
    });

    $.fn.prim_dock.Constructor = Dock;

    $(document)
        .on('click.dock.data-api', '.dock > .collapse-handle', function (e) {
            var $dock = $(e.target).closest('.dock');
            e.preventDefault();
            if ($dock.data('prevent-collapse-handle-click')) {
                $dock.data('prevent-collapse-handle-click', false);
            } else {
                $dock.prim_dock('toggle');
            }
        })
        .on('click.dock.data-api', '.dock .dock-controls button', function (e) {
            var $btn = $(e.currentTarget),
                orient = $btn.data('orientation'),
                $dock = $(e.target).closest('.dock');
            $dock.prim_dock('orient', orient);
        })
        .on('keydown', '.dock .resize-handle', function (e) {
            var $dock = $(e.target).closest('.dock'),
                dock,
                orientation = getOrientation($dock),
                multiplier = ($dock.hasClass('dock-left') || $dock.hasClass('dock-top')) ? -1 : 1,
                diff = 10 * multiplier;

            dock = $dock.prim_dock().data('prim_dock'); // Make sure that dock has been initialized internally

            if ($dock.hasClass('collapsed')) { return; }

            if (orientation === 'horizontal') {
                if (e.keyCode === keyCode.UP) {
                    $dock.css('height', (parseInt($dock.css('height'), 10) + diff) + 'px');
                    e.preventDefault();
                } else if (e.keyCode === keyCode.DOWN) {
                    $dock.css('height', (parseInt($dock.css('height'), 10) - diff) + 'px');
                    e.preventDefault();
                }
            } else if (orientation === 'vertical') {
                if (e.keyCode === keyCode.LEFT) {
                    $dock.css('width', (parseInt($dock.css('width'), 10) + diff) + 'px');
                    e.preventDefault();
                } else if (e.keyCode === keyCode.RIGHT) {
                    $dock.css('width', (parseInt($dock.css('width'), 10) - diff) + 'px');
                    e.preventDefault();
                }
            }

            if (e.isDefaultPrevented()) {
                dock.throttledResize();
            }
        })
        .on('mousedown.dock.data-api', '.dock .resize-handle, .dock > .collapse-handle', function (e) {
            var $target = $(e.target),
                $dock = $target.closest('.dock'),
                isCollapseHandle = $target.hasClass('collapse-handle'),
                moved = false;

            if ($dock.hasClass('collapsed')) { return; }

            $('body').css({
                '-webkit-user-select' : 'none',
                '-moz-user-select' : 'none',
                '-ms-user-select' : 'none',
                '-o-user-select' : 'none',
                'user-select' : 'none'
            });

            var lastX = e.pageX,
                lastY = e.pageY,
                orientation = getOrientation($dock),
                namespace = _.uniqueId('.pgbu-dock-resize-'),
                multiplier = ($dock.hasClass('dock-left') || $dock.hasClass('dock-top')) ? -1 : 1;

            var moveHandler = function (event) {
                moved = true;

                if (orientation === 'horizontal') {
                    $dock.css('height', (parseInt($dock.css('height'), 10) - (event.pageY - lastY) * multiplier) + 'px');
                } else if (orientation === 'vertical') {
                    $dock.css('width', (parseInt($dock.css('width'), 10) - (event.pageX - lastX) * multiplier) + 'px');
                }

                lastX = event.pageX;
                lastY = event.pageY;
            };

            var removeHandler = function () {
                $dock.data('prevent-collapse-handle-click', isCollapseHandle && moved);

                $('body').css({
                    '-webkit-user-select' : '',
                    '-moz-user-select' : '',
                    '-ms-user-select' : '',
                    '-o-user-select' : '',
                    'user-select' : ''
                });
                $target.off('dragstart' + namespace);
                $(document).off('mousemove' + namespace + ' mouseleave' + namespace + ' mouseup' + namespace);
                $dock.trigger('dock-resized');
            };

            $target.on('dragstart' + namespace, _.constant(false));

            $(document)
                .on('mousemove' + namespace, moveHandler)
                .on('mouseleave' + namespace + ' mouseup' + namespace, removeHandler);
        });

    var isResizing = false;
    function resizeAllDocksImpl() {
        // Wait until fixedSizePage has run
        if (!isResizing) {
            isResizing = true;
            _.defer(function () {
                $('.dock').each(function () {
                    $(this).trigger('dock-resized');
                });
                isResizing = false;
            });
        }
    }

    $(window).on('resize.dock.data-api', resizeAllDocksImpl);
});

define('primutils/ui/widgets/modal',[
    'jquery',
    'lodash',

    'bootstrap' // We are "extending" their functionality, so let's make sure that its loaded
], function ($, _) {
    'use strict';

    function startDraggable() {
        var $body = $('body');

        function attachMoveListeners($modal, originalX, originalY) {
            var originalOffset = $modal.offset();

            $modal.find('.modal-header').css('cursor', 'move');

            $body.css({
                '-webkit-user-select' : 'none',
                '-moz-user-select' : 'none',
                '-ms-user-select' : 'none',
                '-o-user-select' : 'none',
                'user-select' : 'none'
            })
                .on('mousemove.modal.data-api', _.throttle(function (ev) {
                    $modal.offset({
                        left : originalOffset.left + (ev.pageX - originalX),
                        top : originalOffset.top + (ev.pageY - originalY)
                    });
                }, 50))
                .on('mouseup.modal.data-api', function () {
                    $body.off('mousemove.modal.data-api mouseup.modal.data-api').css({
                        '-webkit-user-select' : '',
                        '-moz-user-select' : '',
                        '-ms-user-select' : '',
                        '-o-user-select' : '',
                        'user-select' : ''
                    });

                    $modal.find('.modal-header').css('cursor', '');
                });
        }

        $body
            .on('mousedown.modal.data-api', '.modal .modal-header [data-dismiss=modal], .modal .modal-header .control-bar', function (ev) {
                ev.stopPropagation();
            })
            .on('mousedown.modal.data-api', '.modal .modal-header', function (ev) {
                var $modal = $(this).closest('.modal');
                if ($modal.attr('data-draggable') !== 'false' && !$modal.hasClass('modal-full-page')) {
                    attachMoveListeners($modal, ev.pageX, ev.pageY);
                }
            });
    }
    startDraggable();

    var ModalPrototype = $.fn.modal.Constructor.prototype,
        origModalShow = ModalPrototype.show,
        origModalBackdrop = ModalPrototype.backdrop;

    // WARNING! Both this and a11y/bootstrap/modal.js wrap .show. It currently doesn't matter the order, but please note that both wrap.
    ModalPrototype.show = function () {
        // Clear drag position before reopening
        if (!this.isShown) {
            this.$element.css('left', '').css('top', '');
        }
        origModalShow.apply(this, arguments);
    };

    function makeRepositionCallback(originalCallback, $header, $footer, $controlBar) {
        return _.wrap(originalCallback || _.noop, function (origFn) {
            origFn.apply(this, _.rest(arguments));
            var padding = parseInt($header.css('padding-right'), 10);
            $footer.css('right', (2 * padding) + $controlBar.outerWidth(true));
        });
    }

    // WARNING! Both this and a11y/bootstrap/modal.js wrap .backdrop. It currently doesn't matter the order, but please note that both wrap.
    ModalPrototype.backdrop = function (callback) {
        var updatedCallback = callback;

        if (this.isShown && !this._repositionedFSFooter) {
            this._repositionedFSFooter = true;
            if (this.$element.hasClass('modal-full-page')) {
                var $header = this.$element.children('.modal-header');
                var $controlBar = $header.find('.control-bar');
                if ($controlBar.length) {
                    var $footer = this.$element.children('.modal-footer');
                    updatedCallback = makeRepositionCallback(callback, $header, $footer, $controlBar);
                }
            }
        }

        origModalBackdrop.call(this, updatedCallback);
    };

    // Make trays also behave like modals
    $.fn.tray = $.fn.modal;
});

define('primutils/ui/widgets/search',[
    'jquery'
], function ($) {
    'use strict';

    // TODO Implement remainder of ARIA splitter recommendations
    // End - max size
    // Home - min size
    // Enter - Restore previous position

    function Search(el) {
        this.$el = $(el);
    }

    Search.prototype = {
        constructor : Search,

        clear : function (/*state*/) {
            this.$el.find('input').val('');
            this.changed(false);
            this.$el.trigger('search-cleared');
        },

        changed : function (hasText) {
            this.$el.toggleClass('active', hasText);
            this.$el.find('.search-clear-icon').toggle(hasText);
        }
    };

    $.fn.extend({
        prim_search : function (option) {
            var applyArgs = Array.prototype.slice.call(arguments, 1);
            return this.each(function () {
                var $this = $(this),
                    data = $this.data('prim_search');

                if (!data) {
                    data = new Search(this);
                    $this.data('prim_search', data);
                }

                if (typeof option === 'string') {
                    data[option].apply(data, applyArgs);
                }
            });
        }
    });

    $.fn.prim_search.Constructor = Search;

    $(document)
        .on('keyup.search.data-api change.search.data-api', '.filter-div > input, .search-div > input', function () {
            var $this = $(this);
            $this.parent().prim_search('changed', $this.val().length > 0);
        })
        .on('click.search.data-api', '.filter-div > .search-clear-icon, .search-div > .search-clear-icon', function (e) {
            e.preventDefault();
            $(this).parent()
                .prim_search('clear')
                .find('input')
                .focus();
        });
});

define('primutils/ui/widgets/stickyHeader',[
    'jquery',
    'lodash',

    'bootstrap' // We are "extending" their functionality, so let's make sure that its loaded
], function ($, _) {
    'use strict';

    $(document.head).append('<style type="text/css">.affix-parent { position : relative; } .affix-sticky-header-top { top: 0; } .affix-sticky-header-bottom { position: absolute; top: auto; bottom: 0; } </style>');

    var Affix = $.fn.affix.Constructor,
        AffixPrototype = Affix.prototype;

    function StickyHeader(element, options) {
        var $element = $(element);
        this.$parent = options.parent ? $(options.parent) : $element.parent();
        this.$parent.addClass('affix-parent');
        Affix.call(this, element, {
            offset : {
                top : $element.offset().top
            }
        });
        $(window).on('resize.affix.data-api', $.proxy(function () { setTimeout($.proxy(this.resizeElement, this), 1); }, this));
    }

    function hasCSSWidth(styleString) {
        return /width\s*:/.test(styleString);
    }

    StickyHeader.prototype = _.assign(Object.create(AffixPrototype), {
        constructor : StickyHeader,

        checkPosition : function () {
            var originalWidth = this.$element.width(),
                cssWidth = hasCSSWidth(this.$element.attr('style')) ? originalWidth : null,
                previousState = this.affixed,
                scrollTop = this.$window.scrollTop();
            AffixPrototype.checkPosition.apply(this, arguments);
            this.$element.removeClass('affix-sticky-header-bottom');
            if (previousState !== false) {
                if (this.affixed === false) {
                    this._originalCSSWidth = cssWidth;
                    this.$element.width(originalWidth).addClass('affix-sticky-header-top');
                } else if (_.has(this, '_originalCSSWidth')) {
                    this.$element.css('width', this._originalCSSWidth || '').removeClass('affix-sticky-header-top');
                    delete this._originalCSSWidth;
                }
            }

            var isBottom = scrollTop > (this.$parent.outerHeight() + this.options.offset.top - this.$element.outerHeight());
            this.$element.toggleClass('affix-sticky-header-bottom affix-bottom', isBottom);
        },

        resizeElement : function () {
            if (this._originalCSSWidth == null && this.$element.hasClass('affix-sticky-header-top')) {
                this.affixed = null;
                var classes = this.$element.attr('class');
                this.$element.css('width', '').removeClass('affix affix-top affix-bottom affix-sticky-header-top affix-sticky-header-bottom');
                this.checkPosition();
                this.$element.addClass(classes);
            }
        }
    });

    $.fn.prim_stickyHeader = function (option) {
        return this.each(function () {
            var $this = $(this),
                data = $this.data('affix'),
                options = typeof option === 'object' && option;
            if (!data) {
                $this.data('affix', (data = new StickyHeader(this, options)));
            }
            if (typeof option === 'string') {
                data[option]();
            }
        });
    };

    return StickyHeader;
});


define('primutils/markup/primutils/ui/widgets/tabMenuItemTemplate.hbs', ['handlebars','primutils.markup'], function (H, u) {
var imports = u.buildImported(Array.prototype.slice.call(arguments, 2), 'primutils/ui/widgets/tabMenuItemTemplate.hbs'),
    func = H.template({"1":function(depth0,helpers,partials,data) {
  return "checked";
  },"3":function(depth0,helpers,partials,data) {
  return "";
},"5":function(depth0,helpers,partials,data) {
  return "style=\"padding-left: 20px;\"";
  },"7":function(depth0,helpers,partials,data) {
  var stack1, buffer = "<i class=\"check ";
  stack1 = helpers['if'].call(depth0, (depth0 != null ? depth0.isHideAllowed : depth0), {"name":"if","hash":{},"fn":this.program(3, data),"inverse":this.program(8, data),"data":data});
  if (stack1 != null) { buffer += stack1; }
  return buffer + "\"></i>";
},"8":function(depth0,helpers,partials,data) {
  return "hidden";
  },"compiler":[6,">= 2.0.0-beta.1"],"main":function(depth0,helpers,partials,data) {
  var stack1, lambda=this.lambda, buffer = "<li role=\"menuitem\">\n    <a href=\"#\" data-tab-id=\"";
  stack1 = lambda((depth0 != null ? depth0.target : depth0), depth0);
  if (stack1 != null) { buffer += stack1; }
  buffer += "\" class=\"";
  stack1 = helpers['if'].call(depth0, (depth0 != null ? depth0.visible : depth0), {"name":"if","hash":{},"fn":this.program(1, data),"inverse":this.noop,"data":data});
  if (stack1 != null) { buffer += stack1; }
  buffer += "\" ";
  stack1 = helpers['if'].call(depth0, (depth0 != null ? depth0.needToggleSpacing : depth0), {"name":"if","hash":{},"fn":this.program(3, data),"inverse":this.program(5, data),"data":data});
  if (stack1 != null) { buffer += stack1; }
  buffer += ">\n        ";
  stack1 = helpers['if'].call(depth0, (depth0 != null ? depth0.needToggleSpacing : depth0), {"name":"if","hash":{},"fn":this.program(7, data),"inverse":this.noop,"data":data});
  if (stack1 != null) { buffer += stack1; }
  buffer += "\n        <span class=\"tab-title\">";
  stack1 = lambda((depth0 != null ? depth0.title : depth0), depth0);
  if (stack1 != null) { buffer += stack1; }
  return buffer + "</span>\n    </a>\n</li>";
},"useData":true});
return function (context, options) {
    options = u.extendOptions(options || {}, imports);
    return func(context, options).trim();
};
});

define('primutils/ui/widgets/tab',[
    'jquery',
    'lodash',
    'logger',
    'primutils/utils/language',
    'primutils/markup/primutils/ui/widgets/tabMenuItemTemplate.hbs',

    'bootstrap' // We are "extending" their functionality, so let's make sure that its loaded
], function ($, _, logger, lang, tabMenuItemTemplate) {
    'use strict';

    $(document.head).append('<style type="text/css">.nav-tabs-scroll-container .nav-tabs>li.scroll-tab-user-hidden { display: none !important; visibility: hidden !important; }</style>');

    function Tabbar(el) {
        this.$element = $(el);
        this._fixResizeContainer();
    }

    Tabbar.prototype = {
        constructor : Tabbar,

        get $tabs() {
            return this.$element.children(':not([data-noresize])');
        },
        get $nonTabs() {
            return this.$element.children('[data-noresize]');
        },
        get $menuTabs() {
            return this.$tabs.add(this.$nonTabs);
        },

        _fixResizeContainer : function () {
            this.$resizeContainer = this.$element.closest('.nav-tabs-scroll-container');
        },

        resize : function () {
            this.$element.children('li').each(function () {
                if (this.style.display !== 'none') {
                    this.style.display = '';
                }
            });
            var $scrollers = this.$resizeContainer.find('.scrollers');
            var allowHide = this.$element.is('[data-allow-hide]') && this.$menuTabs.length > 1;
            var needsScroll;
            this.$resizeContainer.addClass('has-before-tabs has-after-tabs');
            needsScroll = this.$element.get(0).scrollWidth > this.$element.outerWidth();
            $scrollers.children().toggleClass('available', needsScroll);
            this.$resizeContainer.toggleClass('has-before-tabs', needsScroll);
            this.$resizeContainer.toggleClass('has-after-tabs', needsScroll || allowHide);
            this._scrollActiveIntoView();
        },

        _scrollActiveIntoView : function () {
            this.scrollTabIntoView(this.$element.children('.active'));
        },

        scrollTabIntoView : function ($tabLi) {
            if ($tabLi.length === 1) {
                var activeWidth = $tabLi.outerWidth(),
                    activeLeft = $tabLi.position().left,
                    elWidth = this.$element.outerWidth(),
                    elScroll = this.$element.scrollLeft();
                if (activeLeft + activeWidth < 0) {
                    this.$element.scrollLeft(elScroll + activeLeft);
                } else if (activeLeft > elWidth) {
                    this.$element.scrollLeft(elScroll + (activeLeft + activeWidth - elWidth));
                }
            }
        },

        scrollRight : function () {
            this.$element.scrollLeft(this.$element.scrollLeft() + 53);
        },

        scrollLeft : function () {
            this.$element.scrollLeft(this.$element.scrollLeft() - 53);
        },

        buildMenu : function () {
            var $parent = this.$resizeContainer.find('.tabs-menu'),
                hideAllowed = this.$element.is('[data-allow-hide]'),
                hideAllAllowed = this.$element.is('[data-allow-hide-all="true"]'),
                $nestedMenu, $eligibleTabs, $visibleEligibleTabs;

            if ($parent.hasClass('open')) {
                $nestedMenu = $parent.find('.dropdown-menu .nested-dropdown-menu').empty();
                $eligibleTabs = this.$menuTabs.filter(':visible,.scroll-tab-user-hidden');
                $visibleEligibleTabs = $eligibleTabs.filter(':visible').children('[data-target]');

                $eligibleTabs.each(function () {
                    var $tab = $(this).children('[data-target]'),
                        isVisible = !$(this).hasClass('scroll-tab-user-hidden'),
                        isOnlyVisible = hideAllowed && isVisible && $visibleEligibleTabs.length === 1 && $visibleEligibleTabs.first().is($tab);
                    $nestedMenu.append(tabMenuItemTemplate({
                        title : $tab.text(),
                        target : $tab.attr('data-target'),
                        visible : isVisible,
                        needToggleSpacing : hideAllowed,
                        isHideAllowed : hideAllowed && (hideAllAllowed || !isVisible || !isOnlyVisible)
                    }));
                });
            }
        },

        toggleTab : function (tabTarget) {
            var $tab = this.$menuTabs.find('[data-target="' + tabTarget + '"]').parent(),
                $dropdownMenu = this.$resizeContainer.find('.tabs-menu'),
                $visibleTabs = this.$menuTabs.filter(':visible').children('[data-target]'),
                hideAllAllowed = this.$element.is('[data-allow-hide-all="true"]'),
                shouldTrigger = false,
                doToggleTab = function () {
                    var nowChecked;
                    $tab.toggleClass('scroll-tab-user-hidden');
                    nowChecked = !$tab.hasClass('scroll-tab-user-hidden');
                    $dropdownMenu.find('[data-tab-id="' + tabTarget + '"]').toggleClass('checked', nowChecked);
                    $dropdownMenu.find('.checked > .check').toggleClass('hidden', !hideAllAllowed && ($visibleTabs.length === 2) && !nowChecked);
                    return nowChecked;
                },
                selection;

            if (!$tab.hasClass('active')) {
                var nowChecked = doToggleTab();
                if (nowChecked && $visibleTabs.length === 0) { // There _were_ 0 before
                    selection = document.activeElement;
                    this.selectTab(tabTarget, false);
                    $dropdownMenu.addClass('open');
                    $(selection).focus();
                }
                shouldTrigger = true;
            } else if (!($visibleTabs.length === 1 && $visibleTabs.first().attr('data-target') === tabTarget)) {
                selection = document.activeElement;
                doToggleTab();
                this.selectTab($visibleTabs.not($tab.children('[data-target]')).first().attr('data-target'), false);
                $dropdownMenu.addClass('open');
                $(selection).focus();
                shouldTrigger = true;
            } else if (hideAllAllowed) {
                doToggleTab();
                this.unselectAllTabs();
                shouldTrigger = true;
            }

            if (shouldTrigger) {
                $tab.trigger('tab-user-hidden-toggled');
            }
        },

        unselectAllTabs : function () {
            this.$element.children('.active').removeClass('active').each(function () {
                var $panelContainer = $(this).closest('.panel-container'),
                    $panel = $panelContainer.children('.tab-content').children($(this).children('[data-target]').attr('data-target'));
                $panel.removeClass('active');
            });
        },

        selectTab : function (tabTarget, focusTab) {
            var $tab = this.$menuTabs.find('[data-target="' + tabTarget + '"]'),
                $li = $tab.parent(),
                $dropdownMenu = this.$resizeContainer.find('.tabs-menu');

            if (arguments.length === 1) {
                focusTab = true;
            }

            var canSelect = $li.hasClass('scroll-tab-user-hidden') || $li.css('display') !== 'none';

            if (!$li.hasClass('active') && canSelect) {
                $tab.click();
                if ($li.hasClass('scroll-tab-user-hidden')) {
                    $li.removeClass('scroll-tab-user-hidden');
                    $dropdownMenu.find('[data-tab-id="' + tabTarget + '"]').toggleClass('checked', !$li.hasClass('scroll-tab-user-hidden'));
                    $tab.trigger('tab-user-hidden-toggled');
                }

                if (focusTab) {
                    $tab.focus();
                }
            }
        },

        collapse : function () {
            var e = $.Event('panel-collapsing');
            this.$element.trigger(e);

            if (e.isDefaultPrevented()) {
                return;
            }

            var expandStr = lang.getString('panel.expand');
            this.$element.closest('.panel-container').addClass('collapsed')
                .find('.nav-tabs .collapse-handle a')
                .attr('title', expandStr)
                .find('> i > .spoken-text')
                .text(expandStr);
            this.$element.trigger($.Event('panel-collapsed'));
        },

        expand : function () {
            var e = $.Event('panel-expanding');
            this.$element.trigger(e);

            if (e.isDefaultPrevented()) {
                return;
            }

            var collapseStr = lang.getString('panel.collapse');
            this.$element.closest('.panel-container').removeClass('collapsed')
                .find('.nav-tabs .collapse-handle a')
                .attr('title', collapseStr)
                .find('> i > .spoken-text')
                .text(collapseStr);
            this.$element.trigger($.Event('panel-expanded'));
        },

        toggle : function (state) {
            var $container = this.$element.closest('.panel-container');

            state = (state == null) ? $container.hasClass('collapsed') : state;

            if (state) {
                this.expand();
            } else {
                this.collapse();
            }
        }
    };

    $.fn.extend({
        prim_tabbar : function (option) {
            var applyArgs = Array.prototype.slice.call(arguments, 1);
            return this.each(function () {
                var $this = $(this),
                    data = $this.data('prim_tabbar');

                if (!data) {
                    data = new Tabbar(this);
                    $this.data('prim_tabbar', data);
                }

                if (typeof option === 'string') {
                    data[option].apply(data, applyArgs);
                }
            });
        }
    });

    $.fn.prim_tabbar.Constructor = Tabbar;

    function resizeAllTabbarsImpl() {
        $('.nav-tabs').each(function () {
            $(this).prim_tabbar('resize');
        });
    }
    var resizeAllTabbars = _.throttle(resizeAllTabbarsImpl, 100);

    $(window).on('resize.tab.data-api', resizeAllTabbars);
    $(resizeAllTabbars);

    if (window.MutationObserver) {
        (function () {
            var observer = new window.MutationObserver(resizeAllTabbars);
            var config = { childList : true, subtree : true };
            observer.observe(document.body, config);
        }());
    }

    $(document)
        .on('show.tab.data-api', '[data-toggle="tab"]', function (e) {
            var $next = $(this).parents('li'),
                $prev = $(e.relatedTarget).parents('li'),
                slideLeft = $next.index() < $prev.index();
            $next.add($prev).toggleClass('slide-left', slideLeft);
            $(this).closest('.nav-tabs-scroll-container').children('.nav-tabs').prim_tabbar('scrollTabIntoView', $(this).closest('li'));
        })
        .on('mousedown.tab.data-api', '.nav-tabs-scroll-container .scrollers > .left', function () {
            var $tabs = $(this).closest('.nav-tabs-scroll-container').children('.nav-tabs');
            var interval = window.setInterval(function () {
                $tabs.prim_tabbar('scrollLeft');
            }, 200);
            $tabs.prim_tabbar('scrollLeft');
            $(document).one('mouseup', function () {
                window.clearInterval(interval);
            });
        })
        .on('mousedown.tab.data-api', '.nav-tabs-scroll-container .scrollers > .right', function () {
            var $tabs = $(this).closest('.nav-tabs-scroll-container').children('.nav-tabs');
            var interval = window.setInterval(function () {
                $tabs.prim_tabbar('scrollRight');
            }, 200);
            $tabs.prim_tabbar('scrollRight');
            $(document).one('mouseup', function () {
                window.clearInterval(interval);
            });
        })
        .on('wheel.tab.data-api mousewheel.tab.data-api', '.nav-tabs-scroll-container .nav-tabs', function (e) {
            e.preventDefault();
            if (e.originalEvent.deltaY > 0) {
                $(this).closest('.nav-tabs-scroll-container').children('.nav-tabs').prim_tabbar('scrollRight');
            } else {
                $(this).closest('.nav-tabs-scroll-container').children('.nav-tabs').prim_tabbar('scrollLeft');
            }
        })
        .on('click.tab.data-api', '.nav-tabs-scroll-container .scrollers > .left, .nav-tabs-scroll-container .scrollers > .right', function (e) {
            e.preventDefault();
        })
        .on('click.tab.data-api', '.panel-container .collapse-handle a', function (e) {
            e.preventDefault();
            $(this).closest('.panel-container, .nav-tabs-scroll-container').children('.nav-tabs').prim_tabbar('toggle');
        })
        .on('click.tab.data-api', '.nav-tabs-scroll-container .tabs-menu > button', function (e) {
            e.preventDefault();
            $(this).closest('.nav-tabs-scroll-container').children('.nav-tabs').prim_tabbar('buildMenu');
        })
        .on('click.tab.data-api', '.nav-tabs-scroll-container .tabs-menu [data-tab-id] .check', function (e) {
            if ($(this).closest('.nav-tabs-scroll-container').find('.nav-tabs[data-allow-hide]').length === 0) {
                return;
            }
            e.preventDefault();
            e.stopPropagation();
            $(this).closest('.nav-tabs-scroll-container').children('.nav-tabs')
                .prim_tabbar('toggleTab', $(this).closest('[data-tab-id]').attr('data-tab-id'))
                .prim_tabbar('resize');
        })
        .on('click.tab.data-api', '.nav-tabs-scroll-container .tabs-menu [data-tab-id]', function (e) {
            e.preventDefault();
            $(this).closest('.nav-tabs-scroll-container').children('.nav-tabs')
                .prim_tabbar('selectTab', $(this).attr('data-tab-id'))
                .prim_tabbar('resize');
        })
        .on('dock-expanded.tab.data-api dock-orientation-changed.tab.data-api dock-resized.tab.data-api', '.dock', function () {
            $(this).find('.nav-tabs').prim_tabbar('resize');
        })
        .on('shown.tab.data-api', '.modal, .tray', function () {
            $(this).find('.nav-tabs').prim_tabbar('resize');
        });

    $('body')
        .on('mousedown.tab.data-api', '.nav-tabs[data-sortable]:not(.sort-ready)', function () {
            $(this).addClass('sort-ready').find('> li > [data-toggle="tab"]')
                .attr('draggable', 'false')
                .parent()
                .attr('draggable', 'true');
        })
        .on('dragstart.tab.data-api', '[data-sortable] [draggable="true"]', function (ev) {
            var $tab = $(this);
            var $tabbar = $tab.closest('.nav-tabs');
            var dragId = _.uniqueId('tab-drag-reorder-');
            var dragged = false;

            var beforeReorderEvent = $.Event('tabs-before-reordering');
            $tabbar.trigger(beforeReorderEvent);

            if (beforeReorderEvent.isDefaultPrevented()) {
                $tab = false;
                $tabbar = false;
                dragId = false;
                return;
            }

            var barLeft = $tabbar.offset().left - $(window).scrollLeft(),
                barRight = barLeft + $tabbar.width();

            if (ev.originalEvent.dataTransfer) {
                ev.originalEvent.dataTransfer.setData('text', $(this).text());
            }

            $tab.addClass('dragging');

            $tab.one('dragend.tab.data-api', function () {
                $tab.removeClass('dragging');
                $tabbar.off('.' + dragId);
                if (dragged) {
                    $tabbar.trigger('tabs-reordered');
                }
            });

            $tabbar
                .on('dragover.' + dragId, function (ev) {
                    ev.preventDefault();
                    var $swapTab = $(ev.target).closest('[draggable="true"]'),
                        pageX = ev.originalEvent.clientX - $(window).scrollLeft();

                    if ($swapTab.is($tab) || $swapTab.length === 0) {
                        return;
                    }

                    var swapTabPos = $swapTab.offset(),
                        swapTabHalfway = swapTabPos.left + ($swapTab.outerWidth(true) / 2);

                    if (pageX < swapTabHalfway) {
                        $tab.insertBefore($swapTab);
                    } else {
                        $tab.insertAfter($swapTab);
                    }
                    dragged = true;
                })
                .on('dragover.' + dragId, _.throttle(function (ev) {
                    var pageX = ev.originalEvent.clientX - $(window).scrollLeft();

                    if (pageX < (barLeft + 10)) {
                        $tabbar.prim_tabbar('scrollLeft');
                    } else if (pageX > (barRight - 10)) {
                        $tabbar.prim_tabbar('scrollRight');
                    }
                }, 250))
                .on('drop.' + dragId, function (ev) {
                    ev.preventDefault();
                });

            $tabbar.trigger('tabs-reordering');
        });

    // --------------------
    // Tab enhancements

    var Tab = $.fn.tab.Constructor,
        TabProto = Tab.prototype;

    // Need to be able to pass arguments to Bootstrap tab function
    $.fn.tab = function (option) {
        var applyArgs = Array.prototype.slice.call(arguments, 1);
        return this.each(function () {
            var $this = $(this),
                data = $this.data('tab');

            if (!data) {
                $this.data('tab', (data = new Tab(this)));
            }

            if (typeof option === 'string') {
                data[option].apply(data, applyArgs);
            }
        });
    };

    TabProto.toggle = function (state) {
        state = (state == null) ? !this.element.parent().is(':visible') : state;

        if (state) {
            this.available();
        } else {
            this.unavailable();
        }
    };

    TabProto.available = function () {
        this.element.parent().show();
        this.element.closest('.nav-tabs').prim_tabbar('resize');
    };

    TabProto.unavailable = function () {
        this.element.parent().hide();
        this.element.closest('.nav-tabs').prim_tabbar('resize');
    };
});

define('primutils/ui/widgets/wizard',[
    'jquery',
    'lodash',
    'logger',
    'primutils/templates/helpers/panel',

    'bootstrap' // We are "extending" their functionality, so let's make sure that its loaded
], function ($, _, logger, panelHelpers) {
    'use strict';

    function Wizard(el) {
        this.$element = $(el);
    }

    function getFakePanelContext() {
        return {
            type : 'wizard',
            tabsCreated : [],
            panelsCreated : []
        };
    }

    function noContent() {
        return '';
    }

    Wizard.prototype = {
        constructor : Wizard,

        goToStep : function (stepId, force) {
            if (!(force || this.canChangeStep(stepId))) {
                return;
            }

            var $step = this.$element.find('.steps > li').removeClass('active')
                .filter('[data-wizard-step="' + stepId + '"]').addClass('active');

            this.$element.find('.wizard-panel > .tab-pane').removeClass('active')
                .filter('#' + stepId).addClass('active');

            this._updateButtons();

            this.$element.trigger('wizard-step-changed', { step : stepId });
            $step.trigger('shown');
        },

        _updateButtons : function () {
            var nextStep = this.getNext(),
                prevStep = this.getPrev(),
                canNext = false, canPrev = false;

            if (nextStep) {
                canNext = this.canChangeStep(nextStep);
            }

            if (prevStep) {
                canPrev = this.canChangeStep(prevStep);
            }

            this.$element.find('[data-wizard=prev]').toggle(!!prevStep && canPrev);
            this.$element.find('[data-wizard=next]').toggle(!!nextStep && canNext);
            //this.$element.find('[data-wizard=cancel]').toggle(!prevStep);
            this.$element.find('[data-wizard=save]').toggle(!nextStep);
        },

        canChangeStep : function (newStep) {
            var current = this.getCurrent(),
                eventInfo = {
                    currentStep : current,
                    newStep : newStep
                },
                e = $.Event('wizard-can-change-step', eventInfo);
            this.$element.trigger(e, eventInfo);
            return e.result !== false;
        },

        get $steps() {
            return this.$element.find('.steps');
        },

        get $panels() {
            return this.$element.find('.wizard-panel');
        },

        get $currentLi() {
            return this.$steps.find(' > .active');
        },

        getCurrent : function () {
            return this.$currentLi.attr('data-wizard-step');
        },

        getNext : function () {
            return this.$currentLi.nextAll(':visible').first().attr('data-wizard-step');
        },

        next : function () {
            var nextStep = this.getNext();
            if (nextStep) {
                this.goToStep(nextStep);
            } else {
                this._updateButtons();
            }
        },

        getPrev : function () {
            return this.$currentLi.prevAll(':visible').first().attr('data-wizard-step');
        },

        prev : function () {
            var prevStep = this.getPrev();
            if (prevStep) {
                this.goToStep(prevStep);
            } else {
                this._updateButtons();
            }
        },

        toggleStep : function (step, visible) {
            this.$steps.find('[data-wizard-step="' + step + '"]').toggle(visible);
        },

        addStep : function (stepId, stepName, after) {
            var panelContext = getFakePanelContext(),
                $newStep = panelHelpers.tab.call({}, stepName, stepId, {
                    hash : {},
                    data : {
                        panelContext : panelContext
                    }
                }),
                $newPanel = panelHelpers.panel.call({}, stepId, {
                    hash : {},
                    fn : noContent,
                    data : {
                        panelContext : panelContext
                    }
                });

            if (after) {
                this.$steps.find('[data-wizard-step="' + after + '"]').after($newStep);
            } else {
                this.$steps.append($newStep);
            }

            this.$panels.append($newPanel);
        },

        deleteStep : function (stepId) {
            var $step = this.$steps.find('[data-wizard-step="' + stepId + '"]');
            if ($step.hasClass('active')) {
                if (this.getPrev()) {
                    this.goToStep(this.getPrev(), true);
                } else {
                    this.goToStep(this.getNext(), true);
                }
            }
            $step.remove();
            this.$panels.find('#' + stepId).remove();
        },

        cancel : function () {
            this.$element.trigger('wizard-cancel');
        },

        save : function () {
            this.$element.trigger('wizard-save');
        },

        reset : function () {
            this.$element.trigger('wizard-reset');
        },

        resize : function () {

        }
    };

    $.fn.extend({
        prim_wizard : function (option) {
            var applyArgs = Array.prototype.slice.call(arguments, 1);
            return this.each(function () {
                var $this = $(this),
                    data = $this.data('prim_wizard');

                if (!data) {
                    data = new Wizard(this);
                    $this.data('prim_wizard', data);
                }

                if (typeof option === 'string') {
                    data[option].apply(data, applyArgs);
                }
            });
        }
    });

    $.fn.prim_wizard.Constructor = Wizard;

    $(document).on('click.wizard.data-api', '.wizard-context [data-wizard]', function () {
        var $this = $(this),
            type = $this.attr('data-wizard');
        $this.closest('.wizard-context').prim_wizard(type);
    });
});

define('primutils/ui/widgets/droppable',[
    'jquery',
    'lodash',
    'primutils/utils/language',
    'primutils/utils/emulateTransitionEnd' // For $.fn.emulateTransitionEnd
], function ($, _, localizer) {
    'use strict';

    var DROP_TARGET_CLASS = 'x-drop-target';
    var DROP_TARGET_SELECTOR = '.' + DROP_TARGET_CLASS;

    var INFORMATION = 'x-drop-information';
    var INFORMATION_SELECTOR = '.' + INFORMATION;

    var DROP_LOADING_CLASS = 'x-drop-target-loading';
    var DROP_SUCCESS_CLASS = 'x-drop-target-success';
    var DROP_FAIL_CLASS = 'x-drop-target-fail';

    var HIDE_TIMEOUT = 2000;

    var DEFAULTS = {
        type : 'text/plain'
    };

    function getDefaultStrings() {
        return {
            drop : localizer.getString('droppable.dropTarget'),
            loading : localizer.getString('droppable.loading'),
            success : localizer.getString('droppable.success')
        };
    }

    function success($el, options) {
        $el.removeClass(DROP_LOADING_CLASS);
        $el.addClass(DROP_SUCCESS_CLASS).emulateTransitionEnd(200); //See x-drop-target transition
        $el.children(INFORMATION_SELECTOR).text(options.success);

        $el.on($.support.transition.end, function () {
            setTimeout(function () {
                $el.removeClass([DROP_TARGET_CLASS].concat(DROP_SUCCESS_CLASS).join(' '));
                $el.find(INFORMATION_SELECTOR).remove();
            }, HIDE_TIMEOUT);
        });
    }

    function fail($el, options) {
        $el.removeClass(DROP_LOADING_CLASS);
        $el.addClass(DROP_FAIL_CLASS).emulateTransitionEnd(200); //See x-drop-target transition
        $el.children(INFORMATION_SELECTOR).text(options.fail);

        $el.on($.support.transition.end, function () {
            setTimeout(function () {
                $el.removeClass([DROP_TARGET_CLASS].concat(DROP_FAIL_CLASS).join(' '));
                $el.find(INFORMATION_SELECTOR).remove();
            }, HIDE_TIMEOUT);
        });
    }

    function getData(dataTransfer, options) {
        var data;

        if (_.includes(dataTransfer.types, 'Files')) {
            data = dataTransfer.files;
        } else if (_.includes(dataTransfer.types, options.type)) {
            data = dataTransfer.getData(options.type);
        } else if (options.type === 'text/plain' && _.includes(dataTransfer.types, 'Text')) {
            data = dataTransfer.getData('Text');
        }

        return data;
    }

    /**
     * Droppable module that will allow a containing element to become a drop target.
     *
     * **Note**: also extends the jQuery object with .xdroppable();
     *
     * **Options**:
     *
     *       type: 'text/plain' //Type of data expected to be dropped
     *
     *       drop: localizer.getString('droppable.dropTarget') //Localized drop target information text
     *
     *       loading: localizer.getString('droppable.loading') //Localized loading information text in case the dropped function
     *       returns a deferred.
     *
     *       success: localizer.getString('droppable.success') //Localized success information text when the item has been
     *       dropped.
     *
     *       dropped: null //Function that is called when an item is dropped on the target.
     *
     *       valid: null //Function called when an item is dragged over the target to determine if it is a valid item to drop
     *
     *
     * @class ui.widgets.droppable
     * @constructor
     *
     * @param {jQuery} $el The element to make a drop taget
     * @param {Object} options The options for the droppable configuration
     *
     * @example
     *     var $myElement = $('<modal>'); //Pseudo modal element
     *
     *     droppable($myElement, {
     *         dropped: function (data) {
     *             //Do something with the data
     *         }
     *     });
     *
     * @example
     *     var $myElement = droppable($('<modal>'), {
     *         dropped: function (data) {
     *             //Do something with the data
     *         }
     *     })
     */
    var droppable = function ($el, options) {
        options = _.defaults({}, options, DEFAULTS, getDefaultStrings());

        $el = $el || $(this);

        $el.addClass('droppable');

        $el.on('dragenter', function (event) {
            if (event.currentTarget === $el[0]) {
                event.stopPropagation();
                event.stopImmediatePropagation();

                if (!$el.hasClass(DROP_TARGET_SELECTOR)) {
                    $(DROP_TARGET_SELECTOR).removeClass(DROP_TARGET_CLASS);
                    $(INFORMATION_SELECTOR).remove();
                    $el.addClass(DROP_TARGET_CLASS);

                    //TODO: put information into a template
                    $el.append('<div class="x-drop-information">' + options.drop + '</div>');
                }
            }
        })
        .on('dragleave', function (event) {
            if ($el[0].contains(event.target)) {
                event.stopPropagation();
                event.stopImmediatePropagation();
            }

            if ($el[0] === event.target) {
                event.stopPropagation();
                $el.removeClass(DROP_TARGET_CLASS);
                $el.find(INFORMATION_SELECTOR).remove();
            }
        })
        .on('dragover', function (event) {
            event.preventDefault(); //Stops all native browser drop handlers
        })
        .on('drop', function (event) {
            var dataTransfer = event.originalEvent.dataTransfer;
            var data = getData(dataTransfer, options);
            var valid = data != null && (!options.valid || options.valid(getData(event.originalEvent.dataTransfer, options)));
            var wait;

            event.stopPropagation();

            if (!valid) {
                event.preventDefault(); //Prevent navigating away from on a drop
                fail($el, options);
            } else {
                wait = !options.dropped || options.dropped(data);

                //Promise
                if (wait && _.isFunction(wait.promise)) {
                    event.preventDefault();

                    $el.addClass(DROP_LOADING_CLASS);
                    $el.children(INFORMATION_SELECTOR).text(options.loading);

                    $.when(wait)
                        .done(function () {
                            success($el, options);
                        });
                } else if (wait) {
                    event.preventDefault();
                    success($el, options);
                }
            }
        });

        return $el;
    };

    $.fn.xdroppable = _.partial(droppable, null);

    return droppable;
});



define('primutils/markup/primutils/ui/widgets/statusContainer.hbs', ['handlebars','primutils.markup'], function (H, u) {
var imports = u.buildImported(Array.prototype.slice.call(arguments, 2), 'primutils/ui/widgets/statusContainer.hbs'),
    func = H.template({"1":function(depth0,helpers,partials,data) {
  var stack1, helperMissing=helpers.helperMissing, buffer = "            ";
  stack1 = ((helpers.icon || (depth0 && depth0.icon) || helperMissing).call(depth0, "chevron-down", ((stack1 = (depth0 != null ? depth0.locale : depth0)) != null ? stack1.minimize : stack1), {"name":"icon","hash":{},"data":data}));
  if (stack1 != null) { buffer += stack1; }
  return buffer + "\n";
},"3":function(depth0,helpers,partials,data) {
  var stack1, helperMissing=helpers.helperMissing, buffer = "            ";
  stack1 = ((helpers.icon || (depth0 && depth0.icon) || helperMissing).call(depth0, "chevron-up", ((stack1 = (depth0 != null ? depth0.locale : depth0)) != null ? stack1.restore : stack1), {"name":"icon","hash":{},"data":data}));
  if (stack1 != null) { buffer += stack1; }
  return buffer + "\n";
},"5":function(depth0,helpers,partials,data) {
  var stack1, helperMissing=helpers.helperMissing, buffer = "            ";
  stack1 = ((helpers.icon || (depth0 && depth0.icon) || helperMissing).call(depth0, "dialog-close", ((stack1 = (depth0 != null ? depth0.locale : depth0)) != null ? stack1.hide : stack1), {"name":"icon","hash":{},"data":data}));
  if (stack1 != null) { buffer += stack1; }
  return buffer + "\n";
},"compiler":[6,">= 2.0.0-beta.1"],"main":function(depth0,helpers,partials,data) {
  var stack1, helperMissing=helpers.helperMissing, buffer = "<div class=\"job-container hide fade\">\n    <div class=\"job-controls\">\n";
  stack1 = ((helpers.link || (depth0 && depth0.link) || helperMissing).call(depth0, "#", ((stack1 = (depth0 != null ? depth0.locale : depth0)) != null ? stack1.minimize : stack1), {"name":"link","hash":{
    'class': ("job-minimize")
  },"fn":this.program(1, data),"inverse":this.noop,"data":data}));
  if (stack1 != null) { buffer += stack1; }
  stack1 = ((helpers.link || (depth0 && depth0.link) || helperMissing).call(depth0, "#", ((stack1 = (depth0 != null ? depth0.locale : depth0)) != null ? stack1.restore : stack1), {"name":"link","hash":{
    'class': ("job-restore")
  },"fn":this.program(3, data),"inverse":this.noop,"data":data}));
  if (stack1 != null) { buffer += stack1; }
  stack1 = ((helpers.link || (depth0 && depth0.link) || helperMissing).call(depth0, "#", ((stack1 = (depth0 != null ? depth0.locale : depth0)) != null ? stack1.hide : stack1), {"name":"link","hash":{
    'class': ("job-close")
  },"fn":this.program(5, data),"inverse":this.noop,"data":data}));
  if (stack1 != null) { buffer += stack1; }
  return buffer + "    </div>\n</div>";
},"useData":true});
return function (context, options) {
    options = u.extendOptions(options || {}, imports);
    return func(context, options).trim();
};
});


define('primutils/markup/primutils/ui/widgets/job.hbs', ['handlebars','primutils.markup'], function (H, u) {
var imports = u.buildImported(Array.prototype.slice.call(arguments, 2), 'primutils/ui/widgets/job.hbs'),
    func = H.template({"1":function(depth0,helpers,partials,data) {
  var stack1, helperMissing=helpers.helperMissing, buffer = "            ";
  stack1 = ((helpers.button || (depth0 && depth0.button) || helperMissing).call(depth0, "not-btn", {"name":"button","hash":{
    'class': ("job-cancel")
  },"fn":this.program(2, data),"inverse":this.noop,"data":data}));
  if (stack1 != null) { buffer += stack1; }
  return buffer + "\n";
},"2":function(depth0,helpers,partials,data) {
  var stack1, helperMissing=helpers.helperMissing;
  stack1 = ((helpers.icon || (depth0 && depth0.icon) || helperMissing).call(depth0, "delete", ((stack1 = (depth0 != null ? depth0.locale : depth0)) != null ? stack1.cancel : stack1), {"name":"icon","hash":{},"data":data}));
  if (stack1 != null) { return stack1; }
  else { return ''; }
  },"4":function(depth0,helpers,partials,data) {
  var stack1, helperMissing=helpers.helperMissing, buffer = "            ";
  stack1 = ((helpers.button || (depth0 && depth0.button) || helperMissing).call(depth0, "not-btn", {"name":"button","hash":{
    'class': ("job-retry")
  },"fn":this.program(5, data),"inverse":this.noop,"data":data}));
  if (stack1 != null) { buffer += stack1; }
  return buffer + "\n";
},"5":function(depth0,helpers,partials,data) {
  var stack1, helperMissing=helpers.helperMissing;
  stack1 = ((helpers.icon || (depth0 && depth0.icon) || helperMissing).call(depth0, "synchronize", ((stack1 = (depth0 != null ? depth0.locale : depth0)) != null ? stack1.retry : stack1), {"name":"icon","hash":{},"data":data}));
  if (stack1 != null) { return stack1; }
  else { return ''; }
  },"7":function(depth0,helpers,partials,data) {
  var stack1, helperMissing=helpers.helperMissing;
  stack1 = ((helpers.icon || (depth0 && depth0.icon) || helperMissing).call(depth0, "delete", ((stack1 = (depth0 != null ? depth0.locale : depth0)) != null ? stack1.remove : stack1), {"name":"icon","hash":{},"data":data}));
  if (stack1 != null) { return stack1; }
  else { return ''; }
  },"compiler":[6,">= 2.0.0-beta.1"],"main":function(depth0,helpers,partials,data) {
  var stack1, lambda=this.lambda, helperMissing=helpers.helperMissing, buffer = "<div class=\"job fade in ";
  stack1 = lambda((depth0 != null ? depth0.state : depth0), depth0);
  if (stack1 != null) { buffer += stack1; }
  buffer += " ";
  stack1 = lambda((depth0 != null ? depth0.aggregate : depth0), depth0);
  if (stack1 != null) { buffer += stack1; }
  buffer += "\" id=\"";
  stack1 = lambda((depth0 != null ? depth0.id : depth0), depth0);
  if (stack1 != null) { buffer += stack1; }
  buffer += "\" data-job-id=\"";
  stack1 = lambda((depth0 != null ? depth0.jobId : depth0), depth0);
  if (stack1 != null) { buffer += stack1; }
  buffer += "\" ";
  stack1 = ((helpers['?'] || (depth0 && depth0['?']) || helperMissing).call(depth0, (depth0 != null ? depth0.numericId : depth0), "data-numeric-id", "", {"name":"?","hash":{},"data":data}));
  if (stack1 != null) { buffer += stack1; }
  buffer += ">\n    <span class=\"message\">";
  stack1 = lambda((depth0 != null ? depth0.header : depth0), depth0);
  if (stack1 != null) { buffer += stack1; }
  buffer += "</span>\n    <div class=\"progress\">\n        <div class=\"percentage\">";
  stack1 = lambda((depth0 != null ? depth0.progress : depth0), depth0);
  if (stack1 != null) { buffer += stack1; }
  buffer += "</div>\n        <div class=\"bar ";
  stack1 = ((helpers['?'] || (depth0 && depth0['?']) || helperMissing).call(depth0, (depth0 != null ? depth0.hasAborted : depth0), "bar-error", "", {"name":"?","hash":{},"data":data}));
  if (stack1 != null) { buffer += stack1; }
  buffer += "\" style=\"width: ";
  stack1 = lambda((depth0 != null ? depth0.rawProgress : depth0), depth0);
  if (stack1 != null) { buffer += stack1; }
  buffer += "\">\n            <div class=\"percentage\">";
  stack1 = lambda((depth0 != null ? depth0.progress : depth0), depth0);
  if (stack1 != null) { buffer += stack1; }
  buffer += "</div>\n        </div>\n\n        <div class=\"running-job-toolbar hidden\" style=\"position: absolute; right: 0; color: #333;\">\n";
  stack1 = helpers['if'].call(depth0, (depth0 != null ? depth0.canCancel : depth0), {"name":"if","hash":{},"fn":this.program(1, data),"inverse":this.noop,"data":data});
  if (stack1 != null) { buffer += stack1; }
  buffer += "        </div>\n        <div class=\"aborted-job-toolbar hidden\" style=\"position: absolute; right: 0; color: #333;\">\n";
  stack1 = helpers['if'].call(depth0, (depth0 != null ? depth0.canRetry : depth0), {"name":"if","hash":{},"fn":this.program(4, data),"inverse":this.noop,"data":data});
  if (stack1 != null) { buffer += stack1; }
  buffer += "            ";
  stack1 = ((helpers.button || (depth0 && depth0.button) || helperMissing).call(depth0, "not-btn", {"name":"button","hash":{
    'class': ("job-remove")
  },"fn":this.program(7, data),"inverse":this.noop,"data":data}));
  if (stack1 != null) { buffer += stack1; }
  return buffer + "\n        </div>\n    </div>\n</div>";
},"useData":true});
return function (context, options) {
    options = u.extendOptions(options || {}, imports);
    return func(context, options).trim();
};
});

define('primutils/ui/widgets/status',[
    'jquery',
    'lodash',
    'primutils/utils/language',
    'primutils/utils/objectUtils',
    'primutils/markup/primutils/ui/widgets/statusContainer.hbs',
    'primutils/markup/primutils/ui/widgets/job.hbs',
    'primutils/templates/helpers/icon'
], function ($, _, lang, objUtils, containerTemplate, jobTemplate, icon) {
    'use strict';

    var DEFAULT_OPTIONS = {};
    var userDefaults = {};

    var STATES = {
        RUNNING : 'running',
        PAUSED : 'paused',
        ABORTED : 'aborted',
        CANCELED : 'canceled'
    };

    function getDefaultStrings() {
        return {
            aborted : lang.getString('status.aborted'),
            canceled : lang.getString('status.canceled'),
            retry : lang.getString('status.retry'),
            remove : lang.getString('status.remove'),
            cancel : lang.getString('status.cancel'),
            minimize : lang.getString('status.minimize'),
            restore : lang.getString('status.restore'),
            hide : lang.getString('status.hide'),
            notStarted : lang.getString('status.notStarted')
        };
    }

    var AGGREGATE_PROPERTIES = {
        id : 'job-aggregate',
        jobId : 'aggregate',
        header : '',
        state : STATES.RUNNING,
        aggregate : 'aggregate'
    };

    var getAggregate = function (jobs) {
        return _.reduce(jobs, function (memo, job) {
            return memo + (job.progress || 0);
        }, 0) / jobs.length;
    };

    var jobFinder = function ($element, id) {
        return $element.find('[data-job-id=' + id + ']');
    };

    var toggleVisibility = function ($element, numberOfJobs) {
        $element.toggleClass('in', numberOfJobs > 0);
        $element.toggleClass('hide', numberOfJobs === 0);
    };

    var printPercent = function (val, options) {
        val = val || 0;
        if (options.percentFormatter) {
            return options.percentFormatter.format(val);
        }
        return (val * 100) + '%';
    };

    var aggregator = function (components, $aggregate, progress, numberOfJobs, hasAborted) {
        var $element = components.$el;
        var options = components.options;
        if ($aggregate == null) {
            $aggregate = $(jobTemplate(_.assign({}, AGGREGATE_PROPERTIES, {
                rawProgress : (progress || 0) * 100 + '%',
                progress : printPercent(progress, options),
                hasAborted : hasAborted
            })));

            $element.append($aggregate);
        }

        $aggregate.find('.bar').css({
            width : ((progress || 0) * 100) + '%'
        });
        $aggregate.find('.percentage')
            .text('(' + numberOfJobs + ') ' + printPercent(progress, options));
    };

    /**
     * Long running job status notification tray
     *
     * @class ui.widgets.status
     * @constructor
     *
     * @param {jQuery} container The container to append status notifications
     * @param {Object} [options] The options used within status notifications.  Known keys:
     *   - `percentFormatter`: used to format the progress values displayed
     *   - `cancelJob`: callback when the user tries to manually stop a job. If omitted, the user will not be able to cancel jobs manually.
     *   - `retryJob`: callback when the user tries to restart an aborted or canceled job. If omitted, the user will not be able to retry jobs.
     *   - `locale`: used to change one of the several messages displayed by the widget.
     *   - - `locale.aborted`: displayed when the job is aborted
     *   - - `locale.canceled`: displayed when the job is canceled by the user
     *
     * @example
     *     var jobNotifier = status($('body'));
     *
     *     jobNotifier.addJob({
     *         id: '1',
     *         header: 'Uploading file',
     *         progress: 0, //Will default to 0 if not given
     *     });
     */
    var statusAPI = function ($container, userOptions) {
        var jobs = [];
        var options = objUtils.defaultsDeep({}, userOptions, userDefaults, { locale : getDefaultStrings() }, DEFAULT_OPTIONS);
        var $element = $(containerTemplate({ locale : options.locale }));
        var getJob = _.partial(jobFinder, $element);
        var getJobId = function ($element) {
            var $job = $element.closest('.job');
            return $job.is('[data-numeric-id]') ? parseInt($job.attr('data-job-id'), 10) : $job.attr('data-job-id');
        };
        var toggler = _.partial(toggleVisibility, $element);
        var renderAggregate = _.partial(aggregator, {
            $el : $element,
            options : options
        });
        var handlers = {
            closeJobHandler : function (event) {
                event.preventDefault();

                // native forEach here is fine
                /* eslint-disable lodash/prefer-lodash-chain */
                _.pluck(jobs, 'id').forEach(removeJob);
                /* eslint-enable lodash/prefer-lodash-chain */
            },
            minimizeHandler : function (event) {
                event.preventDefault();

                minimize();
            },
            restoreHandler : function (event) {
                event.preventDefault();

                restore();
            },
            cancelJobHandler : function () {
                var jobId = getJobId($(this)),
                    job = _.find(jobs, 'id', jobId);
                if (options.cancelJob && options.cancelJob(job) !== false) {
                    updateJob({ state : STATES.CANCELED }, jobId, { progress : 0 });
                }
            },
            retryJobHandler : function () {
                var jobId = getJobId($(this)),
                    job = _.find(jobs, 'id', jobId);
                if (options.retryJob && options.retryJob(job) !== false) {
                    updateJob({ state : STATES.RUNNING }, jobId, { progress : 0 });
                } else {
                    removeJob(jobId);
                }
            },
            removeJobHandler : function () {
                removeJob(getJobId($(this)));
            }
        };

        function jobAggregator() {
            if ($element.hasClass('minimize')) {
                renderAggregate(getJob('aggregate'), getAggregate(jobs), jobs.length);
            }
        }

        function minimize() {
            $element.addClass('minimize');

            renderAggregate(null, getAggregate(jobs), jobs.length, _.some(jobs, 'state', STATES.ABORTED));
        }

        function restore() {
            $element.removeClass('minimize');
            $element.removeClass('updated');

            getJob('aggregate').remove();
        }

        function flashUpdate() {
            if ($element.hasClass('minimize')) {
                $element.addClass('updated');
            }
        }

        function removeJob(id) {
            jobs = _.reject(jobs, 'id', id);

            var $job = getJob(id);

            $job.removeClass('in');

            if ($element.hasClass('minimize')) {
                $job.remove();
                toggler(jobs.length);
            } else {
                $job.one($.support.transition.end, function () {
                    $job.remove();

                    toggler(jobs.length);
                });
            }

            jobAggregator();
            flashUpdate();
        }

        function updateJob(state, id, newJobData) {
            var job = _.find(jobs, 'id', id);
            var $job = getJob(id);

            //Remove current state
            $job.removeClass(job.state);

            //Update job with new data as well as state
            job = _.merge(job, newJobData, state);

            //Re-render job properties
            var percentage,
                useHtml = false;
            switch (job.state) {
            case STATES.ABORTED:
                percentage = icon('indicator-critical', '', { hash : {} }) + '<span style="vertical-align: middle; line-height: 20px; margin-left: 4px;">' + options.locale.aborted + '</span>';
                useHtml = true;
                job.progress = 0;
                break;
            case STATES.CANCELED:
                percentage = options.locale.canceled;
                job.progress = 0;
                break;
            case STATES.PAUSED:
                if (job.progress === 0) {
                    percentage = options.locale.notStarted;
                    break;
                }
                    /* falls through */
            default:
                percentage = printPercent(job.progress, options);
                break;
            }
            $job.find('.message').text(job.header);
            $job.find('.percentage')[useHtml ? 'html' : 'text'](percentage);
            $job.find('.bar').css({
                width : (job.progress * 100) + '%'
            });
            $job.addClass(job.state);
            $job.find('.running-job-toolbar').toggleClass('hidden', job.state !== STATES.RUNNING);
            $job.find('.aborted-job-toolbar').toggleClass('hidden', !_.includes([STATES.ABORTED, STATES.CANCELED], job.state));

            //Auto-remove job when progress is 100% (complete)
            if (job.progress === 1) {
                removeJob(job.id);
            }

            jobAggregator();
            flashUpdate();
        }

        function addJob(jobData) {
            var job = _.defaults({}, jobData, {
                uid : _.uniqueId('job'),
                state : STATES.RUNNING
            });
            jobs.push(job);

            var $job = $(jobTemplate({
                id : job.uid,
                jobId : job.id,
                header : job.header,
                state : job.state,
                rawProgress : ((jobData.progress || 0) * 100) + '%',
                progress : (job.state === STATES.PAUSED && (jobData.progress || 0) === 0) ? options.locale.notStarted : printPercent(jobData.progress, options),
                numericId : _.isNumber(job.id),
                canCancel : _.has(options, 'cancelJob'),
                canRetry : _.has(options, 'retryJob'),
                locale : options.locale
            }));
            $job.find('.running-job-toolbar').toggleClass('hidden', job.state !== STATES.RUNNING);
            $job.find('.aborted-job-toolbar').toggleClass('hidden', !_.includes([STATES.ABORTED, STATES.CANCELED], job.state));
            $element.append($job);

            //Show or hide notifications on job length
            toggler(jobs.length);

            jobAggregator();
            flashUpdate();
        }

        //Initialize
        $container.append($element);

        $element
            .on('click', '.job-close', handlers.closeJobHandler)
            .on('click', '.job-minimize', handlers.minimizeHandler)
            .on('click', '.job-restore', handlers.restoreHandler)
            .on('click', '.job-cancel', handlers.cancelJobHandler)
            .on('click', '.job-retry', handlers.retryJobHandler)
            .on('click', '.job-remove', handlers.removeJobHandler);

        //Return the API for the Job Notification
        return Object.freeze({
            /**
             * Adds a job to the status notification tray.
             *
             * If this is the first job added it will also show the notification tray in the container specified during
             * construction. Otherwise it just appends the new job below the existing jobs.
             *
             * @method addJob
             * @param {Object} jobData The job to add to the status notification tray
             */
            addJob : addJob,
            /**
             * Updates an existing job with new data (message, progress).
             *
             * If a jobs progress reaches 1 (100%) it is auto-removed from the list of jobs.
             *
             * @method updateJob
             * @param {Number} id
             * @param {Object} jobData
             */
            updateJob : _.partial(updateJob, null),
            /**
             * Removes a job from the list of jobs.
             *
             * @method removeJob
             * @param {Number} id
             */
            removeJob : removeJob,
            /**
             * Updates a job's state to 'paused'
             *
             * @method pauseJob
             * @param {Number} id
             * @param {Object} [jobData] Additional job data to update the job with
             */
            pauseJob : _.partial(updateJob, {
                state : STATES.PAUSED
            }),
            /**
             * Updates a job's state to 'running'
             *
             * @method unpauseJob
             * @param {Number} id
             * @param {Object} [jobData] Additional job data to update the job with
             */
            unpauseJob : _.partial(updateJob, {
                state : STATES.RUNNING
            }),
            /**
             * Updates a job's state to 'aborted'
             *
             * @method abortJob
             * @param {Number} id
             * @param {Object} [jobData] Additional job data to update the job with
             */
            abortJob : _.partial(updateJob, {
                state : STATES.ABORTED
            }),
            /**
             * Updates a job's state to 'running'
             *
             * @method restartJob
             * @param {Number} id
             * @param {Object} [jobData] Additional job data to update the job with
             */
            restartJob : _.partial(updateJob, {
                state : STATES.RUNNING
            }),
            /**
             * Minimizes the notification tray to the bottom of the original container.
             *
             * @method minimize
             */
            minimize : minimize,
            /**
             * Restores the notification tray.
             *
             * @method restore
             */
            restore : restore
        });
    };

    /**
     * Specifies default options for all future status widgets.
     *
     * @method defaults
     * @static
     */
    statusAPI.defaults = function (opts) {
        _.assign(userDefaults, opts);
    };

    return statusAPI;
});

define('primutils/ui/widgets/widgets',[
    'primutils/ui/widgets/dock',
    'primutils/ui/widgets/modal',
    'primutils/ui/widgets/search',
    'primutils/ui/widgets/stickyHeader',
    'primutils/ui/widgets/tab',
    'primutils/ui/widgets/wizard',
    'primutils/ui/widgets/droppable',
    'primutils/ui/widgets/status'
], function (dock, modal, search, stickyHeader, tab, wizard, droppable, status) {
    'use strict';

    return {
        dock : dock,
        modal : modal,
        search : search,
        stickyHeader : stickyHeader,
        tab : tab,
        wizard : wizard,
        droppable : droppable,
        status : status
    };
});


/**
 * A utility for calculating differences between stard and end dates, accounting for calendar logic like work vs. non-work time, exceptions, and holidays.
 *
 * Primarily intended for internal usage by the gantt - most options mentioned in the Properties tab, below, are focused on how this util/the gantt will access the various data needed for rules to function.
 *
 * `'new'` is not required for instantiation of this utility.
 * The object returned after instantation is itself a function, documented as the `calculate(obj)` method.
 *
 * Example:
 *
 *     var calculator = calendarCalculator({
 *         calendar: {...}, // fuller calendar example below
 *
 *         workweek: _.property('calendar.workweek'),
 *         exceptions: _.property('calendar.exceptions'),
 *         holidays: _.property('calendar.holidays'),
 *         holidayDate: function (options, record) {
 *             return record.holiday;
 *         },
 *         exceptionDate: function (options, record) {
 *             return record.exception;
 *         },
 *         exceptionBlocks: function (options, record) {
 *             return record.workTime;
 *         }
 *
 *         // ...other options...
 *     });
 *     calculator({
 *         start: ...,
 *         end: ...,
 *         record: ...
 *     });
 *     // return val looks like:
 *       { start: ..., end: ..., duration: ... }
 *
 *     calculator.calculate({
 *         start: ...,
 *         duration: ...,
 *         record: ...
 *     });
 *     // return val looks like:
 *       { start: ..., end: ..., duration: ... }
 *
 * The calculator can accept some subset of the `start`/`end`/`duration` properties on the provided object - when in such a state, the calculator will perform a best-effort to make sense of the provided objects, and return a normalized response.
 *
 * An example calendar, usable with the default options, could look something like this:
 *
 *     {
 *       workweek: {
 *         0: [],
 *         1: [{start: '08:00', end: '16:00'}],
 *         2: [{start: '08:00', end: '12:00'}, {start: '13:00', end: '17:00'}],
 *         3, ...,
 *         4: ...,
 *         5: ...,
 *         6: ...
 *       },
 *       exceptions: [{
 *         exception: '2015-03-16T00:00:00',
 *         workTime: [{start: '10:30', end: '13:30'}]
 *       }],
 *       holidays: [{
 *         holiday: '2015-07-03T00:00:00'
 *       }]
 *     }
 *
 * @class calendar.calculator
 * @module calendar
 * @param options {Object} See properties tab
 * @constructor
 */

define('primutils/calendars/calculator',[
    'lodash',
    'logger',
    'moment',
    'primutils/utils/validator'
], function (_, logger, moment, validatorFactory) {
    'use strict';

    var DEFAULT_OPTIONS = {
        workweek : _.property('calendar.workweek'),
        exceptions : _.property('calendar.exceptions'),
        holidays : _.property('calendar.holidays'),

        holidayRepeatDayOfMonth : _.property('dayOfMonth'),
        holidayRepeatDayOfWeek : _.property('dayOfWeek'),
        holidayRepeatMonth : _.property('month'),
        holidayRepeatWeek : _.property('week'),

        holidayDate : function (options, record) {
            return record.holiday;
        },
        exceptionDate : function (options, record) {
            return record.exception;
        },
        exceptionBlocks : function (options, record) {
            return record.workTime;
        }
    };

    var calcWeekOfMonth = function calcWeekOfMonth(mom) {
        return Math.floor((mom.date() - 1) / 7) + 1;
    };

    var blockToHour = _.memoize(function blockToHour(str) {
        // '08:01' -> 8
        return parseInt(str, 10);
    });
    var blockToMinute = _.memoize(function blockToMinute(str) {
        // '08:05' -> 5
        return parseInt(str.split(':')[1], 10);
    });
    var durationInHours = function durationInHours(dur) {
        if (moment.isDuration(dur)) {
            return dur.as('hours');
        }
        return dur;
    };
    var positiveNumber = function positiveNumber(num) {
        return _.isNumber(num) && num > 0;
    };
    var positiveDuration = function positiveDuration(dur) {
        return moment.isDuration(dur) && dur.as('ms') > 0;
    };

    var convertBlockToMoment = function convertBlockToMoment(walkerDay, options, block) {
        var result = {};

        
        result.start = walkerDay.clone().hours(blockToHour(block.start)).minutes(blockToMinute(block.start));

        if (block.end === '00:00') {
            result.end = walkerDay.clone().add(1, 'days');
        } else {
            var minutesEnd = blockToMinute(block.end);
            if (options.roundBlockEnd) {
                var offByMinutes = minutesEnd % 5;
                if (offByMinutes) {
                    minutesEnd += (5 - offByMinutes);
                }
            }

            result.end = walkerDay.clone().hours(blockToHour(block.end)).minutes(minutesEnd);
        }

        return result;
    };
    var mapHolidays = function mapHolidays(options, holidays) {
        return _.map(holidays, function (holiday) {
            if (options.holidayRepeatMonth(holiday) != null) {
                return holiday;
            }
            return options.unserializeHoliday(options.holidayDate(options, holiday));
        });
    };
    var mapExceptions = function mapExceptions(options, exceptions) {
        return _.map(exceptions, function (exception) {
            return {
                date : options.unserializeException(options.exceptionDate(options, exception)),
                blocks : options.exceptionBlocks(options, exception)
            };
        });
    };
    var withStartOfDay = function withStartOfDay(mom) {
                return mom.clone().startOf('day');
    };
    var getNextWorkBlock = function getNextWorkBlock(walker, walkerDay, prevBlock, options, blocks, holidays, exceptions) {
        var block;

        for (var excI = 0, excLen = exceptions.length; excI < excLen; excI++) {
            var exception = exceptions[excI];
            var exceptionDay = exception.date; // normalized by mapExceptions

            // if we're already past the exception, will never matter - splice it out (which is OK because we _.mapped first)
            if (walkerDay.isAfter(exceptionDay, 'day')) {
                exceptions.splice(0, 1);
                excI--;
                excLen--;
                continue;
            }
            if (!walkerDay.isSame(exceptionDay, 'day')) {
                continue;
            }

            var exceptionBlocks = exception.blocks; // normalized by mapExceptions
            for (var blockI = 0, blockLen = exceptionBlocks.length; blockI < blockLen; blockI++) {
                block = convertBlockToMoment(walkerDay, options, exceptionBlocks[blockI]);
                if (block.start >= walker) {
                    return block;
                } else if (walker < block.end) {
                    return {
                        start : walker.clone(),
                        end : block.end // already cloned by convertBlockToMoment
                    };
                }
            }

            // If there's an exception on this day, all of its ON/OFF time replaces the normal blocks
            return null;
        }

        var isHoliday = _.some(holidays, function isHoliday(holiday) {
            if (moment.isMoment(holiday)) {
                return holiday.isSame(walkerDay, 'day');
            }

            var repeatMonth = options.holidayRepeatMonth(holiday);
            if ((walkerDay.month() + 1) !== repeatMonth) {
                return;
            }
            var repeatDOM = options.holidayRepeatDayOfMonth(holiday);
            if (repeatDOM != null) {
                return walkerDay.date() === repeatDOM;
            }

            var repeatWeek = options.holidayRepeatWeek(holiday);
            var repeatDayOfWeek = options.holidayRepeatDayOfWeek(holiday);
            if (repeatWeek != null && repeatDayOfWeek != null) {
                return calcWeekOfMonth(walkerDay) === repeatWeek &&
                    walkerDay.day() === repeatDayOfWeek;
            }
        });
        if (isHoliday) {
            return null;
        }

        var i = 0;
        var lastBlockIdx = blocks.indexOf(prevBlock);
        if (lastBlockIdx !== -1) {
            i = lastBlockIdx + 1;
        }
        for (var len = blocks.length; i < len; ++i) {
            var rawBlock = blocks[i];
            block = convertBlockToMoment(walkerDay, options, rawBlock);
            if (block.start >= walker) {
                return {
                    block : block,
                    rawBlock : rawBlock
                };
            } else if (walker < block.end) {
                return {
                    block : {
                        start : walker.clone(),
                        end : block.end // already cloned by convertBlockToMoment
                    },
                    rawBlock : rawBlock
                };
            }
        }
        return null;
    };
    var MILLISECONDS_IN_HOUR = (1000 * 60 * 60);
    /* Given start/end moments, returns diff in hours */
    var calculateWorkTime = function calculateWorkTime(start, end) {
        // TODO need to round to nearest?
        return (end.valueOf() - start.valueOf()) / MILLISECONDS_IN_HOUR;
        // return end.diff(start, 'hours', true);
    };
    /* Given timeblock as moments, returns diff in hours */
    var calculateWorkTimeFromBlock = function calculateWorkTimeFromBlock(block) {
        return calculateWorkTime(block.start, block.end);
    };

        var __objValidator = validatorFactory({
        name : 'Calculator Object',
        validOptions : {
            start : 'Moment',
            end : 'Moment',
            duration : [positiveNumber, positiveDuration]
        }
    });
    var __validator = validatorFactory({
        name : 'CalculatorFactory',
        validOptions : {
            /**
             * Un-validated set of options, assumed to hold calendar information with the default `workweek`, `exceptions`, and `holidays` options.
             *
             * @property calendar
             * @type {Object}
             * @optional
             */
            calendar : 'StrictObject',
            /**
             * Un-validated set of options that can be used in place of the more specifically named `calendar` option.
             *
             * @property meta
             * @type {Object}
             * @optional
             */
            meta : 'StrictObject',

            /**
             * Accessor function for the relevant calendar workweek.
             * Will be invoked as `workweek(options, record)`, where options are the options passed to the calculator and the record is assume to be passed in by the grid.
             *
             * @property workweek
             */
            workweek : 'Function',
            /**
             * Accessor function for the relevant calendar exceptions.
             * Will be invoked as `exceptions(options, record)`, where options are the options passed to the calculator and the record is assume to be passed in by the grid.
             *
             * @property exceptions
             */
            exceptions : 'Function',
            /**
             * Accessor function for the relevant calendar holidays.
             * Will be invoked as `holidays(options, record)`, where options are the options passed to the calculator and the record is assume to be passed in by the grid.
             *
             * @property holidays
             */
            holidays : 'Function',

            /**
             * Accessor function for the holiday date field for each holiday.
             * See also `holidays` option.
             *
             * @property holidayDate
             */
            holidayDate : 'Function',

            /**
             * Responsible for unserializing holiday date fields from the server to a moment object.
             *
             * This should be a date formatter's `unserialize` function.
             * See also `holidays` and `holidayDate` options.
             *
             * @property unserializeHoliday
             */
            unserializeHoliday : 'Function',

            /**
             * Accessor function for a given calendar exception's date field.
             * See also `exceptions` option.
             *
             * @property exceptionDate
             */
            exceptionDate : 'Function',

            /**
             * Accessor function for a given calendar exception's work time blocks.
             * See also `exceptions`, `exceptionDate`, and `unserializeException` options.
             *
             * @property exceptionBlocks
             */
            exceptionBlocks : 'Function',

            /**
             * Responsible for unserializing exception date fields from the server to a moment object.
             *
             * This should be a date formatter's `unserialize` function.
             * See also `exceptions`, `exceptionDate`, and `exceptionBlocks` options.
             *
             * @property unserializeException
             */
            unserializeException : 'Function',

            /**
             * If set, this option will round up end dates given to the calculator.
             *
             * For example, if you give the calculator a block of `{start: '10:30', end: '11:30'}`, with or without this option, that's considered 60 minutes of work.
             *
             * With this option set,  `{start: '10:30', end: '11:29'}` would give the same result - the default, false, would treat this as 59 minutes of work.
             *
             * @property roundBlockEnd
             * @type {Boolean}
             * @default false
             */
            roundBlockEnd : 'Boolean',

            holidayRepeatDayOfMonth : 'Function',
            holidayRepeatDayOfWeek : 'Function',
            holidayRepeatMonth : 'Function',
            holidayRepeatWeek : 'Function'
        },
        ignoredSubkeys : [
            'calendar', 'meta'
        ],
        required : [
            'unserializeHoliday', 'unserializeException'
        ]
    });
    
    function calcFactory(options) {
        var _options = _.defaults({}, options, DEFAULT_OPTIONS);

                __validator(_options);
        
        var _workweek = _.partial(_options.workweek, _options),
            _exceptions = _.partial(_options.exceptions, _options),
            _holidays = _.partial(_options.holidays, _options);

        function walkFromStartWithDuration(obj) {
            var workweek = _workweek(obj);
            var remainingDuration = durationInHours(obj.duration);
            var walker = obj.start.clone();
            var walkerDay = withStartOfDay(walker);
            var start = null, end = null;
            var holidays = mapHolidays(_options, _holidays(obj));
            var exceptions = mapExceptions(_options, _exceptions(obj));
            var prevRawBlock;

            while (remainingDuration > 0) {
                var workBlocks = workweek[walker.day()];

                var nextBlock = getNextWorkBlock(walker, walkerDay, prevRawBlock, _options, workBlocks, holidays, exceptions);
                if (nextBlock == null) {
                    walker.add(1, 'day').startOf('day');
                    walkerDay.add(1, 'day');
                    prevRawBlock = null;
                    continue;
                }
                if (nextBlock.rawBlock) {
                    prevRawBlock = nextBlock.rawBlock;
                    nextBlock = nextBlock.block;
                }

                var workTime = calculateWorkTimeFromBlock(nextBlock);
                if (start == null) {
                    start = nextBlock.start.clone();
                }
                if (workTime >= remainingDuration) {
                    end = nextBlock.start.clone().add(remainingDuration, 'hours');
                    break;
                }
                remainingDuration -= workTime;

                var nextWalker = nextBlock.end.clone();
                if (walker.day() !== nextWalker.day()) {
                    walkerDay.add(1, 'day');
                }
                walker = nextWalker;
            }

            return {
                start : start || obj.start.clone(),
                duration : durationInHours(obj.duration),
                end : end
            };
        }

        function walkFromStartToEnd(obj) {
            var workweek = _workweek(obj);
            var totalDuration = 0;
            var walker = obj.start.clone();
            var walkerDay = withStartOfDay(walker);
            var holidays = mapHolidays(_options, _holidays(obj));
            var exceptions = mapExceptions(_options, _exceptions(obj));
            var objEndValue = obj.end.valueOf();
            var prevRawBlock;

            while (walker.valueOf() < objEndValue) {
                var workBlocks = workweek[walker.day()];

                var nextBlock = getNextWorkBlock(walker, walkerDay, prevRawBlock, _options, workBlocks, holidays, exceptions);
                if (nextBlock == null) {
                    walker.add(1, 'day').startOf('day');
                    walkerDay.add(1, 'day');
                    prevRawBlock = null;
                    continue;
                }
                if (nextBlock.rawBlock) {
                    prevRawBlock = nextBlock.rawBlock;
                    nextBlock = nextBlock.block;
                }

                if (nextBlock.end >= obj.end) {
                    totalDuration += calculateWorkTime(nextBlock.start, obj.end);
                    break;
                }

                totalDuration += calculateWorkTimeFromBlock(nextBlock);
                var nextWalker = nextBlock.end.clone();
                if (walker.day() !== nextWalker.day()) {
                    walkerDay.add(1, 'day');
                }
                walker = nextWalker;
            }

            return {
                // TODO start and end are not necessarily correct
                start : obj.start,
                end : obj.end,
                duration : totalDuration
            };
        }

        function that(obj) {
                        __objValidator(obj);
                        //   i.e. choke immediately if duration is non-numeric or negative, dates aren't actually moments, and so on

            if (obj.start != null && obj.duration != null) {
                return walkFromStartWithDuration(obj);
            } else if (obj.start != null && obj.end != null) {
                return walkFromStartToEnd(obj);
            } else {
                var err = 'calendarCalculator | invalid set of arguments provided';
                logger.fatal(err);
                throw new Error(err);
            }
        }

        /**
         * Alias for the calculation fundamental calculation method described in the constructor / class description.
         *
         * That is - give this method an object like `{start, end}` and it will return an object with `{start, end, duration}`.
         * Or give an object like `{start, duration}` and it will return an object with `{start, end, duration}`.
         *
         * @method calculate
         * @param {Object}
         * @return {Object}
         */
        that.calculate = that;

        /**
         * Returns a new calculator, with the given options shallow-merged with the options of the current calculator.
         *
         * @method withOptions
         * @param {Object} options
         */
        that.withOptions = function withOptions(newOptions) {
            return calcFactory(_.defaults({}, newOptions, _options));
        };

        return that;
    }

    
    return calcFactory;
});

/**
 * Primary export of primutils.  TODO document me.
 *
 * @class primutils
 * @module primutils
 * @private
 */
define('primutils', [
    'primutils/i18n/Localizer',
    'primutils/formatters/numberFormatter',
    'primutils/formatters/dateFormatter',
    'primutils/formatters/durationFormatter',
    'primutils/validators/validatorFactory',
    'primutils/utils/language',
    'primutils/a11y/accessibilityUtils',
    'primutils/templates/templateUtilities',
    'primutils/ui/ChangeBar',
    'primutils/ui/dialogs/AJAXErrorHandler',
    'primutils/formatters/wrappers',
    'primutils/formatters/formats',
    'primutils/utils/isEqual',
    'primutils/utils/objectUtils',
    'primutils/utils/dates',
    'primutils/utils/escaping',
    'primutils/ui/widgets/widgets',

    'primutils/calendars/calculator'

    // Items that should be available in the compiled file, but are not available as a property of primutils
    // should be listed in Gruntfile.js
], function (Localizer, numberFormatter, dateFormatter, durationFormatter, validatorFactory, language, accessibilityUtils, templateUtilities, ChangeBar, AJAXErrorHandler, formatterWrappers, formatterFormats, isEqual, clone, dates, escaping, widgets) {
    'use strict';

    return {

        version : '5.3.0',

        lang : function (stringMap) {
            language.setStringMap(stringMap);
        },

        numberFormatter : numberFormatter,
        dateFormatter : dateFormatter,
        durationFormatter: durationFormatter,
        validatorFactory : validatorFactory,

        formats: formatterFormats,
        wrappers: formatterWrappers,

        accessibilityUtils : accessibilityUtils,

        registerGlobalHandlebarsImports : templateUtilities.registerImports.bind(templateUtilities),

        protectGlobalHandlebarsImports : templateUtilities.protectHandlebarsImports.bind(templateUtilities),

        Localizer : Localizer,
        setLocalizerForLocalizerHelper : templateUtilities.setLocalizerForLocalizerHelper.bind(templateUtilities),
        setLocalizerForAJAXErrorHandler : AJAXErrorHandler.useLocalizer.bind(AJAXErrorHandler),
        setAJAXErrorHandlerDefaults: AJAXErrorHandler.defaults.bind(AJAXErrorHandler),

        ChangeBar: ChangeBar,

        extendLodashMethods : function () {
            isEqual.mixinToLodash();
            clone.mixinToLodash();
        },

        enableFakeUTCParsing : function () {
            dateFormatter.enableFakeUTCParsing();
            dates.patchMomentWarnings();
            dates.patchMoment();
        },

        enableXSSPatches : function () {
            escaping.patchAll();
        },

        widgets: widgets
    };
});

define('primutils/a11y/bootstrap/button',['jquery', 'bootstrap'], function ($) {
    'use strict';

    var ButtonPrototype = $.fn.button.Constructor.prototype,
        origButtonToggle = ButtonPrototype.toggle;

    ButtonPrototype.toggle = function () {
        if (!this.$element.prop('disabled')) {
            origButtonToggle.apply(this, arguments);
            var $parent = this.$element.closest('[data-toggle="buttons-radio"]');
            if ($parent) {
                $parent.find('[aria-pressed="true"]').attr('aria-pressed', false);
            }
            this.$element.attr('aria-pressed', this.$element.hasClass('active'));
        }
    };
});


define('primutils/a11y/bootstrap/collapse',['jquery',
    'primutils/utils/keyCode',

    'bootstrap'
], function ($, keyCode) {
    'use strict';

    /**
     * Enhancements to Bootstrap's collapse plugin that make it more accessible and usable.
     *
     * You probably do not want to use this directly, use <a href="a11y.bootstrap.accessibilityFixes.html">a11y.bootstrap.accessibilityFixes</a>.
     *
     * @class a11y.bootstrap.collapse
     * @module a11y.bootstrap
     * @private
     */

    var CollapsePrototype = $.fn.collapse.Constructor.prototype,
        origCollapseTransition = CollapsePrototype.transition;

    CollapsePrototype.transition = function (method) {
        origCollapseTransition.apply(this, arguments);

        // Will be set by the original method if preventDefault not called
        if (!this.transitioning) {
            return;
        }

        this.$element.emulateTransitionEnd(350);

        var $accordionGroup = this.$element.closest('.accordion-group');
        var isFoldedAccordion = $accordionGroup.parent().hasClass('folded-accordion');

        $accordionGroup[method]('open');

        //add aria-expanded true if you are adding open class.
        $accordionGroup.find('.accordion-toggle').not($accordionGroup.find('.accordion-group .accordion-toggle'))
            .attr('aria-expanded', method === 'addClass')
            .attr('aria-selected', method === 'addClass');

        $accordionGroup.find('.accordion-inner').not($accordionGroup.find('.accordion-group .accordion-inner'))
            .attr('tabindex', method === 'addClass' ? null : '-1')
            .attr('aria-hidden', method !== 'addClass' && !isFoldedAccordion);
    };

    $(document).on('focusin.collapse.data-api, click.collapse.data-api', '.accordion', function (e) {
        var $accordion = $(e.currentTarget),
            focusedOnce = $accordion.data('focusedOnce'),
            $group = $(e.target).closest('.accordion-group'),
            $selected = $group;

        if (!$group.length) {
            $group = $accordion.find('.accordion-group').first();
        }

        $accordion.data('focusedOnce', true).find('.accordion-toggle')
            .attr('tabindex', '-1') // prevent tabbing, use arrow keys instead
            .attr('aria-selected', false);

        $group.find('.accordion-toggle')
            .attr('tabindex', '0')
            .attr('aria-selected', true);

        if (!focusedOnce && !$selected.closest('.accordion-group').is($group)) {
            $group.find('.accordion-toggle').focus();
        }
    });

    $(document)
        .on('focusout.collapse.data-api', '.accordion', function (ev) {
            $(this).find('.accordion-group-focus').removeClass('accordion-group-focus');
        })
        .on('focusin.collapse.data-api', '.accordion-toggle', function (ev) {
            $(this).closest('.accordion-group').addClass('accordion-group-focus');
        });

    $(document).on('keydown.collapse.data-api', '.accordion-group', function (e) {
        var $target = $(e.target),
            $group = $(e.currentTarget),
            $newGroup;

        if (!$target.closest('.accordion-group').is($group)) { return; }

        if ($target.hasClass('accordion-toggle')) {
            if (e.keyCode === keyCode.UP && e.ctrlKey) {
                $newGroup = $group.parent().closest('.accordion-group');
            } else if (e.keyCode === keyCode.UP || e.keyCode === keyCode.LEFT) {
                $newGroup = $group.prevAll('.accordion-group').first();
            } else if (e.keyCode === keyCode.DOWN || e.keyCode === keyCode.RIGHT) {
                $newGroup = $group.nextAll('.accordion-group').first();
            } else if (e.keyCode === keyCode.HOME) {
                $newGroup = $group.prevAll('.accordion-group').last();
            } else if (e.keyCode === keyCode.END) {
                $newGroup = $group.nextAll('.accordion-group').last();
            } else if (e.keyCode === keyCode.SPACE) {
                $target.click();
                e.preventDefault();
            }

            if ($newGroup) {
                e.preventDefault();
                $newGroup.find('.accordion-toggle').not($newGroup.find('.accordion-group .accordion-toggle')).focus();
            }
        } else if (e.keyCode === keyCode.UP && e.ctrlKey) {
            $group.find('.accordion-toggle').focus();
            e.stopPropagation();
        }
    });

    // Only do this on currently hidden groups so that clicking on a button/link/textbox doesn't recollapse the group
    $(document).on('click.collapse.data-api', '.accordion-body.fold-animated.hidden', function (e) {
        if (!$(e.target).is('[data-toggle="collapse"]')) {
            var $this = $(this);
            $this.find('.accordion-toggle[data-target="#' + $this.attr('id') + '"]').click();
        }
    });

    $(document).on('show.collapse.data-api hidden.collapse.data-api', '.accordion-body', function (e) {
        if (e.target === this) {
            $(this).toggleClass('hidden', e.type !== 'show');
        }
    });
});

define('primutils/a11y/focusManager',['jquery', 'lodash', 'logger', 'primutils/a11y/accessibilityUtils'], function ($, _, logger, utils) {
    'use strict';

    var openModals = [],
        openModalSelections = {},
        defaultFocusFn = function ($el, $newFocus, event) {
            var relatedTarget = event.relatedTarget || STATE.relatedTarget,
                $newSelection = $el.first(),
                $focusables = utils.findFocusable($el),
                isFollowing, isPreceding;

            if (relatedTarget) {
                var $relatedTarget = $(relatedTarget);
                isFollowing = $relatedTarget.is($focusables.first());
                isPreceding = !isFollowing;
            }

            if (isPreceding) {
                $newSelection = $focusables.first();
            } else if (isFollowing) {
                $newSelection = $focusables.last();
            }

            // Doesn't actually do anything despite document.activeElement saying it should have
            //if ($newSelection.is('iframe')) {
            //    var $iframeDocument = $newSelection.contents(),
            //        iframeDocument = $iframeDocument.get(0),
            //        $iframeThings = utils.findFocusable($iframeDocument);
            //    if (isFollowing) {
            //        $iframeThings.last().focus();
            //    } else if (isPreceding) {
            //        $iframeThings.first().focus();
            //    }
            //
            //    // Added because bootstrap modal was not highlighting textfield input when tabbed to.
            //    // The textfield was the focus however its input was not selected.
            //    if (iframeDocument.activeElement.type === 'text' && !iframeDocument.activeElement.readOnly) {
            //        iframeDocument.activeElement.select();
            //    }
            //}

            $newSelection.focus();

            // Added because bootstrap modal was not highlighting textfield input when tabbed to.
            // The textfield was the focus however its input was not selected.
            if (document.activeElement.type === 'text' && !document.activeElement.readOnly) {
                document.activeElement.select();
            }

            return $newSelection;
        };

    var STATE = {
        relatedTarget : null
    };

    $(document).on('focusout.modal-focus', function (e) {
        STATE.relatedTarget = e.target;
    });

    var setupFocus = function ($element) {
        $(document).off('focusin.modal-focus').on('focusin.modal-focus', function (e) {
            if (!$element.is(e.target) && !$element.has(e.target).length && $element[0]._findFocusFn) {
                var $target = $(e.target);
                $element.get(0)._findFocusFn.call(undefined, $element, $target, e);
            }
            STATE.relatedTarget = null;
        });
    };

    return {
        pushFocus : function ($element, findFocusFn) {
            var focusId = [];
            $element.each(function () {
                if (this._findFocusId) {
                    logger.error('Given element was already marked for focus tracking');
                    return;
                }
                this._findFocusFn = findFocusFn || defaultFocusFn;
                this._findFocusId = _.uniqueId('find-focus-');
                focusId.push(this._findFocusId);
            });
            focusId = focusId.join(' ');
            openModals.unshift(focusId);
            openModalSelections[focusId] = $element;
            setupFocus($element);
        },
        popFocus : function ($element) {
            var focusId = [];
            $element.each(function () {
                focusId.push(this._findFocusId);
            });
            focusId = focusId.join(' ');
            if (!_.includes(openModals, focusId)) {
                logger.error('Given element was not registered for focus tracking correctly');
            }

            openModals = _.without(openModals, focusId);
            delete openModalSelections[focusId];

            $element.each(function () {
                delete this._findFocusFn;
                delete this._findFocusId;
            });

            if (openModals.length) {
                setupFocus(openModalSelections[openModals[0]]);
            } else {
                $(document).off('focusin.modal-focus');
            }
        }
    };
});

define('primutils/a11y/bootstrap/distractionFree',[
    'jquery',
    'primutils/a11y/accessibilityUtils',
    'primutils/a11y/focusManager'
], function ($, utils, focusManager) {
    'use strict';

    var $currentDistractionFreeElement,
        $currentNavFreeElement;

    $(document)
        .on('focusin.distraction-free', function (ev) {
            var $html = $('html');
            var $element,
                $breadcrumb = $('.bsky-nav-page-action-area');
            if ($html.hasClass('distraction-free-mode')) {
                $element = $('.distraction-free-element');
                if ($currentDistractionFreeElement == null || $element[0] !== $currentDistractionFreeElement[0]) {
                    $currentDistractionFreeElement = $element;
                    focusManager.pushFocus($element.add($breadcrumb));
                }
            } else if ($html.hasClass('nav-free-mode')) {
                $element = $('.nav-free-element');
                if ($currentNavFreeElement == null || $element[0] !== $currentNavFreeElement[0]) {
                    $currentNavFreeElement = $element;
                    focusManager.pushFocus($element.add($breadcrumb));
                }
            } else {
                if ($currentDistractionFreeElement != null) {
                    focusManager.popFocus($currentDistractionFreeElement.add($breadcrumb));
                    $currentDistractionFreeElement = null;
                }
                if ($currentNavFreeElement != null) {
                    focusManager.popFocus($currentNavFreeElement.add($breadcrumb));
                    $currentNavFreeElement = null;
                }
            }
        });
});


define('primutils/a11y/bootstrap/dropdown',['jquery', 'lodash',
    'primutils/a11y/collapsibleDropdown',
    'primutils/utils/keyCode',
    'bootstrap'
], function ($, _, collapsibleDropdown, keyCode) {
    'use strict';

    /**
     * Enhancements to Bootstrap's dropdown plugin that make it more accessible and usable.
     *
     * You probably do not want to use this directly, use <a href="a11y.bootstrap.accessibilityFixes.html">a11y.bootstrap.accessibilityFixes</a>.
     *
     * @class a11y.bootstrap.dropdown
     * @module a11y.bootstrap
     * @private
     */

    // Copied from Bootstrap (but made to fit our style)
    function getParent($this) {
        var selector = $this.attr('data-target'),
            $parent;

        if (!selector) {
            selector = $this.attr('href');
            selector = selector && /#/.test(selector) && selector.replace(/.*(?=#[^\s]*$)/, ''); //strip for ie7
        }

        $parent = selector && $(selector);

        if (!$parent || !$parent.length) {
            $parent = $this.parent();
        }

        return $parent;
    }

    function shouldDropUp($popup, $parent) {
        var popupHeight = $popup.outerHeight(true),
            parentHeight = $parent.outerHeight(),
            offset = $parent.offset();

        return offset.top + popupHeight + parentHeight > window.pageYOffset + $(window).height();
    }

    function clearMenus() {
        $('[data-toggle=dropdown]').each(function () {
            var $this = $(this);
            var $parent = getParent($this);
            var $popup = $parent.children('.dropdown-menu');
            $this.attr('aria-expanded', false);
            $popup.attr('aria-hidden', true);
        });
    }

    // Needs to be attached _after_ bootstrap
    $(document)
        .on('click.dropdown.data-api', '[data-toggle=dropdown]', function () {
            var $this = $(this),
                $parent = getParent($this),
                $popup = $parent.children('.dropdown-menu'),
                $body = $('body'),
                $menu = $parent.children('.dropdown-menu'),
                isActive = $parent.hasClass('open'),
                $document = $(document);

            $this.attr('aria-expanded', isActive);
            $popup.attr('aria-hidden', !isActive);

            if (isActive) {
                $body.css('overflow', 'hidden'); // Prevent scrollbars while calculating need to dropup
                $parent.toggleClass('dropup', shouldDropUp($popup, $parent));
                $body.css('overflow', '');
            }

            if ($menu.hasClass('pull-center')) {
                var currentMargin = parseFloat($menu.css('margin-right')) || 0;
                $menu.css('margin-right', isActive ? (($menu.outerWidth(true) - currentMargin) / -2) : '');
            }

            if ($menu.hasClass('needs-overflow') && isActive) {
                $menu.css({ position : '', left : '', top : '', bottom : '', right : '' });
                var position = $menu.offset(),
                    scrollLeft = $document.scrollLeft(),
                    scrollTop = $document.scrollTop(),
                    marginTop = parseInt($menu.css('margin-top') || '0', 10);

                $menu.css({ position : 'fixed', left : position.left - scrollLeft, top : position.top - scrollTop - marginTop, bottom : 'auto', right : 'auto' });

                var eventListener = function (e) {
                    if ($parent.closest($(e.target)).length) {
                        $parent.removeClass('open');
                        $this.attr('aria-expanded', false);
                        $popup.attr('aria-hidden', true);
                        $this.focus();
                        window.removeEventListener('scroll', eventListener, true);
                    } else if ($parent.closest('html').length === 0) { // If element has been removed from the DOM, remove this event listener
                        window.removeEventListener('scroll', eventListener, true);
                    }
                };
                window.addEventListener('scroll', eventListener, true);
            }

            if (!$popup.data('collapsibleDropdown')) {
                $popup.data('collapsibleDropdown', collapsibleDropdown($popup));
            }

            return false;
        })
        .on('keydown.dropdown.data-api', '[data-toggle=dropdown]', function (e) {
            if (e.keyCode === keyCode.DOWN) {
                var $parent = getParent($(this)),
                    $items = $parent.find('[role=menu] li:not(.divider):visible').children('a, .dropdown-submenu-item, .dropdown-submenu');
                $items.first().focus();
            }
        })
        .on('click.dropdown.data-api', function () {
            clearMenus();
        })
        .on('click.dropdown.data-api', '[data-toggle=submenu]', function (e) {
            var $target = $(this),
                $menu = $target.next('.dropdown-menu');

            if ($menu.length) {
                e.preventDefault();
                e.stopPropagation();
            }
        });

    $(document).on('keydown.dropdown.data-api', '[role=menu]', function (event) {
        var $target = $(event.target),
            $parent = getParent($(event.currentTarget)),
            $items = $parent.find('[role=menu] li:not(.divider):visible').children('a, .dropdown-submenu-item'),
            $closestSubmenu = $parent.closest('.dropdown-submenu');

        if (event.keyCode === keyCode.ESCAPE && $closestSubmenu.length) {
            getParent($closestSubmenu.parent()).addClass('open');
            $closestSubmenu.removeClass('open').children('.dropdown-submenu-item').focus();
        }

        if ($items.length === 1) {
            if (event.keyCode === keyCode.TAB) { //Both TAB and Shift-TAB should remove open on single child menu
                $target.closest('.open').removeClass('open');
            }
        } else if ($target.parent().is(':first-child')) {
            if (event.keyCode === keyCode.TAB && event.shiftKey) {
                $target.closest('.open').removeClass('open');
            } else if (event.keyCode === keyCode.UP) {
                $items.last().focus();
            }
        } else if ($target.parent().is(':last-child')) {
            if (event.keyCode === keyCode.TAB && !event.shiftKey) {
                $target.closest('.open').removeClass('open');
            } else if (event.keyCode === keyCode.DOWN) {
                $items.first().focus();
            }
        }
    });
});


define('primutils/a11y/bootstrap/modal',[
    'jquery',
    'lodash',
    'logger',
    'primutils/a11y/accessibilityUtils',
    'primutils/a11y/focusManager',
    'primutils/utils/keyCode',
    'bootstrap',
    'primutils/utils/emulateTransitionEnd' // For $.fn.emulateTransitionEnd
], function ($, _, logger, utils, focusManager, keyCode) {
    'use strict';

    /**
     * Enhancements to Bootstrap's modal plugin that make it more accessible and usable.
     *
     * You probably do not want to use this directly, use <a href="a11y.bootstrap.accessibilityFixes.html">a11y.bootstrap.accessibilityFixes</a>.
     *
     * @class a11y.bootstrap.modal
     * @module a11y.bootstrap
     * @private
     */

    function blurInner($elem) {
        return function () {
            try {
                if ($elem.get(0) === document.activeElement) { // If $elem is still the active element after deferring
                    $elem.blur();
                }
            } catch (e) {
                logger.error('Error blurring element: ' + e.message);
            }
        };
    }

    function blur($elem) {
        if ($elem.is('html, body')) {
            return;
        }
        _.defer(blurInner($elem));
    }

    var ModalPrototype = $.fn.modal.Constructor.prototype,
        origModalBackdrop = ModalPrototype.backdrop,
        origModalEscape = ModalPrototype.escape,
        origModalShow = ModalPrototype.show,
        origModalHideModal = ModalPrototype.hideModal;

    // WARNING! Both this and ui/widgets/modal.js wrap .backdrop. It currently doesn't matter the order, but please note that both wrap.
    ModalPrototype.backdrop = function () {
        origModalBackdrop.apply(this, arguments);
        if (this.$backdrop) {
            // If there was no animation, then Bootstrap will have already removed $backdrop
            // and set it to null, so we don't need to do this
            this.$backdrop.emulateTransitionEnd(200);
        }
        if (this.$element) {
            this.$element.emulateTransitionEnd(300);
        }
    };

    // WARNING! Both this and ui/widgets/modal.js wrap .show. It currently doesn't matter the order, but please note that both wrap.
    ModalPrototype.show = function () {
        this._originalFocusedElement = document.activeElement;
        blur($(this._originalFocusedElement));

        origModalShow.apply(this, arguments);

        $('body').addClass('prevent-scrolling');

        //Happens after origModalShow is called so this.$backdrop is set up for moving
        var shownModalsWithoutSelf = $('.modal.in, .tray.in').not(this.$element);
        if (shownModalsWithoutSelf.length > 0) {
            //Shift some z-indices...
            var max = _.chain(shownModalsWithoutSelf)
                .map(function (modal) {
                    return parseInt($(modal).css('z-index'), 10);
                })
                .max()
                .value();

            //Modals and backdrops originally start at 1050 and 1040 respectively
            this.$backdrop.css('z-index', max + 1);
            this.$element.css('z-index', max + 10);

            // Offset additional modals by a little bit in order to make it clear that a new modal is on top
            if (this.$element.hasClass('modal')) {
                var $onlyModals = shownModalsWithoutSelf.filter('.modal');
                if ($onlyModals.length) {
                    /* eslint-disable lodash/no-single-chain */
                    var position = $(_.chain($onlyModals)
                        .max(function (modal) {
                            return parseInt($(modal).css('z-index'), 10);
                        })
                        .value()).position();
                    /* eslint-enable lodash/no-single-chain */
                    this.$element.css('top', position.top + 30);
                }
            }
        }

        // Move the modal backdrop to right after the modal, but only if it's already in the DOM
        // This is necessary due to potentially having modals in new stacking contexts, (e.g. position:absolute)
        // as well as when you have modals inside of modals. Additionally, if z-indexing was touched above
        // backdrop needs to be in same parent as child modal
        if (this.$element.parent().length) {
            this.$element.after(this.$backdrop);
        }
    };

    ModalPrototype.hideModal = function () {
        origModalHideModal.apply(this, arguments);
        if ($('.modal.in').length === 0) {
            $('body').removeClass('prevent-scrolling');
        }
    };

    ModalPrototype.escape = function () {
        var that = this;

        if (this.isShown) {
            this.$element.on('keyup.defaultAction.modal', function (e) {
                var $defaultEsc = that.$element.find('[data-default="esc"]'),
                    $target = $(e.target);
                if (e.which === keyCode.ESCAPE && $defaultEsc.length && !$defaultEsc.prop('disabled') && $target.closest('.modal').is(that.$element)) {
                    $defaultEsc.click();
                }
            });

            this.$element.on('keydown.defaultAction.modal', function (e) {
                var $defaultEnter = that.$element.find('[data-default="enter"]').not($(document.activeElement)),
                    $target = $(e.target);
                if (e.which === keyCode.ENTER && $defaultEnter.length && !$defaultEnter.prop('disabled') && $target.closest('.modal').is(that.$element) && ($target.is($defaultEnter) || $target.not('a, button, input[type=radio], input[type=checkbox]').length)) {
                    $defaultEnter.click();
                }
            });
        } else if (!this.isShown) {
            this.$element.off('.defaultAction.modal');

            this.$element.one('hidden', function () {
                if (that._originalFocusedElement && that._originalFocusedElement.parentNode) {
                    that._originalFocusedElement.focus();
                    delete that._originalFocusedElement;
                }
            });

            this.popFocus();
        }

        origModalEscape.apply(this, arguments);
    };

    ModalPrototype.popFocus = function () {
        if (this._focusPushed) {
            focusManager.popFocus(this.$element);
            this._focusPushed = false;
        }
    };

    ModalPrototype.enforceFocus = function () {
        // Only enforce focus if the modal is still going to be visible after it finishes fading in
        // Cases where this is not true is when .hide is called before the transition finishes
        // e.g. loading spinners which are done before the transition finishes
        if (this.isShown) {
            focusManager.pushFocus(this.$element);
            this._focusPushed = true;
        }
    };
});


define('primutils/a11y/bootstrap/tab',['jquery',
    'primutils/utils/keyCode',

    'bootstrap'
], function ($, keyCode) {
    'use strict';

    var tabSelector = '[data-toggle="tab"]',
        tabishSelector = tabSelector + ',.scrollers > a,.collapse-handle > a',
        tabishAndMenusSelector = tabishSelector + ',[data-toggle=dropdown]';

    /**
     * Enhancements to Bootstrap's tab plugin that make it more accessible and usable.
     *
     * You probably do not want to use this directly, use <a href="a11y.bootstrap.accessibilityFixes.html">a11y.bootstrap.accessibilityFixes</a>.
     *
     * @class a11y.bootstrap.tab
     * @module a11y.bootstrap
     * @private
     */

    $(document).on('focusin.tab.data-api, click.collapse.data-api', tabishSelector, function (e) {
        var $tab = $(e.currentTarget),
            $tabbar = $tab.closest('.nav-tabs'),
            focusedOnce = $tabbar.data('focusedOnce'),
            $selected = $tab;

        if ($tab.attr('data-toggle') !== 'tab') {
            $tab = $tabbar.find(tabSelector).first();
        } else if (e.type === 'focusin') {
            $tab = $tabbar.find('.active').find(tabSelector).first();
            if ($tab.length === 0) {
                $tab = $tabbar.find(tabSelector).first();
            }
        }

        $tabbar.data('focusedOnce', true).find(tabishAndMenusSelector)
            .attr('tabindex', '-1') // prevent tabbing, use arrow keys instead
            .attr('aria-selected', false);

        $tab
            .attr('tabindex', '0')
            .attr('aria-selected', ($tab.attr('data-toggle') === 'tab' && e.type === 'click'));

        $tabbar.find('li.active a').attr('aria-selected', true);

        if ((!focusedOnce && !$selected.is($tab) && $tab.length) || (e.type === 'click' && $tab.length)) {
            $tab.focus();
        }
    });

    $(document).on('keydown.tab.data-api', tabishSelector, function (e) {
        var $tab = $(e.currentTarget),
            $tabbar = $tab.closest('.nav-tabs'),
            $li = $tab.closest('li'),
            isScroller = $li.hasClass('scrollers'),
            fwd,
            $newTab, $swapTab,
            $menu,
            tabbarSortable = $tabbar.is('[data-sortable]');

        if (e.keyCode === keyCode.UP || e.keyCode === keyCode.LEFT) {
            $newTab = (!isScroller || $tab.prevAll().length === 0) ? $li.prevAll(':visible').first() : $li;
            fwd = false;
        } else if (e.keyCode === keyCode.DOWN || e.keyCode === keyCode.RIGHT) {
            $menu = $li.find('[data-toggle=dropdown]');
            if (e.keyCode === keyCode.DOWN && $menu.length > 0) {
                $menu.focus().click();
                e.preventDefault();
            } else {
                $newTab = (!isScroller || $tab.next().length === 0) ? $li.nextAll(':visible').first() : $li;
                fwd = true;
            }
        } else if (e.keyCode === keyCode.HOME) {
            $newTab = $li.prevAll(':visible').find(tabSelector).parent().first();
            fwd = false;
        } else if (e.keyCode === keyCode.END) {
            $newTab = $li.nextAll(':visible').find(tabSelector).parent().last();
            fwd = true;
        } else if (e.keyCode === keyCode.SPACE) {
            $tab.click();
            e.preventDefault();
        } else if (e.keyCode === keyCode.F10 && e.shiftKey) {
            $menu = $li.find('[data-toggle=dropdown]');
            if ($menu.length > 0) {
                $menu.focus().click();
                e.preventDefault();
            }
        } else if (tabbarSortable && e.keyCode === keyCode.PAGEUP && e.shiftKey) {
            $newTab = $li;
            $swapTab = $li.prevAll(':visible').first();
            fwd = false;
        } else if (tabbarSortable && e.keyCode === keyCode.PAGEDOWN && e.shiftKey) {
            $newTab = $li;
            $swapTab = $li.nextAll(':visible').first();
            fwd = true;
        }

        if ($swapTab) {
            e.preventDefault();
            // Backwards because we're working with the opposite tab
            $swapTab.filter(':visible')[fwd ? 'insertBefore' : 'insertAfter']($li);
        }

        if ($newTab) {
            e.preventDefault();
            $newTab.filter(':visible').find('> a:not([data-toggle=dropdown])')[fwd ? 'last' : 'first']().focus();
        }
    });

    $(document).on('keydown.tab.data-api', '.nav-tabs [data-toggle=dropdown]', function (e) {
        var $dropdown = $(this),
            $li = $dropdown.closest('li');
        if (e.keyCode === keyCode.LEFT || e.keyCode === keyCode.RIGHT || (e.keyCode === keyCode.F10 && e.shiftKey)) {
            $li.removeClass('open');
            $li.find('[data-toggle=tab]').focus();
        }
    });

    $(document).on('keydown.tab.data-api', '.tab-pane', function (e) {
        var id = $(this).attr('id');
        if (e.keyCode === keyCode.UP && e.ctrlKey) {
            e.preventDefault();
            $('a[href=#' + id + ']').focus();
            e.stopPropagation();
        }
    });
});

define('primutils/a11y/bootstrap/tooltip',['jquery', 'lodash', 'bootstrap'], function ($, _) {
    'use strict';

    var TooltipPrototype = $.fn.tooltip.Constructor.prototype,
        origTooltipFixTitle = TooltipPrototype.fixTitle,
        origTooltipApplyPlacement = TooltipPrototype.applyPlacement,
        origTooltipShow = TooltipPrototype.show;

    TooltipPrototype.fixTitle = function () {
        origTooltipFixTitle.apply(this, arguments);
        if (_.isEmpty(this.$element.attr('aria-label'))) {
            this.$element.attr('aria-label', this.getTitle());
        }
    };

    TooltipPrototype.applyPlacement = function () {
        if (this.$element.hasClass('tooltip-needs-overflow')) {
            this.tip().css({
                position : 'fixed',
                width : this.$tip.outerWidth() + 1
            });
        }
        origTooltipApplyPlacement.apply(this, arguments);
    };

    TooltipPrototype.show = function () {
        if (this.$element.hasClass('tooltip-needs-overflow')) {
            this.tip().css('width', '');
        }
        // Stop events from bubbling and affecting tabs/etc.
        if (this.hasContent() && this.enabled) {
            this.$element.one('show', function (e) {
                e.stopImmediatePropagation();
                this.$element.one('shown', function (e) {
                    e.stopImmediatePropagation();
                });
                this.$element.one('hide', function (e) {
                    e.stopImmediatePropagation();
                    this.$element.one('hidden', function (e) {
                        e.stopImmediatePropagation();
                    });
                }.bind(this));
            }.bind(this));
        }
        origTooltipShow.apply(this, arguments);
    };
});

/**
 * Methods for requesting and saving downloaded files.
 *
 * The export is itself a function that saves files.  Normally, this is not that useful, and you really want to use the `requestAndSave` method.
 *
 * @class utils.fileSaver
 * @module utils
 */
define('primutils/utils/fileSaver',['lodash'], function (_) {
    'use strict';

    var TIMEOUT = 500;
    var DEFAULTS = {
        type : 'text/plain'
    };

    var fileSaver = function (opts, name, fileContents) {
        var options = _.defaults({}, opts, DEFAULTS);
        var blob = new window.Blob([fileContents], { type : options.type });
        var url = window.URL.createObjectURL(blob);
        var link = window.document.createElement('a');
        var ev;

        link.href = url;
        link.download = name;
        link.textContent = 'save';

        if (window.navigator.msSaveOrOpenBlob) {
            window.navigator.msSaveOrOpenBlob(blob, name || 'download');
        } else if (window.document.createEvent) {
            ev = window.document.createEvent('MouseEvents');
            ev.initMouseEvent('click', true, true, window,
                          0, 0, 0, 0, 0, false, false, false, false, 0, null);
            link.dispatchEvent(ev);
        } else {
            link.click();
        }

        //Arbitrary Timeout for Firefox
        //https://bugzilla.mozilla.org/show_bug.cgi?id=939510
        // retested November 19 2015 - still broken
        // how to reproduce: in FF, press the "download" button in FF in the modals demo in the preview dialog example.
        //  if the download does nothing - because FF revokes the URL too early - broken
        //  if the download does something (which it will with the setTimeout), we're all good!
        setTimeout(function () {
            window.URL.revokeObjectURL(url);
        }, TIMEOUT);
    };

    var DEFAULT_REQUEST_HEADERS = {
    };

    /**
     * Requests and saves files to the user's system, using an XHR request and the arraybuffer response type.
     *
     * Once the file is downloaded, this method will present users with the "save file" dialog, just as if you simply linked to the file.
     * Still, this method has the advantage of downloading via an XHR request, meaning that headers (like for CSRF protections) can be globally set by applications using the setDefaultHeaders method.
     * (Just like other global preparation steps in https://confluence.oraclecorp.com/confluence/display/PRIMEENGINEERING/Global+Prepares )
     *
     * Example:
     *
     *     fileSaver.requestAndSave({
     *         fileUrl: '../files/fileid',
     *         fileName:  'File Name', // overriden by server's Content-Disposition header, if present
     *         headers: { ... }
     *     });
     *
     * @method requestAndSave
     * @static
     * @param opts {Object} fileUrl and fileName required, headers optional
     */
    fileSaver.requestAndSave = function (opts) {
        var fileUrl = opts.fileUrl;
        window.open(fileUrl);
    };

    /**
     * Sets default headers for all future requests with the `requestAndSave` method.
     *
     * @method setDefaultHeaders
     * @static
     */
    fileSaver.setDefaultHeaders = function (headers) {
        _.merge(DEFAULT_REQUEST_HEADERS, headers);
    };

    return fileSaver;
});

define('primutils/a11y/bootstrap/files',[
    'jquery',
    'lodash',
    'primutils/utils/fileSaver'
], function ($, _, fileSaver) {
    'use strict';

    $(document).on('click', '[data-fileDownload]', function (ev) {
        ev.preventDefault();

        var $link = $(this);
        var fileUrl = $link.attr('href');
        var fileName = '';
        var altFileName = $link.attr('data-fileDownload');
        if (altFileName !== true && altFileName !== false && altFileName !== 'true' && altFileName !== 'false') {
            fileName = altFileName;
        }
        fileSaver.requestAndSave({
            fileUrl : fileUrl,
            fileName : fileName
        });
    });
});


define('primutils/bootstrapAccessibilityFixes',[
    'jquery',
    'bootstrap',
    'primutils/a11y/bootstrap/button',
    'primutils/a11y/bootstrap/collapse',
    'primutils/a11y/bootstrap/distractionFree',
    'primutils/a11y/bootstrap/dropdown',
    'primutils/a11y/bootstrap/modal',
    'primutils/a11y/bootstrap/tab',
    'primutils/a11y/bootstrap/tooltip',
    'primutils/a11y/bootstrap/files'
], function ($) {
    'use strict';

    // This needs both Bootstrap to be loaded (line 4) and to wait for DOMReady
    $(function () {
        function stopBubble(e) {
            e.stopPropagation();
            e.target.removeEventListener($.support.transition.end, stopBubble);
        }

        window.document.addEventListener($.support.transition.end, function (e) {
            e.target.addEventListener($.support.transition.end, stopBubble);
        }, true);

        $.event.special[$.support.transition.end] = {
            noBubble : true
        };
    });

    /**
     * Enhancements to Bootstrap that make it more accessible and usable.
     *
     * This module loads all of the <a href="../modules/a11y.bootstrap.html">accessibility fixes</a> for Bootstrap.
     *
     * @class a11y.bootstrap.accessibilityFixes
     * @module a11y.bootstrap
     */
});

define('primutils/store/OrderedMap',[
    'lodash'
], function (_) {
    'use strict';

    /**
     * Notes:
     * collection = ordered list of items
     * indexes = lookup map between item key and collection index
     *
     * @param options
     * @class OrderedMap
     * @module store
     * @private
     * @constructor
     */
    function OrderedMap(options) {
        this.options = _.defaults(options, {
            key : 'key'
        });

        if (_.isFunction(options.key)) {
            this._key = functionKeyImpl;
        }

        this.keys = {};
        this.collection = [];
    }

    var functionKeyImpl = function (item) {
        return this.options.key(item);
    };

    var populateIndexes = function () {
        this.keys = {};
        for (var index = 0, len = this.collection.length; index < len; ++index) {
            var item = this.collection[index];
            if (item == null) {
                this.collection.splice(index, 1);
                len--;
                index--;
                continue;
            }
            this.keys[this._key(item)] = index;
        }
    };

    var isKeyValid = function (key) {
        if (key == null) {
            throw new Error('key undefined', 'OrderedMap.js');
        }
        return true;
    };

    OrderedMap.prototype = {
        constructor : OrderedMap,

        /**
         * Return all items
         * @method items
         * @return {Array(Object)}
         */
        items : function () {
            return this.collection;
        },

        /**
         * Return item referenced by key
         * @method item
         * @param key - string
         * @return object
         */
        item : function (key) {
                        if (!isKeyValid.call(this, key)) {
                return null;
            }
            
            var item = this.collection[this.keys[key]];
            return item == null ? null : item;
        },

        /**
         * Add array of items
         * @method addAll
         * @param items {Array(Object)}
         */
        addAll : function (items) {
            for (var index = 0, len = items.length; index < len; index++) {
                this.add(items[index]);
            }
        },

        addAllProcessed : function (items, fn) {
            for (var index = 0, len = items.length; index < len; index++) {
                this.add(fn(items[index]));
            }
        },

        replaceAllProcessed : function (items, fn) {
            for (var index = 0, len = items.length; index < len; index++) {
                var item = fn(items[index]);
                var oldPos = this.keys[this._key(item)];
                this.collection[oldPos] = item;
            }
        },

        /**
         * Add item
         * @method add
         * @param item {Object}
         */
        add : function (item) {
            var newLen = this.collection.push(item);
            this.keys[this._key(item)] = newLen - 1;
            return newLen - 1;
        },

        /**
         * Insert item into a specific position
         * Note - SLOW due to repopulate of indexes
         * @method insertAt
         * @param item {Object}
         * @param position {Number}
         */
        insertAt : function (item, position) {
            if (position < 0 || position > this.collection.length - 1) {
                throw new RangeError('position out of index bounds', 'orderedmap.js');
            }

            this.collection.splice(position, 0, item);
            populateIndexes.call(this);
        },

        /**
         * Replaces the item in the collection with one of the same key.
         *
         * Throws an Error if an item with that key does not already exist
         *
         * @method replace
         * @param item {Object}
         * @return the old item from the collection
         */
        replace : function (item) {
            var pos = this.position(item[this.options.key]);
            if (pos < 0) {
                throw new Error('Key not found in collection', 'orderedmap.js');
            }
            var oldItem = this.collection[pos];
            this.collection[pos] = item;
            return oldItem;
        },

        /**
         * Remove item
         * Note - SLOW due to repopulate of indexes
         * @method remove
         * @param key {String}
         */
        remove : function (key) {
            if (isKeyValid.call(this, key)) {
                var position = this.position(key);
                if (position > -1) {
                    delete this.keys[key];
                    this.collection.splice(position, 1);
                    populateIndexes.call(this);
                }
            }
        },

        /*
         * Like `remove`, but doesn't repopulate indices, assuming that caller will
         */
        _removeQuick : function (key) {
            if (isKeyValid.call(this, key)) {
                var position = this.position(key);
                if (position > -1) {
                    delete this.keys[key];
                    this.collection[position] = undefined;
                }
            }
        },

        removeMany : function (keys) {
            this.collection = _.difference(this.collection, _.map(keys, this.item.bind(this)));
            populateIndexes.call(this);
        },

        /**
         * Return the position of the item
         * @method position
         * @param key {String}
         * @return {Number}
         */
        position : function (key) {
            if (isKeyValid.call(this, key)) {
                var thisKey = this.keys[key];
                return thisKey == null ? -1 : thisKey;
            }
            return -1;
        },

        /**
         * Return whether the ordered map is empty
         * @method isEmpty
         * @return {Boolean}
         */
        isEmpty : function () {
            return this.count() === 0;
        },

        /**
         * Return the size of the ordered map
         * @method count
         * @return {Number}
         */
        count : function () {
            return this.collection.length;
        },

        /**
         * Remove all items from ordered map
         * @method clear
         */
        clear : function () {
            this.keys = {};
            this.collection = [];
        },

        _key : function (item) {
            return item[this.options.key];
        },

        _repopulateIndexes : function () {
            populateIndexes.call(this);
        },

        sort : function (sorter) {
            this.collection = this.collection.sort(sorter);
            this._repopulateIndexes();
        }
    };

    return OrderedMap;
});

define('primutils/store/Collection',['lodash', 'primutils/store/OrderedMap'], function (_, OrderedMap) {
    'use strict';

    var Collection = function (name, options) {
        this.name = name;
        this.map = {};
        this.options = _.defaults(options, {
            getKey : function (/* type */) {
                return 'id';
            }
        });
    };

    var addNewMap = function (type) {
        this.map[type] = new OrderedMap({
            key : this.options.getKey.call(this, type)
        });
    };

    Collection.prototype.constructor = Collection;

    Collection.prototype.add = function (type, item) {
        if (!this.hasType(type)) {
            addNewMap.call(this, type);
        }
        return this.map[type].add(item);
    };

    Collection.prototype.insertAt = function (type, item, position) {
        if (!this.hasType(type)) {
            addNewMap.call(this, type);
        }
        this.map[type].insertAt(item, position);
        return item;
    };

    Collection.prototype.position = function (type, key) {
        if (this.hasType(type)) {
            return this.map[type].position(key);
        }
        return -1;
    };

    Collection.prototype.remove = function (type, key) {
        if (this.hasType(type)) {
            this.map[type].remove(key);
        }
    };

    Collection.prototype._removeQuick = function (type, key) {
        if (this.hasType(type)) {
            this.map[type]._removeQuick(key);
        }
    };

    Collection.prototype.replace = function (type, item) {
        if (this.hasType(type)) {
            return this.map[type].replace(item);
        }
        return null;
    };

    Collection.prototype.item = function (type, key) {
        if (this.hasType(type)) {
            return this.map[type].item(key);
        }

        return null;
    };

    Collection.prototype.items = function (type) {
        if (this.hasType(type)) {
            return this.map[type].items();
        }
        return [];
    };

    Collection.prototype.hasType = function (type) {
                if (type == null) {
            throw new Error('storeCollection.js - "type" parameter not defined');
        }
        
        return this.map[type] != null;
    };

    Collection.prototype.clear = function (type, keys, purge) {
        if (this.hasType(type)) {
            if (_.isArray(keys)) {
                var that = this;
                var itemKeys = _.pluck(this.items(type), that.options.getKey(type));
                _.forEach(itemKeys, function (itemKey) {
                    if (_.includes(keys, itemKey)) {
                        that.remove(type, itemKey, '', purge);
                    }
                });
            } else {
                delete this.map[type];
            }
        }
    };

    Collection.prototype.clearAll = function () {
        this.map = {};
    };

    Collection.prototype.count = function (type) {
        if (this.hasType(type)) {
            return this.map[type].count();
        }
        return 0;
    };

    Collection.prototype.isEmpty = function (type) {
        if (this.hasType(type)) {
            return this.map[type].isEmpty();
        }
        return true;
    };

    Collection.prototype.isEmptyAll = function () {
        return _.every(this.map, function (coll) {
            return coll.isEmpty();
        });
    };

    Collection.prototype.types = function () {
        return _.keys(this.map);
    };

    Collection.prototype._repopulateIndexes = function (type) {
        if (this.hasType(type)) {
            this.map[type]._repopulateIndexes();
        }
    };

    Collection.prototype.sort = function (type, sorter) {
        if (this.hasType(type)) {
            this.map[type].sort(sorter);
        }
    };

    return Collection;
});

define('primutils/utils/nestedPropertyAccessors',[], function () {
    'use strict';

    /**
     * A collection of static utility methods for dealing with nested objects.
     *
     * Suggested import variable name is 'props'.
     *
     * @class utils.nestedPropertyAccessors
     * @module utils
     */

    var DEFAULT_PATH_DELIMITER = '.';

    var api = {

        /**
         * Gets a value from an object.
         *
         * Example:
         *
         *     define(['primutils/utils/nestedPropertyAccessors'], function (props) {
         *         console.log(props.getValue({
         *             a: {
         *                 b: {
         *                     c: 'the value'
         *                 }
         *            }
         *         }, 'a.b.c', '.')); // would print out 'the value'
         *     });
         *
         * @param {Object} object the object whose nested property to get
         * @param {String} field the field to get
         * @param {String} delimiter the delimiter that separates properties in the given field string - defaults to `'.'`
         * @method getValue
         * @static
         */
        getValue : function (item, field, userDelim) {
            if (item == null) {
                return item;
            }

            var delimiter = userDelim || DEFAULT_PATH_DELIMITER;

            var firstLevelIndex = field.indexOf(delimiter);
            if (firstLevelIndex === -1) {
                // return the value
                return item[field];
            }

            // move down a level in the object
            var depth = field.slice(0, firstLevelIndex);
            return api.getValue(item[depth], field.slice(firstLevelIndex + 1), delimiter);
        },

        /**
         * Sets a value on an object.
         *
         * Example:
         *
         *     define(['primutils/utils/nestedPropertyAccessors'], function (props) {
         *         var object = {
         *             a: {
         *                 b: {
         *                     c: 'the value'
         *                 }
         *            }
         *         };
         *         props.setValue(object, 'a.b.d', 'the d value', '.');
         *         console.log(object.a.b.d); // would print out 'the d value'
         *     });
         *
         * *Note*: if you try to set a field on an object that does not have the required nested objects, this method will throw an error.
         * For example, if `object.a.b.c` would throw an undefined error because a is undefined, `setValue(object, 'a.b.c', 'value')` would also fail.
         *
         * @param {Object} object the object whose nested property to set
         * @param {String} field the field to set
         * @param {*} value the value to set
         * @param {String} delimiter the delimiter that separates properties in the given field string - defaults to `'.'`
         * @method setValue
         * @static
         */
        setValue : function (item, field, value, userDelim) {
            var delimiter = userDelim || DEFAULT_PATH_DELIMITER;

            // get last part of path
            var firstLevelIndex = field.indexOf(delimiter);
            if (firstLevelIndex === -1) {
                // set value
                item[field] = value;
            } else {
                // move down a level in the object
                var depth = field.slice(0, firstLevelIndex);
                if (item[depth] == null) {
                    // no object at this depth
                    throw new Error('nesterPropertyAccessors - no object at this depth: ' + field);
                }
                return api.setValue(item[depth], field.slice(firstLevelIndex + 1), value, delimiter);
            }
        }
    };
    api.get = api.getValue;
    api.set = api.setValue;

    return api;
});

define('primutils/store/ModifyCollection',['lodash', 'primutils/store/Collection', 'primutils/utils/nestedPropertyAccessors', 'primutils/utils/objectUtils', 'primutils/utils/isEqual'], function (_, Collection, propHelpers, objectUtils, isEqual) {
    'use strict';

    var PATH_DELIMITER = '.';

    var BREAK = {};

    var hasOwnProp = Object.prototype.hasOwnProperty;
    function has(item, prop) {
        return hasOwnProp.call(item, prop);
    }

    function pathItemEach(item, path, callback, context) {
        var piece = item,
            parts = path.split(PATH_DELIMITER),
            fullPath = '';
        for (var i = 0; i < parts.length; i++) {
            fullPath += parts[i];
            if (!_.isUndefined(piece)) {
                piece = piece[parts[i]];
            }
            if (callback.call(context, parts[i], fullPath, piece) === BREAK) {
                return;
            }
            fullPath += PATH_DELIMITER;
        }
    }

    function reversePathEach(path, callback, context) {
        var pathPart = path,
            parent, key, delimLoc;

        /* eslint-disable no-cond-assign */
        while ((delimLoc = pathPart.lastIndexOf(PATH_DELIMITER)) > 0) {
            parent = pathPart.substr(0, delimLoc);
            key = pathPart.substr(delimLoc + 1);
            if (callback.call(context, key, pathPart) === BREAK) {
                return;
            }
            pathPart = parent;
        }
        /* eslint-enable no-cond-assign */

        // Root path
        callback.call(context, pathPart, pathPart);
    }

    function groupAllByPath(fields) {
        return _.reduce(fields, function (res, field) {
            reversePathEach(field, function (key, path) {
                res[path] = res[path] || [];
                res[path].push(field);
            });
            return res;
        }, {});
    }

    function ModifyStoreCollection(name, options) {
        Collection.call(this, name, options);
        this._proto = Collection.prototype;
    }

    ModifyStoreCollection.prototype = _.assign(Object.create(Collection.prototype), {
        constructor : ModifyStoreCollection,

        hasOriginalValue : function (type, id, field) {
            var item = this.item(type, id),
                found = false;

            if (item != null) {
                if (has(item.oldValues, field)) {
                    return true;
                }

                reversePathEach(field, function (key, path) {
                    if (has(item.oldValues, path)) {
                        found = true;
                        return BREAK;
                    }
                });
            }

            return found;
        },

        getOriginalValue : function (type, id, field) {
            var item = this.item(type, id),
                valueAtDepth, foundAtPath;

            if (item != null) {
                if (has(item.oldValues, field)) {
                    return item.oldValues[field];
                }

                reversePathEach(field, function (key, path) {
                    if (has(item.oldValues, path)) {
                        valueAtDepth = item.oldValues[path];
                        foundAtPath = path;
                        return BREAK;
                    }
                });

                if (foundAtPath != null && valueAtDepth != null) {
                    if (field.indexOf(foundAtPath + PATH_DELIMITER) === 0) {
                        var value = propHelpers.getValue(valueAtDepth, field.substr((foundAtPath.length + PATH_DELIMITER.length)), PATH_DELIMITER);
                        item.oldValues[field] = value;
                    }
                    return item.oldValues[field];
                }
            }

            return undefined;
        },

        _allModifiedFields : function (type, id) {
            var item = this.item(type, id);
            if (item != null && item.oldValues != null) {
                return _.keys(item.oldValues);
            }
            return [];
        },

        modifiedFields : function (type, id) {
            var item = this.item(type, id);
            if (item != null && item.trueModifiedFields != null) {
                return _.keys(item.trueModifiedFields);
            }
            return [];
        },

        isSameValue : function (type, id, field, other) {
            return isEqual(this.getOriginalValue(type, id, field), other);
        },

        _forceMarkModified : function (type, id, field) {
            var item = this.item(type, id);

            if (item == null) {
                throw new Error('ModifyStore() - _forceMarkModified expects the item to be modified already');
            }

            item.trueModifiedFields[field] = true;
        },

        anyModified : function (type, fields) {
            return _.some(this.items(type), function (changeItem) {
                return _.some(fields, _.partial(has, changeItem.trueModifiedFields));
            });
        },

        add : function (type, id, field, originalItem) {
            var item = this.item(type, id),
                key = this.options.getKey.call(this, type),
                oldValues;

            if (item == null) {
                item = {
                    oldValues : {},
                    trueModifiedFields : {}
                };
                item[key] = id;
                this._proto.add.call(this, type, item);
            }

            oldValues = item.oldValues;

            item.trueModifiedFields[field] = true;

            if (!has(oldValues, field)) {
                if (this.hasOriginalValue(type, id, field)) {
                    oldValues[field] = this.getOriginalValue(type, id, field);
                } else {
                    pathItemEach(objectUtils.cloneDeep(originalItem), field, function (key, path, value) {
                        if (!has(oldValues, path)) {
                            oldValues[path] = value;
                        }
                    });
                }
            }
        },

        remove : function (type, id, field, purge) {
            var item = this.item(type, id);

            if (item != null) {
                var oldValues = item.oldValues,
                    oldValueKeys = _.keys(oldValues);

                if (field != null) {
                    delete oldValues[field];
                    delete item.trueModifiedFields[field];

                    _.forEach(oldValueKeys, function (path) {
                        if (path.indexOf(field + PATH_DELIMITER) === 0) {
                            delete oldValues[path];
                            delete item.trueModifiedFields[path];
                        }
                    });

                    var trueModifiedGroups = groupAllByPath(_.keys(item.trueModifiedFields)),
                        extraOldValues = _.difference(_.keys(oldValues), _.keys(trueModifiedGroups));

                    _.forEach(extraOldValues, function (path) {
                        delete oldValues[path];
                    });
                }

                // remove the item if no modified fields exist
                // If purge is true, get rid of the whole record - this is
                //     because the remove method for this collection is by
                //     field and not by id.
                if (_.isEmpty(oldValues) || purge === true) {
                    this._proto.remove.call(this, type, id);
                }
            }

            // remove the type if type is empty
            if (this.isEmpty(type)) {
                this.clear(type);
            }
        },

        doMarkModified : function (type, key, field, newValue, originalItem) {
            if (this.hasOriginalValue(type, key, field)) {
                // is the updated value the same as the original
                if (this.isSameValue(type, key, field, newValue)) {
                    // remove value from modified collection
                    this.remove(type, key, field);
                } else {
                    this._forceMarkModified(type, key, field);
                }
            } else {
                this.add(type, key, field, originalItem);
            }
        }
    });

    return ModifyStoreCollection;
});

define('primutils/store/IndexManager',[
    'lodash',
    'logger'
], function (_, logger) {
    'use strict';

    var without = function (array, target) {
        var shallow = array.slice();
        shallow.splice(_.indexOf(array, target), 1);
        return shallow;
    };

    function IndexManager(indexDefinitions, options) {
        var that = this;

        this.accessors = options.utilities;
        this.getKeyFn = options.getKey;

        this.indexDefinitions = {
            parentToChild : {},
            childToParent : {},
            unique : {},
            group : {}
        };

        if (indexDefinitions) {
            if (indexDefinitions.foreign) {
                                logger.assert(_.isArray(indexDefinitions.foreign), 'indices must be an array, or not provided');
                                _.forEach(indexDefinitions.foreign, function (index) {
                    var primaryKey = that.getKeyFn(index.primaryType);
                    if (!_.has(index, 'primaryId')) {
                        index.primaryId = primaryKey;
                    } else if (index.primaryId !== primaryKey) {
                        that.indexDefinitions.unique[index.primaryType] = that.indexDefinitions.unique[index.primaryType] || [];
                        that.indexDefinitions.unique[index.primaryType].push(index.primaryId);
                    }

                    that.indexDefinitions.parentToChild[index.primaryType] = that.indexDefinitions.parentToChild[index.primaryType] || {};
                    logger.assert(that.indexDefinitions.parentToChild[index.primaryType][index.foreignType] == null, 'IndexManager | Only one ' + index.primaryType + '->' + index.foreignType + ' relationship can be defined');
                    that.indexDefinitions.parentToChild[index.primaryType][index.foreignType] = index.primaryId;

                    that.indexDefinitions.childToParent[index.foreignType] = that.indexDefinitions.childToParent[index.foreignType] || {};
                    that.indexDefinitions.childToParent[index.foreignType][index.primaryType] = index.foreignId;
                });
            }

            if (indexDefinitions.unique) {
                                logger.assert(_.isArray(indexDefinitions.unique), 'indices must be an array, or not provided');
                                _.forEach(indexDefinitions.unique, function (index) {
                    that.indexDefinitions.unique[index.uniqueType] = that.indexDefinitions.unique[index.uniqueType] || [];
                    that.indexDefinitions.unique[index.uniqueType].push(index.uniqueField);
                });
            }

            if (indexDefinitions.group) {
                                logger.assert(_.isArray(indexDefinitions.group), 'indices must be an array, or not provided');
                                _.forEach(indexDefinitions.group, function (index) {
                    that.indexDefinitions.group[index.groupType] = that.indexDefinitions.group[index.groupType] || [];
                    that.indexDefinitions.group[index.groupType].push(index.groupField);
                });
            }
        }

        this.allCleared();
    }

    var addChildIndex = function (idxManager, childType, parentType, parentId, childId) {
        var typeIndices = idxManager.indices.children[parentType][childType];
        if (!_.get(idxManager.indexExists.children, [parentType, childType, childId])) {
            typeIndices[parentId] = typeIndices[parentId] || [];
            typeIndices[parentId].push(childId);
            _.set(idxManager.indexExists.children, [parentType, childType, childId], true);
        }

        idxManager.indices.parent[childType][parentType][childId] = parentId;
    };

    var removeChildIndex = function (idxManager, childType, parentType, parentId, childId) {
        idxManager.indices.children[parentType][childType][parentId] = without(idxManager.indices.children[parentType][childType][parentId], childId);
        _.set(idxManager.indexExists.children, [parentType, childType, childId], false);
        delete idxManager.indices.parent[childType][parentType][childId];
    };

    IndexManager.prototype = {
        constructor : IndexManager,

        getChildIds : function (primaryType, primaryId, foreignType) {
            logger.assert(this.indices.children[primaryType] && this.indices.children[primaryType][foreignType], 'IndexManager | Relationship ' + primaryType + '->' + foreignType + ' not setup');
            if (primaryId == null) {
                return (this.indices.children[primaryType][foreignType].null || []).concat(this.indices.children[primaryType][foreignType][undefined] || []);
            } else {
                return this.indices.children[primaryType][foreignType][primaryId] || [];
            }
        },

        getParentId : function (foreignType, foreignId, primaryType) {
            logger.assert(this.indices.parent[foreignType] && this.indices.parent[foreignType][primaryType], 'IndexManager | Relationship ' + primaryType + '->' + foreignType + ' not setup');
            return this.indices.parent[foreignType][primaryType][foreignId];
        },

        getTrueParentId : function (foreignType, foreignId, primaryType) {
            var indexedId = this.getParentId(foreignType, foreignId, primaryType);
            if (this.indexDefinitions.parentToChild[primaryType][foreignType] === this.getKeyFn(primaryType)) {
                return indexedId;
            } else {
                return this.getPrimaryKey(primaryType, this.indexDefinitions.parentToChild[primaryType][foreignType], indexedId);
            }
        },

        getPrimaryKey : function (type, uniqueKey, uniqueValue) {
            logger.assert(this.indices.unique[type] && this.indices.unique[type][uniqueKey], 'Unique key ' + type + '/' + uniqueKey + ' not setup');
            return this.indices.unique[type][uniqueKey][uniqueValue];
        },

        getGroupIds : function (type, groupKey, groupValue) {
            logger.assert(this.indices.group[type] && this.indices.group[type][groupKey], 'Group key ' + type + '/' + groupKey + ' not setup');
            if (groupValue == null) {
                return _.union(this.indices.group[type][groupKey].null || [], this.indices.group[type][groupKey][undefined] || []);
            } else {
                return this.indices.group[type][groupKey][groupValue];
            }
        },

        getChildTypes : function (type) {
            return _.keys(this.indexDefinitions.parentToChild[type]);
        },

        itemAdded : function (type, item) {
            var that = this,
                itemId = this.accessors._getValue(item, this.getKeyFn(type));

            _.forEach(this.indexDefinitions.childToParent[type], function (foreignField, primaryType) {
                var value = that.accessors._getValue(item, foreignField);
                addChildIndex(that, type, primaryType, value, itemId);
            });
            _.forEach(this.indexDefinitions.unique[type], function (field) {
                var value = that.accessors._getValue(item, field);
                if (value != null) {
                    logger.assert(!_.has(that.indices.unique[type][field], value), 'IndexManager | Unique constraint violated: ' + type + '/' + field);
                    that.indices.unique[type][field][value] = itemId;
                }
            });
            _.forEach(this.indexDefinitions.group[type], function (field) {
                var value = that.accessors._getValue(item, field);
                var fieldIndices = that.indices.group[type][field];
                if (!_.get(that.indexExists.group, [type, field, itemId])) {
                    fieldIndices[value] = fieldIndices[value] || [];
                    fieldIndices[value].push(itemId);
                    _.set(that.indexExists.group, [type, field, itemId], true);
                }
            });
        },

        itemUpdated : function (type, item, field, value, oldValue) {
            var that = this,
                itemId = this.accessors._getValue(item, this.getKeyFn(type));

            _.forEach(this.indexDefinitions.parentToChild[type], function (/* primaryField, foreignType */) {
                that.itemRemoved(type, itemId, item);
                that.itemAdded(type, item);
            });
            _.forEach(this.indexDefinitions.childToParent[type], function (foreignField, primaryType) {
                if (field === foreignField) {
                    removeChildIndex(that, type, primaryType, oldValue, itemId);
                    addChildIndex(that, type, primaryType, value, itemId);
                }
            });
            if (_.has(this.indices.unique[type], field)) {
                delete this.indices.unique[type][field][oldValue];
                if (value != null) {
                    logger.assert(!_.has(that.indices.unique[type][field], value), 'IndexManager | Unique constraint violated: ' + type + '/' + field);
                    that.indices.unique[type][field][value] = itemId;
                }
            }
            if (_.has(this.indices.group[type], field)) {
                var fieldIndices = that.indices.group[type][field];
                fieldIndices[oldValue] = without(fieldIndices[oldValue], itemId);
                fieldIndices[value] = fieldIndices[value] || [];
                fieldIndices[value].push(itemId);
            }
        },

        itemRemoved : function (type, key, item) {
            var that = this;

            // Do NOT remove child indices because there is no way to get them back when restoring a parent item without repopulating the entire index
            //_.each(this.indexDefinitions.parentToChild[type], function (primaryField, foreignType) {
            //    var value = that.accessors._getValue(item, primaryField);
            //
            //    delete that.indices.children[type][foreignType][value];
            //    that.indices.parent[foreignType][type] = _.reduce(that.indices.parent[foreignType][type], function (memo, primaryId, foreignId) {
            //        if (value !== primaryId) {
            //            memo[foreignId] = primaryId;
            //        }
            //        return memo;
            //    }, {});
            //});
            _.forEach(this.indexDefinitions.childToParent[type], function (foreignField, primaryType) {
                var value = that.accessors._getValue(item, foreignField);
                removeChildIndex(that, type, primaryType, value, key);
            });
            _.forEach(this.indexDefinitions.unique[type], function (field) {
                var value = that.accessors._getValue(item, field);
                delete that.indices.unique[type][field][value];
            });
            _.forEach(this.indexDefinitions.group[type], function (field) {
                var value = that.accessors._getValue(item, field);
                _.set(that.indexExists.group, [type, field, key], false);
                that.indices.group[type][field][value] = without(that.indices.group[type][field][value], key);
            });
        },

        typeCleared : function (type) {
            var that = this;

            // Only clear if the type being deleted is the foreign type
            // Do NOT remove child indices because there is no way to get them back when restoring a parent item without repopulating the entire index
            _.forEach(this.indices.children, function (indices, primaryType) {
                _.forEach(indices, function (map, foreignType) {
                    if (foreignType === type) {
                        indices[foreignType] = {};
                        if (that.indexExists.children[primaryType] != null) {
                            delete that.indexExists.children[primaryType][foreignType];
                        }
                    }
                });
            });
            // Only clear if the type being deleted is the foreign type
            // Do NOT remove child indices because there is no way to get them back when restoring a parent item without repopulating the entire index
            _.forEach(this.indices.parent, function (indices, foreignType) {
                _.forEach(indices, function (map, primaryType) {
                    if (foreignType === type) {
                        indices[primaryType] = {};
                    }
                });
            });
            that.indices.unique[type] = _.reduce(this.indexDefinitions.unique[type], function (memo, field) {
                memo[field] = {};
                return memo;
            }, {});
            that.indices.group[type] = _.reduce(this.indexDefinitions.group[type], function (memo, field) {
                memo[field] = {};
                return memo;
            }, {});
            _.set(that.indexExists.group, [type], {});
        },

        allCleared : function () {
            var that = this;

            this.indices = {
                children : {},
                parent : {},
                unique : {},
                group : {}
            };
            this.indexExists = {
                children : {},
                group : {}
            };
            _.forEach(that.indexDefinitions.parentToChild, function (indices, primaryType) {
                that.indices.children[primaryType] = _.mapValues(indices, function () { return {}; });
            });
            _.forEach(that.indexDefinitions.childToParent, function (indices, foreignType) {
                that.indices.parent[foreignType] = _.mapValues(indices, function () { return {}; });
            });
            _.forEach(that.indexDefinitions.unique, function (fields, type) {
                that.indices.unique[type] = _.reduce(fields, function (memo, field) {
                    memo[field] = {};
                    return memo;
                }, {});
            });
            _.forEach(that.indexDefinitions.group, function (fields, type) {
                that.indices.group[type] = _.reduce(fields, function (memo, field) {
                    memo[field] = {};
                    return memo;
                }, {});
            });
        }
    };

    var noopIdxMgr = _.mapValues(IndexManager.prototype, function () {
        return _.noop;
    });
    delete noopIdxMgr.constructor;
    Object.freeze(noopIdxMgr);

    IndexManager.noopIndexManager = function () {
        return noopIdxMgr;
    };

    return IndexManager;
});

/**
 * @module utils
 */
define('primutils/utils/ajax',[
    'jquery',
    'lodash'
], function ($, _) {
    'use strict';

    /**
     * A collection of AJAX utility functions.
     *
     * Mainly used by the Primutil's store in its utilities namespace.
     *
     * @class utils.ajax
     */
    var ajax = {
        /**
         * Ajax get call
         *
         * ALIASES: GET
         *
         * @method get
         * @static
         * @param {String} url The URL to get
         * @param {Boolean} [async] Whether to send the data asynchronously
         * @return {jQuery Promise} A jQuery promise
         */
        get : function (url, async) {
            return $.ajax({
                url : url,
                dataType : 'json',
                type : 'GET',
                async : (async == null) ? true : async
            });
        },

        /**
         * Ajax post call
         *
         * ALIASES: POST
         *
         * @method post
         * @static
         * @param {String} url The URL to POST to
         * @param {Object} data The data to send
         * @param {Boolean} [async] Whether to send the data asynchronously
         * @return {jQuery Promise} A jQuery Promise
         */
        post : function (url, data, async) {
            return $.ajax({
                url : url,
                type : 'POST',
                data : (_.isObject(data) ? JSON.stringify(data) : data),
                contentType : 'application/json',
                async : (async == null) ? true : async
            });
        },

        /**
         * Ajax put call
         *
         * ALIASES: PUT
         *
         * @method put
         * @static
         * @param {String} url The URL to PUT to
         * @param {Object} data The data to send
         * @param {Boolean} [async] Whether to send the data asynchronously
         * @return {jQuery Promise} A jQuery Promise
         */
        put : function (url, data, async) {
            return $.ajax({
                url : url,
                type : 'PUT',
                data : (_.isObject(data) ? JSON.stringify(data) : data),
                contentType : 'application/json',
                async : (async == null) ? true : async
            });
        },

        /**
         * Ajax delete call
         *
         * ALIASES: DELETE
         *
         * @method delete
         * @static
         * @param {String} url The URL to DELETE to
         * @param {Object} data The data to send
         * @param {Boolean} [async] Whether to send the data asynchronously
         * @return {jQuery Promise} A jQuery Promise
         */
        delete : function (url, data, async) {
            return $.ajax({
                url : url,
                type : 'DELETE',
                data : (_.isObject(data) ? JSON.stringify(data) : data),
                contentType : 'application/json',
                async : (async == null) ? true : async
            });
        }
    };

    // Legacy names
    ajax.GET = ajax.get;
    ajax.PUT = ajax.put;
    ajax.POST = ajax.post;
    ajax.DELETE = ajax.delete;

    return ajax;
});

/**
 * This class is still available as the `primutils/utils/promisifyError` module, but the documentation is now located at [promises](utils.promises.html#method_promisifyError).
 *
 * @private
 * @module utils
 * @class utils.promisifyError
 */
define('primutils/utils/promisifyError',['jquery', 'lodash', 'logger'], function ($, _, logger) {
    'use strict';

    return function (fn) {
        return function () {
            var prom = $.Deferred();
            try {
                var result = fn.apply(this, _.toArray(arguments));
                if (result && result.then && _.isFunction(result.then)) { // we'll assume it's a promise
                    return result;
                }
                return prom.resolve(result).promise();
            } catch (err) {
                logger.error('ERROR occurred, wrapping as a rejected promise: ' + err);
                                if (window.console && window.console.error && err.stack) {
                    window.console.error(err.stack);
                }
                                return prom.reject(err).promise();
            }
        };
    };
});



/**
 * This helper is documented at [promises](utils.promises.html#method_stripResolveData).
 *
 * @private
 * @module utils
 * @class utils.stripPromiseResolveData
 */

define('primutils/utils/stripPromiseResolveData',['lodash', 'jquery'], function (_, $) {
    'use strict';

    return function (promise) {
        var dfd = $.Deferred();

        promise
            .done(function () {
                dfd.resolve();
            })
            .fail(function () {
                dfd.reject.apply(dfd, arguments);
            });

        return dfd.promise();
    };
});



/**
 * A collection of small utilities for use with asynchronous control flow in JS, specifcally methods around dealing with promises.
 *
 * All methods in this module are static.  See methods list, below, for details.
 *
 * @module utils
 * @class utils.promises
 */

define('primutils/utils/promises',[
    'jquery',
    'lodash',
    'primutils/utils/promisifyError',
    'primutils/utils/stripPromiseResolveData'
], function ($, _, promisifyError, stripPromiseResolveData) {
    'use strict';

    return {
        /**
         * Returns a new promise.
         *
         * @method new
         * @static
         */
        new : function () {
            return $.Deferred();
        },

        /**
         * `promisifyError` is a utility designed to improve error handling that may occur in promise chains.
         *
         * To describe it's use, say we have the following:
         *
         *     // define statement, etc.
         *     $.get(...)
         *         .then(appendElementsToPage)
         *         .fail(showAlertDialogOnFailure);
         *
         * This code could work: GET some information from the server, and use it to insert something into the DOM.
         * If the AJAX request fails, the alert dialog would get shown.
         * But if the AJAX request goes through and `appendElementsToPage` throws an uncaught error, showing the alert dialog wouldn't happen (e.g. if the function was `function appendElementsToPage() { throw new Error(); }`)
         * This utility function allows us to address the error.
         * To update the above example, let's try the following:
         *
         *     define([..., 'primutils/utils/promises', function (..., promises) {
         *         $.get(...)
         *            .then(promises.promisifyError(appendElementsToPage))
         *            .fail(showAlertDialogOnFailure);
         *     });
         *
         * With this snippet, both an error in the AJAX request and the `appendElementsToPage` function will cause `showAlertDialogOnFailure` to get called.
         * If both the AJAX request and `appendElementsToPage` operate successfully, the behavior is the same as the first snippet.
         *
         * *Note*: this function will always return a promise, even if the result was passing.
         *
         * *Note to Prime developers*: if you're using [AJAXErrorHandler](http://pgbuci.us.oracle.com:8080/job/ui-platform-cont/Javascript_YUIDocs/classes/shared.dialogs.AJAXErrorHandler.html) as your failure handler, AJAXErrorHandler will propagate the underlying JS error for better debugging.
         *
         * *alias*: `primutils/utils/promisifyError` module.
         *
         * @method promisifyError
         * @param fn {Function} the function to be called, which is given the promise's resolved data (as if a `done` call)
         * @return a promise, which will be rejected() if the given fn throws some error
         * @static
         * @since 2.9.0 (`primutils/utils/promisifyError` module was available earlier)
         */
        promisifyError : promisifyError,

        /**
         * `stripResolveData` is a utility designed to encourage clean contracts among promise consumers.
         *
         * To describe it's use, say we have the following:
         *
         *     // define statement, etc.
         *     function loadData() {
         *         return $.get(...)
         *             .done(function (data) {
         *                 store.addAll(..., data.values);
         *             });
         *     }
         *     // ...
         *     loadData().done(function (data) { // well, I have data, so what they heck... });
         *
         * This function works, but tends to promote a bad practice - since loadData returns the original promise (`done` does not change the promise chaining context), consumers of the loadData function may attempt to use the same data.values member, even though it has already been loaded into the store.
         *
         * `stripResolveData` can be used as such to strip any success data given by the original promise resolution (e.g. the data, for AJAX requests), but keep any failure data:
         *
         *     define([..., 'primutils/utils/promises'], function (..., promises) {
         *         function loadData() {
         *             return promises.stripResolveData($.get(...)
         *                 .done(function (data) {
         *                     store.addAll(..., data.values);
         *                 }));
         *         }
         *         loadData().done(function () { // arguments is an empty list here, can't use it });
         *     });
         *
         * *alias*: `primutils/utils/stripPromiseResolveData` module.
         *
         * @method stripResolveData
         * @param promise {Promise} the promise to mirror in terms of success/failure results, but whose resolved data should not be given to consumers
         * @return a promise, which will be resolved() if the original promise is resolved, and will maintain failure data if the original promise is rejected
         * @static
         * @since 2.9.0
         */
        stripResolveData : stripPromiseResolveData,

        promisifyWebSocket : function (ws) {
            var that = this,
                wasRejected = false,
                dfd = $.Deferred();

            function isCloseErrorCode(code) {
                return (code >= 1001 && code <= 1999) || code === that.webSocketCodes.CLOSE_PROMISE_REJECTED;
            }

            ws.addEventListener('close', function (e) {
                if (!e.wasClean || wasRejected || isCloseErrorCode(e.code)) {
                    dfd.reject();
                } else {
                    dfd.resolve();
                }
            }, false);

            ws.addEventListener('error', function () {
                dfd.reject();
            }, false);

            ws.resolve = (function (closeConnection) {
                if (dfd.state() === 'pending') {
                    if (closeConnection && (this.readyState !== window.WebSocket.CLOSED || this.readyState !== window.WebSocket.CLOSING)) {
                        this.close();
                    } else {
                        dfd.resolve();
                    }
                } else if (closeConnection && (this.readyState !== window.WebSocket.CLOSED || this.readyState !== window.WebSocket.CLOSING)) {
                    this.close();
                }
            }.bind(ws));

            ws.reject = (function (closeConnection) {
                if (dfd.state() === 'pending') {
                    wasRejected = true;
                    if (closeConnection && (this.readyState !== window.WebSocket.CLOSED || this.readyState !== window.WebSocket.CLOSING)) {
                        this.close(that.webSocketCodes.CLOSE_PROMISE_REJECTED, 'Client requested close but as a failure');
                    } else {
                        dfd.reject();
                    }
                } else if (closeConnection && (this.readyState !== window.WebSocket.CLOSED || this.readyState !== window.WebSocket.CLOSING)) {
                    this.close(that.webSocketCodes.CLOSE_PROMISE_REJECTED, 'Client requested close but as a failure');
                }
            }.bind(ws));

            return dfd.promise();
        },

        webSocketCodes : {
            CLOSE_PROMISE_REJECTED : 3901
        }
    };
});

/**
 * @module utils.websockets
 */
define('primutils/utils/websockets/Basic',[
    'jquery',
    'lodash',
    'logger',
    'primutils/utils/Subscribable',
    'primutils/utils/promises'
], function ($, _, logger, Subscribable, promises) {
    'use strict';

    function logWSFailed() {
        logger.error('WebSocket was never opened, calling methods on it will not work');
    }

    // Separate out into own method so JIT can still optimize the constructor
    function getWebSocket(url, protocols) {
        try {
            if (protocols != null) {
                return new window.WebSocket(url, protocols);
            } else {
                return new window.WebSocket(url);
            }
        } catch (e) {
            return null;
        }
    }

    var DEFAULT_OPTIONS = {
        /**
         * Parse the incoming and outgoing messages as JSON objects
         *
         * @property parseJSON
         * @type Boolean
         * @default true
         * @for utils.websockets.Basic
         */
        parseJSON : true,
        /**
         * Automatically close the WebSocket if the server sends a malformed message
         *
         * @property closeOnInvalidMessage
         * @type Boolean
         * @default true
         * @for utils.websockets.Basic
         */
        closeOnInvalidMessage : true
        /**
         * The protocol(s) being used to communicate with the server
         *
         * @property protocol
         * @type String|Array
         * @default null
         * @for utils.websockets.Basic
         */
    };

    /**
     * A simple WebSocket wrapper.
     *
     * Supports sending and receiving data from a WebSockets server optionally (on by default) parsing the data on
     * the wire as JSON.
     *
     * It also implements a jQuery-like promise API which makes it compatible with `$.when`, `$.Deferred.then`, etc.
     *
     * Example:
     *
     *     var ws = new BasicWebSocket('ws://localhost:8080/echo');
     *     ws.on('message', function (e) {
     *         console.log(e.data);
     *         promise.resolveAndClose(); // also ws.resolveAndClose();
     *     });
     *     ws.done(function () {
     *         console.log('connection was closed');
     *     });
     *     ws.send({ foo : 'bar' });
     *
     * Because WebSockets can be long-lived, potentially the whole time the page is open, you can resolve the
     * WebSocket's promise yourself when there is enough data for whatever is waiting on the promise to proceed.
     *
     * This can be useful for pages with a lot of data.
     * Once the metadata has been loaded, you can signal the promise to continue while data continues loading in the
     * background.
     *
     *     var ws = new BasicWebSocket('ws://localhost:8080/load');
     *     ws.on('message', function (e) {
     *         if (e.data.meta) {
     *             addMetadata(edata.meta);
     *             promise.resolve(); // but leave open
     *         }
     *
     *         if (e.data.complete) {
     *             ws.close(); // Close the WebSocket when we're done
     *         }
     *     });
     *     ws.done(function () {
     *         // continue loading page with metadata
     *         // additional data will appear as the page loads
     *     });
     *     ws.send({ projectId : '1' });
     *
     * For a more simplified interface with WebSockets, see the [utils.ajaws](utils.ajaws.html) utility.
     *
     * @class utils.websockets.Basic
     * @extends utils.Subscribable
     * @param url {String} The WebSocket URL to connection to (**MUST** begin with `ws://` or `wss://`, see [utils.urls.templateWebSocket](utils.urls.html#method_templateWebSocket))
     * @param [options] {Object} Additional options (see the Properties tab)
     * @constructor
     */
    function BasicWebSocket(url, options) {
        var that = this,
            logFailed = _.once(logWSFailed),
            promiseFns, ws, dfd;

        logger.assert(window.WebSocket != null, 'This browser does not support WebSockets');
        logger.assert(arguments.length >= 1, 'URL is required');

        Subscribable.call(this);

        this.options = _.merge({}, DEFAULT_OPTIONS, options);

        ws = getWebSocket(url, this.options.protocol);

        if (ws && ws.readyState !== window.WebSocket.CLOSED) {
            /**
             * Triggered when the WebSocket is opened.
             *
             * **Warning**: if you wait too long to attach to this event, the WebSocket will already be opened and this event will not trigger.
             *
             * @event open
             */
            /**
             * Triggered when the WebSocket is closed.
             *
             * @event close
             */
            /**
             * Triggered when the WebSocket encounters an error.
             *
             * @event error
             */
            _.forEach(['open', 'close', 'error'], function (evt) {
                ws.addEventListener(evt, function proxiedEventHandler() {
                    that.trigger(evt);
                }, false);
            });
            ws.addEventListener('message', function basicWebsocketMessageHandler(e) {
                var parseResult = that._parseData(e.data);
                if (parseResult.parsed) {
                    that._triggerMessage(parseResult.value, promiseFns, e.data);
                } else {
                    that._invalidMessage(e.data, promiseFns);
                }
            }, false);
            /**
             * Add a function that is called when the WebSocket closes cleanly or when the developer calls [`.resolve`](#method_resolve).
             * @method done
             * @param fn {Function} Function to call
             */
            /**
             * Add a function that is called when the WebSocket closes uncleanly or when the developer calls [`.reject`](#method_reject).
             * @method fail
             * @param fn {Function} Function to call
             */
            /**
             * Add a function that is called when the WebSocket is closed or the developer marks the WebSocket promise as complete.
             * @method always
             * @param fn {Function} Function to call
             */
            /**
             * The state of the WebSocket's promise.
             * @method state
             * @returns String 'pending', 'resolved', or 'rejected'
             */
            /**
             * Chains a promise to this WebSocket.
             * @method then
             * @param doneFn {Function} Function to call when promise is resolved
             * @param failFn {Function} Function to call when promise is rejected
             */
            /**
             * Gets a read-only version of this WebSocket's promise.
             * @method promise
             * @returns {jQuery.promise}
             */
            /**
             * Add a function that is called when the WebSocket reports a status update.
             *
             * This feature is not currently supported.
             *
             * @method progress
             * @param fn {Function} Function to call
             */
            dfd = promises.promisifyWebSocket(ws);
            promiseFns = {
                /**
                 * Mark the promise as resolved.
                 * @method resolve
                 */
                resolve : function () {
                    ws.resolve();
                },
                /**
                 * Mark the promise as resolved and close the connection to the server.
                 * @method resolveAndClose
                 */
                resolveAndClose : function () {
                    ws.resolve(true);
                },
                /**
                 * Mark the promise as rejected.
                 * @method reject
                 */
                reject : function () {
                    ws.reject();
                },
                /**
                 * Mark the promise as rejected and close the connection to the server.
                 * @method rejectAndClose
                 */
                rejectAndClose : function () {
                    ws.reject(true);
                }
            };

            /**
             * Returns whether the WebSocket is open for communication with the server.
             * @method isOpen
             * @return Boolean
             */
            this.isOpen = function () {
                return ws.readyState === window.WebSocket.OPEN;
            };

            /**
             * Sends a raw string to the server.
             * @method rawSend
             * @param data {String} The data to send
             * @private
             */
            this.rawSend = function (data) {
                if (this.isOpen()) {
                    ws.send(data);
                } else {
                    logger.error('WebSocket is not open');
                }
            };

            /**
             * Closes the connection to the server.
             * @method close
             */
            this.close = function () {
                ws.close();
            };

            _.assign(this, dfd, promiseFns);
        } else {
            dfd = $.Deferred().reject().promise();
            _.assign(this, {
                isOpen : function () { return false; },
                rawSend : logFailed,
                close : logFailed
            }, dfd, {
                resolve : logFailed,
                resolveAndClose : logFailed,
                reject : logFailed,
                rejectAndClose : logFailed
            });
        }

        if (this.options.parseData) {
            this._parseData = this.options.parseData;
        }
        if (this.options.triggerMessage) {
            this._triggerMessage = this.options.triggerMessage;
        }
    }

    BasicWebSocket.prototype = _.create(Subscribable.prototype, {
        constructor : BasicWebSocket,

        /**
         * Sends data to the server.
         * If `parseJSON` is true, calls `JSON.stringify` on the object before sending it across the wire.
         * @method send
         * @param data {*} The data to send to the server
         */
        send : function (data) {
            return this.rawSend(this.options.parseJSON ? JSON.stringify(data) : data);
        },

        /**
         * Parses the raw data sent from the server.
         * If `parseJSON` is true, calls `JSON.parse` on the received string.
         * @method _parseData
         * @param rawData {String} The data to parse
         * @protected
         * @return {{parsed:Boolean,value:*}} Whether parsing was successful and the value that was parsed
         */
        _parseData : function (rawData) {
            if (this.options.parseJSON) {
                try {
                    return { parsed : true, value : JSON.parse(rawData) };
                } catch (e) {
                    logger.fatal('Server message was not JSON');
                    return { parsed : false };
                }
            } else {
                return { parsed : true, value : rawData };
            }
        },

        /**
         * Triggers events based on the received data.
         * @method _triggerMessage
         * @param parsedData {*} The data parsed by the `_parseData` method
         * @param promiseFns {Object} Functions that can be used from within the callback
         * @param rawData {String} The raw data from the server
         * @protected
         */
        _triggerMessage : function (parsedData, promiseFns /*, rawData */) {
            /**
             * Triggered when a message is received from the server
             *
             * The event object contains the following properties:
             *
             * - `data`: The parsed data from the server
             * - `promise`: Functions that can be performed on the WebSocket (`resolve`, `resolveAndClose`, `reject`, `rejectAndClose`)
             *
             * @event message
             * @param e {Object}
             */
            this.trigger('message', {
                data : parsedData,
                promise : promiseFns
            });
        },

        /**
         * Triggers an invalid message event based on the raw data.
         * If `closeOnInvalidMessage` is true, this will also call `rejectAndClose`.
         * @method _invalidMessage
         * @param rawData {String} The raw data from the server
         * @param promiseFns {Object} Functions that can be used from within the callback
         * @protected
         */
        _invalidMessage : function (rawData, promiseFns) {
            /**
             * Triggered when an invalid message is received from the server
             *
             * The event object contains the following properties:
             *
             * - `data`: The raw data from the server
             * - `promise`: Functions that can be performed on the WebSocket (`resolve`, `resolveAndClose`, `reject`, `rejectAndClose`)
             *
             * @event invalid-message
             * @param e {Object}
             */
            this.trigger('invalid-message', {
                data : rawData,
                promise : promiseFns
            });
            if (this.options.closeOnInvalidMessage) {
                this.rejectAndClose();
            }
        }
    });

    return BasicWebSocket;
});

define('primutils/utils/websockets/Command',[
    'lodash',
    'logger',
    'primutils/utils/websockets/Basic'
], function (_, logger, BasicWebSocket) {
    'use strict';

    /**
     * A WebSocket wrapper that supports sending command/message style messages to the server.
     *
     * Messages sent will be in the form:
     *
     *     COMMAND                 The command name (any-case) followed by a new-line
     *     {"foo":"bar"}           The payload, usually a stringified JSON object
     *
     * Example:
     *
     *     var ws = new CommandWebSocket('ws://localhost:8080/command');
     *     ws.on('command', function (e) {
     *         console.log(e.command, e.message);
     *         promise.resolveAndClose(); // also ws.resolveAndClose();
     *     });
     *     ws.done(function () {
     *         console.log('connection was closed');
     *     });
     *     ws.send('COMMAND', { foo : 'bar' });
     *
     * @class utils.websockets.Command
     * @extends utils.websockets.Basic
     * @param url {String} The WebSocket URL to connect to - **MUST** begin with `ws://` or `wss://`, see [utils.urls.templateWebSocket](utils.urls.html#method_templateWebSocket)
     * @param [options] {Object} Additional options (see the Properties tab)
     * @constructor
     */
    function CommandWebSocket(url, options) {
        BasicWebSocket.call(this, url, options);
    }

    CommandWebSocket.prototype = _.create(BasicWebSocket.prototype, {
        constructor : CommandWebSocket,

        /**
         * Sends a command to the server.
         * If `parseJSON` is true, calls `JSON.stringify` on the message before sending it across the wire.
         * @method send
         * @param command {String} The command to send to the server
         * @param message {*} The message to send to the server
         */
        send : function (command, message) {
            if (arguments.length <= 1) {
                this.rawSend(command);
            } else {
                this.rawSend(command + '\n' + (this.options.parseJSON ? JSON.stringify(message) : message));
            }
        },

        /**
         * Parses the raw data sent from the server.
         * If `parseJSON` is true, calls `JSON.parse` on the received message part of the command.
         * @method _parseData
         * @param newData {String} The data to parse
         * @protected
         * @return {{parsed:Boolean,value:*}} Whether parsing was successful and the command that was parsed
         */
        _parseData : function (newData) {
            var newLine = newData.indexOf('\n');
            if (newLine < 0) {
                return { parsed : false };
            }
            try {
                var valuePart = newData.substring(newLine + 1);
                return {
                    parsed : true,
                    value : {
                        command : newData.substring(0, newLine),
                        message : this.options.parseJSON ? JSON.parse(valuePart) : valuePart
                    }
                };
            } catch (e) {
                return { parsed : false };
            }
        },

        /**
         * Triggers events based on the received data.
         *
         * Triggers two additional events in addition to the base `message`.
         * See the Events tab for more info.
         *
         * @method _triggerMessage
         * @param parsedData {Object} The command parsed from the server
         * @param promiseFns {Object} Functions that can be used from within the callback
         * @param rawData {String} The raw data from the server
         * @private
         */
        _triggerMessage : function (parsedData, promiseFns /*, rawData */) {
            BasicWebSocket.prototype._triggerMessage.apply(this, arguments);
            if (parsedData.command) {
                /**
                 * Triggered when a command is received.
                 *
                 * The event object contains the following properties:
                 *
                 * - `command`: The command from the server
                 * - `message`: The message part of the command
                 * - `promise`: Functions that can be performed on the WebSocket (`resolve`, `resolveAndClose`, `reject`, `rejectAndClose`)
                 *
                 * @event command
                 */
                this.trigger('command', _.assign({ promise : promiseFns }, parsedData));
                /**
                 * Triggered when a specific command is received.
                 *
                 * For example, you can listen to the `LOAD-WBS` command with the following code:
                 *
                 *     ws.on('command-LOAD-WBS', function (e) { ... });
                 *
                 * The event object contains the following properties:
                 *
                 * - `message`: The message part of the command
                 * - `promise`: Functions that can be performed on the WebSocket (`resolve`, `resolveAndClose`, `reject`, `rejectAndClose`)
                 *
                 * @event command-*
                 */
                this.trigger('command-' + parsedData.command, {
                    message : parsedData.message,
                    promise : promiseFns
                });
            }
        }
    });

    return CommandWebSocket;
});

/**
 * @module utils
 */
define('primutils/utils/ajaws',[
    'lodash',
    'primutils/utils/websockets/Basic',
    'primutils/utils/websockets/Command'
], function (_, BasicWebSocket, CommandWebSocket) {
    'use strict';

    /**
     * A collection of WebSocket utility functions.
     *
     * Mainly used by the Primutil's store in its utilities namespace.
     *
     * @class utils.ajaws
     */
    return {

        /**
         * A utility function for creating a Primutils WebSocket wrapper.
         *
         * @method createWebSocket
         * @param type {String} The type of WebSocket to create ('basic' or 'command')
         * @param url {String} The WebSocket URL to connection to (**MUST** begin with `ws://` or `wss://`, see [utils.urls.templateWebSocket](utils.urls.html#method_templateWebSocket))
         * @param [options] {Object} Additional options
         * @returns {utils.websockets.BasicWebSocket|utils.websockets.CommandWebSocket}
         * @static
         */
        createWebSocket : function (type, url, options) {
            switch (type) {
            case 'basic':
                return new BasicWebSocket(url, options);
            case 'command':
                return new CommandWebSocket(url, options);
            default:
                throw new Error('Unrecognized WebSocket type');
            }
        },

        /**
         * Creates a WebSocket and sends an initial message.
         * Additional communication must be done through the `onMessageReceived` callback.
         *
         * This wrapper prevents the accidental passing of a live WebSocket so that all communication is contained
         * to where this method was called.
         *
         * Example:
         *
         *     var initialMessage = { test : 'message' };
         *     ajaws.basicWebSocket('ws://localhost:8080/test', initialMessage, function (message, args) {
         *         if (message.continue) {
         *             args.send({ next : 'message' }); // Send additional messages from within the onMessageReceived method only!
         *         } else if (message.done) {
         *             args.resolveAndClose(); // Marks the promise complete and closes the connection
         *                                     // Also resolve, resolveAndClose, reject, and rejectAndClose
         *                                     // Depending on how your server works, your server may close the connection
         *                                     //   automatically when there is no new data to send.
         *         }
         *     });
         *
         * @method basicWebSocket
         * @param url {String} The WebSocket URL to connection to (**MUST** begin with `ws://` or `wss://`, see [utils.urls.templateWebSocket](utils.urls.html#method_templateWebSocket))
         * @param initialMessage {*} The initial message to send to the server
         * @param onMessageReceived {Function} Called when a message is sent from the server, see the example above for additional information
         * @param [options] {Object} Additional options to pass to the created WebSocket
         * @returns {jQuery.promise}
         * @static
         */
        basicWebSocket : function (url, initialMessage, onMessageReceived, options) {
            var ws = new BasicWebSocket(url, options);
            ws.on('message', function (e) {
                onMessageReceived(e.data, _.assign({ send : ws.send.bind(ws) }, e.promise));
            });
            if (ws.isOpen()) {
                ws.send(initialMessage);
            } else {
                ws.on('open', function initialSend() {
                    ws.send(initialMessage);
                    ws.off('open', initialSend);
                });
            }
            return ws.promise();
        },

        /**
         * Creates a WebSocket and sends an initial command.
         * Additional communication must be done through the `commandsReceived` callbacks.
         *
         * This wrapper prevents the accidental passing of a live WebSocket so that all communication is contained
         * to where this method was called.
         *
         * Example:
         *
         *     var command = 'LOAD',
         *         message = { projectId : '7' };
         *     ajaws.basicWebSocket('ws://localhost:8080/load', command, message, {
         *         // List as many commands are the client responds to...
         *         'LOAD-WBS' : function (message, args) {
         *             // Store the data somehow
         *             if (message.needsAdditionalInfo) {
         *                 args.send({ next : 'message' }); // Send additional messages from within the commandsReceived methods only!
         *             }
         *         },
         *         'DONE' : function (message, args) {
         *             // Depending on how your server works, your server may close the connection
         *             //   automatically when there is no new data to send, but you can do so with...
         *             args.resolveAndClose(); // Marks the promise complete and closes the connection
         *                                     // Also resolve, resolveAndClose, reject, and rejectAndClose
         *         }
         *     });
         *
         * @method commandWebSocket
         * @param url {String} The WebSocket URL to connection to (**MUST** begin with `ws://` or `wss://`, see [utils.urls.templateWebSocket](utils.urls.html#method_templateWebSocket))
         * @param initialCommand {String} The initial command to send to the server
         * @param initialMessage {*} The initial message to send to the server
         * @param commandsReceived {Object} A collection of functions called when a command is received from the server, see the example above for additional information
         * @param [options] {Object} Additional options to pass to the created WebSocket
         * @returns {jQuery.promise}
         * @static
         */
        commandWebSocket : function (url, initialCommand, initialMessage, commandsReceived, options) {
            var ws = new CommandWebSocket(url, options);
            ws.on('command', function (e) {
                if (_.has(commandsReceived, e.command)) {
                    commandsReceived[e.command](e.message, _.assign({ send : ws.send.bind(ws) }, e.promise));
                }
            });
            if (ws.isOpen()) {
                ws.send(initialCommand, initialMessage);
            } else {
                ws.on('open', function initialSend() {
                    ws.send(initialCommand, initialMessage);
                    ws.off('open', initialSend);
                });
            }
            return ws.promise();
        },

        /**
         * Creates a WebSocket and sends an initial command.
         * Additional communication must be done through the `onMessageReceived` callback.
         *
         * A "batch command" WebSocket sends data like a CommandWebSocket, but receives data like a BasicWebSocket.
         * This is useful in situations where the client on sends one message to the server,
         * then the server sends several batches of the same result to the client and then disconnects.
         *
         * This wrapper prevents the accidental passing of a live WebSocket so that all communication is contained
         * to where this method was called.
         *
         * Example:
         *
         *     var command = 'LOAD',
         *         message = { projectId : '7' };
         *     ajaws.batchCommandWebSocket('ws://localhost:8080/load', command, message, function (data, args) {
         *         // Store your data somehow
         *     });
         *
         * @method batchCommandWebSocket
         * @param url {String} The WebSocket URL to connection to - warning, **MUST** begin with `ws://` or `wss://`, see [utils.urls.templateWebSocket](utils.urls.html#method_templateWebSocket)
         * @param initialCommand {String} The initial command to send to the server
         * @param initialMessage {*} The initial message to send to the server
         * @param onMessageReceived {Function} Called when a message is sent from the server, see the example above for additional information
         * @param [options] {Object} Additional options to pass to the created WebSocket
         * @returns {jQuery.promise}
         * @static
         */
        batchCommandWebSocket : function (url, initialCommand, initialMessage, onMessageReceived, options) {
            var ws = new CommandWebSocket(url, _.assign({
                parseData : BasicWebSocket.prototype._parseData
            }, options));
            ws.on('message', function (e) {
                onMessageReceived(e.data, _.assign({ send : ws.send.bind(ws) }, e.promise));
            });
            if (ws.isOpen()) {
                ws.send(initialCommand, initialMessage);
            } else {
                ws.on('open', function initialSend() {
                    ws.send(initialCommand, initialMessage);
                    ws.off('open', initialSend);
                });
            }
            return ws.promise();
        }

    };
});

/**
 * Store allows you to work with different models(objects) for your page.  Some of the key benefits of using the store are:
 * <ul>
 *     <li>Ordered keyed list.  Allows for item retrieval via key.</li>
 *     <li>Hold many different models.  EG collection of projects, collection of categories, object of meta data, etc.</li>
 *     <li>Original values are held after modification.  This gives an easy mechanism to revert changes.</li>
 *     <li>Supplies a pre-determined batch save object of any added, deleted and modified objects</li>
 *     <li>Triggers post events on the main operations</li>
 * </ul>
 *
 * *Information about Keys*
 *
 * The store only works with keys that are strings.
 * The reason behind this is that for many operations the key is used as the key in a JavaScript object hash.
 * This key could be strictly compared against your object's key, and if you are not using a string key this comparison will fail.
 *
 * *Updating Fields*
 *
 * Updating fields is done through the `.update` method.
 * You can update entire objects or navigate directly through to the path you want to update.
 * Paths are always separated by periods (.).
 * Avoid using complex keys (those needing bracket notation `foo['bar.bar']`) which contain periods.
 *
 *     .update('type', 'id', 'path.to.update', _value_);
 *
 * You should never update your object's key property directly.
 * The store uses several internal collections to maintain modification (also added/deleted) state of your object.
 * If you change the key of your object you will lose all of the modification information.
 * You should only update your object's key through either of the `.commit` or `.mergeAndCommit` methods.
 *
 * You should also avoid updating specific keys in arrays.
 * Always refresh the entire array instead of adding a new index.
 *
 * Example Object:
 *
 *     {
 *         'foo' : [ 1, 2, 3, 4 ]
 *     }
 *
 * Bad:
 *
 *     .update('type', 'id', 'foo.5', 5);
 *
 * Not Great (but would work):
 *
 *     .update('type', 'id', 'foo.1', 100);
 *
 * Best:
 *
 *     .update('type', 'id', 'foo', [ 100, 2, 3, 4, 5 ]);
 *
 * **What is Committing?**
 *
 * Committing your store indicates that your model is no longer dirty and that you no longer want to be able to revert
 * any of the changes that have already been made.
 * It is also a way of indicating that your model is synchronized with an outside source like a web server.
 *
 * There are two types of commit: plain committing and merge-and-committing.
 * Plain committing means that the previous history about your object is lost and the next time you reset your store
 * modifications will be restored to this point.
 * Merging and committing means that you would also like to take in updates from an outside source as part of your commit.
 * See information about the types of merging that the store supports in the `mergeMode` option.
 *
 * Both types of merging take a list of "new keys".
 * The new keys are a way of safely updating keys on new store objects to a new value from an outside source like a web server.
 * Generally when you add a new record, you give it a negative id to indicate it is new (see `_isItemNew`).
 * Passing new keys allows you to convert those new records to regular modifications on subsequent saves.
 *
 *     //Example
 *     {
 *         // Old Key    New Key
 *         "-1"        : "23",
 *         "-2"        : "24"
 *     }
 *
 * Any objects that do not have a corresponding new key will be left alone, but their modification state will be dropped.
 *
 * **About Resetting**
 *
 * Resetting allows your to return some or all of your objects back to their state before modification (i.e. since the last
 * call to `.commit` or `.add`.
 * Resetting new items will remove them from the store, and resetting deleted items will add them back to their last modified value.
 * There is essentially only one "checkpoint" which is created when you call `.commit`.
 * You can only reset back to the last commit state, so if you plan on resetting to some specific state make sure you call commit first.
 *
 * **Updating Store Options**
 *
 * The store does not provide a way to update its options because of the multitude of side-effects that most of the options have.
 * For example, changing the `key` option could affect unique key constraints and force reindexing on the store.
 * How would changing the options for flex and code values work? Do the old codes get reserialized, dropped, left alone?
 * Do the updated ones get auto-deserialized?
 *
 * The only place this is not enforced is for dates (see the `setDateFields` option).
 *
 * Example:
 *
 *      var META_ID = '1';
 *
 *      // returned server json
 *      var jsonResponse = {
 *          meta : {
 *              columns : {}
 *          },
 *          data : [{projectA}, {projectB}]
 *      };
 *
 *      var store = new Store({key : 'id'});
 *      store.add('projects', jsonResponse.data);
 *      store.add('meta', jsonResponse.meta, function(type, item){
 *          // add an id to the meta for any retrieval and updating
 *          item.id = META_ID;
 *          return item;
 *      });
 *
 *      store.on('update', function(args){
 *          if (this.isDirtyAll()){
 *              // store is dirty
 *          } else {
 *              // store is clean
 *          }
 *      });
 *
 *      // get objects
 *      var project = store.item('project', 'projectA');
 *      var columns = store.item('meta', META_ID).columns;
 *
 *      // modify a value
 *      store.update('project', 'projectA', 'code', 'prj-002');
 *
 *      // get save object
 *      // {
 *      //      added : [],
 *      //      deleted : [],
 *      //      modified : [{projectA}]
 *      // }
 *      var saveObject = store.saveAll();
 *
 *      // commit changes
 *      store.commitAll({
 *          storeTypeFoo: { // new IDs for the items in storeType 'storeTypeFoo'
 *              "-5": "100",
 *              "-6": "101"
 *          }
 *      });
 *
 *      // cancel changes
 *      store.resetAll();
 *
 *
 * @class store.Store
 * @module store
 * @param options Options for the store.
 * @constructor
 */

define('primutils/store/Store',[
    'lodash',
    'logger',
    'primutils/store/Collection',
    'primutils/store/ModifyCollection',
    'primutils/utils/Subscribable',
    'primutils/store/IndexManager',
    'moment',
    'primutils/utils/nestedPropertyAccessors',
    'primutils/utils/objectUtils',
    'primutils/utils/isEqual',
    'primutils/utils/ajax',
    'primutils/utils/ajaws',
    'primutils/utils/promises'
], function (_, logger, Collection, ModifyCollection, Subscribable, IndexManager, moment, propHelpers, objectUtils, isEqual, ajax, ajaws, promises) {
    'use strict';

    var hasOwnProperty = Object.prototype.hasOwnProperty;
    function hasOwnProp(obj, prop) {
        return hasOwnProperty.call(obj, prop);
    }

    function dateFieldsObj(dateFields) {
        if (_.isArray(dateFields)) {
            logger.info('converting date fields array to object');
            return _.reduce(dateFields, function (memo, dateField) {
                memo[dateField] = DATE;
                return memo;
            }, {});
        }
        return dateFields;
    }

    var MERGE_MODES = {
        NONE : 'none',
        REPLACE : 'replace',
        EXTEND_EXISTING : 'extendExisting',
        EXTEND_EXISTING_SHALLOW : 'extendExistingShallow',
        EXTEND_NEW : 'extendNew',
        EXTEND_NEW_SHALLOW : 'extendNewShallow'
    };
    MERGE_MODES.default = MERGE_MODES.EXTEND_EXISTING;

    var DATE = 'date',
        TZ_DATE = 'tzDate';

    var DEFAULT_OPTIONS = {
        /**
         * Field used to index the object
         *
         * @property key
         * @type String
         * @optional
         * @default 'id'
         */
        key : 'id',
        /**
         * Types that cannot be changed through the store
         *
         * @property readOnly
         * @type Array(String)
         * @optional
         */
        readOnly : [],
        /**
         * Field used in specific types to index the object
         *
         * @property typeKeys
         * @type Map(String,String)
         * @Optional
         */
        typeKeys : {},
        /**
         * Override options per type
         *
         * Only some options are supported: `key`, `dateFields`, `recursiveSerialization`, `flex*Key` (e.g. `flexOtherKey`),
         * `flex*Search` (e.g. `flexTestSearch`), `codeKey`, `codePrefix`, `codeSearch`, `cascadeDelete`
         *
         * @property typeOptions
         * @type Map(String, Object)
         * @Optional
         */
        typeOptions : {},
        /**
         * Date formatter to be used to unserialize and serialize dates
         *
         * Must provide two methods:
         *  * `unserialize` - convert string from server into MomentJS
         *  * `serialize` - convert MomentJS into string for server
         *
         * @property dateFormatter
         * @type Object
         */
        dateFormatter : {
            unserialize : function (value) {
                return value;
            },
            serialize : function (value) {
                return value;
            }
        },
        /**
         * Date formatter to be used to unserialize and serialize timezone-sensitive dates
         *
         * Must provide two methods:
         *  * `unserialize` - convert string from server into MomentJS
         *  * `serialize` - convert MomentJS into string for server
         *
         * @property timezoneSensitiveDateFormatter
         * @type Object
         */
        timezoneSensitiveDateFormatter : {
            unserialize : function (value) {
                return value;
            },
            serialize : function (value) {
                return value;
            }
        },

        /**
         * Fields which are automatically deserialized into moment objects on objects
         *
         * The object is a key-value pair of field name to serialization type.
         * The default type is **non**-timezone-sensitive ('date').
         *
         * If an array is passed, it is automatically converted to an object with all the values being the default.
         *
         *     {
         *         "standardField"    : "date",
         *         "tzSensitiveField" : "tzDate"
         *     }
         *
         * @property dateFields
         * @type Object|Array
         */
        dateFields : {},
        /**
         * Recursively do serialization and deserialization
         *
         * Recursive serialization may be useful if you have nested ROs (Rest Objects) coming back from the server.
         *
         * For example, a Scope Item Assignment has its own information as well as the Scope Item itself.
         *
         *     {
         *         "plannedUnits" : 30,
         *
         *         // Scope Item Assignment Flex Fields, Codes, Dates
         *
         *         "scopeItem" : {
         *             "name" : "Scope Item 1",
         *             "availableUnits" : 40,
         *
         *             // Scope Item Flex Fields, Codes, Dates
         *         }
         *     }
         *
         * By default, the store will not attempt to deserialize the scope item's flex fields (code values, date fields) for you.
         * In the past, you had to use the `modifyItemFn` to do subitem deserialization.
         *
         *     store.addAll('scopeItemAssignments', [...], function (item) {
         *         store._deserialize('scopeItemAssignments', item.scopeItem);
         *     }, dateFields);
         *
         * Setting this property to true means you no longer have to do this.
         *
         * @property recursiveSerialization
         * @type Boolean
         */
        recursiveSerialization : false,
        /**
         * How the merge methods work:
         *
         * * Store.mergeModes.NONE : Only add the item to the store if a value with that key (or corresponding "old key") does not already exist. This is the legacy
         * behavior.
         * * Store.mergeModes.REPLACE : Replace the object currently in the store with that from the server.
         * * Store.mergeModes.EXTEND_EXISTING : (Default) Do a deep extends with the current store object as the target. Values from the merge object will take
         * precedence over those currently in the store.
         * * Store.mergeModes.EXTEND_EXISTING_SHALLOW : Do an extends with the current store object as the target.
         * * Store.mergeModes.EXTEND_NEW : Do a deep extends with the merge object as the target. Values from the store object will take precedence over those from
         * the merge object.
         * * Store.mergeModes.EXTEND_NEW_SHALLOW : Do an extends with the merge object as the target.
         *
         * @property mergeMode
         * @type {String}
         * @default EXTEND_EXISTING
         */
        mergeMode : MERGE_MODES.default,

        /**
         * Defines different ways to automatically index the data in the store.
         * The [introductory blog post](http://cegbu.oraclecorp.com/confluence/display/PRIMEENGINEERING/2015/04/07/Faster+child+record+lookup+in+Primutils+Store)
         * may also interest you.
         *
         * Options include parent-child relationships (called 'foreign'), uniqueness constraints ('unique'), and grouped items ('group').
         *
         * The `indices` option is structured like this:
         *
         *     indices : {
         *         foreign : [],
         *         unique : [],
         *         group : []
         *     }
         *
         * Add your particular relationship to the appropriate collection.
         *
         * *NOTE*: You should only index items by something which has a unique value when coerced to a string.
         * The types of both `primaryId` and `foreignId` in parent-child relationships should always be a string to match the type of the item key.
         * Unique and grouped indices should be number or string fields.
         * Other types will give potentially erratic results.
         *
         * Parent-Child ('foreign', Foreign Key constraints):
         *
         *     {
         *         primaryType : 'parent',
         *         primaryId : 'id', // Optional -- will default to the type key of the parent type
         *         foreignType : 'child',
         *         foreignId : 'parentId' // Which field on the child gets us back to the parent
         *     }
         *
         * Uniqueness ('unique'):
         *
         *     {
         *         uniqueType : 'type',
         *         uniqueField : 'alternateId'
         *     }
         *
         * Groups of Items ('group'):
         *
         *     {
         *         groupType : 'type',
         *         groupField : 'groupId'
         *     }
         *
         * @property indices
         * @type {Object}
         * @optional
         */
        indices : {},
        /**
         * Indicates whether deleting a parent item will also delete its child items.
         *
         * You can specify an array of child types that will be deleted instead of all.
         * If `cascadeDelete` is set to `true`, it will delete all child types based on the indices.
         *
         * @property cascadeDelete
         * @type {Boolean|Array}
         * @default true
         */
        cascadeDelete : true,

        /**
         * If this option is set to true, special handling of udfs is disabled - a performance boost - and it is assumed that these relations are simply stored as
         * deep objects in the store, if at all.
         *
         * See https://confluence.oraclecorp.com/confluence/display/PRIMEENGINEERING/Feature+Columns+Metadata for more information.
         *
         * @property ignoreFlex
         * @type Boolean
         * @default false
         */
        ignoreFlex : false,
        /**
         * Key under which other flex fields live
         *
         * @property flexOtherKey
         * @type String
         * @default 'flex'
         */
        flexOtherKey : 'flex',
        /**
         * Key under which text flex fields live
         *
         * @property flexTextKey
         * @type String
         * @default 'textFlex'
         */
        flexTextKey : 'textFlex',
        /**
         * Key under which cost flex fields live
         *
         * @property flexCostKey
         * @type String
         * @default 'costFlex'
         */
        flexCostKey : 'costFlex',

        /**
         * Regular expression for determining whether a field is an other flex field
         *
         * @property flexOtherSearch
         * @type RegExp
         * @default /_FO-/
         */
        flexOtherSearch : /_FO-/,
        /**
         * Regular expression for determining whether a field is a text flex field
         *
         * @property flexTextSearch
         * @type RegExp
         * @default /_FT-/
         */
        flexTextSearch : /_FT-/,
        /**
         * Regular expression for determining whether a field is a cost flex field
         *
         * @property flexCostSearch
         * @type RegExp
         * @default /_FC-/
         */
        flexCostSearch : /_FC-/,

        /**
         * If this option is set to true, special handling of Codes is disabled - a performance boost - and it is assumed that these relations are simply stored as
         * deep objects in the store, if at all.
         *
         * See https://confluence.oraclecorp.com/confluence/display/PRIMEENGINEERING/Feature+Columns+Metadata for more information.
         *
         * @property ignoreCodes
         * @type Boolean
         * @default false
         */
        ignoreCodes : false,
        /**
         * Key under which object code values live
         *
         * @property codeKey
         * @type String
         * @default 'projectCategories'
         */
        codeKey : 'projectCategories',
        /**
         * Prefix applied when deserializing object code values
         *
         * @property codePrefix
         * @type String
         * @default 'CO_CODE_TYPE-'
         */
        codePrefix : 'CO_CODE_TYPE-',

        /**
         * This option is an optional performance optimization for unserialization performance on your page.
         *
         * If this option is set to true, it will:
         * * Create a cache, mapping date strings sent by the server to their Moment.js representations
         * * For all unserialized dates from the server, it will first read from this cache - if the date was already parsed, it will use the same moment instance, or populate the cache otherwise
         *
         * So, in effect, if "Act1" and "Act2" both have the same start/end times, rather than the default behavior (populating the objects with equivalent moments that are not the same instance),
         *  both activities would have the same start/end moment instances.
         *
         * *WARNING*: this option is not defaulted to true because it may not be compatible with all pages.
         * You are responsible for ensuring that no operations are done on these dates without cloning them first in any page-specific logic.
         * (If you identify any issues in our core libraries, do file those as core requests).
         *
         * @property cacheDates
         * @type Boolean
         * @default false
         */
        cacheDates : false,

        asyncAddThreshold : 5000,
        asyncAddDelay : 25,

        /**
         * Function to use when comparing an updated value with the original value to determine if a value was updated.
         *
         * By default it will use a modified version of Lodash's `isEqual` method which also supports Moment objects.
         *
         * If this function returns `true`, then the update method will be aborted, the value on the item will not be modified,
         * and the `'update'` event will *not* be triggered.
         * Compare this to the `isItemModified` option, which will always trigger the `'update'` event.
         *
         * In this function, you will have full information about the store type, item, path, old and new value are.
         *
         *     comparisonFunction: function (newValue, oldValue, type, item, path) {
         *         return newValue.foo === oldValue.foo;
         *     }
         *
         * See also the [`fieldProcessors`](#property_fieldProcessors) option.
         *
         * @property comparisonFunction
         * @type Function
         * @optional
         * @default _.isEqual with enhancements for Moment objects
         */
        comparisonFunction : isEqual,

        /**
         * Callback to determine if the item being updated should be marked as modified.
         *
         * If this function returns `true`, the change will be entered in the Store's modified values.
         * If `false`, the value will still be updated on the item, but it will be treated as if `preventDirty` had been set when calling update.
         * See information in the `update` method about the `preventDirty` parameter.
         * Compare this to the `comparisonFunction` option, which would not trigger the `'update'` event.
         *
         * In this function, you will have full information about the store type, item, path, old and new value are.
         *
         *     isItemModified: function (newValue, oldValue, type, item, path) {
         *         if (type === 'my_type' && path === 'someField') {
         *             return (newValue.foo === 'apple' && oldValue.foo === 'orange');
         *         }
         *         return true;
         *     }
         *
         * See also the [`fieldProcessors`](#property_fieldProcessors) option.
         *
         * @property isItemModified
         * @type Function
         * @optional
         * @default Always returns true
         */
        isItemModified : _.constant(true),

        /**
         * A structure of functions for processing individual fields.
         * Organized by type and field, the following processors are supported:
         *
         * * `unserialize`: can modify the value before first loading in the store in `add`/`addAll`/etc
         *   - signature: `function (value, type, item, field)`
         * * `serialize`: can modify the value before transforming to object in `save`/`saveAll`/etc
         *   - signature: `function (value, type, item, field)`
         * * `update`: can modify the value being updated, before any processing by `comparison`, before-update rules, etc
         *   - signature: `function (value, type, item, field)`
         * * `isItemModified`: overrides the [`isItemModified`](#property_isItemModified) function
         * * `comparison`: overrides the [`comparisonFunction`](#property_comparison) function
         *
         * Example:
         *
         *     fieldProcessors: {
         *         typefoo: {
         *             fieldbar: {
         *                 unserialize: function (value) {
         *                     return {field: value};
         *                 },
         *                 // ...
         *             },
         *             'deep.field': {
         *                 comparison: function (newValue, oldValue, type, item, path) {
         *                     return _.isEqual(_.pick(newValue, 'id', 'name'), _.pick(oldValue, 'id', 'name'));
         *                 }
         *             },
         *         }
         *     }
         *
         * @property fieldProcessors
         * @type Object
         * @optional
         */
        fieldProcessors : {}
    };

    var PATH_DELIMITER = '.';
    var STUB = 'DO-NOT-DELETE';
    var UNSERIALIZED_MARKER = '$unserialized$';

    var getTypedOption = function (type, option) {
        var value = this.options.typeOptions[type] && this.options.typeOptions[type][option];
        return value == null ? this.options[option] : value;
    };

    var _unserialize = function (type, item, dateFields, recursive) {
        var that = this,
            flexCostKey = getTypedOption.call(this, type, 'flexCostKey'),
            flexTextKey = getTypedOption.call(this, type, 'flexTextKey'),
            flexKey = getTypedOption.call(this, type, 'flexOtherKey'),
            projectCategoryKey = getTypedOption.call(this, type, 'codeKey'),
            projectCategoryPrefix = getTypedOption.call(this, type, 'codePrefix'),
            processors = that.utilities._getValue(this, 'options.fieldProcessors.' + type),
            recursiveFields;

        function fieldsOfFlexType(item, flexType) {
            return item[flexType] != null ? _.keys(item[flexType]) : [];
        }

        function hasUnserialzedFlex(item, flexType) {
            return item[flexType] != null && item[flexType][UNSERIALIZED_MARKER] === true;
        }

        // Unserialize flex fields
        if (!this.options.ignoreFlex) {
            _.forEach([flexKey, flexCostKey, flexTextKey], function unserializeFlexFields(flexType) {
                if (!hasUnserialzedFlex(item, flexType)) {
                    _.forEach(fieldsOfFlexType(item, flexType), function (fieldName) {
                        item[fieldName] = item[flexType][fieldName];
                        item[flexType][fieldName] = STUB;
                    });
                }

                if (item[flexType] != null) {
                    item[flexType][UNSERIALIZED_MARKER] = true;
                }
            });
        }

        _.forEach(processors, function (proc, field) {
            var unserializeProc = proc.unserialize;
            if (unserializeProc) {
                var orig = that.utilities._getValue(item, field);
                var swapped = unserializeProc(orig, type, item, field);
                that.utilities._setValue(item, field, swapped);
            }
        });

        if (!this.options.ignoreCodes) {
            // Unserialize project category fields
            if (hasOwnProp(item, projectCategoryKey) && item[projectCategoryKey][UNSERIALIZED_MARKER] !== true) {
                _.forEach(item[projectCategoryKey], function (projectCategory, typeId) {
                    item[projectCategoryPrefix + typeId] = projectCategory;
                    item[projectCategoryKey][typeId] = STUB;
                });
                item[projectCategoryKey][UNSERIALIZED_MARKER] = true;
            }
        }

        // Unserialize date fields
        var dateFormatter = that.options.dateFormatter,
            tzDateFormatter = that.options.timezoneSensitiveDateFormatter;
        function unserializeDateFieldDirect(fieldType, dateField) {
            var value = that.utilities._getValue(item, dateField);
            if (value != null && typeof value === 'string') {
                that.utilities._setValue(item, dateField, fieldType === TZ_DATE ? tzDateFormatter.unserialize(value) : dateFormatter.unserialize(value));
            }
        }
        var dateCache = that._dateCache;
        function unserializeDateFieldCached(fieldType, dateField) {
            var value = that.utilities._getValue(item, dateField);
            if (value != null && typeof value === 'string') {
                var date;
                if (dateCache[value]) {
                    date = dateCache[value];
                } else {
                    date = fieldType === TZ_DATE ? tzDateFormatter.unserialize(value) : dateFormatter.unserialize(value);
                    dateCache[value] = date;
                }
                that.utilities._setValue(item, dateField, date);
            }
        }
        _.forEach(dateFields, that.options.cacheDates ? unserializeDateFieldCached : unserializeDateFieldDirect);

        // Do recursion
        if (recursive) {
            recursiveFields = _.difference(_.keys(item), _.keys(dateFields), [flexCostKey, flexTextKey, flexKey, projectCategoryKey]);
            _.forEach(recursiveFields, function unserializeRecursively(fieldName) {
                if (_.isObject(item[fieldName])) {
                    if (_.isArray(item[fieldName])) {
                        item[fieldName] = _.map(item[fieldName], function (value) {
                            return _unserialize.call(that, type, value, dateFields, recursive);
                        });
                    } else {
                        item[fieldName] = _unserialize.call(that, type, item[fieldName], dateFields, recursive);
                    }
                }
            });
        }

        return item;
    };

    var _serialize = function (type, item, recursive) {
        var that = this,
            flexKey = getTypedOption.call(this, type, 'flexOtherKey'),
            flexCostKey = getTypedOption.call(this, type, 'flexCostKey'),
            flexTextKey = getTypedOption.call(this, type, 'flexTextKey'),
            projectCategoryKey = getTypedOption.call(this, type, 'codeKey'),
            processors = that.utilities._getValue(this, 'options.fieldProcessors.' + type),
            dateFields = getDateFields.call(this, type),
            dateFormatter = that.options.dateFormatter,
            tzDateFormatter = that.options.timezoneSensitiveDateFormatter,
            flexFields, projectCategoryFields, recursiveFields;


        if (!this.options.ignoreFlex) {
            flexFields = _.chain(item).keys().groupBy(function (fieldName) {
                if (isFlexField.call(that, type, fieldName)) {
                    return flexKey;
                } else if (isFlexCostField.call(that, type, fieldName)) {
                    return flexCostKey;
                } else if (isFlexTextField.call(that, type, fieldName)) {
                    return flexTextKey;
                }
                return '__private_not_flex__';
            })
                .omit('__private_not_flex__')
                .value();
        }

        if (!this.options.ignoreCodes) {
            projectCategoryFields = _.filter(_.keys(item), function (fieldName) {
                return isProjectCategoryField.call(that, type, fieldName);
            });
        }

        recursiveFields = recursive ? _.difference(_.keys(item), [flexCostKey, flexTextKey, flexKey, projectCategoryKey, flexFields, projectCategoryFields]) : [];

        _.forEach(processors, function (proc, field) {
            var serializeProc = proc.serialize;
            if (serializeProc) {
                var orig = that.utilities._getValue(item, field);
                var swapped = serializeProc(orig, type, item, field);
                that.utilities._setValue(item, field, swapped);
            }
        });

        // Do known dates first so we can do semi-nested fields like "foo.bar" without having to futz with recursion
        _.forEach(dateFields, function (fieldType, dateField) {
            var value = that.utilities._getValue(item, dateField);
            if (value != null && moment.isMoment(value)) {
                that.utilities._setValue(item, dateField, fieldType === TZ_DATE ? tzDateFormatter.serialize(value) : dateFormatter.serialize(value));
            }
        });

        // Serialize date fields and do recursion
        if (recursive) {
            _.forEach(recursiveFields, function (fieldName) {
                if (moment.isMoment(item[fieldName])) {
                    item[fieldName] = (dateFields[fieldName] === TZ_DATE) ? tzDateFormatter.serialize(item[fieldName]) : dateFormatter.serialize(item[fieldName]);
                } else if (_.isObject(item[fieldName])) {
                    if (_.isArray(item[fieldName])) {
                        item[fieldName] = _.map(item[fieldName], function (value) {
                            return _serialize.call(that, type, value, recursive);
                        });
                    } else {
                        item[fieldName] = _serialize.call(that, type, item[fieldName], recursive);
                    }
                }
            });
        } else { // Or just date fields
            _.forEach(item, function (value, fieldName) {
                if (moment.isMoment(value)) {
                    item[fieldName] = (dateFields[fieldName] === TZ_DATE) ? tzDateFormatter.serialize(item[fieldName]) : dateFormatter.serialize(item[fieldName]);
                }
            });
        }

        if (!this.options.ignoreFlex) {
            // Serialize flex fields
            _.forEach(flexFields, function (fieldNames, flexType) {
                _.forEach(fieldNames, function (fieldName) {
                    item[flexType] = item[flexType] || {};
                    item[flexType][fieldName] = item[fieldName];
                    delete item[fieldName];
                });
            });

            if (item[flexKey]) {
                delete item[flexKey][UNSERIALIZED_MARKER];
            }
            if (item[flexCostKey]) {
                delete item[flexCostKey][UNSERIALIZED_MARKER];
            }
            if (item[flexTextKey]) {
                delete item[flexTextKey][UNSERIALIZED_MARKER];
            }
        }

        if (!this.options.ignoreCodes) {
            // Serialize project categories
            if (projectCategoryFields.length > 0) {
                item[projectCategoryKey] = item[projectCategoryKey] || {};

                _.forEach(projectCategoryFields, function (fieldName) {
                    var projectCategory = item[fieldName];

                    if (projectCategory != null && hasOwnProp(projectCategory, 'typeId')) {
                        item[projectCategoryKey][projectCategory.typeId] = projectCategory;
                    }

                    delete item[fieldName];
                });
            }

            if (item[projectCategoryKey]) {
                // Filter out all of the projectCategories that were deleted from the object
                item[projectCategoryKey] = _.reduce(item[projectCategoryKey], function (result, catVal, typeId) {
                    if (catVal !== STUB) {
                        result[typeId] = catVal;
                    }
                    return result;
                }, {});

                delete item[projectCategoryKey][UNSERIALIZED_MARKER];
            }
        }

        return item;
    };

    var _remove = function (type, item, key, silent, removeMethod) {
        var that = this;
        this._changeStack.push({
            changetype : 'remove',
            type : type,
            key : key
        });

        if (!silent) {
            /**
             * Triggered before an item is removed from the store.
             * This event is triggered before any cascade deletes.
             *
             * Args:
             *  * type: The type of the removed item
             *  * key: The key of the item removed
             *  * preventRemove: Function to call to prevent the delete from occurring
             *  * isRemovePrevented: Function to check if the remove has already been prevented by another listener
             *
             * @event before-remove
             * @param args {Object}
             */
            var preventRemove = false;
            var beforeRemoveArgs = {
                type : type,
                key : key,
                changeStack : this._changeStack,
                preventRemove : function () {
                    preventRemove = true;
                },
                isRemovePrevented : function () {
                    return preventRemove;
                }
            };
            this._subscribable.trigger('before-remove', beforeRemoveArgs);
            if (beforeRemoveArgs.isRemovePrevented()) {
                this._changeStack.pop();
                return item;
            }
        }

        var cascadeDeletes = getTypedOption.call(this, type, 'cascadeDelete');
        if (cascadeDeletes) {
            var typesToDelete = (cascadeDeletes === true) ? this.indexManager.getChildTypes(type) : _.intersection(cascadeDeletes, this.indexManager.getChildTypes(type));
            var removedTypeKey = that._getTypeKey(type);
            _.forEach(typesToDelete, function (childType) {
                var childTypeKey = that.indexManager.indexDefinitions.parentToChild[type][childType];
                var id = (removedTypeKey === childTypeKey) ? key : that.utilities._getValue(item, childTypeKey);
                _.forEach(that.indexManager.getChildIds(type, id, childType), function (childId) {
                    that.remove(childType, childId, silent);
                });
            });
        }

        // if the item is not new then add to deleted store
        if (!this._isItemNew(type, item) && !_.includes(this.options.readOnly, type)) {
            this.reset(type, key, true);
            this.deleteStoreCollection.add(type, item);
        }

        this.indexManager.itemRemoved(type, key, item);

        // remove item from new store
        this.newStoreCollection[removeMethod](type, key);

        // remove item from modify store
        this.modifyStoreCollection[removeMethod](type, key, null, true);

        this.storeCollection[removeMethod](type, key);

        if (!silent) {
            /**
             * Triggered when an item is removed from the store
             *
             * Args:
             *  * type: The type of the removed item
             *  * key: The key of the item removed
             *  * item: The item that was removed
             *
             * @event remove
             * @param args {Object}
             */
            this._subscribable.trigger('remove', {
                type : type,
                key : key,
                item : item,
                changeStack : this._changeStack
            });
        }

        this._changeStack.pop();

        return item;
    };

    var isFlexField = function (type, fieldName) {
        return fieldName.search(getTypedOption.call(this, type, 'flexOtherSearch')) > -1;
    };

    var isFlexCostField = function (type, fieldName) {
        return fieldName.search(getTypedOption.call(this, type, 'flexCostSearch')) > -1;
    };

    var isFlexTextField = function (type, fieldName) {
        return fieldName.search(getTypedOption.call(this, type, 'flexTextSearch')) > -1;
    };

    var isProjectCategoryField = function (type, fieldName) {
        return fieldName.search(getTypedOption.call(this, type, 'codeSearch')) > -1;
    };

    var getDateFields = function (type) {
        if (this._dateFieldsCache[type] == null) {
            var global = this.options.dateFields || {},
                typed = getTypedOption.call(this, type, 'dateFields') || {};
            this._dateFieldsCache[type] = _.assign({}, global, typed);
        }
        return this._dateFieldsCache[type];
    };

    var markItemModified = function (type, key, field, newValue, originalItem) {
        if (!this._isItemNew(type, originalItem)) {
            this.modifyStoreCollection.doMarkModified(type, key, field, newValue, originalItem);
        }
    };

    var Store = function (options) {
        var that = this,
            storeCollectionOptions = {
                getKey : function (type) {
                    return that._getTypeKey(type);
                }
            };

        this.options = objectUtils.extendDeep({}, DEFAULT_OPTIONS, options);
        this.setDateFields(this.options.dateFields);
        _.forEach(this.options.typeOptions, function (opts) {
            if (opts.dateFields) {
                opts.dateFields = dateFieldsObj(opts.dateFields);
            }
        });
        if (this.options.cacheDates) {
            this._dateCache = {};
        }
        this.options.codeSearch = new RegExp('^' + this.options.codePrefix);

        // event handling
        this._subscribable = new Subscribable();
        this.storeCollection = new Collection('store', storeCollectionOptions);
        this.modifyStoreCollection = new ModifyCollection('modify', storeCollectionOptions);
        this.deleteStoreCollection = new Collection('delete', storeCollectionOptions);
        this.newStoreCollection = new Collection('new', storeCollectionOptions);
        this.tempStoreCollection = new Collection('temp', storeCollectionOptions);
        this._changeStack = [];
        if (!_.isEmpty(this.options.indices)) {
            this.indexManager = new IndexManager(this.options.indices, _.assign({ utilities : this.utilities }, storeCollectionOptions));
        } else {
            this.indexManager = IndexManager.noopIndexManager();
        }
    };

    Store.prototype = {
        constructor : Store,

        /**
         * COLLECTION OF METHODS
         *
         * Utilities is a collection of helpful methods when using the store.
         *
         * **Field Helpers**
         * - _getValue - See [utils.nestedPropertyAccessors.getValue](utils.nestedPropertyAccessors.html#method_getValue)
         * - _setValue - See [utils.nestedPropertyAccessors.setValue](utils.nestedPropertyAccessors.html#method_setValue)
         *
         * **AJAX Helpers**
         *
         * The utilities object contains all of the methods of the [AJAX Utilities](utils.ajax.html#methods).
         *
         * **WebSocket Helpers**
         *
         * The utilities object contains all of the methods of the [WebSockets Utilities](utils.ajaws.html#methods).
         *
         * @method utilities
         */
        utilities : _.assign({
            _getValue : function (item, field) { return propHelpers.getValue(item, field, PATH_DELIMITER); },
            _setValue : function (item, field, value) { return propHelpers.setValue(item, field, value, PATH_DELIMITER); }
            // consider someday, but for now, slower
            // _getValue : function (item, field) { return _.get(item, field); },
            // _setValue : function (item, field, value) { return _.set(item, field, value); }
        }, ajax, ajaws),

        /**
         * Unserializes fields in a object that can be used client-side
         *
         * Does all unserializations: date, flex objects, code values
         *
         * @param type {String} Store type being unserialized
         * @param item {Object} Object being unserialized
         * @param dateFields {Object|Array} Date fields to convert to moment objects (For format, see `dateFields` option)
         * @param modifyItemFn {Function} (type, object) -> object: Additional modifications to perform after unserialization
         * @private
         */
        _unserialize : function (type, item, dateFields, modifyItemFn) {
            var recursive = getTypedOption.call(this, type, 'recursiveSerialization');

            dateFields = dateFieldsObj(dateFields);

            if (_.isFunction(modifyItemFn)) {
                var newItem = modifyItemFn(type, item);
                if (newItem != null) {
                    item = newItem;
                }
            }

            _unserialize.call(this, type, item, dateFields, recursive);
        },

        /**
         * Serializes fields in a object
         *
         * Does all serializations: date, flex objects, code values
         *
         * @param type {String} Store type being serialized
         * @param item {Object} Object being serialized
         * @param serializeFn {Function} (type, object) -> object: Additional modifications to perform after serialization
         * @private
         */
        _serialize : function (type, item, serializeFn) {
            var recursive = getTypedOption.call(this, type, 'recursiveSerialization');
            _serialize.call(this, type, item, recursive);
            if (_.isFunction(serializeFn)) {
                serializeFn(type, item);
            }
        },

        /**
         * Determines if the item belongs in the new item collection
         *
         * The default behavior is to treat all items with keys that parse to negative numbers as new
         *
         * -1 : New
         * 1  : Not new
         *
         * @method _isItemNew
         * @param {String} type
         * @param {Object} item
         * @return Boolean
         * @private
         */
        _isItemNew : function (type, item) {
                        if (item == null) {
                throw new Error('store._isItemNew() - item undefined');
            }
            
            var id = String(this._itemKey(type, item));
            return id[0] === '-';
        },

        /**
         * Return item key property
         *
         * @method _itemKey
         * @param {String} type
         * @param {Object} item
         * @return String
         * @private
         */
        _itemKey : function (type, item) {
                        if (item == null) {
                throw new Error('store._itemKey() - item undefined');
            }
            
            var key = this._getTypeKey(type);
            return item[key];
        },

        /**
         * Gets the key property for a given type.
         *
         * @method _getTypeKey
         * @param {String} type
         * @return {String} the key for the given type
         * @protected
         */
        _getTypeKey : function (type) {
            // Can be used by subclasses, like TimePhaseStore/DiscussionStore
            var value = this.options.typeOptions[type] && this.options.typeOptions[type].key;
            if (value != null) {
                return value;
            }
            value = this.options.typeKeys[type];
            if (value != null) {
                return value;
            }
            return this.options.key;
        },

        /**
         * Return whether the store contains any objects of the given type
         *
         * @method isEmpty
         * @param {String} type
         * @return Boolean
         */
        isEmpty : function (type) {
            return this.storeCollection.isEmpty(type);
        },

        /**
         * Return whether the store is completely empty
         *
         * @method isEmptyAll
         * @return Boolean
         */
        isEmptyAll : function () {
            return this.storeCollection.isEmptyAll();
        },

        /**
         * Return the amount of items for a given type
         *
         * @method count
         * @param {String} type
         * @return Integer
         */
        count : function (type) {
            return this.storeCollection.count(type);
        },

        /**
         * Add an array of items to the store to a given type
         *
         * <b>Note</b> - If the item key property meets the `_isItemNew` criteria it is treated as new
         *
         * Triggers the `add` event for each item added to the store.
         * Triggers the `before-add-all` event before all items are added to the store.
         * Triggers the `add-all` after items are added to the store event.
         *
         * Example:
         *
         *      store.addAll('type', [{}, {}], function(type, item){
         *          item.foo = 'bar';   // add another property
         *          return item;
         *      });
         *
         * @method addAll
         * @param {String} type
         * @param {Array} items
         * @param {Function} [modifyItemFn] (type, object) -> object: Additional modifications to perform after unserialization
         * @param {Object|Array} [dateFields] Additional date fields to deserialize (For format, see `dateFields` option)
         */
        addAll : function (type, items, modifyItemFn, dateFields) {
            var that = this;

            /**
             * Triggered before the loop adding all items to the store.
             *
             * Args:
             * * type: the type of which all items were added
             * * items: the items which will be added
             *
             * *WARNING*: items are not guaranteed to be added synchronously - in the event a user calls `addAllAsync` as opposed to `addAll`, the add-all will
             * happen asynchronously and would resolve some time later.
             *
             * @event before-add-all
             */
            var trig = {
                type : type,
                items : items
            };
            this._subscribable.trigger('pre-add-all', trig);
            this._subscribable.trigger('before-add-all', trig);

            _.forEach(items, function addAllLoop(item) {
                that.add(type, item, modifyItemFn, dateFields);
            });

            /**
             * Triggered by the store when bulk adds are performed.
             *
             * Args:
             * * type: the type of which all items were added
             * * items: the items which will be added
             *
             * @event add-all
             * @since 2.1.0
             */
            this._subscribable.trigger('add-all', {
                type : type,
                items : items
            });
        },

        /**
         * Add an array of items to the store to a given type.
         * This method is like [`addAll`](#method_addAll), but unlike `addAll`, will always process added items asynchronously.
         * See [PRIMCR-1637](https://jira.oraclecorp.com/jira/browse/PRIMCR-1637) for information about why this might be desirable.
         *
         * Triggers the same events as `addAll`.
         *
         * Example:
         *
         *      store.addAllAsync('type', [{}, {}], function(type, item){
         *          item.foo = 'bar';   // add another property
         *          return item;
         *      }).then(function () {
         *          // do things with the added items
         *      });
         *
         * @method addAllAsync
         * @param {String} type
         * @param {Array} items
         * @param {Function} [modifyItemFn] (type, object) -> object: Additional modifications to perform after unserialization
         * @param {Object|Array} [dateFields] Additional date fields to deserialize (For format, see `dateFields` option)
         */
        addAllAsync : function (type, items, modifyItemFn, dateFields) {
            var result = promises.new();
            var that = this;
            var currentIdx = 0;
            var options = this.options;

            var trig = {
                type : type,
                items : items
            };
            this._subscribable.trigger('pre-add-all', trig);
            this._subscribable.trigger('before-add-all', trig);

            function scheduleNextGroup() {
                if (currentIdx >= items.length) {
                    result.resolve();
                    that._subscribable.trigger('add-all', {
                        type : type,
                        items : items
                    });
                } else {
                    _.defer(addNextGroup, options.asyncAddDelay);
                }
            }
            function addNextGroup() {
                var iterationFinish = Math.min(items.length, currentIdx + options.asyncAddThreshold);
                for (; currentIdx < iterationFinish; currentIdx++) {
                    var item = items[currentIdx];
                    that.add(type, item, modifyItemFn, dateFields);
                }
                scheduleNextGroup();
            }
            _.delay(scheduleNextGroup);

            return result.promise();
        },

        /**
         * Add item to store for a given type
         *
         * <b>Note</b> - If the item key property meets the `_isItemNew` criteria it is treated as new
         *
         * Triggers Event - 'add'
         *
         * @method add
         * @param {String} type
         * @param {Object} item
         * @param {Function} [modifyItemFn] (type, object) -> object: Additional modifications to perform after unserialization
         * @param {Object|Array} [dateFields] Additional date fields to deserialize (For format, see `dateFields` option)
         */
        add : function (type, item, modifyItemFn, dateFields) {
            if (item == null) {
                return;
            }

            var allDateFields = getDateFields.call(this, type);
            if (dateFields != null) {
                allDateFields = _.assign({}, allDateFields, dateFieldsObj(dateFields));
            }

            this._unserialize(type, item, allDateFields, modifyItemFn);

            var itemKey = this._itemKey(type, item);
                        logger.assert(this.item(type, itemKey, true) == null, 'Store | Item with same key (' + itemKey + ') already exists in store');
                        this.indexManager.itemAdded(type, item);
            var newIndex = this.storeCollection.add(type, item);

            if (this._isItemNew(type, item) && !_.includes(this.options.readOnly, type)) {
                this.newStoreCollection.add(type, item);
            }

            this._changeStack.push({
                changetype : 'add',
                type : type,
                key : itemKey
            });

            /**
             * Triggered when an item is added to the store
             *
             * Args:
             *  * type: The type of the added item
             *  * item: The item added
             *  * position: The position of the item in the store
             *
             * @event add
             * @param args {Object}
             *
             */
            this._subscribable.trigger('add', {
                type : type,
                item : item,
                key : itemKey,
                position : newIndex,
                changeStack : this._changeStack
            });

            this._changeStack.pop();
        },

        /**
         * Add item to temporary store for a given type
         *
         * @method addTemporary
         * @param {String} type
         * @param {Object} item
         * @param {Function} [modifyItemFn] (type, object) -> object: Additional modifications to perform after unserialization
         * @param {Object|Array} [dateFields] Additional date fields to deserialize (For format, see `dateFields` option)
         */
        addTemporary : function (type, item, modifyItemFn, dateFields) {
            if (item == null) {
                return;
            }

            var allDateFields = getDateFields.call(this, type);
            if (dateFields != null) {
                allDateFields = _.assign({}, allDateFields, dateFieldsObj(dateFields));
            }

            this._unserialize(type, item, allDateFields, modifyItemFn);

            var itemKey = this._itemKey(type, item);
                        logger.assert(this.item(type, itemKey, true) == null, 'Store | Item with same key (' + itemKey + ') already exists in store');
                        //We don't need to call the index manager for temporary records because they can't have children and should not be indexed yet
            this.tempStoreCollection.add(type, item);
        },

        /**
         * Combines updated data with an existing record according to the merge mode.
         *
         * **WARNING** This method has been been completely changed in 3.7.0.
         * Previously this method acted as if you had called [`mergeAndCommit`](#method_mergeAndCommit).
         * Since 3.7.0, this method now operates on a single record and updates its data with the new data based on the chosen merge mode.
         * This method is used internally by `mergeAndCommit` and [`mergeAndCommitAll`](#method_methodAndCommitAll).
         *
         * The `options` parameter is an object which has three properties:
         * - `mergeMode`: How the new data should be merged with existing data
         * - `modifyItemFn`: A function that will be used to modify the item before it is added to the store (same as with the [`add`](#method_add) method).
         * - `dateFields`: Fields on the record to automatically deserialize into dates (same as with the `add` method, for format, see `dateFields` option).
         *
         * See information about the merge modes under the `mergeMode` option.
         *
         * This method triggers the [`merge`](#events_merge) event.
         * It does **not** trigger the `update` event because it does not do deep checking to see which fields actually differ between the old and new record.
         * And this behavior could cause adverse functionality to existing Prime grids.
         * The item is still marked as modified, but only top-level properties are checked.
         * This means that all nested properties will report modification even if they weren't changed.
         *
         * Consider the following example:
         *
         *     Before:                       After:
         *     {                             {
         *         id: 1,                        id: 1,
         *         foo: 'foo',                   foo: 'bar',
         *         other: 'other',               other: 'other',
         *         nested: {                     nested: {
         *             kept: 'kept',                 kept: 'kept',
         *             updated: 'original'           updated: 'updated'
         *         }                             }
         *     }                             }
         *
         *     store.modifiedFields(type, '1'); // ['foo', 'nested']
         *     store.isModifiedField(type, '1', 'foo'); // true
         *     store.isModifiedField(type, '1', 'nested.kept'); // true (even though 'kept' === 'kept' because only 'nested' was checked and not equal)
         *     store.isModifiedField(type, '1', 'nested.updated'); // true
         *
         * For this reason, you should only use `merge` if you know that the data you are merging into the record is free of erroneous data.
         *
         * @method merge
         * @param {String} type
         * @param {String|Number} key
         * @param {Object} item
         * @param {Object} options
         * @since 3.7.0
         */
        merge : function (type, key, item, options) {
            var safeOptions = options || {},
                keyField = this._getTypeKey(type),
                mergeMode = safeOptions.mergeMode || getTypedOption.call(this, type, 'mergeMode'),
                allDateFields = getDateFields.call(this, type),
                oldItem = this.storeCollection.item(type, key) || {},
                mergedKeys;

            if (safeOptions.dateFields != null) {
                allDateFields = _.assign({}, allDateFields, dateFieldsObj(safeOptions.dateFields));
            }

            logger.assert(key === item[keyField], 'Merging item\'s key and key parameter must match');

            if (mergeMode === MERGE_MODES.NONE) {
                if (this.storeCollection.item(type, key) == null &&
                    this.deleteStoreCollection.item(type, key) == null &&
                    this.modifyStoreCollection.item(type, key) == null) {
                    // Don't need to use allDateFields here because it's essentially repeated by the add call
                    this.add(type, item, safeOptions.modifyItemFn, safeOptions.dateFields);
                }
            } else if (mergeMode === MERGE_MODES.REPLACE) {
                this._unserialize(type, item, allDateFields, safeOptions.modifyItemFn);
                if (!safeOptions.bypassModify) {
                    mergedKeys = _.keys(item);
                }
                this.indexManager.itemRemoved(type, key, oldItem);
                oldItem = this.storeCollection.replace(type, item); // Recapture oldItem
                this.indexManager.itemAdded(type, item);
            } else { // EXTEND_EXISTING(_SHALLOW)?, // EXTEND_NEW(_SHALLOW)?
                var extendExisting = (mergeMode === MERGE_MODES.EXTEND_EXISTING || mergeMode === MERGE_MODES.EXTEND_EXISTING_SHALLOW);
                var shallowExtend = (mergeMode === MERGE_MODES.EXTEND_EXISTING_SHALLOW || mergeMode === MERGE_MODES.EXTEND_NEW_SHALLOW);
                var newItem;
                this._unserialize(type, item, allDateFields, safeOptions.modifyItemFn);
                if (!safeOptions.bypassModify) {
                    mergedKeys = _.keys(item); // Need keys from pre-merge object to avoid marking all fields in `newItem` as modified
                }
                newItem = extendExisting ? objectUtils[shallowExtend ? 'extend' : 'extendDeep']({}, oldItem, item) : objectUtils[shallowExtend ? 'extend' : 'extendDeep']({}, item, oldItem);
                this.indexManager.itemRemoved(type, key, oldItem);
                oldItem = this.storeCollection.replace(type, newItem); // Recapture oldItem
                this.indexManager.itemAdded(type, newItem);
                item = newItem; // For markModify check below
            }

            if (mergedKeys != null) {
                _.forEach(mergedKeys, function (field) {
                    if (field === keyField) { return; } // Never modifying the key
                    markItemModified.call(this, type, key, field, item[field], oldItem); // Might be `newItem` if EXTEND-y merge (see above)
                }.bind(this));

                /**
                 * Triggered when an item is updated in the store
                 *
                 * Args:
                 *  * type: The type of the merged item
                 *  * key: The key of the item merged
                 *
                 * @event merge
                 * @param args {Object}
                 */
                this._subscribable.trigger('merge', {
                    type : type,
                    key : key
                });
            }
        },

        /**
         * Merges and commits several types at once.
         *
         * See additional information in the `mergeAndCommit` method.
         *
         * This method also triggers the 'commit-all' method.
         *
         * @method mergeAndCommitAll
         * @param newKeyMap {Object}
         * @param newItemsMap {Object}
         * @param {Function} [modifyItemFn] (type, object) -> object: Additional modifications to perform after unserialization
         * @param {Object|Array} [dateFields] Additional date fields to deserialize (For format, see `dateFields` option)
         * @param mergeMode {String}
         */
        mergeAndCommitAll : function (newKeyMap, newItemsMap, modifyItemFn, dateFields, mergeMode) {
            var that = this,
                types = _.union(_.keys(newKeyMap), _.keys(newItemsMap));
            _.forEach(types, function (type) {
                that.mergeAndCommit(type, newKeyMap[type] || {}, newItemsMap[type] || [], modifyItemFn, dateFields, mergeMode);
            });

            // Trigger commit all since we're not actually really calling the commitAll function
            that._subscribable.trigger('commit-all');
        },

        /**
         * Combines new data with existing data according to the merge mode and then performs a commit.
         *
         * This method uses the [`merge`](#method_merge) method under the hood.
         * However, it does not trigger individual [`merge`](#event_merge) events.
         *
         * This method triggers the 'commit' method.
         *
         * @method mergeAndCommit
         * @param {String} type
         * @param {Object} newKeys
         * @param {Array} items
         * @param {Function} [modifyItemFn] (type, object) -> object: Additional modifications to perform after unserialization
         * @param {Object|Array} [dateFields] Additional date fields to deserialize (For format, see `dateFields` option)
         * @param {String} [mergeMode]
         */
        mergeAndCommit : function (type, newKeys, items, modifyItemFn, dateFields, mergeMode) {
            var that = this,
                oldKeys = _.invert(newKeys),
                key = this._getTypeKey(type);

            _.forEach(items, function (item) {
                item[key] = hasOwnProp(oldKeys, item[key]) ? oldKeys[item[key]] : item[key];
                that.merge(type, item[key], item, {
                    mergeMode : mergeMode,
                    modifyItemFn : modifyItemFn,
                    dateFields : dateFields,
                    bypassModify : true
                });
            });

            that.commit(type, newKeys);
        },

        /**
         * Does a merge and commit for specific items from a particular type collection.
         *
         * Like with a regular [`mergeAndCommit`](#method_mergeAndCommit), you can pass a map of old IDs to new IDs.
         * However, to commit existing updated records, you must the same ID as both old and new.
         *
         * There is no equivalent mergeAndCommitAllPartial that would do several types at once.
         *
         * @method mergeAndCommitPartial
         * @param type
         * @param newAndOldKeys
         * @param items
         * @param modifyItemFn
         * @param dateFields
         * @param mergeMode
         */
        mergeAndCommitPartial : function (type, newAndOldKeys, items, modifyItemFn, dateFields, mergeMode) {
            var that = this,
                oldKeys = _.invert(newAndOldKeys),
                key = this._getTypeKey(type);

            _.forEach(items, function (item) {
                item[key] = hasOwnProp(oldKeys, item[key]) ? oldKeys[item[key]] : item[key];
                that.merge(type, item[key], item, {
                    mergeMode : mergeMode,
                    modifyItemFn : modifyItemFn,
                    dateFields : dateFields,
                    bypassModify : true
                });
            });

            that.commitPartial(type, newAndOldKeys);
        },

        /**
         * Add item to store at a particular position to a given type
         *
         * Triggers Event - 'insert'
         *
         * @method insertAt
         * @param {String} type
         * @param {Object} item
         * @param {Integer} position
         * @param {Function} [modifyItemFn] (type, object) -> object: Additional modifications to perform after unserialization
         * @param {Object|Array} [dateFields] Additional date fields to deserialize (For format, see `dateFields` option)
         */
        insertAt : function (type, item, position, modifyItemFn, dateFields) {
            if (item == null) {
                return;
            }

            var allDateFields = getDateFields.call(this, type);
            if (dateFields != null) {
                allDateFields = _.assign({}, allDateFields, dateFieldsObj(dateFields));
            }

            this._unserialize(type, item, allDateFields, modifyItemFn);

            this.storeCollection.insertAt(type, item, position);

            if (this._isItemNew(type, item)) {
                if (!_.includes(this.options.readOnly, type)) {
                    this.newStoreCollection.add(type, item);
                }
            }

            /**
             * Triggered when an item is inserted into the store
             *
             * Args:
             *  * type: The type of the inserted item
             *  * item: The item added
             *  * position: The position of the item in the store
             *
             * @event insert
             * @param args {Object}
             */
            this._subscribable.trigger('insert', {
                type : type,
                item : item,
                position : position
            });
        },

        /**
         * Return position of item in store for a given type
         *
         * @method position
         * @param {String} type
         * @param {String} key
         * @return Integer
         */
        position : function (type, key) {
            return this.storeCollection.position(type, key);
        },

        /**
         * Return an array of items from the store for a given type
         *
         * To include items that have been deleted set the <i>includeDeleted</i> param to true
         *
         * To filter the return list pass a function (since forever) or object (since 2.11.5) in for <i>filterFn</i> param (as
         * [`_.filter`](http://devdocs.io/lodash/index#filter)).
         *
         * Example:
         *
         *      var items = store.items('type', function(type, item) {
         *          // evaluate to true to include item
         *          return item.foo === 'foo';
         *      });
         *
         * @method items
         * @param {String} type
         * @param {Function|Object} [filterFn] (item) -> boolean: Filter items that match the function condition, or objects that match all the given object
         *     properties (as [`_.filter`](http://devdocs.io/lodash/index#filter))
         * @param {Boolean} [includeDeleted] Also return items that are marked for deletion
         * @return Array
         */
        items : function (type, filterFn, includeDeleted) {
            var items = [];
            if (includeDeleted) {
                items = _.union(this.storeCollection.items(type), this.deleteStoreCollection.items(type));
            } else {
                items = this.storeCollection.items(type);
            }

            if (arguments.length >= 2) {
                items = _.filter(items, filterFn);
            }
            return items;
        },

        /**
         * Return item from store for a given type with that key
         *
         * To include items that have been deleted set the <i>includeDeleted</i> param to true
         *
         * If the item is not found, this method returns `null`.
         *
         * @method item
         * @param {String} type
         * @param {String} key
         * @param {Boolean} [includeDeleted] Also try looking for the item in the deleted items collection
         * @return object
         */
        item : function (type, key, includeDeleted) {
            var item = this.storeCollection.item(type, key);
            if (item == null && includeDeleted) {
                item = this.deleteStoreCollection.item(type, key);
            }
            if (item == null) {
                item = this.tempStoreCollection.item(type, key);
            }
            return item;
        },

        /**
         * Update item in store for a given type with that key.
         *
         * The store holds the original value so value can be reverted.
         *
         * Values can be set at any level for the item.
         *
         * If the value is set back to a value which passes the `_.isEqual` test, the field will be reset and will
         * no longer appear in the modified fields list.
         *
         * Example:
         *
         *      var item = {
         *          key : '1',
         *          bar : 'bar',
         *          foo : {
         *              foo : 'foo'
         *          }
         *      };
         *      store.add('type', item);
         *      console.log(store.isModifiedField('type', '1', 'foo.foo'));             // false
         *      store.update('type', '1', 'foo.foo', 'bar');
         *      console.log(store.isModifiedField('type', '1', 'foo.foo'));             // true
         *      console.log(store.isModifiedField('type', '1', 'bar'));                 // false
         *      store.update('type', '1', 'bar', 'foo');
         *      console.log(store.isModifiedField('type', '1', 'bar'));                 // true
         *      store.update('type', '1', 'bar', 'bar');
         *      console.log(store.isModifiedField('type', '1', 'bar'));                 // false
         *
         * <b>Note</b> For the store to work correctly all values must be updated through the store using this function
         *
         * Triggers Event - 'update'
         *
         * <b>Note about using `preventDirty`</b>:
         *
         * Using the `preventDirty` flag prevents the change from being tracked in the modified fields list.
         * This can cause inconsistencies if it is not used consistently for the same field.
         * I consider the inconsistent use of the flag as undefined behavior that could potentially change in future releases.
         * If the behavior does change, there will be a blog post about it.
         *
         * The following example demonstrates what happens based on the order of the use of the flag:
         *
         *     store.add('type', { key : '1', value : 'a' });
         *
         *     store.update('type', '1', 'value', 'b', true); // Updates value to 'b', triggers the update event, but doesn't track the change
         *     store.update('type', '1', 'value', 'c');       // Updates value to 'c', triggers the update event, and tracks the change
         *     store.resetAll();
         *
         *     store.item('type', '1').value === 'b';         // value is reset to 'b' as it was the value when the first tracked change occurred
         *
         *     // ----------
         *
         *     store.add('type', { key : '1', value : 'a' });
         *
         *     store.update('type', '1', 'value', 'b');       // Updates value to 'b', triggers the update event, and tracks the change
         *     store.update('type', '1', 'value', 'c', true); // Updates value to 'c', triggers the update event, but doesn't track the change
         *     store.resetAll();
         *
         *     store.item('type', '1').value === 'a';         // value is reset to 'a' as it was the value when the first tracked change occurred
         *
         * @method update
         * @param {String} type
         * @param {String} key
         * @param {String} field A period-delimited path to the field to update
         * @param {*} value
         * @param {Boolean} [preventDirty] (Default: `false`) True to bypass tracking the change in the modify collection (See note in description)
         */
        update : function (type, key, field, value, preventDirty) {
            if (_.includes(this.options.readOnly, type)) {
                throw new Error('store.update() - You can not update a read only type - ' + type);
            }

            var item = this.storeCollection.item(type, key);
            var isFromTemp = false;
            if (item == null) {
                item = this.tempStoreCollection.item(type, key);
                isFromTemp = item != null;
            }
            if (item != null) {
                // map new value, if necessary
                var processors = this.utilities._getValue(this, 'options.fieldProcessors.' + type);
                var processor = processors && processors[field];
                if (processor && processor.update) {
                    value = processor.update(value, type, item, field);
                }

                // is the new value different from present value
                var currentValue = this.utilities._getValue(item, field);
                var compFn = processor && processor.comparison || this.options.comparisonFunction;
                if (compFn(value, currentValue, type, item, field)) {
                    return;
                }

                this._changeStack.push({
                    changetype : 'update',
                    type : type,
                    key : key,
                    path : field
                });

                /**
                 * Triggered before an item is updated in the store.
                 *
                 * Args:
                 *  * type: The type of the updated item
                 *  * key: The key of the item updated
                 *  * preventUpdate: Function to call to prevent the update from occurring
                 *  * isUpdatePrevented: Function to check if the update has already been prevented by another listener
                 *
                 * @event before-update
                 * @param args {Object}
                 */
                var preventUpdate = false;
                var beforeUpdateArgs = {
                    type : type,
                    key : key,
                    path : field,
                    newValue : value,
                    changeStack : this._changeStack,
                    preventUpdate : function () {
                        preventUpdate = true;
                    },
                    isUpdatePrevented : function () {
                        return preventUpdate;
                    }
                };
                this._subscribable.trigger('before-update', beforeUpdateArgs);
                if (beforeUpdateArgs.isUpdatePrevented()) {
                    this._changeStack.pop();
                    return;
                }

                this.indexManager.itemUpdated(type, item, field, value, currentValue);

                // tell the modify store value has changed
                var modFn = processor && processor.isItemModified || this.options.isItemModified;
                if (!isFromTemp && !preventDirty && modFn(value, currentValue, type, item, field)) {
                    markItemModified.call(this, type, key, field, value, item);
                }

                // change the value
                if (this.utilities._getValue(item, field) == null) {
                    var parts = field.split(PATH_DELIMITER);
                    var itemPart = item;
                    for (var i = 0; i < parts.length; i++) {
                        var token = parts[i];
                        if (itemPart[token] == null) {
                            // create
                            if (i === parts.length - 1) {
                                itemPart[token] = undefined;
                            } else {
                                itemPart[token] = {};
                            }
                        }
                        itemPart = itemPart[token];
                    }
                }

                this.utilities._setValue(item, field, value);

                /**
                 * Triggered when an item is updated in the store
                 *
                 * Args:
                 *  * type: The type of the updated item
                 *  * key: The key of the item updated
                 *  * path: The path to the modified field
                 *
                 * @event update
                 * @param args {Object}
                 */
                this._subscribable.trigger('update', {
                    type : type,
                    key : key,
                    path : field,
                    changeStack : this._changeStack
                });

                this._changeStack.pop();
            }
        },

        /**
         * Remove item from store for a given type with that key
         *
         * If you attempt to remove an item from the store that does not exist, no error is thrown.
         *
         * Triggers Events - 'before-remove', 'remove'
         *
         * @method remove
         * @param {String} type
         * @param {String} key
         * @return {Object} the item removed from the store (post 3.4.1)
         */
        remove : function (type, key, silent) {
            var item = this.item(type, key);
            if (item != null) {
                return _remove.call(this, type, item, key, silent, 'remove');
            }
        },

        /**
         * Removes a batch of items from the store.
         *
         * Triggers events - `before-remove-all`, `remove-all`, and for each record, `before-remove` and `remove`
         *
         * @method removeAll
         * @param {String} type
         * @param {Array<Object>} records the records to remove
         */
        removeAll : function (type, items, silent) {
            this._changeStack.push({
                changetype : 'remove',
                type : type,
                items : items
            });

            if (!silent) {
                /**
                 * Triggered before a batch of items is removed from the store.
                 *
                 * Args:
                 *  * type: The type of the removed items
                 *  * items: The items being removed
                 *  * preventRemove: Function to call to prevent the delete from occurring
                 *  * isRemovePrevented: Function to check if the remove has already been prevented by another listener
                 *
                 * @event before-remove-all
                 * @param args {Object}
                 */
                var preventRemove = false;
                var beforeRemoveArgs = {
                    type : type,
                    items : items,
                    changeStack : this._changeStack,
                    preventRemove : function () {
                        preventRemove = true;
                    },
                    isRemovePrevented : function () {
                        return preventRemove;
                    }
                };
                this._subscribable.trigger('before-remove-all', beforeRemoveArgs);
                if (beforeRemoveArgs.isRemovePrevented()) {
                    this._changeStack.pop();
                    return;
                }
            }

                        logger.assert(_.isArray(items), 'items to remove must be an array, of objects');
                        for (var i = 0, len = items.length; i < len; ++i) {
                var item = items[i];
                var key = this._itemKey(type, item);
                _remove.call(this, type, item, key, silent, '_removeQuick');
            }
            this.newStoreCollection._repopulateIndexes(type);
            this.modifyStoreCollection._repopulateIndexes(type);
            this.storeCollection._repopulateIndexes(type);

            if (!silent) {
                /**
                 * Triggered after a batch of items is removed from the store.
                 *
                 * Args:
                 *  * type: The type of the removed items
                 *  * items: The items being removed
                 *
                 * @event remove-all
                 * @param args {Object}
                 */
                this._subscribable.trigger('remove-all', {
                    type : type,
                    item : items,
                    changeStack : this._changeStack
                });
            }

            this._changeStack.pop();
        },

        /**
         * Remove item from the temporary store for a given type with that key
         *
         * If you attempt to remove an item from the store that does not exist, no error is thrown.
         *
         * @method removeTemporary
         * @param {String} type
         * @param {String} key
         * @return {Object} the item removed from the store (post 3.4.1)
         */
        removeTemporary : function (type, key) {
            var item = this.item(type, key);
            if (item != null) {
                //We don't need to call the index manager for temporary records because they can't have children and should not be indexed yet
                this.tempStoreCollection.remove(type, key);
                return item;
            }
        },

        /**
         * Does the field exist for a given type with that key.
         *
         * If a field has been set to undefined explicitly this method will still return false.
         *
         * Example:
         *
         *      var item = {
         *          key : '1',
         *          foo : {
         *              foo : 'foo',
         *              bar : undefined
         *          }
         *      };
         *      store.add('type', item);
         *      console.log(store.hasField('type', '1', 'key'));                    // true
         *      console.log(store.hasField('type', '1', 'bar'));                    // false
         *      console.log(store.hasField('type', '1', 'foo.foo'));                // true
         *      console.log(store.hasField('type', '1', 'foo.bar'));                // false
         *
         * @method hasField
         * @param {String} type
         * @param {String} key
         * @param {String} field
         * @returns Boolean
         */
        hasField : function (type, key, field) {
            var item = this.item(type, key);
            if (item != null) {
                var value = this.utilities._getValue(item, field);
                return !_.isUndefined(value);
            }
            return false;
        },

        /**
         * Determines whether or not a particular type or type-key pair has been updated in the store.
         * (Adds and removals do not count as updates for this method; for that, see `isDirty` and `isDirtyAll`)
         *
         * If no arguments are provided this behaves exactly like isModifiedAll.
         * If one argument is provided, it checks to see if a particular type has any changes.
         * With both arguments it checks to see if a particular key in a type has been updated.
         *
         * @method isModified
         * @param {String} [type]
         * @param {String} [key]
         * @return Boolean
         */
        isModified : function (type, key) {
            if (!arguments.length) {
                return this.isModifiedAll();
            } else if (arguments.length === 1) {
                return !(this.modifyStoreCollection.isEmpty(type));
            } else {
                return !_.isEmpty(this.modifyStoreCollection.item(type, key));
            }
        },

        /**
         * Has the store had any items updated for all types.
         * (Adds and removals do not count as updates for this method; for that, see `isDirty` and `isDirtyAll`)
         *
         * @method isModifiedAll
         * @return Boolean
         */
        isModifiedAll : function () {
            return !this.modifyStoreCollection.isEmptyAll();
        },

        /**
         * Determines whether or not a particular type or type-key pair has been added in the store.
         *
         * If no arguments are provided this behaves exactly like isNewAll.
         * If one argument is provided, it checks to see if a particular type has any changes.
         * With both arguments it checks to see if a particular key in a type has been added.
         *
         * @method isNew
         * @param {String} [type]
         * @param {String} [key]
         * @return Boolean
         */
        isNew : function (type, key) {
            if (!arguments.length) {
                return this.isNewAll();
            } else if (arguments.length === 1) {
                return !(this.newStoreCollection.isEmpty(type));
            } else {
                return !_.isEmpty(this.newStoreCollection.item(type, key));
            }
        },

        /**
         * Has the store had any items added for all types.
         *
         * @method isNewAll
         * @return Boolean
         */
        isNewAll : function () {
            return !this.newStoreCollection.isEmptyAll();
        },

        /**
         * Determines whether or not a particular type or type-key pair has been deleted in the store.
         *
         * If no arguments are provided this behaves exactly like isDeletedAll.
         * If one argument is provided, it checks to see if a particular type has any changes.
         * With both arguments it checks to see if a particular key in a type has been deleted.
         *
         * @method isDeleted
         * @param {String} [type]
         * @param {String} [key]
         * @return Boolean
         */
        isDeleted : function (type, key) {
            if (!arguments.length) {
                return this.isModifiedAll();
            } else if (arguments.length === 1) {
                return !(this.deleteStoreCollection.isEmpty(type));
            } else {
                return !_.isEmpty(this.deleteStoreCollection.item(type, key));
            }
        },

        /**
         * Has the store had any items deleted for all types.
         *
         * @method isDeletedAll
         * @return Boolean
         */
        isDeletedAll : function () {
            return !this.deleteStoreCollection.isEmptyAll();
        },

        /**
         * Determines whether or not a particular type or type-key pair has been modified in the store; added, updated, and removed items all count as dirty.
         *
         * If no arguments are provided this behaves exactly like isDirtyAll.
         * If one argument is provided, it checks to see if a particular type has any changes.
         * With both arguments it checks to see if a particular key in a type has been modified.
         *
         * @method isDirty
         * @param {String} [type]
         * @param {String} [key]
         * @return Boolean
         */
        isDirty : function (type, key) {
            if (arguments.length === 0) {
                return this.isDirtyAll();
            } else if (arguments.length === 1) {
                return !(this.newStoreCollection.isEmpty(type) && this.modifyStoreCollection.isEmpty(type) && this.deleteStoreCollection.isEmpty(type));
            } else {
                return this.isModified(type, key) || this.isNew(type, key) || this.isDeleted(type, key);
            }
        },

        /**
         * Has the store had any items added, deleted or modified for all types
         *
         * @method isDirtyAll
         * @return Boolean
         */
        isDirtyAll : function () {
            return !(this.newStoreCollection.isEmptyAll() &&
                    this.modifyStoreCollection.isEmptyAll() &&
                    this.deleteStoreCollection.isEmptyAll());
        },

        /**
         * Has the field been modified for a given type with that key
         *
         * Fields which have been updated back to their original value (matching the `_.isEqual` test) will no longer
         * appear in this list.
         *
         * @method isModifiedField
         * @param {String} type
         * @param {String} key
         * @param {String} field
         * @return Boolean
         */
        isModifiedField : function (type, key, field) {
            return this.modifyStoreCollection.hasOriginalValue(type, key, field);
        },

        /**
         * Answers the question "have any of these fields, for all items in this store type, been modified?"
         * Returns true if so, false otherwise.
         *
         * *Note*: if any items have been added for the type, this will return true.
         *
         * @method anyModified
         * @returns true if any fields were modified for items of the given type, false otherwise
         * @param type {String} the type of items to check
         * @param fields {String|Array} the field, or fields, to check for modifications
         */
        anyModified : function (type, fields) {
                        logger.assert(_.isString(fields) || _.isArray(fields), 'must give String/Array for `anyModified` fn');
            
            if (this.isNew(type)) {
                return true;
            }
            if (!this.isModified(type)) {
                return false;
            }

            if (_.isString(fields)) {
                fields = [fields];
            }

            return this.modifyStoreCollection.anyModified(type, fields);
        },

        /**
         * Return a list of all modified fields for a given type with that key
         *
         * Example:
         *
         *      var item = {
         *          key : '1',
         *          bar : 'bar',
         *          foo : {
         *              foo : 'foo'
         *          }
         *      };
         *      store.add('type', item);
         *      store.update('type', '1', 'foo.foo', 'bar');
         *      store.update('type', '1', 'bar', 'foo');
         *      console.log(store.modifiedFields('type', '1'));                 // ['foo.foo', 'bar']
         *
         * @method modifiedFields
         * @param {String} type
         * @param {String} key
         * @return Array
         */
        modifiedFields : function (type, key) {
            return this.modifyStoreCollection.modifiedFields(type, key);
        },

        /**
         * Returns the original value for a particular path on an object before any updates.
         *
         * @param type {String}
         * @param key {String}
         * @param field {String}
         * @returns {*}
         */
        getOriginalValue : function (type, key, field) {
            if (this.modifyStoreCollection.hasOriginalValue(type, key, field)) {
                return this.modifyStoreCollection.getOriginalValue(type, key, field);
            } else {
                return this.utilities._getValue(this.storeCollection.item(type, key), field);
            }
        },

        /**
         * Reset item for a given type with that key or the entire type
         *
         * Reverts any modified values for item
         *
         * Example:
         *
         *      var item = {
         *          key : '1',
         *          bar : 'bar'
         *      };
         *      store.add('type', item);
         *      store.update('type', '1', 'bar', 'foo');
         *      console.log(store.item('type', '1').bar);                 // 'foo'
         *      store.reset('type', '1');
         *      console.log(store.item('type', '1').bar);                 // 'bar'
         *
         * Example whole type:
         *
         *      var items = [
         *          { key : '1', bar : 'bar' }, { key : '2', bar : 'another' }
         *      ];
         *      store.addAll('type', items);
         *      store.update('type', '1', 'bar', 'foo');
         *      store.update('type', '2', 'bar', 'more');
         *      console.log(store.item('type', '1').bar);                 // 'foo'
         *      console.log(store.item('type', '2').bar);                 // 'more'
         *      store.reset('type');
         *      console.log(store.item('type', '1').bar);                 // 'bar'
         *      console.log(store.item('type', '2').bar);                 // 'another'
         *
         * Triggers Event - `reset`
         *
         * @method reset
         * @param {String} type
         * @param {String} [key]
         */
        reset : function (type, key, silent) {
            var that = this;
            if (key != null) {
                // Reset one particular item
                var modifyItem = that.modifyStoreCollection.item(type, key),
                    storeItem = that.storeCollection.item(type, key);

                if (modifyItem != null) {
                    this.indexManager.itemRemoved(type, key, storeItem);

                    _.forEach(that.modifyStoreCollection._allModifiedFields(type, key), function (field) {
                        if (field.indexOf(PATH_DELIMITER) < 0) {
                            that.utilities._setValue(storeItem, field, modifyItem.oldValues[field]);
                            that.modifyStoreCollection.remove(type, key, field);
                        }
                    });

                    this.indexManager.itemAdded(type, storeItem);
                } else {
                    // The item was added or deleted
                    var deleteItem = that.deleteStoreCollection.item(type, key),
                        addItem = that.newStoreCollection.item(type, key);

                    if (deleteItem != null) {
                        // New items do not end up in the deleteStoreCollection so we don't need to check for new here.
                        that.indexManager.itemAdded(type, deleteItem);
                        that.storeCollection.add(type, deleteItem);
                        that.deleteStoreCollection.remove(type, key);
                    } else if (addItem != null) {
                        that.indexManager.itemRemoved(type, key, addItem);
                        that.remove(type, key, true);
                    }
                }

                /**
                 * Triggered when an item is reset back to its original value
                 *
                 * Args:
                 *  * type: The type of the reset item
                 *  * key: The key of the reset item
                 *
                 * @event reset
                 * @param args {Object}
                 */
                if (!silent) {
                    that._subscribable.trigger('reset', {
                        type : type,
                        key : key
                    });
                }
            } else {
                // Reset whole type

                var typeKey = that._getTypeKey(type);
                if (this.newStoreCollection.map[type]) {
                    var cascadeDeleteNew = function (type, keys) {
                        var newMap = that.newStoreCollection.map[type];
                        var storeMap = that.storeCollection.map[type];
                        var cascadeDeletes = getTypedOption.call(that, type, 'cascadeDelete');
                        var cascadeTypesToDelete = (cascadeDeletes === true) ? that.indexManager.getChildTypes(type) : _.intersection(cascadeDeletes, that.indexManager.getChildTypes(type));
                        var cascadeRemovedTypeKey = that._getTypeKey(type);

                        _.forEach(keys, function (key) {
                            var item = storeMap.item(key);
                            that.indexManager.itemRemoved(type, key, item);

                            if (cascadeDeletes) {
                                _.forEach(cascadeTypesToDelete, function (childType) {
                                    if (_.has(that.newStoreCollection.map, childType)) {
                                        var childTypeKey = that.indexManager.indexDefinitions.parentToChild[type][childType];
                                        var id = (cascadeRemovedTypeKey === childTypeKey) ? key : that.utilities._getValue(item, childTypeKey);
                                        cascadeDeleteNew(childType, that.indexManager.getChildIds(type, id, childType));
                                    }
                                });
                            }
                        });
                        newMap.removeMany(keys);
                        storeMap.removeMany(keys);
                    };

                    // We can skip going through the remove method because we want a 'silent' remove and we'll do our own cascade delete check here
                    // This function assumes that all children of new records are themselves new records
                    cascadeDeleteNew(type, _.keys(this.newStoreCollection.map[type].keys));
                }

                if (this.modifyStoreCollection.map[type]) {
                    _.forEach(this.modifyStoreCollection.map[type].items(), function (modifyItem) {
                        var storeItem = that.storeCollection.item(type, modifyItem[typeKey]),
                            itemKey = that._itemKey(type, storeItem);
                        that.indexManager.itemRemoved(type, itemKey, storeItem);
                        _.forEach(that.modifyStoreCollection._allModifiedFields(type, modifyItem[typeKey]), function (field) {
                            if (field.indexOf(PATH_DELIMITER) < 0) {
                                that.utilities._setValue(storeItem, field, modifyItem.oldValues[field]);
                                // No need to call modifyStoreCollection.remove here since we're emptying is below with .clear
                            }
                        });
                        that.indexManager.itemAdded(type, storeItem);
                    });
                }

                if (this.deleteStoreCollection.map[type]) {
                    _.forEach(this.deleteStoreCollection.map[type].items(), function (item) {
                        // New items do not end up in the deleteStoreCollection so we don't need to check for new here.
                        that.indexManager.itemAdded(type, item);
                        that.storeCollection.add(type, item);
                    });
                }

                this.newStoreCollection.clear(type);
                this.modifyStoreCollection.clear(type);
                this.deleteStoreCollection.clear(type);

                if (!silent) {
                    that._subscribable.trigger('reset', {
                        type : type
                    });
                }
            }
        },

        /**
         * Reset all items in store.
         *
         * Removes all new items.
         * Reverts all deleted items.
         * Reverts any modified values for all items.
         *
         * Example:
         *
         *      var newItem = {
         *          key : '-1'
         *      };
         *      var deletedItem = {
         *          key : '2'
         *      };
         *      var modifyItem = {
         *          key : '3'
         *      };
         *      store.add('type', newItem);
         *      store.add('type', deletedItem);
         *      store.add('type', modifyItem);
         *      store.remove('type, '2');
         *      store.update('type, '3', 'bar', 'bar');
         *      console.log(store.isDirtyAll());                                // true
         *      store.resetAll();
         *      console.log(store.isDirtyAll());                                // false
         *
         * Triggers Event - 'reset-all'
         *
         * @method resetAll
         */
        resetAll : function () {
            var that = this,
                types = _.union(
                    this.storeCollection.types(),
                    this.modifyStoreCollection.types(),
                    this.newStoreCollection.types(),
                    this.deleteStoreCollection.types());

            _.forEach(types, function (type) {
                that.reset(type, undefined, true);
            });

            /**
             * Triggered when the entire store is reset with `store.resetAll()`.
             * Doesn't provide any parameters.
             *
             * @event reset-all
             */
            this._subscribable.trigger('reset-all');
        },

        /**
         * Clear items from store for a given type.
         *
         * Pass in an array of keys to only clear items that meet those keys.
         *
         * Triggers the `before-clear` event before cleaing
         * Triggers the `clear` event after items have been cleared from the store and all internal collections
         *
         * @method clear
         * @param {String} type
         * @param {Array} [keys]
         */
        clear : function (type, keys) {
            /**
             * Triggered before an individual store type is cleared, with the keys if specified
             *
             * Args:
             *  * `type`: The type of the cleared store
             *  * `keys`: The specific keys to be cleared, or undefined if all keys were to be cleared
             *
             * @event before-clear
             * @param args {Object}
             */
            this._subscribable.trigger('before-clear', {
                type : type,
                keys : keys
            });

            var that = this;
            if (_.isArray(keys)) {
                _.forEach(keys, function (key) {
                    var item = that.storeCollection.item(type, key, true);
                    if (item != null) {
                        that.indexManager.itemRemoved(type, key, item);
                    }
                });
            } else {
                this.indexManager.typeCleared(type);
            }

            this.storeCollection.clear(type, keys);
            this.modifyStoreCollection.clear(type, keys, true);
            this.deleteStoreCollection.clear(type, keys);
            this.newStoreCollection.clear(type, keys);

            /**
             * Triggered when an individual store type is cleared.
             *
             * Args:
             *  * `type`: The type of the cleared store
             *  * `keys`: The specific keys to be cleared, or undefined if all keys were to be cleared
             *
             * @event clear
             * @param args {Object}
             */
            this._subscribable.trigger('clear', {
                type : type,
                keys : keys
            });
        },

        /**
         * Clear all items from store.
         *
         * Triggers the event - `before-clear-all` before clearing
         * Triggers the event - `clear-all` once done
         *
         * @method clearAll
         */
        clearAll : function () {
            /**
             * Triggered before the entire store is cleared with `store.clearAll()`.
             * Doesn't provide any parameters.
             *
             * @event before-clear-all
             */
            this._subscribable.trigger('before-clear-all');

            this.indexManager.allCleared();
            this.storeCollection.clearAll();
            this.modifyStoreCollection.clearAll();
            this.deleteStoreCollection.clearAll();
            this.newStoreCollection.clearAll();

            /**
             * Triggered when the entire store is cleared with `store.clearAll()`.
             * Doesn't provide any parameters.
             *
             * @event clear-all
             */
            this._subscribable.trigger('clear-all');
        },

        /**
         * Return a save object for all item types.
         *
         * Example:
         *
         *      {
         *          foo : {
         *              added: [],      // array of objects
         *              modified: [],   // array of objects
         *              deleted: []     // array of keys
         *          },
         *          bar : {
         *              added: [],      // array of objects
         *              modified: [],   // array of objects
         *              deleted: []     // array of keys
         *          }
         *      }
         *
         * @param {Function} [serializeFn]
         * @method saveAll
         * @return Object
         */
        saveAll : function (serializeFn) {
            var that = this;
            var saveData = {};

            var types = that.storeCollection.types();
            _.forEach(types, function (type) {
                var saveObj = that.save(type, serializeFn);
                if (saveObj != null && (saveObj[type].added.length || saveObj[type].deleted.length || saveObj[type].modified.length)) {
                    saveData[type] = saveObj[type];
                }
            });

            return saveData;
        },

        /**
         * Return save object for a given type.
         *
         * The type still appears as a top level key on the returned object.
         * See example.
         *
         * Example:
         *
         *      {
         *          foo : {
         *              added: [],      // array of objects
         *              modified: [],   // array of objects
         *              deleted: []     // array of keys
         *          }
         *      }
         *
         * @method save
         * @param {String} type
         * @param {Function} [serializeFn] (type, item) -> item: Additional serialization to perform on the item
         * @return Object
         */
        save : function (type, serializeFn) {
            var that = this,
                key = this._getTypeKey(type);
            if (this.storeCollection.hasType(type)) {
                var saveData = {};
                saveData[type] = {
                    added : [],
                    deleted : [],
                    modified : []
                };

                if (!_.includes(that.options.readOnly, type)) {
                    // new items - whole object
                    _.forEach(this.newStoreCollection.items(type), function (item) {
                        var cloneItem = objectUtils.cloneDeep(item);
                        that._serialize(type, cloneItem, serializeFn);
                        saveData[type].added.push(cloneItem);
                    });

                    // removed items - just the id
                    _.forEach(this.deleteStoreCollection.items(type), function (item) {
                        saveData[type].deleted.push(item[key]);
                    });

                    // modified items
                    _.forEach(this.modifyStoreCollection.items(type), function (item) {
                        var cloneItem = objectUtils.cloneDeep(that.storeCollection.item(type, item[key]));
                        that._serialize(type, cloneItem, serializeFn);
                        saveData[type].modified.push(cloneItem);
                    });
                }
                return saveData;
            }

            return null;
        },

        /**
         * Commit all changes in store.
         *
         * New items are persisted with a real key.
         *
         * To persist new items pass in the following format:
         *
         *      // old and new key map for each type
         *      var newKeys = {
         *          foo : {
         *              '-1' : '17261',
         *              '-2' : '17262'
         *          },
         *          bar : {
         *              '-5' : '2323'
         *          }
         *      };
         *      store.commitAll(newKeys);
         *
         * Deleted items are permanently removed.
         *
         * Modified values on an item are permanent.
         *
         * Triggers the event `commit-all`.
         * Triggers the event `commit` for each committed storeType.
         *
         * @method commitAll
         * @param {Object} newKeys
         */
        commitAll : function (newKeys) {
            var that = this;
            _.forEach(newKeys, function (data, type) {
                that.commit(type, data);
            });
            if (!newKeys) {
                logger.warn('Store.commitAll() should be called with an object describing the storeTypes and keys to be committed.  If not provided, the `commit` event will not get called for each store type that gets cleared.');
            }
            that.modifyStoreCollection.clearAll();
            that.deleteStoreCollection.clearAll();
            that.newStoreCollection.clearAll();

            /**
             * Triggered when `store.commitAll` is called.
             * Doesn't provide any parameters.
             *
             * @event commit-all
             * @since 2.0.9
             */
            that._subscribable.trigger('commit-all');
        },

        /**
         * Commit items to store for a given type.
         *
         * New items are persisted with a real key.
         *
         * To persist new items pass in the following format:
         *
         *      // old and new key map
         *      var newKeys = {
         *          '-1' : '17261',
         *          '-2' : '17262'
         *      };
         *      store.commit('type', newKeys);
         *
         * Deleted items are permanently removed
         *
         * Modified values on an item are permanent
         *
         * Trigger Event - 'commit'
         *
         * @method commit
         * @param {String} type
         * @param {Object} newKeys
         */
        commit : function (type, newKeys) {
            var that = this,
                key = this._getTypeKey(type);
            // update the new item ids
            _.forEach(newKeys, function (newId, oldId) {
                var item = that.storeCollection.item(type, oldId);

                if (item != null) {
                    that.indexManager.itemRemoved(type, oldId, item);
                    item[key] = newId;
                    that.indexManager.itemAdded(type, item);
                    that.storeCollection._repopulateIndexes(type);
                    that._keyChanged(type, oldId, newId);
                }
            });

            // clear new
            that.newStoreCollection.clear(type);

            // clear modified
            that.modifyStoreCollection.clear(type);

            // clear deleted
            that.deleteStoreCollection.clear(type);

            that._subscribable.trigger('commit', { type : type });
        },

        // hook for subclasses
        _keyChanged : function (/* type, oldId, newId */) {
        },

        /**
         * Commits specific items from a particular type collection.
         *
         * Like with a regular [`commit`](#method_commit), you can pass a map of old IDs to new IDs.
         * However, to commit existing updated records, you must the same ID as both old and new.
         *
         * There is no equivalent commitAllPartial that would do several types at once.
         *
         * Example:
         *
         * Store IDs Before:
         * Items: ['-1', '2', '3']
         * Added : ['-1']
         * Modified: ['2', '3']
         * Deleted: ['4']
         *
         *     store.commit(type, {
         *         "-1" : "1",
         *          "2" : "2",
         *          "4" : "4"
         *     });
         *
         * Store IDs After:
         * Items: ['1', '2', '3']
         * Added : []
         * Modified : ['3']
         * Deleted: []
         *
         * @method commitPartial
         * @param type {String}
         * @param newAndOldKeys {Object} Keys to commit (with updated ID as necessary)
         */
        commitPartial : function (type, newAndOldKeys) {
            var that = this,
                key = this._getTypeKey(type),
                oldIds = _.keys(newAndOldKeys);

            // Clear other collections first because the are references to the storeCollection object.

            // clear new
            that.newStoreCollection.clear(type, oldIds);

            // clear modified
            that.modifyStoreCollection.clear(type, oldIds, true);

            // clear deleted
            that.deleteStoreCollection.clear(type, oldIds);

            // update the new item ids
            _.forEach(newAndOldKeys, function (newId, oldId) {
                var item = that.storeCollection.item(type, oldId);

                if (item != null && newId !== oldId) {
                    that.indexManager.itemRemoved(type, oldId, item);
                    item[key] = newId;
                    that.indexManager.itemAdded(type, item);
                    that.storeCollection._repopulateIndexes(type);
                }
            });

            that._subscribable.trigger('commit', { type : type, keys : newAndOldKeys });
        },

        /**
         * Attach an event handler
         *
         * Example:
         *
         *      store.on('add', function(args) {
         *          // fired when an item is added to store
         *      });
         *
         * @method on
         * @param {String} eventName
         * @param {Function} callbackFn
         * @param {Object} [context]
         */
        on : function (eventName, callbackFn, context) {
            this._subscribable.on(eventName, callbackFn, context);
        },

        /**
         * Attach an event handler that will only trigger once.
         *
         *
         * @method one
         * @param {String} eventName
         * @param {Function} callbackFn
         * @param {Object} [context]
         */
        one : function (eventName, callbackFn, context) {
            this._subscribable.one(eventName, callbackFn, context);
        },

        /**
         * Remove event handler
         *
         * Example:
         *
         *      store.remove('add');
         *
         * @method off
         * @param {String} eventName
         */
        off : function (eventName) {
            this._subscribable.off(eventName);
        },

        /**
         * Returns whether or not the store has a collection for the given type.
         * This is different that checking `store.count() > 0` in that even if the store doesn't have a collection for the given type, `.count()` would return 0.
         *
         * @method hasType
         * @param {String} type
         * @returns true if the store has a collection for the given type, false otherwise
         */
        hasType : function (type) {
            return this.storeCollection.hasType(type);
        },

        /**
         * Returns whether or not an index has been applied for the given `parentType`/`childType`, such that `getChildItems` is a supported call.
         *
         * @method hasForeignIndex
         * @param {String} parentType the type of the parent
         * @param {String} childType the type of the parent
         * @returns true if the store has an index for the given parent/child relationship
         */
        hasForeignIndex : function (parentType, childType) {
            var indices = this.indexManager.indexDefinitions;
            if (!indices) {
                return false;
            }

            var indicesOfType = indices.parentToChild[parentType];
            return indicesOfType && indicesOfType[childType];
        },

        /**
         * Returns whether or not an index has been applied for the given `type`/`field`, such that `getItemsByGroupField` is a supported call.
         *
         * @method hasGroupIndex
         * @param {String} type the type
         * @param {String} field the field on the type
         * @returns true if the store has an index for the given type/field relationship
         */
        hasGroupIndex : function (type, field) {
            var indices = this.indexManager.indexDefinitions;
            if (!indices) {
                return false;
            }

            var indicesOfType = indices.group[type];
            return indicesOfType && _.includes(indicesOfType, field);
        },

        /**
         * Sets the known date fields that will automatically be deserialized when an item is added to the store.
         * You can set the global date fields or the date fields per type.
         *
         * Global : `store.setDateFields({...});`
         * Per Type : `store.setDateFields('type', {...});`
         *
         * For the format of the `allFields` parameter, see `dateFields` option.
         *
         * NOTE: You should only call this method when the store (or the specific type) is empty!
         * Calling setDateFields will not modify the existing items in the store. E.g if you've added and new date field
         * and removed another, the old field will not be serialized back to a string and the new field will not be turned
         * into a moment object/
         *
         * NOTE: We generally do not allow updating options on the store because there are often weird side effects.
         * We make an exception for date fields because of how the store generally interacts with the meta data from the server.
         * Generally the store is created before your initial rest call, so we don't known the date fields before that server call.
         * This method allows them to be set after that meta call.
         *
         * @method setDateFields
         * @param [type] {String}
         * @param allFields {Object|Array}
         */
        setDateFields : function (type, allFields) {
            if (arguments.length < 2) {
                allFields = type;
                type = null;
            }

            allFields = dateFieldsObj(allFields);

            if (type == null) {
                this.options.dateFields = allFields || {};
                this._dateFieldsCache = {};
            } else {
                this.options.typeOptions[type] = this.options.typeOptions[type] || {};
                this.options.typeOptions[type].dateFields = allFields || {};
                delete this._dateFieldsCache[type];
            }
        },

        /**
         * Updates the [`fieldProcessors`](#property_fieldProcessors) option based on metadata received from server code.
         *
         * The same NOTEs as described in [`setDateFields`](#method_setDateFields) apply here - this should only be called when the store is empty, and don't
         * normally allow updating options, but defining fieldProcessors is a use case explicitly known in Prime.
         *
         * *NOTE*: this method does a merge on the existing fieldProcessors in the store.
         * It is your responsibility to ensure that no options are overriden with this call.
         *
         * @method mergeFieldProcessors
         * @param processors {Object}
         */
        mergeFieldProcessors : function (processors) {
            _.merge(this.options.fieldProcessors, processors);
        },

        /**
         * Alias for {{#crossLink "store.Store/mergeFieldProcessors:method"}}{{/crossLink}}
         *
         * @method replaceFieldProccessors
         * @param processors {Object}
         */
        replaceFieldProcessors : function (processors) {
            this.mergeFieldProcessors(processors);
        },

        /**
         * Gets the child items of a particular type for a particular item.
         *
         * (You cannot get deleted child items.)
         *
         * @method getChildItems
         * @param type {String} The parent type
         * @param id {String} The primary key of the parent item (this is often, but not always the item key -- see [indices](#property_indices)).
         * @param childType {String} The type of the children to retrieve
         * @returns {Array} The child items of this item
         */
        getChildItems : function (type, id, childType) {
            var that = this,
                ids = this.indexManager.getChildIds(type, id, childType);
            return _.map(ids, function (id) {
                return that.item(childType, id);
            });
        },

        /**
         * Gets the parent item of a particular child.
         *
         * (You cannot get the parent item if it is deleted.)
         *
         * @method getParentItem
         * @param type {String} The type of the child item
         * @param id {String} The id of the child item (This is ALWAYS the item key)
         * @param parentType {String} The type of the parent to retrieve
         * @returns {Object} The parent item or null if it is not found
         */
        getParentItem : function (type, id, parentType) {
            var trueId = this.indexManager.getTrueParentId(type, id, parentType);
            if (trueId == null) {
                return null;
            }
            return this.item(parentType, trueId);
        },

        /**
         * Gets an item by a key other than the item key.
         *
         * @method getItemByUniqueField
         * @param type {String} The type of the item to get
         * @param field {String} The field by which the type is indexed
         * @param value {*} The index value
         * @returns {Object} The item which matches that unique field
         */
        getItemByUniqueField : function (type, field, value) {
            var trueId = this.indexManager.getPrimaryKey(type, field, value);
            if (trueId == null) {
                return null;
            }
            return this.item(type, trueId);
        },

        /**
         * Gets a set of items by a particular field's value.
         *
         * @method getItemsByGroupField
         * @param type {String} The type of the items to get
         * @param field {String} The field by which the type is indexed
         * @param value {*} The index value
         * @returns {Array(Object)} The items which match the group
         */
        getItemsByGroupField : function (type, field, value) {
            var that = this,
                ids = this.indexManager.getGroupIds(type, field, value);
            return _.map(ids, function (id) {
                return that.item(type, id);
            });
        },

        /**
         * Creates a semi-lazy expression that allows for filtering large indexed sets of items before returning a final list.
         *
         * There are five "sub-queries":
         *  - children(type): gets the children of the current set of items of the given type
         *  - childrenTransitive(type): gets all the children recursively of the current set of items of the give type
         *    (Note: This would require a foreign index from type->type, not just parentType->type)
         *  - parents(type): gets the parent of the given type of each item in the current set of items
         *  - fromGroup(field, value): filters the current items to only those that match the given group index condition
         *  - withUnique(field, value): filters the current items to only that which matches the given uniqueness constraint
         *
         * To produce a result use either:
         *  - run(): to get an array of results
         *  - runSingle(): to get the first item result
         *
         * Example:
         *
         *     store.query('parent', '3').children('child').fromGroup('groupingField', 'someValue').run();
         *
         *     store.query('child', '3.3').parents('parent').runSingle();
         *
         *     store.query('child').withUnique('uniqueField', 'foo').parents('parent').runSingle();
         *
         *     store.query('parent').fromGroup('type', 'someType').children('child').run();
         *
         * The `id` parameter is required if your query begins with a `children` or `parents` clause and optional for `fromGroup` and `withUnique`.
         * If `id` is omitted the first `fromGroup` or `withUnique` clause will populate the list, otherwise the list will be filtered from the current collection.
         *
         *     store.query('parent', '3').fromGroup('field', 'value')
         *
         * This will either produce `['3']` if item 3 has { field : 'value' } or `[]` if item 3 does not.
         *
         *     store.query('parent').children('child')
         *
         * This will *always* produce `[]` since a starting id was not provided.
         *
         * @method query
         * @param type {String} The type to start with
         * @param [id] {String|Number} The id of the item to start with
         * @returns {*}
         */
        query : function (type, id) {
            var that = this;
            var ids = arguments.length === 2 ? [id] : [];
            var currentType = type;
            var firstQuery = (ids.length === 0);

            function getChildIds(ids, type) {
                return _.reduce(ids, function (memo, id) {
                    return memo.concat(that.indexManager.getChildIds(currentType, id, type));
                }, []);
            }

            var api = {
                children : function (type) {
                    ids = getChildIds(ids, type);
                    currentType = type;
                    firstQuery = false;
                    return api;
                },
                childrenTransitive : function (type, includeSelf) {
                    var childIds = ids,
                        newIds = (includeSelf && (type === currentType)) ? [].concat(childIds) : [];

                    while (childIds.length) {
                        childIds = getChildIds(childIds, type);
                        newIds = newIds.concat(childIds);
                    }

                    ids = newIds;
                    currentType = type;
                    firstQuery = false;
                    return api;
                },
                parents : function (type) {
                    ids = _.uniq(_.map(ids, function (id) {
                        return that.indexManager.getTrueParentId(currentType, id, type);
                    }));
                    currentType = type;
                    firstQuery = false;
                    return api;
                },
                fromGroup : function (field, value) {
                    var groupIds = that.indexManager.getGroupIds(currentType, field, value);
                    if (firstQuery) {
                        ids = groupIds;
                    } else {
                        ids = _.intersection(ids, groupIds);
                    }
                    firstQuery = false;
                    return api;
                },
                withUnique : function (field, value) {
                    var uniqueId = that.indexManager.getPrimaryKey(currentType, field, value);
                    if (firstQuery && uniqueId != null) {
                        ids = [uniqueId];
                    } else {
                        ids = _.intersection(ids, [uniqueId]);
                    }
                    firstQuery = false;
                    return api;
                },
                run : function () {
                    return _.map(ids || [], function (id) {
                        return that.item(currentType, id);
                    });
                },
                runSingle : function () {
                    return ids.length > 0 ? that.item(currentType, ids[0]) : null;
                }
            };
            return api;
        }
    };

    Store.mergeModes = MERGE_MODES;
    Store.defaultComparisonFunction = isEqual;

    return Store;
});

/**
 * Listens to page resize events and resizes the .content-inner div to be the correct size.
 *
 * Should be called in your page.js file so that it is the first window resize event fired.
 * This is done automatically by bpc, so for most pages in development, this shouldn't be necessary.
 *
 * @class ui.fixedSizePage
 * @module ui
 */

define('primutils/ui/fixedSizePage',['jquery', 'lodash'], function ($, _) {
    'use strict';

    var $inner = $(), $outer = $(), $contentInner, $content, bottomPadding, namespace;

    function fixContentSize() {
        $contentInner = $contentInner && $contentInner.length ? $contentInner : $inner;
        $content = $content && $content.length ? $content : $outer;

        if ($content.length && $contentInner.length) {
            bottomPadding = bottomPadding || parseInt($content.css('padding-bottom').slice(0, -2), 10); // Fixed bottom padding (for save/cancel bar), generally 48

            var windowHeight = $(window).height();
            var topPadding = $contentInner.offset().top; // Generally 132 for pages with headers, 92 for pages without

            var newHeight = Math.floor(windowHeight - topPadding - bottomPadding);
            $contentInner.css('height', newHeight);
        }
    }

    namespace = 'fixed-size-page';

    return {
        setElements : function ($newOuter, $newInner) {
            $outer = $newOuter;
            $inner = $newInner;
        },
        eventNamespace : namespace,
        fixContentSize : fixContentSize,
        activate : _.once(function () {
            $(window).on('resize.' + namespace, fixContentSize);
            fixContentSize();
        })
    };
});


define('primutils/markup/primutils/ui/changeBar/multiViewChangeBarTemplate.hbs', ['handlebars','primutils.markup','primutils/ui/changeBar/helpers','primutils/templates/helpers/button','primutils/templates/helpers/buttonGroup','primutils/templates/helpers/dropdown','primutils/templates/helpers/join','primutils/templates/helpers/caret','primutils/ui/changeBar/helpers'], function (H, u) {
var imports = u.buildImported(Array.prototype.slice.call(arguments, 2), 'primutils/ui/changeBar/multiViewChangeBarTemplate.hbs'),
    func = H.template({"1":function(depth0,helpers,partials,data) {
  var stack1, buffer = "    ";
  stack1 = helpers['change-bar-status'].call(depth0, {"name":"change-bar-status","hash":{},"data":data});
  if (stack1 != null) { buffer += stack1; }
  buffer += "\n    ";
  stack1 = helpers['change-bar-errors'].call(depth0, {"name":"change-bar-errors","hash":{},"data":data});
  if (stack1 != null) { buffer += stack1; }
  buffer += "\n    ";
  stack1 = helpers['change-bar-cancel'].call(depth0, {"name":"change-bar-cancel","hash":{},"data":data});
  if (stack1 != null) { buffer += stack1; }
  buffer += "\n";
  stack1 = helpers.buttonGroup.call(depth0, {"name":"buttonGroup","hash":{
    'class': ("save-view-button-group hidden")
  },"fn":this.program(2, data),"inverse":this.noop,"data":data});
  if (stack1 != null) { buffer += stack1; }
  buffer += "    ";
  stack1 = helpers['change-bar-save'].call(depth0, {"name":"change-bar-save","hash":{},"data":data});
  if (stack1 != null) { buffer += stack1; }
  return buffer + "\n";
},"2":function(depth0,helpers,partials,data) {
  var stack1, buffer = "";
  stack1 = helpers.button.call(depth0, "primary", {"name":"button","hash":{
    'class': ("save-view-button"),
    'disabled': (true)
  },"fn":this.program(3, data),"inverse":this.noop,"data":data});
  if (stack1 != null) { buffer += stack1; }
  buffer += "        ";
  stack1 = helpers.button.call(depth0, "primary", {"name":"button","hash":{
    'dropdownToggle': (true)
  },"fn":this.program(6, data),"inverse":this.noop,"data":data});
  if (stack1 != null) { buffer += stack1; }
  buffer += "\n";
  stack1 = helpers.dropdownMenu.call(depth0, {"name":"dropdownMenu","hash":{},"fn":this.program(8, data),"inverse":this.noop,"data":data});
  if (stack1 != null) { buffer += stack1; }
  return buffer;
},"3":function(depth0,helpers,partials,data) {
  var stack1, buffer = "            ";
  stack1 = helpers['change-bar-save-inner'].call(depth0, {"name":"change-bar-save-inner","hash":{},"fn":this.program(4, data),"inverse":this.noop,"data":data});
  if (stack1 != null) { buffer += stack1; }
  return buffer + "\n";
},"4":function(depth0,helpers,partials,data) {
  var stack1, lambda=this.lambda;
  stack1 = lambda(((stack1 = (depth0 != null ? depth0.locale : depth0)) != null ? stack1.saveView : stack1), depth0);
  if (stack1 != null) { return stack1; }
  else { return ''; }
  },"6":function(depth0,helpers,partials,data) {
  var stack1;
  stack1 = helpers.caret.call(depth0, {"name":"caret","hash":{},"data":data});
  if (stack1 != null) { return stack1; }
  else { return ''; }
  },"8":function(depth0,helpers,partials,data) {
  var stack1, buffer = "            ";
  stack1 = helpers.dropdownItem.call(depth0, {"name":"dropdownItem","hash":{
    'class': ("save-view-as-button"),
    'liClass': ("disabled")
  },"fn":this.program(9, data),"inverse":this.noop,"data":data});
  if (stack1 != null) { buffer += stack1; }
  buffer += "\n            ";
  stack1 = helpers.dropdownItem.call(depth0, {"name":"dropdownItem","hash":{
    'class': ("reset-view-button"),
    'liClass': ("disabled")
  },"fn":this.program(11, data),"inverse":this.noop,"data":data});
  if (stack1 != null) { buffer += stack1; }
  return buffer + "\n";
},"9":function(depth0,helpers,partials,data) {
  var stack1, lambda=this.lambda;
  stack1 = lambda(((stack1 = (depth0 != null ? depth0.locale : depth0)) != null ? stack1.saveViewAs : stack1), depth0);
  if (stack1 != null) { return stack1; }
  else { return ''; }
  },"11":function(depth0,helpers,partials,data) {
  var stack1, lambda=this.lambda;
  stack1 = lambda(((stack1 = (depth0 != null ? depth0.locale : depth0)) != null ? stack1.resetView : stack1), depth0);
  if (stack1 != null) { return stack1; }
  else { return ''; }
  },"compiler":[6,">= 2.0.0-beta.1"],"main":function(depth0,helpers,partials,data) {
  var stack1, buffer = "";
  stack1 = helpers['change-bar-controls'].call(depth0, {"name":"change-bar-controls","hash":{},"fn":this.program(1, data),"inverse":this.noop,"data":data});
  if (stack1 != null) { buffer += stack1; }
  return buffer;
},"useData":true});
return function (context, options) {
    options = u.extendOptions(options || {}, imports);
    return func(context, options).trim();
};
});

define('primutils/ui/changeBar/MultiViewChangeBar',[
    'jquery',
    'lodash',
    'primutils/ui/ChangeBar',
    'primutils/utils/objectUtils',
    'primutils/markup/primutils/ui/changeBar/multiViewChangeBarTemplate.hbs',
    'primutils/utils/language'
], function ($, _, ChangeBar, objectUtils, markup, lang) {
    'use strict';

    var DEFAULT_OPTIONS = {
        canSaveView : _.constant(true)
    };

    function getDefaultStrings() {
        return {
            // This needs to go here because of the customized locale
            changeBarMarkup : markup({
                locale : {
                    saveView : lang.getString('changeBar.saveView'),
                    saveViewAs : lang.getString('changeBar.saveViewAsTriggerModal'),
                    resetView : lang.getString('changeBar.resetView')
                }
            }),
            locale : {
                defaultStatusMessageWithViews : lang.getString('changeBar.defaultStatusMessageWithViews'),
                defaultStatusMessageOnlyViews : lang.getString('changeBar.defaultStatusMessageOnlyViews')
            }
        };
    }

    // In this function, the `this` context is a ChangeBarView
    var extendView = function () {
        var statusMessageOverridden = false;
        var viewSaveShown = false;
        var viewSaveStatus = 'unsaved';
        var viewHideTimeout = null;
        var fixedViewSaveButtonWidth = false;

        var baseShow = this.show;
        var baseHide = this.hide;

        this.$saveView = this.$el.find('.save-view-button-group');
        this.buttons.$saveViewButton = this.$el.find('.save-view-button');
        this.buttons.$saveViewAsButton = this.$el.find('.save-view-as-button');
        this.buttons.$resetViewButton = this.$el.find('.reset-view-button');

        var maybeCancelHide = function () {
            if (viewSaveStatus !== 'unsaved') {
                this._cancelHide();
            }
        }.bind(this);

        var fixSaveViewButtonWidth = function () {
            this.buttons.$saveViewButton.children('.save-button-state').removeClass('hidden');
            this.buttons.$saveViewButton.css('width', this.buttons.$saveViewButton.outerWidth(true));
            updateViewSaveThreeState(viewSaveStatus);
            fixedViewSaveButtonWidth = true;
        }.bind(this);

        var updateViewSaveThreeState = function (state) {
            viewSaveStatus = state;

            var $siblings = this.buttons.$saveViewButton.siblings('.btn')
                .prop('disabled', state !== 'unsaved');
            this.buttons.$saveViewButton
                .prop('disabled', state !== 'unsaved' || !this.controller._options.canSaveView())
                .add($siblings)
                .toggleClass('success', state === 'saved')
                .toggleClass('btn-blue-dark', state === 'saving')
                .toggleClass(this._savingButtonClass, state !== 'saving');
            this.buttons.$saveViewButton.children('.save-button-state').addClass('hidden').filter('.' + state).removeClass('hidden');
            this.$el.toggleClass('view-saving view-saved view-unsaved', false).toggleClass('view-' + state, true);

            maybeCancelHide();
        }.bind(this);

        var saveViewAsState = true;
        this.toggleSaveViewAs = function (state) {
            saveViewAsState = state;
            updateViewSaveButton.call(this);
        };

        var updateViewSaveButton = function () {
            this.buttons.$saveViewAsButton.closest('li').toggleClass('disabled', !(viewSaveShown && saveViewAsState));
        }.bind(this);

        this.showViewSave = function () {
            viewSaveShown = true;
            this.$saveView.removeClass('hidden');
            // Save View Button disabled controlled by updateViewSaveThreeState
            updateViewSaveButton.call(this);
            this.buttons.$resetViewButton.closest('li').removeClass('disabled');

            if (viewHideTimeout != null) {
                clearTimeout(viewHideTimeout);
            }

            if (!statusMessageOverridden) {
                this.clearStatusMessage();
            }
            updateViewSaveThreeState('unsaved');
            baseShow.call(this);
            if (!fixedViewSaveButtonWidth) {
                fixSaveViewButtonWidth();
            }
        };

        this.hideViewSave = function () {
            viewSaveShown = false;
            this.$saveView.addClass('hidden');
            // Save View Button disabled controlled by updateViewSaveThreeState then forced below
            this.buttons.$saveViewAsButton.closest('li').addClass('disabled');
            this.buttons.$resetViewButton.closest('li').addClass('disabled');

            if (viewHideTimeout != null) {
                clearTimeout(viewHideTimeout);
            }

            if (!statusMessageOverridden) {
                this.clearStatusMessage();
            }
            updateViewSaveThreeState('unsaved');
            this.buttons.$saveViewButton.prop('disabled', true);
            if (!this.controller._reallyShown) {
                baseHide.call(this);
            } else if (this._savingState === 'unsaved') {
                baseShow.call(this); // Force button state reset
            }
        };

        this.viewSaving = function () {
            updateViewSaveThreeState('saving');
            maybeCancelHide();
            if (viewHideTimeout != null) {
                clearTimeout(viewHideTimeout);
            }
        };

        this.viewSaveSucceeded = function () {
            updateViewSaveThreeState('saved');
            if (viewHideTimeout == null) {
                viewHideTimeout = setTimeout(function () {
                    this.hideViewSave();
                    viewHideTimeout = null;
                }.bind(this), 2000);
            }
        };

        this.viewSaveFailed = function () {
            updateViewSaveThreeState('unsaved');
            maybeCancelHide();
            if (viewHideTimeout != null) {
                clearTimeout(viewHideTimeout);
            }
        };

        this.show = _.wrap(baseShow, function (origFn) {
            this.controller._reallyShown = true;
            origFn.apply(this, _.rest(arguments));
            if (!statusMessageOverridden) {
                this.clearStatusMessage();
            }
        });

        this.hide = _.wrap(baseHide, function (origFn) {
            this.controller._reallyShown = false;
            if (!viewSaveShown) {
                origFn.apply(this, _.rest(arguments));
            } else {
                baseShow.call(this); // Force button state reset
            }
            if (!statusMessageOverridden) {
                this.clearStatusMessage();
            }
        });

        this.saveSucceeded = _.wrap(this.saveSucceeded, function (origFn) {
            origFn.apply(this, _.rest(arguments));
            maybeCancelHide();
            setTimeout(function () {
                if (viewSaveShown) {
                    baseShow.call(this);
                }
            }.bind(this), 2000);
            this.controller._reallyShown = false;
            if (!statusMessageOverridden) {
                this.clearStatusMessage();
            }
        });

        this.removeListeners = _.wrap(this.removeListeners, function (origFn) {
            origFn.apply(this, _.rest(arguments));
            removeListeners.apply(this, arguments);
        });

        this.setStatusMessage = _.wrap(this.setStatusMessage, function (origFn) {
            origFn.apply(this, _.rest(arguments));
            statusMessageOverridden = true;
        });

        this.clearStatusMessage = _.wrap(this.clearStatusMessage, function (origFn) {
            if (viewSaveShown && this.controller._reallyShown) {
                this.setStatusMessage(this.controller._options.locale.defaultStatusMessageWithViews);
            } else if (viewSaveShown) {
                this.setStatusMessage(this.controller._options.locale.defaultStatusMessageOnlyViews);
            } else {
                origFn.apply(this, _.rest(arguments));
            }
            statusMessageOverridden = false;
        });

        this.updateButtonStates = _.wrap(this.updateButtonStates, function (origFn) {
            origFn.apply(this, _.rest(arguments));
            if (this._suspendButtonState || this._suspendRedraws) { return; }
            var cancelEnabled = !this.buttons.$cancelButton.prop('disabled') && !!this.controller._reallyShown;
            this.buttons.$cancelButton.prop('disabled', !cancelEnabled);
        });
    };

    // In this function, the `this` context is a ChangeBarView
    var initListeners = function (controller) {
        this.buttons.$saveViewButton.on('click', function (e) {
            e.preventDefault();
            controller._saveView();
        });

        this.buttons.$saveViewAsButton.on('click', function (e) {
            e.preventDefault();
            if (!$(this).closest('li').hasClass('disabled')) {
                controller._saveViewAs();
            }
        });

        this.buttons.$resetViewButton.on('click', function (e) {
            e.preventDefault();
            if (!$(this).closest('li').hasClass('disabled')) {
                controller._resetView();
            }
        });
    };

    // In this function, the `this` context is a ChangeBarView
    var removeListeners = function () {
        this.buttons.$saveViewButton.off('click');
        this.buttons.$saveViewAsButton.off('click');
        this.buttons.$resetViewButton.off('click');
    };

    function MultiViewChangeBar(options, actions) {
        options = objectUtils.defaultsDeep({}, options, DEFAULT_OPTIONS, getDefaultStrings());
        actions = actions || {};

        ChangeBar.call(this, options, actions);
        this._subscribableOptions.allowedEvents = this._subscribableOptions.allowedEvents.concat(['saveView', 'saveViewAs', 'resetView']);

        if (_.isFunction(actions.saveView)) {
            this.on('saveView', actions.saveView);
        }
        if (_.isFunction(actions.saveViewAs)) {
            this.on('saveViewAs', actions.saveViewAs);
        }
        if (_.isFunction(actions.resetView)) {
            this.on('resetView', actions.resetView);
        }

        this.visibility.showViewSave = this.showViewSave.bind(this);
        this.visibility.hideViewSave = this.hideViewSave.bind(this);
        this.visibility.viewSaving = this.viewSaving.bind(this);
        this.visibility.viewSaveSucceeded = this.viewSaveSucceeded.bind(this);
        this.visibility.viewSaveFailed = this.viewSaveFailed.bind(this);

        extendView.call(this._view);
        initListeners.call(this._view, this);

        this._checkAllowSave = _.wrap(this._checkAllowSave, function (origFn) {
            return this._reallyShown && origFn.apply(this, _.rest(arguments));
        });
    }

    MultiViewChangeBar.prototype = _.create(ChangeBar.prototype, {
        constructor : MultiViewChangeBar,

        _saveView : function () {
            this.trigger('saveView');
        },

        _saveViewAs : function () {
            this.trigger('saveViewAs');
        },

        _resetView : function () {
            this.trigger('resetView');
        },

        showViewSave : function () {
            this._view.showViewSave();
            return this;
        },

        hideViewSave : function () {
            this._view.hideViewSave();
            return this;
        },

        toggleSaveViewAs : function (state) {
            this._view.toggleSaveViewAs(state);
            return this;
        },

        viewSaving : function () {
            this._view.viewSaving();
            return this;
        },

        viewSaveSucceeded : function () {
            this._view.viewSaveSucceeded();
            return this;
        },

        viewSaveFailed : function () {
            this._view.viewSaveFailed();
            return this;
        }
    });

    return MultiViewChangeBar;
});


define('primutils/markup/primutils/ui/detailPopovers/emptyPopover.hbs', ['handlebars','primutils.markup'], function (H, u) {
var imports = u.buildImported(Array.prototype.slice.call(arguments, 2), 'primutils/ui/detailPopovers/emptyPopover.hbs'),
    func = H.template({"compiler":[6,">= 2.0.0-beta.1"],"main":function(depth0,helpers,partials,data) {
  var stack1, lambda=this.lambda, helperMissing=helpers.helperMissing, buffer = "<div class=\"dropdown-panel disclosure object-detail-popover-panel\" title=\"\" role=\"dialog\" aria-labelledby=\"";
  stack1 = lambda((depth0 != null ? depth0.uniqueId : depth0), depth0);
  if (stack1 != null) { buffer += stack1; }
  buffer += "-label\">\n    <div style=\"display: none !important;\" id=\"";
  stack1 = lambda((depth0 != null ? depth0.uniqueId : depth0), depth0);
  if (stack1 != null) { buffer += stack1; }
  buffer += "-label\">";
  stack1 = lambda((depth0 != null ? depth0.loadingString : depth0), depth0);
  if (stack1 != null) { buffer += stack1; }
  buffer += "</div>\n    ";
  stack1 = ((helpers.icon || (depth0 && depth0.icon) || helperMissing).call(depth0, "throbber", (depth0 != null ? depth0.loadingString : depth0), {"name":"icon","hash":{
    'style': ("margin: 0 auto;"),
    'class': ("square")
  },"data":data}));
  if (stack1 != null) { buffer += stack1; }
  return buffer + "\n</div>";
},"useData":true});
return function (context, options) {
    options = u.extendOptions(options || {}, imports);
    return func(context, options).trim();
};
});

define('primutils/ui/detailPopovers/DetailPopover',[
    'jquery',
    'lodash',
    'logger',
    'primutils/utils/language',
    'primutils/utils/keyCode',
    'primutils/a11y/focusManager',
    'primutils/markup/primutils/ui/detailPopovers/emptyPopover.hbs'
], function ($, _, logger, lang, keyCodes, focusManager, popoverMarkup) {
    'use strict';

    var POPOVER_SHOW_DELAY = 1500;
    var POPOVER_HIDE_DELAY = 750;

    // Enabling this means that if you mouse back-forth over the link for POPOVER_SHOW_DELAY, it will still show instead of
    // waiting for the mouse to be stationary over the element for that time
    var ENABLE_RECATCH = false;
    var KEY_M = 77;

    function isKeyboardTriggered(e) {
        return (e.which === KEY_M && e.altKey);
    }

    function closePopovers() {
        $('.open[data-item-detail-popover]').each(function () {
            var popover = $(this).children('a').data('detail-popover');
            if (popover) {
                popover.hide();
            }
        });
    }

    var initListeners = function (mouseOverControlling) {
        var inControllingElement = mouseOverControlling,
            inPopover = false,
            recatch = false,
            $elements = this._$controllingElement.add(this._$el);

        this._$el.on('click', function (e) {
            e.stopPropagation();
        });

        this._$controllingElement.on('mouseenter mouseleave', function (e) {
            inControllingElement = (e.type === 'mouseenter');
            if (recatch) {
                this.show();
            }
        }.bind(this));

        this._$el.on('mouseenter mouseleave', function (e) {
            inPopover = (e.type === 'mouseenter');
        });

        var maybeShow = function () {
            if (inPopover || inControllingElement) {
                this.show();
            }
        }.bind(this);

        var maybeHide = function () {
            if (!inPopover && (!inControllingElement || this._$controllingElement.closest('html').length === 0)) {
                recatch = ENABLE_RECATCH;
                this.hide(true);
            }
        }.bind(this);

        var endQuickReshow = function () {
            recatch = false;
        }.bind(this);

        var throttledMaybeShow = _.throttle(maybeShow, POPOVER_SHOW_DELAY, { leading : false });

        $elements.on('mouseenter', throttledMaybeShow);
        $elements.on('mouseleave', throttledMaybeShow.cancel);
        $elements.on('mouseleave', _.throttle(maybeHide, POPOVER_HIDE_DELAY, { leading : false }));
        $elements.on('mouseleave', _.throttle(endQuickReshow, POPOVER_SHOW_DELAY + POPOVER_HIDE_DELAY, { leading : false }));

        $elements.on('keydown', function (e) {
            if (isKeyboardTriggered(e)) {
                var isOpen = this._$controllingElement.parent().hasClass('open');
                this[isOpen ? 'hide' : 'show']();
                if (!isOpen) {
                    this.bringFocus();
                }
            } else if (e.keyCode === keyCodes.ESCAPE) {
                this.hide();
            }
        }.bind(this));

        $elements.on('focusout', function () {
            $(document).one('focusin', function (e) {
                var $target = $(e.target);
                if ($target.closest(this._$el).length === 0 && $target.closest(this._$controllingElement).length === 0) {
                    closePopovers();
                }
            }.bind(this));
        }.bind(this));

        // Prevent scrolling in the "Other..." dropdown from scrolling grids or for clicking to select the grid row
        var eventName = 'wheel'; //Standardized event
        if (document.onmousewheel !== undefined && document.onwheel === undefined) {
            eventName = 'mousewheel'; //Only needed for Safari
        }
        this._$el.on(eventName + ' mousedown', function (e) {
            e.stopPropagation();
        });
    };

    function DetailPopover(options) {
        logger.assert(this instanceof DetailPopover, 'DetailPopover | constructor must be called as DetailPopover.apply(this, arguments)');
        logger.assert(_.has(options, '$controllingElement'), 'DetailPopover | constructor must be called as DetailPopover.apply(this, arguments)');
        logger.assert(_.has(options, 'mouseOverControlling'), 'DetailPopover | constructor must be called as DetailPopover.apply(this, arguments)');

        this._loadPromise = null;

        this._uniqueId = _.uniqueId('detail-popover-');

        this._$controllingElement = options.$controllingElement.data('detail-popover', this);
        this._$bodyContainer = $('<div style="position: relative; z-index: 1050;" />');
        this._$el = $(popoverMarkup({
            uniqueId : this._uniqueId,
            loadingString : lang.getString('detailPanels.loading')
        })).attr('id', this._uniqueId);
        this._$el.appendTo(this._$controllingElement.parent());

        this._$el.one('detail-popover-shown', function () {
            this.loadData().done(function () {
                this._displayData();
            }.bind(this));
        }.bind(this));

        initListeners.call(this, options.mouseOverControlling);
        if (options.mouseOverControlling) {
            this._$controllingElement.trigger('mouseenter');
        }
    }

    function shouldDropUp($popup, $parent, offset) {
        var popupHeight = $popup.outerHeight(true),
            parentHeight = $parent.outerHeight();
        return offset.top + popupHeight + parentHeight > window.pageYOffset + $(window).height();
    }

    function shouldPullRight($popup, $parent, offset) {
        var popupWidth = $popup.outerWidth(true);
        return offset.left + popupWidth > window.pageXOffset + $(window).width();
    }

    var positionAndShowPopover = function () {
        var $body = $(document.body),
            $parent = this._$controllingElement.parent(),
            originalStyle = $body.attr('style') || '',
            parentOffset = $parent.offset(),
            isDropUp, isPullRight,
            offsetLeft = 0, offsetTop = $parent.outerHeight(true),
            scrollTop = $body.scrollTop(),
            scrollLeft = $body.scrollLeft(),
            alreadyOpen = $parent.hasClass('open');

        this._$el.removeClass('pull-right');
        $parent.removeClass('dropup-for-panel');

        $body.css('overflow', 'hidden');

        $parent.addClass('open');
        this._$bodyContainer.addClass('open');
        $parent.toggleClass('dropup-for-panel', (isDropUp = shouldDropUp(this._$el, $parent, parentOffset)));
        this._$el.toggleClass('pull-right', (isPullRight = shouldPullRight(this._$el, $parent, parentOffset)));

        if (isDropUp) {
            offsetTop = -(this._$el.outerHeight(true));
        }

        if (isPullRight) {
            offsetLeft = $parent.outerWidth(true) - (this._$el.outerWidth(true));
        }

        this._$el.css({
            position : 'fixed',
            left : parentOffset.left + offsetLeft - scrollLeft,
            bottom : 'auto',
            top : parentOffset.top + offsetTop - scrollTop,
            right : 'auto'
        });

        $body.attr('style', originalStyle);

        if ($parent.is('[data-popover-append-body]')) {
            this._$bodyContainer.appendTo(document.body);
            this._$el.appendTo(this._$bodyContainer);
            if (!alreadyOpen) {
                focusManager.pushFocus(this._$el);
            }
        }

        this._$el.trigger('detail-popover-shown');
    };

    DetailPopover.prototype = {
        constructor : DetailPopover,

        show : function () {
            if (this._$controllingElement.parent().hasClass('open')) {
                return;
            }

            positionAndShowPopover.call(this);
            if (this._data) {
                this._redrawData();
            }
        },

        bringFocus : function () {
            this._$el.attr('tabindex', '-1').focus();
        },

        hide : function (noRefocus) {
            var $parent = this._$controllingElement.parent(),
                alreadyClosed = !$parent.hasClass('open');

            $parent.removeClass('open');
            this._$bodyContainer.removeClass('open');
            if (!alreadyClosed && this._$controllingElement.parent().is('[data-popover-append-body]')) {
                focusManager.popFocus(this._$el);
                this._$bodyContainer.detach();
            }

            this._$el.off('.popover-in-grid').removeAttr('tabindex');

            var e = $.Event('detail-popover-hidden');
            e.refocusAllowed = !noRefocus;
            this._$el.trigger(e);

            if (!noRefocus && !e.isDefaultPrevented()) {
                this._$controllingElement.focus();
            }
        },

        activateForGrid : function ($grid) {
            $grid.attr('aria-activedescendant', this._$el.attr('id'));
            this.bringFocus();
            this._$el.on('keydown.popover-in-grid keyup.popover-in-grid keypress.popover-in-grid', function (e) {
                e.stopPropagation();
            });
        },

        loadData : function () {
            if (!this._loadPromise) {
                this._loadPromise = this._load().then(function (data) {
                    return (this._data = data);
                }.bind(this));
            }

            return this._loadPromise;
        },

        _load : function () {
            return $.when();
        },

        _getMarkup : _.constant(''),

        _redrawData : _.noop,

        _initContent : _.noop,

        _displayData : function () {
            this._$el.empty().append(this._getMarkup());
            this._initContent();
        }
    };

    DetailPopover.closePopovers = closePopovers;
    DetailPopover.isKeyboardTriggered = isKeyboardTriggered;
    DetailPopover.hasPopover = function ($elem) {
        return !!DetailPopover.getPopover($elem);
    };
    DetailPopover.getPopover = function ($elem) {
        return $elem.data('detail-popover');
    };

    return DetailPopover;
});

define('primutils/ui/detailPopovers/detailPopoverActivator',[
    'require',
    'jquery',
    'lodash',
    'logger',
    'primutils/ui/detailPopovers/DetailPopover',
    'primutils/utils/grid'
], function (require, $, _, logger, DetailPopover, gridUtils) {
    'use strict';

    var GRID_SELECTOR = ':' + gridUtils.gridDataName;

    var that = {};

    that._attachScrollListener = _.once(function () {
        $(window).on('scroll', function () {
            DetailPopover.closePopovers();
        });
    });

    that.activatePopoversFor = function ($container, options) {
        logger.assert(_.has(options, 'factory'), 'DetailPopover | missing factory for call to activatePopoversFor');

        var descriptionId = _.uniqueId('object-detail-popover-description-');

        $('<div id="' + descriptionId + '" style="display: none !important;">' + options.description +
            '</div>').appendTo($container);

        that._attachScrollListener();

        function activatePopover($this, e) {
            var $parent = $this.parent(),
                hasPopover = DetailPopover.hasPopover($this),
                byKeyboard = e.type === 'keydown' && DetailPopover.isKeyboardTriggered(e),
                shouldCreate = e.type === 'mouseover' || byKeyboard;

            if (!hasPopover && shouldCreate) {
                var popover = options.factory($parent, $this, e.type === 'mouseover');

                if (byKeyboard) {
                    popover.show();
                    popover.bringFocus();
                }

                return popover;
            } else {
                return DetailPopover.getPopover($this);
            }
        }

        $container.on('mouseover keydown', '[data-item-detail-popover] > a', function (e) {
            activatePopover($(this), e);
        });
        $container.on('focus', '[data-item-detail-popover] > a', function () {
            $(this).attr('aria-describedby', descriptionId);
        });

        function setGridListener() {
            $container.on('keydown', GRID_SELECTOR, function (e) {
                var $this = $(this);
                var $popoverLink = $this.find('.cell-focused [data-item-detail-popover] > a');
                if ($popoverLink.length === 0) {
                    DetailPopover.closePopovers();
                } else if (DetailPopover.isKeyboardTriggered(e)) {
                    var popover = activatePopover($popoverLink, e);
                    popover.show();
                    popover.activateForGrid($this);
                }
            });
        }
        if (gridUtils.gridApiLoaded()) {
            setGridListener();
        } else {
            gridUtils.whenGridApiLoaded(function () {
                setGridListener();
            });
        }

        $container.on('keydown', '.orcl-picker', function (e) {
            var $this = $(this);
            var $target = $(e.target);
            var $popoverLink = $this.find('.item.active [data-item-detail-popover] > a');

            if ($target.closest('[data-item-detail-popover]').length) {
                return;
            }

            if ($popoverLink.length === 0) {
                DetailPopover.closePopovers();
            } else if (DetailPopover.isKeyboardTriggered(e)) {
                var popover = activatePopover($popoverLink, e);
                popover.show();
                popover.bringFocus();
            }
        });

        function refocusGrid($grid) {
            $grid.data(gridUtils.gridDataName).requestBrowserFocus(true);
        }

        $container.on('detail-popover-shown', function (e) {
            var $popover = $(e.target);

            if (gridUtils.gridApiLoaded()) {
                var $closestGridRowset = $popover.closest('.grid-rowset');
                if ($closestGridRowset.length) {
                    $closestGridRowset.addClass('grid-popover-visible');
                }
            }
        });

        $container.on('detail-popover-hidden', function (e) {
            var $popover = $(e.target);

            var isGridLoaded = gridUtils.gridApiLoaded();
            var isOuiLoaded = require.defined('oui');

            if (isGridLoaded) {
                var $closestGridRowset = $popover.closest('.grid-rowset');
                if ($closestGridRowset.length) {
                    $closestGridRowset.removeClass('grid-popover-visible');
                }
            }

            if (!e.refocusAllowed) {
                return;
            }

            if (isGridLoaded) {
                var $closestGrid = $popover.closest(GRID_SELECTOR);
                if ($closestGrid.length) {
                    refocusGrid($closestGrid);
                    e.preventDefault();
                }
            }

            var $closestTreeItem = $popover.closest('[data-treeid]'),
                $treeItems = $popover.closest('.tree-items'),
                $tree = $treeItems.closest('.orcl-widget.tree'),
                $picker = $tree.closest('.orcl-picker');

            if ($treeItems.length && isOuiLoaded) {
                _.defer(function () {
                    var oui = require('oui');
                    $tree.focus();
                    var picker = oui.get($picker, 'picker');
                    var treePanel = oui.get($tree, 'tree-panel');
                    if (picker) {
                        treePanel = picker.treePanel();
                    }
                    if (treePanel) {
                        treePanel._controller.active($closestTreeItem.attr('data-treeid'));
                    }
                });
                e.preventDefault();
            }
        });
    };

    return that;
});


define('primutils/markup/primutils/ui/dialogs/confirm.hbs', ['handlebars','primutils.markup','primutils/templates/helpers/modal','primutils/templates/helpers/button','primutils/templates/helpers/icon','primutils/templates/helpers/join','primutils/templates/helpers/ternary'], function (H, u) {
var imports = u.buildImported(Array.prototype.slice.call(arguments, 2), 'primutils/ui/dialogs/confirm.hbs'),
    func = H.template({"1":function(depth0,helpers,partials,data) {
  var stack1, buffer = "    ";
  stack1 = helpers.modalHeader.call(depth0, {"name":"modalHeader","hash":{},"data":data});
  if (stack1 != null) { buffer += stack1; }
  buffer += "\r\n";
  stack1 = helpers.modalBody.call(depth0, {"name":"modalBody","hash":{
    'id': ("confirm-modal-message")
  },"fn":this.program(2, data),"inverse":this.noop,"data":data});
  if (stack1 != null) { buffer += stack1; }
  stack1 = helpers.modalFooter.call(depth0, {"name":"modalFooter","hash":{},"fn":this.program(7, data),"inverse":this.noop,"data":data});
  if (stack1 != null) { buffer += stack1; }
  return buffer;
},"2":function(depth0,helpers,partials,data) {
  var stack1, lambda=this.lambda, buffer = "";
  stack1 = helpers['if'].call(depth0, (depth0 != null ? depth0.hasIcon : depth0), {"name":"if","hash":{},"fn":this.program(3, data),"inverse":this.noop,"data":data});
  if (stack1 != null) { buffer += stack1; }
  stack1 = helpers['if'].call(depth0, (depth0 != null ? depth0.uiContext : depth0), {"name":"if","hash":{},"fn":this.program(5, data),"inverse":this.noop,"data":data});
  if (stack1 != null) { buffer += stack1; }
  buffer += "        <p class=\"break-long-words\">";
  stack1 = lambda((depth0 != null ? depth0.message : depth0), depth0);
  if (stack1 != null) { buffer += stack1; }
  return buffer + "</p>\r\n";
},"3":function(depth0,helpers,partials,data) {
  var stack1, buffer = "            <span class=\"modal-icon\">";
  stack1 = helpers.icon.call(depth0, (depth0 != null ? depth0.iconType : depth0), "", {"name":"icon","hash":{},"data":data});
  if (stack1 != null) { buffer += stack1; }
  return buffer + "</span>\r\n";
},"5":function(depth0,helpers,partials,data) {
  var stack1, lambda=this.lambda, buffer = "            <h4 class=\"ui-context\">";
  stack1 = lambda((depth0 != null ? depth0.uiContext : depth0), depth0);
  if (stack1 != null) { buffer += stack1; }
  return buffer + "</h4>\r\n";
},"7":function(depth0,helpers,partials,data) {
  var stack1, buffer = "        ";
  stack1 = helpers.button.call(depth0, "secondary", {"name":"button","hash":{
    'data-dismiss': ("modal"),
    'data-default': ("esc"),
    'class': ("cancelModal")
  },"fn":this.program(8, data),"inverse":this.noop,"data":data});
  if (stack1 != null) { buffer += stack1; }
  buffer += "\r\n        ";
  stack1 = helpers.button.call(depth0, "primary", {"name":"button","hash":{
    'data-dismiss': ("modal"),
    'data-default': ("enter"),
    'class': ("proceedModal")
  },"fn":this.program(10, data),"inverse":this.noop,"data":data});
  if (stack1 != null) { buffer += stack1; }
  return buffer + "\r\n";
},"8":function(depth0,helpers,partials,data) {
  var stack1, lambda=this.lambda;
  stack1 = lambda((depth0 != null ? depth0.cancel : depth0), depth0);
  if (stack1 != null) { return stack1; }
  else { return ''; }
  },"10":function(depth0,helpers,partials,data) {
  var stack1, lambda=this.lambda;
  stack1 = lambda((depth0 != null ? depth0.accept : depth0), depth0);
  if (stack1 != null) { return stack1; }
  else { return ''; }
  },"compiler":[6,">= 2.0.0-beta.1"],"main":function(depth0,helpers,partials,data) {
  var stack1, buffer = "";
  stack1 = helpers.modal.call(depth0, (depth0 != null ? depth0.title : depth0), {"name":"modal","hash":{
    'class': (helpers._.call(depth0, helpers['?'].call(depth0, (depth0 != null ? depth0.hasIcon : depth0), "pgbu-modal-with-icon ", "", {"name":"?","hash":{},"data":data}), "confirm-modal", {"name":"_","hash":{},"data":data})),
    'id': ("confirm-modal")
  },"fn":this.program(1, data),"inverse":this.noop,"data":data});
  if (stack1 != null) { buffer += stack1; }
  return buffer;
},"useData":true});
return function (context, options) {
    options = u.extendOptions(options || {}, imports);
    return func(context, options).trim();
};
});


/**
 * Displays a custom message to the user to confirm whether or not they really want to perform the action which triggers this ConfirmDialog.
 * Contrary to most modals, ConfirmDialog does not provide a default `message`, so you will always want to configure that option.
 *
 * By specifying the `doNotShowAgain` option with a key specific to your modal, the end-user will be able to indicate that they do not want to see this modal again (in which case it will be auto-confirmed).
 *
 * To add a handler for when the user clicks "yes", see the `ok` method or `okCallback` property.
 * To add a handler for when the user clicks "no" or closes the modal, see the `cancel` method or `cancelCallback` property.
 *
 * @class ui.dialogs.ConfirmDialog
 * @module ui.dialogs
 * @param options {Object} See properties tab
 * @constructor
 */
define('primutils/ui/dialogs/ConfirmDialog',[
    'jquery',
    'lodash',
    'logger',
    'primutils/markup/primutils/ui/dialogs/confirm.hbs',
    'primutils/markup/primutils/ui/dialogs/doNotShow.hbs',
    'primutils/utils/language',
    'primutils/ui/dialogs/doNotShowHelper',
    'primutils/ui/dialogs/dialogTypes',
    'primutils/a11y/accessibilityUtils'
], function ($, _, logger, confirmTemplate, doNotShowTemplate, lang, doNotShowHelper, dialogTypes, accessibilityUtils) {
    'use strict';

    var DEFAULT_OPTIONS = {
        /**
         * A key stored in local storage which can prevent a particular dialog from displaying again.
         * If the user checks "Do not show again" and clicks Ok, the <a href="#property_okCallback">okCallback</a> will
         * automatically be called on subsequent calls to dialogs using the same key.
         *
         * Using `false` means the dialog will always be shown.
         *
         * @property doNotShowAgain {String|Boolean}
         * @optional
         * @default false
         */
        doNotShowAgain : false,
        /**
         * A method which will be called when the user clicks the Ok button.
         *
         * The callback takes no arguments.
         *
         * @property okCallback {Function}
         * @optional
         */
        //okCallback,
        /**
         * A method which will be called when the user clicks the Cancel button. This method will also be called if the
         * user dismisses the dialog by another means like the Close (X) button or escape.
         *
         * The callback takes no arguments.
         *
         * @property cancelCallback {Function}
         * @optional
         */
        //cancelCallback,
        /**
         * Information about what the UI was doing when the dialog needed to be shown.
         *
         * This is a brief sentence fragment that describes what the user was attempting to do when the dialog was shown.
         * This differs from the `message` which shows the cause and potential resolution steps to the users problem.
         *
         * @property uiContext {String}
         * @optional
         */
        //uiContext,
        /**
         * The type of the dialog.
         *
         * The type of the dialog controls the default title, but it is overridable.
         *
         * Available types are exposed by the static `types` property.
         *
         * @property type
         * @type {ui.dialogs.dialogType}
         * @default CONFIRM
         */
        type : dialogTypes.types.CONFIRM
    };

    function getDefaultStrings() {
        return {
            /**
             * The title of the dialog.
             * @property title {String}
             * @optional
             * @default Depends on the dialog type, see `type` property
             */
            //title,
            /**
             * The message to display inside the dialog.
             * @property message {String}
             * @optional
             * @default ""
             */
            message : '',
            /**
             * The text to display on the default Yes button
             * @property acceptLabel {String}
             * @optional
             * @default "Yes" - localized
             */
            acceptLabel : lang.getString('dialogs.yes'),
            /**
             * The text to display on the No button
             * @property cancelLabel {String}
             * @optional
             * @default "No" - localized
             */
            cancelLabel : lang.getString('dialogs.no')
        };
    }

    function bindOptionsCallbacks(options, dfd) {
        if (_.isFunction(options.okCallback)) {
            dfd.done(options.okCallback);
        }

        if (_.isFunction(options.cancelCallback)) {
            dfd.fail(options.cancelCallback);
        }
    }

    function ConfirmDialog(options) {
        var that = this,
            dfd = $.Deferred(),
            promise = dfd.promise();
        options = _.defaults({}, options, DEFAULT_OPTIONS, getDefaultStrings(), {
            title : dialogTypes.titles[options.type || DEFAULT_OPTIONS.type]
        });

                if (!options.message) {
            logger.warn('message property is falsey - no default message is provided, but you probably want to provide one');
        }
        
        if (!_.isString(options.doNotShowAgain)) {
            options.doNotShowAgain = false;
        }

        /**
         * Adds a method which will be called when the user clicks the Ok button.
         *
         * The callback takes no arguments.
         *
         * @method ok
         * @param fn {Function} The callback function
         * @chainable
         * @returns {ConfirmDialog} The confirm dialog instance
         */
        this.ok = promise.done;
        /**
         * Adds a method which will be called when the user clicks the Cancel button. This method will also be called if
         * the user dismisses the dialog by another means like the Close (X) button or escape.
         *
         * The callback takes no arguments.
         *
         * @method cancel
         * @param fn {Function} The callback function
         * @chainable
         * @returns {ConfirmDialog} The confirm dialog instance
         */
        this.cancel = promise.fail;

        if (options.doNotShowAgain !== false && doNotShowHelper.isDoNotShow(options.doNotShowAgain)) {
            dfd.resolve();
            bindOptionsCallbacks(options, dfd);
            return; // Don't bother constructing the rest of the dialog if we're not showing it
        }

        this.$modal = $(confirmTemplate({
            title : options.title,

            message : options.message,
            uiContext : options.uiContext,

            accept : options.acceptLabel,
            cancel : options.cancelLabel,
            close : lang.getString('dialogs.close'),

            hasIcon : options.type !== dialogTypes.types.NONE,
            iconType : dialogTypes.icons[options.type]
        }));

        if (options.doNotShowAgain !== false) {
            this.$modal.find('.modal-footer').prepend(doNotShowTemplate({
                doNotShowAgain : lang.getString('dialogs.dontShowAgain')
            }));
        }

        this.$modal.find('.proceedModal').on('click', function (event) {
            if (options.doNotShowAgain !== false) {
                doNotShowHelper.toggleDoNotShow(options.doNotShowAgain, that.$modal.find('#doNotShowAgain').prop('checked'));
            }

            dfd.resolve();

            event.stopPropagation();
        });

        this.$modal.find('.cancelModal, [data-dismiss=modal]').on('click', function (event) {
            // Do not toggle the do not show
            dfd.reject({
                fromCloseButton : !$(this).hasClass('cancelModal')
            });
            event.preventDefault();
            event.stopPropagation();
        });

        this.$modal.on('shown', function () {
            $('.proceedModal').focus();
        });

        this.$modal.one('hidden', function () {
            that.$modal.remove();
        });

        dfd.always(function () {
            that.$modal.modal('hide');
        });

        bindOptionsCallbacks(options, dfd);

        accessibilityUtils.modalAppend(this.$modal);

        this.$modal.modal({
            show : (dfd.state() === 'pending'),
            backdrop : 'static'
        });
    }

    ConfirmDialog.prototype = {
        constructor : ConfirmDialog
    };

    /**
     * The types available to this dialog.
     *
     * See also the `type` property and [dialogTypes](ui.dialogs.dialogType.html).
     *
     * @property types
     * @static
     */
    Object.defineProperty(ConfirmDialog, 'types', {
        get : function () {
            return dialogTypes.types;
        }
    });

    return ConfirmDialog;
});


/**
 * Displays a message to the user asking them if they would like to cancel their changes.
 * The user can either cancel (by clicking "yes") or not cancel their changes (by clicking "no").
 *
 * By specifying the `doNotShowAgain` option with a key specific to your modal, the end-user will be able to indicate that they do not want to see this modal again (in which case it will be auto-confirmed).
 *
 * To add a handler for when the user clicks "yes", see the `yes` method or `yesCallback` property.
 * To add a handler for when the user clicks "no" or closes the modal, see the `no` method or `noCallback` property.
 *
 * @class ui.dialogs.CancelDialog
 * @module ui.dialogs
 * @param options {Object} See properties tab
 * @constructor
 */
define('primutils/ui/dialogs/CancelDialog',[
    'jquery',
    'lodash',
    'primutils/ui/dialogs/ConfirmDialog',
    'primutils/ui/dialogs/dialogTypes',
    'primutils/utils/language'
], function ($, _, ConfirmDialog, dialogTypes, lang) {
    'use strict';

    var DEFAULT_OPTIONS = {
        /**
         * A key stored in local storage which can prevent a particular dialog from displaying again.
         * If the user checks "Do not show again", the <a href="#property_okCallback">okCallback</a> will automatically
         * be called on subsequent calls to dialogs using the same key.
         *
         * Using `false` means the dialog will always be shown.
         *
         * @property doNotShowAgain {String|Boolean}
         * @optional
         * @default false
         */
        doNotShowAgain : false,
        /**
         * A method which will be called when the user clicks the Yes button.
         *
         * The callback takes no arguments.
         *
         * @property yesCallback {Function}
         * @optional
         */
        //yesCallback,
        /**
         * A method which will be called when the user clicks the No button. This method will also be called if the
         * user dismisses the dialog by another means like the Close (X) button or escape.
         *
         * The callback takes no arguments.
         *
         * @property noCallback {Function}
         * @optional
         */
        //noCallback,
        /**
         * The type of the dialog.
         *
         * The type of the dialog controls the default title, but it is overridable.
         *
         * Available types are exposed by the static `types` property.
         *
         * @property type
         * @type {ui.dialogs.dialogType}
         * @default WARNING
         */
        type : dialogTypes.types.WARNING
    };

    function getDefaultStrings() {
        return {
            /**
             * The title of the dialog.
             * @property title {String}
             * @optional
             * @default Depends on the dialog type, see `type` property
             */
            title : lang.getString('dialogTitles.cancel'),
            /**
             * Information about what the UI was doing when the dialog needed to be shown.
             *
             * This is a brief sentence fragment that describes what the user was attempting to do when the dialog was shown.
             * This differs from the `message` which shows the cause and potential resolution steps to the users problem.
             *
             * @property uiContext {String}
             * @optional
             */
            uiContext : lang.getString('dialogs.cancelContext'),
            /**
             * The message to display inside the dialog.
             * @property message {String}
             * @optional
             * @default "Are you sure you want to cancel your changes?" - localized
             */
            message : lang.getString('dialogs.cancelMessage'),
            /**
             * The text to display on the default Yes button
             * @property yes {String}
             * @optional
             * @default "Yes" - localized
             */
            yes : lang.getString('dialogs.yes'),
            /**
             * The text to display on the No button
             * @property no {String}
             * @optional
             * @default "No" - localized
             */
            no : lang.getString('dialogs.no')
        };
    }

    function CancelDialog(options) {
        // Map options to Confirm Dialog names
        var defaultStrings = getDefaultStrings(),
            confirmOptions = {
                doNotShowAgain : options.doNotShowAgain || DEFAULT_OPTIONS.doNotShowAgain,
                okCallback : options.yesCallback,
                cancelCallback : options.noCallback,
                title : options.title || defaultStrings.title,
                message : options.message || defaultStrings.message,
                acceptLabel : options.yes || defaultStrings.yes,
                cancelLabel : options.no || defaultStrings.no,
                uiContext : options.uiContext || defaultStrings.uiContext,
                type : options.type || DEFAULT_OPTIONS.type
            };

        ConfirmDialog.call(this, confirmOptions);

        /**
         * Adds a method which will be called when the user clicks the Yes button.
         *
         * The callback takes no arguments.
         *
         * @method yes
         * @param fn {Function} The callback function
         * @chainable
         * @returns {CancelDialog} The cancel dialog instance
         */
        this.yes = this.ok;
        /**
         * Adds a method which will be called when the user clicks the No button. This method will also be called if the
         * user dismisses the dialog by another means like the Close (X) button or escape.
         *
         * The callback takes no arguments.
         *
         * @method no
         * @param fn {Function} The callback function
         * @chainable
         * @returns {CancelDialog} The cancel dialog instance
         */
        this.no = this.cancel;
    }

    CancelDialog.prototype = _.assign(Object.create(ConfirmDialog.prototype), {
        constructor : CancelDialog
    });

    /**
     * The types available to this dialog.
     *
     * See also the `type` property and [dialogTypes](ui.dialogs.dialogType.html).
     *
     * @property types
     * @static
     */
    Object.defineProperty(CancelDialog, 'types', {
        get : function () {
            return dialogTypes.types;
        }
    });

    return CancelDialog;
});


define('primutils/markup/primutils/ui/dialogs/loadingDialog.hbs', ['handlebars','primutils.markup'], function (H, u) {
var imports = u.buildImported(Array.prototype.slice.call(arguments, 2), 'primutils/ui/dialogs/loadingDialog.hbs'),
    func = H.template({"compiler":[6,">= 2.0.0-beta.1"],"main":function(depth0,helpers,partials,data) {
  var stack1, lambda=this.lambda, buffer = "<div class=\"page-spinner page-loading-spinner page-spinner-static\" aria-labelledby=\"loading-modal-label\">\n    <div class=\"spinner-label\" id=\"loading-modal-label\">";
  stack1 = lambda((depth0 != null ? depth0.loading : depth0), depth0);
  if (stack1 != null) { buffer += stack1; }
  return buffer + "</div>\n    <div class=\"spinner-large loading-spinner-large\">\n        <div class=\"dot dot1\"></div>\n        <div class=\"dot dot2\"></div>\n        <div class=\"dot dot3\"></div>\n    </div>\n</div>";
},"useData":true});
return function (context, options) {
    options = u.extendOptions(options || {}, imports);
    return func(context, options).trim();
};
});


/**
 * Creates a loading spinner over a particular component to signify that the content fetched may take a while to load.
 *
 * Unlike the <a href="ui.dialogs.SavingDialog.html">SavingDialog</a>, the LoadingDialog does not create a mask over
 * the page content. The user is free to navigate away from loading content at any time.
 *
 * Example:
 *
 *     require(['jquery', 'primutils/ui/dialogs/LoadingDialog'], function ($, LoadingDialog) {
 *
 *         // creating a new spinner showing the defaults
 *         var spinner = new LoadingDialog({
 *             autoDisplay : true,
 *             target : $('.my-container')
 *         });
 *
 *         $.get('mystuff')
 *             .done(function (result) {
 *                 // do something with result
 *             })
 *             // always close the spinner, even if AJAX call fails
 *             .always(spinner.complete);
 *     });
 *
 * @class ui.dialogs.LoadingDialog
 * @module ui.dialogs
 */
define('primutils/ui/dialogs/LoadingDialog',[
    'jquery',
    'lodash',
    'logger',
    'primutils/markup/primutils/ui/dialogs/loadingDialog.hbs',
    'primutils/utils/language'
], function ($, _, logger, loadingSpinnerTemplate, lang) {
    'use strict';

    var defaultTarget = $('body');

    var DEFAULT_OPTIONS = {
        /**
         * Whether to automatically call `show` when the spinner is created
         *
         * @property autoDisplay {Boolean}
         * @optional
         * @default true
         */
        autoDisplay : true
        /**
         * The jQuery element to display the spinner over
         *
         * @property target {jQuery Object}
         * @optional
         * @default $('body')
         */
        //target
        /**
         * The text to place above the spinner
         *
         * @propery loading {String}
         * @optional
         * @default "Loading" - localized
         */
        //loading
    };

    function calculateOffset($target) {
        var $window = $(window),
            scrollTop = $window.scrollTop(),
            windowHeight = $window.height(),
            offsetTop = $target.offset().top,
            height = $target.height(),
            top = _.max([scrollTop - offsetTop, 0]),
            bottom = _.min([windowHeight + scrollTop - offsetTop, height]);
        return ((((bottom - top) / 2) + top) / height) * 100;
    }

    function LoadingDialog(options) {
        this.options = _.defaults({}, options, DEFAULT_OPTIONS, {
            // Need to specify this here, so it pulls the latest from "global"
            target : defaultTarget
        });

        // So that stop can be passed into promise.always / promise.then.
        this.complete = _.bind(this.complete, this);
        this.done = function () {
                        logger.warn('LoadingDialog.done() is deprecated.  Use .complete() instead.  .done() may be removed without further warning.  Full docs at: http://10.140.183.183:8080/job/primutils/javadoc/classes/ui.dialogs.LoadingDialog.html');
                        this.complete();
        }.bind(this);

        if (this.options.autoDisplay) {
            this.show();
        }
    }

    LoadingDialog.prototype = {
        constructor : LoadingDialog,

        /**
         * Shows the spinners over the current target(s)
         *
         * @method show
         */
        show : function () {
            var that = this,
                $spinner = $();
            this.options.target.each(function () {
                var $this = $(this);

                // Need to set position
                // The $target of the spinner needs to be positioned something non-"static" in order for the spinner to
                // be positioned properly.
                // TODO be smarter about this. Store original position in data so that the proper value can be restored
                // if it was set in inline CSS
                if ($this.css('position') === 'static') {
                    $this.css('position', 'relative');
                    $this.data('setPosition', true);
                }

                var $result = $(loadingSpinnerTemplate({
                    loading : that.options.loading || lang.getString('dialogTitles.loading')
                }))
                    .css({
                        display : 'none',
                        top : calculateOffset($this) + '%'
                    })
                    .appendTo($this)
                    .fadeIn();

                $spinner = $spinner.add($result);
            });

            this.$spinner = $spinner;
        },

        /**
         * Tells the spinner that it is complete and should hide and remove itself.
         *
         * *Note*: `complete` is bound to `this` on instantiation as a convenience.
         * This means you can write code like:
         *
         *     var loading = new LoadingDialog(...);
         *     $.ajax(...)
         *         .always(loading.complete);
         *
         * @method complete
         */
        complete : function () {
            var that = this;
            this.$spinner.fadeOut(function () {
                $(this).remove();
                that.options.target.each(function () {
                    var $this = $(this);
                    // unset position if it was set
                    if ($this.data('setPosition')) {
                        $this.css('position', '').removeData('setPosition');
                    }
                });
            });
            this.$spinner = $();
        }
    };

    /**
     * Sets the default target for all created loading dialogs.
     *
     * Useful when your application has a main content area that isn't the document body
     *
     * @method setDefaultTarget
     * @static
     * @param $target {jQuery Object} The default target of the loading dialog
     */
    LoadingDialog.setDefaultTarget = function ($target) {
        defaultTarget = $target;
    };

    return LoadingDialog;
});


define('primutils/markup/primutils/ui/dialogs/promptMarkup.hbs', ['handlebars','primutils.markup','primutils/templates/helpers/modal','primutils/templates/helpers/button','primutils/templates/helpers/controlGroup','primutils/templates/helpers/input','primutils/templates/helpers/join'], function (H, u) {
var imports = u.buildImported(Array.prototype.slice.call(arguments, 2), 'primutils/ui/dialogs/promptMarkup.hbs'),
    func = H.template({"1":function(depth0,helpers,partials,data) {
  var stack1, buffer = "    ";
  stack1 = helpers.modalHeader.call(depth0, {"name":"modalHeader","hash":{},"data":data});
  if (stack1 != null) { buffer += stack1; }
  buffer += "\r\n";
  stack1 = helpers.modalBody.call(depth0, {"name":"modalBody","hash":{},"fn":this.program(2, data),"inverse":this.noop,"data":data});
  if (stack1 != null) { buffer += stack1; }
  stack1 = helpers.modalFooter.call(depth0, {"name":"modalFooter","hash":{},"fn":this.program(5, data),"inverse":this.noop,"data":data});
  if (stack1 != null) { buffer += stack1; }
  return buffer;
},"2":function(depth0,helpers,partials,data) {
  var stack1, lambda=this.lambda, buffer = "        <div class=\"short-message\">";
  stack1 = lambda((depth0 != null ? depth0.directions : depth0), depth0);
  if (stack1 != null) { buffer += stack1; }
  buffer += "</div>\r\n";
  stack1 = helpers.controlGroup.call(depth0, (depth0 != null ? depth0.message : depth0), {"name":"controlGroup","hash":{},"fn":this.program(3, data),"inverse":this.noop,"data":data});
  if (stack1 != null) { buffer += stack1; }
  return buffer;
},"3":function(depth0,helpers,partials,data) {
  var stack1, buffer = "            ";
  stack1 = helpers.input.call(depth0, "text", {"name":"input","hash":{
    'hintHidden': (true),
    'hintId': ("prompt-help"),
    'hint': ((depth0 != null ? depth0.helpText : depth0)),
    'maxlength': ((depth0 != null ? depth0.maxLength : depth0)),
    'class': ("prompt-text input-block-level"),
    'id': ("prompt-input")
  },"data":data});
  if (stack1 != null) { buffer += stack1; }
  return buffer + "\r\n";
},"5":function(depth0,helpers,partials,data) {
  var stack1, buffer = "        ";
  stack1 = helpers.button.call(depth0, "secondary", {"name":"button","hash":{
    'data-dismiss': ("modal"),
    'data-default': ("esc"),
    'class': ("cancelModal")
  },"fn":this.program(6, data),"inverse":this.noop,"data":data});
  if (stack1 != null) { buffer += stack1; }
  buffer += "\r\n        ";
  stack1 = helpers.button.call(depth0, "primary", {"name":"button","hash":{
    'data-default': ("enter"),
    'class': ("proceedModal")
  },"fn":this.program(8, data),"inverse":this.noop,"data":data});
  if (stack1 != null) { buffer += stack1; }
  return buffer + "\r\n";
},"6":function(depth0,helpers,partials,data) {
  var stack1, lambda=this.lambda;
  stack1 = lambda((depth0 != null ? depth0.cancel : depth0), depth0);
  if (stack1 != null) { return stack1; }
  else { return ''; }
  },"8":function(depth0,helpers,partials,data) {
  var stack1, lambda=this.lambda;
  stack1 = lambda((depth0 != null ? depth0.accept : depth0), depth0);
  if (stack1 != null) { return stack1; }
  else { return ''; }
  },"compiler":[6,">= 2.0.0-beta.1"],"main":function(depth0,helpers,partials,data) {
  var stack1, buffer = "";
  stack1 = helpers.modal.call(depth0, (depth0 != null ? depth0.title : depth0), {"name":"modal","hash":{
    'class': ("prompt-modal"),
    'id': ("prompt-modal")
  },"fn":this.program(1, data),"inverse":this.noop,"data":data});
  if (stack1 != null) { buffer += stack1; }
  return buffer;
},"useData":true});
return function (context, options) {
    options = u.extendOptions(options || {}, imports);
    return func(context, options).trim();
};
});


/**
 * Prompts the user for a value with a textual input.
 * You can supply a message to the user and provide a default value.
 * The user can then either choose to submit the value or cancel entering a value.
 *
 * To add a handler for when the user has entered text and clicks "ok", see the `ok` method or `okCallback` property.
 * To add a handler for when the user clicks "cancel" or closes the modal, see the `cancel` method or `cancelCallback` property.
 *
 * @class ui.dialogs.PromptDialog
 * @module ui.dialogs
 * @param options {Object} See properties
 * @constructor
 */
define('primutils/ui/dialogs/PromptDialog',[
    'jquery',
    'lodash',
    'primutils/markup/primutils/ui/dialogs/promptMarkup.hbs',
    'primutils/utils/keyCode',
    'primutils/utils/language',
    'primutils/a11y/accessibilityUtils'
], function ($, _, confirmTemplate, keyCode, lang, accessibilityUtils) {
    'use strict';

    var DEFAULT_OPTIONS = {
        /**
         * Whether or not the user must entire a value in the dialog before proceeding.
         *
         * @property required {Boolean}
         * @optional
         * @default false
         */
        required : false
        /**
         * A callback to check whether the value entered in the dialog is unique. A unique check is only run when the
         * user attempts to submit the value.
         *
         * The callback takes one parameter, the value entered in the box.
         *
         *     checkUnique : function (value) { ... }
         *
         * The return value can either be an object the returns the following properties or a `$.when` compatible
         * promise that resolves to the same. `message` is optional and will default to the value of the `helpText` option.
         *
         *     return {
         *         result : true // or false; Whether this item already exists, i.e. returns true when the item is *NOT* unique
         *         message : ''  // or missing; A message to show the user when the violated the constraint.
         *     };
         *
         * You can also return a simple `true` or `false` which matches `result` in the object above.
         *
         * @property checkUnique {Function}
         * @optional
         */
        //checkUnique,
        /**
         * A method which will be called when the user clicks the Ok button. This method will *not* be called if the
         * user closes the modal through the Close (X) button in the corner or with the escape button.
         *
         * The callback takes one arguments, the value entered in the box.
         *
         *     okCallback : function (value) { ... }
         *
         * @property okCallback {Function}
         * @optional
         */
        //okCallback,
        /**
         * A method which will be called when the user clicks the Cancel button. This method will also be called if the
         * user dismisses the dialog by another means like the Close (X) button or escape.
         *
         * The callback takes no arguments.
         *
         * @property cancelCallback {Function}
         * @optional
         */
        //cancelCallback
    };

    function getDefaultStrings() {
        return {
            /**
             * The title of the dialog.
             * @property title {String}
             * @optional
             * @default "Prompt" - localized
             */
            title : lang.getString('dialogTitles.prompt'),
            /**
             * The initial value of the text box.
             * @property defaultText {String}
             * @optional
             * @default ""
             */
            defaultText : '',
            /**
             * The label to display above the text box.
             * @property message {String}
             * @optional
             * @default ""
             */
            message : '',
            /**
             * A short paragraph explaining the purpose of the prompt. Displays above the text box and label.
             * @property directions {String}
             * @optional
             * @default ""
             */
            directions : '',
            /**
             * The default error message to show when `checkUnique` or `required` constraints are not met.
             * @property helpText {String}
             * @optional
             * @default ""
             */
            //helpText,
            /**
             * The text to display on the default Ok button
             * @property acceptLabel {String}
             * @optional
             * @default "Ok" - localized
             */
            acceptLabel : lang.getString('dialogs.ok'),
            /**
             * The text to display on the Cancel button
             * @property cancelLabel {String}
             * @optional
             * @default "Cancel" - localized
             */
            cancelLabel : lang.getString('dialogs.cancel'),
            /**
             * The max length of the value in the text box.
             *
             * @property maxLength {Number|String}
             * @optional
             */
            maxLength : ''
        };
    }

    function PromptDialog(options) {
        var that = this,
            dfd = $.Deferred(),
            promise = dfd.promise();

        options = _.defaults({}, options, DEFAULT_OPTIONS, getDefaultStrings());

        /**
         * Adds a method which will be called when the user clicks the Ok button.
         *
         * The callback takes one arguments, the value entered in the box.
         *
         *     function (value) { ... }
         *
         * @method ok
         * @param fn {Function} The callback function
         * @chainable
         * @returns {PromptDialog} The prompt dialog instance
         */
        this.ok = promise.done;
        /**
         * Adds a method which will be called when the user clicks the Cancel button. This method will also be called if
         * the user dismisses the dialog by another means like the Close (X) button or escape.
         *
         * The callback takes no arguments.
         *
         * @method cancel
         * @param fn {Function} The callback function
         * @chainable
         * @returns {PromptDialog} The prompt dialog instance
         */
        this.cancel = promise.fail;

        this.$modal = $(confirmTemplate({
            close : lang.getString('dialogs.close'),
            title : options.title,
            helpText : options.helpText || '',
            message : options.message,
            directions : options.directions,
            accept : options.acceptLabel,
            cancel : options.cancelLabel,
            maxLength : options.maxLength
        }));

        if (options.required) {
            this.$modal.find('.control-group').addClass('required');
            this.$modal.find('.prompt-text').attr('required', 'required').attr('aria-required', 'true').attr('aria-describedby', 'prompt-help');
        }

        function getPromptValue() {
            var val = that.$modal.find('.prompt-text').val();
            // The backend trims every string coming from the UI, without fail.
            // Whitespace is not a valid string value for any non-nullable, min-size 1 string in the application,
            // regardless of area. This is definitely true in Prime and is a good choice to make anywhere else this
            // dialog is being used.
            // For Prime, see: projects/framework-rest/src/com/oracle/pgbu/framework/rest/serializers/StringDeserializer.java
            return val && val.trim();
        }

        function setError(msg) {
            var message = msg || options.helpText;
            that.$modal.find('.control-group')
                    .addClass('error')
                    .find('.help-block')
                    .text(message)
                    .removeClass('hidden');
            that.$modal.find('.proceedModal').prop('disabled', true);
        }

        function clearError() {
            that.$modal.find('.control-group').removeClass('error').find('.help-block').addClass('hidden');
            that.$modal.find('.proceedModal').prop('disabled', false);
        }

        var validate = function () {
            var promptValue = getPromptValue();
            if (_.isFunction(options.checkUnique)) {
                $.when(options.checkUnique(promptValue)).done(function (response) {
                    var result, message;

                    if (_.isObject(response)) {
                        result = response.result;
                        message = response.message || options.helpText;
                    } else {
                        result = response;
                        message = options.helpText;
                    }

                    if (!result) {
                        // TODO change this?
                        // The developer could modify the $input in `checkUnique` but it would be weird to do
                        // and likely no one is actually doing it.
                        dfd.resolve(getPromptValue());
                    } else {
                        setError(message);
                    }
                }).fail(function () {
                    setError(lang.getString('dialogs.serverValidationError'));
                    // Re-enable the proceed button, as we cannot say the value is invalid so let them try again
                    that.$modal.find('.proceedModal').prop('disabled', false);
                });
            } else if (options.required) {
                // Accounts for if user hits enter with a required field left blank
                if (promptValue !== '') {
                    dfd.resolve(promptValue);
                }
            } else {
                dfd.resolve(promptValue);
            }
        };

        that.$modal.find('.prompt-text')
            .val(options.defaultText)
            .on('keydown', function (ev) {
                /* This function was put in strictly for some odd IE10 behavior. Modals
                 * lose focus when users hit 'Enter', and the keydown event was firing on
                 * the 'Close Tab' buttons. This prevents 'Enter' keydown from triggering
                 * the wrong events in IE, and doesn't affect anything else.
                 */
                if (ev.keyCode === keyCode.ENTER) {
                    ev.preventDefault();
                }
            })
            .on('keydown', function () {
                clearError();
            })
            .on('blur', function () {
                if (getPromptValue().length === 0 && options.required) {
                    setError(lang.getFormattedString('dialogs.required', options.message));
                }
            })
            .on('keyup', function (e) {
                var valid = !options.required || $(this).val().length > 0;
                that.$modal.find('.proceedModal').prop('disabled', !valid);
                if (e.keyCode === keyCode.ESC) {
                    dfd.reject({
                        fromCloseButton : false
                    });
                    e.stopPropagation();
                }
            });

        this.$modal.find('.proceedModal')
            .prop('disabled', options.defaultText.length === 0 && options.required)
            .on('click', function () {
                //disable button when clicked so users cannot hit button twice.
                that.$modal.find('.proceedModal').prop('disabled', true);
                validate();
            });

        this.$modal.find('.cancelModal, [data-dismiss=modal]').on('click', function (ev) {
            dfd.reject({
                fromCloseButton : !$(this).hasClass('cancelModal')
            });
            ev.preventDefault();
            ev.stopPropagation();
        });

        this.$modal.on('shown', function () {
            that.$modal.find('.prompt-text').focus().get(0).select();
        });

        this.$modal.one('hidden', function () {
            that.$modal.remove();
        });

        dfd.always(function () {
            that.$modal.modal('hide');
        });

        if (_.isFunction(options.okCallback)) {
            dfd.done(options.okCallback);
        }

        if (_.isFunction(options.cancelCallback)) {
            dfd.fail(options.cancelCallback);
        }

        accessibilityUtils.modalAppend(this.$modal);

        this.$modal.modal({
            show : true,
            backdrop : 'static'
        });
    }

    PromptDialog.prototype = {
        constructor : PromptDialog
    };

    return PromptDialog;
});


define('primutils/markup/primutils/ui/dialogs/SavingDialogProgress.hbs', ['handlebars','primutils.markup','primutils/templates/helpers/modal','primutils/templates/helpers/button','primutils/templates/helpers/join','primutils/templates/helpers/ternary','primutils/templates/helpers/progress'], function (H, u) {
var imports = u.buildImported(Array.prototype.slice.call(arguments, 2), 'primutils/ui/dialogs/SavingDialogProgress.hbs'),
    func = H.template({"1":function(depth0,helpers,partials,data) {
  var stack1, buffer = "    ";
  stack1 = helpers.modalHeader.call(depth0, {"name":"modalHeader","hash":{
    'hideClose': (true)
  },"data":data});
  if (stack1 != null) { buffer += stack1; }
  buffer += "\n";
  stack1 = helpers.modalBody.call(depth0, {"name":"modalBody","hash":{},"fn":this.program(2, data),"inverse":this.noop,"data":data});
  if (stack1 != null) { buffer += stack1; }
  buffer += "    ";
  stack1 = helpers.modalFooter.call(depth0, {"name":"modalFooter","hash":{
    'omit': (true)
  },"data":data});
  if (stack1 != null) { buffer += stack1; }
  return buffer + "\n";
},"2":function(depth0,helpers,partials,data) {
  var stack1, lambda=this.lambda, buffer = "        ";
  stack1 = helpers.progress.call(depth0, {"name":"progress","hash":{
    'indeterminate': ((depth0 != null ? depth0.indeterminate : depth0))
  },"data":data});
  if (stack1 != null) { buffer += stack1; }
  buffer += "\n        <div style=\"display: none !important;\" class=\"saving-aria-message\" id=\"";
  stack1 = lambda((depth0 != null ? depth0.spinnerId : depth0), depth0);
  if (stack1 != null) { buffer += stack1; }
  buffer += "-aria-message\" role=\"alert\">";
  stack1 = lambda((depth0 != null ? depth0.savingInProgress : depth0), depth0);
  if (stack1 != null) { buffer += stack1; }
  return buffer + "</div>\n";
},"compiler":[6,">= 2.0.0-beta.1"],"main":function(depth0,helpers,partials,data) {
  var stack1, buffer = "";
  stack1 = helpers.modal.call(depth0, (depth0 != null ? depth0.message : depth0), {"name":"modal","hash":{
    'data-draggable': (false),
    'aria-describedby': (helpers._.call(depth0, (depth0 != null ? depth0.spinnerId : depth0), "-aria-message", {"name":"_","hash":{},"data":data})),
    'style': ("top: 30% !important;"),
    'id': ("saving-modal")
  },"fn":this.program(1, data),"inverse":this.noop,"data":data});
  if (stack1 != null) { buffer += stack1; }
  return buffer;
},"useData":true});
return function (context, options) {
    options = u.extendOptions(options || {}, imports);
    return func(context, options).trim();
};
});


define('primutils/markup/primutils/ui/dialogs/SavingDialogIndeterminate.hbs', ['handlebars','primutils.markup'], function (H, u) {
var imports = u.buildImported(Array.prototype.slice.call(arguments, 2), 'primutils/ui/dialogs/SavingDialogIndeterminate.hbs'),
    func = H.template({"compiler":[6,">= 2.0.0-beta.1"],"main":function(depth0,helpers,partials,data) {
  var stack1, lambda=this.lambda, buffer = "<div class=\"page-spinner page-saving-spinner page-spinner-static hide fade saving-dialog\" aria-labelledby=\"";
  stack1 = lambda((depth0 != null ? depth0.spinnerId : depth0), depth0);
  if (stack1 != null) { buffer += stack1; }
  buffer += "-saving-modal-label\" aria-describedby=\"";
  stack1 = lambda((depth0 != null ? depth0.spinnerId : depth0), depth0);
  if (stack1 != null) { buffer += stack1; }
  buffer += "-aria-message\" style=\"outline: none;\" tabindex=\"-1\" role=\"dialog\">\n    <div class=\"spinner-label\" id=\"";
  stack1 = lambda((depth0 != null ? depth0.spinnerId : depth0), depth0);
  if (stack1 != null) { buffer += stack1; }
  buffer += "-saving-modal-label\">";
  stack1 = lambda((depth0 != null ? depth0.message : depth0), depth0);
  if (stack1 != null) { buffer += stack1; }
  buffer += "</div>\n    <div class=\"spinner-large saving-spinner-large\">\n        <div class=\"dot dot1\"></div>\n        <div class=\"dot dot2\"></div>\n        <div class=\"dot dot3\"></div>\n    </div>\n    <div style=\"display: none !important;\" class=\"saving-aria-message\" id=\"";
  stack1 = lambda((depth0 != null ? depth0.spinnerId : depth0), depth0);
  if (stack1 != null) { buffer += stack1; }
  buffer += "-aria-message\" role=\"alert\">";
  stack1 = lambda((depth0 != null ? depth0.savingInProgress : depth0), depth0);
  if (stack1 != null) { buffer += stack1; }
  return buffer + "</div>\n</div>";
},"useData":true});
return function (context, options) {
    options = u.extendOptions(options || {}, imports);
    return func(context, options).trim();
};
});


/**
 * Creates a saving dialog in either a progress bar or spinner format.
 *
 * Unlike the <a href="ui.dialogs.LoadingDialog.html">LoadingDialog</a>, the SavingDialog creates a mask over the
 * page content which prevents the user from interacting with the application while the save operation is running.
 * It is imperative that you call `complete` in order to return control of the application back to the user.
 *
 * Normally, `type: 'spinner'` (the default) should be used.
 * If the progress of your operation is known, `type: 'progress'` may be preferred.
 *
 * You can change the name of the spinner using the `message` option.
 *
 *     require(['jquery', 'primutils/ui/dialogs/SavingDialog'], function ($, SavingDialog) {
 *
 *         // creating a new dialog showing the defaults
 *         var dialog = new SavingDialog({
 *             type : 'spinner',
 *             autoDisplay : true
 *         });
 *
 *         $.get('mystuff')
 *             .done(function (result) {
 *                 // do something with result
 *             })
 *             // always close the spinner, even if AJAX call fails
 *             .always(spinner.complete);
 *     });
 *
 * @class ui.dialogs.SavingDialog
 * @module ui.dialogs
 */
define('primutils/ui/dialogs/SavingDialog',['jquery', 'lodash', 'logger',
    'primutils/markup/primutils/ui/dialogs/SavingDialogProgress.hbs',
    'primutils/markup/primutils/ui/dialogs/SavingDialogIndeterminate.hbs',
    'primutils/utils/language',
    'primutils/a11y/accessibilityUtils',

    'primutils/utils/emulateTransitionEnd' // $.fn.emulateTransitionEnd
], function ($, _, logger, progressTemplate, indeterminateTemplate, lang, accessibilityUtils) {
    'use strict';

    var DEFAULT_OPTIONS = {
        /**
         * Is this a progress bar or a spinner.
         *
         * Valid options are 'spinner' or 'progress'
         *
         * @property type {String}
         * @optional
         * @default 'spinner'
         */
        type : 'spinner',
        /**
         * Whether to automatically call `show` when the dialog is created
         *
         * @property autoDisplay {Boolean}
         * @optional
         * @default true
         */
        autoDisplay : true
    };

    function getDefaultStrings() {
        return {
            /**
             * The default message to display, can be overridden. It's probably best to keep this somewhat short.
             *
             * @property message {String}
             * @optional
             * @default "Saving" - localized
             */
            message : lang.getString('dialogTitles.saving')
        };
    }

    function SavingDialog(options) {
        logger.assert(this && this instanceof SavingDialog, 'SavingDialog must be instantiated with new.');

        this.options = _.defaults({}, options, DEFAULT_OPTIONS, getDefaultStrings());

        // So that stop can be passed into promise.always / promise.then.
        this.complete = _.bind(this.complete, this);
        this.done = function () {
                        logger.warn('SavingDialog.done() is deprecated.  Use .complete() instead.  .done() may be removed without further warning.  Full docs at: http://10.140.183.183:8080/job/primutils/javadoc/classes/ui.dialogs.SavingDialog.html');
                        this.complete();
        }.bind(this);

        this._percent = 0.0;
        this._shown = false;

        this.savingProgress = progressTemplate({
            spinnerId : _.uniqueId('saving-spinner-'),
            indeterminate : this.options.type === 'progress-indeterminate',
            message : this.options.message,
            savingInProgress : lang.getString('dialogs.savingInProgress')
        });
        this.savingIndeterminate = indeterminateTemplate({
            spinnerId : _.uniqueId('saving-spinner-'),
            message : this.options.message,
            savingInProgress : lang.getString('dialogs.savingInProgress')
        });


        if (this.options.autoDisplay) {
            this.show();
        }
    }

    SavingDialog.prototype = {
        constructor : SavingDialog,

        /**
         * Shows the dialog
         *
         * @method show
         */
        show : function () {
            // Do not show twice
            if (this._shown) { return; }

            if (this.options.type === 'progress' || this.options.type === 'progress-indeterminate') {
                this._$dialog = $(this.savingProgress);
            } else {
                this._$dialog = $(this.savingIndeterminate);
            }

            // Remove old modal focus traps
            $(document).off('focusin.modal');

            accessibilityUtils.modalAppend(this._$dialog, { allowClosingParent : true });

            this._$dialog.modal({
                show : true,
                backdrop : 'static',
                keyboard : false
            });

            this._shown = true;
        },

        /**
         * Tells the spinner that it is complete and should hide and remove itself.
         *
         * *Note*: if you're closing the spinner and a modal is already up, the default behavior of the `SavingDialog` is to hide the 'parent' modal to which it initially appended.
         * This _may_ not be want you want.
         * For example, in case the user clicks save on your modal and there's a server-side error, it may be better to preserve the state of the modal than to hide it.
         * In this case, call with: `mySpinner.complete({allowClosingParent: false})`
         *
         * *Note*: `complete` is bound to `this` on instantiation as a convenience.
         * This means you can write code like:
         *
         *     var saving = new SavingDialog(...);
         *     $.ajax(...)
         *         .always(saving.complete);
         *
         * @param options {Object} if `options.allowClosingParent` is false, will NOT allow 'hide' and 'hidden' events to propagate to the parent modal.
         *     The default is true (meaning that propagation is allowed).
         *
         * @method complete
         */
        complete : function (options) {
            var that = this;

            if (this.options.type === 'progress' && this._percent !== 1.0) {
                this.percentage(1.0);
            }

            this._$dialog.one('hidden.hide', function (ev) {
                that._$dialog.remove();
                that.announceComplete();

                if (options && options.allowClosingParent === false) {
                    ev.stopPropagation();
                }
            }).modal('hide');

            this._shown = false;
        },

        /**
         * Changes the percentage of the progress bar.
         * The newPercent value should always be greater than the previous percentage otherwise display glitches may
         * result.
         *
         * Only available if type is 'progress'.
         *
         * @method percentage
         * @param newPercent {Number} If specified, set the percent to this value (0.0 - 1.0)
         */
        percentage : function (newPercent) {
            if (this.options.type !== 'progress') {
                this._percent = 0.0;
            }

            if (!arguments.length) {
                return this._percent;
            }

            if (this.options.type === 'progress') {
                // Clamp between 0.0 and 1.0
                this._percent = Math.min(Math.max(0, newPercent), 1);

                this._$dialog.find('.progress .bar').css('width', (this._percent * 100) + '%');

                if (this._percent === 1.0) {
                    this.complete();
                }
            }
        },

        announceComplete : function () {
            var $saveComplete = $('<div style="position: fixed; top: -10000px; left: -10000px;" class="saving-complete-alert" role="alert" aria-live="assertive" aria-relevant="all" aria-atomic="true">' + lang.getString('dialogs.savingCompleted') + '</div>').appendTo(document.body);
            setTimeout(function () {
                $saveComplete.remove();
            }, 0);
        }
    };

    return SavingDialog;
});


/**
 * Like `LoadingDialog`, but includes a blocking backdrop when the dialog is active.
 *
 * @class ui.dialogs.BlockingLoadingDialog
 * @module ui.dialogs
 * @uses ui.dialogs.LoadingDialog
 */
define('primutils/ui/dialogs/BlockingLoadingDialog',['jquery', 'lodash', 'primutils/ui/dialogs/LoadingDialog'], function ($, _, LoadingDialog) {
    'use strict';

    function modalBackdropAbsolute() {
        return $('<div class="modal-backdrop in loading-spinner-backdrop" style="position: absolute;"></div>');
    }

    function BlockingLoadingDialog() {
        LoadingDialog.apply(this, arguments);
    }

    BlockingLoadingDialog.prototype = _.assign(Object.create(LoadingDialog.prototype), {
        constructor : BlockingLoadingDialog,

        show : function () {
            LoadingDialog.prototype.show.apply(this, arguments);
            this.options.target.each(function () {
                var $this = $(this);
                modalBackdropAbsolute().appendTo($this);
            });
        },

        complete : function () {
            this.options.target.each(function () {
                $(this).children('.modal-backdrop.loading-spinner-backdrop').fadeOut(function () {
                    $(this).remove();
                });
            });
            LoadingDialog.prototype.complete.apply(this, arguments);
        }

    });

    return BlockingLoadingDialog;
});

define('primutils/validators/fileValidator',[
    'lodash',
    'primutils/i18n/Localizer'
], function (_, Localizer) {
    'use strict';

    var DEFAULTS = {
        required : false
    };

    /**
     * Validator for a collection of files (FileList API).
     *
     * **Options**:
     *
     *     required: false //Whether or not the file is required.
     *
     *     allowedExtensions: Array<String> //Array of extensions to validate.
     *
     *     maxFileSize: Number //Maximum file size for each file within the file list
     *
     *     validateByFileName: Function //Function called with the file list to validate a collection of files by name
     *
     *     validateFile: Function //Function called with the file list to validate a collection
     *
     *     fileSizePrinter: `Function` called for reporting the max file size in a user-friendly manner
     *
     * @class validators.fileValidator
     * @constructor
     *
     * @param {Object} options
     *
     * @module validation
     *
     */
    return function () {
        var options = _.defaults({}, arguments[0], DEFAULTS);

        if (options.maxFilesize) {
            options.maxFileSize = options.maxFilesize;
        }

        return Object.freeze({
            /**
             * @method validate
             * @param {FileList} fileList Collection of files to validate
             *
             * If at any point one of the validations fail it will short-circuit and return the validity.
             *
             * Validation Order:
             *
             *     Required -> type = required
             *     Extensions -> type = extension-mismatch
             *     MaxFileSize -> type = size
             *     ValidateByFileName -> type = name-mismatch
             *     ValidateFile -> type = file-exception
             *
             * **Note**: result contains two properties: valid (Boolean), and type (String)
             *
             * @return {Object} result
             */
            validate : function (fileList) {
                var allMatch;
                var result;

                if (options.required && (!fileList || fileList.length === 0)) {
                    return {
                        valid : false,
                        message : options.requiredErrorKey,
                        type : 'required'
                    };
                }

                if (options.allowedExtensions) {
                    allMatch = _.every(fileList, function (file) {
                        var fileExtension = file.name.split('.').pop().toLowerCase();
                        return _.includes(options.allowedExtensions, fileExtension);
                    });

                    if (!allMatch) {
                        return {
                            valid : false,
                            message : options.allowedTypesErrorKey, //Keep same option name here but it is already a localized message
                            type : 'extension-mismatch'
                        };
                    }
                }

                if (options.maxFileSize) {
                    allMatch = _.every(fileList, function (file) {
                        return file.size <= options.maxFileSize;
                    });

                    if (!allMatch) {
                        var sizePretty = _.isFunction(options.fileSizePrinter) ?
                            options.fileSizePrinter(options.maxFileSize) :
                            options.maxFileSize;
                        return {
                            valid : false,
                             //Keep same option name here but it is already a localized message
                            message : Localizer.formatString(options.maxFilesizeErrorMessageKey, sizePretty),
                            type : 'size'
                        };
                    }
                }

                if (options.validateFileByName) {
                    result = options.validateFileByName(fileList);

                    if (!result) {
                        return {
                            valid : false,
                            type : 'name-mismatch'
                        };
                    }
                }

                if (options.validateFile) {
                    result = options.validateFile(fileList);

                    if (!result) {
                        return {
                            valid : false,
                            type : 'file-exception'
                        };
                    }
                }

                return {
                    valid : true
                };
            }
        });
    };
});


define('primutils/markup/primutils/ui/dialogs/fileDialog.hbs', ['handlebars','primutils.markup'], function (H, u) {
var imports = u.buildImported(Array.prototype.slice.call(arguments, 2), 'primutils/ui/dialogs/fileDialog.hbs'),
    func = H.template({"1":function(depth0,helpers,partials,data) {
  var stack1, helper, options, functionType="function", helperMissing=helpers.helperMissing, blockHelperMissing=helpers.blockHelperMissing, buffer = "    ";
  stack1 = ((helper = (helper = helpers.modalHeader || (depth0 != null ? depth0.modalHeader : depth0)) != null ? helper : helperMissing),(typeof helper === functionType ? helper.call(depth0, {"name":"modalHeader","hash":{},"data":data}) : helper));
  if (stack1 != null) { buffer += stack1; }
  buffer += "\n";
  stack1 = ((helper = (helper = helpers.modalBody || (depth0 != null ? depth0.modalBody : depth0)) != null ? helper : helperMissing),(options={"name":"modalBody","hash":{},"fn":this.program(2, data),"inverse":this.noop,"data":data}),(typeof helper === functionType ? helper.call(depth0, options) : helper));
  if (!helpers.modalBody) { stack1 = blockHelperMissing.call(depth0, stack1, options); }
  if (stack1 != null) { buffer += stack1; }
  stack1 = ((helper = (helper = helpers.modalFooter || (depth0 != null ? depth0.modalFooter : depth0)) != null ? helper : helperMissing),(options={"name":"modalFooter","hash":{},"fn":this.program(5, data),"inverse":this.noop,"data":data}),(typeof helper === functionType ? helper.call(depth0, options) : helper));
  if (!helpers.modalFooter) { stack1 = blockHelperMissing.call(depth0, stack1, options); }
  if (stack1 != null) { buffer += stack1; }
  return buffer;
},"2":function(depth0,helpers,partials,data) {
  var stack1, helperMissing=helpers.helperMissing, buffer = "";
  stack1 = ((helpers.controlGroup || (depth0 && depth0.controlGroup) || helperMissing).call(depth0, (depth0 != null ? depth0.uploadFileLabel : depth0), {"name":"controlGroup","hash":{
    'class': ("file-upload-control-group")
  },"fn":this.program(3, data),"inverse":this.noop,"data":data}));
  if (stack1 != null) { buffer += stack1; }
  return buffer;
},"3":function(depth0,helpers,partials,data) {
  var stack1, helperMissing=helpers.helperMissing, buffer = "            ";
  stack1 = ((helpers.input || (depth0 && depth0.input) || helperMissing).call(depth0, "file", {"name":"input","hash":{
    'hint': ((depth0 != null ? depth0.helpText : depth0)),
    'class': ("input-block-level"),
    'id': ("fileUpload")
  },"data":data}));
  if (stack1 != null) { buffer += stack1; }
  return buffer + "\n";
},"5":function(depth0,helpers,partials,data) {
  var stack1, helperMissing=helpers.helperMissing, buffer = "        ";
  stack1 = ((helpers.icon || (depth0 && depth0.icon) || helperMissing).call(depth0, "loading", "", {"name":"icon","hash":{
    'style': ("margin-right: 8px;"),
    'class': ("upload-spinner hidden")
  },"fn":this.program(6, data),"inverse":this.noop,"data":data}));
  if (stack1 != null) { buffer += stack1; }
  buffer += "\n        ";
  stack1 = ((helpers.button || (depth0 && depth0.button) || helperMissing).call(depth0, "secondary", {"name":"button","hash":{
    'data-dismiss': ("modal"),
    'data-default': ("esc"),
    'class': ("cancelModal")
  },"fn":this.program(8, data),"inverse":this.noop,"data":data}));
  if (stack1 != null) { buffer += stack1; }
  buffer += "\n";
  stack1 = ((helpers.button || (depth0 && depth0.button) || helperMissing).call(depth0, "primary", (depth0 != null ? depth0.importButtonLabel : depth0), {"name":"button","hash":{
    'disabled': (true),
    'data-default': ("enter"),
    'class': ("accept disabled")
  },"fn":this.program(10, data),"inverse":this.noop,"data":data}));
  if (stack1 != null) { buffer += stack1; }
  return buffer;
},"6":function(depth0,helpers,partials,data) {
  return "";
},"8":function(depth0,helpers,partials,data) {
  var stack1, lambda=this.lambda;
  stack1 = lambda((depth0 != null ? depth0.cancel : depth0), depth0);
  if (stack1 != null) { return stack1; }
  else { return ''; }
  },"10":function(depth0,helpers,partials,data) {
  var stack1, lambda=this.lambda, helperMissing=helpers.helperMissing, buffer = "            <div class=\"default-state\">\n                ";
  stack1 = lambda((depth0 != null ? depth0.importButtonLabel : depth0), depth0);
  if (stack1 != null) { buffer += stack1; }
  buffer += "\n            </div>\n            <div class=\"saving-state hidden\">\n                ";
  stack1 = ((helpers.icon || (depth0 && depth0.icon) || helperMissing).call(depth0, "throbber", (depth0 != null ? depth0.importButtonLabel : depth0), {"name":"icon","hash":{},"data":data}));
  if (stack1 != null) { buffer += stack1; }
  return buffer + "\n            </div>\n";
},"compiler":[6,">= 2.0.0-beta.1"],"main":function(depth0,helpers,partials,data) {
  var stack1, helperMissing=helpers.helperMissing;
  stack1 = ((helpers.modal || (depth0 && depth0.modal) || helperMissing).call(depth0, (depth0 != null ? depth0.title : depth0), {"name":"modal","hash":{
    'class': ("file-upload-modal x-drop-inline")
  },"fn":this.program(1, data),"inverse":this.noop,"data":data}));
  if (stack1 != null) { return stack1; }
  else { return ''; }
  },"useData":true});
return function (context, options) {
    options = u.extendOptions(options || {}, imports);
    return func(context, options).trim();
};
});

define('primutils/ui/dialogs/fileDialog',[
    'jquery',
    'lodash',
    'oui',
    'primutils/validators/fileValidator',
    'primutils/ui/widgets/droppable',
    'primutils/utils/language',
    'primutils/a11y/accessibilityUtils',
    'primutils/a11y/messageAnnouncer',
    'primutils/markup/primutils/ui/dialogs/fileDialog.hbs'
], function ($, _, oui, fileValidator, droppable, localizer, a11yUtils, messageAnnouncer, fileUploadTemplate) {
    'use strict';

    var DEFAULTS = {
        autoShow : true,
        dismissOnSubmit : true,
        dismissOnAction : true,
        optional : false
    };

    function getDefaultStrings() {
        return {
            close : localizer.getString('dialogs.close'),
            title : localizer.getString('dialogTitles.file'),
            fileLabel : localizer.getString('fileDialog.file'),
            submitLabel : localizer.getString('dialogs.ok'),
            cancelLabel : localizer.getString('dialogs.cancel'),
            helpText : localizer.getString('fileDialog.help'),
            requiredHelp : localizer.getString('fileDialog.requiredHelp')
        };
    }

    function getTemplateData(options) {
        return {
            close : options.close,
            title : options.title,
            uploadFileLabel : options.fileLabel,
            importButtonLabel : options.submitLabel,
            cancel : options.cancelLabel,
            helpText : options.helpText
        };
    }

    function getValidatorOptions(options) {
        return _.assign({}, options.file, {
            fileSizePrinter : options.fileSizePrinter
        });
    }

    function attachListeners($el, fns) {
        $el.on('hidden', function (event) {
            if ($el[0] === event.target) {
                fns.hide();
            }
        })
        .on('click', '.accept', fns.submit)
        .on('click', '[data-action]', fns.action);
    }

    function toggleable($modal) {
        return {
            show : function () {
                a11yUtils.modalAppend($modal);
                $modal.modal({ show : true, backdrop : 'static' });
            },
            hide : function () {
                $modal.one('hidden', function () {
                    $modal.detach();
                });

                if ($modal.hasClass('in')) {
                    $modal.modal('hide');
                } else {
                    $modal.detach();
                }
            }
        };
    }

    function applyErrors($el, errors) {
        var $controlGroups = $el.find('[data-field]').closest('.control-group');

        $controlGroups.removeClass('error');

        $controlGroups.find('.help-block').text('');

        _.forEach(errors, function (message, fieldName) {
            var field = $el.find('[data-field=' + fieldName + ']');

            field.closest('.control-group').addClass('error');
            field.siblings('.help-block').text(message);
        });
    }

    /**
     * File dialog with basic selection and validation according to mime types, file size, name, and/or extension.
     *
     * **Options**:
     *
     *     dismissOnSubmit: true //Whether or not the dialog should auto dismiss as soon as accepted
     *
     *     dismissOnAction: true //Whether or not the dialog should auto dismiss as soon as an action has finished
     *
     *     autoShow: true //Whether the dialog should show as soon as constructed
     *
     *     optional: false //Whether the file is optional before submitting.
     *
     *     close: localizer.getString('dialogs.close') //Label for the close text within the dialog
     *
     *     title: localizer.getString('dialogTitles.file') //Default text for the title of the file dialog
     *
     *     fileLabel: localizer.getString('fileDialog.file') //Default text for the label above the file input
     *
     *     submitLabel: localizer.getString('dialogs.ok') //Default text for the accept button handler
     *
     *     cancelLabel: localizer.getString('dialogs.cancel') //Default text for the cancel button handler
     *
     *     helpText: localizer.getString('fileDialog.help') //Default text for the help block for the file input
     *
     *     requiredHelp: localizer.getString('fileDialog.required') //Default required error message text for the file input.
     *
     *     file: {} //Options for the file validator,
     *
     *     submit: function () {} //Accept handler after the button has been pressed. Can return a deferred.
     *
     *     actions: {} //Map of actions that are defined within the custom template data-action attributes to functions.
     *
     *     template: Function //Handlebars template that returns a DOM string
     *
     *     initComponents: Function //Handles the construction of additional components within the modal such as Pickers,
     *     TypeAheads, and Trees
     *
     * @class ui.dialogs.fileDialog
     * @constructor
     *
     * @param {Object} options
     *
     * @example
     *     var fileDialog = fileDialog({
     *         submit: function (files, additionalData) {
     *             //Called after 'Ok', 'Upload' button has been clicked.
     *             //Can be a deferred
     *         }
     *     });
     *
     * @example
     *     var fileDialog = fileDialog({
     *         submit: submitHandler,
     *         autoShow: true,
     *         dismissOnAccept: true,
     *         file: {
     *             allowedExtensions: ['txt', 'csv', 'xls']
     *         }
     *     });
     *
     * @example
     *     var fileDialog = fileDialog({
     *         template: customTemplate, //Handlebars template and not a jquery object
     *         actions: {
     *             remove: removeHandler //Passed files object - useful for 'DELETE' operations
     *         }
     *     });
     *
     *     fileDialog.show();
     *
     * @example
     *     var fileDialog = fileDialog({
     *         template: customTemplate,
     *         formData: function (object) { //Object with modal, and fields properties
     *             return {}; //Return additional data to be passed to submit. Useful when there are additional fields that
     *                        //are necessary for where a file may be uploaded/inserted to like a project
     *         }
     *     });
     */
    return function () {
        var options = _.defaults({}, arguments[0], DEFAULTS, getDefaultStrings());
        var validator = fileValidator(getValidatorOptions(options));
        var files = [];
        var handlers = {
            change : function (event) {
                updateFiles(event.target.files);
            },
            submit : function () {
                var $submitButton = $(this);
                var fields = _.reduce($modal.find('[data-field]'), function (memo, el) {
                    memo[$(el).attr('data-field')] = $(el);
                    return memo;
                }, {});
                var additionalData = _.isFunction(options.formData) ? options.formData({ modal : $modal, fields : fields }) : options.formData;
                var dfd;

                if (options.submit) {
                    dfd = options.submit(files, additionalData);
                }

                if (dfd) {
                    $submitButton.prop('disabled', true);
                    $submitButton.find('.default-state').addClass('hidden');
                    $submitButton.find('.saving-state').removeClass('hidden');
                    $.when(dfd)
                        .always(function () {
                            $submitButton.prop('disabled', false);
                            $submitButton.find('.default-state').removeClass('hidden');
                            $submitButton.find('.saving-state').addClass('hidden');
                        })
                        .done(function () {
                            // Only announce message if there is a deferred to wait for, otherwise something else will need to announce this
                            messageAnnouncer.announce(localizer.getString('fileDialog.uploadComplete'));
                        });
                }

                if (options.dismissOnSubmit) {
                    $.when(dfd)
                        .done(toggle.hide);
                }
            },
            action : function (event) {
                var dfd = options.actions[$(event.target).attr('data-action')](event, files);

                if (_.isFunction(dfd.promise)) {
                    $modal.addClass('busy');

                    $.when(dfd)
                        .always(function () {
                            $modal.removeClass('busy');

                            if (options.dismissOnAction) {
                                toggle.hide();
                            }
                        });
                } else if (options.dismissOnAction) {
                    toggle.hide();
                }
            }
        };
        var toggle;
        var $controlGroup;
        var $modal;
        var $filePicker;
        var userSubmittable;

        function updateFiles(newFiles) {
            validate(newFiles);

            //Called when changed
            files = newFiles;

            //If using OUI there will be an input of type text
            $filePicker.closest('.orcl-widget').find('input[type=text]').val(_.pluck(newFiles, 'name').join(', ') || '');

            //Can only wipe out native file input
            if (newFiles == null) {
                $filePicker.val('');
            }

            return true;
        }

        function submittable(enabled) {
            userSubmittable = enabled;

            validate(files);
        }

        function validate(files) {
            var result = validator.validate(files);
            var valid = result.valid && (options.optional || files && files.length > 0);
            var $help = $controlGroup.find('.help-block');
            var additionalValidation = !options.validator || options.validator(files || [], $modal.find('[data-field]'));
            var disabled = !(valid && (userSubmittable == null || userSubmittable) && additionalValidation);

            $help.text(options[result.type] || result.message || options.helpText);

            $controlGroup.toggleClass('error', !result.valid);

            $modal.find('.accept')
                .prop('disabled', disabled)
                .toggleClass('disabled', disabled);
        }

        //Allow custom templates
        if (options.template) {
            fileUploadTemplate = options.template;
        }

        $modal = $(fileUploadTemplate(getTemplateData(options)));
        toggle = toggleable($modal);
        $filePicker = $modal.find('input[type=file]');
        $controlGroup = $filePicker.closest('.control-group');

        if (options.initComponents) {
            options.initComponents($modal);
        }

        if (oui && $.fn.filePicker) {
            $filePicker.filePicker({
                fluid : true,
                multiple : false //for parity right now
            });
        }

        droppable($modal, {
            inline : true,
            type : 'Files', //Prevent default 'text' from being applied (we are only looking for files)
            valid : function (file) {
                return validator.validate(file).valid;
            },
            dropped : updateFiles
        });

        $filePicker.on('change', handlers.change);

        attachListeners($modal, {
            submit : handlers.submit,
            hide : toggle.hide,
            action : handlers.action
        });

        if (options.autoShow) {
            toggle.show();
        }

        return Object.freeze({
            /**
             * Reruns validation on the file dialog and toggles the submit button based on result.
             *
             * @method revalidate
             */
            revalidate : function () {
                validate(files);
            },
            /**
             * Clears the files within the dialog.
             *
             * @method clear
             */
            clear : _.partial(updateFiles, null),
            /**
             * Shows the dialog.
             *
             * @method show
             */
            show : toggle.show,
            /**
             * Hides the dialog.
             *
             * @method hide
             */
            hide : toggle.hide,

            /**
             * Add event listeners to the modal element.
             *
             * Shortcut since the custom template is a handlebars template and not a jQuery object.
             *
             * @method on
             */
            on : $modal.on.bind($modal),

            /**
             * Find shortcut bound to the internal modal to find custom elements from the template.
             *
             * @method find
             */
            find : $.fn.find.bind($modal),

            /**
             * Attribute for underlying modal element
             *
             * @property $el
             */
            $el : $modal,

            /**
             * Update errors on the dialog
             *
             * If called with no parameters it will remove all errors.
             *
             * @method errors
             * @param {Object} errors The error map of field name to message
             */
            errors : function (errors) {
                applyErrors($modal, errors);
                submittable(_.keys(errors).length === 0);
            },

            /**
             * Disables the file dialog.
             *
             * @method disable
             */
            disable : _.partial(submittable, false),
            /**
             * Enables the file dialog.
             *
             * **Note**: if the dialog is still in an invalid state due to validation it will not force the submit action
             * handler to enabled. Only allow it to be enabled when it becomes valid.
             *
             * @method enable
             */
            enable : _.partial(submittable, true)
        });
    };
});


define('primutils/markup/primutils/ui/dialogs/previewDialog.hbs', ['handlebars','primutils.markup'], function (H, u) {
var imports = u.buildImported(Array.prototype.slice.call(arguments, 2), 'primutils/ui/dialogs/previewDialog.hbs'),
    func = H.template({"1":function(depth0,helpers,partials,data) {
  var stack1, helper, options, helperMissing=helpers.helperMissing, functionType="function", blockHelperMissing=helpers.blockHelperMissing, buffer = "";
  stack1 = ((helpers.modalHeader || (depth0 && depth0.modalHeader) || helperMissing).call(depth0, {"name":"modalHeader","hash":{
    'hideTitle': (true)
  },"fn":this.program(2, data),"inverse":this.noop,"data":data}));
  if (stack1 != null) { buffer += stack1; }
  stack1 = ((helper = (helper = helpers.modalBody || (depth0 != null ? depth0.modalBody : depth0)) != null ? helper : helperMissing),(options={"name":"modalBody","hash":{},"fn":this.program(17, data),"inverse":this.noop,"data":data}),(typeof helper === functionType ? helper.call(depth0, options) : helper));
  if (!helpers.modalBody) { stack1 = blockHelperMissing.call(depth0, stack1, options); }
  if (stack1 != null) { buffer += stack1; }
  stack1 = ((helper = (helper = helpers.modalFooter || (depth0 != null ? depth0.modalFooter : depth0)) != null ? helper : helperMissing),(options={"name":"modalFooter","hash":{},"fn":this.program(22, data),"inverse":this.noop,"data":data}),(typeof helper === functionType ? helper.call(depth0, options) : helper));
  if (!helpers.modalFooter) { stack1 = blockHelperMissing.call(depth0, stack1, options); }
  if (stack1 != null) { buffer += stack1; }
  return buffer;
},"2":function(depth0,helpers,partials,data) {
  var stack1, lambda=this.lambda, helperMissing=helpers.helperMissing, buffer = "        <div class=\"modal-title\" id=\"";
  stack1 = lambda(((stack1 = (data && data.modalContext)) && stack1.labeledBy), depth0);
  if (stack1 != null) { buffer += stack1; }
  buffer += "\">\n            ";
  stack1 = ((helpers.icon || (depth0 && depth0.icon) || helperMissing).call(depth0, "file-generic", (depth0 != null ? depth0.fileLabel : depth0), {"name":"icon","hash":{
    'class': ("doc-information")
  },"data":data}));
  if (stack1 != null) { buffer += stack1; }
  buffer += "\n            <span class=\"doc-information\">\n                ";
  stack1 = lambda((depth0 != null ? depth0.name : depth0), depth0);
  if (stack1 != null) { buffer += stack1; }
  buffer += "\n            </span>\n            <span class=\"doc-information\">\n                ";
  stack1 = lambda((depth0 != null ? depth0.size : depth0), depth0);
  if (stack1 != null) { buffer += stack1; }
  buffer += "\n            </span>\n";
  stack1 = helpers['if'].call(depth0, (depth0 != null ? depth0.downloadable : depth0), {"name":"if","hash":{},"fn":this.program(3, data),"inverse":this.noop,"data":data});
  if (stack1 != null) { buffer += stack1; }
  buffer += "        </div>\n\n";
  stack1 = ((helpers.modalControlBar || (depth0 && depth0.modalControlBar) || helperMissing).call(depth0, {"name":"modalControlBar","hash":{
    'hideClose': (true),
    'class': ("right-controls")
  },"fn":this.program(6, data),"inverse":this.noop,"data":data}));
  if (stack1 != null) { buffer += stack1; }
  return buffer;
},"3":function(depth0,helpers,partials,data) {
  var stack1, helperMissing=helpers.helperMissing, buffer = "                ";
  stack1 = ((helpers.button || (depth0 && depth0.button) || helperMissing).call(depth0, "default", (depth0 != null ? depth0.downloadLabel : depth0), {"name":"button","hash":{
    'class': ("download")
  },"fn":this.program(4, data),"inverse":this.noop,"data":data}));
  if (stack1 != null) { buffer += stack1; }
  return buffer + "\n";
},"4":function(depth0,helpers,partials,data) {
  var stack1, lambda=this.lambda;
  stack1 = lambda((depth0 != null ? depth0.downloadLabel : depth0), depth0);
  if (stack1 != null) { return stack1; }
  else { return ''; }
  },"6":function(depth0,helpers,partials,data) {
  var stack1, helperMissing=helpers.helperMissing, buffer = "            <span class=\"zoom-controls\">\n";
  stack1 = ((helpers.button || (depth0 && depth0.button) || helperMissing).call(depth0, "not-btn", (depth0 != null ? depth0.zoomOut : depth0), {"name":"button","hash":{
    'class': ("zoom-out")
  },"fn":this.program(7, data),"inverse":this.noop,"data":data}));
  if (stack1 != null) { buffer += stack1; }
  buffer += "                <input type=\"range\">\n";
  stack1 = ((helpers.button || (depth0 && depth0.button) || helperMissing).call(depth0, "not-btn", (depth0 != null ? depth0.zoomIn : depth0), {"name":"button","hash":{
    'class': ("zoom-in")
  },"fn":this.program(9, data),"inverse":this.noop,"data":data}));
  if (stack1 != null) { buffer += stack1; }
  buffer += "            </span>\n            <span class=\"rotate-controls\">\n";
  stack1 = ((helpers.button || (depth0 && depth0.button) || helperMissing).call(depth0, "not-btn", (depth0 != null ? depth0.rotateLeft : depth0), {"name":"button","hash":{
    'class': ("rotate-left")
  },"fn":this.program(11, data),"inverse":this.noop,"data":data}));
  if (stack1 != null) { buffer += stack1; }
  stack1 = ((helpers.button || (depth0 && depth0.button) || helperMissing).call(depth0, "not-btn", (depth0 != null ? depth0.rotateRight : depth0), {"name":"button","hash":{
    'class': ("rotate-right")
  },"fn":this.program(13, data),"inverse":this.noop,"data":data}));
  if (stack1 != null) { buffer += stack1; }
  buffer += "            </span>\n            <span class=\"zoom-level\">\n            </span>\n";
  stack1 = ((helpers.link || (depth0 && depth0.link) || helperMissing).call(depth0, "#", (depth0 != null ? depth0.close : depth0), {"name":"link","hash":{
    'data-dismiss': ("modal"),
    'class': ("modal-close")
  },"fn":this.program(15, data),"inverse":this.noop,"data":data}));
  if (stack1 != null) { buffer += stack1; }
  return buffer;
},"7":function(depth0,helpers,partials,data) {
  var stack1, helperMissing=helpers.helperMissing, buffer = "                    ";
  stack1 = ((helpers.icon || (depth0 && depth0.icon) || helperMissing).call(depth0, "zoom-out", (depth0 != null ? depth0.zoomOut : depth0), {"name":"icon","hash":{
    'size': ("pgbu-icon-large")
  },"data":data}));
  if (stack1 != null) { buffer += stack1; }
  return buffer + "\n";
},"9":function(depth0,helpers,partials,data) {
  var stack1, helperMissing=helpers.helperMissing, buffer = "                    ";
  stack1 = ((helpers.icon || (depth0 && depth0.icon) || helperMissing).call(depth0, "zoom-in", (depth0 != null ? depth0.zoomIn : depth0), {"name":"icon","hash":{
    'size': ("pgbu-icon-large")
  },"data":data}));
  if (stack1 != null) { buffer += stack1; }
  return buffer + "\n";
},"11":function(depth0,helpers,partials,data) {
  var stack1, helperMissing=helpers.helperMissing, buffer = "                    ";
  stack1 = ((helpers.icon || (depth0 && depth0.icon) || helperMissing).call(depth0, "external-link", (depth0 != null ? depth0.rotateLeft : depth0), {"name":"icon","hash":{
    'size': ("pgbu-icon-large")
  },"data":data}));
  if (stack1 != null) { buffer += stack1; }
  return buffer + "\n";
},"13":function(depth0,helpers,partials,data) {
  var stack1, helperMissing=helpers.helperMissing, buffer = "                    ";
  stack1 = ((helpers.icon || (depth0 && depth0.icon) || helperMissing).call(depth0, "external-link", (depth0 != null ? depth0.rotateRight : depth0), {"name":"icon","hash":{
    'size': ("pgbu-icon-large")
  },"data":data}));
  if (stack1 != null) { buffer += stack1; }
  return buffer + "\n";
},"15":function(depth0,helpers,partials,data) {
  var stack1, helperMissing=helpers.helperMissing, buffer = "                ";
  stack1 = ((helpers.icon || (depth0 && depth0.icon) || helperMissing).call(depth0, "dialog-close", (depth0 != null ? depth0.close : depth0), {"name":"icon","hash":{},"data":data}));
  if (stack1 != null) { buffer += stack1; }
  return buffer + "\n";
},"17":function(depth0,helpers,partials,data) {
  var stack1, helperMissing=helpers.helperMissing, buffer = "        <div class=\"viewer-container\">\n            <div class=\"image-container\">\n                ";
  stack1 = ((helpers.icon || (depth0 && depth0.icon) || helperMissing).call(depth0, "throbber", (depth0 != null ? depth0.loading : depth0), {"name":"icon","hash":{},"data":data}));
  if (stack1 != null) { buffer += stack1; }
  buffer += "\n            </div>\n";
  stack1 = ((helpers.button || (depth0 && depth0.button) || helperMissing).call(depth0, "not-btn", {"name":"button","hash":{
    'class': ("circle-btn btn-prev page-switch")
  },"fn":this.program(18, data),"inverse":this.noop,"data":data}));
  if (stack1 != null) { buffer += stack1; }
  stack1 = ((helpers.button || (depth0 && depth0.button) || helperMissing).call(depth0, "not-btn", {"name":"button","hash":{
    'class': ("circle-btn btn-next page-switch")
  },"fn":this.program(20, data),"inverse":this.noop,"data":data}));
  if (stack1 != null) { buffer += stack1; }
  return buffer + "        </div>\n        <div class=\"modal-sidebar\"></div>\n";
},"18":function(depth0,helpers,partials,data) {
  var stack1, helperMissing=helpers.helperMissing, buffer = "                    ";
  stack1 = ((helpers.icon || (depth0 && depth0.icon) || helperMissing).call(depth0, "chevron-left", (depth0 != null ? depth0.prevPage : depth0), {"name":"icon","hash":{},"data":data}));
  if (stack1 != null) { buffer += stack1; }
  return buffer + "\n";
},"20":function(depth0,helpers,partials,data) {
  var stack1, helperMissing=helpers.helperMissing, buffer = "                    ";
  stack1 = ((helpers.icon || (depth0 && depth0.icon) || helperMissing).call(depth0, "chevron-right", (depth0 != null ? depth0.nextPage : depth0), {"name":"icon","hash":{},"data":data}));
  if (stack1 != null) { buffer += stack1; }
  return buffer + "\n";
},"22":function(depth0,helpers,partials,data) {
  return "";
},"compiler":[6,">= 2.0.0-beta.1"],"main":function(depth0,helpers,partials,data) {
  var stack1, helperMissing=helpers.helperMissing;
  stack1 = ((helpers.modal || (depth0 && depth0.modal) || helperMissing).call(depth0, (depth0 != null ? depth0.name : depth0), {"name":"modal","hash":{
    'data-draggable': ("false"),
    'class': ("modal-preview")
  },"fn":this.program(1, data),"inverse":this.noop,"data":data}));
  if (stack1 != null) { return stack1; }
  else { return ''; }
  },"useData":true});
return function (context, options) {
    options = u.extendOptions(options || {}, imports);
    return func(context, options).trim();
};
});

define('primutils/ui/dialogs/previewDialog',[
    'jquery',
    'lodash',
    'oui',
    'logger',
    'primutils/formatters/numberFormatter',
    'primutils/utils/language',
    'primutils/utils/keyCode',
    'primutils/utils/fileSaver',
    'primutils/utils/Subscribable',
    'primutils/a11y/accessibilityUtils',
    'primutils/markup/primutils/ui/dialogs/previewDialog.hbs'
], function ($, _, oui, logger, numberFormatter, localizer, keyCode, fileSaver, Subscribable, accessabilityUtils, previewTemplate) {
    'use strict';

    var DISTRACTION_FREE_SELECTOR = '>.modal-header, >.modal-footer, .page-switch, .modal-sidebar';

    var DEFAULTS = {
        downloadable : true,
        startPage : 1,
        rotation : false,
        zoom : true,
        file : {},
        getPage : function () {},
        getPageStatus : function (page, total) {
            return localizer.getFormattedString('previewDialog.pageStatus', page, total);
        },
        toolbar : null,
        sidebar : null,
        distractionFreeMode : true,
        percentFormatter : numberFormatter({
            format : '0%',
            formattedMultiplier : 100
        })
    };

    function getDefaultStrings() {
        return {
            close : localizer.getString('dialogs.close'),
            zoomIn : localizer.getString('previewDialog.zoomIn'),
            zoomOut : localizer.getString('previewDialog.zoomOut'),
            rotateLeft : localizer.getString('previewDialog.rotateLeft'),
            rotateRight : localizer.getString('previewDialog.rotateRight'),
            downloadLabel : localizer.getString('previewDialog.download'),
            fileLabel : localizer.getString('previewDialog.file'),
            loading : localizer.getString('dialogTitles.loading'),
            prevPage : localizer.getString('previewDialog.prevPage'),
            nextPage : localizer.getString('previewDialog.nextPage')
        };
    }

    function getTemplateData(options) {
        return _.pick(options, [
            'close', 'zoomIn', 'zoomOut', 'rotateLeft', 'rotateRight', 'downloadLabel',
            'file', 'loading', 'prevPage', 'nextPage', 'fileLabel'
        ]);
    }

    function addImage($el, url, callback) {
        $el.find('img').remove();
        $el.addClass('loading');
        var $img = $('<img>')
            .one('load', function () {
                $el.removeClass('loading');
                if (_.isFunction(callback)) {
                    callback($img);
                }
            })
            .appendTo($el)
            .attr('src', url);
    }

    function getDeg(deg) {
        return ((deg % 360) + 360) % 360;
    }

    /**
     * The Preview dialog used to view images taking up the entire browser window. Useful for seeing documents rendered as
     * images.
     *
     * **Options**:
     *
     * * `downloadable`: default true, whether or not the document is allowed to be downloaded
     * * `startPage`: 1 //First page to render
     * * `rotation`: false //Whether or not the images are allowed to be rotated
     * * `zoom`: true //Whether or not the images are allowed to be zoomed
     * * `footer`: true //Whether or not the footer should appear
     * * `file`: Object //Metadata for the file being previewed
     * * `file.name`: String //Name of the document
     * * `file.size`: String //Formatted size of the document
     * * `file.pages`: Number //Total number of pages within the document
     * * `getPage`: Function //Returns an href to an image for the page number supplied
     * * `toolbar`: Object //A Dom element or JQuery selection to use as a custom modal head
     * * `sidebar`: Object //A Dom element or JQuery selection to use as a custom sidebar
     * * `distractionFreeMode : Boolean //Whether or not the preview can enter distraction free mode.  Defaults to true`
     * * `getPageStatus`: Function //Callback for rendering the page out of the total pages. Supplied the current page and the total number of pages
     *
     * @class ui.dialogs.previewDialog'
     * @constructor
     * @param {Object} options The options for the preview dialog
     */
    return function () {
        var options = _.defaults({}, arguments[0], DEFAULTS, getDefaultStrings());
        var fileMetaData = options.file;
        logger.assert(fileMetaData && _.isString(fileMetaData.url) && !_.isEmpty(fileMetaData.url),
            'file - including url - must be provided to previewDialog for download functionality to work');
        var currentPage = options.startPage;
        var $el = $(previewTemplate(_.assign({
            name : fileMetaData.name,
            size : fileMetaData.size,
            type : fileMetaData.type,
            downloadable : options.downloadable,
            startPage : options.startPage,
            numberOfPages : fileMetaData.pages
        }, getTemplateData(options))));
        var slider;
        var pageTransforms = {};

        var $imageContainer = $el.find('.image-container');
        var $modalHeader = $el.find('>.modal-header');
        var $modalBody = $el.find('>.modal-body');
        var $modalFooter = $el.find('>.modal-footer');

        if (options.toolbar) {
            $modalHeader.find('.modal-title').empty().append(options.toolbar);
        }

        if (options.sidebar) {
            $modalBody.find('.modal-sidebar').append(options.sidebar);
        }

        //Preview doesn't start with any modals appended to it, but be safe.
        accessabilityUtils.stopModalHidePropagation($el.children());

        function applyTransforms(transforms) {
            pageTransforms = transforms;
            render();
        }

        function showImage(page) {
            addImage($imageContainer, options.getPage(page), function ($image) {
                render(page); //Will trigger image-transform
                api.trigger('image-load', {
                    page : currentPage
                });
            });

            $modalFooter.text(options.getPageStatus(page, fileMetaData.pages));

            //Hide page navigation buttons if there's only 1 page.
            //For now, don't hide either button of the first/last page since the buttons can wrap.
            $el.find('.btn-prev, .btn-next').toggleClass('hidden', fileMetaData.pages === 1);
        }

        function updateTransform(page, data) {
            if (!options.rotate) {
                data.rotate = 0;
            }

            if (!options.zoom) {
                data.scale = 1;
            }

            pageTransforms[page] = _.merge(pageTransforms[currentPage] || {}, data);
        }

        function triggerTransform() {
            var $image = $imageContainer.find('img');
            var position = $image.position();
            var scale = 1;
            if (pageTransforms[currentPage] && pageTransforms[currentPage].scale) {
                scale = pageTransforms[currentPage].scale;
            }

            api.trigger('image-transform', {
                x : Math.max(parseFloat(position.left) + parseFloat($image.css('marginLeft')), 0),
                y : Math.max(parseFloat(position.top) + parseFloat($image.css('marginTop')), 0),
                width : $image.width() * scale,
                height : $image.height() * scale,
                transforms : pageTransforms[currentPage]
            });
        }

        function download() {
            if (options.downloadable) {
                fileSaver.requestAndSave({
                    fileUrl : options.file.url,
                    fileName : options.file.name
                });
            }
        }

        function render() {
            var transform = _.defaults(pageTransforms[currentPage] || {}, {
                scale : 1,
                rotate : 0
            });

            var $img = $imageContainer.find('img');
            var imgTrueHeight = $img.height();
            var imgTrueWidth = $img.width();

            var transformCSS = '';
            var translateX = 0;
            var translateY = 0;

            if (transform.scale) {
                transformCSS += ' scale(' + transform.scale + ')';

                translateX += (imgTrueWidth * transform.scale - imgTrueWidth) / 2;
                translateY += (imgTrueHeight * transform.scale - imgTrueHeight) / 2;
            }

            if (transform.rotate) {
                transformCSS += ' rotate(' + transform.rotate + 'deg)';
            }

            $img.css({
                transform : transformCSS
            });

            if (slider) {
                slider.value({
                    end : transform.scale
                });
            }

            $modalHeader.find('.zoom-level').text(options.percentFormatter.format(transform.scale));

            if (transform.scale < 1) {
                if (imgTrueHeight * transform.scale < $imageContainer.height()) {
                    $img.css({
                        marginTop : -(imgTrueHeight / 2 - $imageContainer.height() / 2),
                        marginLeft : 0
                    });
                } else {
                    $img.css({
                        marginTop : -((imgTrueHeight - imgTrueHeight * transform.scale) / 2),
                        marginLeft : 0
                    });
                }
            } else if (transform.scale === 1) {
                $img.css({
                    marginTop : 0,
                    marginLeft : 0
                });
            } else if (transform.scale > 1) {
                $img.css({
                    marginTop : translateY,
                    marginLeft : translateX
                });
            }

            //Rerender it to force rescroll
            $img.css({
                display : 'none'
            });

            $img.css({
                display : 'block'
            });

            if (pageTransforms[currentPage]) {
                $imageContainer.scrollTop((($img.height() * transform.scale) - $imageContainer.height()) / 2);
                $imageContainer.scrollLeft((($img.width() * transform.scale) - $imageContainer.width()) / 2);
            } else {
                $imageContainer.scrollTop(0);
                $imageContainer.scrollLeft(0);
            }

            triggerTransform();
        }

        if (options.autoShow) {
            $el.modal('show');
        }

        //Only append slider if it exists, allow updating transform normally otherwise.
        if (oui.Slider) {
            slider = new oui.Slider({
                min : 0.25,
                max : 4,
                fluid : false,
                type : 'single',
                range : false
            }, $el.find('.zoom-controls input[type=range]'), { end : 1 }).init();

            slider.element().addClass('input-small');

            //Apply zoom
            slider.on('update-end', function (data) {
                updateTransform(currentPage, { scale : data.end });
                render();
            });
        }

        $el.find('.zoom-controls').toggleClass('hidden', !options.zoom);
        $el.find('.rotate-controls').toggleClass('hidden', !options.rotate);

        $el.on('click', '.zoom-in', function () {
            if (slider) {
                var curData = slider.data();
                if (curData.end % 0.25 === 0) {
                    curData.end = curData.end + 0.25;
                } else {
                    //This operation converts curData to nearest multiple of 0.25
                    curData.end = Math.floor((curData.end) / 0.25) * 0.25;
                }
                slider.data({
                    end : curData.end
                });
            } else {
                var curScale = (pageTransforms[currentPage] || {}).scale || 1;
                updateTransform(currentPage, { scale : curScale + 0.25 });
                render();
            }
        });

        $el.on('click', '.zoom-out', function () {
            if (slider) {
                var curData = slider.data();
                if (curData.end % 0.25 === 0) {
                    curData.end = curData.end - 0.25;
                } else {
                    //This operation converts curData to nearest multiple of 0.25
                    curData.end = Math.floor((curData.end) / 0.25) * 0.25;
                }
                slider.data({
                    end : curData.end
                });
            } else {
                var curScale = (pageTransforms[currentPage] || {}).scale || 1;
                updateTransform(currentPage, { scale : (curScale - 0.25) || 0.25 });
                render();
            }
        });

        $el.on('click', '.rotate-left', function () {
            updateTransform(currentPage, { rotate : getDeg((_.get(pageTransforms, [currentPage, 'rotate']) || 0) - 90) });
            render();
        });

        $el.on('click', '.rotate-right', function () {
            updateTransform(currentPage, { rotate : getDeg((_.get(pageTransforms, [currentPage, 'rotate']) || 0) + 90) });
            render();
        });

        //Panning document
        $imageContainer.on('mousedown', 'img', function (event) {
            event.preventDefault();
            var originalX = event.clientX;
            var originalY = event.clientY;

            $el.focus();

            $el.on('mousemove.panning', function (event) {
                api.pan(originalX - event.clientX, originalY - event.clientY);

                originalX = event.clientX;
                originalY = event.clientY;
            });

            $el.on('mouseup.panning mouseout.panning', function () {
                $el.off('.panning');
            });
        });

        //Don't allow shown to be re-triggered by nested modals.
        $el.one('shown.modal', function () {
            showImage(currentPage);
        });

        $el.one('hidden', function (ev) {
            $el.detach();
        });

        $el.on('click', '.download', download);
        $el.on('click', '.btn-next', function () {
            api.nextPage();
        });
        $el.on('click', '.btn-prev', function () {
            api.prevPage();
        });

        $el.on('keydown', function (event) {
            var $target = $(event.target);
            //Don't trigger for embedded modals or if the target is an input element.
            var isInput = $(event.target).is('input, textarea, select, button, a');
            var isChildModal = $el.find('.modal').has($target).length > 0;
            if (isInput || isChildModal) {
                return;
            }

            switch (event.keyCode) {
            case keyCode.LEFT:
                if (event.altKey) {
                    event.preventDefault();
                    api.pan(-20, 0);
                } else {
                    api.prevPage();
                }
                break;
            case keyCode.RIGHT:
                if (event.altKey) {
                    event.preventDefault();
                    api.pan(20, 0);
                } else {
                    api.nextPage();
                }
                break;
            case keyCode.UP:
                event.preventDefault();
                api.pan(0, -20);
                break;
            case keyCode.DOWN:
            case keyCode.SPACE:
                event.preventDefault();
                api.pan(0, 20);
                break;
            case keyCode.HOME:
                event.preventDefault();
                $imageContainer.scrollTop(0);
                break;
            case keyCode.END:
                event.preventDefault();
                $imageContainer.scrollTop($imageContainer[0].scrollHeight);
                break;
            case keyCode.ENTER:
                if (event.altKey && options.downloadable) {
                    download();
                }
                break;
            case keyCode.PAGEDOWN:
                api.pan(null, $imageContainer.height, false);
                break;
            case keyCode.PAGEUP:
                $imageContainer.scrollTop(0, $imageContainer.height());
                break;
            }
        });
        var timeout;
        var hideDocInformation = function () {
            if (options.distractionFreeMode) {
                $el.find(DISTRACTION_FREE_SELECTOR).addClass('out');
                $modalBody.addClass('no-mouse');
            }
        };

        var showDocInformation = _.throttle(function (event) {
            //reset timeout
            clearTimeout(timeout);
            $el.find(DISTRACTION_FREE_SELECTOR).removeClass('out');
            $modalBody.removeClass('no-mouse');
            timeout = setTimeout(hideDocInformation, 3000);
        }, 1000, { leading : true, trailing : false });

        timeout = setTimeout(hideDocInformation, 3000);
        $el.on('mousemove keydown', showDocInformation);
        $el.on('mousemove mouseenter', DISTRACTION_FREE_SELECTOR, function (event) {
            //This will sometimes prevent showDocInformation from ever being called,
            //see https://github.com/megawac/lodash/commit/cf0b8fbe8501e64c46488f71e748bbc9e201670b
            showDocInformation.cancel();
            clearTimeout(timeout);
        });

        $el.on('mouseleave', DISTRACTION_FREE_SELECTOR, showDocInformation);


        var api = _.create(Subscribable.prototype, {
            /**
             * Shows the preview modal.
             *
             * @method show
             */
            show : function () {
                $el.modal('show');
            },
            /**
             * Hides the preview modal.
             *
             * @method hide
             */
            hide : function () {
                $el.modal('hide');
            },
            /**
            * Getter and setter for the current page
            * - Setting this value will update and redraw the current page
            *
            * @method currentPage
            * @param {number} page Page to navigate to
            */
            currentPage : function (page) { //Using JQuery style setters instead of JS setters
                if (arguments.length && _.isNumber(page)) {
                    currentPage = Math.abs(page % fileMetaData.pages);
                    //Pages are 1 based, not 0 based, so we need some extra adjustment
                    if (currentPage === 0) {
                        currentPage = fileMetaData.pages;
                    }
                    showImage(currentPage);
                }

                return currentPage;
            },
            /**
            * Convenience method for navigating to the next page
            *
            * @method prevPage
            */
            nextPage : function () {
                return api.currentPage(currentPage + 1);
            },
            /**
            * Convenience method for navigating to the previous page
            *
            * @method prevPage
            */
            prevPage : function () {
                return api.currentPage(currentPage - 1);
            },
            /**
            * Downloads the preview (if options.downloadable is true)
            *
            * @method download
            */
            download : download,
            /**
            * Getter and setter for the preview's current pan position
            *
            * @param {number} x X position to pan to
            * @param {number} y Y position to pan to
            * @param {boolean} relative Optional toggle for whether to pan relative to the current position (defaults to true)
            * @method pan
            */
            pan : function (x, y, relative) {
                if (arguments.length > 0) {
                    //Allow null values for relative, x, and y
                    relative = (relative != null) ? relative : true;
                    x = (x != null) ? x : (relative ? 0 : $imageContainer.scrollLeft());
                    y = (y != null) ? y : (relative ? 0 : $imageContainer.scrollTop());

                    $imageContainer.scrollTop(relative ? $imageContainer.scrollTop() + y : y);
                    $imageContainer.scrollLeft(relative ? $imageContainer.scrollLeft() + x : x);

                    triggerTransform();
                }

                return {
                    x : $imageContainer.scrollLeft(),
                    y : $imageContainer.scrollTop()
                };
            },
            /**
             * Set the current transformations on the pages of the document.
             *
             * @method applyPageTransformations
             * @param {Object} transforms Map of page numbers to transformations to apply
             */
            applyPageTransformations : applyTransforms,
            /**
             * Get the current set of transformations applied to the document.
             *
             * @method getPageTransformations
             * @return {Object} pageTransforms Map of page numbers to transformations
             */
            getPageTransformations : function () {
                return pageTransforms;
            }
        });

        Subscribable.call(api);
        return api;
    };
});

define('primutils/ui/dialogs/drawable/shapeFns',[
    'lodash',
    'jquery',
    'logger'
], function (_, $, logger) {
    'use strict';

    var TYPES = {
        RECTANGLE : 'rectangle',
        ELLIPSE : 'ellipse',
        HIGHLIGHT : 'highlight',
        POLYGON : 'polygon',
        PIN : 'pin'
    };

    var validateContext = function (context, methodName) {
        logger.assert(context.x != null && context.y != null, 'Valid shape context must contain an `x` and `y` coordinates');
        logger.assert(shapeFns[context.type], '`' + methodName + '` must be called on a valid shape type: ' + context.type);
    };

    var shapeFns = {
        canSerialize : function () {
            return true;
        },
        preManipulate : function (context) {
            return context; //_.clone(context);
        },
        manipulate : function (context) {
            return context; //_.clone(context);
        }
    };

    shapeFns[TYPES.RECTANGLE] = {
        render : function (context, $group) {
            //It's annoying to do conversion during rendering instead of manipulation, but modifying x and y during manipulation turns out to be ill advised.
            var x = (context.width < 0) ? context.x + context.width : context.x;
            var y = (context.height < 0) ? context.y + context.height : context.y;
            var width = Math.abs(context.width) || 0;
            var height = Math.abs(context.height) || 0;

            $(document.createElementNS('http://www.w3.org/2000/svg', 'rect'))
                .attr('x', x + '%')
                .attr('y', y + '%')
                .attr('width', width + '%')
                .attr('height', height + '%')
                .appendTo($group);
        },
        preManipulate : function (context, x, y) {
            var updated = _.clone(context);

            //Create/update new properties if they don't exist yet.
            //Render is only guaranteed that x and y will exist.
            //Everything else is optional, but you can take advantage of it if context has it.
            updated.width = x - context.x;
            updated.height = y - context.y;

            return updated;
        }
    };

    shapeFns[TYPES.PIN] = {
        icon : (function () {
            var $el = $('<i class="pgbu-icon-small pgbu-icon-push-pin" style="display:none;"></i>').appendTo('body');
            //JQuery doesn't have great ways to manipulate pseudo-selectors like :before since they're technically not part of the dom.
            //Fall back on `getComputedStyle`, which is supported in IE11 up.
            //Content css will always be surrounded by ", so strip off the first and last characters.
            var glyph = window.getComputedStyle($el[0], ':before').getPropertyValue('content').slice(1, -1);
            $el.remove();
            return glyph;
        }()),
        render : function (context, $group) {
            $(document.createElementNS('http://www.w3.org/2000/svg', 'text'))
                .attr('x', context.x + '%')
                .attr('y', context.y + '%')
                .text(this.icon)
                .appendTo($group);
        },

        preManipulate : function (context, x, y) {
            var updated = _.clone(context);

            updated.x = x;
            updated.y = y;

            return updated;
        }
    };

    shapeFns[TYPES.ELLIPSE] = {
        render : function (context, $group) {
            context.xRadius = context.xRadius || 0;
            context.yRadius = context.yRadius || 0;

            var x = context.x + (context.xRadius);
            var y = context.y + (context.yRadius);
            var yRadius = Math.abs(context.yRadius);
            var xRadius = Math.abs(context.xRadius);

            $(document.createElementNS('http://www.w3.org/2000/svg', 'ellipse'))
                .attr('cx', x + '%')
                .attr('cy', y + '%')
                .attr('rx', xRadius + '%')
                .attr('ry', yRadius + '%')
                .appendTo($group);
        },

        preManipulate : function (context, x, y) {
            var updated = _.clone(context);

            updated.xRadius = (x - updated.x) / 2;
            updated.yRadius = (y - updated.y) / 2;

            return updated;
        }
    };

    shapeFns[TYPES.HIGHLIGHT] = {
        render : function (context, $group) {
            var $viewBox = $(document.createElementNS('http://www.w3.org/2000/svg', 'svg'))
                .attr('x', 0)
                .attr('y', 0)
                .appendTo($group);

            //Using JQuery would cause this property to be lowercase.
            $viewBox[0].setAttribute('viewBox', '0 0 100 100');
            $viewBox[0].setAttribute('preserveAspectRatio', 'none');

            var data = 'M' + context.x + ',' + context.y;
            _.forEach(context.points, function (point) {
                data = data + ' L' + point.x + ',' + point.y;
            });

            $(document.createElementNS('http://www.w3.org/2000/svg', 'path'))
                .attr('d', data)
                .appendTo($viewBox);
        },

        preManipulate : function (context, x, y) {
            var updated = _.clone(context); //_.clone will also handle cloning embedded arrays

            updated.points = updated.points || [];
            updated.points.push({ x : x, y : y });
            return updated;
        }
    };

    shapeFns[TYPES.POLYGON] = {
        snappingDistance : 2,

        /*
         * 'r' is not supported css in svg 1.1
         * - in order to read, we take advantage of an unused property
         * - this lets us avoid hard-coding the style here.
         */
        radius : (function () {
            var $el = $('<div class="drawing" style="display:none;">' +
                '<svg><g class="shape" shape-type="' +
                TYPES.POLYGON +
                '"></g></svg></div>'
            ).appendTo('body');

            var radius = $el.find('g').css('margin-top');
            var hoverRadius = $el.find('g').css('margin-bottom');
            $el.remove();
            return {
                base : radius,
                hover : hoverRadius
            };
        }()),

        render : function (context, $group) {
            var that = this;
            var $viewBox = $(document.createElementNS('http://www.w3.org/2000/svg', 'svg'))
                .attr('class', 'fill')
                .attr('x', 0)
                .attr('y', 0);

            /*Created two svg groups points and border, points corresponds to circle (vertex of polygon) on screen and
             * border corresponds to (sides of polygon) line. This is done in order to achieve the overlapping
             * of circle over line and to have the mouse over hover effect on points. */
            var $points = $(document.createElementNS('http://www.w3.org/2000/svg', 'g'))
                .attr('class', 'points');

            var $border = $((document).createElementNS('http://www.w3.org/2000/svg', 'g'))
                .attr('class', 'border');

            //Using JQuery would cause this property to be lowercase.
            $viewBox[0].setAttribute('viewBox', '0 0 100 100');
            $viewBox[0].setAttribute('preserveAspectRatio', 'none');

            //Draw polyline
            var data = context.x + ',' + context.y + ' ';
            _.forEach(context.points, function (point) {
                data = data + point.x + ',' + point.y + ' ';

                //circle drawn on screen will serve as a vertex of the polygon
                $(document.createElementNS('http://www.w3.org/2000/svg', 'circle'))
                    .attr('cx', point.x + '%')
                    .attr('cy', point.y + '%')

                    /*
                     * Necessary hacks for Firefox
                     * - this section can be removed once Firefox supports svg 2 ('r' property in CSS)
                     * - also covers for lack of `transform-origin` support
                     */
                    .attr('r', that.radius.base)
                    .attr('onmouseover', 'this.setAttribute("r", "' + that.radius.hover + '")')
                    .attr('onmouseout', 'this.setAttribute("r", "' + that.radius.base + '")')
                    .appendTo($points);
            });

            // drawing lines between two points.
            var from = { x : context.x, y : context.y };
            _.forEach(context.points, function (to) {
                $(document.createElementNS('http://www.w3.org/2000/svg', 'line'))
                    .attr('x1', from.x + '%')
                    .attr('y1', from.y + '%')
                    .attr('x2', to.x + '%')
                    .attr('y2', to.y + '%')
                    .appendTo($border);
                from = to;
            });

            data = data.trim();
            //to fill the shape polyline is used.
            $(document.createElementNS('http://www.w3.org/2000/svg', 'polyline'))
                .attr('points', data)
                .appendTo($viewBox);

            $viewBox.appendTo($group);
            $border.appendTo($group);
            $points.appendTo($group);
        },

        manipulate : function (context, x, y) {
            var updated = _.clone(context); //_.clone will also handle cloning embedded arrays
            if (!updated.points) { //On initial point creation, synchronize shape x and y
                updated.x = x;
                updated.y = y;
                updated.points = [];
            }

            if (_.size(updated.points) >= 3) {
                var getFirstPoint = _.first(updated.points);
                var distance = Math.sqrt(Math.pow((getFirstPoint.x - x), 2) + Math.pow((getFirstPoint.y - y), 2));
                //imaginary radius if clicked within the range, will snap it to the first point for closing the polygon
                if (distance < this.snappingDistance) {
                    x = getFirstPoint.x;
                    y = getFirstPoint.y;
                }
            }

            updated.points.push({ x : x, y : y });
            return updated;
        },

        canSerialize : function (context) {
            return ((_.size(context.points) > 3) && (_.isEqual(_.first(context.points), _.last(context.points))));
        }
    };

    var shape = {
        render : function (context, $group) {
            logger.assert($group, '`Shape.render` must not be called without passing in a target to render to.');
            if ($group[0].tagName !== 'g') {
                logger.warn('`Shape.render` should be called passing in an svg group, you passed in: ', $group[0].tagName);
            }

            $group.empty();
            validateContext(context, 'render');
            shapeFns[context.type].render(context, $group);
        },

        /*premanipulate and manipulate is use to separate out the logic of mouse drag and generalize the framework*/
        manipulate : function (context, x, y) {
            validateContext(context, 'manipulate');
            var type = shapeFns[context.type];
            return (type.manipulate || shapeFns.manipulate).call(type, context, x, y);
        },

        preManipulate : function (context, x, y) {
            validateContext(context, 'preManipulate');
            var type = shapeFns[context.type];
            return (type.preManipulate || shapeFns.preManipulate).call(type, context, x, y);
        },

        canSerialize : function (context) {
            validateContext(context, 'canSerialize');
            //Roll back to default canSerialize method if shape doesn't have it.
            var type = shapeFns[context.type];
            return (type.canSerialize || shapeFns.canSerialize).call(type, context);
        }
    };

    shape.types = TYPES; //Constants for future reference.
    return Object.freeze(shape);
});

define('primutils/ui/dialogs/drawable/DrawableView',[
    'lodash',
    'jquery',
    'logger',
    'primutils/utils/Subscribable',
    'primutils/ui/dialogs/drawable/shapeFns'
], function (_, $, logger, Subscribable, shapeFns) {
    'use strict';

    var optionDefaults = {
        //You can manually manage resizing the SVG if you want to (we do for Preview, for example)
        //But more convenient is syncing the SVG and the background-image by putting them in the same div and just adjusting the width and height of that.
        height : '100%',
        width : '100%',
        $parent : null,

        requireOverrides : {} //For testing purposes - Don't use this.
    };

    function DrawableView(options) {
        var that = this;
        logger.assert(that && that instanceof DrawableView, 'DrawableView must be instantiated with new');
        Subscribable.call(that);
        _.defaults(options, optionDefaults);

        //Look for a better way to do this.
        shapeFns = options.requireOverrides.shapeFns || shapeFns;

        var _$parent = $(options.$parent); //wrap with jQuery just for safety's sake
        Object.defineProperty(that, '$parent', {
            get : function () { return _$parent; },
            set : function (value) {
                _$parent = value;
                that.$canvas.detach(); //Force a re-attachment. Otherwise, order might get messed up.
                that.$canvas.appendTo(_$parent);
            }
        });
        Object.defineProperty(that, '$canvas', {
            writable : false,
            value : (function () {
                var $svg = $(createSVGElement('svg'))
                    .attr('width', options.width)
                    .attr('height', options.height);

                var $canvas = $('<div></div>')
                    //Use style instead of a class since width/height/top/left are all setable by user.
                    .attr('style', 'left:0; top:0; width:' + options.width + '; height:' + options.height + ';')
                    .addClass('drawing')
                    .append($svg);

                return $canvas;
            }())
        });

        //$parent is not required, so don't try to append to null. If a blank jQuery object was passed in, handle that as well.
        if (that.$parent && that.$parent.length) {
            that.$canvas.appendTo(that.$parent);
        }

        var mouseEvent = function (args, type) {
            var position = convertPercents(that.$canvas, that.$parent, args.clientX, args.clientY);
            var overlaps = getOverlaps(args.clientX, args.clientY);
            that.trigger(type || args.type, {
                type : type || args.type,
                x : position.x,
                y : position.y,
                active : overlaps.active,
                focus : overlaps.focus
            });
        };


        //Throttle mouse tracking.
        var tracking = false;
        that.$canvas.on('mousedown', function (args) {
            if (!tracking) {
                tracking = true;
                that.$canvas.select(); //Avoid dragging previous selections instead of drawing
                that.$canvas.on('mousemove', mouseEvent);
                mouseEvent(args, 'mousedown');
                args.preventDefault();
            }
        });

        /*
        * Don't listen for mouseleave.
        * Basically, there are two behaviors that this could do -
        *
        * If it doesn't listen for mouseleave, then you can drag off of the canvas, release your mouse, and drag back on.
        * - then, the event will continue until you click again.  It's a little annoying.
        *
        * If it does listen for mouseleave, when you drag off the canvas, the event immediately ends
        * - in practice, this is jarring, and breaks things like dragging over a toolbar, highlighting something, etc...
        * - basically, a lot of errors.  So of the two, the first situation is preferable, and what most apps online do.
        */
        that.$canvas.on('mouseup dragend', function (args) {
            tracking = false;
            that.$canvas.off('mousemove', mouseEvent);
            mouseEvent(args, 'mouseup');
        });
    }

    var createSVGElement = function (element, classString) {
        var $group = $(document.createElementNS('http://www.w3.org/2000/svg', element));
        toggleSVGClass($group, classString, true);

        return $group;
    };

    //SVG does not support jQuery classes, add them the old fashioned way
    var toggleSVGClass = function ($selector, classString, toggle) {
        $selector.each(function (index, element) { //Support multi-selections and blank selections
            if (!classString || classString === '') {
                return;
            }
            var $element = $(element);
            var toggling = classString.split(' ');
            var existing = ($element.attr('class') || '').split(' ');
            var result;

            //Start with destination, add everything from source that doesn't appear in filter.
            var filterBy = function (destination, source, filter) {
                result = destination;
                _.forEach(source, function (classInstance) {
                    if (filter.indexOf(classInstance) === -1) {
                        result.push(classInstance);
                    }
                });
            };

            if (toggle) {
                //For each new class, push to result if it doesn't exist in result.
                filterBy(existing, toggling, existing);
            } else {
                //For each existing class, push to result if it doesn't exist in toggling.
                filterBy([], existing, toggling);
            }
            $element.attr('class', result.join(' '));
        });
    };

    //Takes a clientX/clientY from an event, and converts them to percentage coordinates for a given element.
    var convertPercents = function ($elem, $parent, x, y) {
        //TODO: allow caching boundingClientRect somehow?
        var rect = $elem[0].getBoundingClientRect();
        return {
            x : (x - rect.left) / rect.width * 100,
            y : (y - rect.top) / rect.height * 100
        };
    };

    //Convenience function for extracting all of the shapes the mouse is overlapping.
    //Returns an object with a focus (the uppermost element), and all of the active elements (including the focus)
    //Order is not garanteed, but order will be consistent.
    var getOverlaps = function (x, y, active) {
        active = active || [];
        var $target = $(document.elementFromPoint(x, y)).closest('.shape'); //Experimental, but supported in all browsers we support.
        var targetId;
        if ($target.length !== 0) {
            //found.push($target);
            targetId = $target.attr('shape-id');
            active.push(targetId);

            $target.toggle(false); //Hide element so we can check beneath it.
            getOverlaps(x, y, active);
            $target.toggle(true); //Reset dom to original state.
        }

        return {
            focus : targetId,
            active : active
        };
    };

    DrawableView.prototype = _.create(Subscribable.prototype, {
        constructor : DrawableView,

        render : function (shapes) {
            var that = this;

            that.$canvas.find('svg:first').empty();
            _.forEach(shapes, function (shape) {
                var $group = createSVGElement('g', 'shape').appendTo(that.$canvas.find('svg:first'));
                _.forEach(shape.groups, function (group, key) {
                    $group.attr('shape-' + key, group);
                });

                shapeFns.render(shape.data, $group);
            });
        },

        //Call it on one shape.
        //If you call it on a shape that doesn't exist, it will be created.
        redraw : function (shape, dirty) {
            var that = this;

            //Get group if it exists.
            //TODO: remove hard-dependency on id existing?  model does garantee it.
            var $group = that.$canvas.find('[shape-id=' + shape.groups.id + ']');

            //Generate/redraw group if required.
            if ($group.length === 0) {
                $group = createSVGElement('g', 'shape').appendTo(that.$canvas.find('svg:first'));
                shapeFns.render(shape.data, $group);
            } else if (dirty) {
                shapeFns.render(shape.data, $group);
            }

            //Update attributes.
            _.forEach(shape.groups, function (group, key) {
                $group.attr('shape-' + key, group);
            });
        },

        //You can turn everything off by just not passing in a value (or by passing in a blank array)
        //You can use multiple values by passing in an array
        filter : function (filter, value) {
            var that = this;
            var $shapes = that.$canvas.find('.shape');
            toggleSVGClass($shapes, 'hidden', true); //TODO: check to see what the actual class here should be.

            if (value == null) {
                toggleSVGClass($shapes, 'hidden', false);
            } else if (_.isArray(value)) {
                _.forEach(value, function (valueInstance) {
                    toggleSVGClass($shapes.filter('[shape-' + filter.toLowerCase() + '="' + valueInstance + '"]'), 'hidden', false); //TODO: check to see what the actual class here should be.
                });
            } else {
                toggleSVGClass($shapes.filter('[shape-' + filter.toLowerCase() + '="' + value + '"]'), 'hidden', false); //TODO: check to see what the actual class here should be.
            }
        },

        setDimensions : function (/*width, height , x, y */) {
            var that = this;
            var args = arguments;
            //Append px if caller hasn't passed in any units.
            _.forEach(args, function (num, index) {
                if (!_.includes(num, '%') && !_.includes(num, 'px')) {
                    args[index] += 'px';
                }
            });

            //https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Strict_mode/Transitioning_to_strict_mode#arguments_doesn't_alias_named_function_arguments
            var style = 'position:absolute; width:' + args[0] + '; height:' + args[1] + '; ';
            if (arguments.length > 2) {
                style += 'left:' + args[2] + '; top:' + args[3] + ';';
            } else {
                style += 'left:0px; top:0px;';
            }

            that.$canvas.attr('style', style);
        },

        getOffset : function (id) { //TODO: finish
            var that = this;
            var $group = that.$canvas.find('[shape-id=' + id + ']');
            return $group.offset();
        },

        //Toggle should toggle the canvas.  Filter can just filter everything on its end if its so important.
        //Allows you to click right through the canvas.
        toggle : function (show) {
            var that = this;
            toggleSVGClass(that.$canvas, 'clickOff', !show);
        }
    });

    return DrawableView;
});

define('primutils/ui/dialogs/drawable/DrawableModel',[
    'lodash',
    'primutils/utils/Subscribable',
    'logger'
], function (_, Subscribable, logger) {
    'use strict';


    /* *
     * EXAMPLE STRUCTURE:
     *
     * ```
     * _nodes : [], //Unfiltered list of nodes.
     * _sort : //comparitor to use when adding new nodes.
     * _filters : {
     *         __meta__ : {
     *           bindings : ['id'] //list of filters (by key) that are bound to values within data, currently not used.
     *         },
     *
     *         'selected' : {
     *              __meta__ : {
     *                defaultValue : 0, //Value or function, what to use when adding a new item
     *              },
     *
     *             '0' : [ list of nodes ],
     *             '1' : [ list of nodes ],
     *         },
     *
     *         'id' : {
     *             __meta__ : {
     *                 defaultValue : _.uniqueId.bind(_, 'prefix'),
     *                 keyBinding : 'id', //This key on data will be updated when the group is updated
     *             },
     *
     *             'prefix1' : [{node}],
     *             'prefix2' : [{node}],
     *             //etc...
     *         },
     * },
     *
     *
     * node : {
     *         data : {
     *             id : 'prefix1',
     *             customData : 'my other custom data'
     *         },
     *       groups : {
     *           id : 'prefix1',
     *           selected : 0,
     *       },
     * }
     * ```
     *
     * Use the API to generate arrays of nodes (buffers). Pass buffers back into the API to refine them or perform other operations.
     */

    var defaultFilters = {
        id : {
            initialValue : _.uniqueId.bind(_, 'drawable')
        }
    };

     // TODO:
     // - Think about adding a `unique` tag to filters to reduce overhead with ids.
     // TODO:
     // - Think about adding 'living buffers'
    function Model(filters) {
        var that = this;
        Subscribable.call(that); //TODO: check up on some of these options.
        //assert you have to use `new`.
        _.defaults(filters, defaultFilters);

        //TODO: abstract into filter builder.
        that._filters = {};
        Object.defineProperty(that._filters, '__meta__', {
            enumerable : false,
            value : { bindings : [] }
        });

        _.forEach(filters, function (item, key) { //TODO: think about abstracting out into a `buildFilter` method.
            logger.assert(_.isString(item.initialValue) || _.isNumber(item.initialValue) || _.isFunction(item.initialValue), '`filter.initialValue` must be a string, number, or function, you passed: ' + item.initialValue);
            logger.assert(item.keyBinding == null || _.isString(item.keyBinding) || _.isNumber(item.keyBinding), '`filter.keyBinding` must be null, a string, or a number, you passed: ' + item.keyBinding);

            //Store descriptive information about this filter,
            //Once data is added, other keys on this item will be discrete values for the filter.
            that._filters[key] = {};
            Object.defineProperty(that._filters[key], '__meta__', {
                enumerable : false,
                value : item
            });

            //Cache that this is a data-binding for faster updates in the future.
            if (item.keyBinding != null) {
                that._filters.__meta__.bindings.push(key);
            }
        });

        that._nodes = []; //Un-filtered list of all of the nodes in the model.
    }

    //Will break sorts, use with caution
    var fastDelete = function (data, value) {
        if (!data || !data.length) {
            return false;
        }

        var index = data.indexOf(value);

        if (index !== -1) {
            data[index] = data[data.length - 1];
            data.pop();
            return true;
        }

        return false;
    };

    //Slower, preserves sorts.
    var slowDelete = function (data, value) {
        if (!data || !data.length) {
            return false;
        }

        var index = data.indexOf(value);
        if (index !== -1) {
            //Faster mutable version of slice
            var newLength = data.length - 1;
            for (; index < newLength; index++) {
                data[index] = data[index + 1];
            }
            data.pop();
            return true;
        }

        return false;
    };

    //Set a groups value and adjust the model's caches and filters accordingly.
    var setGroup = function (that, node, key, value) {
        logger.assert(_.isString(value) || _.isNumber(value), 'A group must be either a string or a number, you passed: ' + value);

        var filter = that._filters[key];
        var prev = node.groups[key];
        if (value === prev) {
            return value;
        }

        //Swap groups and garbage collect as necessary.
        if (prev != null && slowDelete(filter[prev], node) && filter[prev].length === 0) {
            filter[prev] = undefined;
        }
        if (!filter[value]) {
            that._filters[key][value] = [];
        }

        //Add to new group.
        filter[value].splice(_.sortedIndex(filter[value], that._sort), 0, node);

        //Only trigger update if the previous group was not null (ie. SETTING a group is different than CHANGING a group)
        if (prev != null) {
            that.trigger('group-update', {
                node : node,
                key : key,
                value : value
            });
        }

        return value;
    };

    var buildNode = function (that, data, startingGroups) {
        var groups = {};
        var node = {
            groups : groups
        };

        //Create and set data.
        var _value;
        Object.defineProperty(node, 'data', {
            enumerable : true,
            set : function (value) {
                //Loop through keyBindings and update any relevant groups
                _.forEach(that._filters.__meta__.bindings, function (binding) {
                    var keyReference = that._filters[binding].__meta__.keyBinding;
                    //Don't trigger update if the value hasn't changed, or if it's being changed to the existing value in the group.
                    if (_value[keyReference] !== value[keyReference] && value[keyReference] !== node.groups[binding]) {
                        node.groups[binding] = value[keyReference];
                    }
                });

                _value = value;
                that.trigger('data-update', {
                    node : node,
                    value : value
                });
            },
            get : function () {
                return _value;
            }
        });
        _value = data;

        //Create and propogate groups
        _.forEach(that._filters, function (filter, key) {
            var meta = filter.__meta__;

            var _value;
            Object.defineProperty(groups, key, {
                enumerable : true,
                set : function (value) {
                    _value = setGroup(that, node, key, value);

                    //If a keybinding exists, modifying that group will trigger a change on the data as well.
                    if (meta.keyBinding != null && node.data[meta.keyBinding] !== value) {
                        var newData = _.clone(node.data);
                        newData[meta.keyBinding] = value;
                        node.data = newData; //Will trigger a data-update
                    }
                },
                get : function () {
                    return _value;
                }
            });

            if (meta.keyBinding != null && data[meta.keyBinding] != null) { //If a default keybinding exists, use it, otherwise defer to the keybinding, otherwise defer to the default value.
                logger.assert(_.isString(data[meta.keyBinding]) || _.isNumber(data[meta.keyBinding]), 'Non-string value on `data` for group ' + key + ': ' + data[meta.keyBinding]);
                groups[key] = data[meta.keyBinding];
            } else if (startingGroups && startingGroups[key] != null) {
                logger.assert(_.isString(startingGroups[key]) || _.isNumber(startingGroups[key]), 'Non-string value on `startingGroup` for group ' + key + ': ' + startingGroups[key]);
                groups[key] = startingGroups[key];
            } else {
                groups[key] = _.isFunction(meta.initialValue) ? meta.initialValue() : meta.initialValue;
            }
        });

        return node;
    };

    Model.prototype = _.create(Subscribable.prototype, {
        constructor : Model,

        filter : function (key, value, buffer) {
            var that = this;

            if (arguments.length === 0) {
                return _.clone(that._nodes);
            }

            logger.assert(_.isString(key) || _.isNumber(key), 'Group to filter by must be a string or number, you passed: ' + key);
            logger.assert(_.isString(value) || _.isNumber(value), 'Value to filter by must be a string or number, you passed: ' + value);

            if (!buffer) {
                return _.clone(that._filters[key][value]) || [];
            } else {
                return _.intersection(that._filters[key][value], buffer);
            }

            return that;
        },

        //Convenience method for safely setting a group across every item in the given buffer.
        //Pass in a function for value to do fancy crud.
        setBuffer : function (buffer, key, value) {
            var that = this;
            var methodValue = _.isFunction(value);
            logger.assert(methodValue || _.isString(value) || _.isNumber(value), 'Value must be either a string, number, or function, you passed: ' + value);
            logger.assert(that._filters[key] != null, 'Attempted to set group that does not exist: ' + key);

            _.forEach(buffer, function (node) {
                node.groups[key] = methodValue ? value.call(node) : value;
            });
        },

        add : function (data, groups) {
            var that = this;
            var node = buildNode(that, data, groups);

            that._nodes.splice(_.sortedIndex(that._nodes, that._sort), 0, node);
            that.trigger('add', {
                node : node
            });

            return node;
        },

        addAll : function (dataArray, groups) {
            var that = this;
            _.forEach(dataArray, function (data) {
                that.add(data, groups);
            });
        },

        //Removes every node in the given buffer.
        remove : function (buffer) {
            var that = this;

            if (!buffer) { //rm -rf /
                buffer = that._nodes;
                that._nodes = [];
            }

            _.forEach(buffer, function (node) {
                _.forEach(that._filters, function (filter, key) {
                    //I wonder if it would be faster to do fastDelete than resort,
                    //rather than slow delete.
                    fastDelete(filter[node.groups[key]], node);
                });

                fastDelete(that._nodes, node);
            });

            that.trigger('remove', {
                removed : buffer //Return the nodes that were removed I guess?
            });
        },

        //Slow, try to only call this once.
        sortBy : function (key) {
            var that = this;
            logger.assert(_.isString(key) || _.isNumber(key), 'Key must be either number or string, you passed: ' + key);
            _.sortBy(that._nodes, key);

            //@Huh?  this...did nothing.  @Shumway
            // _.each(that._filters, function (filter) {
            //     _.each(filter, function (value) {
            //         _.sortBy(key);
            //     });
            // });

            that._sort = key;
            return true;
        }
    });

    return Model;
});

define('primutils/ui/dialogs/drawable/Drawable',[
    'lodash',
    'jquery',
    'primutils/utils/Subscribable',
    'primutils/utils/objectUtils',
    'primutils/ui/dialogs/drawable/shapeFns',
    'primutils/ui/dialogs/drawable/DrawableView',
    'primutils/ui/dialogs/drawable/DrawableModel',
    'logger'
], function (_, $, Subscribable, objectUtils, shapeFns, DrawableView, DrawableModel, logger) {
    'use strict';

    var CONSTANTS = {
        colors : {
            BLACK : 'black',
            BLUE : 'blue',
            WHITE : 'white',
            YELLOW : 'yellow',
            RED : 'red',
            GREEN : 'green'
        },
        tools : shapeFns.types
    };

    var DEFAULT_OPTIONS = {

        keys : {
            /**
             * @property keys.id - the key of the unique id attached to every shape.
             * @optional
             * @default 'id'
             */
            id : 'id',

             /**
              * @property keys.color - the key of the color of the shape. Look at Drawable.constants for acceptable values.
              * @optional
              * @default 'color'
              */
            color : 'color',

            /**
             * the key of the type of the shape. Look at Drawable.constants for acceptable values.
             * - rectangle
             * - ellipse
             * - highlight
             * - pin
             *
             * @property keys.type
             * @optional
             * @default 'type'
             */
            type : 'type'
        },

        validation : {
            /**
             * Method called after a new shape has been created.
             * The shape will not be committed to the drawable store until after the method resolves.
             *
             * You may return:
             * - A boolean. If false, the shape will be discarded. If true, the original shape will be used.
             * - An object. This will be committed to the store instead of the original object.
             * - A promise. If this resolves successfully, the object it returns will be committed to the store.
             * @property validation.add
             * @optional
             * @default returns
             */
            add : function (data) {
                return data; //Can return false, or true, or the data to add, or a promise that resolves to the data to add.
            },

            /**
            * Method called before a shape is removed.
            * - The shape will not be removed until after the method resolves.
            *
            * You may return:
            * - A boolean. If false, the shape will not be removed. If true, the shape will be removed.
            * - A promise. If this resolves successfully, the shape will be removed.
            */
            remove : function (/* data */) {
                return true; //Can return true, false, or a promise that is either resolved or rejected
            }
        },

        /**
         * Any additional keys you want to be able to filter by on your shapes.
         * @property filters
         * @optional
         * @type {Array}
         */
        filters : [],

        /**
         * What key to sort shapes by.  Affects ``selectNext`` and ``selectPrevious``.
         * @property stortBy
         * @optional
         * @default 'id'
         * @type {String}
         */
        sortBy : 'id',

        /**
         * Key and default value to group by. Use this if you want a multiple "page" effect.
         * - only shapes that match the given value will be drawn to the canvas.
         *
         * ```
         * {
         *    key : 'the shape key to group by',
         *    value : 'initial value you wish to group by',
         * }
         * ```
         * @property group
         * @optional
         * @default null
         */
        group : null, //Default value to group by. Use this if you want multiple "pages" of drawings.

        /**
         * Whether or not clicking on overlapping shapes will select all of them.
         * @property multiselect
         * @optional
         * @default false
         */
        multiselect : false,

        /**
         * Whether or not the canvas is editable
         * @property editable
         * @optional
         * @default true
         */
        editable : true,

        /**
         * When the user draws a new shape, if it passes validation, it will be auto-selected
         * @property selectOnAdd
         * @optional
         * @default true
         */
        selectOnAdd : true,

        /**
         * Container to attach drawable to.
         * - If not set, you can attach later using ``Drawable.rebuild``
         * @optional
         * @property $container
         */
        $container : null,

        /**
         *
         * @property dimensions
         * @optional
         * @default ``{ width: '100%', height: '100%' }``
         */
        dimensions : {
            width : '100%',
            height : '100%'
        },

        /**
         *
         * @property requireOverrides
         * @private
         */
        requireOverrides : {} //For testing purposes - Don't use this.
    };

    /**
     * Drawable allows a user to mark up an image or structure with custom shapes and tools.
     * -
     * @param {object} drawableOptions - Options for the component. See properties for more information.
     * @class Drawable
     * @constructor
     */
    function Drawable(drawableOptions) {
        drawableOptions = objectUtils.defaultsDeep({}, drawableOptions, DEFAULT_OPTIONS);
        DrawableView = drawableOptions.requireOverrides.DrawableView || DrawableView;
        DrawableModel = drawableOptions.requireOverrides.DrawableModel || DrawableModel;
        shapeFns = drawableOptions.requireOverrides.shapeFns || shapeFns;


        var _that = _.create(ctx); _.assign(_that, {
            _options : objectUtils.defaultsDeep(drawableOptions, DEFAULT_OPTIONS),

            _view : null,
            _model : null,

            //TODO: should use living buffers here.
            _shapes : [], //Buffer of all of the shapes in the model
            _selected : [], //Buffer of all the selected shapes in the model
            _display : [], //Buffer of the nodes currently being displayed.  Not a living buffer.

            _tempShape : null, //A shape in the act of being drawn, but not yet out of validation
            _color : null, //What the current color is.
            _tool : (function () { //What the current tool is
                var _activeTool = null;
                Object.defineProperty(_that, '_tool', {
                    get : function () {
                        return _activeTool;
                    },
                    set : function (value) {
                        _activeTool = value;

                        //Toggle clickthrough based on whether or not the active tool is null.
                        if (_that._view) {
                            _that._view.toggle(_activeTool != null);
                        }
                    }
                });
            }()),


            _trigger : function () { that.trigger.apply(that, arguments); } //Wrapper to give privates the ability to trigger events.
        });

        _that._init();
        _that._initEvents();

        var that = _.create(Subscribable.prototype, {
            /**
             * Adds a new shape to the internal store
             * - throws the ``add`` event
             * @param {object} shape - the shape to add
             * @method add
             */
            add : function (shape) {
                _that._model.add(shape);
            },
            /**
             * Adds an array of shapes to the internal store
             * @param {array} shapes - the shapes to add
             * @method addAll
             */
            addAll : function (shapes) {
                _.forEach(shapes, function (shape) {
                    that.add(shape);
                });
            },
            /**
            * Creates a temporary shape that is returned as a shape object.
            * - the shape will not be persisted to the store until ``shape.serialize`` is called.
            * - once ``shape.serialize`` is called, normal validation will occur.
            * - calling ``Drawable.create`` again before the shape is serialized will discard your old shape.
            * - respects ``options.editable``
            *
            * @param {number} x - x position of the shape
            * @param {number} y - y position of the shape
            * @returns {object} shape - A manipulable shape
            * ```
            * {
            *    id: 'temp id',
            *    manipulate : function (x, y) {}, //Forwards to shape.manipulate, won't work after serialize is called
            *    serialize : function () {}, //Forwards through regular validation
            *    discard : function () {}, //Rejects the shape and deletes it from the canvas
            * }
            * ```
            * @method create
            */
            create : function (x, y) {
                if (!_that._options.editable) {
                    return;
                }

                _that._makeShape(x, y);
                var reference = _that._tempShape;
                return {
                    id : reference.groups.id,
                    manipulate : function (x, y) {
                        if (_that._tempShape === reference) { //None of these functions will work if the reference is out of date.
                            _that._tempShape.data = shapeFns.manipulate(_that._tempShape.data, x, y);
                            _that._view.redraw(_that._tempShape, true);
                        }
                    },
                    serialize : function () {
                        if (_that._tempShape === reference) {
                            _that._serializeShape();
                        }
                    },
                    discard : function () {
                        if (_that._tempShape === reference) {
                            _that._tempShape = null;
                            _that._refreshShapes();
                            _that._view.render(_that._shapes);
                        }
                    }
                };
            },
            /**
             * Selects a shape
             * - throws the ``selected`` event
             * @property {string} id - unique id of the shape to select.
             * @method select
             */
            select : function (id) {
                _that._selectShape(id, []);
            },
            /**
             * Selects the next shape after the provided id.
             * - if no id is provided, will use the currently focused shape.
             * - if no shape is focused, will default to the first shape.
             * - if no shapes exist, will not select anything.
             * @param {string} id - unique id of the shape to use as a reference
             * @method selectNext
             */
            selectNext : function (id) {
                id = id || (_.first(_that._selected) ? _.first(_that._selected).groups.id : null);
                var next = _.findIndex(_that._display, function (shape) {
                    return shape.groups.id === id;
                });

                //If the current selection doesn't exist, try for the last element in the _display array.
                if (next === -1) {
                    next = _that._display.length - 1;
                }

                if (next !== -1) { //If next is still negative 1, then you don't have anything *to* select
                    next = (next + 1) % _that._display.length;
                    that.select(_that._display[next].groups.id);
                }
            },
            /**
             * Selects the previous shape after the provided id.
             * - if no id is provided, will use the currently focused shape.
             * - if no shape is focused, will default to the last shape.
             * - if no shapes exist, will not select anything.
             * @param {string} id - unique id of the shape to use as a reference
             * @method selectPrev
             */
            selectPrev : function (id) {
                id = id || (_.first(_that._selected) ? _.first(_that._selected).groups.id : null);
                var prev = _.findIndex(_that._display, function (shape) {
                    return shape.groups.id === id;
                });

                if (prev !== -1) {
                    prev = Math.abs((prev - 1) % _that._display.length); //Adjust for negative modulus operators
                    that.select(_that._display[prev].groups.id);
                } else if (_that._display.length !== 0) {
                    that.select(_that._display[_that._display.length - 1].groups.id); //If you can't find anything else, try to select the last element.
                }
            },
            /**
             * Removes a shape with a given id
             * @param {string} id - unique id of the shape to remove
             * @method remove
             */
            remove : function (id) {
                var shapes = _that._model.filter('id', id);
                var shape = _.first(shapes);
                logger.assert(shape, 'Attempted to remove a shape that does not exist: ' + id);

                var valid = that._options.validation.remove(shape.data);
                $.when(valid).done(function (result) {
 //Accept promises

                    //Mirror JQuery's approach to filtering out non-promise results
                    if (result || (valid && _.isFunction(valid.promise))) {
                        _that._model.remove(shapes);
                    }
                }).always(function () {
                    _that._refreshShapes();
                    _that._view.render(_that._display);
                });
            },
            /**
            * Hides every shape that does not match a given criteria
            * @param {string} group - the group to filter by
            * @param {string|array|null} values - the values to filter by
            * @method filter
            */
            filter : function (group, values) {
                values = _.isArray(values) ? values : [values];

                //Reset previous filter
                _that._refreshShapes();
                _that._view.render(_that._display);

                //Update display filter to match current view
                //TODO: should ``View.filter`` sometime in the future be able to pass this kind of information back out? At the very least, ``Model.filter`` should think about obeying the same rules
                if (values[0] != null || !values.length) { //If a null value passed in, accept any value (don't filter).  If no values are passed in, hide everything.
                    _that._display = [];
                    _.forEach(values, function (value) {
                        _that._display = _that._display.concat(_that._model.filter(group, value, _that._shapes));
                    });
                }
                //
                _that._view.filter.apply(_that._view, arguments);
            },
            /**
            * Hides every shape on the canvas
            * @param {boolean} hide - ``false`` will hide, ``true`` will show
            * @method toggle
            */
            toggle : _that._view.toggle.bind(_that._view), //Eventually may need to hide other things like toolbar? Throw event?
            /**
             * Gets the matching shape for a given id
             * @param {number} id - id of the shape to return
             * @method getShape
             */
            getShape : function (id) {
                var shape = _.first(_that._model.filter('id', id));
                return shape ? shape.data : null;
            },
            /**
            * Sets the value of the current group.
            * - will trigger a redraw of the canvas
            * @param {string} value
            * @method setGroup
            */
            setGroup : function (value) {
                if (!_that._options.group) {
                    return;
                }

                _that._options.group.value = value;
                _that._refreshShapes();
                _that._view.render(_that._shapes);
            },
            /**
            * Set the current color.
            * - Will clear any temporary shapes if called before serialization
            * @param {string} color
            * @method setColor
            */
            setColor : function (color) {
                if (_that._tempShape && _that._tempShape.groups.color !== color) {
                    _that._tempShape = null;
                }

                _that._color = color;
            },
            /**
            * Set the current tool.
            * - Will be set back to null after the user finishes drawing
            * - While the tool is active, mouse events on the element drawable is covering won't register.
            * - Will clear any temporary shapes if called before serialization and tool does not match
            * @param {string} tool
            * @method setTool
            */
            setTool : function (tool) {
                if (tool != null && _that._tempShape) {
                    _that._tempShape = null;
                }
                _that._tool = tool;
            },

            /**
            * Reattach drawable to a target container and give it a new width and height.
            * @param {jQuery} $container - target to attach to
            * @param {string} width
            * @param {string} height
            * @method rebuild
            */
            rebuild : function ($container, width, height /*, x, y*/) {
                _that._options.$container = $container;
                _that._view.$parent = $container;

                if (arguments.length > 1) {
                    _that._options.dimensions = {
                        width : width,
                        height : height
                    };
                    _that._view.setDimensions.apply(that._view, Array.prototype.slice.call(arguments, 1));
                }
            }

        });

        Subscribable.call(that);
        return _.assign(that, _that);
    }

    var ctx = {
        _init : function () {
            var that = this;

            var filters = {
                id : {
                    initialValue : _.uniqueId.bind(_, '-'),
                    keyBinding : that._options.keys.id
                },
                color : {
                    initialValue : 'black',
                    keyBinding : that._options.keys.color
                },
                type : {
                    initialValue : 'pin',
                    keyBinding : that._options.keys.type
                },
                selected : {
                    initialValue : 'off'
                },
                hover : {
                    initialValue : 'off'
                }
            };

            //Attach custom filters
            _.forEach(that._options.filters, function (key) {
                logger.assert(_.isString(key), 'Attempted to add a non-string filter to Drawable: ' + key);
                logger.assert(!filters[key], 'Attempted to add a filter to Drawable that already exists: ' + key);

                filters[key] = {
                    initialValue : function () {
                        logger.error('No value was found on shape for custom filter: ' + key + '. Every shape must include a key and value for every custom filter.');
                    },
                    keyBinding : key
                };
            });

            //Attach custom group if it exists
            if (that._options.group) {
                filters[that._options.group.key] = {
                    initialValue : function () {
                        logger.error('No value was found on shape for custom group: ' + that._options.group.key + '. Every shape must include a key and value for the base group.');
                    },
                    keyBinding : that._options.group.key
                };
            }

            that._model = new DrawableModel(filters);
            that._view = new DrawableView({
                $parent : that._options.$container,
                height : that._options.dimensions.height,
                width : that._options.dimensions.width
            });

            //Default state for tools.
            that._tool = null;
            that._color = CONSTANTS.colors.BLACK;
        },

        _initEvents : function () {
            var that = this;
            that._view.on('mousedown', function (args) {
                if (that._options.editable && that._tool && !that._tempShape) {
                    that._makeShape(args.x, args.y);
                    that._view.redraw(that._tempShape, true);
                } else if (!that._tempShape) {
                    that._selectShape(args.focus, args.active);
                }
            });

            that._view.on('mousemove', function (args) {
                if (that._options.editable && that._tool && that._tempShape != null) {
                    that._tempShape.data = shapeFns.preManipulate(that._tempShape.data, args.x, args.y);
                    that._view.redraw(that._tempShape, true);
                }
            });

            that._view.on('mouseup', function (args) {
                if (!that._tempShape) {
                    return;
                }

                that._tempShape.data = shapeFns.manipulate(that._tempShape.data, args.x, args.y);
                that._view.redraw(that._tempShape, true);

                if (shapeFns.canSerialize(that._tempShape.data)) {
                    that._serializeShape();
                    that._tool = null;
                }
            });

            that._model.on('add', function (args) {
                that._view.redraw(args.node);
                that._trigger('add', {
                    shape : args.node.data
                });
            });

            that._model.on('remove', function (args) {
                that._refreshShapes(); //Will be obsolete if model ever supports living buffers.
                that._view.render(that._shapes);
            });
        },

        _refreshShapes : function () {
            var that = this;
            if (that._options.group) {
                that._shapes = that._model.filter(that._options.group.key, that._options.group.value);
            } else {
                that._shapes = that._model.filter();
            }

            that._display = that._shapes; //Since we use a functional mode for filtering, it's fine for this assignment to happen.
        },

        _selectShape : function (focus, active) {
            var that = this;
            if (focus != null && !that._model.filter('id', focus || '')[0]) {
                return;
            }

            var setSelected = function (id, value) {
                var shape = that._model.filter('id', id || '')[0];
                if (shape) {
                    shape.groups.selected = value;
                }
            };

            //Reset shape selection status to none.
            that._model.setBuffer(that._selected || [], 'selected', 'off');

            setSelected(focus, 'on');
            if (that._options.multiselect) {
                _.forEach(active, function (shape) {
                    setSelected(shape, 'on');
                });
            }

            that._selected = that._model.filter('selected', 'on', that._shapes); //Right now we need to refilter, but in the future this should use a living buffer.
            that._view.render(that._display);

            //TODO: should only throw this event if the selection *changed*.
            that._trigger('selected', {
                focus : focus,
                selected : that._selected
            });
        },

        _makeShape : function (x, y) {
            var that = this;
            var data = {
                id : _.uniqueId('-'),
                type : that._tool,
                color : that._color
            };

            that._tempShape = {
                data : _.assign({}, data, {
                    x : x,
                    y : y
                }),
                groups : data
            };

             //Remove any previous temp shapes and re-render
            that._refreshShapes();
            that._view.render(that._shapes);
            //Add new temp shape to canvas
            that._view.redraw(that._tempShape);

            //Auto-select new shape
            that._selectShape(that._tempShape.groups.id);
        },

        _serializeShape : function () {
            var that = this;
            var valid = that._options.validation.add(that._tempShape.data);
            var reference = that._tempShape.data;
            that._tempShape = null; //Allow other 'creates' in the meantime.

            $.when(valid).done(function (result) { //Accept promises
                var node;
                if (result) {
                    node = that._model.add(_.isPlainObject(result) ? result : reference);
                    if (that._options.selectOnAdd) {
                        that._selectShape(node[that._options.keys.id]); //Potentially should trigger on every `add`, not just on shapes made through `_makeShape`.
                    }
                }
            }).always(function () {
                that._refreshShapes();
                that._view.render(that._shapes);
            });
        }
    };

    Drawable.constants = CONSTANTS;
    return Drawable;
});


define('primutils/markup/primutils/ui/widgets/dockDisabledEmptyState.hbs', ['handlebars','primutils.markup'], function (H, u) {
var imports = u.buildImported(Array.prototype.slice.call(arguments, 2), 'primutils/ui/widgets/dockDisabledEmptyState.hbs'),
    func = H.template({"1":function(depth0,helpers,partials,data) {
  return "";
},"compiler":[6,">= 2.0.0-beta.1"],"main":function(depth0,helpers,partials,data) {
  var stack1, helperMissing=helpers.helperMissing;
  stack1 = ((helpers.emptyState || (depth0 && depth0.emptyState) || helperMissing).call(depth0, "info", "", {"name":"emptyState","hash":{
    'style': ("height: 100%; overflow:hidden; box-sizing: border-box; padding-top: 0;"),
    'class': ("empty-state-small"),
    'static': (true)
  },"fn":this.program(1, data),"inverse":this.noop,"data":data}));
  if (stack1 != null) { return stack1; }
  else { return ''; }
  },"useData":true});
return function (context, options) {
    options = u.extendOptions(options || {}, imports);
    return func(context, options).trim();
};
});

/* eslint-disable no-nested-ternary */
/**
 * 'primutils/ui/widgets/dockControlledSplitter' module.
 *
 * This module helps manage detail panels which use docks.
 * It automatically resizes the master pane, can programatically expand/collapse the dock, and can lock the dock when it should not be available.
 *
 * This function can be invoked with or without `new`.
 *
 * @class ui.widgets.dockControlledSplitter
 * @module ui.widgets
 * @param {jQuery} $parentElement The jQuery element which contains the master pane and the dock. Those **must** be the only two child elements and the dock must have the `.dock` class.
 * @param {Object} initOptions Initial options for the object, see the Properties tab.
 * @constructor
 */
define('primutils/ui/widgets/dockControlledSplitter',[
    'jquery', 'lodash', 'logger', 'primutils/utils/language', 'primutils/markup/primutils/ui/widgets/dockDisabledEmptyState.hbs', 'primutils/utils/grid'
], function ($, _, logger, localizer, emptyStateMarkup, gridUtils) {
    'use strict';

    var DEFAULT_OPTIONS = {
        /**
         * Is the user allowed to toggle the state of the dock?
         *
         * @property allowCollapse
         * @type Boolean
         * @default true
         */
        allowCollapse : true,

        useEmptyStateLock : true,
        getEmptyStateInfo : _.noop,

        push : {
            bottom : true,
            right : false
        },

        /**
         * Event listeners
         *
         *     {
         *         // Fired when the dock is expanded or collapsed
         *         collapseStateChanged : function (isCollapsed) { ... },
         *
         *         // Fired when the dock is resized (throttled)
         *         splitResizeThrottled : function (masterPaneRatio) { ... },
         *
         *         // Fired when the dock is collapsed
         *         detailPanelCollapsed : function () { ... }
         *     }
         *
         * @property events
         * @type {{collapseStateChanged:Function,splitResizeThrottled:Function,detailPanelCollapsed:Function}}
         * @default No event listeners
         */
        events : {
            collapseStateChanged : $.noop,
            splitResize : $.noop,
            splitResizeThrottled : $.noop,
            detailPanelCollapsed : $.noop,
            disableStateChanged : $.noop
        }
    };

    return function ($parentElement, initOptions) {
        var options = $.extend(true, {}, DEFAULT_OPTIONS, initOptions || {});

        var $children = $parentElement.children(),
            $master = $children.first(),
            $dock = $children.last(),
            $panelContainer = $parentElement.find('.panel-container'),
            $dockEmptyState = $(),
            userExpanded = !$panelContainer.hasClass('collapsed'),
            programaticCollapse = false;

        //TODO convert these to asserts eventually (don't want to break existing pages right now)
        if ($children.length !== 2) {
            logger.error('dockControlledSplitter.js | The $parentElement must only contain a master pane and a dock');
        }
        if (!$dock.hasClass('dock')) {
            logger.error('dockControlledSplitter.js | The second child in the parent does not appear to be a .dock');
        }

        if (options.useEmptyStateLock) {
            $dockEmptyState = $(emptyStateMarkup({})).addClass('hidden').appendTo($dock);
            $dockEmptyState.children().first().css('margin-top', 40); // Weird scrolling problem
        }

        // Ensure parent is .dock-parent
        $parentElement.addClass('dock-parent');

        function triggerResizes() {
            if (gridUtils.gridApiLoaded()) {
                gridUtils.findGrids($parentElement).grid('resize');
            }
            $(window).trigger('resize'); // For other things listening to window resizes
        }

        function isDockBottom() {
            return $dock.hasClass('dock-bottom');
        }

        function dockHeightPercent() {
            // NOTE: Technically this is wrong, but it keeps the master panel from intersecting the dock collapse handle
            return $dock.outerHeight() / $parentElement.height();
        }

        function dockWidthPercent() {
            // NOTE: Technically this is wrong, but it keeps the master panel from intersecting the dock collapse handle
            return $dock.outerWidth() / $parentElement.width();
        }

        function resizeMasterPane() {
            if (isDockBottom()) {
                var masterHeightPercent = 1 - dockHeightPercent();
                $master.css({
                    height : options.push.bottom ? (masterHeightPercent * 100).toString() + '%' : '100%',
                    width : '100%'
                });
            } else {
                var masterWidthPercent = 1 - dockWidthPercent();
                $master.css({
                    height : '100%',
                    width : options.push.right ? (masterWidthPercent * 100).toString() + '%' : '100%'
                });
            }
        }

        var api = {
            /**
             * Gets the ratio of the master pane to the parent
             *
             * @method getSplitterSize
             * @returns {Number}
             */
            getSplitterSize : function () {
                return $master.height() / $parentElement.height();
            },

            getSplitterViewData : function () {
                $dock.prim_dock();
                var dock = $dock.data('prim_dock'),
                    dockedBottom = isDockBottom(),
                    isCollapsed = this.isCollapsed(),
                    storedHeightPercent = dock._storedHeight ? (parseInt(dock._storedHeight, 10) / $parentElement.height()) : undefined,
                    storedWidthPercent = dock._storedWidth ? (parseInt(dock._storedWidth, 10) / $parentElement.width()) : undefined;

                return {
                    orientation : dockedBottom ? 'bottom' : 'right',
                    horizontalSize : dockedBottom ? (isCollapsed ? storedHeightPercent : $dock.height() / $parentElement.height()) : storedHeightPercent,
                    verticalSize : !dockedBottom ? (isCollapsed ? storedWidthPercent : $dock.width() / $parentElement.width()) : storedWidthPercent,
                    collapsed : !userExpanded
                };
            },

            applySplitterViewData : function (viewData, opts) {
                $dock.prim_dock();
                var dock = $dock.data('prim_dock');
                if (viewData.orientation) {
                    dock.orient(viewData.orientation === 'bottom' ? 'horizontal' : 'vertical');
                }
                if (_.has(viewData, 'collapsed')) {
                    userExpanded = !viewData.collapsed;
                    if (!opts || !opts.skipToggleRedraw) {
                        this.toggleBottomPane(userExpanded);
                    }
                }
                if (viewData.horizontalSize) {
                    dock._storedHeight = (viewData.horizontalSize * $parentElement.height());
                    if (!this.isCollapsed() && isDockBottom()) {
                        $dock.height(dock._storedHeight);
                        resizeMasterPane();
                        triggerResizes();
                    }
                } else {
                    delete dock._storedHeight;
                }
                if (viewData.verticalSize) {
                    dock._storedWidth = (viewData.verticalSize * $parentElement.width());
                    if (!this.isCollapsed() && !isDockBottom()) {
                        $dock.width(dock._storedWidth);
                        resizeMasterPane();
                        triggerResizes();
                    }
                } else {
                    delete dock._storedWidth;
                }
            },

            /**
             * Forces the master panel to resize
             *
             * @method forceResize
             */
            forceResize : function () {
                resizeMasterPane();
            },

            /**
             * Expands or collapses the dock (if it's not disabled)
             *
             * @method toggleBottomPane
             * @param {Boolean} expand True to expand, false to collapse
             */
            toggleBottomPane : function toggleSplitter(expand) {
                var isCollapsed = this.isCollapsed();
                if (expand == null) {
                    expand = isCollapsed;
                }

                var delayed = false,
                    callback = function () {
                        programaticCollapse = true;
                        $dock.prim_dock('toggle', expand);
                        programaticCollapse = false;
                    };

                if (expand) {
                    gridUtils.selectGrids().each(function () {
                        if ($(this).data(gridUtils.gridDataName).view.currentEditor != null) {
                            delayed = true;
                            $(this).one('grid-editor-closed', callback);
                        }
                    });
                }
                if (!delayed) {
                    callback();
                }
            },

            /**
             * Enables or disable the dock
             *
             * @method toggleDisable
             * @param {Boolean} disable True to disable, false to re-enable
             */
            toggleDisable : function (disable) {
                var wasAlreadyDisabled = this.isDisabled(),
                    wasAlreadyCollapsed = this.isCollapsed();
                if (options.useEmptyStateLock) {
                    $dock.find('.dock-body').toggleClass('hidden', disable);
                    fixEmptyStateWidth();
                    $dockEmptyState.toggleClass('hidden', !disable);
                    if (disable) {
                        var emptyStateInfo = options.getEmptyStateInfo() || {};
                        $dockEmptyState.find('.empty-state-message').text(emptyStateInfo.message || localizer.getString('message.global_no_details_select_other'));
                        $dockEmptyState.find('.empty-state-header').text(emptyStateInfo.header || localizer.getString('label.global_info'));
                    } else if (userExpanded && wasAlreadyCollapsed) {
                        this.toggleBottomPane(true);
                        var newSplitSize = api.getSplitterSize();
                        triggerResizes();
                        options.events.collapseStateChanged(false);
                        options.events.splitResize(newSplitSize);
                        options.events.splitResizeThrottled(newSplitSize);
                    }
                    options.events.disableStateChanged(disable);
                } else {
                    if (disable) {
                        this.toggleBottomPane(!disable);
                    }
                    $dock.find('.collapse-handle').toggleClass('hidden', disable);
                    if (!disable && userExpanded && wasAlreadyDisabled) {
                        this.toggleBottomPane(true);
                    }
                }
                if (disable && !wasAlreadyDisabled) {
                    $dock.trigger('dock-disabled');
                } else if (!disable && wasAlreadyDisabled) {
                    $dock.trigger('dock-enabled');
                }
            },

            /**
             * Is the dock disabled
             *
             * @method isDisabled
             * @returns {Boolean}
             */
            isDisabled : function () {
                if (options.useEmptyStateLock) {
                    return !$dockEmptyState.hasClass('hidden');
                } else {
                    var $collapseHandle = $dock.find('.collapse-handle');
                    return $collapseHandle.length && $collapseHandle.hasClass('hidden');
                }
            },

            /**
             * Is the dock collapsed
             *
             * @method isCollapsed
             * @returns {Boolean}
             */
            isCollapsed : function () {
                return $dock.hasClass('collapsed');
            },

            /**
             * Is the dock collapsed and was the user the reason (vs. a programatic collapse)
             *
             * @method isUserCollapsed
             * @returns {Boolean}
             */
            isUserCollapsed : function () {
                return this.isCollapsed() && !userExpanded;
            },

            /**
             * (Not an option; exposed on returned object)
             *
             * The collapse handle on the dock
             *
             * @property $collapseHandle
             * @type jQuery
             */
            get $collapseHandle() {
                return $dock.find('.collapse-handle');
            }
        };

        function fixEmptyStateWidth() {
            $dockEmptyState.css('width', (api.isCollapsed() && !isDockBottom()) ? '0' : '');
        }

        api.debouncedToggleDisable = _.debounce(api.toggleDisable, 0);

        if (api.isCollapsed()) {
            // If the dock starts collapsed, update the master pane
            resizeMasterPane();
            triggerResizes();
        } else {
            // If not, resize tabbars (since IE won't do an initial resize)
            $dock.find('.nav-tabs').prim_tabbar('resize');
        }

        // If the master has overflow hidden, do not allow scrolling (IE9 only problem)
        $master.on('scroll', function () {
            if ($master.css('overflow-y') === 'hidden') {
                $master.scrollTop(0);
            }
        });


        $dock.on('dock-collapsing dock-expanding', function (e) {
            if (!options.useEmptyStateLock && api.isDisabled()) { e.preventDefault(); }
        });

        $dock.on('dock-collapsing', function () {
            if (!api.isDisabled()) {
                $dock.prim_dock();
                if (isDockBottom()) {
                    $dock.data('prim_dock')._storedHeight = $dock.css('height');
                } else {
                    $dock.data('prim_dock')._storedWidth = $dock.css('width');
                }
            }
        });

        $dock.on('dock-expanded', function () {
            if (!api.isDisabled()) {
                $dock.prim_dock();
                var dock = $dock.data('prim_dock');
                if (isDockBottom()) {
                    var oldWidth = dock._storedWidth;
                    dock.orient('horizontal');
                    dock._storedWidth = oldWidth;
                } else {
                    var oldHeight = dock._storedHeight;
                    dock.orient('vertical');
                    dock._storedHeight = oldHeight;
                }
            }
        });

        $dock.on('dock-collapsed dock-expanded dock-resized dock-orientation-changed', function (e) {
            var stateChange = e.type === 'dock-collapsed' || e.type === 'dock-expanded',
                isExpanded = e.type !== 'dock-collapsed',
                newSplitSize = api.getSplitterSize();

            fixEmptyStateWidth();

            resizeMasterPane();
            triggerResizes();
            if (stateChange) {
                options.events.collapseStateChanged(!isExpanded);
            }
            options.events.splitResize(newSplitSize);
            options.events.splitResizeThrottled(newSplitSize);

            if (stateChange && !isExpanded) {
                options.events.detailPanelCollapsed();
            }

            if (!programaticCollapse && stateChange) {
                userExpanded = isExpanded;
            }
        });

        //TODO why is this necessary on some pages but not others?
        $dock.on('shown.tab.data-api', '.nav-tabs [data-toggle="tab"]', function () {
            var $this = $(this),
                pane = $this.attr('data-target'),
                $pane;

            if (!pane) {
                pane = $this.attr('href');
                pane = pane && pane.replace(/.*(?=#[^\s]*$)/, ''); //strip for ie7
            }

            $pane = $(pane);
            if (gridUtils.gridApiLoaded()) {
                gridUtils.findGrids($pane).filter(':visible').grid('resize');
                gridUtils.filterGrids($pane).grid('resize');
            }
        });

        return api;
    };
});

define('primutils/ui/widgets/tabOrderManager',['jquery', 'lodash', 'logger'], function ($, _, logger) {
    'use strict';

    function TabOrderManager($tabbar) {
        this.$tabbar = $tabbar;
        logger.assert($tabbar.hasClass('nav-tabs'), 'tab order manager must be run on .nav-tabs');
    }

    var toggleTabs = function (tabs, beVisible) {
        this.$tabbar.find('[data-target]').parent().toggleClass('scroll-tab-user-hidden', beVisible);
        _.forEach(tabs, function (tab) {
            this.$tabbar.find('[data-target="' + tab + '"]').parent().toggleClass('scroll-tab-user-hidden', !beVisible);
        }.bind(this));
        this.$tabbar.prim_tabbar('resize');
    };

    TabOrderManager.prototype = {
        constructor : TabOrderManager,

        getTabOrder : function () {
            return _.toArray(this.$tabbar.find('[data-toggle="tab"]').map(function () {
                return $(this).attr('data-target');
            }));
        },

        setTabOrder : function (tabOrder) {
            var that = this,
                $lastTab;
            _.forEach(tabOrder, function (tabName) {
                var $tab = that.$tabbar.find('[data-target="' + tabName + '"]').parent();
                if ($tab.length) {
                    if ($lastTab) {
                        $tab.insertAfter($lastTab);
                    } else {
                        $tab.prependTo(that.$tabbar);
                    }
                    $lastTab = $tab;
                } else {
                    logger.warn('Could not find a tab for: ' + tabName);
                }
            });
        },

        getActiveTab : function () {
            return this.$tabbar.find('.active [data-toggle="tab"]').attr('data-target');
        },

        setActiveTab : function (tabName) {
            var $tab = this.$tabbar.find('[data-target="' + tabName + '"]');
            if ($tab.length) {
                $tab.click().parent().removeClass('scroll-tab-user-hidden');
            } else {
                logger.warn('Could not find a tab for: ' + tabName);
            }
        },

        getHiddenTabs : function () {
            return _.map(this.$tabbar.find('.scroll-tab-user-hidden'), function (tab) {
                return $(tab).children('[data-target]').attr('data-target');
            });
        },

        getVisibleTabs : function () {
            return _.map(this.$tabbar.find('[data-toggle="tab"]').parent(':not(.scroll-tab-user-hidden)').children('[data-target]'), function (tab) {
                return $(tab).attr('data-target');
            }).concat(_.map(this.$tabbar.find('li.header'), function (tab) {
                return $(tab).attr('data-target');
            }));
        },

        setHiddenTabs : function (tabs) {
            toggleTabs.call(this, tabs, false);
        },

        setVisibleTabs : function (tabs) {
            toggleTabs.call(this, tabs, true);
        },

        supportsReordering : function () {
            return this.$tabbar.is('[data-sortable]');
        },

        supportsHiding : function () {
            return this.$tabbar.is('[data-allow-hide]');
        }
    };

    return TabOrderManager;
});

define('primutils/ui/VirtualScroller',[
    'jquery',
    'lodash',
    'logger'
], function ($, _, logger) {
    'use strict';

    var _virtualScrollUUID = 0;

    var DEFAULT_OPTIONS = {
        initialize : _.noop
    };

    var maybeInitialize = function ($item) {
        var $result;
        if (!$item[0]._primVirtualScroll.initialized) {
            $result = this._options.initialize($item);
            $item[0]._primVirtualScroll.initialized = true;
            if (!_.isUndefined($result)) {
                this.replace($item, $result);
                return $result;
            }
        }
        return $item;
    };

    var redrawDown = function (newScroll, scrollAmount, elHeight) {
        var $toMove = _.toArray(this._$el.children().not(this._$virtualTop).not(this._$virtualBottom)),
            $firstChild = this._$virtualTop.next(),
            firstOffset = ($firstChild.length && !$firstChild.is(this._$virtualBottom)) ? scrollAmount + getRelativeOffset.call(this, $firstChild) : 0,
            $lastChild = this._$virtualBottom.prev(),
            lastNewHeight = ($lastChild.length && !$lastChild.is(this._$virtualTop)) ? $lastChild[0]._primVirtualScroll.height : 0,
            scrollGained = lastNewHeight > 0 ? scrollAmount + getRelativeOffset.call(this, $lastChild) + lastNewHeight : 0;

        var $bottomNext = this._$bottomStorage.children().first(),
            bottomNextHeight = $bottomNext.length && $bottomNext[0]._primVirtualScroll.height,
            bottomHeightLost = 0;

        while ($bottomNext.length && (scrollGained + bottomNextHeight) < scrollAmount) {
            $toMove.push($bottomNext[0]);
            scrollGained += bottomNextHeight;
            bottomHeightLost += bottomNextHeight;

            $bottomNext = $bottomNext.next();
            bottomNextHeight = $bottomNext.length && $bottomNext[0]._primVirtualScroll.height;
        }

        this._$virtualTop.height(this._$virtualTop.height() + scrollGained - firstOffset);
        $bottomNext.insertBefore(this._$virtualBottom);
        maybeInitialize.call(this, $bottomNext);
        this._$virtualBottom.height(this._$virtualBottom.height() - bottomHeightLost - bottomNextHeight);
        this._$topStorage.append($toMove);

        refill.call(this, elHeight);
    };

    var redrawUp = function (newScroll, scrollAmount, elHeight) {
        var $toMove = _.toArray(this._$el.children().not(this._$virtualTop).not(this._$virtualBottom)),
            $firstChild = this._$virtualTop.next(),
            firstOffset = ($firstChild.length && !$firstChild.is(this._$virtualBottom)) ? scrollAmount + getRelativeOffset.call(this, $firstChild) : 0,
            $lastChild = this._$virtualBottom.prev(),
            lastNewHeight = ($lastChild.length && !$lastChild.is(this._$virtualTop)) ? $lastChild[0]._primVirtualScroll.height : 0,
            lastOffsetBottom = scrollAmount + getRelativeOffset.call(this, $lastChild) + lastNewHeight,
            scrollGained = 0;

        var $topPrev = this._$topStorage.children().last(),
            topPrevHeight = $topPrev.length && $topPrev[0]._primVirtualScroll.height,
            topHeightLost = 0;

        while ($topPrev.prev().length && (scrollGained + topPrevHeight) < -scrollAmount) {
            $toMove.unshift($topPrev[0]);
            scrollGained += topPrevHeight;
            topHeightLost += topPrevHeight;

            $topPrev = $topPrev.prev();
            topPrevHeight = $topPrev.length && $topPrev[0]._primVirtualScroll.height;
        }

        this._$virtualTop.height(this._$virtualTop.height() - topHeightLost - topPrevHeight);
        $topPrev.insertAfter(this._$virtualTop);
        maybeInitialize.call(this, $topPrev);
        this._$virtualBottom.height(this._$virtualBottom.height() + topHeightLost + (lastOffsetBottom - firstOffset));
        this._$bottomStorage.prepend($toMove);

        refill.call(this, elHeight);
    };

    var syncHeights = function () {
        this._$el.children().not(this._$virtualTop).not(this._$virtualBottom).each(function () {
            var $this = $(this);
            $this[0]._primVirtualScroll.height = getOuterHeight($this[0]);
        });
    };

    var redraw = function (newScroll, scrollAmount, elHeight) {
        syncHeights.call(this);
        if (scrollAmount > 0) {
            redrawDown.call(this, newScroll, scrollAmount, elHeight);
        } else {
            redrawUp.call(this, newScroll, scrollAmount, elHeight);
        }
    };

    var refill = function (elHeight) {
        var $firstChild = this._$virtualTop.next(),
            firstNewHeight = ($firstChild.length && !$firstChild.is(this._$virtualBottom)) ? ($firstChild[0]._primVirtualScroll.height = getOuterHeight($firstChild[0])) : 0,
            $lastChild = this._$virtualBottom.prev(),
            lastNewHeight = ($lastChild.length && !$lastChild.is(this._$virtualTop)) ? ($lastChild[0]._primVirtualScroll.height = $lastChild.outerHeight(true)) : 0;

        // Remove children above visible area
        while (!$firstChild.is(this._$virtualBottom) && getRelativeOffset.call(this, $firstChild) + firstNewHeight < 0) {
            // Add padding first to prevent not being able to scroll to last item
            this._$virtualTop.height(this._$virtualTop.height() + firstNewHeight);
            $firstChild.appendTo(this._$topStorage);

            $firstChild = this._$virtualTop.next();
            firstNewHeight = ($firstChild.length && !$firstChild.is(this._$virtualBottom)) ? ($firstChild[0]._primVirtualScroll.height = getOuterHeight($firstChild[0])) : 0;
        }

        // Remove children below visible area
        while (!$lastChild.is(this._$virtualTop) && getRelativeOffset.call(this, $lastChild) + lastNewHeight > elHeight) {
            // Add padding first to prevent not being able to scroll to last item
            this._$virtualBottom.height(this._$virtualBottom.height() + lastNewHeight);
            $lastChild.prependTo(this._$bottomStorage);

            $lastChild = this._$virtualBottom.prev();
            lastNewHeight = ($lastChild.length && !$lastChild.is(this._$virtualTop)) ? ($lastChild[0]._primVirtualScroll.height = getOuterHeight($lastChild[0])) : 0;
        }

        // Add children to fill open top space
        while (!$firstChild.is(this._$virtualTop) && getRelativeOffset.call(this, $firstChild) + firstNewHeight > 0) {
            $firstChild = this._$topStorage.children().last();
            if ($firstChild.length === 0) { break; }
            firstNewHeight = $firstChild[0]._primVirtualScroll.height;
            this._$virtualTop.height(this._$virtualTop.height() - firstNewHeight);
            $firstChild.insertAfter(this._$virtualTop);
            $firstChild = maybeInitialize.call(this, $firstChild);
        }

        // Add children to fill open bottom space
        while (!$lastChild.is(this._$virtualTop) && getRelativeOffset.call(this, $lastChild) + lastNewHeight < elHeight) {
            $lastChild = this._$bottomStorage.children().first();
            if ($lastChild.length === 0) { break; }
            lastNewHeight = $lastChild[0]._primVirtualScroll.height;
            this._$virtualBottom.height(this._$virtualBottom.height() - lastNewHeight);
            $lastChild.insertBefore(this._$virtualBottom);
            $lastChild = maybeInitialize.call(this, $lastChild);
        }
    };

    var handleScroll = function (ev) {
        var $prevFocus = $(document.activeElement),
            elHeight = this._$el.height(),
            newScroll = this._$el.scrollTop(),
            scrollAmount = newScroll - this.scrollPosition;

        this._$el.trigger('scroller-before-scroll');

        if (Math.abs(scrollAmount) > elHeight) {
            redraw.call(this, newScroll, scrollAmount, elHeight);
        } else {
            refill.call(this, elHeight);
        }
        this.scrollPosition = newScroll;

        if (this._setRestoreFocusOnScroll && $prevFocus.closest(this._$el).length) {
            $prevFocus.focus();
        }

        this._$el.trigger('scroller-after-scroll');
    };

    var getRelativeOffset = function ($item) {
        return $item[0].getBoundingClientRect().top - this._$el[0].getBoundingClientRect().top;
        //return $item.offset().top - this._$el.offset().top;
    };

    function getOuterHeight(item) {
        var computedStyle = window.getComputedStyle(item, '');
        return item.offsetHeight +
                parseFloat(computedStyle.marginTop) + parseFloat(computedStyle.marginBottom) +
                parseFloat(computedStyle.borderTopWidth) + parseFloat(computedStyle.borderBottomWidth) +
                parseFloat(computedStyle.paddingTop) + parseFloat(computedStyle.paddingBottom);
    }

    var initScrollData = function (item, scrollId, guessHeight) {
        return (item._primVirtualScroll = {
            height : guessHeight != null ? guessHeight : getOuterHeight(item),
            scrollId : scrollId
        });
    };

    function VirtualScroller($el, options) {
        var that = this;
        logger.assert($el.jquery, 'VirtualScroller | container must be jQuery element');

        this._options = _.defaults({}, options, DEFAULT_OPTIONS);
        this._$el = $el;
        this._$topStorage = $('<div />');
        this._$bottomStorage = $('<div />');
        this.scrollPosition = 0;
        this._setRestoreFocusOnScroll = false;

        var $children = $el.children().remove();

        this._$el.on('scroll.virtual-scroller', _.debounce(handleScroll.bind(this), 25, { leading : true, trailing : true }));
        $(window).on('resize.virtual-scroller', _.debounce(this.resize.bind(this), 25, { leading : true, trailing : true }));

        this._$virtualTop = $('<div class="virtual-scroller-buffer virtual-scroller-buffer-top" style="height: 0;"></div>').appendTo(this._$el);
        this._$virtualBottom = $('<div class="virtual-scroller-buffer virtual-scroller-buffer-bottom" style="height: 0;"></div>').appendTo(this._$el);

        $children.each(function () {
            that.append($(this));
        });

        window.scroller = this;
    }

    VirtualScroller.prototype = {
        constructor : VirtualScroller,

        append : function (items, options) {
            var that = this,
                hiddenHeight = 0,
                hasHidden = false,
                elHeight = this._$el.height(),
                itemsIter = _.toArray(items),
                item;

            /* eslint-disable no-cond-assign */
            while (!hasHidden && (item = itemsIter.shift()) != null) {
                var $item = $(item);
                initScrollData.call(that, item, ++_virtualScrollUUID, options && options.guessHeight);
                $item.insertBefore(this._$virtualBottom);
                if (getRelativeOffset.call(that, $item) > elHeight) {
                    $item.detach();
                    hasHidden = true;
                    itemsIter.unshift(item);
                } else {
                    maybeInitialize.call(that, $item);
                    item._primVirtualScroll.height = getOuterHeight(item);
                }
            }
            /* eslint-enable no-cond-assign */

            _.forEach(itemsIter, function (item) {
                var scrollData = initScrollData.call(that, item, ++_virtualScrollUUID, options && options.guessHeight);
                hiddenHeight += scrollData.height;
            });
            $(itemsIter).appendTo(this._$bottomStorage);
            this._$virtualBottom.height(this._$virtualBottom.height() + hiddenHeight);

            // itemsItem.length > 0 iff there were scrolled items
            return itemsIter.length === 0;
        },

        remove : function ($item) {
            var virtualScrollData = $item[0]._primVirtualScroll;
            $item[0]._primVirtualScroll = null;
            if ($item.parent().is(this._$topStorage)) {
                this._$virtualTop.height(this._$virtualTop.height() - virtualScrollData.height);
            } else if ($item.parent().is(this._$bottomStorage)) {
                this._$virtualBottom.height(this._$virtualBottom.height() - virtualScrollData.height);
            }
            $item.remove();
        },

        replace : function ($item, $newItem, isResize) {
            $newItem.insertBefore(this._$virtualBottom);
            var oldVirtualScrollData = $item[0]._primVirtualScroll,
                virtualScrollData = initScrollData.call(this, $newItem[0], oldVirtualScrollData.scrollId);
            $item.replaceWith($newItem);
            if ($item.parent().is(this._$topStorage)) {
                this._$virtualTop.height(this._$virtualTop.height() - oldVirtualScrollData.height + virtualScrollData.height);
            } else if ($item.parent().is(this._$bottomStorage)) {
                this._$virtualBottom.height(this._$virtualBottom.height() - oldVirtualScrollData.height + virtualScrollData.height);
            } else if (!isResize) {
                maybeInitialize.call(this, $newItem);
            }
        },

        resize : function () {
            if (this._$el.is(':visible')) {
                handleScroll.call(this);
            }
        },

        resizeElement : function ($item) {
            var scrollData = $item[0]._primVirtualScroll,
                $temp = $('<div style="height: ' + scrollData.height + 'px;"></div>'),
                isVirtual = $item.parent().is(this._$topStorage) || $item.parent().is($item._$bottomStorage);
            if (isVirtual) {
                this.replace($item, $temp);
                $item.insertBefore(this._$virtualBottom);
            }
            var newScrollData = initScrollData($item[0], scrollData.scrollId);
            newScrollData.initialized = scrollData.initialized;

            if (isVirtual) {
                this.replace($temp, $item, true);
            }
            refill.call(this, this._$el.height());
        },

        find : function (query) {
            var top = _.toArray(this._$topStorage.find(query)),
                el = _.toArray(this._$el.find(query)),
                bottom = _.toArray(this._$bottomStorage.find(query));
            return $(top.concat(el, bottom));
        },

        _getAllItems : function () {
            return _.toArray(this._$topStorage.children())
                .concat(_.toArray(this._$el.children().not(this._$virtualTop).not(this._$virtualBottom)))
                .concat(_.toArray(this._$bottomStorage.children()));
        },

        scrollTo : function ($item) {
            var offset = getRelativeOffset.call(this, $item),
                isVisible = offset >= 0 && (offset + $item[0]._primVirtualScroll.height) <= this._$el.height();
            if (!isVisible) {
                var allItems = this._getAllItems(),
                    indexOf = _.indexOf(allItems, $item[0]),
                    scrollTop = _.sum(_.take(allItems, indexOf), function (item) {
                        return item._primVirtualScroll.height;
                    });
                this._$el.scrollTop(scrollTop);
            }
        },

        isFirstChild : function ($item) {
            return this._$topStorage.children().first()[0] === $item[0] || (this._$topStorage.length === 0 && $item.prev()[0] === this._$virtualTop[0]);
        },

        isLastChild : function ($item) {
            return this._$bottomStorage.children().last()[0] === $item[0] || (this._$bottomStorage.length === 0 && $item.next()[0] === this._$virtualBottom[0]);
        },

        getFirstChild : function () {
            return this._$topStorage.length > 0 ? this._$topStorage.children().first() : this.getFirstVisibleChild();
        },

        getFirstVisibleChild : function () {
            return this._$virtualTop.next().not(this._$virtualBottom);
        },

        getLastChild : function () {
            return this._$bottomStorage.length > 0 ? this._$bottomStorage.children().last() : this.getLastVisibleChild();
        },

        getLastVisibleChild : function () {
            return this._$virtualBottom.prev().not(this._$virtualTop);
        },

        getNextItem : function ($item) {
            return $item.is(this.getLastVisibleChild()) ? this._$bottomStorage.children().first() : $item.next();
        },

        getPreviousItem : function ($item) {
            return $item.is(this.getFirstVisibleChild()) ? this._$topStorage.children().last() : $item.prev();
        },

        setRestoreFocusOnScroll : function (toggle) {
            this._setRestoreFocusOnScroll = toggle;
        }
    };

    return VirtualScroller;
});

define('primutils/utils/fullscreen',['jquery', 'logger'], function ($, logger) {
    'use strict';

    var elementProto = window.Element.prototype;
    elementProto.requestFullscreen = elementProto.requestFullscreen || elementProto.mozRequestFullScreen || elementProto.webkitRequestFullscreen || elementProto.msRequestFullscreen;

    var fullscreenExit = (document.exitFullscreen || document.mozCancelFullScreen || document.webkitExitFullscreen || document.msExitFullscreen || $.noop).bind(document);
    var fullscreenElement = function () { return document.fullscreenElement || document.mozFullScreenElement || document.msFullscreenElement || document.webkitFullscreenElement; };

    var api = {
        exitFullscreen : function () {
            if (this.isFullscreenEnabled()) {
                fullscreenExit();
            }
        },

        enterFullscreen : function (el) {
            if (el != null) {
                if (el instanceof $) {
                    if (el.length !== 1) {
                        logger.warn('fullscreen.js | Fullscreen only works with one element');
                    }
                    el = el.get(0);
                }
                el.requestFullscreen();
            }
        },

        toggleFullscreen : function (el) {
            if (this.isFullscreenEnabled()) {
                this.exitFullscreen();
            } else {
                this.enterFullscreen(el);
            }
        },

        isFullscreenEnabled : function () {
            return this.getFullscreenElement() != null;
        },

        getFullscreenElement : function () {
            return fullscreenElement();
        },

        get$FullscreenElement : function () {
            return $(this.getFullscreenElement());
        }
    };

    $(document).on('webkitfullscreenchange mozfullscreenchange fullscreenchange MSFullscreenChange', function () {
        $('.fullscreen-element').removeClass('fullscreen-element');
        api.get$FullscreenElement().addClass('fullscreen-element');
    });

    return api;
});

/*jshint -W084,-W030*/
define('primutils/utils/searchTermRegExpBuilder',[
    'lodash',
    'logger',
    'primutils/utils/regex',
    'primutils/utils/strings'
], function (_, logger, regexes, strings) {
    'use strict';

    var HTML_TAG_CONTENTS = '(<[^>]*>)';

    /**
     * `Matcher`s are objects capable of testing and manipulating strings, with an interface similar to JavaScript's native `RegExp` object, namely [`exec`](#method_exec) and [`test`](#method_test), along with extensions useful for some cases like [`surroundMatches`](#method_surroundMatches).
     *
     * `Matcher`s are built through the [`utils.searchTermRegExpBuilder`](utils.searchTermRegExpBuilder.html) module.
     *
     * `Matcher`s support UTF-8 text, not terms (user search terms preceded with a `-` character), and in general anything that the [test suite](https://alm.oraclecorp.com/pgbu/#projects/pgbu_core-ui/scm/core-utils.git/blob/assets/test/utils/search.js?revision=develop) covers.
     *
     * @module utils
     * @class utils.Matcher
     */
    function Matcher(components) {
        var that = this;
        function wrapTester(tester) {
            if (!tester) { return null; }

            var term = components.term;
            return function wrappingTester(val) {
                if (_.isFunction(val)) {
                    return val(term, that);
                }
                return tester.test(val);
            };
        }

        var priv = {};
        Object.defineProperty(this, '_priv', {
            enumerable : false,
            value : priv
        });
        priv.matchOnAny = components.matchOnAny;
        priv.positive = wrapTester(components.positive);
        priv.negative = wrapTester(components.negative);
        priv.positiveRaw = components.positive;
        priv.negativeRaw = components.negative;
        priv.passesPos = function (str) {
            return passesPosTerms(priv, str);
        };
        priv.passesNeg = function (str) {
            return passesNotTerms(priv, str);
        };

        if (!components.matchOnAny) {
            priv.all = _.map(components.positiveTerms, function (term) {
                var re = new RegExp(term, 'i');
                var tester = {
                    test : function (string) {
                        return string != null && re.test(string);
                    }
                };
                return wrapTester(tester);
            });
        }
    }

    var testString = function (priv, str) {
        if (!passesNotTerms(priv, str)) {
            return false;
        }
        return passesPosTerms(priv, str);
    };
    var passesPosTerms = function (priv, str) {
        if (priv.matchOnAny) {
            return priv.positive(str);
        } else {
            return _.every(priv.all, function (matcher) {
                return matcher(str);
            });
        }
    };
    var passesNotTerms = function (priv, str) {
        var neg = priv.negative;
        if (neg != null && neg(str)) {
            return false;
        }
        return true;
    };
    var lazyAny = function (chain, fn) {
        var res = chain.filter(fn).first();
        if (res instanceof _) { res = res.value(); }
        return res != null;
    };
    var lazyAll = function (chain, fn) {
        var res = chain.reject(fn).first();
        if (res instanceof _) { res = res.value(); }
        return res == null;
    };

    Matcher.prototype = {
        constructor : Matcher,

        /**
         * Executes a search against the given string, returning the matches - similar to the native [`RegExp#exec`](http://devdocs.io/javascript/global_objects/regexp/exec)
         *
         * @method exec
         * @param str {String} the string to match against
         * @returns matches {Array} the matched text, or `undefined` if no matches were identified
         */
        exec : function (str) {
                        logger.assert(strings.isString(str) || (_.isNumber(str) && !_.isNaN(str)), 'exec must be called with strings or non-NaN numbers');
                        str = String(str);
            if (!str || !testString(this._priv, str)) {
                return null;
            }
            return _.map(this._priv.positiveRaw.exec(str), _.trim);
        },

        /**
         * Tests whether or not the given arg is matched - similar to the native [`RegExp#test`](http://devdocs.io/javascript/global_objects/regexp/test) for Strings.
         *
         * This method, however, also takes either of two alternate types.
         *
         * Given an `Array` of strings, this method treats the array as one record.
         * Such that, if doing an ALL-style search, the terms the user types in could be from different positions in the array, but as long as all are matched, `test` will return true.
         *
         * Given a LoDash chain, the chain will be asked to produce Strings.
         * These strings are treated as if they were given from an `Array`, so the same rules as `Array` matching apply.
         *
         * @method test
         * @param str {String} the string/array/lodash chain to test
         * @returns matches {Boolean} whether or not the string had matches
         */
        test : function (str) {
            if (!str) {
                return false;
            }

            var priv = this._priv;
            if (strings.isString(str) || _.isNumber(str)) {
                return testString(priv, String(str));
            } else if (_.isArray(str)) {
                if (priv.negative && !_.every(str, priv.passesNeg)) {
                    return false;
                }

                if (priv.matchOnAny) {
                    return _.some(str, priv.passesPos);
                } else {
                    return _.every(priv.all, function (tester) {
                        return _.some(str, tester);
                    });
                }
            } else if (str instanceof _) {
                if (priv.negative && !lazyAll(str, priv.passesNeg)) {
                    return false;
                }

                if (priv.matchOnAny) {
                    return lazyAny(str, priv.passesPos);
                } else {
                    var tests = priv.all;
                    return lazyAny(str, function (recordVal) {
                        return _.every(tests, function (tester) {
                            if (tester(recordVal)) {
                                tests = _.without(tests, tester);
                                return true;
                            }
                        });
                    });
                }
            }
                        logger.warn('Unexpected type of input given as test: ' + typeof str);
                        return false;
        },

        /**
         * Surrounds the matches found in the given string with the given `opts.prefix` and `opts.postfix`.
         *
         * @example
         *
         *     var matcher = searchBuilder('text');
         *     matcher.surroundMatches('the text is great', {
         *         prefix: 'yummy ',
         *         postfix: ' really',
         *     });
         *     // would return "the yummy text really is great
         *
         * @method surroundMatches
         * @param str {String} the string whose matches should be surrounded
         * @param opts {Object} an object with `prefix` and `postfix` options used as the surroundings
         * @return newStr {String} a new string with all matches surrounded with the given `prefix`/`postfix`
         */
        surroundMatches : function (str, opts) {
                        logger.assert(strings.isString(str), 'surroundMatches must be called with strings');
                        if (!testString(this._priv, str)) {
                return str;
            }

            var prefix = opts.prefix;
            var postfix = opts.postfix;
            var globalMatcher = new RegExp(HTML_TAG_CONTENTS + '|(' + this._priv.positiveRaw.source + ')', 'gi');
            return str.replace(globalMatcher, function (match) {
                if (!match) { return ''; }
                if (match[0] === '<' && match[match.length - 1] === '>') {
                    return match;
                }

                var parts = /^(\s*)([^]*?)(\s*)$/.exec(match);
                return parts[1] + prefix + parts[2] + postfix + parts[3];
            });
        }
    };

    // \b does NOT recognize boundaries on unicode characters; sub this in instead
    //  not perfect, since the prefix will include the space in the match.  not ideal, but...maybe acceptable?  better than not having exact matches on unicode at all
    var wordBoundaryPrefix = '(?:^|\\s)';
    var wordBoundaryPostfix = '(?=\\s|$)';
    var wordBoundarySearch = /\\w/;
    function markTermBoundary(word) {
        if (!word || !word.trim()) { return; }

        var firstChar = word[0];
        var startsWhitespace = wordBoundarySearch.test(firstChar);
        var endsWhitespace = wordBoundarySearch.test(word[word.length - 1]);
        var prefix = startsWhitespace ? '' : wordBoundaryPrefix;
        var postfix = endsWhitespace ? '' : wordBoundaryPostfix;

        return prefix + word + postfix;
    }

    function swapNumericStartsWith(term) {
        if (/^\d+$/.test(term)) {
            return wordBoundaryPrefix + '-?' + term;
        }
        return term;
    }

    var DEFAULT_OPTIONS = {
        matchOnAny : false,
        numbersUseStartsWith : true
    };

    /**
     * Constructs a new [`Matcher` - link](utils.Matcher.html) instance based on a end-user-entered term and args optionally provided by a developer.
     *
     * The second `args` option is an `Object`, respecting the following keys:
     * * `matchOnAny`: if true, and the user enters multiple terms for their search (e.g. `"foo bar"`), only one of the terms will have to match on a given string.
     *     The default is false, meaning that all user-entered terms have to match for there to be a true match.
     * * `numbersUseStartsWith`: if true - the default - a user typing "50" would match "500" but not "10500".
     *
     * Example:
     *
     *     var searchText = $search.val(); // assume the user entered 'foo bar'
     *     var matcher = searchBuilder(searchText, {matchOnAny: true});
     *     matcher.test('billy bob'); // false
     *     matcher.test('foo is cool'); // true
     *     matcher.test('bar is also cool'); // true
     *
     * For more example usages, refer to the [unit tests](https://alm.oraclecorp.com/pgbu/#projects/pgbu_core-ui/scm/core-utils.git/blob/assets/test//utils/search.js?revision=develop).
     *
     * @module utils
     * @class utils.searchTermRegExpBuilder
     */
    function matcherFactory(term, args) {
        args = _.defaults({}, args, DEFAULT_OPTIONS);
        if (!strings.isString(term) || !term.trim().length) {
            return new Matcher({
                positiveTerms : [],
                positive : /^$/,
                negativeTerms : [],
                negative : false
            });
        }

        term = term.trim();
        term = regexes.escapeLiterals(term);
        term = term.replace(/&quot;/g, '"');

        var notTerms = [];
        var not = /-(?!\d)(?:"(.*?)"|(\S*))/g;
        var notMatch;
        /* eslint-disable no-cond-assign */
        while (notMatch = not.exec(term)) {
            // guard against terms like "foo-thing", which should be searched as a whole word, not considered a negative term
            var notPart = notMatch[0];
            var notInd = term.indexOf(notPart);
            if (notInd > 0 && /\S/.test(term[notInd - 1])) {
                continue;
            }

            term = term.replace(notMatch[0], '');
            not.lastIndex = 0;
                        logger.assert((notMatch[2] == null) !== (notMatch[1] == null), 'matched as both quoted and unquoted...how?');
            
            var termPart = notMatch[2] || markTermBoundary(notMatch[1]);
            if (termPart) {
                notTerms.push(termPart);
            }
        }
        /* eslint-enable no-cond-assign */

        var searchTerms = [];
        var terms = _.groupBy(term.split('"'), function (value, i) {
            return i % 2 === 0 ? 'unquoted' : 'quoted';
        });
        terms.unquoted = _.flatten(_.map(terms.unquoted, function (term) {
            return term.split(' ');
        }));

        _.forEach(terms.quoted, function (word) {
            if (word.length) {
                searchTerms.push(markTermBoundary(word));
            }
        });

        _.forEach(terms.unquoted, function (word) {
            word = word.trim();
            if (word.length) {
                searchTerms.push(word);
            }
        });

        if (args.numbersUseStartsWith) {
            searchTerms = _.map(searchTerms, swapNumericStartsWith);
            notTerms = _.map(notTerms, swapNumericStartsWith);
        }

        var positiveMatcher = new RegExp(searchTerms.join('|'), 'i');
        var negativeMatcher = notTerms.length && (new RegExp(notTerms.join('|'), 'i'));
        return new Matcher({
            term : term,
            matchOnAny : args.matchOnAny,
            numbersUseStartsWith : args.numbersUseStartsWith,
            positiveTerms : searchTerms,
            positive : positiveMatcher,
            negativeTerms : notTerms,
            negative : negativeMatcher
        });
    }
    matcherFactory.Matcher = Matcher;

    return matcherFactory;
});

/* eslint-disable no-nested-ternary */
define('primutils/utils/sorters',[
    'lodash',
    'logger'
], function (_, logger) {
    'use strict';

    // stolen from grid
    // should be in primutils though it is nice to be able to quickly add custom comparators
    function standardCompare(value1, value2, asc) {
        var a, b;
        if (asc) {
            a = value1;
            b = value2;
        } else {
            a = value2;
            b = value1;
        }
        if (a == null) { return 1; }
        if (b == null) { return -1; }
        return a < b ? -1 : a > b ? 1 : 0;
    }

    return {
        standardCompare : standardCompare,

        boolean : function (valueA, valueB, asc) {
            var a = valueA ? 1 : 0;
            var b = valueB ? 1 : 0;
            return standardCompare(a, b, asc);
        },
        // Checkboxes (probably better to use boolean)
        check : function (valueA, valueB, asc) {
            var a = valueA === 'true',
                b = valueB === 'true';
            return standardCompare(a, b, asc);
        },
        float : function (valueA, valueB, asc) {
            var a = parseFloat(valueA),
                b = parseFloat(valueB);
            if (_.isNaN(a)) { a = null; }
            if (_.isNaN(b)) { b = null; }
            return standardCompare(a, b, asc);
        },
        integer : function (valueA, valueB, asc) {
            var a = parseInt(valueA, 10),
                b = parseInt(valueB, 10);
            if (_.isNaN(a)) { a = null; }
            if (_.isNaN(b)) { b = null; }
            return standardCompare(a, b, asc);
        },
        string : function (valueA, valueB, asc) {
                        if ((valueA != null && !_.isString(valueA)) ||
                (valueB != null && !_.isString(valueB))) {
                logger.error('sorters: An attempt was made to sort with the "string" sorter, but one or more values that were not strings.');
            }
            
            var a = valueA == null ? '' : _.isString(valueA) ? valueA.toLowerCase() : valueA.toString().toLowerCase(),
                b = valueB == null ? '' : _.isString(valueB) ? valueB.toLowerCase() : valueB.toString().toLowerCase(),
                match = 0;

            if (asc) {
                match = a.localeCompare(b);
            } else {
                match = b.localeCompare(a);
            }
            if (match < 0) {
                match = -1;
            }

            return match;
        },
        date : function (valueA, valueB, asc) {
            return standardCompare(valueA, valueB, asc);
        },
        datetime : function (valueA, valueB, asc) {
            return standardCompare(valueA, valueB, asc);
        },
        // This is naive and doesn't sort by either sequence number or display value, so implementers should override this
        enum : function (valueA, valueB, asc) {
            return standardCompare(valueA, valueB, asc);
        },
        // TODO sort by item values?
        picker : function (valueA, valueB, asc) {
            var a = valueA != null ? valueA.length : null,
                b = valueB != null ? valueB.length : null;
            return standardCompare(a, b, asc);
        }
    };
});

/**
 * A collection of utilities for dealing with URLs.
 *
 * @class utils.urls
 * @module utils
 */
define('primutils/utils/urls',[
    'lodash',
    'logger'
], function (_, logger) {
    'use strict';

    var ABSOLUTE_URL = /^https?:|^wss?:/i;
    var that = {
        /**
         * Gives the current `window.location.origin`, even in IE10 where this has to be manually determined.
         *
         * @method origin
         * @static
         */
        origin : function () {
            var origin = window.location.origin;
            //Must be IE10
            //http://tosbourn.com/a-fix-for-window-location-origin-in-internet-explorer/
            if (!origin) {
                origin = window.location.protocol + '//' + window.location.hostname + (window.location.port ? ':' + window.location.port : '');
            }
            return origin;
        },

        /**
         * Gets a param from the current `window.location`.
         * Does not handle arrays of data in the URL such as `aParam[]=val1&aParam[]=val2`.
         *
         * For example, if the URL is:
         *
         *     http://HOST:PORT/primavera/p/error?tab=5&loggerLevel=DEBUG&encoded=my%20user#other
         *
         * Then:
         *
         *     getSingleParam('tab', 'default');    // '5'
         *     getSingleParam('nonParam', '567');   // '567'
         *     getSingleParam('loggerLevel', null); // 'DEBUG'
         *     getSingleParam('encoded', null);     // 'my user' (as of 2.7.7, previous versions did not decode parts)
         *
         * @method getSingleParam
         * @static
         * @param paramName {String} the name of the parameter to get
         * @param defaultValue the default return value if the param is not present in the URL - this will not get decoded
         * @returns the parameter value in the URL if found, defaultValue otherwise
         */
        getSingleParam : function (paramName, defaultValue) {
            return that.getSingleParamFromURL(window.location.search, paramName, defaultValue);
        },

        getSingleParamFromURL : function (url, paramName, defaultValue) {
            var urlParams = url.replace(/^.*\?/, '').split('&');

            var result = defaultValue;
            for (var i = 0, len = urlParams.length; i < len; i++) {
                var param = urlParams[i];
                var pair = param.split('=');
                var key = pair[0] || '';
                var value = pair[1] || '';

                if (key === paramName) {
                    result = window.decodeURIComponent(value);
                }
            }

            return result;
        },

        // Note: like getSingleParam, does not handle array data
        parseURL : function (url) {
            var search = _.includes(url, '?') ? url.split('?')[1] : '';
            var searchTokens = _.reduce(search.split('&'), function (memo, token) {
                var pair = token.split('=');
                var key = pair[0];
                var value = pair[1];
                memo[key] = value;
                return memo;
            }, {});

            return {
                href : url,
                search : search,
                searchTokens : searchTokens
            };
        },

        /**
         * Takes the given url and appends a query to it.
         *
         * @example
         *
         *     urls.injectQuery('../rest/foo', ['param', 'value'])
         *         // returns '../rest/foo?param=value'
         *
         * @method injectQuery
         * @static
         * @param url {String}
         * @param query {Array<String>}
         */
        injectQuery : function (url, query) {
                        logger.assert(url && query && _.isArray(query), 'bad params given to injectQuery method - expected something like injectQuery(url, ["paramName", "paramValue"])');
            
            var divider = _.includes(url, '?') ? '&' : '?';
            var paramName = query[0];
            var paramValue = query[1];
            return url + divider + paramName + '=' + paramValue;
        },

        /**
         * Simple templating function, similar to [`strings.template`](utils.strings.html#method_template) but intended for URL-based strings.
         *  As [`strings.template`](strings.html#method_template), but any `'<%=VAR_NAME%>'`-style strings will [encoded](http://devdocs.io/javascript/global_objects/encodeuricomponent) via [encodeURIObject](method_encodeURIObject).
         *  `'<%!VAR_NAME%>'`-style strings, on the other hand, will not be encoded.
         *
         * Examples:
         *
         *     urls.template('rest/<%=type%>/<%=id%>', {type: 'a', id: 'b'});
         *         // would return 'rest/a/b'
         *     urls.template('rest/<%=type%>/<%=id%>', {type: 'document', id: 'uuid://&amp;5'});
         *         // would return 'rest/document/uuid%3A%2F%2F%265'
         *     urls.template('rest/<%!ident%>/<%=type%>/<%=id%>',
         *             {type: 'document', id: 'uuid://&amp;5', ident: 'project/5'});
         *         // would return 'rest/project/5/document/uuid%3A%2F%2F%265'
         *
         * @method template
         * @static
         * @param {String} text the url template, similar to first option of _.template
         * @param {Object} [data] if specified, template will invoke the template
         *   and do replacement based on the arguments provided; if not, template will
         *   return a function that can be called with an Object to do the same thing.
         */
        template : function template(text, origData) {
            function handleReplace(data) {
                var encoded = that.encodeURIObject(data);
                var response = text;
                _.forEach(encoded, function (replaceVal, replaceName) {
                    var reReplace = new RegExp('<%=' + replaceName + '%>', 'g');
                    response = response.replace(reReplace, function () {
                        return replaceVal;
                    });
                    var reReplaceUnencoded = new RegExp('<%!' + replaceName + '%>', 'g');
                    response = response.replace(reReplaceUnencoded, function () {
                        return data[replaceName];
                    });
                });
                return response;
            }

            if (origData) {
                return handleReplace(origData);
            }

            return handleReplace;
        },

        /**
         * Resolves relative URLs given the relative path (e.g. '../rest') and base (e.g. `window.location.href`).
         *
         * @method resolveRelativePath
         * @static
         * @param path
         * @param base
         */
        resolveRelativePath : function resolveRelative(path, base) {
            /* eslint-disable no-useless-escape */
            if (ABSOLUTE_URL.test(path)) {
                return path;
            }

            base = base.split('?')[0];

            // Upper directory
            if (path.indexOf('../') === 0) {
                return resolveRelative(path.slice(3), base.replace(/\/[^\/]*$/, ''));
            }
            // Same directory
            if (path.indexOf('./') === 0) {
                return resolveRelative(path.slice(2), base);
            }
            // Relative to the root
            if (path.indexOf('/') === 0) {
                var match = base.match(/(\w*:\/\/)?[^\/]*\//) || [base];
                return match[0] + path.slice(1);
            }
            //relative to the current directory
            return base.replace(/\/[^\/]*$/, '') + '/' + path;
            /* eslint-enable no-useless-escape */
        },

        /**
         * Simple templating function which converts relative URLs into fully-qualified WebSocket URLs.
         *
         * This method behaves in the same manner as [`template`](#method_template).
         *
         * The method assumes that the server's WebSocket services are hosted on the same port as its HTTP services.
         * It also assumes that the protocol prefix is `ws://` for regular WebSockets and `wss://` for secure WebSockets, by virtue of using [`resolveRelativePath`](#method_resolveRelativePath) to resolve the relative WS path to the current domain.
         *
         * @method templateWebSocket
         * @param {String} text the url template, similar to first option of _.template
         * @param {Object} [data] if specified, template will invoke the template
         *   and do replacement based on the arguments provided; if not, template will
         *   return a function that can be called with an Object to do the same thing.
         * @static
         */
        templateWebSocket : function wsTemplate(text, data) {
            function postProcessUrl(result) {
                return that.resolveRelativePath(result, window.location.href)
                    .replace(/^http/i, 'ws');
            }
            var baseTemplate = that.template(text, data);
            if (_.isFunction(baseTemplate)) {
                return _.flowRight(postProcessUrl, baseTemplate);
            } else {
                return postProcessUrl(baseTemplate);
            }
        },

        /**
         * Given an Object whose values are all strings, this function will return a map with the same keys, where each value is encoded via `encodeURIComponent`.
         *
         * @method encodeURIObject
         * @param {Object} data the Object whose values should be encoded
         * @static
         */
        encodeURIObject : function (data) {
            return _.reduce(data, function (result, value, key) {
                result[key] = _.isString(value) ? encodeURIComponent(value) : value;
                return result;
            }, {});
        }
    };

    return that;
});


define('primutils/utils/History',['jquery', 'lodash', 'logger', 'primutils/utils/urls', 'primutils/utils/Subscribable'], function ($, _, logger, urls, Subscribable) {
    'use strict';

    function generateNewUrl(paramName, paramValue, paramMatcher) {
        var loc = window.location;
        var newLocation;
        if (loc.search.indexOf(paramName) !== -1) {
            newLocation = loc.href.replace(paramMatcher, paramName + '=' + window.encodeURIComponent(paramValue));
        } else {
            var origin = loc.origin || (window.location.protocol + '//' + window.location.hostname + (window.location.port ? ':' + window.location.port : ''));
            var baseLocation = origin + loc.pathname + loc.search;
            var newParamPrefix = loc.search ? '&' : '?';
            newLocation = baseLocation + newParamPrefix + paramName + '=' + window.encodeURIComponent(paramValue) + loc.hash;
        }
        return newLocation;
    }

    /**
     * History is a thin wrapper around the HTML5 history API.
     *
     * It's purpose is to track one param in the URL (as configured by the `param` option).
     *
     * It provides a helper method for updating this param in the URL (`pushParam`).
     *
     * It provides an event (`param-updated`) anytime this param is updated through any method other than the `pushParam` method - for example, the user using their browser's back/forward buttons.
     *
     * ----------
     *
     * Flow example:
     *
     * * User starts on the page (url is now `'/p/scope_items'`)
     *   - Here, your page would have some criteria for identifying a `defaultValue` - perhaps the id of the first project in a *sorted* list of projects
     *   - Alternatively (although using `defaultValue` is preferred!), if your page may not have a stable `defaultValue` on returning to the page when the user re-navigates to it, the `defaultParam` method can be used to replace the param on first load.
     *     - An example where your page may not have a stable `defaultValue` param on subsequent loads is if you use the first project's id from an *unsorted* list of projects.
     * * User makes an interaction that calls `pushParam` (url is now `'/p/scope_items?projectId=5'`)
     * * User clicks the back button (going back to url `'/p/scope_items'`)
     *   - Here, the `param-updated` event is triggered with the value `defaultValue` provided on page load
     *
     * One of the important points of the `defaultValue` is that it works just as well even if the user navigates to a new page entirely - say, `'/p/work_packages'` - and then starts using the back button to re-enter the page and navigate history between the states of the page.
     * In Prime, the social user profile is a good working example of this flow in action (if you can get other users in a user's feed).
     *
     * @class utils.History
     * @module utils
     */

    function History(options) {
        logger.assert(this && this instanceof History, 'History must be created with \'new\'');

        /**
         * The param to be tracked.
         * For example, if the param given is `user`, the URLs `...&user=5` `...&user=6` represent two possible states for the `param-updated` event to switch between.
         *
         * @property param
         * @type {String}
         * @required
         */
        logger.assert(options && _.isString(options.param), 'History must be created with a param to track as part of provided options');
        this._options = _.defaults({}, options, {
            eventNamespace : _.uniqueId('history_')
        }, this._options);
        this._currentValue = this._options.defaultValue;

        /**
         * Triggered every time the param is updated, except when `pushParam` is called.
         *
         * @event param-updated
         * @param param {String} the param that changed (same as options.param)
         * @param newValue {String} the new value of the tracked param
         */
        Subscribable.call(this, {
            allowedEvents : ['param-updated']
        });
        this._init();
    }
    History.prototype = Object.create(Subscribable.prototype);

    var hp = History.prototype;
    hp.constructor = hp;

    hp._options = {};

    hp._init = function () {
        this._supportsHistory = window.history && window.history.pushState;
        this._paramMatcher = new RegExp(this._options.param + '=[^&#]*'); // Use * because it could be followed by null, RegExp is greedy and will always match as much as it can
        this._attach();
    };

    hp._attach = function () {
        var that = this;

        if (!this._supportsHistory) {
            return;
        }

        var ns = this._options.eventNamespace;
        $(window).on('popstate.' + ns, function () {
            var paramValue = urls.getSingleParam(that._options.param, that._options.defaultValue);
            if (paramValue !== that._currentValue) {
                that._currentValue = paramValue;
                /**
                 * A default value provided to the `param-updated` event in the case that the tracked URL param is no longer present.
                 *
                 * For example, on the social user profile page, `options.defaultValue` is the user's ID, meaning that the user's own profile is loaded if the URL no longer contains some user identifier.
                 *
                 * @property defaultValue
                 * @type {String}
                 * @optional
                 */
                that.trigger('param-updated', that._options.param, paramValue);
            }
        });

        $('body').on('click.' + ns, this._linkFilter(), function (ev) {
            var $node = $(this);

            // let users modify keypress if they'd like
            if (ev.ctrlKey || ev.shiftKey) {
                return;
            }

            ev.preventDefault();

            var newParam = $node.attr('data-historyreplace-value');

            if (newParam && !that._skipValue(newParam)) {
                that.trigger('param-updated', that._options.param, newParam, that._getOldValue());
                that.pushParam(newParam);
            }
        });
    };

    hp._getOldValue = function () {
        return urls.getSingleParam(this._options.param);
    };

    hp._skipValue = function (newValue) {
        var oldValue = this._getOldValue();
        return oldValue !== undefined && oldValue === newValue;
    };

    hp._linkFilter = function () {
        return '[data-historyreplace-name="' + this._options.param + '"]';
    };

    /**
     * Cleans up event listeners, etc.
     *
     * @method destroy
     */
    hp.destroy = function () {
        var ns = this._options.eventNamespace;
        $(window).off('popstate.' + ns);
        $('body').off('click.' + ns, this._linkFilter());
    };

    /**
     * Updates this History watcher with a new `defaultValue` configuration setting.
     *
     * This can be useful if, for example, your page loads some metadata that determines the `defaultValue` setting, but you want to create your History instance before that point.
     *
     * @method updateDefaultValue
     * @chainable
     */
    hp.updateDefaultValue = function (newDefault) {
        this._options.defaultValue = newDefault;
        return this;
    };

    /**
     * Pushes a new param state to the URL, if possible (i.e. if the browser supports the HTML5 history API), otherwise does nothing.
     *
     * @method pushParam
     * @param paramValue {String} the new value of the param for the URL
     * @chainable
     */
    hp.pushParam = function (paramValue) {
        if (!this._supportsHistory || this._skipValue(paramValue)) {
            return this;
        }

        this._currentValue = paramValue;
        window.history.pushState(null, null, generateNewUrl(this._options.param, paramValue, this._paramMatcher));

        return this;
    };

    /**
     * Replaces the existing param state in the URL, if possible (i.e. if the browser supports the HTML5 history API), otherwise does nothing.
     *
     * @method replaceParam
     * @param paramValue {String} the new value of the param for the URL
     * @chainable
     */
    hp.replaceParam = function (paramValue) {
        if (!this._supportsHistory || this._skipValue(paramValue)) {
            return this;
        }

        this._currentValue = paramValue;
        window.history.replaceState(null, null, generateNewUrl(this._options.param, paramValue, this._paramMatcher));

        return this;
    };

    /**
     * Replaces the current URL if possible (i.e. the browser supports history API) and if a `param` was not already present in the URL.
     *
     * @method defaultParam
     * @param paramValue {String} the new value of the param for the URL
     * @chainable
     */
    hp.defaultParam = function (paramValue) {
        if (!this._supportsHistory || this._getOldValue() !== undefined) {
            return this;
        }

        this._currentValue = paramValue;
        window.history.replaceState(null, null, generateNewUrl(this._options.param, paramValue, this._paramMatcher));

        return this;
    };

    return History;
});


define('primutils/utils/perfMonitor',['lodash'], function (_) {
    'use strict';

    function PerfMonitor(options) {
        this._nextLandmarkId = 0;
        this._options = _.defaults(options, {
            landmarkStart : _.noop,
            landmarkEnd : _.noop,
            actionTaken : _.noop
        });
        this._openMonitors = {};
    }

    PerfMonitor.prototype = {
        landmarkStart : function (landmarkName) {
                        var id = ++this._nextLandmarkId;
            if (id === Infinity) {
                id = this._nextLandmarkId = 0;
            }
            var info = this._openMonitors[id] = { name : landmarkName };
            this._options.landmarkStart(info);
            info.time = new Date();
            info.start = window.performance.now();
            return id;
                    },
        landmarkEnd : function (landmarkId) {
                        var end = window.performance.now();
            var info = this._openMonitors[landmarkId];
            info.end = end;
            delete this._openMonitors[landmarkId];
            this._options.landmarkEnd(info);
                    },
        actionTaken : function (actionName) {
                        this._options.actionTaken({
                name : actionName,
                msSinceLoad : window.performance.now(),
                time : new Date()
            });
                    }
    };

    return PerfMonitor;
});

/**
 * Utility functions around dealing with lists.
 *
 * @class lists.search
 */
define('primutils/lists/search',[
    'lodash'
], function (_) {
    'use strict';

    /**
     * An implementation of the [binary search algorithm](https://en.wikipedia.org/wiki/Binary_search_algorithm).
     * This method takes a sorted list and a function used to determine whether the target item is before, after, or actually is the target object.
     *
     * @example
     *
     *     search.search([1, 20, 500, 10000], function (val) {
     *         // indicates that the target is in the later part of the list
     *         if (val < 20) { return 1; }
     *         // indicates that the target is in the earlier part of the list
     *         if (val > 20) { return -1; }
     *         // indicates that this is the target
     *         if (val === 20) { return 0; }
     *     }); // returns 20, after making log(N) comparisons, where N is the size of the list
     *
     * *WARNING* this method does not do any verification that the list is sorted prior to running the search.
     * That is your responsibility.
     *
     * *NOTE 1* a `null` or `undefined` result from your callback is used as an indicator that the target item is not in the given list.
     * In this case, `undefined` will be returned.
     *
     * @method search
     * @static
     * @param list {Array} the sorted list of items to search within
     * @param fn {Function} the comparison function used to determine whether the target item is earlier or later in the list than a given item
     * @return the target item if found, `undefined` otherwise
     */
    var search = function (list, fn) {
        var lower = 0, upper = list.length - 1;
        while (upper >= lower) {
            var checkIdx = Math.floor((lower + upper) / 2);
            var val = list[checkIdx];
            var check = fn(val);
            if (check === 0) {
                return val;
            }
            if (check < 0) {
                upper = checkIdx - 1;
            } else if (check > 0) {
                lower = checkIdx + 1;
            } else if (check == null) {
                break;
            }
        }
    };

    return {
        search : search
    };
});

/**
 * All formatters abide by the same core interface and theory; see
 *  [Formatter Theory](formatters.FormatterTheory.html) for more information.
 *
 * The multicurrencyFormatter uses a rather specific set of Objects and metadata to do its work.
 * If you're in Prime, this is all automatic.
 *
 * The options for multicurrencyFormatter are documented in the Properties tab, below.
 * *Note* that most developers in Prime should use other interfaces rather than customizing options on this formatter.
 *
 * @uses formatters.FormatterTheory
 * @class formatters.multicurrencyFormatter
 * @module formatters
 */
define('primutils/formatters/multicurrencyFormatter',['lodash',
    'logger',
    'primutils/formatters/helpers/unformatIgnore',
    'primutils/formatters/helpers/numbers',
    'primutils/formatters/baseFormatter'
], function (_, logger, unformatIgnore, numberUtils, baseFormatter) {
    'use strict';

    var defaults = {
        meta : undefined,
        showDecimals : true,
        hasRecordCurrency : _.constant(false),
        recordCurrencyId : _.constant(null)

        /**
         * A pattern (RegExp or String), or list of patterns, that will be ignored in any string
         *  the user types, for unformat() or validate() calls.
         *
         * @property unformatIgnore
         * @type String|RegExp|Array
         * @optional
         */
        //unformatIgnore = [],

        //serializeWrapper,
        //unserializeWrapper,
        //formatWrapper,
        //unformatWrapper,
        //validateWrapper
    };

    function modOpts(options) {
        if (options.baseCostFormatter) {
            var passthroughOpts = _.omit(options,
                    'meta', 'showDecimals', 'hasRecordCurrency', 'recordCurrencyId',
                    'currencyType', 'currency', 'baseCostFormatter', 'showCurrencySymbol');
            options.baseCostFormatter = options.baseCostFormatter.withOptions(passthroughOpts);
        }
    }

    function getValue(options, opt) {
        if (_.isFunction(options[opt])) { return options[opt](); }
        return options[opt];
    }

    function serialize(options, multicurrencyObject) {
        if (multicurrencyObject == null) {
            return null;
        }

        return multicurrencyObject;
    }

    function unserialize(options, serializedMulticurrencyObject) {
        if (serializedMulticurrencyObject == null || serializedMulticurrencyObject === '') {
            return null;
        }

        return serializedMulticurrencyObject;
    }

    function getCurrencyType(options) {
        var currentCurrencyType = _.result(options, 'currencyType') ||
                _.result(options, 'currency') || // TODO remove case once Doug merges in prime
                options.meta.currencyOfRecord;

                if (!options.meta) {
            var message = 'multicurrencyFormatter cannot be used without applicable metadata.';
            logger.fatal(message);
            throw new Error(message);
        }
        if (options.meta.currencyTypes[currentCurrencyType] == null) {
            logger.warn('options indicated a currency of ' + currentCurrencyType + ', but the only currency types indicated in meta were ' + _.keys(options.meta.currencyTypes));
        }
        
        return currentCurrencyType;
    }

    function getCurrentCurrency(options, multicurrencyObject) {
        var currentCurrencyType = getCurrencyType(options);
        return options.meta.currencies[multicurrencyObject[currentCurrencyType].currency];
    }

    function getUnformatCurrency(options) {
        if (options.hasRecordCurrency()) {
            return options.meta.currencies[options.recordCurrencyId()];
        }
        var currencyType = getCurrencyType(options);
        return options.meta.currencies[options.meta.currencyTypes[currencyType].currencyId];
    }

    // TODO it would be better if this module could hook into withOptions calls;
    //  if it could we could avoid doing baseCostFormatter.withOptions all the time
    function getCostFormatter(options, currentCurrency) {
        var baseCostFormatter = options.baseCostFormatter;
        var baseCostOptions = baseCostFormatter.getOptions();
        var showSymbol = getValue(options, 'showCurrencySymbol');
        return baseCostFormatter.withOptions({
            format : numberUtils.useDecimalsInFormat(baseCostOptions.basePositiveFormat, currentCurrency.decimalsToDisplay, showSymbol, options.showDecimals),
            negativeFormat : numberUtils.useDecimalsInFormat(baseCostOptions.baseNegativeFormat, currentCurrency.decimalsToDisplay, showSymbol, options.showDecimals),
            symbol : currentCurrency.symbol || '',
            printSymbolIfPresent : showSymbol
        });
    }

    function format(options, multicurrencyObject, formatOptions) {
        if (multicurrencyObject == null) {
            return '';
        }

                if (_.isNumber(multicurrencyObject)) {
            var errMsg = 'Given Cost was a Number, but MULTICURRENCY_COSTs must be Objects.';
            logger.fatal(errMsg);
            throw new Error(errMsg);
        }
        
        var currentCurrencyType = getCurrencyType(options);
        if (!multicurrencyObject[currentCurrencyType] || multicurrencyObject[currentCurrencyType].cost == null) {
                        logger.info('No cost information available for ' + currentCurrencyType + ' currency');
                        return '';
        }

        var currentCost = multicurrencyObject[currentCurrencyType].cost;
        var currentCurrency = getCurrentCurrency(options, multicurrencyObject);
        if (currentCurrency == null) {
                        logger.warn('Object has no data for ' + currentCurrencyType + ' cost type; returning empty string (TODO climb up cost type hierarchy until one is found?)');
                        return '';
        }
        var costFormatter = getCostFormatter(options, currentCurrency);
        return costFormatter.format(currentCost, formatOptions);
    }

    function unformat(options, costString, unformatOptions) {
        if (costString == null || costString === '') {
            return null;
        }

        var unformatCurrency = getUnformatCurrency(options);
        var costFormatter = getCostFormatter(options, unformatCurrency);
        var enteredCost = costFormatter.unformat(costString, unformatOptions);
        var currentCurrencyType = getCurrencyType(options);
        var multicurrencyCostData = {
            cost : enteredCost,
            currency : unformatCurrency.id
        };
        var ret = {};
        ret[currentCurrencyType] = multicurrencyCostData;
        return ret;
    }

    function validate(options, costString) {
        if (costString == null || costString === '') {
            return true;
        }

        var currentCurrencyType = getCurrencyType(options);
        if (options.meta.currencyTypes[currentCurrencyType] == null) {
            return false;
        }

        var costFormatter = getCostFormatter(options, getUnformatCurrency(options));
        return costFormatter.validate(costString);
    }

    return baseFormatter(defaults, serialize, unserialize, format, unformat, validate, modOpts);
});


/**
 * All formatters abide by the same core interface and theory; see
 *  [Formatter Theory](formatters.FormatterTheory.html) for more information.
 *
 * The options for this formatter are documented in the Properties tab below.
 *
 * @uses formatters.FormatterTheory
 * @uses formatters.numericOptions
 * @class formatters.ratioFormatter
 * @module formatters
 */
define('primutils/formatters/ratioFormatter',['lodash',
    'logger',
    'primutils/utils/regex',
    'primutils/formatters/durationFormatter',
    'primutils/formatters/helpers/scientific',
    'primutils/formatters/helpers/duration',
    'primutils/formatters/helpers/unformatIgnore',
    'primutils/formatters/baseFormatter'
], function (_, logger, regexUtils, durationFormatter, scientific, durationUtils, unformatIgnore, baseFormatter) {
    'use strict';

    var defaults = {
        // PGBU application settings
        /**
         * Same as the option in [durationFormatter](formatters.durationFormatter.html#property_units).
         *
         * @property units
         * @type Object
         * @optional
         * @default 8 hours/day, 40 hours/week, 172 hours/month, 2000 hours/year
         */
        units : {
            n : { // minutes
                internalType : 'minutes',
                scale : 1 / 60,
                sub : null
            },
            h : { // hours
                internalType : 'hours',
                scale : 1,
                sub : 'n'
            },
            d : { // days
                internalType : 'days',
                scale : 8,
                sub : 'h'
            },
            w : { // weeks
                internalType : 'weeks',
                scale : 40,
                sub : 'd'
            },
            m : { // months
                internalType : 'months',
                scale : 172,
                sub : 'w'
            },
            y : { // years
                internalType : 'years',
                scale : 2000,
                sub : 'm'
            }
        },

        /**
         * Same as the option in [durationFormatter](formatters.durationFormatter.html#property_hoursPer).
         *
         * Aliased as the `scales` option.
         *
         * @property hoursPer
         * @type Object
         * @optional
         * @default null
         */
        hoursPer : null,

        // PGBU user pref

        /**
         * What unit to show as the numerator.
         * As [durationFormatter#showIn](formatters.durationFormatter.html#property_showIn), this option must match one of the keys of this formatter's `unit` config.
         *
         * @property numeratorShowIn
         * @type String
         * @optional
         * @default 'h' (Hours)
         */
        numeratorShowIn : 'h',

        /**
         * What UOM type to use for the numerator.
         * The default is `'units'` - corresponding to the `units` config property.
         * When you have an alternate `numeratorType`, the formatter assumes that you have provided an option with the same name.
         * For example, the following would work together:
         *
         *     baseRatioFormatter.withOptions({
         *         showAsPercentage: false,
         *         numeratorType: 'uom5',
         *         uom5: {
         *             'feet':  { scale: 1 },
         *             'ft':    { scale: 1 },
         *             'yards': { scale: 3 }
         *          }
         *     });
         *
         * *Note*: you are responsible for making sure the type doesn't conflict with any other options in your formatter.
         * To this end, it is advised to use a DB key as part of the type config - like above, as if the length-like unit of measure had an id of 5.
         *
         * @property numeratorType
         * @type String
         * @optional
         * @default 'units' (respects the `units` config option)
         */
        numeratorType : 'units',

        /**
         * What unit to show as the denominator.
         * As [durationFormatter#showIn](formatters.durationFormatter.html#property_showIn), this option must match one of the keys of this formatter's `unit` config.
         *
         * @property denominatorShowIn
         * @type String
         * @optional
         * @default 'h' (Hours)
         */
        denominatorShowIn : 'h',

        /**
         * As the `numeratorType` option, but for the denominator.
         * @property denominatorType
         * @type String
         * @optional
         * @default 'units' (respects the `units` config option)
         */
        denominatorType : 'units',

        /**
         * The number of decimal places to show
         *
         * @property decimalsToShow
         * @type Number (0-2)
         * @optional
         * @default 0
         */
        decimalsToShow : 0,

        /**
         * Whether or not to allow negative values to be entered by the user with this formatter.
         *
         * @property allowNegative
         * @type Boolean
         * @optional
         * @default false
         */
        allowNegative : false,

        // documented in base numericOptions
        thousandsSeparator : ',',

        // documented in base numericOptions
        decimalSeparator : '.',

        // documented in base numericOptions
        thousandsGroupingStyle : 'by_threes',

        // documented in base numericOptions
        workableDecimalPlaces : 6,

        /**
         * Whether or not to show values given to this formatter as a percentage "%", instead of units per duration "h/h".
         *
         * @property showAsPercentage
         * @default false
         * @optional
         */
        showAsPercentage : false,

        /**
         * The assistant percent formatter to use if `showAsPercentage` is true.
         *
         * @property percentFormatter
         * @default null
         * @optional
         */
        percentFormatter : null

        /**
         * A pattern (RegExp or String), or list of patterns, that will be ignored in any string
         *  the user types, for unformat() or validate() calls.
         *
         * @property unformatIgnore
         * @type String|RegExp|Array
         * @optional
         */
        //unformatIgnore = [],

        //serializeWrapper,
        //unserializeWrapper,
        //formatWrapper,
        //unformatWrapper,
        //validateWrapper
    };

    // TODO add validation options on create; showSubUnits: false + showAllNonzeroSubUnits: true
    //  makes no sense.  Probably other such things we could do as well, like check
    //  showIn/flexibleScale validity

    var getTrueScale = durationUtils.getTrueScale;

    function checkOpts(options) {
                durationUtils.checkRatioOpts(options, options.numeratorShowIn, options.numeratorType);
        durationUtils.checkRatioOpts(options, options.denominatorShowIn, options.denominatorType);
            }

    function format(options, val, formatOptions) {
        if (val == null) {
            return '';
        }

        if (!_.isNumber(val)) {
            logger.fatal('format: workableForm for durationFormatter is a Number');
            throw new Error('Must be a number');
        }

        if (_.isNaN(val)) {
            logger.error('numberFormatter should not be passed NaN');
            return '';
        }
        if (val === Number.POSITIVE_INFINITY || val === Number.NEGATIVE_INFINITY) {
            logger.error('numberFormatter should not be passed infinity');
            return '';
        }

        if (options.showAsPercentage) {
            return options.percentFormatter.format(val);
        }

        var numShowIn = options.numeratorShowIn;
        var denomShowIn = options.denominatorShowIn;
        var numUnits = options[options.numeratorType];
        var denomUnits = options[options.denominatorType];
        if (!numUnits[numShowIn] || !denomUnits[denomShowIn]) {
            var error = 'No unit specified for the major scale ' + numShowIn + ' or ' + denomShowIn + ', did you forget to specify the numeratorShowIn or denominatorShowIn option?';
            logger.fatal(error);
            throw new Error(error);
        }

        var isScientific = scientific.isScientific(val);
        if (isScientific && Math.abs(val) < 1) {
            // if it's this small, let's face it, not worth bothering about
            return format(options, 0, formatOptions);
        }

        var decimalsToShow = options.decimalsToShow;
        if (formatOptions && formatOptions.truncate === false) {
            decimalsToShow = Math.max(decimalsToShow, scientific.countDecimalsInNumber(val));
        }

        var numUnit = numUnits[numShowIn];
        var numScale = getTrueScale(options, numUnit);
        var denomUnit = denomUnits[denomShowIn];
        var denomScale = getTrueScale(options, denomUnit);

        var scaledValue = (val * denomScale) / numScale;
        return durationUtils.stringy(options, durationUtils.fixDecimal(scaledValue, decimalsToShow)) + numShowIn + '/' + denomShowIn;
    }

    function unformat(options, valStr) {
        if (valStr == null || valStr === '') {
            return null;
        }

        if (!_.isString(valStr)) {
            logger.fatal('unformat: typed form for all formatters is a String');
            throw new Error('Must be a string');
        }

        var thousands = regexUtils.escapeLiterals(options.thousandsSeparator);
        var originalValStr = valStr;

        valStr = _.unescape(valStr);

        if (thousands !== ' ') {
            valStr = valStr
                            .replace(new RegExp(thousands, 'g'), '')
                           .replace(options.decimalSeparator, '.')
                           .trim();
        } else {
            valStr = valStr
                           .replace(options.decimalSeparator, '.')
                           .trim();
        }

        valStr = unformatIgnore(options, valStr);

        if (valStr.length >= durationUtils.numbers.maxInputLength) {
            return NaN;
        }

        if (valStr === '') {
            return null;
        }

        var percentFormatter = options.percentFormatter;
        var shouldUnformatAsPercent = !!options.showAsPercentage;
        if (shouldUnformatAsPercent && !percentFormatter) {
            throw new Error('showAsPercentage was true, but no percent formatter given');
        }

        if (!shouldUnformatAsPercent && percentFormatter) {
            // We have a percent formatter, so let's make the attempt anyway...
            var percentSymbol = percentFormatter.getOptions().symbol;
            if (_.includes(valStr, percentSymbol)) {
                shouldUnformatAsPercent = true;
            }
        }

        if (shouldUnformatAsPercent && percentFormatter.validate(valStr)) {
            return percentFormatter.unformat(originalValStr);
        }

        var match = durationUtils.reRatioChunk.exec(valStr);
        if (!match) {
            return NaN;
        }

        var userNumber = Number(match[1]);
        if (thousands === ' ') {
            userNumber = Number(match[1].replace(' ', ''));
        }

        var userUnits = match[2];
        var numUnits = options[options.numeratorType];
        var denomUnits = options[options.denominatorType];
        var numUnit, denomUnit;

        if (userUnits === '') {
            numUnit = options.numeratorShowIn;
            denomUnit = options.denominatorShowIn;
        } else if (_.includes(userUnits, '/')) {
            var split = userUnits.split('/');
            if (split.length !== 2) {
                return NaN;
            }

            numUnit = (userUnits.split('/')[0] || '').trim();
            denomUnit = (userUnits.split('/')[1] || '').trim();
        } else if (numUnits[userUnits]) {
            numUnit = userUnits;
            denomUnit = options.denominatorShowIn;
        } else {
            return NaN;
        }

        if (!numUnits[numUnit] || !denomUnits[denomUnit]) {
            return NaN;
        }
        var numScale = getTrueScale(options, numUnits[numUnit]);
        var denomScale = getTrueScale(options, denomUnits[denomUnit]);
        var baseValue = (userNumber * numScale) / denomScale;

        return durationUtils.clipResult(baseValue, options.workableDecimalPlaces);
    }

    function validate(options, valStr) {
        if (valStr == null || valStr === '') {
            return true;
        }

        if (!options.allowNegative && valStr.indexOf('-') !== -1) {
            return false;
        }

        if (valStr.length >= durationUtils.maxInputLength) {
            return false;
        }

        var wrappedUnformat = options.unformatWrapper ? _.wrap(unformat, options.unformatWrapper) : unformat;
        var unformatted = wrappedUnformat(options, valStr);

        if (unformatted == null) {
            return true;
        }
        if (_.isNaN(unformatted)) {
            return false;
        }

        return true;
    }

    var formatter = baseFormatter(defaults, durationUtils.serialize, durationUtils.unserialize, format, unformat, validate, checkOpts);

    /**
     * Same as the method in [durationFormatter](formatters.durationFormatter.html#method_withSpecializedUnits).
     *
     * @method withSpecializedUnits
     * @param formatter {DurationFormatter} an existing formatter to sub-specialize with `withOptions`
     * @param hoursPer {Object} the new `hoursPer` object to apply
     * @static
     */
    formatter.withSpecializedUnits = durationFormatter.withSpecializedUnits;

    return formatter;
});


define('primutils/formatters/formatterFactory',['lodash',
    'primutils/formatters/helpers/numbers',
    'primutils/formatters/wrappers',
    'primutils/formatters/formats',
    'primutils/formatters/numberFormatter',
    'primutils/formatters/dateFormatter',
    'primutils/formatters/durationFormatter',
    'primutils/formatters/multicurrencyFormatter',
    'primutils/formatters/ratioFormatter'
], function (_, numberUtils, builtinWrappers, builtinFormats, numFmt, dateFmt, durFmt, multicurrencyFormatter, ratioFormatter) {
    'use strict';

    function correctShowInUserPref(options, showIn) {
        switch (showIn) {
        case 'n':
            return options.minuteChar;
        case 'h':
            return options.hourChar;
        case 'd':
            return options.dayChar;
        case 'w':
            return options.weekChar;
        case 'm':
            return options.monthChar;
        case 'y':
            return options.yearChar;
        }
        throw new Error('Unrecognized user pref value: ' + showIn);
    }

    var HOURS_PER_MINUTE = 1 / 60;
    var DEFAULT_OPTIONS = {
        fourDigitYear : true,
        monthName : true,
        leadingZeros : true,

        dateSeparator : '/',
        dateFormat : '0',
            // day-month-year: 1
            // month-day-year: 0
            // year-month-day: 2

        showDecimals : true,
        numberDecimals : 2,

        currencyShowSymbol : true,
        currencySymbol : '$',
        currencyDecimalCount : 2,

        decimalSeparator : '.',
        digitSeparator : ',',
        digitGroupingStyle : 'by_threes',

        positiveFormat : '#1.1',

        negativeFormat : '(#1.1)',

        // duration formatting:
        minuteChar : 'n',
        hourChar : 'h',
        dayChar : 'd',
        weekChar : 'w',
        monthChar : 'm',
        yearChar : 'y',

        hoursPerDay : '8',
        hoursPerWeek : '40',
        hoursPerMonth : '172',
        hoursPerYear : '2064',

        displayTime : true,
        is24HourTime : true,
        showMinutes : false,

        displayUnits : 'h',
        unitsSubUnits : false,
        unitsLabel : true,
        unitsDecimalPlaces : true,
        unitsNumberDecimalPlaces : 0,

        displayDuration : 'h',
        durationSubUnits : false,
        durationLabel : true,
        durationDecimalPlaces : true,
        durationNumberDecimalPlaces : 0,

        unitsPerTimeFormat : 'ShowAsUnitsPerDuration'
    };
    var KNOWN_KEYS = _.keys(DEFAULT_OPTIONS);

    function makeZeros(count) {
        var result = '';
        for (var i = 0; i < count; i++) {
            result += '0';
        }
        return result;
    }

    function convertServerFormats(options) {
        options.positiveFormat = options.positiveFormat
            .replace('1#1', '0,0$0')
            .replace('#', '$')
            .replace('1.1', '0,0');

        options.negativeFormat = options.negativeFormat
            .replace('1#1', '0,0$0')
            .replace('#', '$')
            .replace('1.1', '0,0');

        options.displayDuration = correctShowInUserPref(options, options.displayDuration);
        options.displayUnits = correctShowInUserPref(options, options.displayUnits);
    }

    function formatterFactory(givenOptions) {
        var options = _.defaults({}, givenOptions || {}, DEFAULT_OPTIONS);
        var formatters = {};
        // # -> $, 1.1 -> 0,0
        convertServerFormats(options);

        var decimalFormat = '';
        if (options.showDecimals) {
            decimalFormat = '.' + makeZeros(options.numberDecimals);
        }

        var positiveNumberFormat = '0,0' + decimalFormat;
        var positiveIntegerFormat = '0,0';
        var bareIntegerFormat = '0';
        var showSymbolAsDecimalIfFormat = true;

        if (!options.currencyShowSymbol) {
            if (!numberUtils.shouldReplaceDecimalWithSymbol(options.positiveFormat)) {
                options.positiveFormat = options.positiveFormat.replace('$', '');
            } else if (options.showDecimals === false) {
                showSymbolAsDecimalIfFormat = false;
            }

            if (!numberUtils.shouldReplaceDecimalWithSymbol(options.negativeFormat)) {
                options.negativeFormat = options.negativeFormat.replace('$', '');
            } else if (options.showDecimals === false) {
                showSymbolAsDecimalIfFormat = false;
            }
        }

        var positiveCurrencyFormat = numberUtils.useDecimalsInFormat(options.positiveFormat, !!options.showDecimals && options.currencyDecimalCount, showSymbolAsDecimalIfFormat, options.showDecimals);
        var negativeCurrencyFormat = numberUtils.useDecimalsInFormat(options.negativeFormat, !!options.showDecimals && options.currencyDecimalCount, showSymbolAsDecimalIfFormat, options.showDecimals);

        var decimalSeparator = options.decimalSeparator;
        var digitSeparator = options.digitSeparator;
        if (digitSeparator === 'space') {
            digitSeparator = ' ';
        }

        var baseNumberFormatter = numFmt({
            thousandsGroupingStyle : options.digitGroupingStyle,
            alwaysShowNegativeSymbol : false,
            decimalSeparator : decimalSeparator,
            thousandsSeparator : digitSeparator
        });
        var baseCostFormatter = baseNumberFormatter.withOptions({
            workableDecimalPlaces : 5,

            format : positiveCurrencyFormat,
            negativeFormat : negativeCurrencyFormat,
            symbol : options.currencySymbol,

            basePositiveFormat : options.positiveFormat,
            baseNegativeFormat : options.negativeFormat
        });

        var year = options.fourDigitYear ? 'YYYY' : 'YY';
        var month = options.monthName ? 'MMM' : options.leadingZeros ? 'MM' : 'M';
        var day = options.leadingZeros ? 'DD' : 'D';

        var dateOrder =
            options.dateFormat === '0' ? [month, day, year] :
            options.dateFormat === '1' ? [day, month, year] :
                                         [year, month, day];
        var dayMonthOrder =
            options.dateFormat === '1' ? [day, month] : [month, day];
        var monthYearOrder =
            options.dateFormat === '2' ? [year, month] : [month, year];

        var baseDateFormat = dateOrder.join(options.dateSeparator);

        var showMinutes = options.showMinutes ? ':mm' : '';
        var baseTimeFormat = options.is24HourTime ? ('HH' + showMinutes) : 'hh' + showMinutes + ' A';
        var usedTimeFormat = options.displayTime ? (' ' + baseTimeFormat) : '';
        var timezoneSensitiveDateTimeFormat = baseDateFormat + usedTimeFormat;
        var timezoneSensitiveTimeFormat = baseTimeFormat;
        var globalDateFormat = baseDateFormat;
        var globalDateTimeFormat = baseDateFormat + usedTimeFormat;
        var globalTimeFormat = baseTimeFormat;

        var hoursPerDay = parseFloat(options.hoursPerDay);
        var hoursPerWeek = parseFloat(options.hoursPerWeek);
        var hoursPerMonth = parseFloat(options.hoursPerMonth);
        var hoursPerYear = parseFloat(options.hoursPerYear);

        var durationFormatterOptions = {
            workableDecimalPlaces : 6,
            decimalsToShow : options.showDecimals ? options.numberDecimals : 0,

            decimalSeparator : decimalSeparator,
            thousandsSeparator : digitSeparator,

            showIn : options.hourChar,
            units : {
            }
        };
        durationFormatterOptions.units[options.minuteChar] = {
            internalType : 'minute',
            scale : HOURS_PER_MINUTE,
            sub : null
        };
        durationFormatterOptions.units[options.hourChar] = {
            internalType : 'hour',
            scale : 1,
            sub : options.minuteChar
        };
        durationFormatterOptions.units[options.dayChar] = {
            internalType : 'day',
            scale : hoursPerDay,
            sub : options.hourChar
        };
        durationFormatterOptions.units[options.weekChar] = {
            internalType : 'week',
            scale : hoursPerWeek,
            sub : options.dayChar
        };
        durationFormatterOptions.units[options.monthChar] = {
            internalType : 'month',
            scale : hoursPerMonth,
            sub : options.weekChar
        };
        durationFormatterOptions.units[options.yearChar] = {
            internalType : 'year',
            scale : hoursPerYear,
            sub : options.monthChar
        };

        var baseDurationFmt = durFmt.withOptions(durationFormatterOptions);

        var baseDateFmt = dateFmt.withOptions({
            forgiveLeadingZeros : true
        });

        // https://confluence.oraclecorp.com/confluence/pages/viewpage.action?pageId=57204654
        var percentFormatter = baseNumberFormatter.withOptions({
            workableDecimalPlaces : 5,

            formattedMultiplier : 100,

            // Doesn't do negatives in P6, so probably doesn't need any special handling
            //  (though maybe validate() should return false for negatives?)
            format : positiveNumberFormat + '$',
            symbol : '%'
        });

        var baseRatioFormatter = ratioFormatter.withOptions({
            numeratorShowIn : options.displayUnits,
            denominatorShowIn : options.displayDuration,

            workableDecimalPlaces : durationFormatterOptions.workableDecimalPlaces,
            units : durationFormatterOptions.units,

            showAsPercentage : false,

            // TODO...maybe?
            decimalsToShow : options.showDecimals ? options.numberDecimals : 0,

            decimalSeparator : decimalSeparator,
            thousandsSeparator : digitSeparator
        });

        // durationFormatter and close relatives
        var durationFormatter = baseDurationFmt.withOptions({
            showIn : options.displayDuration,
            showSubUnits : options.durationSubUnits,
            showAbbr : options.durationLabel,
            decimalsToShow : options.durationDecimalPlaces ? options.durationNumberDecimalPlaces : 0
        });
        var monthFormatter = durationFormatter.withOptions({
            showIn : options.monthChar, // Show in months
            showAbbr : true,
            showSubUnits : false
        });
        var hourFormatter = durationFormatter.withOptions({
            showIn : options.hourChar,
            allowNegative : true
        });
        var quantityFormatter = durationFormatter.withOptions({
            allowNegative : true,
            showAbbr : false,
            showSubUnits : false,

            // if you override units, you MUST also override showIn
            // and probably should consider toggling `showAbbr` to true
            units : { 'LOCALIZE qty' : { scale : 1 } },
            showIn : 'LOCALIZE qty'
        });
        var unitFormatter = baseDurationFmt.withOptions({
            allowNegative : true,
            showIn : options.displayUnits,
            showSubUnits : options.unitsSubUnits,
            showAbbr : options.unitsLabel,
            decimalsToShow : options.unitsDecimalPlaces ? options.unitsNumberDecimalPlaces : 0
        });

        formatters = {
            options : {
                currency : {
                    showSymbol : options.currencyShowSymbol,
                    symbol : options.currencySymbol,

                    basePositiveFormat : options.positiveFormat,
                    baseNegativeFormat : options.negativeFormat
                },
                numeric : {
                    decimalSeparator : decimalSeparator,
                    thousandsSeparator : digitSeparator,
                    showDecimals : options.showDecimals,
                    decimalPlaces : options.numberDecimals,

                    unitShowIn : options.displayUnits,
                    durationShowIn : options.displayDuration
                },
                datetime : {
                    year : year,
                    month : month,
                    day : day,

                    dayMonth : dayMonthOrder.join(options.dateSeparator),
                    monthYear : monthYearOrder.join(options.dateSeparator),

                    dateSeparator : options.dateSeparator,

                    dateFormat : baseDateFormat,
                    timeFormat : baseTimeFormat,

                    displayTime : !!options.displayTime,
                    showMinutes : !!options.showMinutes
                },
                duration : {
                    minuteChar : options.minuteChar,
                    hourChar : options.hourChar,
                    dayChar : options.dayChar,
                    weekChar : options.weekChar,
                    monthChar : options.monthChar,
                    yearChar : options.yearChar,

                    hoursPerMinute : HOURS_PER_MINUTE,
                    hoursPerDay : hoursPerDay,
                    hoursPerWeek : hoursPerWeek,
                    hoursPerMonth : hoursPerMonth,
                    hoursPerYear : hoursPerYear
                },
                unknown : _.omit(options, KNOWN_KEYS)
            },

            timezoneSensitiveDateTimeFormatter : baseDateFmt.withOptions({
                format : timezoneSensitiveDateTimeFormat,
                unserializeFormat : builtinFormats.timezoneSensitiveDateTimeUnserialization,
                serializeFormat : builtinFormats.timezoneSensitiveDateTimeSerialization,

                                unserializeWrapper : builtinWrappers.timezoneSensitiveDateTimeUnserializeValidation
                            }),

            timezoneSensitiveTimeFormatter : baseDateFmt.withOptions({
                format : timezoneSensitiveTimeFormat,
                unserializeFormat : builtinFormats.timezoneSensitiveDateTimeUnserialization,
                serializeFormat : builtinFormats.timezoneSensitiveDateTimeSerialization,

                                unserializeWrapper : builtinWrappers.timezoneSensitiveDateTimeUnserializeValidation
                            }),

            globalDateFormatter : baseDateFmt.withOptions({
                format : globalDateFormat,
                unserializeFormat : builtinFormats.globalDateUnserialization,
                serializeFormat : builtinFormats.globalDateSerialization,

                                unserializeWrapper : builtinWrappers.globalDateUnserializeValidation
                            }),

            globalDateTimeFormatter : baseDateFmt.withOptions({
                format : globalDateTimeFormat,
                alternateFormats : [
                    baseDateFormat,
                    baseDateFormat + ' hh:mm A',
                    baseDateFormat + ' HH:mm',
                    baseDateFormat + ' hh A',
                    baseDateFormat + ' HH'
                ],

                unserializeFormat : builtinFormats.globalDateTimeUnserialization,
                serializeFormat : builtinFormats.globalDateTimeSerialization,

                                unserializeWrapper : builtinWrappers.globalDateTimeUnseriailzeValidation
                            }),

            globalTimeFormatter : baseDateFmt.withOptions({
                format : globalTimeFormat,
                alternateFormats : [
                    'hh:mm A',
                    'HH:mm',
                    'hh A',
                    'HH'
                ],

                unserializeFormat : builtinFormats.globalDateTimeUnserialization,
                serializeFormat : builtinFormats.globalDateTimeSerialization,

                                unserializeWrapper : builtinWrappers.globalDateTimeUnseriailzeValidation
                            }),

            genericNumberFormatter : baseNumberFormatter.withOptions({
                workableDecimalPlaces : 6,

                format : positiveNumberFormat
            }),

            genericIntegerFormatter : baseNumberFormatter.withOptions({
                validateWrapper : builtinWrappers.integerValidation,
                workableDecimalPlaces : 0,

                format : positiveIntegerFormat
            }),

            bareIntegerFormatter : baseNumberFormatter.withOptions({
                validateWrapper : builtinWrappers.integerValidation,
                workableDecimalPlaces : 0,

                format : bareIntegerFormat
            }),

            percentFormatter : percentFormatter,

            costFormatter : baseCostFormatter,

            durationFormatter : durationFormatter,
            hourFormatter : hourFormatter,
            quantityFormatter : quantityFormatter,
            monthFormatter : monthFormatter,
            longTermDurationFormatter : monthFormatter,
            cppDurationFormatter : monthFormatter,
            unitFormatter : unitFormatter,

            baseRatioFormatter : baseRatioFormatter,

            unitPerDurationFormatter : baseRatioFormatter.withOptions({
                showAsPercentage : options.unitsPerTimeFormat !== 'ShowAsUnitsPerDuration',
                percentFormatter : percentFormatter
            }),

            baseMulticurrencyFormatter : multicurrencyFormatter.withOptions({
                baseCostFormatter : baseCostFormatter,

                showDecimals : options.showDecimals,
                showCurrencySymbol : showSymbolAsDecimalIfFormat
            })
        };

        return formatters;
    }

    return formatterFactory;
});

define('primutils/store/DeepCollection',['lodash', 'primutils/store/OrderedMap'], function (_, OrderedMap) {
    'use strict';

    var DeepCollection = function (name, options) {
        this.name = name;
        this.map = {};
        this.options = _.defaults({}, options, {
            getKey : function (/* type */) {
                return 'id';
            }
        });
    };

    var addNewMap = function (type) {
        _.set(this.map, type, new OrderedMap({
            key : this.options.getKey.call(this, type)
        }));
    };

    DeepCollection.prototype.constructor = DeepCollection;

    DeepCollection.prototype.getMap = function (type) {
        if (!this.hasType(type)) {
            addNewMap.call(this, type);
        }

        return _.get(this.map, type);
    };

    DeepCollection.prototype.add = function (type, item) {
        if (!this.hasType(type)) {
            addNewMap.call(this, type);
        }

        return _.get(this.map, type).add(item);
    };

    DeepCollection.prototype.addAll = function (type, items) {
        if (!this.hasType(type)) {
            addNewMap.call(this, type);
        }

        return _.get(this.map, type).addAll(items);
    };

    DeepCollection.prototype.insertAt = function (type, item, position) {
        if (!this.hasType(type)) {
            addNewMap.call(this, type);
        }
        _.get(this.map, type).insertAt(item, position);
        return item;
    };

    DeepCollection.prototype.position = function (type, key) {
        if (this.hasType(type)) {
            return _.get(this.map, type).position(key);
        }
        return -1;
    };

    DeepCollection.prototype.remove = function (type, key) {
        if (this.hasType(type)) {
            _.get(this.map, type).remove(key);
        }
    };

    DeepCollection.prototype.replace = function (type, item) {
        if (this.hasType(type)) {
            return _.get(this.map, type).replace(item);
        }
        return null;
    };

    DeepCollection.prototype.item = function (type, key) {
        if (this.hasType(type)) {
            return _.get(this.map, type).item(key);
        }

        return null;
    };

    DeepCollection.prototype.items = function (type) {
        if (this.hasType(type)) {
            return _.get(this.map, type).items();
        }
        return [];
    };

    DeepCollection.prototype.hasType = function (type) {
                if (type == null) {
            throw new Error('storeCollection.js - "type" parameter not defined');
        }
        
        return _.get(this.map, type) != null;
    };

    DeepCollection.prototype.clear = function (type, keys, purge) {
        if (this.hasType(type)) {
            if (_.isArray(keys)) {
                var that = this;
                var itemKeys = _.pluck(this.items(type), that.options.getKey(type));
                _.forEach(itemKeys, function (itemKey) {
                    if (_.includes(keys, itemKey)) {
                        that.remove(type, itemKey, '', purge);
                    }
                });
            } else {
                _.set(this.map, type, null);
            }
        }
    };

    DeepCollection.prototype.clearAll = function () {
        this.map = {};
    };

    DeepCollection.prototype.clearPrefix = function (prefix) {
        if (prefix) {
            _.set(this.map, prefix, null);
        } else {
            this.map = {};
        }
    };

    DeepCollection.prototype.count = function (type) {
        if (this.hasType(type)) {
            return _.get(this.map, type).count();
        }
        return 0;
    };

    DeepCollection.prototype.isEmpty = function (type) {
        if (this.hasType(type)) {
            return _.get(this.map, type).isEmpty();
        }
        return true;
    };

    function isEmpty(map) {
        if (map instanceof OrderedMap) {
            return map.isEmpty();
        } else if (typeof map === 'object') {
            return _.every(map, isEmpty);
        }
        return true;
    }

    DeepCollection.prototype.isEmptyAll = function () {
        return isEmpty(this.map);
    };

    function findKeys(prefix, map, key) {
        var trueKey = prefix && key ? prefix + '.' + key : prefix || key;
        if (map instanceof OrderedMap) {
            return trueKey;
        } else if (typeof map === 'object') {
            return _.chain(map)
                .map(_.partial(findKeys, trueKey))
                .values()
                .flatten(true)
                .value();
        }
    }

    DeepCollection.prototype.isEmptyPrefix = function (prefix) {
        return isEmpty(_.get(this.map, prefix));
    };

    DeepCollection.prototype.types = function () {
        return findKeys('', this.map, '');
    };

    DeepCollection.prototype.typesPrefix = function (prefix) {
        var map = prefix && _.get(this.map, prefix) || this.map;
        return findKeys(prefix, map, '');
    };

    DeepCollection.prototype._repopulateIndexes = function (type) {
        if (this.hasType(type)) {
            _.get(this.map, type)._repopulateIndexes();
        }
    };

    DeepCollection.prototype.sort = function (type, sorter) {
        if (this.hasType(type)) {
            _.get(this.map, type).sort(sorter);
        }
    };

    return DeepCollection;
});

define('primutils/store/DeepModifyCollection',[
    'lodash',
    'primutils/store/DeepCollection',
    'primutils/store/ModifyCollection'
], function (_, DeepCollection, ModifyCollection) {
    'use strict';

    function DeepModifyCollection() {
        DeepCollection.apply(this, arguments);
        this._proto = DeepCollection.prototype;
    }

    DeepModifyCollection.prototype = Object.create(DeepCollection.prototype);
    _.assign(DeepModifyCollection.prototype, ModifyCollection.prototype, {
        constructor : DeepModifyCollection,

        anyModifiedPrefix : function (prefix, fields) {
            var that = this;
            var types = this.typesPrefix(prefix);
            return _.some(types, function (type) {
                return that.anyModified(type, fields);
            });
        }
    });

    return DeepModifyCollection;
});

define('primutils/store/rulesEngine',[
    'lodash',
    'logger',
    'primutils/store/Store'
], function (_, logger, PrimutilsStore) {
    'use strict';

    function buildEventArgs(context, store, storeArgs, preventOperation) {
        var prevented = false;
        var changeStack = storeArgs.changeStack;
        if (!_.has(changeStack, '_rules')) {
            changeStack._rules = {};
        }
        var args = {
            stopRules : function () { changeStack._rules[context._rulesNamespace] = true; },
            suspendRulesEngine : function () { context.suspendRules(); },
            areRulesStopped : function () { return changeStack._rules[context._rulesNamespace]; },

            store : store,
            type : storeArgs.type,
            key : storeArgs.key,
            changeStack : changeStack
        };
        if (_.has(storeArgs, 'path')) {
            args.path = storeArgs.path;
        }
        if (preventOperation) {
            args.preventChange = function () { prevented = true; };
            args.isChangePrevented = function () { return prevented; };
        }
        return Object.freeze(args);
    }

    function maybeResumeRules(ns, eventArgs) {
        // Run before last item cleared off changeStack
        if (eventArgs.changeStack.length === 1) {
            eventArgs.changeStack._rules[ns] = false;
        }
    }

    function buildRunRulesFunction(ruleType, store, changePreventionAction) {
        function ruleMatchesPath(updatePath, rulePath) {
            // Checks for exact matches or parent fields (foo.bar -> foo) and is essentially the regular expression: /^rulePath($|\.)/
            // It ensures that field names that match partially (like scopeItem and scopeItemAssignment) do not get caught by the wrong rules
            return (updatePath.indexOf(rulePath) === 0 && (updatePath.length === rulePath.length || updatePath[rulePath.length] === '.'));
        }
        function shouldRunRule(rule, storeArgs) {
            return _.isArray(rule.field) ?
                _.some(rule.field, _.partial(ruleMatchesPath, storeArgs.path)) :
                ruleMatchesPath(storeArgs.path, rule.field);
        }

        function runUpdateRule(rule, storeArgs, eventArgs) {
            //TODO find way to do this w/o running _all_ rules
            if (shouldRunRule(rule, storeArgs)) {
                var item = store.item(storeArgs.type, storeArgs.key),
                    value = store.utilities._getValue(item, storeArgs.path);
                rule.callback.call(undefined, item, value, eventArgs);
            }
        }

        function runBeforeUpdateRule(rule, storeArgs, eventArgs) {
            //TODO find way to do this w/o running _all_ rules
            if (shouldRunRule(rule, storeArgs)) {
                var item = store.item(storeArgs.type, storeArgs.key),
                    oldValue = store.utilities._getValue(item, storeArgs.path);
                rule.callback.call(undefined, item, storeArgs.newValue, oldValue, eventArgs);
            }
        }

        function runRule(rule, storeArgs, eventArgs) {
            rule.callback.call(undefined, storeArgs.item, eventArgs);
        }

        function runSpreadRule(rule, storeArgs, eventArgs) {
            rule.callback.call(undefined, eventArgs);
        }

        var ruleRunner =
            ruleType === 'update' ? runUpdateRule :
            ruleType === 'beforeUpdate' ? runBeforeUpdateRule :
            _.includes(ruleType, '-spread') ? runSpreadRule :
            runRule;

        return function executeRules(storeArgs) {
            if (!this._rulesRunning) { return; }

            var rules = getRulesForType.call(this, storeArgs.type, ruleType);
            var eventArgs = buildEventArgs(this, store, storeArgs, changePreventionAction != null);

            // Rule Checks
            var inChangeStack = _.where(storeArgs.changeStack, { changetype : 'update', type : storeArgs.type, key : storeArgs.key, path : storeArgs.path }).length > 1;
            if (rules.length && (ruleType === 'update' || ruleType === 'beforeUpdate') && inChangeStack) {
                if (_.isFunction(this._options.stopRulesOnCycle) ? this._options.stopRulesOnCycle(storeArgs.changeStack) : this._options.stopRulesOnCycle) {
                    eventArgs.stopRules();
                }
            }

            for (var i = 0, rulesLen = rules.length; (i < rulesLen && !eventArgs.areRulesStopped()); i++) {
                var rule = rules[i];
                ruleRunner(rule, storeArgs, eventArgs);
                if (changePreventionAction != null && eventArgs.isChangePrevented()) {
                    storeArgs[changePreventionAction].call(storeArgs);
                }
            }

            maybeResumeRules(this._rulesNamespace, eventArgs);
        };
    }

    var getRulesForType = function (storeType, type) {
        if (!this._rules.hasOwnProperty(storeType)) {
            this._rules[storeType] = {};
        }
        if (!this._rules[storeType].hasOwnProperty(type)) {
            this._rules[storeType][type] = [];
        }
        return this._rules[storeType][type];
    };

    var addRulesForType = function (storeType, type, ruleDefinition) {
        // Inverse priority so higher items get run first while maintaining order of same-priority rules
        ruleDefinition.priority = -(ruleDefinition.priority || 0);
        _.forEach(_.isArray(storeType) ? storeType : [storeType], function (storeType) {
            getRulesForType.call(this, storeType, type);
            this._rules[storeType][type].push(ruleDefinition);
            this._rules[storeType][type] = _.sortBy(this._rules[storeType][type], 'priority');
        }.bind(this));
    };

    var DEFAULT_BUILDER_OPTIONS = {
        stopRulesOnCycle : true
    };

    return function rulesEngineBuilder(options) {
        /**
         * A collection of rules which can be bound to a {{#crossLink "store.Store"}}{{/crossLink}}.
         *
         * A RulesEngine cannot be instantiated directly.
         * You must use a {{#crossLink "store.rulesEngineBuilder"}}{{/crossLink}} to instantiate a new RulesEngine.
         *
         * The `event` EventArg object contains many useful functions and properties to modify the underlying model.
         *
         * EventArgs:
         *  * `stopRules()` : Call to prevent any additional rules from this Rules Engine from running until we're done with the first rule that triggered the engine
         *  * `suspendRulesEngine()` : Suspends the Rules Engine, same as {{#crossLink "store.RulesEngine/suspendRules:method"}}{{/crossLink}}.
         *  * `areRulesStopped()` : query to see whether any additional rules will run this update
         *  * `store` : The Primutils store that the rules are currently executing on
         *  * `type`: The store type of the rule
         *  * `key` : The primary key of the item being manipulated by the rules
         *  * `changeStack` : The current stack of changes since the first rule was invoked
         *
         * For "before" rules, there is also:
         *  * `preventChange()`: Prevents the actual operation from occurring
         *  * `isChangePrevented()`: query to see whether the pending operation will occur
         *
         * It is important to note that you should always use `event.store` to perform additional cascading actions on the store.
         * Your rules may be running on a store different from the one you expect.
         * It's also helpful, because you should define your rules in a separate file from your model, and you will not have access to your model by capture there.
         *
         * BAD:
         *
         *     var myStore = new Store({});
         *
         *     rulesEngineBuilder()
         *         .onUpdate('foo', 'distance', function (item, value, event) {
         *             // Calculate a new time based on distance
         *             myStore.update(event.type, event.key, 'time', value * item.rate); // BAD - uses captured variable `myStore`
         *         })
         *         .build()
         *         .attachTo(myStore);
         *
         * GOOD:
         *
         *     var myStore = new Store({});
         *
         *     rulesEngineBuilder()
         *         .onUpdate('foo', 'distance', function (item, value, event) {
         *             // Calculate a new time based on distance
         *             event.store.update(event.type, event.key, 'time', value * item.rate); // GOOD - uses `store` from event args
         *         })
         *         .build()
         *         .attachTo(myStore);
         *
         * @class store.RulesEngine
         * @module store
         * @type {{attachTo:Function,detachFrom:Function,suspendRules:Function,resumeRules:Function}}
         * @private
         */
        var rulesEngine = {
            /**
             * Attaches the rules to a Primutils Store.
             *
             * @method attachTo
             * @param {store.Store} store The Primutils Store to listen on
             * @chainable
             * @returns {store.RulesEngine}
             */
            attachTo : function (store) {
                logger.assert(store && store instanceof PrimutilsStore, 'argument `store` must be a Primutils Store');
                store.on(context._rulesNamespace + '.add ' + context._rulesNamespace + '.insert', buildRunRulesFunction('add', store), context);
                store.on(context._rulesNamespace + '.before-update', buildRunRulesFunction('beforeUpdate', store, 'preventUpdate'), context);
                store.on(context._rulesNamespace + '.update', buildRunRulesFunction('update', store), context);
                store.on(context._rulesNamespace + '.before-remove', buildRunRulesFunction('beforeDelete', store, 'preventRemove'), context);
                store.on(context._rulesNamespace + '.remove', buildRunRulesFunction('delete', store), context);
                store.on(context._rulesNamespace + '.add-spread', buildRunRulesFunction('add-spread', store), context);
                store.on(context._rulesNamespace + '.add-all-spread', buildRunRulesFunction('add-all-spread', store), context);
                store.on(context._rulesNamespace + '.update-spread', buildRunRulesFunction('update-spread', store), context);

                return this;
            },
            /**
             * Removes the rules from a Primutils Store.
             *
             * @method detachFrom
             * @param {store.Store} store The Primutils Store to stop listening to
             * @chainable
             * @returns {store.RulesEngine}
             */
            detachFrom : function (store) {
                store.off(context._rulesNamespace + '.add');
                store.off(context._rulesNamespace + '.insert');
                store.off(context._rulesNamespace + '.before-update');
                store.off(context._rulesNamespace + '.update');
                store.off(context._rulesNamespace + '.before-remove');
                store.off(context._rulesNamespace + '.remove');
                store.off(context._rulesNamespace + '.add-spread');
                store.off(context._rulesNamespace + '.add-all-spread');
                store.off(context._rulesNamespace + '.update-spread');

                return this;
            },

            /**
             * Temporarily halts all rules from this Rules Engine from running on any store it's attached to
             *
             * @method suspendRules
             * @chainable
             * @returns {store.RulesEngine}
             */
            suspendRules : function () {
                context._rulesRunning = false;
                return this;
            },
            /**
             * Resumes running all rules from this Rules Engine on all stores it's attached to
             *
             * @method resumeRules
             * @chainable
             * @returns {store.RulesEngine}
             */
            resumeRules : function () {
                context._rulesRunning = true;
                return this;
            }
        };

        var context = {
            _rulesNamespace : _.uniqueId('rules-'),
            _rules : {},
            _rulesRunning : true,
            suspendRules : rulesEngine.suspendRules.bind(rulesEngine),
            _options : _.merge({}, DEFAULT_BUILDER_OPTIONS, options)
        };

        /**
         * Creates a new {{#crossLink "store.RulesEngine"}}{{/crossLink}} that can be attached to a {{#crossLink "store.Store"}}{{/crossLink}}.
         *
         * @class store.rulesEngineBuilder
         */
        var builder = {
            /**
             * Creates a new rule that is run when an item is added.
             *
             * `callback` is of the form:
             *
             *     function onAdd (item, event) {}
             *
             * See {{#crossLink "store.RulesEngine"}}{{/crossLink}} for information on the EventArgs.
             *
             * @method onAdd
             * @param {String} type The store type to operate on
             * @param {Function} callback The rule to run
             * @param {Number} [priority] (Default: 0) The priority of this rule, higher rules happen first
             * @chainable
             * @returns {store.rulesEngineBuilder}
             */
            onAdd : function (type, callback, priority) {
                addRulesForType.call(context, type, 'add', {
                    callback : callback,
                    priority : priority
                });
                return this;
            },
            /**
             * Creates a new rule that is run when before an item is updated.
             *
             * `callback` is of the form:
             *
             *     function onBeforeUpdate (item, newValue, oldValue, event) {}
             *
             * The `newValue` and `oldValue` that are passed to the callback will depend on what actual value was updated in the store.
             * Nested/parent fields can be watched, but the value will be whichever field was truly updated in the store.
             * Also, while updating nested fields cascades up to parent fields, updating a parent object will not trigger the
             * rules for nested fields. See the `onUpdate` method for more information.
             *
             * See {{#crossLink "store.RulesEngine"}}{{/crossLink}} for information on the EventArgs.
             *
             * The actual update can be prevented by calling:
             *
             *     event.preventChange();
             *
             * @method onBeforeUpdate
             * @param {String} type The store type to operate on
             * @param {String} field The field that when changed will trigger the rule
             * @param {Function} callback The rule to run
             * @param {Number} [priority] (Default: 0) The priority of this rule, higher rules happen first
             * @chainable
             * @returns {store.rulesEngineBuilder}
             */
            onBeforeUpdate : function (type, field, callback, priority) {
                addRulesForType.call(context, type, 'beforeUpdate', {
                    field : field,
                    callback : callback,
                    priority : priority
                });
                return this;
            },
            /**
             * Creates a new rule that is run when before an item is updated.
             *
             * `callback` is of the form:
             *
             *     function onUpdate (item, value, event) {}
             *
             * The `value` that is passed to the callback will depend on what actual value was updated in the store.
             * Nested/parent fields can be watched, but the `value` will be whichever field was truly updated in the store.
             * Also, while updating nested fields cascades up to parent fields, updating a parent object will not trigger the
             * rules for nested fields.
             * For example:
             *
             *     rules.onUpdate('type', 'field.nested', function (item, value, event) { ... });
             *     store.update('type', 'id', 'field',        { nested : '...' }); // Result: RULE IS NOT CALLED
             *     store.update('type', 'id', 'field.nested', '...');              // Result: rule called with value === '...'
             *
             *     rules.onUpdate('type, 'field', function (item, value, event) { ... });
             *     store.update('type', 'id', 'field',        { nested : '...' }); // Result: rule called with value === { nested : '...' }
             *     store.update('type', 'id', 'field.nested', '...');              // Result: rule called with value === '...'
             *
             * See {{#crossLink "store.RulesEngine"}}{{/crossLink}} for information on the EventArgs.
             *
             * @method onUpdate
             * @param {String} type The store type to operate on
             * @param {String} field The field that when changed will trigger the rule
             * @param {Function} callback The rule to run
             * @param {Number} [priority] (Default: 0) The priority of this rule, higher rules happen first
             * @chainable
             * @returns {store.rulesEngineBuilder}
             */
            onUpdate : function (type, field, callback, priority) {
                addRulesForType.call(context, type, 'update', {
                    field : field,
                    callback : callback,
                    priority : priority
                });
                return this;
            },
            /**
             * Creates a new rule that is run when before an item is deleted.
             *
             * `callback` is of the form:
             *
             *     function onBeforeDelete (item, event) {}
             *
             * See {{#crossLink "store.RulesEngine"}}{{/crossLink}} for information on the EventArgs.
             *
             * The actual delete can be prevented by calling:
             *
             *     event.preventChange();
             *
             * @method onBeforeDelete
             * @param {String} type The store type to operate on
             * @param {Function} callback The rule to run
             * @param {Number} [priority] (Default: 0) The priority of this rule, higher rules happen first
             * @chainable
             * @returns {store.rulesEngineBuilder}
             */
            onBeforeDelete : function (type, callback, priority) {
                addRulesForType.call(context, type, 'beforeDelete', {
                    callback : callback,
                    priority : priority
                });
                return this;
            },
            /**
             * Creates a new rule that is run when an item is deleted.
             *
             * `callback` is of the form:
             *
             *     function onDelete (item, event) {}
             *
             * See {{#crossLink "store.RulesEngine"}}{{/crossLink}} for information on the EventArgs.
             *
             * @method onDelete
             * @param {String} type The store type to operate on
             * @param {Function} callback The rule to run
             * @param {Number} [priority] (Default: 0) The priority of this rule, higher rules happen first
             * @chainable
             * @returns {store.rulesEngineBuilder}
             */
            onDelete : function (type, callback, priority) {
                addRulesForType.call(context, type, 'delete', {
                    callback : callback,
                    priority : priority
                });
                return this;
            },

            /**
             * Creates a new rule that is run when a spreads are added for a batch of master records.
             *
             * `callback` is of the form:
             *
             *     function onAddAllSpread (event) {}
             *
             * See {{#crossLink "store.RulesEngine"}}{{/crossLink}} for information on the EventArgs.
             *
             * @method onAddAllSpread
             * @param {String} type The store type to operate on
             * @param {Function} callback The rule to run
             * @param {Number} [priority] (Default: 0) The priority of this rule, higher rules happen first
             * @chainable
             * @returns {store.rulesEngineBuilder}
             */
            onAddAllSpread : function (type, callback, priority) {
                addRulesForType.call(context, type, 'add-all-spread', {
                    callback : callback,
                    priority : priority
                });
                return this;
            },

            /**
             * Creates a new rule that is run when a spreads are added for a master record.
             *
             * `callback` is of the form:
             *
             *     function onAddSpread (event) {}
             *
             * See {{#crossLink "store.RulesEngine"}}{{/crossLink}} for information on the EventArgs.
             *
             * @method onAddSpread
             * @param {String} type The store type to operate on
             * @param {Function} callback The rule to run
             * @param {Number} [priority] (Default: 0) The priority of this rule, higher rules happen first
             * @chainable
             * @returns {store.rulesEngineBuilder}
             */
            onAddSpread : function (type, callback, priority) {
                addRulesForType.call(context, type, 'add-spread', {
                    callback : callback,
                    priority : priority
                });
                return this;
            },

            /**
             * Creates a new rule that is run when a spread record cell is updated.
             *
             * `callback` is of the form:
             *
             *     function onUpdateSpread (event) {}
             *
             * See {{#crossLink "store.RulesEngine"}}{{/crossLink}} for information on the EventArgs.
             *
             * @method onUpdateSpread
             * @param {String} type The store type to operate on
             * @param {Function} callback The rule to run
             * @param {Number} [priority] (Default: 0) The priority of this rule, higher rules happen first
             * @chainable
             * @returns {store.rulesEngineBuilder}
             */
            onUpdateSpread : function (type, callback, priority) {
                addRulesForType.call(context, type, 'update-spread', {
                    callback : callback,
                    priority : priority
                });
                return this;
            },

            /**
             * Scopes the rules such that they all belong to a certain type.
             *
             * The following definitions are equivalent:
             *
             *     rulesEngineBuilder()
             *         .onUpdate('foo', 'fieldA', function (item, value, event) {})
             *         .onUpdate('foo', 'fieldB', function (item, value, event) {})
             *         .build();
             *
             *     rulesEngineBuilder()
             *         .for('foo')
             *             .onUpdate('fieldA', function (item, value, event) {})
             *             .onUpdate('fieldB', function (item, value, event) {})
             *         .build();
             *
             * @method for
             * @param {String} type The store type to scope to
             * @returns {store.typedRulesEngineBuilder}
             */
            for : function (type) {
                var typed = _.mapValues(builder, function (fn) {
                    return _.partial(fn, type);
                });
                typed.for = builder.for;
                typed.build = builder.build;
                typed.pop = _.constant(builder);
                return typed;
            },

            /**
             * Builds the Rules Engine.
             *
             * After this call, no additional rules can be added.
             *
             * @method build
             * @returns {store.RulesEngine}
             */
            build : function () {
                //TODO don't leak the builder; technically I can still add new rules to an engine if I keep the builder around
                return rulesEngine;
            }
        };

        /**
         * Please see the function definitions for the {{#crossLink "store.rulesEngineBuilder"}}{{/crossLink}}.
         * These functions are the same, except they omit the `type` parameter.
         *
         * @class store.typedRulesEngineBuilder
         * @private
         */
        /**
         * @method onAdd
         * @param {Function} callback
         * @param {Number} [priority]
         * @chainable
         * @returns {store.typedRulesEngineBuilder}
         */
        /**
         * @method onBeforeUpdate
         * @param {String} field
         * @param {Function} callback
         * @param {Number} [priority]
         * @chainable
         * @returns {store.typedRulesEngineBuilder}
         */
        /**
         * @method onUpdate
         * @param {String} field
         * @param {Function} callback
         * @param {Number} [priority]
         * @chainable
         * @returns {store.typedRulesEngineBuilder}
         */
        /**
         * @method onBeforeDelete
         * @param {Function} callback
         * @param {Number} [priority]
         * @chainable
         * @returns {store.typedRulesEngineBuilder}
         */
        /**
         * @method onDelete
         * @param {Function} callback
         * @param {Number} [priority]
         * @chainable
         * @returns {store.typedRulesEngineBuilder}
         */
        /**
         * @method onAddSpread
         * @param {Function} callback
         * @param {Number} [priority]
         * @chainable
         * @returns {store.typedRulesEngineBuilder}
         */
        /**
         * @method onAddAllSpread
         * @param {Function} callback
         * @param {Number} [priority]
         * @chainable
         * @returns {store.typedRulesEngineBuilder}
         */
        /**
         * @method onUpdateSpread
         * @param {Function} callback
         * @param {Number} [priority]
         * @chainable
         * @returns {store.typedRulesEngineBuilder}
         */
        /**
         * Changes the type scope to a new store type.
         *
         * @method for
         * @param {String} type The store type to scope to
         * @returns {store.typedRulesEngineBuilder}
         */
        /**
         * Returns to the standard rules engine builder scope from a typed scope.
         *
         * @method pop
         * @returns {store.rulesEngineBuilder}
         */
        /**
         * @method build
         * @returns {store.RulesEngine}
         */

        return builder;
    };
});


